// background.js

const HARDWARE_MONITOR_INTERVAL = 500;
const TABS_MONITOR_INTERVAL = 5000; // check tabs every ten seconds
const HTTP_MONITOR_INTERVAL = 4000;
const MAX_QUEUE_LENGTH = 100;
const WATCH_DOG_TIMEOUT = 120000; // 2 min

const HOST_PERMISSIONS = ["https://*/*", "http://*/*"];
const QUEUE_CHUNK_SIZE = 10;
const KEYFRAME_INTERVAL = 120000;

class eventQueue {
  #queue = [];
  markEvent(eventName, eventData) {
    eventData.ts = Date.now();
    this.#queue.unshift([eventName, eventData])
    if (this.#queue.length > MAX_QUEUE_LENGTH) {
      this.#queue.splice(0, QUEUE_CHUNK_SIZE);
    }
  }
  splice() {
    return this.#queue.splice(-QUEUE_CHUNK_SIZE, QUEUE_CHUNK_SIZE)
  }
}


class hardwareMonitor {
  #intervalHandle = null;
  #queue;
  #prevCPUTime;
  constructor(queue) {
    this.#queue = queue;
  }
  #cpuInfo(info) {
    let usedInPercentage;
    let test = [];
    for (let i = 0; i < info.processors.length; i++) {
      let usage = info.processors[i].usage;
      if (this.#prevCPUTime) {
        let oldUsage = this.#prevCPUTime.processors[i].usage;
        usedInPercentage = Math.floor((usage.kernel + usage.user - oldUsage.kernel - oldUsage.user) / (usage.total - oldUsage.total) * 100);
      } else {
        usedInPercentage = Math.floor((usage.kernel + usage.user) / usage.total * 100);
      }
      test.push(usedInPercentage);
    }

    this.#prevCPUTime = info;
    return {
      modelName: info.modelName,
      archName: info.archName,
      cores: info.processors.length,
      usage: test
    };
  }
  #ramInfo(info) {
    let capacity, used, usedInPercentage, leftInPercentage;
    capacity = info.capacity / 1.074e+9;
    used = info.availableCapacity / 1.074e+9;
    leftInPercentage = Math.round((used / capacity) * 100);
    usedInPercentage = 100 - leftInPercentage;

    return {
      'used': usedInPercentage,
      'left': leftInPercentage
    };

  }

  start() {
    if (!this.#intervalHandle) {
      this.#intervalHandle = setInterval(() => {
        chrome.system.cpu.getInfo((info) => {
          let cpuInfo = this.#cpuInfo(info);
          this.#queue.markEvent("cpu.info", { value: cpuInfo });
        });
        chrome.system.memory.getInfo((info) => {
          let ramInfo = this.#ramInfo(info);
          this.#queue.markEvent("ram.info", { value: ramInfo });
        });
      }, HARDWARE_MONITOR_INTERVAL);
    }
  }
  stop() {
    if (this.#intervalHandle) {
      clearInterval(this.#intervalHandle);
      this.#intervalHandle = null;
    }
  }
  isRunning() {
    return this.#intervalHandle !== null;
  }
  setParams(params) {

  }

}
class tabsMonitor {
  #intervalHandle = null;
  #lastUpdate = 0;
  #queue;
  #tabs = {};
  
  constructor(queue) {
    this.#queue = queue;
  }
  start() {
   
    if (!this.#intervalHandle) {
      this.#intervalHandle = setInterval(() => {

        chrome.tabs.query({ url: ["https://*/*", "http://*/*"] })
          .then((tabs) => {
            let queueEntry = [];
//            if (Date.now() - KEYFRAME_INTERVAL > this.#lastUpdate) {
              this.#tabs = {}
              for (let i = 0; i < tabs.length; i++) {
                let tab = tabs[i];
                let tabInfo = { type: "new", id: tab.id, url: tab.url, active: tab.active };
                queueEntry.push(tabInfo);
                this.#tabs[tab.id] = tabInfo;
              }
              console.log("TABS INFO",{
                value: queueEntry,
                keyframe: true
              });
              this.#queue.markEvent("tabs.info", {
                value: queueEntry,
                keyframe: true
              });
              this.#lastUpdate = Date.now();
/*
            } else {
              for (let tabId in this.#tabs) {
                if (this.#tabs.hasOwnProperty(tabId)) {
                  this.#tabs[tabId].type = "closed";
                }
              }
              for (let i = 0; i < tabs.length; i++) {
                let tab = tabs[i];
                let citem = this.#tabs[tab.id];
                if (citem) {
                  if (citem.url !== tab.url || citem.active !== tab.active) {
                    this.#tabs[tab.id].url = tab.url;
                    this.#tabs[tab.id].active = tab.active;
                    this.#tabs[tab.id].type = "update";
                    queueEntry.push(this.#tabs[tab.id]);
                  } else {
                    this.#tabs[tab.id].type = "exist";
                  }
                } else {
                  let tabInfo = { type: "new", id: tab.id, url: tab.url, active: tab.active };
                  queueEntry.push(tabInfo);
                  this.#tabs[tab.id] = tabInfo;
                }
              }
              let closedIds = [];
              for (let tabId in this.#tabs) {
                if (this.#tabs.hasOwnProperty(tabId) && this.#tabs[tabId].type === "closed") {
                  queueEntry.push(this.#tabs[tabId]);
                  closedIds.push(tabId)
                }
              }
              for (let i = 0; i < closedIds.length; i++) {
                delete this.#tabs[closedIds[i]];
              }
              if (queueEntry.length > 0) {
                console.log("TABS INFO 2",{
                  value: queueEntry,
                  keyframe: false
                });
                this.#queue.markEvent("tabs.info", {
                  value: queueEntry,
                  keyframe: false
                })
              }
            }
            */
          })
          .catch((err) => console.log("tabs query error: ", err));
      }, TABS_MONITOR_INTERVAL);
    }
  }
  stop() {
    if (this.#intervalHandle) {
      clearInterval(this.#intervalHandle);
      this.#intervalHandle = null;
    }
  }
  isRunning() {
    return this.#intervalHandle !== null;
  }
  setParams(params) {

  }
}


class httpMonitor {
  #timeoutHandle = null;
  #isRunning = false;
  #sites = []
  #queue;
  constructor(queue) {
    this.#queue = queue;
  }

  async ping(site) {
    let start = Date.now();
    return fetch(site.url)
      .then(() => {
        site.value = Date.now() - start;
        return site;
      })
      .catch((err) => {
        site.value = "error"
        return site;
      })
  }

  start() {
    if (!this.#timeoutHandle && !this.#isRunning) {
      this.#isRunning = true;
      let checkAll = () => {
        this.#timeoutHandle = null;
        let sites = this.#sites.slice();

        let next = (site, values) => {
          return new Promise((resolve, reject) => {
            if (site) {
              this.ping(site)
                .then((result) => {
                  values.push(result);
                  next(sites.pop(), values)
                    .then(resolve)
                    .catch(reject);
                })
                .catch(reject)
            } else {
              resolve(values);
            }
          });
        }
        if(sites.length > 0){
          next(sites.pop(), [])
          .then((values) => {
            this.#queue.markEvent("http.ping", { value: values })
            if (this.#isRunning) {
              this.#timeoutHandle = setTimeout(checkAll, HTTP_MONITOR_INTERVAL);
            }
          })
          .catch((err) => {
            console.log("pings error " + err)
            if (this.#isRunning) {
              this.#timeoutHandle = setTimeout(checkAll, HTTP_MONITOR_INTERVAL);
            }
          })

        }
      }
      checkAll();
    }
  }
  stop() {
    if (this.#timeoutHandle) {
      clearTimeout(this.#timeoutHandle);
      this.#timeoutHandle = null;
    }
    this.#isRunning = false;
  }
  isRunning() {
    return this.#isRunning;
  }
  setParams(params) {
    if (params && params.sites) {
      this.#sites = params.sites;
    }
  }
}
class WatchDogTimeout {
  #timeout = null;
  #services = {};
  stopServices() {
    for (let serviceKey in this.#services) {
      if (this.#services.hasOwnProperty(serviceKey)) {
        let service = this.#services[serviceKey];
        if (service.isRunning()) {
          service.stop();
        }
      }
    }
  }

  setParams(params) {
    for (let serviceKey in this.#services) {
      if (this.#services.hasOwnProperty(serviceKey)) {
        let service = this.#services[serviceKey];
        service.setParams(params[serviceKey]);
      }
    }
  }


  startServices() {

    for (let serviceKey in this.#services) {
      if (this.#services.hasOwnProperty(serviceKey)) {
        let service = this.#services[serviceKey];
        if (!service.isRunning()) {
          service.start();
        }
      }
    }
  }
  reset() {
    if (this.#timeout) {
      clearTimeout(this.#timeout);
    }
    this.startServices();
    this.#timeout = setTimeout(() => {
      this.#timeout = null;
      this.stopServices();
    }, WATCH_DOG_TIMEOUT)
  }
  constructor(services) {
    this.#services = services;
  }
}
class Tick {
  #intervalId = null;
  constructor(){
    this.#intervalId = setInterval(()=>{
      chrome.tabs.query({ url: ["https://*/*", "http://*/*"] })
      .then((tabs) => {
        for (const tab of tabs) {
          chrome.tabs
            .sendMessage(tab.id, { timerTick: new Date().getTime()})
            .then((response) => {})
            .catch((error)=>{});
        }
      })
    },100)
  }
}


let queue = new eventQueue();
let tick = new Tick();
let watchDogTimeout = new WatchDogTimeout({
   hardwaremonitor: new hardwareMonitor(queue),
   tabsmonitor: new tabsMonitor(queue),
   httpmonitor: new httpMonitor(queue)
});

async function getSystemEvents(payload, sender) {
  watchDogTimeout.setParams(payload || {})
  watchDogTimeout.reset();
  return queue.splice();
}



async function dispatchMessage(msgId, payload, sender) {

  switch (msgId) {
    case 'get-system-events':
      return getSystemEvents(payload, sender);
    default:
      return Promise.reject(new Error('unknown message id' + msgId));
  }
}

// message from content script

chrome.runtime.onMessage.addListener((request, sender, sendResponse) => {

  dispatchMessage(request.data.msgid,
    request.data.payload,
    sender)
    .then((result) => {
      sendResponse(result)
    })
    .catch((error) => sendResponse(error));
  return true;
});

/*
  From the user interface we ask the user to install the extension and 
  when the process ends we instrument all open tabs adding a known div 
  with class "iccbamliboverlay". In this way the popup that requests the
   installation of the extension can detect it and close automatically.
*/

chrome.runtime.onInstalled.addListener(() => {
  chrome.tabs.query({ url: HOST_PERMISSIONS })
    .then((tabs) => {
      let tabsInfo = [];
      for (let i = 0; i < tabs.length; i++) {
        let tab = tabs[i];
        chrome.scripting.executeScript(
          {
            target: { tabId: tab.id, allFrames: true },
            files: ['content-script.js'],
          },
          () => { });
      }

    })
    .catch((err) => {
      console.log("tabs query error " + err)
    });
})