local orbit = require("orbit")
require "inconcert.orbit.application"
local Synchronize = require "Synchronize"
local jsonEncoder = require "LuaOrm.Serializers.Json"
local LuaConfigurationControl = require "LuaConfigurationControl"
local cassandra = require "lua_cassandra"
local jwt = require "shared.jwt"
local md5 = require "md5"
local utils = require "shared.apps.utils"
require "LuaOrm.Date"
local api_utils = require "api.utils"
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")

module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)

_M.model = nil

local model = relativeRequire("model", ...)
local m_models = model.GetModels()

local configControl

local modelBuilt = false

---
-- Some components expect a lualiblogger compatible logger instance. 
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function make_compatible_logger ()
	return {
		debug = LogDebug,
		info = LogInfo,
		warn = LogWarning,
		error = LogError,
		fatal = LogFatal
	}
end

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",
	
	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("[API] Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"agent", "repository", "agentservice", "redis", "agentredis", "cassandra", "elasticsearch"})

		app_status.current = newConfig

		if not inconcert.environment.api then
			inconcert.environment.api = {}
		end

		--Configure Agent Service
		local agentServiceDiff = newConfig:DiffState("agentservice")

		inconcert.environment.apps.agent.agentservice = inconcert.environment.apps.agent.agentservice or {}
		inconcert.environment.apps.agent.agentservice.host = newConfig["agentservice"]["host"]
		inconcert.environment.apps.agent.agentservice.port = tonumber(newConfig["agentservice"]["port"])

		--Configure Agent Redis
		inconcert.environment.apps.agent.agentredis = inconcert.environment.apps.agent.agentredis or {}
		inconcert.environment.apps.agent.agentredis.host = newConfig["agentredis"]["host"]
		inconcert.environment.apps.agent.agentredis.port = tonumber(newConfig["agentredis"]["port"])
		inconcert.environment.apps.agent.agentredis.pwd = newConfig["agentredis"]["pwd"] or ""

		--Configure Redis
		inconcert.environment.api.redis = inconcert.environment.api.redis or {}
		inconcert.environment.api.redis.host = newConfig["redis"]["host"]
		inconcert.environment.api.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.api.redis.pwd = newConfig["redis"]["pwd"] or ""

		inconcert.environment.apps.agent.redis = inconcert.environment.apps.agent.redis or inconcert.environment.api.redis

		-- --Configure Elasticsearch (is an array, but for now, we only use the first node)
		inconcert.environment.api.elasticsearch = inconcert.environment.api.elasticsearch or {}
		inconcert.environment.api.elasticsearch.host = newConfig.elasticsearch[1].host
		inconcert.environment.api.elasticsearch.port = tonumber(newConfig.elasticsearch[1].port)

		--Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("[API] Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
			end
		end
		inconcert.environment.api.cassandra = cassandra

		-- require "helpers"
		-- helpers.table:Log(inconcert.environment.api)

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end
		
		app_status.actualStatus = "ready"
		return "ready"
	end
}

-- This function returns the title of the application. It needs to use an explicit translator because _T could be bound to a
-- different application.
function GetApplicationTitle()
	return "api"
end

-- funcion que es llamada cuando el webserver se inicializa.
function Initialize()
	LogInfo("[API:Initialize] Starting application 'api'.")

	configControl = LuaConfigurationControl.Client {
		type = "API",
		address = inconcert.environment.apps.agent.config_server.ip,
		port = inconcert.environment.apps.agent.config_server.port,
		node_address = inconcert.environment.apps.agent.config_server.myIp,
		node_port = inconcert.environment.apps.agent.config_server.myPort,
	}

	configControl:AddServiceDependency("Agent")
	configControl:AddServiceDependency("Repository")
	configControl:AddServiceDependency("AgentService")
	configControl:TriggerNodeReady()

	return true
end

--
-- The user has been authenticated and has permissions for this application. Should we let him enter anyway? (returns false if not)
function CanInitializeSession(user)
	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function InitializeSession()
	
end

--
-- This function gets called when the user logs out. This gets called reliably. The Logout function below may not
-- be called if the user logs off first (apps/login/logout)
function FinalizeSession(appEnvironment)
	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetAppLocation()
	return m_appLocation
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetApplicationName()
	return "api"
end

--///////////////////////////////////////////////////////////////////////////////////
--

-----------------------
--  (cotrolles) urls/apis
local function GetControllers ()
	local controllerDir = m_appLocation .. "/controllers"
	local controllers = {}
	for file in lfs.dir(controllerDir) do
		if file ~= "." and file ~= ".." then
			local f = controllerDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local controllerName = string.match(file, [[(.+).lua]])
				if controllerName then
					controllers[controllerName] = { url = controllerName }
				end
			end
		end
	end
	return controllers
end

local authFilter = api_utils.makeAuthFilter()

local function Login (web)
	--validate request
	local request = web.request
	LogInfo("[API:Login] DOING LOGIN with User '%s'", request.user or "NONE")
	--LogDebug("[API:Login] 1--> Posted User '%s'", json.encode(request.user or {}))
	if not request.user or not request.password then
		return api_utils.replyInvalidRequest(web, "LOGIN - invalid request. username and password is required.")
	end
	--validate credentials
	local userId, vcc = utils.SeparateVCC(request.user)
	if not vcc then
		return api_utils.replyInvalidRequest(web, "LOGIN - invalid user format. vcc user is required.")
	end
	local info, err = api_utils.getJwtInfo(vcc)
	if not info then
		local reason = "LOGIN - invalid vcc"
		if err then
			reason = string.format("%s Err: %s", reason, json.encode(err))
		end
		return api_utils.replyInvalidRequest(web, reason)
	end
	local status, reason = api_utils.checkIpFilters(info, web.vars.HTTP_X_CLIENT_IP, vcc, userId)
	if not status then
		reason = string.format("LOGIN - %s", reason)
		return api_utils.replyForbidden(web, reason)
	end
	local user = m_models.ApiUsers.GetApiUsers(vcc, userId)
	--LogDebug("[API:Login] 2--> API User '%s'", json.encode(user or {}))
	local postedEncryptedPwd = md5.sumhexa(request.password)
	--LogDebug("[API:Login] 3--> postedEncryptedPwd '%s'", postedEncryptedPwd)
	if not user or user.Password ~= postedEncryptedPwd then
		return api_utils.replyForbidden(web, "LOGIN - invalid credentials")
	end
	--make data
	local data = {
		userName = userId,
		virtualContactCenter = vcc,
		userType = user.Type,
		api = true
	}
	local expirationTime = tonumber(user.ExpirationTime)
	if expirationTime > 0 then
		if user.ExpirationUnit == "minutes" then
			expirationTime = expirationTime * 60
		elseif user.ExpirationUnit == "hours" then
			expirationTime = expirationTime * 60 * 60
		elseif user.ExpirationUnit == "days" then
			expirationTime = expirationTime * 60 * 60 * 24
		end
		--expiration date
		data.exp = os.time() + expirationTime
	end
	--make token
	local token, err = jwt.encode(data, info.key, info.algorithm)
	if not token then
		LogError("[API:Login] error making user '%s' jwt token - %s", request.user, err)
		return api_utils.replyServerError(web, "LOGIN - error making token")
	end

	-- OK
	LogInfo("[API:Login] user '%s' logged in", request.user)
	return api_utils.replyOk(web, { token = token .. "@" .. vcc, expiration = data.exp })
end

local function TestNotifications(web)
	LogInfo("[API:TestNotifications] TEST %s", json.encode(web.request))
	return { status = true }
end

local function Definition (web, apiType)
	apiType = apiType:len() > 0 and apiType or "full"
	LogInfo("[API:Definition] Loading API Definition TYPE %s", apiType)

	local apiDefinitionModule = require "api.definition.full"
	local apiDefinition
	if apiType:lower() == "full" then
		apiDefinition = utils.DeepCopy(apiDefinitionModule.GetFullApi())
	else
		apiDefinition = utils.DeepCopy(apiDefinitionModule.GetApiType(apiType))
	end

	--LogDebug("1----------------> apiDefinition: %s", json.encode(apiDefinition))
	--
	-- TODO: Formatear el atributo "body" según lo que se debe POSTEAR.

	return apiDefinition
end

for key, controllerDef in pairs(GetControllers()) do
	--requires y mapeo funciones
	--LogDebug("api init - key %s", key)
	local controller = relativeRequire("controllers." .. key, ...)
	if type(controller.MapUris) == "function" then
		controller.MapUris(_M, authFilter, "/" .. controllerDef.url, "/api")
	end
end

local vr = utils.JsonValidatorReply
_M:dispatch_post (vr .. Login                  , "/login/"             )
_M:dispatch_post (vr .. TestNotifications      , "/test_notifications/")
_M:dispatch_get  (utils.JsonReply .. Definition, "/definition/(.+)"    )
