local http_mapping = require "apps.agent.result_codes_to_http_codes"

module(..., package.seeall, inconcert.orbit.application.Decorate)

local Filter = require "nozzle"
local jwt = require "shared.jwt"
require "LuaOrm.Date"
local utils = require "shared.apps.utils"
local apiModel = require "api.model"
local apiUsersModel = require "api.models.api_users"
local requestModel = require "api.models.request"
local campaignsModel = require "api.models.campaigns"
local m_systemModel = require "api.models.system"

function GetAdmModel (entityName)
	return utils.GetAdmModel(entityName)
end

function GetAgentModel (entityName)
	return utils.GetAgentModel(entityName)
end

function GetBatchModel (entityName)
	return utils.GetBatchModel(entityName)
end

function GetAgentController (controllerInfo)
	return utils.GetAgentController(controllerInfo)
end

function emptyTable(t)
	return not t or next(t) == nil
end

function empty(value)
	return not value or value == ""
end

function matchText(value, text)
	return tostring(value):lower():match(tostring(text):lower())
end

function GetApiConnectionConfig (connectionType)
	--LogDebug("[API.UTILS:GetApiConnectionConfig] -> Get API SDK connectionType.", connectionType)
	assert(type(connectionType) == "string")
	local allowedConnections = {
		cassandra     = true,
		elasticsearch = true,
		redis         = true
	}
	connectionType = connectionType:lower()
	if not allowedConnections[connectionType] then
		LogWarning("[API.UTILS:GetApiConnectionConfig] API SDK does not allow '%s' connections.", connectionType)
		return
	end
	local appId = "api"
	local conn = utils.GetConnectionConfig(connectionType, appId)
	return conn
end

local function HasEnabledApiInVcc (vcc)
	vcc = vcc or inconcert.authUser and inconcert.authUser.virtualContactCenter
	--LogDebug("[API.UTILS:HasEnabledApiInVcc] 1--> Checking if vcc: '%s' has ENABLED - API SDK", vcc)
	if not vcc then
		return true
	end
	local apiConfiguration = apiUsersModel.GetApiConfiguration(vcc)
	--LogDebug("[API.UTILS:HasEnabledApiInVcc] 2=> vcc: '%s', API SDK configuration: '%s'", vcc, json.encode(apiConfiguration or {}))
	if not apiConfiguration then
		-- VCC no has configuration
		LogWarning("[API.UTILS:HasEnabledApiInVcc] VCC '%s does not have API configuration.", vcc)
		local errType = "vcc does not have API configuration"
		return false, {
			status = false,
			code = -1,
			errType = errType,
			reason = errType
		}
	end
	if not apiConfiguration.enabled then
		-- DISABLED
		LogWarning("[API.UTILS:HasEnabledApiInVcc] VCC: '%s' has API configuration DISABLED.", vcc)
		local errType = "vcc has API configuration disabled"
		return false, {
			status = false,
			code = -2,
			errType = errType,
			reason = errType
		}
	end
	-- ENABLED
	return true, apiConfiguration
end

-------------------------------------- AUTH CHECK --------------------------------------

local JWT_INFO_CACHE_SECONDS = 10
local m_jwt_info

local function loadJwtInfo (vcc)
	--LogInfo("[API.UTILS:loadJwtInfo] Loading JWT Information in VCC: '%s'", vcc)
	local success, apiConfiguration = HasEnabledApiInVcc(vcc)
	--LogDebug("[API.UTILS:loadJwtInfo] --> Loading JWT Information in VCC: '%s' apiConfiguration: %s", vcc, json.encode(apiConfiguration or {}))
	if not success then
		--LogDebug("[API.UTILS:loadJwtInfo] 1 ..")
		local err = apiConfiguration
		return nil, err
	end
	if not apiConfiguration then
		--LogDebug("[API.UTILS:loadJwtInfo] 2 ..")
		return nil, err
	end
	m_jwt_info = m_jwt_info or {}
	m_jwt_info[vcc] = {
		info = apiConfiguration,
		timestamp = LuaOrm.Date():time()
	}
	--LogDebug("[API.UTILS:loadJwtInfo] 3 ..")
	return apiConfiguration
end

function getJwtInfo (vcc)
	--LogDebug("[API.UTILS:getJwtInfo] 1 - '%s'", vcc)
	if not m_jwt_info or not m_jwt_info[vcc] then
		--LogDebug("[API.UTILS:getJwtInfo] 2 - '%s'", vcc)
		return loadJwtInfo(vcc)
	end
	if LuaOrm.Date():time() - m_jwt_info[vcc].timestamp > JWT_INFO_CACHE_SECONDS then
		--LogDebug("[API.UTILS:getJwtInfo] 3 - '%s'", vcc)
		return loadJwtInfo(vcc)
	end
	--LogDebug("[API.UTILS:getJwtInfo] 4 - '%s'", vcc)
	return m_jwt_info[vcc].info
end

function checkIpFilters (info, clientIp, vcc, userName)
	if emptyTable(info.ip_filters) then
		return true
	end
	for _, ip in pairs(info.ip_filters) do
		if string.match(clientIp, ip) then
			return true
		end
	end
	LogError("[API.UTILS:checkIpFilters] user %s@%s can't access from ip %s", userName, vcc, clientIp)
	return false, "access not allowed"
end

function authCheck (token, clientIp)
	if not token then
		LogError("[API.UTILS:authCheck] Filter token not found")
		return false, "Auth token not found"
	end
	local token, vcc = utils.SeparateVCC(token)
	if not vcc then
		LogError("[API.UTILS:authCheck] Filter token vcc not found")
		return false, "Auth token vcc not found"
	end
	local info, err = getJwtInfo(vcc)
	if not info then
		LogError("[API.UTILS:authCheck] Filter token vcc %s info not found", vcc)
		local reason = "Auth token vcc info not found."
		if err then
			return false, string.format("%s Err: %s", reason, json.encode(err))
		end
		return false, reason
	end
	token = token:gsub("Bearer ", "")
	local data, err = jwt.decode(token, info.key, info.algorithm)
	if not data then
		LogError("[API.UTILS:authCheck] Filter invalid token '%s' - %s", token, err)
		return false, err
	end
	if data.virtualContactCenter ~= vcc then
		LogError("[API.UTILS:authCheck] Filter token vcc %s data is different than vcc %s token", data.virtualContactCenter, vcc)
		return false, "Auth token vcc invalid"
	end
	local status, reason = checkIpFilters(info, clientIp, vcc, data.userName)
	if not status then
		return false, reason
	end
	inconcert.authUser = data
	-- TODO: ver tema timezone
	inconcert.authUser.timezone = json.decode(
[[{"offset":0,"minutes":0,"HistoricalDaylight":[],"id":"UTC","dst":false}]]
	)
	return true
end

function makeAuthFilter()
	return Filter{
		input = function(self, web, ...)
			local status, reason = authCheck(web.vars.HTTP_AUTHORIZATION, web.vars.HTTP_X_CLIENT_IP)
			if not status then
				web.status = 403
				web:content_type("application/json")
				return makeResponseStr(false, reason)
			end
		end
	}
end

-------------------------------------- ROLES CHECK --------------------------------------

local ROLES_PERMISSIONS = {
	read  = { read = true },
	write = { read = true, write = true },
	full  = { read = true, write = true, full = true }
}

local function CheckPermissions (permission, apiDefinition)
	--LogDebug("[API.utils:CheckPermissions] ENTRA - inconcert.authUser %s", inconcert.authUser)
	if apiDefinition then
		if apiDefinition.wsapi then
			return true
		end
	end
	if not inconcert.authUser then
		local err = {
			code     = 1,
			httpCode = 500,
			reason   = "Unable check validation. invalid inconcert.authUser"
		}
		return false, err
	end
	--
	assert(type(permission) == "table", "permission is required for validate url filter")
	local permissionType = permission.type or ""

	--LogDebug("[API.utils:CheckPermissions] permission: '%s'", json.encode(permission))
	--LogDebug("[API.utils:CheckPermissions] inconcert.authUser.userType: '%s'", inconcert.authUser.userType)
	--LogDebug("[API.utils:CheckPermissions] ROLES_PERMISSIONS: '%s'", json.encode(ROLES_PERMISSIONS))

	if not ROLES_PERMISSIONS[inconcert.authUser.userType] or not ROLES_PERMISSIONS[inconcert.authUser.userType][permissionType] then
		local err = {
			code     = 2,
			httpCode = 403,
			reason   = "user has not permissions to run action"
		}
		return false, err
	end
	return true
end

local function MakeUrlVccResourceRediskKey (resourceType, vcc)
	assert(type(resourceType) == "string")
	assert(type(vcc) == "string")
	local key = string.format("API:URL:RESOURCE:%s:VCC:%s", resourceType:lower(), vcc:lower())
	--.LogDebug("[API.UTILS:MakeUrlVccResourceRediskKey]------------> key: '%s'", key)
	return key
end

local function GetCurrentUrlRateLimit (resourceType, vcc)
	local redis = apiModel.GetRedisClient()
	local vccResourceKey = MakeUrlVccResourceRediskKey(resourceType, vcc)
	local encodedVccResourceValue = redis:get(vccResourceKey)
	--LogDebug("[API.UTILS:GetCurrentUrlRateLimit] 2---> vccResourceKey: '%s', value: '%s'.", vccResourceKey, encodedVccResourceValue)
	return encodedVccResourceValue and json.decode(encodedVccResourceValue) or nil
end

local function SetCurrentUrlRateLimit (def, currentUrlRateLimits, rateLimits, vcc, defaultRateLimits)
	local key = MakeUrlVccResourceRediskKey(def.config.resourceType, vcc)
	local timeStamp = os.time()
	if currentUrlRateLimits then
		--LogDebug("[API.UTILS:SetCurrentUrlRateLimit] 1 - key: '%s'.", key)
		currentUrlRateLimits.totalRequests = currentUrlRateLimits.totalRequests + 1
		currentUrlRateLimits.lastRequestDate = timeStamp
	else
		--LogDebug("[API.UTILS:SetCurrentUrlRateLimit] 2 - key: '%s'.", key)
		currentUrlRateLimits = {
			vcc              = vcc,
			resourceType     = def.config.resourceType,
			--url              = url,
			firstRequestDate = timeStamp,
			lastRequestDate  = timeStamp,
			totalRequests    = 1,
			config           = rateLimits
		}
	end
	local redis = apiModel.GetRedisClient()
	redis:set(key, json.encode(currentUrlRateLimits))
	-- La key expira en xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
	local MINUTE_IN_SEC = defaultRateLimits.timeLimitSec + 5 																--currentUrlRateLimits.config.timeLimitSec + 5
	--LogDebug("[API.UTILS:SetCurrentUrlRateLimit] 3 - url: '%s'. Key expire in %s seconds.", key, MINUTE_IN_SEC)
	redis:expire(key, MINUTE_IN_SEC)
end

local function ValidateRateLimits (def)
	if not inconcert.authUser then
		--LogDebug("[API.utils.ValidateRateLimits] 0.1. OK)  - ---")
		return true
	end
	local vcc = inconcert.authUser.virtualContactCenter
	local vccRateLimits, defaultRateLimits

	if type(def.config.rateLimits) == "function" then
		vccRateLimits, defaultRateLimits = def.config:rateLimits(vcc) -- GetApiURLRateLimit (ejecuta este metodo generico que se ha definido en las URLs)
	else
		local urlRateLimits = def.config.rateLimits
		if urlRateLimits and urlRateLimits.Disabled then
			-- ok - no tiene configuracion de rate limit
			LogWarning("[API.utils.ValidateRateLimits] url: %s. Has DISABLED rate limit configuration.", def.url)
			return true
		end
		-- error - configuracion invalida
		local errMsg = string.format("Rate limits error in URL '%s'. Has not conifgured 'rateLimits' attributo.", def.url)
 		LogWarning("[API.UTILS:ValidateRateLimits] %s.", errMsg)
 		return false, errMsg
	end

	--LogDebug("[API.utils.ValidateRateLimits] ---> vcc: '%s', url: %s,\nvccRateLimits: %s", vcc, def.url, json.encode(vccRateLimits or {}))

	if type(vccRateLimits) ~= "table" then
		-- error - no tiene configuracion de rate limit
		local errMsg = string.format("Rate limits error in URL '%s'. Has not calculated vcc '%s' 'rateLimits'.", def.url, vcc)
 		LogWarning("[API.UTILS:ValidateRateLimits] %s.", errMsg)
 		return false, errMsg
	end
	vccRateLimits.maxRequests = tonumber(vccRateLimits.maxRequests)
	defaultRateLimits.timeLimitSec = tonumber(defaultRateLimits.timeLimitSec)
	if type(vccRateLimits.maxRequests) ~= "number" or type(defaultRateLimits.timeLimitSec) ~= "number" then
		-- error - configuracion invalida
		local errMsg = string.format("Rate limits error in URL '%s'. Vcc '%s' with invalid config '%s'", def.url, vcc, json.encode(vccRateLimits))
 		LogWarning("[API.UTILS:ValidateRateLimits] %s.", errMsg)
 		return false, errMsg
	end
	if vccRateLimits.maxRequests <= 0 then
		-- error - configuracion invalida
		local errMsg = string.format("Rate limits error in URL '%s'. Vcc '%s' with invalid config in attribute maxRequests '%s'", def.url, vcc, vccRateLimits.maxRequests)
 		LogWarning("[API.UTILS:ValidateRateLimits] %s.", errMsg)
 		return false, errMsg
	end

	local currentUrlRateLimits = GetCurrentUrlRateLimit(def.config.resourceType, vcc)
	if not currentUrlRateLimits then
		-- ok - es el primer request
		--LogDebug("[API.utils.ValidateRateLimits] 1. OK) - Is the first request for url %s in vcc: '%s'.", def.url, vcc)
		SetCurrentUrlRateLimit(def, nil, vccRateLimits, vcc, defaultRateLimits)
		return true
	end
	local elapsedTime = utils.GetElapsedTimeSec(currentUrlRateLimits.firstRequestDate)
	if elapsedTime > defaultRateLimits.timeLimitSec then
		-- ok - ha pasado el tiempo limite desde el ultimo request
		--LogDebug("[API.utils.ValidateRateLimits] 2. OK)  - Vcc '%s' - Time limit has passed since the last request. Elapsed Time '%s' seconds, currentUrlRateLimits: %s", vcc, elapsedTime, json.encode(currentUrlRateLimits))
		SetCurrentUrlRateLimit(def, nil, vccRateLimits, vcc, defaultRateLimits)
		return true
	end
	local totalRequests = currentUrlRateLimits.totalRequests + 1
	if totalRequests > vccRateLimits.maxRequests then
		-- error - ha llegado a la cantidad maxima de request en un tiempo determinado.
		local errLimits = {}

		errLimits.msg = string.format("Vcc: '%s' - Rate limits error in URL '%s'. Has exceeded the maximum limit supported in url. Supports %s requests in %s seconds", vcc, def.url, vccRateLimits.maxRequests, defaultRateLimits.timeLimitSec)
		errLimits.nxt = os.time(os.date("!*t")) + (defaultRateLimits.timeLimitSec - elapsedTime)

 		LogWarning("[API.UTILS:ValidateRateLimits] %s. Next retry: %s", errLimits.msg, errLimits.nxt)

 		return false, errLimits
	end
	-- ok
	--LogDebug("[API.utils.ValidateRateLimits] 3. OK)  - Vcc '%s' - Elapsed Time '%s' seconds, currentUrlRateLimits: %s", vcc, elapsedTime, json.encode(currentUrlRateLimits))
	SetCurrentUrlRateLimit(def, currentUrlRateLimits, nil, vcc, defaultRateLimits)
	return true
end

local function CheckRateLimits (def)
	-- if not def.config.rateLimits then
	-- 	return true
	-- end
	local success, errType = ValidateRateLimits(def)
	if not success then
		if type(errType) == "table" then
			local err = {
				code                   = 3,
				httpCode               = 429,
				reason                 = "user request has reached its rate limit",
				errType                = errType.msg,
				nextRetrySecsFromEpoch = errType.nxt
			}
			return false, err
		else
			local err = {
				code                   = 3,
				httpCode               = 429,
				reason                 = "user request has reached its rate limit",
				errType                = errType
			}
			return false, err
		end
	end
	return true
end

local function ReturnWSAPIError (err)
	LogWarning("[API.utils:ReturnWSAPIError] WSAPI Error: %s", json.encode(err))
	local wsapi = require "wsapi"
	-- wsapi.request = require "wsapi.request"
	-- wsapi.request_stream = require "wsapi.request_stream"
	local response = wsapi.response.new(200, { ["Content-Type"]= "text/plain" })
	response:write(json.encode(err))
	response.headers["Response"] = "defaultError"
	response.status = err.httpCode or 500
	return response:finish()
end

local function ChekUrlFilter (apiDef, method, callbackFn)
	assert(type(apiDef) == "table", "apiDef is required for validate url filter")
	assert(type(apiDef.config) == "table", "apiDef.config is required for validate url filter")
	callbackFn = callbackFn or function() end

	return function(web, ...)
		--LogDebug("1 - ChekUrlFilter ENTRA ")

		local success, apiConfiguration = HasEnabledApiInVcc()
		if not success  then
			local err = apiConfiguration
			err.httpCode = 403
			callbackFn(web, false)
			web.status = 403
			web:content_type("application/json")
			return makeResponseStr(false, err)
		end

		local success, err = CheckPermissions(apiDef.config.permission, apiDef)
		--LogDebug("2 - ChekUrlFilter:CheckPermissions     success %s", success)
		if not success then
			callbackFn(web, false)
			web.status = err.httpCode
			web:content_type("application/json")
			return makeResponseStr(false, err)
		end

		local request
		if not apiDef.wsapi then
			local success
			local logMethodName = string.format("API.%s:%s", apiDef.entityType or "", apiDef.config.urlLinkedFunction or "")
			success, request = validateRequest(web, logMethodName, apiDef.body, apiDef)
			--LogDebug("3 - ChekUrlFilter:validateRequest     success %s", success)
			if not success then
				callbackFn(web, false)
				local err = request
				return err
			end
		end
		request = request or {}

		success, err = CheckRateLimits(apiDef)
		--LogDebug("4 - ChekUrlFilter:CheckRateLimits     success %s", success)
		if not success then
			if apiDef.wsapi then
				-- err.httpCode
				--return 500, makeResponseStr(false, "url rate limit error")
				return ReturnWSAPIError(err)
			else
				callbackFn(web, false)
				web.status = 500
				--[[
					TODO: Revisar por qué falla en el webhandler/wsapi el error 429 (Too Many Requests)
				]]--
				--web.status = err.httpCode
				web:content_type("application/json")
				return makeResponseStr(false, err)
			end
		end

		local methodType = (apiDef.method or ""):upper()
		if methodType == "GET" then
			if apiDef.wsapi then
				--LogDebug("--------------------- WSAPI")
				return method(web, ...)
			else
				return method(web, ...)
			end
		elseif methodType == "POST" then
			-- local logMethodName = string.format("API.%s:%s", apiDef.entityType or "", apiDef.config.urlLinkedFunction or "")
			-- local success, request = validateRequest(web, logMethodName, apiDef.body)
			-- LogDebug("4 - ChekUrlFilter:validateRequest     success %s", success)
			-- if not success then
			-- 	callbackFn(web, false)
			-- 	local err = request
			-- 	return err
			-- end
			-- request = request or {}
			request.apiDefinition = apiDef
			return method(web, request, ...)
		else
			assert(false, string.format("Method '%s' not supported", methodType))
		end
	end
end

--[[

]]--
local function GetApiType (apiType)
	local m_apiDefinition = require "api.definition.full"
	return m_apiDefinition.GetApiType(apiType)
end

--[[

]]--
local function GetApiTypeItemDefinition (apiType, itemName)
	if not apiType or not itemName then
		return
	end
	itemName = itemName:lower()
	local api = GetApiType(apiType)
	for _, def in ipairs(api.definition.items) do
		if def.name:lower() == itemName then
			return def
		end
	end
end

--[[
	Mapeo de URLs dinamicas (definidas en \api\definition\full.lua)
]]--
function MapDynamicURIs (self, apiType, m, myModule, replyPostFn, replyGetFn, callback)
	--LogDebug("[API.UTILS:MapDynamicURIs] Mapping dynamic API URIs type '%s'", apiType)
	local api = GetApiType(apiType)
	local Log = require("apps.agent.requestFilters").LogRequest

	local function CanIncludeUrl (def)
		return true
	end

	local function dispatch_get (def, fn)
		m:dispatch_get(replyGetFn .. ChekUrlFilter(def, fn, callback), def.url)
	end

	local function dispatch_post (def, fn)
		m:dispatch_post(replyPostFn .. ChekUrlFilter(def, fn, callback), def.url)
	end

	local function dispatch_wsapi (def, fn)
		--TODO: Realizar validaciones de filtros en urls de wsapi....
		--
		m:dispatch_wsapi(Log .. ChekUrlFilter(def, fn, callback), def.url)
		--m:dispatch_wsapi(Log .. fn, def.url)
	end

	for _, def in ipairs(api.definition.items) do
		if CanIncludeUrl(def) then
			local method = (def.method or ""):upper()
			local url = assert(def.url)
			local fnName = assert(def.config.urlLinkedFunction)
			local fn = myModule[fnName]

			--LogDebug("[API.UTILS:MapDynamicURIs] ---> method '%s', wsapi: '%s', url '%s', = fnName = '%s' fn = '%s'", method, def.wsapi, url, fnName, fn)
			assert(fn, "function '" .. tostring(fnName) .. "' does not exist in module")

			if method == "GET" then
				if def.wsapi then
					dispatch_wsapi(def, fn)
				else
					dispatch_get(def, fn)
				end
			elseif method == "POST" then
				if def.wsapi then
					dispatch_wsapi(def, fn)
				else
					dispatch_post(def, fn)
				end
			else
				assert(false, string.format("Method '%s' not supported", method))
			end
		end
	end
end

--[[

]]--
function CheckOccPermissions (apiType, itemName, vcc, ksAgent)
	local def = GetApiTypeItemDefinition(apiType, itemName)
	if not def then
		return true
	end
	local occPermissions = def.config and def.config.occPermissions or nil
	if not occPermissions then
		return true
	end
	if not occPermissions.appId
		or not occPermissions.actionId
	then
		return true
	end
	local models = apiModel:GetModels()
	local vccInfo = models.VCCs.Get(vcc, ksAgent)
	vccInfo.vccPermissions = vccInfo.Permissions -- el modulo shared/permissions lo espera en este atributo "vccPermssions"
	--LogDebug("[API.UTILS.CheckOccPermissions] --> vccInfo.vccPermissions: %s", json.encode(vccInfo.vccPermissions or {}))
	local model_perm = require "shared.permissions.models.permissions"
	local hasPermission = model_perm.IsEnabledActionAppVccOverride(occPermissions.appId, occPermissions.actionId, vccInfo)
	if not hasPermission then
		return false, "disabled permisson in vcc"
	end
	--ok
	return true
end

-------------------------------------- REQUESTS --------------------------------------

function makeResponse (status, data, response)
	data = data or {}
	if type(data) ~= "table" then
		data = {reason = data}
	end
	data.status = status
	data.response = response or {}
	return data
end

function makeResponseStr (status, data, response)
	return json.encode(makeResponse(status, data, response))
end

local function replyRequest (web, status, data, response)
	status = status or 200
	web.status = status
	return makeResponse(status == 200, data, response)
end

function replyError (web, status, reason, method, response)
	method = method or "NONE"
	reason = type(reason) == "table" and json.encode(reason) or reason
	LogError("[API.UTILS:replyError] --> METHOD: '%s', reason: %s\nrawError: %s", method, reason, json.encode(response or {}))
	return replyRequest(web, status or 500, reason, response)
end

function replyServerError (web, reason)
	return replyError(web, 500, reason, "replyServerError")
end

function replyNotFound (web, reason)
	return replyError(web, 404, reason, "replyNotFound")
end

function replyInvalidRequest (web, reason)
	return replyError(web, 400, reason, "replyInvalidRequest")
end

function replyForbidden (web, reason)
	return replyError(web, 403, reason, "replyForbidden")
end

function replyOtherNode (web, reason)
	return replyRequest(web, 503, reason)
end

function replyOk (web, data)
	return replyRequest(web, 200, data)
end

function getHttpStatus(result, reason)
	if type(result) == "table" then
		reason = reason or result["reason"] or "server error"
		result = result["result"]
	end
	return http_mapping.getHTTPCode(result) or 500, reason
end

local function getRequestAccountByName(request, definition, globalData)
	local accountName = request[definition.name]
	if not accountName then
		return false, "required parameter missing: " .. definition.name
	end
	local campaign = request[definition.campaign]
	if not campaign then
		return false, "required parameter missing: " .. definition.campaign
	end
	if not globalData.accountsByName or not globalData.accountsByName[campaign] then
		local accountsByName = {}
		local vcc = inconcert.authUser.virtualContactCenter
		local campaignAtVCC = utils.AppendVcc(campaign, vcc)
		local accounts = campaignsModel.GetCampaignAccounts(campaignAtVCC)
		for _, account in pairs(accounts) do
			if account.state and account.state == "1" then
				accountsByName[account.name] = account.id
			end
		end
		globalData.accountsByName = globalData.accountsByName or {}
		globalData.accountsByName[campaign] = accountsByName
	end
	for name, id in pairs(globalData.accountsByName[campaign]) do
		if name == accountName then
			return id
		end
	end
	return false, "Invalid account name"
end

local paramTypesLabel = {
	["string"]      = "string",
	["number"]      = "numeric",
	["boolean"]     = "boolean",
	["table"]       = "object",
	["rowsPerPage"] = "numeric"
}

local function __validateRequest (request, definition, globalData, apiDefinition)
	--LogDebug("[API.utils.__validateRequest] --> definition: %s, request: %s", json.encode(definition or {}), json.encode(request or {}))
	if not inconcert.authUser then
		LogDebug("[API.utils.__validateRequest] 0.1. OK)  - ---")
		return true
	end
	local vcc = inconcert.authUser.virtualContactCenter

	for _, param in pairs(definition) do
		local value = request[param.id]
		if value ~= nil then
			local valueType = type(value)
			if param.type and valueType ~= param.type then
				local typeFixed = false
				if valueType == "string" then
					if param.type == "number" then
						--try to parse
						value = tonumber(value)
						if value ~= nil then
							typeFixed = true
							request[param.id] = value
						end
					elseif param.type == "rangeDate" then
						typeFixed = true
						if apiDefinition then
							if param.to then
								local from = LuaOrm.Date(value)
								local to = request[param.to]
								if to then
--	local allowedDateMinutes = apiDefinition.config.filters.rangeDate.minutes
									local vccFilters, defaultFilters = apiDefinition.config:filters(vcc) -- GetApiURLFilters (ejecuta este metodo generico que se ha definido en las URLs)
									local filters = vccFilters or defaultFilters
									local allowedDateMinutes = filters.rangeDate.minutes
									local allowedDateSec = allowedDateMinutes * 60
									to = LuaOrm.Date(to)
									local differenceSec = to:time() - from:time()
									--LogDebug("[API.utils.__validateRequest]   1--> from: %s, to: %s, differenceSec: %s, allowedDateSec: %s", from, to, differenceSec, allowedDateSec)
									if differenceSec > allowedDateSec then
										local text
										local rangeAllowed = "Queries with a maximum range of %s are allowed."
										if allowedDateSec < 60 then
											text = string.format("%s seconds", allowedDateSec)
										else
											if allowedDateMinutes < 60 then
												text = string.format("%s minutes", allowedDateMinutes)
											else
												local allowedDateHours = allowedDateMinutes / 60
												if allowedDateHours < 24 then
													text = string.format("%s hours", allowedDateHours)
												else
													local allowedDateDays = allowedDateHours / 24
													text = string.format("%s days", allowedDateDays)
												end
											end
										end
										rangeAllowed = string.format(rangeAllowed, text)
										return false, {
											httpCode = 400,
											code     = 4,
											reason   = string.format("date parameters out of range. from: %s, to: %s. %s", param.id, param.to, rangeAllowed),
											errType  = "date parameters out of range"
										}
									end
								end
							end
						end

					end
				elseif valueType == "number" then
					if param.type == "rowsPerPage" then
						--LogDebug("rowsPerPage --> value %s", value)
						value = tonumber(value)
						if value ~= nil then
--local allowedRowsPerPage = apiDefinition.config.paged.rowsPerPage
							local vccPaged, defaultPaged = apiDefinition.config:paged(vcc) -- GetApiURLPaged (ejecuta este metodo generico que se ha definido en las URLs)
							local paged = vccPaged or defaultPaged
							local allowedRowsPerPage = paged.rowsPerPage
							if value > allowedRowsPerPage then
								local reason = string.format("only %s rows allowed per page", allowedRowsPerPage)
								return false, {
									httpCode = 400,
									code     = 5,
									reason   = reason,
									errType  = "has exceeded lines allowed at query"
								}
							end
							typeFixed = true
							request[param.id] = value
						end
					end
				end
				if not typeFixed then
					return false, "invalid parameter type: " .. param.id .. " must be " .. paramTypesLabel[param.type]
				end
			end
			if param.limits then
				if param.limits.greater and value <= param.limits.greater then
					return false, "invalid parameter value: " .. param.id .. " must be greater than " .. param.greater
				end
				if param.limits.less and value >= param.limits.less then
					return false, "invalid parameter value: " .. param.id .. " must be less than " .. param.less
				end
			end
			if param.domain and not param.domain[value] then
				return false, "invalid parameter value: " .. param.id
			end
			if param.children then
				if param.array then
					for _, arrayValue in pairs(value) do
						local arrayResult, arrayError = __validateRequest(arrayValue, param.children, globalData, apiDefinition)
						if not arrayResult then
							return false, arrayError
						end
					end
				else
					local childrenResult, childrenError = __validateRequest(value, param.children, globalData, apiDefinition)
					if not childrenResult then
						return false, childrenError
					end
				end
			end
			if param.custom then
				local customResult, customError = param.custom(value)
				if not customResult then
					return false, childrenError
				end
			end
		else
			local required = param.required == true or (type(param.required) == "function" and param.required(request))
			if required then
				if param.account then
					local accountId, err = getRequestAccountByName(request, param.account, globalData)
					if not accountId then
						return false, err
					end
					request[param.id] = accountId
				else
					return false, "required parameter missing: " .. param.id
				end
			elseif param.default then
				request[param.id] = param.default
			end
		end
	end
	return true
end

function validateRequest (web, method, definition, apiDefinition)
	local request = web.request
	local status, reason = __validateRequest(request, definition, {}, apiDefinition)
	if status then
		return true, request
	end
	LogError("[API.UTILS:%s] %s", method, reason)
	return false, replyInvalidRequest(web, reason)
end

function validateGet(web, method, definition)
	local request = web.GET
	local status, reason = __validateRequest(request, definition, {}, apiDefinition)
	if status then
		return true, request
	end
	LogError("[API.UTILS:%s] %s", method, reason)
	return false, replyInvalidRequest(web, reason)
end

function validateParams(params, method, definition)
	local status, reason = __validateRequest(params, definition, {})
	if status then
		return true
	end
	LogError("[API.UTILS:%s] %s", method, reason)
	return false, reason
end

function validateManualRequest (method, token, clientIp, permission, data, definition)
	--authentication
	local status, reason = authCheck(token, clientIp)
	if not status then
		return 403, reason
	end
	--permissions
	local err
	status, err = CheckPermissions({type = permission})
	if not status then
		return 403, err.reason
	end
	--params
	status, reason = validateParams(data, method, definition)
	if not status then
		return 400, reason
	end
	return 200
end

function processAsyncResult (params)
	--required params
	local id = params.id
	local result = params.result
	--optional params
	local async = params.async or false
	local timeout = params.timeout or 60
	local pollingSeconds = params.pollingSeconds or 1
	--1. check if is async
	if not result.async or async then
		return result
	end
	local seconds = 0
	while seconds < timeout do
		local result = requestModel.GetRequest(id)
		if result then
			return result
		end
		--LogDebug("[API.UTILS:processAsyncResult] waiting for request %s result", id)
		socket.sleep(pollingSeconds)
		seconds = seconds + pollingSeconds
	end
	return { result = "TIMEOUT", reason = "timeout waiting for result" }
end

--[[
	Estos valores se obtienen de la configuracion almacenada en la CF "SystemConfiguration" key "API_SDK", Column "URLsDefaultConfig" en el KeySpace "Provisioning"

	{
		permission = {
			type = "read"
		},
		resources = {
			high = {
				rateLimits = {
					-- los valores que se encuentran comentados, son valores de prueba para los desarrolladores.
					timeLimitSec = 60, --10
					maxRequests = 1000 --3
				}
			},
			low = {
				rateLimits = {
					-- los valores que se encuentran comentados, son valores de prueba para los desarrolladores.
					timeLimitSec = 60, --10
					maxRequests = 1000 --3
				}
			}
		},
		filters = {
			rangeDate = {
				minutes = 518400 -- 365 dias
			}
		},
		paged = {
			rowsPerPage = 200
		}
	}
]]--
function GetApiURLsDefaultConfig ()
	local API_SDK_SystemConfig = m_systemModel.Get("API_SDK")
	local defaultUrlsConfig = API_SDK_SystemConfig.URLsDefaultConfig
	assert(defaultUrlsConfig)
	return defaultUrlsConfig
end

function GetApiURLsDefaultResourceConfig (resourceType)
	assert(resourceType)
	resourceType = resourceType:lower()
	local defaultUrlsConfig = GetApiURLsDefaultConfig()
	assert(defaultUrlsConfig.resources)
	local resourceInfo = defaultUrlsConfig.resources[resourceType]
	assert(resourceInfo)
	return resourceInfo
end

-- resources
function GetApiURLResource (resourceType, vcc)
	assert(type(resourceType) == "string", "resourceType is required")
	resourceType = resourceType:lower()
	--LogDebug("[API.UTILS:GetApiURLResource] 1=> vcc: '%s', resourceType: '%s'", vcc, resourceType)

	local apiConfiguration = apiUsersModel.GetApiConfiguration(vcc)
	--LogDebug("[API.UTILS:GetApiURLResource] 3=> vcc: '%s', apiConfiguration: '%s'", vcc, json.encode(apiConfiguration or {}))

	if not apiConfiguration then
		--LogDebug("[API.UTILS:GetApiURLResource] 4=> vcc: '%s', resourceType: '%s'", vcc, resourceType)
		return nil
	end
	if not apiConfiguration.urlsConfig or not apiConfiguration.urlsConfig.resources then
		--LogDebug("[API.UTILS:GetApiURLResource] 5=> vcc: '%s', resourceType: '%s'", vcc, resourceType)
		return nil
	end
	local vccUrlsResource = apiConfiguration.urlsConfig.resources[resourceType]
	--LogDebug("[API.UTILS:GetApiURLResource] 6=> vcc: '%s', vccUrlsResource: '%s'", vcc, json.encode(vccUrlsResource))
	-- ok
	return vccUrlsResource
end

-- rate limit
function GetApiURLsDefaultRateLimit (resourceType)
	assert(resourceType)
	resourceType = resourceType:lower()
	local resourceInfo = GetApiURLsDefaultResourceConfig(resourceType)
	local defaultRateLimits = resourceInfo.rateLimits
	--LogDebug("[API.UTILS:GetApiURLsDefaultRateLimit] resourceType: '%s', defaultRateLimits: '%s'", resourceType, json.encode(defaultRateLimits or {}))
	return defaultRateLimits
end

function GetApiURLRateLimit (self, vcc)
	assert(type(self.resourceType) == "string", "resourceType is required")
	local resourceType = self.resourceType:lower()
	local vccUrlsResource = GetApiURLResource(resourceType, vcc)
	local defaultRateLimits = GetApiURLsDefaultRateLimit(resourceType)
	if not vccUrlsResource or not vccUrlsResource.rateLimits then
		--LogDebug("[API.UTILS:GetApiURLRateLimit] 1=> vcc: '%s', resourceType: '%s'", vcc, resourceType)
		return nil, defaultRateLimits
	end
	--LogDebug("[API.UTILS:GetApiURLRateLimit] 2=> vcc: '%s', vccUrlsResource.rateLimits: '%s'", vcc, json.encode(vccUrlsResource.rateLimits))
	-- ok
	return vccUrlsResource.rateLimits, defaultRateLimits
end

--- filters
function GetApiURLsDefaultFilters (resourceType)
	assert(resourceType)
	resourceType = resourceType:lower()
	local resourceInfo = GetApiURLsDefaultResourceConfig(resourceType)
	local defaultFilters = resourceInfo.filters
	--LogDebug("[API.UTILS:GetApiURLsDefaultFilters] resourceType: '%s', defaultFilters: '%s'", resourceType, json.encode(defaultFilters or {}))
	return defaultFilters
end

function GetApiURLFilters (self, vcc)
	assert(type(self.resourceType) == "string", "resourceType is required")
	local resourceType = self.resourceType:lower()
	local vccUrlsResource = GetApiURLResource(resourceType, vcc)
	local defaultFilters = GetApiURLsDefaultFilters(resourceType)
	if not vccUrlsResource or not vccUrlsResource.filters then
		--LogDebug("[API.UTILS:GetApiURLFilters] 1=> vcc: '%s', resourceType: '%s'", vcc, resourceType)
		return nil, defaultFilters
	end
	--LogDebug("[API.UTILS:GetApiURLFilters] 2=> vcc: '%s', vccUrlsResource.filters: '%s'", vcc, json.encode(vccUrlsResource.filters))
	-- ok
	return vccUrlsResource.filters, defaultFilters
end

-- paged
function GetApiURLsDefaultPaged (resourceType)
	assert(resourceType)
	resourceType = resourceType:lower()
	local resourceInfo = GetApiURLsDefaultResourceConfig(resourceType)
	local defaultPaged = resourceInfo.paged
	--LogDebug("[API.UTILS:GetApiURLsDefaultPaged] resourceType: '%s', defaultPaged: '%s'", resourceType, json.encode(defaultPaged or {}))
	return defaultPaged
end

function GetApiURLPaged (self, vcc)
	assert(type(self.resourceType) == "string", "resourceType is required")
	local resourceType = self.resourceType:lower()
	local vccUrlsResource = GetApiURLResource(resourceType, vcc)
	local defaultPaged = GetApiURLsDefaultPaged(resourceType)
	if not vccUrlsResource or not vccUrlsResource.paged then
		--LogDebug("[API.UTILS:GetApiURLPaged] 1=> vcc: '%s', resourceType: '%s'", vcc, resourceType)
		return nil, defaultPaged
	end
	--LogDebug("[API.UTILS:GetApiURLPaged] 2=> vcc: '%s', vccUrlsResource.paged: '%s'", vcc, json.encode(vccUrlsResource.paged))
	-- ok
	return vccUrlsResource.paged, defaultPaged
end
