local orbit = require "orbit"
local wsapi = require "wsapi"
wsapi.request = require "wsapi.request"
wsapi.request_stream = require "wsapi.request_stream"
local utils = require "shared.apps.utils"
local m_models = require("apps.administration.model"):GetModels()
local permissions = require "shared.permissions.permissions"
local modelRestrictions = require "apps.agent.models.config_restrictions"
require "uuid"
require "LuaOrm.Date"
local cjson = require "cjson"
local Charsets = require "shared.relational_databases.charsets"
local modulename = select(1, ...)
local m_moduleLocation = select(2, ...)

module(modulename, package.seeall, orbit.new, inconcert.orbit.application.Decorate)

local function GetMappingUsersFields (mappingType)
	local vcc = inconcert.authUser.virtualContactCenter

	local function IsRequiredMailField ()
		if utils.HasHelpDeskIntegration(vcc) then
			return true
		end
		return false
	end

	-- LOS REQUERIDOS
	local requiredFields = {
		-- el id de usuario único y válido
		{CSVFieldName = "USERID"         , fieldName = "Id"             , required = true},
		-- un rol del sistema o personalizado válido
		{CSVFieldName = "GROUP"          , fieldName = "UserGroup"      , required = true},
		-- el primer nombre del usuario
		{CSVFieldName = "NAME"           , fieldName = "Name"           , required = true},
		-- el primer apellido del usuario
		{CSVFieldName = "LASTNAME"       , fieldName = "FirstSurname"   , required = true},
		-- el timezone id
		{CSVFieldName = "TIMEZONE"       , fieldName = "TimeZoneId"     , required = true},
		-- el código de país
		{CSVFieldName = "COUNTRYCODE"    , fieldName = "countryCode"    , required = true},
		-- el código de área para un país (countryCode) determinado
		{CSVFieldName = "AREACODE"       , fieldName = "areaCode"       , required = true}
	}
	-- LOS REQUERIDOS de forma opcional (Ej: El mail es requerido en VCCs que tengan integración con HelpDesk)
	local optionallyRequiredFields = {
		-- la dirección de correo electrónico válido y único dentro del vcc
		{CSVFieldName = "MAIL"           , fieldName = "Mail"           , required = IsRequiredMailField()}
	}
	-- LOS NO REQUERIDOS
	local notRequiredFields = {
		-- la contraseña plana
		{CSVFieldName = "PASSWORD"       , fieldName = "Password"       , required = false},
		-- el tipo de lenguaje que se le configura al usuario
		{CSVFieldName = "LANGUAGE"       , fieldName = "LanguageId"     , required = false},
		-- campañas
		{CSVFieldName = "CAMPAIGNS"      , fieldName = "groups"         , required = false},
		-- la campaña por defecto, solo sirve cuando vienen campañas en la lista "groups"
		{CSVFieldName = "DEFAULTCAMPAIGN", fieldName = "DefaultCampaign", required = false}
	}
	-- LOS campos exclusivos de VCCs SIN Integración con Allegro
	local fieldsWithOutAllegroIntegration = {
		-- el estado inicial
		{CSVFieldName = "INITIALSTATE"   , fieldName = "initialState"   , required = false}
	}
	-- LOS campos exclusivos de VCCs CON Integración con Allegro
	local fieldsWithAllegroIntegration = {
		-- el Teléfono formado por el Service + Peer + Technology (TELEPHONY_192_168_24_236 + 2002 + ASTERISK,SIP)
		{CSVFieldName = "PHONE SERVICE"   , fieldName = "PhoneService"   , required = false, onlyAllegro = true},
		{CSVFieldName = "PHONE PEER"      , fieldName = "PhonePeer"      , required = false, onlyAllegro = true},
		{CSVFieldName = "PHONE TECHNOLOGY", fieldName = "PhoneTechnology", required = false, onlyAllegro = true}
	}

	local mapping = {}
	local function Fill (fields)
		for _, field in ipairs(fields) do
			table.insert(mapping, field)
		end
	end

	Fill(requiredFields)
	Fill(optionallyRequiredFields)
	Fill(notRequiredFields)
	local hasAllegroIntegration = utils.HasAllegroIntegration(inconcert.authUser.virtualContactCenter)
	if hasAllegroIntegration then
		Fill(fieldsWithAllegroIntegration)
	else
		Fill(fieldsWithOutAllegroIntegration)
	end

	if mappingType == "HASH_BY_CSV_FIELD_NAME" then
		return utils.ArrayToHash(mapping, "CSVFieldName")
	elseif mappingType == "HASH_BY_CASSANDRA_USER_FIELD_NAME" then
		return utils.ArrayToHash(mapping, "fieldName")
	end

	-- simple array
	return mapping
end

local function ValidateCSVFile (fileId, LoadMappingFieldsFn)
	assert(type(fileId) == "string")
	assert(type(LoadMappingFieldsFn) == "function")

	LogInfo("[ADMINISTRATION.IMPORTS:ValidateCSVFile] Validating CSV File '%s'.", fileId)

	local csv = m_models.Imports.GetTemporalUploadedFile({
		FileId = fileId
	})
	if not csv then
		local msg = "CSV File does not exist"
		return false, {
			type = "CSV FILE DOES NOT EXIST",
			code = -101,
			reason = msg
		}
	end

	--LLevo a utf8 para procesar caracteres especiales
	csv = Charsets.Init().toutf8(csv)
	local csvHeader = utils.ParseCSVHeader(csv)
	if not csvHeader then
		local msg = "Invalid CSV File"
		return false, {
			type = "INVALID CSV FILE",
			code = -105,
			reason = msg
		}
	end

	--LogDebug("1 csvHeader: %s", json.encode(csvHeader))
	local mapping = LoadMappingFieldsFn()
--	LogDebug("2 mapping: %s", json.encode(mapping))
	local missingRequiredFields = {}

	-- Valido los campos requeridos del mapeo contra el CSV
	for _, fieldInfo in ipairs(mapping) do
		if fieldInfo.required then
			local existRequiredField = false
			for index, csvFieldName in ipairs(csvHeader) do
				csvFieldName = utils.alltrim(csvFieldName):upper()
				if fieldInfo.CSVFieldName == csvFieldName then
					existRequiredField = true
					break
				end
			end
			if not existRequiredField then
				table.insert(missingRequiredFields, fieldInfo.CSVFieldName)
			end
		end
	end

	local totalMissingRequiredFields = #missingRequiredFields
	if totalMissingRequiredFields > 0 then
		local msg = string.format("Missing %d required fields in CSV File.", totalMissingRequiredFields)
		return false, {
			type = "MISSING SOME REQUIRED FIELDS",
			code = -102,
			reason = msg,
			missingRequiredFields = missingRequiredFields
		},
		csvHeader
	end

	-- Clean Up de las columnas que contiene el CSV.
	-- Esto es para que no me puedan agregar cualquier columna en la BD, solo columnas válidas y permitidas.
	local mappingUsersFields = LoadMappingFieldsFn("HASH_BY_CSV_FIELD_NAME")
	local allowedHeaderFields = {}

	for index, csvFieldName in ipairs(csvHeader) do
		csvFieldName = utils.alltrim(csvFieldName):upper()
		local fieldInfo = mappingUsersFields[csvFieldName]
		if fieldInfo then
			table.insert(allowedHeaderFields, csvFieldName)
		end
	end
	--LogDebug("3 allowedHeaderFields: %s", json.encode(allowedHeaderFields))

	return true, csv, allowedHeaderFields
end

local function ValidateImportation (info, keyspace)
	LogInfo("[ADMINISTRATION.IMPORTS:ValidateImportation] Validating Importation.")

	if type(info) ~= "table" then
		local msg = "Invalid request"
		return false, {
			type = "INVALID REQUEST",
			code = -100,
			reason = msg
		}
	end

	local isAnyImport = m_models.Imports.IsAnyImportInProgressInVCC()
	--LogDebug("isAnyImport: '%s'", isAnyImport)
	if isAnyImport then
		local msg = string.format("Unable to Import. There another import in progress.")
		return false, {
			type = "THERE ANOTHER IMPORT IN PROGRESS",
			code = -103,
			reason = msg
		}
	end

	local success, csv, csvHeader = ValidateCSVFile(info.fileInfo.fileId, GetMappingUsersFields)
	if not success then
		local err = csv
		return false, err
	end

	local mappingUsersFields = GetMappingUsersFields("HASH_BY_CSV_FIELD_NAME")
	--LogDebug("mappingUsersFields: %s", json.encode(mappingUsersFields))

	local totalCampaigns = 0

	local parsedUsersCSV = utils.Csv2Table(csv, function(csvFieldName, csvFieldValue, user)
		local fieldInfo = mappingUsersFields[utils.alltrim(csvFieldName):upper()]
		if not fieldInfo then
			-- Si la columna del CSV, NO es una columna válida en la BD, o no es un valor permitido a importar, NO lo agrego al usuario
			return
		end
		if fieldInfo.CSVFieldName == "PASSWORD" then
			user.IsPlainPassword = true -- en el CSV si viene una pwd, tiene que ser plana, entonces la flagueo como tal para que el controller user de la app agent, realice el procedimiento que corresponda.
		end

		if fieldInfo.CSVFieldName == "PHONE SERVICE" or fieldInfo.CSVFieldName == "PHONE PEER" or fieldInfo.CSVFieldName == "PHONE TECHNOLOGY" then
			user.AllegroIntegration = user.AllegroIntegration or {peer={}}

			if fieldInfo.CSVFieldName == "PHONE SERVICE" then
				user.AllegroIntegration.peer.service = csvFieldValue    -- "TELEPHONY_192_168_24_236"
			elseif fieldInfo.CSVFieldName == "PHONE PEER" then
				user.AllegroIntegration.peer.peer = csvFieldValue       -- "2002"
			else
				user.AllegroIntegration.peer.technology = csvFieldValue:gsub("_", ",") -- "ASTERISK,IAX2" (el separado viene con _ y hay que reemplazarlo por ,)
			end

			return nil
		end

		--Hago un conteo de las campannas para validar antes de realizarla importacion que no se supere el limte de relacion users/campaigns
		if fieldInfo.CSVFieldName == "CAMPAIGNS" then
			local campaigns = utils.ParseList(csvFieldValue, ";", function(campaign)
				if #utils.alltrim(campaign) == 0 then
					return
				end
				return {
					id = campaign
				}
			end)

			totalCampaigns = totalCampaigns + #campaigns
		end

		return fieldInfo.fieldName
	end)

	LogDebug("parsedUsersCSV: contCampaigns %s ", totalCampaigns)

	local MAXIMUM_NUMBER_OF_USERS_IN_IMPORT = 500
	local totalUsers = #parsedUsersCSV
	if totalUsers > MAXIMUM_NUMBER_OF_USERS_IN_IMPORT then
		local msg = string.format("The import must not exceed %d users. File contains %d Users.", MAXIMUM_NUMBER_OF_USERS_IN_IMPORT, totalUsers)
		return false, {
			type = "IMPORT EXCEEDS THE LIMIT OF USERS SUPPORTED",
			code = -104,
			reason = msg,
			maximumNumberUsersInImport = MAXIMUM_NUMBER_OF_USERS_IN_IMPORT,
			fileRows = totalUsers
		}
	end

	--Valido el limite de campannas permitidas.
	if totalCampaigns > 0 then
		local success, error = modelRestrictions.ValidateConfigurationsLimits(inconcert.authUser.virtualContactCenter, "MaxCampaignsUsers", keyspace, totalCampaigns)
		if not success then
			return false, {
				type = error or "ERROR_CAMPS_USERS_EXCEEDS_LIMIT_ALLOWED",
				code = -106,
				totalCampaigns = totalCampaigns
			}
		end
	end

	return true, parsedUsersCSV, csvHeader
end

local function StartImport (encodedInfo)
	LogInfo("[ADMINISTRATION.IMPORTS:StartImport] Starting Imports. Info: %s", encodedInfo)

	local vcc = inconcert.authUser.virtualContactCenter
	local info = json.decode(encodedInfo)
	
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")

	local success, parsedUsersCSV, csvHeader = ValidateImportation(info, keyspace)

	--LogDebug("---> StartImport success: %s, parsedUsersCSV: %s", success, parsedUsersCSV)
	if success then
		LogInfo("[ADMINISTRATION.IMPORTS:StartImport] Successfully started %s Importation.", encodedInfo)

		local importation = {
			Id = uuid.new(),
			VCC = vcc,
			Name = info.importName,
			StartDate = LuaOrm.Date():toString(),
			User = inconcert.authUser.userName,
			Status = "RUNNING",
			FileInfo = info.fileInfo,
			CSVHeader = csvHeader
		}
		m_models.Imports.UpdateVCCStatusImportation(importation)
		importation.TotalRows = #parsedUsersCSV

		LogDebug("---> parsedUsersCSV: %s", cjson.encode(parsedUsersCSV))
		LogInfo("[ADMINISTRATION.IMPORTS:StartImport] Importing %d users in VCC '%s'.", importation.TotalRows, vcc)
		--

		m_models.Imports.SaveUserImportation(importation, keyspace)

		local moduleAgentUsers = require "apps.agent.controllers.users" -- un asco cambiar, TODO: realizar request!!
		local importedUsers = {}
		local failedUsers = {}
		local passwordHasBeenAutoGenerated = false
		local updateStatusEvery = 10
		local i = 0
		local applicationsActions = permissions.GetApplicationsActions(keyspace, vcc)

		for line, user in ipairs(parsedUsersCSV) do
			local startCurrentProcessTime = os.time()

			--LogDebug("user-----> %s", json.encode(user))
			user.newUser = true
			user.notify = false
			-- campaigns
			---LogDebug("1 ---> user.groups: %s", user.groups)
			if user.groups then
				user.groups = utils.ParseList(user.groups, ";", function(campaign)
					if #utils.alltrim(campaign) == 0 then
						return
					end
					return {
						id = campaign,
						role = user.UserGroup,
						invite = false
					}
				end)
				--LogDebug("2 ---> user.groups: %s", json.encode(user.groups))
				if #user.groups == 0 then
					user.groups = nil
				end
			end

			-- Si la password no viene dada en el CSV, la misma queda igual al ID de Usuario.
			if not user.Password then
				user.Password = "" -- la misma se genera en "UpdateUser" del controller Users de la app "agent"
				user.IsPlainPassword = true
			end

			local csvLine = line + 1 -- le sumo 1 a los rows para que tenga en cuenta el header en el número de línea
			-- Aca manda guardar el Usuario a Cassandra, contra el controller de la App "AGENT"
			local success, errType = moduleAgentUsers.UpdateUser(user, keyspace, applicationsActions, nil, true)
			if success then
				if user.MustChangePassword then
					passwordHasBeenAutoGenerated = true
				end
				table.insert(importedUsers, {id = user.Id, line = csvLine, password = user.plainPassword.plain})
			else
				if type(errType) == "table" then
					local mappingUsersFields = GetMappingUsersFields("HASH_BY_CASSANDRA_USER_FIELD_NAME")
					if mappingUsersFields[errType.fieldId] then
						local csvFieldName = mappingUsersFields[errType.fieldId].CSVFieldName
						errType.csvFieldName = csvFieldName
					end
				elseif type(errType) == "string" then
					local err = {
						result = "VALUE NOT ALLOWED",
						msg = errType,
						--fieldId = fieldId,
						--fieldValue = tostring(fieldValue)
					}
				end
				table.insert(failedUsers, {id = user.Id, errType = errType, line = csvLine})
			end

			-- Actualizo el status de la importacion, cada N loops (usuarios importados)
			i = i + 1
			if i == updateStatusEvery then
				importation.NumberImportedRows = #importedUsers
				importation.NumberFailedRows = #failedUsers
				m_models.Imports.UpdateVCCStatusImportation(importation)
				i = 0
				-- local status = m_models.Imports.GetVCCStatusImportation()
				-- LogDebug("----> line %d,  status: '%s'", line, status)
			end

			LogInfo("[ADMINISTRATION.IMPORTS:StartImport] Save User '%s' ELAPSED Time '%s' seconds.", user.Id, utils.GetElapsedTimeSec(startCurrentProcessTime))
		end

		--LogDebug("failedUsers: %s", json.encode(failedUsers))
		LogInfo("[ADMINISTRATION.IMPORTS:StartImport] Users Importation result for VCC '%s'. %d Imported , %d Failed.", vcc, #importedUsers, #failedUsers)

		importation.EndDate = LuaOrm.Date():toString()
		importation.Status = "FINISHED"
		importation.NumberImportedRows = #importedUsers
		importation.NumberFailedRows = #failedUsers
		importation.DetailFailedRows = failedUsers
		importation.PasswordHasBeenAutoGenerated = passwordHasBeenAutoGenerated

		if passwordHasBeenAutoGenerated then
		-- Si la contraseña fue autogenerada, dejo un detalle de los usuarios y sus contraseñas planas
			importation.DetailImportedRows = importedUsers
		end

		m_models.Imports.SaveUserImportation(importation, keyspace)

		-- local status = m_models.Imports.GetVCCStatusImportation()
		-- LogDebug("----3 status: '%s'", status)

		keyspace:Close()
		return true, importation
	end

	-- Error
	local err = parsedUsersCSV
	return false, err
end

local function __StartImport (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]

	LogInfo("[ADMINISTRATION.IMPORTS:__StartImport] Starting Imports in '%s' vcc. Posted Info: '%s'", vcc, encodedInfo)

	local startTime = os.time()

	web.headers["Content-Type"] = "application/json"
	local success, importation = StartImport(encodedInfo)

	if success then
		LogInfo("[ADMINISTRATION.IMPORTS:__StartImport] Success Started Imports %s. ELAPSED Time '%s' seconds.", encodedInfo, utils.GetElapsedTimeSec(startTime))
		return importation
	end

	-- Error
	local err = importation
	web.status = 500
	LogError("[ADMINISTRATION.IMPORTS:__StartImport] Error Starting Imports %s. Err: %s", encodedInfo, json.encode(err))

	return err
end

local function __DeleteImportation (web, importationId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ADMINISTRATION.IMPORTS:__DeleteImportation] Deleting Importation '%s' in '%s' vcc. User: '%s'", importationId, vcc, userId)

	local success = m_models.Imports.DeleteUserImportation(importationId)
	if success then
		LogInfo("[ADMINISTRATION.IMPORTS:__DeleteImportation] Success Deleted Importation '%s'. User: '%s'", importationId, userId)
		return {result = "SUCCESS", importationId = importationId}
	end
	-- Error
	web.status = 500
	LogError("[ADMINISTRATION.IMPORTS:__DeleteImportation] Error Deleting Importation '%s'. User: '%s'", importationId, userId)

	return {result = "UNABLE_TO_DELETE_USER_IMPORTATION", importationId = importationId}
end

local function __GetUploadedCSVHeader (web, fileId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ADMINISTRATION.IMPORTS:__GetUploadedCSVHeader] Retrieving CSV Header file '%s'. User %s, vcc %s", fileId, userId, vcc)

	local success, info, csvHeader = ValidateCSVFile(fileId, GetMappingUsersFields)

	local response = {
		result = "OK",
		csvHeader = csvHeader,
		info = {}
	}
	if success then
		return json.encode(response)
	end

	response.result = "ERROR"
	response.info = info

	web.status = 404
	return response
end

local function __GetAllowedHeaderInfo (web)
	LogInfo("[ADMINISTRATION.IMPORTS:__GetAllowedHeaderInfo] Loading Allowed Users Importation Header Columns.")
	local mapping = GetMappingUsersFields()
	return json.encode(mapping)
end

--[[
	Para tener permisos de importar usuarios, debe de tener permisos sobre las 2 acciones (importar usuarios y agregar usuarios)

	Esta funcion es invocada antes de subir a Redis el archivo uploaded
]]--
local function HasPermissionToImportUsers ()
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	if not permissions.IsPermissionEnabledForLoggedUser("administration" , "import_users", nil, ksAgent) then
		ksAgent:Close()
		return false, {errType = "NOT HAS PERMISION", actionId = "import_users"}
	end
	if not permissions.IsPermissionEnabledForLoggedUser("administration" , "saveuser", nil, ksAgent) then
		ksAgent:Close()
		return false, {errType = "NOT HAS PERMISION", actionId = "saveuser"}
	end
	-- enabled permssion
	ksAgent:Close()
	return true
end

--[[
	Guarda el archivo temporal en redis para que luego sea enviado al repositorio de 5.5
]]--
local function UploadTemporalFile (wsapi_env)
	local response = wsapi.response.new(200, { ["Content-Type"]= "text/plain" })
	response:write("")
	response.headers["Response"] = "defaultError"
	if wsapi_env.REQUEST_METHOD:lower() ~= "post" then
		response.status = 405
		return response:finish()
	end

	local hasPermission, err = HasPermissionToImportUsers()
	if not hasPermission then
		LogError("[ADMINISTRATION.IMPORTS:UploadTemporalFile] Unable to upload file. Err: %s - Action: %s ", err.errType, err.actionId)
		response.status = 403
		response:write(err.errType)
		return response:finish()
	end

	local request = wsapi.request_stream.new(wsapi_env)
	local section = tonumber(request.params.section)
	local timestamp = os.time(os.date("!*t")) -- current date in UTC

	local uploadedFile = {
		FileId      = request.params.id,
		Section     = section                                ,
		Format      = request.params.format                  ,
		ContentType = request.params.contentType             ,
		FileName    = request.params.fileName                ,
		ServerPath  = request.params.path                    ,
		FileSize    = tonumber(wsapi_env.CONTENT_LENGTH) or 0,
		Campaign    = request.params.campaign                ,
		Account     = request.params.account                 ,
		TimeStamp   = timestamp
	}

	LogInfo("[ADMINISTRATION.IMPORTS:UploadTemporalFile] Upload temporal file. File info: %s", json.encode(uploadedFile))

	if uploadedFile.FileSize <= 0 then
		LogError("[ADMINISTRATION.IMPORTS:UploadTemporalFile] Unable to upload file. Current file size is %d", uploadedFile.FileSize)
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("Invalid file. Zero length")
		return response:finish()
	end

	if not uploadedFile.FileId or not uploadedFile.Section or not uploadedFile.Format or not uploadedFile.ContentType then
		LogError("[ADMINISTRATION.IMPORTS:UploadTemporalFile] Unable to upload file. Incomplete file information. Info: %s", json.encode(uploadedFile))
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("Invalid file data")
		return response:finish()
	end

	-- aca piso la metatabla para poder encargarme yo de la leida del POST data
	setmetatable(request.POST, nil)

	local chunkSize = 2048
	local table_file = {}

	local left = uploadedFile.FileSize
	local blocksize = math.min(chunkSize, left)
	while left > 0 do
		local chunk = wsapi_env.input:read(blocksize)
		if not chunk then break end
		table.insert(table_file, chunk)
		left = left - #chunk
		blocksize = math.min(blocksize, left)
	end

	local temporalFile = table.concat(table_file)
	local fileKey = m_models.Imports.SetTemporalUploadedFile(uploadedFile, temporalFile)

	LogInfo("[ADMINISTRATION.IMPORTS:UploadTemporalFile] Successfully uploaded temporal file. File key: '%s', file info: %s", fileKey, json.encode(uploadedFile))

	local result = {fileId = uploadedFile.FileId, fileName = uploadedFile.FileName, fileSize = uploadedFile. FileSize}
	response:write(json.encode(result))

	return response:finish()
end

local function RemoveTemporalFile (web)
	web.headers["Response"] = "defaultError"

	local file = web.GET
	LogInfo("[ADMINISTRATION.IMPORTS:RemoveTemporalFile] Removing file: %s.", json.encode(file))

	local uploadedFile = {
		FileId  = file.id,
		Section = file.section
	}
	local success = m_models.Imports.DeleteTemporalUploadedFile(uploadedFile)

	return json.encode( success )
end

local function RemoveAllTemporalFiles (web)
	web.headers["Response"] = "defaultError"

	LogInfo("[ADMINISTRATION.IMPORTS:RemoveAllTemporalFiles] Removing multiples files: '%s'", web.input["post_data"])

	local result = { deleted = {}, failed  = {} }
	local files = json.decode(web.input["post_data"])

	for _, file in ipairs(files or {}) do
		local success = m_models.Imports.DeleteTemporalUploadedFile(file)
		if success then
			table.insert(result.deleted, file)
		else
			table.insert(result.failed, file)
		end
	end

	return json.encode(result)
end

local function __LoadImportations (web)
	local vcc = inconcert.authUser.virtualContactCenter

	LogInfo("[ADMINISTRATION.IMPORTS:__LoadImportations] Loading Users Importations for vcc %s", vcc)

	local usersImportations = m_models.Imports.LoadUsersImportations()
	-- Aca recorro las importaciones y valido que las que se encuentren en estado RUNNING en cassandra, se encuentre vivas en REDIS.
	--  En el caso de que la importación NO se encuentre viva en REDIS, la marco con el status ERROR
	for importId, importation in pairs(usersImportations) do
		if (importation.Status or ""):upper() == "RUNNING" then
			-- La busco en REDIS
			local encodedImportationStatus = m_models.Imports.GetVCCStatusImportation()
			--LogDebug("---> BD Import: '%s', running importation status: '%s'", importId, encodedImportationStatus)
			local ok, importationStatus = pcall(json.decode, encodedImportationStatus)
			if not ok then
				importationStatus = nil
			end
			if not importationStatus or (importationStatus.Id or ""):lower() ~= importId:lower() then
				LogInfo("[ADMINISTRATION.IMPORTS:__LoadImportations] Mark import '%s' with error in VCC '%s'.", importId, vcc)
				-- La marco como Erronea
				importation.Status = "ERROR"
				m_models.Imports.SaveUserImportation(importation)
			end
		end
	end

	local encodedUsersImportations = json.encode(usersImportations)
	--LogDebug("---__LoadImportations------> usersImportations: %s", encodedUsersImportations)

	return encodedUsersImportations
end


---------------------------
-- SHARED FILES
---------------------------
local function GetMappingSharedFilesFields (mappingType)
	local vcc = inconcert.authUser.virtualContactCenter

	-- Name,Description,Language,AgentVisible,AccountMessagesVisible,FlowDesignerVisible
	-- LOS REQUERIDOS
	local requiredFields = {
		-- el nombre/id del archivo único y válido
		{CSVFieldName = "NAME"                  , fieldName = "Name"                  , required = true },
		-- un rol del sistema o personalizado válido
		{CSVFieldName = "LANGUAGE"              , fieldName = "Language"              , required = true }
	}
	-- LOS NO REQUERIDOS
	local notRequiredFields = {
		-- la descripcion
		{CSVFieldName = "DESCRIPTION"           , fieldName = "Description"           , required = false},
		-- si es visible por el agente
		{CSVFieldName = "AGENTVISIBLE"          , fieldName = "AgentVisible"          , required = false},
		-- si es visible para las cuentas de los mensajes
		{CSVFieldName = "ACCOUNTMESSAGESVISIBLE", fieldName = "AccountMessagesVisible", required = false},
		-- si es visible en el flow designer
		{CSVFieldName = "FLOWDESIGNERVISIBLE"   , fieldName = "FlowDesignerVisible"   , required = false}
	}

	local mapping = {}
	local function Fill (fields)
		for _, field in ipairs(fields) do
			table.insert(mapping, field)
		end
	end

	Fill(requiredFields)
	Fill(notRequiredFields)

	if mappingType == "HASH_BY_CSV_FIELD_NAME" then
		return utils.ArrayToHash(mapping, "CSVFieldName")
	elseif mappingType == "HASH_BY_CASSANDRA_FIELD_NAME" then
		return utils.ArrayToHash(mapping, "fieldName")
	end

	-- simple array
	return mapping
end

local function __GetAllowedHeaderSharedFiles (web)
	LogInfo("[ADMINISTRATION.IMPORTS:__GetAllowedHeaderSharedFiles] Loading Allowed Shared Files Header Columns.")
	local mapping = GetMappingSharedFilesFields()
	return json.encode(mapping)
end

local SHARED_FILES_MAXIMUM_NUMBER_OF_FILES_IN_IMPORT = 500
function __SharedFilesLoadUploadedCSVToJson (web, fileId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ADMINISTRATION.IMPORTS:__SharedFilesLoadUploadedCSVToJson] Shared Files Loading CSV file to JSON '%s'. User %s, vcc %s", fileId, userId, vcc)

	local response = LoadUploadedCSVToJson(web, fileId)
	if response.result ~= "OK" then
		return response
	end

	local totalRows = #response.jsonValues
	if totalRows > SHARED_FILES_MAXIMUM_NUMBER_OF_FILES_IN_IMPORT then
		web.status = 400
		local msg = string.format("The import must not exceed %d files. File contains %d files.", SHARED_FILES_MAXIMUM_NUMBER_OF_FILES_IN_IMPORT, totalRows)
		LogError("[ADMINISTRATION.IMPORTS:__SharedFilesLoadUploadedCSVToJson] Shared Files Error. %s", msg)
		return {
			result = "ERROR",
			info = {
				type = "IMPORT EXCEEDS THE LIMIT OF FILES SUPPORTED",
				code = -104,
				reason = msg,
				maximumNumberRowsInImport = SHARED_FILES_MAXIMUM_NUMBER_OF_FILES_IN_IMPORT,
				fileRows = totalRows
			}
		}
	end

	-- OK
	return response
end

-------------------------------------------
-- GENERALES
-------------------------------------------

function LoadUploadedCSVToJson (web, fileId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ADMINISTRATION.IMPORTS:LoadUploadedCSVToJson] Loading CSV file to JSON '%s'. User %s, vcc %s", fileId, userId, vcc)

	local success, csvValues, csvHeader = ValidateCSVFile(fileId, GetMappingSharedFilesFields)
	if not success then
		local response = {}
		response.result = "ERROR"
		response.info = csvValues
		web.status = 404
		return response
	end

	local jsonValues = utils.Csv2Table(csvValues)

	local response = {
		result = "OK",
		csvHeader = csvHeader,
		jsonValues = jsonValues,
		csvValues = csvValues,
		info = {}
	}

	return response
end

--Funcion que devuelve un json con los datos del csv y valida por columnas pasada por parametros
--CSVFieldName debe andar en upper case
		--[[{
			{CSVFieldName = "NAME", required = true }	
		}]]--
function LoadCustomUploadedCSVToJson (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local request = json.decode(web.input["post_data"] or "{}")
	LogInfo("[ADMINISTRATION.IMPORTS:LoadCustomUploadedCSVToJson] Loading CSV file to JSON '%s'. User %s, vcc %s", web.input["post_data"], userId, vcc)

	if not request.fileId or not request.CSVFields then
		local response = {}
		response.result = "ERROR"
		response.info = {code = -100}
		web.status = 404
		return response
	end

	local success, csvValues, csvHeader = ValidateCSVFile(request.fileId, function (mappingType)
		if mappingType == "HASH_BY_CSV_FIELD_NAME" then
			return utils.ArrayToHash(request.CSVFields, "CSVFieldName")
		end
		return request.CSVFields
	end)
	
	if not success then
		local response = {}
		response.result = "ERROR"
		response.info = csvValues
		response.csvHeader = csvHeader
		web.status = 404
		return response
	end

	local jsonValues = utils.Csv2Table(csvValues, function (column, value, row)
		return column:lower()
	end)
	local totalValues = #jsonValues
		
		if request.Limit and totalValues -1 > tonumber(request.Limit) then
			local response = {}
			response.result = "ERROR"
			response.info = {
				code = -104,
				maximumNumberRowsInImport = tonumber(request.Limit),
				fileRows = totalValues
			}
			response.csvHeader = csvHeader
			web.status = 404
			return response
		end

	local response = {
		result = "OK",
		csvHeader = csvHeader,
		jsonValues = jsonValues,
		csvValues = csvValues,
		info = {}
	}

	return response
end

m_htmlRender = nil
m_appLocation = nil

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest

	local replyFn = utils.JsonReply
	local replyValFn = utils.JsonValidatorReply

	local function HasPermission (fn, actionId, responseFn)
		actionId = actionId or "import_users"
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. permissions:CheckPermission("administration", actionId, fn)
	end

	local jsonValReply = utils.JsonValidatorReply .. LogRequest
	local jsonReply = utils.JsonReply .. LogRequest
	local p = prefix

	-- Importation functions
	module:dispatch_get   (jsonReply .. __LoadImportations                , p .. "/load_all/"                                  )
	module:dispatch_post  (
		HasPermission(
			__StartImport,
			-- para tener permisos de importar usuarios, debe de tener permisos sobre las 2 acciones (importar usuarios y agregar usuarios)
			{"import_users", "saveuser"}
		)
		, p .. "/start/"
	)
	module:dispatch_delete(HasPermission(__DeleteImportation)             , p .. "/delete/(.*)"                                )
	module:dispatch_get   (jsonReply .. __GetUploadedCSVHeader            , p .. "/get_uploaded_csv_header/(.+)"               )
	module:dispatch_get   (jsonReply .. LoadUploadedCSVToJson             , p .. "/load_csv_uploaded_to_json/(.+)"             )
	module:dispatch_get   (jsonReply .. __GetAllowedHeaderInfo            , p .. "/get_allowed_header_info/"                   )
	module:dispatch_get   (jsonReply .. __GetAllowedHeaderSharedFiles     , p .. "/get_allowed_header_shared_files/"           )
	module:dispatch_post  (jsonReply .. LoadCustomUploadedCSVToJson       , p .. "/load_custom_csv_uploaded_to_json/"                                     )
	-- CSV functions
	module:dispatch_wsapi (UploadTemporalFile                             , p .. "/upload/([^.]*)"                             )
	module:dispatch_get   (RemoveTemporalFile                             , p .. "/remove/(.*)"                                )
	module:dispatch_post  (RemoveAllTemporalFiles                         , p .. "/remove_all/([^.]*)"                         )

	-- SharedFiles
	module:dispatch_get   (jsonReply .. __SharedFilesLoadUploadedCSVToJson, p .. "/shared_files/load_csv_uploaded_to_json/(.+)")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end

function Public ()
	return {
		SharedFilesGetMappingFields = GetMappingSharedFilesFields,
		SharedFilesGetMaxNumberSupportedFilesImport = function() return SHARED_FILES_MAXIMUM_NUMBER_OF_FILES_IN_IMPORT end
	}
end
