module(..., package.seeall)
local utils = require "shared.apps.utils"
local agent_model = require("apps.agent.model"):GetModels()
local m_datasource

local DIALING_SCHEMA_SCOPES = {
    "Local",
    "SameCountry",
    "Global"
}

function GetName ()
    return "PhoneValidator"
end

function GetDefinition (datasource)
    m_datasource = datasource
    return _M
end

function GetSystemSchemas(countryId)
    local path = "/usr/share/inconcert/helpers/phonecodes.lua"
	local runner = loadfile(path)
	runner()
    if countryId then
        return codeData[countryId] or {}
    else
        return codeData
    end
end

function GetVccSchemas (vcc, countryId, keyspace)
    local schemas = utils.GetAllColumns {
        keyspace = keyspace or "Agent",
        columnFamily = "PhoneValidator",
        superColumns = true,
        key = vcc,
        columns = countryId and {countryId} or {}
    }
    if countryId then
        return schemas and schemas[countryId] or {}
    else
        return schemas or {}
    end
end

local function MergePrefix(systemData, vccData)
    if not vccData then
        return {
            data = systemData,
            status = "original"
        }
    end

    if not systemData then
        vccData.status = "added"
        return vccData
    end

    vccData.status = "updated"
    vccData.original = systemData
    return vccData
end

local function MergePrefixes(systemPrefixes, vccPrefixes)
    local mergedPrefixes = {}
    for prefix, systemData in pairs(systemPrefixes) do
        mergedPrefixes[prefix] = MergePrefix(systemData, vccPrefixes[prefix])
    end
    for prefix, vccData in pairs(vccPrefixes) do
        if not systemPrefixes[prefix] then
            mergedPrefixes[prefix] = MergePrefix(nil, vccData)
        end
    end
    return mergedPrefixes
end

local function MergeDialingSchemas(systemSchemas, vccSchemas)
    local mergedSchemas = {}
    for _, scope in pairs(DIALING_SCHEMA_SCOPES) do
        mergedSchemas[scope] = MergePrefixes(systemSchemas[scope] or {}, vccSchemas[scope] or {})
    end
    return mergedSchemas
end

local function MergeArea(systemArea, vccArea)
    return {
        description = vccArea.description or systemArea.description,
        dialingSchemas = MergeDialingSchemas(systemArea.dialingSchemas or {}, vccArea.dialingSchemas or {}),
        prefixes = MergePrefixes(systemArea.prefixes or {}, vccArea.prefixes or {}),
        timezones = MergePrefixes(systemArea.timezones or {}, vccArea.timezones or {}),
    }
end

local function MergeAreas(systemAreas, vccAreas)
    local mergedAreas = {}
    for areaCode, systemData in pairs(systemAreas) do
        mergedAreas[areaCode] = MergeArea(systemData, vccAreas[areaCode] or {})
        mergedAreas[areaCode].status = vccAreas[areaCode] and "updated" or "original"
    end
    for areaCode, vccData in pairs(vccAreas) do
        if not systemAreas[areaCode] then
            mergedAreas[areaCode] = MergeArea({}, vccData)
            mergedAreas[areaCode].status = "added"
        end
    end
    return mergedAreas
end

local function MergeCountrySchemas(systemSchema, vccSchema)
    return {
        description = systemSchema.description,
        dialingSchemas = MergeDialingSchemas(systemSchema.dialingSchemas or {}, vccSchema.dialingSchemas or {}),
        prefixes = MergePrefixes(systemSchema.prefixes or {}, vccSchema.prefixes or {}),
        areas = MergeAreas(systemSchema.areas or {}, vccSchema.areas or {}),
    }
end

function GetMergedSchemas(vcc)
    local systemSchemas = GetSystemSchemas()
    local vccSchemas = GetVccSchemas(vcc)
    
    local mergedSchemas = {}
    for countryId, systemSchema in pairs(systemSchemas) do
        mergedSchemas[countryId] = MergeCountrySchemas(systemSchema, vccSchemas[countryId] or {})
    end
    return mergedSchemas
end

local function StorePhoneValidatorCountry (vcc, countryId, country, keyspace)
    utils.SaveRows {
        keyspace = keyspace or "Agent",
        columnFamily = "PhoneValidator",
        superColumns = true,
        rows = {
            [vcc] = { [countryId] = country }
        }
    }
end

local function GetCountryAreas (vcc, countryId, keyspace)
    local countryKey = utils.AppendVcc(countryId, vcc)
    local data = utils.GetAllColumns {
        keyspace = keyspace or "Provisioning",
        columnFamily = "Countries",
        key = countryKey,
        columns = {"Areas"}
    }
    return data and data.Areas or {}
end

local function StoreCountryAreas (vcc, countryId, areas, keyspace)
    local countryKey = utils.AppendVcc(countryId, vcc)
    utils.SaveRows {
        keyspace = keyspace or "Provisioning",
        columnFamily = "Countries",
        rows = {
            [countryKey] = { Areas = areas }
        }
    }
end

local function StorePhoneValidatorAreaDescription(vcc, countryId, areaCode, description)
    local countryAreas = GetVccSchemas(vcc, countryId).areas or {}
    countryAreas[areaCode] = countryAreas[areaCode] or {}
    countryAreas[areaCode].description = description
    StorePhoneValidatorCountry(vcc, countryId, { areas = countryAreas })
end

local function StoreCountryAreaDescription(vcc, countryId, areaCode, description)
    local countryAreas = GetCountryAreas(vcc, countryId).areas or {}
    countryAreas[areaCode] = {
        id = areaCode,
        name = description
    }
    StoreCountryAreas(vcc, countryId, countryAreas)
end

local function StoreAreaDescription(vcc, countryId, areaCode, description)
    StorePhoneValidatorAreaDescription(vcc, countryId, areaCode, description)
    StoreCountryAreaDescription(vcc, countryId, areaCode, description)
end

function SaveVccArea (vcc, countryId, area)
    local areaCode = area.areaCode
    local description = area.description
    local timezone = area.timezone
    StoreAreaDescription(vcc, countryId, areaCode, description)
    if timezone ~= nil then
        SaveVccRule(vcc, countryId, {
            areaCode = areaCode,
            type = "timezone",
            prefix = "",
            description = "",
            data = timezone
        })
    end
end

local function DeletePhoneValidatorArea(vcc, countryId, areaCode)
    local countryAreas = GetVccSchemas(vcc, countryId).areas or {}
    countryAreas[areaCode] = nil
    StorePhoneValidatorCountry(vcc, countryId, { areas = countryAreas })
end

local function DeleteCountryArea(vcc, countryId, areaCode)
    local countryAreas = GetCountryAreas(vcc, countryId).areas or {}
    countryAreas[areaCode] = nil
    StoreCountryAreas(vcc, countryId, countryAreas)
end

local function RemoveAreaDialing(vcc, countryId, areaCode)
    local provisioningKS = utils.GetDatasource():CassandraConnection("Provisioning")
    provisioningKS:RemoveColumn(
        utils.AppendVcc(countryId.."-"..areaCode, vcc),
        "PhoneNumberNormalization"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc(countryId.."-"..areaCode, vcc),
        "CountryCodeTimeZones"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc("Local:"..countryId.."/"..areaCode, vcc),
        "PhoneDialingSchemas"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc("SameCountry:"..countryId.."/"..areaCode, vcc),
        "PhoneDialingSchemas"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc("Global:"..countryId.."/"..areaCode, vcc),
        "PhoneDialingSchemas"
    )
    provisioningKS:Close()
end

function DeleteVccArea(vcc, countryId, area)
    local areaCode = area.areaCode
    DeletePhoneValidatorArea(vcc, countryId, areaCode)
    DeleteCountryArea(vcc, countryId, areaCode)
    RemoveAreaDialing(vcc, countryId, areaCode)
end

local function DeletePhoneValidatorAreas(vcc, countryId, keyspace)
    utils.RemoveColumns {
        keyspace = keyspace or "Agent",
        columnFamily = "PhoneValidator",
        superColumns = true,
        rows = {
            [vcc] = { [countryId] = { "areas" } }
        }
    }
end

local function DeleteCountryAreas(vcc, countryId)
    local countryKey = utils.AppendVcc(countryId, vcc)
    utils.RemoveColumns {
        keyspace = keyspace or "Provisioning",
        columnFamily = "Countries",
        rows = {
            [countryKey] = "Areas"
        }
    }
end

function DeleteVccAreas(vcc, countryId)
    local countryAreas = GetVccSchemas(vcc, countryId).areas or {}
    for areaCode, _ in pairs(countryAreas) do
        RemoveAreaDialing(vcc, countryId, areaCode)
    end
    DeletePhoneValidatorAreas(vcc, countryId)
    DeleteCountryAreas(vcc, countryId)
end

local function MakePhoneNumberNormalizationKey(vcc, countryId, rule)
    local key = countryId
    if rule.areaCode then
        key = key .. "-" .. rule.areaCode
    end
    return utils.AppendVcc(key, vcc)
end

local function MakePhoneDialingSchemaKey(vcc, countryId, rule)
    local key = rule.schemaType .. ":" .. countryId
    if rule.areaCode then
        key = key .. "/" .. rule.areaCode
    end
    return utils.AppendVcc(key, vcc)
end

local function MakeCountryCodeTimeZoneKey(vcc, countryId, rule)
    local key = countryId
    if rule.areaCode then
        key = key .. "-" .. rule.areaCode
    end
    return utils.AppendVcc(key, vcc)
end

local function EmptyPrefixes(prefixes)
    if not prefixes then return true end

    for _, data in pairs(prefixes) do
        if data then return false end
    end
    return true
end

local RULE_METHODS = {
    validation = {
        append = function(data, rule)
            local prefixes = data.prefixes or {}
            prefixes[rule.prefix] = { data = rule.data, description = rule.description }
            return "prefixes", prefixes
        end,
        remove = function(data, rule)
            local prefixes = data.prefixes or {}
            prefixes[rule.prefix] = nil
            return "prefixes", prefixes
        end,
        appendDialing = function(vcc, countryId, rule)
            local key = MakePhoneNumberNormalizationKey(vcc, countryId, rule)
            utils.SaveRows {
                keyspace = keyspace or "Provisioning",
                columnFamily = "PhoneNumberNormalization",
                rows = {
                    [key] = { ["." .. rule.prefix] = rule.data }
                }
            }
        end,
        removeDialing = function(vcc, countryId, rule)
            local key = MakePhoneNumberNormalizationKey(vcc, countryId, rule)
            utils.RemoveColumns {
                keyspace = keyspace or "Provisioning",
                columnFamily = "PhoneNumberNormalization",
                rows = {
                    [key] = "." .. rule.prefix
                }
            }
        end,
    },
    schema = {
        append = function(data, rule)
            local prefixes = data.dialingSchemas or {}
            prefixes[rule.schemaType] = prefixes[rule.schemaType] or {}
            prefixes[rule.schemaType][rule.prefix] = { data = rule.data, description = rule.description }
            return "dialingSchemas", prefixes
        end,
        remove = function(data, rule)
            local dialingSchemas = data.dialingSchemas or {}
            dialingSchemas[rule.schemaType] = dialingSchemas[rule.schemaType] or {}
            dialingSchemas[rule.schemaType][rule.prefix] = nil
            local cleanSchemas = {}
            for schema, prefixes in pairs(dialingSchemas) do
                if not EmptyPrefixes(prefixes) then
                    cleanSchemas[schema] = prefixes
                end
            end
            return "dialingSchemas", cleanSchemas
        end,
        appendDialing = function(vcc, countryId, rule)
            local key = MakePhoneDialingSchemaKey(vcc, countryId, rule)
            utils.SaveRows {
                keyspace = keyspace or "Provisioning",
                columnFamily = "PhoneDialingSchemas",
                rows = {
                    [key] = { ["." .. rule.prefix] = rule.data }
                }
            }
        end,
        removeDialing = function(vcc, countryId, rule)
            local key = MakePhoneDialingSchemaKey(vcc, countryId, rule)
            utils.RemoveColumns {
                keyspace = keyspace or "Provisioning",
                columnFamily = "PhoneDialingSchemas",
                rows = {
                    [key] = "." .. rule.prefix
                }
            }
        end,
    },
    timezone = {
        append = function(data, rule)
            local prefixes = data.timezones or {}
            prefixes[rule.prefix] = { data = rule.data, description = rule.description }
            return "timezones", prefixes
        end,
        remove = function(data, rule)
            local prefixes = data.timezones or {}
            prefixes[rule.prefix] = nil
            return "prefixes", prefixes
        end,
        appendDialing = function(vcc, countryId, rule)
            local key = MakeCountryCodeTimeZoneKey(vcc, countryId, rule)
            utils.SaveRows {
                keyspace = keyspace or "Provisioning",
                columnFamily = "CountryCodeTimeZones",
                rows = {
                    [key] = { ["." .. rule.prefix] = rule.data }
                }
            }
        end,
        removeDialing = function(vcc, countryId, rule)
            local key = MakeCountryCodeTimeZoneKey(vcc, countryId, rule)
            utils.RemoveColumns {
                keyspace = keyspace or "Provisioning",
                columnFamily = "CountryCodeTimeZones",
                rows = {
                    [key] = "." .. rule.prefix
                }
            }
        end,
    },
}

local function AppendRule(data, rule)
    return RULE_METHODS[rule.type].append(data, rule)
end

local function RemoveRule(data, rule)
    return RULE_METHODS[rule.type].remove(data, rule)
end

local function AppendDialingRule(vcc, countryId, rule)
    return RULE_METHODS[rule.type].appendDialing(vcc, countryId, rule)
end

local function RemoveDialingRule(vcc, countryId, rule)
    return RULE_METHODS[rule.type].removeDialing(vcc, countryId, rule)
end

local function StorePhoneValidatorAreaRule(vcc, countryId, areaCode, rule)
    local countryAreas = GetVccSchemas(vcc, countryId).areas or {}
    countryAreas[areaCode] = countryAreas[areaCode] or {}
    local column, data = AppendRule(countryAreas[areaCode], rule)
    countryAreas[areaCode][column] = data
    StorePhoneValidatorCountry(vcc, countryId, { areas = countryAreas })
end

local function StorePhoneValidatorGlobalRule(vcc, countryId, rule)
    local country = GetVccSchemas(vcc, countryId)
    local column, data = AppendRule(country, rule)
    StorePhoneValidatorCountry(vcc, countryId, { [column] = data })
end

local function StorePhoneValidatorRule(vcc, countryId, rule)
    if rule.areaCode then
        StorePhoneValidatorAreaRule(vcc, countryId, rule.areaCode, rule)
    else
        StorePhoneValidatorGlobalRule(vcc, countryId, rule)
    end
end

function SaveVccRule(vcc, countryId, rule)
    StorePhoneValidatorRule(vcc, countryId, rule)
    AppendDialingRule(vcc, countryId, rule)
end

local function DeletePhoneValidatorAreaRule(vcc, countryId, areaCode, rule)
    local countryAreas = GetVccSchemas(vcc, countryId).areas or {}
    countryAreas[areaCode] = countryAreas[areaCode] or {}
    local column, data = RemoveRule(countryAreas[areaCode], rule)
    countryAreas[areaCode][column] = not EmptyPrefixes(data) and data or nil
    if EmptyPrefixes(countryAreas[areaCode]) then
        countryAreas[areaCode] = nil
        if EmptyPrefixes(countryAreas) then
            return DeletePhoneValidatorAreas(vcc, countryId)
        end
    end
    StorePhoneValidatorCountry(vcc, countryId, { areas = countryAreas })
end

local function DeletePhoneValidatorGlobalRule(vcc, countryId, rule)
    local country = GetVccSchemas(vcc, countryId)
    local column, data = RemoveRule(country, rule)
    if EmptyPrefixes(data) then
        return utils.RemoveColumns {
            keyspace = "Agent",
            columnFamily = "PhoneValidator",
            superColumns = true,
            rows = {
                [vcc] = { [countryId] = { column } }
            }
        }
    end
    StorePhoneValidatorCountry(vcc, countryId, { [column] = data })
end

local function DeletePhoneValidatorRule(vcc, countryId, rule)
    if rule.areaCode then
        DeletePhoneValidatorAreaRule(vcc, countryId, rule.areaCode, rule)
    else
        DeletePhoneValidatorGlobalRule(vcc, countryId, rule)
    end
end

function DeleteVccRule(vcc, countryId, rule)
    DeletePhoneValidatorRule(vcc, countryId, rule)
    RemoveDialingRule(vcc, countryId, rule)
end

local function DeletePhoneValidatorGlobalRules(vcc, countryId, keyspace)
    utils.RemoveColumns {
        keyspace = keyspace or "Agent",
        columnFamily = "PhoneValidator",
        superColumns = true,
        rows = {
            [vcc] = { [countryId] = { "dialingSchemas", "prefixes" } }
        }
    }
end

local function RemoveGlobalDialingRules(vcc, countryId)
    local provisioningKS = utils.GetDatasource():CassandraConnection("Provisioning")
    provisioningKS:RemoveColumn(
        utils.AppendVcc(countryId, vcc),
        "PhoneNumberNormalization"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc("Local:"..countryId, vcc),
        "PhoneDialingSchemas"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc("SameCountry:"..countryId, vcc),
        "PhoneDialingSchemas"
    )
    provisioningKS:RemoveColumn(
        utils.AppendVcc("Global:"..countryId, vcc),
        "PhoneDialingSchemas"
    )
    provisioningKS:Close()
end

function RemoveGlobalDialing(vcc, countryId)
    DeletePhoneValidatorGlobalRules(vcc, countryId)
    RemoveGlobalDialingRules(vcc, countryId)
end
