module(..., package.seeall)
local utils = require "shared.apps.utils"
local agent_model = require("apps.agent.model"):GetModels()
local allegro_skills = require "apps.allegro.controllers.skills"
local m_datasource

function GetName ()
	return "Skills"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function GetSkills (vcc, skillId, keyspace)
	local skills = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCSkills",
		superColumns = true,
		key = vcc,
		columns = skillId and {skillId} or {}
	}
	if skillId then
		return skills and skills[skillId] or {}
	else
		return skills or {}
	end
end

--[[
	-- Guarda en Allegro los skills
]]--
local function UpdateAllegroSkillsIn55( skills, vcc, keyspace )

	if not utils.HasAllegroIntegration( vcc, keyspace ) then
		LogInfo("[ADMINISTRATION:UpdateAllegroSkillsIn55] - Allegro integration. Updating skills '%s'. Vcc '%s' has not configured integration with allegro.", json.encode(skills), vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	local success, err = allegro_skills.Save( skills, vcc )
	if not success then
		LogError("[ADMINISTRATION:UpdateAllegroSkillsIn55] - Allegro integration error. Updating skills '%s'. Error: '%s'.", json.encode(skills), json.encode(err))
		return false, { type = "ALLEGRO_INTEGRATION_ERROR", err = err }
	end

	return true
end

function SaveSkills (vcc, skills, keyspace)

	-- Guarda en Allegro los skills
	-- Debo mandar a actualizar aca, en este lugar, antes de guardar algo en cassandra, asi me evito el Rollback y toda esas cosas!!!
	--
	local success, err = UpdateAllegroSkillsIn55( skills, vcc, keyspace )
	if not success then
		-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
		--  la actualizacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			if close_ks then keyspace:Close() end
			return false, { result = err.type }
		end
	end

	utils.SaveRows {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCSkills",
		superColumns = true,
		rows = {
			[vcc] = skills
		}
	}

	return true
end


--[[
	-- Delete en Allegro los skills
]]--
local function DeleteAllegroSkillsIn55( skills_to_remove, keyspace )

	local vcc = inconcert.authUser.virtualContactCenter

	if not utils.HasAllegroIntegration( vcc, keyspace ) then
		LogInfo("[ADMINISTRATION:DeleteAllegroSkillsIn55] - Allegro integration. Deleting skills '%s'. Vcc '%s' has not configured integration with allegro.", json.encode(skills_to_remove), vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	local success, err = allegro_skills.Delete( skills_to_remove, vcc )
	if not success then
		if err.type == "ALLEGRO_INTEGRATION_ERROR" then
			LogError("[ADMINISTRATION:DeleteAllegroSkillsIn55] - Allegro integration error. Deleting skills '%s'. Response: '%s'.", json.encode(skills_to_remove), json.encode(err))
		else
			LogInfo("[ADMINISTRATION:DeleteAllegroSkillsIn55] - Allegro integration. Deleting skills '%s'. Response: '%s'.", json.encode(skills_to_remove), json.encode(err))
		end
		return false, err
	end

	return true
end

function RemoveSkills (vcc, skills_to_remove, keyspace)

	local close_ks = false
	if not keyspace then
		keyspace = m_datasource.cassandra:CassandraConnection("Agent")
		close_ks = true
	end

	local success, err = DeleteAllegroSkillsIn55( skills_to_remove, keyspace )
	if not success then
		-- Solo tiene que retornar sin eliminar el skill en i6, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar la eliminacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			keyspace:Close()
			return false, { result = err.type, err = (err.err or err.sub_type) }
		end
	end

	utils.RemoveColumns {
		keyspace = keyspace,
		columnFamily = "VCCSkills",
		superColumns = true,
		rows = {
			[vcc] = skills_to_remove
		}
	}

	-- 1- Elimino los skills utilizados por el usuario.
	--
	-- Recorro la lista de skills a remover y valido si el skill a remover se encuentra utilizado por algun usuario.
	-- Si se encuentra utilizado por algun usuario, lo quito de la lista de skills del usuario, y mando actualizar la lista a cassandra.
	local users_skills = GetAllAgentSkills(vcc, keyspace) or {}
	-- aca recorro la lista de skills a remover
	for _, skill_id_to_remove in ipairs( skills_to_remove ) do
		-- aca recorro la lista de todos los usuarios que tengan skills
		for user_id, user_skills in pairs(users_skills) do
			-- aca recorro los skills de un usuario en particular
			for user_skill_id in pairs(user_skills) do
				if user_skill_id == skill_id_to_remove then
					-- aca se encontro que el skill a remover que se encuentra utilizado por el usuario,
					-- lo quito de la lista y actualizo los skills del usuario.
					user_skills[ user_skill_id ] = nil
					-- tengo que agregarlos en un array a los skills asociados al usuario que quedan, ya que en este formato
					-- los espera la funcion "UpdateUserSkills"
					local arr_user_skills = {}
					for id, value in pairs(user_skills) do
						local skill = {
							skill  = id,
							weight = value
						}
						table.insert(arr_user_skills, skill)
					end
					agent_model.Users.UpdateUserSkills( keyspace, user_id, arr_user_skills )
					break
				end
			end
		end
	end

	-- 2- Elimino los skills utilizados por las cuentas.
	--
	-- Recorro la lista de skills a remover, luego por cada uno de ellos recorro la lista de accounts que tienen el skill y mando a actualizar
	--  la cuenta con los skills que le han quedado asociados.
	--
	for _, skill_id_to_remove in ipairs( skills_to_remove ) do
		local skill_id_without_vcc = utils.GetId(skill_id_to_remove)
		-- obtengo la lista de ids cuentas por skill (los skills a remover)
		local accounts_skill = GetAccountsBySkill( vcc, skill_id_to_remove, keyspace ) or {}
		-- ahora recorro esta lista de ids de cuentas.
		for skill_account_id, account_skill in pairs( accounts_skill ) do
			-- obtengo la campana y recorro sus cuentas buscando en el array de cuentas los skills a eliminar.
			local accounts = agent_model.Campaigns.GetCampaignAccounts(account_skill.campaign_id, keyspace)
			for _, account in ipairs( accounts or {} ) do
				if skill_account_id == account.id then
					-- recorro la lista de skills de la cuenta y se encuentro el skill a elimminar, actualizo la cuenta
					--  eliminando este skill que ya no pertenece a la misma
					for i, skill in ipairs( account.Skills or {} ) do
						if skill.skill == skill_id_without_vcc then
							-- aca remuevo del array el skill que ya no pertenece y mando a guardar
							table.remove( account.Skills, i )
							agent_model.Campaigns.SaveAccount(
								keyspace,
								account_skill.campaign_id,
								skill_account_id,
								account,
								false -- false == actualizando la account
							)
							break
						end
					end
				end
			end
		end
	end

	if close_ks then keyspace:Close() end

	return true
end

function GetAgentSkills (vcc, user_id, keyspace)
	user_id = utils.AppendVcc( user_id, vcc )
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "AgentSkills",
		key = user_id
	}
end

function GetAllAgentSkills (vcc, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "AgentSkills",
		startKey = "",
		endKey = ""
	}
end

function GetAccountsBySkill (vcc, skill_id, keyspace)
	skill_id = utils.AppendVcc( skill_id, vcc )
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "AccountsBySkill",
		key = skill_id
	}
end
