module(..., package.seeall)

local utils = require "shared.apps.utils"
local utf8 = require "icu.utf8"
local model_skills = require "apps.administration.models.skills"
local model_campaigns = require "apps.administration.models.campaigns"
local model_perm = require "shared.permissions.models.permissions"
local cjson = require "cjson"

local m_datasource

function GetName ()
	return "Users"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function GetDataSource ()
	if not m_datasource then
		local model = require "apps.administration.model"
		model.BuildModels(true)
	end
	return m_datasource
end

function GetVCCUsers (vcc, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCUsers",
		key = vcc,
		columnCount = 50000
	} or {}
end

function IsEnabled ( full_user_id, keyspace )
	local isEnabled = utils.GetColumnValue({
		keyspace = keyspace or "Agent",
		columnFamily = "Users",
		key = full_user_id,
		column = "Status"
	})

	return ( isEnabled == nil or tonumber(isEnabled) == 1 )
end

function GetUsersByRole ( fullRoleId, keyspace )
	local vcc = inconcert.authUser.virtualContactCenter
	local users = {}
	local count = 0
	local simpleRoleId = utils.GetId( fullRoleId )

	-- El System Role Id tiene que ser buscado en minúsculas ya que el Alta del mismo se realiza en minúsuclas.
	local systemGroupId = utf8.lower( utils.AppendVcc( simpleRoleId, "system" ) )
	local usersByGroup = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UsersByGroup",
		key = systemGroupId,
		columnCount = 50000
	} or {}

	-- agrego los usuarios del role system, pero solo los del VCC
	for user_id, value in pairs(usersByGroup) do
		local pos = user_id:find("@")
		if pos then
			local vcc_user = user_id:sub( pos + 1 )
			if vcc_user and vcc_user:lower() == vcc:lower() then
				users[ user_id ] = value
				count = count + 1
			end
		end
	end

	-- El Custom Role Id tiene que ser buscado sin convertir a minúsuclas, ya que cuando el mismo se da Alta como el Usuario haya ingresado.
	local customGroupId = utils.AppendVcc( simpleRoleId, vcc )
	usersByGroup = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UsersByGroup",
		key = customGroupId,
		columnCount = 50000
	} or {}

	-- agrego los usuarios del role custom, o del role system que se ha modificado en el current vcc
	for user_id, value in pairs(usersByGroup) do
		if not users[ user_id ] then
			users[ user_id ] = value
			count = count + 1
		end
	end

	return users, count
end

function GetUser (vcc, simpleId, keyspace, columns, retrieveRelations)
	simpleId = simpleId:lower()
	local user       = utils.GetAllColumns({
		keyspace     = keyspace or "Agent",
		columnFamily = "Users",
		key          = utils.AppendVcc(simpleId, vcc),
		columns      = columns
	}) or {}
	user.Avatar = nil
	if retrieveRelations then
		user.skills = model_skills.GetAgentSkills(vcc, simpleId, keyspace) or utils.EMPTY_OBJECT
		user.campaigns = model_campaigns.GetUserCampaigns(vcc, simpleId, keyspace) or utils.EMPTY_OBJECT
	end
	return user
end

function SaveUser (vcc, user, keyspace)
	utils.SaveRows {
		keyspace =  keyspace or "Agent",
		columnFamily = "Users",
		rows = {
			[utils.AppendVcc(user.Id, vcc)] = user
		}
	}

	local fullUser = GetUser(vcc, user.Id, keyspace)
	local closeKeyspace = false
	if not keyspace then
		keyspace = utils.GetDatasource():CassandraConnection("Agent")
		closeKeyspace = true
	end
	keyspace:InsertColumn(vcc, "VCCUsers", user.Id, cjson.encode(fullUser))
	if closeKeyspace then
		keyspace:Close()
	end
end

function MailOwner (mail, keyspace)
	local owners = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UsersByEmail",
		key = mail
	}
	for owner in pairs(owners or {}) do
		return owner
	end
	return nil
end

function GetAddresses (vcc, userId, keyspace)
	local full_user_id = utils.AppendVcc(utils.GetId(userId), vcc)
	local addresses = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "AgentAddresses",
		superColumns = true,
		key = full_user_id
	}
	return addresses
end

function GetMailFromAddresses (vcc, userId, keyspace)
	local addresses = GetAddresses(vcc, userId, keyspace)
	return addresses and addresses.MAIL and addresses.MAIL.address
end

function EnableUser (vcc, simple_user_id, enabled, mail, peerData, keyspace, allegroOtherInfo)
	local close_keyspace = false
	if not keyspace then
		keyspace = utils.GetDatasource():CassandraConnection("Agent")
		close_keyspace = true
	end

	local agentUserModel = require "apps.agent.models.users"
	local isEnabled = enabled and "1" or "0"
	local hasHelpDeskIntegration = utils.HasHelpDeskIntegration(vcc, keyspace)
	local hasAllegroIntegration = utils.HasAllegroIntegration(vcc, keyspace)
	local userToSaveInIntegration
	local success, helpDeskInfo

	if hasHelpDeskIntegration then
		-- Si tiene integracion obtiene toda la informacion del usuario necesaria en HelpDesk
		--   relacion con usercampaigns, skills, etc
		local retrieveRelations = true
		userToSaveInIntegration = userToSaveInIntegration or GetUser(vcc, simple_user_id, keyspace, nil, retrieveRelations)

		if userToSaveInIntegration then
		-- Si el usuario existe en cassandra, primero actualizo el estado y el mail y luego actualizo toda la informacion en HelpDesk.
		--   En si en un deactivate/activate en HelpDesk se alamacena todo el usuario.

			userToSaveInIntegration.Status = isEnabled
			userToSaveInIntegration.Mail = mail or userToSaveInIntegration.Mail

			-- Guarda en HelpDesk el usuario
			-- Para este tipo de integración, NO realizamos Rollback en el caso de que falle la actualización en HelpDesk.
			-- Esto se realiza así, para que en i6 igualmente se almacene el Usuario, y que i6 NO dependa de este tipo de integraciones.
			--  Entonces si por algún motivo no se ha almacenado, solamente logueamos la información.
			success, helpDeskInfo = agentUserModel.UpdateHelpDeskUser(userToSaveInIntegration, vcc, keyspace)
			if not success then
				-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "HELPDESK_INTEGRATION_ERROR",
				--  ya que otro tipo de error, puede ser que no tenga integracion con HelpDesk y por lo tanto en si, no es un error y debe de continuar
				--  la actualizacion en i6.
				if helpDeskInfo and helpDeskInfo.type == "HELPDESK_INTEGRATION_ERROR" then
					if close_keyspace then keyspace:Close() end
					return false, {result = helpDeskInfo.type, message_error = helpDeskInfo.sub_type}
				end
			end
		end
	end

	local allegroInfo
	if hasAllegroIntegration then
		-- Si tiene integracion obtiene toda la informacion del usuario necesaria en 5.5
		--   relacion con usercampaigns, skills, etc
		local retrieveRelations = true
		userToSaveInIntegration = userToSaveInIntegration or GetUser(vcc, simple_user_id, keyspace, nil, retrieveRelations)

		if userToSaveInIntegration then
		-- Si el usuario existe en cassandra, primero actualizo el estado y el mail y luego actualizo toda la informacion en 5.5.
		--   En si en un deactivate/activate en 5.5 se alamacena todo el usuario.

			userToSaveInIntegration.Enabled = isEnabled
			userToSaveInIntegration.Mail = mail or userToSaveInIntegration.Mail

			-- Guarda en Allegro el usuario
			-- Debo mandar a actualizar aca, en este lugar, antes de guardar algo en cassandra,
			--   asi me evito el Rollback y toda esas cosas!!!
			--
			success, allegroInfo = agentUserModel.UpdateAllegroUserIn55(userToSaveInIntegration, vcc, keyspace, allegroOtherInfo)
			if not success then
				-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
				--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
				--  la actualizacion en i6.
				if allegroInfo and allegroInfo.type == "ALLEGRO_INTEGRATION_ERROR" then
					if close_keyspace then keyspace:Close() end
					return false, {result = allegroInfo.type}
				end
			end
		end
	end

	local full_user_id = utils.AppendVcc(simple_user_id, vcc)

	utils.SaveRows({
		keyspace = keyspace,
		columnFamily = "Users",
		rows = {
			[full_user_id] = {
				Status = isEnabled,
				StatusReason = "ADMIN",
				StatusTimestamp = tostring(os.time()),
				Mail = mail,
				LoginFailedAttemps = "0"
			}
		}
	})

	local vccUsers = GetVCCUsers(vcc, keyspace)
	local vccUser = vccUsers[simple_user_id] or {}
	vccUser.Status = isEnabled
	if (not vccUser.StatusReason or vccUser.StatusReason == "") and isEnabled == "0" then
		vccUser.StatusReason = "ADMIN"
	elseif isEnabled == "1" then
		vccUser.StatusReason = ""
	end
	
	keyspace:InsertColumn(vcc, "VCCUsers", simple_user_id, cjson.encode(vccUser))

	if enabled then
		if peerData then
			utils.SaveRows {
				keyspace = keyspace,
				columnFamily = "AgentAddresses",
				superColumns = true,
				rows = {
					[full_user_id] = {
						CALL = peerData
					}
				}
			}
			utils.SaveRows {
				keyspace = keyspace,
				columnFamily = "UsersByPeer",
				rows = {
					[peerData.id] = {
						[full_user_id] = ""
					}
				}
			}
		end
		if mail then
			utils.SaveRows {
				keyspace = keyspace,
				columnFamily = "AgentAddresses",
				superColumns = true,
				rows = {
					[full_user_id] = {
						MAIL = {
							address = mail,
							address_id = mail,
							resource_address = ""
						}
					}
				}
			}
			utils.SaveRows {
				keyspace = keyspace,
				columnFamily = "UsersByEmail",
				rows = {
					[mail] = {
						[full_user_id] = ""
					}
				}
			}
		end
	else
		-- Si tiene integracion, al usuario ya lo obtuvo de cassandra "userToSaveInIntegration"
		local user = userToSaveInIntegration or GetUser(vcc, simple_user_id, keyspace, { "Mail" })
		local email
		if user and user.Mail and user.Mail ~= "" then
			email = user.Mail
		else
			email = GetMailFromAddresses(vcc, simple_user_id, keyspace)
		end
		if email then
			LogInfo("Removing email %s from user %s", email, simple_user_id)
			utils.RemoveColumns {
				keyspace = keyspace,
				columnFamily = "UsersByEmail",
				rows = {email}
			}
		end
		--find addresses to remove peer
		local addresses = GetAddresses(vcc, simple_user_id, keyspace)
		if addresses and addresses.CALL then
			local peer = addresses.CALL.id
			LogInfo("Removing peer %s from user %s", peer, simple_user_id)
			--remove from Users
			utils.RemoveColumns {
				keyspace = keyspace,
				columnFamily = "UsersByPeer",
				rows = {peer}
			}
			utils.RemoveColumns {
				keyspace = keyspace,
				columnFamily = "AgentAddresses",
				superColumns = true,
				rows = {
					[full_user_id] = {"CALL"}
				}
			}
		end
	end

	if close_keyspace then
		keyspace:Close()
	end

	local updateInfo = {
		StatusReason = vccUser.StatusReason,
		PasswordChangeTimestamp = vccUser.PasswordChangeTimestamp,
		integrations = {
			helpdesk = helpDeskInfo,
			allegro  = allegroInfo
		}
	}
	return true, updateInfo
end

local function FilterAuditFields (user)
	--[[
		key = true|false| { alias = "key_alias"}
	
	]]--
	local auditFields = {
		Id = true,
		FirstSurname = true,
		LastSurname = true,
		Name = true,
		LastName = true,
		Mail = true,
		UserGroup = true,
		Password = true,
		WorkgroupId = true,
		countryCode = true,
		areaCode = true,
		initialState = true,
		TimeZoneId = {
			alias = "TimeZone"
		},
		CompanyInformation = true,
		CustomInformation = true,
		DefaultCampaign = true,
		LanguageId = true,
		campaigns = true,
		Status = true,
		overridePermissions = true,
		skills = true,
		PasswordHistory = false,
		LoginFailedAttemps = false,
		ValidationOrigin = false,
		MustChangePassword = false,
		StatusTimestamp = false,
		PasswordChangeTimestamp = false,
		IsSystemUser = false,
		FreeSecondsAfterInteraction = false,
		creationDate = false,
		HasRoleOverride = false,
		StatusReason = false,
		Enabled = false,
		Clients = false,
		VccGroup = false,
		VCC = false,
		DateFormat = true
	}

	local result = {}
	for key, value in pairs(user) do
		local auditField = auditFields[key]
		if auditField and type(auditField) == "table" then
			result[auditField.alias or key] = value
		elseif auditField then
			result[key] = value
		end
	end

	return result
end

function GetUserAuditData (fullUserId, keyspace, vcc, usersOverridePermissions, userColumns)
	assert(type(fullUserId) == "string")
	assert(type(keyspace) == "userdata")
	assert(type(usersOverridePermissions) == "table")
	local vcc = vcc or inconcert.authUser.virtualContactCenter
	local simpleUserId = utils.GetId(fullUserId)
	local user = GetUser(vcc, simpleUserId, keyspace, userColumns, true)
	if not user then
		return nil
	end
	user.overridePermissions = usersOverridePermissions[fullUserId] and usersOverridePermissions[fullUserId].actions or nil
	local camp = {}
	for key, value in pairs(user.campaigns or {}) do
		camp[key] = {
			id = utils.GetId(key)
		}
	end
	user.campaigns = not utils.TableIsEmpty(camp) and camp or nil
	user.skills = not utils.TableIsEmpty(user.skills) and user.skills or nil
	--LogDebug("[ADMINISTRATION.USERS:GetUserAuditData]  1- Loaded User '%s' ", json.encode(user))
	user = FilterAuditFields(user)
	if user and not utils.TableIsEmpty(user) then
		-- ok
		--LogDebug("[ADMINISTRATION.USERS:GetUserAuditData]  2- Filtered User '%s' ", json.encode(user))
		return user
	end
	-- err
	--LogDebug("[ADMINISTRATION.USERS:GetUserAuditData]  3- EMTPY Filtered User")
	return nil
end

function GetUserValue (vcc, userId, retrieveColumn, keyspace)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local simpleUserId = utils.GetId(userId)
	local fullUserId   = utils.AppendVcc(simpleUserId, vcc)
	return utils.GetColumnValue({
									keyspace     = keyspace or "Agent",
									columnFamily = "Users",
									key          = fullUserId,
									column       = retrieveColumn
								})
end

function SetUserDefaultCampaign (vcc, userId, campaignId, keyspace)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local simpleUserId     = utils.GetId(userId)
	local fullUserId       = utils.AppendVcc(simpleUserId, vcc)
	local simpleCampaignId = utils.GetId(campaignId)
	local fullCampaignId   = utils.AppendVcc(simpleCampaignId, vcc)

	if campaignId then -- Set DefaultCampaign
		utils.SaveRows {
			keyspace = keyspace or "Agent",
			columnFamily = "Users",
			rows = {
				[fullUserId] = {DefaultCampaign = fullCampaignId}
			}
		}
	else               -- Remove DefaultCampaign
		utils.RemoveColumns {
			keyspace = keyspace or "Agent",
			columnFamily = "Users",
			rows = {
				[fullUserId] = {"DefaultCampaign"}
			}
		}
	end
end

--[[------------------------------------------------------------------------------------------
	SUPER USERS
]]--
function GetVCCSuperUsers (vcc, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCSuperUsers",
		key = vcc,
		columnCount = 50000
	} or {}
end

function GetSuperUser (vcc, simpleId, keyspace, columns)
	simpleId = simpleId:lower()
	local superUser  = utils.GetAllColumns({
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCSuperUsers",
		key          = utils.AppendVcc(simpleId, vcc),
		columns      = columns
	})
	return user
end

function SaveSuperUser (vcc, user, keyspace)
	assert(vcc)
	assert(user)
	local closeKeyspace = false
	if not keyspace then
		keyspace = utils.GetDatasource():CassandraConnection("Agent")
		closeKeyspace = true
	end
	keyspace:InsertColumn(vcc, "VCCSuperUsers", user.Id, cjson.encode({}))
	if closeKeyspace then
		keyspace:Close()
		keyspace = nil
	end
end
