local notificationsService = require "lua_notificationsserviceclient"
local agentservice = require "apps.agent.agentservice"
local utils = require "shared.apps.utils"
local cache = require "shared.cache.cache"
local utf8 = require "icu.utf8"
local m_models = require("apps.administration.model"):GetModels()

module(..., package.seeall, inconcert.orbit.application.Decorate)

local m_client

local function GetModel (entityName)
	assert(type(entityName) == "string")
	if not m_models or not m_models[entityName] then
	-- Si el modelo no se ha creado aun, debo crearlo.
	--  Esto sucede en algunas ocasiones si es que me han invocado desde otra app Ej: "Allegro" y aun no se ha creado el modelo.
	--  Entonces tengo que crear el modelo.
		local model = require "apps.administration.model"
		model.CreateModelTable()
		m_models = model:GetModels()
	end
	local entityModel = m_models[entityName]
	assert(type(entityModel) == "table")
	return entityModel
end

local function GetClient()
	if not m_client then
		LogInfo("[agentNotfications.GetClient] Creating Notifications Service. Cluster: %s", inconcert.environment.apps.agent.cluster)
		m_client = notificationsService.LuaNotificationsServiceClient(inconcert.environment.apps.agent.cluster, "webhandler")
		for _, address in pairs(inconcert.environment.apps.agent.notificationsservice) do
			LogInfo("[agentNotfications.GetClient] Add Known Notifications Service. Cluster. IP: '%s', Port: '%s'.", 
				address.ip, address.port)
			m_client:AddKnownNotificationsService(address.ip, address.port)
		end
	end
	return m_client
end

local m_scopeNotifications = cache.New(function(scope)
	LogDebug("[agentNotfications.FetchNotifications] scope %s", scope)
	local notifications = GetModel("AgentNotifications").GetAgentScopeNotifications(scope)
	return notifications or {}
end, 30)

local function GetEventConfiguration(scope, eventType)
	local scopeNotifications = m_scopeNotifications.get(scope)
	return scopeNotifications[eventType]
end

local function HasToNotifyAgent(vcc, agentId)
	if not agentservice.IsAgentLoggedIn(agentId, vcc) then
		return false
	end
	return true
end

local function GetAgentsToNotify(vcc, fixedAgents, findLevel, message, keyspace, excludeAgents)
	--find agents
	local foundAgents = {}
	excludeAgents = excludeAgents or {}
	
	if findLevel == "VCC_USERS" then
		foundAgents = GetModel("Users").GetVCCUsers(vcc, keyspace)
	elseif findLevel == "CAMPAIGN_USERS" then
		foundAgents = GetModel("Campaigns").GetCampaignMembers(vcc, message.campaignId, keyspace)
	elseif findLevel == "ROLE_USERS" then
		local full_role_id = message.role_id .. "@" .. ( message.is_system_group and "system" or vcc )
		foundAgents, count = GetModel("Users").GetUsersByRole(full_role_id, keyspace)
	end
	--add available agents
	local availableAgents = {}
	local availableAgents_d = {}
	for _, agentAtVcc in pairs(fixedAgents) do
		local agentId = utils.GetId(agentAtVcc)
		if HasToNotifyAgent(vcc, agentId) and not excludeAgents[agentAtVcc] then
			table.insert(availableAgents, utils.AppendVcc(agentId, vcc))
			availableAgents_d[agentId] = true
		end
	end
	for agentAtVcc in pairs(foundAgents) do
		local agentId = utils.GetId(agentAtVcc)
		if not availableAgents_d[agentId] and HasToNotifyAgent(vcc, agentId) and not excludeAgents[agentAtVcc] then
			table.insert(availableAgents, utils.AppendVcc(agentId, vcc))
		end
	end
	return availableAgents
end

local SCOPE_CHANNELS = {
	agent = "AGENT",
	administration = "AGENT"
}

local function MakePollerMessage(scope, eventType, reference, message)
	message = message or {}
	message.reference = message.reference or reference
	return json.encode({
		channel = SCOPE_CHANNELS[scope] or scope,
		type = eventType,
		message = json.encode(message)
	})
end

local function SendPollerMessage(scope, eventType, reference, level, agents, pollerMessage)
	local result = GetModel("AgentNotifications").QueueAgentsMessage(scope, eventType, reference, level, agents, pollerMessage)
	local pollerCount = 0
	for agentId, agentResult in pairs(result.notifiedAgents) do
		LogInfo("[agentNotfications.SendPollerMessage] Scope %s event %s agent %s notified with messageId %s", 
			scope, eventType, agentId, agentResult.messageId)
		pollerCount = pollerCount + 1
	end
	return result.serverAgents, pollerCount
end

local function MakeServerMessage(scope, eventType, reference, message, pollerMessage)
	message.scope = scope
	message.type = eventType
	message.reference = reference
	message.campaignId = message.campaignId or pollerMessage.campaignId or nil
	message.campaignId = message.campaignId and utils.GetId(message.campaignId) or nil
	message.role_id = message.role_id or pollerMessage.role_id or nil
	message.role_id = message.role_id and utils.GetId(message.role_id) or nil
	local account = pollerMessage.account or message.account
	message.account = nil
	message.accountName = message.accountName or (account and account.name) or nil
	message.channel = message.channel or (account and account.proxyType) or nil
	return message
end

local function SendServerMessage(vcc, serverAgents, serverMessage)
	local request = json.encode({
		vcc = vcc,
		agents = serverAgents,
		notification = { 
			message = serverMessage
		}
	})
	local status, err = GetClient():SendAgentsNotification(request)
	if not status then
		LogError("[agentNotfications.SendServerMessage] error sending notification - %s", err)
	end
end

function NotifyAgents(request, keyspace)
	local vcc = request.vcc or inconcert.authUser.virtualContactCenter
	local scope = request.scope or "administration"
	local eventType = request.eventType
	local reference = request.reference or nil
	local pollerEventType = request.pollerEventType or eventType
	local result = { pollerCount = 0 }
	--find agents
	local agents = GetAgentsToNotify(vcc, request.agents or {}, request.agentsLevel, request.pollerMessage, keyspace, request.excludeAgents)
	if #agents == 0 then
		return result
	end
	--get configuration
	local configuration = GetEventConfiguration(scope, eventType)
	if not configuration then
		LogWarning("[agentNotfications.NotifyAgents] scope %s event %s not found", scope, eventType)
		return result
	end
	local level = configuration.level or "GLOBAL"
	--send message to poller
	local pollerMessage = MakePollerMessage(scope, pollerEventType, reference, request.pollerMessage)
	local serverAgents, pollerCount = SendPollerMessage(scope, eventType, reference, level, agents, pollerMessage)
	result.pollerCount = pollerCount
	if utils.TableIsEmpty(serverAgents) then
		return result
	end
	--send message to server
	local serverMessage = MakeServerMessage(scope, eventType, reference, request.serverMessage or {}, request.pollerMessage or {})
	SendServerMessage(vcc, serverAgents, serverMessage)
	return result
end

function CleanNotifications(userName, vcc)
	userName = userName or inconcert.authUser.userName
	local vcc = vcc or inconcert.authUser.virtualContactCenter
	GetModel("AgentNotifications").CleanNotifications(userName, vcc)
end
