local gravatar = require "apps.agent.gravatar"
require "base64"
local models = require("apps.agent.model"):GetModels()
local ir = require "lua_iragentlink"
local m_moduleLocation = select(2, ...)

module(..., package.seeall, inconcert.orbit.application.Decorate)

local function OpenFile(fileName)
	local fullName = m_moduleLocation .. "\\images\\" .. fileName
	fullName = webHandler.path.normalize(fullName)
	local file, err = io.open(fullName, "rb")
	if not file then 
		LogWarning("Generic avatar does not exist. Path: %s", fullName)
		return nil, err
	end
	local blob = file:read("*a")
	file:close()
	return {
		image = blob,
		size  = #blob
	}
end

local m_cachedFileGenericAvatar
function GetGenericAvatar()
	if m_cachedFileGenericAvatar then
		return m_cachedFileGenericAvatar
	end
	m_cachedFileGenericAvatar = OpenFile("genericAvatar.png")
	return m_cachedFileGenericAvatar
end

local m_cachedFileAgentGenericAvatar
function GetAgentGenericAvatar()
	if m_cachedFileAgentGenericAvatar then
		return m_cachedFileAgentGenericAvatar
	end
	m_cachedFileAgentGenericAvatar = OpenFile("genericProfileAvatar.png")
	return m_cachedFileAgentGenericAvatar
end

local m_cachedFileCampaignGenericAvatar
function GetCampaignGenericAvatar()
	if m_cachedFileCampaignGenericAvatar then
		return m_cachedFileCampaignGenericAvatar
	end
	m_cachedFileCampaignGenericAvatar = OpenFile("campaignGenericAvatar.png")
	return m_cachedFileCampaignGenericAvatar
end

function FindBestCampaignAvatar(campaign)
	local facebookAccounts = {}
	local twitterAccounts = {}
	local emailAccounts = {}
	local imAccounts = {}

	--prefilter
	for k, account in ipairs(campaign.accounts) do
		if (account.proxyType == "FACEBOOK") then
			table.insert(facebookAccounts, account)
		elseif (account.proxyType == "TWITTER") then
			table.insert(twitterAccounts, account)
		elseif (account.proxyType == "MAIL") then
			table.insert(emailAccounts, account)
		elseif (account.proxyType == "IM") then
			table.insert(imAccounts, account)
		end
	end

	--search facebook avatar
	for i, account in ipairs(facebookAccounts) do
		local imagesrc = string.format("http://graph.facebook.com/%s/picture/", account.id)
		LogDebug("Searching for facebook avatar on url %s", imagesrc)
		local result, image, imageSize = gravatar.GetRequest(imagesrc)
		if (result) then
			return { image = image, size = imageSize}
		end
	end
	
	--search twitter avatar
	for i, account in ipairs(twitterAccounts) do
		local avatar = GetTwitterAvatar(account.id, campaign.Id, account.id)
		if (avatar) then
			return avatar
		end
	end
	
	--search email gravatar
	if (inconcert.environment.apps.agent.disableGravatar ~= true) then
		for i, account in ipairs(emailAccounts) do
			local avatar = GetMailAvatar(account.id)
			if (avatar) then
				return avatar
			end
		end
	end
	
end

--[[
	iterate over different contact address by avatar priorty, using this order:
	1) facebook
	2) IM
	3) gravatar/email
--]]
function FindBestAvatar (contact)
	--LogDebug("[AVATAR:FindBestAvatar] Searching Best Avatar for '%s'.", contact.Id)

	local facebookAddresses = {}
	local twitterAddresses = {}
	local emailAddresses = {}
	local imAddresses = {}

	--prefilter
	for addressType, address in pairs(contact.addresses) do
		if (address.Type == "Facebook") then
			table.insert(facebookAddresses, address)
		elseif (address.Type == "Twitter") then
			table.insert(twitterAddresses, address)
		elseif (address.Type == "Mail" or address.Type == "MAIL") then
			table.insert(emailAddresses, address)
		elseif (address.Type == "IM") then
			table.insert(imAddresses, address)
		end
	end

	--search facebook avatar
	--LogDebug("[AVATAR:FindBestAvatar] Searching Best FACEBOOK Avatar for '%s'.", contact.Id)
	for i, address in ipairs(facebookAddresses) do
		local imagesrc = string.format("http://graph.facebook.com/%s/picture/", address.UID)
		--LogDebug("[AVATAR:FindBestAvatar] Searching for FACEBOOK avatar on url %s", imagesrc)
		local result, image, imageSize = gravatar.GetRequest(imagesrc)
		if result then
			return { image = image, size = imageSize}
		end
	end

	--search twitter avatar
	--LogDebug("[AVATAR:FindBestAvatar] Searching Best TWITTER Avatar for '%s'.", contact.Id)
	for i, address in ipairs(twitterAddresses) do
		local imagesrc = address.profile_image_url
		if imagesrc then
			--LogDebug("[AVATAR:FindBestAvatar] Searching for TWITTER avatar on url %s", imagesrc)
			local result, image, imageSize = gravatar.GetRequest(imagesrc)
			if result then
				return { image = image, size = imageSize}
			end
		end
	end

	--search IM avatar
	--LogDebug("[AVATAR:FindBestAvatar] Searching Best IM Avatar for '%s'.", contact.Id)
	for i, address in ipairs(imAddresses) do
		--LogDebug("[AVATAR:FindBestAvatar] Searching for IM avatar on buddy %s, protocol %s", address.buddy, address.protocol)
		local imAvatar = GetIMAvatar(address.buddy, address.protocol)
		if imAvatar then
			return imAvatar
		end
	end

	--search email gravatar
	if inconcert.environment.apps.agent.disableGravatar ~= true then
		--LogDebug("[AVATAR:FindBestAvatar] Searching Best MAIL Avatar for '%s'.", contact.Id)
		for i, address in ipairs(emailAddresses) do
			--LogDebug("[AVATAR:FindBestAvatar] Searching for MAIL avatar on address %s", address.address)
			local avatar = GetMailAvatar(address.address)
			if avatar then
				return avatar
			end
		end
	end
end

function GetIMAvatar(buddy, protocol)
	local buddyStatus = models["Interactions"].IMGetBuddyStatus(buddy, protocol)
	if (buddyStatus and buddyStatus.icon and string.len(buddyStatus.icon) > 0) then
		local image = base64.decode(buddyStatus.icon)
		return { image = image, size = #image}
	end
end

function GetMailAvatar(address)
	return gravatar.GetAvatar(address)
end

function GetTwitterAvatar(userId, campaignAtVCC, account)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:TwitterGetUser(userId, campaignAtVCC, account)
	if result["result"] ~= "OK" then
		LogError("Error retrieving twitter user info for user %s on account %s, avatar won't be loaded, reason: %s", userId, account, result["reason"])
		return
	end
	local user = result["user_info"]
	if not user then
		return
	end
	local imagesrc = user.profile_image_url
	local result, image, imageSize = gravatar.GetRequest(imagesrc)
	if result then
		return {image = image, size = imageSize}
	end
end
