local oo = require "loop.base" 
require "inconcert.translators"
local agentservice = require "apps.agent.agentservice"
local avatar = require "apps.agent.avatar"
local utf8 = require "icu.utf8"
local mail = require "apps.agent.mail"
local permissions = require "shared.permissions.permissions"
local model_perm = require "shared.permissions.models.permissions"
local users = require "apps.agent.controllers.users"
local helpers = require "apps.agent.helpers"
local utils = require "shared.apps.utils"
local models = require("apps.agent.model"):GetModels()
local agentModel = require "apps.agent.model"
local m_adm_models = require("apps.administration.model"):GetModels()
local http_mapping = require "apps.agent.result_codes_to_http_codes"
local ir = require "lua_iragentlink"
local JsonValidator = require "nozzle.stock".json_validator()
local notificationsEngine = require("apps.agent.notifications")
local m_helpers = require "helpers"
local m_moduleLocation = select(2, ...)

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil

local function GetModel (entityName)
	assert(type(entityName) == "string")
	return utils.GetAgentModel(entityName)
end

local function ReplyJSON (web, data)
	assert(type(data) ~= "string", "The parameter can not be string")
	web.headers["Content-Type"] = "application/json"
	return json.encode(data)
end

local function WhatsappSetCloudApiPublicRepositoryFileUrl (campaignIdAtVCC, accountId, fileId, usedById)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:WhatsappSetCloudApiPublicRepositoryFileUrl({
		campaignAtVCC = campaignIdAtVCC,
		account = accountId,
		fileId = fileId,
		usedById = usedById
	})

	if res.result ~= "OK" then
		LogError("Error setting public url for '%s' with repository file '%s' on whatsapp account '%s' and campaign '%s' -> '%s'", usedById, fileId, accountId, campaignIdAtVCC, res.reason)
	end
end

local function WhatsappDelCloudApiPublicRepositoryFileUrl (campaignIdAtVCC, accountId, fileId, usedById)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:WhatsappDelCloudApiPublicRepositoryFileUrl({
		campaignAtVCC = campaignIdAtVCC,
		account = accountId,
		fileId = fileId,
		usedById = usedById
	})

	if res.result ~= "OK" then
		LogError("Error deleting public url for '%s' with repository file '%s' on whatsapp account '%s' and campaign '%s' -> '%s'", usedById, fileId, accountId, campaignIdAtVCC, res.reason)
	end
end

--[[
	funcion que valida que la cuenta de Whatsapp que se esta configurando, coincida con el mismo id de cuenta de whatsapp que tiene el dibujo (flow)
]]--
local function ValidateOneWhatsappAccountInFlow (flowId, accountsByFlows, currentAccountId)
	LogDebug("[AGENT.CAMPAIGNS:ValidateOneWhatsappAccountInFlow]  (ENTRA)--> - currentAccountId: '%s', flowId '%s'.", currentAccountId, flowId)
	for currentFlowId, state in pairs(accountsByFlows or {}) do
		-- LogDebug("--> currentFlowId: %s", currentFlowId)
		for stateId, stateInfo in pairs(state or {}) do
			-- LogDebug("  --> stateId: %s", stateId)
			-- LogDebug("   --> stateInfo: %s", json.encode(stateInfo))
			for activityId, activityInfo in pairs(stateInfo or {}) do
				-- LogDebug("      --> activityId: %s", activityId)
				-- LogDebug("        --> activityInfo: %s", json.encode(activityInfo))
				for _, account in ipairs(activityInfo.accounts or {}) do
					-- LogDebug("      --> %s - account: %s", _, json.encode(account))
					if account.proxyType == "WHATSAPP" then
						if account.accountId ~= currentAccountId then
							return false, {accountId = account.accountId, campaignId = utils.GetId(account.campaignId), accountName = account.accountName}
						end
					end
				end
			end
		end
	end
	-- ok
	LogDebug("[AGENT.CAMPAIGNS:ValidateOneWhatsappAccountInFlow]  (SALE)--> - OK - currentAccountId: '%s', flowId '%s'.", currentAccountId, flowId)
	return true
end

--[[
	Funcion encargada de llamar a las funciones que actualizan la key de redis REPOSITORY_FILE_PUBLIC_URL_USED_BY, la cual determina que shared files con URL publica asociada se estan usando y desde donde

	flowConfig es una tabla que contiene 
		{
			"<flowId>": {
				"<flowState>": {
					"activityId": {
						"RepositoryIds": ["<RepositoryId1>", "<RepositoryId2>", ...],
						"type": "send_and_wait_chat_message",
						"description": "Enviar mensaje y esperar",
						"sharedFiles": ["<sharedFileId1>", "<sharedFileId2>", ...]
					}
				}
			}
		}	
]]--
local function WhatsappUpdateCloudApiPublicRepositoryFileUrlUsedByFlow(campaignIdAtVCC, accountId, sharedFiles, isOld)
	for flowId, states in pairs(sharedFiles) do
		for stateId, activities in pairs(states) do
			for activityId, activity in pairs(activities) do
				for _, repositoryId in pairs(activity.RepositoryIds or {}) do
					local usedById = flowId .. "_" .. stateId .. "_" .. activityId
					if isOld then
						WhatsappDelCloudApiPublicRepositoryFileUrl(campaignIdAtVCC, accountId, repositoryId, usedById)
					else
						WhatsappSetCloudApiPublicRepositoryFileUrl(campaignIdAtVCC, accountId, repositoryId, usedById)
					end
				end
			end
		end
	end
end

--[[
	Funcion encargada de invocar a una funcion "ExternalProcessAllActivities" del controller de process flow para obtener las cuentas de whatsapp que se encuentran configuradas
	en las diferentes actividades del flow.

	Luego invoca a una funcion para validar que los flows "entrantes" y "fuera de hora", tengan configurada la misma cuenta que la cuenta que se esta configurando en la campana->cuenta
]]--
local function ValidateWhatsappAccountInFlows (account, keyspace, isOld)
	if not account then
		return true
	end
	if not account.FlowInformation then
		return true
	end
	LogInfo("[AGENT.CAMPAIGNS:ValidateWhatsappAccountInFlows] Validating Process Flows in account.")
	local vcc = inconcert.authUser.virtualContactCenter
	--Se obtiene controller del flow_designer. Tambien el modelo, por si no se han contruidos. 
	local flowsController = utils.GetControllerOthersApps("flow_designer", {controllerName = "process_flows", modelName = "ProcessFlows"})
	local function StartValidation (config, errorCode, isOld)
		if not config or not config.Process then
			-- ok
			return true
		end
		local flowId = config.Process
		-- ejecuto la funcion del controller del process flow para poder obtener la lista de cuentas que están configuradas en las diferentes actividades del flow.
		local success, flowResponse, flow = flowsController.ExternalProcessAllActivities(flowId, vcc, keyspace)
		if not success then
			-- error
			local err = flowResponse
			return false, err
		end

		local acc, campaignIdAtVCC = GetModel("Campaigns").FindAccountById(account.id, keyspace)
		WhatsappUpdateCloudApiPublicRepositoryFileUrlUsedByFlow(campaignIdAtVCC, account.id, flowResponse.sharedFiles, isOld)

		-- Cuando se llama a esta function con la configuracion previa de la account (lo que tiene antes de guardar) no valido contra las cuentas de Whatsapp
		if not isOld then
			-- ahora con estas cuentas, realizo las validaciones contra las cuentas de WhatsApp
			local success, resp = ValidateOneWhatsappAccountInFlow(flowId, flowResponse.accountsByFlows, account.id)
			if not success then
				-- error
				return false, {
					result      = errorCode,
					reason      = "The account configured in the process flow is different from the current account",
					campaignId  = resp.campaignId,
					accountId   = resp.accountId,
					accountName = resp.accountName
				}
			end
		end
		-- ok
		return true
	end

	-- valido los flow process entrantes de la nueva config de la account
	local success, err = StartValidation(account.FlowInformation.on_schedule, "FLOW_INBOUND_CONFLICT_ACCOUNT_RELATION", isOld)
	if not success then
		return false, err
	end
	-- valido los flow process fuera de hora de la nueva config de la account
	local success, err = StartValidation(account.FlowInformation.out_of_schedule, "FLOW_OUT_OF_HOURS_CONFLICT_ACCOUNT_RELATION", isOld)
	if not success then
		return false, err
	end

	-- ok
	return true
end

local function GetUserCampaigns (web)
	web.headers["Content-Type"] = "application/json"

	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName

	LogInfo("User %s on vcc %s retrieving his campaigns", name, vcc)

	local userCampaigns = GetModel("Campaigns").GetUserCampaigns(name, vcc)
	return json.encode(userCampaigns)
end

local outgoingChannels = {
	CALL = function(account)
		if not account.outgoingConfig then
			return false
		end
		if type(account.outgoingConfig) == "string" then
			account.outgoingConfig = json.decode(account.outgoingConfig)
		end
		return account.outgoingConfig.enabled and account.outgoingConfig.ContactCenter and 
			tonumber(account.outgoingConfig.ContactCenter.AllowManualCalls) == 1 or false
	end,
	MAIL = true,
	SMS = true,
	WHATSAPP = true,
	RCS = true
}
local function IsOutgoingAccount(account)
	if not outgoingChannels[account.proxyType] then
		return false
	end
	if type(outgoingChannels[account.proxyType]) == "function" then
		return outgoingChannels[account.proxyType](account)
	end
	return true
end

local function GetUserOutgoingCampaigns (web, channel)
	web.headers["Content-Type"] = "application/json"

	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName

	LogInfo("User %s on vcc %s retrieving his outgoing campaigns", name, vcc)

	local outgoingCampaigns = {}
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local userCampaigns = GetModel("Campaigns").GetUserCampaigns(name, vcc, keyspace)
	channel = channel ~= "" and channel or nil
	for campaignIdAtVCC, _ in pairs(userCampaigns) do
		local campaign = GetModel("Campaigns").Get(campaignIdAtVCC, keyspace)
		if campaign then
			local outgoingAccounts = {}
			for _, account in pairs(campaign.accounts) do
				if (not channel or channel == account.proxyType) and IsOutgoingAccount(account) then
					outgoingAccounts[account.id] = { 
						name = account.name, 
						uid = account.uid 
					}
				end
				if not utils.TableIsEmpty(outgoingAccounts) and 
					model_perm.IsPermissionEnabledForUser("interaction", "int_create_outgoing", campaignIdAtVCC, keyspace) 
				then
					outgoingCampaigns[campaignIdAtVCC] = {
						name = campaign.Name,
						accounts = outgoingAccounts
					}
				end
			end
		end
	end
	keyspace:Close()
	return json.encode(outgoingCampaigns)
end

--[[
	Only updates campaign basic data
--]]
function SaveCampaign (updatedCampaign, keyspace, allegroOtherInfo)
	assert(type(updatedCampaign) == "table", "SaveCampaign error. Campaign does not a table.")

	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local campaign_with_vcc = utils.AppendVcc( utils.GetId(updatedCampaign.Id), vcc )

	LogInfo("[AGENT:SaveCampaign] Updating campaign '%s'. User %s on vcc %s.", campaign_with_vcc, name, vcc)

	local current_campaign = GetModel("Campaigns").Get(campaign_with_vcc, keyspace)
	if current_campaign then
		for k, v in pairs(current_campaign) do
			current_campaign[k] = updatedCampaign[k]
			updatedCampaign[k] = nil
		end
		for k, v in pairs(updatedCampaign) do
			current_campaign[k] = v
		end
		current_campaign.Id = campaign_with_vcc

		local success, err = GetModel("Campaigns").ShallowSave(current_campaign, keyspace, allegroOtherInfo)
		if not success then
			return false, err
		end
	else
		LogError("[AGENT:SaveCampaign] Tried to save an unexistent campaign %s", campaign_with_vcc)
		return false, { result = "UNEXISTENT CAMPAIGN" }
	end

	--notifications
	notificationsEngine.SendNotification("CampaignUpdated", 0, { campaign = current_campaign })

	return true
end

function CreateCampaign (campaign, keyspace, allegroOtherInfo)
	assert(type(campaign) == "table", "CreateCampaign error. Campaign does not a table.")

	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName

	LogInfo("[AGENT:CreateCampaign] Creating campaign. User %s on vcc %s campaign %s", name, vcc, campaign.Id)

	local campaignId = helpers.StripVCC(campaign.Id, vcc)
	local campaignIdAtVCC = campaignId.."@"..vcc
	campaign.Id = utf8.lower(campaignIdAtVCC)
	campaign.LastUpdate = LuaOrm.Date():toString()
	campaign.VCC = vcc
	campaign.AlwaysValid = 1
	campaign.defaultAgentCriteria = "CIRCULAR"

	-- user id validation ------------------
	if GetModel("Campaigns").CampaignExists(campaign.Id, keyspace) then
		local reason = string.format("Campaign ID '%s' Not available.", campaign.Id)
		return false, { result = "MUST_HAVE_UNIQUE_NAME", err_type = "ALREADY EXISTS", reason = reason }
	end
	--need to place a reservation for the campaign name, if possible then fully save it
	local reserved = GetModel("Campaigns").ReserveName(campaign.Id, name)
	if not reserved then
		local reason = string.format("User %s does not own the lock on campaign %s, will not save", name, campaign.Id)
		return false, { result = "MUST_HAVE_UNIQUE_NAME", err_type = "USED BY ANOTHER USER", reason = reason }
	end

	--default to a full time scheduler for the campaign
	campaign.scheduler = {
		scheduleDays = {
			{dayOfWeek = 1, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"},
			{dayOfWeek = 2, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"},
			{dayOfWeek = 3, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"},
			{dayOfWeek = 4, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"},
			{dayOfWeek = 5, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"},
			{dayOfWeek = 6, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"},
			{dayOfWeek = 7, startTime = "1970-01-01 00:00:00", endTime = "1970-01-01 00:00:00"}
		},
		scheduleSpecialDays = {},
		scheduleSpecialDayAllYearss = {}
	}
	campaign.ValidationTimeZone = campaign.ValidationTimeZone or "GMT"
	local success, saved_campaign = GetModel("Campaigns").CreateCampaign(campaign, keyspace, allegroOtherInfo)
	if not success then
		local err = saved_campaign
		GetModel("Campaigns").UnreserveName(campaign.Id)
		LogWarning("[AGENT:CreateCampaign] Unable to save campaign '%s'. Internal error.", campaign.Id)
		return false, saved_campaign
	end

	GetModel("Campaigns").UnreserveName(campaign.Id)

	--notifications
	notificationsEngine.SendNotification("CampaignCreated", 0, { campaign = campaign })

	return true, json.encode({campaign = campaign})
end

local ChangeAccountStateTasks_d = {
	FACEBOOK = function(campaignId, account, enabled, ksAgent)
		local pages = {}
		local pagesMessenger = account.pagesMessenger or {}
		if type(pagesMessenger) == "string" then
			pagesMessenger = json.decode(pagesMessenger)
		end
		for pageId, pageData in pairs(pagesMessenger) do
			pages[pageId] = true
		end
		local status, err = GetModel("Campaigns").RegisterUnregisterPagesWebhooks("FACEBOOK", account.id, account.accessToken, 
			enabled and pages or {}, not enabled and pages or {})
		return status, err
	end,
	INSTAGRAM = function(campaignId, account, enabled, ksAgent)
		local pagesFbIds = account.pagesFbIds or {}
		if type(pagesFbIds) == "string" then
			pagesFbIds = json.decode(pagesFbIds)
		end
		local status, err
		for igId, fbId in pairs(pagesFbIds or {}) do
			if enabled then
				err = GetModel("Campaigns").AddInstagramAccountByPage(campaignId, account.id, igId, ksAgent)
				if err then
					return false, err
				end
			else
				GetModel("Campaigns").RemoveInstagramAccountByPage(campaignId, account.id, igId, ksAgent)
			end
		end
		local status, err = GetModel("Campaigns").RegisterUnregisterPagesWebhooks("INSTAGRAM", account.id, account.accessToken, 
			enabled and pagesFbIds or {}, not enabled and pagesFbIds or {})
		return status, err
	end,
	YOUTUBE = function(campaignId, account, enabled, ksAgent)
		local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(account, enabled)
		return status, err
	end,
	APPSTORE = function(campaignId, account, enabled, ksAgent)
		local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(account, enabled)
		return status, err
	end,
	GOOGLEPLAY = function(campaignId, account, enabled, ksAgent)
		local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(account, enabled)
		return status, err
	end,
	RCS = function(campaignId, account, enabled, ksAgent)
		local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(account, enabled)
		return status, err
	end,
	TWITTER = function(campaignId, account, enabled, ksAgent)
		local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(account, enabled)
		return status, err
	end,
	WHATSAPP = function(campaignId, account, enabled, ksAgent)
		local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(account, enabled)
		return status, err
	end
}

local function ChangeAccountStateTasks (campaignId, account, enabled, ksAgent)
	if not ChangeAccountStateTasks_d[account.proxyType] then
		return true
	end
	return ChangeAccountStateTasks_d[account.proxyType](campaignId, account, enabled, ksAgent)
end

--Para activar una cuenta de WHATSAPP Cloud API no debe haber ninguna otra cuenta activa en este ambiente OCC con el mismo telefono
--En el webhook server se comprueba esto mismo pero para otros ambientes OCC
local function ValidateWhatsappCloudApiAccountStateChange(campaignAccount, ksAgent)
	--Validate only for whatsapp account activation
	if not campaignAccount.state or campaignAccount.proxyType ~= "WHATSAPP" then
		return true
	end

	local campaignToChange, vccToChange = utils.SeparateVCC(campaignAccount.campaignId)
	local accountToChange = GetModel("Campaigns").GetAccount(campaignAccount.accountId, campaignAccount.campaignId, ksAgent)
	--Validate only for Cloud API integration type
	if accountToChange.integrationType ~= "ICC_CLOUD_API" then
		return true
	end

	local allAccounts = GetModel("Campaigns").GetAllVCCsAccounts(ksAgent)
	for campaignAtVcc, campaignAccounts in pairs( allAccounts ) do
		local campaignId, vcc = utils.SeparateVCC(campaignAtVcc)
		for accountId, account in pairs(campaignAccounts) do
			if accountId ~= campaignAccount.accountId and account.state == "1" and account.proxyType == "WHATSAPP" and account.integrationType == "ICC_CLOUD_API" and account.phoneId == accountToChange.phoneId then
				if vcc ~= vccToChange then
					return false, { result = "WHATSAPP_PHONE_IN_USE_IN_ANOTHER_VCC", reason = "Whatsapp phone is in use in another vcc", phone_number = account.phoneNumber }
				end
				if campaignId ~= campaignToChange then
					local campaign = GetModel("Campaigns").Get(campaignAtVcc, ksAgent)
					local campaign_name = campaign and campaign.Name or ""
					return false, { result = "WHATSAPP_PHONE_IN_USE_IN_ANOTHER_CAMPAIGN", reason = "Another active account is using the same phone", phone_number = account.phoneNumber, account_name = account.name, campaign_name = campaign_name }
				end
				return false, { result = "WHATSAPP_PHONE_IN_USE_IN_ANOTHER_ACCOUNT_CURRENT_CAMPAIGN", reason = "Another active account is using the same phone", phone_number = account.phoneNumber, account_name = account.name }
			end
		end
	end
	return true
end

function ChangeAccountState (campaignAccount, campaignId, ksAgent)
	local success, error = ValidateWhatsappCloudApiAccountStateChange(campaignAccount,ksAgent)
	if not success then
		return false, error
	end
	
	local vcc = inconcert.authUser.virtualContactCenter
	local campaign = GetModel("Campaigns").Get(campaignId, ksAgent)

	if campaign then
		for i, account in pairs(campaign.accounts) do
			if account.id == campaignAccount.accountId and account.proxyType == campaignAccount.proxyType then
				if utils.HasAllegroIntegration(vcc, ksAgent) then
					if GetModel("Campaigns").IsAllegroAccount(account) then
						return false, {result = "UNABLE_TO_CHANGE_STATE_ALLEGRO_ACCOUNT", reason = "Unable to change state to allegro accounts."}
					end
				end
				local status, err = ChangeAccountStateTasks(campaignId, account, campaignAccount.state == true, ksAgent)
				if not status then
					return status, err
				end
				campaign.accounts[i].state = campaignAccount.state == true and "1" or "0"
				local result = GetModel("Campaigns").ChangeStateOmnichannelAccount(campaign)
				--notifications
				notificationsEngine.SendNotification(campaignAccount.state == true and "AccountEnabled" or "AccountDisabled", 0, { campaignId = utils.GetId(campaignId), accountId = account.id })
				return true
			end
		end
		return false, {result = "ACCOUNT_DOES_NOT_EXIST", reason = "Tried to change account status for unexistent account"}
	else
		return false, {result = "CAMPAIGN_DOES_NOT_EXIST", reason = "Tried to change account status for unexistent campaign " .. campaignId}
	end
end

local function RemoveSocialNetworksAccountValidation(accountId)
	local activePosts = GetModel("Campaigns").GetAccountSocialNetworkActivePosts(accountId)
	if not utils.TableIsEmpty(activePosts) then
		LogError("RemoveSocialNetworksAccountValidation: account %s has active posts", accountId)
		return false, "HAS_ACTIVE_POSTS"
	end
	return true
end

local function RemoveOutgoingBatchesAccountValidation(accountId)
	local activeBatches = GetModel("Campaigns").GetAccountActiveOutgoingBatches(accountId)
	if not utils.TableIsEmpty(activeBatches) then
		LogError("RemoveOutgoingBatchesAccountValidation: account %s has active batches", accountId)
		return false, "HAS_ACTIVE_BATCHES"
	end
	return true
end

local function FacebookRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	local pages = account.pages or {}
	if type( pages ) == "string" then
		pages = json.decode(pages)
	end
	for _, pageId in pairs(pages or {}) do
		local status, err = RemoveSocialNetworksAccountValidation(account.id .. ":" .. pageId)
		if err then
			return false, { result = err, page = pageId }
		end
	end
	return true
end

local function InstagramRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	local pages = account.pages or {}
	if type( pages ) == "string" then
		pages = json.decode(pages)
	end
	for _, pageId in pairs(pages or {}) do
		local status, err = RemoveSocialNetworksAccountValidation(account.id .. ":" .. pageId)
		if err then
			return false, { result = err, page = pageId }
		end
	end
	return true
end

local function YoutubeRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	return RemoveSocialNetworksAccountValidation(account.id)
end

local function TwitterRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	return RemoveSocialNetworksAccountValidation(account.id)
end

local function MailRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	return RemoveOutgoingBatchesAccountValidation(account.id)
end

local function SmsRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	return RemoveOutgoingBatchesAccountValidation(account.id)
end

local function WhatsappRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	ValidateWhatsappAccountInFlows(account, keyspace, true)

	return RemoveOutgoingBatchesAccountValidation(account.id)
end

local function RcsRemoveAccountValidation(campaignIdAtVCC, account, keyspace)
	return RemoveOutgoingBatchesAccountValidation(account.id)
end

local function ValidateCampaignAndAccountsRelations (info, response, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	response.errors = response.errors or {}
	--local eacdRelatedCampaigns
	--local rulesRelatedCampaigns
	local vccCampaigns = info.vccCampaigns or m_adm_models.Campaigns.GetVCCCampaigns(vcc, keyspace)
	local currentCampaign = info.campaignIdAtVCC:lower()

	LogInfo("[AGENT:ValidateCampaignAndAccountsRelations] Validating relations of campaign '%s' and account '%s'.", info.campaignIdAtVCC, json.encode(info.account or {}))
	--LogDebug("[AGENT:ValidateCampaignAndAccountsRelations] info '%s'.", json.encode(info))

	--LogDebug("[AGENT:ValidateCampaignAndAccountsRelations] 1-----> entra - response.errType '%s'.", response.errType)

	local currentAccountId = (info.account and info.account.id or ""):lower()

	for campaignIdWithoutVcc, vccCampaign in pairs(vccCampaigns) do
		local fullCampaignId = utils.AppendVcc(campaignIdWithoutVcc, vcc)
		vccCampaign = type(vccCampaign) == "string" and m_adm_models.Campaigns.GetCampaign(vcc, campaignIdWithoutVcc, keyspace) or vccCampaign
		-- channels
		local accounts = vccCampaign.accounts or m_adm_models.Campaigns.GetCampaignAccounts(fullCampaignId, keyspace)
		vccCampaign.accounts = accounts
		vccCampaigns[campaignIdWithoutVcc] = vccCampaign
	end

	--LogDebug("---- vccCampaigns %s", json.encode(vccCampaigns))
	for campaignIdWithoutVcc, vccCampaign in pairs(vccCampaigns) do
		local fullCampaignId = utils.AppendVcc(campaignIdWithoutVcc, vcc)
		if currentCampaign ~= fullCampaignId:lower() then
			-- vccCampaign = type(vccCampaign) == "string" and m_adm_models.Campaigns.GetCampaign(vcc, campaignIdWithoutVcc, keyspace) or vccCampaign
			-- -- channels
			-- local accounts = vccCampaign.accounts or m_adm_models.Campaigns.GetCampaignAccounts(fullCampaignId, keyspace)
			-- vccCampaign.accounts = accounts

--LogDebug("1 - response.errors %s", json.encode(response.errors or {}))
			local accounts = vccCampaign.accounts
			--LogDebug("-------------> accounts %s ", json.encode(accounts))
			for _, account in pairs(accounts or {}) do
				if account.ActionOnAcdWaitTimeout then
					if (account.ActionOnAcdWaitTimeout.action or ""):lower() == "transfertocampaign" then
						if (account.ActionOnAcdWaitTimeout.campaign or ""):lower() == currentCampaign then
							if (account.ActionOnAcdWaitTimeout.account or ""):lower() == currentAccountId then

								response.errors[campaignIdWithoutVcc] = response.errors[campaignIdWithoutVcc] or {rulesRelatedCampaigns = {}, acdRelatedCampaigns = {}}

								local campaignToTransfer = vccCampaigns[utils.GetId(account.ActionOnAcdWaitTimeout.campaign)]
								local accountToTransfer = campaignToTransfer.accounts[account.ActionOnAcdWaitTimeout.account]

								response.errors[campaignIdWithoutVcc].acdRelatedCampaigns = {
									relatedCampaignId = campaignIdWithoutVcc,
									to = {
										campaignId = utils.GetId(currentCampaign),
										accountId = accountToTransfer.id,
										accountName = accountToTransfer.name,
									},
									action = "transfer"
								}

								if not response.errType then
									response.errType = "CAMPAIGN IS USED IN ACD TIMEOUT OF OTHER CAMPAIGNS"
								elseif response.errType == "CAMPAIGN IS USED IN RULES OF OTHER CAMPAIGNS" then
									response.errType = "CAMPAIGN IS USED IN ACD TIMEOUT AND RULES OF OTHER CAMPAIGNS"
								end
							end
						end
					end
				end
			end
--LogDebug("2 - response.errors %s", json.encode(response.errors or {}))
			-- rules
			local campaignRules = vccCampaign.rules or m_adm_models.Campaigns.GetRules(fullCampaignId, keyspace)
			vccCampaign.rules = campaignRules
			if not utils.TableIsEmpty(campaignRules) then
				for _, rule in pairs(campaignRules) do
					local actions = rule.actions
					if actions and actions.transfer and (actions.transfer.destination or ""):lower() == "campaign" then
						if (actions.transfer.campaign or ""):lower() == currentCampaign then
							if (actions.transfer.account or ""):lower() == currentAccountId then
								--rulesRelatedCampaigns = rulesRelatedCampaigns or {}
								--LogDebug(" 1 currentCampaign %s", currentCampaign)
								--LogDebug(" 2 actions.transfer.account %s", actions.transfer.account)
								--LogDebug(" 3 vccCampaign.accounts %s", vccCampaign.accounts)
								local campaignToTransfer = vccCampaigns[utils.GetId(actions.transfer.campaign)]
								local account = campaignToTransfer.accounts[actions.transfer.account]
								--LogDebug(" 3 accountId %s, account: %s", actions.transfer.account, json.encode(account or {}))
								response.errors[campaignIdWithoutVcc] = response.errors[campaignIdWithoutVcc] or {rulesRelatedCampaigns = {}, acdRelatedCampaigns = {}}

								local rulesRelatedCampaigns = response.errors[campaignIdWithoutVcc].rulesRelatedCampaigns

								rulesRelatedCampaigns[rule.name] = {
									relatedCampaignId = campaignIdWithoutVcc,
									to = {
										campaignId = utils.GetId(currentCampaign),
										accountId = actions.transfer.account,
										accountName = account.name
									},
									action = "transfer",
									ruleName = rule.name
								}

								if not response.errType then
									response.errType = "CAMPAIGN IS USED IN RULES OF OTHER CAMPAIGNS"
								elseif response.errType == "CAMPAIGN IS USED IN ACD TIMEOUT OF OTHER CAMPAIGNS" then
									response.errType = "CAMPAIGN IS USED IN ACD TIMEOUT AND RULES OF OTHER CAMPAIGNS"
								end
							end
						end
					end
				end
			end
		end
	end

	info.vccCampaigns = vccCampaigns

	--LogDebug("[AGENT:ValidateCampaignAndAccountsRelations] 2-----> sale - response.errType '%s'.", response.errType)

	if response.errType then
		return false, {errType = response.errType}
	end

	-- ok
	return true, info
end

local function ValidateAccountToRemove (campaignIdAtVCC, account, response, keyspace)
	local info = {
		campaignIdAtVCC = campaignIdAtVCC,
		account = account,
		vccCampaigns = response.vccCampaigns
	}
	local success
	success, info = ValidateCampaignAndAccountsRelations(info, response, keyspace)
	if not success then
		local err = info
		return false, err, "RELATIONS ERROR"
	end
	local specificsAccountsTypeValidation = {
		FACEBOOK  = FacebookRemoveAccountValidation,
		INSTAGRAM = InstagramRemoveAccountValidation,
		YOUTUBE   = YoutubeRemoveAccountValidation,
		TWITTER   = TwitterRemoveAccountValidation,
		MAIL      = MailRemoveAccountValidation,
		SMS       = SmsRemoveAccountValidation,
		WHATSAPP  = WhatsappRemoveAccountValidation,
		RCS       = RcsRemoveAccountValidation
	}
	if specificsAccountsTypeValidation[account.proxyType] then
		return specificsAccountsTypeValidation[account.proxyType](campaignIdAtVCC, account, keyspace)
	end
	return true, info
end

function ValidateCampaignAccountsToRemove (campaignIdAtVCC, campaignAccounts, keyspace)
	--LogDebug("[AGENT.Campaigns.ValidateCampaignAccountsToRemove] ENTRA --> campaignIdAtVCC: %sm accounts: %s", campaignIdAtVCC, json.encode(campaignAccounts or {}))
	local isOk = true
	local response = {
		errors = {}
	}
	for _, account in pairs(campaignAccounts or {}) do
		local success, result, errType = ValidateAccountToRemove(campaignIdAtVCC, account, response, keyspace)
		if not success then
			isOk = false
			if "RELATIONS ERROR" ~= errType then
				local err = result
				if type(err) == "string" then
					err = {result = err}
				end
				err.account = account.id
				if not err.errType then
					err.errType = err.result
				end
				--LogDebug("[AGENT.Campaigns.ValidateCampaignAccountsToRemove]    1-SALE --> isOk: %s err: %s", false, json.encode(err))
				return false, err
			end
		end
	end
	--LogDebug("[AGENT.Campaigns.ValidateCampaignAccountsToRemove] SALE --> isOk: %s response: %s", isOk, json.encode(response))
	return isOk, response
end

function RemoveCampaignAccount (account_to_remove, campaignId, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local campaign = GetModel("Campaigns").Get(campaignId, keyspace)
	local response = {
		errors = {}
	}
	if campaign then
		for i, account in pairs(campaign.accounts) do
			if account.id == account_to_remove.accountId and account.proxyType == account_to_remove.proxyType then
				if utils.HasAllegroIntegration(vcc) then
					if GetModel("Campaigns").IsAllegroAccount(account) then
						return false, {result = "UNABLE_TO_DELETE_ALLEGRO_ACCOUNT", reason = "Unable to delete to allegro accounts."}
					end
				end
				local success, info, errType = ValidateAccountToRemove(campaignId, account, response, keyspace)
				if not success then
					local err = info
					if type(err) == "string" then
						err = {result = err}
					end
					err.errors = response.errors
					if not err.errType then
						err.errType = err.result
					end
					return false, err
				end
				campaign.accounts[i] = nil
				campaign.removedAccounts = campaign.removedAccounts or {}
				campaign.removedAccounts[account.id] = account_to_remove.data
				local result = GetModel("Campaigns").RemoveOmnichannelAccount(campaign, keyspace)
				--notifications
				notificationsEngine.SendNotification("AccountDeleted", 0, {campaignId = utils.GetId(campaignId), accountId = account.id})
				return true, account
			end
		end
		return false, {result = "ACCOUNT_DOES_NOT_EXIST", reason = "Tried to delete account for unexistent account"}
	else
		return false, {result = "CAMPAIGN_DOES_NOT_EXIST", reason = "Tried to delete account for unexistent campaign " .. campaignId}
	end
end

local function CallAccountValidation ( posted_account, _posted_campaignId, keyspace, _original_account )
	local vcc = inconcert.authUser.virtualContactCenter
	if utils.HasAllegroIntegration(vcc, keyspace) then
		return true,""
	end

	didsPool = posted_account.didsPool or {}
	GetModel("Campaigns").UpdateAccountDidsPool(posted_account.id, didsPool.dids or {})
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Facebook"
	1 - Se esta validando que al crear/actualizar una cuenta,
		sus PAGINAS seleccionados no se encuentren siendo utilizadas por otra cuenta,
		en la misma campana, otra campana, o en otro vcc.
--]]
local function FacebookAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "FacebookAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "FacebookAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Facebook Account: '%s'.", json.encode(posted_account))

	--make pages lists
	local originalPages = original_account and original_account.pages or {}
	if type(originalPages) == "string" then
		originalPages = json.decode(originalPages)
	end
	local posted_pages_d = {}
	local removedPages = {}
	for _, pageId in pairs(posted_account.pages or {}) do
		posted_pages_d[pageId] = true
	end
	for _, pageId in pairs(originalPages) do
		if not posted_pages_d[pageId] then
			removedPages[pageId] = true
		end
	end

	--check removed pages
	for pageId, _ in pairs(removedPages) do
		local pagePosts = GetModel("Campaigns").GetAccountSocialNetworkActivePosts(posted_account.id .. ":" .. pageId)
		if not utils.TableIsEmpty(pagePosts) then
			return false, { 
				result = "HAS_ACTIVE_POSTS", 
				reason = string.format("removed page %s has active posts", pageId) 
			}
		end
	end

	--process pages
	for pageId, _ in pairs(posted_pages_d) do
		GetModel("Campaigns").AddFacebookAccountByPage(posted_campaignId, posted_account.id, pageId)
	end
	--process removed pages
	for pageId, _ in pairs(removedPages) do
		GetModel("Campaigns").DeleteAccountStatistics(posted_account.id .. ":" .. pageId)
		GetModel("Campaigns").RemoveFacebookAccountByPage(pageId)
	end
	
	local mustCleanMessenger = posted_account.mustCleanMessenger
	posted_account.mustCleanMessenger = nil
	if mustCleanMessenger then
		GetModel("Campaigns").FacebookCleanMessengerConfiguration(posted_account.id, posted_account.accessToken, removedPages)
	end

	--check messenger registration
	local registerPages = {}
	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	if enabled then
		for pageId, pageData in pairs(posted_account.pagesMessenger or {}) do
			if pageData and pageData.processFlow then
				registerPages[pageId] = true
			end
		end
	end
	local unregisterPages = {}
	local originalPagesMessenger = original_account and original_account.pagesMessenger or {}
	if type(originalPagesMessenger) == "string" then
		originalPagesMessenger = json.decode(originalPagesMessenger)
	end
	for pageId, pageData in pairs(originalPagesMessenger) do
		if pageData and pageData.processFlow and not registerPages[pageId] then
			unregisterPages[pageId] = true
		end
	end
	if not utils.TableIsEmpty(registerPages) or not utils.TableIsEmpty(unregisterPages) then
		local status, err = GetModel("Campaigns").RegisterUnregisterPagesWebhooks("FACEBOOK", posted_account.id, posted_account.accessToken, registerPages, unregisterPages)
		if not status then
			return status, err
		end
	end

	--clean page tokens if account was authorized
	if posted_account.authorized then
		GetModel("Campaigns").CleanFacebookPagesData(posted_account.uid)
		posted_account.authorized = nil
	end

	local posted_pages = posted_account.pagesNames
	if not posted_pages then
		return true
	end

	-- Todas las cuentas de todas las campana de todos los vccs
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts()
	if not all_accounts then
		return true
	end

	-- recorro todas las campanas
	for current_campaign, current_accounts in pairs( all_accounts ) do
		-- recorro todas las accounts
		for full_acc_id, current_account in pairs( current_accounts ) do

			-- si la current account es distinta a la que se ha recibido por parametro, continua con la validacion, sino continua en el loop
			if full_acc_id ~= posted_account.id then
				-- recorre el hash de pages de la current account
				if current_account.pagesNames then

					-- ahora recorre las pages de la account que se ha recibido por parametro
					--  y valida si alguna de ellas se encuentra utilizada en las pages de la current account.
					--  si es asi, y alguna se encuentra utilizada, falla!!!, sino continuna buscando en las demas accounts de las campanas
					for page_id, page_name in pairs( posted_pages ) do

						if current_account.pagesNames[ page_id ] then
						-- Error (already exists)!!!

							local simple_posted_campaignId, posted_vcc = helpers.SeparateVCC( posted_campaignId )
							local camp_id, current_vcc = helpers.SeparateVCC( current_campaign )

							local err = { page_id = page_id, page_name = page_name, campaign = camp_id, account_id = full_acc_id }

							if posted_campaignId:lower() == current_campaign:lower() then
							-- se encuentra utilizada en una cuenta de la misma campana
								err.result = "FACEBOOK_PAGE_IN_USE_IN_ANOTHER_ACCOUNT_CURRENT_CAMPAIGN"
								err.reason = string.format("Unable to save facebook account '%s' on campaign '%s'. Facebook page id: '%s', name: '%s' in use in another account '%s' of the current campaign.", posted_account.id, posted_campaignId, page_id, page_name, full_acc_id)
							else
								if ( posted_vcc or "" ):lower() == ( current_vcc or "" ):lower() then
								-- se encuentra utilizada en una cuenta de otra campana del mismo VCC
									err.result = "FACEBOOK_PAGE_IN_USE_IN_ANOTHER_CAMPAIGN"
									err.reason = string.format("Unable to save facebook account '%s' on campaign '%s'. Facebook page id: '%s', name: '%s' in use in another account '%s' of the '%s' campaign.", posted_account.id, posted_campaignId, page_id, page_name, full_acc_id, current_campaign )
								else
								-- se encuentra utilizada en una cuenta que pertenece a otro VCC
									err.result = "FACEBOOK_PAGE_IN_USE_IN_ANOTHER_VCC"
									err.reason = string.format("Unable to save facebook account '%s' on campaign '%s'. Facebook page id: '%s', name: '%s' in use in another account '%s' of the '%s' campaign of vcc '%s'.", posted_account.id, posted_campaignId, page_id, page_name, full_acc_id, current_campaign, current_vcc )
								end
							end

							return false, err
						end
					end
				end
			end
		end
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Instagram"
	1 - Se esta validando que al crear/actualizar una cuenta,
		sus PAGINAS seleccionados no se encuentren siendo utilizadas por otra cuenta,
		en la misma campana, otra campana, o en otro vcc.
--]]
local function InstagramAccountValidation ( posted_account, posted_campaignId, ksAgent, original_account )
	assert(type(posted_account) == "table", "InstagramAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "InstagramAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Instagram Account: '%s'.", json.encode(posted_account))

	--clean deleted pages data
	local originalPagesFbIds = original_account and original_account.pagesFbIds or {}
	if type(originalPagesFbIds) == "string" then
		originalPagesFbIds = json.decode(originalPagesFbIds)
	end
	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	
	local registerPages = {}
	local unregisterPages = {}
	
	--removed pages
	for igId, fbId in pairs(originalPagesFbIds) do
		if not posted_account.pagesFbIds[igId] then
			local pagePosts = GetModel("Campaigns").GetAccountSocialNetworkActivePosts(posted_account.id .. ":" .. igId)
			if not utils.TableIsEmpty(pagePosts) then
				return false, { 
					result = "HAS_ACTIVE_POSTS", 
					reason = string.format("removed page %s has active posts", igId) 
				}
			end
			unregisterPages[igId] = fbId
		end
	end
	for igId, fbId in pairs(unregisterPages) do
		GetModel("Campaigns").DeleteAccountStatistics(posted_account.id .. ":" .. igId, ksAgent)
		GetModel("Campaigns").RemoveInstagramAccountByPage(posted_campaignId, posted_account.id, igId, ksAgent)
	end
	
	--added pages
	local pagesByFbId = {}
	for igId, fbId in pairs(posted_account.pagesFbIds or {}) do
		pagesByFbId[fbId] = igId
		if enabled then
			registerPages[igId] = fbId
		end
		local err = GetModel("Campaigns").AddInstagramAccountByPage(posted_campaignId, posted_account.id, igId, ksAgent)
		if err then
			return false, err
		end
	end
	
	if not utils.TableIsEmpty(registerPages) or not utils.TableIsEmpty(unregisterPages) then
		local status, err = GetModel("Campaigns").RegisterUnregisterPagesWebhooks("INSTAGRAM", posted_account.id, posted_account.accessToken, registerPages, unregisterPages)
		if not status then
			return status, err
		end
	end

	--clean page tokens if account was authorized
	if posted_account.authorized then
		GetModel("Campaigns").CleanInstagramPagesData(posted_account.uid)
		posted_account.authorized = nil
	end

	-- Todas las cuentas de todas las campana de todos los vccs
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts(ksAgent)
	if not all_accounts then
		return true
	end

	-- recorro todas las campanas
	for current_campaign, current_accounts in pairs( all_accounts ) do
		-- recorro todas las accounts
		for full_acc_id, current_account in pairs( current_accounts ) do

			-- si la current account está habilitada y es distinta a la que se ha recibido por parametro, continua con la validacion, sino continua en el loop
			if current_account.state == "1" and full_acc_id ~= posted_account.id then
				-- recorre el hash de pages de la current account
				if current_account.pagesFbIds then
					-- ahora recorre las pages de la account que se ha recibido por parametro
					--  y valida si alguna de ellas se encuentra utilizada en las pages de la current account.
					--  si es asi, y alguna se encuentra utilizada, falla!!!, sino continuna buscando en las demas accounts de las campanas
					for ig_id, fb_id in pairs( current_account.pagesFbIds ) do
						local existing_page_id
						if posted_account.pagesFbIds[ig_id] then
							existing_page_id = ig_id
						elseif pagesByFbId[fb_id] then
							existing_page_id = pagesByFbId[fb_id]
						end
						
						if existing_page_id then
						-- Error (already exists)!!!

							local simple_posted_campaignId, posted_vcc = helpers.SeparateVCC( posted_campaignId )
							local camp_id, current_vcc = helpers.SeparateVCC( current_campaign )
							local page_name = posted_account.pagesNames[existing_page_id]

							local err = { page_id = existing_page_id, page_name = page_name, campaign = camp_id, account_id = full_acc_id }

							if posted_campaignId:lower() == current_campaign:lower() then
							-- se encuentra utilizada en una cuenta de la misma campana
								err.result = "INSTAGRAM_USER_IN_USE_IN_ANOTHER_ACCOUNT_CURRENT_CAMPAIGN"
								err.reason = string.format("Unable to save instagram account '%s' on campaign '%s'. Instagram page id: '%s', name: '%s' in use in another account '%s' of the current campaign.", posted_account.id, posted_campaignId, existing_page_id, page_name, full_acc_id)
							else
								if ( posted_vcc or "" ):lower() == ( current_vcc or "" ):lower() then
								-- se encuentra utilizada en una cuenta de otra campana del mismo VCC
									err.result = "INSTAGRAM_USER_IN_USE_IN_ANOTHER_CAMPAIGN"
									err.reason = string.format("Unable to save instagram account '%s' on campaign '%s'. Instagram page id: '%s', name: '%s' in use in another account '%s' of the '%s' campaign.", posted_account.id, posted_campaignId, existing_page_id, page_name, full_acc_id, current_campaign )
								else
								-- se encuentra utilizada en una cuenta que pertenece a otro VCC
									err.result = "INSTAGRAM_USER_IN_USE_IN_ANOTHER_VCC"
									err.reason = string.format("Unable to save instagram account '%s' on campaign '%s'. Instagram page id: '%s', name: '%s' in use in another account '%s' of the '%s' campaign of vcc '%s'.", posted_account.id, posted_campaignId, existing_page_id, page_name, full_acc_id, current_campaign, current_vcc )
								end
							end

							return false, err
						end
					end
				end
			end
		end
	end

	-- success
	return true
end

local function MakeAccountConflictError(posted_campaignId, posted_account, campaignId, accountId, err, result, reason)
	local simple_posted_campaignId, posted_vcc = helpers.SeparateVCC( posted_campaignId )
	local camp_id, current_vcc = helpers.SeparateVCC( campaignId )
	err.campaign = campaignId
	err.account_id = accountId
	if posted_campaignId:lower() == campaignId:lower() then
		--same campaign
		err.result = result .. "_IN_ANOTHER_ACCOUNT_CURRENT_CAMPAIGN"
		err.reason = string.format("Unable to save account '%s' on campaign '%s'. %s in another account '%s' of the current campaign.", 
			posted_account.id, posted_campaignId, reason, accountId)
	elseif ( posted_vcc or "" ):lower() == ( current_vcc or "" ):lower() then
		--same vcc
		err.result = result .. "_IN_ANOTHER_CAMPAIGN"
		err.reason = string.format("Unable to save account '%s' on campaign '%s'. %s in another account '%s' of the '%s' campaign.", 
			posted_account.id, posted_campaignId, reason, accountId, campaignId)
	else
		err.result = result .. "_IN_ANOTHER_VCC"
		err.reason = string.format("Unable to save account '%s' on campaign '%s'. %s in another account '%s' of the '%s' campaign in '%s' vcc.", 
			posted_account.id, posted_campaignId, reason, accountId, campaignId, current_vcc)
	end
	return err
end

local function WhatsappWebAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )
	return true
end

local function WhatsappWavyAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )
	--validate username is unique
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts()
	for campaignId, campaign_accounts in pairs(all_accounts or {}) do
		for accountId, account in pairs(campaign_accounts) do
			if accountId ~= posted_account.id then
				if account.integrationType == "WAVY" and (account.userName or ""):lower() == (posted_account.userName or ""):lower() then
					return false, MakeAccountConflictError(posted_campaignId, posted_account, campaignId, accountId, 
						{ user = posted_account.userName },
						"WHATSAPP_WAVY_USERNAME_IN_USE", 
						string.format("Wavy username '%s' in use", posted_account.userName))
				end
			end
		end
	end
	return true
end

local function WhatsappSmoochAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )
	--validate phoneNumber is unique
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts()
	for campaignId, campaign_accounts in pairs(all_accounts or {}) do
		for accountId, account in pairs(campaign_accounts) do
			if accountId ~= posted_account.id then
				if account.integrationType == "SMOOCH" and (account.phoneNumber or ""):lower() == (posted_account.phoneNumber or ""):lower() then
					return false, MakeAccountConflictError(posted_campaignId, posted_account, campaignId, accountId, 
						{ user = posted_account.phoneNumber },
						"WHATSAPP_PHONE_IN_USE", 
						string.format("Smooch phone '%s' in use", posted_account.phoneNumber))
				end
			end
		end
	end
	return true
end

local function WhatsappInfobipAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )
	--validate phoneNumber is unique
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts()
	for campaignId, campaign_accounts in pairs(all_accounts or {}) do
		for accountId, account in pairs(campaign_accounts) do
			if accountId ~= posted_account.id then
				if account.integrationType == "INFOBIP" and (account.phoneNumber or ""):lower() == (posted_account.phoneNumber or ""):lower() then
					return false, MakeAccountConflictError(posted_campaignId, posted_account, campaignId, accountId, 
						{ user = posted_account.phoneNumber },
						"WHATSAPP_PHONE_IN_USE", 
						string.format("Infobip phone '%s' in use", posted_account.phoneNumber))
				end
			end
		end
	end
	return true
end

local function WhatsappBusinessAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )
	--validate phoneNumber is unique
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts()
	for campaignId, campaign_accounts in pairs(all_accounts or {}) do
		for accountId, account in pairs(campaign_accounts) do
			if accountId ~= posted_account.id then
				if account.integrationType == "BUSINESS" and (account.phoneNumber or ""):lower() == (posted_account.phoneNumber or ""):lower() then
					return false, MakeAccountConflictError(posted_campaignId, posted_account, campaignId, accountId, 
						{ user = posted_account.phoneNumber },
						"WHATSAPP_PHONE_IN_USE", 
						string.format("Business phone '%s' in use", posted_account.phoneNumber))
				end
			end
		end
	end
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Whatsapp"
--]]
local function WhatsappAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )
	assert(type(posted_account) == "table", "WhatsappAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "WhatsappAccountValidation error. CampaignId does not a string.")

	local integrationType = posted_account.integrationType or "WEB"
	LogInfo("[AGENT.CAMPAIGNS:WhatsappAccountValidation] Validating Add/Update Whatsapp Account. integrationType: '%s', campaign '%s'.", integrationType, posted_campaignId)

	-- validaciones contra los Process Flows. Primero con la original_account
	local success, resp = ValidateWhatsappAccountInFlows(original_account, keyspace, true)
	if not success then
		local err = resp
		LogError("[AGENT.CONTROLLER.WhatsappAccountValidation] Error Validating Add/Update Whatsapp Account %s, campaign %s. Flow validation error.", posted_account.id, posted_campaignId)
		return false, err
	end

	-- validaciones contra los Process Flows. Ahora con la account que se quiere guardar
	success, resp = ValidateWhatsappAccountInFlows(posted_account, keyspace)
	if not success then
		local err = resp
		LogError("[AGENT.CONTROLLER.WhatsappAccountValidation] Error Validating Add/Update Whatsapp Account %s, campaign %s. Flow validation error.", posted_account.id, posted_campaignId)
		return false, err
	end

	local specificsIntegrationTypeValidation = {
		WEB      = WhatsappWebAccountValidation,
		WAVY     = WhatsappWavyAccountValidation,
		SMOOCH   = WhatsappSmoochAccountValidation,
		INFOBIP  = WhatsappInfobipAccountValidation,
		BUSINESS = WhatsappBusinessAccountValidation
	}
	local integrationValidation = specificsIntegrationTypeValidation[integrationType]
	if integrationValidation then
		local status, err = integrationValidation( posted_account, posted_campaignId, keyspace, original_account )
		if not status then
			LogError("[AGENT.CAMPAIGNS:WhatsappAccountValidation] - Error Validating Add/Update Whatsapp Account. integrationType: '%s', campaign '%s', status '%s', status '%s'.", integrationType, posted_campaignId, status, err)
			return status, err
		end
	end

	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(posted_account, enabled, original_account)
	if not status then
		LogError("[AGENT.CAMPAIGNS:WhatsappAccountValidation] -- Error Validating Add/Update Whatsapp Account. integrationType: '%s', campaign '%s', status '%s', status '%s'.", integrationType, posted_campaignId, status, utils.JSON_Encode(err))
		return status, err
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Twitter"
--]]
local function TwitterAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "TwitterAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "TwitterAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Twitter Account: '%s'.", json.encode(posted_account))

	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(posted_account, enabled)
	if not status then
		return status, err
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Youtube"
--]]
local function YoutubeAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "YoutubeAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "YoutubeAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Youtube Account: '%s'.", json.encode(posted_account))

	if original_account and original_account.uid ~= posted_account.uid then
		return false, { 
			result = "YOUTUBE_UID_CHANGED", 
			reason = "Can not change account channel"
		}
	end

	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(posted_account, enabled)
	if not status then
		return status, err
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Appstore"
--]]
local function AppstoreAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "AppstoreAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "AppstoreAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Appstore Account: '%s'.", json.encode(posted_account))

	if original_account and original_account.uid ~= posted_account.uid then
		return false, { 
			result = "APPSTORE_UID_CHANGED", 
			reason = "Can not change account channel"
		}
	end

	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(posted_account, enabled)
	if not status then
		return status, err
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Googleplay"
--]]
local function GoogleplayAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "GoogleplayAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "GoogleplayAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Googleplay Account: '%s'.", json.encode(posted_account))

	if original_account and original_account.uid ~= posted_account.uid then
		return false, { 
			result = "GOOGLEPLAY_UID_CHANGED", 
			reason = "Can not change account channel"
		}
	end

	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(posted_account, enabled)
	if not status then
		return status, err
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Rcs"
--]]
local function RcsAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "RcsAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "RcsAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Rcs Account: '%s'.", json.encode(posted_account))

	if original_account and original_account.uid ~= posted_account.uid then
		return false, { 
			result = "RCS_UID_CHANGED", 
			reason = "Can not change account channel"
		}
	end

	local enabled = not original_account or original_account.state == true or original_account.state == "1" or false
	local status, err = GetModel("Campaigns").RegisterUnregisterWebhook(posted_account, enabled)
	if not status then
		return status, err
	end

	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Mail"
	1 - Se esta validando que al crear/actualizar una cuenta,
		sus PAGINAS seleccionados no se se encuentren siendo utilizadas por otra cuenta,
		en la misma campana, otra campana, o en otro vcc.
--]]
function MailAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "MailAccountValidation error. Account does not a table.")

	LogInfo("Validating Add/Update Mail Account: '%s'.", json.encode(posted_account))

	local posted_incoming = posted_account.incomingConfig
	local posted_outgoing = posted_account.outgoingConfig

	if utils.TableIsEmpty(posted_incoming) and utils.TableIsEmpty(posted_outgoing) then
		return true
	end

	-- Todas las cuentas de todas las campanas de todos los vccs
	local all_accounts = m_adm_models.Campaigns.GetAllVCCsAccounts( keyspace )
	if not all_accounts then
		return true
	end

	local IsOk = function( current_config, posted_config )
		-- valida el user name
		if utils.AreEqualStrings( current_config.user, posted_config.user ) then
			-- valida la url del server (entrante o saliente)
			if utils.AreEqualStrings( current_config.url, posted_config.url ) then
				return false
			end
		end
		return true
	end

	local vcc = inconcert.authUser.virtualContactCenter

	posted_campaignId = posted_campaignId or ""
	local GetError = function( current_campaign, current_account, typeof )

		local camp_id, current_vcc = helpers.SeparateVCC( current_campaign )
		local simple_posted_campaignId, posted_vcc  = helpers.SeparateVCC( posted_campaignId )
		posted_vcc = posted_vcc or vcc

		local err = { campaign = camp_id, account_id = current_account.id, typeof = typeof, mail = current_account.address }
		if typeof == "INCOMING" then
			err.user = current_account.incomingConfig.user or ""
			err.url = current_account.incomingConfig.url or ""
		else
			err.user = current_account.outgoingConfig.user or ""
			err.url = current_account.outgoingConfig.url or ""
		end
		local err_user_server = string.format("%s - %s", err.user, err.url)

		if posted_campaignId:lower() == current_campaign:lower() then
		-- se encuentra utilizada en una cuenta de la misma campana
			err.result = "MAIL_%s_USER_IN_USE_IN_ANOTHER_ACCOUNT_CURRENT_CAMPAIGN"
			err.reason = string.format("1.Unable to save mail account '%s' on campaign '%s'. %s Mail user/url: '%s', in use in another account '%s' of the current campaign.", posted_account.id, posted_campaignId, typeof, err_user_server, current_account.id)
		else
			if ( posted_vcc or "" ):lower() == ( current_vcc or "" ):lower() then
			-- se encuentra utilizada en una cuenta de otra campana del mismo VCC
				err.result = "MAIL_%s_USER_IN_USE_IN_ANOTHER_CAMPAIGN"
				err.reason = string.format("2.Unable to save mail account '%s' on campaign '%s'. %s Mail user/url: '%s' in use in another account '%s' of the '%s' campaign.", posted_account.id, posted_campaignId, typeof, err_user_server, current_account.id, current_campaign )
			else
			-- se encuentra utilizada en una cuenta que pertenece a otro VCC
				err.result = "MAIL_%s_USER_IN_USE_IN_ANOTHER_VCC"
				err.reason = string.format("3.Unable to save mail account '%s' on campaign '%s'. %s Mail user/url: '%s' in use in another account '%s' of the '%s' campaign of vcc '%s'.", posted_account.id, posted_campaignId, typeof, err_user_server, current_account.id, current_campaign, current_vcc )
			end
		end

		err.result = string.format( err.result, typeof )

		return err
	end

	-- 1 - Valido contra las accounts de todas las campanas de todos los VCCs
	for current_campaign, current_accounts in pairs( all_accounts ) do
		-- recorro todas las accounts
		for full_acc_id, current_account in pairs( current_accounts ) do

			-- si la current account es distinta a la que se ha recibido por parametro, continua con la validacion, sino continua en el loop
			if full_acc_id ~= posted_account.id then
				-- valido la incoming configuracion
				local has_incoming_config = not utils.TableIsEmpty(current_account.incomingConfig)
				if has_incoming_config and not IsOk( current_account.incomingConfig, posted_incoming ) then
				-- Incoming Error (already exists)!!!
					return false, GetError( current_campaign, current_account, "INCOMING" )
				end
				-- valido la outgoing configuracion
				local has_outgoing_config = not utils.TableIsEmpty(current_account.outgoingConfig)
				if has_outgoing_config and not IsOk( current_account.outgoingConfig, posted_outgoing ) then
				-- Outgoing Error (already exists)!!!
					return false, GetError( current_campaign, current_account, "OUTGOING" )
				end
			end
		end
	end

	-- 2 - Valido contra las accounts salientes a nivel VCC de todos los VCCs
	local mail_accounts_by_vcc = m_adm_models.VCCs.GetAccountsOfAllVCCs( posted_account.proxyType, keyspace )
	if mail_accounts_by_vcc then
		-- recorro todas las campanas
		for current_vcc, vcc_accounts in pairs( mail_accounts_by_vcc ) do
			-- recorro todas las accounts
			for full_acc_id, vcc_account in pairs( vcc_accounts ) do
				-- si la current account es distinta a la que se ha recibido por parametro, continua con la validacion, sino continua en el loop
				if vcc_account.id ~= posted_account.id then
					-- valido la outgoing configuracion
					if vcc_account.outgoingConfig and not IsOk( vcc_account.outgoingConfig, posted_outgoing ) then
					-- Outgoing Error (already exists)!!!

						local err = {
							account_id = vcc_account.id,
							mail       = vcc_account.address,
							typeof     = "OUTGOING",
							user       = vcc_account.outgoingConfig.user,
							url        = vcc_account.outgoingConfig.url or ""
						}
						local err_user_server = err.user .. " - " .. err.url

						if vcc == current_vcc then
							err.result = "MAIL_OUTGOING_USER_IN_USE_VCCS_ACCOUNTS"
						else
							err.result = "MAIL_OUTGOING_USER_IN_USE_IN_ANOTHER_VCC"
						end

						err.reason = string.format("4.Unable to save mail account '%s' on campaign '%s'. OUTGOING Mail user/url: '%s' in use in another outgoing account '%s' of vcc '%s'.", posted_account.id, posted_campaignId, err_user_server, vcc_account.id, current_vcc )

						return false, err
					end
				end
			end
		end
	end
	-- success
	return true
end

--[[
	Esta funcion realiza validaciones especificas sobre el tipo de cuenta "Webcontact"
--]]
local function WebcontactAccountValidation ( posted_account, posted_campaignId, keyspace, original_account )

	assert(type(posted_account) == "table", "WebcontactAccountValidation error. Account does not a table.")
	assert(type(posted_campaignId) == "string", "WebcontactAccountValidation error. CampaignId does not a string.")

	LogInfo("Validating Add/Update Webcontact Account: '%s'.", json.encode(posted_account))

	GetModel("Campaigns").InsertAccountToken(posted_account.Token, posted_campaignId, posted_account.id)

	-- success
	return true
end

local function GetCampaignAvatar (web, id)
	local vcc = inconcert.authUser.virtualContactCenter

	LogInfo("Retrieving campaign avatar id %s ", id)
	web.headers["Content-Type"] = "image/png"
	web.headers["Cache-Control"] = "max-age=172800, public, must-revalidate"
	--this is done in order to avoid api manglers retrieving contacts from another vcc from the one they're logged in
	local campaignIdAtVCC = helpers.AddVCC(id, vcc)

	local campaign = GetModel("Campaigns").Get(campaignIdAtVCC)
	if (campaign) then
		if (campaign.Avatar == nil) then
			local avatar = avatar.FindBestCampaignAvatar(campaign)
			if (avatar) then
				GetModel("Campaigns").SaveAvatar(campaign, avatar)
				campaign.Avatar = avatar.image
				return campaign.Avatar
			end
		else
			return campaign.Avatar
		end
	end
	local genericAvatar = avatar.GetGenericAvatar()
	if not genericAvatar then
		return
	end
	return genericAvatar.image
end

local function GetCampaignDispositions (web, id)
	local vcc = inconcert.authUser.virtualContactCenter

	LogInfo("Retrieving campaign dispositions id %s ", id)
	local campaignIdAtVCC = helpers.AddVCC(id, vcc)
	local campaign = GetModel("Campaigns").Get(campaignIdAtVCC)
	local dispositions = campaign and campaign.dispositionCodes or "[]"
	return dispositions
end

local function SetAccountDefaultValues (account, campaign, is_new_account, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter

	if not utils.HasAllegroIntegration(vcc, ksAgent) then
		if account.proxyType == "CALL" and (not account.outgoingConfig or not account.outgoingConfig.outboundPrefix) then
			if not account.outgoingConfig then
				account.outgoingConfig = {}
			end
			account.outgoingConfig.outboundPrefix = utils.MakeOutgoingPrefixForCalli6(vcc, ksAgent)
		end
	end

	if is_new_account then
		if account.proxyType == "MAIL" then
			if account.type == "IMAP" then
				account.mustKeepFolder = "0"
			end
		end
	end

	-- Siempre que se actualiza una cuenta tiene que quedar con el state en "1" (activa)
	account.state = "1"
end

---
--
local function SaveNewAccount (campaignId, account, allegroOtherInfo, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter
	local campaign = GetModel("Campaigns").Get(campaignId, ksAgent)

	if not campaign then
		LogError("[AGENT:CAMPAIGNS:SaveNewAccount] Unable to add account %s. Campaign %s does not exist.", account.id, campaignId)
		return false, {result = "CAMPAIGN_DOES_NOT_EXIST", reason = string.format("Campaign %s does not exist", campaignId)}
	end

	local responseAccount, responseCampaign = GetModel("Campaigns").FindAccountByUID( account, ksAgent )

	if responseAccount then
		if responseCampaign == campaignId then
			LogError("[AGENT:CAMPAIGNS:SaveNewAccount] Unable to add account %s on campaign %s. Account already exists on current campaign.", account.id, campaignId)
			return false, {result = "ALREADY_EXISTS_IN_CURRENT_CAMPAIGN", reason = "account already exists in current campaign"}
		end
		local currentId,  currentVCC  = helpers.SeparateVCC(campaignId)
		local responseId, responseVCC = helpers.SeparateVCC(responseCampaign)
		if currentVCC ~= responseVCC then
			LogError("[AGENT:CAMPAIGNS:SaveNewAccount] Unable to add account %s on campaign %s. Account already exists in another vcc %s.", account.id, campaignId, responseVCC)
			return false, {result = "ALREADY_EXISTS_IN_ANOTHER_VCC", reason = "account already exists in another vcc"}
		else
			LogError("[AGENT:CAMPAIGNS:SaveNewAccount] Unable to add account %s on campaign %s. Account already exists in %s campaign.", account.id, campaignId, responseId)
			return false, {result = "ALREADY_EXISTS_IN_ANOTHER_CAMPAIGN", reason = "account already exists in another campaign", campaign = responseId}
		end
	end

	local existingAccountInVCC = m_adm_models.VCCs.GetAccounts( vcc, account.proxyType, account.uid, ksAgent )
	if existingAccountInVCC then
		local err = {
			result = "ACCOUNT_ALREADY_EXISTS_IN_CURRENT_VCC",
			reason = string.format( "Unable to add account. Account id: '%s', type: '%s', Already exists in current VCC: %s", account.uid, account.proxyType, vcc )
		}
		LogError("[AGENT:CAMPAIGNS:SaveNewAccount] %s", err.reason)
		return false, err
	end

	-- extender con la validaciones de cada tipo de interaccion ( si es necesario )
	local specificsAccountsTypeValidation = {
		CALL   	   = CallAccountValidation,
		FACEBOOK   = FacebookAccountValidation,
		INSTAGRAM  = InstagramAccountValidation,
		WHATSAPP   = WhatsappAccountValidation,
		TWITTER    = TwitterAccountValidation,
		YOUTUBE    = YoutubeAccountValidation,
		APPSTORE   = AppstoreAccountValidation,
		GOOGLEPLAY = GoogleplayAccountValidation,
		RCS        = RcsAccountValidation,
		MAIL       = MailAccountValidation,
		WEBCONTACT = WebcontactAccountValidation
	}
	local specificProxyValidation = specificsAccountsTypeValidation[ account.proxyType ]
	if type( specificProxyValidation ) == "function" then
		local success, err = specificProxyValidation( account, campaignId, ksAgent )
		if not success then
			if type(err) == "string" then
				err = { result = err }
			end
			LogError("[AGENT:CAMPAIGNS:SaveNewAccount] %s", err.reason or err.result)
			return false, err
		end
	end

	if not GetModel("Campaigns").ReserveAccount(account) then
		--GetModel("Campaigns").UnreserveAccount(account)
		LogWarning("[AGENT:CAMPAIGNS:SaveNewAccount] User %s does not own the lock on account id %s, will not save", inconcert.authUser.userName, account.id)
		return false, {result = "ACCOUNT_IDENTIFIER_IN_USE", reason = "account identifier in use"}
	end

	SetAccountDefaultValues(account, campaign, true, ksAgent)

	local success, errInfo = GetModel("Campaigns").SaveCampaignAccount(campaignId, account, true, ksAgent, allegroOtherInfo)
	GetModel("Campaigns").UnreserveAccount(account)
	if not success then
		LogError("[AGENT:CAMPAIGNS:SaveNewAccount] Unable to add account %s on campaign %s. Failed add on database.", account.id, campaignId)
		local err = {result = "FAILED_SAVING_ON_DATABASE", reason = "failed saving on database."}
		if errInfo and errInfo.result then
			err = errInfo
		end
		return false, err
	end

	GetModel("Campaigns").PrepareAccountForRead(account)

	--notifications
	notificationsEngine.SendNotification("AccountCreated", 0, { campaignId = utils.GetId(campaignId), account = account }, ksAgent)

	return true, account
end

local function UpdateAccount (campaignId, account, allegroOtherInfo, ksAgent)
	local campaign = GetModel("Campaigns").Get(campaignId, ksAgent)
	if not campaign then
		LogError("[AGENT.CONTROLLER.UpdateAccount] Unable to update account %s. Campaign %s does not exist.", account.id, campaignId)
		return false, {result = "CAMPAIGN_DOES_NOT_EXIST", reason = string.format("Campaign %s does not exist", campaignId)}
	end

	-- extender con la validaciones de cada tipo de interaccion ( si es necesario )
	local specificsAccountsTypeValidation = {
		CALL 	   = CallAccountValidation,
		FACEBOOK   = FacebookAccountValidation,
		INSTAGRAM  = InstagramAccountValidation,
		WHATSAPP   = WhatsappAccountValidation,
		TWITTER    = TwitterAccountValidation,
		YOUTUBE    = YoutubeAccountValidation,
		APPSTORE   = AppstoreAccountValidation,
		GOOGLEPLAY = GoogleplayAccountValidation,
		RCS        = RcsAccountValidation,
		MAIL       = MailAccountValidation,
		WEBCONTACT = WebcontactAccountValidation
	}
	local specificProxyValidation = specificsAccountsTypeValidation[ account.proxyType ]
	if type( specificProxyValidation ) == "function" then
		local originalAccount = GetModel("Campaigns").GetCampaignAccount(campaign, account.id)
		local success, err = specificProxyValidation( account, campaignId, ksAgent, originalAccount )
		if not success then
			LogError("[AGENT.CONTROLLER.UpdateAccount] %s", err.reason)
			return false, err
		end
	end

	SetAccountDefaultValues(account, campaign, false, ksAgent)

	local success, errInfo = GetModel("Campaigns").SaveCampaignAccount(campaignId, account, false, ksAgent, allegroOtherInfo)
	if not success then
		LogError("[AGENT.CONTROLLER.UpdateAccount] Unable to update account %s on campaign %s. Failed updating on database. Err: %s", account.id, campaignId, errInfo)
		local err = {result = "FAILED_SAVING_ON_DATABASE", reason = "failed saving on database."}
		if errInfo and errInfo.result then
			err = errInfo
		end
		return false, err
	end

	GetModel("Campaigns").PrepareAccountForRead(account)

	--notifications
	notificationsEngine.SendNotification("AccountUpdated", 0, { campaignId = utils.GetId(campaignId), account = account }, ksAgent)
	
	return true, account
end

local function IsAccountTypeLicensed (accountType, vcc, ksAgent)
	if m_helpers.utils.HasLicenseSystem() then
		local options = utils.Get_m_helpers_agentservice_options()
		local channelLicenses = m_helpers.agentservice.GetChannelLicenses(options, vcc)
		return channelLicenses[accountType]
	end
	local vcc = utils.GetAllColumns {
		keyspace = ksAgent or "Agent",
		columnFamily = "VCCs",
		key = vcc,
	} or {}
	if vcc[accountType.."_licensed"] == "0" then
		return false
	else --if it's "1" or nil then it's true
		return true
	end	
end

function AddCampaignAccount (account, campaignId, allegroOtherInfo, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter

	if not IsAccountTypeLicensed(account.proxyType, vcc, ksAgent) then
		local err = { err_type = 400, result = "NOT_LICENSES_IN_VCC", reason = "Not licenses in vcc: " .. json.encode(account)}
		LogError("[AGENT.CONTROLLER.AddCampaignAccount] Unable to add account. Account Type '%s' is not licensed in this vcc (%s). Error: %s", account.proxyType, vcc, json.encode(err))
		return false, err
	end

	local saved, account = SaveNewAccount(campaignId, account, allegroOtherInfo, ksAgent)
	if not saved then
		local err = account or {result = "UNABLE_TO_SAVE_ACCOUNT", reason = "Unable to add account"}
		err.err_type = 500
		LogError("[AGENT.CONTROLLER.AddCampaignAccount] Unable to add account. Account Type '%s' vcc (%s). Error: %s", account.proxyType, vcc, json.encode(err))
		return false, err
	end

	return true, account
end

function UpdateCampaignAccount (account, campaignId, allegroOtherInfo, ksAgent)
	local vcc = inconcert.authUser.virtualContactCenter

	if not IsAccountTypeLicensed(account.proxyType, vcc, ksAgent) then
		local err = {err_type = 400, result = "NOT_LICENSES_IN_VCC", reason = "Not licenses in vcc: " .. account.proxyType}
		LogError("[AGENT.CONTROLLER.UpdateCampaignAccount] Unable to update account. Account Type '%s' is not licensed in this vcc (%s). Error: %s", account.proxyType, vcc, json.encode(err))
		return false, err
	end

	local saved, updated_account = UpdateAccount(campaignId, account, allegroOtherInfo, ksAgent)
	if not saved then
		local err = updated_account or {result = "UNABLE_TO_SAVE_ACCOUNT", reason = "Unable to update account"}
		err.err_type = 500
		LogError("[AGENT.CONTROLLER.UpdateCampaignAccount] Unable to update account. Account Type '%s' vcc (%s). Error: %s", account.proxyType, vcc, json.encode(err))
		return false, err
	end

	return true, updated_account
end

--[[
	campaign_id               (opcional) Si se quiere o no unca campana especifica. Si es nil, se busca en todas las campanas del usuario logueado.
	can_add_member_column_fn  (opcional) Funcion que se pasa al model de "User" en la funcion "Get" para determinar que columnas se incluiran o no en la entidad user que nos retorna la funcion.
	rows_per_page                  (opcional) La cantidad maxima de rows (usuarios) que retornara la funcion.
	filter_fn                 (opcional) Funcion que realiza ciertas validaciones para saber si el usuario aplica o no con el filtro de quien me ha invocado.
--]]
local m_cached_members = {}
-- callback que lo invoca el modelo de usuario para determinar las columnas que se incluiran en el member
local CanAddMemberColumn = function( column_name )
	local allowed_columns = { Id = true, Name = true, LastName = true, FirstSurname = true, LastSurname = true, Status = true, UserGroup = true }
	return allowed_columns[column_name]
end

local function BuildTeamMembersHash ( params )

	local vcc = inconcert.authUser.virtualContactCenter
	local user_name = inconcert.authUser.userName

	local can_add_member_column_fn = params.can_add_member_column_fn
	local campaignId = params.campaign_id
	local filter_fn  = params.filter_fn

	-- obtengo el keyspace para utilizar el mismo en todas las conexiones a cassandra de esta funcion.
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	filter_fn = filter_fn or function() return true end
	-- Funcion que valida si el usuario puede ser agregado y devuelto en la lista de usuarios.
	-- Por defecto se valida que sean unicamente usuarios activos "Enabled" distinto a "0", luego invoca a la funcion recibida por parametro para que esta 
	--  determine si se debe incluir o no.
	-- 
	local can_add = function( member )
		if not member or member.Enabled == "0" then
			return false
		end
		return filter_fn( member, keyspace )
	end

	local myselfAtVCC = user_name .. "@" .. vcc
	local count_added_members = 0
	local teamMembers = {}
	-- obtengo las campanas a las cuales pertenece el usuario
	local userCampaigns = GetModel("Campaigns").GetUserCampaigns(user_name, vcc, keyspace)

	for campaignAtVCC, role in pairs(userCampaigns) do

		if role ~= "invited" and (campaignId == campaignAtVCC or campaignId == nil)  then

			-- obtengo los members de la campana
			local campaignMembers = GetModel("Campaigns").GetCampaignMembers(campaignAtVCC, keyspace) or {}
			for userAtVCC in pairs(campaignMembers) do
				--the querying user will belong to all the campaigns, should not be returned
				if myselfAtVCC ~= userAtVCC then

					local member
					if not m_cached_members[userAtVCC] then
						m_cached_members[userAtVCC] = models.Users.Get(userAtVCC, can_add_member_column_fn, keyspace)
					end
					member = m_cached_members[userAtVCC]
					table.insert(member.campaigns or {}, campaignAtVCC)

					if can_add( member, userAtVCC, campaignAtVCC ) then

						member.Avatar = nil
						teamMembers[userAtVCC] = member
						count_added_members = count_added_members + 1
					end

				end
			end
		end
	end

	keyspace:Close()
	return teamMembers, count_added_members 
end

--[[
	- Esta funcion retorna todas las columnas de la CF User, y no realiza un filtro por usuario, sino que los retorna a todos.
]]--
local function GetTeamMembers (web)

	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local response, request = pcall(json.decode, web.input["post_data"])
	if not response or not request then
		web.status = 400
		return "INVALID_REQUEST"
	end
	if not request.campaignId then
		LogInfo("[GetTeamMembers] Retrieving team members for all campaigns for user %s@%s.", name, vcc)
	else
		LogInfo("[GetTeamMembers] Retrieving team members of campaign '%s' for user %s@%s.", tostring(request.campaignId), name, vcc)
	end

	local teamMembers, count_members = BuildTeamMembersHash( { campaign_id = request.campaignId, can_add_member_column_fn = CanAddMemberColumn } )
	local encoded_members = json.encode(teamMembers)
	LogInfo("[GetTeamMembers] Recovered %d team members of campaign '%s'.", count_members, tostring(request.campaignId))

	return encoded_members
end

--[[
	can_add_member_column_fn  (opcional) Funcion que se pasa al model de "User" en la funcion "Get" para determinar que columnas que se incluiran o no en la entidad user que nos retorna la funcion.
	rows_per_page                  (opcional) La cantidad maxima de rows (usuarios) que retornara la funcion.
	filter_fn                 (opcional) Funcion que realiza ciertas validaciones para saber si el usuario aplica o no con el filtro de quien me ha invocado.
--]]

local function BuildPaginatedTeamMembers ( params )

	local vcc = inconcert.authUser.virtualContactCenter
	local user_name = inconcert.authUser.userName

	local can_add_member_column_fn = params.can_add_member_column_fn
	local filter_fn                = params.filter_fn
	local rows_per_page            = params.rows_per_page
	local page_number              = params.page_number or 1

	local start_page_member, end_page_member
	if page_number == 1 then
		end_page_member = rows_per_page
		start_page_member = 1
	else
		-- Si no es la primer pagina:
		--
		-- La ultima fila, es la cantidad filas por pagina multiplicado por el numero de pagina que se ha solicitado
		end_page_member = rows_per_page * page_number
		-- La primer fila, se obtiene restando 1 a la pagina solicitada y multiplicandolo por la cantidad de filas por pagina,
		--  para asi obtener el numero de la ultima fila de la pagina anterior, luego le sumamos 1 para poder saber la primer fila de la pagina que se ha
		--  solicitado.
		start_page_member = (rows_per_page * (page_number - 1)) + 1
	end

	--LogDebug("[BuildPaginatedTeamMembers] Retrieving page: '%d', start page: '%d', end page: '%d'", page_number, start_page_member, end_page_member)

	-- obtengo el keyspace para utilizar el mismo en todas las conexiones a cassandra de esta funcion.
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")

	filter_fn = filter_fn or function() return true end
	-- Funcion que valida si el usuario puede ser agregado y devuelto en la lista de usuarios.
	-- Por defecto se valida que sean unicamente usuarios activos "Enabled" distinto a "0", luego invoca a la funcion recibida por parametro para que esta 
	--  determine si se debe incluir o no.
	--
	local count_matched_members = 0
	local members_previous_pages = {}
	local page_members = {}

	local can_add = function( member, full_user_id, camp_id )
		if not member then
			--LogDebug("[BuildPaginatedTeamMembers] can_add 1")
			return false
		end
		if page_members[full_user_id] then
			--LogDebug("[BuildPaginatedTeamMembers] can_add 2")
			return false -- ya existe en la misma pagina
		end
		if not member or member.Status == "0" then
			--LogDebug("[BuildPaginatedTeamMembers] can_add 3")
			return false
		end
		if not filter_fn( member, keyspace ) then
			--LogDebug("[BuildPaginatedTeamMembers] can_add 4")
			return false
		end

		if members_previous_pages[full_user_id] then
			--LogDebug("[BuildPaginatedTeamMembers] can_add 5")
			return false -- ya existe en paginas anteriores
		end
		count_matched_members = count_matched_members + 1
		members_previous_pages[full_user_id] = { campaign = camp_id }

		local member_in_page = (start_page_member <= count_matched_members and end_page_member >= count_matched_members)
		--LogDebug("[BuildPaginatedTeamMembers]  --> member_in_page: %s, count_matched_members: %d, start_page_member: %s, end_page_member: %s", member_in_page, count_matched_members, start_page_member, end_page_member)
		if member_in_page then
			--LogDebug("[BuildPaginatedTeamMembers] (IS OK) can_add 6 ")
			return true
		end
		--LogDebug("[BuildPaginatedTeamMembers] can_add 7")
		return false
	end

	local count_added_members_in_page = 0
	local return_values = function()
		--LogDebug("[BuildPaginatedTeamMembers] Retrieving count_added_members_in_page: %d", count_added_members_in_page)
		keyspace:Close()
		return {
			members = page_members,
			count_users_in_pages = count_added_members_in_page
		}
	end

	local myselfAtVCC = user_name .. "@" .. vcc
	-- obtengo las campanas a las cuales pertenece el usuario
	local userCampaigns = GetModel("Campaigns").GetUserCampaigns(user_name, vcc, keyspace)

	local continue = true
	for campaignAtVCC, role in pairs(userCampaigns) do

		if role ~= "invited" then

			-- obtengo los members de la campana
			local campaignMembers = GetModel("Campaigns").GetCampaignMembers(campaignAtVCC, keyspace) or {}
			for userAtVCC in pairs(campaignMembers) do
				--the querying user will belong to all the campaigns, should not be returned
				if myselfAtVCC ~= userAtVCC then
					--LogDebug("[BuildPaginatedTeamMembers] %s -> %s", campaignAtVCC, userAtVCC)

					local member = agentModel.GetRedisClient():hget("CachedTeamMembers", userAtVCC)
					if not member then
						member = models.Users.Get(userAtVCC, can_add_member_column_fn, keyspace)
						if member then
							local createdKey = agentModel.GetRedisClient():exists("CachedTeamMembers")
							agentModel.GetRedisClient():hset("CachedTeamMembers", userAtVCC, json.encode(member))
							--Establecer caducidad solo cuando no existe la key
							if not createdKey then
								agentModel.GetRedisClient():expire("CachedTeamMembers", 60 * 60 * 24) -- 1 day
							end
						end
					else
						member = json.decode(member)
					end
					
					table.insert(member.campaigns or {}, campaignAtVCC)

					if can_add( member, userAtVCC, campaignAtVCC ) then

						member.Avatar = nil
						page_members[userAtVCC] = member
						count_added_members_in_page = count_added_members_in_page + 1
						continue = count_added_members_in_page < rows_per_page

						--LogDebug("[BuildPaginatedTeamMembers] Adding member: %s, count_added_members_in_page: %d, count_matched_members: %d, %s", userAtVCC, count_added_members_in_page, count_matched_members, continue)
					end

					if not continue then
						return return_values()
					end
				end
			end

		end
	end

	return return_values()
end

--[[
	- Esta funcion retorna las columnas que se hayan especificado de la CF User, y realiza un filtro por usuario.
	- Retorna una maximo de 40 usuarios o lo que se especifique en el request.
]]--
local function GetTeamMembersWithState (web)

	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	local response, request = pcall(json.decode, web.input["post_data"])

	--LogDebug("GetTeamMembersWithState post_data: %s", web.input["post_data"])

	if not response or not request then
		web.status = 400
		return "INVALID_REQUEST"
	end
	local filter_text = utf8.lower(request.filter and request.filter.query or "")
	local page_number = tonumber(request.filter.page_number or 1)

	--LogInfo("[GetTeamMembersWithState] Retrieving team members for all campaigns for user %s@%s , By filter: '%s', page: %d", name, vcc, filter_text, page_number)

	--funcion que valida si el texto que se ha recibido client side se encuentra en la palabra
	local pass_filter_text = function( text )
		return string.find(utf8.lower(text or ""), filter_text)
	end
	-- callback que invoca la funcion que genera el hash de members, para determinar si el usuario debe ser agregado o no en la lista de members.
	local filter_member_fn = function( member )

		if filter_text:len() == 0 then
			return true
		end
		-- busca en el id
		if pass_filter_text(member.Id) then
			return true
		end
		-- busca en el primer nombre
		if pass_filter_text(member.Name) then
			return true
		end
		-- busca en el segundo nombre
		if pass_filter_text(member.LastName) then
			return true
		end
		-- busca en el primer apellido
		if pass_filter_text(member.FirstSurname) then
			return true
		end
		-- busca en el segundo apellido
		if pass_filter_text(member.LastSurname) then
			return true
		end
		-- busca en el primer y segundo nomobre
		local name_lastname = (member.Name or "") .. " " .. (member.LastName or "")
		if pass_filter_text(name_lastname) then
			return true
		end
		-- busca en el primer nombre y primer apellido
		local name_firstsurname = (member.Name or "") .. " " .. (member.FirstSurname or "")
		if pass_filter_text(name_firstsurname) then
			return true
		end
		-- busca en el primer y segundo nombre y en el primer apellido
		local name_lastname_firstsurname = name_lastname .. " " .. (member.FirstSurname or "")
		if pass_filter_text(name_lastname_firstsurname) then
			return true
		end
		-- busca en el nombre completo (primer y segundo nombre y primer y segundo apellido)
		local full_name = name_lastname_firstsurname .. " " .. (member.LastSurname or "")
		if pass_filter_text(full_name) then
			return true
		end

		return false
	end

	-- valida si el usuario tiene permisos sobre las apps "interaction" o "supervisor"
	local applicationsActions = permissions.GetApplicationsActions(keyspace, vcc)
	local has_permisson_to_apps = function( member, keyspace )
		local user_at_vcc = utils.AppendVcc(member.Id, vcc)
		local hasPermissionToAppInteraction = permissions.HasPermissionToApplication(
			{"interaction", "supervisor"},
			user_at_vcc,
			member.UserGroup,
			nil, -- campaign id
			keyspace,
			{applicationsActions = applicationsActions}
		)
		return hasPermissionToAppInteraction
	end

	local rows_per_page = tonumber(request.filter.rows_per_page) or 40 -- cantidad de members que retorna la funcion al cliente

	local response = BuildPaginatedTeamMembers(
	{
		page_number              = page_number,
		rows_per_page            = rows_per_page,
		can_add_member_column_fn = CanAddMemberColumn,
		filter_fn                = function( member, keyspace ) 
			if filter_member_fn( member ) then
			-- si pasas el filtro que se ha posteado, valida permisos sobre las apps
				return has_permisson_to_apps( member, keyspace )
			end
			return false
		end
	})

	--query the agentservice to get each member status
	local statuses = agentservice.GetAgentsStatuses(response.members, vcc) or {}
	for userAtVCC, status in pairs(statuses) do
		response.members[userAtVCC].status = status
	end

	--Add to each member if has or not access to supervisor application
	for memberId, member in pairs(response.members) do
		response.members[memberId].isSupervisor = permissions.HasPermissionToApplication(
			{"supervisor"},
			memberId,
			member.UserGroup,
			nil, -- campaign id
			keyspace
		)
	end

	LogInfo("[GetTeamMembersWithState] Recovered %d team members on page %d.", response.count_users_in_pages, page_number)

	return json.encode(response)
end

function GetCampaignScheduler (campaignId, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local campaignName, campaignVCC = string.match(campaignId, "([^@]+)@([^@]+)")
	campaignId = utf8.lower((campaignName or campaignId) .. "@" .. vcc)
	local scheduler = GetModel("Campaigns").GetCampaignScheduler(campaignId, keyspace)
	--LogInfo("[AGENT:GetCampaignScheduler] Retrieved Scheduler for campaign '%s'", campaignId)
	--LogDebug("[AGENT:GetCampaignScheduler] Retrieved Scheduler for campaign '%s' -> %s", campaignId, json.encode(scheduler))
	return scheduler 
end

local function GetCampaignInfo (web, campaignIdAtVCC)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName

	-- Avoid data manglers spoofing vccs
	local campaignId = helpers.StripVCC(campaignIdAtVCC, vcc)
	campaignIdAtVCC = campaignId.."@"..vcc

	LogInfo("[AGENT.CAMPAIGNS:GetCampaignInfo] User %s on vcc %s retrieving info of campaign %s", name, vcc, campaignIdAtVCC)

	-- obtengo el keyspace para utilizar el mismo en todas las conexiones a cassandra de esta funcion.
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	local result = GetModel("Campaigns").Get(campaignIdAtVCC, keyspace)
	if not result then
		keyspace:Close()
		LogError("[AGENT.CAMPAIGNS:GetCampaignInfo] Tried to get info from unexistent campaign %s", campaignIdAtVCC)
		web.status = 404
		return json.encode{ message = "Unexistent campaign" }
	end
	result.role = GetModel("Campaigns").IsMember(campaignIdAtVCC, name.."@"..vcc, keyspace)
	result.status = agentservice.GetStatusOnCampaign(name, campaignId, vcc)

	keyspace:Close()
	return json.encode(result)
end

local function LoadVccCampaigns (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local name = inconcert.authUser.userName
	LogInfo("[AGENT.CAMPAIGNS:LoadVccCampaigns] User %s on vcc %s retrieving all campaigns.", name, vcc)
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local result = GetModel("Campaigns").GetVCCCampaignsAndAccounts(vcc, keyspace)
	keyspace:Close()
	return json.encode(result)
end

local function GetTemplateMessageOptions(web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local userName = inconcert.authUser.userName
	local response, request = pcall(json.decode, web.input["post_data"])
	if not response or not request then
		web.status = 400
		return "INVALID_REQUEST"
	end

	LogInfo("[GetTemplateMessageOptions] Retrieving template message options for user %s@%s , By type: %s campaign: %s account: %s page: %s",
		userName, vcc, request.proxyType, request.campaignAtVCC, request.account, request.page or "")

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:GetTemplateMessageOptions({
		proxyType = request.proxyType,
		campaignAtVCC = request.campaignAtVCC,
		account = request.account,
		page = request.page,
		cache = request.cache
	})
	result.options = json.decode(result.options)
	web.status = http_mapping.getHTTPCode(result.result) or 500
	
	return json.encode(result)
end

local function WhatsappListGroups(web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local userName = inconcert.authUser.userName
	local response, request = pcall(json.decode, web.input["post_data"])
	if not response or not request then
		web.status = 400
		return "INVALID_REQUEST"
	end

	LogInfo("[WhatsappListGroups] Retrieving groups for user %s@%s , By campaign: %s account: %s",
		userName, vcc, request.campaignAtVCC, request.account)

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:WhatsappListGroups({
		campaignAtVCC = request.campaignAtVCC,
		account = request.account,
		cache = true --agent get groups using cache
	})
	web.status = http_mapping.getHTTPCode(result.result) or 500
	return json.encode(result)
end

function MapUris(module, htmlRender, prefix)
	local LogRequest = require("apps.agent.requestFilters").LogRequest

	-- esta URL es invocada desde el agent_poller
	module:dispatch_post( LogRequest .. GetCampaignInfo          , prefix .. "/get_campaign_info/(.*)"              )

	--this functions are public for all
	module:dispatch_post( LogRequest .. GetTeamMembers           , prefix .. "/get_members/([^.]*)"                 )
	module:dispatch_post( LogRequest .. GetTeamMembersWithState  , prefix .. "/get_members_with_state/([^.]*)"      )
	module:dispatch_post( LogRequest .. GetTemplateMessageOptions, prefix .. "/get_template_message_options/([^.]*)")
	module:dispatch_get(  LogRequest .. GetUserCampaigns         , prefix .. "/get_campaigns"                       )
	module:dispatch_get(  LogRequest .. GetUserOutgoingCampaigns , prefix .. "/get_outgoing_campaigns/(.*)"         )
	module:dispatch_get(  LogRequest .. GetCampaignAvatar        , prefix .. "/get_avatar/(.*)"                     )
	module:dispatch_get(  LogRequest .. GetCampaignDispositions  , prefix .. "/get_dispositions/(.*)"               )
	module:dispatch_get(  LogRequest .. LoadVccCampaigns         , prefix .. "/load_vcc_campaigns/"                 )

	--whatsapp groups
	module:dispatch_post( LogRequest .. WhatsappListGroups       , prefix .. "/whatsapp/list_groups/(.*)"           )

	m_htmlRender = htmlRender
end
