module(..., package.seeall, inconcert.orbit.application.Decorate)

local Ticketek = require "ticketek"
local utf8 = require "icu.utf8"
local utils = require "shared.apps.utils"
local ir = require "lua_iragentlink"
local models = require("apps.agent.model"):GetModels()

local function ReturnErr( err )
	if not err or not err.msg then
		return err
	end
	LogError("code: '%s', reason: '%s'", err.code, err.msg)
	return err
end

local function HasTicketConfiguredInAccount( interaction_id )

	local interaction = models.Interactions.GetInteraction( interaction_id )
	if utils.TableIsEmpty(interaction) then
	-- error
		return false, { code = "INTERACTION_DOES_NOT_EXIST", reason = string.format("Interaction '%s' does not exist", interaction_id) }
	end

	local campaign = models.Campaigns.Get( interaction.campaign )
	if not campaign then
	-- error
		return false, { code = "CAMPAIGN_DOES_NOT_EXIST", reason = string.format("Campaign '%s' does not exist", interaction.campaign) }
	end

	local account_id = interaction.account
	local accounts = campaign.accounts
	if not accounts or not account_id then
	-- error
		return false, { code = "ACCOUNT_DOES_NOT_EXIST", reason = string.format("Account '%s' does not exist", account_id) }
	end

	local proxy_type = interaction.type

	for _, account in ipairs( accounts ) do
		if account.id == account_id and utils.TypeBelongsToProxy(proxy_type, account.proxyType) then
			local ticket = account.TicketEmitter
			if ticket then
				if type(ticket) == "string" then
					ticket = json.decode( ticket )
				end
				if not utils.TableIsEmpty( ticket ) then
				-- SUCCESS
					return true
				else
				-- error
					break
				end
			else
			-- error
				break
			end
		end
	end

	-- error
	return false, { code = "ACCOUNT_DOES_NOT_HAVE_TICKET_CONFIG", reason = string.format("Account '%s' don't have ticket configured.", account_id) }
end

---
-- emits and assigns a new ticket to the given interaction
--
local function Emit (web, interaction_id)
	local vcc = inconcert.authUser.virtualContactCenter
	local user_name = inconcert.authUser.userName

	LogInfo("User '%s@%s' requested to emit a ticket for interaction '%s'", user_name, vcc, interaction_id)

	local has_ticket_emitter, err = HasTicketConfiguredInAccount( interaction_id )
	if not has_ticket_emitter then
		return ReturnErr( { status = false, code = err.code, msg = err.reason } )
	end

	local irLink = ir.IRAgentLink( inconcert.environment.apps.agent.cluster )

	local response = irLink:SendCommand({
		interactionID = interaction_id,
		RID           = web.vars.HTTP_X_ICC_RID,
		proxyType     = "AGENT",
		agentAtVCC    = user_name .. "@" .. vcc,
		command       = "AssignTicketToInteraction"
	})

	if response.result == "OK" then
		-- if no ticket emitter is configured, ticket is nil
		return { status = true, ticket = response.ticket }
	end
	-- bugger off
	return ReturnErr ( { status = false, code = response.result, msg = response.reason } )
end

local function check_allowed_campaign (user_campaigns, campaign_id)
	for campaign in pairs(user_campaigns) do
		if campaign_id == campaign then
			return true
		end
	end
	return false
end

local function Search (web)

	local user_name = inconcert.authUser.userName
	local vcc = inconcert.authUser.virtualContactCenter
	local filter = web.request.filter
	local current_interaction_id = web.request.interaction_id or ""
	local campaign_id = filter.campaignId -- optional filter

	if campaign_id then
		campaign_id = utils.AppendVcc( utils.GetId( campaign_id ), vcc )
	end

	LogInfo("[Tickets.Search] User '%s@%s' searching for filter ticket '%s', current interaction: '%s'", user_name, vcc, json.encode(filter), current_interaction_id)

	local molule_interactions = require "apps.agent.controllers.interactions"
	local rules = {}
	table.insert(rules, {id = "ticket", label = "Ticket", value = filter.tickets, operator = "equal"})
	-- El query como se necesita para Elastic Search.
	local query = {
		last = "",
		last_by_page = {
			["1"] = 15
		},
		rows = 25,
		page = 1,
		output = "interactions",
		paginating = false,
		sort = {
			{
				agent = {
				order = "asc"
				}
			}
		},
		query = {
			condition = "AND",
			rules = rules
		},		
		mustValidateSearchPermission = true,
		exact = true
	}

	local status, response = molule_interactions.SearchInteractionsHelper(query)
	web.status = status

	if status ~= 200 then
		return ReturnErr( { status = false, code = response.result, msg = response.reason } )
	end

	return { status = true, interactions = response.interactions }
end

function MapUris(module, htmlRender, prefix)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	module:dispatch_post( LogRequest .. utils.JsonReply .. Emit, prefix .. "/emit/(.+)" )
	module:dispatch_post( LogRequest .. utils.JsonValidatorReply .. Search, prefix .. "/search" )
end
