require("orbit")
require "inconcert.translators" 
require "inconcert.orbit.application"
local jsonEncoder = require "LuaOrm.Serializers.Json"
local LuaConfigurationControl = require "LuaConfigurationControl"
local cassandra = require "lua_cassandra"
local Ticketek = require "ticketek"
local ir = require "lua_iragentlink"
local authFilter = (require "authfilter")("agent")
local luakeystoreclient = require "luakeystoreclient"
local utils = require "shared.apps.utils"
local allegro = require "apps.allegro.proxy"
local permissions = require "shared.permissions.permissions"
local model_perm = require "shared.permissions.models.permissions"
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")
local models = require("apps.agent.model"):GetModels()
local audits = require "audits"
local m_helpers = require "helpers"
local md5 = require "md5"
local jwt = require "authfilter.jwt"
module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)

_M.model = nil

local ksCore = luakeystoreclient.KSCore(inconcert.environment.apps.agent.privateKey, inconcert.environment.apps.agent.keystorePublicKey, "")
local model = relativeRequire("model", ...)
local agent = relativeRequire("agent", ...)
local agentNotifications = relativeRequire("agentNotifications", ...)
local configControl
local modelBuilt = false

---
-- Some components expect a lualiblogger compatible logger instance.
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function make_compatible_logger ()
	return {
		debug = LogDebug,
		info  = LogInfo,
		warn  = LogWarning,
		error = LogError,
		fatal = LogFatal
	}
end

---
-- Returns a string with a suitable value to be used as version in files which needs it.
-- For instance, the offline.appcache, or version fields in querystrings.
--
local function get_version_string ()
	local hourAlignedTime = os.time()
	hourAlignedTime = hourAlignedTime - (hourAlignedTime % 3600)
	local version = os.getenv("ICC_VERSION") or tostring(hourAlignedTime)
	return version
end

model.SetKSCore (ksCore)

-- App status control table for the appReadyFilter
-- It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		--LogDebug("[AGENT:CheckStatus] 1--------> Start checking status for app 'AGENT'.") -- TODO: comentar esa linea, esta solo para diagnostico
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"keystore", "irLinks", "agentservice", "repository", "redis", "cassandra", "elasticsearch", "agentredis", "facebook", "instagram", "youtube", "twitter"})

		app_status.current = newConfig

		--Configure facebook
		inconcert.environment.apps.agent.facebook = inconcert.environment.apps.agent.facebook or {}
		inconcert.environment.apps.agent.facebook.appId = newConfig["facebook"]["appId"]

		--Configure instagram
		inconcert.environment.apps.agent.instagram = inconcert.environment.apps.agent.instagram or {}
		inconcert.environment.apps.agent.instagram.appId = newConfig["instagram"]["appId"]

		--Configure youtube
		inconcert.environment.apps.agent.youtube = inconcert.environment.apps.agent.youtube or {}
		inconcert.environment.apps.agent.youtube.appId = newConfig["youtube"]["appId"]

		--Configure twitter
		inconcert.environment.apps.agent.twitter = inconcert.environment.apps.agent.twitter or {}

		--Configure Allegro
		inconcert.environment.apps.agent.allegroproxy = inconcert.environment.apps.agent.allegroproxy or {}
		for key, value, state in newConfig["allegroproxy"]:childs() do
			if state ~= "deleted" then
				table.insert(inconcert.environment.apps.agent.allegroproxy, { ip = value.ip, port = value.port })
				LogInfo("[AGENT.app_status.CheckStatus] Connecting to allegro proxy on %s:%s", value.ip, tostring(value.port))
			end
		end
		if #inconcert.environment.apps.agent.allegroproxy == 0 then
			LogError("[AGENT.app_status.CheckStatus] Dependency 'allegroproxy' is not met")
			app_status.actualStatus = "invalid"
			error("Unmet dependency allegroproxy")
		end

		--Configure notificationsservice
		inconcert.environment.apps.agent.notificationsservice = inconcert.environment.apps.agent.notificationsservice or {}
		for key, value, state in newConfig["notificationsservice"]:childs() do
			if state ~= "deleted" then
				table.insert(inconcert.environment.apps.agent.notificationsservice, { ip = value.ip, port = value.port })
				LogInfo("[AGENT.app_status.CheckStatus] Connecting to notifications service on %s:%s", value.ip, tostring(value.port))
			end
		end
		if #inconcert.environment.apps.agent.notificationsservice == 0 then
			LogError("[AGENT.app_status.CheckStatus] Dependency 'notificationsservice' is not met")
			app_status.actualStatus = "invalid"
			error("Unmet dependency notificationsservice")
		end

		--Configure IR links
		if newConfig:DiffState("irLinks") == "updated" or newConfig:DiffState("irLinks") == "added" then
			local links = {}
			for key, value, state in newConfig["irLinks"]:childs() do
				if state ~= "deleted" then
					table.insert (links, value)
				end
			end
			LogInfo("[AGENT.app_status.CheckStatus] Adding new agent links for cluster %s", inconcert.environment.apps.agent.cluster)
			local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
			irLink:DiscoverIRs(links)
		end

		ksCore:SetServer(newConfig["keystore"][1]["ip"], newConfig["keystore"][1]["port"])

		--Configure Agent Service
		local agentServiceDiff = newConfig:DiffState("agentservice")

		inconcert.environment.apps.agent.agentservice = inconcert.environment.apps.agent.agentservice or {}
		inconcert.environment.apps.agent.agentservice.host = newConfig["agentservice"]["host"]
		inconcert.environment.apps.agent.agentservice.port = tonumber(newConfig["agentservice"]["port"])

		--Configure outboundengineservice
		inconcert.environment.apps.agent.outboundengineservice = {}
		local outboundengineservice = inconcert.environment.apps.agent.outboundengineservice
		for key, value, state in newConfig["outboundengineservice"]:childs() do
			if state ~= "deleted" then
				outboundengineservice[#outboundengineservice + 1] = {host = value.ip, port = value.port}
			end
		end

		--Configure Agent Redis
		inconcert.environment.apps.agent.agentredis = inconcert.environment.apps.agent.agentredis or {}
		inconcert.environment.apps.agent.agentredis.host = newConfig["agentredis"]["host"]
		inconcert.environment.apps.agent.agentredis.port = tonumber(newConfig["agentredis"]["port"])
		inconcert.environment.apps.agent.agentredis.pwd = newConfig["agentredis"]["pwd"] or ""

		--Configure Redis
		inconcert.environment.apps.agent.redis = inconcert.environment.apps.agent.redis or {}
		inconcert.environment.apps.agent.redis.host = newConfig["redis"]["host"]
		inconcert.environment.apps.agent.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.apps.agent.redis.pwd = newConfig["redis"]["pwd"] or ""

		--Configure OutboundEngine Redis
		inconcert.environment.apps.agent.outboundengineredis = inconcert.environment.apps.agent.outboundengineredis or {}
		inconcert.environment.apps.agent.outboundengineredis.host = newConfig["outboundengineredis"]["host"]
		inconcert.environment.apps.agent.outboundengineredis.port = tonumber(newConfig["outboundengineredis"]["port"])
		inconcert.environment.apps.agent.outboundengineredis.pwd = newConfig["outboundengineredis"]["pwd"] or ""

		--Configure Elasticsearch (is an array, but for now, we only use the first node)
		inconcert.environment.apps.agent.elasticsearch = inconcert.environment.apps.agent.elasticsearch or {}
		inconcert.environment.apps.agent.elasticsearch.host = newConfig.elasticsearch[1].host
		inconcert.environment.apps.agent.elasticsearch.port = tonumber(newConfig.elasticsearch[1].port)
		-- elascitsearch user credentials
		inconcert.environment.apps.agent.elasticsearch.credentials = {
			userId    = newConfig.elasticsearch[1].userId,   -- or "inconcert",
			userPwd   = newConfig.elasticsearch[1].userPwd  -- or "12345678"
		}
		--Configure Elasticsearch Contact Dialer (is an array, but for now, we only use the first node)
		inconcert.environment.apps.agent.elasticsearch_dialer = inconcert.environment.apps.agent.elasticsearch_dialer or {}
		inconcert.environment.apps.agent.elasticsearch_dialer.host = "127.0.0.1"
		inconcert.environment.apps.agent.elasticsearch_dialer.port = -1
		for index, value in newConfig["elasticsearch_dialer"]:childs() do
			inconcert.environment.apps.agent.elasticsearch_dialer.host = value.host
			inconcert.environment.apps.agent.elasticsearch_dialer.port = tonumber(value.port)
			-- inconcert.environment.apps.agent.elasticsearch_dialer.numberOfShards = tonumber(value.number_of_shards)
			-- inconcert.environment.apps.agent.elasticsearch_dialer.numberOfReplicas = tonumber(value.number_of_replicas)
			break
		end
		inconcert.environment.apps.agent.elasticsearch_dialer.credentials = inconcert.environment.apps.agent.elasticsearch.credentials

		--Configure time servers
		inconcert.environment.apps.agent.timeServers = {}
		local timeservers = inconcert.environment.apps.agent.timeServers
		for key, value, state in newConfig["timeServers"]:childs() do
			if state ~= "deleted" then
				timeservers[#timeservers + 1] = { IP = value.IP, port = value.port }
			end
		end

		--LogDebug("[AGENT.app_status.CheckStatus] B)---> inconcert.environment.apps.agent.elasticsearch_dialer: %s", json.encode(inconcert.environment.apps.agent.elasticsearch_dialer))

		-- Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("[AGENT.app_status.CheckStatus] Connecting to cassandra on %s:%s", value.host, value.port)
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port

				Ticketek.init({
					cassandra = { host = value.host, port = value.port },
					redis = inconcert.environment.apps.agent.redis,
					logger = make_compatible_logger()
				})

				audits.init({
					cassandra = { host = value.host, port = value.port }
				})
				LogInfo("[AGENT.app_status.CheckStatus] Done connecting to cassandra")
			end
		end
		inconcert.environment.apps.agent.cassandra = cassandra
		--LogDebug("[AGENT:CheckStatus] 2--------> Start checking status for app --> inconcert.environment.apps.agent: %s.", json.encode(inconcert.environment.apps.agent)) -- TODO: comentar esa linea, esta solo para diagnostico
		InitializeFullcontact()

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end

		app_status.actualStatus = "ready"
		return "ready"
	end
}

function InitializeFullcontact (callback)
	--LogDebug("[AGENT:InitializeFullcontact] Initializing Fullcontact for 'AGENT'.")
	local Fullcontact = require "fullcontact"
	local cassandraConfig = {
		host = inconcert.environment.apps.agent.cassandra.ip,
		port = inconcert.environment.apps.agent.cassandra.port
	}
	Fullcontact.init({
		owner               = "AGENT",
		cassandra           = cassandraConfig,
		cluster             = inconcert.environment.apps.agent.cluster,
		elasticsearch       = inconcert.environment.apps.agent.elasticsearch,
		elasticsearchDialer = inconcert.environment.apps.agent.elasticsearch_dialer,
		timeServers         = inconcert.environment.apps.agent.timeServers,
		redis               = inconcert.environment.apps.agent.redis,
		logger              = make_compatible_logger()
	})
	if callback then
		callback()
	end
end

local function GetVcc (vcc, keyspace, columns)
	local vccInfo = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCs",
		key = vcc,
		columns = columns
	} or {}
	if m_helpers.utils.HasLicenseSystem() then
		local options = utils.Get_m_helpers_agentservice_options()
		vccInfo = m_helpers.agentservice.GetVccWithLicenseInfo(options, vccInfo)
	end
	return vccInfo
end

-- Returns a function bound to a translator for this application
local function makeTextTranslator()
	local translator = inconcert.translators.agent
	return function(...)
		return translator:Text(...)
	end, i18nlib.LCID[translator.language]
end

-- This function returns the title of the application. It needs to use an explicit translator because _T could be bound to a
-- different application.
function GetApplicationTitle()
	local _T = makeTextTranslator()
	return _T("agent")
end

local htmlRender = require"inconcert.render".new("agent", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("agent") end,
	model = model.GetModels(),
	makeTextTranslator = makeTextTranslator
})

-- funcion que es llamada cuando el webserver se inicializa.
function Initialize ()
	LogInfo("[AGENT.Initialize] Starting application 'agent'.")

	--OJO que si se cambia en controllers\interactions.lua se debe cambiar aqui tambien!
	configControl = LuaConfigurationControl.Client {
		type = "AGENT",
		address = inconcert.environment.apps.agent.config_server.ip,
		port = inconcert.environment.apps.agent.config_server.port,
		node_address = inconcert.environment.apps.agent.config_server.myIp,
		node_port = inconcert.environment.apps.agent.config_server.myPort
	}

	configControl:AddServiceDependency("Repository")
	configControl:AddServiceDependency("AgentService")
	configControl:AddServiceDependency("InteractionRouter")
	configControl:AddServiceDependency("KeyStore")
	configControl:TriggerNodeReady()

	return true
end

--
-- The user has been authenticated and has permissions for this application. Should we let him enter anyway? (returns false if not)
local function CanInitializeSession(user)
	return true
end

---
--
local function InitializeSession (authUser) end

--
local function LogoutHelpDesk (fullUserId, loggedUserInfo)
	assert(fullUserId)
	if not loggedUserInfo or not loggedUserInfo.AppsPermissions or not loggedUserInfo.AppsPermissions.helpdesk then
		return
	end
	local helpdeskLogin = require "apps.helpdesk.login"
	helpdeskLogin.Logout(fullUserId)
end

local function LogoutCleanup (vcc, userName, ksAgent)
	assert(vcc)
	assert(userName)
	LogInfo("[AGENT.LogoutCleanup] Logout Cleanup for user %s@%s.", userName, vcc)
	-- Limpia las notificaciones
	agentNotifications.CleanNotifications(userName, vcc)
	-- Limipa las interacciones de Call de i6 en estado Wrapup
	local agentInteractions = require "apps.agent.controllers.interactions"
	agentInteractions.CleanupI6CallInteractions(vcc, userName, ksAgent)
end
--
-- This function gets called when the user logs out. This gets called reliably. The Logout function below may not
-- be called if the user logs off first (apps/login/logout)
function FinalizeSession (authUser, isPresenceFilterCheck)
	-- if the user does not especify an authUser, we use inconcert.authUser as default.
	local authUser = authUser or inconcert.authUser
	local vcc = authUser.virtualContactCenter
	local userName = authUser.userName
	local fullUserId = utils.AppendVcc(userName, vcc)
	local sessionId = authUser.authToken
	local models = model.GetModels()
	local loggedUserInfo = authFilter:GetLoggedUserInfo(userName, vcc, sessionId)

	LogInfo("[AGENT.FinalizeSession] Detail of the logged out user: userName: (%s) - virtualContactCenter: (%s)\nLogged User Info: '%s'.", userName, vcc, json.encode(loggedUserInfo or {}))

	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")

	agentNotifications.NotifyAgents({
		vcc = vcc,
		scope = "agent",
		eventType = "agent_server_logout",
		agents = { fullUserId },
		pollerMessage = {
			status = 403
		}
	}, ksAgent)

	-- una pausa para que el mensaje pueda llegar via poller al agente, antes de que cierre sesion
	if not isPresenceFilterCheck then
		socket.sleep(0.5)
	end

	if app_status:CheckStatus() ~= "ready" then
		ksAgent:Close()
		LogWarning("[AGENT.FinalizeSession] App status is not ready. Can not logout agent %s on session %s", userName, sessionId)
		return false
	end

	LogInfo("[AGENT.FinalizeSession] Removing temporary user files for session %s", sessionId)
	if models["TemporaryFiles"] then
		models["TemporaryFiles"].DeleteAll(userName, vcc, ksAgent)
	end
	local kickoutData = json.encode(authUser.kickout or {})

	LogInfo("[AGENT.FinalizeSession] Login out agent %s on session %s with kickout info '%s'", userName, sessionId, kickoutData)
	agent.LogoutAgent(authUser)
	LogoutCleanup(vcc, userName, ksAgent)
	ksAgent:Close()

	-- Realizo Logout en helpdesk
	LogoutHelpDesk(fullUserId, loggedUserInfo)

	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetAppLocation()
	return m_appLocation
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetApplicationName()
	return "agent"
end

--///////////////////////////////////////////////////////////////////////////////////
--
function serve_file(app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
--		LogInfo("[AGENT.serve_file] Serving file: %s", m_appLocation.."/"..filename)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, m_appLocation.."/"..filename)
	end
end

--///////////////////////////////////////////////////////////////////////////////////
--

--cache manifest file
_M:dispatch_get(function(web, language)
	web:content_type("text/cache-manifest")
	web.headers["Cache-Control"] = "no-cache"
	local version = get_version_string()
	local filename = string.sub(web.path_info, 2, #web.path_info)
	local env = { 
		_L = language,
		ICC_VERSION = version .. "." .. language
	}
	manifest = htmlRender:RenderPageLP("/apps/agent/offline.appcache", web, env)
	return manifest
end, "/offline.(.+).appcache")


---
--
function render_main (web)
	local url = "pages/inboxMaster.html"
	local _T, _L = makeTextTranslator()

	local manifest = ""
--	if os.getenv("INCONCERT_ENV_MODE") == "PRODUCTION" then
--		manifest = "/inconcert/apps/agent/offline." .. _L .. ".appcache"
--	else
--		LogInfo("Not using appCache")
--	end
	LogInfo("Not using appCache")

	-- WIDGETS INTEGRATIONS
	local widgetsIntegrationsAPI = require "apps.integrations.api"
	local widgetsDefinitions = widgetsIntegrationsAPI.GetAvailablesWidgetsDefinitions()

	local env = {
		content = function()
			local url = url
			local captures = {
			}
			return self.htmlRender:IncludeLP(url, captures)
		end,
		Title = "Titulo de la pagina",
		ImageTitleClass = 'Image Title Class',
		AppCacheManifest = manifest,
		ICC_VERSION = get_version_string(),
		_T = _T,
		_L = _L,
		fb_appId = inconcert.environment.apps.agent.facebook.appId,
		twitter_domain = inconcert.environment.apps.agent.twitter.subdomainHost,
		m_utils = utils,
		m_permissions = permissions,
		m_customInfo = {
			widgets = {
				definitions = widgetsDefinitions
			}
		}
	}
	_M.response.DontCache(web)
	return htmlRender:RenderPageLP(url, web, env)
end

--///////////////////////////////////////////////////////////////////////////////////////////////////
--Mapeo urls /apis(vistas, controlles, scripts y styles) --------------------------------------------------------

--main url
_M:dispatch_get(render_main, "/%d*")

-- dummy url. el usuario redis@system la postea cada tanto para mantenerse vivo
-- esto es un hack y va a volar
_M:dispatch_post(function()
	return ""
end, "/dummy_ping")

-----------------------
--  (controllers) urls/apis
function GetControllers ()
	local controllerDir = m_appLocation .. "/controllers"
	local controllers = {}
	for file in lfs.dir(controllerDir) do
		if file ~= "." and file ~= ".." then
			local f = controllerDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local controllerName = string.match(file, [[(.+).lua]])
				if controllerName then
					controllers[controllerName] = { url = controllerName }
				end
			end
		end
	end
	return controllers
end

--main application module
agent.MapUris(_M, htmlRender)

for key, controllerDef in pairs(GetControllers()) do
	--requires y mapeo funciones
	local controller = relativeRequire("controllers." .. key, ...)
	controller.MapUris(_M, htmlRender, "/" .. controllerDef.url)
end

--Login APIs
local function ReplyJSON (web, data, ksAgent)
	assert(type(data) == "table", "Parameter data needs to be a table")
	web.headers["Content-Type"] = "application/json"
	if ksAgent then
		ksAgent:Close()
	end
	return json.encode(data)
end

function RenderLoginFormWSAPI (wsapi_env)
	--magic! invoke 'RenderLoginForm' as if the request came through Orbit
	local web = orbit.make_web_object(_M, wsapi_env)
	return web.status, web.headers, RenderLoginForm(web)
end

--[[
	Attempts to detect the language using the language Cookie (set by the login process) or the Accept-Language header sent by the browser
--]]
local function DetectLanguage (web)
	--LogInfo("[AGENT.DetectLanguage] Trying to detect language => Accept-Language: %s - Language Cookie: %s", web.vars.HTTP_ACCEPT_LANGUAGE, web.cookies.LULID)
	local httpProtocol =  web.vars and web.vars.HTTP_X_HTTP_PROTOCOL or "https"
	local acceptHeader, languageCookie = web.vars.HTTP_ACCEPT_LANGUAGE or "", web.cookies.LULID
	if languageCookie then
		return languageCookie
	else
		local vLULID = {
				path = "/", 
				domain = m_cookies_domain
			}
			
		if httpProtocol == "https" then
			vLULID.httponly = true
			vLULID.sameSite = "None"
			vLULID.secure = true
		end

		local langCode = string.match(acceptHeader, "^([^,]+)")		
		if langCode then
			langCode = langCode:match("(%a%a)[%-.+]?")
			vLULID.value = langCode
			web:set_cookie("LULID", vLULID)
			return langCode
		else
			vLULID.value = "en_US"
			web:set_cookie("LULID", vLULID)
			return "en"	-- defaulting to english
		end
	end
end

---
-- Shows the 'login' form
--
function RenderLoginForm (web)
	local env = {}
	env.required_login_script = "/inconcert/lib/webtoolkit/webtoolkit.base64.js"

	inconcert.authUser = { languageId = DetectLanguage(web) }
	env._T, env._L = makeTextTranslator()
	env.m_utils = utils

	inconcert.authUser = nil

	web:delete_cookie("AUTH_SID_AGENT", "/", authFilter:GetCookiesDomain())

	return htmlRender:RenderPageLP("/apps/agent/pages/login.html", web, env)
end

---
-- Shows the 'agent_api_iframe' form
--
function RenderAgentAPIForm (web)
	local env = {}
	return htmlRender:RenderPage("/apps/agent/pages/agentApiIframe.html", web, env)
end

--[[
	Obtiene del VCC (cassandra), la configuracion de seguridad que este tenga.
	Ej:
		{
			"captcha": {
				"enabled": true,
				"publicKey": "6Lf-J0UUAAAAAGmR_IKYvcsFhjsu9v_AXeTnsg87",
				"secretKey": "6Lf-J0UUAAAAAFFGy0pjfQgF0tZRtaGYOc2NuWdw"
			}
		}
]]--
local function LoadVccSecurityInfo (vcc, keyspace)
	local columns = {"Security"}
	local vccInfo = GetVcc(vcc, keyspace, columns)
	--LogDebug("[AGENT.LoadVccSecurityInfo] Loading vcc %s security: %s", vcc, json.encode(vccInfo.Security or{}))
	return vccInfo.Security
end

local function GetNonceToken (web)
	local username = web.input.username
	local vcc = web.input.callcenter

	if not(username and vcc) then
		LogError("[AGENT.GetNonceToken] Error in GetNonceToken, invalid parameters: user (%s), vcc (%s)", username, vcc)
		--web.status = 400
		return ReplyJSON(web, {
			status = false,
			message = "invalid request"
		})
	end

	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	--expires in 30 seconds
	local expireTokenSec = 30 -- en los casos normales el token expira en 30 segundos, para dar tiempo a que escriba la password (login en dos pasos).

	-- obtenemos la configuracion de seguridad del vcc y la enviamos al lado del cliente.
	-- para que por ejemplo sea utilizada en el login con captcha
	local vccSecurityInfo = LoadVccSecurityInfo(vcc, ksAgent)
	if not vccSecurityInfo then
		ksAgent:Close()
		LogError("[AGENT.GetNonceToken] Error in GetNonceToken, vcc security info not found on vcc '%s'", vcc)
		--web.status = 400
		return ReplyJSON(web, {
			status = false,
			--message = "invalid request"
			message = { message = "invalid username", field_id = "username"}
		})
	end

	if vccSecurityInfo and vccSecurityInfo.captcha then
		vccSecurityInfo.captcha.secretKey = nil -- la clave secreta NO puede viajar al lado del cliente.
		if tostring(vccSecurityInfo.captcha.enabled) == "1" then
			-- en el caso de captcha debemos aumentar la expiracion del token, ya que sino no le da el tiempo al usuario a responder las preguntas de captcha.
			expireTokenSec = 60
		end
	end
	--mfa
	if vccSecurityInfo and vccSecurityInfo.mfa and vccSecurityInfo.mfa then
		local mfa = vccSecurityInfo.mfa
		vccSecurityInfo.mfa = nil
		if tostring(mfa.enabled) == "1" then
			expireTokenSec = mfa.expirationSeconds + 60 --add some extra time to process mfa initial step
		end
	end

	if vccSecurityInfo and vccSecurityInfo.authentication and (vccSecurityInfo.authentication.type == "BASIC" or vccSecurityInfo.authentication.type == "MFA") then
		vccSecurityInfo.passwordChecks = {}
		vccSecurityInfo.passwordChecks["LEN"] = { range = {vccSecurityInfo.authentication.passwordChecks.minLen, vccSecurityInfo.authentication.passwordChecks.maxLen} }
		vccSecurityInfo.passwordChecks["UPPERS"] = { range = {vccSecurityInfo.authentication.passwordChecks.minUpperChars} }
		vccSecurityInfo.passwordChecks["LOWERS"] = { range = {vccSecurityInfo.authentication.passwordChecks.minLowerChars} }
		vccSecurityInfo.passwordChecks["NUMBERS"] = { range = {vccSecurityInfo.authentication.passwordChecks.minNumericChars} }
		vccSecurityInfo.passwordChecks["SYMBOLS"] = { range = {vccSecurityInfo.authentication.passwordChecks.minSpecialChars} }
		vccSecurityInfo.passwordChecks["HISTORY"] = { range = {vccSecurityInfo.authentication.passwordChecks.history} }

		-- authentication data is not needed client side, so we remove it in the response for security reasons
		vccSecurityInfo.authentication = nil
	end

	local token, err, pwdExpirationDays, hasChangePasswordPermission, userInfo = authFilter:getNonceToken(web, expireTokenSec, ksAgent)
	if not token then
		ksAgent:Close()
		LogError("[AGENT.GetNonceToken] Error in GetNonceToken for user %s on vcc %s: '%s'", username, vcc, err)
		--web.status = 400
		local err_msg = {
			["migrating or rollbacking vcc"] = true
		}
		local resp_err = type(err) == "table" and err.message or err
		return ReplyJSON(web, {
			status  = false,
			message = {message = err_msg[resp_err] and resp_err or "invalid username", field_id = "username"}
		})
	end

	local mustChangePassword = err == "change password"
	local encodeMode = "md5"
	if utils.IsAllegroAuthenticationRequired(vcc, ksAgent) then
		encodeMode = "plain"
	end
	web.headers["Cache-Control"] = "no-cache"

	authFilter:OpenIdSetAuthEndPoint(web, token, expireTokenSec, vccSecurityInfo)
	ksAgent:Close()

	local info = {
		status = true,
		loginToken = token,
		encodeMode = encodeMode,
		info = {
			vcc = vccSecurityInfo
		},
		mustChangePassword = mustChangePassword,
		pwdExpirationDays = pwdExpirationDays,
		hasChangePasswordPermission = hasChangePasswordPermission,
		userInfo = {
			IsFullSystemAccess = utils.IsUserFullSystemAccess(userInfo)
		}
	}

	LogDebug("[AGENT.GetNonceToken] Agent %s@%s is Token And VCC information: '%s'", username, vcc, json.encode(info))

	return ReplyJSON(web, info)
end

local function GetConcurrentUserLimit (vcc, keyspace)
	local vccInfo = GetVcc(vcc, keyspace)
	--LogDebug("[AGENT.GetConcurrentUserLimit] Loading vcc %s OnlineUsers: %s", vcc, vccInfo.OnlineUsers or "none")
	return tonumber(vccInfo.OnlineUsers)
end

--[[
	Funcion utilizada (para desarrollo modo debug) para forzar un logout (Ej: Si falla el codigo del authFilter y el usuario ha quedado logueado.)
	La misma se utiliza en modo de desarrollo
]]--
-- local function ForceLogout (web)
-- 	LogDebug("ForceLogout 1")
-- 	authFilter:Logout(web)
-- 	LogDebug("ForceLogout 2")
-- 	authFilter:RevokeFriendCredentials(web)
-- 	LogDebug("ForceLogout 3")
-- 	local options = {hashLogin = web.input.hash, nonceToken = web.input.loginToken, username =  web.input.username, vcc = web.input.callcenter, web = web}
-- 	LogDebug("ForceLogout 4")
-- 	authFilter:RevokeAllSessions(web.input.username.."@"..web.input.callcenter, authToken)
-- 	LogDebug("ForceLogout 5")
-- end

--[[
	Funcion encargada de invocar al módulo de "reCaptcha" para que realice las validaciones del token que ha recibido del lado del cliente.
]]--
local function ValidateReCaptcha (username, vcc, recaptchaToken, secretKey)
	LogInfo("[AGENT.ValidateReCaptcha] Validating ReCaptcha. Vcc: '%s', User: '%s',\nRecaptcha Token: '%s'\nSecret Key: '%s'.", username, vcc, recaptchaToken, secretKey)
	local reCaptcha = require "shared.captcha.recaptcha"
	local clientIP = nil
	local success, err = reCaptcha.ValidateCaptcha(recaptchaToken, clientIP, secretKey)
	if success then
		return true
	end
	-- codigo de error retornado al cliente
	local _T = makeTextTranslator()
	return false, _T("Authentication with reCaptcha has failed.")
end

--[[ MFA ]]

--[[ MFA:IsRequired ]]
local function MfaIsRequired(params)
	--always required
	return true
end

--[[ MFA:InitialStep ]]

local function mfaSendEmail(params)
	--required params
	local username = params.username
	local vcc = params.vcc
	local fullUserId = params.fullUserId
	local mfa = params.mfa
	local code = params.code
	local _T = params._T
	--1. get user email
	local models = model.GetModels()
	local userEmail = models.Users.GetSpecificColumn(fullUserId, "Mail")
	if utils.StringIsEmpty(userEmail) then
		return { status = false, message = _T("Must have a configured e-mail to login. Please contact to your administrator.") }
	end
	--2. get account
	local account = models.VCCs.GetAccounts(vcc, "MAIL", mfa.account)
	if not account then
		return { status = false, message = _T("Invalid mfa configuration. Please contact to your administrator.") }
	end
	--3. send email
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:SendIsolatedMail({
		account = account,
		to = userEmail,
		subject = mfa.subject,
		body = mfa.body,
		content_type = "text/html",
		attachments = mfa.attachments,
		params = {
			agentAtVCC = fullUserId,
			CODE = code
		}
	})
	if result.result == "OK" then
		return { status = true, address = userEmail }
	else
		return { 
			status = false, 
			message = string.format(_T("Error sending e-mail to %s. Please contact to your administrator."), userEmail) 
		}
	end
end

local function mfaMessageInitialStep(params)
	--required params
	local username = params.username
	local vcc = params.vcc
	local fullUserId = params.fullUserId
	local mfa = params.mfa
	local _T = params._T
	local sendMessage = params.sendMessage
	--1. generate code
	local code = utils.RandomNumber(mfa.keyLength)
	--2. send message
	params.code = code
	local result = sendMessage(params)
	if not result.status then
		return result
	end
	--3. store message
	local models = model.GetModels()
	models.Users.SetMfaCode(fullUserId, code, mfa.expirationSeconds)
	--4. return
	result.mfa = mfa.type
	result.length = mfa.keyLength
	return result
end

local mfaInitialStepByType = {
	MAIL = function(params)
		params.sendMessage = mfaSendEmail
		return mfaMessageInitialStep(params)
	end
}

local function MfaInitialStep(params)
	if mfaInitialStepByType[params.mfa.type] then
		return mfaInitialStepByType[params.mfa.type](params)
	else
		LogError("[AGENT.MfaInitialStep] - invalid mfa type %s", params.mfa.type)
		return { status = false, message = params._T("invalid request") }
	end
end

--[[ MFA:FinalStep ]]

local function mfaMessageFinalStep(params)
	--required params
	local username = params.username
	local vcc = params.vcc
	local fullUserId = params.fullUserId
	local mfa = params.mfa
	local data = params.data
	local validateOnly = params.validateOnly or false
	local _T = params._T
	--1. check code
	local code = data.code
	if utils.StringIsEmpty(code) then
		--1.1 code not found
		return { status = false, message = _T("invalid request") }
	end
	local models = model.GetModels()
	local storedCode = models.Users.GetMfaCode(fullUserId)
	if utils.StringIsEmpty(storedCode) then
		--1.2 stored code not found: expired
		return { status = false, message = _T("Authentication code expired. Please retry.") }
	end
	if storedCode == code then
		if not validateOnly then
			models.Users.RemoveMfaCode(fullUserId)
		end
		return { status = true }
	else
		return {
			status = false,
			retry = true,
			message = _T("Invalid code."),
		}
	end
end

local mfaFinalStepByType = {
	MAIL = function(params)
		return mfaMessageFinalStep(params)
	end
}

local function MfaFinalStep(params)
	if mfaFinalStepByType[params.mfa.type] then
		return mfaFinalStepByType[params.mfa.type](params)
	else
		LogError("[AGENT.MfaFinalStep] - invalid mfa type %s", params.mfa.type)
		return { status = false, message = params._T("invalid request") }
	end
end

--[[
	Funcion que valida el limte de usuario concurrentes que han iniciado sesion no supere lo que se ha configurado en el VCC..
	Util solo para ambientes de DEBIAN ya que para UBUNTU la validacion se realiza en el AgentService.
]]--
local function ValidateConcurrentUserLimit (userInfo, keyspace)
	local resp = {status = true, message = "The validation will be done in the AgentService"}

	if m_helpers.utils.HasLicenseSystem() then
		--[[-----------------------------------------------
			UBUNTU
		[-------------------------------------------------]]
		--LogInfo("[AGENT.ValidateConcurrentUserLimit] OK - User: '%s@%s' - message: '%s'.", userInfo.Id, userInfo.VCC, resp.message)
		return resp -- (OK) - para el caso de los ambientes de UBUNTU, la validacion se realiza en el AS.
	end
	--[[-----------------------------------------------
		DEBIAN
	[-------------------------------------------------]]

	-- 1. Valida si es un Super Usuario.
	if utils.IsUserFullSystemAccess(userInfo) then
		resp.message = "The User is a Super User"
		LogInfo("[AGENT.ValidateConcurrentUserLimit] OK - User: '%s@%s' - message: '%s'.", userInfo.Id, userInfo.VCC, resp.message)
		return resp -- (OK) - Es un Super Usuario, NO debe de validar.
	end

	-- 2. Obtiene los valores configurados en el VCC.
	local limit = GetConcurrentUserLimit(userInfo.VCC, keyspace)
	if not limit then
		resp.message = "No limits set on vcc"
		LogInfo("[AGENT.ValidateConcurrentUserLimit] OK - User: '%s@%s' - message: '%s'.", userInfo.Id, userInfo.VCC, resp.message)
		return resp -- (OK) - si no tiene configurado en el VCC.
	end

	-- 3. Valida si se encuentra dentro de los límites permitidos en el VCC.
	local currentSessions = authFilter:GetAllSessions(userInfo.VCC)
	local totalLoggedIn = utils.Cardinality(currentSessions)
	if totalLoggedIn < limit then
		resp.message = "It is within the limit allowed in the vcc"
		LogInfo("[AGENT.ValidateConcurrentUserLimit] OK - User: '%s@%s' - message: '%s'.", userInfo.Id, userInfo.VCC, resp.message)
		return resp -- (OK) - Se encuentra dentro del límite permitido en el vcc.
	end

	-- error - llego al limite
	LogWarning("[AGENT.ValidateConcurrentUserLimit] Concurrent user limit reached: user '%s@%s'. Max online users (%s) reached on vcc. totalLoggedIn: %d", userInfo.Id, userInfo.VCC, limit, totalLoggedIn)
	local _T = makeTextTranslator()
	resp.status   = false
	resp.message  = _T("Concurrent user limit reached")
	resp.vccLimit = limit
	return resp
end

--it exists to show error messages with allegro integration
--remove when remove it
local function ValidateAuthentication (web)
	local username = web.input.username
	local vcc = web.input.callcenter
	local hash = web.input.hash
	local loginToken = web.input.loginToken
	inconcert.authUser = { languageId = DetectLanguage(web) }
	local _T = makeTextTranslator()
	inconcert.authUser = nil

	LogInfo("[AGENT.ValidateAuthentication] Validating Authentication for username '%s@%s'.", username, vcc)

	local ok, mfaData = pcall(json.decode, web.input.mfa or "{}")
	if not ok then
		LogError("[AGENT.ValidateAuthentication] Error in Authenticate, invalid login mfa data %s", web.input.mfa)
		return ReplyJSON(web, {
			status = false,
			message = "invalid request"
		})
	end
	if not(username and vcc and hash) then
		LogError("[AGENT.ValidateAuthentication] Error in Authenticate, invalid parameters: user (%s), vcc (%s), hash (%s)", username, vcc, hash)
		return ReplyJSON(web, {
			status = false,
			message = _T("invalid request")
		})
	end

	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	username = username:lower()
	vcc = vcc:lower()
	local user = models.Users.Get(username.."@"..vcc, nil, keyspace)
	if not user then
		LogError("[AGENT.ValidateAuthentication] Error retrieving user '%s'", username.."@"..vcc)
		return ReplyJSON(web, {
			status = false,
			message = _T("invalid username or password")
		}, keyspace)
	end

	local fullUserId = utils.AppendVcc(username, vcc)
	local requiresMfa = false
	local vccSecurityInfo = LoadVccSecurityInfo(vcc, keyspace)

	--mfa
	if vccSecurityInfo and vccSecurityInfo.mfa and tostring(vccSecurityInfo.mfa.enabled) == "1" then
		if not utils.TableIsEmpty(mfaData) then
			--mfa final step
			local result = MfaFinalStep({
				username = username,
				vcc = vcc,
				fullUserId = fullUserId,
				data = mfaData,
				mfa = vccSecurityInfo.mfa,
				validateOnly = true,
				_T = _T
			})
			if not result.status then
				LogError("[AGENT.ValidateAuthentication] Error running mfa final step for user %s@%s: %s", username, vcc, result.message)
				return ReplyJSON(web, result, keyspace)
			end
		else
			--mfra initial step
			requiresMfa = MfaIsRequired({
				username = username, 
				vcc = vcc,
				fullUserId = fullUserId,
				mfa = vccSecurityInfo.mfa
			})
		end
	end

	local sessionCookie = web.cookies[authFilter.AUTH_SID]
	if sessionCookie then
		LogInfo("Client has session '%s' attached to it. Verifying if it's still valid", sessionCookie)
		local userData = authFilter:GetUserFromToken(sessionCookie)
		if userData then
			LogInfo("Session '%s' is still valid. Can not continue", sessionCookie)
			local message
			if username == userData.userName then
				message = _T("Already logged in the same machine")
			else
				message = _T("Another user is logged in the same machine")
			end
			return ReplyJSON(web, {
				status = false,
				message = message
			}, keyspace)
		end
	end

	-- Only let one agent in at the same time
	-- (This is only needed for the Agent application)
	if authFilter:GetTokenFromUser(username, vcc) then
		LogInfo("[AGENT.ValidateAuthentication] Agent %s@%s is already logged in", username, vcc)
		return ReplyJSON(web, {
			status = false,
			message = _T("invalid username or password")
		}, keyspace)
	end

	-- (DEDIAN) valido el limite de usuarios contectados en el VCC.
	local resp = ValidateConcurrentUserLimit(user, keyspace)
	if not resp.status then
		return ReplyJSON(web, resp, keyspace)
	end

	local options = {hashLogin = hash, nonceToken = loginToken, username = username, vcc = vcc, web = web}
	if utils.IsAllegroAuthenticationRequired(vcc, keyspace) then
		options.allegro = allegro
	end
	local result, message = authFilter:ValidateAuthentication(options, keyspace)
	if not result then
		LogWarning("[AGENT.ValidateAuthentication] Error authenticating user %s@%s: %s", username, vcc, message)
		return ReplyJSON(web, {
			status = false,
			message = _T(message)
		}, keyspace)
	end
	
	if requiresMfa then
		local result = MfaInitialStep({
			username = username,
			vcc = vcc,
			fullUserId = fullUserId,
			mfa = vccSecurityInfo.mfa,
			_T = _T
		})
		if result.status then
			LogInfo("[AGENT.ValidateAuthentication] Authentication for user %s@%s requires mfa.", username, vcc)
			return ReplyJSON(web, result, keyspace)
		else
			LogError("[AGENT.ValidateAuthentication] Error running mfa initial step for user %s@%s: %s", username, vcc, result.message)
			return ReplyJSON(web, result, keyspace)
		end
	end
	
	return ReplyJSON(web, {
		status = true,
		message = _T("Login succeeded")
	}, keyspace)
end

local function OpenIdValidateAuthentication (web)
	local options = web and web.input or {}
	local languageId = DetectLanguage(web)
	inconcert.authUser = {languageId = languageId}
	local _T = makeTextTranslator()

	local function showError(error)
		LogWarning("[AGENT.OpenIdValidateAuthentication] Authenticating user with openId options %s: %s", json.encode(options), error or "")
		error = error == "invalid local_user" and "There is a session signed in to Azure with another user" or "Login Error"
		return htmlRender:RenderPageLP("pages/openIdError.html", web, {
			code = -1,
			message = _T(error),
			_T = _T
		})
	end
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local result, message = authFilter:OpenIdValidateAuthentication(options, ksAgent)
	ksAgent:Close()
	if not result then
		return showError(message)
	end
	inconcert.authUser.userName = options.username
	inconcert.authUser.userGroup = options.userGroup
	return htmlRender:RenderPageLP("pages/openid_Ok.html", web, {
		code = 0,
		message = _T("Login succeeded"),
		username = options.username,
		loginToken = options.loginToken,
		callcenter = options.callcenter,
		authenticationFrom = "api",
		app = "agent",
		_T = _T
	})
end

local function Authenticate (web)
	local username = web.input.username
	local vcc = web.input.callcenter
	local hash = web.input.hash
	local loginToken = web.input.loginToken
	local recaptchaToken = web.input.recaptchaToken
	local jwtToken = web.input.jwtToken

	-- optional new password protected with md5
	local hashNew = web.input.hashNew
	-- optional new password protected with jwt
	local jwtTokenNew = web.input.jwtTokenNew

	LogInfo("[AGENT.Authenticate] LogginIn username '%s@%s' - web.input.data: '%s', web.input.mfa: '%s'.", username, vcc, web.input.data, web.input.mfa)

	local sessionData = utils.JSON_Decode(web.input.data) or {}
	-- sessionData =  {
	-- 	type = "MOBILE"
	-- --	type = "WEB"
	-- }
	if not sessionData then
		LogError("[AGENT.Authenticate] Error in Authenticate, invalid login data %s", web.input.data)
		return ReplyJSON(web, {
			status = false,
			message = "invalid request"
		})
	end

	local mfaData = utils.JSON_Decode(web.input.mfa) or {}
	if not mfaData then
		LogError("[AGENT.Authenticate] Error in Authenticate, invalid login mfa data %s", web.input.mfa)
		return ReplyJSON(web, {
			status = false,
			message = "invalid request"
		})
	end
	--[[
		atributo utilizado para saber desde donde me han invocado.
			webuser 
				desde el navegador web (caso habitual), en el cual SI puede utilizar captcha si es que el vcc lo tiene configurado.
			webuser_with_auto_login
				desde el navegador con auto login, el caso en el se invoca desde el api de usuarios (NUNCA utiliza captcha)
			api
				el valor utilizado cuando nos invocan desde scripts lua (ej: el new_user.lua) (NUNCA utiliza captcha)
	]]--
	local authenticationFrom = web.input.authenticationFrom or "api"
	--
	local languageId = DetectLanguage(web)
	inconcert.authUser = {languageId = languageId}
	local _T = makeTextTranslator()
	inconcert.authUser = nil

	LogInfo("[AGENT.Authenticate] Authenticating FROM '%s' - User '%s' VCC '%s' Hash '%s' Login Token '%s' Language '%s'.\nRecaptcha Token: '%s'.", authenticationFrom, username, vcc, hash, loginToken, languageId, recaptchaToken)

	--[[
		Funcion utilizada (para desarrollo modo debug) para forzar un logout (Ej: Si falla el codigo del authFilter y el usuario ha quedado logueado.)
		La misma queda comentada.
	]]--
	--ForceLogout(web)

	if not(username and vcc and hash) then
		LogError("[AGENT.Authenticate] Error in Authenticate, invalid parameters: user (%s), vcc (%s), hash (%s)", username, vcc, hash)
		return ReplyJSON(web, {
			status = false,
			message = _T("invalid request")
		})
	end

	username = username:lower()
	vcc = vcc:lower()
	local sessionRejoin = nil
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")

	local user = models.Users.Get(username.."@"..vcc, nil, keyspace)
	if not user then
		LogError("[AGENT.Authenticate] Error retrieving user '%s'", username.."@"..vcc)
		return ReplyJSON(web, {
			status = false,
			message = _T("invalid username or password")
		}, keyspace)
	end

	local now = os.time()
	user.LastLoginAttempTimestamp = tostring(now)
	--This is to avoid update user data on external applications if the vcc has them (helpdesk, allegro, ...)
	user.saveFromLogin = true

	local function SaveUser ()
		--Recuperar los skills del user para que no se pierdan al guardar
		local userSkills = models.Users.GetAgentSkills (vcc, username, keyspace) or {}
		user.skills = {}
		for skillId, skillValue in pairs(userSkills) do
			local skill = {
				skill = skillId,
				weight = skillValue
			}
			table.insert(user.skills, skill)
		end
		local saveUserResult, saveUserErr = models.Users.Save(username, vcc, user, keyspace)
		if not saveUserResult then
			LogError("[AGENT.Authenticate] Error saving user '%s':'%s'", username.."@"..vcc, saveUserErr)
		end
		return saveUserResult
	end

	-- Si el usuario es admin@system y viene desde la web lo cancelo.
	if username == "admin" and vcc == "system" and authenticationFrom == "webuser" then
		LogInfo("[AGENT.Authenticate] Agent %s@%s is not valid from webuser.", username, vcc)
		SaveUser()
		return ReplyJSON(web, {
			status = false,
			message = _T("invalid username or password")
		}, keyspace)
	end 
	
	-- Only let one agent in at the same time
	-- (This is only needed for the Agent application)
	local token = authFilter:GetTokenFromUser(username, vcc)
	if token then
		local userData = authFilter:GetUserFromToken(token)
		if (userData and userData.sessionData and userData.sessionData.type == "MOBILE") and sessionData.type == "MOBILE" then
			LogInfo("[AGENT.Authenticate] Agent %s@%s is already logged with MOBILE platform type, session rejoin", username, vcc)
			sessionRejoin = token
		elseif (userData and userData.sessionData and userData.sessionData.type == "MOBILE") and sessionData.type ~= "MOBILE" then
			LogInfo("[AGENT.Authenticate] Agent %s@%s is already logged with MOBILE platform type, changing to %s type", username, vcc, sessionData.type)
			FinalizeSession(userData)
			authFilter:RevokeAllSessions(username.."@"..vcc)
		else
			LogInfo("[AGENT.Authenticate] Agent %s@%s is already logged in", username, vcc)
			SaveUser()
			return ReplyJSON(web, {
				status = false,
				message = _T("invalid username or password")
			}, keyspace)
		end
	end

	local sessionCookie = web.cookies[authFilter.AUTH_SID]
	if sessionCookie then
		LogInfo("[AGENT.Authenticate] Client has session '%s' attached to it. Verifying if it's still valid. authFilter.AUTH_SID: '%s'.", sessionCookie, authFilter.AUTH_SID)
		local userData = authFilter:GetUserFromToken(sessionCookie)

		if userData then
			LogInfo("[AGENT.Authenticate] Session '%s' is still valid. Can not continue", sessionCookie)
			local message
			if username == userData.userName then
				message = _T("Already logged in the same machine")
			else
				message = _T("Another user is logged in the same machine")
			end
			SaveUser()
			return ReplyJSON(web, {
				status = false,
				message = message
			}, keyspace)
		end
	end

	local fullUserId = utils.AppendVcc(username, vcc)
	local skipCaptcha = authenticationFrom:lower() ~= "webuser"
	local requiresMfa = false
	local vccSecurityInfo = LoadVccSecurityInfo(vcc, keyspace)

	if not sessionRejoin then
		--mfa
		if vccSecurityInfo and vccSecurityInfo.mfa and tostring(vccSecurityInfo.mfa.enabled) == "1" then
			if not utils.TableIsEmpty(mfaData) then
				skipCaptcha = true
				--mfa final step
				local result = MfaFinalStep({
					username = username,
					vcc = vcc,
					fullUserId = fullUserId,
					data = mfaData,
					mfa = vccSecurityInfo.mfa,
					_T = _T
				})
				if not result.status then
					LogError("[AGENT.Authenticate] Error running mfa final step for user %s@%s: %s", username, vcc, result.message)
					SaveUser()
					return ReplyJSON(web, result, keyspace)
				end
			else
				--mfra initial step
				requiresMfa = MfaIsRequired({
					username = username, 
					vcc = vcc,
					fullUserId = fullUserId,
					mfa = vccSecurityInfo.mfa
				})
			end
		end

		--[[
			Si la función fue invocada desde el navegador, realiza la validaciones del token de captcha (si es que el vcc así lo tiene configurado.)
		]]--
		if not skipCaptcha and vccSecurityInfo and vccSecurityInfo.captcha and tostring(vccSecurityInfo.captcha.enabled) == "1" then
			local success, translatedErr = ValidateReCaptcha(username, vcc, recaptchaToken, vccSecurityInfo.captcha.secretKey)
			if not success then
				SaveUser()
				return ReplyJSON(web, {
					status = false,
					message = translatedErr
				}, keyspace)
			end
		end
		-- (DEDIAN) valido el limite de usuarios contectados en el VCC.
		local resp = ValidateConcurrentUserLimit(user, keyspace)
		if not resp.status then
			SaveUser()
			return ReplyJSON(web, resp, keyspace)
		end
	end

	--[[
		Funcion encargada de validar los permisos sobre las APPs.
		Las APPs que se validan sus permisos se encuentran definidos en la tabla "appsDefinitions".

		El resultado de esta validación, se enviará a la rocka "authFilter" para que esta almacene en el usuario que ha iniciado sesión, los permisos
		que el mismo tiene sobre las APPs definidas.
	]]--

	local function LoadAppsPermissions ()
		local models = model.GetModels()
		local userGroup = models.Users.GetSpecificColumn(fullUserId, "UserGroup", keyspace)
		if not userGroup then
			return
		end
		local applicationsActions = permissions.GetApplicationsActions(keyspace, vcc)
		-- Aca validamos permisos sobre una App completa
		local function HasPermissionToApp(appId)
			return permissions.HasPermissionToApplication(appId, fullUserId, userGroup, nil, keyspace, {applicationsActions = applicationsActions})
		end
		-- Aca validamos permisos sobre una acción determinada de una App (app "apps_integrations" action "helpdesk")
		local function HasPermissionToHelpDesk(appId)
			local custom_permissions = model_perm.GetUserPermissions(username, keyspace, vcc)
			local hasPermissiontToAction = permissions.IsPermissionEnabledForSpecificUser(appId, "helpdesk_full_access", fullUserId, userGroup, custom_permissions, nil, nil, keyspace, applicationsActions, true)

			--Si no tiene permisos para helpdesk_full_access, valido que tenga para tickets
			if not hasPermissiontToAction then
				-- Permiso de interaccion app
				local appInteraction = permissions.HasPermissionToApplication("interaction", fullUserId, userGroup, nil, keyspace, {
					applicationsActions = applicationsActions,
					user_vcc_permissions = custom_permissions})
				if appInteraction then
					-- tickets: Permisos de tickets y al menos una campana con integracion helpdesk o no tener encuenta integracion en campanas(helpdesk_tickets_without_integration)
					local hasTicketsForceLogin = permissions.IsPermissionEnabledForSpecificUser(appId, "helpdesk_tickets_without_integration", fullUserId, userGroup, custom_permissions, nil, nil, keyspace, applicationsActions, true)
					hasPermissiontToAction = hasTicketsForceLogin or (permissions.IsPermissionEnabledForSpecificUser(appId, "helpdesk_tickets_in_app_interaction", fullUserId, userGroup, custom_permissions, nil, nil, keyspace, applicationsActions, true) and (#models.Campaigns.GetUserCampaignsWithHelpdeskIntegration(vcc, username, keyspace) > 0))
				end
				
			end
			--LogDebug("[AGENT.Authenticate] user %s HasPermissionToHelpDesk '%s'.", fullUserId, hasPermissiontToAction)
			return hasPermissiontToAction
		end
		--[[
			Validamos:
				1 - Si tenemos permisos sobre la APP "apps_integrations" action "helpdesk"; se utiliza para validar las licencias contra HD.
				2 - Si tenemos permisos sobre la App "WebDesigner" en modo Diseño. El modo Runtime quedó a modo ejemplo, pero siempre es true.
		]]--
		-- Validamos si tenemos permisos sobre la APP "apps_integrations" action "helpdesk".
		-- 
		local appsDefinitions = {
			helpdesk             = {permissionAppName = "apps_integrations", fn = HasPermissionToHelpDesk   },
			webdesigner_designer = {permissionAppName = "administration"   , fn = HasPermissionToApp        },
			webdesigner_runtime  = {permissionAppName = "interaction"      , fn = function() return true end}
		}
		local apps = {}
		for appName, def in pairs(appsDefinitions) do
			apps[appName] = def.fn(def.permissionAppName)
			--LogDebug("--> appName '%s' permission = '%s', permissionAppName: '%s'", appName, apps[appName], def.permissionAppName)
		end
		return apps
	end

	local options = {
		hashLogin = hash,
		nonceToken = loginToken,
		username = username,
		vcc = vcc,
		web = web,
		appsPermissions = not requiresMfa and LoadAppsPermissions(), -- Aca lo pasamos al "authfilter"
		sessionData = sessionData,
		sessionRejoin = sessionRejoin,
		jwtToken = jwtToken,
		jwtTokenNew = jwtTokenNew,
		hashNew = hashNew
	}

	if not sessionRejoin then
		if utils.IsAllegroAuthenticationRequired(vcc, keyspace) then
			options.allegro = allegro
		end
		if requiresMfa then
			--validates authentication and run mfa initial step
			local result, message = authFilter:ValidateAuthentication(options, keyspace)
			if not result then
				LogWarning("[AGENT.Authenticate] Error authenticating user %s@%s: %s", username, vcc, message)
				SaveUser()
				return ReplyJSON(web, {
					status = false,
					message = _T(message)
				}, keyspace)
			end
			local result = MfaInitialStep({
				username = username,
				vcc = vcc,
				fullUserId = fullUserId,
				mfa = vccSecurityInfo.mfa,
				_T = _T
			})
			if result.status then
				LogInfo("[AGENT.Authenticate] Authentication for user %s@%s requires mfa.", username, vcc)
				SaveUser()
				return ReplyJSON(web, result, keyspace)
			else
				LogError("[AGENT.Authenticate] Error running mfa initial step for user %s@%s: %s", username, vcc, result.message)
				SaveUser()
				return ReplyJSON(web, result, keyspace)
			end
		end
	end

	local isInternalUser = vccSecurityInfo.authentication.type == "MFA" or vccSecurityInfo.authentication.type == "BASIC"
	local result, authToken, additionalErrorInfo
	result, authToken, additionalErrorInfo = authFilter:AuthenticateUser(options)

	if result and hashNew and options.nonceToken and utils.HasAllegroIntegration(vcc, keyspace) then
		-- Se valida que el password no tenga caracteres reservados que se modifican durante la serialización de la URL (#, &, ?, =), porque da error en el login.
		local pattern = "[?=&#]"
		local jwtData = jwt.decode(jwtTokenNew, md5.sumhexa(options.nonceToken), "HS256")
		if jwtData and jwtData.password and string.match(jwtData.password, pattern) then
			LogError("[AGENT.Authenticate] The password value must not contain reserved characters (#, &, ?, =).")
			result = nil
			authToken = "new password reserved characters error"
			additionalErrorInfo = {}
		end
	end

	if result then
		if isInternalUser then
			--We have a new, already validated, password 
			if hashNew then
				user.Password = hashNew
				if not user.PasswordHistory or #user.PasswordHistory == 0 then
					user.PasswordHistory = {}
				end
				table.insert(user.PasswordHistory, hashNew)
				user.PasswordChangeTimestamp = tostring(os.time())
				user.MustChangePassword = "0"
			end
			--reset user's failed attemps and update user status if apply 
			if user.LoginFailedAttemps and tonumber(user.LoginFailedAttemps) > 0 or hashNew then
				user.LoginFailedAttemps = "0"
				user.Status = "1"
				user.StatusReason = ""
				user.StatusTimestamp = tostring(os.time())
			end
		end

		--Revoke all previous friend credentials
		--if it's not done this way, other user in the same machine could still be logged
		--to these applications and it's not a desired behaviour
		authFilter:RevokeFriendCredentials(web)

		--Revokes all the sessions (except the created above) for this user
		--this is done in order to logout user from other applications
		authFilter:RevokeAllSessions(username.."@"..vcc, authToken)

		local authUser = authFilter:GetUserFromToken(authToken)
		InitializeSession(authUser)

		LogInfo("[AGENT.Authenticate] --> Authentication for user %s@%s is OK. authUser: %s", username, vcc, json.encode(authUser))

		user.LastLoginSuccessTimestamp = tostring(now)
		--If password has changed we must inform to external application when apply
		if jwtTokenNew and hashNew then
			user.saveFromLogin = false
		end
		SaveUser()

		return ReplyJSON(web, {
			status = true,
			message = _T("Login succeeded"),
			authToken = authToken,
			languageId = i18nlib.LCID[authUser.languageId]
		}, keyspace)
	else
		authToken = authToken or _T("Login Error")
		local authFilterKnownErrors = {
			["invalid user token"]           = "time to login has expired",
			["new password error"]           = "new password error",
			["migrating or rollbacking vcc"] = _T("administrative tasks are running"),
			["new password reserved characters error"] = _T("The new password value must not contain reserved characters (#, &, ?, =)."),
		}
		local messageError = authFilterKnownErrors[authToken] or _T(authToken)
		LogWarning("[AGENT.Authenticate] Error authenticating user %s@%s: Error code: '%s', description: '%s'.", username, vcc, authToken, messageError)

		if isInternalUser and additionalErrorInfo and additionalErrorInfo.IncrLoginFailedAttemps then
			additionalErrorInfo = nil
			if user.LoginFailedAttemps then
				local loginFailedAttemps = tonumber(user.LoginFailedAttemps) + 1
				if tonumber(vccSecurityInfo.authentication.passwordChecks.blockOnLoginFailures) > 0 and loginFailedAttemps >= tonumber(vccSecurityInfo.authentication.passwordChecks.blockOnLoginFailures) then
					user.Status = "0"
					user.StatusReason = "MAX_LOGIN_RETRIES"
					user.StatusTimestamp = tostring(os.time())
				end
				user.LoginFailedAttemps = tostring(loginFailedAttemps)
				SaveUser()
			end
		end

		if additionalErrorInfo and not additionalErrorInfo.dontSaveUser then
			SaveUser()
		end
		return ReplyJSON(web, {
			status = false,
			message = messageError,
			additionalErrorInfo = additionalErrorInfo
		}, keyspace)
	end
end

--[[
]]--
local function Logout (web)
	local kickoutData
	if web.input then
		encodedData = web.input["post_data"] or ""
		local ok, postedData = pcall(json.decode,encodedData)
		if ok then
			for name,value in pairs (postedData) do
				if name ~= "callFrom" then
					-- Parametros no esperados, posible problema de seguridad.
					LogWarning("Logout - receive invalid data %s", encodedData)
					web.status = 400       
					return ReplyJSON(web, {status = false, message = "invalid request"})
				end
			end
		end
		
	end
	if inconcert.authUser then
		if utils.HasAllegroIntegration() then
			local data, err = allegro.generic.GetAgentLastKickout(inconcert.authUser.userName)
			if data then
				LogInfo("[AGENT.Logout] kickout data from Allegro '%s'", json.encode(data))
				kickoutData = data.KickoutData
			end
			inconcert.authUser.kickout = kickoutData
		end
		local encodedData = ""
		if web.input then
			encodedData = web.input["post_data"] or ""
		end
		LogInfo("[AGENT.Logout] Session:'%s' vcc:'%s' userName:'%s' Extra Data:'%s'", inconcert.authUser.authToken, inconcert.authUser.virtualContactCenter, inconcert.authUser.userName, encodedData)
		FinalizeSession()
		authFilter:Logout(web)
	end
	return ReplyJSON(web, {status = true, kickout = kickoutData})
end

--login related stuff

_M:dispatch_get(RenderLoginForm, "/login/")
_M:dispatch_get(RenderAgentAPIForm, "/agent_api/")
_M:dispatch_post(GetNonceToken, "/login/getToken/")
_M:dispatch_post(Authenticate, "/login/authenticate/")
_M:dispatch_post(ValidateAuthentication, "/login/validate_authentication/")
_M:dispatch_post(Logout, "/login/logout/")
_M:dispatch_post(OpenIdValidateAuthentication, "/login/openid/validate_authentication/")

--scripts and pages

--paginas
local function render_pages (web, name, versionCode, languageCode)
	if languageCode ~= "" then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()

	-- Las funciones de la tabla _MAIL son utilizadas en el html "print.html".
	-- La funcion "Noop" (empty) debe retornar un string vacio, ya que es lo que espera el archivo "print.html" en cada una de estas funciones.
	local Noop = function()
		return ""
	end

	local env = {
		_T = _T,
		_L = _L,
		m_utils = utils,
		_MAIL = {
			CanPrint   = Noop,
			GetDate    = Noop,
			GetFrom    = Noop,
			GetTo      = Noop,
			GetCC      = Noop,
			GetBCC     = Noop,
			GetSubject = Noop,
			GetBody    = Noop
		}
	}

	local page = "pages/" .. name .. ".html"

	LogInfo("AGENT:render_pages] Start rendering luapage. Page: '%s', version: '%s', language: '%s', URI: '%s'.", name, versionCode, languageCode, page)

	local luapage = htmlRender:RenderPageLP(page, web, env)
	return luapage
end

--scripts
function render_script(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:ScriptLp("scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")

--
--  styles and resources
--  NOTA: ESTO SE PROCESA EN XAVANTE
_M:dispatch_get(_M:serve_file(), "/images/.+", "/styles/.+", "/css/.+", ".+%.mp3")
