require "uuid"

module(..., package.seeall)

local m_datasource

function GetName()
	return "AutoAttendant"
end
	
function GetDefinition(datasource)
	m_datasource = datasource
	return _M
end

--deletes a menu and its submenus recursivley
function DeleteMenu(menu, keyspace)
	if menu then
		LogInfo("Deleting auto attendant menu %s", menu.id)
		if menu.options then
			for selector, destination in pairs(menu.options) do
				if destination.type == "submenu" then
					DeleteMenu(destination.menu, keyspace)
				end
			end
		end
		keyspace:RemoveColumn(menu.id, "AutoAttendantMenus")
	end
end

--saves a menu and its submenus recursively
function SaveMenu(menu, keyspace)
	if not menu then
		return
	end
	local menuId = uuid.new()
	LogInfo("Saving new auto attendant menu %s for main message %s", menuId, json.encode(menu.mainMessage))
	--save main message in each language
	for languageId, message in pairs(menu.mainMessage) do
		keyspace:InsertSuperColumn(menuId, "AutoAttendantMenus",  "mainMessage", languageId, json.encode(message))
	end
	--save options
	if menu.options then
		for selector, option in pairs(menu.options) do
			if option.type == "submenu" then
				--save each submenu recursively
				option.menuID = SaveMenu(option.menu, keyspace)
				option.menu = nil
			end
			keyspace:InsertSuperColumn(menuId, "AutoAttendantMenus",  "options", selector, json.encode(option))
		end
	end
	--save settings
	if menu.settings.allowDirectConnectionWithAgent then
		keyspace:InsertSuperColumn(menuId, "AutoAttendantMenus",  "settings", "allowDirectConnectionWithAgent",  "1")
	else
		keyspace:RemoveColumn(menuId, "AutoAttendantMenus",  "settings", "allowDirectConnectionWithAgent")
	end
	keyspace:InsertSuperColumn(menuId, "AutoAttendantMenus",  "settings", "defaultOption", menu.settings.defaultOption or "")
	--save wrong input message in each language
	if menu.wrongInputMessage then
		for languageId, message in pairs(menu.wrongInputMessage) do
			keyspace:InsertSuperColumn(menuId, "AutoAttendantMenus",  "wrongInputMessage", languageId, json.encode(message))
		end
	end
	--save no input message in each language
	if menu.noInputMessage then
		for languageId, message in pairs(menu.noInputMessage) do
			keyspace:InsertSuperColumn(menuId, "AutoAttendantMenus",  "noInputMessage", languageId, json.encode(message))
		end
	end
	return menuId
end

--[[
	Saves the auto attendant
]]--
function Save(attendant)
	local keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	LogInfo("Saving auto attendant for address %s type %s", attendant.addressId, attendant.addressType)

	local key = attendant.addressType .. "_" .. attendant.addressId

	local currentAutoAttendant = Get(attendant.addressId, attendant.addressType)
	
	--create all the menus starting from root, returns the root menu id created
	local rootMenuID = SaveMenu(attendant.rootMenu, keyspace)
	local languageMenuID = SaveMenu(attendant.languageMenu, keyspace)

	--overwrite the current menu IDs
	if languageMenuID then
		keyspace:InsertColumn(key, "AccountAutoAttendants", "languageMenuID", languageMenuID)
	end
	if rootMenuID then
		keyspace:InsertColumn(key, "AccountAutoAttendants", "rootMenuID", rootMenuID)
	end

	--recursively delete the current menus
	DeleteMenu(currentAutoAttendant.rootMenu, keyspace)
	DeleteMenu(currentAutoAttendant.languageMenu, keyspace)
	
	keyspace:Close()
	return true
end


local function GetMenu(menuId, keyspace)
	local menu = {}
	local rows = keyspace:GetSuperRangeSlices("AutoAttendantMenus",  --column family
											 "",  --supercolumn
											 menuId, --startkey
											 menuId, --endkey
												{ 
												column_names = {},
												slice_range = { 
															start = "", 
															finish = "", 
															reversed = false, 
															count = 200}
												},
											200 --total keys
								) or {}		
	for key, entry in rows:Iterator() do
		if (key == menuId) then
			for k, v in pairs(entry) do
				local entryColumns = {}
				for colid, column in pairs(v.columns) do
					local result, value = pcall(json.decode, column.value)
					if result then
						entryColumns[column.name] = value
					else
						entryColumns[column.name] = column.value
					end
				end	
				menu[v.name] = entryColumns
			end
			--iterate over submenu hierarchy and return all the submenu tree
			if menu.options then
				for selector, destination in pairs(menu.options) do
					if destination.type == "submenu" then
						destination.menu = GetMenu(destination.menuID, keyspace)
						destination.menu.name = destination.menuName
					end
				end
			end
			break
		end
	end
	menu.id = menuId
	return menu
end

--[[
	Returns the autoattendant
]]--
function Get(addressId, addressType)
	local autoAttendant = {}
	local keyspace = m_datasource.cassandra:CassandraConnection("Agent")
	
	local key = addressType.."_"..addressId

	local aaColumns = keyspace:GetSliceRange(key, 
										   "AccountAutoAttendants", 
										   "", { 
												column_names = {},
												slice_range = { 
															start = "", 
															finish = "", 
															reversed = true, 
															count = 200
														}
												}
											)
	if (aaColumns) then
		for k, v in ipairs(aaColumns) do
			autoAttendant[v.name] = v.value
		end
		if autoAttendant.languageMenuID then
			autoAttendant.languageMenu = GetMenu(autoAttendant.languageMenuID, keyspace)
		end
		if autoAttendant.rootMenuID then
			autoAttendant.rootMenu = GetMenu(autoAttendant.rootMenuID, keyspace)
		end
	end
	keyspace:Close()
	return autoAttendant
end
