local helpers = require "apps.agent.helpers"
local utils = require "shared.apps.utils"
local ir = require "lua_iragentlink"
local PM = require "provisioningmanager"

module(..., package.seeall)

local m_datasource
local m_ksCore
local agentservice = require "apps.agent.agentservice"
local allegro_campaigns = require "apps.allegro.controllers.campaigns"

local function is_table (t)
	return type(t) == "table"
end

local function is_string (v)
	return type(v) == "string"
end

local function is_empty (t)
	return next(t) == nil
end

--The fields listed here are saved ciphered, and never are returned to the client
local SensibleData = {
	IM = {password = true},
	MAIL = {password = true}
}

function GetName ()
	return "Campaigns"
end

function GetDefinition (datasource, ksCore)
	--LogDebug("[AGENT.CAMPAIGNS.GetDefinition] 1-----------------------> SET datasource def. datasource: %s, ksCore: %s", datasource, ksCore)
	m_datasource = datasource
	m_ksCore = ksCore
	return _M
end

local function GetDatasource ()
	if m_datasource then
		--LogDebug("[AGENT.CAMPAIGNS.GetDatasource] a-----------------------> m_datasource: %s", m_datasource)
		return m_datasource
	end
	local connections = utils.GetConnections()
	--LogDebug("[AGENT.CAMPAIGNS.GetDatasource] b-----------------------> connections: %s", connections)
	return connections
end

-- TO-DO: this is duplicated on administration, some campaigns helpers needs
-- to be shared between apps
function GetVCCCampaigns (vcc, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCCampaigns",
		key = vcc
	} or {}
end

function GetAllVCCsAccounts ( keyspace )
	local params = {
		keyspace = keyspace or "Agent",
		superColumns = true,
		columnFamily = "CampaignAccounts",
		startKey = "",
		endKey = ""
	}

	return utils.GetAllColumns( params ) or {}
end

function GetVccAccounts (vcc, keyspace)
	assert(vcc, "vcc is required")
	vcc = vcc:lower()
	local all_accounts = GetAllVCCsAccounts( keyspace )
	local vcc_accounts = {}

	for campaign_id, campaign_accounts in pairs( all_accounts ) do
		local _, campaign_vcc = utils.SeparateVCC(campaign_id)
		if campaign_vcc and (campaign_vcc:lower() == vcc) then
			vcc_accounts[campaign_id] = campaign_accounts
		end
	end

	return vcc_accounts
end

local function GetAccountId (account)
	if type(account) == "string" then
		return account
	end
	return account.id
end

local function GetCompleteNameLoggedUser ()
	return inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter
end

local function Lock (key, name, value)
	value = value or GetCompleteNameLoggedUser()
	assert( type(key) == "string" and type(name) == "string")
	local locked = GetDatasource().redis:hsetnx(key, name, value)
	if not locked then
		LogError("Failed to lock key: %s, name: %s by user: %s.", key, name, value)
	end
	return locked
end

local function Unlock (key, name)
	assert( type(key) == "string" and type(name) == "string")
	local unlocked = GetDatasource().redis:del(key, name)
	if not unlocked then
		LogError("Failed to unlock key: %s, name: %s.", key, name)
	end
	return unlocked
end


local function SaveAttentionLevelsAccountUpdateAgentService (campaign_id, account_id, account)
	-- ATENCION: EN ESTA FUNCION SE DEBE ACTUALIZAR TAMBIEN O TENER EN CUENTA LA PRIORIDAD DE LA CUENTA.
	--           QUE LLEGA EN EL ATRIBUTO "Account.Priority (1/100)"
	LogInfo("Save Account, Update Assigned Attention Levels AgentService -> campaign:'%s', account_id:'%s', AttentionLevels:'%s'.",
		campaign_id, account_id, json.encode(account.AttentionLevels or{}) )

	-- Map the data to the format expected by the agentservice
	local levels = {}
	for _, attn_level in pairs(account.AttentionLevels or {}) do
		table.insert(levels, {
			id = assert(attn_level.attention_level_id),
			distribution = attn_level.distribution,
			priority = tonumber(attn_level.priority) or 0
		})
	end

	agentservice:GetAgentClient():UpdateAccountAttentionLevels(campaign_id, account_id, levels)
end

local function DeleteAttentionLevelsAccountUpdateAgentService (campaign_id, account_id, account)
	LogInfo("Delete Account, Assigned Attention Levels AgentService -> campaign:'%s', account_id:'%s', AttentionLevels:'%s'.",
		campaign_id, account_id, json.encode(account.AttentionLevels or{}) )

	agentservice:GetAgentClient():DeleteAccountAttentionLevels(campaign_id, account_id)
end

local function UpdateSkillsAccount (keyspace, campaign_id, account_id, account, action)

	LogInfo("Update Index Accounts by Skill -> campaign:'%s', account_id:'%s', action: '%s'.", campaign_id, account_id, action )

	local close_keyspace
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end

	local function MakeSkillId( skill )
		return skill .. "@" .. inconcert.authUser.virtualContactCenter
	end
	local function Add()
		local skills = account.Skills or {}
		if type(skills) == "string" then
			skills = json.decode(skills)
		end
		for _, skill in ipairs( skills ) do
			keyspace:InsertColumn( MakeSkillId( skill.skill ), "AccountsBySkill", account_id,
				json.encode({
					campaign_id = campaign_id,
					account_name = account.name,
					skill = skill
				})
			)
		end
	end

	local function Remove( remove_type )

		local function Del( skills )
			skills = skills or {}
			if is_string( skills ) then
				skills = json.decode(skills)
			end
			for _, skill in ipairs( skills ) do
				keyspace:RemoveColumn( MakeSkillId( skill.skill ), "AccountsBySkill", account_id )
			end
		end

		if remove_type == "all" then
			local current_campaign = Get(campaign_id, keyspace)
			if not current_campaign or not current_campaign.accounts then
				return false
			end
			for _, current_account in ipairs( current_campaign.accounts ) do
				if current_account.id == account_id then
					Del(current_account.Skills)
				end
			end
		else
			Del(account.Skills)
		end
		return true
	end

	if action == "NEW" then
		Add()
	elseif action == "UPDATE" then
		if Remove( "all" ) then
			Add()
		end
	else
		Remove()
	end

	if close_keyspace then
		keyspace:Close()
	end
end

local function UpdateSignatureAccount (keyspace, campaignIdWithVCC, account_id, account, action)

	LogInfo("[UpdateSignatureAccount] Update Index Accounts by Signatures -> campaign:'%s', account_id:'%s', action: '%s'.", campaignIdWithVCC, account_id, action )

	local close_keyspace
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end

	local _, vcc = utils.SeparateVCC(campaignIdWithVCC)

	local function Add()
		local signatureIdWithOutVCC = account.Signature
		if type(signatureIdWithOutVCC) == "string" and signatureIdWithOutVCC:len() > 0 then
			local signatureIdWithVCC = utils.AppendVcc(signatureIdWithOutVCC, vcc)
			keyspace:InsertColumn( signatureIdWithVCC, "AccountsBySignature", account_id,
				json.encode({
					campaign_id  = campaignIdWithVCC,
					account_id   = account_id,
					account_name = account.name,
					signature_id = signatureIdWithOutVCC
				})
			)
		end
	end

	local function Remove( remove_type )

		local function Del( signatureIdWithOutVCC )
		if type(signatureIdWithOutVCC) == "string" and signatureIdWithOutVCC:len() > 0 then
				local signatureIdWithVCC = utils.AppendVcc(signatureIdWithOutVCC, vcc)
				keyspace:RemoveColumn( signatureIdWithVCC, "AccountsBySignature", account_id )
			end
		end

		if remove_type == "all" then
			local current_campaign = Get(campaignIdWithVCC, keyspace)
			if not current_campaign or not current_campaign.accounts then
				return false
			end
			for _, current_account in ipairs( current_campaign.accounts ) do
				if current_account.id == account_id then
					Del(current_account.Signature)
				end
			end
		else
			Del(account.Signature)
		end
		return true
	end

	if action == "NEW" then
		Add()
	elseif action == "UPDATE" then
		if Remove( "all" ) then
			Add()
		end
	else
		Remove()
	end

	if close_keyspace then
		keyspace:Close()
	end
end

local function UpdateFlowAccount (keyspace, campaignIdWithVCC, account_id, account, action)
	LogInfo("[AGENT.CAMPAIGNS:UpdateFlowAccount] Update Index Accounts by Flow -> campaign:'%s', account_id:'%s', action: '%s', flowInfo: '%s'.", campaignIdWithVCC, account_id, action, json.encode(account.FlowInformation))
	--LogInfo("[AGENT.CAMPAIGNS:UpdateFlowAccount] Update Index Accounts by Flow -> account:'%s'.", json.encode(account))

	local close_keyspace
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end

	local _, vcc = utils.SeparateVCC(campaignIdWithVCC)

	local function GetFlowIds(flowInfo)
		local flowIds = {}
		for flowType, flowDefinition in pairs(flowInfo or {}) do
			if flowDefinition and flowDefinition.Process then
				flowIds[flowDefinition.Process] = true
			end
		end
		return flowIds
	end

	--La funcion debe devolver los flujos de las paginas tal cual se tiene en el resto de tipos
	-- de cuenta, esto es: {clave: {Process: <IdProcessFlow>}}
	local function makeStandardFlowInformationForInstagramAndFBAccounts(nonStandardFlowInformation)
		if type(nonStandardFlowInformation) == "string" then
			nonStandardFlowInformation = json.decode(nonStandardFlowInformation)
		end
		local standard = nonStandardFlowInformation
		for key, value in pairs(standard) do
			standard[key].Process = value.processFlow
		end

		return standard
	end

	local function GetExtendedFlowInformation(acc)
		local extendedFlowInformation = acc.FlowInformation or acc.pagesMessenger and makeStandardFlowInformationForInstagramAndFBAccounts(acc.pagesMessenger) or {}
		local currentIds = GetFlowIds(extendedFlowInformation)
		local botsToTransferFromAgent = account.BotsToTransferFromAgent and type(account.BotsToTransferFromAgent) == "table" and account.BotsToTransferFromAgent or {}
		for _, flow in pairs(botsToTransferFromAgent) do
			if not currentIds[flow.id] then
				extendedFlowInformation[flow.id] = {}
				extendedFlowInformation[flow.id].Process = flow.id
			end
		end
		return extendedFlowInformation
	end
	
	local accountFlowInformation = GetExtendedFlowInformation(account)

	local function Add()
		local flowInfo = utils.JSON_Decode(accountFlowInformation)
		local flowIds = GetFlowIds(flowInfo)

		for flowId, _ in pairs(flowIds) do
			local flowIdWithVcc = utils.AppendVcc(flowId, vcc)
			local assignment_type = ""

			if flowInfo and flowInfo.on_schedule and flowInfo.on_schedule.Process and flowInfo.on_schedule.Process == flowId then
				assignment_type = "on_schedule"
			end

			if flowInfo and flowInfo.out_of_schedule and flowInfo.out_of_schedule.Process and flowInfo.out_of_schedule.Process == flowId then
				assignment_type = assignment_type == "on_schedule" and "on_schedule, out_of_schedule" or "out_of_schedule"
			end

			keyspace:InsertColumn( flowIdWithVcc, "AccountsByProcessFlow", account_id,
				json.encode({
					campaign_id  = campaignIdWithVCC,
					account_id   = account_id,
					account_name = account.name,
					flow_id      = flowId,
					assignment_type = assignment_type
				})
			)
		end
	end

	local function Remove( remove_type )

		local function Del( flowInfo )
			local flowIds = GetFlowIds(flowInfo)
			for flowId, _ in pairs(flowIds) do
				local flowIdWithVcc = utils.AppendVcc(flowId, vcc)
				keyspace:RemoveColumn( flowIdWithVcc, "AccountsByProcessFlow", account_id )
			end
		end

		if remove_type == "all" then
			local current_campaign = Get(campaignIdWithVCC, keyspace)
			if not current_campaign or not current_campaign.accounts then
				return false
			end
			for _, current_account in ipairs( current_campaign.accounts ) do
				if current_account.id == account_id then
					local currentAccountFlowInformation = GetExtendedFlowInformation(current_account)
					Del(currentAccountFlowInformation)
				end
			end
		else
			Del(accountFlowInformation)
		end

		return true
	end

	if action == "NEW" then
		Add()
	elseif action == "UPDATE" then
		if Remove( "all" ) then
			Add()
		end
	else
		Remove()
	end

	if close_keyspace then
		keyspace:Close()
	end
end

--[[
	Función pública ya que también se accede desde el model de administration.
]]--
function RemoveCalli6AccountPackagesRelations (campaignIdWithVCC, accountId, currentAccount, ksAgent, ksProv)
	assert(campaignIdWithVCC)
	assert(accountId)
	assert(ksAgent)
	assert(ksProv)

	local campaignId, vcc = utils.SeparateVCC(campaignIdWithVCC)
	if utils.HasAllegroIntegration(vcc, ksAgent) then
		--LogDebug("[RemoveCalli6AccountPackagesRelations] VCC '%s' without Allegro Integration..", vcc)
		return false
	end

	LogInfo("[RemoveCalli6AccountPackagesRelations] Removing Packages relations in Call Campaign '%s' accountId:'%s'.",
			campaignIdWithVCC, accountId)

	currentAccount = currentAccount or GetAccount(accountId, campaignIdWithVCC, ksAgent)
	--LogDebug("---> currentAccount %s", json.encode(currentAccount))

	if currentAccount and currentAccount.packages then

		--#11602 - LicenseSystem - PM.UpdateVCC and PM.GetVCC now require a previous call to InitHelpersAgentServiceOptions
		local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
		local success_agentservice_options, err_agentservice_options =
			pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)
		
		local success, resp = PM.UpdateCampaignAccountPackages(ksProv, ksAgent, vcc, campaignId, accountId)
		if not success then
			local err = resp
			LogWarning("[RemoveCalli6AccountPackagesRelations] Error removing relation Vcc '%s' campaign '%s'. Err: %s",
					   vcc, campaignId, json.encode(err))
		end
	end
	-- ok
	return true
end

local function UpdateCalli6AccountPackagesRelations (
	ksAgent, campaignIdWithVCC, accountId, account, currentAccount, ksProv, mustValidateAllegroAccount)
	--LogDebug("[UpdateCalli6AccountPackagesRelations] Updating Packages relations in Call Campaign '%s' without Allegro Integration. accountId:'%s', account: '%s'.", campaignIdWithVCC, accountId, json.encode(account))

	local campaignId, vcc = utils.SeparateVCC(campaignIdWithVCC)
	if mustValidateAllegroAccount then
		if utils.HasAllegroIntegration(vcc, ksAgent) then
			--LogDebug("[UpdateCalli6AccountPackagesRelations] VCC '%s' without Allegro Integration..", vcc)
			return
		end
		if account.proxyType ~= "CALL" then
			--LogDebug("[UpdateCalli6AccountPackagesRelations] Account '%s' is not Call.", accountId)
			return
		end
	end

	LogInfo("[UpdateCalli6AccountPackagesRelations] Updating Packages relations in Call Campaign '%s'. accountId:'%s', packages: '%s'.",
			campaignIdWithVCC, accountId, json.encode(account.packages or {}))

	local closeKs = false
	if not ksProv then
		ksProv = GetDatasource().cassandra:CassandraConnection("Provisioning")
	end

	--#11602 - LicenseSystem - PM.UpdateVCC and PM.GetVCC now require a previous call to InitHelpersAgentServiceOptions
	local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
	local success_agentservice_options, err_agentservice_options =
		pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)

	local success, resp = PM.UpdateCampaignAccountPackages(ksProv, ksAgent, vcc, campaignId, accountId, account.packages)
	if not success then
		local err = resp
		LogWarning("[UpdateCalli6AccountPackagesRelations] Error adding relation Vcc '%s' campaign '%s'. Err: %s",
				   vcc, campaignId, json.encode(err))
	end
	if closeKs then
		ksProv:Close()
	end
end

--[[
	Esta funcion valida que un paquete de tipo I o I/O de la cuenta de telefonía, se encuentre en una única cuenta/campaña del vcc.
]]--
local function MustUpdateCalli6AccountPackagesRelations (ksAgent, campaignIdWithVCC, accountId, account, ksProv, mustValidateAllegroAccount)
	--LogDebug("[MustUpdateCalli6AccountPackagesRelations] Validating Packages I or I/O relations in Call Campaign '%s' without Allegro Integration. accountId:'%s', account: '%s'.", campaignIdWithVCC, accountId, json.encode(account))

	local campaignId, vcc = utils.SeparateVCC(campaignIdWithVCC)
	if mustValidateAllegroAccount then
		if utils.HasAllegroIntegration(vcc, ksAgent) then
			--LogDebug("[MustUpdateCalli6AccountPackagesRelations] VCC '%s' without Allegro Integration..", vcc)
			return true
		end
		if account.proxyType ~= "CALL" then
			--LogDebug("[MustUpdateCalli6AccountPackagesRelations] Account '%s' is not Call.", accountId)
			return true
		end
	end

	LogInfo("[MustUpdateCalli6AccountPackagesRelations] Validating Packages I or I/O relations in Call Campaign '%s' without Allegro Integration. accountId:'%s', packages: '%s'.", campaignIdWithVCC, accountId, json.encode(account.packages or {}))

	if not account.packages then
		--LogDebug("[MustUpdateCalli6AccountPackagesRelations] Account '%s' has no packages assigned.", accountId)
		return false, {result = "CALL_I6_PACKAGE_IS_REQUIRED", reason = "Package is required"}
	end
	--LogDebug("[MustUpdateCalli6AccountPackagesRelations] ----> Vcc: '%s', posted account.packages: '%s'.", vcc, json.encode(account.packages))

	-- Obtenemos los paquetes asignados al VCC.
	local releaseKsProv = false
	if not ksProv then
		releaseKsProv = true
		ksProv = GetDatasource().cassandra:CassandraConnection("Provisioning")
	end
	local currentVccPackages = PM.GetVCCPackages(ksProv, vcc) or {}
	if releaseKsProv then
		ksProv:Close()
	end

	--LogDebug("[MustUpdateCalli6AccountPackagesRelations] ----> Vcc: '%s', currentVccPackages: '%s'.", vcc, json.encode(currentVccPackages))

	local success = true
	local err
	accountId = accountId:lower()

	for	postedPackageId in pairs(account.packages) do
		-- Si el paquete que han posteado NO existe dentro de los packages asigandos al vcc, fallo.
		local packageInfo = currentVccPackages[postedPackageId]
		if not packageInfo then
			success = false
			err = {result = "CALL_I6_PACKAGE_DOES_NOT_EXIST", packageId = postedPackageId, reason = ("Package %s does not exist"):format(postedPackageId)}
			break
		end

		-- ahora valido que no exista asociado a alguna otra campaña/cuenta del vcc
		local ioMode = (packageInfo.ioMode or ""):upper()
		if ioMode == "I" or ioMode == "I/O" then
			for campaignId, vccPackageAccounts in pairs(packageInfo.campaigns or {}) do
				for vccPackageAccountId in pairs(vccPackageAccounts or {}) do
					if vccPackageAccountId:lower() ~= accountId then
						success = false
						err = {
							result = "CALL_I6_INBOUND_PACKAGES_CAN_ONLY_BE_IN_ONE_CAMPAIGN",
							packageId = postedPackageId,
							accountId = vccPackageAccountId,
							campaignId = campaignId,
							reason = ("Inbound package '%s' can only be in one campaign. Found in the %s' campaign."):format(postedPackageId, campaignId)
						}
						break
					end
				end
				if not success then
					break
				end
			end
			if not success then
				break
			end
		end
	end

	if not success then
		-- Error
		return false, err
	end
	-- is Ok
	return true
end

--[[
	Esta funcion devuelve la cuenta de whatsapp actualmente asignada a un codigo de activacion
]]--
function GetWhatsappInconcertCloudApiCode (activationCode)
	--#11602 - LicenseSystem - PM.GetWhatsappInconcertCloudApiCode and PM.MustUpdateWhatsappAccountActivationCode now require a previous call to InitHelpersAgentServiceOptions
	local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
	local success_agentservice_options = pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)	
	if not success_agentservice_options then
		return false
	end
	local ksProv = GetDatasource().cassandra:CassandraConnection("Provisioning")
	local result = PM.GetWhatsappInconcertCloudApiCode(ksProv, activationCode)
	ksProv:Close()
	return result
end

--[[
	Esta funcion valida que el codigo de activacion de la cuenta de whatsapp se encuentre disponible
]]--
local function MustUpdateWhatsappAccountActivationCode (ksAgent, campaignIdWithVCC, accountId, account)
	local campaignId, vcc = utils.SeparateVCC(campaignIdWithVCC)
	if account.proxyType ~= "WHATSAPP" then
		--LogDebug("[MustUpdateWhatsappAccountActivationCode] Account '%s' is not Whatsapp.", accountId)
		return true
	end
	if account.integrationType ~= "ICC_CLOUD_API" then
		--LogDebug("[MustUpdateWhatsappAccountActivationCode] Whatsapp Account '%s' integration Type is not inConcert Cloud API.", accountId)
		return true
	end

	LogInfo("[MustUpdateWhatsappAccountActivationCode] Validating Activation Code in Whatsapp Campaign '%s'. accountId:'%s', code: '%s'.", campaignIdWithVCC, accountId, account.activation_code)

	if not account.activation_code then
		--LogDebug("[MustUpdateWhatsappAccountActivationCode] Account '%s' has no Activation Code assigned.", accountId)
		return false, {result = "WHATSAPP_ACTIVATION_CODE_IS_REQUIRED", reason = "Activation Code is required"}
	end

	-- Comprobamos que el activation_code esta lockeado para este accountId.
	local activation_code_lock_key = "WhatsappActivationCode:" .. account.activation_code
	local accountIdLocked = GetDatasource().redis:get(activation_code_lock_key)
	if not accountIdLocked or accountIdLocked ~= accountId then
		--LogDebug("[MustUpdateWhatsappAccountActivationCode] Account '%s' has no locked Activation Code '%s'.", accountId, account.activation_code)
		return false, {result = "WHATSAPP_ACTIVATION_CODE_NOT_LOCKED", reason = "Activation Code is not locked"}
	end

	--#11602 - LicenseSystem - PM.GetWhatsappInconcertCloudApiCode and PM.MustUpdateWhatsappAccountActivationCode now require a previous call to InitHelpersAgentServiceOptions
	local helpers_agentservice_options = utils.Get_m_helpers_agentservice_options()
	local success_agentservice_options = pcall(PM.InitHelpersAgentServiceOptions, helpers_agentservice_options)
	if not success_agentservice_options then
		return false
	end

	local user = inconcert.authUser.userName
	local ksProv = GetDatasource().cassandra:CassandraConnection("Provisioning")
	local activationCodeResult = PM.SetWhatsappInconcertCloudApiCode(ksProv, account.activation_code, vcc, accountId, user)
	ksProv:Close()	
	
	if not activationCodeResult then
		LogWarning("[MustUpdateWhatsappAccountActivationCode] Activation Code '%s' could not be assigned to Account '%s'.", account.activation_code, accountId)
		return false, {result = "WHATSAPP_ACTIVATION_CODE_INVALID", reason = "Activation Code is invalid"}
	end

	-- is Ok
	return true
end

local function IncludeAllegroOtherInfo (allegro_info, allegroOtherInfo)
	if not allegro_info then
		return
	end
	allegro_info.otherInfo = allegroOtherInfo
end

--[[
	-- Crea en Allegro la campana
]]--
local function CreateAllegroCampaignIn55 (campaign, keyspace, allegro_info)
	local vcc = inconcert.authUser.virtualContactCenter
	local campaignIdWithVCC = campaign.Id

	if not utils.HasAllegroIntegration( vcc, keyspace ) then
		LogInfo("[AGENT.CAMPAIGNS:CreateAllegroCampaignIn55] - Allegro integration. Creating campaign '%s'. Vcc '%s' has not configured integration with allegro.", campaignIdWithVCC, vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	local success, err = allegro_campaigns.Create(campaign, vcc, allegro_info)
	if not success then
		if err.type == "ALLEGRO_INTEGRATION_ERROR" then
			LogError("[AGENT.CAMPAIGNS:CreateAllegroCampaignIn55] - Allegro integration error. Creating campaign '%s'. Response: '%s'.", campaignIdWithVCC, json.encode(err))
		else
			LogInfo("[AGENT.CAMPAIGNS:CreateAllegroCampaignIn55] - Allegro integration. Creating campaign '%s'. Response: '%s'.", campaignIdWithVCC, json.encode(err))
		end
		return false, err
	end

	return true
end
 
--[[
	-- Actualiza en Allegro la campana
]]--
function UpdateAllegroCampaignIn55 (campaignIdWithVCC, keyspace, allegro_info)
	local vcc = inconcert.authUser.virtualContactCenter

	if not utils.HasAllegroIntegration( vcc, keyspace ) then
		LogInfo("[AGENT.CAMPAIGNS:UpdateAllegroCampaignIn55] - Allegro integration. Updating campaign '%s'. Vcc '%s' has not configured integration with allegro.", campaignIdWithVCC, vcc)
		return false, { type = "WITHOUT_ALLEGRO_INTEGRATION" }
	end

	local success, err = allegro_campaigns.Update(campaignIdWithVCC, vcc, allegro_info)
	if not success then
		if err.type == "ALLEGRO_INTEGRATION_ERROR" then
			LogError("[AGENT.CAMPAIGNS:UpdateAllegroCampaignIn55] - Allegro integration error. Updating campaign '%s'. Response: '%s'.", campaignIdWithVCC, json.encode(err))
		else
			LogInfo("[AGENT.CAMPAIGNS:UpdateAllegroCampaignIn55] - Allegro integration. Updating campaign '%s'. Response: '%s'.", campaignIdWithVCC, json.encode(err))
		end
		return false, err
	end

	return true
end

local Fullcontact = require "fullcontact"
local function ExecuteFullcontactMethod (callback)
	local agent = require "apps.agent"
	agent.InitializeFullcontact(callback)
end

local m_blackListAPI
local function GetBlackListApi ()
	m_blackListAPI = m_blackListAPI or Fullcontact.Dialer:GetBlackListApi()
	return m_blackListAPI
end

local function GetBlacklistEnabledProxies (proxy)
	local proxies = {
		CALL = {
			IsEnabled = function(vcc, ksAgent)
				if utils.HasAllegroIntegration(vcc, ksAgent) then
					--LogInfo("[AGENT.CAMPAIGNS:GetBlacklistEnabledProxies] --> Unable to update blacklists relations for channel 'CALL' of VCC with Allegro, campaign: '%s', account: '%s'.", account.proxyType, campaignIdWithVCC, accountId)
					return false
				end
				return true
			end,
			GetAccountBlacklists = function(currentAccount)
				if not currentAccount then
					return {}
				end
				if not currentAccount.outgoingConfig then
					return {}
				end
				if not currentAccount.outgoingConfig.Blacklists then
					return {}
				end
				return currentAccount.outgoingConfig.Blacklists.scopes or {}
			end
		}
	}
	return proxies[proxy]
end

--[[
	Funcion que actualiza la relacion de las blacklists con la cuenta.
	Para realizar esta tarea, lo realiza invocando a la rocka "fullcontact"
]]--
local function UpdateBlacklistsAccount (ksAgent, campaignIdWithVCC, accountId, account, currentAccount, action)
	LogInfo("[AGENT.CAMPAIGNS:UpdateBlacklistsAccount] --> Updating blacklists in campaign: '%s' account: '%s'.", campaignIdWithVCC, accountId)
	local channelConfig = GetBlacklistEnabledProxies(account.proxyType)
	if not channelConfig then
		LogInfo("[AGENT.CAMPAIGNS:UpdateBlacklistsAccount] --> Unable to update blacklists relations for channel '%s', campaign: '%s', account: '%s'.", account.proxyType, campaignIdWithVCC, accountId)
		return false
	end
	local vcc = inconcert.authUser.virtualContactCenter
	if not channelConfig.IsEnabled(vcc, ksAgent) then
		return false
	end
	local newBlacklists = channelConfig.GetAccountBlacklists(account)
	--LogDebug("[AGENT.CAMPAIGNS:UpdateBlacklistsAccount] 1- new blacklists relations for channel '%s', campaign: '%s', account: '%s'. scopes: %s", account.proxyType, campaignIdWithVCC, accountId, json.encode(newBlacklists))
	
	local oldBlacklists = channelConfig.GetAccountBlacklists(currentAccount)
	--LogDebug("[AGENT.CAMPAIGNS:UpdateBlacklistsAccount] 2- old blacklists relations for channel '%s', campaign: '%s', account: '%s'. scopes: %s", account.proxyType, campaignIdWithVCC, accountId, json.encode(oldBlacklists))

	local fullUserId = utils.AppendVcc( inconcert.authUser.userName, vcc)
	local success, response = GetBlackListApi():AddRelatedCampaignsAccounts(vcc, campaignIdWithVCC, accountId, oldBlacklists, newBlacklists, fullUserId)
	-- ok
	return true
end

function RemoveBlacklistsAccount (ksAgent, campaignIdWithVCC, accountId, account)
	LogInfo("[AGENT.CAMPAIGNS:RemoveBlacklistsAccount] .--> Removing blacklists of campaign: '%s' account: '%s'.", campaignIdWithVCC, accountId)
	local channelConfig = GetBlacklistEnabledProxies(account.proxyType)
	if not channelConfig then
		LogInfo("[AGENT.CAMPAIGNS:RemoveBlacklistsAccount] --> Unable to remove blacklists relations for channel '%s', campaign: '%s', account: '%s'.", account.proxyType, campaignIdWithVCC, accountId)
		return false
	end
	local blacklists = channelConfig.GetAccountBlacklists(account)
	--LogDebug("[AGENT.CAMPAIGNS:RemoveBlacklistsAccount] current blacklists relations for channel '%s', campaign: '%s', account: '%s'. scopes: %s", account.proxyType, campaignIdWithVCC, accountId, json.encode(blacklists))
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = utils.AppendVcc( inconcert.authUser.userName, vcc)
	local success, response = GetBlackListApi():RemoveRelatedCampaignsAccounts(vcc, campaignIdWithVCC, accountId, blacklists, fullUserId)
	-- ok
	return true
end

function SaveAccount (ksAgent, campaignIdWithVCC, account_id, account, is_new_account, allegroOtherInfo, ksProv, pMustValidateAllegroAccount, pMustValidateNativeAccount)
	local mustValidateAllegroAccount = true
	if type(pMustValidateAllegroAccount) == "boolean" then
		mustValidateAllegroAccount = pMustValidateAllegroAccount
	end
	local mustValidateNativeAccount = true
	if type(pMustValidateNativeAccount) == "boolean" then
		mustValidateNativeAccount = pMustValidateNativeAccount
	end

	LogInfo("[AGENT.MODELS.SaveAccount] Updating Account campaignId %s, account: %s, mustValidateAllegroAccount: %s, mustValidateNativeAccount: %s.", campaignIdWithVCC, account_id, mustValidateAllegroAccount, mustValidateNativeAccount)

	if mustValidateAllegroAccount then
		local allegro_info = { action_id = "SAVE_ACCOUNT", account = account, account_id = account_id, is_new = is_new_account }
		IncludeAllegroOtherInfo(allegro_info, allegroOtherInfo)
		local success, err = UpdateAllegroCampaignIn55( campaignIdWithVCC, ksAgent, allegro_info )
		if not success then
			-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
			--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
			--  la actualizacion en i6.
			if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
				return false, { result = err.type, err = (err.err or err.sub_type) }
			end
		end
	end

	local close_ksAgent
	if not ksAgent then
		ksAgent = GetDatasource().cassandra:CassandraConnection("Agent")
		close_ksAgent = true
	end

	-- Validamos que los paquetes de tipo I o I/O de la cuenta de telefonía, NO se encuentren en otra campaña/cuenta del vcc.
	if mustValidateNativeAccount then
		local success, err = MustUpdateCalli6AccountPackagesRelations(ksAgent, campaignIdWithVCC, account_id, account, ksProv, mustValidateAllegroAccount)
		if not success then
			if close_ksAgent then
				ksAgent:Close()
			end
			LogError("[AGENT.MODELS.SaveAccount] Error Updating i6 Call Account campaignId %s, account: %s\nErr: %s.", campaignIdWithVCC, account_id, json.encode(err))
			return false, err
		end
	end

	local action = (is_new_account and "NEW" or "UPDATE")

	if action == "NEW" then
		--Validamos que el activation_code en cuentas de whatsapp con conector inConcert Cloud API sea valido
		local success, err = MustUpdateWhatsappAccountActivationCode(ksAgent, campaignIdWithVCC, account_id, account)
		if not success then
			if close_ksAgent then
				ksAgent:Close()
			end
			LogError("[AGENT.MODELS.SaveAccount] Error Updating Whatsapp Account campaignId %s, account: %s\nErr: %s.", campaignIdWithVCC, account_id, json.encode(err))
			return false, err
		end
	end

	-- actualizo el indice de skills x cuenta
	UpdateSkillsAccount( ksAgent, campaignIdWithVCC, account_id, account, action)

	-- actualizo el indice de firma x cuenta de mail
	UpdateSignatureAccount( ksAgent, campaignIdWithVCC, account_id, account, action)

	-- actualizo el indice del process flow x cuenta
	UpdateFlowAccount( ksAgent, campaignIdWithVCC, account_id, account, action)

	-- actualizo la relación de la campaña de Call de i6 (sin allegro), con los paquetes (vccpackages)
	if mustValidateNativeAccount then
		local currentAccount = GetAccount(account_id, campaignIdWithVCC, ksAgent)
		UpdateCalli6AccountPackagesRelations(ksAgent, campaignIdWithVCC, account_id, account, currentAccount, ksProv, mustValidateAllegroAccount)

		-- actualizo la relacion de blacklists con la cuenta
		UpdateBlacklistsAccount(ksAgent, campaignIdWithVCC, account_id, account, currentAccount, action)
	end

	-- Acualizo la CF CampaignAccounts
	ksAgent:InsertSuperColumn(campaignIdWithVCC, "CampaignAccounts", account_id, "id", account_id)
	for key, value in pairs(account) do
		ksAgent:InsertSuperColumn(
			campaignIdWithVCC,
			"CampaignAccounts",
			account_id,
			key,
			type(value) ~= "table" and tostring(value) or json.encode(value)
		)
	end
	ksAgent:InsertSuperColumn(campaignIdWithVCC, "CampaignAccounts", account_id,
		is_new_account and "created_at" or "updated_at",
		LuaOrm.Date():toString()
	)

	-- Acualizo la CF indice AccountsById
	ksAgent:InsertColumn(account_id, 'AccountsById', campaignIdWithVCC, json.encode(account))

	if close_ksAgent then
		ksAgent:Close()
	end

	-- actualizo la informacion de skills en campaign de la account para la utilizacion en el AgentService
	-- arreglar nombres de propiedades
	local skills = {}
	for _, skill in ipairs(account.Skills or {}) do
		table.insert(skills, { name = skill.skill, weight = skill.weight, required = skill.required })
	end
	agentservice:GetAgentClient():UpdateAccountSkills(account_id, skills)

	-- actualizo la informacion de asignacion attention levels en campaign de la account para la utilizacion en el AgentService
	SaveAttentionLevelsAccountUpdateAgentService( campaignIdWithVCC, account_id, account, is_new_account )

	-- actualizo entrada en redis
	local activeKey = "ActiveAccountsByChannel:" .. account.proxyType .. "@" .. inconcert.authUser.virtualContactCenter
	if account.state == "1" then
		GetDatasource().redis:hset(activeKey, account_id, campaignIdWithVCC)
	elseif account.state == "0" then
		GetDatasource().redis:hdel(activeKey, account_id)
	end

	return true
end

local function NotifyCampaignChanged (campaignIdWithVCC, action)
	local _, vcc = helpers.SeparateVCC(campaignIdWithVCC)
	if not vcc or (action ~= "ADD" and action ~= "REMOVE" and action ~= "UPDATE") then
		error("Invalid parameters "..tostring(campaignIdWithVCC).." action "..tostring(action))
		return
	end
	local campaign = Get(campaignIdWithVCC) or { Id = campaignIdWithVCC }
	if action == "REMOVE" then
		campaign.deleted = true
	end
	local message = {
		campaign = campaign,
		action = action,
	}
	GetDatasource().redis:publish("Agent::CampaignChannel", json.encode(message))
end

local RESERVE_ACCOUNT_KEY = "Agent::Account"
function ReserveAccount (account)
	local accountId = GetAccountId(account)
	return Lock(RESERVE_ACCOUNT_KEY, accountId)
end

function UnreserveAccount (account)
	local accountId = GetAccountId(account)
	return Unlock(RESERVE_ACCOUNT_KEY, accountId)
end

function CampaignExists (campaignAtVCC, keyspace)
	assert(type(campaignAtVCC) == "string")
	local close_connection = false
	if not keyspace then
		close_connection = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end
	campaignAtVCC = campaignAtVCC:lower()
	local res, column = pcall(keyspace.GetColumnValue, keyspace, campaignAtVCC, "Campaigns", "Id")
	if close_connection then
		keyspace:Close()
	end
	return res and column
end

function IsMember (campaignAtVCC, userAtVCC, keyspace)
	local release = false
	if not keyspace then
		release = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end
	local res, column = pcall(keyspace.GetColumn, keyspace, campaignAtVCC, "CampaignMembers", userAtVCC)
	if release then
		keyspace:Close()
	end
	if res then
		return column and column.value
	end
end

--[[
returns all the users belonging to a campaign  with its correspoding role(can be agent, member or moderator)
{
user:role,
user:role
}
]]--
function GetCampaignMembers (campaignAtVCC, keyspace)
	local result = {}
	local release = false
	if not keyspace then
		release = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local campaignMembers = keyspace:GetSliceRange(campaignAtVCC, 
		"CampaignMembers", 
		"",
		{
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 50000
			}
		}
	) 
	if campaignMembers then
		for i, column in ipairs(campaignMembers) do
			result[column.name] = column.value
		end
	end
	if release then
		keyspace:Close()
	end
	return result
end

function GetCampaignLabels (campaignIdWithVCC, keyspace)
	local release
	if not keyspace then
		release = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end
	local rows = keyspace:GetSuperRangeSlices("CampaignLabels",  --column family
		"",  --supercolumn
		campaignIdWithVCC, --startkey
		campaignIdWithVCC, --endkey
		{
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 200
			}
		},
		200 --total keys
	) or {}
	local entries = {}
	for key, campaignLabels in rows:Iterator() do
		if (key == campaignIdWithVCC) then
			--estoy en la que pedi, no deberia volver mas de un row anyway
			for k, v in pairs(campaignLabels) do
				local campaignLabel = {}
				--v es la supercolumn de cada entrada con la pareja name/subcolumns
				for colid, column in pairs(v.columns) do
					campaignLabel[column.name] = column.value
				end
				entries[v.name] = campaignLabel
			end
			break
		end
	end
	if release then
		keyspace:Close()
	end
	return entries
end

local function GetCampaignSchedulerSpecialDays (campaignIdWithVCC, keyspace)
	local specialDays = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "CampaignSchedulersSpecialDays",
		key = campaignIdWithVCC
	} or {}
	return specialDays
end

function GetCampaignScheduler (campaignIdWithVCC, keyspace)
	local closeKeyspace = false
	if not keyspace then
		closeKeyspace = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end
	local schedulerColumns = keyspace:GetSliceRange(campaignIdWithVCC, 
		"CampaignSchedulers",
		"", {
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = true,
				count = 300
			}
		}
	)
	local activationScheduler = {}
	if schedulerColumns then
		activationScheduler.scheduleDays = {}
		for k, v in pairs(schedulerColumns) do
			local scheduler = {}
			local day, startDate, endDate = string.match(v.name, "(.*)_(.*)_(.*)")
			scheduler.startTime = startDate
			scheduler.endTime = endDate
			scheduler.dayOfWeek = day
			table.insert(activationScheduler.scheduleDays, scheduler)
		end
		-- Aca cargo los special days
		local specialDays = GetCampaignSchedulerSpecialDays(campaignIdWithVCC, keyspace)
		if specialDays then
			activationScheduler.scheduleSpecialDays = specialDays.scheduleSpecialDays
			activationScheduler.scheduleSpecialDayAllYearss = specialDays.scheduleSpecialDayAllYearss
		end
	end
	if closeKeyspace then
		keyspace:Close()
	end
	return activationScheduler
end

function GetCannedMessages (campaignIdWithVCC, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "CampaignCannedMessages",
		key = campaignIdWithVCC
	} or {}
end

function GetAttentionLevels (campaignIdWithVCC, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "CampaignAttentionLevels",
		key = campaignIdWithVCC
	} or {}
end

function GetCampaignThresholds (campaignIdWithVCC, keyspace)
	local closeKeyspace = false
	if not keyspace then
		closeKeyspace = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local all_thresholds = {}
	local channels = {"CALL", "CHAT", "FACEBOOK", "INSTAGRAM", "YOUTUBE", "APPSTORE", "GOOGLEPLAY", "RCS", "IM", "MAIL", "SMS", "WHATSAPP", "TWITTER", "WEBCONTACT"}
	for i, channel in ipairs(channels) do
		local threshold_key = string.format("%s:%s", channel, campaignIdWithVCC)
		local thresholds_by_channel = keyspace:GetSliceRange(
			threshold_key, 
			"CampaignThresholds",
			"", {
				column_names = {},
				slice_range = {
					start = "",
					finish = "",
					reversed = true,
					count = 300
				}
			}
		)
		for i, column in pairs(thresholds_by_channel) do
			local th_name  = column.name
			local th_value = column.value
			all_thresholds[channel] = all_thresholds[channel] or {}
			all_thresholds[channel][th_name] = tonumber(th_value)
		end
	end

	if closeKeyspace then
		keyspace:Close()
	end
	return all_thresholds
end

---
-- Returns all the campaigns that the user belongs to. with its corresponding role.
-- The role can be agent, member, administrator, supervisor, etc
-- { campaign: role, campaign: role }
--
--	keyspace          (opcional)  el keyspace de quien me invoca, es utilizado para utilizar la misma conexion a cassandra de quien me ha invocado.
--
function GetUserCampaigns (user, vcc, keyspace)
	local result = {}
	local release = false
	if not keyspace then
		release = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local id = user .. "@" .. vcc

	local userCampaigns = keyspace:GetSliceRange(id,
		"UserCampaigns",
		"", {
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 5000
			}
		}
	)
	if userCampaigns then
		for _, column in ipairs(userCampaigns) do
			result[column.name] = column.value
		end
	end

	if release then
		keyspace:Close()
	end

	return result
end

function GetUserCampaignsWithHelpdeskIntegration (vcc, simpleUserId, keyspace)
	local campaigns = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "UserCampaigns",
		key = utils.AppendVcc(simpleUserId, vcc)
	} or {}

	local result = {}
	for k in pairs(campaigns) do

		local campInfo = utils.GetAllColumns({
			keyspace     = keyspace or "Agent",
			columnFamily = "Campaigns",
			key          = k,
			columns = {"Integration"}
		}) or {}
		if campInfo.Integration and campInfo.Integration.othersIntegrations and campInfo.Integration.othersIntegrations.HELPDESK 
		and campInfo.Integration.othersIntegrations.HELPDESK.Enable then
			table.insert( result, k)
		end
	end
	return result
end

function SaveAccountStatusCache (statuses)
	GetDatasource().redis:set("Agent::AccountStatuses", statuses)
	GetDatasource().redis:publish("Agent::AccountStatusesChannel", statuses)
end

--Los campos sensibles de la account se guardan cifrados
function PrepareAccountForWrite (campaignIdWithVCC, account)
	if SensibleData[account.proxyType] then
		for k, v in pairs (account) do
			if SensibleData[account.proxyType][k] then
				local ciphered, err = m_ksCore:Cipher (campaignIdWithVCC, account[k])
				if not ciphered then
					v = "***************" --para no quemar la password en el stacktrace
					error ("Error ciphering data: "..err)
				end
				account[k] = ciphered
			end
		end
	end
end

--Los campos sensibles de la account nunca se devuelven
function PrepareAccountForRead (account)
	if SensibleData[account.proxyType] then
		for k, v in pairs (account) do
			if SensibleData[account.proxyType][k] then
				account[k] = nil
			end
		end
	end
end

local campaign__index = function(campaign, attribute)
	if attribute == "dispositions" then
		local dispositions = getmetatable(campaign).dispositions
		if not dispositions then
			return
		end
		dispositions = json.decode(dispositions)
		rawset(campaign, "dispositions", dispositions)
		return dispositions
	end
	return nil
end

local function DecodeAccountInformation ( account )
	if not account then
		return
	end
	if is_string(account.ContactManagement) then
		account.ContactManagement = json.decode(account.ContactManagement)
	end
	if is_string(account.FlowInformation) then
		account.FlowInformation = json.decode(account.FlowInformation)
	end
	if is_string(account.TicketEmitter) then
		local t = json.decode(account.TicketEmitter)
		if is_empty(t) then
			account.TicketEmitter = nil
		else
			account.TicketEmitter = t
		end
	end
	if is_string(account.Skills) then
		account.Skills = json.decode(account.Skills)
	end
	if is_string(account.Survey) then
		local t = json.decode(account.Survey)
		if is_empty(t) then
			account.Survey = nil
		else
			account.Survey = t
		end
	end
	if is_string(account.AttentionLevels) then
		account.AttentionLevels = json.decode(account.AttentionLevels)
	end
	if is_string(account.QnAMaker) then
		account.QnAMaker = json.decode(account.QnAMaker)
	end
	if is_string(account.HSMConfig) then
		account.HSMConfig = json.decode(account.HSMConfig)
	end
	if is_string(account.incomingConfig) then
		account.incomingConfig = json.decode(account.incomingConfig)
	end
	if is_string(account.outgoingConfig) then
		account.outgoingConfig = json.decode(account.outgoingConfig)
	end
	if is_string(account.packages) then
		account.packages = json.decode(account.packages)
	end
	if is_string(account.GUI) then
		account.GUI = json.decode(account.GUI)
	end
	return account
end

function GetCampaignAccounts (campaignIdWithVCC, keyspace)
	local release = false
	if not keyspace then
		release = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local rows = keyspace:GetSuperRangeSlices("CampaignAccounts",  --column family
		"",  --supercolumn
		campaignIdWithVCC, --startkey
		campaignIdWithVCC, --endkey
			{
				column_names = {},
				slice_range = {
					start = "",
					finish = "",
					reversed = false,
					count = 2000
				}
			},
		2000 --total keys
	) or {}

	local accounts = {}
	for key, campaignAccounts in rows:Iterator() do
		if key == campaignIdWithVCC then
			--estoy en la que pedi, no deberia volver mas de un row anyway
			for k, v in pairs(campaignAccounts) do
				local account = {}
				--v es la supercolumn de cada entrada con la pareja name/subcolumns
				for colid, column in pairs(v.columns) do
					account[column.name] = column.value
				end
				PrepareAccountForRead(account)
				account = DecodeAccountInformation(account)
				--las accounts tiene un id v.name que no me interesa, armo array nomas
				table.insert(accounts, account)
			end
			break
		end
	end

	if release then
		keyspace:Close()
	end

	return accounts
end

function GetAccount (accountId, campaignIdWithVCC, keyspace)
	local accounts = GetCampaignAccounts(campaignIdWithVCC, keyspace)
	for _, account in pairs(accounts) do
		if account.id == accountId then
			return account
		end
	end
end

--[[
	Returns the campaign with account information included
]]--
function Get (campaignIdWithVCC, keyspace)
	local campaign = {}
	local release = false
	if not keyspace then
		release = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local campaignColumns = keyspace:GetSliceRange(campaignIdWithVCC,
		"Campaigns",
		"", {
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = true,
				count = 2000
			}
		}
	)

	if campaignColumns then
		for k, v in ipairs(campaignColumns) do
			--campaign[v.name] = utils.JSON_Decode(v.value)
			if v.name == "Integration" or v.name == "InteractionNotifications" then
				campaign[v.name] = json.decode(v.value)
			else
				campaign[v.name] = v.value
			end
		end
		if not campaign.Id then
			--this is a stub!
			if release then
				keyspace:Close()
			end
			return
		end
		campaign.Avatar = nil --TODO this is UGLY, fix it

		setmetatable(campaign, {dispositions = campaign.dispositionCodes, __index = campaign__index} )

		-- campaign accounts
		campaign.accounts        = GetCampaignAccounts( campaignIdWithVCC, keyspace )
		campaign.timezoneOffset  = utils.GetTimeZoneInstance():CalculateTimezoneOffsetMin(campaign.ValidationTimeZone, keyspace)
		campaign.scheduler       = GetCampaignScheduler(campaignIdWithVCC, keyspace)
		campaign.Thresholds      = GetCampaignThresholds(campaignIdWithVCC, keyspace)
		campaign.CannedMessages  = GetCannedMessages(campaignIdWithVCC, keyspace)
		campaign.AttentionLevels = GetAttentionLevels(campaignIdWithVCC, keyspace)
		campaign.Labels          = GetCampaignLabels(campaignIdWithVCC, keyspace)
	end
	if release then
		keyspace:Close()
	end
	return campaign
end

function HasDispositionCodes (campaign)
	assert(campaign, "campaign is required")
	if not campaign.dispositions then
		return false
	end
	return #campaign.dispositions > 0
end

function FindDisposition (dispositions, dispositionTreePath, index)
	if not dispositions or #dispositions == 0 or not dispositionTreePath or #dispositionTreePath == 0 then
		return false
	end
	index = index or 1
	for _, disposition in ipairs(dispositions) do
		if disposition.name == dispositionTreePath[index] then
			if index == #dispositionTreePath then
				if not disposition.childs or #disposition.childs == 0 then
					return disposition
				end
				return false
			end
			return FindDisposition(disposition.childs, dispositionTreePath, index + 1)
		end
	end
	return false
end

--[[
	Saves an avatar for the campaign
]]--
function SaveAvatar (campaign, avatar)
	local keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	keyspace:InsertColumnWithSize(campaign.Id, 'Campaigns', 'Avatar', avatar.image, avatar.size)
	keyspace:Close()
end

--[[
	Saves only the attributes of the campaign, doesn't touch accounts or members
]]--
function ShallowSave (campaign, keyspace, allegroOtherInfo)
	local release_keyspace = false
	if not keyspace then
		release_keyspace = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local allegro_info = {action_id = "BASIC_INFO", campaign = campaign}
	IncludeAllegroOtherInfo(allegro_info, allegroOtherInfo)

	local success, err = UpdateAllegroCampaignIn55( campaign.Id, keyspace, allegro_info )
	if not success then
		-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
		--  la actualizacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			if release_keyspace then
				keyspace:Close()
			end
			return false, { result = err.type, err = (err.err or err.sub_type) }
		end
	end

	local campaignIdWithVCC = campaign.Id
	for k, v in pairs(campaign) do
		if (type(v) ~= "table") then
			keyspace:InsertColumn(campaignIdWithVCC, 'Campaigns', k, tostring(v))
		end
	end

	if release_keyspace then
		keyspace:Close()
	end

	NotifyCampaignChanged(campaignIdWithVCC, "UPDATE")

	return true
end

function SaveThresholds (campaignIdWithVCC, all_thresholds, keyspace, allegroOtherInfo)
	if type(campaignIdWithVCC) ~= "string" or type(all_thresholds) ~= "table" then
		LogError("models.SaveThresholds - Error saving thresholds - campaignId %s, thresholds: %s", campaignIdWithVCC, all_thresholds)
		return false
	end

	local closeKeyspace = false
	if not keyspace then
		closeKeyspace = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	local camp_exists = CampaignExists(campaignIdWithVCC, keyspace)
	if not camp_exists then
		LogError("models.SaveThresholds - Error saving thresholds - campaignId %s, does not exist", campaignIdWithVCC)
		if closeKeyspace then
			keyspace:Close()
		end
		return false
	end

	local allegro_info = { action_id = "SAVE_THRESHOLDS", thresholds = all_thresholds }
	IncludeAllegroOtherInfo(allegro_info, allegroOtherInfo)

	local success, err = UpdateAllegroCampaignIn55( campaignIdWithVCC, keyspace, allegro_info )
	if not success then
		-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
		--  la actualizacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			return false, { result = err.type, err = (err.err or err.sub_type) }
		end
	end

	keyspace:RemoveColumn(campaignIdWithVCC, "CampaignThresholds")
	for channel, thresholds_by_channel in pairs(all_thresholds) do
		local threshold_key = string.format("%s:%s", channel, campaignIdWithVCC)
		keyspace:RemoveColumn(threshold_key, "CampaignThresholds")
		for th_name, th_value in pairs(thresholds_by_channel) do
			keyspace:InsertColumn(threshold_key, "CampaignThresholds", th_name, tostring(th_value))
		end
	end

	if closeKeyspace then
		keyspace:Close()
	end
	return true
end

function ReserveName (campaignIdWithVCC, userName)
	return GetDatasource().redis:hsetnx("Agent::CampaignNames", campaignIdWithVCC, userName)
end

function UnreserveName (campaignIdWithVCC)
	return GetDatasource().redis:hdel("Agent::CampaignNames", campaignIdWithVCC)
end

function RemoveAccountsSkills( keyspace, campaignIdWithVCC, account )
	LogInfo("[campaigns.RemoveAccountsSkills] Removing account skills. Campaign: '%s', Account: '%s'.", campaignIdWithVCC, account.id )
	UpdateSkillsAccount( keyspace, campaignIdWithVCC, account.id, account, "REMOVE" )
	agentservice:GetAgentClient():DeleteAccountSkills(account.id)
end

function RemoveAccountsSignature( keyspace, campaignIdWithVCC, account )
	LogInfo("[campaigns.RemoveAccountsSignature] Removing account signature. Campaign: '%s', Account: '%s'.", campaignIdWithVCC, account.id )
	UpdateSignatureAccount( keyspace, campaignIdWithVCC, account.id, account, "REMOVE" )
end

function RemoveAccountsFlow( keyspace, campaignIdWithVCC, account )
	LogInfo("[campaigns.RemoveAccountsFlow] Removing account flow. Campaign: '%s', Account: '%s'.", campaignIdWithVCC, account.id )
	UpdateFlowAccount( keyspace, campaignIdWithVCC, account.id, account, "REMOVE" )
end

function CleanFacebookPagesData (accountUid)
	GetDatasource().redis:del("FBPROXY:uid:" .. accountUid .. ":pages")
end

function AddFacebookAccountByPage(campaignIdWithVCC, accountId, page)
	GetDatasource().redis:set("FacebookAccountByPage:" .. page, json.encode({ campaign = campaignIdWithVCC, account = accountId }))
end

function RemoveFacebookAccountByPage(page)
	GetDatasource().redis:del("FacebookAccountByPage:" .. page)
end

function CleanInstagramPagesData (accountUid)
	GetDatasource().redis:del("INSTAGRAMPROXY:uid:" .. accountUid .. ":pages")
end

function AddInstagramAccountByPage (campaignIdWithVCC, accountId, page, ksAgent)
	local campaignId, vcc = helpers.SeparateVCC(campaignIdWithVCC)
	local pageAssignmentStr = GetDatasource().redis:get("InstagramAccountByPage:" .. page)
	local pageAssignment = pageAssignmentStr and json.decode(pageAssignmentStr or "{}")
	if pageAssignment and pageAssignment.campaign and pageAssignment.account then
		local currentCampaign = Get(pageAssignment.campaign, ksAgent)
		local currentcampaignId, currentVcc = helpers.SeparateVCC(pageAssignment.campaign)
		local currentAccountId = pageAssignment.account
		local currentAccount = GetAccount(currentAccountId, pageAssignment.campaign, ksAgent)

		if currentCampaign and currentAccount and currentAccount.state == "1" then
			local pageName = ""
			if currentAccount and currentAccount.pagesNames then
				pageNames = type(currentAccount.pagesNames) == "string" and json.decode(currentAccount.pagesNames) or currentAccount.pagesNames
				pageName = pageNames[page] or ""
			end
			local err = { page_id = page, page_name = pageName, campaign = currentcampaignId, campaign_name = currentCampaign.Name, account_id = currentAccountId, account_name = currentAccount.name}

			if vcc == currentVcc and campaignId == currentcampaignId and currentAccountId ~= accountId then
				err.result = "INSTAGRAM_USER_IN_USE_IN_ANOTHER_ACCOUNT_CURRENT_CAMPAIGN"
				err.reason = string.format("Unable to save instagram account '%s' on campaign '%s'. Instagram page id: '%s', name: '%s'  in use in account '%s' of the current campaign.", accountId, campaignIdWithVCC, page, pageName, currentcampaignId)
				return err
			elseif vcc == currentVcc and campaignId ~= currentcampaignId then
				err.result = "INSTAGRAM_USER_IN_USE_IN_ANOTHER_CAMPAIGN"
				err.reason = string.format("Unable to save instagram account '%s' on campaign '%s'. Instagram page id: '%s', name: '%s'  in use in account '%s' of the '%s' campaign.", accountId, campaignIdWithVCC, page, pageName, currentAccountId, currentcampaignId)
				return err
			elseif vcc ~= currentVcc then
				err.result = "INSTAGRAM_USER_IN_USE_IN_ANOTHER_VCC"
				err.reason = string.format("Unable to save instagram account '%s' on campaign '%s'. Instagram page id: '%s', name: '%s'  in use in account '%s' of the '%s' campaign of vcc '%s'.", accountId, campaignIdWithVCC, page, pageName, currentAccountId, currentcampaignId, currentVcc)
				return err
			end
		end
	end
	GetDatasource().redis:set("InstagramAccountByPage:" .. page, json.encode({ campaign = campaignIdWithVCC, account = accountId }))
end

function RemoveInstagramAccountByPage (campaignIdWithVCC, accountId, page, ksAgent)
	local campaignId, vcc = helpers.SeparateVCC(campaignIdWithVCC)
	local pageAssignmentStr = GetDatasource().redis:get("InstagramAccountByPage:" .. page)
	local pageAssignment = pageAssignmentStr and json.decode(pageAssignmentStr or "{}")
	if pageAssignment and pageAssignment.campaign and pageAssignment.account then
		local currentcampaignId, currentVcc = helpers.SeparateVCC(pageAssignment.campaign)
		local currentAccountId = pageAssignment.account
		local currentAccount = GetAccount(currentAccountId, pageAssignment.campaign, ksAgent)
		if currentAccount then
			if vcc ~= currentVcc or campaignId ~= currentcampaignId or currentAccountId ~= accountId then
				return
			end
		end
	end
	GetDatasource().redis:del("InstagramAccountByPage:" .. page)
end

function DeleteAccountStatistics (id, keyspace)
	utils.RemoveColumns {
		keyspace = keyspace or "Interactions",
		columnFamily = "SocialNetworkAccounts",
		rows = {
			[id] = { "statistics" }
		}
	}
	utils.RemoveColumns {
		keyspace = keyspace or "Interactions",
		columnFamily = "OutgoingBatchesAccounts",
		rows = {
			[id] = { "statistics" }
		}
	}
end

function FacebookCleanMessengerConfiguration(accountId, accessToken, pages)
	LogInfo("Cleaning messenger configuration for facebook pages %s in account %s", json.encode(pages), accountId)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:CleanMessengerConfiguration({
		proxyType = "FACEBOOK",
		account = accountId, 
		accessToken = accessToken, 
		pages = pages
	})
	if result.result ~= "OK" then
		LogError("Error cleaning messenger configuration for facebook pages %s in account %s: %s",
			json.encode(pages), accountId, json.encode(result))
		return false, { result = "ERROR_CLEANING_PAGES_MESSENGER", reason = result.reason }
	end
	LogInfo("End cleaning messenger configuration for facebook pages %s in account %s", json.encode(pages), accountId)
	return true
end

function RegisterUnregisterPagesWebhooks(proxyType, accountId, accessToken, registerPages, unregisterPages)
	LogInfo("Registering webhooks for %s pages %s and unregistering for pages %s in account %s",
		proxyType, json.encode(registerPages), json.encode(unregisterPages), accountId)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:RegisterUnregisterWebhookPages({
		proxyType = proxyType,
		account = accountId, 
		accessToken = accessToken, 
		registerPages = registerPages, 
		unregisterPages = unregisterPages
	})
	if result.result ~= "OK" then
		LogError("Error registering webhooks for %s pages %s and unregistering for pages %s in account %s: %s",
			proxyType, json.encode(registerPages), json.encode(unregisterPages), accountId, json.encode(result))
		return false, { result = "ERROR_REGISTERING_UNREGISTERING_PAGES", reason = result.reason }
	end
	LogInfo("End registering webhooks for %s pages %s and unregistering for pages %s in account %s",
		proxyType, json.encode(registerPages), json.encode(unregisterPages), accountId)
	return true
end

function RegisterUnregisterWebhook(account, register, old)
	LogInfo("%s webhooks for account %s", register and "Registering" or "Unregistering", account.id)
	-- Hago este hack porque para cuentas con integrationType ICC_CLOUD_API esta dumpeando el WH al pasarle los datos de account y de old, y de todas formas no se usa el HSMConfig para RegisterAccountWebhook / UnregisterAccountWebhook
	local account_copy = json.decode(json.encode(account))
	local old_copy = old and json.decode(json.encode(old)) or nil
	
	if account_copy.integrationType == "ICC_CLOUD_API" then
		--#19211: Si la cuenta esta desactivada no se hace el unregister, porque podriamos tener otra cuenta activa en otra campana, vcc o ambiente con el mismo numero de telefono y la estariamos dejando huerfana en el webhook
		if not register and account.state == "0" then
			return true
		end
		account_copy.HSMConfig = nil
		if old_copy then
			old_copy.HSMConfig = nil
		end
	end

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink[register and "RegisterAccountWebhook" or "UnregisterAccountWebhook"](irLink, {
		proxyType = account_copy.proxyType,
		account = account_copy, 
		old = old_copy
	})
	if result.result ~= "OK" then
		LogError("Error %s webhooks for account %s: %s",
			register and "registering" or "unregistering", account.id, json.encode(result))
		return false, { result = "ERROR_REGISTERING_UNREGISTERING_PAGES", reason = result.reason }
	end
	LogInfo("End %s webhooks for account %s", register and "registering" or "unregistering", account.id)
	return true
end

local removeAccountChannelTasks_d = {
	CALL = function(keyspace, campaignIdWithVCC, account, _removedAccountData)
		local vcc = inconcert.authUser.virtualContactCenter
		if utils.HasAllegroIntegration(vcc, keyspace) then
			return
		end

		RemoveAccountDidsPool(account.id)
	end,
	FACEBOOK = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		--clean account by page
		local pages = account.pages or {}
		if type( pages ) == "string" then
			pages = json.decode(pages)
		end
		for _, pageId in pairs(pages or {}) do
			RemoveFacebookAccountByPage(pageId)
		end
		--clean messenger configuration
		if removedAccountData and removedAccountData.mustCleanMessenger then
			local removedPages = {}
			for _, pageId in pairs(pages or {}) do
				removedPages[pageId] = true
			end
			FacebookCleanMessengerConfiguration(account.id, account.accessToken, removedPages)
		end
		--unregister pages
		local pagesMessenger = account.pagesMessenger or {}
		if type( pagesMessenger ) == "string" then
			pagesMessenger = json.decode(pagesMessenger)
		end
		local unregisterPages = {}
		for pageId, pageData in pairs(pagesMessenger or {}) do
			if pageData and pageData.processFlow then
				unregisterPages[pageId] = true
			end
		end
		if #unregisterPages then
			RegisterUnregisterPagesWebhooks("FACEBOOK", account.id, account.accessToken, {}, unregisterPages)
		end
		--delete statistics
		local keyspaceInteractions = utils.GetDatasource():CassandraConnection( "Interactions" )
		local statistics = utils.GetAllColumns {
			keyspace = keyspaceInteractions,
			columnFamily = "SocialNetworkAccounts",
			columns = {},
			startKey = ""
		}
		for id, _ in pairs(statistics) do
			if id:match(account.id .. ":") then
				DeleteAccountStatistics(id, keyspaceInteractions)
			end
		end
		keyspaceInteractions:Close()
	end,
	INSTAGRAM = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		--clean account by page
		local pagesFbIds = account.pagesFbIds or {}
		if type( pagesFbIds ) == "string" then
			pagesFbIds = json.decode(pagesFbIds)
		end
		for igId, fbId in pairs(pagesFbIds or {}) do
			RemoveInstagramAccountByPage(campaignIdWithVCC, account.id, igId)
		end
		--unregister pages
		if #pagesFbIds then
			RegisterUnregisterPagesWebhooks("INSTAGRAM", account.id, account.accessToken, {}, pagesFbIds)
		end
		--delete statistics
		local keyspaceInteractions = utils.GetDatasource():CassandraConnection( "Interactions" )
		local statistics = utils.GetAllColumns {
			keyspace = keyspaceInteractions,
			columnFamily = "SocialNetworkAccounts",
			columns = {},
			startKey = ""
		}
		for id, _ in pairs(statistics) do
			if id:match(account.id .. ":") then
				DeleteAccountStatistics(id, keyspaceInteractions)
			end
		end
		keyspaceInteractions:Close()
	end,
	TWITTER = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		DeleteAccountStatistics(account.id)
		RegisterUnregisterWebhook(account, false)
	end,
	YOUTUBE = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		DeleteAccountStatistics(account.id)
		RegisterUnregisterWebhook(account, false)
	end,
	APPSTORE = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		RegisterUnregisterWebhook(account, false)
	end,
	GOOGLEPLAY = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		RegisterUnregisterWebhook(account, false)
	end,
	RCS = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		DeleteAccountStatistics(account.id)
		RegisterUnregisterWebhook(account, false)
	end,
	WHATSAPP = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		DeleteAccountStatistics(account.id)
		RegisterUnregisterWebhook(account, false)
	end,
	WEBCONTACT = function( keyspace, campaignIdWithVCC, account, removedAccountData )
		DeleteAccountToken(account.Token)
	end
}

function RemoveAccountChannelTasks( keyspace, campaignIdWithVCC, account, removedAccountData )
	if removeAccountChannelTasks_d[account.proxyType] then
		removeAccountChannelTasks_d[account.proxyType]( keyspace, campaignIdWithVCC, account, removedAccountData )
	end
	local activeKey = "ActiveAccountsByChannel:" .. account.proxyType .. "@" .. inconcert.authUser.virtualContactCenter
	GetDatasource().redis:hdel(activeKey, account.id)
end

local function __Update (campaign, ksAgent, allegroOtherInfo)
	local campaignIdWithVCC = campaign.Id
	local currentCampaign = Get(campaignIdWithVCC, ksAgent)
	local is_new_campaign = not currentCampaign and true or false

	ksAgent:InsertColumn(campaignIdWithVCC, "Campaigns", 
		is_new_campaign and "created_at" or "updated_at",
		LuaOrm.Date():toString()
	)

	for k, v in pairs(campaign) do
		if type(v) ~= "table" then
			ksAgent:InsertColumn(campaignIdWithVCC, 'Campaigns', k, tostring(v))
		end
	end

	--vcc campaigns needs the id without vcc
	local campaignIdWithOutVCC, campaignVCC = helpers.SeparateVCC(campaignIdWithVCC)
	ksAgent:InsertColumn(campaign.VCC, 'VCCCampaigns', campaignIdWithOutVCC, campaign.Name)

	local currentAccounts = {}
	for k, account in pairs(campaign.accounts or {}) do
		local accountId = GetAccountId(account)
		if accountId then
			currentAccounts[accountId] = true
			PrepareAccountForWrite (campaignIdWithVCC, account)
			SaveAccount(ksAgent, campaignIdWithVCC, accountId, account, is_new_campaign, allegroOtherInfo)
		end
	end

	local scheduler = campaign.scheduler
	if scheduler then
		ksAgent:RemoveColumn(campaignIdWithVCC, 'CampaignSchedulers')
		for _, scheduleDay in pairs(scheduler.scheduleDays or {}) do
			local schedKey = string.format( "%s_%s_%s", scheduleDay.dayOfWeek, scheduleDay.startTime, scheduleDay.endTime)
			ksAgent:InsertColumn(campaignIdWithVCC, 'CampaignSchedulers', schedKey, "")
		end

		-- Aqui almacenamos el Special Days ----------
		local scheduleSpecialDays = scheduler.scheduleSpecialDays or {}
		local scheduleSpecialDayAllYearss = scheduler.scheduleSpecialDayAllYearss or {}

		local specialDays = {
			scheduleSpecialDays = json.encode(scheduleSpecialDays),
			scheduleSpecialDayAllYearss = json.encode(scheduleSpecialDayAllYearss)
		}

		utils.SaveRows {
			keyspace     = ksAgent or "Agent",
			columnFamily = "CampaignSchedulersSpecialDays",
			removeBefore = true,
			rows = {
				[campaignIdWithVCC] = specialDays
			}
		}
	end

	if currentCampaign then
		local counter_acc_by_channel = {} -- aca se encuentra la cantidad de cuentas por channel
		local ksProv = GetDatasource().cassandra:CassandraConnection("Provisioning")

		for i, account in ipairs(currentCampaign.accounts) do
			local proxy = account.proxyType
			counter_acc_by_channel[ proxy ] = ( counter_acc_by_channel[ proxy ] or 0 ) + 1 -- incremento una
			if account.id and not currentAccounts[account.id] then
				counter_acc_by_channel[ proxy ] = counter_acc_by_channel[ proxy ] - 1 -- decremento una

				LogWarning("Removing campaign account "..account.id)
				-- tareas particulares del canal
				local removedAccountData = campaign.removedAccounts and campaign.removedAccounts[account.id] or nil
				RemoveAccountChannelTasks( ksAgent, campaignIdWithVCC, account, removedAccountData)

				-- actualizo el indice de skills x cuenta ( los elimino )
				RemoveAccountsSkills( ksAgent, campaignIdWithVCC, account )

				-- actualizo el indice de signature x cuenta ( los elimino )
				RemoveAccountsSignature( ksAgent, campaignIdWithVCC, account )

				-- actualizo la relación de la campaña de Call de i6 (sin allegro), con los paquetes (vccpackages)
				RemoveCalli6AccountPackagesRelations(campaignIdWithVCC, account.id, nil, ksAgent, ksProv)

				-- actualizo el indice de flow x cuenta ( los elimino )
				RemoveAccountsFlow( ksAgent, campaignIdWithVCC, account )

				-- asdfasdfasdfasdf adf asdfasdfasd
				RemoveBlacklistsAccount(ksAgent, campaignIdWithVCC, account.id, account)

				ksAgent:RemoveColumn(account.id, "AccountsById", campaignIdWithVCC)
				ksAgent:RemoveSuperColumn(campaignIdWithVCC, "CampaignAccounts", account.id)
				DeleteAttentionLevelsAccountUpdateAgentService(campaignIdWithVCC, account.id, account)
			end
		end

		ksProv:Close()

		-- aca recorro la cantidad de accounts que quedaron por channel
		--  y en los canales que haya 0 accounts, se debe limpiar la informacion de sus thresholds
		local empty_channel_thresholds
		for proxy, count in pairs( counter_acc_by_channel ) do
			if count == 0 then
				empty_channel_thresholds = empty_channel_thresholds or {}
				empty_channel_thresholds[ proxy ] = {}
			end
		end
		if empty_channel_thresholds then
			-- aca limpio la info de los thresholds
			SaveThresholds( campaignIdWithVCC, empty_channel_thresholds, ksAgent )
		end

		NotifyCampaignChanged(campaignIdWithVCC, "UPDATE")
	else
		NotifyCampaignChanged(campaignIdWithVCC, "ADD")
	end

	return true
end

function CreateCampaign (campaign, keyspace, allegroOtherInfo)
	local release_keyspace = false
	if not keyspace then
		release_keyspace = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end
	local allegro_info = {}
	IncludeAllegroOtherInfo(allegro_info, allegroOtherInfo)
	
	local success, err = CreateAllegroCampaignIn55(campaign, keyspace, allegro_info)
	if not success then
		-- Solo tiene que retornar sin guardar en cassandra, si el error es de tipo "ALLEGRO_INTEGRATION_ERROR",
		--  ya que otro tipo de error, puede ser que no tenga integracion con Allegro y por lo tanto en si, no es un error y debe de continuar
		--  la actualizacion en i6.
		if err and err.type == "ALLEGRO_INTEGRATION_ERROR" then
			if release_keyspace then
				keyspace:Close()
			end
			return false, { result = err.type, err = (err.err or err.sub_type) }
		end
	end

	success = __Update(campaign, keyspace)

	if release_keyspace then
		keyspace:Close()
	end

	return success, err
end

local function __UpdateOmnichannelAccount (campaign, keyspace)
	local release_keyspace = false
	if not keyspace then
		release_keyspace = true
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
	end

	if utils.IsAllegroCampaign( campaign.Id, keyspace ) then
		LogWarning("[AGENT.CAMPAIGNS:__UpdateOmnichannelAccount] - Unable to update allegro campaign '%s'.", campaign.Id)
		if release_keyspace then
			keyspace:Close()
		end
		return false
	end

	local success, err = __Update(campaign, keyspace)
	if release_keyspace then
		keyspace:Close()
	end

	return success, err
end

function ChangeStateOmnichannelAccount (campaign, keyspace)
	LogInfo("[AGENT.CAMPAIGNS:ChangeStateOmnichannelAccount] campaignId: %s", campaign.Id)
	return __UpdateOmnichannelAccount(campaign, keyspace)
end

function RemoveOmnichannelAccount (campaign, keyspace)
	LogInfo("[AGENT.CAMPAIGNS:RemoveOmnichannelAccount] campaignId: %s", campaign.Id)
	return __UpdateOmnichannelAccount(campaign, keyspace)
end

function FindAccountByUID ( accountToFind, keyspace )

	local close_ks = false
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		close_ks = true
	end

	local accounts = keyspace:GetRangeSlices("AccountsById",  --column family
		"",  --supercolumn
		"", --startkey
		"", --endkey
		{ --predicate
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 1000
			}
		},
		1000 --total keys
	)
	if close_ks then
		keyspace:Close()
	end

	if accounts then
		local proxy = accountToFind.proxyType
		local uid = accountToFind.uid or accountToFind.name

		for account_id, accountRow in accounts:Iterator() do
			local campaign_id, account
			for i, v in ipairs( accountRow ) do
				campaign_id = v.name
				account = v.value
				if type( account ) == "string" then
					account = json.decode(account)
				end
				break
			end

			if account then
				local acc_uid = account.uid or account.name
				if proxy == account.proxyType and uid == acc_uid then
					return account, campaign_id
				end
			end
		end
	end

	-- does not exist
	return
end

function FindAccountById (accountToFind, keyspace)
	assert(accountToFind, "account is required")


	local close_keyspace = false
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end

	local accountId = GetAccountId(accountToFind)
	local accountColumns = keyspace:GetSliceRange(accountId, "AccountsById", "", 
		{
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = true,
				count = 2000
			}
		}
	)
	if close_keyspace then
		keyspace:Close()
	end

	local account, campaign
	for i, column in ipairs(accountColumns) do
		campaign = column.name
		account = json.decode(column.value)
		account = DecodeAccountInformation(account)
		break
	end
	return account, campaign
end

---
-- Saves an account for a given campaign
--
-- @tparam string campaignId the id of the campaign
-- @tparam table account a table with account data. Each key is saved as a column in Cassandra.
-- @tparam boolean is_new a flag indicating if this is a new account (which is a hack, I should check in Cassandra)
-- TODO: document what is in the table
--
function SaveCampaignAccount (campaignIdWithVCC, account, is_new, keyspace, allegroOtherInfo)
	is_new = not not is_new -- coerce a possible nil value to a boolean
	local close_keyspace = false
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		close_keyspace = true
	end
	local account_id = GetAccountId(account)

	LogInfo("[AGENT.MODEL.SaveCampaignAccount] Saving campaign account. Campaign: '%s', Account: '%s', Action: '%s'.", campaignIdWithVCC, account_id, (account_id and "Update" or "Create") )

	local success, errInfo
	if account_id then
		PrepareAccountForWrite(campaignIdWithVCC, account)
		success, errInfo = SaveAccount(keyspace, campaignIdWithVCC, account_id, account, is_new, allegroOtherInfo)
		if success then
			NotifyCampaignChanged(campaignIdWithVCC, "UPDATE")
		end
	end

	if close_keyspace then keyspace:Close() end

	if not success then
		LogError("[AGENT.MODEL.SaveCampaignAccount] Unable to save campaign account id for account %s", json.encode(account))
	end

	return success, errInfo
end

function FindAll ()
	local results = {}
	local keyspace = GetDatasource().cassandra:CassandraConnection("Agent")

	--TODO this should be paginating cause ain't working with more than 1000 campaigns configured
	local campaignList = keyspace:GetRangeSlices("Campaigns",  --column family
		"",  --supercolumn
		"", --startkey
		"", --endkey
		{ --predicate
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 1000
			}
		},
		10000 --total keys
	)
	if campaignList then
		for campaignIdWithVCC, campaignRow in campaignList:Iterator() do
			results[campaignIdWithVCC] = Get(campaignIdWithVCC, keyspace)
		end
	end
	keyspace:Close()
	return results
end

function IsAllegroAccount(account)
	return account.proxyType == "CALL"
end

function IsAllegroCampaign(vcc, campaignIdWithVCC, keyspace)
	if not utils.HasAllegroIntegration(vcc, keyspace) then
		return false
	end
	local accounts = GetCampaignAccounts(campaignIdWithVCC, keyspace)
	for _, account in pairs(accounts) do
		if IsAllegroAccount(account) then
			return true
		end
	end
	return false
end

function GetCampaignAccount(campaign, accountId)
	for _, account in pairs(campaign.accounts or {}) do
		if account.id == accountId then
			return account
		end
	end
	return nil
end

function UserIsMember (campaignId, userId, keyspace)
	local closeKeyspace = false
	if not keyspace then
		keyspace = GetDatasource().cassandra:CassandraConnection("Agent")
		closeKeyspace = true
	end

	local res, column = pcall(keyspace.GetColumn, keyspace, campaignId, "CampaignMembers", userId)
	if closeKeyspace then
		keyspace:Close()
	end

	if res then
		return column and column.value
	end
end

function GetAccountSocialNetworkActivePosts(accountId)
	return GetDatasource().redis:hgetall("SocialNetworkActivePosts:" .. accountId)
end

function GetAccountActiveOutgoingBatches(accountId)
	return GetDatasource().redis:hgetall("OutgoingBatchesActive:" .. accountId)
end

function InsertAccountToken(accountToken, campaignAtVCC, account)
	GetDatasource().redis:set("AccountToken:" .. accountToken, 
		json.encode({ campaignAtVCC = campaignAtVCC, account = account }))
end

function DeleteAccountToken(accountToken)
	GetDatasource().redis:del("AccountToken:" .. accountToken)
end

function GetVCCCampaignsAndAccounts (vcc, keyspace)
	assert(type(vcc) == "string")
	assert(keyspace)
	--LogDebug("[AGENT.CAMPAIGNS.GetVCCCampaignsAndAccounts] ENTRA-----------------------> loading all campaigns in vcc '%s'", vcc)
	local campaigns = {}
	local vccCampaigns = GetVCCCampaigns(vcc, keyspace)
	for campaignId in pairs(vccCampaigns or {}) do
		local fullCampaignId = utils.AppendVcc(utils.GetId(campaignId), vcc)
		local campaign = Get(fullCampaignId, keyspace)
		table.insert(campaigns, campaign)
	end
	--LogDebug("[AGENT.CAMPAIGNS.GetVCCCampaignsAndAccounts] SALE-----------------------> loaded all campaigns in vcc '%s'. campaigns: %s", vcc, json.encode(campaigns))
	return campaigns
end

local function AccountDidsPoolKey(accountId, section)
	return string.format("AccountDidsPool::%s::%s", accountId, section)
end

local function SplitDidsPoolByStatus(accountId, didsPool)
	local byStatus = { ACTIVE = {}, DISABLED = {}, REJECTED = {} }
	for _, did in pairs(didsPool) do
		if did.status then
			local statusKey = did.status:upper()
			byStatus[statusKey] = byStatus[statusKey] or {}
			table.insert(byStatus[statusKey], did.number)
		end
	end
	return byStatus
end

local function GetUpdateAccountDidsPoolScript()
	return [[
local accountId = ARGV[1]
local didsPoolByStatus = cjson.decode(ARGV[2])

local function AccountDidsPoolKey(status)
	return string.format("AccountDidsPool::%s::%s", accountId, status)
end
local consecutiveKey = AccountDidsPoolKey("CONSECUTIVE_FAILS")

local function UpdateDidsPoolByStatus(status, didsPool)
	local key = AccountDidsPoolKey(status)
	redis.call("del", key)
	for _, did in pairs(didsPool) do
		redis.call("sadd", key, did)
		if status == "ACTIVE" then
			redis.call("hdel", consecutiveKey, did)
		end
	end
end

for status, didsPool in pairs(didsPoolByStatus) do
	UpdateDidsPoolByStatus(status, didsPool)
end

return "OK"
	]]
end

function UpdateAccountDidsPool(accountId, didsPool)
	local byStatus = SplitDidsPoolByStatus(accountId, didsPool)
	local redis = GetDatasource().redis
	redis:eval(GetUpdateAccountDidsPoolScript(), 0, accountId, json.encode(byStatus))
end

function RemoveAccountDidsPool(accountId)
	local redis = GetDatasource().redis
	redis:del(AccountDidsPoolKey(accountId, "ACTIVE"))
	redis:del(AccountDidsPoolKey(accountId, "DISABLED"))
	redis:del(AccountDidsPoolKey(accountId, "REJECTED"))
	redis:del(AccountDidsPoolKey(accountId, "CONSECUTIVE_FAILS"))
end

local function AppendAccountDidsFromPackages(accountName, accountPackages, vccDids, vccPackages)
	for packageId, _ in pairs(accountPackages) do
		local did = vccPackages[packageId] and vccPackages[packageId].DID
		if did then
			vccDids[did] = vccDids[did] or {}
			vccDids[did][accountName] = "package"
		end
	end
end

local function AppendAccountDidsFromPool(accountName, didsPool, vccDids)
	for _, did in pairs(didsPool) do
		if did.number then
			vccDids[did.number] = vccDids[did.number] or {}
			vccDids[did.number][accountName] = "pool"
		end
	end
end

local function AppendAccountDids(account, vccDids, vccPackages)
	if account.proxyType ~= "CALL" then return end

	AppendAccountDidsFromPackages(account.name, account.packages or {}, vccDids, vccPackages)
	local didsPool = account.didsPool or {}
	AppendAccountDidsFromPool(account.name, didsPool.dids or {}, vccDids)
end

function GetVccDids(vcc, vccPackages)
	local vccDids = {}
	local vccAccounts = GetVccAccounts(vcc)
	for _, campaignAccounts in pairs(vccAccounts) do
		for _, account in pairs(campaignAccounts) do
			AppendAccountDids(account, vccDids, vccPackages)
		end
	end
	return vccDids
end
