module(..., package.seeall)

local utils = require "shared.apps.utils"
local json = require "cjson"
local agentModels = require("apps.agent.model"):GetModels()
local m_datasource
local redisKey = "TOTAL_NUMBER_CONFIGURATIONS_PER_VCC"

function GetName ()
	return "config_restrictions"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

local function GetCassandra ()
	return m_datasource.cassandra
end

local function __AddToRedisVCCConfigurationTotals(vcc, redis, keyspace)
    assert(keyspace)
    assert(vcc)
    assert(redis)
    --LogInfo("AddToRedisVCCConfigurationTotals --> processing vcc: %s", vcc)
    local totals = {
        MaxCampaigns = 0,
        MaxCampaignsUsers = 0,
        MaxAttsLevel = 0,
        MaxAttsLevelUsers = 0,
        MaxWorkgroups = 0,
        MaxWorkgroupsUsers = 0
    }
    
    --Workgroups
    local workgroups = utils.GetAllColumns({
        keyspace = keyspace,
        columnFamily = "Workgroup",
        superColumns = true,
        key = vcc
    }) or {}

    for _, workgroup in pairs(workgroups) do
        totals.MaxWorkgroups = totals.MaxWorkgroups + 1
        totals.MaxWorkgroupsUsers = totals.MaxWorkgroupsUsers + utils.GetRowCount(workgroup.Agents or {})
    end

    --Campaigns
    local vccCampaigns = utils.GetAllColumns {
        keyspace = keyspace,
        columnFamily = "VCCCampaigns",
        key = vcc
    } or {}

    for campaignId in pairs(vccCampaigns) do
        local full_campaign_id = utils.AppendVcc(campaignId, vcc )

        local campaign =  utils.GetCampaign(full_campaign_id, {"Id", "SystemType"}, keyspace)
        if campaign and next(campaign) and campaign.SystemType ~= "sandbox" then
            totals.MaxCampaigns = totals.MaxCampaigns + 1
            
            local campUsers = utils.GetAllColumns {
                                    keyspace = keyspace,
                                    columnFamily = "CampaignMembers",
                                    key = full_campaign_id
                                } or {}

            totals.MaxCampaignsUsers = totals.MaxCampaignsUsers + utils.GetRowCount(campUsers)
        
            --AttsLevel
            local als = utils.GetAllColumns {
                                keyspace = keyspace,
                                columnFamily = "CampaignAttentionLevels",
                                key = full_campaign_id
                            } or {}

            for _, al_value in pairs(als) do
                totals.MaxAttsLevel = totals.MaxAttsLevel + 1
                totals.MaxAttsLevelUsers = totals.MaxAttsLevelUsers + utils.GetRowCount(al_value.users or {})
            end
        end
    end

    local encodeTotals = json.encode(totals)
   -- LogInfo("AddToRedisVCCConfigurationTotals --> adding total %s to redis for vcc %s ", json.encode(encodeTotals), vcc)
    local success = pcall(redis.hset, redis, redisKey, vcc, encodeTotals)
    if not success then
        LogError("__AddToRedisVCCConfigurationTotals -- > ERROR: Redis")
        return false, "ERROR_CONNECTION_REDIS"
    end
    return totals
end

-- Agregar a redis los totales de configuraciones actuales de los VCCs existentes para validar limites configurado
function AddToRedisEmptyVCCConfigurationTotals(vcc)
    assert(vcc)
    local redis = utils.GetRedisClient()
    if not redis then
        LogError("AddToRedisEmptyVCCConfigurationTotals -- > ERROR: Failed to connect to redis")
        return
    end
    
    local totals = {
        MaxCampaigns = 0,
        MaxCampaignsUsers = 0,
        MaxAttsLevel = 0,
        MaxAttsLevelUsers = 0,
        MaxWorkgroups = 0,
        MaxWorkgroupsUsers = 0
    }
    local encodeTotals = json.encode(totals)
	--LogInfo("AddToRedisEmptyVCCConfigurationTotals --> adding totals %s to redis for vcc %s ", json.encode(encodeTotals), vcc)
	redis:hset(redisKey, vcc, encodeTotals)
end

function AddToRedisVCCsConfigurationTotals()
	local keyspace = GetCassandra():CassandraConnection( "Agent" )
    local redis = utils.GetRedisClient()
    if not redis then
        LogError("AddToRedisVCCsConfigurationTotals -- > ERROR: Failed to connect to redis")
        return
    end
    
	local startSubProcessTime = os.time()
	--LogInfo("AddToRedisVCCsConfigurationTotals --> START")

	local vccs = utils.GetAllColumns {
		keyspace     = keyspace,
		columnFamily = "VCCs",
		startKey = "",
		endKey = "",
		columns = {"Id", "IsSystemVCC"}
	} or {}

	for vcc, vccData in pairs(vccs) do
		if tonumber(vccData.IsSystemVCC) == 0 then
			__AddToRedisVCCConfigurationTotals(vcc, redis)
		end
	end
	--LogInfo("AddToRedisVCCsConfigurationTotals --> END Elapsetime %s", utils.GetElapsedTimeSec(startSubProcessTime))
end

--[[
    configKey: MaxCampaigns,
			   MaxCampaignsUsers,
			   MaxAttsLevel,
			   MaxAttsLevelUsers,
			   MaxWorkgroups,
			   MaxWorkgroupsUsers

    Dada la key, se busca en redis y se valida si el total de configKey + 'increment' excede el limite configurado para el VCC.
    Retorna true si estan en rango, false en caso negativo
]]--
function ValidateConfigurationsLimits (vcc, configKey, agentKS, increment, vccInfo)
    assert(configKey)
    vcc = vcc or inconcert.authUser.virtualContactCenter
    if type(increment) == "nil" then
        increment = 1
    end
    --LogInfo("ValidateConfigurationsLimits --> INIT  vcc %s, configKey %s, increment %s", vcc, configKey, increment)
    local closeAgentKS
    if not agentKS then
        closeAgentKS = true
        agentKS =  GetCassandra():CassandraConnection( "Agent" )
    end

    local vccInfo =  vccInfo or agentModels.VCCs.Get(vcc, agentKS, {"ConfigLimits"}, true) or {}
    local vccConfigLimits = vccInfo.ConfigLimits
    --LogInfo("ValidateConfigurationsLimits --> vccConfigLimits %s", json.encode(vccConfigLimits or {}))
    local result = false
    local error
    if vccConfigLimits and next(vccConfigLimits) then
        
        local redis = utils.GetRedisClient()
        if redis then
            local success, totalsConfig = pcall(redis.hget, redis, redisKey, vcc)
            if success then
                if totalsConfig then
                    totalsConfig = json.decode(totalsConfig)
                    result = vccConfigLimits[configKey] >= (totalsConfig[configKey] + increment)
                end

                --Si es false o no tiene totalsConfig en redis, recalculo por inconsistencias
                if not totalsConfig or not result then
                    totalsConfig, error =  __AddToRedisVCCConfigurationTotals(vcc, redis, agentKS)
                    --LogInfo("ValidateConfigurationsLimits --> __AddToRedisVCCConfigurationTotals, totalsConfig %s, error %s", json.encode(totalsConfig or {}), error)
                    if totalsConfig then
                        result = vccConfigLimits[configKey] >= (totalsConfig[configKey] + increment)
                    end
                end
            else
                LogError("ValidateConfigurationsLimits -- > ERROR: Redis")
                error = "ERROR_CONNECTION_REDIS"
            end
        end
    else
         --Si no tiene configurados limites por VCC, salto la validacion
        result = true
    end
    LogInfo("ValidateConfigurationsLimits --> result %s, error %s", result, error)
   
    if closeAgentKS then
        agentKS:Close()
    end
    return result, error
end

function UpdateRedisConfigTotals (vcc, map_keyconfig_values, agentKS)
    assert(map_keyconfig_values)
    assert(agentKS)
    vcc = vcc or inconcert.authUser.virtualContactCenter
   -- LogInfo("UpdateRedisConfigTotals -- > vcc %s values %s", vcc, json.encode(map_keyconfig_values))
    local redis = utils.GetRedisClient()
    local success, totalsConfig = pcall(redis.hget, redis, redisKey, vcc)
    if not success then
        LogError("UpdateRedisConfigTotals -- > ERROR: Redis")
        return false
    end
    local error
    totalsConfig, error = totalsConfig and json.decode(totalsConfig) or __AddToRedisVCCConfigurationTotals(vcc, redis, agentKS)

    if error then
        LogError("UpdateRedisConfigTotals -- > ERROR: Redis")
        return false
    end

    for key, value in pairs(map_keyconfig_values) do
        totalsConfig[key] = totalsConfig[key] + value
    end
    local success = pcall(redis.hset, redis, redisKey, vcc, json.encode(totalsConfig))
    --LogInfo("UpdateRedisConfigTotals -- > totalsConfig to update %s, success %s",json.encode(totalsConfig), success)
   
    if not success then
        LogError("UpdateRedisConfigTotals -- > ERROR updating total")
    end
end