local json = require "cjson"
local utils = require "shared.apps.utils"
local moses = require "moses"
local interactionmanip = require "interactionmanip"
local utf8 = require "icu.utf8"
local date = require "LuaOrm.Date"

module(..., package.seeall)

json.encode_empty_table_as_array("on")

local m_datasource

function GetName ()
	return "Interactions"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

-- FIX-ME: This is stupid, its repeated all over the scripts
-- should be a rock or something.
function GetVCCCampaigns (vcc)
	return utils.GetAllColumns({ keyspace = "Agent", columnFamily = "VCCCampaigns", key = vcc }) or {}
end

local m_interactionTypes = {
	"CALL",
	"MAIL",
	"IM",
	"SMS",
	"WHATSAPP",
	"CHAT",
	"FACEBOOK",
	"INSTAGRAM",
	"YOUTUBE",
	"APPSTORE",
	"GOOGLEPLAY",
	"RCS",
	"TWITTER",
	"WEBCONTACT",
	"VOICEMAIL"
}

---
-- Helper to perform the actual search on ES.
-- Needs more work.
--
function search_elastic (vcc, query, withPit, fromAPI, elasticSearchConnectionConfig)
	elasticSearchConnectionConfig = elasticSearchConnectionConfig or inconcert.environment.apps.agent.elasticsearch
	if not elasticSearchConnectionConfig or not elasticSearchConnectionConfig.host or not elasticSearchConnectionConfig.port then
		LogError("[INTERACTIONS.models.search_elastic] Unable to execute ES Request. Invalid elasticsearch config: %s", json.encode(elasticSearchConnectionConfig or {}))
		return false, {error = "invalid elasticsearch config", status = 500}
	end
	
	local requestInfo = utils.GetElasticSearchInstance():MakeRequestInfo({
		query   = query,
		fromAPI = fromAPI
	})

	local http = require "socket.http"
	local ltn12 = require "ltn12"
	local response_body = {}
	local index_name = ("vcc-%s"):format(vcc)
	local url = ("/%s/_search"):format(index_name)

	-- A search request with the pit parameter must not specify index, routing, and preference.
	if withPit then
		url = "/_search"
	end

	--LogDebug("[INTERACTIONS.models.search_elastic] 1. url: %s", url)

	local ok, response_code, response_headers, response_status_line = http.request({
		host = elasticSearchConnectionConfig.host,
		port = elasticSearchConnectionConfig.port,
		url = url,
		headers = requestInfo.headers,
		method = "POST",
		source = requestInfo.source,
		sink = ltn12.sink.table(response_body),
		redirect = false
	})
	if not ok then
		LogError("%s\n%s\n%s\n%s", response_code, json.encode(response_headers), response_status_line, json.encode(response_body))
		return false, { error = response_status_line, status = response_code }
	end

	--LogDebug("[INTERACTIONS.models.search_elastic] 2. response_headers: %s", response_headers)

	if not response_headers["content-type"]:match("^application/json") then
		LogError("[INTERACTIONS.models.search_elastic] Unknown response from elasticsearch. headers: %s\nresponse: %s", json.encode(response_headers), raw_response)
		return false, { error = "inconcert unknown", status = 500 }
	end

	local raw_response = table.concat(response_body)
	--LogDebug("[INTERACTIONS.models.search_elastic] 3. response_body: %s", raw_response)
	local response = json.decode(raw_response)
	if response.error then
		LogError("[INTERACTIONS.models.search_elastic] Request failed with the following error: '%s'", json.encode(response.error))
		return false, response
	end

	return true, response
end

-- shallow clone
local function clone(t)
	local new = {}
	for k,v in pairs(t) do
		new[k] = v
	end
	return new
end

--Get user role in the campaign
--return nil if user doesn't belong to campaign

function GetCampaignUserRole (vcc, user, campaign)
	local userRole = utils.GetAllColumns {
		keyspace = "Agent",
		columnFamily = "UserCampaigns",
		key = utils.AppendVcc(user, vcc)
	}
	return userRole and userRole[utils.AppendVcc(campaign, vcc)]
end

function GetInteractionGroup (vcc, groupId, ks_interactions)
	if not groupId then
		return
	end
	vcc = vcc or inconcert.authUser and inconcert.authUser.virtualContactCenter or nil
	if not vcc then
		return
	end
	local groups = utils.GetAllColumns {
		keyspace = ks_interactions or "Interactions",
		columnFamily = "InteractionGroup",
		key = vcc,
		columns = {groupId}
	}
	return groups and groups[groupId]
end

 -- set relation interaction group
local function SetInteractionGroup (interaction, ks_interactions)
	if not interaction or not interaction.groupId then
		return
	end
	interaction.interactionsGroup = GetInteractionGroup( nil, interaction.groupId, ks_interactions )
end

--retrieves a full interaction JSON from redis
function GetInteraction (id, ks_interactions, datasource)
	assert(type(id) == "string")
	local data = (datasource or m_datasource).redis:get("InteractionDetail:"..id)
	--LogDebug("[INTERACTIONS.model:GetInteraction] ---> Load active interaction id: %s.\ndata: '%s'", id, data)
	if not data or data == "" then
		return
	end
	local res, decodedInteraction = pcall(json.decode, data)
	if res and decodedInteraction then
		SetInteractionGroup( decodedInteraction, ks_interactions ) -- set relation interaction group
		return decodedInteraction
	end
	LogError("Interaction '%s' detail is undecodable: '%s'", id, data)
end


---
-- Builds a summary of interaction counts per attention level, for a given campaign.
-- The 'na' attention level is also included, since interactions without attention level are
-- visible to everyone.
--
-- @tparam string campaignId the id of the campaign
-- @tparam array attention_levels an array of attention levels
--
function GetInteractionsSummary (campaign_id, attention_levels)
	attention_levels = clone(attention_levels or {})
	assert(type(campaign_id) == "string")
	assert(type(attention_levels) == "table")

	local summary = {}

	local params = {
		keyspace = keyspace or "Agent",
		superColumns = true,
		columnFamily = "CampaignAccounts",
		key = campaign_id,
		columns = ""
	}
	local campaignAccounts = utils.GetAllColumns( params ) or {}
		local ChannelOnCampaign = {}
	for campId,accounts in pairs (campaignAccounts) do
		for colName,colValue in pairs(accounts) do
			if colName == "proxyType" then
				ChannelOnCampaign[colValue] = true
			end
		end
		
	end
	
	-- Use the appropiate (regular) sets and also query for
	-- "not available" attention level in order to also get counters for interactions in accounts with no
	-- attention level.
	table.insert(attention_levels, "na")
	for _, level_id in ipairs(attention_levels) do
		for _, t in ipairs(m_interactionTypes) do
			-- dejo los nombres enteros para facilitar el grep en lugar de armarlos al vuelo
			-- e.g. AssignedInteractionsByAttentionLevelTypeAndCampaign:36A47042576403DA5BB773B13549FB32:CHAT:camp1@inconcert
			local waiting_count = 0
			local queued_count = 0
			local unassigned_queued_count = 0
			local assigned_count = 0
			local unassigned_abandoned_count = 0
			if ChannelOnCampaign[t] then
				-- dejo los nombres enteros para facilitar el grep en lugar de armarlos al vuelo
				-- e.g. AssignedInteractionsByAttentionLevelTypeAndCampaign:36A47042576403DA5BB773B13549FB32:CHAT:camp1@inconcert
				suffix = ("%s:%s:%s"):format(level_id, t, campaign_id)
				waiting_count = m_datasource.redis:scard("WaitingInteractionsByAttentionLevelTypeAndCampaign:"..suffix)
				queued_count = m_datasource.redis:scard("QueuedInteractionsByAttentionLevelTypeAndCampaign:"..suffix)
				unassigned_queued_count = m_datasource.redis:scard("UnassignedQueuedInteractionsByAttentionLevelTypeAndCampaign:"..suffix)
				assigned_count = m_datasource.redis:scard("AssignedInteractionsByAttentionLevelTypeAndCampaign:"..suffix)
				unassigned_abandoned_count = m_datasource.redis:scard("UnassignedAbandonedInteractionsByAttentionLevelTypeAndCampaign:"..suffix)
			end

			summary[t] = summary[t] or {}
			summary[t].Waiting = (summary[t].Waiting or 0) + waiting_count
			summary[t].Queued = (summary[t].Queued or 0) + queued_count
			summary[t].UnassignedQueued = (summary[t].UnassignedQueued or 0) + unassigned_queued_count
			summary[t].Assigned = (summary[t].Assigned or 0) + assigned_count
			summary[t].UnassignedAbandoned = (summary[t].UnassignedAbandoned or 0) + unassigned_abandoned_count
		end
	end
	return summary
end

--retrieve all the interactions for a user
--TODO: make this function paginate over the set
function GetUserInteractions (vcc, user, ksInteractions)
	user = utils.GetId(user)
	local fullUserId = utils.AppendVcc(user, vcc)
	local interactions = {}
	-- Assigned interactions (TAKEN)
	local key = "AssignedInteractionsByAgent:"..fullUserId
	local IDs = m_datasource.redis:zrange(key, 0, -1)
	--LogDebug("[INTERACTIONS:GetUserInteractions] key: '%s', IDs: '%s'", key, json.encode(IDs))
	for _, member in pairs(IDs) do
		local ID, type, campaign = string.match(member, "([^#]+)#(.*)#(.*)")
		local interaction = GetInteraction(ID, ksInteractions)
		if not utils.TableIsEmpty(interaction) then
			interactions[ID] = interaction
		else
			LogError("[INTERACTIONS:GetUserInteractions] Interaction %s is in set for user %s but json is undecodable", ID, fullUserId)
		end
	end
	--LogDebug("[INTERACTIONS:GetUserInteractions] key: '%s', interactions: '%s'", key, json.encode(interactions))
	return interactions
end

--retrieve all the interactions for a user
--TODO: make this function paginate over the set
function GetUserOpenInteractions (vcc, user, ksInteractions)
	local userId = user.."@"..vcc
	local interactions = {}
	-- Reserved interactions (OPEN)
	local queue = "OpenInteractionsByAgent:"..userId
	local IDs = m_datasource.redis:zrange(queue, 0, -1)
	for _, ID in pairs(IDs) do
		local interaction = GetInteraction(ID, ksInteractions)
		if not utils.TableIsEmpty(interaction) then
			interactions[ID] = interaction
		else
			LogError("Interaction %s is in set for user %s but json is undecodable", ID, userId)
		end
	end
	return interactions
end

--retrieve all the interactions for a user
--TODO: make this function paginate over the set
function GetUserAbandonedInteractions (vcc, user, ksInteractions)
	local userId = user.."@"..vcc
	local interactions = {}

	local function CanCountInteractionAsAbandoned(interaction)
		local excludeAllegroDialerAbandonedByContact = interaction.type == "CALL" and interaction.state == "ABANDONED" and interaction.initiative == "OUTBOUND" and interaction.interactionData and interaction.interactionData.isAllegro == true and interaction.interactionData.ender == "Contact" and interaction.interactionData.dialer or false
		return not excludeAllegroDialerAbandonedByContact
	end
	-- Reserved interactions (OPEN)
	local queue = "AbandonedInteractionsByAgent:"..userId
	local IDs = m_datasource.redis:zrange(queue, 0, -1)
	for _, ID in pairs(IDs) do
		local interaction = GetInteraction(ID, ksInteractions)		
		if not utils.TableIsEmpty(interaction) then
			--#16463: Allegro dialer abandoned by contact interactions should not count on agent's abandoned panel
			interactions[ID] = CanCountInteractionAsAbandoned(interaction) and interaction or nil
		else
			LogError("Interaction %s is in set for user %s but json is undecodable", ID, userId)
		end
	end
	return interactions
end

--retrieve all the interactions for a user
--TODO: make this function paginate over the set
function GetUserQueuedInteractions (vcc, user, ksInteractions)
	local userId = user.."@"..vcc
	local interactions = {}
	-- Reserved interactions (OPEN)
	local queue = "QueuedInteractionsByAgent:"..userId
	local IDs = m_datasource.redis:zrange(queue, 0, -1)
	for _, member in pairs(IDs) do
		local timestamp, ID = string.match(member, "([^@]+)@(.*)")
		local interaction = GetInteraction(ID, ksInteractions)
		if not utils.TableIsEmpty(interaction) then
			interactions[ID] = interaction
		else
			LogError("Interaction %s is in set for user %s but json is undecodable", ID, userId)
		end
	end
	return interactions
end

function GetAgentBests (vcc, user, ksInteractions)
	local agentBests = {}
	local userId = user.."@"..vcc

	local queryResult = m_datasource.redis:zrange("QueuedInteractionsByAgent:"..userId, 0, 0, "WITHSCORES")
	if type(queryResult) ~= "table" then
		return agentBests -- redis returns a boolean if the sorted set does not exist
	end
	for _, tuple in ipairs(queryResult) do
		local key, priority = tuple[1], -tonumber(tuple[2]) -- the priority is negated here because it was inserted negated by IR
		local timestamp, id = string.match(key, "([^@]+)@(.*)")
		-- Must get the type from the interaction detail.
		local detail = GetInteraction(id, ksInteractions)
		local campaign
		local interactionType = nil
		if not utils.TableIsEmpty(detail) then
			campaign = detail.campaign
			interactionType = detail.type
		else
			LogError("Interaction %s is on an interaction queue, but has no detail.", id)
		end
		if interactionType then
			table.insert(agentBests, { priority = priority, id = id, timestamp = timestamp, type = interactionType, campaign = campaign })
		end
	end
	return agentBests
end

function GetCampaignBests (campaignId, interactionType, ksInteractions)
	local campaignBests = {}

	local function AddToCampaignBests(setName, typeName)
		local queryResult = m_datasource.redis:zrange(setName, 0, 0, "WITHSCORES")
		if type(queryResult) ~= "table" then
			return -- redis returns a boolean if the sorted set does not exist
		end
		for _, tuple in ipairs(queryResult) do
			local key, priority = tuple[1], -tonumber(tuple[2]) -- the priority is negated here because it was inserted negated by IR
			local timestamp, id = string.match(key, "([^@]+)@(.*)")
			if not typeName then
				-- Must get the type from the interaction detail.
				local detail = GetInteraction(id, ksInteractions)
				if not utils.TableIsEmpty(detail) then
					typeName = detail.type
				else
					LogError("Interaction %s is on an interaction queue, but has no detail.", id)
				end
			end
			if typeName then
				table.insert(campaignBests, { priority = priority, id = id, timestamp = timestamp, type = typeName })
			end
		end
	end

	if not interactionType then
		-- Get the best queued interaction of this campaign
		AddToCampaignBests("UnassignedQueuedInteractionsByCampaign:"..campaignId)
	else
		-- Get the best queued interaction of this campaign of a specific type
		AddToCampaignBests("UnassignedQueuedInteractionsByTypeAndCampaign:"..interactionType..":"..campaignId, interactionType)
	end

	return campaignBests
end

---
-- Retrieve all the interactions queued for the given campaign.
--
-- params.campaign_id -- required
-- params.start       -- optional
-- params.total       -- optional
-- params.filter      -- optional
--
function GetCampaignInteractions (params, ksInteractions)
	-- Tope máximo de interacciones a retornar
	local maximumTotalRowsToRetrieve = params.total
	-- Variables del Chunk contra REDIS (inicio y fin)
	local start = params.start or 0
	local finish = maximumTotalRowsToRetrieve and ((start + maximumTotalRowsToRetrieve) - 1) or -1

	-- Retrieve the interaction ids from Redis, then retrieve each interaction from Redis.
	local result = {
		interactions = {}
	}
	-- Si el finish no viene dado, se deben de retornar todas las interacciones.
	local isInfiniteRows = finish == -1
	--
	--[[
		Funcion encargada de Obtener las Interacciones de a Chunks, hasta llegar al límite de interacciones permitidas "maximumTotalRowsToRetrieve"

		Ej:
			En la primera invocación a la función va a obtener desde REDIS 75 interacciones.
			Luego las va a pasar por el filtro, para saber si pueden ser retornadas.
			Si luego de pasar el filtro las interacciones a retornar no llega al tope máximo de interacciones a retornar "maximumTotalRowsToRetrieve", se va invocar nuevamente a la misma función, para que obtenga las siguientes 75 Interacciones,
			y así sucesivamente hasta llegar al tope máximo a retornar.
	]]--
	local function RecursiveLoad (chunk)
		chunk = chunk + 1

		--LogDebug("[AGENT.INTERACTIONS:GetCampaignInteractions] + Loading interactions for campaign '%s'. Chunk '%d', Rows Range Between '%d' and '%d'.", params.campaign_id, chunk, start, finish)

		-- Obtiene un rango dado (start / finish) de interacciones (ej: 0 a 74)
		local queue = "UnassignedQueuedInteractionsByCampaign:" .. params.campaign_id
		local interactionsIDs = m_datasource.redis:zrange(queue, start, finish)

		local totalRowsInChunk -- numero total de interacciones que se han obtenido desde REDIS en el chunk
		local totalAddedInChunk = 0 -- numero total de interacciones que han pasado el filtro en el chunk

		if type(interactionsIDs) == "table" then
			local Filter = params.filter or function() return {success = true} end
			totalRowsInChunk = #interactionsIDs -- la cantidad de interacciones que ha retornado el chunk desde REDIS

			for _, infoId in ipairs(interactionsIDs) do
				local timestamp, interactionId = string.match(infoId, "([^@]+)@(.*)")
				--- obtengo la interaccion
				local interaction = GetInteraction(interactionId, ksInteractions)
				if not utils.TableIsEmpty(interaction) then
					local filterResult = Filter(interaction)
					if filterResult and filterResult.success then
						-- La interacción pasó el filtro, la agrego en lista de interacciones a retornar
						totalAddedInChunk = totalAddedInChunk + 1
						table.insert(result.interactions, interaction)
					end
				end
			end
		end

		-- Si la cantidad de Rows que se han obtenido desde REDIS, es la misma que la cantidad de interacciones a retornar, debo procedar a validar.
		-- En el caso de que la cantidad sea menor, ahi quiere decir que ya no hay mas interacciones a validar en REDIS y tengo que salir de la función.
		if totalRowsInChunk == maximumTotalRowsToRetrieve then
			if not isInfiniteRows then
				local totalRowsToRetrieve = #result.interactions
				-- Si la cantidad de rows que se van a retornar "totalRowsToRetrieve" es menor al Tope máximo de interacciones a retornar "maximumTotalRowsToRetrieve",
				--  tengo que continuar con el siguiente Chunk en REDIS.
				-- En el caso de que la cantidad NO sea menor, es decir es igual, ahi quiere decir que ya se ha alcanzado al Tope máximo de interacciones a retornar "maximumTotalRowsToRetrieve", y tengo que retornar las mismas.
				if totalRowsToRetrieve < maximumTotalRowsToRetrieve then
					-- Acá proceso a invocar de forma recursiva a la función para obtner desde REDIS el siguiente chunk.
					start = finish + 1
					finish = ((start + maximumTotalRowsToRetrieve) - 1)
					RecursiveLoad(chunk)
				end
			end
		end

		--LogDebug("[AGENT.INTERACTIONS:GetCampaignInteractions] - Loaded '%d' interactions for campaign '%s' in Chunk '%d'.", totalAddedInChunk, params.campaign_id, chunk)
	end

	--LogInfo("[AGENT.INTERACTIONS:GetCampaignInteractions] TOTAL Loaded '%d' interactions for campaign '%s'.", #result.interactions, params.campaign_id)

	local chunk = 0
	RecursiveLoad(chunk)

	return result
end

--[[
	Retrieves the contact status from their IM accounts
	this is the status format: { "buddy" : "fpardo@internet.com.uy", "IM" : "MSN", "nick" : "Fernickk the Bunny Master", "personalMessage" : "", "status" : "DISCONNECTED" }
]]--
function IMGetBuddyStatus (buddy, protocol)
	local data = m_datasource.redis:get("AccountInfo:IM:"..buddy.."@"..protocol)
	if (type(data) == "string") then
		return json.decode(data)
	end
	return nil
end

function GetFromArchive (interactionID, ks_interactions)
	--LogDebug("[INTERACTIONS:GetFromArchive] Load archive interaction id: %s", interactionID)
	local release = false
	if not ks_interactions then
		release = true
		local cassandra = m_datasource and m_datasource.cassandra or utils.GetDatasource()
		ks_interactions = cassandra:CassandraConnection("Interactions")
	end

	--basic data
	local basicColumns = ks_interactions:GetSliceRange(interactionID,
		"Interactions",
		"", {
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 200
			}
		}
	)

	local interaction = {}
	for _, v in ipairs(basicColumns or {}) do
		-- Most archived interactions have an "original" field with the interaction as JSON. Some very very old have all their fields separated, so we must check those.
		if v.name == "original" then
			interaction = json.decode(v.value)
			break
		elseif v.name == "interactionData" or v.name == "steps" or v.name == "events" or v.name == "comments" then
			interaction[v.name] = json.decode(v.value)
		else
			interaction[v.name] = v.value
		end
	end

	SetInteractionGroup( interaction, ks_interactions ) -- set relation interaction group

	if release then
		ks_interactions:Close()
	end

	return interaction
end

function UpdateArchivedInteractionData (interaction, ksInteractions, origin)
	assert(interaction)
	--LogDebug("[INTERACTIONS.models:UpdateArchivedInteractionData] --> Updating archived Interaction id: %s, state %s, origin: %s", interaction.id, interaction.state, origin)
	if utils.TableIsEmpty(interaction) then
		--LogDebug("[INTERACTIONS.UpdateArchivedInteractionData] SALE 0.5 - interaction.id: %s", interaction.id)
		return false
	end
	if interaction.state ~= "FINISHED" then
		--LogDebug("[INTERACTIONS.UpdateArchivedInteractionData] SALE 0.6 - interaction.id: %s", interaction.id)
		return false
	end

	local release = false
	if not ksInteractions then
		release = true
		local cassandra = m_datasource and m_datasource.cassandra or utils.GetDatasource()
		ksInteractions = cassandra:CassandraConnection("Interactions")
	end

	local function UpdateRedisInteraction ()
		local encodedInteraction = json.encode(interaction)
		--LogDebug("[INTERACTIONS.models:UpdateArchivedInteractionData] 1--> REDIS Updating archived")
		--LogDebug("[INTERACTIONS.models:UpdateArchivedInteractionData] 1.1--> REDIS Updating archived encodedInteraction %s", encodedInteraction)
		m_datasource.redis:set("InteractionDetail:" .. interaction.id, encodedInteraction)
	end
	local function UpdateCassandraInteraction ()
		local encodedInteractionData = json.encode(interaction.interactionData)
		--LogDebug("[INTERACTIONS.models:UpdateArchivedInteractionData] 2--> CASSANDRA Updating archived")
		---LogDebug("[INTERACTIONS.models:UpdateArchivedInteractionData] 2.1--> CASSANDRA Updating archived encodedInteractionData %s", encodedInteractionData)
		utils.SaveRows({
			keyspace = ksInteractions,
			columnFamily = "Interactions",
			rows = {
				[interaction.id] = {["interactionData"] = encodedInteractionData}
			}
		})
	end
	if origin == "redis" then
		UpdateRedisInteraction()
		local cassandraInteraction = GetFromArchive(interaction.id, ksInteractions)
		--LogDebug("[INTERACTIONS.models:UpdateArchivedInteractionData] 3--> cassandraInteraction %s", cassandraInteraction)
		if not utils.TableIsEmpty(cassandraInteraction) then
			UpdateCassandraInteraction()
		end
	elseif origin == "cassandra" then
		UpdateCassandraInteraction()
	end

	if release then
		ksInteractions:Close()
	end
end


function SaveDraftForPending (rows, keyspace)

	local closeKeyspace = false
	if not keyspace then
		keyspace = utils.GetDatasource():CassandraConnection("Interactions")
		closeKeyspace = true
	end

	utils.SaveRows {
		keyspace = keyspace,
		columnFamily = "DraftsInteractionsByAgents",
		superColumns = true,
		rows = rows
	}
	if closeKeyspace then
		keyspace:Close()
	end
end

function GetDraftsInteractions(key, keyspace)
	local closeKeyspace = false
	if not keyspace then
		keyspace = utils.GetDatasource():CassandraConnection("Interactions")
		closeKeyspace = true
	end
	local interactions = utils.GetAllColumns({
		keyspace =  keyspace or "Interactions",
		columnFamily = "DraftsInteractionsByAgents",
		superColumns = true,
		startKey     = key or "",
		endKey       = "",
		columns = {}
	}) or {}
	if closeKeyspace then
		keyspace:Close()
	end
	if key then
		return interactions[key]
	end
	return interactions
end

function DeleteDraftForPending (rows, keyspace)

	local closeKeyspace = false
	if not keyspace then
		keyspace = utils.GetDatasource():CassandraConnection("Interactions")
		closeKeyspace = true
	end
	utils.RemoveColumns {
		keyspace =  keyspace,
		columnFamily = "DraftsInteractionsByAgents",
		superColumns = true,
		rows = rows
	}
	if closeKeyspace then
		keyspace:Close()
	end
end

---
-- TODO: Description
--
-- @tparam string contact_id
-- @tparam string vcc
-- @treturn {table, string}
--
function GetInteractionsFromContact (contact_id, vcc)
	--LogDebug("[INTERACTIONS:GetInteractionsFromContact] contact: '%s' on vcc: '%s'", contact_id, vcc)

	local vccCampaigns = {}
	local campaigns = GetVCCCampaigns(vcc)
	for campaignId in pairs(campaigns) do
			table.insert(vccCampaigns, utf8.lower(campaignId).."@"..vcc)
	end

	local MAX_INTERACTIONS = 250
	-- No pueden ser mas antiguas de 29 dias.
	local timestamp = date():time()
	local daysInSeconds =  24 * 60 * 60 * 29
	local start_date_string = date(timestamp - daysInSeconds):toString("%Y-%m-%d 00:00:00")
	local range = { start_timestamp = { format = "yyyy-MM-dd HH:mm:ss" } }
	range.start_timestamp.gte = start_date_string
	
	-- Query for an interactions matching the given contact_id and the given array of campaigns.
	local query = {
		query = {
			bool = {
				must = {
					{ range = range }
				},
				filter = { 
					{ terms = { ["campaign"] = vccCampaigns } },
					{ term = { ["contact"] = contact_id } }
				}
			}
		},
		sort = { { start_timestamp = { order = "desc" } } },
		from = 0,
		size = MAX_INTERACTIONS
	}

	LogWarning("[GetInteractionsFromContact] Query against elasticsearch. %s", json.encode(query))
	local ok, result = search_elastic(inconcert.authUser.virtualContactCenter, query)
	if not ok then
		return nil, { code = "E_REQUEST_FAILED", msg = result }
	end
	if result.timed_out then
		LogWarning("[GetInteractionsFromContact] Query against elasticsearch timed out. %s", json.encode(query))
		return nil, { code = "E_REQUEST_TIMEDOUT", msg = "Query timed out" }
	end
	LogInfo("[GetInteractionsFromContact] Query took %s ms and returned %s hits out of %s", result.took, #result.hits.hits, result.hits.total.value)
	--LogDebug("%s", json.encode(result, true))
	local result_interactions = {}

	--local ks_interactions = m_datasource.cassandra:CassandraConnection("Interactions")
	-- retrieve the full interaction
	moses.each(result.hits.hits, function(_, hit)
		--local interaction = GetInteraction(hit._id) or GetFromArchive(hit._id, ks_interactions)
		--if not utils.TableIsEmpty(interaction) then table.insert(result_interactions, interaction) end
		table.insert(result_interactions, { id = hit._id })
	end)

	--ks_interactions:Close()

	--interactionmanip.NormalizeInteractions(result_interactions)

	return result_interactions
end

function SetTemporalFile (id, data)
	m_datasource.redis:set("InteractionTemporalFile:" .. id, data)
	m_datasource.redis:expire("InteractionTemporalFile:" .. id, 180)
end

function SetDisposition (id, data)
	m_datasource.redis:set("InteractionDispositionCode:" .. id, json.encode(data))
	m_datasource.redis:expire("InteractionDispositionCode:" .. id, 86400)
end

function GetDisposition (id)
	local data_s = m_datasource.redis:get("InteractionDispositionCode:" .. id)
	return data_s and data_s ~= "" and json.decode(data_s) or nil
end

function RemoveDisposition (id)
	m_datasource.redis:del("InteractionDispositionCode:" .. id)
end

function SaveApiAdditionalData (id, data)
	m_datasource.redis:set("InteractionApiAdditionalData:" .. id, data)
	m_datasource.redis:expire("InteractionApiAdditionalData:" .. id, 86400)
end

function SetPreInteractionData (id, data)
	m_datasource.redis:set("PreInteractionDetail:" .. id, data)
	m_datasource.redis:expire("PreInteractionDetail:" .. id, 180)
end

function GetCampaignUsers(vcc, campaignAtVCC, interactionPermissions)
	local agents = m_datasource.agentredis:smembers("AgentService::CampaignAgents:" .. campaignAtVCC) or {}
	if not interactionPermissions then return agents end

	local agentsWithPermission = {}
	for _, agentId in pairs(agents) do
		local agentKey = "AgentService::AgentInteractionCampaigns:" .. agentId .. "@" .. vcc
		if m_datasource.agentredis:sismember(agentKey, campaignAtVCC) then
			table.insert(agentsWithPermission, agentId)
		end
	end
	return agentsWithPermission
end

function GetProcessFlowInteractions (ksInteractions, campaignId, interactionType, needHelp)
	local key = "ProcessFlowInteractionsByCampaign:"..campaignId

	if needHelp == "true" then
		key = "ProcessFlowInteractionsByCampaignNeedHelp:"..campaignId
	end

	local IDs = m_datasource.redis:zrange(key, 0, -1)
	local interactions = {}
	
	for _, ID in pairs(IDs) do
		local interaction = GetInteraction(ID, ksInteractions)

		if interactionType == nil or interaction.type == interactionType then
			if not utils.TableIsEmpty(interaction) then
				interactions[ID] = interaction
			else
				LogError("[INTERACTIONS:GetUserInteractions] Interaction %s is in set for user %s but json is undecodable", ID, fullUserId)
			end
		end
	end

	local function getInternalConsultationFlag(interaction)
		if interaction.interactionData and interaction.interactionData.need_help then
			return interaction.interactionData.need_help or false
		end
		return false
	end

	local function compareInteractions(a, b)
		local internal_consultation_flag_a = getInternalConsultationFlag(a)
		local internal_consultation_flag_b = getInternalConsultationFlag(b)

		if internal_consultation_flag_a and not internal_consultation_flag_b then
			return true
		elseif not internal_consultation_flag_a and internal_consultation_flag_b then
			return false
		else
			return false
		end
	end

	local keysTable = {}
	for key, item in pairs(interactions) do
		table.insert(keysTable, item)
	end

	table.sort(keysTable, compareInteractions)

	local sortedInteractions = {}
	for _, item in ipairs(keysTable) do
		table.insert(sortedInteractions, item)
	end

	return sortedInteractions
end
