/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:false, unused:vars, curly:false, browser:true, indent:false, maxerr:50, quotmark:false, -W014, esversion: 11 */

/* global inConcert, PubSub, uuid, AudioContext, AudioWorkletNode, MediaStream, Resampler*/

/*
Main application module
*/

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
	this.inConcert = {};
}

(function () {
	"use strict";

	const _T = inConcert.i18n.getString;
	const format = inConcert.Utils.format;

	//My only one application object, just for encapsulation purposes
	//constructor

	function Application(params) {
		if(typeof(params) !== "object") {
			throw "Application.Constructor - expected an object with configuration";
		}

		var defaultProtocol = window.location.protocol === "http:" ? "http" : "https";
		this.m_mainWindow = null;
		this.m_loggedUser = null;
		this.m_desktopNotifications = null;

		const buildAddressFn = inConcert.Utils.MakeFnServiceAddressBuilder();

		// agent server address
		this.m_serverAddress        = buildAddressFn("");
		this.m_protocol             = defaultProtocol;

		// tweety server address
		this.m_tweetyDesignerServerAddress = this.m_serverAddress;
		this.m_tweetyRuntimeServerAddress = this.m_serverAddress;
		this.m_tweetyRuntimeWithSubDomainServerAddress  = buildAddressFn("webdesigner");
		this.m_tweetyProtocol       = defaultProtocol;

		// chat server address
		this.m_chatServerAddress    = buildAddressFn("chat");
		this.m_chatProtocol         = defaultProtocol;

		// poller server address
		this.m_pollerServerAddress  = buildAddressFn("poller");
		this.m_pollerProtocol       = defaultProtocol;

		// webchat server address
		this.m_webchatServerAddress = buildAddressFn("webchat");
		this.m_webchatProtocol      = defaultProtocol;

		// trackme server address and protocol
		this.m_trackmeServerAddress = buildAddressFn("trackme");
		this.m_trackmeProtocol      = defaultProtocol;

		// webcontactproxy server address and protocol
		this.m_webcontactServerAddress = buildAddressFn("webcontact");
		this.m_webcontactProtocol      = defaultProtocol;

		// allegro server address and protocol
		this.m_allegroWebServerAddress = buildAddressFn("allegro");
		this.m_allegroWebProtocol      = defaultProtocol;

		// sms server address
		this.m_smsServerAddress = buildAddressFn("sms");
		this.m_smsProtocol      = defaultProtocol;

		this.m_current_app = "interaction"; // interaction, configuration, supervisor, reports
		this.m_campaigns = [];
		this.m_settings = null;
		this.m_user = null;
		this.m_applicationTimer = null;
		this.m_lastAccountRefresh = 0;
		this.m_lastWindowRefresh = 0;
		this.loggedOut = true;
		this.m_authToken = null;
		this.m_sessionRejoin = null;
		this.m_country = null;
		this.m_area = null;
		this.m_timezone_id = null;
		this.m_user_timezone_object = null;
		this.cached_roles = null;
		this.cached_mail_signatures = null;
		this.cached_members = null;
		this.m_serverInfo = null;
		this.m_IRInfo = null;
		this.m_passwordExpirationDays = null;
		this.m_enterAsKeyToSend = false;
		this.m_campaignsWarnings = null;
		this.m_vccWarnings = null;
		this.AgentScreenService = null;
		this.m_userIdleThreshold = null;
		this.m_userSoundNotifications = null;
		this.m_userVisualNotifications = null;
		this.m_userIsFullSystemAccess = null;

		this.m_telephone_codes = new inConcert.TelephoneCodes({
			uri: this.getBaseUrl() /* app base uri */
		});

		this.tweety_api = new inConcert.TweetyAPI({
			uri: this.getTweetyDesignerUrl() /* tweety base uri */
		});

		this.allegro_api = null;
		if (inConcert.AllegroAPI) {
			this.allegro_api = new inConcert.AllegroAPI({
				uri: this.getAllegroUrl() /* allegro base uri */
			});
		}

		this.rasa_api = new inConcert.RasaAPI({
			uri: this.getCognitiveServicesUrl() /* cognitive services base uri */
		});

		this.luis_api = new inConcert.LuisAPI({
			uri: this.getCognitiveServicesUrl() /* cognitive services base uri */
		});

		this.qna_maker_api = new inConcert.QnAMakerAPI({
			uri: this.getKnowledgeBaseUrl() /* KnowledgeBase base uri */
		});

		this.qna_v2_api = new inConcert.QnaV2Api({
			uri: this.getQnaV2Url() /* administration base uri */
		});

		this.flow_designer_api = null;
		if (inConcert.FlowDesignerAPI) {
			this.flow_designer_api = new inConcert.FlowDesignerAPI({
				uri: this.getFlowDesignerUrl(), /* flow_designer base uri */
				uri_admin: this.getAdministrationUrl() /* administration base uri */
			});
		}

		this.phone_validator_api = null;
		if (inConcert.PhoneValidatorAPI) {
			this.phone_validator_api = new inConcert.PhoneValidatorAPI({
				uri: this.getAdministrationUrl() /* administration base uri */
			});
		}

		this.shared_files_api = new inConcert.SharedFilesAPI({
			uri: this.getAdministrationUrl() /* administration base uri */
		});

		this.outbound_engine_api = null;
		if (inConcert.OutboundEngineAPI) {
			this.outbound_engine_api = new inConcert.OutboundEngineAPI({
				uri: this.getOutboundEngineUrl() /* outbound engine base uri */
			});
		}

		this.provisioning_api = null;
		if (inConcert.ProvisioningAPI) {
			this.provisioning_api = new inConcert.ProvisioningAPI({
				uri: this.getProvisioningUrl() /* provisioning base uri */
			});
		}

		this.importation_api = null;
		if (inConcert.ImportationAPI) {
			this.importation_api = new inConcert.ImportationAPI({
				uri: this.getAdministrationUrl() /* administration base uri */
			});
		}

		this.helpdesk_api = new inConcert.HelpDeskAPI({
			uri: this.getHelpDeskUrl() /* helpdesk base uri */
		});

		this.repository_api = new inConcert.RepositoryAPI({
			path : this.getRepositoryPath(),
			uri: this.getRepositoryUrl() /* repository base uri */
		});

		this.tickets_api = new inConcert.TicketsAPI({
			uri: this.getBaseUrl() /* agent base uri */
		});

		this.batches_mng_api = new inConcert.BatchesMngAPI({
			uri: this.getBatchesMngUrl() /* batches mng base uri */
		});

		this.social_networks_api = new inConcert.SocialNetworksAPI({
			uri: this.getSocialNetworksUrl() /* batches mng base uri */
		});

		this.agent_api_server = null;

		// cache de todas las campanas del VCC.
		this.m_cachedVccCampaigns = {
			data          : null,
			timestamp     : null,
			maxCacheTimeMs: 30 * 1000 // 30 segundos
		};

		// This variable contains the list of cached timezones
		this.m_timezones = null;
		// This variable contains the list of cached packages
		this.m_packages = null;
		// Matches a string ending in @VCC. VCC cannot contain @ symbols and must be non-empty.
		// The part left to the @ must be non-empty, and conforms the id.
		// Match results are { id, VCC }
		this.idAtVCCRegExp = /^(.+)@([^@]+)$/;

		this.settings = { development: false };
		// TODO: this has been forced true in mcwarren branch
		// set to false when this lands on master (ping burguer)
		this.settings = { development: true };
		
		this.m_pollerCallbacks = {};

		// Store errors in the server as soon as they occur.
		inConcert.logger.enableAjaxLog( this.getBaseDomain() );

		this.messagebus = PubSub;

		this.ScreenCaptureObj = { capturing: false };

		/*-- INCO-64 --*/
		this.AgentMonitorObj = {};
	}

	if (typeof inConcert.Application !== 'function') {
		inConcert.Application = Application;
	}

	Application.prototype.separate_vcc = function( id_with_vcc )
	{
		if ( typeof id_with_vcc !== "string" ) {
			return;
		}
		if ( !id_with_vcc.match( this.idAtVCCRegExp ) ) {
			return;
		}
		return {
			id  : RegExp.$1,
			vcc : RegExp.$2
		};
	};

	/*
		Obtiene que Skin se ha aplicado en el Body para la current App (interaction, administration, oe config, etc, etc)
	*/
	Application.prototype.get_skin_name_of_current_app = function()
	{
		let skins = [
			"skinNew",   // El de siempre
			"skinOCCNew" // El del 2021
		];
		const $body = $("body");
		for (let i = 0; i < skins.length; i++) {
			let skin = skins[i];
			if ($body.hasClass(skin)) {
				return skin;
			}
		}
		return false;
	};

	Application.prototype.get_id = function( id_with_vcc )
	{
		var parsed = this.separate_vcc( id_with_vcc || "");
		if (!parsed) {
			return id_with_vcc;
		}
		return parsed.id;
	};

	Application.prototype.get_domain = function( id_with_vcc )
	{
		var parsed = this.separate_vcc( id_with_vcc );
		if (!parsed) {
			return id_with_vcc;
		}
		return parsed.vcc;
	};

	Application.prototype.get_os_distro = function()
	{
		return this.getServerInfo().OSDistro;
	};

	Application.prototype.is_os_debian = function()
	{
		return this.get_os_distro() === "debian";
	};

	Application.prototype.is_os_ubuntu = function()
	{
		return this.get_os_distro() === "ubuntu";
	};

	Application.prototype.hasLicenseSystem = function()
	{
		return this.is_os_ubuntu();
	};

	Application.prototype.hasRepositoryService = function()
	{
		return this.hasLicenseSystem();
	};

	Application.prototype.getCurrentApp = function() {
		return this.m_current_app;
	};

	Application.prototype.setCurrentApp = function(current_app) {
		this.m_current_app = current_app;
	};

	Application.prototype.is_adobe_air = function() {
		return navigator.userAgent.search("AdobeAIR", "i") > -1;
	};

	// returns the global window manager in order to
	// make it work like a singleton
	Application.prototype.getWindowManager = function() {
		return inConcert.DialogManager;
	};

	Application.prototype.getTweetyAPI = function() {
		return this.tweety_api; // instance of
	};

	Application.prototype.getAllegroAPI = function() {
		return this.allegro_api; // instance of
	};

	Application.prototype.getRasaAPI = function() {
		return this.rasa_api; // instance of
	};

	Application.prototype.getLuisAPI = function() {
		return this.luis_api; // instance of
	};

	Application.prototype.getQnAMakerAPI = function() {
		return this.qna_maker_api; // instance of
	};

	Application.prototype.getQnAV2API = function() {
		return this.qna_v2_api; // instance of
	};

	Application.prototype.getFlowDesignerAPI = function() {
		return this.flow_designer_api; // instance of
	};

	Application.prototype.getPhoneValidatorAPI = function() {
		return this.phone_validator_api; // instance of
	};

	Application.prototype.getSharedFilesAPI = function() {
		return this.shared_files_api; // instance of
	};

	Application.prototype.getRepositoryAPI = function() {
		return this.repository_api; // instance of
	};

	Application.prototype.getTicketsAPI = function() {
		return this.tickets_api; // instance of
	};

	Application.prototype.getOutboundEngineAPI = function() {
		return this.outbound_engine_api; // instance of
	};

	Application.prototype.getProvisioningAPI = function() {
		return this.provisioning_api; // instance of
	};

	Application.prototype.getImportationAPI = function() {
		return this.importation_api; // instance of
	};

	Application.prototype.getHelpDeskAPI = function() {
		return this.helpdesk_api; // instance of
	};

	Application.prototype.getBatchesMngAPI = function() {
		return this.batches_mng_api; // instance of
	};

	Application.prototype.getSocialNetworksAPI = function() {
		return this.social_networks_api; // instance of
	};

	Application.prototype.getAgentApiServer = function() {
		return this.agent_api_server; // instance of
	};

	Application.prototype.getProtocol = function() {
		return this.m_protocol;
	};

	// sets all protocols
	Application.prototype.setProtocol = function(protocol) {
		this.m_protocol = this.m_chatProtocol = this.m_pollerProtocol = this.m_webchatProtocol = protocol;
	};

	Application.prototype.getBaseUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/agent";
	};

	Application.prototype.getBaseDomain = function() {
		return this.getProtocol() + "://" + this.m_serverAddress;
	};

	Application.prototype.getPollerUrl = function() {
		return this.m_pollerProtocol + "://" + this.m_pollerServerAddress + "/inconcert/apps/agent";
	};

	Application.prototype.getTrackmeUrl = function() {
		return this.m_trackmeProtocol + "://" + this.m_trackmeServerAddress + "/inconcert/apps/trackme";
	};

	Application.prototype.getTrackmeServerAddress = function() {
		return this.m_trackmeServerAddress;
	};

	Application.prototype.getTweetyDesignerProtocol = function() {
		return this.m_tweetyProtocol;
	};

	Application.prototype.getTweetyDesignerServerAddress = function() {
		return this.m_tweetyDesignerServerAddress;
	};

	Application.prototype.getTweetyRuntimeServerAddress = function() {
		return this.m_tweetyRuntimeServerAddress;
	};

	Application.prototype.getTweetyRuntimeWithSubDomainServerAddress = function() {
		return this.m_tweetyRuntimeWithSubDomainServerAddress;
	};

	Application.prototype.getTweetyDesignerDomain = function() {
		return "/inconcert/apps/webdesigner";
	};

	Application.prototype.get_interactions_windows_mode = function()
	{
		return "NORMAL WINDOWS";
	//	return "DOCKED WINDOWS";
	};

	Application.prototype.is_enable_docked_windows_interactions = function()
	{
		let windows_mode = this.get_interactions_windows_mode();
		if (windows_mode === "DOCKED WINDOWS") {
			return true;
		}
		return false;
	};

	/*
		Funcion encargada de abrir una ventana del telefono de tipo Live.
		La misma es invocada desde la clase "PhoneClient" en el archivo "phone_client.js"
	*/
	Application.prototype.openLiveTelephoneWindow = function(interaction, on_ready = $.noop, on_close = $.noop)
	{
		const id = interaction && interaction.id || "empty";
		let options = {};
		if (interaction && interaction.isLocked() && interaction.belongsToMe()) {
			const account_gui_config = interaction.getAccountConfigGUI();
			if (account_gui_config.window_open_mode === "MAXIMIZED") {
				// Cuando en la cuenta de la campana se ha configurado que la apertura inicial de la ventana es maximizada, la misma debe abrir en este modo.
				options.maximizable = true;
				options.create_maximized = true;
				options.resizable = true;
				options.allowed_restore_button = true;
			}
		}

		let idIFrameContainer = null;
		// TODO: #9778 - CANY - Docked.
		if (this.is_enable_docked_windows_interactions()) {
			idIFrameContainer = "iFramesWindowsDailogsContainerForInteractionAppLivePanel";
			if (interaction && interaction.isInternal()) {
				idIFrameContainer = null;
			}
		}

		let windowOptions = $.extend({
			windowID   : 'builtInTelephone_' + id,
			page       : "telephonyClientWindow",
			title      : "Web Phone",
			// interaction: interaction,  // TODO: #9778 - CANY - Docked. ????? esto no se si va
			maximizable: true,
			resizable  : true,
			// el minimizar y el drag se habilita una vez que la carga del iframe, la sidebar y el GoldenLayout han finalizado.
			minimizable: false,
			draggable  : false,
			from_app   : "interaction",
			// TODO: #9778 - CANY - Docked.
			id_iframe_windows_dialog_container: idIFrameContainer,
			// TODO: #9778 - CANY - Docked.
			on_ready: function(clientWindow, windowSession) {
				if (clientWindow.switchTo) {
					clientWindow.switchTo(interaction);
				}
				on_ready(windowSession);
			},
			on_close: function() {
				on_close();
			},
			size: {
				// Estas dimensiones deben de coincidir con lo que se ha definido en "phone_client_window.js" ej: "DEFAULT_WINDOW_WIDTH_PX"
				width : 950,
				height: 600
			},
		}, options);

		const window_mamager = this.getWindowManager();
		const session_window = window_mamager.createWindow(windowOptions);
		return session_window;
	};

	Application.prototype.getTweetyUrlParts = function() {
		return {
			protocol: this.getTweetyDesignerProtocol(),
			serverAddress: this.getTweetyDesignerServerAddress(),
			serverRuntimeAddress: this.getTweetyRuntimeServerAddress(),
			serverRuntimeWithSubDomainAddress: this.getTweetyRuntimeWithSubDomainServerAddress(),
			domain: this.getTweetyDesignerDomain()
		};
	};

	Application.prototype.getTweetyDesignerUrl = function() {
		var url_parts = this.getTweetyUrlParts();
		return url_parts.protocol + "://" + url_parts.serverAddress + url_parts.domain;
	};

	Application.prototype.getTweetyRuntimeUrl = function() {
		var url_parts = this.getTweetyUrlParts();
		return url_parts.protocol + "://" + url_parts.serverRuntimeAddress + url_parts.domain;
	};

	Application.prototype.getTweetyRuntimeUrlWithSubDomain = function() {
		var url_parts = this.getTweetyUrlParts();
		return url_parts.protocol + "://" + url_parts.serverRuntimeWithSubDomainAddress + url_parts.domain;
	};

	Application.prototype.getWebchatUrl = function() {
		return this.m_webchatProtocol + "://" + this.m_webchatServerAddress + "/v3";
	};

	Application.prototype.getWebchatRootUrl  = function() {
		return this.m_webchatProtocol + "://" + this.m_webchatServerAddress;
	};

	Application.prototype.getWebcontactUrl = function() {
		return this.m_webcontactProtocol + "://" + this.m_webcontactServerAddress;
	};

	Application.prototype.getChatProtocol = function() {
		return this.m_chatProtocol;
	};

	Application.prototype.getChatServer = function() {
		return this.m_chatServerAddress;
	};

	Application.prototype.getAllegroWebServer = function() {
		return this.m_allegroWebServerAddress;
	};

	Application.prototype.getAllegroWebProtocol = function() {
		return this.m_allegroWebProtocol;
	};

	Application.prototype.getAllegroWebUrl = function() {
		return this.getAllegroWebProtocol() + "://" + this.getAllegroWebServer();
	};

	Application.prototype.redirectAllegroWebUrl = function(url) {
		return url.replace(/.*:\/\/[^/]*\//, this.getAllegroWebUrl() + "/");
	};

	Application.prototype.getAllegroUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/allegro";
	};

	Application.prototype.getAdministrationUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/administration";
	};

	Application.prototype.getKnowledgeBaseUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/knowledge_base";
	};

	Application.prototype.getFlowDesignerUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/flow_designer";
	};

	Application.prototype.getCognitiveServicesUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/cognitive_services";
	};

	Application.prototype.getOutboundEngineUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/outbound_engine";
	};

	Application.prototype.getProvisioningUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/provisioning";
	};

	Application.prototype.getRepositoryPath = function() {
		return "/inconcert/apps/repository";
	};

	Application.prototype.getRepositoryUrl = function() {
		return this.getBaseDomain() + this.getRepositoryPath();
	};

	Application.prototype.getWebSmsWebhooksUrl = function() {
		// SAMPLE: http://sms-i6.inconcert.cc/webhooks/new_message/<id_cuenta>
		return this.m_smsProtocol + "://" + this.m_smsServerAddress + "/webhooks";
	};

	Application.prototype.getHelpDeskUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/helpdesk";
	};

	Application.prototype.getQnaV2Url = function() {
		return this.getKnowledgeBaseUrl();
	};

	Application.prototype.getBatchesMngUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/batch_management/";
	};

	Application.prototype.getSocialNetworksUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/social_networks/";
	};

	Application.prototype.getCampaignAnalyticsUrl = function() {
		return this.getBaseDomain() + "/inconcert/apps/campaign_analytics/";
	};

	Application.prototype.getAgentScreenServiceEnabled = function() {
		let enabled = this.m_vcc_info.ScreenRecording && this.m_vcc_info.ScreenRecording.Enabled;
		if (enabled && ((this.is_os_debian() && !this.has_allegro_integration()) || this.has_allegro_integration())) {// No agentscreenservice sobre debian
			enabled = false;
		}
		return enabled;
	};

	Application.prototype.getAgentScreenServiceBaseDomain = function(subdomain) {
		return "wss://" + (subdomain ? subdomain : "agentscreenservice-" + this.m_serverAddress);
	};

	Application.prototype.getAgentScreenLocalServiceURL = function() {
		return this.m_vcc_info.ScreenRecording && this.m_vcc_info.ScreenRecording.Enabled ? "wss://agentscreenservice.i6.inconcert.cc:" + this.m_vcc_info.ScreenRecording.PORT : undefined;
	};

	Application.prototype.getAgentScreenServiceFPS = function() {
		return this.m_vcc_info.ScreenRecording && this.m_vcc_info.ScreenRecording.Enabled ? parseInt(this.m_vcc_info.ScreenRecording.FPS, 10) : undefined;
	};

	Application.prototype.getAgentScreenServiceBitrate = function() {
		return this.m_vcc_info.ScreenRecording && this.m_vcc_info.ScreenRecording.Enabled ? parseInt(this.m_vcc_info.ScreenRecording.BITRATE, 10) : undefined;
	};

	/* INCO-64: AgentMonitorService */
	Application.prototype.getAgentMonitorServiceBaseDomain = function(subdomain) {
		return "wss://" + ((subdomain ? subdomain : "agentmonitorservice-") + this.m_serverAddress);
	};

	Application.prototype.getAgentMonitorLocalServiceURL = function() {
		return (this.m_vcc_info.AgentMonitoring && this.m_vcc_info.AgentMonitoring.Enabled ? "wss://agentmonitorservice.i6.inconcert.cc:" + this.m_vcc_info.AgentMonitoring.PORT : undefined);
	};

	Application.prototype.getAgentMonitorServiceEnabled = function() {
		return !this.is_os_debian();
	};

	/*-----------------------*/

	Application.prototype.getAuthApplications = function() {
		return this.m_vcc_info.AuthApplications || {};
	};

	Application.prototype.getMaxDraftsAllowedToSave = function() {
		return this.m_vcc_info.DraftsConfig && this.m_vcc_info.DraftsConfig.Enabled && this.m_vcc_info.DraftsConfig.MaxDrafts;
	};

	Application.prototype.getMaxDraftsSizeAllowed = function() {
		return (this.m_vcc_info.DraftsConfig && this.m_vcc_info.DraftsConfig.Enabled && this.m_vcc_info.DraftsConfig.MaxSizeDrafts) || 1024;
	};

	Application.prototype.getDetabasesConfig = function() {
		return this.m_vcc_info.Databases && this.m_vcc_info.Databases.Enabled && this.m_vcc_info.Databases || {
			LimitNumber : 3,
			Sizing : 20480,
			RowsNumber : 3000,
			TimeStamp : 365
		};
	};

	Application.prototype.getMaxConfigurationsAllowed = function(key) {
		let limit;
		if (this.m_vcc_info.ConfigLimits) {
			limit = this.m_vcc_info.ConfigLimits[key];
		}
		return limit;
	};

	//loads settings for the logged user
	Application.prototype.loadSettings = function(settings) {
		this.m_settings = settings;
	};

	Application.prototype.getSettings = function() {
		return this.m_settings;
	};

	/*
		Variables obtenidas desde el IR.
		Lo que se haya definido en el LUA "ir.lua" en la variable "m_formatHandlers"

		Ej:
		{
			{
				id: "{AGENT}", name: "Agent Id"
			}
		}
	*/
	Application.prototype.loadIRVariables = function(sorted, clientSideVariables, canIncludeFn, callback = $.noop) {
		canIncludeFn = canIncludeFn || function() {return true;};
		clientSideVariables = clientSideVariables || [];

		let IRInfo = this.getIRInfo();
		// Name : Variable
		let IRVariables = [];
		IRInfo.variables.forEach(variable => {
			let clonedVariable = inConcert.Utils.cloneObject(variable);
			IRVariables.push(clonedVariable);
		});

		// Filtra las variables que se obtuvieron del IR
		let map = {};
		let filteredVariables = [];
		IRVariables.forEach(variable => {
			let swhowInAgent = variable.showIn && variable.showIn.agent ? variable.showIn.agent : null;
			if (swhowInAgent) {
				if (!swhowInAgent.everywhere || canIncludeFn(swhowInAgent, variable.id.toUpperCase())) {
					filteredVariables.push(variable);
					map[variable.id.toUpperCase()] = variable;
				}
			}
		});

		// Inserto las variables client side
		clientSideVariables.forEach(variable => {
			if (!map[variable.id.toUpperCase()]) {
				filteredVariables.push(variable);
			}
		});

		// Ordeno las variables segun la traduccion del name
		let variables = filteredVariables;
		if (sorted) {
			variables = inConcert.Utils.sortArray(variables, "name", "asc");
		}

		callback(true, variables);
	};

	/*
	*/
	Application.prototype.loadVariablesForAccountMessages = function(sorted, callback = $.noop, extra_can_include) {
		extra_can_include = extra_can_include || function() {
			return true;
		};
		const canIncludeFn = function(swhowInAgent, variableId) {
			if (swhowInAgent.where && swhowInAgent.where.accountsMessages && extra_can_include(swhowInAgent.where.accountsMessages, variableId)) {
				return true;
			}
			return false;
		};
		this.loadIRVariables(sorted, null, canIncludeFn, callback);
	};

	/*
	*/
	Application.prototype.loadVariablesForFlowDesigner = function(sorted, callback = $.noop) {
		const canIncludeFn = function(swhowInAgent) {
			if (swhowInAgent.where && swhowInAgent.where.flowDesigner) {
				return true;
			}
			return false;
		};
		this.loadIRVariables(sorted, null, canIncludeFn, callback);
	};

	/*
	*/
	Application.prototype.loadVariablesForSignatures = function(sorted, callback = $.noop) {
		const canIncludeFn = function(swhowInAgent) {
			if (swhowInAgent.where && swhowInAgent.where.signatures) {
				return true;
			}
			return false;
		};
		this.loadIRVariables(sorted, undefined, canIncludeFn, callback);
	};

	Application.prototype.loadVariablesForBatches = function(sorted, callback = $.noop) {
		const canIncludeFn = function(swhowInAgent) {
			if (swhowInAgent.where && swhowInAgent.where.batches) {
				return true;
			}
			return false;
		};
		this.loadIRVariables(sorted, undefined, canIncludeFn, callback);
	};

	Application.prototype.updateBasicSettings = function(user){
		//Only name
		this.m_userFirstName = user.Name;
		this.m_userLastName = user.LastName;
		this.m_userFirstSurname = user.FirstSurname;
		this.m_userLastSurname = user.LastSurname;
		this.m_initialState = user.initialState;
	};

	Application.prototype.updateSettings = function(newSetting)
	{
		if (!newSetting || !this.m_settings) {
			return;
		}
		
		this.updateBasicSettings(newSetting.personal);
		this.m_country = newSetting.personal.countryCode;
		this.m_area = newSetting.personal.areaCode;
		this.m_enterAsKeyToSend = newSetting.personal.EnterAsKeyToSend; // This is the unique advance value added in issue #15982
		this.m_userIdleThreshold       = newSetting.personal.idleThreshold;
		this.m_userSoundNotifications  = newSetting.personal.soundNotifications === "1" ? true : false;
		this.m_userVisualNotifications = newSetting.personal.visualNotifications === "1" ? true : false;

		if (this.m_settings.InteractionClients) {
			if (this.m_settings.InteractionClients.CALL && newSetting.telephony) {
				this.m_settings.InteractionClients.CALL.phoneType = newSetting.telephony.phoneType;
				this.m_settings.InteractionClients.CALL.external_phone = newSetting.telephony.external_phone;
			}
		}

		if (!this.m_userClients) {
			this.m_userClients = {};
		}
		if (newSetting.personal.layouts) {
			this.m_userClients.Layouts = inConcert.Utils.cloneObject(newSetting.personal.layouts);
		}
	};

	/*
		Summarizes queued interactions by type accross all the user campaigns
		TODO, this belongs to the InteractionManager class
	*/
	Application.prototype.summarizeQueuedInteractions = function() {
		var summary = {};
		var campaigns = this.getCampaigns();
		for (var i = 0; i < campaigns.length; i++) {
			var counters = campaigns[i].interactionCounters;
			//this loop just accumulate totals by counter name
			for (var interactionType in counters) {
				if (counters.hasOwnProperty(interactionType)) {
					for (var counterName in counters[interactionType]) {
						if (counters[interactionType].hasOwnProperty(counterName)) {
							var total = parseInt(counters[interactionType][counterName], 10);
							if (!isNaN(total)) {
								switch (counterName) {
									case "Waiting":
									case "Assigned":
										break;
									case "Queued":
										if (!summary[interactionType]) {
											summary[interactionType] = 0;
										}
										summary[interactionType] += total;
										break;
								}
							}
						}
					}
				}
			}
		}
		return summary;
	};

	//loads user groups for the logged user
	Application.prototype.loadCampaigns = function(campaigns) {
		this.m_campaigns = [];
		for (var campaign in campaigns) {
			if (campaigns.hasOwnProperty(campaign)) {
				var campaignObj = new inConcert.Campaign(campaigns[campaign]);
				this.m_campaigns.push(campaignObj);
			}
		}
	};

	//returns a team member
	//TODO when pagination for members is enabled this should change to get only ONE member from the server not the whole bunch
	Application.prototype.getTeamMember = function( member_id, onLoadedFn, fromCache )
	{
		const self = this;

		onLoadedFn = onLoadedFn || $.noop;

		let member_id_without_vcc = member_id;

		if( member_id.match( inConcert.app.idAtVCCRegExp ) ) {
			member_id_without_vcc = RegExp.$1;
		}

		var searchFn = function( ) {
			if ( !self.cached_members ) {
				return;
			}
			let team_member;
			$.each( self.cached_members, function( _, tm ) {
				if ( tm.Id.toLowerCase() === member_id_without_vcc.toLowerCase() ) {
					team_member = tm;
					return false; // break
				}
			});
			return team_member;
		};

		this.getTeamMembers(
			function( members ) {
				if ( !members ) {
					onLoadedFn( false );
				}
				var member = searchFn();
				if (member) {
					if (!member.avatar) {
						member.avatar = self.__makeAgentAvatar(member_id_without_vcc);
					}
					if (!member.name) {
						member.name = self.getMemberName(member);
					}
				}
				onLoadedFn( member );
			},
			fromCache
		);
	};
	
	Application.prototype.getMemberName = function(member) {
		if (!member) {
			return "";
		}
		let name = member.Id;
		let can_name = false;
		if ( ( member.Name || "" ).trim().length > 0 ) {
			name = member.Name;
			can_name = true;
		}
		if ( ( member.FirstSurname || "" ).trim().length > 0 ) {
			name = ( can_name ? ( name + " " ) : "" ) + member.FirstSurname;
		}
		return name;
	};

	Application.prototype.getInstanceTelephonyCodes = function()
	{
		return this.m_telephone_codes;
	};

	Application.prototype.loadSMSBindTypes = function(callback)
	{
		callback = callback || $.noop;
		var list = [
			{ id: "R" , description: "<%=_T('Receiver')%>"    },
			{ id: "TR", description: "<%=_T('Transceiver')%>" },
			{ id: "T" , description: "<%=_T('Transmitter')%>" }
		];
		var sorted_list = inConcert.Utils.sortArray(list, "description", "asc");
		window.setTimeout(function() {
			callback(true, sorted_list);
		});
	};

	Application.prototype.loadSMSDataCoding = function(callback)
	{
		callback = callback || $.noop;
		var list = [
			{ id: "Default" , code: -1, description: "<%=_T(' Server Default')%>"      }, // uses server config
			{ id: "GSM7"    , code: 0 , description: "<%=_T('GSM 03.38')%>"            }, // always uses GSM7
			{ id: "LATIN1"  , code: 3 , description: "<%=_T('Latin 1 (ISO-8859-1)')%>" },
			{ id: "UNICODE" , code: 8 , description: "<%=_T('Unicode (UCS2)')%>"       },
			{ id: "UTF8"    , code: 11, description: "<%=_T('UTF-8')%>"                }
		];

		var sorted_list = inConcert.Utils.sortArray(list, "description", "asc");
		window.setTimeout(function() {
			callback(true, sorted_list);
		});
	};

	Application.prototype.loadSMSUnicodeEndianess = function(callback)
	{
		callback = callback || $.noop;
		var list = [
			{ id: "LittleEndian", description: "<%=_T('Little Endian')%>" },
			{ id: "BigEndian"   , description: "<%=_T('Big Endian')%>"    }
		];
		var sorted_list = inConcert.Utils.sortArray(list, "description", "asc");
		window.setTimeout(function() {
			callback(true, sorted_list);
		});
	};

	Application.prototype.loadLanguages = function(callback)
	{
		const self = this;
		callback = callback || $.noop;

		if ( this.cached_languages ) {
			callback(true, this.cached_languages);
			return;
		}

		inConcert.invoke({
			url: this.getBaseDomain() + "/inconcert/apps/administration/spell_checker/search/",
			type: "POST",
			data: JSON.stringify({columns: [ "id", "name", "description" ]}),
			dataType: "json",
			contentType: "application/json",
			success: function( languages )
			{
				callback(true, languages.rows);
				self.cached_languages = languages.rows;
			},
			error: function(/*err_type, text_status*/)
			{
				callback(false);
			}
		});
	};

	Application.prototype.loadAgentScreenService = function(callback)
	{
		const self = this;
		callback = callback || $.noop;
		this.AgentScreenService = {
			enabled : false,
			status : "not_found",
			version : "",
			capturing : false,
			needsUpdate : false
		};

		// Comprobar que el usuario actual tenga habilitada la app de interacciones. Caso contrario llamar a callback y salir
		if (!inConcert.permissions.to_application("interaction")) {
			callback();
			return;
		}

		self.AgentScreenService.enabled = inConcert.app.getAgentScreenServiceEnabled();
		if(!self.AgentScreenService.enabled)
		{
			callback();
			return;
		}

		const WS_URL = inConcert.app.getAgentScreenLocalServiceURL();

		self.AgentScreenService.needsUpdate = true;

		var checkRetry = 1;
		const maxRetries = 3;
		let timeoutExpired = false;

		// Bucle de espera hasta que se actualice el AgentScreenService
		// NOTA: Si en la primera llamada se detecta que se necesita actualizar, el servicio comienza el proceso de pararse/desinstalarse/reinistalarse/rearrancarse, 
		//	y no no respondera al WebSocket durante dicho proceso, lo cual generara errores 'normales' en la consola del navegador
		// Se agrega un timeout externo pues el websocket no tiene forma de indicarle uno. Si expira el timeout, no se volvera a intentar la conexion y
		// posteriormente en los eventos onopen, onclose, onerror ignorara la funcionalidad establecida.
		const needsUpdateCheckInterval = 10000;
		function needsUpdateCheck() {
			if(checkRetry < maxRetries)
			{
				let socket    = new WebSocket(WS_URL);
				let timeoutId = setTimeout(() =>
				{
					console.error("[AgentScreenService] Connection tiemout to the server: ", WS_URL);
					timeoutExpired = true;
					callback();
				}, needsUpdateCheckInterval);

				socket.binaryType = "arraybuffer";
				socket.onopen = () =>
				{
					if (!timeoutExpired) {
						clearTimeout(timeoutId);
						console.log(`[AgentScreenService:onopen] Connected to ${WS_URL}`);
						var message = {
							command : "GET_STATUS"
						};
						socket.send(JSON.stringify(message));
					}
					else {
						socket.close();
					}
				};

				// Listen for messages. If timeout this will never happen
				socket.onmessage = (event) =>
				{
					if (!timeoutExpired)
					{
						clearTimeout(timeoutId);
						var message = JSON.parse(event.data);
						if(!message.success)
						{
							console.error(" [AgentScreenService:onmessage] message.error: ", message.error);
							self.AgentScreenService.needsUpdate = false;
							callback();
						}
						else
						{
							self.AgentScreenService.status = message.status;
							self.AgentScreenService.version = message.version;
							self.AgentScreenService.capturing = message.capturing;
							self.AgentScreenService.needsUpdate = message.needsUpdate;
						}
						socket.close();
						if(self.AgentScreenService.needsUpdate) {
							checkRetry++;
							window.setTimeout(needsUpdateCheck, needsUpdateCheckInterval);
						}
						else {
							callback();
						}
					}
				};

				socket.onclose = (event) =>
				{
					if (!timeoutExpired)
					{
						clearTimeout(timeoutId);
						let code    = (event && event.code)  ? event.code : 0;
						let reason  = (event && event.reason) ? event.reason : "";
						console.log(`[AgentScreenService:onclose] Close socket with code: ${code} and reason: ${reason}`);
					}
				};

				socket.onerror = (event) =>
				{
					if (!timeoutExpired)
					{
						clearTimeout(timeoutId);
						console.log("[AgentScreenService:onerror] WebSocket Error: ",event);
						if(self.AgentScreenService.needsUpdate && checkRetry > 1) { //Avoid retries if first failed
							window.setTimeout(needsUpdateCheck, needsUpdateCheckInterval);
						}
						else {
							callback();
						}
					}
				};
			}
		}
		needsUpdateCheck();
	};

	Application.prototype.loadSkills = function(callback)
	{
		callback = callback || $.noop;
		inConcert.invoke({
			url: this.getBaseDomain() + "/inconcert/apps/administration/skills/search/",
			type: "POST",
			data: JSON.stringify({columns: ["Id", "Name", "Description", "DefaultValue"]}),
			dataType: "json",
			contentType: "application/json",
			success: function( skills )
			{
				callback(true, skills.rows);
			},
			error: function(/*err_type, text_status*/)
			{
				callback(false);
			}
		});
	};

	Application.prototype.loadTimezones = function(callbackFn = $.noop)
	{
		const self = this;

		if (this.m_timezones) {
			callbackFn(true, this.m_timezones);
			return;
		}

		inConcert.invoke({
			url: this.getBaseUrl() + "/timezones/find_all/",
			dataType : "json",
			success: function(timezones) {
				self.m_timezones = timezones;
				callbackFn(true, timezones);
			},
			error: function(/*response, error*/) {
				callbackFn(false);
			}
		});
	};

	Application.prototype.loadPackages = function(interaction_type, from_cache, callbackFn = $.noop)
	{
		const self = this;

		if (from_cache && this.m_packages && this.m_packages[interaction_type]) {
			callbackFn( true, this.m_packages[interaction_type] );
			return;
		}

		let url = this.getBaseDomain() + "/inconcert/apps/administration/channels/load_packages/" + interaction_type;
		inConcert.invoke({
			url     : url,
			dataType: "json",
			success : function(packages) {
				self.m_packages = self.m_packages || {};
				self.m_packages[interaction_type] = packages;
				callbackFn(true, packages);
			},
			error: function(_, error) {
				callbackFn(false, error);
			}
		});
	};

	Application.prototype.loadAccountDids = function(interaction_type, callbackFn = $.noop)
	{
		const url = this.getBaseDomain() + "/inconcert/apps/administration/channels/load_account_dids/" + interaction_type;
		inConcert.invoke({
			url     : url,
			dataType: "json",
			success : function(packages) {
				callbackFn(true, packages);
			},
			error: function(_, error) {
				callbackFn(false, error);
			}
		});
	};

	Application.prototype.loadSystemGroups = function( callbackFn )
	{
		callbackFn = callbackFn || $.noop;
		var uri = inConcert.app.getBaseDomain() + "/inconcert/apps/administration/groups/get_system_groups/";

		var request = inConcert.invoke({
			url: uri,
			dataType : "json",
			success: function(system_groups) {
				callbackFn(true, system_groups);
			},
			error: function() {
				callbackFn(false);
			}
		});
		return request;
	};

	Application.prototype.loadAllRoles = function( callbackFn, from_cache, include_allusers, mustIncludeRoleFullSystemAccess = false)
	{
		const self = this;

		callbackFn = callbackFn || $.noop;
		include_allusers = include_allusers || 0;
		/*
		*/
		const filterFn = (all_roles)=> {
			if (mustIncludeRoleFullSystemAccess) {
				return all_roles;
			}
			return Object.keys(all_roles)
						.reduce((currentRole, roleId) => {
							if (all_roles[roleId].IsFullSystemAccess !== "1")
								currentRole[roleId] = all_roles[roleId];
							return currentRole;
						},
						{}
					);
		};
		if ( from_cache && this.cached_roles ) {
			callbackFn(true, filterFn(this.cached_roles));
			return;
		}
		const uri = inConcert.app.getBaseDomain() + "/inconcert/apps/administration/groups/get_all_groups/" + include_allusers;
		const request = inConcert.invoke({
			url: uri,
			dataType : "json",
			success: function( all_roles ) {
				callbackFn(true, filterFn(all_roles));
				self.cached_roles = all_roles;
			},
			error: function() {
				callbackFn(false);
			}
		});
		return request;
	};

	/*
		Carga Roles de los cuales sus permisos puedan ser sobrescritos.
		Funcion utilizada por el ABM de scripting launcher.
	*/
	Application.prototype.loadRolesThatAllowOverwriting = function(callback = $.noop)
	{
		/*
			Valida si el rol se puede incluir
		*/
		const canAdd = (role) => {
			if (this.has_allegro_integration()) {
				if (role.IsSystemGroup === "1") return false; // NO
			}
			// if (role.IsFullSystemAccess === "1") {
			// 	if (!this.getLoggedUserInfo().IsFullSystemAccess) return false; // NO agrego el rol full-system-access, si El Usuario NO es un SuperUser
			// }
			return true; // YES add
		};
		// ------------------
		const roles = {};
		return this.loadAllRoles(
			function(success, all_roles) {
				if (!success)
					return callback(false);
				for (let role_id in all_roles) {
					if (all_roles.hasOwnProperty(role_id)) {
						const role = all_roles[role_id];
						if (canAdd(role)) roles[role_id] = role;
					}
				}
				callback(true, roles);
			},
			false, // from_cache
			0,     // include_allusers
			false   // mustIncludeRoleFullSystemAccess
		);
	};

	Application.prototype.loadDistributionAlgorithms = function(callback = $.noop)
	{
		const algorithms = [
			{id: "CIRCULAR", description: "<%=_T('Circular')%>"   },
			{id: "QUEUED"  , description: "<%=_T('Queued')%>"     },
			{id: "SKILL"   , description: "<%=_T('Skill based')%>"}
		];

		const proxyTypes_algorithms = {
			"CALL"      : algorithms,
			"CHAT"      : algorithms,
			"FACEBOOK"  : algorithms,
			"APPSTORE"  : algorithms,
			"GOOGLEPLAY": algorithms,
			"RCS"       : algorithms,
			"IM"        : algorithms,
			"INSTAGRAM" : algorithms,
			"MAIL"      : algorithms,
			"SMS"       : algorithms,
			"TWITTER"   : algorithms,
			"WEBCONTACT": algorithms,
			"WHATSAPP"  : algorithms,
			"YOUTUBE"   : algorithms
		};

		callback(true, proxyTypes_algorithms);
	};

	Application.prototype.loadIMProtcols = function(callbackFn = $.noop)
	{
		var protocols = [
//			{id: "others" ,        description: "<%=_T('Others')%>"        } , ??? 
			{id: "prpl-jabber", description: "<%=_T('GTalk')%>"       },
			{id: "prpl-msn"   , description: "<%=_T('Windows Live')%>"},
			{id: "prpl-yahoo" , description: "<%=_T('Yahoo!')%>"      }
		];

		callbackFn( true, inConcert.Utils.sortArray( protocols, "description", "asc") );
	};

	Application.prototype.getDefaultFont = function()
	{
		return {
			family:  {id: "Arial,Helvetica", name: "Arial, Helvetica"},
			size  : "12",     // px
			color : "#000000" // Negro en Hexadecimal
		};
	};

	Application.prototype.loadFontsList = function()
	{
		return [
			{id: "Arial,Helvetica"      , name: "<%=_T('Arial, Helvetica')%>"},
			{id: "Calibri"              , name: "<%=_T('Calibri')%>"         },
			{id: "Courier New,Courier"  , name: "<%=_T('Courier New')%>"     },
			{id: "Georgia"              , name: "<%=_T('Georgia')%>"         },
			{id: "Times New Roman,Times", name: "<%=_T('Times New Roman')%>" },
			{id: "Verdana,Geneva"       , name: "<%=_T('Verdana')%>"         }
		];
	};

	Application.prototype.loadTranslatedChannels = function()
	{
		return {
			APPSTORE  : "<%=_T('Appstore')%>",
			CALL      : "<%=_T('Phone')%>",
			CHAT      : "<%=_T('Web Chat')%>",
			FACEBOOK  : "<%=_T('Facebook')%>",
			GOOGLEPLAY: "<%=_T('Googleplay')%>",
			IM        : "<%=_T('IM')%>",
			INSTAGRAM : "<%=_T('Instagram')%>",
			MAIL      : "<%=_T('E-mail')%>",
			RCS       : "<%=_T('Rcs')%>",
			SMS       : "<%=_T('SMS')%>",
			TWITTER   : "<%=_T('Twitter')%>",
			VOICEMAIL : "<%=_T('Voice Mail')%>",
			WEBCONTACT: "<%=_T('Contact Form')%>",
			WHATSAPP  : "<%=_T('Whatsapp')%>",
			YOUTUBE   : "<%=_T('Youtube')%>"
		};
	};

	Application.prototype.getTranslatedChannel = function(channel = "")
	{
		channel = channel.toUpperCase();
		let translatedChannels = this.loadTranslatedChannels();
		return translatedChannels[channel] || channel;
	};

	Application.prototype.loadMailSignatures = function(callback, from_cache, all_columns)
	{
		const self = this;
		callback = callback || $.noop;

		if ( from_cache && this.cached_mail_signatures ) {
			callback(true, this.cached_mail_signatures);
			return;
		}
		var post_data = {};
		if ( !all_columns ) {
			post_data.columns = [ "Id", "Name", "Description" ]; // default columns
		}

		inConcert.invoke({
			url: this.getBaseDomain() + "/inconcert/apps/administration/signatures/search/",
			type: "POST",
			data: JSON.stringify( post_data ),
			dataType: "json",
			contentType: "application/json",
			success: function( signatures )
			{
				callback(true, signatures.rows);
				self.cached_mail_signatures = signatures.rows;
			},
			error: function(/*err_type, text_status*/)
			{
				callback(false);
			}
		});
	};

	Application.prototype.getMailSignature = function(signature_id)
	{
		if ( typeof signature_id !== "string" ) {
			return;
		}
		var signature;
		$.each(this.cached_mail_signatures, function(_, current_signature) {
			if ( current_signature.Id === signature_id ) {
				signature = current_signature;
				return false; //break
			}
		});
		return signature;
	};

	Application.prototype.getCampaignAccountsForTransfer = function(callback)
	{
		callback = callback || $.noop;

		return inConcert.invoke({
			url: this.getBaseDomain() + "/inconcert/apps/administration/campaigns/get_campaign_accounts_for_transfer/",
			type: "GET",
			timeout: 300000, // 5 minutes
			success: function( response )
			{
				callback( true, response );
			},
			error:function( err /*, text_status*/) {
				callback( false, err );
			}
		});
	};

	Application.prototype.GetWebcontactCampaignAccounts = function(callback = $.noop)
	{
		return inConcert.invoke({
			url: this.getBaseDomain() + "/inconcert/apps/administration/campaigns/get_webcontact_campaign_accounts/",
			type: "GET",
			timeout: 300000, // 5 minutes
			success: function( response ) {
				callback( true, response );
			},
			error:function( err /*, text_status*/) {
				callback( false, err );
			}
		});
	};

	/*
	*/
	Application.prototype.getVccCampaigns = function(forceReload, callback = $.noop)
	{
		const self = this;
		if (!forceReload && this.m_cachedVccCampaigns.timestamp) {
			// si el tiempo del cache NO ha expirado, retorna las campanas cacheadas.
			if (this.m_cachedVccCampaigns.timestamp + this.m_cachedVccCampaigns.maxCacheTimeMs > new Date().getTime()) {
				// console.log("Application.prototype.getVccCampaigns 1 -> from cache");
				callback(true, this.m_cachedVccCampaigns.data);
				return;
			}
		}
		// console.log("Application.prototype.getVccCampaigns 2 -> from database");
		// realizo request para obtener las campanas
		return inConcert.invoke({
			url: this.getBaseUrl() + "/campaigns/load_vcc_campaigns/",
			type: "GET",
			timeout: 300000, // 5 minutes
			success: function(response) {
				let campaigns = [];
				for (let i = 0; i < response.length; i++) {
					let campaignObj = new inConcert.Campaign(response[i]);
					campaigns.push(campaignObj);
				}
				self.m_cachedVccCampaigns.data = campaigns;
				self.m_cachedVccCampaigns.timestamp = new Date().getTime();
				callback(true, campaigns);
			},
			error:function(err /*, text_status*/) {
				callback(false, err);
			}
		});
	};

	/*
	*/
	Application.prototype.getVccCampaign = function(campaignId, forceReload, callback = $.noop)
	{
		this.getVccCampaigns(forceReload, function(success, campaigns) {
			if (!success) {
				callback(false);
				return;
			}
			const campaign = campaigns.find((c) => {
				return c.Id === campaignId;
			});
			callback(true, campaign);
		});
	};

	/*
		Returns the joined list of teams for all the campaigns of an agent
	*/
	Application.prototype.getTeamMembers = function(callback, fromCache)
	{
		const self = this;

		callback = callback || $.noop;

		if ( fromCache && this.cached_members ) {
			callback(this.cached_members);
			return;
		}
		return inConcert.invoke({
			url: this.getBaseUrl() + "/campaigns/get_members/",
			type: 'POST',
			data: JSON.stringify({}),
			dataType: "json",
			contentType: "application/json",
			timeout: 300000, // 5 minutes
			success: function(response)
			{
				self.cached_members = [];
				self.cached_members.find = function(id) {
					for (var tm_i = 0; tm_i < this.length; tm_i++) {
						var member = this[tm_i];
						if (member.Id === id) {
							return member;
						}
					}
					return;
				};
				for (var member in response) {
					if (response.hasOwnProperty(member)) {
						var memberObj = new inConcert.TeamMember(response[member]);
						self.cached_members.push(memberObj);
					}
				}

				callback(self.cached_members);
			},
			error:function() {
				callback( false );
			}
		});
	};

	Application.prototype.getTeamMembersWithState = function(filter, callback)
	{
		callback = callback || $.noop;

		return inConcert.invoke({
			url: this.getBaseUrl() + "/campaigns/get_members_with_state/",
			type: 'POST',
			data: JSON.stringify({ filter: filter }),
			dataType: "json",
			contentType: "application/json",
			timeout: 300000, // 5 minutes
			success: function(response)
			{
				var members = [];
				for (var member_id in response.members) {
					if (response.members.hasOwnProperty(member_id)) {
						var memberObj = new inConcert.TeamMember(response.members[member_id]);
						members.push(memberObj);
					}
				}
				response.members = members;
				callback(true, response);
			},
			error:function() {
				callback( false );
			}
		});
	};

	Application.prototype.getCampaigns = function() {
		return this.m_campaigns;
	};

	Application.prototype.getCampaignsByAddress = function(addressType, enabled) {
		return this.m_campaigns.filter(function(campaign) {
			return campaign.has_accounts_of_proxy(addressType, enabled);
		});
	};

	Application.prototype.hasInteractionCampaignForAddress = function(addressType) {
		const self = this;
		return !!this.m_campaigns.find(function(campaign) {
			return campaign.has_accounts_of_proxy(addressType, true) &&
				(inConcert.permissions.to_action("interaction", "int_conversations_mng", "view", campaign.Id) || 
					!self.has_allegro_integration() && inConcert.permissions.to_action("supervisor", "sup_monitoring", "interaction", campaign.Id));
		});
	};

	Application.prototype.getCampaignAddresses = function(campaignId, addressType, accountId = "") {
		let campaignInfo = this.getCampaign(campaignId);
		if (!campaignInfo) {
			return null;
		}
		let accounts_channel = campaignInfo.get_accounts_channel(addressType);
		if (!accountId) {
			return accounts_channel;
		}
		for (let i = 0; i < accounts_channel.length; i++) {
			let account = accounts_channel[i];
			if (accountId.toLocaleLowerCase() === account.id.toLowerCase()) {
				return [account];
			}
		}
		return null;
	};

	Application.prototype.getCampaignPhoneAccount = function(campaignId, addressType, accountId)
	{
		let phone_accounts = this.getCampaignAddresses(campaignId, addressType, accountId);
		return phone_accounts[0];
	};

	Application.prototype.getCampaignPhonePrefix = function(campaignId, accountId)
	{
		let outboundPrefix = "";
		let phone_account = this.getCampaignPhoneAccount(campaignId, "CALL", accountId);
		if (!phone_account) {
			return outboundPrefix;
		}
		let outgoingConfig = phone_account.outgoingConfig;
		// Si es un Vcc SIN integracion con allegro el prefijo que se ha configurado en la GUI se encuentra en "outboundPrefix" (comienza en minuscula)
		outboundPrefix = outgoingConfig.outboundPrefix || "";
		if (this.has_allegro_integration()) {
			// Si es un Vcc CON integracion con Allegro el prefijo que se ha configurado en la GUI se encuentra en "OutboundPrefix" (comienza en mayuscula)
			outboundPrefix = outgoingConfig.OutboundPrefix || outboundPrefix;
		}
		return outboundPrefix;
	};

	/*
		allow the agent to call a cab, but not Uruguay's Speaking clock
	*/
	Application.prototype.validatePhoneNumber = function(number, interactionType)
	{
		if (!number || number.length <= 2) {
			return false;
		}
		return number.match(/^[\+ \- \(\)0-9_.*#]+$/);
	};

	/*
		Processes the notification of campaign counters changed (new queued interactions, etc)
		the counters received are only the ones that changed formmated as campaign.interaction.state.total
	*/
	Application.prototype.updateCampaignCounters = function(counters) {
		var shouldRefresh = false;
		for (var campaign in counters) {
			if (counters.hasOwnProperty(campaign)) {
				var campaignCounters = counters[campaign];
				for (var i = 0; i < this.m_campaigns.length; i++) {
					if (this.m_campaigns[i].Id === campaign) {
						if (!this.m_campaigns[i].interactionCounters) {
							this.m_campaigns[i].interactionCounters = {};
						}
						//only the changed indicators by type are received in the notifications, so, only update those ones
						for (var interactionType in campaignCounters) {
							if (campaignCounters.hasOwnProperty(interactionType)) {
								for (var indicatorName in campaignCounters[interactionType]) {
									if (campaignCounters[interactionType].hasOwnProperty(indicatorName)) {
										if (!this.m_campaigns[i].interactionCounters[interactionType]) {
											this.m_campaigns[i].interactionCounters[interactionType] = {};
										}
										this.m_campaigns[i].interactionCounters[interactionType][indicatorName] = campaignCounters[interactionType][indicatorName];
									}
								}
							}
						}
						shouldRefresh = true;
						break;
					}
				}
			}
		}
		if (shouldRefresh) {
			this.m_mainWindow.updateCampaignCounters();
		}
	};

	Application.prototype.cleanCredentials = function() {
		inConcert.Storage.remove("credentials");
	};

	Application.prototype.getCampaign = function(id) {
		if (id) {
			id = id.toUpperCase();
		}
		for (var i = 0; i < this.m_campaigns.length; i++) {
			if (this.m_campaigns[i].Id.toUpperCase() === id) {
				return this.m_campaigns[i];
			}
		}
		return null;
	};
	
	Application.prototype.getDomainId = function(id) {
		return id + "@" + this.m_domain;
	};

	Application.prototype.getCompleteId = function() {
		return this.getDomainId(this.m_user);
	};

	Application.prototype.getDomain = function() {
		return this.m_domain;
	};

	Application.prototype.getLicensedAccountType = function () {
		return this.m_licensedAccountTypes;
	};

	Application.prototype.hasLicensesForChannel = function ( channel ) {
		channel = ( channel || "" ).toUpperCase();
		var licensed = this.getLicensedAccountType();
		if ( licensed && licensed [ channel ] ) {
			return true;
		}
		return false;
	};

	Application.prototype.getLoggedUserId = function() {
		return this.m_user;
	};
	Application.prototype.getLoggedUserLanguage = function() {
		return this.m_userLanguage;
	};

	Application.prototype.getLoggedUserFullId = function() {
		return this.getLoggedUserId() + "@" + this.get_vcc_id();
	};

	Application.prototype.getUserAvatarHTML = function(params) {
		if (!params)
			throw("Application.prototype.getAvatarHTML - params is required");
		const user = params.user;
		if (!user)
			throw("Application.prototype.getAvatarHTML - user is required");

		if (user.HasAvatar || user.Avatar || params.avatarUrl) {
			// Si tiene Avatar muestra la imagen.
			const imgDomId = params.imgDomId
				? `id="${params.imgDomId}"`
				: "";
			const width = params.width
				? `width="${params.width}"`
				: "";
			const height = params.height
				? `height="${params.height}"`
				: "";
			const className = params.className
				? `class="${params.className}"`
				: "";
			const src = params.avatarUrl
				|| this.__makeAgentAvatar(user.Id);
			//
			return `
				<img ${imgDomId} src="${src}" ${width} ${height} ${className} />
			`;
		}
		// Si NO tiene Avatar muestra las iniciales.
		const GetFirstLetter = (text = "") => text.substring(0, 1).toUpperCase();
		//
		const firstCharName = GetFirstLetter(user.Name);
		const firstCharFurstSurname = GetFirstLetter(user.FirstSurname);
		//
		return `
			<label class="avatarInitials initial${firstCharName}">
				<span>${firstCharName}${firstCharFurstSurname}</span>
			</label>
		`;
	};

	Application.prototype.getAvatarHTMLOfLoggedUser = function(params = {width: 24, height: 24}) {
		params.user = this.getLoggedUserInfo();
		return this.getUserAvatarHTML(params);
	};

	Application.prototype.__makeAgentAvatar = function(agent_id) {
		let agent_id_without_vcc = this.get_id(agent_id);
		return this.getBaseUrl() + "/users/get_avatar/" + agent_id_without_vcc;
	};

	Application.prototype.getLoggedUserAvatar = function() {
		return this.__makeAgentAvatar(this.getLoggedUserId());
	};

	Application.prototype.getUserGenericAvatar = function() {
		return this.getBaseUrl() + "/images/agentGenericAvatar.png";
	};

	Application.prototype.getContactGenericAvatar = function() {
		return this.getBaseUrl() + "/images/genericAvatar.png";
	};

	Application.prototype.getUserGroup = function() {
		return this.m_userRole;
	};

	Application.prototype.getUserAddresses = function() {
		return this.m_settings.InteractionClients;
	};

	Application.prototype.userHasAddress = function(channel) {
		var addresses = this.getUserAddresses();
		return addresses[channel]? true : false;
	};

	Application.prototype.getUserAddressesNames = function() {
		var addresses = this.getUserAddresses();
		var addresses_names = [];
		for (var address in addresses){
			if (addresses.hasOwnProperty(address)){
				addresses_names.push(address);
			}
		}
		return addresses_names;
	};

	/*
		Changes user global disposition-state (beware there is a campaign level disposition/status code)
	*/
	Application.prototype.changeUserState = function(stateId, callback = $.noop)
	{
		const self = this;
		if ( this.m_currentStatus === stateId ) {
			callback({
				success: false,
				reason : "user is already in this state",
				state  : stateId
			});
			return;
		}
		const url = this.getBaseUrl() + "/users/change_state/";
		inConcert.invoke({
			url: url,
			type: 'POST',
			data: JSON.stringify({ state : stateId}),
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				if (response.success) {
					self.m_currentStatus = stateId;
					self.__notifyUserStatusChange();
				}
				callback(response);
			},
			error:function(_, error) {
				callback(error);
			}
		});
	};
	
	Application.prototype.__notifyUserStatusChange = function() {
		let state = this.getStateObject(this.m_currentStatus) || { Id : this.m_currentStatus };
		this.agent_api_server.UserStateChanged(state);
	};

	Application.prototype.activateChannelAddress = function(channel, callbackFn) {
		const self = this;
		var channels = [];
		inConcert.logger.action({action_name : "activate_address", channel: channel, action_type: "agent"});
		var url = self.getBaseUrl() + "/activate_address/";
		if (typeof(channel) === "string") {
			channels.push(channel);
		}
		else {
			channels = channel;
		}
		inConcert.invoke({
			url: url,
			type : "POST",
			data : JSON.stringify(channels),
			contentType: "application/json",
			success: function(/*response, textStatus*/) {
				if (callbackFn) {
					callbackFn(true);
				}
			},
			error: function(/*response, error*/) {
				if (typeof(callbackFn) === "function") {
					callbackFn(false);
				}
			}
		});
	};

	Application.prototype.get_vcc_info = function() {
		return this.m_vcc_info;
	};

	Application.prototype.is_enabled_nlu = function() {
		let vcc_info = this.get_vcc_info();
		if (!vcc_info) {
			return false;
		}
		let bot_info = vcc_info.BotInfo;
		if (!bot_info || bot_info.Type.toUpperCase() === "LUIS") {
			if (bot_info && !bot_info.Enabled) {
				return false;
			}
			if (!vcc_info.MsCognitiveServices) {
				return false;
			}
			if (!vcc_info.MsCognitiveServices.SubscriptionsKeys) {
				return false;
			}
			let total_subscriptions = Object.keys(vcc_info.MsCognitiveServices.SubscriptionsKeys).length;
			if (total_subscriptions === 0) {
				return false;
			}
			return true;
		}
		else if (bot_info.Type.toUpperCase() === "RASA") {
			if (!bot_info.Enabled) {
				return false;
			}
			return true;
		}
		throw("Application.prototype.is_enabled_nlu - Unrecognized NLU type " + bot_info.Type);
	};

	Application.prototype.get_luis_info = function() {
		return this.m_luis_info;
	};

	Application.prototype.get_luis_api_version = function() {
		const luis_info = this.get_luis_info();
		if (!luis_info) {
			return;
		}
		return luis_info.API_VERSION;
	};

	Application.prototype.get_rasa_info = function() {
		return this.m_rasa_info;
	};

	Application.prototype.get_rasa_api_version = function() {
		const rasa_info = this.get_rasa_info();
		if (!rasa_info) {
			return;
		}
		return rasa_info.API_VERSION;
	};

	Application.prototype.has_configured_qna_maker_in_vcc = function() {
		const vcc_info = this.get_vcc_info();
		if (!vcc_info) {
			return false;
		}
		if (!vcc_info.QnAMaker) {
			return false;
		}
		if (vcc_info.QnAMaker.TotalSubscriptions === 0) {
			return false;
		}	
		return true;
	};

	Application.prototype.has_configured_knowledge_fusion_in_vcc = function() {
		const vcc_info = this.get_vcc_info();
		if (!vcc_info) {
			return false;
		}
		if (!vcc_info.QnAMakerV2) {
			return false;
		}
		if (vcc_info.QnAMakerV2.TotalSubscriptions === 0) {
			return false;
		}		
		return true;
	};

	Application.prototype.get_widgets_integrations_info = function() {
		return this.m_widgets_integrations_info;
	};

	Application.prototype.make_widgets_integrations_info = function(arr_widgets_integrations) {
		const self = this;
		this.m_widgets_integrations_info = null;
		if (!arr_widgets_integrations || arr_widgets_integrations.length === 0) {
			return;
		}
		this.m_widgets_integrations_info = {};
		arr_widgets_integrations.forEach(widget => {
			let widget_id = widget.widgetId.toLowerCase();
			self.m_widgets_integrations_info[widget_id] = widget;
		});
	};

	Application.prototype.get_vcc_tts_config = function() {
		let resp = {
			has_configrued: false
		};
		const vcc_info = this.get_vcc_info();
		if (!vcc_info) {
			return resp;
		}
		if ((vcc_info.ttsasrEngines || []).length === 0) {
			return resp;
		}
		// SI tiene TTS configurdo en el VCC
		resp.has_configrued = true;
		resp.ttsasrEngines = vcc_info.ttsasrEngines;
		return resp;
	};

	Application.prototype.get_vcc_all_apps_integrations = function() {
		const vcc_info = this.get_vcc_info();
		if (!vcc_info) {
			return false;
		}
		return vcc_info.AppsIntegrations || false;
	};

	Application.prototype.get_vcc_all_widgets_integrations = function() {
		let widgets = [];
		const vcc_apps_integrations = this.get_vcc_all_apps_integrations();
		if (!vcc_apps_integrations) {
			return widgets;
		}
		let widgets_integrations = this.get_widgets_integrations_info();
		if (!widgets_integrations) {
			return widgets;
		}
		for (let integration_id in vcc_apps_integrations) {
			if (vcc_apps_integrations.hasOwnProperty(integration_id)) {
				let vcc_integration = vcc_apps_integrations[integration_id];
				if ((vcc_integration.IsWidget || "0").toString() === "1" && (vcc_integration.Enabled || "0").toString() === "1") {
					let widget_id = integration_id.toLowerCase();
					let widget_info = widgets_integrations[widget_id];
					if (widget_info) {
						let vcc_widget = vcc_integration;
						vcc_widget.DisplayName = widget_info.displayName || vcc_widget.DisplayName; // traduccion del server segun el usuario que ha iniciado sesion.
						widgets.push(vcc_widget);
					}
				}
			}
		}
		return widgets;
	};

	Application.prototype.get_vcc_app_integration_config = function(type) {
		if (!type) {
			return false;
		}
		const apps_integrations = this.get_vcc_all_apps_integrations();
		if (!apps_integrations) {
			return false;
		}
		const apps_integrations_types = {
			helpdesk: "HelpDesk",
			mas     : "MarketingAndSales"
		};
		const app_integration_type = apps_integrations_types[type] || type;
		return apps_integrations[app_integration_type];
	};

	Application.prototype.is_enabled_apps_integrations = function(type) {
		const app_integration = this.get_vcc_app_integration_config(type);
		if (app_integration && app_integration.Enabled === "1") {
			return true;
		}
		return false;
	};

	//Funcion que se llama por notif de polling cuando se consume una licencia de helpdesk
	// despues de logueado el usuario (serverside)
	Application.prototype.enabled_helpdesk_integration = function() {
		if (this.has_helpdesk_integration()) {
			const vcc_info = this.get_vcc_info();
			//Seteo en true la disponibilidad de tickets y FullApp.
			//Esto es lo mismo que se hace al loguearse el usuario serverside 'GetVCCInfo > SetUserLoggedInfo'
			vcc_info.AppsIntegrations.HelpDesk.Licenses = vcc_info.AppsIntegrations.HelpDesk.Licenses || {};
			vcc_info.AppsIntegrations.HelpDesk.Licenses.Tickets = { isAvailable: true };
			this.m_vcc_info = vcc_info;

			//Habilitar panel de tickets de la app de Interacciones.
			const navigation = this.m_mainWindow.get_navigation();
			const $folder_tickets = navigation && navigation.$container && navigation.$container.find('[folder_id="integration_helpdesk_tickets"]');
			if ($folder_tickets) {
				$folder_tickets.removeClass('disabledFolderItem')
						.removeClass('infoFolderItem');
				$folder_tickets.prop('disabled', false);
				$folder_tickets.inconcert_tooltip({ message: "<%=_T('Tickets of HelpDesk')%>" });
			}

			this.m_mainWindow.showBannerMessage({
				message: "<%=_T('A helpdesk license was consumed.')%>",
				messageType: "Info",
				timeout: 10000
			});
		}
	};

	Application.prototype.has_mas_integration = function() {
		return this.is_enabled_apps_integrations("mas");
	};

	Application.prototype.has_helpdesk_integration = function() {
		return this.is_enabled_apps_integrations("helpdesk");
	};

	Application.prototype.has_widget_integration = function(widget_id) {
		widget_id = (widget_id || "").toLowerCase();
		return this.is_enabled_apps_integrations(widget_id);
	};

	Application.prototype.get_helpdesk_licenses = function() {
		const licenses = (this.get_vcc_app_integration_config("helpdesk") || {}).Licenses;
		return licenses || {};
	};

	Application.prototype.has_license_for_helpdesk_tickets = function() {
		const licenses = this.get_helpdesk_licenses();
		if (licenses.Tickets && licenses.Tickets.isAvailable) {
			return true;
		}
		return false;
	};

	Application.prototype.has_license_for_helpdesk_full_app = function() {
		const licenses = this.get_helpdesk_licenses();
		if (licenses.FullApp && licenses.FullApp.isAvailable) {
			return true;
		}
		return false;
	};

	Application.prototype.is_allegro_channel = function(channel) {
		if (!this.has_allegro_integration()) {
			return false;
		}
		return this.is_call_channel(channel);
	};

	Application.prototype.is_call_channel = function(channel = "") {
		channel = channel.toUpperCase();
		return channel === "CALL" || channel === "VOICEMAIL";
	};

	Application.prototype.has_allegro_integration = function() {
		if (this.m_vcc_info.AllegroIntegration && this.m_vcc_info.AllegroIntegration.enabled) {
			return true;
		}
		return false;
	};

	/*
		Funcion que retorna informacion del VCC de Allegro que ha sido migrado a Nativo
		result:
			IS NATIVE VCC
			IS ALLEGRO VCC
			IS MIGRATED ALLEGRO VCC
			IS MIGRATED ALLEGRO VCC
	*/
	Application.prototype.allegroVccMigrationInfo = function() {
		const info = {
			result            : null, // "IS NATIVE VCC" o "IS ALLEGRO VCC" o "IS MIGRATED ALLEGRO VCC" o "IS ROLLBACKED ALLEGRO VCC"
			isAllegroVcc      : this.has_allegro_integration(),
			AllegroIntegration: this.m_vcc_info.AllegroIntegration || {}
		};
		if (typeof info.AllegroIntegration.enabled === "undefined") {
			// es un VCC que nunca fue migrado, es decir siempre fue Nativo
			info.result = "IS NATIVE VCC";
			return info;
		}
		if (info.AllegroIntegration.enabled) {
			// Un Vcc de Allegro que fue rollbacked o que nunca fue migrado, es decir siempre fue de Allegro
			info.result = info.AllegroIntegration.Migrated ? "IS ROLLBACKED ALLEGRO VCC" : "IS ALLEGRO VCC";
			return info;
		}
		// un vcc de Allegro que fue migrado a Nativo o un Vcc Nativo que siempre fue Vcc Nativo
		info.result = info.AllegroIntegration.Migrated ? "IS MIGRATED ALLEGRO VCC" : "IS NATIVE VCC";
		return info;
	};

	Application.prototype.has_allegro_baragent = function() {
		return this.has_allegro_integration() && inConcert.permissions.to_application("interaction") && this.userHasAddress("CALL");
	};

	Application.prototype.loadAllegro = function() {
		const self = this;
		if (!this.has_allegro_baragent()) {
			return;
		}
		inConcert.BaragentPolling.start(function(success, info) {
			self.__on_baragent_polling(success, info);
		});
	};

	Application.prototype.__on_baragent_polling = function(success, info) {
		const self = this;
		if (!success) {
			// TODO: Hacer algo ???
			return;
		}
		if (!info.agent.notifications) {
			return;
		}
		var change_user_state = function() {
			if (!info.agent.notifications.changeState.newAgentStatus) {
				return;
			}
			let allegro_new_user_status = info.agent.notifications.changeState.newAgentStatus.state;

			const mapping_allegro_user_statuses_with_omnichannel = {
				"ACTIVO"  : {Id: "ACTIVE"  },
				"BREAK"   : {Id: "BREAK"   },
				"PAUSE"   : {Id: "NOACD"   },
				"PERSONAL": {Id: "AWAY"    },
				"TOILETTE": {Id: "BATHROOM"},
				"TRAINING": {Id: "BREAK"   }
			};

			let omnichannel_new_user_status = mapping_allegro_user_statuses_with_omnichannel[allegro_new_user_status.toUpperCase()];
			let state_id = omnichannel_new_user_status ? omnichannel_new_user_status.Id : allegro_new_user_status;
			let state_def = self.getStateObject(state_id);

			if (!state_def) {
				inConcert.logger.warn({
					action_name : "Application.prototype.__on_baragent_polling",
					action_type : "Change User State",
					message     : "Unrecognized to change Allegro User State. Unrecognized Allegro User State '" + allegro_new_user_status + "."
				});
				return;
			}
			state_id = state_def.Id || state_id;

			self.changeUserState(state_id, function(response)
			{
				if (!response.success) {
					inConcert.logger.warn({
						action_name : "Application.prototype.__on_baragent_polling",
						action_type : "Change User State",
						message     : "Updating availables states by poller!! Unable to change current state."
					});
					return;
				}
				self.m_mainWindow.setUserStatus(state_id);
				let new_state_name = state_def.Description || state_id;
				self.m_mainWindow.showBannerMessage({
					messageType : "Info",
					message     : format("<%=_T('Your current state has been changed to {0}.')%>", new_state_name),
					timeout     : 5000
				});
			});
		};

		change_user_state();
	};

	Application.prototype.getBaragentPolling = function() {
		if (!this.has_allegro_baragent()) {
			throw "allegro baragent not enabled";
		}
		return inConcert.BaragentPolling;
	};

	Application.prototype.getBaragentActions = function() {
		if (!this.has_allegro_baragent()) {
			throw "allegro baragent not enabled";
		}
		return inConcert.BaragentActions;
	};

	Application.prototype.is_system_vcc = function() {
		const vcc = this.get_vcc_id();
		if ((vcc || "").toLowerCase() === "system") {
			return true;
		}
		return false;
	};

	Application.prototype.get_vcc_max_proc_manual_rate = function() {
		return this.m_vcc_info.ProcessMaxManualRate || 8;
	};

	Application.prototype.get_vcc_id = function() {
		return this.m_vcc_info.Id || "";
	};

	Application.prototype.get_vcc_company_name = function() {
		return this.m_vcc_info.Company || "";
	};

	Application.prototype.get_vcc_company_address = function() {
		return this.m_vcc_info.CompanyAddress || "";
	};

	Application.prototype.get_vcc_contact_email = function() {
		return this.m_vcc_info.ContactEmail || "";
	};

	Application.prototype.get_vcc_contact_name = function() {
		return this.m_vcc_info.ContactName || "";
	};

	Application.prototype.get_vcc_description = function() {
		return this.m_vcc_info.Description || "";
	};

	Application.prototype.get_vcc_company_telephones = function() {
		var telephones = [];
		telephones.push(this.m_vcc_info.Telephone1 || "");
		telephones.push(this.m_vcc_info.Telephone2 || "");
		return telephones;
	};

	Application.prototype.has_configured_notifications_in_vcc = function() {
		if (this.m_vcc_info.InteractionNotifications && this.m_vcc_info.InteractionNotifications.Enabled === "1") {
			return true;
		}
		return false;
	};

	Application.prototype.getLoggedUserFirstName = function() {
		return this.m_userFirstName || "";
	};

	Application.prototype.getLoggedUserLastName = function() {
		return this.m_userLastName || "";
	};

	Application.prototype.getLoggedUserFirstSurname = function() {
		return this.m_userFirstSurname || "";
	};

	Application.prototype.getLoggedUserLastSurname = function() {
		return this.m_userLastSurname || "";
	};

	Application.prototype.getLoggedUserName = function() {
	/* return first name + first surname */
		var firstName = this.m_userFirstName;
		if ( !firstName ) {
			return this.m_user; // user id
		}
		var firstSurname = this.m_userFirstSurname;
		if ( typeof firstSurname !== "string" || firstSurname.trim().length === 0 ) {
			return firstName;
		}
		return firstName + " " + firstSurname;
	};

	Application.prototype.getLoggedUserCompleteName = function(params)
	{
		params = params || {
			include_user_name: true,
			include_user_last_name: true,
			include_user_first_surname: true,
			include_user_last_surname: true,
			include_user_id: false
		};
		const logged_user_info = this.getLoggedUserInfo();
		return inConcert.Utils.formatUserName(logged_user_info, params);
	};

	Application.prototype.getLoggedUserInfo = function() {
		return {
			Id                : this.m_user,
			Name              : this.m_userFirstName,
			LastName          : this.m_userLastName,
			FirstSurname      : this.m_userFirstSurname,
			LastSurname       : this.m_userLastSurname,
			Mail              : this.m_userEmail,
			DefaultCampaign   : this.m_userDefaultCampaign,
			DateFormat        : this.m_userDateFormat,
			IsFullSystemAccess: this.m_userIsFullSystemAccess,
			HasAvatar         : this.m_userHasAvatar
		};
	};

	Application.prototype.setLoggedUserHasAvatar = function(hasAvatar = false) {
		this.m_userHasAvatar = hasAvatar;
	};

	Application.prototype.getLoggedUserEmail = function() {
		return this.m_userEmail || "";
	};

	Application.prototype.getLoggedUserDefaultCampaign = function() {
		return this.m_userDefaultCampaign || "";
	};

	Application.prototype.getLoggedUserDefaultCampaignByChannel = function(channel) {
		if (!channel) {
			return;
		}
		const user_default_campaign_id = this.getLoggedUserDefaultCampaign();
		if (!user_default_campaign_id) {
			return;
		}
		const campaign = this.getCampaign(user_default_campaign_id);
		if (!campaign) {
			return;
		}
		const belongs_to_channel = campaign.has_accounts_of_proxy(channel);
		if (!belongs_to_channel) {
			return;
		}
		// ok
		return user_default_campaign_id;
	};

	Application.prototype.loadLoggedUserClientsLayoutConfig = function() {
		const has_permission = inConcert.permissions.to_action("settings", "change_personal_settings");
		if (!has_permission) {
			// Si no tiene permisos para cambiar el personal settings, El usuario NO puede cargar su Layout personalizado, si es que antes lo habia configurado.
			// En este caso se cargara el layout personalizado para el VCC o el por defecto.
			return;
		}
		if (!this.m_userClients) {
			return;
		}
		if (!this.m_userClients.Layouts) {
			return;
		}
		return inConcert.Utils.cloneObject(this.m_userClients.Layouts);
	};

	Application.prototype.getLoggedUserCompanyInformation = function() {
		return this.m_userCompanyInformation || {};
	};

	Application.prototype.getLoggedUserCustomInformation = function() {
		return this.m_userCustomInformation || {};
	};

	Application.prototype.getLoggedUserToken = function() {
		return this.m_authToken;
	};

	Application.prototype.getLoggedUserDateFormat = function() {
		return this.m_userDateFormat || "dd/mm/yy";
	};

	Application.prototype.isSessionRejoin = function() {
		return this.m_sessionRejoin ? true : false;
	};

	Application.prototype.getLoggedUserMessage = function() {
		return this.m_statusMessage;
	};

	//only update the status if it changed
	Application.prototype.setLoggedUserMessage = function( message, callback ) {
		const self = this;

		callback = callback || $.noop;

		if (this.m_statusMessage === message) {
			callback( true );
			return;
		}

		var has_permission_to_change_nick = inConcert.permissions.to_action( "settings", "changenick" );
		if ( !has_permission_to_change_nick) {
			callback( false, 403 ); // forbidden
			return;
		}

		inConcert.logger.action({action_name : "change_status_message", message : message, action_type:"agent"});
		var url = this.getBaseUrl() + "/users/set_status_message/";
		inConcert.invoke({
			url: url,
			type: 'POST',
			data: JSON.stringify({ message : message}),
			dataType: "json",
			contentType: "application/json",
			success: function(/*response*/) {
				self.m_statusMessage = message;
				callback( true );
			},
			error: function( response /*, error*/ ) {
				callback( false, response.status );
			}
		});
	};

	Application.prototype.getCurrentStatus = function() {
		return this.m_currentStatus;
	};

	Application.prototype.getSendKeyStatus = function() {
		return this.m_enterAsKeyToSend;
	};

	Application.prototype.getLoggedUserStatus = function()
	{
		return this.getStatusDescription(this.m_currentStatus);
	};

	Application.prototype.getStatusDescription = function(status)
	{
		var state;
		if (typeof status === "string") {
			state = this.m_availableStates[status];
			if (!state) {
				return "";
			}
		}
		else if (typeof status === "object") {
			state = status;
		}
		else {
			return ""; 
		}

		var translatedState = state.DescriptionLabel || state.Description;
		if (typeof state.ShortMessage !== "string") {
			return translatedState;
		}
		var translatedShortMessage = state.ShortMessageLabel || state.ShortMessage;
		return translatedState + 
			format(" ({0})", translatedShortMessage);
	};

	Application.prototype.getContactManager = function() {
		return this.m_contactManager;
	};

	Application.prototype.canExit = function() {
		if (this.loggedOut || this.m_hadErrors) {
			return true;
		}
		var interactions = this.m_interactionManager.getInteractions();
		for (var id in interactions) {
			if (interactions.hasOwnProperty(id)) {
				//los voice mails no son live pero en allegro al cerrar el baragent se cancelan
				if (interactions[id].belongsToMe() && !interactions[id].isAppointment() && 
				!interactions[id].isQueued() && !interactions[id].isAbandoned() && 
				(interactions[id].isLive() || interactions[id].isAllegro()) && !interactions[id].isInternal() && !interactions[id].isFinished()
				) {
					return false;
				}
			}
		}
		return true;
	};

	Application.prototype.getInteractionManager = function() {
		return this.m_interactionManager;
	};

	Application.prototype.getSNPostsManager = function() {
		return this.m_snPostsManager;
	};

	Application.prototype.getBatchesManager = function() {
		return this.m_batchesManager;
	};

	//loads interactions for logged user campaigns(queued to the campaign and to this user)
	Application.prototype.loadInteractions = function(interactions, callback)
	{
		const self = this;
		callback = callback || $.noop;

		var main = inConcert.app.getMainWindow();
		main.showGenericWaiting("<%=_T('Loading conversations, wait moment please...')%>");

		inConcert.Utils.foreach_timeout({
			rows : interactions,
			pause_between_loop_ms: 30,
			append_fn : function( interaction ) {
				self.m_interactionManager.create(interaction);
			},
			complete_fn : function() {
				callback();
				// El hide del Waiting se realiza despues de iniciar los paneles dentro de la funcion "ConversationsPanel.prototype.initialize"
				// Esto porque en el medio hace 20 cosas mas, y sino el waiting se esta visualizando y ocultando a cada rato mientras se realiza el startup del agente.
				// En la funcion nombrada, es donde termina de iniciar la app.
				//
				//main.hideGenericWaiting();
			}
		});
	};

	Application.prototype.windowRefresh = function(now) {
		var delta_ms = now - this.m_lastWindowRefresh;
		//TODO settings should be coming from somewhere with this data set
		if (delta_ms >= 1100) {
			this.m_lastWindowRefresh = new Date().getTime();
			inConcert.app.m_mainWindow.refresh();
		}
	};
	Application.prototype.getInitialState = function() {
		if (inConcert.app.has_allegro_integration()) {
			return "NOACD";
		}
		return this.m_initialState;
	};

	Application.prototype.getCountryCode = function() {
		return this.m_country;
	};

	Application.prototype.getAreaCode = function() {
		return this.m_area;
	};

	Application.prototype.getTimezone = function() {
		return this.m_timezone_id;
	};

	Application.prototype.getUserTimezoneObject = function() {
		return this.m_user_timezone_object;
	};

	Application.prototype.getUserdIdleThreshold = function() {
		return this.m_userIdleThreshold;
	};

	Application.prototype.getUserSoundNotifications = function() {
		return this.m_userSoundNotifications;
	};

	Application.prototype.getUserVisualNotifications = function() {
		return this.m_userVisualNotifications;
	};

	Application.prototype.getAccountStatusLabel = function(proxyType, status) {
		status = status || "";
		switch(status.toLowerCase())
		{
			case 'connection timeout':
				if (proxyType === "MAIL") {
					// For mail. The emailproxy throws and not generate object account so always is connection timeout but the error may a reauthorize the application
					return "<%=_T('Connection timeout. Reauthorize the application.')%>";
				}
				return "<%=_T('Connection timeout')%>";
			case 'fail':
				return "<%=_T('Failure')%>";
			case 'starting':
				return "<%=_T('Starting')%>";
			case 'disconnected':
				switch(proxyType) {
					case "WHATSAPP":
						return "<%=_T('Phone disconnected. Check your phone is turned on and connected to internet.')%>";
				}
				break;
			case 'logged_out':
				switch(proxyType) {
					case "WHATSAPP":
						return "<%=_T('Driver logged out. Needs to be authorized with a phone qr.')%>";
				}
				break;
			case 'without_app':
				switch(proxyType) {
					case "WHATSAPP":
						return "<%=_T('App not found. Needs to be authorized.')%>";
				}
				break;
			case 'without_integration':
				switch(proxyType) {
					case "WHATSAPP":
						return "<%=_T('App integration not found. Needs to be authorized.')%>";
				}
				break;
			case 'integration_inactive':
				switch(proxyType) {
					case "WHATSAPP":
						return "<%=_T('App integration inactive. Needs to be authorized.')%>";
				}
				break;
			case 'integration_error':
				switch(proxyType) {
					case "WHATSAPP":
						return "<%=_T('App integration error. Needs to be authorized.')%>";
				}
				break;
			default:
				if (proxyType === "MAIL" && status === "authentication_error") {
					// This is because some token has problems so it's necessary to reauthorize the app
					return "<%=_T('Authentication error ocurred. Reauthorize the application.')%>";
				}
		}
		return status || "<%=_T('Unrecognized state')%>";
	};

	Application.prototype.setAccountStatus = function(account, status) {
		if (account.proxyType === "CALL") {
			if (this.has_allegro_integration()) {
				if (this.has_allegro_baragent()) {
					status = inConcert.BaragentPolling.getAccountStatus(account);
				}
				else if (!this.userHasAddress("CALL")) {
					status = { status : "<%=_T('User has not configured phone')%>" };
				}
			}
			else if (status && status.status === "OK") {
				if (!this.userHasAddress("CALL")) {
					status = { status : "<%=_T('User has not configured phone')%>" };
				}
				else {
					let client = this.getInteractionManager().getChannelClient("CALL");
					if (client) {
						status = client.getAccountStatus();
					}
					else {
						status = { status : "<%=_T('User has not configured phone')%>" };
					}
				}
			}
		}
		account.status = status;
	};

	Application.prototype.updateCampaigns = function(deltaCampaigns, accountsStatus)
	{
		var i, campaign;
		let new_campaigns = [];
		let isUpdateOrAdd = false;
		for (var id in deltaCampaigns) 
		{
			if (deltaCampaigns.hasOwnProperty(id)) {
				var deltaCampaign = deltaCampaigns[id];
				deltaCampaign.Id = id;
				var campaignObj;

				if (deltaCampaign.deleted || deltaCampaign.role === "") {
					for (i = 0; i < this.m_campaigns.length; i++) {
						campaign = this.m_campaigns[i];
						if (campaign.Id === deltaCampaign.Id) {
							campaign.Dispose();
							this.m_campaigns.splice(i, 1);
							campaignObj = new inConcert.Campaign(deltaCampaign);
							inConcert.app.m_mainWindow.onCampaignRemoved(campaignObj);
							// publish campaign removal
							inConcert.app.messagebus.publish("/application/campaigns/removed", deltaCampaign);
							break;
						}
					}
				}
				else {
					isUpdateOrAdd = true;
					var isUpdate = false;
					for (i = 0; i < this.m_campaigns.length; i++) {
						campaign = this.m_campaigns[i];
						if (campaign.Id === deltaCampaign.Id) {
							for (var k in deltaCampaign) {
								if (deltaCampaign.hasOwnProperty(k)) {
									campaign[k] = deltaCampaign[k];
								}
							}
							campaign.refresh();
							inConcert.app.m_mainWindow.onCampaignUpdated(campaign);
							inConcert.app.messagebus.publish("/application/campaigns/updated", deltaCampaign);
							isUpdate = true;
						}
					}
					if (!isUpdate) {
						campaignObj = new inConcert.Campaign(deltaCampaign);
						this.m_campaigns.push(campaignObj);
						inConcert.app.m_mainWindow.onCampaignAdded(campaignObj);
						inConcert.app.messagebus.publish("/application/campaigns/added", deltaCampaign);
						new_campaigns.push(deltaCampaign);
					}
				}
			}
		}

		if (isUpdateOrAdd && this.getCurrentStatus() === "ACTIVE") {
			this.ValidateChannelCall(true);
		}

		if (new_campaigns.length > 0) {
			// Se notifican las campanas agregadas en modo bulk, es decir se publican todas las campanas nuevas.
			// Ej: lo utiliza la App Supervisor para agregar nuevas campanas.
			inConcert.app.messagebus.publish("/application/campaigns/bulk_add", new_campaigns);
		}

		if (accountsStatus) {
			for (var campaignId in accountsStatus) {
				if (accountsStatus.hasOwnProperty(campaignId)) {
					var statuses = accountsStatus[campaignId];
					for (var j = 0; j < this.m_campaigns.length; j++) {
						campaign = this.m_campaigns[j];
						if (campaign.Id === campaignId) {
							// FIX-ME : This is done to track an error, should be deleted in future
							// for further details : http://redmine.inconcert/issues/2633
							if ( !campaign.accounts ) {
								inConcert.logger.error({
									method:      "Application.updateCampaigns",
									campaign_id:   campaignId,
									message:     "Account Status delta was changed but comes with no accounts",
									accountsStatus : JSON.stringify(accountsStatus)
								});
							}
							else {
								for (i = 0; i < campaign.accounts.length; i++) {
									var account = campaign.accounts[i];
									let new_account_status = statuses[account.id];
									this.setAccountStatus(account, new_account_status);
								}
							}
						}
					}
				}
			}
			inConcert.app.m_mainWindow.updateCampaignAccounts(); // TODO: cambiar este mecanismo, y dejar publicado el de abajo

			inConcert.app.messagebus.publish( "/application/campaigns/update_accounts", accountsStatus );
		}
	};

	//returns a list of interaction ids and event indexes to mark as read
	Application.prototype.getReadMarks = function() {
		return [];
	};

	Application.prototype.onUnrecoverableError = function()
	{
		var app = this;

		// on error, user always will be offline
		app.m_mainWindow.setUserStatus('OFFLINE');
		app.loggedOut = true;
		app.m_waitingData = false;
		app.m_hadErrors = true;

		var msg_options = {
			message : "<%=_T('There was an network error which we can not recover, please reload the page.')%>",
			messageTitle : "<%=_T('Something went wrong')%>",
			popUpTitle : "<%=_T('Unable to log you in automatically')%>",
			messageType : "Error",
			hideCloseButton : true
		};

		app.m_mainWindow.showPopupMessage(msg_options);
		// close chat/call clients
		app.m_interactionManager.closeAllClients();
	};

	/**
	private: Helper for dataPoller. Handles the "custom messages" part of the notifications.
			@param response
	*/
	Application.prototype.handle_custom_messages = function (response)
	{
		var app = this;
		if (!response.custom){
			return;
		}

		response.custom.forEach(function(e)
		{
			var event_message = typeof e.message === "string"  ? JSON.parse( e.message ) : e.message;

			switch (e.type)
			{
			// SUPERVISOR poller notifications
				case "supervisor_broadcast":
					app.m_mainWindow.showPopupMessage({
						message : event_message.message,
						messageTitle : format("<%=_T('Supervisor {0} sent a new message!')%>", event_message.supervisor),
						popUpTitle : "<%=_T('Massive message from the supervisor')%>",
						messageType : "Info"
					});
				break;

				case "supervisor_logout":
					app.m_kickedOut = true;
					app.logoutWithMessage({
						callFrom : "Application-supervisor_logout",
						description : "<%=_T('You have been logged out by supervisor')%><%=_T(':')%>" + " '" + 
							event_message.supervisor + "'"
					});
				break;

			// AGENTSERVICE poller notifications
				case "agentservice_kickout":
					app.m_kickedOut = true;
					app.logoutWithMessage({
						callFrom : "Application-agentservice_logout",
						description : "<%=_T('You have been logged out by strike limit')%>"
					});
				break;

			// GENERIC (AGENT) poller notifications
				// TO-DO: This name doesnt say anything.
				// Change to something more informative, what is server state?
				case "server_changed_state":
					var status_id = event_message.state;
					var message;
					var new_state_description = app.getStatusDescription(status_id);
					if (event_message.type === "NOT_ALLOWED_TO_RECEIVE_CONVERSATIONS") {
						message = format("<%=_T('Current state has changed to: {0}.')%>", new_state_description);
					}
					else if (event_message.type === "SUPERVISOR_HAS_MODIFIED") {
						message = format("<%=_T('Current state has changed to: {0}, by supervisor: {1}.')%>", new_state_description, event_message.supervisor);
					}
					// Update user status.
					app.m_currentStatus = status_id;
					app.m_mainWindow.setUserStatus(status_id, message);
					app.__notifyUserStatusChange();
				break;

				case "campaign_status_change":
					var campaign_id = event_message.campaign;
					var enabled = event_message.status;

					app.m_mainWindow.showBannerMessage({
						messageType : "Info",
						message     : format("<%=_T('You have been set to active: {0} on campaign: {1}, by supervisor: {2}.')%>", enabled, campaign_id, event_message.supervisor),
						timeout     : 5000
					});

					app.m_mainWindow.setCampaignLiveStatus(campaign_id, enabled);
				break;

			// ADMINISTRATION poller notifications

				case "administrator_added_new_account":
					app.added_new_account( event_message );
				break;

				case "administrator_updated_account":
					app.updated_account( event_message );
				break;

				case "administrator_removed_account":
					app.removed_account( event_message );
				break;

				case "administrator_state_change":
					app.update_availables_states( event_message );
				break;

				case "administrator_disposition_change":
					app.update_dispositions( event_message );
				break;

				case "administrator_thresholds_change":
					app.update_thresholds( event_message );
				break;

				case "administrator_canned_messages_change":
					app.update_canned_messages( event_message );
				break;

				case "administrator_labels_change":
					app.update_labels( event_message );
				break;

				case "administrator_integration_change":
					app.update_integration_cci( event_message );
				break;

				case "administrator_interaction_notifications_change":
					app.update_interaction_notifications( event_message );
				break;

				case "administrator_campaign_permissions_change":
					app.update_campaign_permissions( event_message );
					app.update_logged_user_permissions();
				break;

				case "administrator_user_permissions_change":
					app.update_logged_user_permissions();
				break;

				case "administrator_role_permissions_change":
					app.update_logged_user_permissions();
				break;

				case "administrator_applications_actions_change":
					app.update_permissions_applications_actions(event_message);
				break;

				case "administrator_signatures_changed":
					app.update_signatures( event_message );
				break;

				case "administrator_attention_levels_change":
					app.update_attention_levels( event_message );
				break;

				case "administrator_user_information_changed": //Se actualiza name del usuario cuando un admin cambia otro user
					const userData = event_message.userData;
					const responsible_agent = event_message.responsible_agent;

					//Si un no es un mismo admin se cambio la info, actualizo la info basica.
					if (userData.Id === app.getLoggedUserId() && responsible_agent !== userData.Id ) {
						app.updateBasicSettings(userData);
						app.m_mainWindow.updateUserInformation();
					}
					else if(userData.Id !== app.getLoggedUserId()){
						app.update_cached_member(userData);
					}

					app.update_user_information(userData);
					
				break;

			// AGENT poller notifications
				case "agent_server_logout":
					app.on_server_logout( event_message );
				break;

			// IR poller facebook notifications
				case "ir_facebook_post_failed":
				case "ir_facebook_private_message_failed":
				case "ir_facebook_post_comment_failed":
					app.ir_facebook_failures( e.type, event_message );
					break;

			// IR poller instagram notifications
				case "ir_instagram_post_failed":
				case "ir_instagram_private_message_failed":
				case "ir_instagram_post_comment_failed":
					app.ir_instagram_failures( e.type, event_message );
					break;

			// IR poller youtube notifications
				case "ir_youtube_post_failed":
				case "ir_youtube_private_message_failed":
				case "ir_youtube_post_comment_failed":
					app.ir_youtube_failures( e.type, event_message );
					break;

			// IR poller appstore notifications
				case "ir_appstore_post_failed":
				case "ir_appstore_private_message_failed":
				case "ir_appstore_post_comment_failed":
					app.ir_appstore_failures( e.type, event_message );
					break;

			// IR poller googleplay notifications
				case "ir_googleplay_post_failed":
				case "ir_googleplay_private_message_failed":
				case "ir_googleplay_post_comment_failed":
					app.ir_googleplay_failures( e.type, event_message );
					break;

			// Poller callback suscriptions
				case "async_message_result":
				case "async_new_outgoing_batch_result":
					app.__onPollerCallback(e.type, event_message);
				break;

			// Screen Streaming
				case "start_screen_capture":
					inConcert.app.StartScreenCapture(event_message.subdomain);
				break;
				case "stop_screen_capture":
					inConcert.app.StopScreenCapture(event_message.supervisor);
				break;

				case "start_speedtest_capture":
					inConcert.app.ExecuteSpeedTest(event_message);
				break;

				case "start_browser_health_capture":
					inConcert.app.EnableLoggingCapture();
					inConcert.app.EnableWebRtcCapture();
					inConcert.app.EnableBrowserHealthCapture();
				break;

				case "stop_browser_health_capture":
					inConcert.app.DisableBrowserHealthCapture();
					inConcert.app.DisableWebRtcCapture();
					inConcert.app.DisableLoggingCapture();
				break;
				// License Helpdesk
				case "consume_helpdesk_licenses":
					const status = event_message.status;
					if (status === "ENABLE") {
						app.enabled_helpdesk_integration();
					}
				break;

			// OUTBOUNDENGINE poller notifications
				case "outboundengine_start_progressive":
					app.startProgressiveTry(event_message);
				break;
				case "outboundengine_progressive_success":
				case "outboundengine_progressive_fail":
					app.stopProgressiveTry();
				break;

			// UNDEFINED poller notifications
				default:
					inConcert.logger.warn({
						action_name : "Application.prototype.onDataReceived",
						action_type : "message poller",
						message     : format("Unimplemented '{0}' poller message type.", e.type )
					});
			}
		});
	};

	/**
	private:
	Data has been received from the agent_poller. This data includes notifications, custom messages, changes in 
	campaigns or counters, etc.
	@param response
	*/
	Application.prototype.onDataReceived = function(response)
	{
		var app = inConcert.app;
		inConcert.logger.debug({
					action_name : "Application.prototype.onDataReceived",
					message     : format("MainLoop - begin onDataReceived. main_loop_rid '{0}'",app.main_loop_rid)
		});
		if (app.m_hadErrors) {
			app.m_hadErrors = false;
			app.m_mainWindow.hideBannerMessage();
			app.m_mainWindow.setUserStatus(app.getCurrentStatus());
		}
		app.m_waitingData = false;

		// publish application status
		inConcert.app.messagebus.publish("/application/connected", true);

//		try {
			if (response.refresh_campaigns && response.campaign_counters){
				//these contains the summary of total interactions by state on each campaign
				app.updateCampaignCounters(response.campaign_counters);
				app.updateCampaigns(response.refresh_campaigns.delta, response.refresh_campaigns.accounts);
			}
			else {
				inConcert.logger.warn({
					method : "Application.prototype.onDataReceived",
					message: "Invalid response from Agent Poller: " + JSON.stringify(response)
				});
			}

			//process the changed interactions
			var changes = response.refresh_inbox;
			var updates = app.m_interactionManager.updateInteractions(changes);
			if (updates) {
				app.m_mainWindow.updateInteractions(updates);
			}

			app.handle_custom_messages(response);
			
			app.checkAvailableAddresses();

			//TODO:
			//process changed permissions (new campaigns, role changes)
			//contact updates? status? online?
		// }
		// catch(e) {
			// RearmPollerRequest();
			// throw e;
		// }

		app.lastMessageSeen = response.last_message_id;

		app.__rearmPollerRequest();
	};

	Application.prototype.cleanAllInteractions = function() {
		var updates = inConcert.app.m_interactionManager.deleteAllInteractions();
		if (updates) {
			inConcert.app.m_mainWindow.updateInteractions(updates);
		}
	};

	Application.prototype.onDataError = function( err )
	{
		var app = inConcert.app;

		if ( app.loggedOut === true ) {
			return;
		}

		app.m_waitingData = false;
		app.m_hadErrors = true;

		var retryPoller = function( message, timeout_secs )
		{
			clearTimeout(app.pollerTimeout);
			var displayConnecting = function()
			{
				if (timeout_secs > 0) {
					app.m_mainWindow.showBannerMessage({
						messageType : "Alert",
						message : format(message, timeout_secs),
						showOverlay: true,
						hideClose : true
					});
					timeout_secs -= 1;
					setTimeout(displayConnecting, 1000);
				} else {
					app.m_mainWindow.showBannerMessage({ 
						messageType : "Process",
						message : "<%=_T('Connecting...')%>",
						showOverlay: true,
						hideClose : true
					});
					app.retryPoller();
				}
			};
			displayConnecting();
		};

		// update status on main window.
		app.m_mainWindow.setUserStatus("OFFLINE");

		// publish error
		inConcert.app.messagebus.publish("/application/connected", false);

		switch ( err.status ) {
			case 0:
				//retryPoller("<%=_T('Server unreachable, reconnecting in {0} seconds...')%>", 7);
				app.retryPoller();
				break;
			case 403:
				// hide banner if any
				app.m_mainWindow.hideBannerMessage();

				// close chat/call clients
				app.m_interactionManager.closeAllClients();

				// set logged out flag, this is going to stop requests to the server
				app.loggedOut = true;

				// this is done to handle supervisor logout.
				if (app.m_kickedOut){
					return;
				}
				
				app.logoutWithMessage({
					callFrom : "Application-poller",
					title : "<%=_T('Server unreachable')%>",
					description : "<%=_T('Press OK to retry login.')%>"
				});

				break;
			case 503:
				retryPoller("<%=_T('Server overload, reconnecting in {0} seconds...')%>", 7);
				break;
			case 504:
				retryPoller("<%=_T('Internet gateway timeout, reconnecting in {0} seconds...')%>", 7);
				break;
			case 502:
				retryPoller("<%=_T('Bad internet gateway, reconnecting in {0} seconds...')%>", 5);
				break;
			default:
				app.onUnrecoverableError();
		}
	};

	/*
		Data polling, it blocks until there's data or every 10 seconds
	*/
	Application.prototype.dataPoller = function(callback, data = {})
	{
		callback = callback || $.noop;

		if (this.loggedOut) {
			return callback({ result : "LOGGED_OUT", reason : "user logged out" });
		}
		if ( this.m_waitingData ) {
			return callback({ result : "BUSSY", reason : "another polling is running" });
		}

		this.m_lastSentRequest = new Date().getTime();
		this.m_waitingData = true;
		let url = this.getPollerUrl() + "/main_loop/" + (typeof(this.lastMessageSeen) !== "undefined" ? this.lastMessageSeen : "");
		this.main_loop_rid = uuid.v4();
		inConcert.logger.debug({
					action_name : "Application.prototype.dataPoller",
					message     : format("MainLoop - Begin request. main_loop_rid: '{0}'",this.main_loop_rid)
		});
		this.m_pollerRequest = inConcert.invoke({
			url: url,
			type: 'POST',
			data: JSON.stringify(data),
			dataType: "json",
			contentType: "application/json",
			headers: {'cache': 'false', 'main_loop_rid': this.main_loop_rid},
			success: function(response) {
				inConcert.app.onDataReceived(response);
				callback({ result : "OK", data : response });
			},
			error: function( response, text_status/*, error_thrown*/ ) {
				let now = new Date().getTime();
				let vLastDataReceived_ms = 0;
				let vLastSendRequest_ms = 0;
				if (inConcert.app.m_lastDataReceived) {
					vLastDataReceived_ms = now - inConcert.app.m_lastDataReceived;
				}
				if (inConcert.app.m_lastSentRequest) {
					vLastSendRequest_ms = now - inConcert.app.m_lastSentRequest;
				}
				let err = response.jqXHR;
				inConcert.logger.warn({
					action_name : "Application.prototype.dataPoller",
					message     : format("MainLoop - ERROR!! Request onError. main_loop_rid: '{0}' Code: '{1}'. Delta Data Received: '{2}'. Delta Sent Request: '{3}' ", 
						inConcert.app.main_loop_rid, err.status, vLastDataReceived_ms, vLastSendRequest_ms )
				});
				inConcert.app.onDataError( err );
				callback({ result : "FAIL", code : err.status, reason : text_status });
			},
			timeout: 30 * 1000
		});
	};

	/*
		This function is called for the app to update and fetch interactions on the non-socket version
		warning in this case THIS is not the application, kinda static method
	*/
	Application.prototype.onTimer = function()
	{
		if (inConcert.app.isAutomaticPollerDisabled()) {
			return;
		}
		if (inConcert.app.loggedOut) {
			//TODO: reintentar el login con un back off
			clearInterval(inConcert.app.m_applicationTimer);
			return;
		}
		var now = new Date().getTime();
		//update windows (displayed timers and whatever)
		inConcert.app.windowRefresh(now);
		if (inConcert.app.m_lastDataReceived) {
			var delta_ms = now - inConcert.app.m_lastDataReceived;
			var delta_last_send_ms = 0;
			var vMainLoopRid = '';
			if (inConcert.app.main_loop_rid){
				vMainLoopRid = inConcert.app.main_loop_rid;
			}
			if (inConcert.app.m_lastSentRequest) {
				delta_last_send_ms = now - inConcert.app.m_lastSentRequest;
			}
			if (delta_last_send_ms > 25000 && inConcert.app.m_waitingData) {
				inConcert.logger.debug({
						action_name : "Application.prototype.onTimer",
						message     : format("MainLoop - ATTENTION!! Out of time. main_loop_rip '{0}'. Delta Last received data: '{1}'. Delta Last Send Request: '{2}' .", vMainLoopRid, delta_ms, delta_last_send_ms)
				});
			}
			if (delta_ms > 65000 && inConcert.app.m_waitingData) {
				inConcert.logger.error({
						action_name : "Application.prototype.onTimer",
						message     : format("MainLoop - ERROR!! Update dataPoller over 65 secs. main_loop_rip '{0}'. Delta Data Received: '{1}'. Delta Sent Request: '{2}' ", vMainLoopRid, delta_ms, delta_last_send_ms )
					});
				inConcert.app.m_hadErrors = true;
				inConcert.app.m_mainWindow.showBannerMessage({ messageType : "Process", message : "<%=_T('Please check you are properly connected to the internet, retrying connection...')%>"});
				inConcert.app.m_lastDataReceived = undefined;
				inConcert.app.retryPoller();
			}
		}
	};

	//reload available states
	Application.prototype.refreshAvailableStates = function(callback)
	{
		const self = this;
		callback = callback || $.noop;
		var url = self.getBaseUrl() + "/states/find_all/";
		var request = inConcert.invoke({
			url: url,
			dataType: "json",
			contentType: "application/json",
			success: function(response) {
				self.m_availableStates = response;
				callback(true, self.m_availableStates);
			},
			error: function() {
				callback(false);
			}
		});
		return request;
	};

	Application.prototype.setAppsActions = function( apps_actions ) {
		this.m_apps_actions = apps_actions;
	};

	Application.prototype.getAppsActions = function() {
		return this.m_apps_actions;
	};

	Application.prototype.isBusyState = function(stateId) {
		var state = this.m_availableStates[stateId];
		if (state) {
			if (state.isBreak === "1") {
				if (state.overrideable && state.overrideable === "1") {
					return true;
				}
			}
		}
		return false;
	};

	Application.prototype.isAwayState = function(stateId) {
		var state = this.m_availableStates[stateId];
		if (state) {
			if (state.isBreak === "1") {
				if (state.overrideable !== "1") {
					return true;
				}
			}
		}
		return false;
	};

	Application.prototype.isCustomState = function(stateId) {
		var state = this.m_availableStates[stateId];
		if (state && this.isBusyState(stateId) && state.UserState === "1")  {
			return true;
		}
		return false;
	};

	Application.prototype.canTakeWithCurrentState = function() {
		var state = this.m_availableStates[this.m_currentStatus];
		if (state) {
			return state.isBreak !== "1";
		}
		return false;
	};

	/*
		Returns the available states for the user to switch to
		coming out sorted in the most logical way, available states first, and case sensitive for the rest
	*/
	Application.prototype.getAvailableStates = function()
	{
		var availableStates = [];
		var breakStates = [];
		for (var stateId in this.m_availableStates) {
			if (this.m_availableStates.hasOwnProperty(stateId)) {
				var state = this.m_availableStates[stateId];
				state.Id = stateId;
				if (state.isBreak !== "1") {
					availableStates.push(state);
				}
				else {
					breakStates.push(state);
				}
			}
		}
		//Sort states based on the lexicographical order of its descriptions
		availableStates = inConcert.Utils.sortArray(availableStates, "Description", "asc", _T);
		breakStates = inConcert.Utils.sortArray(breakStates, "Description", "asc", _T);

		return availableStates.concat(breakStates);
	};

	Application.prototype.getCustomStates = function()
	{
		var customStates = [];

		for (var stateId in this.m_availableStates) {
			if (this.m_availableStates.hasOwnProperty(stateId)) {
				var state = this.m_availableStates[stateId];
				state.Id = stateId;
				if (state.UserState === "1") {
					customStates.push(state);
				}
			}
		}

		return inConcert.Utils.sortArray(customStates, "Description", "asc", _T);
	};

	Application.prototype.getStateObject = function( stateId )
	{
		if (typeof stateId !== "string" || typeof this.m_availableStates !== "object") {
			return;
		}
		var state = this.m_availableStates[ stateId ];
		if ( state ) {
			return state;
		}
		return this.m_availableStates[ stateId + "@" + this.getDomain() ];
	};

	Application.prototype.on_server_logout = function( err )
	{
		this.onDataError({
			status: err.status
		});
	};

	Application.prototype.added_new_account = function( message )
	{
		inConcert.app.messagebus.publish( "/accounts/" + message.campaignId + "/added", message.account );
	};

	Application.prototype.updated_account = function( message )
	{
		inConcert.app.messagebus.publish( "/accounts/" + message.campaignId + "/updated", message.account );
	};

	Application.prototype.removed_account = function( message )
	{
		inConcert.app.messagebus.publish( "/accounts/" + message.campaignId + "/removed", message.accountId );
	};

	Application.prototype.update_dispositions = function( message )
	{
		inConcert.app.messagebus.publish( "/dispositions/" + message.campaignId + "/updated", message.dispositions );
	};

	Application.prototype.update_thresholds = function( message )
	{
		inConcert.app.messagebus.publish( "/thresholds/" + message.campaignId + "/updated", message.thresholds );
	};

	Application.prototype.update_canned_messages = function( message )
	{
		inConcert.app.messagebus.publish( "/canned_messages/" + message.campaignId + "/updated", message.canned_messages );
	};

	Application.prototype.update_labels = function( message )
	{
		inConcert.app.messagebus.publish( "/update_labels/" + message.campaignId + "/updated", message.labels );
	};

	Application.prototype.update_integration_cci = function( message )
	{
		inConcert.app.messagebus.publish( "/integration_cci/" + message.campaignId + "/updated", message.integration );
	};

	Application.prototype.update_interaction_notifications = function( message )
	{
		inConcert.app.messagebus.publish( "/interaction_notifications/" + message.campaignId + "/updated", message.integration );
	};

	Application.prototype.update_attention_levels = function( message )
	{
		inConcert.app.messagebus.publish( "/update_attention_levels/" + message.campaignId + "/updated", message.attention_levels );
	};

	Application.prototype.update_user_information = function(userData) {
		inConcert.app.messagebus.publish( "/update_user_information/" + this.getLoggedUserFullId() + "/updated", userData );
	};

	Application.prototype.update_permissions_applications_actions = function(apps_actions)
	{
		// 1 - actualizo las acciones en la propia clase
		this.setAppsActions(apps_actions);
		// 2 - actualizo las acciones en la clase de permisos
		const force = true;
		inConcert.permissions.load_apps_actions(null, force);
		// 3 - actualizo los permisos del usuario.
		this.update_logged_user_permissions();
	};

	Application.prototype.update_campaign_permissions = function( message )
	{
		inConcert.app.messagebus.publish( "/permissions/" + message.campaignId + "/updated", message.permissions );
	};

	Application.prototype.update_logged_user_permissions = function()
	{
		inConcert.permissions.load_logged_user_permissions( function() { /* TODO: #4639 Hacer algo ? */});
	};

	Application.prototype.ir_facebook_failures = function( type, message )
	{
		inConcert.app.messagebus.publish( "/facebook/" + message.interactionId + "/failure", message, type );
	};

	Application.prototype.ir_instagram_failures = function( type, message )
	{
		inConcert.app.messagebus.publish( "/instagram/" + message.interactionId + "/failure", message, type );
	};

	Application.prototype.ir_youtube_failures = function( type, message )
	{
		inConcert.app.messagebus.publish( "/youtube/" + message.interactionId + "/failure", message, type );
	};

	Application.prototype.ir_appstore_failures = function( type, message )
	{
		inConcert.app.messagebus.publish( "/appstore/" + message.interactionId + "/failure", message, type );
	};

	Application.prototype.ir_googleplay_failures = function( type, message )
	{
		inConcert.app.messagebus.publish( "/googleplay/" + message.interactionId + "/failure", message, type );
	};

	Application.prototype.update_signatures = function( update_info )
	{
		const self = this;
		var log = function(method, value, message)
		{
			method = method || "info";
			inConcert.logger[method]({
				action_name : "Application.prototype.update_signatures",
				action_type : "update mail signature by poller",
				value       : value || "",
				message     : message
			});
		};

		if (typeof update_info !== "object" || !this.cached_mail_signatures ) {
			return;
		}

		// add new signature / update current signature
		var add = function(signature_id, new_signature)
		{
			var can_add = true;
			$.each(self.cached_mail_signatures || [], function( i, current_signature ) {
				if ( current_signature.Id === signature_id ) {
					self.cached_mail_signatures[i] = new_signature;
					can_add = false;
					return false; // break
				}
			});
			if ( can_add ) {
				self.cached_mail_signatures.push(new_signature);
			}
		};
		// remove signature
		var erase = function(signature_id)
		{
			var index_to_delete = -1;
			$.each(self.cached_mail_signatures || [], function( i, current_signature ) {
				if ( current_signature.Id === signature_id ) {
					index_to_delete = i;
					return false; // break
				}
			});
			if (index_to_delete > -1) {
				self.cached_mail_signatures.splice(index_to_delete, 1);
			}
		};

		var action = update_info.action;
		$.each(update_info.signatures, function(signature_id, signature)
		{
			if (typeof action !== "string") {
				log("warn", null, "Updating e-mail signatures by poller!! Incomplete poller information. Action is undefined");
				return true; // continue
			}
			switch(action.toUpperCase())
			{
				case "DELETE":
					erase(signature_id);
				break;

				case "NEW_UPDATE":
					add(signature_id, signature);
				break;

				default:
					log("warn", action, "Updating e-mail signatures by poller!! Unrecognized action");
			}
		});

		inConcert.Utils.sortArray(this.cached_mail_signatures, "Name", "asc");
	};

	Application.prototype.update_cached_member = function (member) {
		for (let index = 0; index < (this.cached_members || []).length; index++) {
			if (this.cached_members[index].Id === member.Id) {
				this.cached_members[index].Name = member.Name;
				this.cached_members[index].LastName = member.LastName;
				this.cached_members[index].FirstSurname = member.FirstSurname;
				this.cached_members[index].LastSurname = member.LastSurname;
				break;
			}
		}
	};

	Application.prototype.update_availables_states = function(states)
	{
		const self = this;

		var log = function(method, value, message)
		{
			method = method || "info";
			inConcert.logger[method]({
				action_name : "Application.prototype.update_availables_states",
				action_type : "update states by poller",
				value       : value || "",
				message     : message
			});
		};

		log("info", JSON.stringify(states || {}), "Updating availables state by poller!!");
		if (typeof states !== "object") {
			return;
		}

		// add new state
		var add = function(state_id, state)
		{
			if (typeof state_id !== "string" || typeof state !== "object") {
				log("warn", null, "Updating availables state by poller!! Incomplete information adding new state");
				return;
			}
			if (self.m_availableStates.hasOwnProperty(state_id)) {
				log("warn", state_id, "Updating availables state by poller!! State already exist");
				return;
			}
			self.m_availableStates[state_id] = state;
		};

		// remove state
		var erase = function(state_id)
		{
			if (typeof state_id !== "string") {
				log("warn", null, "Updating availables state by poller!! Incomplete information deleting state");
				return;
			}
			if (!self.m_availableStates.hasOwnProperty(state_id)) {
				log("warn", state_id, "Updating availables state by poller!! State to remove does not exist");
				return;
			}
			var current_state = self.getCurrentStatus();
			if (current_state.toUpperCase() === state_id.toUpperCase()) {
			// si elimina el estado actual, muestra un mensaje y cambia de estado al usuario.
				var current_state_description = self.getStatusDescription(state_id, "TU_CLASE_STATE");
				self.m_mainWindow.showBannerMessage({
					messageType: "Info",
					message    : format("<%=_T('Your current state {0} has been deleted')%>", current_state_description),
					timeout    : 5000
				});
				var system_state = "BREAK"; // lo dejo en el estado del sistema BREAK
				self.changeUserState(system_state, function(response)
				{
					if (!response.success) {
						log("warn", null, "Updating availables states by poller!! Unable to change current state.");
						return;
					}
					self.m_mainWindow.setUserStatus(system_state);
				});
			}
			delete self.m_availableStates[state_id];
		};

		// update state
		var update = function(state_id, state)
		{
			if (typeof state_id !== "string" || typeof state !== "object") {
				log("warn", null, "Updating availables states by poller!! Incomplete information updating state");
				return;
			}
			if (!self.m_availableStates.hasOwnProperty(state_id)) {
				log("warn", state_id, "Updating availables states by poller!! State to update does not exist");
				return;
			}
			self.m_availableStates[state_id] = state;
			var current_state = self.getCurrentStatus();
			if (current_state.toUpperCase() !== state_id.toUpperCase()) {
				return;
			}
			// si esta actualizando al estado actual muestra un mensaje y reasigna el estado actual.
			self.m_mainWindow.showBannerMessage({
				messageType : "Info",
				message     : "<%=_T('Your current state has been updated.')%>",
				timeout     : 3000
			});
			self.m_mainWindow.setUserStatus(state_id);
		};

		// close the state menu
		if (self.m_mainWindow.userStatePopup) {
			self.m_mainWindow.userStatePopup.finalize();
		}

		// loop on  status list to update
		$.each(states, function(state_id, update_info)
		{
			var action = update_info.action;
			if (typeof action !== "string") {
				log("warn", null, "Updating availables states by poller!! Incomplete poller information");
				return true; // continue
			}
			switch(action.toUpperCase())
			{
				case "DELETE":
					erase(state_id);
				break;

				case "NEW":
					add(state_id, update_info.state);
				break;

				case "UPDATE":
					update(state_id, update_info.state);
				break;

				default:
					log("warn", action, "Updating availables states by poller!! Unrecognized action");
			}
		});
	};

	Application.prototype.getTimeOffset = function() {
		return this.m_timeDifference || 0;
	};

	Application.prototype.advise = function(callbackFn) {
		$("body").addClass("loading");
		var url = this.getPollerUrl() + "/advise_agent";
		inConcert.invoke({
			url: url,
			type : "POST",
			dataType: "json",
			contentType: "application/json",
			data: JSON.stringify({}),
			success: function() {
				callbackFn(true);
			},
			error: function() {
				callbackFn(false);
			}
		});
	};

	/*
	*/
	Application.prototype.popupShow = function({title, message, callback = $.noop}) {
		$.inconcertPopup.Show(
			title,
			message,
			$.inconcertPopup.ButtonSets.BS_OK,
			callback,
			undefined,
			{
				popup_id       : "agent_inconcertPopup",
				container_class: "confirmationPopUp",
				overlay_class  : "confirmationOverlay"
			}
		);
	};

	/*
	*/
	Application.prototype.validateAllegroWarningsCampaigns = function(warnings = []) {
		const resp = {
			campaignsWithServerDownErrors: []
		};
		if (warnings.length === 0)
			return resp;
		for (let i = 0; i < warnings.length; i++) {
			const warning = warnings[i];
			const {error} = warning;
			if (!error || error.finish === "OK")
				continue;
			if (!error || !error.errType === "ALLEGRO_INTEGRATION_ERROR")
				continue;
			// solo muestra el mensaje del el tipo de error ha sido porque el Servidor de Allegro o el Wehandler se ha caido.
			const {rawErr} = error;
			if (!rawErr
				|| !rawErr.sub_type
				|| (rawErr.sub_type.result || "").toUpperCase() !== "ALLEGRO SERVER OR WEBHANDLER IS DOWN"
			) {
				continue;
			}
			// si tiene al menos una campana con este error
			resp.campaignsWithServerDownErrors.push(warning);
		}
		return resp;
	};

	/*
	*/
	Application.prototype.validateLoggedUserInfo = function() {
		const title  = "<%=_T('Update Problems.')%>";
		if (this.HasExternalUsersSync() && this.m_vccWarnings) {
			let message = format("<%=_T('There is a user synchronization problem with {0}. Please increment Max Campaign/Users Vcc limit to correct it.')%>", this.m_vccWarnings);
			// muestro el popup
			return this.popupShow({title, message});
		}
		if (!this.has_allegro_integration())
			return;
		// Error de sincronizacion de Usuarios con Allegro
		if (!this.m_campaignsWarnings)
			return;
		const {warnings} = this.m_campaignsWarnings;
		// 
		const resp = this.validateAllegroWarningsCampaigns(warnings);
		if (resp.campaignsWithServerDownErrors.length === 0)
			return;
		// valida que tenga permisos para la App "administration"
		const has_permission_to_adm_app = inConcert.permissions.to_application("administration");
		if (!has_permission_to_adm_app)
			return;
		const campaignsIds = resp.campaignsWithServerDownErrors.map(item => item.campaignId).join(",");
		// si tiene permisos para la App "Administration"
		let message = format("<%=_T('There is a User synchronization problem with Allegro in {0} campaign/s: {1}.')%>", resp.campaignsWithServerDownErrors.length, campaignsIds);
		if (this.m_vcc_info.AllegroIntegration.disabledSynchronizeAllegroMembersInLogin === true) {
			// si NO esta habilitada la sincronizacion automatica de usuarios.
			message = message + "<br>"
				+ "<%=_T('To solve the problem you must:')%>"
				+ "<br><br>"
				+ "1- " + "<%=_T('Validate the services in Allegro.')%>"
				+ "<br>"
				+ "2- " + "<%=_T('Then in the Adminstration application you must edit the campaign, go to the section of Users, and save the list.')%>";
		}
		else {
			// si esta habilitada la sincronizacion automatica de usuarios.
			if (this.m_campaignsWarnings.statusAutoSyncAllegroMembers === "UNABLE TO CONNECT WITH ALLEGRO") {
				// si NO pudo comenzar la sincronizacion ya que no se pudo realizar comunicacion con el servidor de Allegro.
				message = message + "<br><br>"
					+ "<%=_T('Unable to automatic users synchronization as communication with the Allegro server could not be made.')%>";
			}
			else {
				// si pudo comenzar con la sincronizaciion automatica.
				message = message + "<br>"
					+ "<%=_T('To resolve the problem, there is currently a background autorun against the Allegro server.')%>"
					+ "<br><br>"
					+ "<%=_T('You can validate if the synchronization has finished in the Campaigns view of Administration app.')%>";
			}
		}
		// muestro el popup
		return this.popupShow({title, message});
	};

	/*
		Realiza requests Iniciales luego de que el Usuario ha iniciado sesion.
		Son request desacoplados de la carga inicial, que no es necesario que se obtengan en la carga inicial
		Con esta funcion nos evitamos recargar las acciones servers side en el inicio de sesion, y que los datos retornados en la carga inicial no sean muy grandes.
	*/
	Application.prototype.initialRequests = function()
	{
		// 1 - Obtengo datos iniciales de la App OE Nativa (sin integracion con Allegro)
		const outbound_engine_api = this.getOutboundEngineAPI();
		if (!outbound_engine_api) {
			return;
		}
		if (this.has_allegro_integration()) {
			return;
		}
		const allowed_app_oe = inConcert.permissions.to_application("outbound_engine");
		if (!allowed_app_oe) {
			return;
		}
		// Estos 2 requests se utilizan en el Frm de Procesos de la App OE.
		// es mejor realizar la carga en este momento y cachear los mismos ya que son datos estaticos (del sistema) que nunca cambiaran.
		outbound_engine_api.systemRulesLoadChannels(function(success) {
			if (!success) {
				console.error("Application.prototype.initialRequests - Unable to load OE initial data [systemRulesLoadChannels].");
				return;
			}
			outbound_engine_api.systemRulesSystemActions(function(success) {
				if (!success) {
					console.error("Application.prototype.initialRequests - Unable to load OE initial data [systemRulesSystemActions].");
					return;
				}
			});
		});
	};

	/*
		Assuming a valid login, initializes everything, from settings and interaction clients to pending interactions..
	*/
	Application.prototype.bootstrap = function(callback = $.noop)
	{
		const app = this;

		app.loggedOut = false;
		inConcert.logger.debug("Bootstraping application");

		//this is where the credentials are stored or removed according to user selection
		if (app.lastCredentials) {
			inConcert.Storage.put("credentials", app.lastCredentials);
		}
		else {
			inConcert.Storage.remove("credentials");
		}

		//after advising myself to the data poller for it to acummulate notifications, get the initial data
		app.advise(function(advised) {
			if (!advised) {
				app.m_mainWindow.close();
				callback(false);
				return;
			}
			//armo la fecha como UTC para enviar al server
			function getUTCDateStruct() {
				var now = new Date();
				return {
					year: now.getUTCFullYear(),
					month: now.getUTCMonth() + 1,
					day: now.getUTCDate(),
					hours: now.getUTCHours(),
					minutes: now.getUTCMinutes(),
					seconds: now.getUTCSeconds()
				};
			}
			var initialRequest = {
				utcDate: getUTCDateStruct()
			};

			var url = app.getBaseUrl() + "/initialize/";
			inConcert.invoke({
				url: url,
				type: 'POST',
				data: JSON.stringify(initialRequest),
				dataType: "json",
				contentType: "application/json",
				timeout: 50 * 1000, // Raise the timeout to 50 seconds (to deal with lots of fat interactions sitting on the inbox)
				success: function(response) {
					if (response.result === true) {
						$("body").removeClass("loading");
						let relogin = app.m_user !== null;
						// vcc info
						app.m_vcc_info = response.vccInfo;
						app.m_domain = response.domain;
						app.m_licensedAccountTypes = response.licensedAccountTypes;
						// luis info
						app.m_luis_info = response.luisInfo;
						// rasa info
						app.m_rasa_info = response.rasaInfo;
						// widgets integrations
						app.make_widgets_integrations_info(response.widgetsIntegrationsInfo);
						// user
						app.m_user = response.user;
						app.m_userHasAvatar =  response.userHasAvatar;
						app.m_userFirstName = response.userName;
						app.m_userLastName = response.userLastName;
						app.m_userFirstSurname = response.userFirstSurname;
						app.m_userLastSurname = response.userLastSurname;
						app.m_userLanguage = response.userLanguage;
						app.m_userEmail =  response.userEmail;
						app.m_userDefaultCampaign = response.userDefaultCampaign;
						app.m_userCompanyInformation = response.userCompanyInfo;
						app.m_userCustomInformation = response.userCustomInfo;
						app.m_userDateFormat = response.userDateFormat;
						app.m_userClients = response.userClients;
						app.m_userIdleThreshold       = response.userIdleThreshold;
						app.m_userSoundNotifications  = response.userSoundNotifications;
						app.m_userVisualNotifications = response.userVisualNotifications;
						app.m_authToken = response.userAccessToken;
						app.m_sessionRejoin = response.sessionRejoin;
						app.m_initialState = response.initialState || "";
						app.m_timeDifference = response.timeDifference;
						app.m_currentStatus = response.status;
						app.m_enterAsKeyToSend = response.enterAsKeyToSend;
						// app.m_group_permissions = response.group_permissions;
						app.m_userRole = response.userRole;
						app.m_availableStates = response.availableStates;
						app.m_country = response.country;
						app.m_area = response.area;
						app.m_timezone_id = response.timezone;
						app.m_user_timezone_object = response.userTimezoneInfo;
						app.m_userIsFullSystemAccess = response.userIsFullSystemAccess;
						app.m_statusMessage = response.statusMessage;
						app.loadSessionData(response.sessionData || {});
						app.loadSettings(response.settings);
						app.loadCampaigns(response.campaigns);
						app.setServerInfo(response.serverInfo);
						app.setIRInfo(response.IRInfo);
						app.m_passwordExpirationDays = response.passwordExpirationDays;
						app.m_internalCampaignId = response.internalCampaignId;
						app.m_campaignsWarnings = response.campaignsWarnings;
						app.m_vccWarnings = response.vccWarnings;
						if (relogin) {
							app.m_interactionManager.closeAllClients();
						}
						/* INCO-64 - Carga objeto AgentMonitor */
						app.m_agentCaptureStatisticsStatus = response.agentCaptureStatisticsStatus;
						if(app.getAgentMonitorServiceEnabled()){
							app.loadAgentMonitorService(function (success, e) {
								if (success) {
									/* INCO-64 - Enviar geolocalizacion del agente */
									app.SendAgentGeolocation(undefined /*subdomain*/, function(success, data) {
										if (!success) {
											console.error(`SendAgentGeolocation error: ${data}`);
										}
									});
									/* Valida status de captura */
									if (app.m_agentCaptureStatisticsStatus.browser_health) {
										app.EnableBrowserHealthCapture();
										app.EnableWebRtcCapture();
										app.EnableLoggingCapture();
									}
								}
							});
						}
						app.m_interactionManager = new inConcert.InteractionManager(app.getBaseUrl(), app.m_user, app.m_domain, app.m_settings.InteractionClients);
						if (inConcert.SNPostsManager) {
							app.m_snPostsManager = new inConcert.SNPostsManager();
						}
						if (inConcert.batchManagementManager) {
							app.m_batchesManager = new inConcert.batchManagementManager();
						}
						app.loadInteractions(response.interactions, function()
						{
							let assignedChatInteractions = app.m_interactionManager.getAssignedInteractionsByChannel("CHAT");
 							if (!$.isEmptyObject(assignedChatInteractions)) {
								let chatClient = app.m_interactionManager.getChannelClient("CHAT");
  								_.each(assignedChatInteractions, function(interaction) {
									chatClient.m_chatClient.RejoinSession(interaction.id);
  								});
							}

							app.m_contactManager = new inConcert.ContactManager(app.getBaseUrl());
							app.setAppsActions( response.applicationsActions );

							// Set Users & Group Permissions.
							inConcert.permissions.set( response.groupPermissions, response.userPermissions );
							// Ejecuto Request de datos Iniciales.
							app.initialRequests();
							// realiza validaciones sobre los datos recibidos en el login del usuario.
							app.validateLoggedUserInfo();
							//
							app.loadAgentScreenService(function (){
								if (!relogin) {
									app.m_mainWindow.initGUI();
								}
								else {
									app.m_mainWindow.resetPanels();
								}

								//the next function triggers gui updating, so it must be called after initGUI
								//start the poller
								app.__rearmPollerRequest();
								inConcert.app.updateCampaignCounters(response.campaignCounters);
								app.m_applicationTimer = window.setInterval(app.onTimer, 100);
								app.loadTimezones();
								app.loadLanguages();
								app.loadMailSignatures(null, null, true /* all columns */);
								app.loadAllegro();
								app.agent_api_server = new inConcert.AgentAPIServer();
								app.agent_api_server.UserLoggedIn();
								app.ValidateChannelCall(app.m_initialState === "ACTIVE", true);
								app.loadAngular();
								callback(true);
							});
						});
					}
					else {
						callback(false, response);
					}
				},
				error: function(/*response, error*/) {
					callback(false);
				}
			});
		});
	};

	//tries to login the user using the credentials passed by parameter
	Application.prototype.login = function(credentials, storeOnSuccess, callback)
	{
		inConcert.logger.action({action_name : "login", user : credentials.user, remember : storeOnSuccess, action_type:"agent"});

		var application = this;
		application.lastCredentials = storeOnSuccess ? credentials : null;
		application.currentCredentials = credentials;

		var authenticator = inConcert.InconcertAuth;
		authenticator.login(
			application.getBaseDomain(),
			credentials.user,
			credentials.password,
			callback
		);
	};

	//tries to register the user with the specified credentials
	Application.prototype.register = $.noop;

	Application.prototype.getServer = function() {
		return this.m_serverAddress;
	};

	Application.prototype.getMainWindow = function() {
		return this.m_mainWindow;
	};

	Application.prototype.getDesktopNotifications = function() {
		return this.m_desktopNotifications;
	};

	Application.prototype.logout = function(callback, callFrom) {
		inConcert.logger.action({action_name : "logout", action_type:"agent"});
		if (!this.loggedOut) {
			if (this.agent_api_server) {
				this.agent_api_server.UserLoggedOut();
			}
			this.m_desktopNotifications.dispose();
			this.loggedOut = true;
			if (this.lastRequest) {
				this.lastRequest.abort();
			}
			var authenticator = inConcert.InconcertAuth;
			authenticator.logout(callback, callFrom);
			inConcert.app.getMainWindow().destroy();
		}
	};
	
	Application.prototype.logoutWithMessage = function(params)
	{
		const self = this;
		//optional params
		let callFrom = params.callFrom || "Application-logout";
		let title = params.title || "<%=_T('Logout')%>";
		let description = params.description || "<%=_T('You have been logged out')%>";
		let logoutFirst = params.logoutFirst || false;
		let async = params.async || false;
		
		let showPopup = function(callback) {
			//1. embedded: avoid popup
			if (self.isEmbedded()) {
				return callback();
			}
			//2. show popup and wait for OK click
			$.inconcertPopup.Show(
				title,
				description,
				$.inconcertPopup.ButtonSets.BS_OK,
				callback,
				undefined,
				{
					popup_id        : "agent_inconcertPopup",
					container_class : "confirmationPopUp",
					overlay_class   : "confirmationOverlay"
				}
			);
		};
		
		let doLogout = function(callback) {
			if (async) {
				inConcert.app.logout(callback, callFrom);
			}
			else {
				inConcert.app.logout("", callFrom);
				callback(true);
			}
		};
		
		let onLoggedOut = function() {
			window.location = "/inconcert/apps/agent/login/";
		};
		
		if (logoutFirst) {
			//1. logout first
			doLogout(function() {
				showPopup(function() {
					onLoggedOut();
				});
			});
		}
		else {
			//2. message first
			showPopup(function() {
				doLogout(function() {
					onLoggedOut();
				});
			});
		}
	};

	Application.prototype.SendCustomMessage = function(interactionID, proxyType, message, waitForResponse, timeout, callbackFn )
	{
		callbackFn = callbackFn || $.noop;
		var uri = inConcert.app.getBaseDomain() + "/inconcert/apps/agent/interactions/send_custom_message/";

		var request = inConcert.invoke({
			url: uri,
			type: 'POST',
			data: JSON.stringify({
				interactionID: interactionID,
				proxyType: proxyType,
				message: message,
				waitForResponse: waitForResponse? true : false,
				timeout: timeout
			}),
			dataType: "json",
			contentType: "application/json",
			timeout: timeout * 1000 + 2000, //priority to ir timeout
			success: function(response) {
				callbackFn(true, response.data || null);
			},
			error: function() {
				callbackFn(false);
			}
		});
		return request;
	};

	/*
		Initializes application, will ask for login if user not already logged and credentials not previously stored
	*/
	Application.prototype.start = function() {
		inConcert.logger.debug("Starting application");
		this.m_desktopNotifications = new inConcert.DesktopNotifications();
		this.m_mainWindow = new inConcert.MainWindow(this);
		this.m_mainWindow.create();
	};

	/**
	 * @param callback
	*/
	Application.prototype.uploadLog = function (callback)
	{
		// TODO: rewrite the logger
		callback();
	};

	const channelsWithPages = {
		FACEBOOK  : true,
		INSTAGRAM : true
	};

	Application.prototype.channelHasPages = function(channel)
	{
		return channelsWithPages[channel]? true : false;
	};

	const sortedChannels = {
		CALL      : {name: "<%=_T('Telephony')%>"   , value: 1 },
		CHAT      : {name: "<%=_T('Web Chat')%>"    , value: 2 },
		MAIL      : {name: "<%=_T('E-mail')%>"        , value: 3 },
		SMS       : {name: "<%=_T('SMS')%>"         , value: 4 },
		RCS       : {name: "<%=_T('Rcs')%>"         , value: 5 },
		WHATSAPP  : {name: "<%=_T('WhatsApp')%>"    , value: 6 },
		FACEBOOK  : {name: "<%=_T('Facebook')%>"    , value: 7 },
		INSTAGRAM : {name: "<%=_T('Instagram')%>"   , value: 8 },
		TWITTER   : {name: "<%=_T('Twitter')%>"     , value: 9 },
		YOUTUBE   : {name: "<%=_T('Youtube')%>"     , value: 10},
		WEBCONTACT: {name: "<%=_T('Contact Form')%>", value: 11},
		APPSTORE  : {name: "<%=_T('Appstore')%>"    , value: 12},
		GOOGLEPLAY: {name: "<%=_T('Googleplay')%>"  , value: 13}
	};

	Application.prototype.load_vcc_channels = function()
	{
		return sortedChannels;
	};

	Application.prototype.channelSortNumber = function(channel)
	{
		return sortedChannels[channel] ? sortedChannels[channel].value : 100;
	};

	Application.prototype.isInRedial = function()
	{
		if (this.has_allegro_integration()) {
			return this.has_allegro_baragent() && inConcert.BaragentPolling.isRedial();
		}
		else {
			let client = this.getInteractionManager().getChannelClient("CALL");
			if (client) {
				return client._interaction && client._interaction.in_redial || false;
			}
		}
		return false;
	};

	Application.prototype.isInAppointment = function()
	{
		if (!this.has_allegro_integration()) {
			let client = this.getInteractionManager().getChannelClient("CALL");
			if (client) {
				return client._interaction && client._interaction.in_appointment || false;
			}
		}
		return false;
	};

	Application.prototype.loadVccClientsLayoutConfig = function()
	{
		if (!this.m_vcc_info.Clients) {
			return;
		}
		if (!this.m_vcc_info.Clients.Layouts) {
			return;
		}
		return inConcert.Utils.cloneObject(this.m_vcc_info.Clients.Layouts);
	};

	Application.prototype.checkAvailableAddresses = function()
	{
		this.checkAvailablePeer();
	};

	Application.prototype.checkAvailablePeer = function()
	{
		if (!this.m_peerAssigned && !this.has_allegro_integration() && !this.userHasAddress("CALL") && this.hasInteractionCampaignForAddress("CALL")) {
			this.assignPeer();
		}
	};

	Application.prototype.canVisualizeAddress = function(action_type /* opcional */, campaign_id /* opcional */, no_log /* opcional */, has_validate_all_overrides /* opcional */)
	{
		const current_app = this.getCurrentApp();
		if(current_app === "supervisor")
		{
			return true;
		}
		return inConcert.permissions.to_action( 'interaction', 'int_visualize_address', action_type, campaign_id, no_log, has_validate_all_overrides );
	};

	Application.prototype.assignPeer = function(callback = $.noop) {
		const self = this;
		this.m_peerAssigned = true;
		this.__assignAddress("CALL", function(status) {
			if (!status) {
				self.m_mainWindow.showBannerMessage({
					messageType: "Error",
					message    : "<%=_T('You were be assigned to a telephony campaign but a peer could not be assigned to you.')%>",
					timeout    : 5000
				});
			}
			callback(status);
		});
	};

	Application.prototype.__assignAddress = function(proxyType, callback = $.noop) {
		const self = this;
		inConcert.invoke({
			url: this.getBaseUrl() + "/assign_address/",
			type: "POST",
			data: JSON.stringify({ channel : proxyType }),
			dataType: "json",
			contentType: "application/json",
			success: function( address ) {
				self.__setAssignedAddress(proxyType, address, callback);
			},
			error: function(/*err_type, text_status*/) {
				callback(false);
			}
		});
	};

	Application.prototype.__setAssignedAddress = function(proxyType, address, callback = $.noop) {
		this.m_settings.InteractionClients[proxyType] = address;
		this.m_interactionManager.setChannelClient(proxyType, address);
		callback(true);
	};

	Application.prototype.loadSessionData = function(sessionData = {}) {
		this.m_sessionData = sessionData;
		this.m_automaticPollerDisabled = !!sessionData.automaticPollerDisabled;
		this.m_embedded = sessionData.embedded || false;
	};

	Application.prototype.isMobile = function() {
		if (!this.m_sessionData) {
			return false;
		}
		return (this.m_sessionData.type || "").toUpperCase() === "MOBILE";
	};

	Application.prototype.getSessionData = function() {
		return this.m_sessionData;
	};

	Application.prototype.setServerInfo = function(serverInfo) {
		this.m_serverInfo = serverInfo;
	};

	Application.prototype.getServerInfo = function() {
		return this.m_serverInfo;
	};

	Application.prototype.setIRInfo = function(IRInfo) {
		this.m_IRInfo = IRInfo;
	};

	Application.prototype.getIRInfo = function() {
		return this.m_IRInfo;
	};

	Application.prototype.isEmbedded = function() {
		return this.m_embedded;
	};

	Application.prototype.getSystemConfiguration = function(element_id /*no requerido*/) {
		let server_info = this.getServerInfo();
		let system_configuration = server_info.systemConfiguration;
		if (!element_id) {
			// si el elemento especifico no llega, retorno toda la informacion.
			return system_configuration;
		}
		let config = system_configuration[element_id];
		if (!config) {
			throw("Application.prototype.getSystemConfiguration - System configuration has not element " + element_id);
		}
		return config;
	};

	Application.prototype.isAutomaticPollerDisabled = function() {
		return this.m_automaticPollerDisabled;
	};

	Application.prototype.enableAutomaticPoller = function() {
		let wasDisabled = this.m_automaticPollerDisabled;
		this.m_automaticPollerDisabled = false;
		if (wasDisabled) {
			this.retryPoller();
		}
	};

	Application.prototype.disableAutomaticPoller = function() {
		this.m_automaticPollerDisabled = true;
	};

	Application.prototype.__cancelPoller = function() {
		this.m_waitingData = false;
		if (this.m_pollerRequest) {
			this.m_pollerRequest.abort();
		}
	};

	Application.prototype.retryPoller = function() {
		this.__cancelPoller();
		this.__rearmPollerRequest();
	};

	Application.prototype.firePoller = function(data, callback = $.noop) {
		data = data || {};
		if (data.force) {
			this.__cancelPoller();
		}
		this.dataPoller(callback, data);
	};

	/**
	Private. Schedules the next poller request.
	*/
	Application.prototype.__rearmPollerRequest = function() {
		const self = this;
		if (this.isAutomaticPollerDisabled()) {
			return;
		}
		inConcert.Utils.zeroTimeout(function() {
			self.dataPoller();
		});
		this.m_lastDataReceived = new Date().getTime();
	};

	Application.prototype.addPollerEventCallback = function(event, id, callback) {
		this.m_pollerCallbacks[event] = this.m_pollerCallbacks[event] || {};
		this.m_pollerCallbacks[event][id] = { callback : callback };
	};
	
	Application.prototype.__onPollerCallback = function(event, data) {
		if (data && data.id && this.m_pollerCallbacks[event] && this.m_pollerCallbacks[event][data.id]) {
			let callback = this.m_pollerCallbacks[event][data.id].callback;
			delete this.m_pollerCallbacks[event][data.id];
			if ($.isEmptyObject(this.m_pollerCallbacks[event])) {
				delete this.m_pollerCallbacks[event];
			}
			inConcert.Utils.zeroTimeout(callback, data);
		}
	};

	Application.prototype.StartScreenCapture = function(subdomain, onStartScreenCapture = $.noop) {
		const self = this;
		if(!self.AgentScreenService.enabled)
		{
			onStartScreenCapture(false, "AgentScreenService not enabled");
			return;
		}
		if(this.ScreenCaptureObj.capturing) {
			onStartScreenCapture(true);
			return;
		}

		try {
			let sendCaptureToUri = inConcert.app.getAgentScreenServiceBaseDomain(subdomain) + "/capture?channel=" + inConcert.app.getLoggedUserFullId();
			const WS_URL = inConcert.app.getAgentScreenLocalServiceURL();
			let socket = new WebSocket(WS_URL);
			socket.binaryType = "arraybuffer";
			socket.onopen = () => {
				console.log(`Connected to ${WS_URL}`);
				let message = {
					command : "START_CAPTURE",
					sendCaptureToUri : sendCaptureToUri,
					captureFps: inConcert.app.getAgentScreenServiceFPS(),
					captureBitrate: inConcert.app.getAgentScreenServiceBitrate()
				};
				socket.send(JSON.stringify(message));
				self.ScreenCaptureObj.capturing = true;
				self.ScreenCaptureObj.screenCaptureSocket = socket;
				onStartScreenCapture(true);
				var filterCallTakenInteraction;
				filterCallTakenInteraction = function(interaction) {
					return interaction.type === "CALL" && interaction.state === "TAKEN";
				};
				let assignedCallTakenInteractions = this.m_interactionManager.getAssignedInteractions(filterCallTakenInteraction);
				for(let assignedCallTakenInteraction in assignedCallTakenInteractions) {
					if (assignedCallTakenInteractions.hasOwnProperty(assignedCallTakenInteraction)) {
						this.AddAudioToScreenCapture(assignedCallTakenInteractions[assignedCallTakenInteraction].type, assignedCallTakenInteractions[assignedCallTakenInteraction].id, this.m_interactionManager.getChannelClient('CALL')._sip_phone);
						break;
					}
				}
			};

			// Listen for messages
			socket.onmessage = (event) => {
				let message = JSON.parse(event.data);
				if(!message.success) {
					console.error(message.error);
				}
			};

			socket.onclose = (code, reason) => {
				code = code === undefined ? 0 : code;
				reason =  reason === undefined ? "" : reason;
				console.log(`Close socket ${code} - ${reason}`);
				self.ScreenCaptureObj.capturing = false;
				if(self.ScreenCaptureObj.screenCaptureSocket && self.ScreenCaptureObj.screenCaptureSocket.readyState === 1)
				{
					self.ScreenCaptureObj.screenCaptureSocket.close();
				}
				if(self.ScreenCaptureObj.audioContext) 	{
					self.ScreenCaptureObj.audioContext.close();
					self.ScreenCaptureObj.audioContext = null;
				}
			};

			socket.onerror = (_) => {
				console.log('Error socket ${error}');
			};
		}
		catch(e){
			onStartScreenCapture(false, e);
			return;
		}
	};

	Application.prototype.StopScreenCapture = function(_, onStopScreenCapture = $.noop) {
		const self = this;
		if(!self.AgentScreenService.enabled)
		{
			onStopScreenCapture(false, "AgentScreenService not enabled");
			return;
		}
		this.ScreenCaptureObj.capturing = false;
		if(this.ScreenCaptureObj.screenCaptureSocket && this.ScreenCaptureObj.screenCaptureSocket.readyState === 1)
		{
			this.ScreenCaptureObj.screenCaptureSocket.close();
		}
		if(this.ScreenCaptureObj.audioContext)
		{
			this.ScreenCaptureObj.audioContext.close();
			this.ScreenCaptureObj.audioContext = null;
		}
		onStopScreenCapture();
	};	

	Application.prototype.InitScreenCaptureAudioContext = function(sip_phone, socket) {
		let sampleRate = 48000;
		const bufSize = 4096;
		let audioCtx = new AudioContext({sampleRate: sampleRate}); //Esto debe ser una variable, ya que se cambia el sampleRate que contiene

		// Loads module script via AudioWorklet.
		var pcmWorkerURL = inConcert.app.getBaseDomain() + "/inconcert/lib/pcmWorker/pcmWorker.js";
		audioCtx.audioWorklet.addModule(pcmWorkerURL)
			.then(() => getLiveAudio(audioCtx))
			.then((liveIn) => {
				//Initialize resampler
				var resampler = new Resampler(sampleRate, 16000, 1, bufSize);

				// After the resolution of module loading, an AudioWorkletNode can be constructed.
				let pcmWorker = new AudioWorkletNode(audioCtx, 'pcm-worker');

				// AudioWorkletNode can be interoperable with other native AudioNodes.
				liveIn.connect(pcmWorker);
				pcmWorker.port.onmessage = function (event) {
					if(socket.readyState === 1)
					{
						//Resample data before send
						const f32OutBuf = resampler.resample(event.data);
						const i16OutBuf = new Int16Array(f32OutBuf);
						socket.send(i16OutBuf);
					}
				};
				pcmWorker.port.start(); 

			})
			.catch(e => console.error(e));

		function getLiveAudio(audioCtx) {
			var recordStream = new MediaStream();
			var pc = sip_phone.m_userAgent.session.sessionDescriptionHandler.peerConnection;
			pc.getSenders().forEach(function (RTCRtpSender) {
				if(RTCRtpSender.track && RTCRtpSender.track.kind === "audio") {
					console.log("Adding sender audio track to record:", RTCRtpSender.track.label);
					recordStream.addTrack(RTCRtpSender.track);
				}
			});
			pc.getReceivers().forEach(function (RTCRtpReceiver) {
				if(RTCRtpReceiver.track && RTCRtpReceiver.track.kind === "audio") {
					console.log("Adding receiver audio track to record:", RTCRtpReceiver.track.label);
					recordStream.addTrack(RTCRtpReceiver.track);
				}
			});
			var mixedAudioVideoRecordStream = new MediaStream();

			function MixAudioStreams(MultiAudioTackStream){
				// Takes in a MediaStream with any mumber of audio tracks and mixes them together

				var audioContext = null;
				try {
					window.AudioContext = window.AudioContext || window.webkitAudioContext;
					audioContext = new AudioContext();
					sampleRate = audioContext.sampleRate;
				}
				catch(e){
					console.warn("AudioContext() not available, cannot record");
					return MultiAudioTackStream;
				}
				var mixedAudioStream = audioContext.createMediaStreamDestination();
				MultiAudioTackStream.getAudioTracks().forEach(function(audioTrack){
					var srcStream = new MediaStream();
					srcStream.addTrack(audioTrack);
					var streamSourceNode = audioContext.createMediaStreamSource(srcStream);
					streamSourceNode.connect(mixedAudioStream);
				});

				return mixedAudioStream.stream;
			}
			mixedAudioVideoRecordStream.addTrack(MixAudioStreams(recordStream).getAudioTracks()[0]);
			return audioCtx.createMediaStreamSource(mixedAudioVideoRecordStream);
		}
		return audioCtx;
	};

	Application.prototype.StartScreenRecording = function(proxyType, interactionId, sip_phone, onStartScreenRecording = $.noop) {
		const self = this;
		if(!self.AgentScreenService.enabled)
		{
			onStartScreenRecording(false, "AgentScreenService not enabled");
			return;
		}
		if(!proxyType || !interactionId)
		{
			onStartScreenRecording(false, "proxyType and interactionId required");
			return;
		}
		if(proxyType === "CALL" && !sip_phone)
		{
			onStartScreenRecording(false, "proxyType CALL requires sip_phone object");
			return;
		}
		inConcert.app.StartScreenCapture(undefined /*subdomain*/, function(success, error) {
			if(success)
			{
				if(proxyType === "CALL" && !self.ScreenCaptureObj.audioContext)
				{
					self.ScreenCaptureObj.audioContext = self.InitScreenCaptureAudioContext(sip_phone, self.ScreenCaptureObj.screenCaptureSocket);	
				}
				var message = {
					command : "START_RECORDING",
					proxyType : proxyType,
					interactionId: interactionId
				};
				self.ScreenCaptureObj.screenCaptureSocket.send(JSON.stringify(message));
			}
			onStartScreenRecording(success, error);			
		});
	};

	Application.prototype.StopScreenRecording = function(interactionId, onStopScreenRecording = $.noop) {
		const self = this;
		if(!self.AgentScreenService.enabled)
		{
			onStopScreenRecording(false, "AgentScreenService not enabled");
			return;
		}
		if(!interactionId)
		{
			onStopScreenRecording(false, "interactionId required");
			return;
		}
		if(self.ScreenCaptureObj.capturing)
		{
			let message = {
				command : "STOP_RECORDING",
				interactionId: interactionId
			};
			self.ScreenCaptureObj.screenCaptureSocket.send(JSON.stringify(message));

			if(self.ScreenCaptureObj.audioContext)
			{
				self.ScreenCaptureObj.audioContext.close();
		    	self.ScreenCaptureObj.audioContext = null;
	    	}			
			onStopScreenRecording(true);
		}
		else
		{
			onStopScreenRecording(false, "Capture socket not ready");
		}
	};

	Application.prototype.AddAudioToScreenCapture = function(proxyType, interactionId, sip_phone) {
		const self = this;
		if(!self.AgentScreenService.enabled || !proxyType || !interactionId || proxyType !== "CALL" || !sip_phone || !self.ScreenCaptureObj.capturing || self.ScreenCaptureObj.audioContext)
		{
			return;
		}
		self.ScreenCaptureObj.audioContext = this.InitScreenCaptureAudioContext(sip_phone, self.ScreenCaptureObj.screenCaptureSocket);
	};

	Application.prototype.HasExternalUsersAuth = function() {
		return this.m_vcc_info.HasExternalUsersAuth;
	};

	Application.prototype.HasExternalUsersSync = function() {
		return this.m_vcc_info.HasExternalUsersSync;
	};

	Application.prototype.PasswordExpirationDays = function() {
		return this.m_passwordExpirationDays;
	};

	Application.prototype.getExternalGroupFilters = function(callbackFn) {
		var url = this.getBaseUrl() + "/refresh_external_group_filters/";
		inConcert.invoke({
			url: url,
			type : "POST",
			dataType: "json",
			contentType: "application/json",
			data: JSON.stringify({}),
			success: function(result) {
				callbackFn(true, result);
			},
			error: function() {
				callbackFn(false);
			}
		});
	};

	/*-- BEGIN INCO-64 ---------------------------------------------------*/
	Application.prototype.loadAgentMonitorService = function(callbackFn = $.noop) {
		let self = this;
		try
		{
			// Todo esto habria que sacarlo a otro javascript dejando disponible
			// sobre Inconcert.app un api de monitoreo

			class MonitorSocket {
				#m_URL;
				#m_wsHandle = null;
		
				constructor(url){
					this.#m_URL = url;
				}
				getSendedMessages(){
					let store = window.localStorage;
					try {
						let items = store.getItem("websocket-"+ this.#m_URL) || '{}';
						let msgs = JSON.parse(items);
						return msgs;
					}catch(e){
						return {};
					}
				}
				storeSendedMessages(msgs){
					let store = window.localStorage;
					try {
						store.setItem("websocket-"+ this.#m_URL,JSON.stringify(msgs));
					}catch(e){
					}
				}
				storeMessage(msg){
					try {
						let msgs = this.getSendedMessages();
						msgs[msg._msgid]=msg;
						this.storeSendedMessages(msgs);
					}catch(e){

					}
				}

				deleteSendedMessage(msgid){
					let msgs=this.getSendedMessages();
					delete msgs[msgid];
					this.storeSendedMessages(msgs);
				}

				getInstance(socketReadyFn = $.noop){
					let wsState = this.#m_wsHandle?this.#m_wsHandle.readyState:-1;
					switch(wsState){
						case 0: 
								setTimeout(()=>{
									this.getInstance(socketReadyFn);
								},10);
						break;
						case 1:
							socketReadyFn(true, this.#m_wsHandle);
						break;

						default:
						/* Socket is unavailable. Create a new instance */
							this.#m_wsHandle = new WebSocket(this.#m_URL);
							this.#m_wsHandle.onopen = ()=> {
								let queue = this.getSendedMessages()
								let c = [];
								for (let msgid in queue){
									if(queue.hasOwnProperty(msgid)){
										c.push(queue[msgid])
									}
								}
								c.sort((a,b)=> a._msgid>b._msgid);
								for(let i=0;i<c.length;i++){
									this.#m_wsHandle.send(JSON.stringify(c[i]));
								}
								socketReadyFn(true, this.#m_wsHandle);
							};
	
							this.#m_wsHandle.onmessage = (e) => {
								try {
									let m = JSON.parse(e.data);
									this.deleteSendedMessage(m._msgid);
									this.onMessage(e);
	
								}catch(err){

								}
							};
							this.#m_wsHandle.onerror = (e) => {
								socketReadyFn(false, null);
								this.#m_wsHandle = null;
							};
							
							this.#m_wsHandle.onclose = ()=> {
								this.onClose();
								this.#m_wsHandle = null;
							};
					}
					
				}
				send(data) {
					data._msgid = Date.now()+Math.floor(Math.random() * 1e5);
					this.getInstance((success, wsHandle) => {
						if (success) {
							this.storeMessage(data);
							wsHandle.send(JSON.stringify(data))
						}
					});
				}

				static sendOneShot (wsURL, wsData) {
					
					return new Promise((resolve,reject)=>{
						let ws = new WebSocket(wsURL);
						wsData._msgid = Date.now();
						ws.onopen = function(e) {
							ws.send(JSON.stringify(wsData));
						};
						
						ws.onmessage = function(e) {
							resolve(e.data);
							/* Since this is a oneshot socket, when a message is received, I must close it */
							ws.close();
						};
		
						ws.onerror = function(e) {
							reject (e);
							ws.close();
						};
		
						ws.onclose = function (e){
							console.log("On close received");
						};
		
					});
				}
				/* WebSocket events */
		
				onError(){}
				onOpen(){}
				onMessage(){}
				onClose(){}
		
			}
		
		
			class CaptureSocket extends MonitorSocket {
				onMessage = $.noop;

				onError = function (e) {
					console.error("[AgentMonitor]: ERROR -> " + e.data);
				}
				onClose = function (code, reason) {
					code = code === undefined ? 0 : code;
					reason = reason === undefined ? "" : reason;
					console.info(`[AgentMonitor] Close socket ${code} - ${reason}`);
				};

			}
			class WatchSocket extends MonitorSocket {
				#listenerLastId = 0;
				#listeners = {};
				#pingInterval = 0;
				constructor(url){
					super(url)
					this.#pingInterval = setInterval(()=>{
						this.send({command: 'ping'})
					},30000)
				}
				onMessage (e) {
					for(let listenerId in this.#listeners){
						if(this.#listeners.hasOwnProperty(listenerId)){
							try{
								this.#listeners[listenerId](true,JSON.parse(e.data));
							}catch(e){}
						}
					}
				};
				setMessageListener(listenerFn){
					let id = ''+ ++this.#listenerLastId;
					this.#listeners[id]  = listenerFn;
					return id;
				}
				clearMessageListener(listenerId){
					delete this.#listeners[listenerId];
				}

				onError (e) {
					console.error("[AgentMonitorWatch]: ERROR -> " + e.data);
				};
				onClose (code, reason) {
					code = code === undefined ? 0 : code;
					reason = reason === undefined ? "" : reason;
					console.info(`[AgentMonitorWatch] Close socket ${code} - ${reason}`);
				};

			}

			inConcert.app.AgentMonitorObj = {
				watchs:{}, 
				captureSocket: new CaptureSocket(this.getAgentMonitorServiceBaseDomain() + "/capture?channel=" +this.getLoggedUserFullId()),
				sendOneShot: MonitorSocket.sendOneShot,
				watch: function(agent_at_vcc, message){
					return new Promise((resolve,reject)=>{
						try{
							this.watchs[agent_at_vcc] = this.watchs[agent_at_vcc] || new WatchSocket(self.getAgentMonitorServiceBaseDomain() + "/watch?channel=" +agent_at_vcc);
							resolve(this.watchs[agent_at_vcc].send(message));
						}catch(e){
							reject(e)
						}
	
					})
				},
				watchMessages: function (agent_at_vcc,onMessage){
					this.watchs[agent_at_vcc] = this.watchs[agent_at_vcc] || new WatchSocket(self.getAgentMonitorServiceBaseDomain() + "/watch?channel=" +agent_at_vcc);
					let id = this.watchs[agent_at_vcc].setMessageListener(onMessage)
					let dummy_command = {
						command: "ping"
					}
					this.watchs[agent_at_vcc].send(dummy_command);

					return id;
				},
				unwatchMessages: function (agent_at_vcc,listenerId){

					if(this.watchs[agent_at_vcc]){
						this.watchs[agent_at_vcc].clearMessageListener(listenerId);
					}
				},
				WebRTC: null,
				SpeedTest: null,
				Logging: null
			};
			/* init WebRTC instance */
			inConcert.app.AgentMonitorObj.WebRTC = {
				capturing: false,
				receiveLastValueTimestamp:0,
				receiveLastValueBytes:0,
				receiveLastValuePackets:0,
				sentLastValueTimestamp:0,
				sentLastValueBytes:0,
				sentLastValuePackets:0,



				statsReportCallback: function(interactionId, stats) {
					const webSocket = inConcert.app.AgentMonitorObj.captureSocket;

					let webrtc_json = {
						
					};

					if (!interactionId) {
						console.warn("[WebRTC] Received statistics, but no interactionId provided.");
						return;
					}

					let webrtc_data = []

					stats.forEach((report)=> {
						var now = new Date().getTime();
						switch(report.type){
							case "outbound-rtp":{
								if(this.sentLastValueTimestamp == 0) {
									this.sentLastValueTimestamp = report.timestamp;
									this.sentLastValuePackets = report.packetsSent;
									this.sentLastValueBytes = report.bytesSent;
									return;
								}
								let kbitsPerSec = (8 * (report.bytesSent - this.sentLastValueBytes))/1000;
								let PacketsPerSec = (report.packetsSent - this.sentLastValuePackets);

								this.sentLastValueBytes = report.bytesSent;
								this.sentLastValuePackets = report.packetsSent;
								this.sentLastValueTimestamp = report.timestamp;

								webrtc_data.push(['webrtc.sent',
								{
									ts: now,
									value: {
										kbitsPerSec: kbitsPerSec,
										packetsPerSec: PacketsPerSec,
									}
								}
							])
								
							}
							break;
							case "inbound-rtp":{
								if(this.receiveLastValueTimestamp == 0) {
									this.receiveLastValueTimestamp = report.timestamp;
									this.receiveLastValueBytes = report.bytesReceived;
									this.receiveLastValuePackets = report.packetsReceived;
									this.receiveLastValuePacketLoss = report.packetsLost;
									return;
								}
		
								// Receive Kilobits Per Second
								let kbitsPerSec = (8 * (report.bytesReceived - this.receiveLastValueBytes))/1000;
								
								// Receive Packets Per Second
								let PacketsPerSec = (report.packetsReceived - this.receiveLastValuePackets);
								
								// Receive Packet Loss
								let PacketsLost = (report.packetsLost - this.receiveLastValuePacketLoss);
	
								this.receiveLastValueTimestamp = report.timestamp;
								this.receiveLastValueBytes = report.bytesReceived;
								this.receiveLastValuePackets = report.packetsReceived;
								this.receiveLastValuePacketLoss = report.packetsLost;
	
								webrtc_data.push(['webrtc.received',
									{
										ts: now,
										value: {
											kbitsPerSec: kbitsPerSec,
											packetsPerSec: PacketsPerSec,
											packetsLost: PacketsLost,
											jitter: report.jitter
										}
									}
								]);
	
							}
								
						}
					});
					if(webrtc_data.length>0){
						webrtc_json.webRTCStats = webrtc_data;
						/* Envio el objeto 'webrtc_json' para el websocket */
						webSocket.send(webrtc_json);
					}

				},
				release: function() {
					this.capturing = false;
				}
			};
			
			/* init SpeedTest instance */
			inConcert.app.AgentMonitorObj.SpeedTest = {
				speedTestInstance: null,
				release: function() {
					// getState == 3 -> 'test running' || getState == 4 -> 'done'
					if (this.SpeedTest.speedTestInstance.getState() == 3 || this.SpeedTest.speedTestInstance.getState() == 4) {
						this.SpeedTest.speedTestInstance.abort();
					}
					this.SpeedTest.speedTestInstance = null;
				}
			};
			
			/* init Logging instance */
			inConcert.app.AgentMonitorObj.Logging = {
				capturing: false,
				allowedMethods: ['log', 'warn', 'error'],
				originalConsole: {},
				logMessages: [],
				addLogMessage: function(message,method) {
					/* get current stack trace using the Error object */
					let _e = new Error();
					var ts_now = Math.floor(Date.now());
					var log_data = {
						message: message,
						timestamp: ts_now,
						method: method
					};

					this.logMessages.push(log_data);
				},
				flushLogMessages: function() {					
					let self = this;
					icc.bm.helpers.setTimeout(function(){
						try
						{
							if(self.logMessages.length>0){
								const webSocket = inConcert.app.AgentMonitorObj.captureSocket;
								let json_data = {
									agentLogs: self.logMessages
								};
							
															/* Clear logMessages array */
								self.logMessages = [];
								webSocket.send(json_data);
	
							}
	
						}
						catch (e)
						{
						}
					},100)
				},
				shouldFlush: function() {
					return (this.logMessages.length >= 2);
				},
				release: function() {
					this.flushLogMessages();
				}
			};
			inConcert.app.AgentMonitorObj.BrowserHealthMonitor = new icc.bm.browserMonitor();
			inConcert.app.AgentMonitorObj.WebRTC.statsReportCallback = 
					inConcert.app.AgentMonitorObj.WebRTC.statsReportCallback.bind(
							inConcert.app.AgentMonitorObj.WebRTC
					);
			/* initialize the mainSocket instance */

			callbackFn(true);
		}
		catch (e)
		{
			console.error(e);
			callbackFn(false, e);
		}
	}

	Application.prototype.SendAgentGeolocation = function(subdomain, callbackFn = $.noop) {
		if (navigator.geolocation)
		{
			const self = this;

			/* Recupero las informaciones de la geolocalizacion */
			navigator.geolocation.getCurrentPosition( function(pos) {
				var message = {
					timestamp: pos.timestamp,
					latitude: pos.coords.latitude,
					longitude: pos.coords.longitude
				};

				let ws_url = self.getAgentMonitorServiceBaseDomain() + "/geolocation?channel=" + inConcert.app.getLoggedUserFullId();
				self.AgentMonitorObj.sendOneShot(ws_url, message)
				.then((message) => console.log("[AgentMonitor] send geolocation data was responded: ",message))
				.catch((error) => console.log("[AgentMonitor] send geolocation data error: ",error));
			});
		}
	};

	Application.prototype.EnableWebRtcCapture = function(onEnableWebRtcCapture = $.noop)
	{
		const self = this;
		try
		{
			if (this.AgentMonitorObj.WebRTC.capturing) {
				onEnableWebRtcCapture(false, "WebRTC capturing is enabled already.");
				return;
			}

			var channel_client = self.getInteractionManager().getChannelClient("CALL");
			if (!channel_client) {
				onEnableWebRtcCapture(false, "Could not obtain a valid phone client.");
				return;
			}

			/* Envio el mesaje de 'enable statistics' para el objeto phone_client */
			if (channel_client.StartStatisticsCapture)
			{
				channel_client.StartStatisticsCapture(this.AgentMonitorObj.WebRTC.statsReportCallback, function(success, message) {
					if (!success) {
						onEnableWebRtcCapture(false, message);
					}
				});

				this.AgentMonitorObj.WebRTC.capturing = true;
				onEnableWebRtcCapture(true);
			}
			else
			{
				this.AgentMonitorObj.WebRTC.release();
				onEnableWebRtcCapture(false, "Phone client is not valid.");
			}
		}
		catch(e)
		{
			onEnableWebRtcCapture(false, e);
		}
	};

	Application.prototype.DisableWebRtcCapture = function(onDisableWebRtcCapture = $.noop)
	{
		try
		{
			if (!this.AgentMonitorObj.WebRTC.capturing) {
				onDisableWebRtcCapture(true, "WebRTC capturing is disabled already.");
				return;
			}

			let channel_client = this.getInteractionManager().getChannelClient("CALL");
			if (!channel_client) {
				onDisableWebRtcCapture(false, "Could not obtain a valid phone client.");
				return;
			}

			/* Envio el mesaje de 'stop statistics' para el objeto phone_client */
			if (channel_client.StopStatisticsCapture)
			{
				channel_client.StopStatisticsCapture(function(success, message) {
					if (!success) {
						onDisableWebRtcCapture(false, message);
					}
				});

				this.AgentMonitorObj.WebRTC.release();
				onDisableWebRtcCapture(true);
			}
			else
			{
				onDisableWebRtcCapture(false, "Phone client is not valid.");
				return;
			}
		}
		catch(e)
		{
			onDisableWebRtcCapture(false, e);
			return;
		}
	};

	Application.prototype.StartWebRtcCapture = function (agent_at_vcc, onStartWebRtcCapture = $.noop) {
		/* Send 'start_webrtc_capture' to the service */
		let message = {
			command: "start_webrtc_capture"
		};
		this.AgentMonitorObj.watch(agent_at_vcc,message)
			.then(() => onStartWebRtcCapture(true))
			.catch((e) => onStartWebRtcCapture(false, e));
	};

	Application.prototype.StopWebRtcCapture = function (agent_at_vcc, onStopWebRtcCapture = $.noop) {
		let message = {
			command: "stop_webrtc_capture"
		};
		this.AgentMonitorObj.watch(agent_at_vcc, message)
			.then(() => onStopWebRtcCapture(true))
			.catch((e) => onStopWebRtcCapture(false, e));
	};

	/* Metodo para capturar (wrapper) window.console */
	Application.prototype.interceptWindowConsole = function(method)
	{
		
		var console = window.console;
		if (!console) {
			alert("No existe la consola !");
			return;
		}
		const loggingObj = this.AgentMonitorObj.Logging;

		/* Guardo la instancia original para poder restaurarla despues */
		loggingObj.originalConsole[method] = console[method];
		let original = loggingObj.originalConsole[method];

		console[method] = function() {
			var message = Array.prototype.slice.apply(arguments).join(" ");
			loggingObj.addLogMessage(message,method);

			if (loggingObj.shouldFlush()) {
				loggingObj.flushLogMessages();
			}
			
			original(message);
		};
	};

	Application.prototype.restoreWindowConsole = function(method)
	{
		var console = window.console;
		if (!console)
		{
			return;
		}

		if (typeof(this.AgentMonitorObj.Logging.originalConsole[method]) === "function")
		{
			console[method] = this.AgentMonitorObj.Logging.originalConsole[method];
		}

		this.AgentMonitorObj.Logging.originalConsole[method] = null;
	};

	/* INCO73: Crear metodos del administrador para Start/Stop logging capture */
	Application.prototype.StartLoggingCapture = function(agent_at_vcc, onStartLoggingCapture = $.noop){
		let message = {
			command: "start_logging_capture"
		};
		this.AgentMonitorObj.watch(agent_at_vcc, message)
			.then(() => onStartLoggingCapture(true))
			.catch((e) => onStartLoggingCapture(false, e));
	};

	Application.prototype.StopLoggingCapture = function(agent_at_vcc, onStopLoggingCapture = $.noop){
		let message = {
			command: "stop_logging_capture"
		};
		this.AgentMonitorObj.watch(agent_at_vcc, message)
			.then(() => onStopLoggingCapture(true))
			.catch((e) => onStopLoggingCapture(false, e));
	};

	Application.prototype.EnableLoggingCapture = function(onEnableLoggingCapture = $.noop)
	{
		try
		{
			if (this.AgentMonitorObj.Logging.capturing === true) {
				onEnableLoggingCapture(true, "Logging capture has already been enabled.");
				return;
			}
			this.errListener = (event, source, lineno, colno, error) => {
				try {
					console.error(event.message + " " + event.filename + " line " + event.lineno)
				}catch(e){
				}
			}
			window.addEventListener("error",this.errListener);

			this.interceptWindowConsole("log");
			this.interceptWindowConsole("warn");
			this.interceptWindowConsole("error");
			this.interceptWindowConsole("debug");
			this.interceptWindowConsole("info");
			this.AgentMonitorObj.Logging.capturing = true;

			onEnableLoggingCapture(true);
		}
		catch(e)
		{
			onEnableLoggingCapture(false, e);
			return;
		}
	};

	Application.prototype.DisableLoggingCapture = function(onDisableLoggingCapture = $.noop)
	{
		try
		{
			var loggingObj = this.AgentMonitorObj.Logging;
			if(this.AgentMonitorObj.Logging.capturing)
			{
				window.removeEventListener("error",this.errListener);
				this.restoreWindowConsole("log");
				this.restoreWindowConsole("warn");
				this.restoreWindowConsole("error");
				this.restoreWindowConsole("debug");
				this.AgentMonitorObj.Logging.capturing = false;

				loggingObj.release();
				onDisableLoggingCapture(true);
			}
			else
			{
				onDisableLoggingCapture(true, "Logging capture has already been disabled.");
			}
		}
		catch (e)
		{
			onDisableLoggingCapture(false, e);
		}
	};

	Application.prototype.ExecuteSpeedTest = function(test_data,onExecuteSpeedTest = $.noop)
	{
		const self = this;
		try
		{
			const socket = this.AgentMonitorObj.captureSocket;

			/* Create a new instance of the speedtest library */
			var speedTest = self.AgentMonitorObj.SpeedTest.speedTestInstance;
			if (!speedTest) {
				speedTest = new Speedtest();
				speedTest.setParameter("url_dl", inConcert.app.getBaseUrl() + "/users/get_download_sample/?size=10240");
				speedTest.setParameter("url_getIp", inConcert.app.getBaseUrl() + "/users/get_client_ip/");
				speedTest.setParameter("mpot", false);
				speedTest.setParameter("getIp_ispInfo", false);
				self.AgentMonitorObj.SpeedTest.speedTestInstance = speedTest;
			}
			speedTest.start();

			var speedtest_json = {
				speedTest: {}
			};

			speedTest.onupdate = function(data) {
				{
					speedtest_json.speedTest = {
						speedtest_id:test_data.speedtest_id,
						dlProgress: data.dlProgress,
						ulProgress: data.ulProgress,
						download: data.dlStatus,
						upload: data.ulStatus,
						ping: data.pingStatus,
						ip: data.clientIp,
						jitter: data.jitterStatus
					};
					
					socket.send(speedtest_json);
				}
			};

			onExecuteSpeedTest(true);
		}
		catch(e)
		{
			onExecuteSpeedTest(false, e);
			return;
		}
	};
	
	Application.prototype.EnableBrowserHealthCapture = function () {
		console.log("Browser Health Monitoring was started !");
		if (!this.AgentMonitorObj.browserHealthIsMonitoring){
			let opts = {httpmonitor:{sites:[{name: "Main Server", url:  inConcert.app.getBaseUrl()+"/ping"}]}};
			this.AgentMonitorObj.BrowserHealthMonitor.startHealthCapture(opts, (data_chunk) => {
				const socket = this.AgentMonitorObj.captureSocket;
				socket.send({browserHealthStats: data_chunk});
			});
			this.AgentMonitorObj.browserHealthIsMonitoring = true;
		}
	};
	Application.prototype.DisableBrowserHealthCapture = function () {
		if(this.AgentMonitorObj.browserHealthIsMonitoring){
			this.AgentMonitorObj.BrowserHealthMonitor.stopHealthCapture();
			this.AgentMonitorObj.browserHealthIsMonitoring=false;
		}
	};


	Application.prototype.StartBrowserHealthCapture = function (agent_at_vcc, onStartBrowserHealthCapture = $.noop) {
		/* Send 'start_webrtc_capture' to the service */
		let message = {
			command: "start_browser_health_capture"
		};
		this.AgentMonitorObj.watch(agent_at_vcc, message)
			.then(() => onStartBrowserHealthCapture(true))
			.catch((e) => onStartBrowserHealthCapture(false, e));
	};

	
	Application.prototype.StartSpeedTest = function (agent_at_vcc, onStartSpeedTest = $.noop) {
		/* Send 'start_webrtc_capture' to the service */
		let message = {
			command: "execute_speedtest"
		};
		this.AgentMonitorObj.watch(agent_at_vcc, message)
			.then(() => onStartSpeedTest(true))
			.catch((e) => onStartSpeedTest(false, e));
	};

	Application.prototype.StopBrowserHealthCapture = function (agent_at_vcc, onStopBrowserHealthCapture = $.noop) {
		/* Send 'stop_browser_health_capture' to the service */
		let message = {
			command: "stop_browser_health_capture"
		};

		this.AgentMonitorObj.watch(agent_at_vcc, message)
			.then(() => onStopBrowserHealthCapture(true))
			.catch((e) => onStopBrowserHealthCapture(false, e));
	};

	/*-- END INCO-64 ---------------------------------------------------*/
	
	Application.prototype.getListCurrentsCategories = function() {
		return this.listCurrentsCategories;
	};

	Application.prototype.isWebMicrophonePermissionGranted = function() {
		return this.m_web_microphone_permission;
	};

	Application.prototype.ValidateChannelCall = function (canChangeStatus, validatePhoneClient) {
		const self = this;
		//No para allegro
		if (this.has_allegro_integration()) {
			return;
		}

		//Valido los permisos del micrifono
		if (!this.isWebMicrophonePermissionGranted()) {
			//Se chequea el permiso y se agrega evento de cambio
			this.CheckMicrophonePermissions(function (isGranted, isChanged) {

				//Si no se tiene Cuentas de call o no esta en estado activo o es forzado a no realizar cambio, no hago nada
				if ((!isChanged && !canChangeStatus) || (inConcert.app.getCampaignsByAddress("CALL") || []).length === 0 || (isChanged && self.getCurrentStatus() !== "ACTIVE") || isGranted) {
					return;
				}
				self.execChangeUserState("NOACD", "<%=_T('Microphone permissions are required to make and receive phone calls.')%>");
			});
		}

		//Validar registro de phone client
		if (validatePhoneClient) {
			this.m_validatePhoneClientTimer = window.setInterval(this.ValidatePeerRegistrationAndChangeToNOACD, 5000);
		}
	};

	Application.prototype.execChangeUserState = function(state, message, callback = $.noop) {
		this.changeUserState(state, (response)=> {
			if (!response.success) {
				console.log("warn", null, message);
				callback(response);
				return;
			}
			if (message) {
				message = "<%=_T('Your current state has been updated.')%>" + " " + message;
				this.m_mainWindow.showBannerMessage({
					messageType : "Info",
					message     : message,
				}, undefined, true);
			}
			this.m_mainWindow.setUserStatus(state);
			callback(response);
		});
	};

	Application.prototype.ValidatePeerRegistrationAndChangeToNOACD = function(canChangeStatus) {
		if (inConcert.app.loggedOut) {
			clearInterval(inConcert.app.m_validatePhoneClientTimer);
			return;
		}

		if (inConcert.app.has_allegro_integration() || (!canChangeStatus && inConcert.app.getCurrentStatus() !== "ACTIVE") || (inConcert.app.getCampaignsByAddress("CALL") || []).length === 0) {
			return true;
		}
		const clientCall = inConcert.app.getInteractionManager().getChannelClient("CALL");
		if (clientCall && !clientCall.isPeerRegistered()) {
			inConcert.app.execChangeUserState("NOACD", "<%=_T('The phone client is not registered.')%>");
			return false;
		}
		return true;
	};

	Application.prototype.CheckMicrophonePermissions = function(callback = $.noop) {
		const self = this;
		self.m_web_microphone_permission = false;
		if (!navigator || !navigator.permissions) {
			return callback(false);
		}
		
		navigator.permissions.query({name:'microphone'}).then(function(permission) {
			self.m_web_microphone_permission = permission.state === "granted";
			callback(self.m_web_microphone_permission);
			permission.onchange = function(){
				self.m_web_microphone_permission = this.state === "granted";
				callback(self.m_web_microphone_permission, true);
			};
		   })
		   .catch(_ => {
			callback(false);
		   });
	};

	Application.prototype.get_internal_campaign_id = function() {
		return this.m_internalCampaignId + "@" + this.m_vcc_info.Id;
	};

	Application.prototype.loadAngular = function() {
		if (window.loadAngularReloadButton) loadAngularReloadButton();
		if (window.loadPartialAngularReloadButton) loadPartialAngularReloadButton()
	};

	Application.prototype.toggleConfNavigation = function(show, $container_grid, $container_nav) {
		$container_nav = $container_nav || $("#conf_navigation");
		$container_nav.toggle(show);
		$container_grid = $container_grid || $("#conf_grid_scroll");
		$container_grid.toggleClass("confEditFullScreenMode", !show);
	};

	Application.prototype.startProgressiveTry = function(message) {
		this.stopProgressiveTry();

		const { address: { Number: number }, campaignAtVCC, contactID: id, contactName } = message;
		const campaignName = inConcert.Utils.getIdWithoutVCC(campaignAtVCC);
		$("body").append(`
			<div id="oe-progressive-try-notification" class="oe-progressive-try-notification">
				<i class="oe-progressive-try-icon far fa-phone-rotary"></i>
				<div>
					<p class="oe-progressive-try-title">
						<%=_T([[Calling]])%> ${contactName || id}
					</p>
					<p class="oe-progressive-try-message">
						${campaignName} +${number}
					</p>
				</div>
			</div>
		`);
	};

	Application.prototype.stopProgressiveTry = function() {
		$("#oe-progressive-try-notification").remove();
	};
	
	//prevent context menu
	window.oncontextmenu = function() {
		return inConcert.app && inConcert.app.settings && inConcert.app.settings.development;
	};

	//prevent backward navigation
	window.history.pushState( null, document.title );
	window.onpopstate = function(/*event*/) {
		window.history.pushState( null, document.title );
	};

	//AppCache stuff
	function onUpdateReady() {
		if (window.applicationCache.status === window.applicationCache.UPDATEREADY) {
			// Browser downloaded a new app cache.
			// Swap it in and reload the page to get the new hotness.
			sessionStorage.loggedFromForm = true;
			inConcert.app.reloadForUpdate = true;
			sessionStorage.reloadForUpdate = true;
			sessionStorage.removeItem("alreadyInitialized");
			window.applicationCache.swapCache();

			inConcert.invoke({
				url: inConcert.app.getAllegroUrl() + "/login/has_baragent/",
				dataType: "json",
				contentType: "application/json",
				success: function(response) {
					if (response && response.baragent) {
						inConcert.app.logoutWithMessage({
							callFrom : "Application-onUpdateReady",
							title : "<%=_T('Application updated')%>",
							description : "<%=_T('Please login again to continue.')%>",
							logoutFirst : true,
							async : true
						});
					}
					else {
						window.location.reload();
					}
				},
				error: function(/*response, error*/) {
					window.location.reload();
				}
			});
		}
		else if (window.applicationCache.status === window.applicationCache.IDLE) {
			$("body").removeClass("downloading");
		}
		else {
			setTimeout(onUpdateReady, 1000);
		}
	}

	if (sessionStorage.reloadForUpdate) {
		sessionStorage.removeItem("reloadForUpdate");
		setTimeout(function() {
			//only for testing environment
			if (inConcert.app.m_serverAddress === "i6.inconcert.cc") {
				setTimeout(function(){
					//celebration!
					(function(){if($("#celebrationContainer").length===0){$("body").append("<div id='celebrationContainer'></div>");$("body").append("<style> .celebration_paper {z-index:10000000;min-height: 20px;min-width: 20px;max-width: 20px;max-height: 20px; position: absolute;opacity: 1;-webkit-transition-property: top, left, transform, opacity;  -webkit-transition-timing-function: linear, linear, linear, cubic-bezier(1,-0.87,.93,.57);}</style>");}function random(a,b){return Math.round(Math.random()*(b-a)+a);}$("#celebrationContainer").html("");
					for(var i=0;i<100;i++){var $paper=$('<div class="celebration_paper"></div>');$paper.css("left",random(400,800)+"px");$paper.css("top",random(-100,-500)+"px");$paper.css("background-color","rgb("+random(50,255)+","+random(50,255)+","+random(50,255)+")");$("#celebrationContainer").append($paper);}setTimeout(function(){$(".celebration_paper").each(function(){var $e=$(this);var x=random(0,1E3)+"px";var y=random(400,600)+"px";var rx=random(2E3,5E3)+"deg";var ry=random(2E3,5E3)+"deg";var rz=random(2E3,5E3)+
					"deg";$e.css({"transition-duration":random(4E3,8E3)+"ms"});$e.css({"transform":"rotateX("+rx+") rotateY("+ry+") rotateZ("+rz+")","top":y,"left":x,"opacity":0});$e.one("webkitTransitionEnd otransitionend oTransitionEnd msTransitionEnd transitionend",function(){$e.remove();});});},100);})();
				}, 1000);


				var url = "sounds/app_update.mp3";
				var updateSound = new Audio(url);
				updateSound.play();
			}
		}, 2000);
	}
	
	if (window.applicationCache) {
		window.applicationCache.addEventListener('error', function(e) {
			console.error(e);
			inConcert.app.logoutWithMessage({
				callFrom : "Application-cache",
				title : "<%=_T('Error updating application')%>",
				description : "<%=_T('There was an error updating the application. Please retry in a few seconds.<br>If the error persists, contact with your administrator.<br><br>Press OK to retry login.')%>"
			});
		}, false);

		window.applicationCache.addEventListener('downloading', function(/*e*/) {
			if (window.applicationCache.status === window.applicationCache.DOWNLOADING) {
				$("body").addClass("downloading");
				$(".downloadingWaitingAnimationAux").css("width", "0%");
				onUpdateReady();
			}
		}, false);

		window.applicationCache.addEventListener('progress', function(e) {
			$(".downloadingWaitingAnimationAux").css("width", 100 * e.loaded/e.total + "%");
		}, false);

		window.applicationCache.addEventListener('updateready', onUpdateReady, false);
	}

	//in case of tab duplication
	if (sessionStorage.alreadyInitialized) {
		sessionStorage.removeItem("loggedFromForm");
		sessionStorage.removeItem("alreadyInitialized");
	}

	if (!sessionStorage.loggedFromForm) {
		//user has arrived here without having logged in from login form
		//redirect to login form
		inConcert.InconcertAuth.logout();
		window.location = "/inconcert/apps/agent/login/";
	}
	else {
		sessionStorage.alreadyInitialized = true;
	}

	inConcert.ApplicationScript = document.currentScript;

}());
