/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:false, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, quotmark: false, esversion: 11 */

/* global inConcert, InconcertChat, MediaSource, SourceBuffer */

/*
	This is where the chat client is implemented.
*/

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
    this.inConcert = {};
}
 

(function () {

/**
	 * Class InteractionScreenSharing
	 * 
	 *
	 * @param {Object} 
	 *		baseUrl: wss://agentscreenservice[NodeId]-<domain>
	 *		clientType: 'watch'
	 *		session_id: Chat session id
	 *		videoContainerId: DOM object id which will contain the remote video		
	 */
	class InteractionScreenSharing
	{
		constructor (params, callback) {
			try {
				this.m_baseUrl = params.baseUrl;
				this.m_clientType = params.clientType;
				this.m_session_id = params.session_id;
				this.m_videoContainerId = params.videoContainerId;
				this.m_videoContainer = document.getElementById(this.m_videoContainerId);
				
				this.onReadyScreenSharing = params.onReadyScreenSharing || function(){};
				this.onStartScreenSharing = params.onStartScreenSharing || function(){};
				this.onStopScreenSharing = params.onStopScreenSharing || function(){};
				this.onCloseScreenSharing = params.onCloseScreenSharing || function(){};
				
				//State of this defined on next three values
				this.m_isInitialized = false;
				this.m_isReadyToReceive = false;
				this.m_isSharing = false;
				
				this.m_videoMimeType = 'video/webm; codecs="vp8"';
				this.m_wsPingIntervalMilliseconds = 30000;
				this.m_wsPingIntervalObj = undefined;
				
				//Websocket commands
				this.m_wsPingCommand = {command: "ping"};
				this.m_wsReadyCommand = {command: "ready_to_receive"};
				this.m_wsResetCommand = {command: "reset_screen_sharing"};
				
				//DOM video Object
				this.m_remoteVideo = undefined;
				
				this.m_mediaSource = undefined;
				this.m_currentMediaTrack = undefined;
				this.m_mediaRecorder = undefined;
				this.m_buffer = undefined;
				this.m_queue = [];
				
				var self = this;
				this.m_ws_url = this.m_baseUrl + '/' + this.m_clientType + '?channel=' + this.m_session_id;
				//Websocket object. If WebSocket cannot be opened a non catchable exception is thrown and callback will not be called
				this.m_ws = new WebSocket(this.m_ws_url);	
				this.m_ws.binaryType = "arraybuffer";

				var processCommand = function(buffer) {
					var commandStr = "";
					if("TextDecoder" in window) {
						var enc = new TextDecoder("utf-8");
						commandStr = enc.decode(buffer);
					}
					else {	//Manual decode
						for(var i = 0;i<buffer.length;i++){
							commandStr += String.fromCharCode.apply(null, buffer.subarray(i,i+1));
						}
					}
					try {
						var command = JSON.parse(commandStr);
						console.log("Command received: '" + command.command + "'");
						switch(command.command) {
							case 'pong':	//Nothing to do							
								break;
							case 'ready_to_receive':
								self.m_isReadyToReceive = true;
								if(self.m_isInitialized) {
									self.onReadyScreenSharing();
								}
								break;
							case 'reset_screen_sharing':
								self.StopSharing(false);
								break;
							default:
								console.error("Unknown command '" + command.command + "'");
						}
					} catch (e) {
						console.error("Parsing received command '" + commandStr + "' as json");
					}
				};		
				
				var wsPingIntervalFn = function() {
					if(self.m_isSharing) {
						clearInterval(self.m_wsPingIntervalObj);
					}
					else if(self.m_ws.readyState === 1) {
						self.m_ws.send(JSON.stringify(self.m_wsPingCommand));
					}
				};
				
				this.m_ws.onopen = () => {
					console.log(`Connected to ${this.m_ws_url}`);
					self.m_wsPingIntervalObj = setInterval(wsPingIntervalFn, self.m_wsPingIntervalMilliseconds);
				};
			
				// Listen for messages
				this.m_ws.onmessage = (event) => {
					//Detect if its a command or binary data
					try {
						var bufView = new Int8Array(event.data);
						//Check 11 first bytes looking for '{"command":'
						var length = bufView.length < 11 ? bufView.length : 11;
						if(length === 11) {
							var result = '';
							for(var i = 0;i<length;i++){
								result += String.fromCharCode.apply(null, bufView.subarray(i,i+1));
							}
							if(result === "{\"command\":") {
								processCommand(bufView);
								return;
							}
						}
					} catch (e) {}
					
					if (self.m_isReadyToReceive && event.data instanceof ArrayBuffer) {
						try {
							if (self.m_buffer.updating || self.m_queue.length > 0) {
								self.m_queue.push(event.data);
							} else {
								self.m_buffer.appendBuffer(event.data);
							}
						}  catch (e) {
							console.error(e);
						} 
					}
				};		
				
				//WebSocket was opened
				callback(true);
			}
			catch (e) {
				callback(false, e);
			} 
		}

		get IsReadyToReceive() {
			return !!this.m_isReadyToReceive;
		}	

		get IsSharing() {
			return !!this.m_isSharing;
		}
		
		Initialize() {
			var self = this;
			this.m_remoteVideo = document.createElement("video");
			this.m_remoteVideo.setAttribute("id", "webchat-remote-video");
			this.m_remoteVideo.setAttribute("class", "customer-video");
			this.m_remoteVideo.setAttribute("width", "100%");
			this.m_remoteVideo.setAttribute("height", "100%");
			//TODO: Acomodar cuando se tenga el layout
			//this.m_remoteVideo.setAttribute("style", "width: 400px; height: 250px; object-fit:fill;");
			this.m_videoContainer.appendChild(this.m_remoteVideo);
			
			this.m_mediaSource = new MediaSource();
			this.m_mediaSource.addEventListener('sourceopen', function(/*e*/) {
				var playPromise = self.m_remoteVideo.play();
				if (playPromise !== undefined) {
					playPromise.then(_ => {}) //playPromise is fulfilled
					.catch(error => {}); // Auto-play was prevented
				}
				self.m_buffer = self.m_mediaSource.addSourceBuffer(self.m_videoMimeType);

				self.m_buffer.addEventListener('update', function() { 
					if (self.m_queue.length > 0 && !self.m_buffer.updating) {
						self.m_buffer.appendBuffer(self.m_queue.shift());
					}
				});
			}, false);
				
			this.m_remoteVideo.src=window.URL.createObjectURL(this.m_mediaSource);
			this.m_remoteVideo.load(); // must call after setting/changing source 	

			this.m_isInitialized = true;
			if(this.m_isReadyToReceive) {
				this.onReadyScreenSharing();
			}
			
			var sendReady = function() {
				if(self.m_ws.readyState === 1) {
					self.m_ws.send(JSON.stringify(self.m_wsReadyCommand));
				}
				else {
					setTimeout(sendReady, 500);
				}
			};
			sendReady();			
		}

		StartSharing() {
			if(!this.m_isInitialized) {
				console.error("Object not initilized");
			}
			let self = this;
		
			var switchToShareScreen = function() {
				if (navigator.mediaDevices.getDisplayMedia) {	//Works on Chrome, Firefox, Edge and Opera
					var screenShareConstraints = { video: true, audio: false };
					navigator.mediaDevices.getDisplayMedia(screenShareConstraints).then(function(newStream) {
						self.m_currentMediaTrack = newStream.getVideoTracks()[0];
						self.m_currentMediaTrack.addEventListener("ended", function() {
							self.StopSharing();
						});
						var options = {
							mimeType : self.m_videoMimeType
						};
						self.m_mediaRecorder = new MediaRecorder(newStream, options);
						self.m_mediaRecorder.ondataavailable = function(e) {
							if(self.m_ws.readyState === 1) {
								self.m_ws.send(e.data);
							}
						};
						self.m_mediaRecorder.start(100);
						self.m_isSharing = true;
						self.onStartScreenSharing();
					}).catch(function (err) {
						console.error("Error on getUserMedia: " + err);
						return;		            
					});
				} 
				else {
					console.error("Error on getUserMedia. Not supported browser");
				}
			};
			
			switchToShareScreen();
		}

		StopSharing(hasToNotify = true) {
			if(this.m_mediaRecorder) {
				this.m_mediaRecorder.stop();
				this.m_mediaRecorder = undefined;
			}
			if(this.m_currentMediaTrack) {
				this.m_currentMediaTrack.stop();
				this.m_currentMediaTrack = undefined;
			}
			if(this.m_mediaSource && this.m_buffer && this.m_buffer instanceof SourceBuffer) {
				this.m_mediaSource.removeSourceBuffer(this.m_buffer);
			}
			this.m_buffer = undefined;
			this.m_queue = [];
			if(hasToNotify && this.m_ws && this.m_ws.readyState === 1) {
				this.m_ws.send(JSON.stringify(this.m_wsResetCommand));	
			}
			this.m_isInitialized = false;					
			this.m_isReadyToReceive = false;
			
			if(this.m_videoContainer && this.m_remoteVideo && this.m_remoteVideo.parentNode === this.m_videoContainer) {
				this.m_videoContainer.removeChild(this.m_remoteVideo);
			}

			if(this.m_ws && this.m_ws.readyState === 1) {
				this.m_ws.close();
			}
			this.onStopScreenSharing(hasToNotify);
		}
		
		Close() {
			this.StopSharing(false);
			this.onCloseScreenSharing();
		}
	}

	var format = inConcert.Utils.format;

	const VideoViewStates = {
		MINIMIZED  : "videoCallDocked",
		MAXIMIZED  : "videoCallFullHeight",
		FULL_SCREN : "videoCallFullScreen"
	};

	//represents a chat conversation
	function ChatSession(client, session) {
		this.m_session = session;
		this.m_client = client;
		this.m_participants = {};
		this.m_sessionId = session.id;
		this.m_owner = client.GetUserID();
		this.m_messages = [];
		this.m_fileTransfers = {};
	}

	/**
	* Retrieves all messages in the session.
	*/
	ChatSession.prototype.getMessages = function()
	{
		var orderedMessages = this.m_messages
			.filter(function(message) {
				return message.type === "text";
			})
			.map(function(message) {
				return message.content;
			});

		return orderedMessages;
	};

	//finishes chat session
	ChatSession.prototype.close = function() {
		if ( this.m_interaction && !this.m_interaction.canHangup() ) {
			return false;
		}
		return this.m_client.CloseSession(this.m_sessionId);
	};

	ChatSession.prototype.onUserJoined = function(user_id, user_name, user_role) {
		if (this.m_interaction) {
			if (user_role === "coach") {
				//this interaction is now coached
				this.m_interaction.coached_by = user_id;
			}
			var view = this.m_interaction.getView();
			if (view) {
				view.addUser(user_id, user_name, user_role);
			}
		}
	};

	ChatSession.prototype.onSessionUpdateProperties = function(values)
	{
		// we check if a user changed role within this session
		if (this.m_interaction && values.session_change_reason === "user_changed_role") {
			var view = this.m_interaction.getView();
			// monitor -> coach
			if (values.changed_old_user_role === "monitor" && values.changed_user_role === "coach") {
				this.m_interaction.coached_by = values.changed_user_id;
				if (view) {
					view.addUser(values.changed_user_id, values.changed_user_name, values.changed_user_role);
				}
			}
			// coach -> monitor
			else if(values.changed_old_user_role === "coach" && values.changed_user_role === "monitor") {
				if (view) {
					view.removeUser(values.changed_user_id, values.changed_user_display_name);
				}
				delete this.m_interaction.coached_by;
			}
		}
	};

	ChatSession.prototype.onClosed = function() {
		if (this.m_interaction) {			
			var view = this.m_interaction.getView();
			if (view) {
				view.onSessionClosed();
			}
			delete this.m_interaction.coached_by;
		}
	};

	ChatSession.prototype.onUserLeft = function(values)
	{
		if (!this.m_interaction) {
			return;
		}
		var user_id = values.user_id;
		var view = this.m_interaction.getView();
		if (view) {
			// when the web client leaves, use the pretty name instead of the ugly, autogenerated one.
			var user_name = values.left_user_type === "webclient" ? this.m_interaction.getClientName() : values.left_user_display_name;
			view.removeUser(user_id, user_name);
		}
		if (this.m_interaction.coached_by === user_id) {
			delete this.m_interaction.coached_by;
		}
	};

	//send a message to the current session
	ChatSession.prototype.sendMessage = function(message, callback = $.noop) {
		this.m_client.SendMessage(this.m_sessionId, message, callback);
	};

	//accepts the file transfer request
	ChatSession.prototype.acceptFileTransfer = function(transferId) {
		this.m_client.SendIndication(this.m_sessionId, "AcceptFileTransfer", {transferId: transferId});
	};

	//cancel the file transfer
	ChatSession.prototype.cancelFileTransfer = function(transferId) {
		var self = this;
		var event = self.m_fileTransfers [transferId];
		event.indication_type = "FileTransferFail";
		self.onEventReceived("fileTransfer", event);
		this.m_client.SendIndication(this.m_sessionId, "CancelFileTransfer", {transferId: transferId});
	};

	ChatSession.prototype.requestFileTransfer = function(fileId, fileName){
		this.m_client.SendIndication(this.m_sessionId, "RequestFileTransfer", {fileId: fileId, filename: fileName});
	};

	ChatSession.prototype.sendStartTyping = function() {
		this.m_client.SendIndication(this.m_sessionId, "start_typing", {}, false);
	};

	ChatSession.prototype.sendStopTyping = function() {
		this.m_client.SendIndication(this.m_sessionId, "stop_typing", {}, false);
	};


	ChatSession.prototype.getOwner = function() {
		return this.m_owner;
	};

	//triggered when a message is received for this session, just forward the message to the view to be displayed and store it locally in the message rolodex
	ChatSession.prototype.onMessageReceived = function(currentMessage)
	{
		var message = {
			user: currentMessage.user_id,
			name: currentMessage.display_name,
			text: currentMessage.text,
			msgtime: currentMessage.msgtime,
			originalMessage: currentMessage
		};
		//the message is from myself, tag it accordingly
		if(this.m_owner === message.user) {
			message.owned = true;
		}
		//send it to the view if it's there
		if (this.m_interaction) {
			if ((message.originalMessage.only_to &&
					message.originalMessage.only_to === this.m_interaction.coached_by) ||
					message.user === this.m_interaction.coached_by) {
				message.coach_session = true;
			}
			//only now update the message's rolodex
			this.onEventReceived("text", message);
			var view = this.m_interaction.getView();
			if (view) {
				view.addMessage(message);
			}
			else {
				inConcert.app.getMainWindow().open_chat_window(this.m_interaction);
			}
			this.notifyDesktopNewMessage(message);
			this.m_interaction.notifyNewMessage(message);
		}
		else {
			// Hack: Send the message to the rolodex
			this.onEventReceived("text", message);
		}
	};

	ChatSession.prototype.notifyDesktopNewMessage = function(message)
	{
		var desktopNotifications = inConcert.app.getDesktopNotifications();
		desktopNotifications.notify_chat_message(this.m_interaction, message);
	};

	//An event is the abstraction of a message, or a file transfer operation
	ChatSession.prototype.onEventReceived = function(type, content)
	{
		var event = {
			type: type,
			content: JSON.parse(JSON.stringify(content)) //esto crea una copia en profundidad
		};

		//queue the event locally
		this.m_messages.push(event);
	};

	//triggered when received a new indication from chat server
	ChatSession.prototype.onNewIndication = function(user, name, indication_type, values)
	{
		if (this.m_owner === user) {
			return; //si es un indication propio lo ignoro
		}

		if (!this.m_interaction){
			//no hay interaccion creada, en teoria nunca llegaria aqui (nunca digas nunca!!!)
			return;
		}

		var view = this.m_interaction.getView();
		if (view && (indication_type === "start_typing" || indication_type === "stop_typing") )
		{
			if (typeof view.processTypingEvent === "function") {
				view.processTypingEvent(indication_type, user, name, values);
			}
			return;
		}

		var event = this.m_fileTransfers [values.transferId];

		if (!event) {
			event = {
				indication_type: indication_type,
				values: values
			};
			this.m_fileTransfers [values.transferId] = event;
		}

		event.indication_type = indication_type;
		event.values = values;

		this.onEventReceived("fileTransfer", event);

		if (view) {
			view.processFileTransfer(event);
		}
	};

	ChatSession.prototype.getInteraction = function() {
		return this.m_interaction;
	};
	
	ChatSession.prototype.hasScreenRecording = function() {
		return this.m_interaction? this.m_interaction.hasScreenRecording() : false;
	};	

	//relates the session to a workitem/interaction
	ChatSession.prototype.setInteraction = function(interaction) {
		this.m_interaction = interaction;
	};

	ChatSession.prototype.CanEnableScreenSharing = function() {
		return this.m_client.CanEnableScreenSharing(this);
	};

	ChatSession.prototype.enableScreenSharing = function() {
		this.m_client.enableScreenSharing(this);
	};

	ChatSession.prototype.disableScreenSharing = function() {
		this.m_client.disableScreenSharing(this);
	};



	function ChatClient(manager, config, statusCallbackFn = $.noop) {
		const self = this;

		this.m_initialized = false;
		this.m_statusCallbackFn = statusCallbackFn;
		this.m_invitedUsers = {
			count: 0
		};
		this.m_sessions = {};//sessions are indexed by id
		this.m_manager = manager;
		this.m_userid = config.username + "@" + config.vcc;
		// Instantiates the Chat object
		this.m_chatClient = new InconcertChat({
			username               : config.username,
			vcc                    : config.vcc,
			displayUsername        : config.alias,
			automaticPollerDisabled: config.automaticPollerDisabled,
			RetrieveToken: function() {
				return inConcert.app.getLoggedUserToken();
			},
			onNewSession: function(session, values) {
				self.onNewSession(session, values);
				inConcert.logger.debug("Session properties updated id:" + session.id + " internal:" +
																(typeof(values.internal) === "string" && values.internal === "1") ? "1" : "0" +
																" origin: " + typeof(values.origin) === "string" ? values.origin : "");
				if (values.internal === "1") {
					var mSession = self.m_sessions[session.id];
					if (mSession && !mSession.getInteraction()) {
						if (values.origin !== inConcert.app.getCompleteId()) {
							//I'm being invited to a new internal chat, update local conversations lists
							self.m_invitedUsers[values.origin] = session.id;
							self.m_invitedUsers.count++;
						}
						self.m_manager.onInternalChat(mSession, values.origin, values.localID);
					}
				}
			},
			onCoordinates: function(session, values, isLast){
				self.onNewMessage(session, values, isLast, true);
			},
			onMessage: function(session, values, isLast) {
				self.onNewMessage(session, values, isLast, true);
			},
			onIndication: function(session, values/*, isLast*/) {
				self.onNewIndication(session, values);
			},
			onPreviousMessage: function(session, values, isLast) {
				self.onNewMessage(session, values, isLast, false);
			},
			onSessionClosed: function(session, values) {
				self.onSessionClosed(session, values);
				if(self.interactionScreenSharing) {
					self.interactionScreenSharing.Close();
					self.interactionScreenSharing = null;
					self.screenshareActiveInteraction = null;
					self.__hideVideoControls();
				}
			},
			onUserJoined: function(session, values) {
				self.onUserJoined(session, values);
			},
			onUserLeft: function(session, values) {
				self.onUserLeft(session, values);
			},
			onUrlPush: function(session, values) {
				self.onUrlPush(session, values);
			},
			onServerDisconnected: function() {
				return self.onServerDisconnected();
			},
			onServerReconnected: function() {
				self.onServerReconnected();
			},
			onSessionUpdateProperties: function(session, message) {
				self.onSessionUpdateProperties(session, message);
			}
		});
		this.m_url = inConcert.app.getChatProtocol() + "://" + inConcert.app.getChatServer() + "/inconcert/apps/chat/";
		this.m_chatClient.SetChatUrl(self.m_url);
		if (inConcert.app.isSessionRejoin()) {
			this.m_chatClient.LoginWithRejoin();
		}
		else {
			this.m_chatClient.Login();
		}

		this._bindScreenshareActions();
		this.m_initialized = true;
	}

	if (typeof inConcert.ChatClient !== 'function') {
		inConcert.ChatClient = ChatClient;
	}

	ChatClient.prototype.close = function() {
		if (this.m_chatClient) {
			this.m_chatClient.Logout();
		}
	};

	ChatClient.prototype.reconnect = function(callback) {
		var self = this;
		var on_client_reconnected = typeof(callback) === "function" ? callback : $.noop;
		if (this.m_chatClient) {
			// try to login the user
			this._status = "RECONNECTING";
			this.m_chatClient.Login(function(ok){
				self.m_initialized = ok;
				on_client_reconnected(ok);
			});
		}
		else {
			on_client_reconnected(false);
		}
	};

	ChatClient.prototype.activateAddress = function() {
		var self = this;
		inConcert.app.activateChannelAddress(
			inConcert.app.getUserAddressesNames(),
			function(activated) {
					if (activated) {
						if ( self._status === "RECONNECTING" ) {
						// solo se envia callback al main window, cuando se estuvo reconectado.
						// cuando se conecta x primera vez, no es necesario ya que las actualizaciones correspondientes llegan
						// en el main loop;
							self.m_statusCallbackFn("READY");
						}
						self._status = "READY";
					}
					else {
						self.m_statusCallbackFn(false, "<%=_T('Unable to activate address')%>");
					}
				}
		);
	};

	ChatClient.prototype.CloseSession = function(sessionId) {
		this.m_chatClient.CloseSession(sessionId);
		return true;
	};

	ChatClient.prototype.GetUserID = function() {
		return this.m_userid;
	};

	ChatClient.prototype.SendMessage = function(sessionid, message, callback = $.noop) {
		return this.m_chatClient.SendMessage(sessionid, message, callback);
	};

	ChatClient.prototype.SendMessageInternal = function(sessionid, message, add_to_qna, callback = $.noop) {
		return this.m_chatClient.SendMessageInternal(sessionid, message, add_to_qna, callback);
	};

	ChatClient.prototype.SendIndication = function(sessionid, indication_type, values, persist)
	{
		if(typeof persist === "undefined") {
			persist = true;
		}
		return this.m_chatClient.SendIndication(sessionid, indication_type, persist, values);
	};

	ChatClient.prototype.getSession = function(sessionId) {
		return this.m_sessions[sessionId];
	};

	/**
	Private:
	A new session was created and we're participating in it somehow.
	@param session The newly created session.
	@param values An object with the session data.
	*/
	ChatClient.prototype.onNewSession = function(session/*, values*/) {
		inConcert.logger.debug("New session received " + session.id);
		var newSession = new ChatSession(this, session);
		this.m_sessions[session.id] = newSession;
		this.m_manager.onNewSession("CHAT", newSession);
	};

	/**
	Private:
	A new message was received for a session we're participating into.
	@param session The session for which the message belongs to.
	@param values A bag with the properties of the message.
	@param isLast A boolean flag indicating if this message is the last of a batch.
	@param isNew A boolean flag indicating if the message is brand new or false when this message is from the "history" of the session.
	*/
	ChatClient.prototype.onNewMessage = function(session, values/*, isLast, isNew*/) {
		//just forward the message to the corresponding session
		var objSession = this.m_sessions[session.id];
		if (objSession) {
			objSession.onMessageReceived(values);
		}
		else {
			inConcert.logger.error(format("Message received for nonexistent session '{0}'", session.id));
		}
	};

	ChatClient.prototype.onNewIndication = function(session, values) {
		//just forward the message to the corresponding session
		var objSession = this.m_sessions[session.id];
		if (objSession) {
			objSession.onNewIndication(values.user_id, values.display_name, values.indication_type, values.values);
		}
		else {
			inConcert.logger.error(format("Message received for nonexistent session '{0}'", session.id));
		}
	};


	/**
	Private:
	A session we were participating into was closed.
	@param session The session.
	@param values An object with the session data.
	*/
	ChatClient.prototype.onSessionClosed = function(session/*, values*/) {
		var self = this;
		var objSession = self.m_sessions[session.id];
		if (objSession) {
			inConcert.logger.debug("Session closed " + session.id);
			//forward the message to the corresponding session
			objSession.onClosed();
			var interaction = objSession.getInteraction();
			if (interaction && interaction.isLocal()) {
				self.m_manager.onInternalChatEnded(objSession);
				//search the corresponding session and clean it up
				for (var user in self.m_invitedUsers) {
					if (self.m_invitedUsers.hasOwnProperty(user)) {
						if (self.m_invitedUsers[user] === session.id) {
							inConcert.logger.debug("Deleting from session " + session.id + " user " + user);
							delete self.m_invitedUsers[user];
							self.m_invitedUsers.count--;
							break;
						}
					}
				}
			}

			delete self.m_sessions[session.id];
		}
		else {
			inConcert.logger.error(format("Message received for nonexistent session '{0}'", session.id));
		}
	};

	/**
	Private:
	A user was added to the given session.
	@param session The session.
	@param values An object with the session data.
	*/
	ChatClient.prototype.onUserJoined = function(session, values) {
		//just forward the message to the corresponding session
		var objSession = this.m_sessions[session.id];
		if (objSession) {
			inConcert.logger.debug("User joined on session " + session.id + " id: " + values.user_id);
			objSession.onUserJoined(values.user_id, values.user_name, values.joined_user_role);
		}
		else {
			inConcert.logger.error(format("Message received for nonexistent session '{0}'", session.id));
		}
	};

	/**
	Private:
	A user was removed from the given session.
	@param session The session.
	@param values An object with the session data.
	*/
	ChatClient.prototype.onUserLeft = function(session, values) {
		//just forward the message to the corresponding session
		var objSession = this.m_sessions[session.id];
		if (objSession) {
			inConcert.logger.debug("User left session " + session.id + " id: " + values.user_id);
			objSession.onUserLeft(values);
			//if it's a local interaction and I'm the only one left just close the session
			var interaction = objSession.getInteraction();
			if (interaction && interaction.isLocal()) {
				var valid = false;
				for (var name in objSession.participants) {
					if (objSession.hasOwnProperty(name)) {
						if (name !== "undefined" && name !== inConcert.app.getCompleteId()) {
							valid = true;
							break;
						}
					}
				}
				if (!valid) {
					objSession.close();
				}
			}
		}
		else {
			inConcert.logger.error(format("Message received for nonexistent session '{0}'", session.id));
		}
	};

	/**
	Private:
	The session properties has been changed.
	@param session The session.
	@param values An object with the session data.
	*/
	ChatClient.prototype.onSessionUpdateProperties = function(session, values)
	{
		//just forward the message to the corresponding session
		var objSession = this.m_sessions[session.id];
		if (objSession) {
			inConcert.logger.debug("Session properties updated id:" + session.id + " internal:" +
				(session.properties.internal === "1" ? "1" : "0") + " origin: " +
				session.properties.origin ? session.properties.origin : ""
			);
			objSession.onSessionUpdateProperties(values);
		}
		else {
			inConcert.logger.error(format("Message received for nonexistent session '{0}'", session.id));
		}
	};

	/**
	Private:
	A user sent a url. This are treated in different way than text messages.
	@param session The session for which the message belongs to.
	@param values A bag with the properties of the message.
	*/
	ChatClient.prototype.onUrlPush = function(/*session, values*/) {};

	/**
	Private:
	The chat server connection has been dropped. Can't do anything else with the server until it reconnects (either automatically or manually)
	*/
	ChatClient.prototype.onServerDisconnected = function() {
		if (this.m_statusCallbackFn) {
			this.m_statusCallbackFn(false, "<%=_T('Chat server disconnected')%>");
		}
	};

	/**
	Private:
	The chat server connection has been re-established
	*/
	ChatClient.prototype.onServerReconnected = function() {
		this.activateAddress();
	};

	//Finds best session for interaction, chat sessions are related to workitems using the sessionid
	ChatClient.prototype.findInteractionSession = function(interaction) {
		if (this.m_sessions[interaction.id]) {
			return this.m_sessions[interaction.id];
		}
		else {
			return null;
		}
	};

	ChatClient.prototype.getUserConversation = function(userId) {
		var fullUserId = userId + "@" + this.m_chatClient.GetVcc();
		var sessionId = this.m_invitedUsers[fullUserId];
		if (sessionId) {
			inConcert.logger.debug("GetUserConversation for " + userId + " found session " + sessionId);
		}
		return sessionId;
	};

	ChatClient.prototype.monitorSession = function(sessionId, callback)
	{
		this.m_chatClient.MonitorSession(sessionId, callback);
	};

	ChatClient.prototype.coachSession = function(sessionId, coachedUser, callback)
	{
		this.m_chatClient.CoachUser(coachedUser, sessionId, callback);
	};

	ChatClient.prototype.leaveSession = function(sessionId, callback) {
		this.m_chatClient.RemoveUserFromSession(sessionId, callback);
	};

	ChatClient.prototype.inviteUser = function(userId, interactionID, callbacks) {
		// look up userId in a list of established sessions
		var self = this;
		var options = {
			internal : "1",
			origin : inConcert.app.getCompleteId(),
			localID : interactionID
		};
		userId = userId.toLowerCase();
		var fullUserId = userId + "@" + this.m_chatClient.GetVcc();
		inConcert.logger.debug("Inviting user " + fullUserId + " to chat");
		if(!self.m_invitedUsers[fullUserId]) {
			self.m_chatClient.InviteUser(userId, function(succeeded, message_or_sessionID) {
				// if succeeded, message_or_sessionID is the new sessionId
				inConcert.logger.debug("User invitation for " + fullUserId + " returned " + succeeded + " and value " + message_or_sessionID);
				if(succeeded) {
					self.m_invitedUsers[fullUserId] = message_or_sessionID;
					self.m_invitedUsers.count++;
				}
				if(typeof(callbacks) === "object") {
					if(succeeded) {
						if(typeof(callbacks.success) === "function") {
							callbacks.success(message_or_sessionID);
						}
					}
					else {
						if(typeof(callbacks.error) === "function") {
							callbacks.error(message_or_sessionID);
						}
					}
				}
			}, options);
		}
		else {
			//return existing session id to chat with this user
			return self.m_invitedUsers[fullUserId];
		}
	};

	//returns whether the client is ready to operate or not.
	ChatClient.prototype.isInitialized = function() {
		return this.m_initialized;
	};

	//returns whether a particular interaction is related to a session
	ChatClient.prototype.associationValid = function(interaction, session) {
		return interaction.id === session.m_sessionId;
	};
	
	ChatClient.prototype.firePoller = function(data = {}, callback = $.noop) {
		this.m_chatClient.firePoller(data, callback);
	};
	
	ChatClient.prototype.isAutomaticPollerDisabled = function() {
		return this.m_chatClient.isAutomaticPollerDisabled();
	};
	
	ChatClient.prototype.enableAutomaticPoller = function() {
		this.m_chatClient.enableAutomaticPoller();
	};
	
	ChatClient.prototype.disableAutomaticPoller = function() {
		this.m_chatClient.disableAutomaticPoller();
	};

	ChatClient.prototype.switchTo = function(interaction) {
		Object.values(this.m_sessions).forEach(function(ses) { 
			if(ses.m_interaction.id === interaction.id) {
				ses.m_interaction.getView().show();
			}
		});
	};

	ChatClient.prototype._bindScreenshareActions = function() {
		var self = this;
		$("#screenShareStartButton").click(function() {
			if(self.interactionScreenSharing) {
				self.interactionScreenSharing.StartSharing();
			}
		});

		$("#screenShareStopButton").click(function() {
			if(self.interactionScreenSharing) {			
				self.interactionScreenSharing.StopSharing();
			}
		});

		$("#webchat-video-maximizeButton").click(function() {
			self._setVideocallViewState(VideoViewStates.MAXIMIZED);
		});

		$("#webchat-video-minimizeButton").click(function() {
			self._setVideocallViewState(VideoViewStates.MINIMIZED);
		});

		$("#webchat-video-presentButton").click(function() {
			self._setVideocallViewState(VideoViewStates.FULL_SCREN);
		});

		$("#webchat-video-goInteractionButton").click(function() {
			self.switchTo(self.screenshareActiveInteraction);
			self._setVideocallViewState(VideoViewStates.MINIMIZED);
		});
	};

	ChatClient.prototype.CanEnableScreenSharing = function(/*session*/) {
		return this.interactionScreenSharing === undefined || this.interactionScreenSharing === null;
	};

	ChatClient.prototype.disableScreenSharing = function(/*session*/) {
		if(this.interactionScreenSharing) {			
			this.interactionScreenSharing.StopSharing();
		}
	};

	ChatClient.prototype.enableScreenSharing = function(session) {
		var self = this;
		if(!self.interactionScreenSharing) {
			var screenSharingOptions = {
				baseUrl					: inConcert.app.getAgentScreenServiceBaseDomain(),
				clientType				: "watch",
				session_id				: session.m_sessionId,
				videoContainerId		: "webchat-remote-video-container",
				onReadyScreenSharing	: function() {
					$("#screenShareStartButton").show();
					$("#screenShareStopButton").hide();
					self._setVideocallTitle("<%=_T('Screen share - Ready to share')%>");
				},
				onStartScreenSharing	: function() {
					$("#screenShareStartButton").hide();
					$("#screenShareStopButton").show();
					session.sendMessage("<%=_T('I start to share screen')%>");
					self._setVideocallTitle("<%=_T('Screen share - Sharing')%>");
				},
				onStopScreenSharing		: function(sendMessage = false) {
					Object.values(self.m_sessions).forEach(function(ses) {
						ses.m_interaction.getView().toggleScreenShare(true, session.m_interaction.id);
						if(ses.m_interaction.id === self.screenshareActiveInteraction.id && sendMessage) {
							ses.sendMessage("<%=_T('I am done with screen sharing')%>");
						}
					});
					self.interactionScreenSharing = null;
					self.screenshareActiveInteraction = null;
					$("#screenShareStartButton").hide();
					$("#screenShareStopButton").hide();
					self._setVideocallTitle("Screen share");
					self.__hideVideoControls();
				},
				onCloseScreenSharing	: function() {
					$("#screenShareStartButton").hide();
					$("#screenShareStopButton").hide();
					self.interactionScreenSharing = null;
					self.__hideVideoControls();
				},
			};		
			self.interactionScreenSharing = new InteractionScreenSharing(screenSharingOptions, function(success/*, error*/) {
				if(success) {
					//Dont show screen share controls until interaction is set
					var waitInteraction = function() {
						if(session.m_interaction) {
							self._setVideocallViewState(VideoViewStates.MINIMIZED);
							self.__showVideoControls();
							self.interactionScreenSharing.Initialize();
							self.screenshareActiveInteraction = session.m_interaction;
							session.sendMessage("<%=_T('I enable screen sharing')%>");
							self._setVideocallTitle("<%=_T('Screen share - Waiting for client ready')%>");

							Object.values(self.m_sessions).forEach(function(ses) {
								ses.m_interaction.getView().toggleScreenShare(false, session.m_interaction.id);
							});
						}
						else {
							setTimeout(waitInteraction, 500);
						}
					};
					setTimeout(waitInteraction, 500);
				}
			});
		}
	};

	ChatClient.prototype.$videocall = function(query) {
		const $videocall = $("#webchat-video-container");
		return query? $videocall.find(query) : $videocall;
	};

	ChatClient.prototype.$remoteVideo = function() {
		return this.$videocall("#webchat-remote-video");
	};

	ChatClient.prototype._setVideocallAction = function(className, enabled) {
		this.$videocall(`.${className}`).toggle(!!enabled);
	};

	ChatClient.prototype._setVideocallActions = function() {
		this._setVideocallAction("maximizeButton",
			this._videoViewState !== VideoViewStates.MAXIMIZED);
		this._setVideocallAction("minimizeButton",
			this._videoViewState !== VideoViewStates.MINIMIZED);
		this._setVideocallAction("presentButton",
			this._videoViewState !== VideoViewStates.FULL_SCREN);
			
		this._setVideocallAction("goInteractionButton", this.interactionScreenSharing);		
	};
	
	ChatClient.prototype._setVideocallVideo = function(enabled, videoClassName, avatarClassName) {
		this.$videocall(`.${videoClassName}`).toggle(enabled);
		this.$videocall(`.${avatarClassName}`).toggle(!enabled);
	};

	ChatClient.prototype.__showVideoControls = function() {
		this.$videocall().show();
		this._setVideocallActions();
	};
	
	ChatClient.prototype.__hideVideoControls = function() {
		this.$videocall().hide();
	};

	ChatClient.prototype._setVideocallViewState = function(newState) {
		this._videoViewState = newState;
		const $container = this.$videocall();
		Object.keys(VideoViewStates).forEach(stateKey => {
			const state = VideoViewStates[stateKey];
			$container.toggleClass(state, state === newState);
		});
		this._setVideocallActions();
	};

	ChatClient.prototype._setVideocallTitle = function(title) {
		$("#webchat-video-container-header-text").text(title);
	};
}());
