/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:false, undef:true, unused:vars, curly:false, browser:true, indent:false, maxerr:50,
  quotmark: false, -W014, esversion: 11
*/

/* global inConcert, CInconcertPlayer, html_sanitize, console, Mustache, _, EmojiPicker */

/*
	Handling GUI events to interact with chat client session
*/

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
	this.inConcert = {};
}

/*

Herencia de clases de la ventana de chat

ClientWindow                           (*) NO se instancia
  |
  |_ChatClientWindow                   (*) NO se instancia
     |
     |_ArchivedClientWindow            (*) cuando se abre desde las archivadas.
     |
     |_WrapupClientWindow              (*) por la app "agent" cuando realiza wrapup
     |
     |_LiveClientWindow                (*) NO se instancia
        |
        |_LiveInternalChatClientWindow (*) las interacciones entre agentes 'INTERNAL_CHAT' (sin proxy)
        |
        |_LiveIMClientWindow           (*) por el proxy 'IM'
        |
        |_LiveWebChatClientWindow      (*) por el proxy 'CHAT'
        |
        |_MonitorClientWindow          (*) por la app "supervisor" cuando monitorea
        |
        |_QueuedMonitorClientWindow    (*) por las apps "agent" y "supervisor" cuando abre la interaccion de la bandeja de encoladas.
        |
        |_WaitingMonitorClientWindow   (*) por las apps "agent" y "supervisor" cuando abre la interaccion de la bandeja de waiting.
*/

(function () {

	var _T = inConcert.i18n.getString;
	var format = inConcert.Utils.format;

	// Mustache templates
	//
	var templates = {
		coach_dialog_tpl : ''.concat(
			'<div id="coachingDialogContainer" class="coachingChatContainer">',
				'<div id="coachingHeaderSection" class="coachingChatWith">',
					'<div class="coachingChatWithAvatar">',
						'<img src="{{ avatar_uri }}"/>',
					'</div>',
					'<div class="coachingWithInfo">',
						'<div class="coachingChatWithAux">', "<%=_T('Coach')%>", '</div>',
						'<div class="coachingChatWithName">{{ name }}</div>',
					'</div>',
				'</div>',
				'<div class="coachingChatReadControl">',
					'<div class="coachingChatRead"></div>',
				'</div>',
				'<div class="coachingChatWriteControl">',
					'<div class="coachingChatWrite"><textarea id="coachMessageText"></textarea></div>',
					'<button id="coachSendMessage" class="coachingChatSendButton">',
						'<i class="far fa-paper-plane"></i>',
					'</button>',
				'</div>',
			'</div>'
		)
	};

	/**
	A finite state machine that describes the transitions between typing states. It is used to send notifications
	 when a user starts or stops typing.

	message_sent
	  ---
	  |  |
	  |  v
	--------                ---------- <-.
	|      |                |        |---| key
	| idle | ------key----->| typing |           -----------
	|      |                |        |---------->| stopped |
	--------                ----------    stop   -----------
		^                          |
		|-- message_sent, timeout--|
				  spellcheck
		@param context
	*/
	function TypingFSM (context)
	{
		this.m_current_state      = 'idle';
		this.m_context            = context;
		this.m_typing_timeout     = undefined;	// a timer function that will be eventually set
		this.m_spellcheck_timeout = undefined;
	}

	TypingFSM.prototype.__makeTimeout = function (event_name, timeout) {
		var self = this;
		event_name = event_name || 'timeout';
		timeout = timeout || 10000;
		return setTimeout(function() {
			self.event(event_name);
		}, timeout);
	};

	TypingFSM.prototype.event = function (evt)
	{
		if(this.m_current_state === 'idle') {
			if(evt === 'key') {
				this.m_current_state = 'typing';
				this.m_context.SendStartTyping();
				this.m_typing_timeout = this.__makeTimeout();

				this.m_context.UpdateSpellcheckUI();

				if(this.m_context.HasToSpellcheck()) {
					clearTimeout(this.m_spellcheck_timeout);
					this.m_spellcheck_timeout = this.__makeTimeout('spellcheck', 2000);
				}
			}
			else if(evt === 'spellcheck') {
				this.m_context.Spellcheck();
				delete this.m_spellcheck_timeout;
			}
			else if(evt === 'message_sent') {
				this.m_context.UpdateSpellcheckUI();
			}
		}
		else if(this.m_current_state === 'typing') {
			switch(evt) {
				case 'key':
					clearTimeout(this.m_typing_timeout);
					this.m_typing_timeout = this.__makeTimeout();

					this.m_context.UpdateSpellcheckUI();

					if(this.m_context.HasToSpellcheck() && typeof(this.m_spellcheck_timeout) === 'undefined') {
						clearTimeout(this.m_spellcheck_timeout);
						this.m_spellcheck_timeout = this.__makeTimeout('spellcheck', 2000);
					}
				break;

				case 'timeout':
					this.m_current_state = 'idle';
					this.m_context.SendStopTyping();
					delete this.m_typing_timeout;
				break;

				case 'message_sent':
					this.m_current_state = 'idle';
					this.m_context.SendStopTyping();
					clearTimeout(this.m_typing_timeout);
					delete this.m_typing_timeout;
					clearTimeout(this.m_spellcheck_timeout);
					delete this.m_spellcheck_timeout;
					this.m_context.UpdateSpellcheckUI();
				break;

				case 'stop':
					this.m_current_state = 'stopped';
					this.m_context.SendStopTyping();
					clearTimeout(this.m_typing_timeout);
					delete this.m_typing_timeout;
					clearTimeout(this.m_spellcheck_timeout);
					delete this.m_spellcheck_timeout;
				break;

				case 'spellcheck':
					this.m_context.Spellcheck();
					delete this.m_spellcheck_timeout;
				break;

				default:
					// You most probably fired an event that does not exist, or you made a typo.
					throw 'Unhandled event "' + evt + '"';
			}
		}
	};

	// Estas dimensiones deben de coincidir con lo que se ha definido en "base_panel.js" funcion "openInteractionClient"
	const DEFAULT_WINDOW_WIDTH_PX = 950;

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Chat Client Handler
	//
	function ChatClientWindow(interaction, options)
	{
		if (!interaction) {
			return;
		}
		// TODO: la inicializacion de interaction tiene que estar en la clase padre (ClientWindow)
		//relate the interaction to the view
		this.m_interaction = interaction;

		// Must call constructor after having set m_interaction. The base class depends on that.
		ChatClientWindow.uber.constructor.call(this, options);

		this.lastName = '';
		this.lastDirection = '';
		this.setHorizontalResizingLimits(DEFAULT_WINDOW_WIDTH_PX);
		this.updateContactDisplay();
	}

	inConcert.Utils.inherit(ChatClientWindow, inConcert.ClientWindow);

	ChatClientWindow.prototype.getWindowDefaultTitle = function()
	{
		return "<%=_T('Chat Window')%>";
	};

	ChatClientWindow.prototype.toggleScreenShare = function(showEnable, showDisableOnInteractionId) {
		const vShowEnable = showEnable && this.m_interaction.state === "TAKEN";
		const vShowDisable = !vShowEnable && this.m_interaction.id === showDisableOnInteractionId;
		$("#enableScreenShare").toggle(vShowEnable);
		$("#disableScreenShare").toggle(vShowDisable);
		$("#enableScreenShareContainer").toggle(showEnable || vShowDisable);
	};

/*
-----------
Send Message
-----------

TODO: (REFACTORING) Revisar todas estas funciones referentes a "send message", las mismas se repiten con el mismo codigo en los clientes "webchat", "whatsapp", "sms", "rcs".

Seria lo mejor moverlas a la clase "clientWindow" asi no repetimos codigo. Aunque la clase client window, tambien la utiliza "mail" el "phone" etc, y estas funciones no son utiles ahi, habria que reviarlo.

Quizas lo mejor es crear un clase a parte, algo asi tipo "clientSendMessage" y quien la necesite que la instancie. VER!!!

*/
	ChatClientWindow.prototype.__updateCharsCounter = _.throttle(function() {
		var currentText = this.$sendMessageText().val();
		var currentChars = currentText && currentText.length || 0;
		this.enableSendMessageButton(currentChars > 0 &&
			!this.m_interaction.isWrappedUp() && this.m_interaction.state !== 'QUEUED');
	}, 100);

	ChatClientWindow.prototype.$sendMessageContainer = function() {
		return $("#sendMessageContainer");
	};

	ChatClientWindow.prototype.toggleSendMessage = function(show) {
		this.$sendMessageContainer().toggle(!!show);
	};

	ChatClientWindow.prototype.$sendMessageText = function() {
		return $("#chatMessageText");
	};

	ChatClientWindow.prototype.$sendMessageEmojiPicker = function() {
		return $("#chatMessageEmojiPicker");
	};

	ChatClientWindow.prototype.startEmojiPicker = function() {
		let self = this;
		let $text = self.$sendMessageText();
		let $picker = self.$sendMessageEmojiPicker();

		self._emojiPicker = new EmojiPicker({
			assetsPath : "/inconcert/lib/emoji-picker/lib/img/"
		});

		$text.emojiarea({
			emojiPopup: self._emojiPicker,
			wysiwyg : false,
			$button: $picker
		});
	};

	ChatClientWindow.prototype.clearSendMessageText = function() {
		this.$sendMessageText().val("");
	};

	ChatClientWindow.prototype.focusSendMessageText = function() {
		/**
		 * Los seteo de focus a los controles, se deben de realizar en un pequeno timeout, ya que sino el plugin "GoldenLayout" se despatarra, si es que aun no finalizo su render.
		 * Ref issue: http://redmine.inconcert/issues/8534
		*/
		const self = this;
		if (this.m_txt_msg_timeout_hnd) {
			window.clearTimeout(this.m_txt_msg_timeout_hnd);
			this.m_txt_msg_timeout_hnd = null;
		}
		this.m_txt_msg_timeout_hnd = window.setTimeout(function() {
			self.m_txt_msg_timeout_hnd = null;
			self.$sendMessageText().focus();
		}, 800); // quizas haya que aumentar el timeout!!
	};

	ChatClientWindow.prototype.enableSendMessageText = function(enabled) {
		if (this.textDisabled) {
			enabled = false;
		}
		this.$sendMessageText().prop("disabled", !enabled);
		this.$sendMessageEmojiPicker()
			.prop("disabled", !enabled)
			.toggleClass("emoji-picker-button-disabled", !enabled);
	};

	ChatClientWindow.prototype.appendToSendMessageText = function(text, afterCursor) {
		let $input = this.$sendMessageText();
		if (!afterCursor) {
			$input.val($input.val() + " " + text);
		}
		this.__updateCharsCounter();
	};

	ChatClientWindow.prototype.bindTextChange = function(onChange = $.noop, onEnter = $.noop) {
		let self = this;
		self.$sendMessageText()
			.bind("paste", null, function() {
				window.setTimeout(function() {
					onChange();
				}, 10);
			})
			.keyup(function(event) {
				if (event.which === 13) {
					return false;
				}
				onChange();
			})
			.change(function(event) {
				if (event.which === 13) {
					return false;
				}
				onChange();
			})
			.keypress(function(event) {
				if (event.which === 13) {
					window.setTimeout(function() {
						onEnter();
					}, 10);
				}
			});
	};

	ChatClientWindow.prototype.getSendMessageText = function() {
		return this.$sendMessageText().val();
	};

	ChatClientWindow.prototype.setSendMessageText = function(text) {
		this.$sendMessageText()
			.val(text || "")
			.keyup();
	};

	ChatClientWindow.prototype.$sendMessageButton = function() {
		return $("#chatButtonSend");
	};

	ChatClientWindow.prototype.enableSendMessageButton = function(enabled) {
		if (this.textDisabled) {
			enabled = false;
		}
		this.$sendMessageButton().prop("disabled", !enabled);
	};

	//triggered to update the view, timers, data and whatsoever
	ChatClientWindow.prototype.onTimer = function() {
		if (this.m_interaction) {
			var time = this.m_interaction.currentTime();
			$('#elapsedTime').text(time);
		}
	};

	ChatClientWindow.prototype.onInteractionGone = function(interaction) {
		ChatClientWindow.uber.onInteractionGone.call( this, interaction ); //invokes the base class method
		this.toggleSendMessage(false);
		$('#internalMessageInfo').hide();
		this.m_interaction.setView(null);
	};

	ChatClientWindow.prototype.showInternalMessage = function(message, timeout, progress)
	{
		$(".chatClientRead").scrollTop(0);
		var $progressContainer = $("#progressBar");
		var $progress = $progressContainer.children("div");

		if (progress) {
			$progress.html(progress);
			$progress.css("width", progress);
			if (progress === "0%") {
				$progressContainer.show();
			}
		}
		else {
			$progressContainer.hide();
		}
		$('#internalMessageInfo')
			.show()
			.find('.internalMessageInfoText')
			.text(message);
		if (timeout) {
			setTimeout(function() {
				$('#internalMessageInfo').hide();
			}, timeout);
		}
	};

	ChatClientWindow.prototype.addUser = function(undefined, user_name)
	{
		this.showInternalMessage(format("<%=_T('{0} has joined the conversation')%>", user_name), 5000);
	};

	ChatClientWindow.prototype.removeUser = function(undefined, user_name)
	{
		this.showInternalMessage(format("<%=_T('{0} has left the conversation')%>", user_name), 5000);
	};

	ChatClientWindow.prototype.__buildButton = function(button) {
		var $button = $("".concat(
			"<button class='msgButton'>",
				"<label class='msgButtonLabel'>",
					button.title,
				"</label>",
			"</button>"
		));

		switch(button.type) {
			case "web_url":
				$button
					.data("url", button.url)
					.click(function() {
						window.open($(this).data("url"), "_blank");
					});
				break;
		}

		return $button;
	};

	ChatClientWindow.prototype.__buildButtons = function(buttons) {
		var self = this;
		var $container = $("<div class='msgButtonsContainer'></div>");
		$.each(buttons, function(_, button) {
			$container.append(self.__buildButton(button));
		});
		return $container;
	};

	ChatClientWindow.prototype.__buildQuickreply = function(quickreply) {
		var $quickreply = $("".concat(
			"<button class='msgButton'>",
				"<label class='msgButtonLabel'>",
					quickreply.title,
				"</label>",
			"</button>"
		));
		return $quickreply;
	};

	ChatClientWindow.prototype.__buildQuickreplies = function(quickreplies) {
		var self = this;
		var $container = $("<div class='msgButtonsContainer'></div>");
		$.each(quickreplies, function(_, quickreply) {
			$container.append(self.__buildQuickreply(quickreply));
		});
		return $container;
	};

	ChatClientWindow.prototype.__buildGalleryItem = function(item) {
		var self = this;
		var $container = $("<div class='galleryItem'></div>");
		if (item.image_url) {
			$container.append("".concat(
				"<div class='galleryItemImageContainer'>",
					"<img src='" + item.image_url + "' class='galleryItemImage'>",
				"</div>"
			));
		}
		if (item.title) {
			$container.append("<label class='galleryItemTitle'>" + item.title + "</label>");
		}
		if (item.subtitle) {
			$container.append("<label class='galleryItemSubtitle'>" + item.subtitle + "</label>");
		}
		if (item.link) {
			$container.append("<a target='_blank' href='" + item.link_url + "' class='galleryItemLink'>" +
				item.link_url + "</a>");
			$container.find(".galleryItemImage")
				.addClass("withLink")
				.data("url", item.link_url)
				.click(function() {
					window.open($(this).data("url"), "_blank");
				});
		}
		if (item.buttons && item.buttons.length) {
			$container.append(self.__buildButtons(item.buttons));
		}
		return $container;
	};

	ChatClientWindow.prototype.__buildGalleryMessage = function(message) {
		var self = this;
		var $container = $("<div class='chatMessageGalleryContainer'></div>");
		$container.addClass(message.image_aspect_ratio + "Images");
		$.each(message.options, function(_, item) {
			$container.append(self.__buildGalleryItem(item));
		});

		return $container;
	};

	ChatClientWindow.prototype.__buildListItem = function(item) {
		var self = this;
		var $container = $("<div class='listItem'></div>");
		var $textContainer = $("<div class='listItemTextContainer'></div>");
		$container.append($textContainer);
		if (item.image_url) {
			$container.append("".concat(
				"<div class='listItemImageContainer'>",
					"<img src='" + item.image_url + "' class='listItemImage'>",
				"</div>"
			));
		}
		if (item.title) {
			$textContainer.append("<label class='listItemTitle'>" + item.title + "</label>");
		}
		if (item.subtitle) {
			$textContainer.append("<label class='listItemSubtitle'>" + item.subtitle + "</label>");
		}
		if (item.link) {
			$textContainer.append("<a target='_blank' href='" + item.link_url + "' class='listItemLink'>" +
				item.link_url + "</a>");
			$container.find(".listItemImage")
				.addClass("withLink")
				.data("url", item.link_url)
				.click(function() {
					window.open($(this).data("url"), "_blank");
				});
		}
		if (item.buttons && item.buttons.length) {
			$container.append(self.__buildButtons(item.buttons));
		}
		return $container;
	};

	ChatClientWindow.prototype.__buildListMessage = function(message) {
		var self = this;
		var $container = $("<div class='chatMessageListContainer'></div>");
		$container.addClass(message.top_element_style + "TopElementStyle");
		$.each(message.options, function(_, item) {
			$container.append(self.__buildListItem(item));
		});
		if (message.buttons) {
			$container.append(self.__buildButtons(message.buttons));
		}
		return $container;
	};

	ChatClientWindow.prototype.__buildImageMessage = function(_, url)
	{
		var $message = $("".concat(
			"<div class='chatMessageImageContainer'>",
				"<img src='", url, "' class='chatMessageImage'>",
			"</div>"
		));
		return $message;
	};

	ChatClientWindow.prototype.__getAttachmentMessageUrl = function(message) {
		switch(message.attachmentType) {
			case "repository":
				let link = inConcert.app.getBaseDomain() + "/inconcert/apps/repository/download/?id=" +
					encodeURIComponent(message.fileId) + "&section=-1&part=-1";
				if (message.fileName) {
					link += "&name=" + encodeURIComponent(message.fileName);
				}
				return link;
			case "url":
				return message.url;
		}
		return "";
	};

	ChatClientWindow.prototype.__buildGeolocMessage = function(message)
	{
		let coords = message.coordinates || [];

		if (coords.length === 0 && message.coords_var && typeof message.coords_var === "string" && message.coords_var !=="") {
			const split = message.coords_var.split(',');
			if (split.length >= 2) {
				coords = [{ lat: split[0], lon: split[1]}];
			}
		}
		const style = {
			"version": 8,
			"sources": {
				"osm": {
					"type": "raster",
					"tiles": [ inConcert.app.getBaseDomain() + "/maps/{z}/{x}/{y}.png"],
					"tileSize": 256,
					"attribution": "&copy; OpenStreetMap Contributors",
					"maxzoom": 19
				}
			},
			"layers": [ { "id": "osm", "type": "raster", "source": "osm" } ]
		};
		var map_id = "map_" + Date.now();
		var $messageNode =  $(`<div id=${map_id} class='chatMessageText' style='width: 200px; height:200px'></div>` +
		`<script type='text/javascript'>` +
			`var map = new maplibregl.Map({` +
				`container: ${map_id},` +
				`style: ${JSON.stringify(style)},` +
				`center: [${coords[0].lon}, ${coords[0].lat}],` +
				`zoom: 3` +
			`});` +
			`var marker = new maplibregl.Marker()`+
				`.setLngLat([${coords[0].lon}, ${coords[0].lat}])`+
				`.addTo(map);`+
		`</script>`
		);
		return $messageNode;
	};

	ChatClientWindow.prototype.__buildAttachmentMessage = function(message)
	{
		var self = this;
		var url = self.__getAttachmentMessageUrl(message);

		switch(message.fileType) {
			case "image":
				return self.__buildImageMessage(message, url);
		}

		var name = message.fileName || url;
		var $messageNode = $('<label class="chatMessageText"></label>');
		$messageNode.html(format(
				'<div class="filesharingContainer">' +
					'<i class="far fa-paperclip"></i>' +
					'<label>{1}</label>' +
					'<a href="{0}" title="{1}" target="_blank" class="filesharingLink"><i class="far fa-download"></i></a>' +
				'</div>' +
				'<span class="filesharingMessage filesharing{2}"><i class="far fa-check-double"></i></span>',
			url,
			name,
			message.owned? 'Received' : 'Sent',
			message.owned? "<%=_T('File received')%>" : "<%=_T('File sent')%>"
		));
		return $messageNode;
	};

	ChatClientWindow.prototype.__buildTextMessage = function(message)
	{
		const self = this;
		let source = this.__getMessageSource(message);

		function buildTextMessage()
		{
			//try to parse xml/html nodes and leave only text if present
			let parsedText = message.text;
			if (MESSAGE_SOURCE.AGENT !== source && !(message.hyperlinks && message.hyperlinks.length)) {
				// Cuando el origen del mensaje NO es el AGENT, el mensaje se sanitiza.
				parsedText = inConcert.Utils.nl2br(html_sanitize(parsedText));
			}
			let linkifiedText = inConcert.Utils.linkify(parsedText, {exclude: {email: true}});
			return linkifiedText;
		}

		function buildFilesharingMessage()
		{
			let textParts = message.text.replace("filesharing:", "").split("|");
			//never happend
			if (textParts.length !== 2) {
				return "";
			}
			let namePart = textParts[0];
			let hrefPart = textParts[1];
			let html = format(
					'<a href="{0}" title="{1}" target="_blank" class="filesharingLink">{1}</a>' +
					'<span class="filesharingMessage filesharing{2}">{3}</span>',
				hrefPart,
				namePart,
				message.owned? 'Sent' : 'Received',
				message.owned? "<%=_T('File sent')%>" : "<%=_T('File received')%>"
			);
			return html;
		}

		const $messageNode = $('<label class="chatMessageText"></label>');
		//DEPRECATED SINCE 2019-04-23
		const contentType = message.text.split(":")[0];
		switch(contentType) {
			case "filesharing":
				$messageNode.html(buildFilesharingMessage());
				break;
			default:
				let message = buildTextMessage();
				$messageNode.html(message);
				if (MESSAGE_SOURCE.AGENT !== source) {
					$messageNode
						.off('dblclick')
						.on('dblclick', function() {
							self.setMessageInSideBar(message);
						}
					);
				}
		}
		return $messageNode;
	};

	ChatClientWindow.prototype.__buildMessageContent = function(message)
	{
		switch(message.msgtype_content) {
			case "attachment":
				return this.__buildAttachmentMessage(message);
			case "gallery":
				return this.__buildGalleryMessage(message);
			case "list":
				return this.__buildListMessage(message);
			case "geo":
				return this.__buildGeolocMessage(message);
			default:
				return [
					this.__buildTextMessage(message),
					message.buttons? this.__buildButtons(message.buttons) : null,
					message.quickreplies? this.__buildQuickreplies(message.quickreplies) : null
				];
		}
	};

	const MESSAGE_SOURCE = {
		BOT   : "bot",
		AGENT : "agent",
		CLIENT: "client"
	};

	ChatClientWindow.prototype.__getMessageSource = function(message) {
		if (message.is_bot) {
			return MESSAGE_SOURCE.BOT;
		}
		//return message.user_id && (message.user_id.match(/-/g) || []).length > 3 && MESSAGE_SOURCE.CLIENT || MESSAGE_SOURCE.AGENT;
		return message.user_id && message.user_id === inConcert.app.getLoggedUserFullId() && MESSAGE_SOURCE.AGENT || MESSAGE_SOURCE.CLIENT;
	};

	//creates an HTML node from a chat message
	ChatClientWindow.prototype.buildMessageNode = function(message, author, customClass)
	{
		var contentMessage = message.originalMessage || message;
		var source = this.__getMessageSource(contentMessage);

		//make timestamp
		var $timestamp = null;
		if (typeof(message.msgtime) !== 'undefined') {
			if (message.originalMessage && message.originalMessage.msgTimeStamp && message.originalMessage.msgtype === 'previous_text') {
				//adjust message time when msgtype is "previous_text"
				var offset = inConcert.app.getTimeOffset() * 1000;
				var t = new Date(message.originalMessage.msgTimeStamp - offset);
				var z = inConcert.Utils.zeropad;
				message.msgtime = z(t.getHours()) + ':' + z(t.getMinutes()) + ':' + z(t.getSeconds());
			}
			$timestamp = $('<label class="chatMessageTime">' + message.msgtime + '</label>');
		}
		//make author
		var $author = null;
		if (author) {
			$author = $('<label class="chatMessageAutor"></label>');
			$author.text(author);
		}
		//make content
		var $messageContent = this.__buildMessageContent(contentMessage);
		//make container
		var chatMessageClass = (source === MESSAGE_SOURCE.CLIENT ? 'chatMessageOther' : 'chatMessageOwned') +
				(customClass? (' ' + customClass) : '') +
				' ' + source + "MessageSource";
		var $messageNode = $(''.concat(
			'<div class="chatMessage ' + chatMessageClass + '">',
				'<div class="chatMessageTimeContainer">',
				'</div>',
			'</div>'
		));
		//add time and content
		$messageNode.find(".chatMessageTimeContainer")
			.append($messageContent)
			.append($timestamp);
		//add author
		if (source === MESSAGE_SOURCE.CLIENT) {
			$messageNode.prepend($author);
		}
		else {
			$messageNode.prepend($author);
		}

		return $messageNode;
	};

	//Adds a message to the window
	ChatClientWindow.prototype.addMessage = function(message, customClass) {
		this.showPeerStoppedTyping();
		var currentName = message.name;
		var messageNode = this.buildMessageNode(
			message,
			currentName !== this.lastName || message.owned !== this.lastDirection ?
			message.name :
			null,
			customClass
		);
		this.lastName = currentName;
		this.lastDirection = message.owned;

		var readMessagesPanel = $('.chatClientRead');
		readMessagesPanel.append(messageNode);
		readMessagesPanel[0].scrollTop = readMessagesPanel[0].scrollHeight;
		if (this.onMessageAdded) {
			this.onMessageAdded(message);
		}
	};

	ChatClientWindow.prototype.processTypingEvent = function (event_type, user_id, display_name, event_data)
	{
		if(event_type === 'start_typing') {
			this.showPeerStartedTyping(user_id, display_name, event_data);
		}
		else {
			this.showPeerStoppedTyping(user_id, display_name, event_data);
		}
	};

	/**
	Generate the desired markup to signal that the peer started typing.
	@param undefined
	@param display_name
	*/
	ChatClientWindow.prototype.showPeerStartedTyping = function (undefined, display_name) {
		let self = this;

		if (self.__endTypingTimmer) {
			window.clearTimeout(self.__endTypingTimmer);
		}
		self.__endTypingTimmer = window.setTimeout(function() {
			self.showPeerStoppedTyping();
			self.__endTypingTimmer = null;
		}, 10000);

		var message = format( "<%=_T('{0} is typing')%>", display_name);
		$('.chatFeedback').text(message).show();
	};

	/**
	Generate the desired markup to signal that the peer stopped typing.
	*/
	ChatClientWindow.prototype.showPeerStoppedTyping = function () {
		if (this.__endTypingTimmer) {
			window.clearTimeout(this.__endTypingTimmer);
			this.__endTypingTimmer = null;
		}
		$('.chatFeedback').text('').hide();
	};

	/**
	@param interaction : ChatInteraction
	@param session
	@param customClass
	*/
	ChatClientWindow.prototype.populateListWithSessionMessages = function (interaction, session, customClass)
	{
		var clientName;

		var text_messages = function(msg) {
			return msg.msgtype === "text";
		};

		session.messages.filter(text_messages).forEach(function(msg, i) {
			// will asume the first message origin to be NOT owned, and the rest to be owned
			if (i === 0) {
				clientName = msg.display_name;
			}
			var message = JSON.parse(JSON.stringify(msg));
			message.name = msg.display_name;
			message.owned = clientName !== msg.display_name;
			message.msgtime = interaction.getMessageTime(msg);
			this.addMessage(message, customClass);
		}, this);
	};

	ChatClientWindow.prototype.getInteractionFromRepository = function (interaction, callback)
	{
		var self = this;
		inConcert.app.getInteractionManager().getFromRepository(interaction, function(session, error_response, error_msg){
			self.hideStatusMessage();
			if (!session) {
				// handle file not found
				if (error_response.status === 404) {
					self.displayStatusMessage( "<%=_T('Conversation not found on the repository')%>", 'Alert', true);
				}
				else {
					self.displayStatusMessage( format("<%=_T('Failed to get conversation from repository with error: {0}')%>", error_msg.result), 'Error', false, 5000);
				}
				callback(null);
			}
			else {
				if(typeof(session) === 'string') {
					try {
						session = JSON.parse(session);
					} catch(e) {
						self.displayStatusMessage( "<%=_T('Failed to get conversation from repository with error: Unable to parse incoming data')%>", 'Error', false, 5000);
						callback(null);
						return;
					}
				}
				callback(session);
			}
		});
	};

	ChatClientWindow.prototype.getTransferredInteractionsFromRepository = function (interaction, callback)
	{
		var self = this;
		let transferrerIds = [];
		if (interaction) {
			transferrerIds = interaction.getTransferrerIds();
		}
		
		(function loadTransferId(index, result) {
			if (index === (transferrerIds.length || 0)) {
				return callback(result);
			}
			self.getInteractionFromRepository({ id : transferrerIds[index] }, function(session) {
				if (session) {
					result.push(session);
				}
				loadTransferId(index + 1, result);
			});
		})(0, []);
	};

	ChatClientWindow.prototype.loadTransferrerSessionsMessages = function(interaction, callback)
	{
		var self = this;
		self.getTransferredInteractionsFromRepository(interaction, function(sessions) {
			$.each(sessions, function(_, session) {
				self.populateListWithSessionMessages(interaction, session, "transferrerMessage");
			});
			callback();
		});
	};

	ChatClientWindow.prototype.getDefaultDialogOptions = function (){
		return {
			dialog : {
				closeOnEscape: false,
				resizable: false
			},
			extend : {
				'closable' : true,
				'maximizable' : false,
				'minimizable' : true,
				'minimizeLocation' : 'left',
				'collapsable' : false,
				'titlebar' : false
			}
		};
	};

	ChatClientWindow.prototype.clientUnusable = function()
	{
		ChatClientWindow.uber.clientUnusable.call( this ); //invokes the base class method
		this.clearSendMessageText();
		this.enableSendMessageButton(false);
		this.enableSendMessageText(false);

		$("#coachMessageText").prop("disabled", true).val("");
		$("#coachSendMessage").prop("disabled", true);
	};

	ChatClientWindow.prototype.onReadyFinalization = function()
	{
		this.focusSendMessageText();
	};

	ChatClientWindow.prototype.__canCloseWindow = function(params)
	{
		if (params && params.interactive) {
			if (this.hasAnyIntegration()) {
				return false;
			}
		}
		return true;
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Archived Client Handler
	//
	function ArchivedClientWindow(interaction, options)
	{
		console.debug("ArchivedClientWindow.Constructor");
		var self = this;
		options = options || {};
		options.archive = true;
		ArchivedClientWindow.uber.constructor.call(self, interaction, options);
		this.toggleSendMessage(false);
		$('#contactCall').hide();

		self.setTitle(format("<%=_T('{0}')%>", interaction.getDisplayName()));

		$('#elapsedTime').text(self.m_interaction.timeSinceLastEvent());

		// show waiting
		$('.chatClientWaiting').show();

		// Go and search repository.
		self.loadTransferrerSessionsMessages(interaction, function() {
			self.getInteractionFromRepository(interaction, function(session){
				// hide waiting
				$('.chatClientWaiting').hide();

				if (session){
					// load parts list and player
					self.loadPlayer(interaction, session);
					// append messages
					self.populateListWithSessionMessages(interaction, session);
				}
			});
		});
	}

	inConcert.Utils.inherit(ArchivedClientWindow, ChatClientWindow);

	ArchivedClientWindow.prototype.sendMessage = $.noop;

	ArchivedClientWindow.prototype.loadPlayer = function(interaction) {
		const self = this;

		let archive_header_tpl = ''.concat(
			'<tr>',
				'<th><%=_T("Type")%></th>',
				'<th><%=_T("Part")%></th>',
				'<th><%=_T("Format")%></th>',
				'<th class="gridColumnActions"></th>',
			'</tr>'
		);

		let archive_row_tpl = ''.concat(
				'<tr part="{{ part }}" agent="{{ agent }}" file_type="{{ file_type }}" format="{{ format }}" class="{{ classes }}" >' +
					'<td>{{ agentText }}</td>' +
					'<td>{{ part }}</td>' +
					'<td>{{ format_prefix }}<%=_T(":")%> {{ format }}</td>' +
					'<td>' +
						'{0}' +
						'{1}' +
					'</td>' +
				'</tr>'
			);

		self.templates = {
			archive_row_tpl : archive_row_tpl,
			archive_row_tpl_play : ''.concat(
				'<button type="button" action="play" class="buttonColorGray gridPlayerButton gridButPlay" title="<%=_T("Play")%>">' +
					'<i></i>' +
				'</button>'
			),
			archive_row_tpl_download : ''.concat(
				'<button type="button" action="download" class="buttonColorGray gridPlayerButton gridButDownload" title="<%=_T("Download")%>">' +
					'<i></i>' +
				'</button>'
			)
		};

		$("body").addClass("callClientPlayer");
		$(".playerPartsGridHeader").append(archive_header_tpl);
		interaction.getRecords(function(success, records) {
			if (success) {
				self.setMediaPlayer(records);
			}
		});
	};

	/** Set media player
	* @param  {object}		sections		-Record sections.
	*/
	ArchivedClientWindow.prototype.setMediaPlayer = function(sections) {
		const self = this;

		function setEmptyPlayer() {
			$("#noRecordsFoundMessage").show();
		}

		//TODO: Si no se empezo a reproducir, el mainPlayerContainer no se deberia mostrar (aparece su borde, con contenido vacio)
		$("#viewConversation").click(function() {
			$("#partContainer").toggle();
			$("#conversationContainer").toggle();
			$("#conversationClientRead").toggle();
			$("#mainPlayerContainer").hide();
			$("#viewScreenrecording").toggle();
			$("#viewConversation").toggle();
		});

		$("#viewScreenrecording").click(function() {
			$("#partContainer").toggle();
			$("#conversationContainer").toggle();
			$("#conversationClientRead").toggle();
			$("#viewScreenrecording").toggle();
			$("#viewConversation").toggle();
		}).show();

		if ($.isEmptyObject(sections)) {
			setEmptyPlayer();
			return;
		}

		let html = "";
		let listen = this.m_interaction.hasPermission("LISTEN_RECORD", this.from_app) && this.templates.archive_row_tpl_play || "";
		let download = this.m_interaction.hasPermission("DOWNLOAD_RECORD", this.from_app) && this.templates.archive_row_tpl_download || "";
		let agentNames = {};

		function getAgentName(agentAtVCC, callback) {
			if (!agentAtVCC) {
				return callback("");
			}
			let agentId = inConcert.Utils.getIdWithoutVCC(agentAtVCC);
			if (agentNames[agentId]) {
				return callback(agentNames[agentId]);
			}
			else if (agentId === inConcert.app.getLoggedUserId()) {
				let agentName = inConcert.app.getLoggedUserName();
				agentNames[agentId] = agentName;
				callback(agentName);
			}
			else {
				inConcert.app.getTeamMember(agentId, function(member) {
					let agentName = inConcert.app.getMemberName(member);
					agentNames[agentId] = agentName;
					callback(agentName);
				}, true);
			}
		}

		function getAgent(section) {
			let agent = "";
			switch(section.Type) {
				case "Agent":
					agent = section.agent;
					break;
				case "TransferConsultation":
				case "ConferenceConsultation":
				case "Conference":
				case "ConferenceWithoutClient":
					agent = section.from;
					break;
				case "Bot":
					agent = "bot";
			}
			return inConcert.Utils.getIdWithoutVCC(agent);
		}

		function getAgentText(section, callback) {
			//first get all names
			inConcert.Utils.asyncIterate(["agent", "from", "to"], function(key, callback) {
				getAgentName(section[key], function(agentName) {
					section[key + "Name"] = agentName;
					callback();
				});
			}, function() {
				let toName = section.toName || section.number;
				switch(section.Type) {
					case "Agent":
						return callback(format("<%=_T('Agent {0}')%>", section.agentName));
					case "TransferConsultation":
					case "ConferenceConsultation":
						return callback(format("<%=_T('Consultation from {0} to {1}')%>",
							section.fromName, toName));
					case "Conference":
						return callback(format("<%=_T('Conference with {0} and {1}')%>",
							section.fromName, toName));
					case "ConferenceWithoutClient":
						return callback(format("<%=_T('Conference with {0} and {1} (after contact hangup)')%>",
							section.fromName, toName));
					case "Bot":
						return callback("<%=_T('Bot')%>");
					default:
						return callback("<%=_T('Record')%>");
				}
			});
		}

		inConcert.Utils.asyncIterate(sections, function(section, callback, i) {
			let previousSection = i > 0? sections[i -1].Section : null;
			let nextSection = i < sections.length - 1? sections[i + 1] : null;
			let currentSection = section.Section;
			let classes = "";
			if (previousSection !== currentSection) {
				classes += " firstPart";
			}
			if (nextSection !== currentSection) {
				classes += " lastPart";
			}
			if (!classes) {
				classes = "middlePart";
			}

			getAgentText(section, function(agentText) {
				html += format(Mustache.render(self.templates.archive_row_tpl, {
					agent : getAgent(section),
					section : section.Section,
					agentText : agentText,
					sectionText : section.Section,
					part : section.Part,
					format : section.Format,
					format_prefix : section.Video? "<%=_T('Video')%>" : "<%=_T('Audio')%>",
					listen : listen,
					download : download,
					classes : classes,
					file_type : section.Video? (section.Mode === "screen_recording" ? "screen_recording" : "video") : null
				}), listen, download);
				callback();
			});
		});

		$("#partContainer")
				.find("tbody")
				.html(html)
				.find("button")
				.click(function() {
					let $row = $(this).parent().parent();
					let section = $row.attr("section");
					let part = $row.attr("part");
					let agent = $row.attr("agent");
					let format = $row.attr("format");
					let fileType = $row.attr("file_type");
					let action = $(this).attr("action");
					switch(action) {
						case "download":
							return self.downloadRecord(agent, section, part, format, fileType);
						case "play":
							return self.playRecord(agent, section, part, format, fileType);
					}
				}
		);
	};

	/** Open record download popup
	* @param  {string}		agent			-Record agent.
	* @param  {string}		section			-Record section.
	* @param  {string}		part			-Record part.
	* @param  {string}		fileFormat		-Record format.
	*/
	ArchivedClientWindow.prototype.downloadRecord = function(agent, section, part, fileFormat, fileType) {
		let url = this.getRecordUrl(section, part, true, fileType) + "/" +
			this.getRecordFileName(agent, section, part, fileFormat);
		window.open(url);
	};

	/** Plays record
	* @param  {string}		agent			-Record agent.
	* @param  {string}		section			-Record section.
	* @param  {string}		part			-Record part.
	* @param  {string}		fileFormat		-Record format.
	*/
	ArchivedClientWindow.prototype.playRecord = function(_/*agent*/, section, part, fileFormat, fileType) {
		const self = this;

		$("#mainPlayerContainer").show();

		let url = self.getRecordUrl(section, part, false, fileType);
		let player = new CInconcertPlayer({
			containerId : "mainPlayerContainer",
			url : url,
			fileType : fileType,
			fileFormat : fileFormat,
			success : function() {
				player.play();
			},
			error : function() {
				self.displayStatusMessage("<%=_T('Error playing record')%>", "Error", false, 2000);
			}
		});

		player.load();
		self.m_player = player;

		$("#partContainer")
				.find("tbody tr")
				.removeClass("selected")
				.filter(format("[section={0}][part={1}]", section || 1, part))
				.addClass("selected");
	};

	/** Get record filename
	* @param  {string}		agent			-Record agent.
	* @param  {string}		section			-Record section.
	* @param  {string}		part			-Record part.
	* @param  {string}		fileFormat		-Record format.
	*/
	ArchivedClientWindow.prototype.getRecordFileName = function(agent, section, part, fileFormat) {
		let campaign = this.m_interaction.campaign && this.m_interaction.campaign.split("@")[0] || "";
		let datetime_aux = inConcert.Utils.convertLocalDateToTimezoneDate(new Date(this.m_interaction.startTimestamp* 1000) ,inConcert.app.m_user_timezone_object);
        let datetime= inConcert.Utils.getStringFromDate(datetime_aux, false, "full_datetime", "yyyy-mm-dd" ,"-");

		let date = datetime.split(" ")[0];
		let time = datetime.split(" ")[1].split(".")[0];
		let fileType = fileFormat;

		let filename = format("{0}_{1}_{2}_{3}_{4}_{5}.{6}",
			campaign, agent, date, time, section, part, fileType);
		return filename;
	};

	/** Get record url.
	* @param  {string}		section			-Record section.
	* @param  {string}		part			-Record part.
	* @param  {boolean}		download		-Indicates if url is for download or play.
	*/
	ArchivedClientWindow.prototype.getRecordUrl = function(_/*section*/, part, _1/*download*/, fileType) {
		let url = inConcert.app.getBaseDomain();
		url += "/inconcert/apps/repository/download/?id=" + this.m_interaction.id +
			(fileType === "video"? "_video" : (fileType === "screen_recording" ? "_screen_recording" : "")) +
			"&section=" + part + "&part=" + part;
		return url;
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Wrapped up Client Handler
	//
	function WrapupClientWindow(interaction, options)
	{
		var self = this;

		WrapupClientWindow.uber.constructor.call(self, interaction, options);

		// nothing to do with a wrapped up conversation.
		this.toggleSendMessage(false);
		$('#contactCall').hide();

		// add wrapup specific options
		self.change_client_state("wrapping_up");

		self.setTitle(format("<%=_T('{0}')%>", interaction.getDisplayName()));

		// show waiting
		$('.chatClientWaiting').show();

		// Go and search repository.
		self.loadTransferrerSessionsMessages(interaction, function() {
			self.getInteractionFromRepository(interaction, function(session){
				// hide waiting
				$('.chatClientWaiting').hide();

				if (session){
					// append messages
					self.populateListWithSessionMessages(interaction, session);
				}
			});
		});
	}

	inConcert.Utils.inherit(WrapupClientWindow, ChatClientWindow);

	WrapupClientWindow.prototype.canCloseWindow = function(params) {
		return this.__canCloseWindow(params);
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Live (base) Client Handler
	//
	function LiveClientWindow(interaction, options)
	{
		var self = this;
		if ( !interaction ) {
			return;
		}
		LiveClientWindow.uber.constructor.call(self, interaction, options);

		if (self.m_interaction.isWrappedUp()) {
			self.change_client_state("wrapping_up");
		}
		else {
			self.change_client_state("connected");
		}

		self.interaction_previous_state = self.m_interaction.state;
		self.m_unread_messages = 0;
		self.m_unread_timeout_handle = null;
		// set window title
		self.m_window_title = self.m_window_title || {
			normal: "<%=_T('{0}')%>",
			minimized_unread: "<span class='minimizedUnreadCounter'>{0}</span>{1}"
		};
		var display_name = self.m_interaction.getDisplayName();
		self.setTitle(format(self.m_window_title.normal, display_name));

		if (self.m_interaction.coached_by) {
			self.createCoachWindow(self.m_interaction.coached_by);
		}
		//coaching and coached all use the same hooks
		self.hookMonitorCommands();

		self.loadTransferrerSessionsMessages(interaction, function() {

			//load queued messages in case this window is reopening, only if we're not monitoring
			if(!options || !options.monitor) {
				var session = interaction.getSession();
				if (session) {
					var messages = session.getMessages() || [];
					if ( messages.length === 0 ) {
						// Cuando se crea la interaccion, y si el contacto aun no ha escrito ningun mensaje
						// se debe dejar un blink en la ventana minimizada
						self.updateWindowHeader();
					}
					else {
						for (var i = 0; i < messages.length; i++) {
							// si la ventana se esta creando minimizada ( this.create_minimized = true )
							// debe quedar un blink en la ventana minimizada, con el numero 1
							// por esto pone la variable can_update_window_title = true
							var can_update_window_title = i === 0 && self.create_minimized;
							//TODO this should be added as already read
							self.addMessage(messages[i], can_update_window_title);
						}
					}
				}
			}

			self.setEvents();
			self.m_timer = window.setInterval(function() {
				self.onTimer();
			}, 1000);

			self.m_typing_fsm = new TypingFSM(self);

			if ('SpellChecker' in $) {
				var $body = $("body");
				var validate_spellchecker_permissions = function()
				{
					// add to dictionary permission
					var allowed_add_to_dictionary = self.hasPermissionToAction( 'SPELL_CHECKER_ADD_WORD' );
					var $add_dictionary = $body.find('.ignore-forever');
					if ( allowed_add_to_dictionary ) {
						$add_dictionary.show();
					}
					else {
						$add_dictionary.hide();
					}
				};

				var vcc = inConcert.app.get_vcc_id();
				var language = self.get_language();

				self.m_spellchecker = new $.SpellChecker(self.$sendMessageText(), {
					lang: language,
					parser: 'text',
					webservice: {
						path: inConcert.app.getBaseDomain() + '/inconcert/apps/spellchecker/spellcheck/' + vcc,
						driver: 'pspell'
					},
					suggestBox: {
						position: 'above'
					},
					incorrectWords: {
						container: '#incorrect-word-list'
					}
				});

				var $incorrect_words = $('.spellchecker-incorrectwords');
				// Bind spellchecker handler functions
				self.m_spellchecker.on('check.success', function() {
					$incorrect_words.empty().parent().hide();
				});

				var $incorrect_words_list = $('#incorrect-word-list');
				self.m_spellchecker.on('replace.word', function() {
					// if there are no more words to be replaced, hide the container
					if( $incorrect_words.children().length === 0 ) {
						$incorrect_words_list.hide();
					}
				});

				var $spellcheck_feedback = $('.spellcheck_feedback');
				// each time we send a message, we increment this counter, so when a spellcheck completes
				// we can check if the line number we sent to check is the current line we are on.
				self.m_spellchecker.current_line = 0;
				self.m_spellchecker.on('check.fail', function() {
					if (self.m_spellchecker.last_checked_line !== self.m_spellchecker.current_line) {
						$incorrect_words.empty().parent().hide();
					}
					else {
						$incorrect_words_list.show();
						$spellcheck_feedback.hide();
					}
					validate_spellchecker_permissions();
				});

				$('#chatButtonSpellcheck').click(function() {
					self.m_spellchecker.check(null, function(words){
						if(_.isArray(words[0]) && words[0].length > 0) {
							$spellcheck_feedback.hide();
						}
						else {
							$spellcheck_feedback.show();
							window.setTimeout(function() {
								$spellcheck_feedback.hide();
							}, 3000);
						}
					});
				});
			}

		});

		if(self instanceof LiveWebChatClientWindow && self.m_interaction && self.m_interaction.hasPermission("SCREEN_SHARE", "interaction") && self.m_interaction.getSession() && self.m_interaction.getSession().CanEnableScreenSharing()) {
			$('#enableScreenShare').click(function() {
				if(self.m_interaction && self.m_interaction.getSession()) {
					self.m_interaction.getSession().enableScreenSharing();
				}
			});
			$('#disableScreenShare').click(function() {
				if(self.m_interaction && self.m_interaction.getSession()) {
					self.m_interaction.getSession().disableScreenSharing();
				}
			});
			self.toggleScreenShare(true);
		}
		else {
			self.toggleScreenShare(false);
		}
	}

	inConcert.Utils.inherit(LiveClientWindow, ChatClientWindow);

	LiveClientWindow.prototype.hasAttachment = function() {
		return true;
	};

	LiveClientWindow.prototype.hasAttachmentDescriptionEmojies = function() {
		return true;
	};

	LiveClientWindow.prototype.getAttachmentOptions = function() {
		var options = LiveClientWindow.uber.getAttachmentOptions.call(this);
		options.image.description = false;
		options.video.description = false;
		options.audio.description = false;
		options.file.description = false;
		return options;
	};

	LiveClientWindow.prototype.hasDragAndDropFiles = function() {
		return true;
	};

	LiveClientWindow.prototype.getDragAndDropControl = function() {
		return this.$sendMessageText();
	};

	LiveClientWindow.prototype.onload = function() {
		let self = this;
		window.setTimeout(function() {
			self.focusSendMessageText();
		}, 0);
	};

	LiveClientWindow.prototype.onSessionClosed = function()
	{
		$('#internalMessageInfo').hide();
		this.hideStatusMessage();
		if (this.m_interaction && this.m_interaction.isLocal()) {
			this.change_client_state("ended");
			this.showInternalMessage("<%=_T('The conversation has ended')%>");
		}
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		this.toggleScreenShare(false);
	};

	LiveClientWindow.prototype.canCloseWindow = function(params) {
		return this.__canCloseWindow(params);
	};

	LiveClientWindow.prototype.onWindowClosed = function()
	{
		this.m_typing_fsm.event('stop');
		return LiveClientWindow.uber.onWindowClosed.call(this);
	};

	LiveClientWindow.prototype.setCannedMessage = function(canned)
	{
		let message, must_automatic_send;
		let format_type = this.sidebar.get_canned_format_type(canned);

		if (format_type.id === "TEXT PLAIN") {
			message = canned.plain_message || canned.message;
			must_automatic_send = false;
		}
		else {
			message = canned.message;
			must_automatic_send = true;
		}

		this.appendToSendMessageText(message || "");

		if (must_automatic_send) {
			this.sendMessage(true);
			return;
		}
		// Como no es un envio automatico del mensaje, se queda posicionado en el control de escritura
		this.focusSendMessageText();
	};

	LiveClientWindow.prototype.setQnAMakerAnswer = function(answer)
	{
		let message = answer.answer || "";
		this.appendToSendMessageText(message);
		this.focusSendMessageText();
	};

	LiveClientWindow.prototype.getWindowSizeOnSidebarShown = function()
	{
		var size = { width: null, height: null };
		var $body = $("body");
		if ( $body.is(".underCoachingClient") || $body.is(".coachingClient") ) {
			size.width = DEFAULT_WINDOW_WIDTH_PX;
		}
		else {
			size.width = DEFAULT_WINDOW_WIDTH_PX;
		}
		return size;
	};

	LiveClientWindow.prototype.getWindowSizeOnSidebarHidden= function()
	{
		var size = { width: null, height: null };
		var $body = $("body");
		if ( $body.is(".underCoachingClient") || $body.is(".coachingClient") ) {
			size.width = DEFAULT_WINDOW_WIDTH_PX;
		}
		else {
			size.width = DEFAULT_WINDOW_WIDTH_PX;
		}
		return size;
	};

	/*
		Creates the side-window for the private conversation with the coach
	*/
	LiveClientWindow.prototype.createCoachWindow = function(coachAtVCC)
	{
		if (!coachAtVCC.match(/(.+)@([^@]+)$/)) {
			return;
		}
		// la funcion activate() lo que hace es un restore de la ventana si la misma se encuentra minimizada.
		this.activate();
		var coachId = RegExp.$1;

		this.setHorizontalResizingLimits(DEFAULT_WINDOW_WIDTH_PX);

		var options = this.getDefaultDialogOptions();
		options.dialog.title = format("<%=_T('Coached by: {0}')%>", coachId);
		options.dialog.dialogClass = "noClose";
		options.dialog.beforeclose = function (/*event, ui*/) { return false; };

		var dialog_container = $(Mustache.render(templates.coach_dialog_tpl, { send_text :  "<%=_T('SEND')%>" }));
		dialog_container.css('z-index', $('body').css('z-index') + 1);
		dialog_container.dialog(options.dialog).dialogExtend(options.extend);

		//coaching and coached all use the same hooks
		this.hookMonitorCommands();
		this.m_coach = coachAtVCC;
		this.change_client_state("being_coached");

		inConcert.app.getTeamMember(
			coachId,
			function(coach) {
				if (coach) {
					var coachingHeaderNode = $('#coachingHeaderSection');
					var userAvatarURI = inConcert.app.getBaseUrl() + '/users/get_avatar/' + coachId;
					coachingHeaderNode.find('.coachingChatWithAvatar > img').attr('src', userAvatarURI);
					coachingHeaderNode.find('.coachingChatWithName').text(coach.Name);
				}
			},
			true //from cache
		);
	};

	LiveClientWindow.prototype.__canAddUser = function( _, callback )
	{
		var self = this;
		callback = callback || $.noop;

		var not_found_msg = _T('Unable join user to conversation. Conversation not found.');
		var response = {success: true};
		if (!this.m_interaction) {
			response.success = false;
			response.message = not_found_msg;
			response.action = "CLOSE_WINDOW";
			callback(response);
			return;
		}
		if (this.interaction_previous_state) {
			if (this.interaction_previous_state === "WAITING") {
				var interactionManager = inConcert.app.getInteractionManager();
				interactionManager.getInteraction(this.m_interaction.id, function( interaction_server )
				{
					if (!interaction_server) {
					// la interaccion no existe
						response.success = false;
						response.message = not_found_msg;
						response.action = "CLOSE_WINDOW";
						callback(response);
						return;
					}
					if (!interaction_server.isWaiting()) {
					// la interaccion ha cambiado de estado
						self.m_interaction.state = interaction_server.state;
						response.success = false;
						response.message = _T('Unable join user to conversation. The conversation is not in waiting.');
						response.action = "CLOSE_WINDOW";
						callback(response);
						return;
					}
					// Puede abrir la ventana con la interaccion en Waiting.
					callback(response);
				});
				return;
			}
		}
		this.interaction_previous_state = this.m_interaction.state;

		callback(response);
	};

	/*
		If the user entering the session has the role coach, asume it's coaching me and create the coach window
	 */
	LiveClientWindow.prototype.addUser = function(user_id, user_name, user_role)
	{
		var self = this;

		this.__canAddUser(user_id, function(response)
		{
			if (!response.success) {
				inConcert.app.m_mainWindow.showBannerMessage({ message: response.message, timeout: 5000 });
				if (response.action === "CLOSE_WINDOW") {
					self.close();
				}
				return;
			}
			if (user_role === 'coach') {
				self.createCoachWindow(user_id);
				return;
			}
			ChatClientWindow.prototype.addUser.call(self, user_id, user_name, user_role); // invoco al metodo de la clase base
		});
	};

	/*
		If the user leaving the session is the coach destroy the coaching window
	 */
	LiveClientWindow.prototype.removeUser = function(user_id, user_name)
	{
		ChatClientWindow.prototype.removeUser.call(this, user_id, user_name);
		if (this.m_interaction.coached_by === user_id) {
			this.destroyCoachWindow();
			this.change_client_state("connected");
		}
		// if the interaction was coached by me, disable coach actions.
		if (this.m_interaction.coached_by === inConcert.app.getCompleteId()){
			this.__disable_coach_actions();
		}
	};

	LiveClientWindow.prototype.destroyCoachWindow = function() {
		$('#coachingDialogContainer').remove();
	};

	LiveClientWindow.prototype.hookMonitorCommands = function() {
		var self = this;
		var send_message_button = $('#coachSendMessage');
		var text_area = $('#coachMessageText');

		send_message_button.unbind('click').bind('click', function() {
			self.sendCoachMessage();
		});

		text_area.unbind('keypress').bind('keypress', function(){
			// enter send message
			if (event.which === 13) {
				self.sendCoachMessage();
				return false;
			}
		});
	};

	LiveClientWindow.prototype.setEvents = function()
	{
		var self = this;

		var language_timeout = null;
		$('#cboLanguage').change(function()
		{
			if ( language_timeout ) {
				window.clearTimeout(language_timeout);
				language_timeout = null;
			}
			var $cbo = $(this);
			language_timeout = window.setTimeout(function()
			{
				window.clearTimeout(language_timeout);
				language_timeout = null;
				var new_language = $cbo.val();
				self.ChangeSpellcheckLanguage(new_language);
			}, 1000);
		});

		self.$sendMessageButton().click(function() {
			self.sendMessage();
			self.m_typing_fsm.event('message_sent');
		});

		self.startEmojiPicker();

		//ENTER triggers message send on message input
		self.bindTextChange(
			function() {
				self.__updateCharsCounter();
				if (self.getSendMessageText().length) {
					self.m_typing_fsm.event('key');
				}
			},
			function() {
				self.sendMessage();
				self.m_typing_fsm.event('message_sent');
			});
		self.__updateCharsCounter();
	};

	LiveClientWindow.prototype.onMaximize = function(/*event*/)
	{
		LiveClientWindow.uber.onMaximize.call( this );
		// si la ventana se esta creando aun (this.create_minimized = true)
		//  no debe de actualizar el titulo (el iframe aun no se creo)
		// Esto porque sino me cambia el blink con el contador cuando esta minimizada
		if ( !this.create_minimized ) {
			this.updateWindowHeader();
			this.focusSendMessageText();
		}
	};

	LiveClientWindow.prototype.onRestore = function(/*event*/)
	{
		LiveClientWindow.uber.onRestore.call( this );
		// si la ventana se esta creando aun (this.create_minimized = true)
		//  no debe de actualizar el titulo (el iframe aun no se creo)
		// Esto porque sino me cambia el blink con el contador cuando esta minimizada
		if ( !this.create_minimized ) {
			this.updateWindowHeader();
			this.focusSendMessageText();
		}
	};

	LiveClientWindow.prototype.updateWindowHeader = function()
	{
		var self = this;
		var display_name = this.m_interaction.getDisplayName();
		var window_title;

		// si o si tengo que realizar un cleartimeout
		if (this.m_unread_timeout_handle) {
			window.clearTimeout(this.m_unread_timeout_handle);
		}

		if ( this.isMinimized() ) {
			// 1- actualizo la cantidad de unread de la ventana minimizada (el titulo de la ventana)
			if ( this.m_unread_messages > 0 ) {
				window_title = format(this.m_window_title.minimized_unread, this.m_unread_messages, display_name);
			}
			else {
				window_title = format(this.m_window_title.normal, display_name);
			}
			// 2- agrego una clase a la ventana minimizada para que realice el blink por N segundos
			this.updateWindowClasses("minimizedUnreadBlink");
			// 3- agrego un timeout de N segundos, luego cambio la clase de unread de la ventana minimizada
			this.m_unread_timeout_handle = window.setTimeout(function()
			{
				var classes = "";
				if ( self.isMinimized() ) {
					classes = "minimizedUnread";
				}
				self.updateWindowClasses(classes);
				self.m_unread_timeout_handle = null;
			}, 30000); // lo dejo 30 segundos al blink de unread, despues lo cambio una clase estatica de unread
		}
		else {
		// si la ventana no se encuentra minimizada debe dejar el titulo sin el mensaje de unread
			window_title = format(this.m_window_title.normal, display_name);
			if (this.m_unread_messages > 0 ) {
				this.m_unread_messages = 0;
			}
			this.updateWindowClasses();
		}

		this.setTitle(window_title);
	};

	LiveClientWindow.prototype.__canUpdateMinimizedWindowHeader = function()
	{
		if ( this.isMinimized() ) {
			return true;
		}
		return false;
	};

	LiveClientWindow.prototype.__canAddMessage = function( _, callback )
	{
		callback = callback || $.noop;

		var response = {success: true};

		if (!this.m_interaction) {
		// error
			response.success = false;
			response.message = _T('Unable to add message conversation. Conversation not found.');
			callback(response);
			return;
		}
		// Aca se podria agregar mas validaciones / filtros.

		// success
		callback(response);
	};

	/*
		This function gets called when there's a new message on this session(i.e. for this window)
		If the message is from/to the coach it will be forwarded to the coach window and avoid messing with the main conversation
	*/
	LiveClientWindow.prototype.addMessage = function( message, can_update_window_title )
	{
		var self = this;

		this.__canAddMessage(message, function(result)
		{
			if (!result.success) {
				inConcert.app.m_mainWindow.showBannerMessage({ message: result.message, timeout: 5000 });
				return;
			}
			can_update_window_title = can_update_window_title === undefined ? true : can_update_window_title;
			if ( can_update_window_title && self.__canUpdateMinimizedWindowHeader() ) {
				self.m_unread_messages++;
				const $interaction = inConcert.app.getMainWindow().getPanel('AssignedInteractions').$findInteraction( self.m_interaction );
				const $counterInteraction = $interaction.find(`#${self.m_interaction.id}_unreadEventCounter`);
				$counterInteraction.prop("title", self.m_unread_messages);
				$counterInteraction.text(self.m_unread_messages);
				$counterInteraction.show();
				self.m_interaction.m_unread_messages = self.m_unread_messages;
				self.updateWindowHeader();
			}
			if (message.coach_session) {
				//it's a message from/to the coach session
				var messageNode = self.buildMessageNode(message, message.name);
				var $coachMessagesPanel = $('.coachingChatRead');
				if ($coachMessagesPanel.length > 0) {
					$coachMessagesPanel.append(messageNode);
					if ($coachMessagesPanel[0]){
						$coachMessagesPanel[0].scrollTop = $coachMessagesPanel[0].scrollHeight;
					}
				}
				//blink it if the message is not mine
				if (self.onMessageAdded) {
					self.onMessageAdded(message);
				}
				return;
			}
			ChatClientWindow.prototype.addMessage.call(self, message);
		});
	};

	LiveClientWindow.prototype.onMessageAdded = function(message) {
		if (!message.owned) {
			this.notifyTaskbar();
		}
	};

	// this function implement on the specific client
	LiveClientWindow.prototype.clientUpdate = function(interaction) {
		var currentInteraction = interaction ? interaction : this.m_interaction;
		if (!currentInteraction) {
			return;
		}
		this.updateContactDisplay(currentInteraction);
		this.updateWindowHeader();
	};

	/*
		Sends a message to the coach
	*/
	LiveClientWindow.prototype.sendCoachMessage = function()
	{
		var $textbox = $('#coachMessageText');
		var message = $.trim($textbox.val());
		if (message.length === 0) {
			return;
		}
		var session = this.m_interaction.getSession();
		if (session) {
			//the agent needs to prefix the message with ## in order to send it only to the coach
			var text = '##' + message;
			session.sendMessage(inConcert.Utils.html2text(text));
		}
		else {
			this.displayStatusMessage("<%=_T('Failed sending message, please try again in a few seconds...')%>", 'Error');
		}
		$textbox.val('');
	};

	/*
		Sends a message to the session
	*/
	LiveClientWindow.prototype.sendMessage = function(send_as_HTML, callback = $.noop)
	{
		if (this.__sending) {
			return callback(false, "already sending another message");
		}
		this.__sending = true;
		var session = this.m_interaction.getSession();
		if (this.m_interaction.isWrappedUp()) {
			this.__sending = false;
			return callback(false, "interaction is in wrapup");
		}
		if (session) {
			var text = this.getSendMessageText();
			if (!send_as_HTML) {
				text = inConcert.Utils.html2text(text);
			}
			session.sendMessage(text, callback);
		}
		else {
			this.displayStatusMessage("<%=_T('Failed sending message, please try again in a few seconds...')%>", 'Error');
			this.__sending = false;
			return callback(false, "session not found");
		}

		if(this.m_spellchecker) {
			this.m_spellchecker.current_line++;
		}

		this.clearSendMessageText();
		this.__updateCharsCounter();
		this.__sending = false;
	};

	/**
	Sends an indication to notify the user started typing.
	*/
	LiveClientWindow.prototype.SendStartTyping = function ()
	{
		var session = this.m_interaction.getSession();
		if(session) {
			session.sendStartTyping();
		}
	};

	/**
	Sends an indication to notify the user stopped typing.
	*/
	LiveClientWindow.prototype.SendStopTyping = function ()
	{
		var session = this.m_interaction.getSession();
		if(session) {
			session.sendStopTyping();
		}
	};

	LiveClientWindow.prototype.ChangeSpellcheckLanguage = function( new_language )
	{
		if( !this.m_spellchecker ) {
			return;
		}
		this.m_spellchecker.config.lang = new_language;
		this.m_spellchecker.webservice.config.lang = new_language;
		this.m_spellchecker.webservice.defaultConfig.data.lang = new_language;

		$('#incorrect-word-list').hide();
		this.Spellcheck();
	};

	LiveClientWindow.prototype.HasToSpellcheck = function ()
	{
		if(!this.m_spellchecker) {
			return false;
		}

		var current_text = this.getSendMessageText();
		return current_text !== '';
	};

	LiveClientWindow.prototype.UpdateSpellcheckUI = function ()
	{
		if(!this.m_spellchecker) {
			return;
		}

		if(!this.HasToSpellcheck()) {
			$('.spellchecker-incorrectwords').empty().parent().hide();
		}
	};

	LiveClientWindow.prototype.Spellcheck = function ()
	{
		if(this.HasToSpellcheck()) {
			this.m_spellchecker.last_checked_line = this.m_spellchecker.current_line;
			this.m_spellchecker.check();
		}
		else {
			$('.spellchecker-incorrectwords').empty().parent().hide();
		}
	};

	LiveClientWindow.prototype.change_client_state = function( new_state )
	{
		new_state = LiveClientWindow.uber.change_client_state.call( this, new_state );

		var $body = $("body");
		var $state_flap = $(".clientStateFlap");
		$body.removeClass("underCoachingClient");

		var $spell = $("#chatButtonSpellcheck");
		var $ctrls_spell = $("[for='chatButtonSpellcheck']");
		var $language = $("#cboLanguage");
		var $ctrls_language = $("[for='cboLanguage']");

		var show_spell_ctrls = function( show )
		{
			var method = show ? "show" : "hide";
			$spell[method]();
			$ctrls_spell[method]();
			$language[method]();
			$ctrls_language[method]();
		};

		switch( new_state )
		{
			case "being_coached":
				$state_flap.text("<%=_T('Being Coached')%>");
				$body.addClass("underCoachingClient");
				show_spell_ctrls( true );
			break;

			case "connected":
				$state_flap.text("");
				show_spell_ctrls( true );
			break;

			case "ended":
				$state_flap.text("<%=_T('Ended')%>");
				show_spell_ctrls( false );
			break;

			case "wrapping_up":
				this.clearSendMessageText();
				this.enableSendMessageText(false);
				$("#coachMessageText").prop("disabled", true).val("");
				$("#coachSendMessage").prop("disabled", true);
				show_spell_ctrls( false );
			break;
		}

		return new_state;
	};

	LiveClientWindow.prototype.__addDraftTextMessage = function(message, callback) {
		this.setSendMessageText(message.text || "");
		callback(true);
	};

	LiveClientWindow.prototype.__sendDraftTextMessage = function(_, callback) {
		this.sendMessage(null, callback);
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Live Internal Chat Client Handler
	//
	function LiveInternalChatClientWindow(interaction, options) {
		LiveInternalChatClientWindow.uber.constructor.call(this, interaction, options);
		this.updateToolbarChat();
	}

	inConcert.Utils.inherit(LiveInternalChatClientWindow, LiveClientWindow);

	LiveInternalChatClientWindow.prototype.updateToolbarChat = function() {
		$('.clientContainerActions input, .clientContainerActions img ').hide();
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Live WebChat
	//
	function LiveWebChatClientWindow(interaction, options) {
		LiveWebChatClientWindow.uber.constructor.call(this, interaction, options);
	}

	inConcert.Utils.inherit(LiveWebChatClientWindow, LiveClientWindow);

	LiveWebChatClientWindow.prototype.toggleScreenShare = function(showEnable, showDisableOnInteractionId) {
		const vShowEnable = showEnable && this.m_interaction.state === "TAKEN";
		const vShowDisable = !vShowEnable && this.m_interaction.id === showDisableOnInteractionId;
		$("#enableScreenShare").toggle(vShowEnable);
		$("#disableScreenShare").toggle(vShowDisable);
		$("#enableScreenShareContainer").toggle(showEnable || vShowDisable);
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Live IM Client Handler
	//
	function LiveIMClientWindow(interaction, options)
	{
		LiveIMClientWindow.uber.constructor.call(this, interaction, options);
		var status = 'connected';

		// if you have a single campaign, automatically starting chat
		if (this.m_interaction) {
			if (this.m_interaction.initiative === 'OUTBOUND') {
				if (this.m_interaction.contact) {
					var campaigns = this.m_interaction.getAllowedCampaigns(inConcert.app.getCampaigns());
					if (campaigns && campaigns.length === 1) {
						if (this.m_interaction.state === 'PENDING' && !this.m_interaction.getSession()) {
							this.startSession();
						}
					}
				}
				else {
					status = 'disconnected';
					$('#contactCall').click();
				}
			}
		}
		this.updateToolbarChat(status);
	}

	inConcert.Utils.inherit(LiveIMClientWindow, LiveClientWindow);

	LiveIMClientWindow.prototype.setEvents = function() {
		var self = this;
		var startChatAndSendMessage = function() {
			/*
			// TODO: Esta opcion se debe dejear habilitada cuando
					se agrega en el IM proxy/IR el atributo 'account' con la cuenta correspondiente.
			if (!self.canExecuteInteractionAction()) {
				return;
			}
			*/
			if (self.m_interaction.state === 'PENDING' && !self.m_interaction.getSession()) {
				self.startSession(true);
			}
			else {
				self.sendMessage();
			}
		};

		self.$sendMessageButton().click(function() {
			startChatAndSendMessage();
			self.m_typing_fsm.event('message_sent');
		});

		self.startEmojiPicker();

		self.bindTextChange(
			function() {
				self.__updateCharsCounter();
				if (self.getSendMessageText().length) {
					self.m_typing_fsm.event('key');
				}
			},
			function() {
				self.sendMessage();
				self.m_typing_fsm.event('message_sent');
			});
		self.__updateCharsCounter();

		$('#contactCall')
			.unbind('click')
			.bind('click', function()
			{
				var selector = new inConcert.ContactSelector($('#contactSelector'),
					self.m_interaction.type,
					undefined,
					function() {
						return $('#cboCampaigns').val();
					});

				var $this = $( this );
				selector.show({
					parentNode: $this,
					callbacks: {
						onOpen  : $.noop,
						onClose : $.noop
					},
					onContactSelectedFn: function(contactInfo, contact) {
						if (typeof(contactInfo) !== 'object' || !contactInfo.addresses || !contact) {
							return;
						}
						//rather than address 0 use the one of type IM
						for (var i=0; i<contactInfo.addresses.length; i++) {
							if (contactInfo.addresses[i].Type === self.m_interaction.type) {
								self.m_interaction.completeAddress = contactInfo.addresses[i];
								break;
							}
						}
						self.m_interaction.contact = contactInfo.Id;
						self.m_selectedContact = true;
						self.m_interaction.interactionData = self.m_interaction.interactionData ? self.m_interaction.interactionData : {};
						self.m_interaction.interactionData.contactName = contactInfo.Name;
						self.m_interaction.contactAddress = contact;
						self.clientUpdate();
						self.disableCampaignAccountsControls();
						self.focusSendMessageText();
					}
				});
			}
		);
	};

	LiveIMClientWindow.prototype.updateToolbarChat = function(status) {
		if (status === 'connected') {
			$('#contactCall').prop('disabled', true);
			$('#contactCall').hide();
			$('#contactCall').next().hide();
		}
		else {
			$('#contactCall').prop('disabled', false);
			$('#contactCall').show();
			$('#contactCall').next().show();
		}
	};

	LiveIMClientWindow.prototype.startSession = function(sendMessage)
	{
		var self = this;
		if (!this.m_interaction.contact || this.m_interaction.contact.length === 0) {
			this.displayStatusMessage(
				"<%=_T('You have to select a contact from the list!')%>",
				'Info', false, 1500, function() {
					$('#contactCall').click();
				});
			return;
		}
		this.updateToolbarChat('waiting');
		this.displayStatusMessage("<%=_T('Requesting chat...')%>", 'Process', true);
		this.m_interaction.createConversation(this.m_interaction.completeAddress,
			//The callback should receive a boolean (successful or not) and a message to display (not used when successful)
			function(successful, message) {
				self.hideStatusMessage();
				self.disableCampaignAccountsControls();
				self.updateToolbarChat('connected');
				if (successful) {
					self.displayStatusMessage("<%=_T('Chat started!')%>", 'Info', false, 1500);
					if (sendMessage) {
						self.sendMessage();
					}
				}
				else {
					self.displayStatusMessage(_T(message), 'Error');
				}
			});
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Monitor Client Handler
	//
	function MonitorClientWindow(interaction, options)
	{
		var self = this;
		options = options || {};
		options.monitor = true;
		// window title
		this.m_window_title = {
			normal: "<%=_T('{0}')%>",
			minimized_unread: "<span class='minimizedUnreadCounter'>{0}</span>{1}"
		};
		MonitorClientWindow.uber.constructor.call(this, interaction, options);

		if (!interaction) {
			return;
		}

		var $coach_agent_button = $('#coachAgent');
		var $owner_node = $('#ownerSection');

		// if the owner section is hidden the coach button is not bound
		if ( $owner_node.is(':visible') && interaction.state !== "CREATED") {
			var agentAtVCC = this.m_interaction.assignedAgents[0];

			$coach_agent_button.show();

			$coach_agent_button.click(function()
			{
				if (self.startingCoach) {
					return;
				}
				self.startingCoach = true;

				// going from MONITOR to COACH
				var chat_client = inConcert.app.getInteractionManager().getChannelClient("CHAT");
				var agent = agentAtVCC.match(/(.+)@([^@]+)$/)[1];

				chat_client.coachSession(self.m_interaction.id, agent, function(ok, response)
				{
					self.startingCoach = false;

					if (ok) {
						var options = self.getDefaultDialogOptions();
						var dialog_view = {
							send_text : "<%=_T('SEND')%>",
							avatar_uri : inConcert.app.getBaseUrl() + '/users/get_avatar/' + agent,
							name : self.m_monitoredMember ? self.m_monitoredMember.Name : agent
						};

						options.dialog.title = format("<%=_T('Coaching: {0}')%>", dialog_view.name);

						// setup handler to go back to monitor mode
						options.dialog.close = function(){
							self.__monitorInteraction(self.m_interaction);
							$coach_agent_button.show();
						};

						var dialog_container = $(Mustache.render(templates.coach_dialog_tpl, dialog_view));
						dialog_container.css('z-index', $('body').css('z-index') + 1);
						dialog_container.dialog(options.dialog).dialogExtend(options.extend);

						self.setHorizontalResizingLimits(DEFAULT_WINDOW_WIDTH_PX);

						self.change_client_state("coaching");
						self.__state = "coaching";

						// setup comands
						self.hookMonitorCommands();

						// you cannot coach again this agent
						$coach_agent_button.hide();
					}
					else {
						if (!response || !response.reason) {
							self.displayStatusMessage("<%=_T('Unable to coach conversation: Unhandled error')%>", 'Error', true);
							self.__disable_chat_actions();
							return;
						}
						if (response.reason === "USER_ALREADY_COACHED") {
							inConcert.app.getTeamMember(
								response.previousCoach,
								function(teamMember) {
									self.displayStatusMessage(_T('This conversation is already being coached by ' +
									(teamMember ? teamMember.Name : response.previous_coach_id)), 'Info', false, 3000);
								},
								true // from cache
							);
						}
						else {
							self.__handle_chat_errors("<%=_T('Unable to coach conversation: ')%>", response);
						}
					}
				});
			});
		}
		else {
			$coach_agent_button.hide();
			$owner_node.hide();
		}

		this.__state = "unknown"; // possible states: unknown, monitoring, coaching, ended
		// default conversation monitor on startup is monitor only (no coaching)
		this.__monitorInteraction(this.m_interaction);
	}

	inConcert.Utils.inherit(MonitorClientWindow, LiveClientWindow);

	MonitorClientWindow.prototype.__disable_chat_actions = function() {
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		$('#assignLabels').hide();
		$('#coachAgent').remove();
	};

	MonitorClientWindow.prototype.__disable_coach_actions = function() {
		$('#coachMessageText, #coachSendMessage').prop('disabled', true);
	};

	MonitorClientWindow.prototype.addMessage = function(message, can_update_window_title ) {

		if (this.__state === "ended") {
			return;
		}
		return MonitorClientWindow.uber.addMessage.call(this, message, can_update_window_title);
	};

	/**
	Handles error that can be dealt with in a generic way.
	@param message
	@param response
	*/
	MonitorClientWindow.prototype.__handle_chat_errors = function(message, response) {
		var self = this;

		switch(response.reason) {
			case 'NONEXISTENT_SESSION':
				// if the conversation has ended, show a read-only view to the user
				self.change_client_state("ended");

				self.displayStatusMessage("<%=_T('The conversation has ended.')%>", 'Info');
				self.__disable_chat_actions();
				// Go and search repository.
				self.loadTransferrerSessionsMessages(self.m_interaction, function() {
					self.getInteractionFromRepository(self.m_interaction, function(session) {
						if (session) {
							// append messages
							self.populateListWithSessionMessages(self.m_interaction, session);
						}
					});
				});
				break;
			case 'USER_NOT_FOUND':
				// the user does not have valid credentials,
				// try to relogin, if the relogin fail, logout from agent, everything is broken.
				inConcert.app.getInteractionManager().reconnectClient('CHAT', function(ok) {
					if (!ok) {
						inConcert.app.onUnrecoverableError();
					}
				});
				break;
			case 'NETWORK_ERROR':
				self.displayStatusMessage(format(_T('There is a network error: {0}', response.message), 'Error'));
				self.__disable_chat_actions();
				break;
			default:
				self.displayStatusMessage(message + (response.message ? response.message : response.reason), 'Error', true);
				self.__disable_chat_actions();
				break;
		}
	};

	MonitorClientWindow.prototype.__monitorInteraction = function(interaction) {
		var self = this;
		var interactionManager = inConcert.app.getInteractionManager();

		interactionManager.startMonitoringInteraction(interaction, null, function(result, response) {
			if(result) {
				self.__state = "monitoring";
				self.change_client_state("monitoring");
			}
			else {
				if (!response || !response.reason) {
					self.displayStatusMessage("<%=_T('Unable to monitor conversation: Unhandled error')%>", 'Error', true);
					self.__disable_chat_actions();
					return;
				}
				self.__handle_chat_errors("<%=_T('Unable to monitor conversation: ')%>", response);
			}
		});
	};

	/**
	The session had closed. If coaching, hide its panel.
	*/
	MonitorClientWindow.prototype.onSessionClosed = function() {
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		$('#assignLabels').hide();
		$('#coachAgent').remove();
		if (this.__state === "coaching") {
			this.change_client_state("monitoring");
		}
		else {
			this.change_client_state("ended");
		}
		this.__state = "ended";
	};

	/**
	Before closing the window, we need to stop either monitoring or coaching. Doing so in onWindowClosed is too late.
	@param params
	*/
	MonitorClientWindow.prototype.onWindowClosing = function(params) {
		inConcert.app.getInteractionManager().stopMonitoringInteraction(this.m_interaction, $.noop);
		return MonitorClientWindow.uber.onWindowClosing.call(this, params);
	};

	/**
	When the monitoring window is closed automatically stop monitoring or coaching the
	current conversation since there is no other handler to this window to do so later
	*/
	MonitorClientWindow.prototype.onWindowClosed = function()
	{
		inConcert.app.getInteractionManager().stopMonitoringInteraction(this.m_interaction, $.noop);
		return MonitorClientWindow.uber.onWindowClosed.call(this);
	};

	/*
		If there's another user monitoring the conversation I do not care
		She's not even coaching ME
	*/
	MonitorClientWindow.prototype.addUser = function(undefined, user_name) {
		this.showInternalMessage(format( "<%=_T('{0} has joined the conversation')%>", user_name), 5000);
	};

	MonitorClientWindow.prototype.sendCoachMessage = function()
	{
		var textbox = $('#coachMessageText');
		var message = $.trim(textbox.val());
		if (message.length === 0) {
			return;
		}
		var session = this.m_interaction.getSession();
		if (session) {
			session.sendMessage(inConcert.Utils.escapeAttrib(message));
		}
		else {
			this.displayStatusMessage("<%=_T('Failed sending message, please try again in a few seconds...')%>", 'Error');
		}
		textbox.val('');
	};

	MonitorClientWindow.prototype.change_client_state = function( new_state )
	{
		if ( new_state === "connected" ) {
			return;
		}
		if ( new_state !== "monitoring" ) {
			new_state = MonitorClientWindow.uber.change_client_state.call( this, new_state );
		}
		var $body = $("body");
		$body.removeClass("gossipingClient monitoringClient monitoringQueuedClient monitoringWaitingClient coachingClient");

		var label = "<%=_T('Monitoring')%>";
		var monitoring_class = "monitoringClient";

		if ( this.from_app === "interaction" ) {
			if ( (this.m_interaction.isTaken() && !this.m_interaction.belongsToMe()) || (this.m_interaction.isFinished() && !this.m_interaction.archivedBelongsToMe()) ) {
				label = "<%=_T('Owned by another user')%>";
				monitoring_class = "gossipingClient";
				this.__disable_chat_actions();
			}
		}

		//a conversation monitored may have no assigned agent, maybe it's on the IVR yet
		//if ( !this.m_interaction.assignedAgents || this.m_interaction.assignedAgents.length === 0 ) {
		if ( this.m_interaction && this.m_interaction.isQueued() ) {
			monitoring_class = "monitoringQueuedClient";
		}

		var $state_flap = $(".clientStateFlap");
		switch( new_state )
		{
			case "coaching":
				$state_flap.text("<%=_T('Coaching')%>");
				$body.addClass(monitoring_class);
				$body.addClass("coachingClient");
			break;

			case "monitoring":
				$state_flap.text(label);
				$body.addClass(monitoring_class);
			break;
		}
	};

	MonitorClientWindow.prototype.on_before_transfer_execute = function(transfer_to)
	{
		const transfer_to_user = this.m_interaction && (this.m_interaction.isInflated() || this.inflating) && this.from_app === "supervisor" && transfer_to && transfer_to.to === "member" && transfer_to.user;
		/*
			Cuando desde monitoreo se trasnfiere asi mismo, la interaccion queda con la vista de monitoreo en estado ended, lo que provoca errores cuando se crea la nueva session.
			Para esto se cierra la ventana de monitoreo que tambien hace un stop del monitoreo.
		*/
		if (transfer_to_user && transfer_to_user === inConcert.app.getLoggedUserId() && this.m_interaction.isAssigned()) {
			this.close();
		}
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Queued monitor
	//
	function QueuedMonitorClientWindow(interaction, options)
	{
		options = options || {};
		options.monitor = true;
		// window title
		this.m_window_title = {
			normal: "<%=_T('{0}')%>",
			minimized_unread: "<span class='minimizedUnreadCounter'>{0}</span>{1}"
		};
		QueuedMonitorClientWindow.uber.constructor.call(this, interaction, options);

		if (!interaction) {
			return;
		}

		var $coach_agent_button = $('#coachAgent');
		var $owner_node = $('#ownerSection');
		$coach_agent_button.hide();
		$owner_node.hide();
		this.__state = "unknown"; // possible states: unknown, monitoring, coaching, ended
		// default conversation monitor on startup is monitor only (no coaching)
		this.__monitorInteraction(this.m_interaction);
	}

	inConcert.Utils.inherit(QueuedMonitorClientWindow, LiveClientWindow);

	QueuedMonitorClientWindow.prototype.__disable_chat_actions = function()
	{
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		$('#assignLabels, #addComments').hide();
		$('#coachAgent').remove();
	};

	QueuedMonitorClientWindow.prototype.__disable_coach_actions = function()
	{
		$('#coachMessageText, #coachSendMessage').hide();
	};

	/**
	Handles error that can be dealt with in a generic way.
	@param message
	@param response
	*/
	QueuedMonitorClientWindow.prototype.__handle_chat_errors = function(message, response) {
		var self = this;

		switch(response.reason) {
			case 'NONEXISTENT_SESSION':
				// if the conversation has ended, show a read-only view to the user
				self.change_client_state("ended");

				self.displayStatusMessage("<%=_T('The conversation has ended.')%>", 'Info');
				self.__disable_chat_actions();
				// Go and search repository.
				self.loadTransferrerSessionsMessages(self.m_interaction, function() {
					self.getInteractionFromRepository(self.m_interaction, function(session) {
						if (session) {
							// append messages
							self.populateListWithSessionMessages(self.m_interaction, session);
						}
					});
				});
				break;
			case 'USER_NOT_FOUND':
				// the user does not have valid credentials,
				// try to relogin, if the relogin fail, logout from agent, everything is broken.
				inConcert.app.getInteractionManager().reconnectClient('CHAT', function(ok) {
					if (!ok) {
						inConcert.app.onUnrecoverableError();
					}
				});
				break;
			case 'NETWORK_ERROR':
				self.displayStatusMessage(format(_T('There is a network error: {0}', response.message), 'Error'));
				self.__disable_chat_actions();
				break;
			default:
				self.displayStatusMessage(message + (response.message ? response.message : response.reason), 'Error', true);
				self.__disable_chat_actions();
				break;
		}
	};

	QueuedMonitorClientWindow.prototype.__monitorInteraction = function(interaction)
	{
		var self = this;

		this.toggleSendMessage(false);

		inConcert.app.getInteractionManager().startMonitoringInteraction(interaction, null, function(result, response)
		{
			if (result) {
				self.__state = "monitoring";
				self.change_client_state("monitoring");
			}
			else {
				if (!response || !response.reason) {
					self.displayStatusMessage("<%=_T('Unable to monitor queued conversation: Unhandled error')%>", 'Error', true);
					self.__disable_chat_actions();
					return;
				}
				self.__handle_chat_errors("<%=_T('Unaeble to monitor queued conversation: ')%>", response);
			}
		});
	};

	/**
	The session had closed. If coaching, hide its panel.
	*/
	QueuedMonitorClientWindow.prototype.onSessionClosed = function()
	{
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		$('#assignLabels, #addComments').hide();
		$('#coachAgent').remove();
		this.change_client_state("ended");
		this.__state = "ended";
	};

	/**
	Before closing the window, we need to stop either monitoring or coaching. Doing so in onWindowClosed is too late.
		@param params
	*/
	QueuedMonitorClientWindow.prototype.onWindowClosing = function(params)
	{
		inConcert.app.getInteractionManager().stopMonitoringInteraction(this.m_interaction, $.noop);
		return QueuedMonitorClientWindow.uber.onWindowClosing.call(this, params);
	};

	/**
	When the monitoring window is closed automatically stop monitoring or coaching the
	current conversation since there is no other handler to this window to do so later
	*/
	QueuedMonitorClientWindow.prototype.onWindowClosed = function()
	{
		inConcert.app.getInteractionManager().stopMonitoringInteraction(this.m_interaction, $.noop);
		return QueuedMonitorClientWindow.uber.onWindowClosed.call(this);
	};

	QueuedMonitorClientWindow.prototype.sendCoachMessage = $.noop; // empty base class function

	QueuedMonitorClientWindow.prototype.change_client_state = function( new_state )
	{
		if ( new_state === "ended" ) {
			this.close();
			return;
		}
		if ( new_state === "connected" ) {
			return;
		}
		if ( new_state !== "monitoring" ) {
			new_state = MonitorClientWindow.uber.change_client_state.call( this, new_state );
		}
		var $body = $("body");
		$body.removeClass("gossipingClient monitoringClient monitoringQueuedClient monitoringWaitingClient coachingClient");
		var $state_flap = $(".clientStateFlap");

		//a conversation monitored may have no assigned agent, maybe it's on the IVR yet
		var monitoring_class, label;
		if ( new_state === "monitoring" && (!this.m_interaction.assignedAgents || this.m_interaction.assignedAgents.length === 0) ) {
			label = "<%=_T('Queued')%>";
			if (this.from_app === "supervisor") {
				monitoring_class = "monitoringQueuedClient";
			}
			else {
				monitoring_class = "interactionQueued";
			}
		}
		else {
			label = "<%=_T('Monitoring')%>";
			monitoring_class = "monitoringClient";
		}

		switch( new_state )
		{
			case "coaching":
				$state_flap.text("<%=_T('Coaching')%>");
				$body.addClass(monitoring_class);
				$body.addClass("coachingClient");
			break;

			case "monitoring":
				$state_flap.text(label);
				$body.addClass(monitoring_class);
			break;
		}
	};

	////////////////////////////////////////////////////////////////////////////////////////////////////////
	// Waiting monitor
	//
	function WaitingMonitorClientWindow(interaction, options)
	{
		options = options || {};
		options.monitor = true;
		// window title
		this.m_window_title = {
			normal: "<%=_T('{0}')%>",
			minimized_unread: "<span class='minimizedUnreadCounter'>{0}</span>{1}"
		};
		WaitingMonitorClientWindow.uber.constructor.call(this, interaction, options);

		if (!interaction) {
			return;
		}

		var $coach_agent_button = $('#coachAgent');
		var $owner_node = $('#ownerSection');
		$coach_agent_button.hide();
		$owner_node.hide();
		this.__state = "unknown"; // possible states: unknown, monitoring, coaching, ended
		// default conversation monitor on startup is monitor only (no coaching)
		this.__monitorInteraction(this.m_interaction);
	}

	inConcert.Utils.inherit(WaitingMonitorClientWindow, LiveClientWindow);

	WaitingMonitorClientWindow.prototype.__disable_chat_actions = function()
	{
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		$('#assignLabels, #addComments').hide();
		$('#coachAgent').remove();
	};

	WaitingMonitorClientWindow.prototype.__disable_coach_actions = function()
	{
		$('#coachMessageText, #coachSendMessage').hide();
	};

	/**
	Handles error that can be dealt with in a generic way.
	@param message
	@param response
	*/
	WaitingMonitorClientWindow.prototype.__handle_chat_errors = function(message, response) {
		var self = this;

		switch(response.reason) {
			case 'NONEXISTENT_SESSION':
				// if the conversation has ended, show a read-only view to the user
				self.change_client_state("ended");

				self.displayStatusMessage("<%=_T('The conversation has ended.')%>", 'Info');
				self.__disable_chat_actions();
				// Go and search repository.
				self.loadTransferrerSessionsMessages(self.m_interaction, function() {
					self.getInteractionFromRepository(self.m_interaction, function(session) {
						if (session) {
							// append messages
							self.populateListWithSessionMessages(self.m_interaction, session);
						}
					});
				});
				break;
			case 'USER_NOT_FOUND':
				// the user does not have valid credentials,
				// try to relogin, if the relogin fail, logout from agent, everything is broken.
				inConcert.app.getInteractionManager().reconnectClient('CHAT', function(ok) {
					if (!ok) {
						inConcert.app.onUnrecoverableError();
					}
				});
				break;
			case 'NETWORK_ERROR':
				self.displayStatusMessage(format("<%=_T('There is a network error: {0}')%>", response.message), 'Error');
				self.__disable_chat_actions();
				break;
			default:
				self.displayStatusMessage(message + (response.message ? response.message : response.reason), 'Error', true);
				self.__disable_chat_actions();
				break;
		}
	};

	WaitingMonitorClientWindow.prototype.__monitorInteraction = function(interaction)
	{
		var self = this;

		this.toggleSendMessage(false);

		inConcert.app.getInteractionManager().startMonitoringInteraction(interaction, null, function(result, response)
		{
			if (result) {
				self.__state = "monitoring";
				self.change_client_state("monitoring");
			}
			else {
				if (!response || !response.reason) {
					self.displayStatusMessage("<%=_T('Unable to monitor waiting conversation: Unhandled error')%>", 'Error', true);
					self.__disable_chat_actions();
					return;
				}
				self.__handle_chat_errors("<%=_T('Unable to monitor waiting conversation: ')%>", response);
			}
		});
	};

	/**
	The session had closed. If coaching, hide its panel.
	*/
	WaitingMonitorClientWindow.prototype.onSessionClosed = function()
	{
		this.showPeerStoppedTyping();
		this.toggleSendMessage(false);
		$('#assignLabels, #addComments').hide();
		$('#coachAgent').remove();
		this.change_client_state("ended");
		this.__state = "ended";
	};

	/**
	Before closing the window, we need to stop either monitoring or coaching. Doing so in onWindowClosed is too late.
	@param params
	*/
	WaitingMonitorClientWindow.prototype.onWindowClosing = function(params)
	{
		inConcert.app.getInteractionManager().stopMonitoringInteraction(this.m_interaction, $.noop);
		return WaitingMonitorClientWindow.uber.onWindowClosing.call(this, params);
	};

	/**
	When the monitoring window is closed automatically stop monitoring or coaching the
	current conversation since there is no other handler to this window to do so later
	*/
	WaitingMonitorClientWindow.prototype.onWindowClosed = function()
	{
		inConcert.app.getInteractionManager().stopMonitoringInteraction(this.m_interaction, $.noop);
		return WaitingMonitorClientWindow.uber.onWindowClosed.call(this);
	};

	WaitingMonitorClientWindow.prototype.sendCoachMessage = $.noop; // empty base class function

	WaitingMonitorClientWindow.prototype.change_client_state = function( new_state )
	{
		if ( new_state === "connected" ) {
			return;
		}
		if ( new_state !== "monitoring" ) {
			new_state = MonitorClientWindow.uber.change_client_state.call( this, new_state );
		}
		var $body = $("body");
		$body.removeClass("gossipingClient monitoringClient monitoringQueuedClient monitoringWaitingClient coachingClient");
		var $state_flap = $(".clientStateFlap");

		//a conversation monitored may have no assigned agent, maybe it's on the IVR yet
		var monitoring_class, label;
		if ( new_state === "monitoring" && (!this.m_interaction.assignedAgents || this.m_interaction.assignedAgents.length === 0) ) {
			label = "<%=_T('Waiting')%>";
			if (this.from_app === "supervisor") {
				monitoring_class = "monitoringWaitingClient";
			}
			else {
				monitoring_class = "interactionWaiting";
			}
		}
		else {
			label = "<%=_T('Monitoring')%>";
			monitoring_class = "monitoringClient";
		}

		switch( new_state )
		{
			case "coaching":
				$state_flap.text("<%=_T('Coaching')%>");
				$body.addClass(monitoring_class);
				$body.addClass("coachingClient");
			break;

			case "monitoring":
				$state_flap.text(label);
				$body.addClass(monitoring_class);
			break;
		}
	};

	////////////////////////////////////////////////////////////////////////////////////
	//when jquery says so, start it all up
	$(document).ready(function() {

		window.hook_environment();
		inConcert.i18n.updateScreen();

		var callback = function() {
			let client;
			// TODO: #9778 - CANY - Docked.
			let options = {from_app: window.from_app, panel_name: window.panel_name};
			if (window.archive) {
				client = new ArchivedClientWindow( window.interaction, options );
			}
			else if (window.queued) {
				client = new QueuedMonitorClientWindow( window.interaction, options );
			}
			else if (window.waiting) {
				client = new WaitingMonitorClientWindow( window.interaction, options );
			}
			else if(window.monitor || window.interaction.isBotInternalConsultation()) {
				client = new MonitorClientWindow( window.interaction, options );
			}
			else if (window.wrapup){
				client = new WrapupClientWindow( window.interaction, options );
			}
			else {
				var interactionType = window.interaction.type;
				if (interactionType === 'IM') {
					client = new LiveIMClientWindow( window.interaction, options );
				}
				else if (interactionType === 'CHAT') { // web chat
					client = new LiveWebChatClientWindow( window.interaction, options );
				}
				else if (interactionType === 'INTERNAL_CHAT') {
					client = new LiveInternalChatClientWindow( window.interaction, options );
				}
				else {
					throw 'Chat type "' + interactionType + '" does not exit!!';
				}
			}

			window.onReady(client);

			if (window && window.view && window.view instanceof inConcert.ClientWindowProxy) {
				window.view.apply(client);
			}
			window.view = client;

			// Prevent middle mouse button clicks.
			// In Chrome, at least with 30.0.1599.101 m, starting a scroll with the middle button will "freeze".
			// Everything would still work except that you can't click anywhere. It seems like the mouseup event
			// got lost. This has something to do with iframes.
			// There are bug reports here:
			//
			//   https://code.google.com/p/chromium/issues/detail?id=308989
			//   https://code.google.com/p/chromium/issues/detail?id=269917
			//
			$("body").mousedown(function(e){
				/* 0=left, 1=middle, 2=right*/
				if( e.button === 1 ) {
					return false;
				}
			});
		};

		if(window.interaction.isBeingComposed()) {
			callback();
		}
		else {
			window.interaction.inflate(callback);
		}
	});

}());
