/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, quotmark:double, esversion: 9 */

/*global $, inConcert, _ */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
	this.inConcert = {};
}

(function () {

	"use strict";

	const format = inConcert.Utils.format;

	var MakeSearchRegExp = function(word) {
		if (typeof word !== "string") {
			return "";
		}
		let search_pattern = inConcert.Utils.removeSpecialChars(word);
		search_pattern = search_pattern.replace("+", "");
		return new RegExp(search_pattern.replace(/[^\w\s]/gi, ""), "i");
	};

	/**
	Class ContactSelectorModel.
	*/
	function ContactSelectorModel (addressType, getInteractionCampaignFn)
	{
		this.m_rowsPerPage = 30;
		this.m_iteratorIndex = 0;
		this.m_lastSearchStart = "";
		this.m_noMoreContacts = false;
		this.m_filterRegex = null;
		this.m_lastSearchQuery = "";
		this.m_currentRequest = null;

		this.m_contacts = [];
		this.m_addressType = addressType;
		this.getInteractionCampaignFn = getInteractionCampaignFn;
		this.m_contactManager = new inConcert.ContactManager(inConcert.app.getBaseUrl());
	}

	//Define functions and objects in the namespace in case they don't already exist
	if (typeof inConcert.ContactSelectorModel !== "function") {
		inConcert.ContactSelectorModel = ContactSelectorModel;
	}

	ContactSelectorModel.prototype.filterFunction = function(/*contact*/) {
		return true;
	};

	ContactSelectorModel.prototype.getAddressType = function() {
		return this.m_addressType;
	};

	ContactSelectorModel.prototype.filterFromCache = function(pattern, onSearchComplete) {
		this.m_iteratorIndex = 0;
		this.m_noMoreContacts = false;
		this.m_filterRegex = new RegExp(pattern, "i");

		if (typeof(onSearchComplete) === "function") {
			onSearchComplete(true);
		}
	};

	/*!
	 * \private
	 *
	 * Initializes contact iterator
	 *
	 * This function is intended to be called inside the \c search() function
	 */
	ContactSelectorModel.prototype.onBeforeSearch = function(pattern) {
		this.m_iteratorIndex = 0;
		this.m_lastSearchStart = "";
		this.m_noMoreContacts = false;
		this.m_filterRegex = null;
		this.m_lastSearchQuery = pattern || "";
		this.m_searchCanceled = false;
	};

	/*!
	 * This functions searchs for those contacts whose name matches
	 * the patter \p pattern.
	 *
	 * When the request finished \p onSearchComplete is called with no
	 * arguments, the user shall call \c getNextContact to iterate over
	 * the contact list.
	 */
	ContactSelectorModel.prototype.search = function(pattern, onSearchComplete) {
		var self = this;

		this.onBeforeSearch(pattern);
		var filter = {
			last : self.m_lastSearchStart,
			rows  : self.m_rowsPerPage,
			query : self.m_lastSearchQuery
		};
		self.m_contactManager.search(filter, function(contacts, lastContact) {
			self.m_lastSearchStart = lastContact;
			if (contacts) {
			//only clear on successful search, event when result is empty
				self.clearContacts();
			}

			// appendContacts calls onSearchComplete()
			self.appendContacts(contacts || [], onSearchComplete);
		});
	};

	ContactSelectorModel.prototype.cancelSearch = function() {
		this.m_searchCanceled = true;
		this.m_contactManager.cancelCurrentRequest();
	};

	/*!
	 * \private
	 *
	 * Incremental contact search, searches from contacts starting
	 * from the last contact found
	 */
	ContactSelectorModel.prototype.incrementalSearch = function(onSearchComplete) {
		var self = this;

		if(self.m_lastSearchStart.length === 0 || self.m_searchCanceled)  {
			onSearchComplete(false);
			return;
		}

		// TODO Fix this with issue 1259: http://redmine.inconcert/issues/1259
		if(self.m_contacts.length > 200) {
			onSearchComplete(false);
			return;
		}

		var filter = {
			last : self.m_lastSearchStart,
			rows  : self.m_rowsPerPage,
			query : self.m_lastSearchQuery
		};
		self.m_contactManager.search(filter, function(contacts, lastContact) {
			self.m_lastSearchStart = lastContact;
			contacts = contacts && contacts.length > 0 ? contacts : null;
			if( contacts ) {
				self.appendContacts(contacts);
			}
			if ( typeof(onSearchComplete) === "function" ) {
				onSearchComplete(contacts ? true : false);
			}
		});
	};

	/*!
	 * Returns the next contact, if there are no more contacts on the list
	 * then more contacts are pulled from the server, if there are no more contacts
	 * in the serverto this function calls contactHandler(false)
	 *
	 * \note \c search() should be called first
	 *
	 * \param contactHandler a function which gets a boolean as first parameter, and
	 * a contact as second
	 *
	 * \example
	 *
	 * \code
	 *		contacts.search("lastrator", function() {
	 *			var contactCount = 0;
	 *			var showNextContact = function(ok, contact) {
	 *				if ( ok === false ) {
	 *					alert(contactCount > 0 ? "No more contacts" : "No contacts found");
	 *				} else {
	 *					contactCount++;
	 *					alert("Got contact: " + contact.Name);
	 *					contacts.getNextContact(showNextContact);
	 *				}
	 *			}
	 *
	 *		contacts.getNextContact(showNextContact);
	 *	});
	 *\endcode
	 */
	ContactSelectorModel.prototype.getNextContact = function(contactHandler) {
		var self = this;

		// This function calls contactHandler
		var invokeCallback = function(withContact) {
			if (typeof(contactHandler) === "function") {
				if (withContact) {
					var contactIndex = self.m_iteratorIndex++;
					contactHandler(true, self.m_contacts[contactIndex]);
				} else {
					contactHandler(false);
				}
			}
		};

		// This function looks for the next contact to return, aplying a filter when required
		var findNextValidIndex = function() {
			if (self.m_filterRegex !== null) {
				for ( ; self.m_iteratorIndex < self.m_contacts.length; ++self.m_iteratorIndex ) {
					var contact = self.m_contacts[self.m_iteratorIndex];
					if (contact.Name.match(self.m_filterRegex) ||
						(contact.contactAddress.length > 0 &&
						contact.contactAddress.match(self.m_filterRegex))
					) {
						break;
					}
				}
			}

			return self.m_iteratorIndex;
		};

		// This function searchs from more contacts in the server
		var searchMoreContacts = function() {
			self.incrementalSearch(function(gotContacts) {
				if ( !gotContacts ) {
					self.m_noMoreContacts = true;
					invokeCallback(false);
					return;
				}

				if ( findNextValidIndex() < self.m_contacts.length) {
					invokeCallback(true);
				} else {
					searchMoreContacts();
				}
			});
		};

		if (self.m_searchCanceled || (self.m_iteratorIndex === 0 && self.m_contacts.length === 0) || self.m_noMoreContacts) {
			invokeCallback(false);
		}

		if ( findNextValidIndex() < self.m_contacts.length) {
			invokeCallback(true);
		}
		else {
			searchMoreContacts();
		}
	};

	ContactSelectorModel.prototype.clearContacts = function() {
		var self = this;
		self.m_contacts = [];
		if (typeof(self.onContactsCleared) === "function") {
			self.onContactsCleared();
		}
	};

	ContactSelectorModel.prototype.getOnlineStatus = function(contact) {
		return contact.getOnlineStatus();
	};

	ContactSelectorModel.prototype.getAvatar = function(contact) {
		return inConcert.app.getBaseUrl() + contact.getAvatarURI();
	};

	ContactSelectorModel.prototype.getDisplayName = function(contact) {
		return contact.Name;
	};

	ContactSelectorModel.prototype.getContactAddress = function(contact) {
		return contact.contactAddress;
	};

	ContactSelectorModel.prototype.getContactDescription = function(contact) {
		return contact.Description;
	};

	ContactSelectorModel.prototype.getContacts = function() {
		return this.m_contacts;
	};

	ContactSelectorModel.prototype.appendContacts = function(contacts, onSearchComplete) {
		var self = this;
		inConcert.logger.action({action_name:"appendContacts", count:contacts.length});
		for (var i = 0; i < contacts.length; i++) {
			var contact = contacts[i];
			if (self.filterFunction(contact)) {
				//if no address returned assume this contact has no address for contact on this technology type
				var bestAddress = self.findBestAddress(contact);
				if (typeof(bestAddress) !== "undefined") {
					contact.contactAddress = bestAddress;
					if (self.m_addressType && typeof(contact.addressType) === "undefined") {
						contact.addressType = self.m_addressType;
					}
					self.m_contacts.push(contact);
				}
			}
		}

		if (typeof(onSearchComplete) === "function") {
			onSearchComplete(!self.m_searchCanceled);
		}
	};

	ContactSelectorModel.prototype.findBestAddress = function(contact) {
		//if the user have not asked for a specific address type all contacts match
		if (!this.m_addressType) {
			return "";
		}
		var addressType = this.m_addressType.toLowerCase();
		for (var i = 0; i < (contact.addresses || []).length; i++) {
			var address = contact.addresses[i];
			if (address.Type.toLowerCase() === addressType) {
				return contact.getAddressDisplayText(address);
			}
		}
		return; // address is undefined, contact does not match
	};

	ContactSelectorModel.prototype.mustAddAllRowsClientSide = function() {
		return false;
	};

	/**
	Class FavoritesSelectorModel.
	*/
	function FavoritesSelectorModel (addressType)
	{
		FavoritesSelectorModel.uber.constructor.call(this, addressType, null);
	}

	inConcert.Utils.inherit(FavoritesSelectorModel, ContactSelectorModel);

	if (typeof inConcert.FavoritesSelectorModel !== "function") {
		inConcert.FavoritesSelectorModel = FavoritesSelectorModel;
	}

	FavoritesSelectorModel.prototype.search = function(pattern, onSearchComplete) {
		const self = this;
		this.onBeforeSearch(pattern);
		inConcert.app.getContactManager().getFavorites(function(favorites) {
			if( !favorites ) {
				favorites = [];
			}
			else if (self.m_lastSearchQuery.length > 0) {
				var index = 0;
				var regex = new RegExp(self.m_lastSearchQuery, "i");
				while ( index < favorites.length) {
					if (!favorites[index].Name.match(regex)) {
						favorites.splice(index, 1);
					} else {
						index++;
					}
				}
			}
			self.clearContacts();
			self.appendContacts(favorites, onSearchComplete);
		});
	};

	FavoritesSelectorModel.prototype.incrementalSearch = function(onSearchComplete) {
		if ( typeof(onSearchComplete) === "function") {
			onSearchComplete(false);
		}
	};

	/**
	Class TeamMemberSelectorModel.
	*/
	function TeamMemberSelectorModel (addressType, interactions)
	{
		this.m_interactions = _.flatten([interactions]);
		TeamMemberSelectorModel.uber.constructor.call(this, addressType, null);
	}

	inConcert.Utils.inherit(TeamMemberSelectorModel, ContactSelectorModel);

	if (typeof inConcert.TeamMemberSelectorModel !== "function") {
		inConcert.TeamMemberSelectorModel = TeamMemberSelectorModel;
	}

	TeamMemberSelectorModel.prototype.search = function(pattern, onSearchComplete) {
		var self = this;
		self.onBeforeSearch(pattern);

		var loaderFn = function(members, checkOnline) {
			if( !members ) {
				members = [];
			}
			self.clearContacts();
			inConcert.logger.action({action_name:"loadTeamMembers", count:members.length});
			//only online members that pass the filter are good enough
			var passedList = [];
			var regex = new RegExp(self.m_lastSearchQuery, "i");
			for (var i = 0; i < members.length; i++) {
				var member = members[i];
				if (!checkOnline || member.isOnline()) {
					if (self.m_lastSearchQuery.length === 0 || member.Name.match(regex)) {
						if (typeof(member.addressType) === "undefined") {
							member.addressType = "agent"; // override address type
						}
						passedList.push(member);
					}
				}
			}
			self.appendContacts(passedList, onSearchComplete);
		};
		if (self.m_interactions) {
			var ids = _.pluck(self.m_interactions, "id");
			inConcert.logger.action({ action_name: "getMembersForTransfer", interactions: ids });
			self.m_currentRequest = inConcert.app.getInteractionManager().getMembersForTransfer(ids, function(members) {
				self.m_currentRequest = null;
				if (members) {
					loaderFn(members, false);
				}
			});
		} else {
			//TODO this should be called pre filtered for performance, now its retrieving the fucking list every fucking time
			inConcert.logger.action({action_name:"getTeamMembers"});
			self.m_currentRequest = inConcert.app.getTeamMembers(function(members) {
				self.m_currentRequest = null;
				loaderFn(members, true);
			});
		}
	};

	TeamMemberSelectorModel.prototype.cancelSearch = function() {
		this.m_searchCanceled = true;
		var current_request = this.m_currentRequest;
		if ( current_request ) {
			current_request.abort();
		}
	};

	TeamMemberSelectorModel.prototype.incrementalSearch = function(onSearchComplete) {
		if ( typeof(onSearchComplete) === "function") {
			onSearchComplete(false);
		}
	};

	TeamMemberSelectorModel.prototype.findBestAddress = function(member) {
		return member.Id;
	};

	TeamMemberSelectorModel.prototype.getAvatar = function(member) {
		return inConcert.app.getBaseUrl() + "/users/get_avatar/" + member.Id;
	};

	TeamMemberSelectorModel.prototype.getOnlineStatus = function(member) {
		return member.isOnline();
	};


	/**
	Class AddressesBookSelectorModel.
	*/
	function AddressesBookSelectorModel (channel)
	{
		this.channel = channel.toUpperCase();
		const address_type_by_channel = {
			"CALL"    : "Phone",
			"SMS"     : "Phone",
			"WHATSAPP": "Phone",
			"MAIL"    : "Mail"
		};

		let addressType = address_type_by_channel[this.channel] || this.channel;
		AddressesBookSelectorModel.uber.constructor.call(this, addressType, null);
	}

	inConcert.Utils.inherit(AddressesBookSelectorModel, ContactSelectorModel);

	if (typeof inConcert.AddressesBookSelectorModel !== "function") {
		inConcert.AddressesBookSelectorModel = AddressesBookSelectorModel;
	}

	AddressesBookSelectorModel.prototype.search = function(_, onSearchComplete = $.noop) {
		const self = this;
		//this.onBeforeSearch(pattern);
		inConcert.app.getContactManager().getAddressesBookByChannel(this.channel, function(addressesBook) {
			addressesBook = addressesBook || [];
			let sortedAddressesBook = inConcert.Utils.sortArray(addressesBook, "Name");
			self.clearContacts();
			self.appendContacts(sortedAddressesBook, function() {
				const ok = true;
				onSearchComplete(ok);
			});
		});
	};

	AddressesBookSelectorModel.prototype.incrementalSearch = function(onSearchComplete) {
		if ( typeof(onSearchComplete) === "function") {
			onSearchComplete(false);
		}
	};

	AddressesBookSelectorModel.prototype.findBestAddress = function(addressesBook) {
		const fullPhoneNumber = true;
		return addressesBook.getAddressDisplayText(addressesBook, fullPhoneNumber);
	};

	AddressesBookSelectorModel.prototype.getDisplayName = function(addressesBook) {
		return addressesBook.Name || "";
	};

	AddressesBookSelectorModel.prototype.getAvatar = function(/*addressesBook*/) {
		return inConcert.app.getBaseUrl() + "/images/address-book-icon.png";
	};

	AddressesBookSelectorModel.prototype.mustAddAllRowsClientSide = function() {
		return true;
	};

	/**
	Class ContactSelector.
	*/
	function ContactSelector (node, addressType, createContactFn, getCampaignInteractionFn)
	{
		this.parent = node;
		ContactSelector.uber.constructor.call(this);

		this.clickEventName = this.clickEventName + "contact";
		this.ownerId = this.parent.attr("id");
		this.canReposition = true;

		this.createContactFn = createContactFn;
		this.getCampaignInteractionFn = getCampaignInteractionFn;
		this.addressType = addressType;
		this.parent.hide();

		this.m_contactManager = new inConcert.ContactManager(inConcert.app.getBaseUrl());
		this.m_pageStart = "";
		this.m_rowsPerPage = 40;
		this.m_scrollOnPercentage = 80;
	}

	inConcert.Utils.inherit( ContactSelector, inConcert.PopupWindow );

	if (typeof inConcert.ContactSelector !== "function") {
		inConcert.ContactSelector = ContactSelector;
	}

	ContactSelector.prototype.close = function() {
		var self = this;
		if (parent.window.view && (typeof parent.window.view.onContactsClose==="function")) {
			parent.window.view.onContactsClose();
		} else {
			inConcert.logger.guiAction({action_name : "close_contact_window", window_id: window.id});
		}
		this.parent.children().remove();
		self.finalize();
		self.onClose();
		self.parent.hide();
	};

	ContactSelector.prototype.filterFunction = function(/*contact*/) {
		return true;
	};

	ContactSelector.prototype.findBestAddress = function(contact, dealingWithAgents) {
		if (dealingWithAgents) {
			return (contact.Id + "@" + contact.VccGroup);
		}
		//if the user have not asked for a specific address type all contacts match
		if (!this.addressType) {
			return "";
		}
		var addressType = this.addressType.toLowerCase();
		for (var i = 0; i < (contact.addresses || []).length; i++) {
			var address = contact.addresses[i];
			if (address.Type.toLowerCase() === addressType) {
				return contact.getAddressDisplayText(address);
			}
		}
		return;
	};

	ContactSelector.prototype.createMarkup = function() {
		let title = this.m_title;
		if (!title) {
			if (this.getCampaignInteractionFn) {
				const campaignInteraction = this.getCampaignInteractionFn();
				if (campaignInteraction) {
					const campaign = inConcert.app.getCampaign(campaignInteraction);
					title = format("<%=_T('Choose Contact from {0}')%>", campaign.getDescription());
				}
			}
			if (!title) {
				title = "<%=_T('Choose Contact')%>";
			}
		}
		this.parent.children().remove();
		let markup = "" +
			"<div class='windowTop'>" +
				"<div class='windowTitle windowMember'>" +
					"<span>" + title + "</span>" +
					"<img src='/inconcert/apps/agent/images/blank.png' class='windowAddAddresseeButClose' hidden/>" +
				"</div>" +
				"<div class='windowControl'>" +
					"<input class='windowControlField contactSearchField' id='selectorSearchField'/>" +
					"<button class='butControlRefresh buttonColorGray' type='button' value=''></button>" +
				"</div>" +
			"</div>" +
			"<div class='windowMiddle'>" +
				"<div waiting_container='true' class='windowAddAddresseeListContainer'>" +
					"<span class='windowTargetNoResultsText'>" + "<%=_T('Loading, wait moment please')%>" + "</span>" +
				"</div>" +
				"<div contact_list_container='true' class='windowAddAddresseeListContainer' style='display:none;'>" +
					"<div contact_list_area_container='true' class='windowAddAddresseeListScrollArea'></div>" +
				"</div>" +
				"<div not_found_container='true' class='windowTargetNoResults' style='display:none;'>" +
					"<span class='windowTargetNoResultsText'>" + "<%=_T('No results found.')%>" + "</span>" +
				"</div>" +
			"</div>";
		this.parent.append(markup);
		if (this.createContactFn) {
			this.parent.addClass("windowAddNewAddressee");
		}
		else {
			this.parent.removeClass("windowAddNewAddressee");
		}
	};

	ContactSelector.prototype.search = function(pattern, addresses_book_contacts, callback = $.noop) {
		const self = this;
		let withAddressesBook = typeof(addresses_book_contacts) !== "undefined";

		let filterText = pattern || "";
		this.m_lastSearch = filterText;
		let filter = {
			last : "",
			rows : 30,
			query : filterText
		};

		this.m_contactManager.search(filter,
			function(contacts/*, lastContact*/) {
				//only clear on successful search
				self.clearListContainer();
				if (!contacts && !addresses_book_contacts) {
					callback();
					return;
				}

				let contacts_list = [];
				let unordered = false;

				if (withAddressesBook) {
					for (let i = 0; i < (addresses_book_contacts || []).length; i++) {
						contacts_list.push(addresses_book_contacts[i]);
					}
					// (**) Si tenemos que incluir la lista addresses book la misma ya llega ordenada, pero la lista contactos (no la de address book)
					//  se tiene que ordenar aca y no en el proceso de Append de la clase base.
					//  Esto porque se visualizan ambas listas (address books y contactos) ya que primero tienen que ir las addresses book y despues los contactos.
					unordered = true;
					contacts = inConcert.Utils.sortArray(contacts, "Name");
				}
				for (let i = 0; i < (contacts || []).length; i++) {
					contacts_list.push(contacts[i]);
				}
				self.appendContacts(contacts_list, undefined /*dealingWithAgents*/, unordered, function() {
					callback();
				});
			}
		);
	};

	ContactSelector.prototype.clearListContainer = function() {
		this.m_pageStart = "";
		this.parent.find("[contact_list_area_container]").children().remove();
	};

	ContactSelector.prototype.getOnlineStatus = function(contact) {
		return contact.getOnlineStatus();
	};

	ContactSelector.prototype.getAvatar = function(contact) {
		return inConcert.app.getBaseUrl() + contact.getAvatarURI();
	};

	ContactSelector.prototype.getContactName = function(contact)
	{
		let contact_name = contact.Id;
		let can_name = false;
		if ( ( contact.Name || "" ).trim().length > 0 ) {
			contact_name = contact.Name;
			can_name = true;
		}
		if ( ( contact.FirstSurname || "" ).trim().length > 0 ) {
			contact_name = ( can_name ? ( contact_name + " " ) : "" ) + contact.FirstSurname;
		}
		return contact_name;
	};

	ContactSelector.prototype.getContactType = function( contact )
	{
		let contact_type = (contact.contactType || "").toUpperCase();
		if (contact_type === "ADDRESSES BOOK") {
			return "<%=_T('Addresses Book')%>";
		}
		return "<%=_T('Contact')%>";
	};

	ContactSelector.prototype.appendContacts = function(contacts, dealingWithAgents, unordered, appendCompletedFn = $.noop)
	{
		const self = this;

		const $waiting_conatiner = this.parent.find("[waiting_container]");
		const $list_container = this.parent.find("[contact_list_container]");
		const $not_found_container = this.parent.find("[not_found_container]");

		var showWithTimeout = function (timeout_ms, callback = $.noop) {
			timeout_ms = timeout_ms || 300;
			window.clearTimeout(self.m_timeout_shown_hnd);
			self.m_timeout_shown_hnd = window.setTimeout(function() {
				window.clearTimeout(self.m_timeout_shown_hnd);
				self.m_timeout_shown_hnd = null;
				callback();
				appendCompletedFn();
			}, timeout_ms);
		};
		this.m_timeout_shown_hnd = null;

		if (!contacts || contacts.length === 0) {
			showWithTimeout(null, function() {
				$not_found_container.show();
				$waiting_conatiner.hide();
				$list_container.hide();
			});
			return;
		}

		var get_tooltip = function(contact) {
			if (!contact.Description) {
				return;
			}
			return contact.Description;
		};

		var on_finished = function() {
			if (total_added > 0) {
				showWithTimeout(null, function() {
					$list_container.show();
					$not_found_container.hide();
					$waiting_conatiner.hide();
					window.setTimeout(function() {
						self.setPosition();
					});
				});
			}
			else {
				showWithTimeout(null, function() {
					$waiting_conatiner.hide();
					$list_container.hide();
					$not_found_container.show();
					window.setTimeout(function() {
						self.setPosition();
					});
				});
			}
	
			$(".windowAddAddresseeListItem")
				.unbind("click.selected")
				.bind("click.selected", function() {
					const contact = $(this).data("contact");
					const address = $(this).data("address");
					if (self.onContactSelected) {
						self.onContactSelected(contact, address);
						self.close();
					}
			});
		};

		let total_added = 0;
		const $list = this.parent.find("[contact_list_area_container]");
		// (**) Inserto la lista ordena o desordenada, segun se haya especificado por parametros.
		let sorted_contacts = unordered ? contacts : inConcert.Utils.sortArray(contacts, "Name");

		// se carga en un foreach que corre dentro de un timeout para que el navegador no quede colgado
		inConcert.Utils.foreach_timeout({
			rows : sorted_contacts,
			max_append_per_loop: 500,
			pause_between_loop_ms: 2,
			append_fn : function( contact )
			{
				if (!self.filterFunction(contact)) {
					return;
				}
				//if no address returned assume this contact has no address for contact on this technology type
				let bestAddress = self.findBestAddress(contact, dealingWithAgents);
	
				// TO-DO: Why?
				if (typeof(bestAddress) === "undefined") {
					return;
				}
				const $contactNode = $(
					"<div class='windowAddAddresseeListItem'>" +
						"<img class='addresseeListItemAvatar' src='" + self.getAvatar(contact) + "' width='26' height='26' />" +
						"<label class='addresseeListItemName'>" + self.getContactName(contact) + "</label>" +
						"<label class='addresseeListItemAddress'>" + bestAddress + "</label>" +
						"<label class='addresseeListItemContactType'>" + self.getContactType(contact) + "</label>" +
					"</div>"
				);
				let tooltip = get_tooltip(contact);
				if (tooltip) {
					$contactNode.attr("title", tooltip);
				}
				$contactNode.data("contact", contact);
				$contactNode.data("address", bestAddress);
				$list.append($contactNode);
				total_added++;
			},
			waiting_fn : function(/*message*/) {
				if (total_added > 0) {
					let message = format("<%=_T('Please wait, adding {0} of many.')%>", total_added);
					$waiting_conatiner.text(message);
				}
			},
			complete_fn : function() {
				on_finished();
			}
		});
	};

	ContactSelector.prototype.setEvents = function() {
		const self = this;
		$(".windowAddAddresseeButClose").click(function() {
			self.close();
		});
		$("#selectorSearchField").keyup(function(/*event*/) {
			var text = $("#selectorSearchField").val();
			self.search(text);
		});

		this.parent.find(".butFormAddNew").click(function() {
			self.createContactFn(function(contact) {
				var address = self.findBestAddress(contact);
				if (typeof(address) !== "undefined") {
					self.onContactSelected(contact, address);
					self.close();
				}
				//TODO avisarle que creo un contacto sin una direccion valida para lo que queria??
			});
		});

		this.m_listContainer = self.parent.find("[contact_list_area_container]");
		// this function it is called when ocurred the event scroll
		var onScroll = function() {
			var scroll = this;
			var scrollTop = scroll.scrollTop;
			var scrollHeight = scroll.scrollHeight;
			var clientHeight = scroll.clientHeight;
			var scrolledPercentage = Math.round((scrollTop / (scrollHeight - clientHeight)) * 100);
			if (scrolledPercentage < self.m_scrollOnPercentage) {
				return;
			}
			//start from the current loaded
			self.m_listContainer.unbind("scroll");
			var filter = {
				start: self.m_pageStart,
				rows: self.m_rowsPerPage,
				query: self.m_lastSearch
			};
			self.m_contactManager.search(filter, function(contacts) {
				if (!contacts) {
					return;
				}
				self.appendContacts(contacts, undefined /*dealingWithAgents*/, undefined /*unordered*/, function() {
					self.m_listContainer.bind("scroll", onScroll);
				});
			});
		};
		// unbind/bind scroll event.
		this.m_listContainer.unbind("scroll").bind("scroll", onScroll);
	};

	ContactSelector.prototype.initialize = function() {
		var self = this;
		self.createMarkup();
		this.onContactSelected = self.data.onContactSelectedFn;
		this.search();
		this.parent.show();
		self.data.container = this.parent;
		self.setEvents();
		$("#selectorSearchField").focus();
	};


	/**
	Class TeamMemberSelector.
	*/
	function TeamMemberSelector (node, addressType, interactions, parameters)
	{
		this.m_title = "<%=_T('Choose Team Member')%>";
		this.m_interactions = _.flatten([interactions]);
		TeamMemberSelector.uber.constructor.call(this, node, addressType, undefined, undefined, parameters);
	}

	inConcert.Utils.inherit(TeamMemberSelector, ContactSelector);

	if (typeof inConcert.TeamMemberSelector !== "function") {
		inConcert.TeamMemberSelector = TeamMemberSelector;
	}

	TeamMemberSelector.prototype.search = function(pattern) {
		var self = this;
		var filterText = pattern || "";
		self.m_lastSearch = filterText;
		var loaderFn = function(members, checkOnline) {
			if ( !members ) {
				members = [];
			}
			self.clearListContainer();
			//only online members that pass the filter are good enough
			var passedList = [];
			var re = new RegExp(filterText, "i");
			for (var i = 0; i < members.length; i++) {
				var member = members[i];
				if (!checkOnline || member.isOnline()) {
					if (filterText.length === 0 || member.Name.match(re)) {
						passedList.push(member);
					}
				}
			}
			self.appendContacts(passedList, true /*dealingWithAgents*/);
		};
		if (self.m_interactions) {
			var ids = _.pluck(self.m_interactions, "id");
			inConcert.logger.action({ action_name:"getMembersForTransfer", interactions:ids });
			inConcert.app.getInteractionManager().getMembersForTransfer(ids, function(members) {
				if (members) {
					loaderFn(members, false);
				}
			});
		}
		else {
			//TODO this should be called pre filtered for performance, now its retrieving the fucking list every fucking time
			inConcert.app.getTeamMembers(function(members) {
				loaderFn(members, true);
			});
		}
	};

	TeamMemberSelector.prototype.findBestAddress = function(contact, dealingWithAgents) {
		if (dealingWithAgents) {
			return (contact.Id + "@" + contact.VccGroup);
		}
		return "";
	};

	TeamMemberSelector.prototype.getAvatar = function(member) {
		return inConcert.app.getBaseUrl() + "/users/get_avatar/" + member.Id;
	};

	TeamMemberSelector.prototype.getOnlineStatus = function(member) {
		return member.isOnline();
	};

	// --------------------------------------
	/**
	Class AutoSearchContactSelector.
	*/
	function AutoSearchContactSelector (node, addressType, interactions, parameters, search_value, addresses_book_model)
	{
		this.m_interactions = _.flatten([interactions]);
		this.m_search_value = null;
		this.m_addresses_book_model = addresses_book_model;
		this.m_cached_addresses_book_contacts = null;
		this.set_search_value(search_value);
		AutoSearchContactSelector.uber.constructor.call(this, node, addressType, undefined, undefined, parameters);
	}

	inConcert.Utils.inherit(AutoSearchContactSelector, ContactSelector);

	if (typeof inConcert.AutoSearchContactSelector !== "function") {
		inConcert.AutoSearchContactSelector = AutoSearchContactSelector;
	}

	AutoSearchContactSelector.prototype.get_search_value = function() {
		return this.m_search_value;
	};

	AutoSearchContactSelector.prototype.set_search_value = function(search_value) {
		this.m_search_value = search_value;
	};

	AutoSearchContactSelector.prototype.search = function() {
		const self = this;

		var search = function (addresses_book_contacts = []) {
			const search_value = self.get_search_value();
			self.m_filterRegex = MakeSearchRegExp(search_value);
			// (**) La lista de addresses book se tiene que ordenar aca y no en el proceso de Append de la clase base.
			//  Esto porque se visualizan ambas listas (address books y contactos) ya que primero tienen que ir las addresses book y despues los contactos.
			let sorted_addresses_book_contacts = inConcert.Utils.sortArray(addresses_book_contacts, "Name");
			AutoSearchContactSelector.uber.search.call(self, search_value, sorted_addresses_book_contacts, function() {
				// window.setTimeout(function() {
				// 	self.setPosition();
				// }, 600);
			});
		};

		if (this.m_addresses_book_model) {
			// Si tenemos libreta de direcciones, obtenemos las mismas y luego realizo el search contra los contactos.
			if (this.m_cached_addresses_book_contacts) {
				search(this.m_cached_addresses_book_contacts);
			}
			else {
				this.m_addresses_book_model.search("", function(success) {
					let addresses_book_contacts;
					if (success) {
						addresses_book_contacts = self.m_addresses_book_model.getContacts();
						self.m_cached_addresses_book_contacts = addresses_book_contacts;
					}
					search(addresses_book_contacts);
				});
			}
			return;
		}
		// SI NO tenemos libreta de direcciones, realizo el search solo contra los contactos.
		search();
	};

	AutoSearchContactSelector.prototype.createMarkup = function() {
		this.parent.children().remove();
		let markup = "" +
			"<div class='windowMiddle windowList'>" +
				"<div waiting_container='true' class='windowAddAddresseeListContainer'>" +
					"<span class='windowTargetNoResultsText'>" + "<%=_T('Loading, wait moment please')%>" + "</span>" +
				"</div>" +
				"<div contact_list_container='true' class='windowAddAddresseeListContainer' style='display:none;'>" +
					"<div contact_list_area_container='true' class='windowAddAddresseeListScrollArea'></div>" +
				"</div>" +
				"<div not_found_container='true' class='windowTargetNoResults  messageInfo inlineMessagesContainer' style='display:none;'>" +
					"<span class='windowTargetNoResultsText'>" + "<%=_T('No results found.')%>" + "</span>" +
				"</div>" +
			"</div>";
		this.parent.append(markup);
		this.parent.addClass("windowAddNewAddressee");
	};

	AutoSearchContactSelector.prototype.findBestAddress = function(contact, dealingWithAgents) {
		if (dealingWithAgents) {
			return (contact.Id + "@" + contact.VccGroup);
		}
		if (!this.addressType) {
			return;
		}
		if (contact.contactType === "ADDRESSES BOOK") {
			const fullPhoneNumber = true;
			return contact.getAddressDisplayText(contact, fullPhoneNumber);
		}

		let addressType = this.addressType.toLowerCase();
		for (var i = 0; i < (contact.addresses || []).length; i++) {
			var address = contact.addresses[i];
			if (address.Type.toLowerCase() === addressType) {
				return contact.getAddressDisplayText(address);
			}
		}
		return;
	};

	/*
		Filter Client Side
	*/
	AutoSearchContactSelector.prototype.filterFunction = function(contact) {
		if (!contact) {
		// NO lo visualiza
			return false;
		}
		if (!this.m_filterRegex) {
		// SI lo visualiza
			return true;
		}
		let contactName = this.getContactName(contact);
		if (contactName) {
			let match = contactName.match(this.m_filterRegex);
			if (match) {
			// SI lo visualiza
				return true;
			}
		}
		let bestAddress = this.findBestAddress(contact);
		if (bestAddress) {
			let match = bestAddress.match(this.m_filterRegex);
			if (match) {
			// SI lo visualiza
				return true;
			}
		}
		/*
			Aca podriamos filtrar por tipo de contacto (libreta de direcciones "funciona", o por contacto "NO funciona")
			El de contactos NO funciona ya que la busqueda la filtra server side y ese valor no llega.
			Para hacer esto, habria que implementarlo en la rocka fullcontact a la busqueda, con un nuevo campo mas + un migration
			(NO VALE LA PENA REALIZARLO!! igual dejo el comentario)
		*/
		// let contactType = this.getContactType(contact);
		// if (contactType) {
		// 	let match = contactType.match(this.m_filterRegex);
		// 	if (match) {
		// 		return true;
		// 	}
		// }

		// NO lo visualiza
		return false;
	};

	AutoSearchContactSelector.prototype.setEvents = function() {
		AutoSearchContactSelector.uber.setEvents.call(this);
		// sin el evento scroll (no lo necesito ya que los datos NO tienen paginado)
		this.m_listContainer.unbind("scroll");
	};

	AutoSearchContactSelector.prototype.getAvatar = function(contact) {
		let contact_type = (contact.contactType || "").toUpperCase();
		if (contact_type === "ADDRESSES BOOK") {
			return inConcert.app.getBaseUrl() + "/images/address-book-icon.png";
		}
		return AutoSearchContactSelector.uber.getAvatar.call(this, contact);
	};

}());
