/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:false, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, quotmark: false, esversion: 11 */

/* global inConcert: true */

(function () {

function emptyCallback() {}

/**
Session object constructor.
*/
function Session(sessionId) {
	var m_userdata = null;

	this.id = sessionId;
	this.participants = {};
	this.participantsCount = 0;
	this.is_closed = false;
	this.properties = {};

	this.GetUserdata = function() {
		return m_userdata;
	};
	this.SetUserdata = function(ud) {
		m_userdata = ud;
	};
}

/**
InconcertChat Constructor
*/
function InconcertChat(options) {
	// default chat url
	var m_chatUrl = '/inconcert/apps/chat/';

	function zeropad(num) {
		return ((num < 10) ? '0' : '') + num;
	}

	// set callbacks
	var callbackOnMessage = options.onMessage || emptyCallback;
	var callbackOnCoordinates = options.onCoordinates || emptyCallback; 
	var callbackOnPrevMessage = options.onPreviousMessage || emptyCallback;
	var callbackOnIndication = options.onIndication || emptyCallback;
	var callbackOnNewSession = options.onNewSession || emptyCallback;
	var callbackOnSessionClosed = options.onSessionClosed || emptyCallback;
	var callbackOnUserJoined = options.onUserJoined || emptyCallback;
	var callbackOnUserLeft = options.onUserLeft || emptyCallback;
	var callbackOnUrlPush = options.onUrlPush || emptyCallback;
	var callbackOnSessionUpdateProperties = options.onSessionUpdateProperties || emptyCallback;
	var callbackOnServerDisconnected = options.onServerDisconnected;
	var callbackOnServerReconnected = options.onServerReconnected || emptyCallback;

	this.callbackOnServerReconnected = callbackOnServerReconnected;

	if(typeof(options.username) === 'undefined') {
		throw 'You must supply a username';
	}
	if(typeof(options.vcc) === 'undefined') {
		throw 'You must supply a virtualcallcenter';
	}
	this.m_userID = options.username + '@' + options.vcc;	// ej: supervisor@el_vcc
	this.m_userName = options.username;						// ej: supervisor
	this.m_vcc = options.vcc;
	this.m_displayableUsername = options.displayUsername;
	this.m_authTokenFn = options.RetrieveToken;
	this.m_automaticPollerDisabled = options.hasOwnProperty("automaticPollerDisabled") ? options.automaticPollerDisabled : false;
	this.m_xhrGetMessages = undefined;
	this.m_waiting_data = false;	// flag that indicates if there is a running request
	this.m_last_request_sent = undefined;	// timestamp of the last request sent
	this.m_request_watchdog = undefined;	// a timer used to check if there is a running request when it is supposed to be
	this.m_xhr_error_counter = 0;
	this.m_xhr_max_error_count = 10;
	this.m_message_queue = [];	// message queue, to ensure proper ordering

	// holds the current sessions
	var m_sessions = {
		count: 0,
		currentSession: null,
		sessions: {}
	};

	/**
	Getter and Setter for chat url.
	*/
	this.GetChatUrl = function() {
		return m_chatUrl;
	};
	this.SetChatUrl = function(newUrl) {
		m_chatUrl = newUrl;
	};

	/**
	Public
	Returns the number of current sessions.
	*/
	this.GetSessionCount = function() {
		return m_sessions.count;
	};

	/**
	Returns the current sessions.
	*/
	this.GetSessions = function() {
		return m_sessions.sessions;
	};

	/**
	Returns the session with the given id
	*/
	this.GetSession = function(id) {
		return m_sessions.sessions[id];
	};

	/**
	Private mesage handlers.
	*/
	function HandleOnConnect(sessionId, message) {
		var session;
		if(typeof(m_sessions.sessions[sessionId]) === 'undefined') {
			// This adds a new 'Session' object
			session = new Session(sessionId);
			m_sessions.sessions[sessionId] = session;
			m_sessions.count++;

			// now add the participants
			if(typeof(message.participants) !== 'undefined') {
				for (var participantId in message.participants) {
					if(message.participants.hasOwnProperty(participantId)) {
						session.participants[participantId] = 'normal';	//TODO, use the real role here
						session.participantsCount++;
					}
				}
			}
		}
		session = m_sessions.sessions[sessionId];
		callbackOnNewSession(session, message);
	}

	function HandleOnNewMessage(sessionId, message, isLast) {
		var session = m_sessions.sessions[sessionId];
		if(!session) {
			// ignore unknown sessions, there's nothing to do
			inConcert.logger.debug('Ignoring OnNewMessage message for session ' + sessionId);
			return;
		}
		if(message.msgtype === 'text') {
			callbackOnMessage(session, message, isLast);
		}
		else if(message.msgtype === 'previous_text') {
			callbackOnPrevMessage(session, message, isLast);
		}
		else if(message.msgtype === 'indication') {
			callbackOnIndication(session, message, isLast);
		}
		else if(message.msgtype === 'geo') {
			callbackOnCoordinates(session, message, isLast);
		}
	}

	function HandleOnUserJoined(sessionId, message) {
		var session = m_sessions.sessions[sessionId];
		if(!session) {
			// ignore unknown sessions, there's nothing to do
			inConcert.logger.debug('Ignoring OnUserJoined message for session ' + sessionId);
			return;
		}
		session.participants[message.userid] = message.joined_user_role || 'normal';
		session.participantsCount++;
		callbackOnUserJoined(session, message);
	}

	function HandleOnUserLeft(sessionId, message) {
		var session = m_sessions.sessions[sessionId];
		if(!session) {
			inConcert.logger.debug('Ignoring OnUserLeft message for session ' + sessionId);
			return;
		}
		if(typeof(session.participants[message.user_id]) !== 'undefined') {
			session.participantsCount--;
			delete session.participants[message.user_id];
		}
		callbackOnUserLeft(session, message);
	}

	function HandleOnUrlPush(sessionId, message) {
		var session = m_sessions.sessions[sessionId];
		if(!session) {
			inConcert.logger.debug('Ignoring OnPushUrl message for session ' + sessionId);
			return;
		}
		callbackOnUrlPush(session, message);
	}

	function HandleOnDisconnect(sessionId, message) {
		var session = m_sessions.sessions[sessionId];
		if(!session) {
			inConcert.logger.debug('Ignoring OnSessionClosed message for session ' + sessionId);
			return;
		}
		m_sessions.count--;
		session.is_closed = true;
		callbackOnSessionClosed(session, message);
		delete m_sessions.sessions[sessionId];
	}

	function HandleOnServerDisconnected() {
		var sessions = m_sessions.sessions;
		for(var key in sessions) {
			if(sessions.hasOwnProperty(key)) {
				var session = sessions[key];
				session.is_closed = true;	// eso lo tengo que hacer en inconcertchat.js y no aca
			}
		}
		if(typeof(callbackOnServerDisconnected) !== 'undefined') {
			callbackOnServerDisconnected();
		}
	}

	/**
		Timer that checks if we need to resend a request. In Adobe AIR, requests vanishes into thin air from time to time, so
		check: if we're waiting for data and the request started more than 15 seconds ago, cancel it and start anew.
	*/
	function OnRequestWatchdog() {
		if( typeof(this.m_last_request_sent) === 'undefined' ) {
			return;
		}

		// there is nothing to do with a logged out user.
		if (this.m_loggedOut) {
			return;
		}
		
		if (this.isAutomaticPollerDisabled()) {
			return;
		}

		var now = new Date().getTime();

		var delta = now - this.m_last_request_sent;
		if(delta > 15000 && this.m_waiting_data) {
			inConcert.logger.debug('[inConcertChat] OnRequestWatchdog: cancelling request started "' + delta + '" ms ago.');
			this.m_last_request_sent = undefined;
			this.retryPoller();
		}
	}

	this.onServerDisconnected = HandleOnServerDisconnected;


	this._login = function(url, callback) {
		var self = this;
		callback = callback || emptyCallback;

		inConcert.invoke({
			type: 'POST',
			url: url,
			data: JSON.stringify({}),
			dataType: 'json',
			contentType: 'application/json',
			headers: {
				'x-inconcert-sid': 'AGENT:' + self.m_authTokenFn()
			},
			success: function(response) {
				self.callbackOnServerReconnected(response.status, response);
				self.__rearmPollerRequest();
				// Attempto to flush pending messages
				if(self.m_message_queue.length > 0) {
					self.__SendMessage(self.m_message_queue[0]);
				}
				callback(true);
			},
			error: function(/*http_response, error*/) {
				self.onServerDisconnected();
				callback(false);
			}
		});
	};

	/**
	Privileged
	*/
	this.m_onMessagesAvailable = function(messages, textStatus) {
		var self = this;
		var ids_to_acknowledge = [];
		let isLast;
		self.m_current_msg_ids_to_ack = null;

		try {
			messages = messages.messages;
			for(var i = 0; i < messages.length; i++) {
				var message = messages[i];

				// Acknowledge the message even if processing it causes an error.
				// Avoids getting into a loop when the message arrives again and again.
				ids_to_acknowledge.push(message.__message_id);
				if(typeof(message.msgtime) !== 'undefined') {
					var vDate;
					if (message.msgtype === 'previous_text') {
						//es un mensaje previo, debo mostrar la hora que viene del server
						vDate = new Date(parseInt(message.msgtime, 10) * 1000);
					} else {
						vDate = new Date(); // toma, te muestro la hora actual. Al diablo con la del server!
					}
					message.msgTimeStamp = parseInt(message.msgtime, 10) * 1000;
					message.msgtime = zeropad(vDate.getHours()) + ':' + zeropad(vDate.getMinutes()) + ':' + zeropad(vDate.getSeconds());
				}
				inConcert.logger.debug(message);
				// Since I use the sessionID to as part of the ID of DOM nodes, I need to change the curly braces to something else (valid)
				var sessionId = message.session_id.replace(/\{|\}/g, '_');
				switch(message.message.toLowerCase()) {
					case 'onconnect':
						HandleOnConnect(sessionId, message);
					break;

					case 'onnewmessage':
					case 'onindication':
					case 'oncoordinatesmessage':
						isLast = (messages.length - 1 === i);
						HandleOnNewMessage(sessionId, message, isLast);
					break;
					
					case 'onuserjoined':
						HandleOnUserJoined(sessionId, message);
					break;

					case 'onuserleft':
						HandleOnUserLeft(sessionId, message);
					break;

					case 'onurlpush':
						HandleOnUrlPush(sessionId, message);
					break;

					case 'ondisconnect':
					case 'onsessionclosed':
						HandleOnDisconnect(sessionId, message);
					break;

					case 'onsessionupdateproperties':
						var session = m_sessions.sessions[sessionId];
						if(!session) {
							inConcert.logger.debug('Ignoring OnSessionUpdateProperties message for session ' + sessionId);
							// ignore unknown sessions, there's nothing to do
							continue;
						}
						for (var prop in message) {
							if(message.hasOwnProperty(prop) && prop !== 'session_id' && prop !== 'msgtime' && prop !== 'message') {
								session.properties[prop] = message[prop];
							}
						}
						callbackOnSessionUpdateProperties(session, message);
					break;

					case 'onserverdisconnected':
						HandleOnServerDisconnected();
					break;

					case 'onserverreconnected':
						callbackOnServerReconnected(message);
					break;
				}
			}
		}
		catch(e) {
			inConcert.logger.error({
				where: 'inConcertChat.m_onMessagesAvailable',
				exception: e,
				messages: messages,
				textStatus: textStatus
			});
		}
		self.__rearmPollerRequest(ids_to_acknowledge);
	};


	/**
	Privileged:
	Asks for messages to the server. Receives a Json with an array of messages.
	ids_to_acknowledge is an (optional) array of message ids to acknowledge on the server.
	*/
	this.__RequestMessages = function(ids_to_acknowledge, callback = $.noop, sync = false) {
		var self = this;

		// If no ids array is given, retry the acknowledge with the last set of ids.
		// If there is none, use an empty array.
		if(ids_to_acknowledge) {
			self.m_current_msg_ids_to_ack = ids_to_acknowledge;
		}
		else if(!self.m_current_msg_ids_to_ack) {
			self.m_current_msg_ids_to_ack = [];
		}

		if (self.m_loggedOut) {
			return callback({ result : "LOGGED_OUT", reason : "user logged out" });
		}

		// Start watchdog
		if(typeof(self.m_request_watchdog) === 'undefined') {
			self.m_request_watchdog = window.setInterval(function() {
				OnRequestWatchdog.apply(self);
			}, 1000);
		}

		// If already waiting for data, don't issue a new request
		if(self.m_waiting_data) {
			return callback({ result : "BUSSY", reason : "another polling is running" });
		}
		self.m_waiting_data = true;
		self.m_last_request_sent = new Date().getTime();
		// tiro el request asincronico y me guardo el objeto XHR para poder cancelar el request eventualmente
		self.m_xhrGetMessages = inConcert.invoke({
			type: 'post',
			url: m_chatUrl + 'get_user_messages',
			data: JSON.stringify({
				ack_ids : self.m_current_msg_ids_to_ack,
				sync : sync
			}),
			dataType: 'json',
			contentType: 'application/json',
			headers: {
				'x-inconcert-sid': 'AGENT:' + self.m_authTokenFn()
			},
			success: function(data, textStatus) {
				self.m_waiting_data = false;
				self.m_xhr_error_counter = 0;

				if (!data.status){
					setTimeout(function() {
						self.Login();
					}, 2000);
					return;
				}
				self.m_onMessagesAvailable(data, textStatus);
				callback({ result : "OK", data : data, status : textStatus });
			},
			error: function(http_response, text_status) {
				if (self.m_loggedOut) {
					return callback({ result : "LOGGED_OUT", reason : "user logged out" });
				}
				self.m_waiting_data = false;
				self.m_xhr_error_counter++;
				// Reintento por siempre hasta que obtenga conexion o se pierda la sesion..
				//if (self.m_xhr_error_counter === self.m_xhr_max_error_count) {
				//	// we don't request more messages
				//	HandleOnServerDisconnected();
				//}
				//else {
				//	self.__rearmPollerRequest(self.m_current_msg_ids_to_ack);
				//}
				self.__rearmPollerRequest(self.m_current_msg_ids_to_ack);
				let err = http_response.jqXHR;
				callback({ result : "FAIL", code : err.status, reason : text_status });
			}
		});
	};
}
	
InconcertChat.prototype.__cancelPoller = function() {
	this.m_waiting_data = false;
	if (this.m_xhrGetMessages) {
		this.m_xhrGetMessages.abort();
	}
};
	
InconcertChat.prototype.retryPoller = function(ids_to_acknowledge = null) {
	this.__cancelPoller();
	this.__rearmPollerRequest(ids_to_acknowledge);
};

InconcertChat.prototype.firePoller = function(data = {}, callback = $.noop) {
	if (data.force) {
		this.__cancelPoller();
	}
	this.__RequestMessages(this.m_pending_ids_to_acknowledge, callback, data.sync);
};


InconcertChat.prototype.__rearmPollerRequest = function(ids_to_acknowledge = null) {
	let self = this;
	if (self.isAutomaticPollerDisabled()) {
		self.m_pending_ids_to_acknowledge = ids_to_acknowledge;
		return;
	}
	inConcert.Utils.zeroTimeout(function(ids_to_acknowledge) {
		self.__RequestMessages(ids_to_acknowledge);
	}, ids_to_acknowledge);
};
	
InconcertChat.prototype.isAutomaticPollerDisabled = function() {
	return inConcert.app.isAutomaticPollerDisabled() || this.m_automaticPollerDisabled;
};
	
InconcertChat.prototype.enableAutomaticPoller = function() {
	let wasDisabled = this.m_automaticPollerDisabled;
	this.m_automaticPollerDisabled = false;
	if (wasDisabled) {
		this.retryPoller();
	}
};

InconcertChat.prototype.disableAutomaticPoller = function() {
	this.m_automaticPollerDisabled = true;
};

InconcertChat.prototype.toString = function() { return '[object inConcertChat]'; };

/**
Public
Logs the current user in. When the user is logged in, messages start to be requested.
*/
InconcertChat.prototype.Login = function(callback) {
	var self = this;
	var url = this.GetChatUrl() + 'login';

	callback = callback || emptyCallback;

	self._login(url, callback);
};

/**
Public
Rejoin the current user in. When the user is logged in, messages start to be requested.
*/
InconcertChat.prototype.LoginWithRejoin = function(callback) {
	var self = this;
	var url = this.GetChatUrl() + 'loginwithrejoin';

	callback = callback || emptyCallback;

	self._login(url, callback);
};


/**
Public
Logs Rejoin session chat to current user in.
*/
InconcertChat.prototype.RejoinSession = function(sessionId, callback) {
	var self = this;
	callback = callback || emptyCallback;

	inConcert.invoke({
		type: 'POST',
		url: this.GetChatUrl() + 'rejoinsession/' + encodeURIComponent(sessionId),
		data: JSON.stringify({}),
		dataType: 'json',
		contentType: 'application/json',
		headers: {
			'x-inconcert-sid': 'AGENT:' + self.m_authTokenFn()
		},
		success: function(data, textStatus) {
			if (data.status){
				self.m_onMessagesAvailable(data, textStatus);
				callback({ result : "OK", data : data, status : textStatus });
			} else {
				callback({ result : "FAIL", code : -1, reason : textStatus });
			}
		},
		error: function(http_response, text_status) {
			if (self.m_loggedOut) {
				return callback({ result : "LOGGED_OUT", reason : "user logged out" });
			}
			let err = http_response.jqXHR;
			callback({ result : "FAIL", code : err.status, reason : text_status });
		}
	});
};

/**
Public
Logs the current user out.
*/
InconcertChat.prototype.Logout = function() {
	if (this.m_loggedOut) {
		return;
	}
	this.m_loggedOut = true;
	if(typeof(this.m_xhrGetMessages) !== 'undefined') {
		this.m_xhrGetMessages.abort();
	}
	if(typeof(this.m_request_watchdog) !== 'undefined') {
		window.clearInterval(this.m_request_watchdog);
	}
	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + 'logout',
		data: '',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn()
		},
		async: false, // Safari will ignore any XHR executed during a onbeforeunload or onunload handler unless it is executed synchronously
		error: function(/*http_response, error*/) {}
	});
};

/**
Public
Returns the current user's ID
*/
InconcertChat.prototype.GetUserID = function() {
	return this.m_userID;
};

/**
Public
Returns the current user's VCC
*/
InconcertChat.prototype.GetVcc = function() {
	return this.m_vcc;
};

/**
Public
Invites a user to chat.
@param username is the username of the user we're inviting
@param callback is an optional callback function to be called either on success or on error. The function signature is:
		callback(Boolean status, String message_or_SessionId)
*/
InconcertChat.prototype.InviteUser = function(username, callback, parameters) {
	var self = this;

	callback = callback || emptyCallback;

	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + 'invite_user/' + encodeURIComponent(username),
		data: JSON.stringify(parameters || {}),
		dataType: 'json',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn()
		},
		contentType: 'application/json',
		success: function(response) {
			var newSessionId = response.message.replace(/\{|\}/g, '_');
			response.message = newSessionId;
			var newSession = self.GetSession(newSessionId);
			// it is possible that we haven't received the OnSessionCreatedMessage yet
			if (newSession) {
				newSession.properties.invited_user = username;
			}
			callback(true, response.message);
		},
		error: function(http_response, error) {
			inConcert.logger.error('inConcertChat.InviteUser - http error: ' + http_response.status);
			callback(false, error.result);
		}
	});
};

/**
Public
Monitors a given session.
@param sessionId is the id of the session to monitor.
*/
InconcertChat.prototype.MonitorSession = function (sessionId, callback)
{
	callback = callback || emptyCallback;

	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + 'monitor_session/' + encodeURIComponent(sessionId),
		data: '',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn()
		},
		dataType: 'json',
		contentType: 'text/plain',
		success: function(response) {
			callback(response.status, response);
		},
		error: function(http_response, error) {
			inConcert.logger.error('inConcertChat.MonitorSession - http error' + http_response.status);
			callback(false, {
				reason: "NETWORK_ERROR",
				message: error.message
			});
		}
	});
};

/**
Public
Coaches a user in a given session.
@param username is the username of the user we'll coach
@param sessionId is the id of the session the user is in
*/
InconcertChat.prototype.CoachUser = function (username, sessionId, callback)
{
	callback = callback || emptyCallback;

	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + 'coach_user/' + encodeURIComponent(username) + '/' + encodeURIComponent(sessionId),
		data: '',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn()
		},
		dataType: 'json',
		contentType: 'text/plain',
		success: function(response) {
			if (response.reason === "USER_ALREADY_COACHED") {
				// parse previous coach
				response.previousCoach = response.previous_coach_id.match(/(.+)@([^@]+)$/)[1];
			}
			callback(response.status, response);
		},
		error: function(http_response, error) {
			inConcert.logger.error('inConcertChat.CoachUser - http error: ' + http_response.status);
			callback(false, error.result);
		}
	});
};

/**
Public
Removes a user from a given session
@param sessionId is the Id of the session you want to remove the logged user from
@callback return function confirmation
*/
InconcertChat.prototype.RemoveUserFromSession = function(sessionId, callback) {
	callback = callback || emptyCallback;

	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + 'remove_user_from_session/' + encodeURIComponent(sessionId),
		data: '',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn()
		},
		dataType: 'json',
		contentType: 'application/json',
		success: function(response) {
			callback(response.status, response.reason, response.message);
		},
		/* http_response: { status : number (http status), headers : {} (http headers) } } */
		/* error: { response: body : string } */
		error: function(http_response, error) {
			inConcert.logger.error('inConcertChat.RemoveUserFromSession - http error: ' + http_response.status);
			callback(false, error.result);
		}
	});
};

/**
Public
Sends a message to the given session. The message gets queued to ensure proper ordering. (#3450)
@param sessionId is the Id of the session to send the message to.
@param message is the message to send.
*/
InconcertChat.prototype.SendMessage = function(sessionId, message, callback = $.noop) {
	if(message === '' || typeof(message) !== 'string') {
		return callback(false, "must enter a message");
	}
	var session = this.GetSession(sessionId);
	if(!session || session.is_closed) {
		return callback(false, "session not found or closed");
	}
	var currentChatText = encodeURIComponent(message);
	// Convert underscores back to curly braces
	if(sessionId.charAt(0) === '_' && sessionId.charAt( sessionId.length - 1) === '_') {
		sessionId = '{' + sessionId.substring(1, sessionId.length - 1) + '}';
	}

	// queue the message. If it is the only message on the queue, send it right away.
	this.m_message_queue.push({ session_id: sessionId, text: currentChatText});
	
	if(this.m_message_queue.length === 1) {
		this.__SendMessage(this.m_message_queue[0], callback);
	}
	else {
		callback(true);
	}
};

InconcertChat.prototype.SendMessageInternal = function(sessionId, message, add_to_qna, callback = $.noop) {
    
	if(message === '' || typeof(message) !== 'string') {
		return callback(false, "must enter a message");
	}
	var session = this.GetSession(sessionId);
	if(!session || session.is_closed) {
		return callback(false, "session not found or closed");
	}
	var currentChatText = encodeURIComponent(message);

	// Convert underscores back to curly braces
	if(sessionId.charAt(0) === '_' && sessionId.charAt( sessionId.length - 1) === '_') {
		sessionId = '{' + sessionId.substring(1, sessionId.length - 1) + '}';
	}

	// queue the message. If it is the only message on the queue, send it right away.
	this.m_message_queue.push({ session_id: sessionId, text: currentChatText, add_to_qna: add_to_qna});

	if(this.m_message_queue.length === 1) {
		this.__SendMessageInternal(this.m_message_queue[0], callback);
	}
	else {
		callback(true);
	}
};

/**
* Private:
* Actually sends the message over the wire. When the message is sent, sends the next one in queue (if any).
* @param message is a hash with session_id and text.
*/
InconcertChat.prototype.__SendMessage = function(message, callback = $.noop)
{
	var self = this;
	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + 'send_message/' + encodeURIComponent(message.session_id),
		data: message.text,
		contentType: 'application/xml',
		processData: false,
		dataType: 'json',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn(),
			'Content-Encoding': 'utf-8'
		},
		success: function(response) {
			var message_sent = self.m_message_queue.shift();
			if(response.status) {
				callback(true);
			}
			else {
				inConcert.logger.info("Message ('" + message_sent.text + "') was delivered but failed with reason '" +
					response.reason + "' for session '" + message_sent.session_id + "'");
				callback(false, "error sending message");
			}

			// TODO: send the whole queue in batch. Need to add support for that in both the gateway and the server.
			if(self.m_message_queue.length >= 1) {
				self.__SendMessage(self.m_message_queue[0]);
			}
		},
		error: function(http_response/*, error*/) {
			callback(false, "error sending message");
			var message_failed = self.m_message_queue[0];
			inConcert.logger.info("Message ('" + message_failed.text + "') was not delivered, it failed with '" +
				"http status '" + http_response.status + "' for session '" + message_failed.session_id + "'");
			if(self.m_message_queue.length >= 1) {
				self.__SendMessage(self.m_message_queue[0]);
			}
		}
	});
};

InconcertChat.prototype.__SendMessageInternal = function(message, callback = $.noop)
    {
        var self = this;
		inConcert.invoke({
			type: 'post',
			url: this.GetChatUrl() + 'send_message_internal/' + encodeURIComponent(message.session_id),
			data: JSON.stringify(message),
			contentType: 'application/xml',
			processData: false,
			dataType: 'json',
			headers: {
				'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn(),
				'Content-Encoding': 'utf-8'
			},
			success: function(response) {

				var message_sent = self.m_message_queue.shift();
				if(response.status) {
					callback(true);
				}
				else {
					inConcert.logger.info("Message ('" + message_sent.text + "') was delivered but failed with reason '" +
						response.reason + "' for session '" + message_sent.session_id + "'");
					callback(false, "error sending message");
				}
			},
			error: function(/*, error*/) {
				callback(false, "error sending message");
			}
		});
    };

/**
Public
Sends an indication to the given session.
@param sessionId is the Id of the session to send the indication to.
@param indication_type is the type of the indication to send.
@param persist is a flag to store the indication in the log or not.
@param values is a hash with values to be sent.
*/
InconcertChat.prototype.SendIndication = function(sessionId, indication_type, persist, values) {
	if(indication_type === '' || typeof(indication_type) !== 'string') {
		return;
	}
	if(typeof(persist) !== 'boolean') {
		return;
	}

	var session = this.GetSession(sessionId);
	if(!session || session.is_closed) {
		return;
	}

	// Convert underscores back to curly braces
	if(sessionId.charAt(0) === '_' && sessionId.charAt( sessionId.length - 1) === '_') {
		sessionId = '{' + sessionId.substring(1, sessionId.length - 1) + '}';
	}
	inConcert.invoke({
		type: 'post',
		url: this.GetChatUrl() + (persist ? 'send_indication/' : 'send_np_indication/') +
			encodeURIComponent(sessionId) +  '/' + encodeURIComponent(indication_type),
		data: JSON.stringify(values),
		contentType: 'application/json',
		processData: false,
		dataType: 'json',
		headers: {
			'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn(),
			'Content-Encoding': 'utf-8'
		},
		success: function(/*response*/) {}
	});
};

/**
Public
Closes a given session.
@param sessionId is the ID of the session to close.
*/
InconcertChat.prototype.CloseSession = function(sessionId) {
	var session = this.GetSession(sessionId);
	if(!session) {
		return;
	}
	if(!session.is_closed) {
		// Convert underscores back to curly braces
		if(sessionId.charAt(0) === '_' && sessionId.charAt( sessionId.length - 1) === '_') {
			sessionId = '{' + sessionId.substring(1, sessionId.length - 1) + '}';
		}
		inConcert.invoke({
			type: 'POST',
			url: this.GetChatUrl() + 'close_session/' + encodeURIComponent(sessionId),
			data: JSON.stringify({}),
			dataType: 'json',
			contentType: 'application/json',
			headers: {
				'x-inconcert-sid': 'AGENT:' + this.m_authTokenFn()
			},
			success: function(/*response*/) {
				session.is_closed = true;
			},
			error: function(http_response/*, error*/) {
				inConcert.logger.error('inConcertChat.CloseSession - http error: ' + http_response.status);
			}
		});
	}
};

// TODO: Both these classes should be in inConcert namespace and not global
window.InconcertChat = InconcertChat;
window.Session = Session;

}());
