/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 9 */

/* global inConcert */

if (!this.inConcert) {
	this.inConcert = {};
}
if (!inConcert.interactions) {
	inConcert.interactions = {};
}

/**
 * @namespace
 * @name SmsInteraction
 */
(function () {

	"use strict";
	var format = inConcert.Utils.format;

	/**
	 * Class SmsInteraction.
	 * @param {Object} rawInteraction
	 */
	function SmsInteraction (rawInteraction)
	{
		SmsInteraction.uber.constructor.call(this, rawInteraction);
		this.events = this.events || [];

		// Pending interactions have no properties yet. Provide some sane defaults
		if (this.isBeingComposed()) {
			this.properties = this.properties || {};
			this.properties.display_name = "";
			this.properties.summary = "";
		}
	}

	inConcert.Utils.inherit(SmsInteraction, inConcert.interactions.base);

	if (typeof inConcert.interactions.sms !== "function") {
		inConcert.interactions.sms = SmsInteraction;
	}

	SmsInteraction.prototype.inflate = function (callback)
	{
		var self = this;
		SmsInteraction.uber.inflate.call(this, function(data){
			if(data) {
				self.__processEvents();
				self.sortEvents();
				callback(data);
			}
		});
	};

	SmsInteraction.prototype.__processEvents = function() {
		$.each(this.events || [], function(_, event) {
			if (event.message) {
				//all messages need an id
				event.message.id = event.message.id || event.message.fake_id;
			}
		});
	};

	/*
		Funcion que ordena los eventos primero el evento orgigen, luego por fecha de creacion
	*/
	SmsInteraction.prototype.__sort_by_creation_date = function()
	{
		this.events.sort( function( a, b )
		{
			if (!a.message.timestamp || !b.message.timestamp) {
				return a.__index - b.__index;
			}
			if (a.message.timestamp === b.message.timestamp) {
				return a.__index - b.__index;
			}
			return a.message.timestamp - b.message.timestamp;
		});
	};

	SmsInteraction.prototype.get_error_message = function( error_code, reason, type, error )
	{
		switch ( (error_code || "" ).toUpperCase() )
		{
			case "INTERACTION_ALREADY_EXIST_ADDRESS":
				return "<%=_T('Already exists an open interaction with same phone number')%>";
			case "TIMEOUT":
				return "<%=_T('Waiting time expired to send message')%>";
			case "THREAD_OPEN":
				if (error.agent) {
					let agent = inConcert.Utils.getIdWithoutVCC(error.agent);
					return format("<%=_T('There is an open interaction with this user assigned to agent {0}.')%>", agent);
				}
				else {
					var state = error.state || "";
					switch(state) {
						case "WAITING":
							state = "<%=_T('Waiting')%>";
							break;
						case "QUEUED":
							state = "<%=_T('Queued')%>";
							break;
					}
					return format("<%=_T('There is an open interaction with this user in state {0}.')%>", state);
				}
				break;
			case "INVALID":
				return "<%=_T('Invalid phone number')%>";
			case "BLACKLIST":
				return "<%=_T('Blacklisted addresses')%><%=_T(':')%> " +
					(error.addresses || []).join(", ");
			default:
				return "<%=_T('There was an error while sending message')%>";
		}
	};

	SmsInteraction.prototype.createAndSendMessage = function (params)
	{
		const self = this;
		const attention_level_id = params.attention_level_id;
		const address = params.address;
		const message = params.message;
		const fake_event_id = params.fake_event_id || inConcert.Utils.generateGuid();
		const callback = params.callback || $.noop;

		inConcert.logger.action({
			action_name : "new_sms_message",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		var url = this.m_manager.getBaseUrl() + "/interactions/sms/create_and_send/";
		var request = {
			fake_event_id		: fake_event_id,
			interactionID		: this.id,
			parentInteractionId : this.parentInteractionId,
			campaign			: this.campaign,
			account				: this.account,
			attention_level_id  : attention_level_id,
			contactID			: this.contact || null,
			phoneNumber			: address,
			message				: message
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true, "OK" );
			},
			error: function(_, error){
				inConcert.logger.action({ action_name : "new_sms_message", error_message : error.result });
				var message = self.get_error_message( error.result, error.reason, "SEND_MESSAGE", error );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	SmsInteraction.prototype.sendMessage = function( fake_event_id, message, callback )
	{
		var self = this;

		callback = callback || $.noop;
		fake_event_id = fake_event_id || inConcert.Utils.generateGuid();

		inConcert.logger.action({
			action_name : "new_sms_message",
			id : this.id,
			type : this.type,
			campaign : this.campaign,
			fake_event_id: fake_event_id
		});

		var url = this.m_manager.getBaseUrl() + "/interactions/sms/send_message/";
		var request = {
			fake_event_id		: fake_event_id,
			interactionID		: this.id,
			campaign			: this.campaign,
			account				: this.account,
			message				: message
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify(request),
			dataType: "json",
			contentType: "application/json",
			success: function() {
				callback( true, "OK" );
			},
			error: function(_, error){
				inConcert.logger.action({ action_name : "new_sms_message", error_message : error.result });
				var message = self.get_error_message( error.result, error.reason, "SEND_MESSAGE", error );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};

	// Just returns the summary
	SmsInteraction.prototype.getExtendedInfo = function( max_length )
	{
		return inConcert.Utils.truncateWord( this.properties.summary, max_length );
	};

	SmsInteraction.prototype.getMessageContent = function(event)
	{
		var data = { type : "default", text : "" };
		var message = event.message;
		data.text = message.text || "";
		return data;
	};

	SmsInteraction.prototype.getMessageText = function(event)
	{
		return this.getMessageContent(event).text;
	};

	SmsInteraction.prototype.getEventShortDescription = function(sorted_index) {
		var ev = this.getEvent(sorted_index);
		return this.getMessageText(ev);
	};

	SmsInteraction.prototype.getMessageStatus = function(event) {
		return event.message && event.message.status || null;
	};

	SmsInteraction.prototype.getMessageStatusReason = function(event) {
		return event.message && event.message.reason || null;
	};

	SmsInteraction.prototype.getMessageStatusDate = function(event) {
		var status = this.getMessageStatus(event);
		var timestamp = status && event.message["timestamp_" + status] || null;
		return timestamp && new Date(timestamp * 1000) || null;
	};

	/**
	* @param index
	*/
	SmsInteraction.prototype.getDateObject = function(index) {
		index = index || 0;
		const event = this.getEvent(index);
		if (!event || !event.message) {
			return "";
		}
		const event_date = event.message.timestamp * 1000;
		return this.convertDateStringToDate(event_date);
	};

	SmsInteraction.prototype.getRawAddress = function(index) {
		this.assert_inflated();
		return this.events[index].address && this.events[index].address.Number || "";
	};

	SmsInteraction.prototype.getSource = function( sorted_index )
	{
		this.assert_inflated();
		if ( !this.events || sorted_index === undefined || sorted_index === -1 ) {
			return "";
		}
		var event = this.events[ sorted_index ];
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		if (event.message && event.message.name) {
			return event.message.name;
		}
		if (event.address && event.address.Number) {
			return event.address.Number;
		}
		return "";
	};

	SmsInteraction.prototype.getEventUserAvatarURI = function( event )
	{
		if (typeof(event) !== "object" || typeof(event.message) !== "object") {
			return "undefined";
		}
		if ( event.creator === "agent" && event.agent ) {
			var user_id = event.agent;
			if ( !user_id ) {
				user_id = inConcert.app.getCompleteId();
			}
			if ( user_id.match(inConcert.app.idAtVCCRegExp) ) {
				user_id = RegExp.$1;
			}
			return inConcert.app.getBaseUrl() + "/users/get_avatar/" + user_id;
		}
		else if ( event.creator === "bot" ) {
			return "/inconcert/apps/agent/images/botAvatar.png";
		}
		else {
			return "undefined";
		}
	};

	//returns the conversation contact or member display text
	SmsInteraction.prototype.getUserName = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}

		if ( event.creator === "agent" && event.agent ) {
			return this.get_agent_name( event );
		}
		if ( event.creator === "bot" ) {
			return this.getBotName();
		}
		if (event.creator === "contact" && this.interactionData.batchAddressName) {
			return this.interactionData.batchAddressName;
		}
		if (event.message && event.message.name) {
			return event.message.name;
		}
		if (event.address && event.address.Number) {
			return event.address.Number;
		}
		return "";
	};

	SmsInteraction.prototype.getUserNumber = function ( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}

		if (event.address && event.address.Number) {
			return event.address.Number;
		}
		return "";
	};

	SmsInteraction.prototype.get_user_event_id = function( event )
	{
		this.assert_inflated();

		if ( typeof( event ) !== "object" ) {
			return "";
		}
		if ( event.creator === "agent" ) {
			return event.agent || inConcert.app.getCompleteId();
		}
		if (event.userid) {
			return event.userid;
		}
		return "";
	};

	/**
	* Returns a displayable name, which was pre computed on the server. It extracts the user name from the first event 
	* of the interaction or the agent id, depending on which one originated the interaction.
	* The computed property has a prefixed value to it. If the prefix is "text:", then the rest is to be taken 
	* verbatim. It is the name of a FB user.
	* If the prefix is "agent_id:", then we need to perform a lookup to translate the id to an agent name.
	* 
	* @return {String}
	*/
	SmsInteraction.prototype.getDisplayName = function()
	{
		var display_name = this.properties.display_name;
		if(display_name.match(/text:(.+)/)) {
			return RegExp.$1;
		}
		else if(display_name.match(/phone:(.+)/)) {
			return RegExp.$1;
		}
		else if(display_name.match(/agent_id:/)) {
			var agent_id = RegExp.$1;
			if (agent_id === inConcert.app.getCompleteId()) {
				return inConcert.app.getLoggedUserName();
			}
			else {
				var agent_name;

				inConcert.app.getTeamMember( agent_id, function(agent)
					{
						if (!agent) {
							return;
						}
						// Esto por qué carajo no lo hago una vez sola?
						var have_name = false;
						if ((agent.Name || "").trim().length > 0) {
							agent_name = agent.Name;
							have_name = true;
						}
						if ((agent.FirstSurname || "").trim().length > 0) {
							agent_name = (have_name ? (agent_name + " ") : "") + agent.FirstSurname;
						}
					},
					true // from cache
				);

				agent_name = agent_name || agent_id;
				return agent_name;
			}
		}
		return display_name;
	};

	SmsInteraction.prototype.getFirstUnreadEventIndex = function() {
		this.assert_inflated();

		var eventIndex = inConcert.interactions.base.prototype.getFirstUnreadEventIndex.call(this);
		if (typeof(eventIndex) === "number" && this.events[eventIndex]) {
			var event = this.events[eventIndex];
			return (typeof(event.parentIndex) !== "undefined") ? event.parentIndex : eventIndex;
		}
		else {
			return eventIndex;
		}
	};

	/* @return {Boolean} */
	// IB: ESTA FUNCION SE PUEDE CALCULAR UNA VEZ EN EL SERVER Y LISTO
	SmsInteraction.prototype.eventsAreFromTheSameClient = function() {
		this.assert_inflated();

		let events_from_the_same_client = true;
		let last_event_client = null;

		for (let i = 0; i < this.events.length; i++) {
			let current_event_client = this.getRawAddress(i);
			// first iteration
			if (!last_event_client){
				last_event_client = current_event_client;
			} else {
				if (current_event_client !== last_event_client){
					events_from_the_same_client = false;
					break;
				}
			}
		}

		return events_from_the_same_client;
	};

	SmsInteraction.prototype.isValid = function() {
		/*this.assert_inflated();
		// IB: Revisar cual es el sentido de esto

		if (this.state === "PENDING") {
			return true;
		}
		if (!$.isArray(this.events) || this.events.length === 0) {
			return false;
		}*/
		return true;
	};

	SmsInteraction.prototype.forward = function( address, messages, callback ) 
	{
		callback = callback || $.noop;
		let self = this;
		let fake_event_id = inConcert.Utils.generateGuid();
		
		inConcert.logger.action({action_name : "sms_forward", id : this.id, type : this.type,
			campaign : this.campaign, address : address, account : this.account, action_type : "conversation", fake_event_id: fake_event_id});

		let url = this.m_manager.getBaseUrl() + "/interactions/sms/forward/";
		var request = {
			fake_event_id : fake_event_id,
			campaignId	  : this.campaign,
			account       : this.account,
			page          : this.getPage(),
			interactionID : this.id,
			address		  : address,
			messages      : messages
		};
		inConcert.invoke({
			type: "POST",
			url: url,
			data: JSON.stringify( request ),
			dataType: "json",
			contentType: "application/json",
			success: function( result ) {
				var message;
				var status = result.result === "OK";
				if (!status) {
					message = self.get_error_message( result.result, result.reason, "FORWARD", result );
				}
				callback( status, message, result.result );
			},
			error: function( _, error ) {
				inConcert.logger.action({ action_name : "sms_forward", error_message : error.result });
				var message = self.get_error_message( error.result, error.reason, "FORWARD", error );
				callback( false, message, error );
			}
		});

		return fake_event_id;
	};
	
	let internal_message_keys = {
		chattype : true,
		chatid : true,
		wappid : true,
		id : true,
		name : true,
		forward : true,
		status : true,
		origin : true
	};
	
	SmsInteraction.prototype.__isForwardMessageKey = function(key) {
		return !internal_message_keys[key] && key.indexOf("timestamp") < 0;
	};
	
	SmsInteraction.prototype.getForwardMessage = function(index) {
		let self = this;
		let event = this.getEvent(index);
		return event && event.message && Object.keys(event.message).reduce(function(message, key) {
			if (self.__isForwardMessageKey(key)) {
				message[key] = event.message[key];
			}
			return message;
		}, {}) || null;
	};
	
	SmsInteraction.prototype.getForwarded = function(index) {
		let event = this.getEvent(index);
		return event && event.message && event.message.forward || null;
	};
	
	SmsInteraction.prototype.hasForward = function() {
		return true;
	};

}());
