/*jshint bitwise:false, browser:true, curly:false, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, -W014, esversion: 11 */
/*jshint quotmark:false*/
/*global _, inConcert, EventEmitter, $*/

this.inConcert = this.inConcert || {};

(function() {
	"use strict";

	const format = inConcert.Utils.format;
	const MAX_VIDEO_BANDWIDTH = 64000;
	const EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND = "not_interaction_found"; // valor utilizado para suscribir eventos de close de ventanas cuando la interaccion ha sido eliminada y no existe en el interactionManager.

	/**
	 * Known actions, which may be supported (or not) by the phone
	 */
	const PhoneAction = {
		DIAL       :   1,  // dial() is available
		ANSWER     :   2,  // answer() and reject() are available
		HANGUP     :   4,  // hangup() is available
		MUTE       :   8,  // mute() is available
		HOLD       :  16,  // hold() is available
		TRANSFER   :  32,
		CONFERENCE :  64,
		_SWITCH    : 128
	};
	
	const VideoViewStates = {
		MINIMIZED  : "videoCallDocked",
		MAXIMIZED  : "videoCallFullHeight",
		FULL_SCREN : "videoCallFullScreen"
	};

	/*--------------------------------------------------------------------
		Agent Poller Interaction Events:
			Objeto que contiene funciones que subscribe a la escucha de eventos de las interacciones que llegan desde el agentpoller.

			Utiliza "pubsub.js"
		
		Nota:
			Este objeto por el momento soporta unicamente la baja de interacciones.
	----------------------------------------------------------------------*/
	const m_agentPollerEvents = {
		hnd_remove : {},
		subscribe  : {},
		unsubscribe: {},
		/* Funcion que invoca a funciones que escuchan eventos de actualizacion de las interacciones via AgentPoller */
		init       : (phoneClient, interaction) => {
			console.log("PhoneClient.m_agentPollerEvents.init -> interaction.id: " + interaction.id);
			m_agentPollerEvents.subscribe.interactionIsRemoved(phoneClient, interaction);
		}
	};

	/*
		Funcion que suscribe un evento para escuchar cuando las interacciones se remueven.
	*/
	m_agentPollerEvents.subscribe.interactionIsRemoved = (phoneClient, interaction) => {
		m_agentPollerEvents.unsubscribe.interactionIsRemoved(interaction);
		m_agentPollerEvents.hnd_remove[interaction.id] = inConcert.app.messagebus.subscribe(
			`/interactions/removed/${interaction.id}`,
			(removedInteraction) => phoneClient.onInteractionRemoved(removedInteraction, ()=> {
				//console.log("PhoneClient.m_agentPollerEvents.subscribe.interactionIsRemoved -> callback removedInteraction.id: " + removedInteraction.id);
				// TODO: si llega a ser necesario realizar unsubsribe del evento de la interaccion. 
			})
		);
	};

	/*
		Funcion que da de baja el evento que escucha cuando las interacciones se remueven.
	*/
	m_agentPollerEvents.unsubscribe.interactionIsRemoved = (interaction) => {
		if (!m_agentPollerEvents.hnd_remove[interaction.id])
			return;
		inConcert.app.messagebus.unsubscribe(m_agentPollerEvents.hnd_remove[interaction.id]);
		delete m_agentPollerEvents.hnd_remove[interaction.id];
	};
	// ----------------------------------------------------------------------

	/**
	 * This class is intended to work as a bridge between the low-level sip client
	 * and the inConcert"s specific domain. It should handle telephony and interaction
	 * events in a seamless manner.
	 *
	 * The main problem to solve here is the sychronization between both worlds, mostly when
	 * the network is not as reliable as we wish it was. As an example, an interaction may
	 * have been assigned to the agent, but the telephony link hasn"t come yet, or the other
	 * way around! The phone may be ringing but the interaction is not here yet!
	 *
	 * events: statusChanged(status, prev_status)
	 * @param manager
	 * @param config
	 * @param statusCallbackFn
	 */
	function PhoneClient(manager, config, statusCallbackFn)
	{
		function createSipPhone(params) {
			// note: inConcert.app.newPhone is injested by the API mockup when running tests
			// TODO: Force API mockup to define the class inConcert.SipPhone
			return inConcert.app.newPhone && inConcert.app.newPhone(params) ||
					(inConcert.app.has_allegro_integration()? 
						new inConcert.AllegroSipPhone(params) :
						new inConcert.SipPhone(params));
		}

		_.extend(this, {
			_status           : "STARTING",
			_initialized      : false,
			_builtin          : false,
			_byed             : false,
			_sip_phone        : null,
			_manager          : manager,
			_phone_type       : config.phoneType || "builtin",
			_status_callback  : statusCallbackFn || function() { window.console.info(arguments); },
			_available_actions: {},
			_view             : {},
		});
		
		config = {
			...config,
			localVideo : this.$localVideo()[0],
			remoteVideo : this.$remoteVideo()[0],
			remoteAudio : this.$remoteAudio()[0],
			bitrate: inConcert.app.get_vcc_info().VideoConfig && inConcert.app.get_vcc_info().VideoConfig.BITRATE || MAX_VIDEO_BANDWIDTH
		};

		switch (this._phone_type) {
			case "builtin":
				this._builtin = true;
				this._sip_phone = createSipPhone(config);
				this._available_actions = PhoneAction.DIAL + PhoneAction.HANGUP +
					PhoneAction.TRANSFER +
					PhoneAction.CONFERENCE +
					PhoneAction.HOLD +
					PhoneAction.MUTE +
					PhoneAction._SWITCH +
					PhoneAction.ANSWER;
				this.__registerEventHandlers();
				break;
			case "desktop":
			case "remote":
				this._available_actions = PhoneAction.DIAL + PhoneAction.HANGUP;
				break;
			default:
				throw "Unknown phone type: " + this.m_phoneType;
		}
		
		this._bindVideocallActions();
		this._setEventDisableMinimizeClient();
		this._setVideocallViewState(VideoViewStates.MINIMIZED);

		this.appPollingInterval = null;

		// info del render actual de la ventana de Tipo LiveWindow.
		this.m_renderInfo = {
			interactionId        : null,
			interactionInitiative: null,
			interactionState     : null,
			logFrom              : null,
			phoneStatus          : null,
			status               : "0-WINDOW_NEVER_OPENED",
			timestamp            : null
		};

		//_.defer(_.bind(this.connect, this));
		inConcert.Utils.zeroTimeout(() => this.connect());
	}

	inConcert.Utils.inherit(PhoneClient, EventEmitter);

	if (typeof inConcert.PhoneClient !== "function") {
		inConcert.PhoneClient = PhoneClient;
	}

	PhoneClient.prototype.status = function() {
		return this._status;
	};

	PhoneClient.prototype.builtin = function() {
		return this._builtin;
	};

	/**
	 * @brief Interfaces requirement, returns @c true iff the phone is ready to use
	 */
	PhoneClient.prototype.isInitialized = function() {
		return this._initialized;
	};

	PhoneClient.prototype.createSession = function(interaction, onIdChanged) {
		return {
			interaction: interaction,
			close : $.noop,
			setInteraction : function(interaction) {
				if (this.interaction && this.interaction.id !== interaction.id) {
					onIdChanged(this, this.interaction.id, interaction.id);
				}
				this.interaction = interaction;
			}
		};
	};

	/**
	 * @brief Interface requirement, implemented only for remote phone
	@param interaction
	 */
	PhoneClient.prototype.onNewOutgoing = function(interaction) {
		if (this._builtin === false) {
			this.setInteraction(interaction);
			throw "unused";
		}
	};

	PhoneClient.prototype.setInteraction = function(interaction) {
		console.log("PhoneClient.prototype.setInteraction -> PARAMETER interaction.id: " + (interaction || {}).id + ", current --> this._interaction.id: " + (this._interaction || {}).id);
		this._interaction = interaction;
		if (interaction) {
			m_agentPollerEvents.init(this, interaction); // suscribe eventos de la interaccion (que llegan del agent poller)
		}
	};

	PhoneClient.prototype.getInteraction = function() {
		return this._interaction;
	};

	/**
	 * @brief Interface requirement, returns whether a particular interaction is
	 * related to a session
	 * @param interaction
	 * @param session
	 */
	PhoneClient.prototype.associationValid = function(interaction, session) {
		if (inConcert.app.has_allegro_integration()) {
			return interaction.id === session.id;
		}
		return interaction.callerid === session.callerid ||
			(session.callerid.match(new RegExp("sip:("+interaction.contactAddress+")@","g"))||[]).length===1;
	};

	/**
	 * @brief Interface requirement. Since we only support one call at a time, this
	 * function returns the active call session.
	 * @param interaction
	 */
	PhoneClient.prototype.findInteractionSession = function(/*interaction*/) {
		if (this._builtin === false) {
			return {
				close         : _.identity,
				setInteraction: _.identity
			};
		}
		// TODO: Implement a reasonable comparison
		return this._active_call;
	};

	/**
	 * @brief Sets the default view
	 *
	 * @param view An object supporting one of the following operations
	 * @li clientUpdate(interaction)
	 * @li switchTo(interaction)
	 * @li update(status, prev_status)
	 */
	PhoneClient.prototype.setView = function(view) {
		let interaction_id = (this._interaction || {}).id;
		console.log("PhoneClient.prototype.setView -> interactionId: " + interaction_id + ", view: " + view);
		this._view = this._default_view = view;
		let result = this.__onViewChanged();
		return result;
	};
	
	PhoneClient.prototype.getView = function() {
		return this._view;
	};

	/**
	 * @brief Temporarily changes the view
	 *
	 * @return zero on failure, > 0 if the callback was changed, < 0 if the callback was already set
	 * @param view
	 * @param options
	 */
	PhoneClient.prototype.changeView = function(view, options) {
		if (this._view === view) {
			return -1;
		}

		options = _.extend({
			sound_enabled: true
		}, options);

		if (this.hasTelephonyWindow() && this._status !== "CONNECTED") {
			return 0;
		}
		this._sound_enabled = options.sound_enabled;
		let result = this.setView(view);
		return result; // returns 1
	};

	/**
	 * @brief Restores the default view
	 * @return the same as @c changeView
	 */
	PhoneClient.prototype.restoreView = function() {
		return this.changeView(this._default_view);
	};

	PhoneClient.prototype.__playRingback = function(status, interaction) {
		
		let play = false;
		//Aca las condiciones
		const transfer_type = interaction?.transferInfo?.target_data.type || interaction?.transferInfo?.target_data.to;
		if (status === "TRANSFER TRYING" && interaction.state === "TRANSFER TRYING" && (transfer_type === "member" || transfer_type === "number") ) {
			play = true;
		}

		if (play) {//Aca play
			this.__playSound("sounds/phone_ringback.mp3");
		}
	};

	/**
	 * @brief Called by the interaction manager in order to notify changes in the interaction.
	 * @param interaction
	 */
	PhoneClient.prototype.clientUpdate = function(interaction, callback = $.noop) {
		if (interaction && (!this._interaction || this._interaction.id !== interaction.id)) {
			let assignedInteractions = inConcert.app.getInteractionManager().getAssignedInteractionsByChannel("CALL");
			if (!$.isEmptyObject(assignedInteractions) && assignedInteractions[interaction.id] !== undefined) {
				console.info("PhoneClient.prototype.clientUpdate 1-> INVOKE -> this.switchTo(interaction, callback); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
				this.switchTo(interaction, callback);
			}
		}

		if (interaction.state === "QUEUED" || interaction.state === "OPEN") {
			return;
		}

		if (this._builtin === false) {
			return this.setInteraction(interaction);
		}

		if (!interaction || !this.hasTelephonyWindow() || !this._view.clientUpdate) {
			return;
		}

		if (this._interaction) {
			if (this._interaction.id !== interaction.id) {
				if (this._interaction.state === "PENDING" && !_.contains(["RESERVED", "TRANSFER LOCKED", "CONFERENCE LOCKED"], interaction.state)) {
					return;
				}
				let assignedInteractions = inConcert.app.getInteractionManager().getAssignedInteractionsByChannel("CALL");
				if (interaction.state === "TAKEN" && assignedInteractions[interaction.id] === undefined) {
					return;
				}
			}
			if (this._interaction.state === "PENDING" && _.contains(["TRANSFER LOCKED", "CONFERENCE LOCKED"], interaction.state)) {
				console.info("PhoneClient.prototype.clientUpdate 2-> INVOKE -> this.switchTo(interaction, callback); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
				this.switchTo(interaction, callback);
			}
		}
		if (!inConcert.app.has_allegro_baragent()) {
			// The SIPPhone doesn"t notify transferring, so we rely on the interaction status
			if (this._status === "ACTIVE" && interaction.state === "TRANSFER LOCKED") {
				this._interaction.changeState("TRANSFER TRYING");
				this.clientUpdate(this._interaction, callback);
				this.__changeStatus("TRANSFER TRYING");
				this.__playRingback(this._status, interaction);
				inConcert.app.getAgentApiServer().CallTransferring(this._interaction);
			}
			else if (this._status === "TRANSFER TRYING" && interaction.state === "TRANSFER TAKEN") {
				if (this._interaction.__last_step.from === this._manager.m_user + "@" + this._manager.m_domain) {
					this.__changeStatus("TRANSFER TAKEN");
					inConcert.app.getAgentApiServer().CallTransferConnected(this._interaction);
				} else {
					this.__changeStatus("INCOMING TRANSFER TAKEN");
				}
			}
			else if (_.contains(["TRANSFER TRYING", "TRANSFER TAKEN", "INCOMING TRANSFER TAKEN"], this._status) &&
				interaction.state === "TAKEN")
			{
				this.__changeStatus("ACTIVE");
				inConcert.app.getAgentApiServer().CallTransferCanceled(this._interaction);
				inConcert.app.getAgentApiServer().CallTaken(this._interaction);				
			}
			else if (this._status === "ACTIVE" && interaction.state === "CONFERENCE LOCKED") {
				this._interaction.changeState("CONFERENCE TRYING");
				this.clientUpdate(this._interaction, callback);
				this.__changeStatus("CONFERENCE TRYING");
				inConcert.app.getAgentApiServer().CallConferenceConsultationTrying(this._interaction);
			}
			else if (this._status === "CONFERENCE TRYING" && interaction.state === "CONFERENCE TAKEN") {
				if (this._interaction.__last_step.from === this._manager.m_user + "@" + this._manager.m_domain) {
					this.__changeStatus("CONFERENCE TAKEN");
					inConcert.app.getAgentApiServer().CallConferenceConsultationConnected(this._interaction);
				} else {
					this.__changeStatus("INCOMING CONFERENCE TAKEN");
				}
			}
			else if (_.contains(["CONFERENCE TRYING", "CONFERENCE TAKEN", "INCOMING CONFERENCE TAKEN"], this._status) &&
				interaction.state === "TAKEN")
			{
				this.__changeStatus("ACTIVE");
				inConcert.app.getAgentApiServer().CallConferenceConsultationCanceled(this._interaction);
				inConcert.app.getAgentApiServer().CallTaken(this._interaction);
			}
			else if (this._status === "CONNECTED" && interaction.state === "LOCKED" && 
					interaction.isPreview()) {
				this.__changeStatus("PREVIEW");
				inConcert.app.getAgentApiServer().CallOnPreview(interaction);
			}
			else if (this._status === "CONNECTED" && interaction.state === "WAITING" && 
					(interaction.initiative === "INBOUND" || interaction.isDialer())) {
				console.error("PhoneClient status is 'CONNECTED' while 'INBOUND' interaction is 'WAITING'");
				this._view.close();
				return;
			}
		}
		
		this._setVideocallActions();

		this._view.clientUpdate(interaction);
	};

	/*
		Funcion encargada de abrir una ventana del telefono de tipo Live.
		Invoca a una funcion recursiva, con un timeout, esperando que el telefono sea cargado y este ready.
		Esto lo realiza asi, para despues poder invocar a la funcion "switchTo" de la clase "LiveClientWindow" en el archivo "phone_client_window.js"
	*/
	PhoneClient.prototype.__openLiveTelephoneWindow = function(interaction, callback = $.noop)
	{
		const self = this;
		const timemout_ms = 500;
		const max_retries = 50;
		let without_phone_retries = 0;
		let with_phone_retries = 0;
		let timemout_hnd;
		let session_window;
		let startOpenLiveTelephoneWindow;

		let currentInteractionLog = "Current Interaction - Id: " + interaction.id + ", state: " + interaction.state + ", initiative: " + interaction.initiative;
		const renderInfo = this.__getRender("__openLiveTelephoneWindow INICIO");
		console.log("PhoneClient.prototype.__openLiveTelephoneWindow - Start - Render status: " + JSON.stringify(renderInfo) + ". " + currentInteractionLog);

		if (this.__existAnotherRenderInProgress(currentInteractionLog)) {
			let err_type = renderInfo.interactionId === interaction.id ? "1 - ANOTHER RENDER IN PROGRESS FOR THIS INTERACTION" : "2 - ANOTHER RENDER IN PROGRESS FOR OTHER INTERACTION";
			console.log("PhoneClient.prototype.__openLiveTelephoneWindow - Another render in progress- Render status: " + JSON.stringify(renderInfo) + ". " + currentInteractionLog + ", ERR_TYPE: " + err_type);
			callback(false, {
				err: {
					err_type      : err_type,
					current_render: renderInfo
				}
			});
			return;
		}

		/*
			Flag utilizado para determinar si la ventana se debe cerrar de forma automatica, luego de haber culmiando el render.
			Ej: para el caso en el cual la interaccion se ha cancelado mientras se realizaba el render.
		*/
		let must_force_window_close = false; 

		const isGone = function(it) {
			const statesOfGoneInteraction = ["ABANDONED", "CANCELED", "FINISHED"];
			let isGoned = $.inArray(it.state, statesOfGoneInteraction) >= 0;
			if (isGoned) {
				console.info("[PhoneClient.prototype.__openLiveTelephoneWindow] 2--> Is Gone interaction '" + it.id + "', state "  + it.state);
			}
			return isGoned;
		};

		//Funcion recursiva
		//
		startOpenLiveTelephoneWindow = function() {
			// Valida si el telefono ha comenzado el render.
			if (self.hasTelephonyWindow(true)) {
				console.log("PhoneClient.prototype.__openLiveTelephoneWindow 1 ---> With PHONE - tried number: " + (with_phone_retries + 1) + " of " + max_retries + ". Current Interaction Id: " + interaction.id);

				window.clearTimeout(timemout_hnd);
				// si la funcion "switchTo" se encuentra disponible, quiere decir que la carga de la ventana esta ready, y podemos proceder
				// a actualizar el telefono, invocando a esta funcion para que actualice el mismo segun la interaccion.
				if (self._view.switchTo) {
					console.log("PhoneClient.prototype.__openLiveTelephoneWindow - Phone Windows is READY. Interaction Id: " + interaction.id);
					if (!inConcert.app.has_allegro_integration() && interaction.initiative === "INBOUND" && interaction.state !== "WRAPUP" && self._sip_phone.currentState !== "inbound" && self._sip_phone.currentState !== "active") {
						let err_type = "3 - INBOUND INTERACTION WITHOUT 'INBOUND' STATE ON SIPPHONE";
						console.error("PhoneClient.prototype.__openLiveTelephoneWindow - " + err_type);
						callback(true, {
							session_window: session_window,
							callback: function() {
								console.info("PhoneClient.prototype.__openLiveTelephoneWindow 1-> Phone Windows FORCE CLOSE for Interaction Id: " + interaction.id);
								inConcert.app.getInteractionManager().cancelInteraction(interaction);
								interaction.reject();
								self._view.close();
								self.__changeStatus("CONNECTED", err_type);
							},
							err: {
								err_type      : err_type,
								current_render: self.__getRender("__openLiveTelephoneWindow " + err_type)
							}
						});
						self.__setRender("5-WINDOW_LOADING_INIT_ERR_EXEC_CANCEL_CURRENT_RENDER", interaction, "FROM __openLiveTelephoneWindow FUNC - err_type: " + err_type);
						return;
					}

					let is_goned = isGone(interaction);
					//is_goned = true;  // para buscar problemas cuando la interaccion se ha cancelado justo antes del ready.
					if (must_force_window_close || is_goned) {
						// Se debe cerrar la ventana de forma forzada (ejemplo: si mientras estaba renderizando la misma, la interaccion fue cancelada)
						let err_type = "8 - UNABLE TO OPEN LIVEPHONECLIENTWINDOW - INTERACTION IS GONE";
						console.info("PhoneClient.prototype.__openLiveTelephoneWindow - " + err_type + ", must_force_window_close:" + must_force_window_close + ", is_goned: " + is_goned + ", session_window: " + session_window);
					}

					// OK - El telefono esta listo, ahora comienza la carga de la sidebar, integraciones, etc !!!!!
					self.__setRender("8-WINDOW_IS_READY_WAITING_LOADING_INTERACTION_INFO", interaction, "IS OK - FROM __openLiveTelephoneWindow FUNC");
					self._view.switchTo(interaction);
					self.clientUpdate(interaction);
					// TODO: #9778 - CANY - Docked.
					callback(true, {
						session_window: session_window,
						callback: function() {
							console.info("PhoneClient.prototype.__openLiveTelephoneWindow -> Phone Windows is READY AND OPEN for Interaction Id: " + interaction.id + ", state: " + interaction.state + ", phoneStatus: " + self._status);
						}
					});
					//callback(false);
					return;
				}

				with_phone_retries++;
				self.__setRender("3-WINDOW_INIT_FRM_IN_PROGRESS_OK", interaction, "FROM __openLiveTelephoneWindow FUNC - RETRY " + with_phone_retries.toString());

				if (with_phone_retries > max_retries) {
					let err_type = "4 - OPENING WINDOW. YOU HAVE REACHED THE MAXIMUM NUMBER OF ATTEMPTS";
					console.error("PhoneClient.prototype.__openLiveTelephoneWindow - " + err_type + ". Render status: " + JSON.stringify(self.__getRender("__openLiveTelephoneWindow log")) + ". Current Interaction Id: " + interaction.id);
					without_phone_retries = 0;
					with_phone_retries = 0;
					// llego al limite maximo de reintentos, aca no deberia llegar nunca.
					callback(false, {
						err: {
							err: {
								err_type      : err_type,
								current_render: self.__getRender("__openLiveTelephoneWindow " + err_type)
							}
						}
					});
					self.__setRender("6-WINDOW_LOADING_INIT_ERR_TIMEOUT_EXEC_CANCEL_CURRENT_RENDER", interaction, "FROM __openLiveTelephoneWindow FUNC - err_type: " + err_type);
					return;
				}
				const _interaction = inConcert.app.getInteractionManager().findInteraction(interaction.id);
				if (!_interaction || _interaction.isGone()) { // === "CANCELED"
					let err_type = "5 - UNABLE TO OPEN LIVEPHONECLIENTWINDOW - INTERACTION IS GONE";
					console.log("PhoneClient.prototype.__openLiveTelephoneWindow a- " + err_type + ". Interaction Id: " + interaction.id + ", interaction.state: " + interaction.state + ", interaction.initiative: " + interaction.initiative);
					callback(true, {
						session_window: session_window,
						callback: function() {
							console.info("PhoneClient.prototype.__openLiveTelephoneWindow 3-> Phone Windows FORCE CLOSE for Interaction Id: " + interaction.id);
							self._view.close();
						},
						err: {
							err_type      : err_type,
							current_render: self.__getRender("__openLiveTelephoneWindow " + err_type)
						}
					});
					self.__setRender("11-WINDOW_LOADING_INIT_ERR_INTERACTION_IS_GONE_EXEC_CANCEL_CURRENT_RENDER", interaction, "FROM __openLiveTelephoneWindow FUNC - err_type: " + err_type);
					return;
				}
				// ejecuto un timeout para esperar a invocar nuevamente a la funcion recursiva.
				timemout_hnd = window.setTimeout(function() {
					// invoco a la funcion recursiva para que valide si el telefono se encuentra ready.
					startOpenLiveTelephoneWindow();
				}, timemout_ms);
			}
			else {
				window.clearTimeout(timemout_hnd);
				console.log("PhoneClient.prototype.__openLiveTelephoneWindow 2 ===>  Without PHONE - tried number: " + (without_phone_retries + 1) + " of " + max_retries + ". Interaction Id: " + interaction.id + ", state: " + interaction.state);

				if (!self.__renderRunning().isRunning) {
					// El telefono no ha comenzado el render, entonces invoca a la funcion para que cree la ventana.
					let is_goned = isGone(interaction);
					const _interaction = inConcert.app.getInteractionManager().findInteraction(interaction.id);
					if (is_goned || !_interaction || _interaction.isGone()) {
						let err_type = "7 - UNABLE TO OPEN LIVEPHONECLIENTWINDOW - INTERACTION IS GONE";
						must_force_window_close = true;
						console.log("PhoneClient.prototype.__openLiveTelephoneWindow 1- Must Force Window Close. - " + err_type + ". Interaction Id: " + interaction.id + ", interaction.state: " + interaction.state + ", interaction.initiative: " + interaction.initiative + ", must_force_window_close=true");
						self.__setRender("11-WINDOW_LOADING_INIT_ERR_INTERACTION_IS_GONE_EXEC_CANCEL_CURRENT_RENDER", interaction, "FROM __openLiveTelephoneWindow FUNC - err_type: " + err_type);
						return;
					}
					else {
						self.__setRender("1-WINDOW_LOADING_START_OK", interaction, "FROM __openLiveTelephoneWindow FUNC - RETRY " + without_phone_retries.toString());
						console.log("PhoneClient.prototype.__openLiveTelephoneWindow a- Opening window. Render status: " + JSON.stringify(self.__getRender("log")) + ". Current Interaction Id: " + interaction.id);
	
						// TODO: #9778 - CANY - Docked.
						session_window = inConcert.app.openLiveTelephoneWindow(interaction,
							// on_ready fn
							function() {
								console.log("PhoneClient.prototype.__openLiveTelephoneWindow - on_ready callback");
							},
							// on_close fn
							function() {
								console.log("PhoneClient.prototype.__openLiveTelephoneWindow - on_close callback");
								self.onWindowClosed(interaction, "PHONE_CLIENT");
							}
						);
						console.log("PhoneClient.prototype.__openLiveTelephoneWindow b- Opening window. Render session_window: " + session_window + ". Current Interaction Id: " + interaction.id);
					}
				}
				else {
					// el render esta en proceso pero aun no ha finalizado.
					self.__setRender("2-WINDOW_LOADING_IN_PROGRESS_OK", interaction, "FROM __openLiveTelephoneWindow FUNC - RETRY " + without_phone_retries.toString());
					console.log("PhoneClient.prototype.__openLiveTelephoneWindow c- Opening window. Render status: " + JSON.stringify(self.__getRender("log")) + ". Current Interaction Id: " + interaction.id);
				}
				//window.clearTimeout(timemout_hnd);
				without_phone_retries++; // incremento el retry
				if (without_phone_retries > max_retries) {
					let err_type = "6 - OPENING WINDOW. YOU HAVE REACHED THE MAXIMUM NUMBER OF ATTEMPTS";
					console.error("PhoneClient.prototype.__openLiveTelephoneWindow - " + err_type + ". Render status: " + JSON.stringify(self.__getRender("log")) + ". Current Interaction Id: " + interaction.id);
					// llego al limite maximo de reintentos, aca no deberia llegar nunca.
					without_phone_retries = 0;
					with_phone_retries = 0;
					callback(false, {
						err: {
							err_type      : err_type,
							current_render: self.__getRender("__openLiveTelephoneWindow " + err_type)
						}
					});
					self.__setRender("4-WINDOW_LOADING_INIT_ERR_BUT_EXEC_ACTION_CONTINUE_WITH_CURRENT_RENDER", interaction, "FROM __openLiveTelephoneWindow FUNC - err_type: " + err_type);
					//TODO: HACER ALGO
					return;
				}
				let is_goned = isGone(interaction);
				const _interaction = inConcert.app.getInteractionManager().findInteraction(interaction.id);
				if (is_goned || !_interaction || _interaction.isGone()) {
					let err_type = "7 - UNABLE TO OPEN LIVEPHONECLIENTWINDOW - INTERACTION IS GONE";
					must_force_window_close = true;
					console.log("PhoneClient.prototype.__openLiveTelephoneWindow 2- Must Force Window Close. - " + err_type + ". Interaction Id: " + interaction.id + ", interaction.state: " + interaction.state + ", interaction.initiative: " + interaction.initiative + ", must_force_window_close=true");
					self.__setRender("4-WINDOW_LOADING_INIT_ERR_BUT_EXEC_ACTION_CONTINUE_WITH_CURRENT_RENDER", interaction, "FROM __openLiveTelephoneWindow FUNC - err_type: " + err_type);
				}
				// ejecuto un timeout para esperar a invocar nuevamente a la funcion recursiva.
				if (!must_force_window_close) {
					timemout_hnd = window.setTimeout(function() {
						// invoco a la funcion recursiva para que valide si el telefono se encuentra ready.
						startOpenLiveTelephoneWindow();
					}, timemout_ms);
				}
			}
		};

		// // invoco a la funcion que comienza con el proceso de creacion de la ventana.
		startOpenLiveTelephoneWindow();
	};

	PhoneClient.prototype.switchTo = function(interaction, callback)
	{
		const self = this;

		let currentInteraction = (this._interaction || {});
		console.info("PhoneClient.prototype.switchTo 1---> START Switch To interaction.Id: " + interaction.id + ", interaction.state: " +  interaction.state +
			" - currentInteraction.id: " + currentInteraction.id + ", currentInteraction.state: " + currentInteraction.state + ", callback: " + callback);

		callback = callback || function(session_window, on_callback = $.noop) {
			console.log("[PhoneClient.prototype.switchTo] --> Default Callback --> interactionId: " + interaction.id + ", state: " + interaction.state + ", session_window: " + session_window + ", on_callback:(" + (session_window ? true : false) + ")");
			if (!session_window) {
				on_callback(false);
				return;
			}
			on_callback(true);
		};

		if (this._builtin === false) {
			this.setInteraction(interaction);
			return {
				getWindow: function() {
					return {
						activate: _.identity,
						close   : _.identity
					};
				},
				on: function() { }
			};
		}

		let switchTo;
		let retry = 0;
		const max_retries = 50;
		switchTo = function() {
			retry++;
			console.info("PhoneClient.prototype.switchTo    N.1-> switchTo (invoke __openLiveTelephoneWindow) Interaction: " + interaction.id + ", state: " + interaction.state + ", phoneStatus: " + self.status() + ", retry: " + retry + " of " + max_retries);
			// valida si la nueva interaccion a abrir existe en el "interactionManager"
			const newRenderInteraction = inConcert.app.getInteractionManager().findInteraction(interaction.id);
			if (!newRenderInteraction) {
				if (!self.__renderRunning().isRunning) {
					// si no existe la interaccion en el "interactionManager" y no esta corriendo un runnig, debe marcar como canceled el current render.
					console.info("PhoneClient.prototype.switchTo N.1.1-----> interaction not found !! current render IS NOT RUNNING - retry: " + retry);
					self.__setRender("5-WINDOW_LOADING_INIT_ERR_EXEC_CANCEL_CURRENT_RENDER", null, "FROM SWITCHTO FUNC - RENDER NEW INTERACTION NOT FOUND");
				}
				else {
					console.info("PhoneClient.prototype.switchTo N.1.2-----> NEW interaction not found !! current render IS RUNNING");
				}
			}

			// TODO: Revisar si esto no tiene que ir antes de la validacion anterior
			self._active_call = null;
			self.setInteraction(interaction);

			// Abro o actualizo el telefono.
			self.__openLiveTelephoneWindow(interaction, function(success, info = {}) {
				let session_window = info.session_window;
				if (!success) {
					let err = info.err;
					console.info("PhoneClient.prototype.switchTo N.2-> CALLBACK **> NO Switch To Interaction: " + interaction.id + ", err: " + JSON.stringify(err));
					if (err.err_type === "2 - ANOTHER RENDER IN PROGRESS FOR OTHER INTERACTION") {
						let mustRetry;
						const current_render_interaction = inConcert.app.getInteractionManager().findInteraction(err.current_render.interactionId);
						if (current_render_interaction) {
							let current_render_interaction_state = current_render_interaction ? current_render_interaction.state : "WITHOUT STATE";
							console.info("PhoneClient.prototype.switchTo N.3-> CALLBACK **> Validate FORCE Switch RENDER to Interaction: " + interaction.id + ". (Current Render Interaction: " + err.current_render.interactionId + ", state: " + current_render_interaction_state + ")");
							const states =  ["PENDING", "CANCELED", "CREATED", "ABANDONED", "FINISHED"];
							mustRetry = $.inArray(current_render_interaction_state, states) >= 0;
						}
						else {
							console.info("PhoneClient.prototype.switchTo N.3.1 -*----> CALLBACK ---> Validate FORCE Switch RENDER to Interaction: " + interaction.id + ". (Current Render Interaction: " + err.current_render.interactionId + " DOES NOT EXIST !!!!!)");
							mustRetry = true;
							if (!self.__renderRunning().isRunning) {
								console.info("PhoneClient.prototype.switchTo N.3.2-----> interaction not found !! current render IS NOT RUNNING - retry: " + retry);
								self.__setRender("5-WINDOW_LOADING_INIT_ERR_EXEC_CANCEL_CURRENT_RENDER", null, "FROM SWITCHTO FUNC - RENDER CURRENT INTERACTION NOT FOUND");
							}
							else {
								console.info("PhoneClient.prototype.switchTo N.3.3-----> CURRENT interaction not found !! current render IS RUNNING, please Wait - retry: " + retry);
							}
						}
						if (mustRetry) {
							if (retry > max_retries) {
								console.info("PhoneClient.prototype.switchTo    N.4-> switchTo (invoke __openLiveTelephoneWindow) Interaction: " + interaction.id + ", has reached the maximum number of retries. max_retries: " + max_retries);
								callback(false);
								return;
							}
							const timemout_ms = 200;
							window.setTimeout(function() {
								switchTo();
							}, timemout_ms);
							return;
						}
						else {
							console.info("PhoneClient.prototype.switchTo    N.5-> switchTo (invoke __openLiveTelephoneWindow) Interaction: " + interaction.id + ", Unable to RENDER");
							callback(false);
							return;
						}
					}
					callback(false);
					return;
				}
				console.info("PhoneClient.prototype.switchTo N.6 ----> CALLBACK --> YES Switch To Interaction: " + interaction.id + ", session_window: " + session_window + ", info.callback: " + info.callback);
				callback(session_window, info.callback);
			});
		};

		this.m_pubsub_close_phone_windows = this.m_pubsub_close_phone_windows || {};
		const cancel_result = function() {
			if (!self._interaction) {
				return {cancel: false, type: "NO CURRENT INTERACTION", _interaction: null};
			}
			if (currentInteraction.id === interaction.id) {
				return {cancel: false, type: "SAME INTERACTION", interaction_id: interaction.id, currentInteraction_id: currentInteraction.id};
			}
			if (currentInteraction.initiative === "OUTBOUND") {
				if (currentInteraction.isBeingComposed()) {
					return {cancel: true, type: "EXIST ANOTHER INTERACTION & CURRENT INTERACTION IS OUTBOUND", interaction_id: interaction.id, currentInteraction_id: currentInteraction.id, currentInteractionIsBeingComposed: true};
				}
			}
			return {cancel: false, type: "DEFAULT", interaction_id: interaction.id, currentInteraction_id: currentInteraction.id};
		}();

		console.info("PhoneClient.prototype.switchTo ****--->  CANCEL_RESULT: " + JSON.stringify(cancel_result));

		if (cancel_result.cancel) {
			// //Valido si el estado actual del phone client es Offering y la otra interaccion existente esta en locked (is phone_ringback) 
			const isOfferingLocked = self.status() === "OFFERING" && interaction.state === "LOCKED";
			console.log("PhoneClient.prototype.switchTo ------> A- First Action is Cancel Interaction: " + currentInteraction.id + ". Second action is Switch To Interaction: " + interaction.id + ". isOfferingLocked: " + isOfferingLocked);
			inConcert.app.getInteractionManager().cancelCallPendingInteraction(currentInteraction, function(mustCanceled) {
				let currentId = currentInteraction.id;
				if (!mustCanceled) {
					console.info("PhoneClient.prototype.switchTo -> a.1 NO Cancel Interaction: " + currentId);
					switchTo();
					return;
				}
				//-----------------------------------------------
				// 1) Suscribo un evento en el callback del Panel "base_panel.js" (Panel.prototype.onClientClosed), para que cuando se cierre la ventana de la interaction,
				//	nos llegue el evento y se pueda proceder a realizar el switchTo (open de la ventana)
				//
				console.info("PhoneClient.prototype.switchTo -> a.2 Suscribe event of close windows interaction: " + currentId);
				if (self.m_pubsub_close_phone_windows[currentId]) {
					inConcert.app.messagebus.unsubscribe(self.m_pubsub_close_phone_windows[currentId]);
					delete self.m_pubsub_close_phone_windows[currentId];
				}
				self.m_pubsub_close_phone_windows[currentId] = inConcert.app.messagebus.subscribe(
					"/live_phone_client_window/interaction/" + currentId + "/windows_closed/",
					function(closedInteractionWindow) {
						console.info("PhoneClient.prototype.switchTo -> a.2.1 Received event close windows interaction: " + closedInteractionWindow.id);
						if (self.m_pubsub_close_phone_windows[currentId]) {
							inConcert.app.messagebus.unsubscribe(self.m_pubsub_close_phone_windows[currentId]);
							delete self.m_pubsub_close_phone_windows[currentId];
						}
						if (isOfferingLocked && interaction.state === "LOCKED" && self.status() === "CONNECTED") {
						//En este pto el estado del phoneclient queda en CONNECTED, despues de cancelar el outbound, lo que hace que para la otra interaccion no se muestre el panel de contestar
						// Para esto se vuelve a llamar al ring y se agrega el estado en OFFERING del client phone
							console.log("PhoneClient.prototype.switchTo ------> B- LOCKED and CONNECTED -> First Action is Cancel Interaction: " + currentInteraction.id + ". Second action is Switch To Interaction: " + interaction.id + ". isOfferingLocked: " + isOfferingLocked);
							self.__changeStatus("OFFERING");
							self.__playSound("sounds/phone_ringback.mp3");
						}
						// Timeout para que la ejecucion quede desacoplada (por las dudas)
						window.setTimeout(function() {
							switchTo();
						});
					}
				);
				//-----------------------------------------------
				// 2) Se suscribe un evento para el caso en que la interaccion NO existe en el "interactionManager".
				//    Sino estas ventanas NO se cierran de forma automatica.
				//
				console.info("PhoneClient.prototype.switchTo -> a.3 Suscribe event of close windows for " + EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND);
				self.m_pubsub_close_phone_windows[EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND] = inConcert.app.messagebus.subscribe(
					`/live_phone_client_window/interaction/${ EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND }/windows_closed/`,
					function() {
						console.info("PhoneClient.prototype.switchTo -> a.3.1 Received event close windows interaction: " + EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND);
						if (self.m_pubsub_close_phone_windows[EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND]) {
							inConcert.app.messagebus.unsubscribe(self.m_pubsub_close_phone_windows[EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND]);
							delete self.m_pubsub_close_phone_windows[EVENT_SUBSCRIBE_NOT_INTERACTION_FOUND];
						}
						if (isOfferingLocked && interaction.state === "LOCKED" && self.status() === "CONNECTED") {
						//En este pto el estado del phoneclient queda en CONNECTED, despues de cancelar el outbound, lo que hace que para la otra interaccion no se muestre el panel de contestar
						// Para esto se vuelve a llamar al ring y se agrega el estado en OFFERING del client phone
							console.log("PhoneClient.prototype.switchTo ------> B- LOCKED and CONNECTED -> First Action is Cancel Interaction: " + currentInteraction.id + ". Second action is Switch To Interaction: " + interaction.id + ". isOfferingLocked: " + isOfferingLocked);
							self.__changeStatus("OFFERING");
							self.__playSound("sounds/phone_ringback.mp3");
						}
						// Timeout para que la ejecucion quede desacoplada (por las dudas)
						window.setTimeout(function() {
							switchTo();
						});
					}
				);
			});
		}
		else {
			console.info("PhoneClient.prototype.switchTo 3-> Switch To Interaction: " + interaction.id);
			switchTo();
		}
	};

	/*
		Esta funcion lo unico que realiza es invocar a la funcion switchTo,
			pero la misma queda implementa para que quede mas legible el codigo, desde donde lo invocan para abrir un telefono.
			Ej: desde la funcion "Panel.prototype.openInteractionClient" en el archivo "base_panel.js" al momento de crear una saliente.
	*/
	PhoneClient.prototype.openLiveTelephoneWindow = function(interaction, callback = $.noop)
	{
		console.info("PhoneClient.prototype.openLiveTelephoneWindow -> INVOKE -> this.switchTo(interaction, callback); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
		this.switchTo(interaction, callback);
	};

	/**
	 * @brief Used to test the phone in order to known whether an operation is
	 * available or not (some buttons will be shown, others will be hidden)
	 *
	 * @params operation The name of a known operation, possible values are:
	 *   @list dial       (   1 )
	 *   @list answer     (   2 )
	 *   @list hangup     (   4 )
	 *   @list mute       (   8 )
	 *   @list hold       (  16 )
	 *   @list transfer   (  32 )
	 *   @list conference (  64 )
	 *   @list switch     ( 128 )	 
	 *   @param operation
	 */
	PhoneClient.prototype.supports = function(operation) {
		var bitval = typeof(operation) === "number" ? operation : PhoneAction[operation.toUpperCase()];
		return bitval && ((this._available_actions & bitval) === bitval);
	};

	/**
	 * @brief Starts registering with the SIP proxy
	 */
	PhoneClient.prototype.connect = function() {
		if ( this._builtin ) {
			this._active_call = null;
			this.__changeStatus("CONNECTING"); // Should replace status callback
			this._sip_phone.connect();
		} else {
			this.__activateAddress();
		}
	};

	PhoneClient.prototype.reconnect = function() {
		this.__changeStatus("RECONNECTING");
		if ( this._builtin ) {
			this._status_callback("CONNECTING");
		}
		this.connect();
	};

	PhoneClient.prototype.isPeerRegistered = function() {
		
		if ( this._sip_phone && this._sip_phone.getPhoneState() !== "connecting" && this._sip_phone.m_initialized && this._sip_phone.isUserRegistered) {
			return this._sip_phone.isUserRegistered();
		} 

		return true;
	};

	/**
	 * @brief Disconnects from the SIP proxy, called at exit
	 */
	PhoneClient.prototype.close = function() {
		this._initialized = false;
		return this._builtin && this._sip_phone.disconnect();
	};

	/**
	 * @brief Sends a INVITE request the the number specified by @p destination
	 * @param campaign
	 * @param prefix
	 * @param destination
	 * @param contact
	 */
	PhoneClient.prototype.dial = function(params) {
		const interactionID = params.interactionID;
		const campaignAtVCC = params.campaignAtVCC;
		const accountId     = params.accountId;
		const prefix = params.prefix;
		const destination = params.destination;
		const contact = params.contact;
		const isInternal = params.isInternal || false;
		const callback = params.callback || $.noop;
		
		if (this._builtin) {
			this.__changeStatus("DIALING");
			this._sip_phone.invite({
				interactionID : interactionID,
				campaignAtVCC : campaignAtVCC,
				accountId    : accountId,
				prefix : prefix,
				destination : destination,
				contact : contact,
				isInternal : isInternal,
				callback : callback
			});
		}
		else {
			callback(false, "phone not found");
		}
	};

	/**
	 * @brief Accepts an incoming call, does not affect the interaction
	 */
	PhoneClient.prototype.answer = function(interaction, callback = $.noop) {
		if (this._builtin) {
			return this._sip_phone.accept(callback);
		}
	};

	PhoneClient.prototype.reject = function(interaction, reason, callback = $.noop) {
		if (this._builtin) {
			return this._sip_phone.reject(reason, callback);
			// TODO: byed?
		} else {
			this.hangup(interaction, callback);
		}
	};

	PhoneClient.prototype.hold = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.hold(function(ok, err) {
					callback(ok, err);
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("in_hold", true);
					}
				});
			}
			else
			{
				this._sip_phone.hold(callback);
			}
		}
	};

	PhoneClient.prototype.unhold = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.unhold(function(ok, err) {
					callback(ok, err);
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("in_hold", false);
					}
				});
			}
			else
			{
				this._sip_phone.resume(callback);
			}
		}
	};

	PhoneClient.prototype.mute = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._sip_phone.mute(function(ok, err) {
					callback(ok, err);
					if (ok) {
						self._sip_phone.emit("muted", true);
					}
				});
			}
			else
			{
				this._sip_phone.mute(callback);
			}
		}
	};

	PhoneClient.prototype.unmute = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._sip_phone.unmute(function(ok, err) {
					callback(ok, err);
					if (ok) {
						self._sip_phone.emit("muted", false);
					}
				});
			}
			else
			{
				this._sip_phone.unmute(callback);
			}
		}
	};

	PhoneClient.prototype.disableVideo = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.disableVideo(function(ok, err) {
					if (callback) {
						callback(ok, err);
					}
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("video_disabled", true);
					}
				});
			}
			else
			{
				this._sip_phone.disableVideo(callback);
			}
		}
	};

	PhoneClient.prototype.enableVideo = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.enableVideo(function(ok, err) {
					if (callback) {
						callback(ok, err);
					}
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("video_disabled", false);
					}
				});
			}
			else
			{
				this._sip_phone.disableVideo(callback);
			}
		}
	};

	PhoneClient.prototype._switch = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.switchConversation(function(ok, err) {
					callback(ok, err);
					if (ok) {
						self._sip_phone.emit("switched", true);
					}
				});
			}
			else
			{
				this._sip_phone._switch(callback);
			}
		}
	};

	PhoneClient.prototype.unswitch = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.unswitchConversation(function(ok, err) {
					callback(ok, err);
					if (ok) {
						self._sip_phone.emit("switched", false);
					}
				});
			}
			else
			{
				this._sip_phone.unswitch(callback);
			}
		}
	};

	/**
	 * @brief Sends a single digit to the remote endpoint
	 * @param digit
	 */
	PhoneClient.prototype.sendDigit = function(interaction, digit, callback) {
		if (this._builtin) {
			this._sip_phone.sendDigit(digit, callback);
		}
	};

	/**
	 * @brief Hangs up an active call
	 */
	PhoneClient.prototype.hangup = function(interaction, callback = $.noop) {
		const self = this;
		const has_allegro_integration = inConcert.app.has_allegro_integration();

		// --Callback onHungup--------------------------------
		const onHangup = !has_allegro_integration ?
			/*
				Nativo:
					Se asigna directamente la Fn de callback que se recibe por parametro,
						ya que el SipPhone maneja internamente el evento "rejected" que termina invocando a la funcion onRejected
						que es la Fn que termina cancelando la interaccion y haciendo que en las salientes (pending) el telefono se cierre automaticamente.
			*/
			callback :
			/*
				Allegro:
					Hay que implementar la Fn de onHangup, para que termine invocando a la funcion .emit("rejected")->onRejected
			*/
			(ok, err) => {
				const isInRedial = inConcert.app.isInRedial();
				const { id, initiative, state } = interaction || {};
				console.log(`[PhoneClient.prototype.hangup] 1-> Allegro - onHangup. id: ${id}, initiative: ${initiative}, state: ${state}, phoneStatus: ${self._status}, isInRedial: ${isInRedial}. isOk: ${ok}, err: ${err}, _builtin: ${self._builtin}.`);
				if (ok) {
					if (self._builtin &&
						initiative === "OUTBOUND" &&
						$.inArray(state || "", ["PENDING", "LOCKED"]) >= 0 &&
						!isInRedial
					) {
						console.log(`[PhoneClient.prototype.hangup] 2-> Allegro - onHangup. id: ${id}, initiative: ${initiative}, state: ${state}, phoneStatus: ${self._status}, isInRedial: ${isInRedial}. --> Execute: self._sip_phone.emit('rejected')'.`);
						self._sip_phone.emit("rejected");
					}
				}
				callback(ok, err);
			};
		// ----------------------------------

		if (this._builtin) {
			if (this._sip_phone.currentState === "inbound") {
				this.reject(interaction, null, onHangup);
			} 
			else {
				this._sip_phone.bye(onHangup);
				// TODO: on byed?
			}
		}
		else {
		// TODO: Perform some kind of refactoring
			return this._interaction && this._interaction.hangup(function(ok, err) {
				if (!ok) {
					// the request could will fail if the client hangs-up just before the agent does so we do not
					//take much care about error conditions here, the interaction is going to be closed anyway
					inConcert.app.getMainWindow().showBannerMessage({
						messageType: "Alert",
						message    : "<%=_T('Something went wrong trying to hangup the call')%>"
					});
				}
				onHangup(ok, err);
			});
		}
	};

	PhoneClient.prototype.record = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.startRecord(function(ok, err) {
					callback(ok, err);
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("recording", true);
					}
				});
			}
			else
			{
				this._sip_phone.record(callback);
			}
		}
	};

	PhoneClient.prototype.stopRecord = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.stopRecord(function(ok, err) {
					callback(ok, err);
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("recording", false);
					}
				});
			}
			else
			{
				this._sip_phone.stopRecord(callback);
			}
		}
	};

	PhoneClient.prototype.acceptPreview = function(interaction, callback) {
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.acceptPreview(callback);
			}
			else
			{
				this._sip_phone.acceptPreview(callback);
			}
		}
	};

	PhoneClient.prototype.cancelPreview = function(interaction, callback) {
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.cancelPreview(callback);
			}
			else
			{
				this._sip_phone.cancelPreview(callback);
			}
		}
	};

	PhoneClient.prototype.confirmTransfer = function(interaction, callback) {
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.commitAttendedTransfer(callback);
			}
			else
			{
				this._sip_phone.confirmTransfer(callback);
			}
		}
	};

	PhoneClient.prototype.cancelTransfer = function(interaction, callback) {
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.cancelAttendedTransfer(callback);
			}
			else
			{
				this._sip_phone.cancelTransfer(callback);
			}
		}
	};

	PhoneClient.prototype.confirmConference = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.commitAttendedConference(function(ok, err) {
					callback(ok, err);
					if (ok) {
						self._sip_phone.emit("in_conference");
					}
				});
			}
			else
			{
				this._sip_phone.confirmConference(callback);
			}
		}
	};

	PhoneClient.prototype.cancelConference = function(interaction, callback) {
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				if (this._interaction.isInConference()) {
					this.hangup(interaction, callback);
				}
				else {
					this._interaction.cancelAttendedConference(callback);
				}
			}
			else
			{
				this._sip_phone.cancelConference(callback);
			}
		}
	};

	PhoneClient.prototype.replayVoicemail = function(interaction, callback) {
		if (this._builtin) {
			this._sip_phone.replayVoicemail(callback);
		}
	};

	PhoneClient.prototype.redial = function(interaction, callback = $.noop) {
		let self = this;
		let assignedPanel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
		assignedPanel.hideMenuActions();
		if (!self._builtin || self._status !== "WRAPUP" || !interaction.isPhoneValidToRedial()) {
			return callback(false);
		}

		interaction.in_redial = true;
		interaction.state = "LOCKED";
		self.__changeStatus("DIALING", "REDIAL");

		var redialCallback = function(ok, err) {
			if (!ok) {
				interaction.in_redial = false;
				interaction.state = "WRAPUP";
				self.__changeStatus("WRAPUP");
			}
			callback(ok, err);
			if (ok && !inConcert.app.has_allegro_baragent()) {
				self._sip_phone.emit("redial");
			}				
		};

		if (inConcert.app.has_allegro_baragent()) {
			self._sip_phone.redial(redialCallback);
		} else {
			self._sip_phone.invite({
				prefix : "_REDIAL_",
				destination : interaction.id,
				callback : redialCallback
			});
		}
	};

	PhoneClient.prototype.takeAppointment = function(interaction, callback = $.noop) {
		let self = this;
		let assignedPanel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
		if (self._builtin) {
			interaction.in_appointment = true;
			interaction.state = "LOCKED";
			self.__changeStatus("DIALING", "APPOINTMENT");
			var appointmentCallback = function(ok, err) {
				if (!ok) {
					interaction.in_appointment = false;
					interaction.state = "WRAPUP";
					self.__changeStatus("WRAPUP");
				}
				callback(ok, err);
				if (ok && !inConcert.app.has_allegro_baragent()) {
					self._sip_phone.emit("appointment");
				}
			};
			self._sip_phone.invite({
				prefix : "_APPOINTMENT_",
				destination : interaction.id,
				callback : appointmentCallback
			});
		}
		assignedPanel.hideMenuActions();
	};

	PhoneClient.prototype.__activateAddress = function() {
		var self = this;
		this._initialized = true;
		inConcert.app.activateChannelAddress("CALL", function(activated) {
			if (activated) {
				if ( self._status === "RECONNECTING" ) {
					// solo se envia callback al main window, cuando se estuvo reconectado.
					// cuando se conecta x primera vez, no es necesario ya que las actualizaciones correspondientes llegan
					// en el main loop;
					self._status_callback(self._builtin ? "READY" : "DISABLED");
				}
			} else {
				self._status_callback(false, "<%=_T('Unable to activate address')%>");
			}
		});
	};

	PhoneClient.prototype.__stopSound = function(type) {
		if (this._sound_enabled === false) {
			return;
		}
		if (this._current_sound && (type === undefined || this._current_sound_type === type)) {
			this._current_sound.pause();
			this._current_sound.removeAttribute('src');
			this._current_sound.load();
			this._current_sound = null;
		}
	};

	PhoneClient.prototype.__playSound = function(type) {
		let self = this;
		if (this._sound_enabled === false) {
			return;
		}
		if (inConcert.app.has_allegro_baragent()) {
			return;
		}
		this.__stopSound();
		this._current_sound_type = type;
		this._current_sound = new Audio("/inconcert/apps/agent/" + type);
		this._current_sound.preload = "auto";
		this._current_sound.loop = true;
		this._current_sound.oncanplaythrough = function() {
			self._current_sound.play().then(function() {
				// plays until stopped*/
			}).catch(function(e) {
				console.warn("Unable to play audio file.", e);
			}); 
		};
	};

	PhoneClient.prototype.__changeStatus = function(status, event_name) {
		this.__stopSound();

		const prev_status = this._status;
		this._status = status;

		if (this.hasTelephonyWindow() && this._view.clientUpdate) {
			this._view.clientUpdate(null, event_name);
		}

		this.emit("statusChanged", status, prev_status);
	};

	PhoneClient.prototype.__onViewChanged = function() {
		if (this.hasTelephonyWindow() && this._view.clientUpdate) {
			_.defer( _.bind(this._view.clientUpdate, this._view) );
		}
		return 1; // view changed
	};

	PhoneClient.prototype.canMakeAnOutgoingCall = function() {
		const interaction = this._interaction;
		const returnErr = (err)=> {
			err.interaction = {
				id        : interaction?.id,
				state     : interaction?.state,
				initiative: interaction?.initiative
			};
			err.phoneStatus = this._status;
			console.info("PhoneClient.prototype.canMakeAnOutgoingCall - err: " + JSON.stringify(err));
			return err;
		};
		if (this.__existAnotherRenderInProgress()) {
			return returnErr({
				title  : "<%=_T('Phone is bussy with another conversation')%>",
				message: "<%=_T('CANT MAKE A NEW CALL WHEN ANOTHER CALL IS ACTIVE')%>",
				errType: "ANOTHER WINDOW PHONE RENDER IN PROGRESS"
			});
		}
		if (this._status === "STARTING" || this._status === "CONNECTING") {
			return returnErr({
				title  : "<%=_T('Phone connecting')%>",
				message: "<%=_T('Phone connecting')%>",
				errType: "PHONE IS STARTING OR CONNECTING"
			});
		}
		else if (this._status === "ERROR") {
			return returnErr({
				title  : "<%=_T('Phone disconnected')%>",
				message: this.__errorMessage,
				errType: "PHONE IS DISCONNECTED"
			});
		}
		else if (this._status !== "CONNECTED") {
			return returnErr({
				title  : "<%=_T('Phone is bussy with another conversation')%>",
				message: "<%=_T('CANT MAKE A NEW CALL WHEN ANOTHER CALL IS ACTIVE')%>",
				errType: "PHONE IS BUSY IN ANOTHER CONVERSATION"
			});
		}
		else if (this._interaction && this._interaction.isBeingComposed()) {
			return returnErr({
				title: "<%=_T('Phone is bussy with another conversation')%>",
				message: "<%=_T('CANT MAKE A NEW CALL WHEN ANOTHER CALL IS ACTIVE')%>",
				errType: "INTERACTION IS COMPOSED/PENDING"
			});
		}
	};

	PhoneClient.prototype.canStartOutgoingCall = function() {
		if (this.__existAnotherRenderInProgress()) {
			return {
				success: false,
				title: "<%=_T('Phone is bussy with another conversation')%>",
				message: "<%=_T('CANT MAKE A NEW CALL WHEN ANOTHER CALL IS ACTIVE')%>"
			};
		}
		const phoneClientStatus = this._status?.toUpperCase();
		if (phoneClientStatus === "STARTING" || phoneClientStatus === "CONNECTING") {
			return {
				success: false,
				title: "<%=_T('Phone connecting')%>",
				message: "<%=_T('Phone connecting')%>"
			};
		}
		else if (phoneClientStatus === "ERROR") {
			return {
				success: false,
				title: "<%=_T('Phone disconnected')%>",
				message: this.__errorMessage
			};
		}

		const sipPhoneState = this.getPhoneState().toLowerCase();
		if (sipPhoneState !== "connected") {
			return {
				success: false,
				title: "<%=_T('Phone is bussy with another conversation')%>",
				message: "<%=_T('CANT MAKE A NEW CALL WHEN ANOTHER CALL IS ACTIVE')%>"
			};
		}
		// ok
		return {success: true};
	};

	PhoneClient.prototype.getAccountStatus = function() {
		if (this._status === "STARTING" || this._status === "CONNECTING" || this._status === "CONNECTING") {
			return { status : "<%=_T('Phone connecting')%>" };
		}
		else if (this._status === "ERROR") {
			return { status : this.__errorMessage };
		}
		else {
			return { status : "OK" };
		}
	};

	PhoneClient.prototype.hasTelephonyWindow = function(must_shown_log) {
		let interaction_id = this._interaction ? this._interaction.id : "NO INTERACTION ID";
		if (!this._view) {
			if (must_shown_log) {
				console.info("PhoneClient.prototype.hasTelephonyWindow: this._view is not initialized. Wait moment please!! " + interaction_id);
			}
			return false;
		}
		if ($.isEmptyObject(this._view)) {
			if (must_shown_log) {
				console.info("PhoneClient.prototype.hasTelephonyWindow: this._view is empty. Wait moment please!! " + interaction_id);
			}
			return false;
		}
		if (this._view.is_closed) {
			console.warn("PhoneClient.prototype.hasTelephonyWindow: view is closed " + interaction_id);
			this.setView(null);
			return false;
		}
		if (must_shown_log) {
			console.log("PhoneClient.prototype.hasTelephonyWindow: this._view is READY. " + interaction_id);
		}
		return true;
	};

	PhoneClient.prototype.attendingInteraction = function(interaction) {
		return this._interaction && interaction && this._interaction.id === interaction.id;
	};

	/*
		Funcion invocada desde el objeto 'm_agentPollerEvents' que escucha eventos de interacciones via agent poller.
	*/
	PhoneClient.prototype.onInteractionRemoved = function(interaction, callback = $.noop) {
		interaction = interaction || {};
		const phoneInteraction = (this._interaction || {});
		const sipPhoneState = this.getPhoneState().toLowerCase();
		const info = {
			interactionId        : interaction.id,
			interactionState     : interaction.state,
			phoneStatus          : this._status,
			sipPhoneState        : sipPhoneState,
			phoneInteractionId   : phoneInteraction.id || "none",
			phoneInteractionState: phoneInteraction.state || "none"
		};
		console.info("[PhoneClient.prototype.onInteractionRemoved] ===> onInteractionRemoved info: " + JSON.stringify(info));
		/*
			Funcion que decide si el flag "this._status" del estado actual del telefono debe de cambiar a CONNECTED
		*/
		const canChangeStatusToConnected = () => {
			if (this._status === "CONNECTED")
				return false; // NO, lo debo cambiar, ya se encuentra en connected
			const interactionStates = {
				"WAITING": true
			};
			if (interactionStates[interaction.state])
				return true; // SI, lo debo cambiar en el estado actual de la interaction
			
			return false;    // NO, lo debo cambiar
		};
		if (canChangeStatusToConnected()) {
			console.debug("[PhoneClient.prototype.onInteractionRemoved] =======> onInteractionRemoved EXECUTE: this.__changeStatus('CONNECTED', 'BYED')");
			this.__changeStatus("CONNECTED", "BYED");
		}
		callback();
	};

	/*

	*/
	PhoneClient.prototype.__registerEventHandlers = function() {
		var self = this;
		var reconnectTimer = null;

		function onConnected() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onConnected interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self.__activateAddress();
			self.__changeStatus("CONNECTED");
		}

		function onWarning(message) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onWarning interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (!self.hasTelephonyWindow()) {
				return;
			}
			self._view.displayStatusMessage(message, "Alert", true, 2000); // not proud of it, no no
			if (self._view.playErrorSound) {
				self._view.playErrorSound();
			}
		}

		function onError(message) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onError interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._sip_phone.currentState === inConcert.PhoneStates.IDLE) {
				if (self._status !== "ERROR") { // notify only once ...
					self._initialized = false;
					self.__changeStatus("ERROR");
					self.__errorMessage = message;
					self._status_callback(false, format("<%=_T('There was an error with the phone client: {0}')%>", message));
				}
				if (!reconnectTimer) {
					reconnectTimer = setTimeout(function() {
						reconnectTimer = null;
						self.reconnect();
					}, 10000);
				}
			} else { // phone is still connected
				onWarning(message);
			}
		}

		//TODO: review when remove allegro
		function loadActiveCall(data) {
			if (!inConcert.app.has_allegro_baragent()) {
				return;
			}
			console.info("PhoneClient.prototype.__registerEventHandlers->loadActiveCall -> entra");
			self._active_call = {
				close: $.noop,
				setInteraction: function(interaction) {
					self.setInteraction(interaction);
					if (this === self._active_call) {
						console.info("PhoneClient.prototype.__registerEventHandlers->loadActiveCall -> INVOKE -> this.switchTo(interaction); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
						self.switchTo(interaction);
					}
				}
			};
			_.extend(self._active_call, data);
			self._manager.onNewSession(data.type === "VOICEMAIL"? "VOICEMAIL" : "CALL", self._active_call);
		}

		function changeInteractionState(newState, notManual) {
			let myInt = (self._interaction || {});
			console.info("PhoneClient.prototype.__registerEventHandlers->changeInteractionState interaction.id " + myInt.id + ", interaction.state: " + myInt.state + ", newState: " + newState);
			if (!self._interaction) {
				return;
			}
			if (newState === "PENDING") {
				self._interaction.cleanSession();
			}
			self._interaction.changeState(newState, notManual);
			self.clientUpdate(self._interaction);
		}

		function onMessage(message) {
			try {
				var jmessage = JSON.parse(message);
				if (jmessage && jmessage.result) {
					switch (jmessage.result) {
						case "CANNOT VALIDATE NUMBER":
							self._view.displayStatusMessage(
								"<%=_T('Invalid phone number')%>", "Alert", true, 3000, null,
								{
									concatenateValue: true,
									messageSeparator: "..."
								});
							break;
						case "NO OUTGOING PACKAGES":
							self._view.displayStatusMessage(
								"<%=_T('No package allows you to call this number at this time.')%>", "Alert",
								true, 3000, null,
								{
									concatenateValue: true,
									messageSeparator: "..."
								});
							break;
						case "REJECT CALL":
							self._view.displayStatusMessage(
								"<%=_T('Call Rejected')%>", "Alert", true, 3000, null,
								{
									concatenateValue: true,
									messageSeparator: "..."
								});
							break;
						case "BLACKLIST":
							self._view.displayStatusMessage(
								"<%=_T('number is in blacklist')%>", "Alert", true, 3000, null,
								{
									concatenateValue: true,
									messageSeparator: "..."
								});
							break;
						default:
							self._view.displayStatusMessage(jmessage.reason, "Alert", true, 3000, null,
								{
									concatenateValue: true,
									messageSeparator: "..."
								});
							break;
					}
				}

			} catch (SyntaxError) {}
		}

		function onInvited(data) {
			let myInt = (data || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onInvited interaction.id: " + myInt.id + ", state: " + myInt.state + ", type: " + myInt.type + ", phone._status: " + self._status);
			self._byed = false;
			self.__changeStatus("OFFERING");
			self.__playSound("sounds/phone_ringback.mp3");
			let interaction = inConcert.app.getInteractionManager().findInteraction(data.id);
			if (inConcert.app.has_allegro_baragent()) {
				if (!interaction) {
					interaction = inConcert.AllegroInteractionManager.createCallInteraction(
						data.id,
						data.campaign,
						data.account,
						data.initiative,
						data.contactId,
						data.contactName,
						data.type === "VOICEMAIL",
						"onInvited"
					);
				}
			}
			else {
				if (!interaction) {
					return;
				}
			}
			console.info("PhoneClient.prototype.__registerEventHandlers->onInvited -> INVOKE -> this.switchTo(interaction); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
			self.switchTo(interaction);
			loadActiveCall(data);
			changeInteractionState("LOCKED");
			if (interaction && interaction.isPreview()) {
				inConcert.app.getAgentApiServer().CallOnPreview(interaction);
			}
		}

		function onDialing(data) {
			let myInt = (data || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onDialing interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self._byed = false;
			if (self._status === "DIALING") {
				return;
			}
			//it happen when we create a new manual call from outside
			self.__changeStatus("DIALING", "DIALING");
			//self.__playSound("sounds/phone_ringback.mp3");
			var interaction = inConcert.app.getInteractionManager().findInteraction(data.id);
			if (!interaction) {
				interaction = inConcert.AllegroInteractionManager.createCallInteraction(
					data.id,
					data.campaign,
					data.account,
					data.initiative,
					data.contactId,
					data.contactName,
					data.type === "VOICEMAIL",
					"onDialing"
				);
			}
			console.info("PhoneClient.prototype.__registerEventHandlers->onDialing -> INVOKE -> this.switchTo(interaction); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
			self.switchTo(interaction);
			loadActiveCall(data);
			changeInteractionState("LOCKED");
		}

		const has_allegro_integration = inConcert.app.has_allegro_integration();

		function onOutbound(data) {
			let myInt = (data || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onOutbound interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self._byed = false;
			//self.__playSound("sounds/phone_ringback.mp3");
			loadActiveCall(data);
			if (self.hasTelephonyWindow() && has_allegro_integration) {
				// funcion definida en "client_window.js"
				// esta funcion para los demas canales y para el mismo canal sin integracion con allegro, se ejecuta en la funcion "onInteractionUpdated" "client_window.js"
				self._view.updateLayout("PhoneClient.__registerEventHandlers - onOutbound - has_allegro_integration = true");
			}
		}

		function onRinging() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onRinging interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self._byed = false;
			self.__playSound("sounds/phone_ringback.mp3");
		}

		function onAccepted() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onAccepted interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self._byed = false;
			if (self.hasVideo()) {
				self.__showVideoControls();
			}
			if (!inConcert.app.has_allegro_baragent() && self._interaction && self._interaction.in_redial) {
				self.__changeStatus("ACTIVE", "REDIAL ACCEPTED");
				inConcert.app.getAgentApiServer().CallRedialAccepted(self._interaction);
				// The new interaction will be TAKEN (if it"s not already assigned)
			} else {
				changeInteractionState("TAKEN");
				self.__changeStatus("ACTIVE", "ACCEPTED");
				inConcert.app.getAgentApiServer().CallTaken(self._interaction);
				if (self._interaction && self._interaction.in_appointment) {
					inConcert.app.getAgentApiServer().CallAppointmentAccepted(self._interaction);
				}
			}
		}

		function onRejected (/*reason*/) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onRejected interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self.__hideVideoControls();
			if (!inConcert.app.has_allegro_baragent() && self._interaction && (self._interaction.in_redial || self._interaction.in_appointment)) {
				if (self._interaction.in_redial) {
					inConcert.app.getAgentApiServer().CallRedialCanceled(self._interaction);
				} else {
					inConcert.app.getAgentApiServer().CallAppointmentCanceled(self._interaction);
				}
				return onWrapup(true);
			}
			changeInteractionState("PENDING");
			self.__changeStatus("CONNECTED", "REJECTED");
			if (self._interaction) {
				if (self.hasTelephonyWindow()) {
					self._view.cancelInteraction("REJECTED");
				}
				else {
					inConcert.app.getInteractionManager().cancelInteraction(self._interaction);
				}
			}
			inConcert.app.getInteractionManager().cancelPendingInteractions("CALL");
			inConcert.app.getInteractionManager().cancelPendingInteractions("VOICEMAIL");
		}

		function onCancelled () {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onCancelled interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self.__hideVideoControls();
			if (!inConcert.app.has_allegro_baragent() && self._interaction && (self._interaction.in_redial || self._interaction.in_appointment)) {
				if (self._interaction.in_redial) {
					inConcert.app.getAgentApiServer().CallRedialCanceled(self._interaction);
				} else {
					inConcert.app.getAgentApiServer().CallAppointmentCanceled(self._interaction);
				}
				return onWrapup(true);
			}
			changeInteractionState("PENDING");
			self.__changeStatus("CONNECTED", "CANCELLED");
			if (self._interaction) {
				if (self.hasTelephonyWindow()) {
					self._view.cancelInteraction("CANCELLED");
				}
				else {
					inConcert.app.getInteractionManager().cancelInteraction(self._interaction);
				}
			}
			inConcert.app.getInteractionManager().cancelPendingInteractions("CALL");
			inConcert.app.getInteractionManager().cancelPendingInteractions("VOICEMAIL");
		}

		function onWrapup (clean_in_redial, callback = $.noop) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onWrapup interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			let was_in_redial = false;
			if (self._interaction) {
				was_in_redial = self._interaction.in_redial;
				self._interaction.in_redial = clean_in_redial ? false : self._interaction.in_redial;
				self._interaction.in_appointment = false;
			}
			changeInteractionState("WRAPUP", was_in_redial);
			self.__changeStatus("WRAPUP", "WRAPUP");
			callback();
		}

		/*
		*/
		const canExecuteWarpupInNativeTel = () =>
		{
			const phoneInteraction = this._interaction;
			// 1 - NO puede ejecutar el Wrapup si la interaccion NO existe
			if (!phoneInteraction)
				return {canExecute: false, reason: "this._interaction does not exist"};
			// 2 - NO puede ejecutar el Wrapup si es Allegro
			if (inConcert.app.has_allegro_baragent())
				return {canExecute: false, reason: "is allegro interaction"};
			// 3 - NO puede ejecutar el Wrapup si el estado del TEL en PhoneClient es OFFERING o PREVIEW
			if (["OFFERING", "PREVIEW"].includes(this._status))
				return {canExecute: false, reason:  `PhoneStatus - this._status: ${ this._status }`};
			// 4 - NO puede ejecutar el Wrapup si la interaccion se encuentra en proceso de ENCOLADO
			const { in_appointment, in_redial, state, inProgressType } = phoneInteraction;
			if (["IN_PROGRESS_QUEUE", "IN_PROGRESS_TRANSFER"].includes(inProgressType))
				return {canExecute: false, reason:  `interaction ${inProgressType}`};
			// 5 - SI puede ejecutar el Wrapup si la INTERACTION se encuentra en APPOINTMENT o en REDIAL
			if (in_appointment || in_redial)
				return {canExecute: true, reason: `interaction in ${ in_appointment ? "APPOINTMENT" : "REDIAL"}`};
			// 6 - SI puede ejecutar el Wrapup si el estado de la INTERACTION es uno de los siguientes:
			if (_.contains(
				[
					"BOT",
					"CONFERENCE LOCKED",
					"CONFERENCE TAKEN",
					"CONFERENCE",
					"TAKEN",
					"TRANSFER LOCKED",
					"TRANSFER TAKEN",
					"TRANSFERRING",
					"WRAPUP"
				],
				state
			)) return {canExecute: true, reason: `interaction.state: ${ state }`};
			// 7 - NO puede ejecutar el Wrapup en cualquier otro caso
			return {canExecute: false, reason: `Unable to execute tel native wrapup`};
		};

		function onByed (params = {invokedFrom: "UNRECOGNIZED"})
		{
			params.callback = params.callback || $.noop;
			params.canExecuteWarpupInNativeTel = params.canExecuteWarpupInNativeTel || function() { return true; };
			const phoneInteraction = (self._interaction || {});
			const has_allegro_baragent = inConcert.app.has_allegro_baragent();

			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onByed interaction.id: " + phoneInteraction.id +
				", state: " + phoneInteraction.state +
				", initiative: " + phoneInteraction.initiative +
				", phone._status: " + self._status +
				", in_redial: " + phoneInteraction.in_redial +
				", in_appointment: " + phoneInteraction.in_appointment +
				", has_allegro_baragent: " + has_allegro_baragent +
				", self._byed: " + self._byed,
				", invokedFrom: " + params.invokedFrom
			);

			self.__hideVideoControls();
			const wrapupResponse = canExecuteWarpupInNativeTel();
			console.info(`[PhoneClient.prototype.__registerEventHandlers] -> onByed - canExecuteWarpupInNativeTel() - wrapupResponse: ${ JSON.stringify(wrapupResponse) }`);
			if (wrapupResponse.canExecute) {
				if (params.canExecuteWarpupInNativeTel()) {
					console.info("[PhoneClient.prototype.__registerEventHandlers] ->a onByed interaction.id: " + phoneInteraction.id + ", Execute Native onWrapup(false)");
					onWrapup(false, function() {
						//params.callback();
					});
					return;
				}
				console.info("[PhoneClient.prototype.__registerEventHandlers] ->a.1 onByed interaction.id: " + phoneInteraction.id + ", No execute Native onWrapup(false)");
			}
			if (!has_allegro_baragent && self._byed) {
				console.info("[PhoneClient.prototype.__registerEventHandlers] ->b onByed interaction.id: " + phoneInteraction.id);
				self.__changeStatus("CONNECTED", "BYED");
				//params.callback();
				return;
			}
			self._byed = true;
			changeInteractionState("FINISHED");
			self.__changeStatus("CONNECTED", "BYED");
			if (self._interaction) {
				let isCancelled = false;
				if (self.hasTelephonyWindow()) {
					console.info("[PhoneClient.prototype.__registerEventHandlers] ->c onByed interaction.id: " + phoneInteraction.id);
					isCancelled = self._view.cancelInteraction("BYED");
				}
				else {
					console.info("[PhoneClient.prototype.__registerEventHandlers] ->d onByed interaction.id: " + phoneInteraction.id);
					isCancelled = inConcert.app.getInteractionManager().cancelInteraction(phoneInteraction);
				}
				if (isCancelled) {
					/*
						(ALLEGRO)
							Si la interaccion fue cancelada y el telefono se estaba abriendo y se ha cerrado, debe subscribir un evento en "pubsub"
								para que cuando se termine de cerrar la ventana, invoque al callback de SipPhone y proceda a ejecutar las acciones pertinentes (si es que amerita)

							Ref Issue #18744
					*/
					console.info("[PhoneClient.prototype.__registerEventHandlers] >d.1 onByed> subscribe event onWindowClosed - isCancelled interaction: " + phoneInteraction.id);
					const pubsub_close_phone_window = inConcert.app.messagebus.subscribe(
						`/live_phone_client_window/interaction/${phoneInteraction.id}/windows_closed/`,
						function(closedInteractionWindow) {
							console.info("[PhoneClient.prototype.__registerEventHandlers] >d.2 onByed> onWindowClosed interaction: " + closedInteractionWindow.id);
							params.callback();
							if (pubsub_close_phone_window) {
								inConcert.app.messagebus.unsubscribe(pubsub_close_phone_window);
							}
						}
					);
				}
			}
			else {
				console.info("[PhoneClient.prototype.__registerEventHandlers] ->d1 onByed interaction.id: " + phoneInteraction.id);
				//params.callback();
			}
			console.info("[PhoneClient.prototype.__registerEventHandlers] ->e onByed interaction.id: " + phoneInteraction.id);
			inConcert.app.getInteractionManager().cancelPendingInteractions("CALL");
			inConcert.app.getInteractionManager().cancelPendingInteractions("VOICEMAIL");
			self._interaction = null;
		}

		this.__byedEvent = function() {
			onByed({
				invokedFrom: "PhoneClient.__byedEvent",
				canExecuteWarpupInNativeTel: () => {
					//callback que determina si el WRAPUP en TEL Nativa se debe ejecutar al finalizar la interaccion. Esto para que NO quede en loop.
					if (inConcert.app.has_allegro_baragent()) return false;
					// NO debe ejecutar: si ya se encuentra en WRAPUP
					if (this._interaction?.state === "WRAPUP") return false;
					// SI, debe de ejecutar
					return true;
				}
			});
		};

		function onRecording(recording) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onRecording interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._interaction && self.hasTelephonyWindow()) {
				self._view.clientUpdate(null, "RECORDING");
			}
			inConcert.app.getAgentApiServer().CallOnRecording(self._interaction, recording);
		}

		function onConference() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onConference interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._interaction && self.hasTelephonyWindow()) {
				self._view.clientUpdate(null, "CONFERENCE");
			}
			inConcert.app.getAgentApiServer().CallOnConference(self._interaction);
		}

		function onHold(hold) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onHold interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._interaction && self.hasTelephonyWindow()) {
				self._view.clientUpdate(null, "HOLD");
			}
			inConcert.app.getAgentApiServer().CallOnHold(self._interaction, hold);
		}

		function onMute(mute) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onMute interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._interaction && self.hasTelephonyWindow()) {
				self._view.clientUpdate(null, "MUTE");
			}
			inConcert.app.getAgentApiServer().CallOnMute(self._interaction, mute);
		}

		function onVideoDisabled(disabled) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onVideoDisabled interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			self._setVideocallActions();
			inConcert.app.getAgentApiServer().CallOnVideoDisabled(self._interaction, disabled);
		}

		function onSwitch(_switch) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onSwitch interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._interaction && self.hasTelephonyWindow()) {
				self._view.clientUpdate(null, "_SWITCH");
			}
			inConcert.app.getAgentApiServer().CallOnSwitch(self._interaction, _switch);
		}

		function onPreview() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onPreview interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._interaction && self.hasTelephonyWindow()) {
				self._view.clientUpdate(null, "PREVIEW");
			}
			if (self._interaction) {
				inConcert.app.getAgentApiServer().CallOnPreview(self._interaction);
			}
		}

		function onRedial() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onRedial interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._builtin && self._interaction && !self._interaction.in_redial) {
				self._interaction.in_redial = true;
				self._interaction.state = "LOCKED";
				self.__changeStatus("DIALING", "REDIAL");
				var assignedPanel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
				assignedPanel.hideMenuActions();
			}
			inConcert.app.getAgentApiServer().CallRedialing(self._interaction);
		}

		function onAppointment() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onAppointment interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			if (self._builtin && self._interaction && !self._interaction.in_appointment) {
				self._interaction.in_appointment = true;
				self._interaction.state = "LOCKED";
				self.__changeStatus("DIALING", "APPOINTMENT");
				var assignedPanel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
				assignedPanel.hideMenuActions();
			}
		}

		function onTransferring() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onTransferring interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			changeInteractionState("TRANSFER TRYING");
			self.__changeStatus("TRANSFER TRYING", "TRANSFERRING");
			inConcert.app.getAgentApiServer().CallTransferring(self._interaction);
		}

		function onTransferringConnected() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onTransferringConnected interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			changeInteractionState("TRANSFER TAKEN");
			self.__changeStatus("TRANSFER TAKEN", "TRANSFERRING CONNECTED");
			inConcert.app.getAgentApiServer().CallTransferConnected(self._interaction);
		}

		function onConferenceConsultationTrying() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onConferenceConsultationTrying interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			changeInteractionState("CONFERENCE TRYING");
			self.__changeStatus("CONFERENCE TRYING", "CONFERENCE TRYING");
			inConcert.app.getAgentApiServer().CallConferenceConsultationTrying(self._interaction);
		}

		function onConferenceConsultationConnected() {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onConferenceConsultationConnected interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			changeInteractionState("CONFERENCE TAKEN");
			self.__changeStatus("CONFERENCE TAKEN", "CONFERENCE CONNECTED");
			inConcert.app.getAgentApiServer().CallConferenceConsultationConnected(self._interaction);
		}

		function onRedialAccepted (data) {
			let myInt = (data || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onRedialAccepted interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status);
			let interaction = inConcert.app.getInteractionManager().findInteraction(data.id);
			if (!interaction) {
				interaction = inConcert.AllegroInteractionManager.createCallInteraction(
					data.id,
					data.campaign,
					data.account,
					data.initiative,
					data.contactId,
					data.contactName,
					data.type === "VOICEMAIL",
					"onRedialAccepted"
				);
			}
			// El cancel de la interaccion se realiza en la propia funcion "switchTo". Si es que se debe de realizar.
			console.info("PhoneClient.prototype.__registerEventHandlers->onRedialAccepted -> INVOKE -> this.switchTo(interaction); for interaction: " + (interaction || {}).id + ", state: " + (interaction || {}).state);
			self.switchTo(interaction);
			loadActiveCall(data);
			changeInteractionState("TAKEN");
			self.__changeStatus("ACTIVE", "REDIAL ACCEPTED");
			inConcert.app.getAgentApiServer().CallRedialAccepted(interaction);
		}

		function onCheckAllegroPopup(popup) {
			let myInt = (self._interaction || {});
			console.info("[PhoneClient.prototype.__registerEventHandlers] -> onCheckAllegroPopup interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + self._status + ", popup: " + popup);
			if (!self.hasTelephonyWindow()) {
				return;
			}
			self._view.checkAllegroPopup(popup);
		}

		function onStopScreenShare() {
			self.StopScreenShare();
		}

		this._sip_phone
			.on("connected", onConnected)
			.on("error", onError)
			.on("recording", onRecording)
			.on("in_conference", onConference)
			.on("in_hold", onHold)
			.on("muted", onMute)
			.on("video_disabled", onVideoDisabled)
			.on("switched", onSwitch)
			.on("preview", onPreview)
			.on("redial", onRedial)
			.on("appointment", onAppointment)
			.on("message", onMessage)
			.on("invited", onInvited)
			.on("outbound", onOutbound)
			.on("dialing", onDialing)
			.on("ringing", onRinging)
			.on("accepted", onAccepted)
			.on("rejected", onRejected)
			.on("cancelled", onCancelled)
			.on("byed", (_ /*callSession*/, params)=> {
				params = params || {invokedFrom: "SipPhone"};
				const myInt = (this._interaction || {});
				console.debug("[PhoneClient.prototype.__registerEventHandlers] -> this._sip_phone('byed') interaction.id: " + myInt.id + ", state: " + myInt.state + ", phone._status: " + this._status);
				onByed(params);
			})
			.on("wrapup", onWrapup)
			.on("transferring", onTransferring)
			.on("transferring_connected", onTransferringConnected)
			.on("conference_consultation_trying", onConferenceConsultationTrying)
			.on("conference_consultation_connected", onConferenceConsultationConnected)
			.on("redial_accepted", onRedialAccepted)
			.on("warning", onWarning)
			.on("check_allegro_popup", onCheckAllegroPopup)
			.on("stopscreenshare", onStopScreenShare);
	};

	PhoneClient.prototype.interactionFinished = function(interaction, pending) {
		const myInt = (interaction || {});
		const myInt1 = (this._interaction || {});
		console.info("[PhoneClient.prototype.interactionFinished] -> onFinished this._interaction.id: " + myInt.id +
			", state: " + myInt.state +
			", phone._status: " + this._status +
			", interaction.id: " + myInt1.id +
			", state: " + myInt1.state +
			", pending: " + pending
		);
		if (inConcert.app.has_allegro_baragent() ||
			!this.__byedEvent ||
			(pending && this._status === "DIALING") ||
			(this._interaction && this._interaction.id !== interaction.id)
		) {
			console.debug("[PhoneClient.prototype.interactionFinished] -> onFinished -> 1 - EXEC: NOTHING");
			return;
		}
		if (!this._interaction) {
			if (this._status === "WRAPUP" || this._status === "ACTIVE") {
				console.debug("[PhoneClient.prototype.interactionFinished] -> onFinished -> 2 -> EXEC: this.__changeStatus('CONNECTED', 'BYED')");
				this.__changeStatus("CONNECTED", "BYED");
			}
			console.debug("[PhoneClient.prototype.interactionFinished] -> onFinished -> 3 - EXEC: NOTHING");
			return;
		}
		if (interaction.state !== "FINISHED") {
			// solo se puede invocar el evento forzado onByed del phoneClient, si la interaccion NO se ha finalizado ya que sino el telefono queda inconsistente.
			console.debug("[PhoneClient.prototype.interactionFinished] -> onFinished -> 4 - EXEC: this.__byedEvent()");
			this.__byedEvent();
			return;
		}
		if (!interaction.in_redial && (this._status === "WRAPUP" || this._status === "ACTIVE")) {
			console.debug("[PhoneClient.prototype.interactionFinished] -> onFinished -> 5 - EXEC: this.__changeStatus('CONNECTED', 'BYED')");
			this.__changeStatus("CONNECTED", "BYED");
			return;
		}
		console.debug("[PhoneClient.prototype.interactionFinished] -> onFinished -> 6 - EXEC: NOTHING");
	};

	PhoneClient.prototype.interactionWrapup = function() {
		this.__changeStatus("WRAPUP", "WRAPUP");
	};

	PhoneClient.prototype.interactionTaken = function() {
		this.__changeStatus("ACTIVE", "ACCEPTED");
	};	
	
	PhoneClient.prototype.getPhoneState = function() {
		if (this._sip_phone) {
			return this._sip_phone.getPhoneState();
		}
		else {
			return "disconnected";
		}
	};
	
	PhoneClient.prototype.hasVideo = function() {
		return this._interaction? this._interaction.hasVideo() : false;
	};
	
	PhoneClient.prototype.hasScreenRecording = function() {
		return this._interaction? this._interaction.hasScreenRecording() : false;
	};
	
	PhoneClient.prototype.$videocall = function(query) {
		const $videocall = $("#telephony-video-container");
		return query? $videocall.find(query) : $videocall;
	};
	
	PhoneClient.prototype.$localVideo = function() {
		return this.$videocall("#telephony-local-video");
	};
	
	PhoneClient.prototype.$remoteVideo = function() {
		return this.$videocall("#telephony-remote-video");
	};
	
	PhoneClient.prototype.$remoteAudio = function() {
		return $("#telephony-remote-audio");
	};
	
	PhoneClient.prototype._bindVideocallAction = function(className, action) {
		this.$videocall(`.${className}`)
			.off("click")
			.on("click", () => this[action]());
	};

	PhoneClient.prototype._setEventDisableMinimizeClient = function() {
		const self = this;
		this.$videocall()
			.off("click")
			.on("click", function (){
				if (self._view.m_enabledMinimizeWhenOutside) {
					self._view.m_enabledMinimizeWhenOutside = false;
				}
			});
	};

	PhoneClient.prototype._bindVideocallActions = function() {
		this._bindVideocallAction("maximizeButton", "maximizeVideo");
		this._bindVideocallAction("minimizeButton", "minimizeVideo");
		this._bindVideocallAction("presentButton", "fullScreenVideo");
		
		this._bindVideocallAction("turnOffCamera", "disableVideo");
		this._bindVideocallAction("turnOnCamera", "enableVideo");
		
		this._bindVideocallAction("shareButton", "screenShare");
		this._bindVideocallAction("stop-shareButton", "StopScreenShare");
		this._bindVideocallAction("goInteractionButton", "openInteractionClient");
	};
	
	PhoneClient.prototype._setVideocallAction = function(className, enabled) {
		this.$videocall(`.${className}`).toggle(!!enabled);
	};

	PhoneClient.prototype.openInteractionClient = function(interaction, callback) {
		console.info("PhoneClient.prototype.openInteractionClient -> INVOKE -> this.switchTo(interaction, callback); for this._interaction: " + (this._interaction || {}).id + ", state: " + (this._interaction || {}).state);
		this.switchTo(this._interaction, callback);
		this.minimizeVideo();
	};	
	
	PhoneClient.prototype._setVideocallActions = function() {
		const videoDisabled = this._interaction && 
			this._interaction.videoDisabled();
		const remoteVideoDisabled = this._interaction && 
			this._interaction.remoteVideoDisabled();
		const hasScreenSharePermission = this._interaction && 
			this._interaction.hasPermission("SCREEN_SHARE", "interaction");
		const isScreenSharing = this._interaction && 
			this._interaction.isScreenSharing();
		
		this._setVideocallAction("maximizeButton",
			this._videoViewState !== VideoViewStates.MAXIMIZED);
		this._setVideocallAction("minimizeButton",
			this._videoViewState !== VideoViewStates.MINIMIZED);
		this._setVideocallAction("presentButton",
			this._videoViewState !== VideoViewStates.FULL_SCREN);
			
		this._setVideocallAction("turnOffCamera", !videoDisabled && !isScreenSharing);
		this._setVideocallAction("turnOnCamera", videoDisabled && !isScreenSharing);
			
		this._setVideocallAction("shareButton", hasScreenSharePermission && !videoDisabled && !isScreenSharing);
		this._setVideocallAction("stop-shareButton", hasScreenSharePermission && !videoDisabled && isScreenSharing);
		
		this._setVideocallAction("goInteractionButton", this._interaction);
		
		this._setVideocallVideo(!videoDisabled, "agent-video", 
			"agent-avatar");
		this._setVideocallVideo(!remoteVideoDisabled, "customer-video", 
			"customer-avatar");
	};
	
	PhoneClient.prototype._setVideocallVideo = function(enabled, videoClassName, avatarClassName) {
		this.$videocall(`.${videoClassName}`).toggle(enabled);
		this.$videocall(`.${avatarClassName}`).toggle(!enabled);
	};

	PhoneClient.prototype.__showVideoControls = function() {
		this.$videocall().show();
		this._setVideocallActions();
	};
	
	PhoneClient.prototype.__hideVideoControls = function() {
		this.$videocall().hide();
	};
	
	PhoneClient.prototype._setVideocallViewState = function(newState) {
		this._videoViewState = newState;
		const $container = this.$videocall();
		Object.keys(VideoViewStates).forEach(stateKey => {
			const state = VideoViewStates[stateKey];
			$container.toggleClass(state, state === newState);
		});
		this._setVideocallActions();
	};
	
	PhoneClient.prototype.maximizeVideo = function() {
		this._setVideocallViewState(VideoViewStates.MAXIMIZED);
	};
	
	PhoneClient.prototype.minimizeVideo = function() {
		this._setVideocallViewState(VideoViewStates.MINIMIZED);
	};
	
	PhoneClient.prototype.fullScreenVideo = function() {
		this._setVideocallViewState(VideoViewStates.FULL_SCREN);
	};

	PhoneClient.prototype.screenShare = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.screenShare(function(ok, err) {
					if (callback) {
						callback(ok, err);
					}
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("screenShare", true);
						self._setVideocallActions();
					}
				});
			}
			else
			{
				this._sip_phone.screenShare(callback);
			}
		}
	};

	PhoneClient.prototype.StopScreenShare = function(interaction, callback) {
		let self = this;
		if (this._builtin) {
			if (!inConcert.app.has_allegro_baragent()) {
				this._interaction.StopScreenShare(function(ok, err) {
					if (callback) {
						callback(ok, err);
					}
					if (ok && ok.result === "OK") {
						self._sip_phone.emit("StopScreenShare", true);
						self._setVideocallActions();
					}
				});
			}
		}
	};

	PhoneClient.prototype.StartStatisticsCapture = function (statsCallback, callback = $.noop) {
		const self = this;
		try {
			clearInterval(this.appPollingInterval); // Just in case
			this.appPollingInterval = window.setInterval(function () {
				if (self._sip_phone && self._sip_phone.getRTCStatistics) {
					self._sip_phone.getRTCStatistics(function (success, stats) {
						if (success) {
							statsCallback(self._interaction && self._interaction.id ? self._interaction.id:"", stats);
						}
					});
				}
			}, 1000);

			callback(true);
		}
		catch (e) {
			callback(false, e);
		}
	};

	PhoneClient.prototype.StopStatisticsCapture = function(callback = $.noop)
	{
		try
		{
			if (this.appPollingInterval !== null) {
				clearInterval(this.appPollingInterval);
			} else {
				callback(false, "Statistics capture was not active.");
				return;
			}

			callback(true);
		}
		catch (e)
		{
			callback(false, e);
		}
	};

	PhoneClient.prototype.onWindowIsReady = function(interaction = {})
	{
		console.log("[PhoneClient.prototype.onWindowIsReady] interaction: " + interaction.id + ", state: " + interaction.state + ", initiative: " + interaction.initiative + ", phoneStatus: " + this._status);
		this.__setRender("10-WINDOW_READY", interaction, "IS OK - FROM onWindowIsReady FUNC");
	};

	PhoneClient.prototype.onWindowClosed = function(interaction = {}, invokedFrom = "WINDOW_DIALOG")
	{
		console.log("[PhoneClient.prototype.onWindowClosed] invokedFrom: " + invokedFrom + ", interaction: " + interaction.id + ", state: " + interaction.state + ", initiative: " + interaction.initiative + ", phoneStatus: " + this._status);
		this.__setRender("9-WINDOW_CLOSED", interaction, "IS OK - FROM onWindowClosed FUNC");
		if (!$.isEmptyObject(interaction)) {
			if (interaction.state === "ABANDONED" || interaction.state === "CANCELED") { // TODO: initiative === "OUTBOUND"
				console.info("[PhoneClient.prototype.onWindowClosed] 1-> Publish event - /live_phone_client_window/interaction/" + interaction.id + ", state: " + interaction.state);
				inConcert.app.messagebus.publish_async("/live_phone_client_window/interaction/" + interaction.id + "/windows_closed/", interaction);
			}
			else {
				console.info("[PhoneClient.prototype.onWindowClosed] 2-> NO Publish event - /live_phone_client_window/interaction/" + interaction.id + ", state: " + interaction.state);
			}
		}
		else {
			console.log("[PhoneClient.prototype.onWindowClosed] 3-> Publish event Interaction does not exist.");
			inConcert.app.messagebus.publish_async("/live_phone_client_window/interaction/not_interaction_found/windows_closed/");
		}
	};

	// --------------------------------------
	// RENDERING
	//---------------------------------------


	const MAX_RENDERING_WINDOW_TIME_MS = 30000; // mas de 30 segundos NO puede en demorar abrir la ventana del telefono

	/*
		Tipos de RENDER (atributo status)
			status = "4-WINDOW_LOADING_INIT_ERR_BUT_EXEC_ACTION_CONTINUE_WITH_CURRENT_RENDER"
			stauts = "6-WINDOW_LOADING_INIT_ERR_TIMEOUT_EXEC_CANCEL_CURRENT_RENDER"
			stauts = "7-WINDOW_LOADING_INIT_ERR_INVALID_RENDER_INFO_EXEC_CANCEL_CURRENT_RENDER"
			status = "5-WINDOW_LOADING_INIT_ERR_EXEC_CANCEL_CURRENT_RENDER"
			status = "3-WINDOW_INIT_FRM_IN_PROGRESS_OK"
			status = "2-WINDOW_LOADING_IN_PROGRESS_OK"
			status = "0-WINDOW_NEVER_OPENED"
			status = "8-WINDOW_IS_READY_WAITING_LOADING_INTERACTION_INFO"
			status = "1-WINDOW_LOADING_START_OK"
			status = "10-WINDOW_READY"
			status = "11-WINDOW_LOADING_INIT_ERR_INTERACTION_IS_GONE_EXEC_CANCEL_CURRENT_RENDER"
			status = "9-WINDOW_CLOSED"
	*/
	PhoneClient.prototype.__setRender = function( status, interaction, logFrom = "NONE IS REQUIRED !!" )
	{
		console.log("");
		console.log("------------------------------------------------------------------------------------------------");
		console.log("[PhoneClient.__setRender] 1) -----------> SET -> OLD Render FROM: '" + logFrom + " - RENDER_INFO: " + JSON.stringify(this.m_renderInfo) + " - interaction: " + interaction + "\n");
		// if (!interaction) {
		// 	console.log("[PhoneClient.__setRender] 1.1) =================================================> NOOOOOOOOOOOOOOT interaction FOOUUUUUNDDDDDDD." + interaction);
		// }
		interaction = interaction || {};
		this.m_renderInfo = {
			interactionId        : interaction.id || null,
			interactionInitiative: interaction.initiative || null,
			interactionState     : interaction.state || null,
			logFrom              : logFrom || null,
			status               : status,
			phoneStatus          : this._status,
			timestamp            : new Date().getTime()
		};
		console.log("[PhoneClient.__setRender] 2) -----------> SET -> NEW Render FROM: '" + logFrom + " - RENDER_INFO: " + JSON.stringify(this.m_renderInfo));
		console.log("------------------------------------------------------------------------------------------------");
		console.log("");
	};

	/*

	*/
	PhoneClient.prototype.__getRender = function(/*logFrom = "NONE"*/)
	{
		//console.log("[PhoneClient.__getRender] -----------> GET -> CURRENT Render FROM: '" + logFrom + " - RENDER_INFO: " + JSON.stringify(this.m_renderInfo));
		return this.m_renderInfo;
	};

	/*
		runningStatuses:
			1) los estados en de tipo "SI running" (true ) NO debe de comenzar con el siguiente render ya que el actual aun no ha terminado.
			2) los estados en de tipo "NO running" (false) DEBE cortar el running actuali y comenzar con el siguiente.
	*/
	PhoneClient.prototype.__renderRunning = function()
	{
		const renderInfo = this.__getRender("__renderRunning");
		const runningStatuses = {
		// SI running
			["4-WINDOW_LOADING_INIT_ERR_BUT_EXEC_ACTION_CONTINUE_WITH_CURRENT_RENDER"]   : { isRunning: true  },
			["3-WINDOW_INIT_FRM_IN_PROGRESS_OK"]                                         : { isRunning: true  },
			["2-WINDOW_LOADING_IN_PROGRESS_OK"]                                          : { isRunning: true  },
			["1-WINDOW_LOADING_START_OK"]                                                : { isRunning: true  },
		// NO running
			["5-WINDOW_LOADING_INIT_ERR_EXEC_CANCEL_CURRENT_RENDER"]                     : { isRunning: false },
			["6-WINDOW_LOADING_INIT_ERR_TIMEOUT_EXEC_CANCEL_CURRENT_RENDER"]             : { isRunning: false },
			["7-WINDOW_LOADING_INIT_ERR_INVALID_RENDER_INFO_EXEC_CANCEL_CURRENT_RENDER"] : { isRunning: false },
			["0-WINDOW_NEVER_OPENED"]                                                    : { isRunning: false },
			["8-WINDOW_IS_READY_WAITING_LOADING_INTERACTION_INFO"]                       : { isRunning: false },
			["9-WINDOW_CLOSED"]                                                          : { isRunning: false },
			["11-WINDOW_LOADING_INIT_ERR_INTERACTION_IS_GONE_EXEC_CANCEL_CURRENT_RENDER"]: { isRunning: false },

		//TODO: validar este si hay que hacer algo ? asi como esta es como si tuviera el valor en--> false (es decir cuenta como un estado NO running)
			//["10-WINDOW_READY"]                  : { isRunning: true },
		};
		const info = runningStatuses[renderInfo.status] || { isRunning: false };
		const response = {
			isRunning : info.isRunning,
			renderInfo: renderInfo
		};
		console.log("[PhoneClient.__renderRunning] -----------> Is Render Running: " + response.isRunning +
			" status: " + response.renderInfo.status +
			" interaction.id: " + response.renderInfo.interactionId +
			" interaction.state: " + response.renderInfo.interactionState +
			" interaction.initiative: " + response.renderInfo.interactionInitiative +
			" phone.Status: " + response.renderInfo.phoneStatus
		);
		return response;
	};

	/*

	*/
	PhoneClient.prototype.__existAnotherRenderInProgress = function(currentInteractionLog = "")
	{
		const { isRunning, renderInfo } = this.__renderRunning();
		if (!isRunning) {
			return false;
		}
		if (renderInfo.timestamp) {
			let now = new Date();
			let elapsed_time_ms = now.getTime() - renderInfo.timestamp;
			if (elapsed_time_ms <= MAX_RENDERING_WINDOW_TIME_MS) {
				return true;
			}
			else {
				console.log("PhoneClient.prototype.__existAnotherRenderInProgress - Cancel current render. " + JSON.stringify(renderInfo || {}) + ". " + currentInteractionLog + ", Elapsed Time Sec: " + elapsed_time_ms / 60);
				this.__setRender("6-WINDOW_LOADING_INIT_ERR_TIMEOUT_EXEC_CANCEL_CURRENT_RENDER", {
						id        : renderInfo.interactionId,
						initiative: renderInfo.interactionInitiative,
						state     : renderInfo.interactionState
					},
					"FROM __existAnotherRenderInProgress FUNC - err_type: TIMEOUT. max_rendering_window_time_seconds: " + (MAX_RENDERING_WINDOW_TIME_MS / 1000).toString()
				);
				return false;
			}
		}
		// aca no deberia llegar nunca, pero dejo un log por las dudas.
		console.log("PhoneClient.prototype.__existAnotherRenderInProgress - Invalid current render info. " + JSON.stringify(renderInfo || {}) + ". " + currentInteractionLog + ", Missing timestamp attribute.");
		this.__setRender("7-WINDOW_LOADING_INIT_ERR_INVALID_RENDER_INFO_EXEC_CANCEL_CURRENT_RENDER", {
				id        : renderInfo.interactionId,
				initiative: renderInfo.interactionInitiative,
				state     : renderInfo.interactionState
			},
			"FROM __existAnotherRenderInProgress FUNC - err_type: Missing timestamp attribute."
		);
		return false;
	};
	//
	// END RENDERING -----------------------------------------------------------------------------
}());
