local templates = require "apps.allegro.templates"

module(..., package.seeall, inconcert.orbit.application.Decorate)

local proxy = require "apps.allegro.proxy"

function LoadPaginatedProcesses (encodedInfo)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:LoadPaginatedProcesses] Loading Processes.")

	local success, data = proxy.outboundEngine.LoadProcesses(encodedInfo)
	if success and data then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:LoadPaginatedProcesses] Successfully loaded %d of %s Processes.", #data.rows, data.total_rows)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_LOADING_PROCESSES"
	}
	return false, errType
end

function GetProcess (processId, vccId)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE:GetProcess] Retrieving Process: '%s', vccId: '%s'.", processId, vccId)

	local response, err = proxy.outboundEngine.GetProcess(processId, vccId)

	if response then
		local process = response.process
		LogInfo("[OUTBOUNDENGINE_PROCESSES:GetProcess] Successfully retrieving %s Process. Info: %s", processId, process)
		return true, process
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_PROCESS"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:GetProcess] Error retrieving %s Process. Info: %s", processId, errType)
	return false, errType
end

function DeleteProcess (encodedInfo, allegroOtherInfo)
	allegroOtherInfo = allegroOtherInfo or {}

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:DeleteProcess] Deleting Process. info: %s, allegroOtherInfo: %s", encodedInfo, allegroOtherInfo)

	--[[
		Informacion adicional que tiene que viajar en el POST.
		Ej: La IP del Server Remoto que se utiliza en el history.
	]]--
	local omnichannelProcess = json.decode(encodedInfo)
	if allegroOtherInfo then
		omnichannelProcess.otherInfo = allegroOtherInfo
		encodedInfo = json.encode(omnichannelProcess)
	end

	local success, data = proxy.outboundEngine.DeleteProcess(encodedInfo)
	--LogDebug("---> DeleteProcess success: %s, data: %s", success, data)
	if success then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:DeleteProcess] Successfully deleted %s Process.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_DELETING_PROCESS"
	}
	return false, errType
end

function StartProcess (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:StartProcess] Starting Process.", encodedInfo)

	local success, data = proxy.outboundEngine.StartProcess(encodedInfo)
	--LogDebug("---> StartProcess success: %s, data: %s", success, data)
	if success then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:StartProcess] Successfully started %s Process.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_STARTING_PROCESS"
	}
	return false, errType
end

function StopProcess (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:StopProcess] Stopping Process.", encodedInfo)

	local success, data = proxy.outboundEngine.StopProcess(encodedInfo)
	--LogDebug("---> StopProcess success: %s, data: %s", success, data)
	if success then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:StopProcess] Successfully stopped %s Process.", encodedInfo)
		return true
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_STOPPING_PROCESS"
	}
	return false, errType
end

function SaveProcess (encodedInfo, allegroOtherInfo)
	allegroOtherInfo = allegroOtherInfo or {}
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:SaveProcess] Saving Process.", encodedInfo)

	--[[
		Informacion adicional que tiene que viajar en el POST.
		Ej: La IP del Server Remoto que se utiliza en el history.
	]]--
	local omnichannelProcess = json.decode(encodedInfo)
	if allegroOtherInfo then
		omnichannelProcess.otherInfo = allegroOtherInfo
		encodedInfo = json.encode(omnichannelProcess)
	end

	local success, data = proxy.outboundEngine.SaveProcess(encodedInfo)
	--LogDebug("---> SaveProcess success: %s, data: %s", success, data)
	if success then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:SaveProcess] Successfully saved %s Process.", encodedInfo)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_SAVING_PROCESS"
	}
	return false, errType
end

function LoadAvailableSchedulers ()

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:LoadAvailableSchedulers] Loading Available Schedulers for Process.")

	local response, err = proxy.outboundEngine.LoadAvailableSchedulers()

	if response then
		local availableSchedulers = response.availableSchedulers
		LogInfo("[OUTBOUNDENGINE_PROCESSES:LoadAvailableSchedulers] Successfully Loaded Available Schedulers for Process. Info: %s", availableSchedulers)
		return true, availableSchedulers
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_PROCESS_AVAILABLE_SCHEDULERS"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:LoadAvailableSchedulers] Error loading Available Schedulers for Process. Info: %s", json.encode(errType))
	return false, errType
end

function GetScheduler (schedulerId, vccId)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:GetScheduler] Loading Scheduler: %s, vcc: %s", schedulerId, vccId)
	local response, err = proxy.outboundEngine.GetScheduler(schedulerId, vccId)
	if response then
		local allegro_schedule = response.schedule
		LogInfo("[OUTBOUNDENGINE_PROCESSES:GetScheduler] Successfully Loaded Scheduler: %s. Allegro schedule: %s", schedulerId, json.encode(allegro_schedule))
		local omnichannel_schedule = templates.MakeOmnichannelScheduler( allegro_schedule )
		LogDebug("[OUTBOUNDENGINE_PROCESSES:GetScheduler] Omnichannel schedule: %s, Info: %s", schedulerId, json.encode(omnichannel_schedule))
		return true, omnichannel_schedule, allegro_schedule
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_SCHEDULER"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:GetScheduler] Error loading Scheduler: %s. Info: %s", schedulerId, json.encode(errType))
	return false, errType
end

function SaveScheduler (encodedInfo, allegroOtherInfo)
	allegroOtherInfo = allegroOtherInfo or {}
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:SaveScheduler] Saving Scheduler.", encodedInfo)

	--[[
		Informacion adicional que tiene que viajar en el POST.
		Ej: La IP del Server Remoto que se utiliza en el history.
	]]--
	local omnichannelScheduler = json.decode(encodedInfo)
	if allegroOtherInfo then
		omnichannelScheduler.otherInfo = allegroOtherInfo
		encodedInfo = json.encode(omnichannelScheduler)
	end

	local success, data = proxy.outboundEngine.SaveScheduler(encodedInfo)
	--LogDebug("---> SaveScheduler success: %s, data: %s", success, data)
	if success then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:SaveScheduler] Successfully saved %s Scheduler.", encodedInfo)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_SAVING_SCHEDULER"
	}
	return false, errType
end

function LoadBatchFilters (vccId)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:LoadBatchFilters] Loading Batch Filters for Process.")
	local response, err = proxy.outboundEngine.LoadBatchFilters(vccId)
	if response then
		local filters = response.filters
		LogInfo("[OUTBOUNDENGINE_PROCESSES:LoadBatchFilters] Successfully Loaded Batch Filters for Process. Info: %s", filters)
		return true, filters
	end
	-- Error
	local errType = {
		type  = "ALLEGRO_INTEGRATION_ERROR",
		err   = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_PROCESS_BATCH_FILTERS",
		vccId = vccId
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:LoadBatchFilters] Error loading Batch Filters for Process. Info: %s", json.encode(errType))
	return false, errType
end

function LoadBatchContactFields (processId)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:LoadBatchContactFields] Loading Batch Contact Fields for Process: %s", processId)

	local response, err = proxy.outboundEngine.LoadBatchContactFields(processId)

	if response then
		local fields = response.fields
		LogInfo("[OUTBOUNDENGINE_PROCESSES:LoadBatchContactFields] Successfully Loaded Batch Filters for Process: %s. Info: %s", processId, fields)
		return true, fields
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_PROCESS_BATCH_FIELDS"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:LoadBatchContactFields] Error loading Batch Filters for Process: %s. Info: %s", processId, json.encode(errType))
	return false, errType
end

function SaveBatchFilter (encodedInfo)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:SaveBatchFilter] Saving Batch Filter: %s", encodedInfo)

	local response, data = proxy.outboundEngine.SaveBatchFilter(encodedInfo)
	--LogDebug("---> SaveBatchFilter success: %s, err: %s", response, err)
	if response then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:SaveBatchFilter] Successfully Saved Batch Filter: %s. Response: %s, %s", encodedInfo, response, data)
		return true, data
	end
	local err = data
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_SAVING_BATCH_FILTER"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:SaveBatchFilter] Error Saving Batch Filter: %s. Info: %s", encodedInfo, json.encode(errType))
	return false, errType
end

function DeleteBatchFilter (encodedInfo)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:DeleteBatchFilter] Deleting Batch Filter: %s", encodedInfo)

	local response, err = proxy.outboundEngine.DeleteBatchFilter(encodedInfo)
	--LogDebug("---> DeleteBatchFilter success: %s, err: %s", response, err)
	if response then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:DeleteBatchFilter] Successfully Deleted Batch Filter: %s. Response: %s, %s", encodedInfo, response, err)
		return true, response
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_DELETING_BATCH_FILTER"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:DeleteBatchFilter] Error Deleting Batch Filter: %s. Info: %s", encodedInfo, json.encode(errType))
	return false, errType
end

function CreateBatch (queryId, encodedInfo)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:CreateBatch] Creating contacts batch: queryId: %s", queryId)

	local response, data = proxy.outboundEngine.CreateBatch(queryId, encodedInfo)
	--LogDebug("---> CreateBatch success: %s, err: %s", response, err)
	if response then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:CreateBatch] Successfully Created contacts batch: queryId: %s. Response: %s, %s", queryId, response, data)
		return true, data
	end
	local err = data
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_CREATING_CONTACTS_BATCH"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:CreateBatch] Error Creating contacts batch: queryId: %s. Info: %s", queryId, json.encode(errType))
	return false, errType
end

function UpdateBatchByQuery (queryId, processId, encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:UpdateBatchByQuery] Updating contacts batch: queryId: %s, processId: %s. Info: %s.", queryId, processId, encodedInfo)
	local response, data = proxy.outboundEngine.UpdateBatchByQuery(queryId, processId, encodedInfo)
	--LogDebug("---> UpdateBatchByQuery success: %s, err: %s", response, err)
	if response then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:UpdateBatchByQuery] Successfully Updated contacts batch: queryId: %s, processId: %s. Info: %s. Response: %s, %s", queryId, processId, encodedInfo, response, data)
		return true, data
	end
	local err = data
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_UPDATING_CONTACTS_BATCH_BY_QUERY"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:UpdateBatchByQuery] Error Updating contacts batch: queryId: %s, processId: %s. Info: %s", queryId, processId, json.encode(errType))
	return false, errType
end

function UpdateBatchByAction (batchId, processId, encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:UpdateBatchByAction] Updating contacts batch: batchId: %s, processId: %s. Info: %s.", batchId, processId, encodedInfo)
	local response, data = proxy.outboundEngine.UpdateBatchByAction(batchId, processId, encodedInfo)
	--LogDebug("---> UpdateBatchByAction success: %s, err: %s", response, err)
	if response then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:UpdateBatchByAction] Successfully Updated contacts batch: batchId: %s, processId: %s. Info: %s. Response: %s, %s", batchId, processId, encodedInfo, response, data)
		return true, data
	end
	local err = data
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_UPDATING_CONTACTS_BATCH_BY_BATCH"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:UpdateBatchByAction] Error Updating contacts batch: batchId: %s, processId: %s. Info: %s", batchId, processId, json.encode(errType))
	return false, errType
end

function DeleteBatch (encodedInfo)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:DeleteBatch] Deleting contacts batch: %s", encodedInfo)
	local response, err = proxy.outboundEngine.DeleteBatch(encodedInfo)
	--LogDebug("---> DeleteBatch success: %s, err: %s", response, err)
	if response then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:DeleteBatch] Successfully Deleted contacts batch: %s. Response: %s, %s", encodedInfo, response, err)
		return true, response
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_DELETING_CONTACTS_BATCH"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:DeleteBatch] Error Deleting contacts batch: %s. Info: %s", encodedInfo, json.encode(errType))
	return false, errType
end

function LoadBatchContactsStatus (processId, batchId)

	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:LoadBatchContactsStatus] Loading Batch %s Contact Fields for Process: %s", batchId, processId)

	local response, err = proxy.outboundEngine.LoadBatchContactsStatus(processId, batchId)

	if response then
		local status = response.contacts_status
		LogInfo("[OUTBOUNDENGINE_PROCESSES:LoadBatchContactsStatus] Successfully Loaded Batch %s Filters for Process: %s. Info: %s", batchId, processId, status)
		return true, status
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_PROCESS_BATCH_CONTACT_STATUS"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:LoadBatchContactsStatus] Error loading Batch %s Filters for Process: %s. Info: %s", batchId, processId, json.encode(errType))
	return false, errType
end

function RunBatchOrContactsQuery (encodedInfo, rowsPerPage, source)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:RunBatchOrContactsQuery] Execute Contact Query. Info: '%s', Source: '%s', rowsPerPage: '%s' .", encodedInfo, source, rowsPerPage)

	local success, data = proxy.outboundEngine.RunBatchOrContactsQuery(encodedInfo, rowsPerPage, source)
	LogDebug("---> RunBatchOrContactsQuery success: %s, data: %s", success, json.encode(data))
	if success then
		LogInfo("[OUTBOUNDENGINE_PROCESSES:RunBatchOrContactsQuery] Successfully Execute Contact Query %s.", source)
		return true, data
	end
	-- Error
	local err = data
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_EXECUTING_CONTACT_QUERY"
	}
	return false, errType
end

function LoadContactsFromQuery (queryId, start, count, requestId)
	LogInfo("[ALLEGRO.OUTBOUNDENGINE_PROCESSES:LoadContactsFromQuery] Loading Contacts From Query '%s', Start: '%s', Count: '%s', requestId: '%s'.", queryId, start, count, requestId)

	local response, err = proxy.outboundEngine.LoadContactsFromQuery(queryId, start, count, requestId)

	if response then
		local contacts = response.contacts
		LogInfo("[OUTBOUNDENGINE_PROCESSES:LoadContactsFromQuery] Successfully Loaded Contacts From Query. Info: %s", contacts)
		return true, contacts
	end
	-- Error
	local errType = {
		type = "ALLEGRO_INTEGRATION_ERROR",
		err  = type(err) == "table" and err.reason or err or "ERROR_RETRIEVING_CONTACTS_FROM_QUERY"
	}
	LogError("[OUTBOUNDENGINE_PROCESSES:LoadContactsFromQuery] Error loading Contacts From Query. Info: %s", json.encode(errType))
	return false, errType
end
