local m_models = require("apps.flow_designer.model"):GetModels()
module(..., package.seeall, inconcert.orbit.application.Decorate)
require "uuid"
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local classesProcessor = require "apps.flow_designer.process_flows.classes"

m_htmlRender = nil
m_appLocation = nil

local function TraslateAttribute (attribute, _T)
	_T = _T or m_htmlRender.extension.makeTextTranslator()
	attribute.description = _T(attribute.description)
end

local function TraslateMethod (method, _T)
	_T = _T or m_htmlRender.extension.makeTextTranslator()
	method.description = _T(method.description)
end

local function TraslateClass (class, _T)
	if not class.System then
		return
	end
	_T = _T or m_htmlRender.extension.makeTextTranslator()
	class.Description = _T(class.Description)
	for _, attribute in pairs(class.Editor.attributes) do
		TraslateAttribute(attribute, _T)
	end
	TraslateMethod(class.Editor.classConstructor, _T)
	for _, method in pairs(class.Editor.methods) do
		TraslateMethod(method, _T)
	end
end

local function TraslateClasses (classes, _T)
	_T = _T or m_htmlRender.extension.makeTextTranslator()
	for _, class in pairs(classes) do
		TraslateClass(class, _T)
	end
end

--Returns all classes in the VCC
local function GetClasses (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local classes = m_models.ProcessFlowClasses.GetFlowClassess(vcc)
	TraslateClasses(classes)
	return classes
end

local function Get (web, classId)
	local vcc = inconcert.authUser.virtualContactCenter
	local classes = m_models.ProcessFlowClasses.GetFlowClassess(vcc, classId)

	if next(classes) then
		classes.Name = classes.Name or classId
		return classes
	else
		web.status = 404
		return {result = "NOT_FOUND"}
	end
end

local function ValidateClasses (classes, keyspace, closeKs)
	local vcc = inconcert.authUser.virtualContactCenter
	local existingClasses = m_models.ProcessFlowClasses.GetFlowClassess(vcc, nil, keyspace)

	local classesByName = {}
	for classId, flowClass in pairs(existingClasses) do
		classesByName[flowClass.Name] = classId
	end

	for classId, flowClass in pairs(classes) do
		if type(flowClass.Name) ~= "string" or flowClass.Name == "" then
			keyspace:Close()
			LogError("Class %s doesnt have a Name", classId)
			return 400, {result = "MUST_HAVE_A_NAME", reason = "Class must have a name"}
		end
		if classesByName[flowClass.Name] and classesByName[flowClass.Name] ~= classId then
			keyspace:Close()
			LogError("Class %s name %s already exists", classId, flowClass.Name)
			return 400, {result = "NAME_ALREADY_EXISTS", reason = "Class name already exists", classId = classesByName[flowClass.Name] }
		end
		classesByName[flowClass.Name] = true
	end

	if closeKs then
		keyspace:Close()
	end

	return 200, { result = "OK" }
end

local function MakeExecutable (class)
	local status, result = classesProcessor.MakeExecutable(class)
	LogDebug("Flow Class %s executable: %s", class.Name, class.Executable)
	return status, result
end

--[[
	Add/Modifies classes in a vcc
	posted data must have the form:
{
	classId1 = {
		Description = "class description 1",
	},
	classId2 = {
		Description = "class description 2",
	},
	...
}

]]
local function SaveClasses (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:Save]Classes, userId %s", userId)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:Save]Classes, userId %s, post data:  %s", userId, web.input["post_data"])

	local classes = web.request
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	local status, result = ValidateClasses(classes, keyspace)

	if status == 200 then
		for classId, flowClass in pairs(classes) do
			status, result = MakeExecutable(flowClass)
			if status ~= 200 then
				break
			end
		end
		if status == 200 then
			m_models.ProcessFlowClasses.SaveFlowClassess(vcc, classes, keyspace)
		end
		keyspace:Close()
	end

	web.status = status
	return result
end

-- local function count(list)
	-- local count = 0
	-- for _ in pairs(list or {}) do count = count + 1 end
	-- return count
-- end

local function ValidateRemoveClasses (classes, keyspace, closeKs)
	local vcc = inconcert.authUser.virtualContactCenter
	local flows = m_models.ProcessFlows.GetFlows(vcc, nil, nil, keyspace)

	local flowsByClass = {}
	for flowId, flow in pairs(flows) do
		for classId, _ in pairs(flow.Classes or {}) do
			flowsByClass[classId] = flowsByClass[classId] or {}
			table.insert(flowsByClass[classId], { id = flowId, name = flow.Name })
		end
	end

	for _, classId in pairs(classes) do
		if flowsByClass[classId] then
			keyspace:Close()
			LogError("Class %s is in flow: %s", classId, json.encode(flowsByClass[classId]))
			return 400, {result = "CLASS_IN_USE", reason = "Class in use", flows = flowsByClass[classId] }
		end
	end

	if closeKs then
		keyspace:Close()
	end

	return 200, { result = "OK" }
end

--[[
Remove classes from a vcc
postedData must be of the form: ["classId1", "classId2", ...]
]]
local function RemoveClasses (web)
	local vcc = inconcert.authUser.virtualContactCenter
	LogInfo("RemoveClasses -> user '%s', vcc '%s', classes: '%s'.", inconcert.authUser.userName, vcc, web.input["post_data"])

	local classes_to_remove = web.request
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	local status, result = ValidateRemoveClasses(classes_to_remove, keyspace)

	if status == 200 then
		m_models.ProcessFlowClasses.RemoveFlowClassess(vcc, classes_to_remove, keyspace)
		keyspace:Close()
	end

	web.status = status
	return result
end

local function ExportClasses (web, classes_s)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ExportClasses] User '%s' Exporting processes flow classes '%s' vcc '%s'.", userId, classes_s, vcc)

	local classes = json.decode(classes_s)
	local data = {}
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	for _, classId in pairs(classes) do
		local class = m_models.ProcessFlowClasses.GetFlowClassess(vcc, classId)
		if not next(class) then
			keyspace:Close()
			web.status = 404
			return {result = "NOT_FOUND", classId = classId}
		end
		class.Executable = nil
		data[classId] = class
	end
	keyspace:Close()
	web.headers["Content-Disposition"] = [[attachment; filename="]] .. 'flowClasses.json"'
	return data
end

local function ValidateClassToBeImported (flowClass, existingClasses, existingClassesByName)
	if type(flowClass.Id) ~= "string" or flowClass.Id == "" then
		LogError("Class doesnt have an id")
		return 400, {result = "MUST_HAVE_A_DESCRIPTION", reason = "Class must have a description"}
	end
	if type(flowClass.Name) ~= "string" or flowClass.Name == "" then
		LogError("Class %s doesnt have a Name", flowClass.Id)
		return 400, {result = "MUST_HAVE_A_NAME", reason = "Class must have a name"}
	end
	if flowClass.System then
		return 200, { data = flowClass }
	end
	local existingId = existingClasses[flowClass.Id] and existingClasses[flowClass.Id].Name or false
	local existingName = existingClassesByName[flowClass.Name] and existingClassesByName[flowClass.Name].Id or false
	if existingName and existingClassesByName[flowClass.Name].System then
		LogError("Class %s has system class name %s", flowClass.Id, flowClass.Name)
		return 400, {result = "SYSTEM_CLASS", reason = "A system class has same name"}
	end
	return 200, { data = flowClass, existingId = existingId, existingName = existingName }
end

function ImportClasses (classesToImport, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local existingClasses = m_models.ProcessFlowClasses.GetFlowClassess(vcc, nil, keyspace)
	local existingClassesByName = {}
	for _, class in pairs(existingClasses) do
		existingClassesByName[class.Name] = class
	end

	local classes = {}
	for classId, class in pairs(classesToImport) do
		local status, result = ValidateClassToBeImported(class, existingClasses, existingClassesByName)
		if status ~= 200 then
			return status, result
		end
		classes[classId] = result
	end
	
	return 200, { result = "OK", classes = classes }
end

--[[
	Funcion que actualiza el código de las clases, de versiones anteriores.
	Ej: clase "luis" ahora soporta NLU de "RASA" y de "LUIS", por lo que los nombres de invocaciones de las funciones ha cambiado.
]]--
function UpdateOldVersionFlow (flow)
	assert(type(flow) == "table")

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:UpdateOldVersionFlow] Updating Classes of Old Version Flow.")

	local columnsToProcess = {}

	-- 1 - Columna "Draw"
	columnsToProcess.Draw = {
		{oldValue = "classHelpers.luis", newValue = "classHelpers.nlu"},
		{oldValue = "ExecuteApiLUIS"   , newValue = "ExecuteApiNLU"   }
	}
	-- 2 - Columna "ClassesMigration"
	columnsToProcess.ClassesMigration = columnsToProcess.Draw

	-- 3 - Recorro la lista de columnas a evaluar el reemplazo
	local sanitizedColumns = {}
	for columnName, columnValuesToProcess in pairs(columnsToProcess) do
		local columnValue =  flow[columnName]
		if columnValue then
			--LogDebug("---> %s %s", columnName, json.encode(columnValue))
			local encodedColumnValue = json.encode(columnValue)
			--LogDebug("    1 - Antes de reemplazar -  encodedColumnValue: '" .. tostring(encodedColumnValue) .. "'")

			if type(encodedColumnValue) == "string" then
				-- recorro la definicion de las columnas a reemplazar y obtengo los valores con los cuales se va a reemplazar
				local isReplacedValuesInColumn = false
				for _, valuesToReplace in pairs(columnValuesToProcess) do

					-- (***) Aca realizo el Replace !!!!! --------
					local newColumnValue, totalReplaced = encodedColumnValue:gsub(valuesToReplace.oldValue, valuesToReplace.newValue)

					if totalReplaced > 0 then
						isReplacedValuesInColumn = true
						encodedColumnValue = newColumnValue

						--LogDebug("      + Column '" .. tostring(columnName) .. "' --> Total replaced values: " .. tostring(totalReplaced))
						--LogDebug("           ---> Replaced Values: '" .. json.encode(valuesToReplace) .. "'")
						--LogDebug("")
					--else
						--LogDebug("      - NO replaced values in Column " .. tostring(columnName))
					end
				end
				if isReplacedValuesInColumn then
					--LogDebug("    2 - Despues de reemplazar -  encodedColumnValue: '" .. tostring(encodedColumnValue) .. "'")
					flow[columnName] = json.decode(encodedColumnValue)
					table.insert(sanitizedColumns, columnName)
				end
			else
				LogWarning("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:UpdateOldVersionFlow] Unable to updated column '%s' in Flow. Column not is a JSON.", columnName)
			end
		else
			LogWarning("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:UpdateOldVersionFlow] Unable to updated column '%s' in Flow. Column does not exist.", columnName)
		end
	end

	if #sanitizedColumns > 0 then
		LogInfo("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:UpdateOldVersionFlow] %d Columns have been Updated. Columns: %s", #sanitizedColumns, json.encode(sanitizedColumns))
	else
		LogInfo("[FLOW_DESIGNER.PROCESS_FLOW_CLASSES:UpdateOldVersionFlow] No Columns has been Updated.")
	end

	return 200, {result = "OK", flow = flow}
end

local function ImportClassesWeb (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("Upload Classes to be imported, userId %s, post data:  %s", userId, web.input["post_data"])

	local status, result = ImportClasses(web.request)
	web.status = status
	return result
end

function MapUris (module, htmlRender, prefix, appLocation)

	local LogRequest = require("apps.agent.requestFilters").LogRequest

	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "view_flow_classes", Get), prefix.."/get/(.+)")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "view_flow_classes", GetClasses), prefix.."/get_all/")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "export_flow_classes", ExportClasses), prefix.."/export/(.+)")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flow_classes", SaveClasses), prefix.."/save/")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "delete_flow_classes", RemoveClasses), prefix.."/delete/")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flow_classes", ImportClassesWeb), prefix.."/import/")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
