local m_models = require("apps.flow_designer.model"):GetModels()
module(..., package.seeall, inconcert.orbit.application.Decorate)
require "uuid"
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local m_moduleLocation = select(2, ...)

m_htmlRender = nil
m_appLocation = nil

--Returns all datasources in the VCC
local function GetDatasources (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local datasources = m_models.ProcessFlowDatasources.GetFlowDatasources(vcc, nil, nil, true)
	return datasources
end

local function Get (web, datasourceId)
	local vcc = inconcert.authUser.virtualContactCenter
	local datasources = m_models.ProcessFlowDatasources.GetFlowDatasources(vcc, datasourceId)

	if next(datasources) then
		datasources.Name = datasources.Name or datasourceId
		return datasources
	else
		web.status = 404
		return {result = "NOT_FOUND"}
	end
end

-- Returns all bultin databases for Vcc
local function GetBuiltinDatabases (web)
	local vcc = inconcert.authUser.virtualContactCenter

	if not vcc then
		web.status = 404
		return {result = "NOT_FOUND"}
	end

	local databases = utils.GetAdmModel("Databases").GetDatabases(nil, nil, vcc)
	local formatDatabases = {}
	for key, value in pairs(databases) do
		table.insert(formatDatabases, {id = key, name = value.Name })
	end
	return formatDatabases
end

local function ValidateDatasources (datasources, keyspace, closeKs)
	local vcc = inconcert.authUser.virtualContactCenter
	local existingDatasources = m_models.ProcessFlowDatasources.GetFlowDatasources(vcc, nil, keyspace)

	local datasourcesByName = {}
	for dsId, flowDs in pairs(existingDatasources) do
		datasourcesByName[flowDs.Name] = dsId
	end

	for dsId, flowDs in pairs(datasources) do
		if type(flowDs.Type) ~= "string" or flowDs.Type == "" then
			keyspace:Close()
			LogError("Datasource %s doesnt have a Type", dsId)
			return 400, {result = "MUST_HAVE_A_TYPE", reason = "Datasource must have a type"}
		end
		if type(flowDs.Name) ~= "string" or flowDs.Name == "" then
			keyspace:Close()
			LogError("Datasource %s doesnt have a Name", dsId)
			return 400, {result = "MUST_HAVE_A_NAME", reason = "Datasource must have a name"}
		end
		if datasourcesByName[flowDs.Name] and datasourcesByName[flowDs.Name] ~= dsId then
			keyspace:Close()
			LogError("Datasource %s name %s already exists", dsId, flowDs.Name)
			return 400, {result = "NAME_ALREADY_EXISTS", reason = "Datasource name already exists", dsId = datasourcesByName[flowDs.Name] }
		end
		datasourcesByName[flowDs.Name] = true
	end

	if closeKs then
		keyspace:Close()
	end

	return 200, { result = "OK" }
end

local function MakeExecutable (datasource)
	local code = ""
	
	datasource.Executable = code
	LogDebug("Flow Datasource %s executable: %s", datasource.Name, datasource.Executable)
	return 200, { result = "OK" }
end

--[[
	Add/Modifies datasources in a vcc
	posted data must have the form:
{
	dsId1 = {
		Description = "ds description 1",
	},
	dsId2 = {
		Description = "ds description 2",
	},
	...
}

]]
local function SaveDatasources (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("Save Datasources, userId %s, post data:  %s", userId, web.input["post_data"])

	local datasources = web.request
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	local status, result = ValidateDatasources(datasources, keyspace)

	if status == 200 then
		for dsId, flowDs in pairs(datasources) do
			status, result = MakeExecutable(flowDs)
			if status ~= 200 then
				break
			end
		end
		if status == 200 then
			m_models.ProcessFlowDatasources.SaveFlowDatasources(vcc, datasources, keyspace)
		end
		keyspace:Close()
	end

	web.status = status
	return result
end

local function ValidateRemoveDatasources (datasources, keyspace, closeKs)
	local vcc = inconcert.authUser.virtualContactCenter
	local flows = m_models.ProcessFlows.GetFlows(vcc, nil, nil, keyspace)

	local flowsByDatasource = {}
	for flowId, flow in pairs(flows) do
		for dsId, _ in pairs(flow.Datasources or {}) do
			flowsByDatasource[dsId] = flowsByDatasource[dsId] or {}
			table.insert(flowsByDatasource[dsId], { id = flowId, name = flow.Name })
		end
	end

	for _, dsId in pairs(datasources) do
		if flowsByDatasource[dsId] then
			keyspace:Close()
			LogError("Datasource %s is in flow: %s", dsId, json.encode(flowsByDatasource[dsId]))
			return 400, {result = "DATASOURCE_IN_USE", reason = "Datasource in use", flows = flowsByDatasource[dsId] }
		end
	end

	if closeKs then
		keyspace:Close()
	end

	return 200, { result = "OK" }
end

--[[
Remove datasources from a vcc
postedData must be of the form: ["dsId1", "dsId2", ...]
]]
local function RemoveDatasources (web)
	local vcc = inconcert.authUser.virtualContactCenter
	LogInfo("RemoveDatasources -> user '%s', vcc '%s', datasources: '%s'.", inconcert.authUser.userName, vcc, web.input["post_data"])

	local datasources_to_remove = web.request
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	local status, result = ValidateRemoveDatasources(datasources_to_remove, keyspace)

	if status == 200 then
		m_models.ProcessFlowDatasources.RemoveFlowDatasources(vcc, datasources_to_remove, keyspace)
		keyspace:Close()
	end

	web.status = status
	return result
end

local function ExportDatasources (web, datasources_s)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[ExportDatasources] User '%s' Exporting processes flow datasources '%s' vcc '%s'.", userId, datasources_s, vcc)

	local datasources = json.decode(datasources_s)
	local data = {}
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	for _, dsId in pairs(datasources) do
		local datasource = m_models.ProcessFlowDatasources.GetFlowDatasources(vcc, dsId)
		if not next(datasource) then
			keyspace:Close()
			web.status = 404
			return {result = "NOT_FOUND", dsId = dsId}
		end
		datasource.Executable = nil
		data[dsId] = datasource
	end
	keyspace:Close()
	web.headers["Content-Disposition"] = [[attachment; filename="]] .. 'flowDatasources.json"'
	return data
end

local function ValidateDatasourceToBeImported (flowDs, existingDatasources, existingDatasourcesByName)
	if type(flowDs.Id) ~= "string" or flowDs.Id == "" then
		LogError("Datasource doesnt have an id")
		return 400, {result = "MUST_HAVE_A_DESCRIPTION", reason = "Datasource must have a description"}
	end
	if type(flowDs.Type) ~= "string" or flowDs.Type == "" then
		LogError("Datasource %s doesnt have a Type", flowDs.Id)
		return 400, {result = "MUST_HAVE_A_TYPE", reason = "Datasource must have a type"}
	end
	if type(flowDs.Name) ~= "string" or flowDs.Name == "" then
		LogError("Datasource %s doesnt have a Name", flowDs.Id)
		return 400, {result = "MUST_HAVE_A_NAME", reason = "Datasource must have a name"}
	end
	local existingId = existingDatasources[flowDs.Id] and existingDatasources[flowDs.Id].Name or false
	local existingName = existingDatasourcesByName[flowDs.Name] and existingDatasourcesByName[flowDs.Name].Id or false
	if existingName and existingDatasourcesByName[flowDs.Name].System then
		LogError("Datasource %s has system datasource name %s", flowDs.Id, flowDs.Name)
		return 400, {result = "SYSTEM_DATASOURCE", reason = "A system datasource has same name"}
	end
	return 200, { data = flowDs, existingId = existingId, existingName = existingName }
end

function ImportDatasources (datasourcesToImport, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local existingDatasources = m_models.ProcessFlowDatasources.GetFlowDatasources(vcc, nil, keyspace)
	local existingDatasourcesByName = {}
	for _, datasource in pairs(existingDatasources) do
		existingDatasourcesByName[datasource.Name] = datasource
	end
	
	local datasources = {}
	for dsId, datasource in pairs(datasourcesToImport) do
		local status, result = ValidateDatasourceToBeImported(datasource, existingDatasources, existingDatasourcesByName)
		if status ~= 200 then
			return status, result
		end
		datasources[dsId] = result
	end
	
	return 200, { result = "OK", datasources = datasources }
end

local function ImportDatasourcesWeb (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("Upload Datasources to be imported, userId %s, post data:  %s", userId, web.input["post_data"])

	local status, result = ImportDatasources(web.request)
	web.status = status
	return result
end

function MapUris(module, htmlRender, prefix, appLocation)

	local LogRequest = require("apps.agent.requestFilters").LogRequest

	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "view_flow_datasources", Get), prefix.."/get/(.+)")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "view_flow_datasources", GetBuiltinDatabases), prefix.."/get_builtin_databases/")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "view_flow_datasources", GetDatasources), prefix.."/get_all/")
	module:dispatch_get(utils.JsonReply..LogRequest..permissions:CheckPermission("flow_designer", "export_flow_datasources", ExportDatasources), prefix.."/export/(.+)")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flow_datasources", SaveDatasources), prefix.."/save/")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "delete_flow_datasources", RemoveDatasources), prefix.."/delete/")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flow_datasources", ImportDatasourcesWeb), prefix.."/import/")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
