local m_models = require("apps.flow_designer.model"):GetModels()
local ir = require "lua_iragentlink"
module(..., package.seeall, inconcert.orbit.application.Decorate)
require "uuid"
local utils = require "shared.apps.utils"
local permissions = require "shared.permissions.permissions"
local apiLuis = require("shared.cognitive_services.luis.luis").GetAPI() -- rocka
local classesController = require "apps.flow_designer.controllers.process_flow_classes"
local datasourcesController = require "apps.flow_designer.controllers.process_flow_datasources"
local repositoryController = require("apps.repository")
local http_mapping = require "apps.agent.result_codes_to_http_codes"

local m_moduleLocation = select(2, ...)

local mainProcessor = require "apps.flow_designer.process_flows.main"
local stateProcessor = require "apps.flow_designer.process_flows.state"
local auxProcessor = require "apps.flow_designer.process_flows.auxiliar"
local m_activitiesProcessors = {}

local function StartRepository ()
	repositoryController.ManualStart(inconcert.environment.apps.agent.cassandra.ip, inconcert.environment.apps.agent.cassandra.port)
end

local function GetActivities ()
	local activityDir = m_moduleLocation .. "/../process_flows/activities"
	local activities = {}
	for file in lfs.dir(activityDir) do
		if file ~= "." and file ~= ".." then
			local f = activityDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local activityName = string.match(file, [[(.+).lua]])
				if activityName then
					activities[activityName] = { url = activityName }
				end
			end
		end
	end
	return activities
end

for key, activityDef in pairs(GetActivities()) do
	m_activitiesProcessors[key] = require("apps.flow_designer.process_flows.activities." .. key)
end

m_htmlRender = nil
m_appLocation = nil

--Returns all flows in the VCC
local function Search (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local fullUserId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	local startTime = os.time()

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOWS:SEARCH] 1 - (START) Search rows for user '%s', group '%s'. Posted data: '%s'", fullUserId, inconcert.authUser.userGroup, web.input["post_data"])

	local postData = web.request
	--Get all flows from model

	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	local rows = m_models.ProcessFlows.GetFlows(vcc, nil, {"Id", "Name", "Type", "Description"}, keyspace)
	local totalUnfilteredRows = utils.GetRowCount(rows)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOWS:SEARCH] 2 - Filtering rows")
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOWS:SEARCH] ===> 2.1 - rows: %s", json.encode(rows or {}))

	for flowId, flow in pairs(rows) do
		flowId = utils.GetId(flowId)
		flow.Id = flowId
		flow.Name = flow.Name or flowId
		flow.Accounts = {}
		local accountsList = m_models.ProcessFlows.GetAccountsByFlow(vcc, flowId, keyspace) or {}
		for _, account in pairs(accountsList) do
			table.insert(flow.Accounts, { name = account.account_name, campaign = utils.GetId(account.campaign_id), assignment_type = account.assignment_type })
		end
		flow.InUse = #flow.Accounts > 0
		if utils.MatchFilter(flow, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, flow)
		end
	end
	keyspace:Close()

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOWS:SEARCH] 3 - Sorting rows %d", #filteredRows)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOWS:SEARCH] ===> 3.1 - filteredRows: %s", json.encode(filteredRows or {}))

	--include only rows that belong to logged user vcc
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOWS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

local function Get (web, flowId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:Get] Retrieving flowId %s, user %s, vcc %s", flowId, userId, vcc)
	local flows = m_models.ProcessFlows.GetFlows(vcc, flowId)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:Get] --> Process flow %s: %s", flowId, json.encode(flows))
	if next(flows) then
		-- ok
		flows.Name = flows.Name or flowId
		return flows
	end
	-- error
	web.status = 404
	return {result = "NOT_FOUND"}
end

local function GetTrace (web, flowId)
	local vcc = inconcert.authUser.virtualContactCenter
	local trace = m_models.ProcessFlows.GetTrace(vcc, flowId)
	return trace
end

local function GetEvent (web, reference)
	local trace = m_models.ProcessFlows.GetEvents(reference)
	return { events = trace }
end

----------------------------------------------------------------------------

--[[
	Incluye la informacion del sharedfile de la actividad en el array de actividades que usan sharedfiles.
]]--
local function IncludeSharedFile (drawSharedFiles, stateId, activityId, activity, sharedFile, RepositoryId)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:IncludeSharedFile] ---->  stateId '%s', activityId '%s', activity '%s', sharedFile '%s'", stateId, activityId, activity, sharedFile)
	drawSharedFiles[stateId] = drawSharedFiles[stateId] or {}
	drawSharedFiles[stateId][activityId] = drawSharedFiles[stateId][activityId] or {}
	drawSharedFiles[stateId][activityId].type = activity.data.type
	drawSharedFiles[stateId][activityId].description = activity.data.description
	drawSharedFiles[stateId][activityId].sharedFiles = drawSharedFiles[stateId][activityId].sharedFiles or {}
	table.insert(drawSharedFiles[stateId][activityId].sharedFiles, sharedFile)
	if RepositoryId then
		drawSharedFiles[stateId][activityId].RepositoryIds = drawSharedFiles[stateId][activityId].RepositoryIds or {}
		table.insert(drawSharedFiles[stateId][activityId].RepositoryIds, RepositoryId)
	end
end

--[[
	Valida si la actividad tiene cofigurado shared files y si es asi, invoca a la funcion correspondiente para que la inserte en el Array.
]]--
local function ProcessSharedFiles (flowId, stateId, activityId, activity, drawSharedFiles)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessSharedFiles]   --> -  flowId '%s', stateId '%s', activityId '%s', activity '%s'", flowId, stateId, activityId, json.encode(activity or {}))
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessSharedFiles]     --> - ENTRA  drawSharedFiles '%s'", json.encode(drawSharedFiles or {}))
	if not activity.data then
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessSharedFiles]     --> - SALE (1)")
		return
	end
	--activity message & sharedfile info
	if not activity.data.message then
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessSharedFiles]     --> - SALE (2)")
		return
	end
	--
	if activity.data.message.sharedFile then
		IncludeSharedFile(drawSharedFiles, stateId, activityId, activity, activity.data.message.sharedFile, activity.data.message.RepositoryId)
	else
		--LogDebug("      - activity.data.message.msgtype %s", activity.data.message.msgtype)
		if activity.data.message.msgtype == "attachment" then
			-- actividades NO menu
			for _, component in ipairs(activity.data.message.options or {}) do
				if component.sharedFile then
					IncludeSharedFile(drawSharedFiles, stateId, activityId, activity, component.sharedFile)
				end
			end
		elseif activity.data.message.msgtype == "menu" then
			-- actividad menu
			--LogDebug("        -- activity.data.message.options %s", json.encode(activity.data.message.options))
			for _, menuOption in ipairs(activity.data.message.options or {}) do
				if type(menuOption.message) == "table" and menuOption.message.msgtype == "attachment" then
					for _, component in ipairs(menuOption.message.options or {}) do
						if component.sharedFile then
							IncludeSharedFile(drawSharedFiles, stateId, activityId, activity, component.sharedFile)
						end
					end
				end
			end
		end
	end
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessSharedFiles]     --> - SALE (3)  drawSharedFiles '%s'", json.encode(drawSharedFiles or {}))
end

--[[
	Incluye la cuenta de Whatsapp de la actividad en el array de actividades que usan cuentas.
]]--
local function IncludeAccount (accounts, stateId, activityId, activity, info, keyspace)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:IncludeAccount] ---->  stateId '%s', activityId '%s', activity '%s', info '%s'", stateId, activityId, activity, info)
	local account, campaignId =  utils.GetAdmModel("Campaigns").FindAccountById(info.account, keyspace)
	if not account then
		LogWarning("[FLOW_DESIGNER.PROCESS_FLOW:IncludeAccount] ---->  Unable to include account '%s', account does not exist. stateId '%s', activityId '%s', activity '%s', info '%s'", info.account, stateId, activityId, activity, info)
		return
	end
	accounts[stateId] = accounts[stateId] or {}
	accounts[stateId][activityId] = accounts[stateId][activityId] or {}
	accounts[stateId][activityId].type = activity.data.type
	accounts[stateId][activityId].description = activity.data.description
	accounts[stateId][activityId].accounts = accounts[stateId][activityId].accounts or {}
	table.insert(accounts[stateId][activityId].accounts, {
		accountId   = info.account,
		accountData = account,
		accountName = account.name or info.account,
		proxyType   = account.proxyType,
		campaignId  = campaignId,
		templateId  = info.id
	})
end

--[[
	Valida si la actividad tiene una cuenta de Whatsapp, y si es asi invoca a la funcion correspondiente para que la inserte en el Array.
]]--
local function ProcessAccounts (flowId, stateId, activityId, activity, accounts, keyspace)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAccounts]   --> -  flowId '%s', stateId '%s', activityId '%s', activity '%s'", flowId, stateId, activityId, json.encode(activity or {}))
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAccounts]     --> - ENTRA  accounts '%s'", json.encode(accounts or {}))
	if not activity.data then
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAccounts]     --> - SALE (1)")
		return
	end
	--activity message
	if not activity.data.message then
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAccounts]     --> - SALE (2)")
		return
	end
	--
	--LogDebug("      - activity.data.message.messageSource %s, activity.data.message.msgtype %s", activity.data.message.messageSource, activity.data.message.msgtype)
	-- #14143: No estabamos incluyendo los msgtype "reply_buttons", los cuales no llevan messageSource
	if activity.data.message.messageSource == "account" and activity.data.message.msgtype == "template" then
		local template = activity.data.message.template or {}
		IncludeAccount(accounts, stateId, activityId, activity, template, keyspace)
	elseif activity.data.message.accountId and activity.data.message.msgtype == "reply_buttons" then
		local account_info = { account = activity.data.message.accountId }
		IncludeAccount(accounts, stateId, activityId, activity, account_info, keyspace)
	end
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAccounts]     --> - SALE (3)  drawSharedFiles '%s'", json.encode(accounts or {}))
end

--[[
	Funcion que valida que NO se puedan configurar cuentas de Whatsapp diferentes en todo el dibujo.
	Es decir puede configurarse la misma cuenta de WhtasApp en todas las actividades necesarias, pero no es posible configurar cuentas diferentes.
]]--
local function ValidateFlowAccounts (flowId, accountsByFlows, keyspace)
	accountsByFlows = accountsByFlows or {}
--	LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]  (ENTRA)--> -  flowId '%s'\naccountsByFlows", flowId, json.encode(accountsByFlows))

	------------------------
	-- 1 - Primero recorro las actividades del dibujo, validando que solo exista 1 cuenta de Whatsapp entre las diferentes actividades.
	--		Es decir la cuenta whastaspp_1 pueda estar N veces en las diferentes actividades y estaria Ok, pero si configuro 2 cuentas diferentes fallaria.
	--
	local configuredWhatsappAccountId = nil
	local newWhatsappAccounts = {}
	local newWhatsappAccountsErrors = {}
	local newTotalWhatsappAccounts = 0
	for currentFlowId, state in pairs(accountsByFlows) do
		--LogDebug("--> currentFlowId: %s", currentFlowId)
		for stateId, stateInfo in pairs(state or {}) do
			--LogDebug("  --> stateId: %s", stateId)
			--LogDebug("   --> stateInfo: %s", json.encode(stateInfo))
			for activityId, activityInfo in pairs(stateInfo or {}) do
				--LogDebug("      --> activityId: %s", activityId)
				--LogDebug("        --> activityInfo: %s", json.encode(activityInfo))
				for _, account in ipairs(activityInfo.accounts or {}) do
					--LogDebug("      --> %s - account: %s", _, json.encode(account))
					if account.proxyType == "WHATSAPP" then
						configuredWhatsappAccountId = account.accountId
						if not newWhatsappAccounts[account.accountId] then
							newWhatsappAccounts[account.accountId] = account
							newWhatsappAccounts[account.accountId].usage = {
								total = 0,
								states = {}
							}
							newTotalWhatsappAccounts = newTotalWhatsappAccounts + 1
						end
						newWhatsappAccounts[account.accountId].usage.total = newWhatsappAccounts[account.accountId].usage.total + 1
						newWhatsappAccounts[account.accountId].usage.states = newWhatsappAccounts[account.accountId].usage.states or {}
						newWhatsappAccounts[account.accountId].usage.states[stateId] = newWhatsappAccounts[account.accountId].usage.states[stateId] or {}
						newWhatsappAccounts[account.accountId].usage.states[stateId][activityId] = (newWhatsappAccounts[account.accountId].usage.states[stateId][activityId] or 0) + 1
						if newWhatsappAccounts[account.accountId].usage.total > 1 then
							newWhatsappAccountsErrors[account.accountId] = newWhatsappAccounts[account.accountId]
						end
					end
				end
			end
		end
	end
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]  (SALE) 1 --> -  flowId '%s', newWhatsappAccounts '%s'", flowId, json.encode(newWhatsappAccounts or {}))
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]  (SALE) 2 --> -  flowId '%s', newWhatsappAccountsErrors '%s'", flowId, json.encode(newWhatsappAccountsErrors or {}))
	if newTotalWhatsappAccounts > 1 then
		-- Si existen 2 cuentas diferentes, fallo.
		local resason = string.format("The process can only have only one WhatsApp account between the different activities. You have %s WhatsApp accounts configured.", newTotalWhatsappAccounts)
		return false, {result = "ONLY_ALLOWS_ONE_WHATSAPP_ACCOUNT", reason = resason, flowId = flowId, totalError = newTotalWhatsappAccounts, dataError = newWhatsappAccountsErrors}
	end

	-- Validaciones contra la Database, si es que hemos configurado cuenta de WhatsApp.
	if configuredWhatsappAccountId then
		-- 2 - (Si se encuentra en uso) Aca valido si el dibujo se ha configurado en alguna cuenta /campaña de la BD (dibujos existentes) y si es así, no lo permito remover del dibujo la cuenta.
		--       Primero debería de dejar de utilizarlo en las cuentas/campañas para después poder quitarlo del dibujo.
		--
		local vcc = inconcert.authUser.virtualContactCenter
		-- obtengo todas las cuentas (sin importar el canal) que el dibujo tiene guardadas en la BD.
		local databaseAccountsByFlow = m_models.ProcessFlows.GetAccountsByFlow(vcc, flowId, keyspace) or {}
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]       b--> -  flowId '%s' databaseAccountsByFlow: %s.", flowId, json.encode(databaseAccountsByFlow))
		-- si tenemos cuentas configuradas en la BD
		if next(databaseAccountsByFlow) then
			local databaseAccountInfo = databaseAccountsByFlow[configuredWhatsappAccountId] or {}
			--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]       b1--> -  flowId '%s' databaseAccountsByFlow: %s.", flowId, json.encode(databaseAccountInfo))
			if not next(databaseAccountInfo) then
				local resason = "Unable to configure accounts in some activities. The process is in use different accounts of campaigns."
				return false, {result = "UNABLE_TO_CHANGE_OR_REMOVE_CONFIGURED_ACCOUNTS_FLOW_IN_USE", reason = resason, flowId = flowId, dataError = databaseAccountsByFlow}
			end
		end

		-- Fix #19500 - It should be allowed to create N drawings related to an account, then only 1 can be assigned to it
		---------
		-- 3 - Ahora valido que la cuenta / campana NO tenga configurado otro dibujo.
		--      Si llega a tener configurado otro dibujo, debo fallar.
		--
		--local newAccount = newWhatsappAccounts[configuredWhatsappAccountId]
		----LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]       c1--> -  flowId '%s' configuredWhatsappAccountId: '%s' newAccount: %s.", flowId, configuredWhatsappAccountId, json.encode(newAccount or {}))
		--if newAccount then
		--	--LogDebug("1")
		--	if newAccount.proxyType == "WHATSAPP" then
		--		--LogDebug("2")
		--		if newAccount.accountData.FlowInformation then
		--			--LogDebug("3")
		--			local onScheduleProcessId = newAccount.accountData.FlowInformation.on_schedule and newAccount.accountData.FlowInformation.on_schedule.Process
		--			--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]       c2--> -  flowId '%s' configuredWhatsappAccountId: '%s' onScheduleProcessId: %s.", flowId, configuredWhatsappAccountId, onScheduleProcessId)
		--			-- valido contra los dibujos en scheduler.
		--			if onScheduleProcessId and onScheduleProcessId ~= flowId then
		--				local resason = "Unable to configure campaign account. The selected account has another process associated with it. On schedule."
		--				return false, {result = "UNABLE_TO_ASSIGN_FLOW_ACCOUNT_HAS_ASSIGNED_OTHER_FLOW_ON_SCHEDULE", reason = resason, flowId = flowId, dataError = databaseAccountsByFlow}
		--			end
		--			-- valido contra los dibujos fuera de hora.
		--			local outOfScheduleProcessId = newAccount.accountData.FlowInformation.out_of_schedule and newAccount.accountData.FlowInformation.out_of_schedule.Process
		--			--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowAccounts]       c3--> -  flowId '%s' configuredWhatsappAccountId: '%s' outOfScheduleProcessId: %s.", flowId, configuredWhatsappAccountId, outOfScheduleProcessId)
		--			if outOfScheduleProcessId and outOfScheduleProcessId ~= flowId then
		--				local resason = "Unable to configure campaign account. The selected account has another process associated with it. Out of schedule."
		--				return false, {result = "UNABLE_TO_ASSIGN_FLOW_ACCOUNT_HAS_ASSIGNED_OTHER_FLOW_OUT_OF_SCHEDULE", reason = resason, flowId = flowId, dataError = databaseAccountsByFlow}
		--			end
		--		end
		--	end
		--end
	end

	return true
end

--[[
	Funcioin que recorre todas las actividades y obtiene los sharedfiles y cuentas configuradas en las actividades.
]]--
local function ProcessAllActivities (flowId, flow, responseData, keyspace)
	if type(flow.Draw) ~= "table" then
		return
	end
	--[[
		callback generico utilizado en la funcion "ProcessAllActivities", para poder procesar y validar las distintas actividades (sharedfiles, accounts, etc)
	]]--
	local callbacksLists = {
		{
		--shared files
			iterFn = function(flowId, stateId, activityId, activity, currentProcessData)
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     1 - sharedfiles callback iterFn -->")
				ProcessSharedFiles(flowId, stateId, activityId, activity, currentProcessData)
			end,
			getFn  = function(flowId)
				responseData.sharedFiles[flowId] = responseData.sharedFiles[flowId] or {}
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     2 - sharedfiles callback getFn --> flowId '%s', retrieving-> sharedFiles: '%s'.", flowId, json.encode(responseData.sharedFiles[flowId]))
				return responseData.sharedFiles[flowId]
			end,
			setFn  = function(flowId, drawSharedFiles)
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     3 - sharedfiles callback setFn --> flowId '%s', setting-> drawSharedFiles: '%s'.", flowId, json.encode(drawSharedFiles or {}))
				responseData.sharedFiles[flowId] = drawSharedFiles or {}
			end
		},
		{
			--accounts
			iterFn = function(flowId, stateId, activityId, activity, currentProcessData)
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     1 - accounts callback iterFn -->")
				ProcessAccounts(flowId, stateId, activityId, activity, currentProcessData, keyspace)
			end,
			getFn  = function(flowId)
				responseData.accountsByFlows[flowId] = responseData.accountsByFlows[flowId] or {}
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     2 - accounts callback getFn --> flowId '%s', retrieving-> accountsByFlows: '%s'.", flowId, json.encode(responseData.accountsByFlows or {}))
				return responseData.accountsByFlows[flowId]
			end,
			setFn  = function(flowId, drawAccounts)
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     3 - accounts callback setFn --> flowId '%s', setting-> drawAccounts: '%s'.", flowId, json.encode(drawAccounts or {}))
				responseData.accountsByFlows[flowId] = drawAccounts or {}
			end
		}
	}

	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities] --> ProcessAllActivities -  flowId '%s', flow.Draw '%s'", flowId, json.encode(flow.Draw))

	-- recorro los diferentes estados
	for stateId, state in pairs(flow.Draw.states or {}) do
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]   --> ProcessAllActivities -  flowId '%s', stateId '%s', state: '%s'", flowId, stateId, json.encode(state or {}))
		-- recorro las actividades
		for activityId, activity in pairs(state.model or {}) do
			--activity data
			--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ProcessAllActivities]     ProcessAllActivities -  flowId '%s', activityId '%s', activity.data '%s'", flowId, activityId, json.encode(activity.data or {}))
			-- recorro la lista de callbacks e infoca a sus funciones
			for _, callbackInfo in ipairs(callbacksLists) do
				-- callback para obtener la informacion que hemos insertado en el array temporal (sharedfiles o cuentas)
				local currentProcessData = callbackInfo.getFn(flowId)
				-- callback para iterar sobre la actividad y buscar los diferentes sharedfiles o cuentas
				callbackInfo.iterFn(flowId, stateId, activityId, activity, currentProcessData)
				-- callback para asignar la informacion que hemos obtenido en la interacion del callback anterior e insertarlo en el array temporal de sharedfiles y cuentas.
				callbackInfo.setFn(flowId, currentProcessData)
			end
		end
	end
end

local function WhatsappSetCloudApiPublicRepositoryFileUrl (campaignIdAtVCC, accountId, fileId, usedById)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:WhatsappSetCloudApiPublicRepositoryFileUrl({
		campaignAtVCC = campaignIdAtVCC,
		account = accountId,
		fileId = fileId,
		usedById = usedById
	})

	if res.result ~= "OK" then
		LogError("Error setting public url for '%s' with repository file '%s' on whatsapp account '%s' and campaign '%s' -> '%s'", usedById, fileId, accountId, campaignIdAtVCC, res.reason)
	end
end

local function WhatsappDelCloudApiPublicRepositoryFileUrl (campaignIdAtVCC, accountId, fileId, usedById)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:WhatsappDelCloudApiPublicRepositoryFileUrl({
		campaignAtVCC = campaignIdAtVCC,
		account = accountId,
		fileId = fileId,
		usedById = usedById
	})

	if res.result ~= "OK" then
		LogError("Error deleting public url for '%s' with repository file '%s' on whatsapp account '%s' and campaign '%s' -> '%s'", usedById, fileId, accountId, campaignIdAtVCC, res.reason)
	end
end

local function WhatsappProcessCloudApiPublicRepositoryFileUrlUsedByFlow(data, isOld)
	for flowId, states in pairs(data.sharedFiles or {}) do
		for stateId, activities in pairs(states) do
			for activityId, activity in pairs(activities) do
				for _, repositoryId in pairs(activity.RepositoryIds or {}) do
					local usedById = flowId .. "_" .. stateId .. "_" .. activityId
					if data.accountsByFlows and data.accountsByFlows[flowId] and data.accountsByFlows[flowId][stateId] and data.accountsByFlows[flowId][stateId][activityId] and data.accountsByFlows[flowId][stateId][activityId].accounts then
						for _, account in pairs(data.accountsByFlows[flowId][stateId][activityId].accounts) do
							if account.proxyType == "WHATSAPP" then
								if isOld then
									WhatsappDelCloudApiPublicRepositoryFileUrl(account.campaignId, account.accountId, repositoryId, usedById)
								else
									WhatsappSetCloudApiPublicRepositoryFileUrl(account.campaignId, account.accountId, repositoryId, usedById)
								end
							end
						end
					end
				end
			end
		end
	end
end

--[[

]]--
local function ValidateProcessFlows (flows, keyspace, pMustValidateRelationsAccounts)
	local mustValidateRelationsAccounts = pMustValidateRelationsAccounts or false
	local vcc = inconcert.authUser.virtualContactCenter
	local existingFlows = m_models.ProcessFlows.GetFlows(vcc, nil, {"Id", "Name", "Type", "Description"}, keyspace)

	local flowsByName = {}
	for flowId, flow in pairs(existingFlows) do
		flowsByName[flow.Name] = {flowId = flowId, oldFlow = flow}
	end

	local responseData = {
		vcc             = vcc,
		sharedFiles     = {},
		accountsByFlows = {}
	}

	for flowId, flow in pairs(flows) do
		if type(flow.Description) ~= "string" or flow.Description == "" then
			keyspace:Close()
			LogError("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlows] Flow %s doesnt have a description", flowId)
			return 400, {result = "MUST_HAVE_A_DESCRIPTION", reason = "Flow must have a description"}
		end
		if type(flow.Name) ~= "string" or flow.Name == "" then
			keyspace:Close()
			LogError("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlows] Flow %s doesnt have a Name", flowId)
			return 400, {result = "MUST_HAVE_A_NAME", reason = "Flow must have a name"}
		end
		if flowsByName[flow.Name] and flowsByName[flow.Name].flowId ~= flowId then
			keyspace:Close()
			LogError("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlows] Flow %s name %s already exists", flowId, flow.Name)
			return 400, {result = "NAME_ALREADY_EXISTS", reason = "Flow name already exists", flowId = flowsByName[flow.Name]}
		end
		-- obtengo el flow actual
		local current_flow = m_models.ProcessFlows.GetFlows(vcc, flowId)
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:Get] --> Process flow %s: %s", flowId, json.encode(flows))
		if next(current_flow) then
			-- proceso todas las actividades y obtengo los sharedfiles y cuentas
			ProcessAllActivities(flowId, current_flow, responseData, keyspace)
			-- actualizo key de redis REPOSITORY_FILE_PUBLIC_URL_USED_BY con las actividades que usan url public de shared files 
			WhatsappProcessCloudApiPublicRepositoryFileUrlUsedByFlow(responseData or {}, true)
		end
		-- reseteo responseData para procesar el flow a guardar
		responseData = {
			vcc             = vcc,
			sharedFiles     = {},
			accountsByFlows = {}
		}

		-- proceso todas las actividades y obtengo los sharedfiles y cuentas
		ProcessAllActivities(flowId, flow, responseData, keyspace)
		-- actualizo key de redis REPOSITORY_FILE_PUBLIC_URL_USED_BY con las actividades que usan url public de shared files 
		WhatsappProcessCloudApiPublicRepositoryFileUrlUsedByFlow(responseData or {})
		-- valido las cuentas.
		if mustValidateRelationsAccounts then
			local success, err = ValidateFlowAccounts(flowId, responseData.accountsByFlows, keyspace)
			if not success then
				keyspace:Close()
				LogError("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlows] Flow %s has conflicts with accounts. Err: %s", flowId, json.encode(err))
				return 400, err
			end
		end
		--
		if flowsByName[flow.Name] then
			flowsByName[flow.Name].flowId = true
		end
	end
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlows] ===> Relations in flow: %s", json.encode(responseData))

	return 200, { result = "OK" }, responseData
end

--[[
	Funcion invocada de forma externa desde otra app "agent" controller "campaigns" para poder obtener la lista de cuentas que están configuradas en las diferentes actividades del flow.
]]--
function ExternalProcessAllActivities (flowId, vcc, keyspace)
	assert(type(flowId) == "string")
	assert(type(vcc) == "string")
	local flow = m_models.ProcessFlows.GetFlows(vcc, flowId, nil, keyspace)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ExternalProcessAllActivities] --> Process flow '%s' vcc '%s', flow: %s", flowId, vcc, json.encode(flow))
	if not next(flow) then
		-- error does not exist.
		LogError("[FLOW_DESIGNER.PROCESS_FLOW:ExternalProcessAllActivities] Process flow does not exist. Flow'%s' vcc '%s'", flowId, vcc)
		return false, {result = "PROCESS FLOW DOES NOT EXIST", id = flowId, vcc = vcc}
	end
	flow.Name = flow.Name or flowId
	local responseData = {
		flowId          = flowId,
		vcc             = vcc,
		sharedFiles     = {},
		accountsByFlows = {}
	}
	ProcessAllActivities(flowId, flow, responseData, keyspace)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ExternalProcessAllActivities] ===> Relations in flow: %s", json.encode(responseData))
	return true, responseData, flow
end
---------------------------------------


local function ValidateProcessFlowsWeb (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlowsWeb] Validate Flows, userId %s", userId)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ValidateProcessFlowsWeb] Validate Flows, post data:  %s", web.input["post_data"])

	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	local mustValidateRelationsAccounts = false
	local status, result = ValidateProcessFlows(web.request, keyspace, mustValidateRelationsAccounts)
	keyspace:Close()

	web.status = status
	return result
end

local function MakeExecutable (flow)
	local activitiesProcessors = m_activitiesProcessors

	--DEV MODE (reload on demand)
	--activitiesProcessors = {}
	--for key, activityDef in pairs(GetActivities()) do
	--	local path = string.format(
	--		"/usr/lib/inconcert/agent/inconcert/apps/flow_designer/process_flows/activities/%s.lua",
	--		key
	--	)
	--	local runner = loadfile(path)
	--	activitiesProcessors[key] = runner()
	--end

	local status, result = mainProcessor.MakeExecutable(flow.Name, flow.Draw, flow.Classes or {}, flow.Datasources or {}, stateProcessor, activitiesProcessors, auxProcessor)
	if status == 200 then
		flow.Executable = json.encode(result)
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:MakeExecutable] Flow %s executable: %s", flow.Name, flow.Executable)
		result = { result = "OK" }
	else
		LogError("[FLOW_DESIGNER.PROCESS_FLOW:MakeExecutable] Error making executable process flow %s - %s", flow.Name, result.reason)
	end
	return status, result
end

--[[
	Add/Modifies flows in a vcc
	posted data must have the form:
{
	flowId1 = {
		Description = "flow description 1",
	},
	flowId2 = {
		Description = "flow description 2",
	},
	...
}

]]
local function SaveProcessFlows (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:SaveProcessFlows] Save Flows, userId %ss", userId)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:SaveProcessFlows] ----> Save Flows, post data:  %s", web.input["post_data"])

	local flows = web.request
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	local mustValidateRelationsAccounts = true
	local status, result, otherInfo = ValidateProcessFlows(flows, keyspace, mustValidateRelationsAccounts)

	if status == 200 then
		for flowId, flow in pairs(flows) do
			status, result = MakeExecutable(flow)
			if status ~= 200 then
				break
			end
		end
		if status == 200 then
			m_models.ProcessFlows.SaveFlows(vcc, flows, keyspace, otherInfo)
		end
	end
	keyspace:Close()

	web.status = status
	return result
end

local function CreateTestSession (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:CreateTestSession] Create Test Session, userId %ss", userId)
	-- LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:CreateTestSession] ----> Create Test Session, post data:  %s", json.encode(web.request))
	
	local session = web.request
	session.userId = userId
	session.vcc = vcc	
	m_models.ProcessFlows.CreateTestSession(session)
	return { result ="OK" }
end

local function CloseTestBot (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:CloseTestBot] Close Test Bot, userId %ss", userId)
	
	local session = web.request
	session.userId = userId
	session.vcc = vcc

	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local res = irLink:CloseTestBot({
		proxyType = session.channel,
		sessionId = session.id,
	})

	--Convert result code to HTTP code
	local http_code = http_mapping.getHTTPCode(res["result"])
	web.status = http_code or 500 --unknown result code

	if res.result ~= "OK" then
		LogError("Error closing test bot", sessionId, res.reason)
		return { result = res.result, reason = res.reason }
	end
	return { result = "OK" }
end

--[[
	Clones flow in a vcc
	posted data must have the form:
{
	id = "cloned flow id",
	name = "new flow name"
}

]]
local function CloneProcessFlow (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:CloneProcessFlow] Clone Flows, userId %s", userId)
--	LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:CloneProcessFlow] Clone Flows, post data:  %s", web.input["post_data"])

	local status, result, otherInfo
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local flow = m_models.ProcessFlows.GetFlows(vcc, web.request.Id, nil, keyspace)
	if next(flow) then
		flow.Id = uuid.new()
		flow.Name = web.request.Name
		local flows = { [flow.Id] = flow }
		local mustValidateRelationsAccounts = false
		status, result, otherInfo = ValidateProcessFlows(flows, keyspace, mustValidateRelationsAccounts)
		if status == 200 then
			m_models.ProcessFlows.SaveFlows(vcc, flows, keyspace, otherInfo)
		end
	else
		status = 404
		result = {result = "NOT_FOUND"}
	end

	keyspace:Close()
	web.status = status
	return result
end

local function count(list)
	local count = 0
	for _ in pairs(list or {}) do count = count + 1 end
	return count
end

--[[
Remove flows from a vcc
postedData must be of the form: ["flowId1", "flowId2", ...]
]]
local function RemoveProcessFlows (web)
	local vcc = inconcert.authUser.virtualContactCenter
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:RemoveProcessFlows] Remove Process Flows -> user '%s', vcc '%s'.", inconcert.authUser.userName, vcc)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:RemoveProcessFlows] Remove Process Flows -> flows: '%s'.", web.input["post_data"])

	local flows_to_remove = web.request
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )

	for _, flowId in pairs(flows_to_remove) do
		local accounts = m_models.ProcessFlows.GetAccountsByFlow( vcc, flowId, keyspace )
		local accountsNumber = count(accounts)
		if accountsNumber > 0 then
			LogWarning("[FLOW_DESIGNER.PROCESS_FLOW:RemoveProcessFlows] Flow '%s'. Is being used for '%d' Accounts: '%s'.", flowId, accountsNumber, json.encode(accounts) )
			local accountsList = {}
			for _, account in pairs(accounts) do
				table.insert(accountsList, { name = account.account_name, campaign = utils.GetId(account.campaign_id) })
			end
			keyspace:Close()
			web.status = 400
			return { result = "FLOW_BEING_USED_FOR_ACCOUNT", accounts = accountsList }
		end
	end

	m_models.ProcessFlows.RemoveFlows(vcc, flows_to_remove, keyspace)

	keyspace:Close()
	return {result = "OK"}
end

local function LoadFlowsByType (web, proxyType)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:LoadFlowsByType] User '%s' Retrieving All processes flows by type '%s' vcc '%s'.", userId, proxyType, vcc)
	local flowsByType = m_models.ProcessFlows.LoadFlowsByType(vcc, proxyType)
	--LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:LoadFlowsByType] Processes flows by type %s: %s", proxyType, json.encode(flowsByType))
	return flowsByType
end

local function AddCognitiveServicesMigrationData (exportedData, appId)
	if exportedData[appId] then
		return
	end
	local vcc = inconcert.authUser.virtualContactCenter
	local app = apiLuis.GetAppFromDataBase(vcc, appId)
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:AddCognitiveServicesMigrationData] AppId: '%s', vcc: '%s', app data: '%s'.", appId, vcc, app)
	exportedData[appId] = app and json.decode(app).name or appId
end

local function MakeRepositoryFileId (file)
	assert(type(file) == "table")
	if file.section == nil then
		file.section = -1
	end
	if file.part == nil then
		file.part = -1
	end
	local repoFileId = string.format("%s[%d][%d]", file.id, file.section, file.part)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:MakeRepositoryFileId] Maked Repository File Id: '%s'", repoFileId)
	return repoFileId
end

local function GetRepositoryFileInfo (file)
	StartRepository()
	local repositoryFile, err = repositoryController.getFile(file.id, file.section, file.part, true)
	if repositoryFile then
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:AddRepositoryOneFileMigration] Retrieved File: '%s' does not exist, Section: '%s', Part: '%s'", file.id, file.section, file.part)
		--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:AddRepositoryOneFileMigration] File Content: '%s'", json.encode(repositoryFile))
		return repositoryFile
	end
	LogWarning("[FLOW_DESIGNER.PROCESS_FLOW:AddRepositoryOneFileMigration] Unable to export Repository File. File: '%s' does not exist, Section: '%s', Part: '%s'.", file.id, file.section, file.part)
end

local function AddRepositoryFilesMigrationData (exportedData, files)
	for _, file in pairs(files) do
		local repositoryFileId = MakeRepositoryFileId(file)
		if not exportedData[repositoryFileId] then
			local repositoryFile = GetRepositoryFileInfo(file)
			if repositoryFile then
				exportedData[repositoryFileId] = repositoryFile
			end
		end
	end
end

local function AddSharedFilesMigrationData (exportedData, files)
	local vcc = inconcert.authUser.virtualContactCenter
	for _, file in pairs(files) do
		if not exportedData[file.id] then
			local sharedFile = utils.GetAdmModel("SharedFiles").GetFiles(vcc, file.id)
			if sharedFile then
				local repoFileInfo = {
					id      = sharedFile.RepositoryId,
					section = sharedFile.Section,
					part    = sharedFile.Part
				}
				local repositoryFile = GetRepositoryFileInfo(repoFileInfo)
				if repositoryFile then
					sharedFile.RespositoryFileInfo = repositoryFile
				else
					LogWarning("[FLOW_DESIGNER.PROCESS_FLOW:AddSharedFilesMigrationData] Shared File: '%s' vcc: '%s'. Does not have the related repo file: %s.", sharedFile.Id, vcc, json.encode(repoFileInfo))
				end
				exportedData[file.id] = sharedFile
				--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:AddSharedFilesMigrationData] Retrieved Shared File: '%s' vcc: '%s'", file.id, vcc)
			else
				LogWarning("[FLOW_DESIGNER.PROCESS_FLOW:AddSharedFilesMigrationData] Shared File: '%s' vcc: '%s' dos not exist.", file.id, vcc)
			end
		end
	end
end

local addMigrationTasks = {
	cognitiveServicesApp = AddCognitiveServicesMigrationData,
	repositoryFiles      = AddRepositoryFilesMigrationData,
	sharedFiles          = AddSharedFilesMigrationData
}

local function AddMigrationData (exportedData, migrationType, migrationData)
	exportedData[migrationType] = exportedData[migrationType] or {}
	if addMigrationTasks[migrationType] then
		addMigrationTasks[migrationType](exportedData[migrationType], migrationData)
	else
		LogWarning("[AddMigrationData] Unknown task: %s", migrationType)
	end
end

--makes data to be migrated
--example: store actual cognitive services app name to be migrated easily
local function ExportMigrationData (draw)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ExportMigrationData] Exporting Migration Data. Flow.Draw '%s'.", json.encode(draw))
	local exportedData = {}
	--states
	for _, state in pairs(draw.states or {}) do
		--activities migrations
		for _, activityMigration in pairs(state.migration or {}) do
			--migration
			for migrationType, migrationData in pairs(activityMigration or {}) do
				AddMigrationData(exportedData, migrationType, migrationData)
			end
		end
	end
	return 200, exportedData
end

local function ExportClassesMigrationData (classes, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local data = {}
	for classId, _ in pairs(classes) do
		local class = m_models.ProcessFlowClasses.GetFlowClassess(vcc, classId, keyspace)
		if not class then
			return 404, {result = "NOT_FOUND", classId = classId}
		end
		if not class.System then
			class.Executable = nil
			data[classId] = class
		else
			data[classId] = {
				Id = classId,
				Name = classId,
				System = true
			}
		end
	end
	return 200, data
end

local function ExportDatasourcesMigrationData (datasources, keyspace)
	local vcc = inconcert.authUser.virtualContactCenter
	local data = {}
	for dsId, _ in pairs(datasources) do
		local datasource = m_models.ProcessFlowDatasources.GetFlowDatasources(vcc, dsId, keyspace)
		if not datasource then
			return 404, {result = "NOT_FOUND", dsId = dsId}
		end
		datasource.Executable = nil
		data[dsId] = datasource
	end
	return 200, data
end

local function ExportProcessFlow (web, flowId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:ExportProcessFlow] User '%s' Exporting processes flow '%s' vcc '%s'.", userId, flowId, vcc)
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local flows = m_models.ProcessFlows.GetFlows(vcc, flowId, nil, keyspace)
	if next(flows) then
		flows.Executable = nil
		--migration
		local status, migration = ExportMigrationData(flows.Draw or {})
		if status ~= 200 then
			web.status = status
			return migration
		end
		flows.Migration = migration
		--classes migration
		local status, classesMigration = ExportClassesMigrationData(flows.Classes or {}, keyspace)
		if status ~= 200 then
			keyspace:Close()
			web.status = status
			return classesMigration
		end
		flows.ClassesMigration = classesMigration
		--datasources migration
		local status, datasourcesMigration = ExportDatasourcesMigrationData(flows.Datasources or {}, keyspace)
		if status ~= 200 then
			keyspace:Close()
			web.status = status
			return datasourcesMigration
		end
		keyspace:Close()
		flows.DatasourcesMigration = datasourcesMigration
		--headers
		web.headers["Content-Disposition"] = [[attachment; filename="]] .. flows.Name .. '.json"'
		return flows
	else
		web.status = 404
		return {result = "NOT_FOUND"}
	end
end

local function UpdateOldVersionFlow (flow)
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:UpdateOldVersionFlow] Updating Old Version Flow")
	-- 1 - Actualiza el código de clases del que haya cambiado en las versiones.
	local status, result = classesController.UpdateOldVersionFlow(flow)
	return status, result
end

local function ValidateFlowToBeImported (flow)
	local vcc = inconcert.authUser.virtualContactCenter
	local keyspace = utils.GetDatasource():CassandraConnection( "Agent" )
	local existingFlows = m_models.ProcessFlows.GetFlows(vcc, nil, nil, keyspace)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:ValidateFlowToBeImported] Validating Uploaded Flows to be import. Vcc:  %s", vcc)

	if type(flow.Id) ~= "string" or flow.Id == "" then
		keyspace:Close()
		LogError("Flow doesnt have an id")
		return 400, {result = "MUST_HAVE_A_DESCRIPTION", reason = "Flow must have a description"}
	end
	if type(flow.Description) ~= "string" or flow.Description == "" then
		keyspace:Close()
		LogError("Flow %s doesnt have a description", flow.Id)
		return 400, {result = "MUST_HAVE_A_DESCRIPTION", reason = "Flow must have a description"}
	end
	if type(flow.Name) ~= "string" or flow.Name == "" then
		keyspace:Close()
		LogError("Flow %s doesnt have a Name", flow.Id)
		return 400, {result = "MUST_HAVE_A_NAME", reason = "Flow must have a name"}
	end

	local status, result = classesController.ImportClasses(flow.ClassesMigration or {}, keyspace)
	if status ~= 200 then
		keyspace:Close()
		return status, result
	end
	flow.ClassesMigration = result.classes

	local status, result = datasourcesController.ImportDatasources(flow.DatasourcesMigration or {}, keyspace)
	if status ~= 200 then
		keyspace:Close()
		return status, result
	end
	keyspace:Close()
	flow.DatasourcesMigration = result.datasources

	-- Actualiza Flows de Versiones anteriores.
	local status, result = UpdateOldVersionFlow(flow)
	if status ~= 200 then
		local err = result
		return 400, err
	end

	local sanitizedFlow = result.flow
	return 200, {result = "OK", flow = sanitizedFlow, override = existingFlows[sanitizedFlow.Id] and existingFlows[sanitizedFlow.Id].Name or false}
end

local function ImportProcessFlow (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)
	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:ImportProcessFlow] Upload Flows to be imported, userId:  %s", userId)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ImportProcessFlow] Upload Flows to be imported, post data:  %s", web.input["post_data"])
	local status, result = ValidateFlowToBeImported(web.request)
	--LogDebug("[FLOW_DESIGNER.PROCESS_FLOW:ImportProcessFlow] Uploaded Flows to be imported, status: %s, result: %s", status, json.encode(result))
	web.status = status
	return result
end

local function GetLlmProfiles (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[FLOW_DESIGNER.PROCESS_FLOW:GetLlmProfiles] Get LlmProfiles, userId %s", userId)

	local flows = web.request
	local keyspace = utils.GetDatasource():CassandraConnection("Agent")
	local mustValidateRelationsAccounts = true
	
	local vccInfo = utils.GetAllColumns {
		keyspace     = keyspace or "Agent",
		key = vcc,
		columnFamily = "VCCs"
	} or {}

	keyspace:Close()
	local LlmProfiles = {}

	for _, profile in pairs(vccInfo.LlmProfiles or {}) do
		if profile.enabled == "1" then
			table.insert(LlmProfiles, profile)
		end		
	end
		
	return LlmProfiles or {}
end

function MapUris(module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest

	module:dispatch_get (utils.JsonReply         ..LogRequest..permissions:CheckPermission("flow_designer", "view_flows"  , Get)                    , prefix.."/get/(.+)"            )
	module:dispatch_get (utils.JsonReply         ..LogRequest..permissions:CheckPermission("flow_designer", "view_flows"  , GetTrace)               , prefix.."/get_trace/(.+)"      )
	module:dispatch_get (utils.JsonReply         ..LogRequest..permissions:CheckPermission("flow_designer", "view_flows"  , GetEvent)               , prefix.."/get_event/(.+)"      )
	module:dispatch_get (utils.JsonReply         ..LogRequest..permissions:CheckPermission("flow_designer", "export_flows", ExportProcessFlow)      , prefix.."/export/(.+)"         )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..Search                                                                                , prefix.."/search/"             )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , CreateTestSession)      , prefix.."/create_test_session/")
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , CloseTestBot)           , prefix.."/close_test_bot/"     )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , SaveProcessFlows)       , prefix.."/save/"               )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , CloneProcessFlow)       , prefix.."/clone/"              )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , ValidateProcessFlowsWeb), prefix.."/validate/"           )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , ImportProcessFlow)      , prefix.."/import/"             )
	module:dispatch_post(utils.JsonValidatorReply..LogRequest..permissions:CheckPermission("flow_designer", "delete_flows", RemoveProcessFlows)     , prefix.."/delete/"             )
	module:dispatch_get (utils.JsonReply         ..LogRequest..permissions:CheckPermission("flow_designer", "save_flows"  , GetLlmProfiles)         , prefix.."/get_llm_profiles/"   )
	module:dispatch_get(utils.JsonReply .. LogRequest .. LoadFlowsByType, prefix.."/load_by_type/(.+)")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
