local orbit = require("orbit")
require "inconcert.orbit.application"
require("inconcert.translators")
local LuaConfigurationControl = require "LuaConfigurationControl"
local cassandra = require "lua_cassandra"
local utils = require "shared.apps.utils"
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")

module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)
_M.model = nil

local model = relativeRequire("model", ...)
local m_models = model.GetModels()
local configControl
local modelBuilt = false

---
-- Some components expect a lualiblogger compatible logger instance.
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function make_compatible_logger ()
	return {
		debug = LogDebug,
		info = LogInfo,
		warn = LogWarning,
		error = LogError,
		fatal = LogFatal
	}
end

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		--LogDebug("1---------> CheckStatus")
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()

		--LogDebug("2---------> CheckStatus --> hasChanged: '%s', app_status.actualStatus: '%s'", hasChanged, app_status.actualStatus)

		if not hasChanged and app_status.actualStatus ~= "invalid" then
		--LogDebug("3---------> CheckStatus")
			return app_status.actualStatus
		end

		--LogDebug("4---------> CheckStatus")

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"agent", "redis", "cassandra"})

		app_status.current = newConfig

		if not inconcert.environment.apps.helpdesk then
			inconcert.environment.apps.helpdesk = {}
		end

		--Configure Redis
		inconcert.environment.apps.helpdesk.redis = inconcert.environment.apps.helpdesk.redis or {}
		inconcert.environment.apps.helpdesk.redis.host = newConfig["redis"]["host"]
		inconcert.environment.apps.helpdesk.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.apps.helpdesk.redis.pwd = newConfig["redis"]["pwd"] or ""

		--Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
			end
		end

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end

		app_status.actualStatus = "ready"
		return "ready"
	end
}

-- Returns a function bound to a translator for this application
local function makeTextTranslator()
	local translator = inconcert.translators.helpdesk
	return function(...)
		return translator:Text(...)
	end, i18nlib.LCID[translator.language]
end

-- This function returns the title of the application. It needs to use an explicit translator because _T could be bound to a
-- different application.
function GetApplicationTitle()
	local _T = makeTextTranslator()
	return _T("helpdesk")
end

local htmlRender = require"inconcert.render".new("helpdesk", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("helpdesk") end,
	model = m_models,
	makeTextTranslator = makeTextTranslator
})

-- funcion que es llamada cuando el webserver se inicializa.
function Initialize()
	LogInfo("Starting application 'helpdesk'.")
	--OJO que si se cambia en controllers\interactions.lua se debe cambiar aqui tambien!
	configControl = LuaConfigurationControl.Client {
		type = "HELPDESK",
		address = inconcert.environment.apps.agent.config_server.ip,
		port = inconcert.environment.apps.agent.config_server.port,
		node_address = inconcert.environment.apps.agent.config_server.myIp,
		node_port = inconcert.environment.apps.agent.config_server.myPort,
	}
	configControl:AddServiceDependency("Agent")
	configControl:TriggerNodeReady()
	return true
end

--
-- The user has been authenticated and has permissions for this application. Should we let him enter anyway? (returns false if not)
function CanInitializeSession(user)
	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function InitializeSession (authUser) end

--
-- This function gets called when the user logs out. This gets called reliably. The Logout function below may not
-- be called if the user logs off first (apps/login/logout)
function FinalizeSession()
	if app_status:CheckStatus() ~= "ready" then
		LogWarning("App status is not ready. Can not logout agent %s on session %s", inconcert.authUser.userName, inconcert.authUser.authToken)
		return false
	end
	LogInfo("apps.helpdesk: Detail of the logged out user:  userName: (%s) - userGroup: (%s) - virtualContactCenter: (%s)",
		inconcert.authUser.userName, inconcert.authUser.userGroup, inconcert.authUser.virtualContactCenter)
	local sessionId = "HELPDESKSESSION:" .. inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter
	SharedSessions.Discard(sessionId)
	return true
end

function CheckStatus ()
	if not configControl then
		Initialize()
	end
	app_status:CheckStatus()
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetAppLocation()
	return m_appLocation
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetApplicationName()
	return "helpdesk"
end

--///////////////////////////////////////////////////////////////////////////////////
--
function serve_file(app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, m_appLocation.."/"..filename)
	end
end

--///////////////////////////////////////////////////////////////////////////////////
--

-----------------------
--  (cotrolles) urls/apis
local function GetControllers ()
	local controllerDir = m_appLocation .. "/controllers"
	local controllers = {}
	for file in lfs.dir(controllerDir) do
		if file ~= "." and file ~= ".." then
			local f = controllerDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local controllerName = string.match(file, [[(.+).lua]])
				if controllerName then
					controllers[controllerName] = { url = controllerName }
				end
			end
		end
	end
	return controllers
end

for key, controllerDef in pairs(GetControllers()) do
	--requires y mapeo funciones
	local controller = relativeRequire("controllers." .. key, ...)
	if type(controller.MapUris) == "function" then
		controller.MapUris(_M, htmlRender, "/" .. controllerDef.url, "/apps/helpdesk")
	end
end
local controller_perm = require "shared.permissions.controllers.permissions"
controller_perm.MapUris(_M, htmlRender, "/permissions")

--scripts and pages

--paginas
local function render_pages(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:RenderPageLP("pages/" .. name .. ".html", web, env)
end

--scripts
function render_script(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:ScriptLp("scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")

--
--  styles and resources
--  NOTA: ESTO SE PROCESA EN XAVANTE
_M:dispatch_get(_M:serve_file(), "/images/.+", "/styles/.+", "/css/.+", ".+%.mp3")
