local json = require "cjson"
require "socket.url"
local m_request = require "apps.helpdesk.request"
local utils = require "shared.apps.utils"
local m_model = require "apps.helpdesk.model"

module(..., package.seeall)

local function __MakeVccJWTTokenRedisKey (vcc)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local key = string.format(
		"Helpdesk:VCC:%s:JWTToken",
		vcc
	)
	--LogDebug("[__MakeVccJWTToken]------------> key: '%s'", key)
	return key
end

local function __SetVccJWTToken (encodedToken, vccId)
	local key = __MakeVccJWTTokenRedisKey(vccId)
	local redis = m_model.GetDatasource().redis
	redis:set(key, encodedToken)
	-- La key expira a los 5 minutos
	local MINUTE_IN_SEC = 300
	redis:expire(key, MINUTE_IN_SEC)
	--LogDebug("[HELPDESK.LOGIN:_SetVccJWTToken]---SET-token--------> key: '%s' = '%s'.", key, encodedToken)
end

local function __GetVccJWTToken (vccId)
	local key = __MakeVccJWTTokenRedisKey(vccId)
	local redis = m_model.GetDatasource().redis
	local encodedToken = redis:get(key)
	--LogDebug("[HELPDESK.LOGIN:__GetVccJWTToken]---GET-token--------> key: '%s' = '%s'.", key, encodedToken)
	return encodedToken
end

function EraseVccJWTToken (vccId)
	local key = __MakeVccJWTTokenRedisKey(vccId)
	local redis = m_model.GetDatasource().redis
	--LogDebug("[HELPDESK.LOGIN:EraseVccJWTToken]---DELETING-token--------> key: '%s'.", key)
	redis:del(key)
end

local function GetDefualtRequestTimeoutSec ()
	return {timeoutSec = 5}
end

--[[
	Ejemplos de la URL para obtener el User Token

	https://13.82.93.183:2443/ichelpdeskdesarrollo/awsintegrationocc.aspx
	https://168.62.161.232:2443/ichelpdeskdesarrollo/awsintegrationocc.aspx"

	https://{IP}:{PORT}/ichelpdeskdesarrollo/awsintegrationocc.aspx

	Cuando retorna el Token de forma exitosa, el mismo llega en un array con el siguiente formato:
		{"key=E3B4B9AA3EC1475739F33198F01FCBEE"}
	Es una tabla con un string adentro, por lo que se debe de eliminar el atributo name y nos quedamos con el GUID del Token.
]]--

local m_tokenURI = "awsintegrationocc.aspx"

function GetUserToken (fullUserId, config)
	assert(type(fullUserId) == "string")
	assert(type(config) == "table")

	LogInfo("[HELPDESK.LOGIN:GetUserToken] Retrieving User '%s' Token. config: '%s'", fullUserId, json.encode(config))

	local params = string.format("agent=%s", socket.url.escape(fullUserId))
--	LogDebug("params: '%s'", params)
	local contentType = "application/x-www-form-urlencoded"
	local headers = m_request.GetHeaderBasicAuthorization()
	-- Si en 5 segundos NO realizó la obtención del TOKEN, me voy por timeout. Esto para NO quedar colgados esperando.
	local options = GetDefualtRequestTimeoutSec()
	local success, response = m_request.Post(m_tokenURI, params, contentType, config, headers, options)
	--LogDebug("success: '%s'\nresponse: '%s'", success, json.encode(response))

	if not success then
		local err = response
		LogError("[HELPDESK.LOGIN:GetUserToken] Error Retrieving User '%s' Token. Response Err '%s'.", fullUserId, err)
		return false, err
	end

	-- Acá eliminamos del array el nombre del atributo y nos quedamos con el Token
	local key = response
	local token = key:gsub("key=", "")
	if token:len() == 0 then
		local err = key
		LogError("[HELPDESK.LOGIN:GetUserToken] Error Retrieving User '%s' Token. Response Err '%s'.", fullUserId, err)
		return false, err
	end

	LogInfo("[HELPDESK.LOGIN:GetUserToken] Retrieved User '%s' Token '%s'.", fullUserId, token)

	return true, token
end

---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------

local m_tokenUriTemplate = "oauth/access_token"
local m_tokenPostDataTemplate = "client_id={CLIENT_ID}&granttype={GRANTTYPE}&scope={SCOPE}&username={USER_NAME}&password={PASSWORD}"

--[[
	JWT User Token
	--------------
		Send:
			URL         : https://168.62.161.232:2443/ichelpdeskdesarrollo/oauth/access_token
			Content-Type: application/x-www-form-urlencoded
			Post data   : client_id=51507d0c6a714f54bb23375c5eb4cabc&granttype=password&scope=FullControl&username=admin&password=Inc0nc3r710
		Response:
			{ "access_token": "3624627e-1bb6-4902-9549-ce09954fa09b!LTyJgOd9GGfe7vlv7m7FqLt2uEdw65iWvziqMHhpKfPP1Ps2A6GYBSd4fYLhoX1SY8737FphKQYR6q", "scope": "FullControl", "refresh_token": "", "user_guid": "5c942780-9a27-4de8-88e7-9e97dcd176c6 " }
]]--
function GetJWTUserToken (config, noGetFromCache, vccId)
	assert(type(config) == "table")
	assert(type(config.UserAuthentication) == "table")

	LogInfo("[HELPDESK.LOGIN:GetJWTUserToken] Retrieving JWT User Token. config: '%s', noGetFromCache: '%s'.", json.encode(config), noGetFromCache)

	if noGetFromCache then
		EraseVccJWTToken(vccId)
	else
		-- EraseVccJWTToken(vccId) -- A esto lo dejo por si en algún momento de desarrollo quiero pedir siempre el TOKEN contra HELPDESK
		local cachedEncodedJWTToken = __GetVccJWTToken(vccId)
		if cachedEncodedJWTToken then
		-- Lo retorno desde el CACHE, así no realizo siempre el mismo request para obtener el token
			LogInfo("[HELPDESK.LOGIN:GetJWTUserToken] Retrieved from CACHE JWT User Token '%s'.", cachedEncodedJWTToken)
			cachedJWTToken = json.decode(cachedEncodedJWTToken)
			--cachedJWTToken.access_token = cachedJWTToken.access_token .. "____EMMA" -- para forzar un error en token "code error 112"
			return true, cachedJWTToken
		end
	end

	-- El post Info, donde paso USER Y PWD.
	--
	local postData = m_tokenPostDataTemplate
	local function Replace (key, value, defaultValue)
		value = (value or defaultValue)
		assert(value)
		postData = postData:gsub(key, value)
	end

	Replace("{CLIENT_ID}", config.UserAuthentication.ClientId , "51507d0c6a714f54bb23375c5eb4cabc") -- 1 - {CLIENT_ID}
	Replace("{GRANTTYPE}", config.UserAuthentication.GrantType, "password"                        ) -- 2 - {GRANTTYPE}
	Replace("{SCOPE}"    , config.UserAuthentication.Scope    , "FullControl"                     ) -- 3 - {SCOPE}
	Replace("{USER_NAME}", config.UserAuthentication.User     , nil                               ) -- 4 - {USER_NAME}
	Replace("{PASSWORD}" , config.UserAuthentication.Password , nil                               ) -- 5 - {PASSWORD}

	--LogDebug("[HELPDESK.LOGIN:GetJWTUserToken] -->postData: '%s'", postData)

	-- El CONTENT TYPE
	--
	local contentType = "application/x-www-form-urlencoded"

	-- Realizo el REQUEST
	--
	-- Si en 5 segundos NO realizó la obtención del TOKEN, me voy por timeout. Esto para NO quedar colgados esperando.
	local options = GetDefualtRequestTimeoutSec()
	local success, encodedToken = m_request.Post(m_tokenUriTemplate, postData, contentType, config, nil, options)
	--LogDebug("success: '%s'\ntoken: '%s'", success, json.encode(encodedToken))

	if not success then
		EraseVccJWTToken(vccId)
		local err = encodedToken
		LogError("[HELPDESK.LOGIN:GetJWTUserToken] Error Retrieving JWT User Token '%s'.", err)
		return false, err
	end

	local jwtToken = type(encodedToken) == "string" and json.decode(encodedToken) or encodedToken
	LogInfo("[HELPDESK.LOGIN:GetJWTUserToken] Retrieved JWT User Token '%s'.", encodedToken)

	__SetVccJWTToken(encodedToken, vccId)

	return true, jwtToken
end

---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------

local m_logoutURI = "awslogout.aspx?{USER_ID}"
--
--[[
	Logout
	--------------
		Send:
			URL         : https://168.62.161.232:2443/ichelpdeskdesarrollo/awslogout.aspx?agentee1@borrame23
			Content-Type: application/x-www-form-urlencoded

		Response:

]]--

function Logout (fullUserId, keyspace)
	assert(type(fullUserId) == "string")

	LogInfo("[HELPDESK.LOGIN:Logout] HelpdDesk Logout User '%s'.", fullUserId)

	local _, vcc = utils.SeparateVCC(fullUserId)
	if not vcc then
		return false
	end
	local enabled, vccHelpDeskConfig = utils.HasHelpDeskIntegration(vcc, keyspace)
	if not enabled then
		LogInfo("[HELPDESK.LOGIN.Logout] VCC '%s' has not configured HelpdDesk Integrations.", vcc)
		return false , {result = "VCC WITHOUT ALLEGRO INTEGRAION"}
	end

	local headers = {
		["Content-Type"] = "application/x-www-form-urlencoded"
	}
	--[[
		Ahora a la URL le reemplazo los valores que correspondan, {USER_ID}
	]]--
	-- 1 - {USER_ID}
	local suffixURL = m_logoutURI:gsub("{USER_ID}", fullUserId)
	-- Si en 5 segundos NO realizo el Logout, me voy por timeout. Esto para NO quedar colgados esperando.
	local options = GetDefualtRequestTimeoutSec()
	local success, response = m_request.Get(suffixURL, vccHelpDeskConfig, headers, options)

	if not success then
		local err = response
		LogError("[HELPDESK.LOGIN:Logout] HelpdDesk Error logout User '%s'. Response Err '%s'.", fullUserId, err)
		return false, err
	end

	LogInfo("[HELPDESK.LOGIN:Logout] HelpdDesk Succes Logout User '%s'.", fullUserId)

	return true, {result = "OK"}
end
