	
local json = require "cjson"
local utils = require "shared.apps.utils"
local cosmo  = require "cosmo"
local m_moduleLocation = select(2, ...)

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local function RequireController (controllerName)
	assert(controllerName) -- integrations
	local controllerModule = require("apps.integrations.controllers." .. controllerName)
	return controllerModule
end

local function Render (web, htmlFileName, params)
	assert(htmlFileName)
	if web then
		_M.response.DontCache(web)
	end

	LogInfo("[INTEGRATIONS.api:Render] ---> Rendering form  %s", htmlFileName)

	if not m_htmlRender and params then
		m_htmlRender = params.htmlRender
	end
	assert(m_htmlRender)
	m_htmlRender.web = m_htmlRender.web or {}
	m_htmlRender.env = m_htmlRender.env or {_T = m_htmlRender.extension.makeTextTranslator()}

	local fullFileName = string.format("/usr/lib/inconcert/agent/inconcert%s", htmlFileName)
	local htmlContent = m_htmlRender:IncludeLP(fullFileName, {})
	if not htmlContent then
		assert(htmlContent, "[INTEGRATIONS.api:Render] HTML does not exist: %s", htmlFileName)
	end

	--LogDebug("[INTEGRATIONS.api:Render] 4----------------------> Render form htmlContent: %s", htmlContent)

	return htmlContent
end

local function IsAllowedWidgetInVcc (widgetId)
	local integrations = RequireController("integrations")
	return integrations.IsAllowedWidgetInVcc(widgetId)
end

local function IsAllowedWidgetInCampaign (widgetId, m_interactionId)
	local integrations = RequireController("integrations")
	return integrations.IsAllowedWidgetInCampaign(widgetId, m_interactionId)
end

local function HasPermissionToInteractionApp (campaignId, keyspace)
	local integrations = RequireController("integrations")
	return integrations.HasPermissionToInteractionApp(campaignId, keyspace)
end

local m_definitionsModules = {}
local function GetWidgetDefinition (params, widgetId)
	assert(type(widgetId) == "string")
	widgetId = widgetId:lower()

	local moduleName = string.format("apps.integrations.widgets.%s.definitions", widgetId)

	--LogDebug("[INTEGRATIONS.api:GetWidgetDefinition] Loading Integration '%s' definition. moduleName: %s", widgetId, moduleName)

	if not m_definitionsModules[widgetId] then
		local success, m = pcall(require, moduleName)
		if not success then
			LogError("[INTEGRATIONS.api:GetWidgetDefinition] Require '%s' error. Widget '%s' module does not exist.", moduleName, widgetId)
			return false
		end
		m_definitionsModules[widgetId] = m
	end

	local definition = m_definitionsModules[widgetId].Get()
	--LogDebug("[INTEGRATIONS.api:GetWidgetDefinition] Integration '%s' definition: %s", widgetId, json.encode(definition or {}))

	if params and params.htmlRender then
		local htmlProvVccFileName = definition.provisioning.html.body.vcc.fileName
		local htmlFileName = string.format("/apps/integrations/pages/widgets/%s/%s", widgetId, htmlProvVccFileName)
		local htmlContent = Render(web, htmlFileName, params)
		htmlContent = htmlContent:gsub("WIDGET_ID", widgetId)
		definition.provisioning.html.body.vcc.content = htmlContent
	end

	return true,  definition
end

local function GetAvailableWidgetDefinition (params, widgetId)
	local success, definition = GetWidgetDefinition(params, widgetId)
	if not success then
		LogWarning("[INTEGRATIONS.api:GetAvailableWidgetDefinition] Unavailable - Widget '%s' does not exist.", widgetId)
		return false, "WIDGET DOES NOT EXIST"
	end

	local INCONCERT_ENV_MODE = RequireController("integrations").GetEnvironmentVariable()
	local widgetStatus = (definition.status or ""):upper()

	--LogDebug("[INTEGRATIONS.api:GetAvailableWidgetDefinition] ----------> INCONCERT_ENV_MODE: %s, widget: %s, status: %s", INCONCERT_ENV_MODE, widgetId, widgetStatus)

	if INCONCERT_ENV_MODE == "PRODUCTION" then
		if INCONCERT_ENV_MODE ~= widgetStatus then
			LogInfo("[INTEGRATIONS.api:GetAvailableWidgetDefinition] Unavailable - Widget '%s' in DEVELOPMENT state.", widgetId)
			return false, "WIDGET IN DEVELOPMENT"
		end
	end

	definition.widgetId = definition.widgetId or widgetId:lower()
	return true, definition
end

local function GetAvailablesWidgetsDefinitions (params)
	LogInfo("[INTEGRATIONS.api:GetAvailablesWidgetsDefinitions] Retrieving availables Widgets Definitions ...")
	local widgetsDefinitionsBaseDir = m_moduleLocation .. "/widgets"
	local directories = utils.GetDirectories(widgetsDefinitionsBaseDir)
	local definitions = {}
	for _, directory in ipairs(directories) do
		local widgetId = directory
		local success, definition = GetAvailableWidgetDefinition(params, widgetId)
		if success then
			table.insert(definitions, definition)
		end
	end
	--LogDebug("[INTEGRATIONS.api:GetAvailablesWidgetsDefinitions] Loaded availables Widgets Definitions ... %s", json.encode(definitions))
	return definitions
end

local function GetMainHeadCss (self, widgetId)
	assert(widgetId)
	local success, definition = GetAvailableWidgetDefinition(nil, widgetId)
	if not success then
		return ""
	end
	return definition.html.head.css
end

local function GetMainHeadJs (self, widgetId)
	assert(widgetId)
	local success, definition = GetAvailableWidgetDefinition(nil, widgetId)
	if not success then
		return ""
	end
	return definition.html.head.js
end

local function GetMainBodyHtml (self, web, widgetId, m_interactionId, urlParameters)
	assert(widgetId)
	assert(m_interactionId)
	assert(type(urlParameters) == "table")
	widgetId = widgetId:lower()

	-- 1. Se valida contra el directorio del Widget en la app Integration
	local success, definition = GetAvailableWidgetDefinition(nil, widgetId)
	if success then
		-- 2. Se valida contra la configuración del VCC en la CF correspondiente.
		local err, result
		success, result = IsAllowedWidgetInVcc(widgetId)
		if not success then
			err = result
			definition = err
		else
			-- 3. Se valida contra la configuración de la Campana en la CF correspondiente, columna Integration.
			success, result = IsAllowedWidgetInCampaign(widgetId, m_interactionId)
			if not success then
				err = result
				definition = err
			else
				local campaignId = result
				-- 4. Valida si el Usuario tiene permisos sobre la App Interaction
				success, result = HasPermissionToInteractionApp(campaignId)
				if not success then
					err = result
					definition = err
				end
			end
		end
	end
	if not success then
		local _T, _L = utils.MakeTextTranslator("integrations")
		local errType = definition
		local errorsMsg = {
			["WIDGET DOES NOT EXIST"]                      = _T("Widget does not exist"),
			["WIDGET IN DEVELOPMENT"]                      = _T("Widget in development"),
			["WIDGET NOT ALLOWED FOR VCC"]                 = _T("The Widget not allowed for the Vcc"),
			["WIDGET NOT ALLOWED FOR CAMPAIGN"]            = _T("The Widget not allowed for the Campaign"),
			["USER HAS NOT PERMISSION TO INTERACTION APP"] = _T("The user has not permission to Interaction App")
		}
		local errMsg = errorsMsg[errType] or errType
		local htmlError = [[
<div>
	<label>$title</label>
	<br>
	<label>$error</label>
</div>
]]
		htmlError = cosmo.fill(htmlError, {title = _T("Unavailable Widget"), error = errMsg})
		return htmlError
	end
	local htmlFileName = string.format("%s/pages/widgets/%s/mainBody.html", m_appLocation, widgetId)
	local htmlContent = Render(web, htmlFileName)
	return htmlContent
end

----------------------------------
-- Mapeo de las URLS
----------------------------------
local function MapUris (m, htmlRender, prefix, appLocation)
	--LogDebug("[INTEGRATIONS.api:MapUris] Mapping API URIs: m: %s, htmlRender: %s, prefix: %s, appLocation: %s", m, htmlRender, prefix, appLocation)
	m_htmlRender = htmlRender
	m_appLocation = appLocation
end

return {
	MapUris                         = MapUris,
	GetWidgetDefinition             = GetWidgetDefinition,
	GetAvailableWidgetDefinition    = GetAvailableWidgetDefinition,
	GetAvailablesWidgetsDefinitions = GetAvailablesWidgetsDefinitions,
	main = {
		GetHeadCss  = GetMainHeadCss,
		GetHeadJs   = GetMainHeadJs,
		GetBodyHtml = GetMainBodyHtml
	}
}
