local ir = require "lua_iragentlink"
local utils = require "shared.apps.utils"

module(..., package.seeall, inconcert.orbit.application.Decorate)

local pageFlows = {
	FACEBOOK = true,
	INSTAGRAM = true,
}

local accountFlows = {
	SMS = true,
	WHATSAPP = true,
	MAIL = true,
	TWITTER = true
}

local function HasToCloseInactiveAccountPageProcesses(redis, campaignId, accountId, page)
	local key = string.format("ProcessFlowInteractionsByCampaignAccountPage:%s:%s:%s", campaignId, accountId, page)
	return redis:exists(key)
end

local function CloseInactiveAccountPageProcesses(redis, irLink, account, page)
	local accountId = account.accountId
	local campaignId = account.campaignId
	local proxyType = account.data.proxyType
	if not HasToCloseInactiveAccountPageProcesses(redis, campaignId, accountId, page) then return end

	--LogDebug("CloseInactiveAccountPageProcesses: processing account %s page %s in campaign %s",
	--		 accountId, page, campaignId)
	local result = irLink:CloseInactiveProcessFlowInteractions({
		proxyType = proxyType,
		campaignAtVCC = campaignId,
		account = accountId,
		page = page,
	})
	if result.result == "OK" then
		LogDebug("CloseInactiveAccountPageProcesses: (END) processing account %s page %s in campaign %s",
				 accountId, page, campaignId)
	else
		LogError("CloseInactiveAccountPageProcesses: Error processing account %s page %s in campaign %s - %s",
				 accountId, page, campaignId, result.reason)
	end
end

local function CloseInactiveAccountPagesProcesses(redis, irLink, account)
	local pages = account.data and account.data.pagesMessenger or {}
	if type(pages) == "string" then pages = json.decode(pages) end
	for page, _ in pairs(pages) do
		CloseInactiveAccountPageProcesses(redis, irLink, account, page)
	end
end

local function HasToCloseInactiveAccountProcesses(redis, campaignId, accountId)
	local key = string.format("ProcessFlowInteractionsByCampaignAccount:%s:%s", campaignId, accountId)
	return redis:exists(key)
end

local function CloseInactiveAccountProcesses(redis, irLink, account)
	local accountId = account.accountId
	local campaignId = account.campaignId
	local proxyType = account.data.proxyType
	if pageFlows[proxyType] then return CloseInactiveAccountPagesProcesses(redis, irLink, account) end
	if not accountFlows[proxyType] then return end
	if not HasToCloseInactiveAccountProcesses(redis, campaignId, accountId) then return end

	--LogDebug("CloseInactiveAccountProcesses: processing account %s in campaign %s", accountId, campaignId)
	local result = irLink:CloseInactiveProcessFlowInteractions({
		proxyType = proxyType,
		campaignAtVCC = campaignId,
		account = accountId,
	})
	if result.result == "OK" then
		LogDebug("CloseInactiveAccountProcesses: (END) processing account %s in campaign %s", accountId, campaignId)
	else
		LogError("CloseInactiveAccountProcesses: Error processing account %s in campaign %s - %s",
				 accountId, campaignId, result.reason)
	end
end

local function GetAllAccounts()
	local cassandra = inconcert.environment.apps.jobs.cassandra
	local keyspace = cassandra:CassandraConnection("Agent")
	local accountsByCampaign = utils.GetAllColumns({
		keyspace = keyspace,
		superColumns = true,
		columnFamily = "CampaignAccounts",
		startKey = "",
		endKey = ""
	}) or {}
	keyspace:Close()

	local accounts = {}
	for campaignId, campaignAccounts in pairs(accountsByCampaign) do
		for accountId, account in pairs(campaignAccounts) do
			table.insert(accounts, {
				campaignId = campaignId,
				accountId = accountId,
				data = account
			})
		end
	end
	return accounts
end

local function CloseInactiveProcesses(web)
	--LogInfo("[JOB.PROCESS_FLOWS:CloseInactiveProcesses] Running task")
	local redis = inconcert.environment.apps.jobs.redis
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	if not utils.CanRunJob(redis, "CloseInactiveProcesses", 60) then
		LogInfo("[JOB.PROCESS_FLOWS:CloseInactiveProcesses] (END) Process was already running.")
		return
	end

	local accounts = GetAllAccounts()
	for _, account in pairs(accounts) do
		CloseInactiveAccountProcesses(redis, irLink, account)
	end
	--LogInfo("[JOB.PROCESS_FLOWS:CloseInactiveProcesses] (END) Running task")
end

local function FireProcessTimeoutEvent(irLink, entry)
	local proxyType, interactionID = entry:match("([^:]+):(.+)")
	LogInfo("FireProcessTimeoutEvent type %s interaction %s", proxyType, interactionID)
	irLink:FireProcessTimeoutEvent({
		proxyType = proxyType,
		interactionID = interactionID
	})
end

local function FireProcessTimeoutEvents(web)
	local redis = inconcert.environment.apps.jobs.redis
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)

	if not utils.CanRunRealtimeJob(redis, "FireProcessTimeoutEvents", 15) then
		return
	end
	
	local entries = redis:zrangebyscore("InstanceProcessTimeoutEvents", "-inf", os.time())
	for _, entry in pairs(entries) do
		FireProcessTimeoutEvent(irLink, entry)
		redis:zrem("InstanceProcessTimeoutEvents", entry)
	end

	utils.EndRealtimeJob(redis, "FireProcessTimeoutEvents")
end

function MapUris(module, prefix)
	module:dispatch_post(CloseInactiveProcesses, prefix.."/close_inactive_processes/(.*)")
	module:dispatch_get(CloseInactiveProcesses, prefix.."/close_inactive_processes/(.*)")
	module:dispatch_post(FireProcessTimeoutEvents, prefix.."/fire_process_timeout_events/(.*)")
	module:dispatch_get(FireProcessTimeoutEvents, prefix.."/fire_process_timeout_events/(.*)")
end