-- Twitter OAuth controller

require "inconcert.orbit.application"

local cosmo = require "cosmo"
local json = require "json"
local uuid = require "uuid"
local wsapi_util = require "wsapi.util"
local OAuth = require "OAuth"
local utils = require "shared.apps.utils"
local ssl = ssl
local Synchronize = require "Synchronize"
local http_request = require "shared.http.request"
local https = Synchronize("luasec_mutex", function()
	return require ("ssl.https")
end)


module(..., package.seeall, inconcert.orbit.application.Decorate)

local m_htmlRender = nil
local m_config
local m_redis_client

function SetConfig (redis_client, conf)
	m_config = conf
	m_redis_client = redis_client
end

---
-- Returns a new instance of a Twitter OAuth enabled client
--
local function MakeTwitterClient(oauth_token, oauth_verifier)
	local twitter = OAuth.new(m_config.consumerKey, m_config.consumerSecret, {
		RequestToken = "https://api.twitter.com/oauth/request_token",
		AuthorizeUser = { "https://api.twitter.com/oauth/authorize", method = "GET" },
		AccessToken = "https://api.twitter.com/oauth/access_token"
	}, { OAuthToken = oauth_token, OAuthVerifier = oauth_verifier })
	return twitter
end

---
-- Given a 'authorization_code' (granted when the user gave permissions to this application), obtain a valid access token
--
local function RenewAccessToken(authorization_code)
	LogError("[oauth] - Renew acess token was called but no implementation here")
end


local function RegisterAuthWebhook(webhook_host, subdomain_host, tracking_code)

	local url = cosmo.fill([[https://$webhook_proxy_host/registeroauth?callback_env=$subdomain_host&state=$state]], {
    	webhook_proxy_host = webhook_host,
        subdomain_host = subdomain_host,
        state = tracking_code
	})

    response, code, headers, status = https.request(url)
    if code ~= 200 then
    	LogError("[oauth] - Error in https request RegisterAuthWebhook: %s , code: %s", response,code )
        return false
	end
	return true
end

---
-- Shows the authorization form for a given oauth provider
--
local function RenderAuthorizationForm(web)
	if m_config == nil then
		web.status = 500
	    LogError("[oauth] - Error in RenderAuthorizationForm: Must configure application oauth in apps_config.lua")
	    return "Must configure application oauth in apps_config.lua"
	end
	
	local twitter = MakeTwitterClient()
	local postData = json.decode(web.input.post_data)
	local webhook_host = m_config.webhook_host

	local tracking_code = uuid.new()
	local callback_url = "https://" .. webhook_host .. "/twitter/auth_redirect?state=" .. wsapi_util.url_encode(tracking_code)

	if not RegisterAuthWebhook(webhook_host, postData.host, tracking_code) then
		web.status = 500
		LogError("[oauth] - Error register twitter token, code %s body %s headers %s", err, body, json.encode(headers))
		return "Error requesting twitter token"
	end

	local values, err, headers, _, body = twitter:RequestToken({ oauth_callback = callback_url })	
	if not values then
		web.status = 500
		LogError("[oauth] - Error requesting twitter token, code %s body %s headers %s", err, body, json.encode(headers))
		return "Error requesting twitter token"
	end

	m_redis_client:set("OAuthTokens:"..tracking_code, json.encode({
		token_secret = values.oauth_token_secret,
		callback_url = callback_url,
		post_data = web.input.post_data,
		SID = inconcert.authUser and inconcert.authUser.authToken or nil
	}))

	local new_url = twitter:BuildAuthorizationUrl({ oauth_callback = callback_url, state = tracking_code })
	web.status = 302
	web.headers["Location"] = new_url
	return ""
end

---
-- Shows the page that the oauth provider redirects to when the user gaves permission
--
local function RenderFormAuthRedirect(web)
	local env = { _T = m_htmlRender.extension.makeTextTranslator() }
	local tracking_code = web.input.state
	if not tracking_code then
		LogError("[oauth] - Can not obtain tracking code from postback")
		return m_htmlRender:RenderPageLP("pages/twitterAccountError.html", web, env)
	end

	local stored_code = m_redis_client:get("OAuthTokens:"..tracking_code)
	if not stored_code then
		LogError("[oauth] - Can not obtain twitter tracking code %s from redis", tracking_code)
		return m_htmlRender:RenderPageLP("pages/twitterAccountError.html", web, env)
	end

	local decoded_transaction = json.decode(stored_code)
	local data = json.decode(decoded_transaction.post_data)
	if web.input.denied then
		-- Twitter a diferencia de Facebook no me llama al callback si el tipo no me autoriza la aplicacion (el maldito MAY del spec)
		data.result = false
		data.error_description = "auth rejected by the user"
		m_redis_client:del("OAuthTokens:" .. tracking_code)
		return m_htmlRender:RenderPageLP("pages/twitterAccountError.html", web, env)
	end

	local oauth_token = web.input.oauth_token
	-- El twitter trolo no manda oauth_verifier aunque el RFC dice que es un MUST mandarlo (solo lo manda si paso un oauth_callback explicito (si uso el implicito no manda nada))
	local oauth_verifier = web.input.oauth_verifier
	local twitter = MakeTwitterClient(oauth_token, oauth_verifier)
	twitter:SetTokenSecret(decoded_transaction.token_secret)

	local values, code, headers, status, body = twitter:GetAccessToken()
	if not values then
		LogError("[oauth] - Unable to create twitter client for token %s verifier %s http status:%s headers:%s status line:%s body:%s", oauth_token, oauth_verifier, code, json.encode(headers), status, body)
		data.result = false
		data.error_description = "failure creating oauth twitter client"
		m_redis_client:del("OAuthTokens:" .. tracking_code)
		return m_htmlRender:RenderPageLP("pages/twitterAccountError.html", web, env)
	end

	m_redis_client:del("OAuthTokens:" .. tracking_code)

	return m_htmlRender:RenderPageLP("pages/twitterAccountOk.lp", web, {
		oauth_token = values.oauth_token,
		oauth_token_secret = values.oauth_token_secret,
		oauth_verifier = oauth_verifier,
		user_id = values.user_id,
		screen_name = values.screen_name,
		origin = data.origin,
		consumerKey = m_config.consumerKey,
		consumerSecret = m_config.consumerSecret,
		_T = env._T
	})
end

--
-- Given a 'authorization_code' (granted when the user gave permissions to this application), obtain a valid access token
function HandleRenewAccessToken(web)
	local code = web.input.authorization_code
	local access_token, values = RenewAccessToken(code)
	
	web:content_type("application", "json")
	return json.encode(values)
end

function MapUris(module, prefix)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	module:dispatch_get(RenderAuthorizationForm, prefix.."/authorize_user/(.*)")
	module:dispatch_get(RenderFormAuthRedirect, prefix.."/auth_redirect")
	module:dispatch_get(HandleRenewAccessToken, prefix.."/renew_access_token")
end

function SetViewEnvironment(env)
	assert(env and env.htmlRender)
	m_htmlRender = env.htmlRender
end
