-- WhatsApp OAuth controller

require "inconcert.orbit.application"
require "AsyncTasks"

local http = require "socket.http"
local ssl = ssl
local Synchronize = require "Synchronize"
local https = Synchronize("luasec_mutex", function()
	return require ("ssl.https")
end)
local cosmo = require "cosmo"
local json = require "json"
local wsapi_util = require "wsapi.util"

local campaigns = require("apps.agent.models.campaigns")

module(..., package.seeall, inconcert.orbit.application.Decorate)

local m_htmlRender = nil
local m_config = nil
local m_redis_client

local DEFAULT_SCOPE = 
	"business_management," ..					-- manage business actives
	"whatsapp_business_management," ..			-- manage whatsapp actives
	"whatsapp_business_messaging"				-- manage whatsapp messages

local ICC_CLOUD_API_VERSION = "v16.0"

---
-- Given a 'authorization_code' (granted when the user gave permissions to this application)
-- obtain a valid access token
--
-- @tparam string authorization_code
-- @tparam table codes
-- @tparam string codes.client_id
-- @tparam string codes.redirect_uri
-- @tparam string codes.client_secret
--
-- @treturn boolean ok           - If the renew went well or not
-- @treturn string message       - If the renew went well wrong, an error message (optional)
-- @treturn string access_token
-- @treturn table values
--
local function RenewAccessToken(_T, authorization_code, codes)
	local new_url = cosmo.fill([[https://graph.facebook.com/$apiVersion/oauth/access_token?format=json&client_id=$client_id&redirect_uri=$redirect_uri&code=$authorization_code&client_secret=$client_secret]], {
		apiVersion = ICC_CLOUD_API_VERSION,
		client_id = codes.client_id,
		redirect_uri = codes.redirect_uri,
		authorization_code = authorization_code,
		client_secret = codes.client_secret
	})
	local response, code, headers, status = https.request(new_url)
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(graph.facebook.com/".. ICC_CLOUD_API_VERSION .. "/oauth/access_token): " .. (response or "")
		LogError("[oauth] - %s", message )
		return false, message
	end
	
	local values = json.decode(response)
	
	new_url = cosmo.fill([[https://graph.facebook.com/$apiVersion/me?access_token=$access_token]], {apiVersion = ICC_CLOUD_API_VERSION, access_token = values.access_token})
	response, code, headers, status = https.request(new_url)
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(graph.facebook.com/" .. ICC_CLOUD_API_VERSION .. "/me): " .. response
		LogError("[oauth] - %s", message )
		return false, message
	end
	
	local userData = json.decode(response)
	values.uid = userData.id
	values.name = userData.name
	return true, "", values.access_token, values
end

local function GetWabas(_T, input_token, system_user_token)
	local new_url = cosmo.fill([[https://graph.facebook.com/$apiVersion/debug_token?input_token=$input_token&access_token=$access_token]], {
		apiVersion = ICC_CLOUD_API_VERSION,
		input_token = input_token,
		access_token = system_user_token
	})
	local response, code, headers, status = https.request(new_url)
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(graph.facebook.com/" .. ICC_CLOUD_API_VERSION .. "/debug_token: " .. (response or "")
		LogError("[oauth] - %s", message )
		return false, message
	end
	
	LogInfo("[oauth] - GetWabas - debug_token response: %s", response)

	local values = json.decode(response)
	local wabas = {}

	if values.data and values.data.granular_scopes then
		for _, scope in pairs(values.data.granular_scopes) do
			if scope.scope == "whatsapp_business_management" and scope.target_ids then
				for _, wabaId in pairs(scope.target_ids) do
					new_url = cosmo.fill([[https://graph.facebook.com/$apiVersion/$wabaId?fields=id,name,message_template_namespace&access_token=$input_token]], {
							apiVersion = ICC_CLOUD_API_VERSION,
							wabaId = wabaId,
							input_token = input_token
						})
					response, code, headers, status = https.request(new_url)
					if code ~= 200 then
						local message = _T("Error in https request ") .. "(graph.facebook.com/" .. ICC_CLOUD_API_VERSION .. "/{wabaId}: " .. (response or "")
						LogError("[oauth] - %s", message )
						return false, message
					end	
					local waba_values = json.decode(response)
					local waba = {}
					waba.id = waba_values.id
					waba.name = waba_values.name
					waba.message_template_namespace = waba_values.message_template_namespace
					table.insert(wabas, waba)
				end
			end
		end
	end
	return true, "", wabas
end

local function GetPhoneNumbers(_T, access_token, waba_id, data)
	local accountId = data.authorizingId
	local operation = data.operation
	local new_url = cosmo.fill([[https://graph.facebook.com/$apiVersion/$waba_id/phone_numbers?fields=id,status,code_verification_status,display_phone_number,quality_rating,is_pin_enabled,messaging_limit_tier&access_token=$access_token]], {
		apiVersion = ICC_CLOUD_API_VERSION,
		waba_id = waba_id,
		access_token = access_token
	})
	local response, code, headers, status = https.request(new_url)
	if code ~= 200 then
		local message = _T("Error in https request ") .. "(graph.facebook.com/" .. ICC_CLOUD_API_VERSION .. "/" .. waba_id .. "/phone_numbers): " .. (response or "")
		LogError("[oauth] - %s", message )
		return false, message
	end

	local messaging_limit_by_tier = {
		TIER_250 = 250,
		TIER_1K = 1000,
		TIER_10K = 10000,
		TIER_100K = 100000,
		TIER_UNLIMITED = -1
	}

	local values = json.decode(response)
	local phoneNumbers = {}

	-- Todas las cuentas de todas las campana de todos los vccs
	local allAccounts = campaigns.GetAllVCCsAccounts()	

	if values.data then
		for _, phoneObj in pairs(values.data) do
			--El proceso de onboarding obliga a seleccionar un telefono verificado (VERIFIED) o a verificarlo durante el onboarding si no lo esta, por lo que al llegar aqui admitiremos unicamente telefonos que esten verificados y no migrados
			--El proceso de migración obliga a que el telefono este registrado (CONNECTED) y que el factor de verificacion no este activado (is_pin_enabled debe ser false)
			if operation == "fullOnboarding" and phoneObj.code_verification_status == "VERIFIED" and phoneObj.status ~= "MIGRATED" or phoneObj.status == "CONNECTED" then
				local phoneNumber = phoneObj.display_phone_number
				local phoneId = phoneObj.id
				local phoneStatus = phoneObj.status
				local codeVerificationStatus = phoneObj.code_verification_status
				local displayPhoneNumber = phoneNumber
				local is_pin_enabled = phoneObj.is_pin_enabled
				local messaging_limit = messaging_limit_by_tier[phoneObj.messaging_limit_tier] or 250
				local canUsePhoneId = true

				-- Elimino telefonos usados en otras cuentas de whatsapp de este ambiente. 
				-- El uso de un telefono en otro ambiente se controla en el webhook
				if allAccounts then
					-- recorro todas las campanas
					for current_campaign, current_accounts in pairs( allAccounts ) do
						-- recorro todas las accounts
						for full_acc_id, current_account in pairs( current_accounts ) do
							if current_account.state == "1" and current_account.proxyType == "WHATSAPP" and current_account.phoneId == phoneId and full_acc_id ~= accountId then
								--No permito usar un telefono que actualmente se encuentre configurado en otra cuenta de whatsapp
								canUsePhoneId = false
								LogDebug("[oauth] - phone %s (%s) cant be used because is in use in account %s (%s)", phoneNumber, phoneId, current_account.name, current_account.id)
								break
							end
						end
						if not canUsePhoneId then
							break
						end
					end			
				end	

				--Para migrar un telefono este debe tener deshabilitado el two-factor authentication
				canUsePhoneId = canUsePhoneId and ((operation == "migratePhone" and not is_pin_enabled) or operation == "fullOnboarding" or operation == "useRegisteredPhone")

				if canUsePhoneId then
					table.insert(phoneNumbers, { phoneNumber = phoneNumber, phoneId = phoneId, displayPhoneNumber = displayPhoneNumber, phoneStatus = phoneStatus, codeVerificationStatus = codeVerificationStatus, messaging_limit = messaging_limit})
				end
			end
		end
	end
	return true, "", phoneNumbers
end

local function RegisterAuthWebhook(webhook_host, subdomain_host, tracking_code)

	local url = cosmo.fill([[https://$webhook_proxy_host/registeroauth?callback_env=$subdomain_host&state=$state]], {
    	webhook_proxy_host = webhook_host,
        subdomain_host = subdomain_host,
        state = tracking_code
	})

    response, code, headers, status = https.request(url)
    if code ~= 200 then
    	LogError("[oauth] - Error in https request RegisterAuthWebhook: %s , code: %s", response,code )
        return false
	end
	return true
end

--
-- Shows the authorization form for a given oauth provider
local function RenderAuthorizationForm(web)
	local _T = m_htmlRender.extension.makeTextTranslator()

	if m_config == nil then
		local message = "Error in RenderAuthorizationForm: Must configure application oauth in apps_config.lua"
		LogError("[oauth] - %s ", message)
		return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end

	local scope = web.input.scope or DEFAULT_SCOPE
	
	local system_user_token = m_config.system_user_token
	local client_id = m_config.client_id
	local client_secret = m_config.client_secret
	local postData = json.decode(web.input.post_data)
	local webhook_host = m_config.webhook_host
	
	-- We use the tracking code to identify the data to be saved in the session
	local tracking_code = uuid.new()

	if not RegisterAuthWebhook(webhook_host, postData.host, tracking_code) then
		web.status = 500
		LogError("[oauth] - Error register twitter token, code %s body %s headers %s", err, body, json.encode(headers))
		return "Error requesting whatsapp token"
	end
	
	local redirect_uri = "https://" .. webhook_host .. "/whatsapp/auth_redirect"
	
	LogInfo("[oauth] - Requesting token for whatsapp access, postback data is %s", web.input.post_data)
		
	m_redis_client:set("OAuthTokens:"..tracking_code, json.encode({
		system_user_token = system_user_token,
		client_id = client_id,
		client_secret = client_secret,
		redirect_uri = redirect_uri,
		post_data = web.input.post_data,
		SID = inconcert.authUser and inconcert.authUser.authToken or nil
	}))
	
	local params = "display=popup"
	local extras = postData.operation == "fullOnboarding" and "extras=%7B%22feature%22%3A%22whatsapp_embedded_signup%22%7D" or nil
	
	local new_url = cosmo.fill([[https://www.facebook.com/dialog/oauth?client_id=$client_id&redirect_uri=$redirect_uri&state=$state&scope=$scope&$params&$extras]], {
		client_id = client_id,
		redirect_uri = redirect_uri,
		state = tracking_code,
		scope = scope,
		extras = extras,
		params = params
	})

	web.status = 302
	web.headers["Location"] = new_url
	return ""
end

--
-- Shows the page that the oauth provider redirects to when the user gaves permission
local function RenderFormAuthRedirect(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	local tracking_code = web.input.state

	if not tracking_code then
		local message = "Can not obtain whatsapp tracking code from postback"
		LogError("[oauth] - %s", message)
		return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	LogInfo("[oauth] - Got whatsapp postback redirect with tracking_code: '%s'", tracking_code)
	
	local stored_code = m_redis_client:get("OAuthTokens:"..tracking_code)
	if not stored_code then
		local message = "Can not obtain whatsapp tracking code from Redis"
		LogError("[oauth] - %s: %s", message, tracking_code)
		return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end

	local decoded_transaction = json.decode(stored_code)
	local data = json.decode(decoded_transaction.post_data)

	if web.input.error_reason == "user_denied" then
		m_redis_client:del("OAuthTokens:"..tracking_code)
		return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
			messageError = "Auth rejected by the user",
			_T = _T
		})
	end

	local authorization_code = web.input.code
	local ok, message, access_token, values = RenewAccessToken(_T, authorization_code, decoded_transaction)
	if not ok then
		return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
			messageError = message,
			_T = _T
		})
	end
	
	LogInfo("[oauth] - Account: '%s' authorization succeeded.", values.uid)

	local wabas
	local wabasToReturn = {}
	ok, message, wabas = GetWabas(_T, access_token, decoded_transaction.system_user_token)
	if not ok then
		return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
			messageError = message,
			_T = _T
		})	
	end
	LogDebug("[oauth] - WhatsApp Business Accounts for user %s:'%s'", values.uid, json.encode(wabas))
	for _, waba in pairs(wabas) do
		local phoneNumbers
		ok, message, phoneNumbers = GetPhoneNumbers(_T, access_token, waba.id, data)
		if not ok then
			return m_htmlRender:RenderPageLP("pages/whatsappAccountError.lp", web, {
				messageError = message,
				_T = _T
			})
		end
		LogDebug("[oauth] - Account: '%s' Phone numbers for WhatsApp Business Account '%s' (%s):'%s'", values.uid, waba.name, waba.id, json.encode(phoneNumbers))
		waba.phoneNumbers = phoneNumbers
		table.insert(wabasToReturn, waba)
	end

	return m_htmlRender:RenderPageLP("pages/whatsappAccountOk.lp", web, {
		authorization_code = authorization_code,
		access_token = access_token,
		wabas = json.encode(wabasToReturn),
		uid = values.uid,
		name = values.name,
		origin = data.origin,
		_T = _T
	})
end

--
-- Given a 'authorization_code' (granted when the user gave permissions to this application), obtain a valid access token
function HandleRenewAccessToken(web)
	local _T = m_htmlRender.extension.makeTextTranslator()
	local code = web.input.authorization_code
	local ok, access_token, values = RenewAccessToken(_T, code)
	
	web:content_type("application", "json")
	return json.encode(values)
end

function MapUris(module, prefix)
	module:dispatch_get(RenderAuthorizationForm, prefix.."/authorize_user/(.*)")
	module:dispatch_get(RenderFormAuthRedirect, prefix.."/auth_redirect")
	module:dispatch_get(HandleRenewAccessToken, prefix.."/renew_access_token")
end

function SetViewEnvironment(env)
	assert(env and env.htmlRender)
	m_htmlRender = env.htmlRender
	
end

function SetConfig (redis_client, conf)
	m_config = conf
	m_redis_client = redis_client
end
