local utils = require "shared.apps.utils"
local modelCarriers = require("apps.provisioning.models.carriers")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form CARRIER with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_carrier_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING:Render] HTML CARRIER does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function GetCarrier (web, carrierId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.CARRIERS:GetCarrier] Retrieving CARRIER '%s'. User %s.", carrierId, userId)

	local success, carrierInfo = modelCarriers.Get(carrierId)
	if not success then
		LogError("[PROVISIONING.CARRIERS:GetCarrier] Error Retrieving CARRIER %s. Not Found.", carrierId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return carrierInfo
end

local function ValidateAndMakeInfoToUpdate (carrierId, carrierInfo)
	assert(carrierId)
	assert(carrierInfo)

	LogInfo("[PROVISIONING.CARRIERS:ValidateAndMakeInfoToUpdate] Validating CARRIER '%s' info to updated.", carrierId)

	local carrierExist, currentCarrierInfo = modelCarriers.Get(carrierId)

	-- 1 - clean up valores que llegan del lado del cliente
	local isNew = carrierInfo.isNew

	if carrierExist then
		-- Si estoy creando el Carrier y el mismo ya exsite en la BD, fallo!!
		if isNew then
			return false, "MUST_HAVE_UNIQUE_NAME"
		end
	end

	carrierInfo.isNew = nil

	return true, carrierInfo
end

local function SaveCarrier (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.CARRIERS:SaveCarrier] Saving CARRIER. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, carriers = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.CARRIERS:SaveCarrier] Error Updating CARRIER. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, carrierInfo = next(carriers)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> carrierInfo %s", carrierInfo)

	local carrierId = carrierInfo.id
	local success, carrierInfo = ValidateAndMakeInfoToUpdate(carrierId, carrierInfo)
	if not success then
		local errMsg = carrierInfo
		LogError("[PROVISIONING.CARRIERS:SaveCarrier] Error Validating CARRIER '%s'. Err: %s", carrierId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local success, errMsg = modelCarriers.Save(carrierId, carrierInfo)
	if not success then
		LogError("[PROVISIONING.CARRIERS:SaveCarrier] Error Updating CARRIER '%s'. Err: %s", carrierId, errMsg)
		web.status = 500
		result.result = errMsg -- Este código de error llega desde la Rocka "provisioningmanager"
		return result
	end

	LogInfo("[PROVISIONING.CARRIERS:SaveCarrier] '%s' saved successfully.", carrierId)

	web.status = 200
	result.result = "OK"
	return result
end

local function DeleteCarrier (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.CARRIERS:DeleteCarrier] Deleting CARRIER. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, carriers = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.CARRIERS:DeleteCarrier] Error Deleting CARRIER. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local carrierId = carriers[1]
	local success, vccInfo = modelCarriers.Get(carrierId)
	if not success then
		LogError("[PROVISIONING.CARRIERS:DeleteCarrier] Error Deleting CARRIER %s. Not Found.", carrierId)
		web.status = 404
		result.result = "CARRIER DOES NOT EXIST"
		return result
	end

	local response = modelCarriers.Delete(carrierId)
	if not response.success then
		LogError("[PROVISIONING.CARRIERS:DeleteCarrier] Error Deleting CARRIER %s. Error: %s.", carrierId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.CARRIERS:DeleteCarrier] Carrier '%s' deleting successfully.", carrierId)

	web.status = 200
	result.result = "OK"
	return result
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.CARRIERS:Search] 1 - (START) Loading Carriers. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, carriers = modelCarriers.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = carriers
		LogWarning("[PROVISIONING.CARRIERS:Search] Error loading Carriers. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--carriers = utils.FilterEntities(carriers, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(carriers)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.CARRIERS:SEARCH] 2 - Filtering rows")

	for id, carrier in pairs(carriers) do
		local carrierId = utils.GetId(id)
		carrier.Id = carrierId
		carrier.Name = carrier.Name or carrierId
		if utils.MatchFilter(carrier, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, carrier)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.CARRIERS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user carriers
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.CARRIERS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadCarriers (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.CARRIERS.:LoadCarriers] Loading Carriers, user %s, vcc %s", userId, vcc)

	local success, carriers = modelCarriers.ListAll()
	if not success then
		local err = carriers
		LogWarning("[PROVISIONING.CARRIERS:LoadCarriers] Error loading Carriers. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return carriers
end

--[[
]]--
local function LoadyCarriersWithTrunks (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.CARRIERS.:LoadyCarriersWithTrunks] Loading Carriers With Trunks, user %s, vcc %s", userId, vcc)

	local success, trunks = modelCarriers.ListCarriersWithTrunks()
	if not success then
		local err = trunks
		LogWarning("[PROVISIONING.CARRIERS:LoadyCarriersWithTrunks] Error loading Carriers With Trunks. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return trunks
end

----------------------------------
-- Mapas de las URLS
----------------------------------
function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix

	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- CARRIER i6
	module:dispatch_post(fn .. Search, p .. "/search/"                                          )
	module:dispatch_get (replyFn .. LoadyCarriersWithTrunks, p .. "/load/with_trunks/"          )
	module:dispatch_get (HasPermission(GetCarrier   , "view_carriers"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(SaveCarrier  , "save_carrier", fn  ), p .. "/save/"      )
	module:dispatch_post(HasPermission(DeleteCarrier, "delete_carrier", fn), p .. "/delete/(.*)")

	module:dispatch_get (replyFn .. LoadCarriers, p .. "/load/"                                 )

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end