local utils = require "shared.apps.utils"
local modelProcessors = require("apps.provisioning.models.dialplan_processors")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil
local m_helpers = nil

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:Render] Render form DIALPLAN_PROCESSORS with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_dialplan_processor_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING.DIALPLAN_PROCESSORS:Render] HTML DIALPLAN_PROCESSORS does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function ValidateAndMakeInfoToUpdate (id, info)
	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:ValidateAndMakeInfoToUpdate] Validating DIALPLAN_PROCESSORS '%s' info to updated. Info %s", id, info)

	if id then
		local exist, currentInfo = modelProcessors.Get(id)
		-- 1 - clean up valores que llegan del lado del cliente
		local isNew = info.isNew

		if exist then
			-- Si estoy creando la Entidad y el mismo ya exsite en la BD, fallo!!
			if isNew then
				return false, "MUST_HAVE_UNIQUE_NAME"
			end
		end
	end
	-- id
	if not info.Id or info.Id == "" then
		return false, "ID_IS_REQUIRED"
	end
	-- description
	if not info.description or info.description == "" then
		return false, "DESCRIPTION_IS_REQUIRED"
	end
	-- code
	if not info.code or info.code == "" then
		return false, "CODE_IS_REQUIRED"
	end

	info.isNew = nil

	return true, info
end

local function SaveProcessor (web)
	local postedData = web.input["post_data"]

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:SaveProcessor] Saving processor. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, rows = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:SaveProcessor] Error Updating processor. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local rowId, rowInfo = next(rows)
	-- LogDebug("--> rowId %s", rowId)
	-- LogDebug("--> rowInfo %s", rowInfo)

	local success, rowInfo = ValidateAndMakeInfoToUpdate(rowId, rowInfo)
	if not success then
		local errMsg = rowInfo
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:SaveProcessor] Error Validating processor '%s'. Err: %s", rowId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local success, err = modelProcessors.Save(rowId, rowInfo)
	if not success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:SaveProcessor] Error Updating processor '%s'. Err: %s", rowId, err)
		web.status = 500
		result.result = err.code -- Este código de error llega desde la Rocka "provisioningmanager"
		result.rawData = err.rawData
		return result
	end

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:SaveProcessor] '%s' saved successfully.", rowId)

	result.result = "OK"
	return result
end

local function DeleteProcessor (web)
	local postedData = web.input["post_data"]
	local success, processorInfo = pcall(json.decode, postedData)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:DeleteProcessor] Deleting processor. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, processors = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:DeleteProcessor] Error Deleting processor. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local processorId = processors[1]
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")

	local success, vccInfo = modelProcessors.Get(processorId, ksProv)
	if not success then
		ksProv:Close()
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:DeleteProcessor] Error Deleting processor %s. Not Found.", processorId)
		web.status = 404
		result.result = "PROCESSOR DOES NOT EXIST"
		return result
	end

	local response = modelProcessors.Delete(processorId, ksProv)
	if not response.success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:DeleteProcessor] Error Deleting processor %s. Error: %s.", processorId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:DeleteProcessor] Processor '%s' deleting successfully.", processorId)

	ksProv:Close()
	web.status = 200
	result.result = "OK"

	return result
end

local function GetProcessor (web, processorId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:GetProcessor] Retrieving processor '%s'. User %s.", processorId, userId)

	local success, processorInfo = modelProcessors.Get(processorId)
	if not success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:GetProcessor] Error Retrieving processor %s. Not Found.", processorId)
		web.status = 404
		return { result = "NOT FOUND" }
	end

	web.status = 200
	return processorInfo
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:Search] 1 - (START) Loading processors. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, processors = modelProcessors.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = processors
		LogWarning("[PROVISIONING.DIALPLAN_PROCESSORS:Search] Error loading processors. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--processors = utils.FilterEntities(processors, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(processors)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:SEARCH] 2 - Filtering rows")

	for id, processor in pairs(processors) do
		local processorId = utils.GetId(id)
		processor.Id = processorId
		processor.Name = processor.Name or processorId
		if utils.MatchFilter(processor, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, processor)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user processors
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

local function ExportProcessor (web, processorId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:ExportProcessor] Retrieving processor '%s'. User %s.", processorId, userId)

	local success, processorInfo = modelProcessors.Get(processorId)
	if not success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:GetProcessor] Error Retrieving processor %s. Not Found.", processorId)
		web.status = 404
		return "NOT FOUND"
	end

	web.status = 200
	web.headers["Content-Disposition"] = [[attachment; filename="]] .. processorInfo.Id .. '.lua"'
	return processorInfo.code
end

--[[
]]--
local function LoadProcessors (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS.:LoadProcessors] Loading processors, user %s, vcc %s", userId, vcc)

	local success, processors = modelProcessors.ListAll()
	if not success then
		local err = processors
		LogWarning("[PROVISIONING.DIALPLAN_PROCESSORS:LoadProcessors] Error loading processors. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return processors
end

--[[
]]--
local function GetHelpers (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS.:GetHelpers] Loading helpers, user %s, vcc %s", userId, vcc)

	local helpers = require("apps.provisioning.dialplan_processor_helpers")
	local _T = m_htmlRender.extension.makeTextTranslator()
	for _, helper in pairs(helpers) do
		helper.description = _T(helper.description)
		for _, method in pairs(helper.methods) do
			method.description = _T(method.description)
		end
	end
	
	-- OK
	web.status = 200
	return helpers
end

--[[
]]--
local function GetTrace (web, processorId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIALPLAN_PROCESSORS:GetTrace] Retrieving processor '%s' trace. User %s.", processorId, userId)

	local success, trace = modelProcessors.GetTrace(processorId)
	if not success then
		LogError("[PROVISIONING.DIALPLAN_PROCESSORS:GetTrace] Error Retrieving processor %s trace. Not Found.", processorId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return trace
end

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix
	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- Processors i6
	module:dispatch_post(fn .. Search, p .. "/search/")
	module:dispatch_get (HasPermission(GetProcessor   , "view_dialplan_processors"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(DeleteProcessor, "delete_dialplan_processors", fn), p .. "/delete/(.*)")
	module:dispatch_post(HasPermission(SaveProcessor  , "save_dialplan_processors", fn  ), p .. "/save/"      )
	module:dispatch_get (HasPermission(ExportProcessor   , "export_dialplan_processors"     ), p .. "/export/(.+)"   )
	module:dispatch_get (replyFn .. LoadProcessors, p .. "/load/")
	module:dispatch_get (replyFn .. GetHelpers, p .. "/get_helpers/")
	module:dispatch_get (replyFn .. GetTrace, p .. "/get_trace/(.+)")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end
