local utils = require "shared.apps.utils"
local modelDIDs = require("apps.provisioning.models.dids")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form DID with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_did_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING:Render] HTML DID does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function ValidateAndMakeInfoToUpdate (didId, didInfo)
	assert(didId)
	assert(didInfo)

	LogInfo("[PROVISIONING.DIDS:ValidateAndMakeInfoToUpdate] Validating DID '%s' info to updated. Info %s", didId, didInfo)

	local exist, currentDidInfo = modelDIDs.Get(didId)

	-- 1 - clean up valores que llegan del lado del cliente
	local isNew = didInfo.isNew

	if exist then
		-- Si estoy creando el DID y el mismo ya exsite en la BD, fallo!!
		if isNew then
			return false, "MUST_HAVE_UNIQUE_NAME"
		end
	end

	didInfo.isNew = nil

	return true, didInfo
end

local function SaveDID (web)
	local postedData = web.input["post_data"]

	LogInfo("[PROVISIONING.DIDS:SaveDID] Saving DID. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, rows = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.DIDS:SaveDID] Error Updating DID. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local rowId, rowInfo = next(rows)
	--LogDebug("--> rowId %s", rowId)
	--LogDebug("--> rowInfo %s", rowInfo)

	local success, rowInfo = ValidateAndMakeInfoToUpdate(rowId, rowInfo)
	if not success then
		local errMsg = rowInfo
		LogError("[PROVISIONING.DIDS:SaveDID] Error Validating DID '%s'. Err: %s", rowId, errMsg)
		web.status = 500
		result.result = errMsg
		return result
	end

	local success, err = modelDIDs.Save(rowId, rowInfo)
	if not success then
		LogError("[PROVISIONING.DIDS:SaveDID] Error Updating DID '%s'. Err: %s", rowId, err)
		web.status = 500
		result.result = err.code -- Este código de error llega desde la Rocka "provisioningmanager"
		return result
	end

	LogInfo("[PROVISIONING.DIDS:SaveDID] '%s' saved successfully.", rowId)

	web.status = 200
	result.result = "OK"
	return result
end

local function DeleteDID (web)
	local postedData = web.input["post_data"]
	local success, didInfo = pcall(json.decode, postedData)

	LogInfo("[PROVISIONING.DIDS:DeleteDID] Deleting DID. User %s@%s posted data '%s'.", inconcert.authUser.userName, inconcert.authUser.virtualContactCenter, postedData)

	local result = {}

	local success, dids = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.PACKAGES:DeleteDID] Error Deleting DID. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local didId = dids[1]
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")

	local success, didInfo = modelDIDs.Get(didId, ksProv)
	if not success or not didInfo then
		ksProv:Close()
		LogError("[PROVISIONING.PACKAGES:DeleteDID] Error Deleting DID %s. Not Found.", didId)
		web.status = 404
		result.result = "DID DOES NOT EXIST"
		return result
	end

	local response = modelDIDs.Delete(didId, ksProv)
	if not response.success then
		LogError("[PROVISIONING.DIDS:DeleteDID] Error Deleting DID %s. Error: %s.", didId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.DIDS:DeleteDID] DID '%s' deleting successfully.", didId)

	ksProv:Close()
	web.status = 200
	result.result = "OK"

	return result
end

local function GetDID (web, didId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIDS:GetDID] Retrieving DID '%s'. User %s.", didId, userId)

	local success, didInfo = modelDIDs.Get(didId)
	if not success or not didInfo then
		LogError("[PROVISIONING.DIDS:GetDID] Error Retrieving DID %s. Not Found.", didId)
		web.status = 404
		local result = {}
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return didInfo
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.DIDS:Search] 1 - (START) Loading DIDs. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, dids = modelDIDs.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = dids
		LogWarning("[PROVISIONING.DIDS:Search] Error loading DIDs. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--dids = utils.FilterEntities(dids, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(dids)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.DIDS:SEARCH] 2 - Filtering rows")

	for id, did in pairs(dids) do
		local didId = utils.GetId(id)
		did.Id = didId
		did.Name = did.Name or didId
		if utils.MatchFilter(did, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, did)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.DIDS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user dids
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.DIDS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadDIDs (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.DIDS.:LoadDIDs] Loading DIDs, user %s, vcc %s", userId, vcc)

	local success, dids = modelDIDs.ListAll()
	if not success then
		local err = dids
		LogWarning("[PROVISIONING.DIDS:LoadDIDs] Error loading DIDs. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return dids
end

function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix
	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- DID i6
	module:dispatch_post(fn .. Search, p .. "/search/")
	module:dispatch_get (HasPermission(GetDID   , "view_dids"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(DeleteDID, "delete_did", fn), p .. "/delete/(.*)")
	module:dispatch_post(HasPermission(SaveDID  , "save_did", fn  ), p .. "/save/"      )

	module:dispatch_get (replyFn .. LoadDIDs, p .. "/load/")

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end