local utils = require "shared.apps.utils"
local modelHosts = require("apps.provisioning.models.hosts")
local modelPeers = require("apps.provisioning.models.peers")
local m_permissions = require "shared.permissions.permissions"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil
m_appLocation = nil

local function Render (web)
	_M.response.DontCache(web)
	local env = {_T = m_htmlRender.extension.makeTextTranslator()}
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING:Render] Render form HOST with sections for user %s, group %s", userId, inconcert.authUser.userGroup)

	local htmlFile = m_appLocation .. "/pages/" .. "frm_host_content.html"

	local htmlContent = m_htmlRender:RenderPageLP(htmlFile, web, env)
	if not htmlContent then
		assert(htmlContent, "[PROVISIONING:Render] HTML HOST does not exist: %s", htmlFile)
	end

	return htmlContent
end

local function GetHost (web, hostId)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.HOSTS:GetHost] Retrieving HOST '%s'. User %s.", hostId, userId)

	local result = {}
	
	local success, hostInfo = modelHosts.Get(hostId)
	if not success then
		LogError("[PROVISIONING.HOSTS:GetHost] Error Retrieving HOST %s. Not Found.", hostId)
		web.status = 404
		result.result = "NOT FOUND"
		return result
	end

	web.status = 200
	return hostInfo
end

local function ValidateAndMakeInfoToUpdate (hostId, hostInfo, keyspace)
	assert(hostId)
	assert(hostInfo)

	LogInfo("[PROVISIONING.HOSTS:ValidateAndMakeInfoToUpdate] Validating HOST '%s' info to updated.", hostId)

	local hostExist, currentHostInfo = modelHosts.Get(hostId, keyspace)

	-- 1 - clean up valores que llegan del lado del cliente
	local isNew = hostInfo.isNew

	if hostExist then
		-- Si estoy creando el Host y el mismo ya exsite en la BD, fallo!!
		if isNew then
			return false, "MUST_HAVE_UNIQUE_NAME"
		end
	end
	-- host id
	if utils.StringIsEmpty(hostInfo.hostId) then
		return false, "HOST_ID_IS_REQUIRED"
	end
	hostInfo.hostId = utils.alltrim(hostInfo.hostId)
	-- private ip
	if utils.StringIsEmpty(hostInfo.sipServer) then
		return false, "PRIVATE_IP_IS_REQUIRED"
	end
	hostInfo.sipServer = utils.alltrim(hostInfo.sipServer)
	-- public ip
	if utils.StringIsEmpty(hostInfo.rtmpServer) then
		return false, "PUBLIC_IP_IS_REQUIRED"
	end
	hostInfo.rtmpServer = utils.alltrim(hostInfo.rtmpServer)
	-- telProxyPort
	if utils.StringIsEmpty(hostInfo.telProxyPort) then
		return false, "TEL_PROXY_PORT_IS_REQUIRED"
	end
	hostInfo.telProxyPort = utils.alltrim(hostInfo.telProxyPort)
	hostInfo.isNew = nil

	return true, hostInfo, currentHostInfo
end

local function SaveHost (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.HOSTS:SaveHost] Saving HOST. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, hosts = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.HOSTS:SaveHost] Error Updating HOST. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local _, hostInfo = next(hosts)
	-- LogDebug("--> _ %s", _)
	-- LogDebug("--> hostInfo %s", hostInfo)

	assert(hostInfo.hostId)
	local isNew = hostInfo.isNew
	hostInfo.hostId = hostInfo.hostId:lower()
	local hostId = hostInfo.hostId
	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, hostInfo, currentHostInfo = ValidateAndMakeInfoToUpdate(hostId, hostInfo, keyspace)

	if not success then
		local errMsg = hostInfo
		LogError("[PROVISIONING.HOSTS:SaveHost] Error Validating HOST '%s'. Err: %s", hostId, errMsg)
		web.status = 500
		result.result = errMsg
		keyspace:Close()
		return result
	end
	
	local success, errMsg = modelHosts.Save(hostId, hostInfo, keyspace)
	if not success then
		LogError("[PROVISIONING.HOSTS:SaveHost] Error Updating HOST '%s'. Err: %s", hostId, errMsg)
		web.status = 500
		result.result = errMsg -- Este código de error llega desde la Rocka "provisioningmanager"
		keyspace:Close()
		return result
	end

	LogInfo("[PROVISIONING.HOSTS:SaveHost] '%s' saved successfully.", hostId)

	--En este punto el host se salvo. Si no es new y se modifico la ip se actualizan todos los telefonos ip.
	if not isNew and currentHostInfo and (hostInfo.sipServer ~= currentHostInfo.sipServer) then
		local success, peers = modelPeers.ListAll(keyspace)
		--LogInfo("[PROVISIONING.HOSTS:SaveHost] 1- updated %s peers '%s'", success, json.encode(peers))
		if success then
			for id, peer in pairs(peers) do
				if peer.hostId:lower() == hostId:lower() then
					local success, err = modelPeers.Save(peer.name, nil, peer, keyspace)
					if not success then
						LogError("[PROVISIONING.PEERS:SaveHost] Error Updating Peer '%s'. Err: %s", id, err)
					end
				end
			end
		end
	end
	keyspace:Close()
	web.status = 200
	result.result = "OK"
	return result
end

local function DeleteHost (web)
	local postedData = web.input["post_data"]
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.HOSTS:DeleteHost] Deleting HOST. User %s posted data '%s'.", userId, postedData)

	local result = {}

	local success, hosts = pcall(json.decode, postedData)
	if not success then
		LogError("[PROVISIONING.HOSTS:DeleteHost] Error Deleting HOST. Invalid Request.")
		web.status = 400
		result.result = "INVALID REQUEST"
		return result
	end

	local hostId = hosts[1]
	local success, vccInfo = modelHosts.Get(hostId)
	if not success then
		LogError("[PROVISIONING.HOSTS:DeleteHost] Error Deleting HOST %s. Not Found.", hostId)
		web.status = 404
		result.result = "HOST DOES NOT EXIST"
		return result
	end

	local response = modelHosts.Delete(hostId)
	if not response.success then
		LogError("[PROVISIONING.HOSTS:DeleteHost] Error Deleting HOST %s. Error: %s.", hostId, response.errMsg or "")
		web.status = 400
		result.result = response.errType -- Este código de error llega desde la Rocka "provisioningmanager"
		result.relatedId = response.relatedId
		return result
	end

	LogInfo("[PROVISIONING.HOSTS:DeleteHost] Host '%s' deleting successfully.", hostId)

	web.status = 200
	result.result = "OK"
	return result
end

local function Search (web)
	web.headers["Content-Type"] = "application/json"
	local vcc = inconcert.authUser.virtualContactCenter
	local encodedInfo = web.input["post_data"]
	local postData = web.request
	local startTime = os.time()

	LogInfo("[PROVISIONING.HOSTS:Search] 1 - (START) Loading Hosts. Posted Info: '%s'", encodedInfo)

	local keyspace = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, hosts = modelHosts.ListAll(keyspace)

	if not success then
		keyspace:Close()
		local err = hosts
		LogWarning("[PROVISIONING.HOSTS:Search] Error loading Hosts. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	--hosts = utils.FilterEntities(hosts, {IsSystemVCC = "0"})

	local totalUnfilteredRows = utils.GetRowCount(hosts)
	local totalAdded = 0
	local filteredRows = {}

	LogInfo("[PROVISIONING.HOSTS:SEARCH] 2 - Filtering rows")

	for id, host in pairs(hosts) do
		local hostId = utils.GetId(id)
		host.Id = hostId
		host.Name = host.Name or hostId
		if utils.MatchFilter(host, postData) then
			totalAdded = totalAdded + 1
			table.insert(filteredRows, host)
		end
	end

	keyspace:Close()

	LogInfo("[PROVISIONING.HOSTS:SEARCH] 3 - Sorting rows")
	--include only rows that belong to logged user hosts
	local response = {
		rows = filteredRows,
		last = "",
		total_rows = totalAdded,
		total_unfiltered_rows = totalUnfilteredRows
	}
	local responseParams = {
		MakeSortingValue = function(row)
			assert(row)
			return tostring(row.Name or row.Id):upper()
		end
	}
	local resultData = utils.MakeSearchResponse(response, postData, responseParams)

	local endTime = os.time()
	local elapsedTime = endTime - startTime
	LogInfo("[PROVISIONING.HOSTS:SEARCH] 4 - (END) Search rows. Retrieve %d of %d matched rows. Total Unfiltered Rows: %d. ELAPSED Time '%s' seconds.", #resultData.rows, resultData.total_rows, resultData.total_unfiltered_rows, elapsedTime)

	return resultData
end

--[[
]]--
local function LoadHosts (web)
	local vcc = inconcert.authUser.virtualContactCenter
	local userId = utils.AppendVcc(inconcert.authUser.userName, vcc)

	LogInfo("[PROVISIONING.HOSTS.:LoadHosts] Loading Hosts, user %s, vcc %s", userId, vcc)

	local success, hosts = modelHosts.ListAll()
	if not success then
		local err = hosts
		LogWarning("[PROVISIONING.HOSTS:LoadHosts] Error loading Hosts. Err'%s'.", err)
		web.status = 500
		return {result = "SERVER ERROR", err = err}
	end

	-- OK
	web.status = 200
	return hosts
end

----------------------------------
-- Mapas de las URLS
----------------------------------
function MapUris (module, htmlRender, prefix, appLocation)
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	local replyFn = utils.JsonReply

	local function HasPermission (fn, actionId, responseFn)
		responseFn = responseFn or replyFn
		return responseFn .. LogRequest .. m_permissions:CheckPermission("provisioning", actionId, fn)
	end

	local fn = utils.JsonValidatorReply .. LogRequest
	local p = prefix

	-- RENDER FRM
	module:dispatch_get(Render, p .. "/frm/render/")

	-- HOST i6
	module:dispatch_post(fn .. Search, p .. "/search/"                                    )
	module:dispatch_get (HasPermission(GetHost   , "view_hosts"     ), p .. "/get/(.+)"   )
	module:dispatch_post(HasPermission(SaveHost  , "save_host", fn  ), p .. "/save/"      )
	module:dispatch_post(HasPermission(DeleteHost, "delete_host", fn), p .. "/delete/(.*)")
	module:dispatch_get (replyFn .. LoadHosts, p .. "/load/"                              )

	m_htmlRender = htmlRender
	m_appLocation = appLocation
end