local utils = require "shared.apps.utils"
local modelVCCs = require "apps.provisioning.models.vccs"
module(..., package.seeall, inconcert.orbit.application.Decorate)

local LOG_PREFIX = "M I G R A T I O N ==> "
--------------------------------------------------------------------------------------
local m_migration = {}
local m_migrationsSteps = {}

--------------------------------------------------------------------------------------
-- STEP PROVISIONING - VCC ------
--------------------------------------------------------------------------------------

local m_vccStep = {
	id = "vcc",
	migrationId = nil, -- es el id de la definicion del archivo "/migrations/init.lua" tabla m_migrationsConfig[0,...].id
	description = "VCC Step Module Migration"
}
--[[

]]--
m_vccStep.Start = function(info)
	LogInfo(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.Start] Start Migration vcc '%s'.", info.vccId)
	-- 1. Asigno un rango de peers al VCC. Los mismos tienen que haber sido creados previamente desde la App Provisioning.
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "assign peers"
	})
	local success, result = m_migration.UpdatePeersInVCC(info)
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "assign peers",
		success = success,
		warningList = result
	})
	-- 2. Asigno N paquetes al VCC. Los mismos tienen que haber sido creados previamente desde la App Provisioning.
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "assign packages"
	})
	local success, result = m_migration.UpdatePackagesInVCC(info)
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "assign packages",
		success = success,
		warningList = result
	})
	-- ok
	return true
end

--[[

]]--
m_vccStep.Rollback = function(info)
	LogInfo(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.Rollback] Rollback Migration vcc '%s'.", info.vccId)
	-- 1. Remuevo todos los peers asignados al VCC.
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "free peers"
	})
	local success, result = m_migration.FreePeerFromVCC(info)
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "free peers",
		success = success,
		warningList = result
	})
	-- 2. Remuevo todos los paquetes asignados al VCC.
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "free peers"
	})
	local success, result = m_migration.FreePackageFromVCC(info)
	info.runningModule.UpdateStepProcess(m_vccStep, {
		stepProcessId = "free peers",
		success = success,
		warningList = result
	})
	-- ok
	return true
end

table.insert(m_migrationsSteps, m_vccStep) -- agrego step al array
--------------------------------------------------------------------------------------

--[[
	Asigno N peers al VCC.
	Parametros:
	{
		telephony = {
			peers = {
				from = 20,
				total = 3
			},
			..
		},
		..
	}
]]--
m_migration.UpdatePeersInVCC = function(info)
	local peers = info.telephony and info.telephony.peers
	LogInfo(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.UpdatePeersInVCC] Assign SIP Peers to Vcc '%s'. info: %s", info.vccId, utils.JSON_Encode(peers))

	if not peers then
		return false, {errType = "UNABLE TO ASSIGN PEERS TO VCC - PEERS NOT FOUND", vccId = info.vccId}
	end
	assert(type(peers.from) == "number")
	assert(type(peers.total) == "number")

	local currentPeer = peers.from
	local to = currentPeer + peers.total
	local peersToUpdate = {}

	while currentPeer < to do
		table.insert(peersToUpdate, {
			Id = string.format("SIP_%d", currentPeer)
		})
		currentPeer = currentPeer + 1
	end

	--LogDebug(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.UpdatePeersInVCC] Assign SIP Peers to Vcc '%s'.--> peersToUpdate: %s", info.vccId, utils.JSON_Encode(peersToUpdate))

	return modelVCCs.Peers().UpdatePeersInVCC(info.vccId, peersToUpdate, "agent", info.ksAgent, info.ksProv)
end

--[[

]]--
m_migration.FreePeerFromVCC = function(info)
	LogInfo(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.FreePeerFromVCC] Removing SIP Peers from Vcc '%s'.", info.vccId)
	return modelVCCs.Peers().FreePeerFromVCC(info.vccId, "agent", info.ksAgent, info.ksProv)
end

--[[
	Asigno N paquetes al VCC.
	Parametros:
	{
		telephony = {
			packages = {
				i_mode  = {}, -- entrantes
				io_mode = {"packageborrame_sal_ent1"}, -- entrantes y salientes
				o_mode  = {} -- salientes
			},
			..
		},
		..
	}
]]--
m_migration.UpdatePackagesInVCC = function(info)
	local packages = info.telephony and info.telephony.packages
	LogInfo(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.UpdatePackagesInVCC] Assign Packages to Vcc '%s'. info: %s", info.vccId, utils.JSON_Encode(packages))

	if not packages then
		return false, {errType = "UNABLE TO ASSIGN PACKAGES TO VCC - PEERS NOT FOUND", vccId = info.vccId}
	end

	local packagesToUpdate = {}
	local function MakePackagesList (list)
		assert(type(list) == "table")
		for _, packageId in ipairs(list) do
			table.insert(packagesToUpdate, {
				Id = packageId
			})
		end
	end

	MakePackagesList(packages.i_mode)  -- agrego los paquetes entrantes
	MakePackagesList(packages.io_mode) -- agrego los paquetes entrantes y salientes
	MakePackagesList(packages.o_mode)  -- agrego los paquetes salientes

	--LogDebug(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.UpdatePackagesInVCC] Assign Packages to Vcc '%s'.--> packagesToUpdate: %s", info.vccId, utils.JSON_Encode(packagesToUpdate))

	return modelVCCs.Packages().UpdatePackagesInVCC(info.vccId, packagesToUpdate, info.ksAgent, info.ksProv)
end

--[[

]]--
m_migration.FreePackageFromVCC = function(info)
	LogInfo(LOG_PREFIX .. "[PROVISIONING.MIGRATIONS.VCC.FreePackageFromVCC] Removing Packages from Vcc '%s'.", info.vccId)
	return modelVCCs.Packages().FreePackageFromVCC(info.vccId, info.ksAgent, info.ksProv)
end

--[[

]]--
m_migration.GetSteps = function()
	return m_migrationsSteps
end

--[[

]]--
m_migration.Init = function(config)
	for _, step in ipairs(m_migration.GetSteps()) do
		step.migrationId = config.id
	end
end

------------------------------------------
--
-- Exported functions
--
return {
	Init     = m_migration.Init,
	GetSteps = m_migration.GetSteps
}
