local orbit = require("orbit")
require "inconcert.orbit.application"
require("inconcert.translators")
local LuaConfigurationControl = require "LuaConfigurationControl"
local cassandra = require "lua_cassandra"
local utils = require "shared.apps.utils"
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")
local audits = require "audits"
module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)
_M.model = nil

local model = relativeRequire("model", ...)
local m_models = model.GetModels()
local configControl
local modelBuilt = false

local m_helpers = require "helpers"
local m_hasRepositoryService = nil

local function HasRepositoryService()
	if m_hasRepositoryService == nil then
		m_hasRepositoryService = m_helpers.utils.HasRepositoryService()
	end
	return m_hasRepositoryService
end
---
-- Some components expect a lualiblogger compatible logger instance.
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function make_compatible_logger ()
	return {
		debug = LogDebug,
		info = LogInfo,
		warn = LogWarning,
		error = LogError,
		fatal = LogFatal
	}
end

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current
		app_status.CheckDependencies(newConfig, {"agent", "redis", "outboundengineredis", "cassandra"})

		app_status.current = newConfig

		if not inconcert.environment.apps.provisioning then
			inconcert.environment.apps.provisioning = {}
		end

		--Configure Redis
		inconcert.environment.apps.provisioning.redis = inconcert.environment.apps.provisioning.redis or {}
		inconcert.environment.apps.provisioning.redis.host = newConfig["redis"]["host"]
		inconcert.environment.apps.provisioning.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.apps.provisioning.redis.pwd = newConfig["redis"]["pwd"] or ""

		--Configure OutboundEngine Redis
		inconcert.environment.apps.provisioning.outboundengineredis = inconcert.environment.apps.provisioning.outboundengineredis or {}
		inconcert.environment.apps.provisioning.outboundengineredis.host = newConfig["outboundengineredis"]["host"]
		inconcert.environment.apps.provisioning.outboundengineredis.port = tonumber(newConfig["outboundengineredis"]["port"])
		inconcert.environment.apps.provisioning.outboundengineredis.pwd = newConfig["outboundengineredis"]["pwd"] or ""

		--Configure Elasticsearch (is an array, but for now, we only use the first node)
		inconcert.environment.apps.provisioning.elasticsearch = inconcert.environment.apps.provisioning.elasticsearch or {}
		inconcert.environment.apps.provisioning.elasticsearch.host = newConfig.elasticsearch[1].host
		inconcert.environment.apps.provisioning.elasticsearch.port = tonumber(newConfig.elasticsearch[1].port)
		-- elascitsearch user credentials
		inconcert.environment.apps.provisioning.elasticsearch.credentials = {
			userId    = newConfig.elasticsearch[1].userId,   -- or "inconcert",
			userPwd   = newConfig.elasticsearch[1].userPwd  -- or "12345678"
		}

		--Configure Elasticsearch Contact Dialer (is an array, but for now, we only use the first node)
		inconcert.environment.apps.provisioning.elasticsearch_dialer = inconcert.environment.apps.provisioning.elasticsearch_dialer or {}
		inconcert.environment.apps.provisioning.elasticsearch_dialer.host = "127.0.0.1"
		inconcert.environment.apps.provisioning.elasticsearch_dialer.port = -1
		for index, value in newConfig["elasticsearch_dialer"]:childs() do
			inconcert.environment.apps.provisioning.elasticsearch_dialer.host = value.host
			inconcert.environment.apps.provisioning.elasticsearch_dialer.port = tonumber(value.port)
			-- inconcert.environment.apps.provisioning.elasticsearch_dialer.numberOfShards = tonumber(value.number_of_shards)
			-- inconcert.environment.apps.provisioning.elasticsearch_dialer.numberOfReplicas = tonumber(value.number_of_replicas)
			break
		end
		inconcert.environment.apps.provisioning.elasticsearch_dialer.credentials = inconcert.environment.apps.provisioning.elasticsearch.credentials

		--Configure call
		inconcert.environment.apps.provisioning.call = {}
		local call = inconcert.environment.apps.provisioning.call
		for key, value, state in newConfig["call"]:childs() do
			if state ~= "deleted" then
				call[#call + 1] = {host = value.ip, port = value.port}
			end
		end

		--Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
				audits.init({
					cassandra = { host = value.host, port = value.port }
				})
			end
		end
		inconcert.environment.apps.provisioning.cassandra = cassandra

		--Configure time provisioning
		inconcert.environment.apps.provisioning.timeServers = {}
		local timeservers = inconcert.environment.apps.provisioning.timeServers
		for key, value, state in newConfig["timeServers"]:childs() do
			if state ~= "deleted" then
				timeservers[#timeservers + 1] = {IP = value.IP, port = value.port}
			end
		end

		if HasRepositoryService() then
			inconcert.environment.apps.provisioning.repository = inconcert.environment.apps.provisioning.repository or {}
			for key, value, state in newConfig["repository"]:childs() do
				if state ~= "deleted" then
					table.insert(inconcert.environment.apps.provisioning.repository, { ip = value.ip, port = value.port })
					LogInfo("Connecting to repository service on %s:%s", value.ip, tostring(value.port))
				end
			end
		end

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end

		InitializeFullcontact()

		app_status.actualStatus = "ready"
		return "ready"
	end
}

local function GetAppEnvironment ()
	if inconcert.environment.apps.provisioning then
		-- TODO: Revisar si el cluster llega en "inconcert.environment.apps.provisioning"
		inconcert.environment.apps.provisioning.cluster = inconcert.environment.apps.provisioning.cluster or inconcert.environment.apps.agent.cluster
 		return inconcert.environment.apps.provisioning
	end
	-- si no existe retorno el del agent
	local agentEnvroment = inconcert.environment.apps.agent
	assert(agentEnvroment)
	return agentEnvroment
end

function InitializeFullcontact (callback)
	--LogInfo("[PROVISIONING:InitializeFullcontact] Initializing Fullcontact for 'PROVISIONING'.")
	local appEnvironment = GetAppEnvironment()
	local Fullcontact = require "fullcontact"
	Fullcontact.init({
		owner               = "PROVISIONING",
		cassandra           = {
			host = appEnvironment.cassandra.ip,
			port = appEnvironment.cassandra.port
		},
		cluster             = appEnvironment.cluster,
		elasticsearch       = appEnvironment.elasticsearch,
		elasticsearchDialer = appEnvironment.elasticsearch_dialer,
		redis               = appEnvironment.redis,
		timeServers         = appEnvironment.timeServers,
		logger              = make_compatible_logger()
	})
	if callback then
		callback()
	end
end

-- Returns a function bound to a translator for this application
local function makeTextTranslator ()
	return utils.MakeTextTranslator("provisioning")
end

-- This function returns the title of the application. It needs to use an explicit translator because _T could be bound to a
-- different application.
function GetApplicationTitle ()
	local _T = makeTextTranslator()
	return _T("provisioning")
end

local htmlRender = require"inconcert.render".new("provisioning", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("provisioning") end,
	model = m_models,
	makeTextTranslator = makeTextTranslator,
	appLocation = m_appLocation,
})

-- funcion que es llamada cuando el webserver se inicializa.
function Initialize()
	LogInfo("Starting application 'provisioning'.")
	--OJO que si se cambia en controllers\interactions.lua se debe cambiar aqui tambien!
	configControl = LuaConfigurationControl.Client {
		type = "PROVISIONING",
		address = inconcert.environment.apps.agent.config_server.ip,
		port = inconcert.environment.apps.agent.config_server.port,
		node_address = inconcert.environment.apps.agent.config_server.myIp,
		node_port = inconcert.environment.apps.agent.config_server.myPort,
	}
	configControl:AddServiceDependency("Agent")
	if HasRepositoryService() then
		configControl:AddServiceDependency("Repository")
	end
	configControl:TriggerNodeReady()
	return true
end

--
-- The user has been authenticated and has permissions for this application. Should we let him enter anyway? (returns false if not)
function CanInitializeSession(user)
	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function InitializeSession (authUser) end

--
-- This function gets called when the user logs out. This gets called reliably. The Logout function below may not
-- be called if the user logs off first (apps/login/logout)
function FinalizeSession(appEnvironment)
	if app_status:CheckStatus() ~= "ready" then
		LogWarning("App status is not ready. Can not logout agent %s on session %s", inconcert.authUser.userName, inconcert.authUser.authToken)
		return false
	end
	LogInfo("apps.provisioning: Detail of the logged out user:  userName: (%s) - userGroup: (%s) - virtualContactCenter: (%s)",
		inconcert.authUser.userName, inconcert.authUser.userGroup, inconcert.authUser.virtualContactCenter)
	local sessionId = "PROVISIONINGSESSION:" .. inconcert.authUser.userName .. "@" .. inconcert.authUser.virtualContactCenter
	SharedSessions.Discard(sessionId)
	return true
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetAppLocation()
	return m_appLocation
end

--///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
--
function GetApplicationName()
	return "provisioning"
end

--///////////////////////////////////////////////////////////////////////////////////
--
function serve_file(app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, m_appLocation.."/"..filename)
	end
end

--///////////////////////////////////////////////////////////////////////////////////
--

-----------------------
--  (cotrolles) urls/apis
local function GetControllers ()
	local controllerDir = m_appLocation .. "/controllers"
	local controllers = {}
	for file in lfs.dir(controllerDir) do
		if file ~= "." and file ~= ".." then
			local f = controllerDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local controllerName = string.match(file, [[(.+).lua]])
				if controllerName then
					controllers[controllerName] = { url = controllerName }
				end
			end
		end
	end
	return controllers
end

for key, controllerDef in pairs(GetControllers()) do
	--requires y mapeo funciones
	local controller = relativeRequire("controllers." .. key, ...)
	if type(controller.MapUris) == "function" then
		controller.MapUris(_M, htmlRender, "/" .. controllerDef.url, "/apps/provisioning")
	end
end
local controller_perm = require "shared.permissions.controllers.permissions"
controller_perm.MapUris(_M, htmlRender, "/permissions")

local function CheckDependencies(web)
	app_status.CheckStatus()

	return true
end

--scripts and pages

--paginas
local function render_pages(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:RenderPageLP("pages/" .. name .. ".html", web, env)
end

--scripts
function render_script(web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:ScriptLp("scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")
_M:dispatch_post(CheckDependencies, "/check_dependencies/")

--
--  styles and resources
--  NOTA: ESTO SE PROCESA EN XAVANTE
_M:dispatch_get(_M:serve_file(), "/images/.+", "/styles/.+", "/css/.+", ".+%.mp3")
