module(..., package.seeall)

local utils = require "shared.apps.utils"
local PM = require("provisioningmanager")
local m_datasource

function GetName ()
	return "Gateways"
end

function GetDefinition (datasource)
	m_datasource = datasource
	PM.SetRedis(datasource.redis)
	return _M
end

----------------------------------
-- Gateways
----------------------------------

function ListAll (ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local gateways = PM.ListGateways(ksProv)

	if releaseConn then
		ksProv:Close()
	end
	gateways = gateways or {}

	return true, gateways
end

local function ProcessNodeTrunks(gateway, ksProv)
	local trunkStrings = gateway.nodeTrunks or {}
	local trunks = {}
	for trunkId, trunkStr in pairs(trunkStrings) do
		trunks[trunkId] = json.decode(trunkStr)
		local nodeTrunk = PM.GetTrunk(ksProv, trunkId)
		trunks[trunkId].Host = nodeTrunk.hostId
	end
	return trunks
end

local function ProcessCarrierTrunks(gateway, _ksProv)
	local trunkStrings = gateway.carrierTrunks or {}
	local trunks = {}
	for trunkId, trunkStr in pairs(trunkStrings) do
		trunks[trunkId] = json.decode(trunkStr)
	end
	return trunks
end

local function ProcessDialplanContexts(gateway, _ksProv)
	local contextStrings = gateway.dialplanContexts or {}
	local contexts = {}
	for contextId, contextStr in pairs(contextStrings) do
		contexts[contextId] = json.decode(contextStr)
	end
	return contexts
end

function Get (gatewayId, ksProv)
	--LogDebug("[provisioning:models.gateways.Get] Retrieving GATEWAY: '%s'.", gatewayId)

	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local currentGateway = PM.GetGateway(ksProv, gatewayId)

	if not utils.TableIsEmpty(currentGateway) then
		currentGateway.nodeTrunks = ProcessNodeTrunks(currentGateway, ksProv)
		currentGateway.carrierTrunks = ProcessCarrierTrunks(currentGateway, ksProv)
		currentGateway.dialplanContexts = ProcessDialplanContexts(currentGateway, ksProv)
	end

	if releaseConn then
		ksProv:Close()
	end

	if utils.TableIsEmpty(currentGateway) then
		local errMsg = "GATEWAY DOES NOT EXIST"
		--LogWarning("[provisioning:models.gateways.Get] Error retrieving GATEWAY '%s'. GATEWAY does not exist.'", gatewayId)
		return false, errMsg
	end

	if currentGateway.error then
		return false, currentGateway.error
	end

	return true, currentGateway
end

function Save (gatewayId, gatewayInfo, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	-- Clean Up ---
	gatewayInfo.isNew = nil

	local result = PM.UpdateGateway(ksProv, gatewayId, gatewayInfo)
	if not result.success then
		LogWarning("[provisioning:models.gateways.Save] Error saving GATEWAY '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function Delete (gatewayId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.DeleteGateway(ksProv, gatewayId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.Delete] Error deleting GATEWAY '%s' -> '%s'.",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

----------------------------------
-- Node trunks
----------------------------------

function ListAllNodeTrunks (gatewayId, ksProv)
	local gateway = Get(gatewayId, ksProv)
	if not success then return success, gateway end

	return true, gateway.nodeTrunks or {}
end

function GetNodeTrunk (gatewayId, trunkId, ksProv)
	local success, trunks = ListAllNodeTrunks(gatewayId, ksProv)
	if not success then return success, trunks end

	if not trunks[trunkId] then
		return false, "TRUNK DOES NOT EXIST"
	end
	
	return true, trunks[trunkId]
end

function SaveNodeTrunk (gatewayId, trunkId, trunkInfo, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	-- Clean Up ---
	trunkInfo.isNew = nil

	local result = PM.UpdateGatewayNodeTrunk(ksProv, gatewayId, trunkId, trunkInfo)
	if not result.success then
		LogWarning("[provisioning:models.gateways.SaveNodeTrunk] Error saving TRUNK '%s' -> '%s'.",
				   trunkId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function DeleteNodeTrunk (gatewayId, trunkId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.DeleteGatewayNodeTrunk(ksProv, gatewayId, trunkId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.DeleteNodeTrunk] Error deleting TRUNK '%s' -> '%s'.",
				   trunkId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

----------------------------------
-- Carrier trunks
----------------------------------

function ListAllCarrierTrunks (gatewayId, ksProv)
	local success, gateway = Get(gatewayId, ksProv)
	if not success then return success, gateway end

	return true, gateway.carrierTrunks or {}
end

function GetCarrierTrunk (gatewayId, trunkId, ksProv)
	local success, trunks = ListAllCarrierTrunks(gatewayId, ksProv)
	if not success then return success, trunks end

	if not trunks[trunkId] then
		return false, "TRUNK DOES NOT EXIST"
	end
	
	return true, trunks[trunkId]
end

function SaveCarrierTrunk (gatewayId, trunkId, trunkInfo, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	-- Clean Up ---
	trunkInfo.isNew = nil

	local result = PM.UpdateGatewayCarrierTrunk(ksProv, gatewayId, trunkId, trunkInfo)
	if not result.success then
		LogWarning("[provisioning:models.gateways.SaveCarrierTrunk] Error saving TRUNK '%s' -> '%s'.",
				   trunkId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function DeleteCarrierTrunk (gatewayId, trunkId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.DeleteGatewayCarrierTrunk(ksProv, gatewayId, trunkId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.DeleteCarrierTrunk] Error deleting TRUNK '%s' -> '%s'.",
				   trunkId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

----------------------------------
-- Dialplan Contexts
----------------------------------

function ListAllDialplanContexts (gatewayId, ksProv)
	local success, gateway = Get(gatewayId, ksProv)
	if not success then return success, gateway end

	return true, gateway.dialplanContexts or {}
end

function GetDialplanContext (gatewayId, contextId, ksProv)
	local success, contexts = ListAllDialplanContexts(gatewayId, ksProv)
	if not success then return success, contexts end

	if not contexts[contextId] then
		return false, "CONTEXT DOES NOT EXIST"
	end
	
	return true, contexts[contextId]
end

function ValidateDialplanContext (gatewayId, contextId, contextInfo, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	-- Clean Up ---
	contextInfo.isNew = nil

	local result = PM.ValidateGatewayDialplanContext(ksProv, gatewayId, contextId, contextInfo)
	if not result.success then
		LogWarning("[provisioning:models.gateways.ValidateDialplanContext] Error updating CONTEXT '%s' -> '%s'.",
				   contextId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function SaveDialplanContext (gatewayId, contextId, contextInfo, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	-- Clean Up ---
	contextInfo.isNew = nil

	local result = PM.UpdateGatewayDialplanContext(ksProv, gatewayId, contextId, contextInfo)
	if not result.success then
		LogWarning("[provisioning:models.gateways.UpdateGatewayDialplanContext] Error updating CONTEXT '%s' -> '%s'.",
				   contextId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function DeleteDialplanContext (gatewayId, contextId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.DeleteGatewayDialplanContext(ksProv, gatewayId, contextId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.DeleteDialplanContext] Error deleting CONTEXT '%s' -> '%s'.",
				   contextId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function SaveMonitoring(gatewayId, monitoring, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.UpdateGatewayMonitoring(ksProv, gatewayId, monitoring)
	if not result.success then
		LogWarning("[provisioning:models.gateways.SaveMonitoring] '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function MonitoringGrafanaWidgets(gatewayId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.GatewayMonitoringGrafanaWidgets(ksProv, gatewayId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.MonitoringGrafanaWidgets] '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function MonitoringSipCaptureWidgets(gatewayId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.GatewayMonitoringSipCaptureWidgets(ksProv, gatewayId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.MonitoringSipCaptureWidgets] '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function MonitoringStopSipCaptureWidgets(gatewayId, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.GatewayMonitoringStopSipCaptureWidgets(ksProv, gatewayId)
	if not result.success then
		LogWarning("[provisioning:models.gateways.MonitoringStopSipCaptureWidgets] '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function MonitoringSendConsoleCommand(gatewayId, request, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.GatewayMonitoringSendConsoleCommand(ksProv, gatewayId, request)
	if not result.success then
		LogWarning("[provisioning:models.gateways.MonitoringSendConsoleCommand] '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end

function MonitoringConfigurationFile(gatewayId, fileName, ksProv)
	local releaseConn = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConn = true
	end

	local result = PM.GatewayMonitoringConfigurationFile(ksProv, gatewayId, fileName)
	if not result.success then
		LogWarning("[provisioning:models.gateways.MonitoringConfigurationFile] '%s' -> '%s'",
				   gatewayId, result.errMsg)
	end

	if releaseConn then
		ksProv:Close()
	end

	return result
end
