module(..., package.seeall)

local utils = require "shared.apps.utils"
local modelProvPeers = require("apps.provisioning.models.peers")
local PM = require("provisioningmanager")
local m_helpers = require "helpers"
local http = require("socket.http")

local m_datasource

-------------
-- VCC (exporta funciones)
-------------
local m_vcc = {}

function VCC ()
	return m_vcc
end

--[[
	Crea indices y templates especificos para un VCC.
	Se utiliza desde la migracion de VCC de Allegro a Nativos.
]]--
m_vcc.CreateESIndexesForVCC = function(vcc, indexes, ksAgent)
	PM.CreateESIndexesForVCC(vcc, indexes, ksAgent)
end

--Al devolver datos al cliente, usamos este fake password para no devolver el password real.
--Al guardar comparamos con este valor para saber si el password ha cambiado
local m_externalAuthFakePassword = "SnvbgV.C0hg7Rlm9"

local m_hasRepositoryService = nil

local function HasRepositoryService()
	if m_hasRepositoryService == nil then
		m_hasRepositoryService = m_helpers.utils.HasRepositoryService()
	end
	return m_hasRepositoryService
end

local function CreateS3BucketName(vccId, timestamp)
	local vccId4S3 = vccId:lower()
	local timestamp_str = tostring(timestamp)
	--Only letters and numbers on bucket name
	vccId4S3 = string.gsub(vccId4S3, "%W", "-")
	--Bucket name must start with a letter or number
	if string.sub(vccId4S3, 1, 1) == "-" then
		vccId4S3 = "a" .. string.sub(vccId4S3, 2, -1)
	end
	--Max length on a bucket name is 63, so the length of the vcc part of that name must be lower than vccMaxLen
	local vcvMaxLen = 63 - string.len(timestamp_str + 1)
	vccId4S3 = string.sub(vccId4S3, 1, vcvMaxLen - 1)

	local s3Bucket = vccId4S3 .. "-" .. timestamp_str
	return s3Bucket
end

local function UpdateProcessMaxRate (vcc, maxManualRate, ksAgent)
	local sucess          = true
	local ksOE            = utils.GetDatasource():CassandraConnection( "OutboundEngine" )
	local apiProccess     = utils.GetOEApi().native.processes
	local nativeProcesses = apiProccess.LoadProcessesAndBatches(ksOE,vcc)

	for procId, currentProc in pairs(nativeProcesses) do
		if  currentProc.configuration               and currentProc.configuration.channels and
			currentProc.configuration.channels.CALL and currentProc.configuration.channels.CALL.correctionManualRate
		then
			local manualRate = tonumber(currentProc.configuration.channels.CALL.correctionManualRate)
			if manualRate and manualRate > maxManualRate then
				LogInfo("[provisioning:models.vccs.UpdateProcessMaxRate] Updating process manual rate to max value: %d", maxManualRate)
				currentProc.configuration.channels.CALL.correctionManualRate = maxManualRate
				currentProc.__vcc  = currentProc.vcc

				local result, data = apiProccess.SaveProcess(currentProc, true, ksAgent, nil, ksOE)
				if not result then
					sucess = false
					LogError("[provisioning:models.vccs.UpdateProcessMaxRate] Error updating porcess manual rate: %s", data)
				end
			end
		end
	end

	ksOE:Close()

	return sucess
end

function GetName ()
	return "VCCs"
end

function GetDefinition (datasource)
	m_datasource = datasource
	PM.SetRedis(datasource.redis)
	PM.InitAuthfilter(datasource.cassandra, datasource.redis)
	return _M
end

local function InitHelpersAgentServiceOptions ()
	--#11602 - LicenseSystem - PM.UpdateVCC and PM.GetVCC now require a previous call to InitHelpersAgentServiceOptions
	local options = utils.Get_m_helpers_agentservice_options()
	local success, err = pcall(PM.InitHelpersAgentServiceOptions, options)
	if not success then
		LogError("[provisioning:models.InitHelpersAgentServiceOptions] Unable to initialize agent service helpers: '%s'", err)
		return false
	end
	return true
end

function ListAll (ksAgent)
	local release_connection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection( "Agent" )
		release_connection = true
	end

	local success, vccs = pcall(PM.ListVCCs, ksAgent)
	if not success then
		local errMsg = vccs
		LogWarning("[provisioning:models.vccs.ListAll] Error retrieving vccs  -> '%s'", errMsg)
	end

	if release_connection then
		ksAgent:Close()
	end
	vccs = vccs or {}

	for _, vcc in pairs(vccs) do
		-- (Security captcha) Elimino las claves publicas y privadas si las mismas son las de inconcert (esto para que no sean enviadas hacia el navegador)
		CleanUpInconcertCaptchaInfo(vcc)
		-- Elimino el password de acceso al servidor externo de autenticacion para que no sea enviada hacia el navegador
		CleanUpExternalAuthenticationPassword(vcc)
		-- Cleanup StorageConfig & StorageLifecycle
		vcc.StorageConfig    = nil
		vcc.StorageLifecycle = nil
	end

	return success, vccs
end

local function GetFirstAdminUser (vccId, keyspace)
	local vccUsers = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCUsers",
		key = vccId,
		columnCount = 5000
	} or {}
	for vccUserId, vccUserData in pairs(vccUsers) do
		if vccUserData.UserGroup == "administrator" then
			return vccUserId
		end
	end
end

local function HasUsers (vccId, keyspace)
	local vccUsers = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCUsers",
		key = vccId,
		columnCount = 1
	} or {}
	return not utils.TableIsEmpty(vccUsers)
end

function Get (vccId, ksAgent, withCustomFields)
	 -- Este parámetro en true es para que la rocka calcule ciertos atributos (ej: HasAllegroIntegration, HasHelpDeskIntegration, etc)
	withCustomFields = withCustomFields or false
	--LogDebug("[provisioning:models.vccs.Get] Retrieving VCC: '%s'.", vccId)

	local releaseConnection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection( "Agent" )
		releaseConnection = true
	end

	local success, currentVcc = GetVCC(ksAgent, vccId, withCustomFields)
	--LogDebug("[provisioning:models.vccs.Get] Retrieved vcc '%s' -> success: '%s', success: '%s'", vccId, success, currentVcc)
	if not success then
		local errMsg = currentVcc
		LogWarning("[provisioning:models.vccs.Get] Error retrieving vcc '%s' -> '%s'", vccId, errMsg)
		if releaseConnection then
			ksAgent:Close()
		end
		return false, errMsg
	end

	-- VCCUsers
	local hasUsers = HasUsers(vccId, ksAgent)
	currentVcc.hasUsers = hasUsers

	-- Devolvemos el primero de los usuarios administradores del VCC, para posibilitar el reseteo de su password
	currentVcc.adminUser = GetFirstAdminUser(vccId, ksAgent)
	-- Este fake pwd sirve para que al guardar comprobemos si nos lo cambiaron, y resetear en ese caso la password del adminUser
	currentVcc.adminPwd = "0xfab43345ff5d90"

	if releaseConnection then
		ksAgent:Close()
	end

	-- Session Timeouts
	if type(currentVcc.SessionTimeouts) == "string" then
		currentVcc.SessionTimeouts = json.decode(currentVcc.SessionTimeouts)
	end

	-- Interaction Notifications
	if type(currentVcc.InteractionNotifications) == "string" then
		currentVcc.InteractionNotifications = json.decode(currentVcc.InteractionNotifications)
	end

	-- Security captcha
	if type(currentVcc.Security) == "string" then
		currentVcc.Security = json.decode(currentVcc.Security)
		-- Elimino las claves publicas y privadas si las mismas son las de inconcert (esto para que no sean enviadas hacia el navegador)
		CleanUpInconcertCaptchaInfo(currentVcc)
	end

	--Security External authentication
	local authType = currentVcc.Security.authentication.type
	if authType ~= "BASIC" and authType ~= "MFA" then
		-- Elimino el password de acceso al servidor externo de autenticacion para que no sea enviada hacia el navegador
		CleanUpExternalAuthenticationPassword(currentVcc)
	end

	-- Outgoing Batches Limit
	if type(currentVcc.OutgoingBatchesLimit) == "string" then
		currentVcc.OutgoingBatchesLimit = json.decode(currentVcc.OutgoingBatchesLimit)
	end

	-- Social Networks Post Limit
	if type(currentVcc.SocialNetworkPostsLimit) == "string" then
		currentVcc.SocialNetworkPostsLimit = json.decode(currentVcc.SocialNetworkPostsLimit)
	end

	-- BOT Type (LUIS o RASA)
	if type(currentVcc.BotInfo) == "string" then
		currentVcc.BotInfo = json.decode(currentVcc.BotInfo)
	end

	-- Screen Recording
	if type(currentVcc.ScreenRecording) == "string" then
		currentVcc.ScreenRecording = json.decode(currentVcc.ScreenRecording)
	end

	-- Video Config (webcall)
	if type(currentVcc.VideoConfig) == "string" then
		currentVcc.VideoConfig = json.decode(currentVcc.VideoConfig)
	end

	-- Audit
	if type(currentVcc.Audit) == "string" then
		currentVcc.Audit = json.decode(currentVcc.Audit)
	end

	-- ApiConfiguration
	if type(currentVcc.ApiConfiguration) == "string" then
		currentVcc.ApiConfiguration = json.decode(currentVcc.ApiConfiguration)
	end

	-- Databases
	if type(currentVcc.Databases) == "string" then
		currentVcc.Databases = json.decode(currentVcc.Databases)
	end

	-- Telephony Native
	if type(currentVcc.TelephonyConfig) == "string" then
		currentVcc.TelephonyConfig = json.decode(currentVcc.TelephonyConfig)
	end

	-- Drafts
	if type(currentVcc.DraftsConfig) == "string" then
		currentVcc.DraftsConfig = json.decode(currentVcc.DraftsConfig)
	end

	-- Phone Validation
	if type(currentVcc.PhoneValidation) == "string" then
		currentVcc.PhoneValidation = json.decode(currentVcc.PhoneValidation)
	end

	-- Limits
	if type(currentVcc.ConfigLimits) == "string" then
		currentVcc.ConfigLimits = json.decode(currentVcc.ConfigLimits)
	end

	-- Load LUIS information -------------------
	--
	-- DEMO
	local isDemo, demoSubscription =  Luis().HasConfiguredDemoSubscriptionsInVcc(vccId)
	local isCustomDemo
	if demoSubscription then 
		local defaultDemoSubscriptionKey = Luis().GetDefaultDemoSubscriptionKey()
		-- Si la Demo que se ha configurado es la x Default de Inconcert, la marco como Custom y elimino la key, ya que la misma NO se tiene que retornar al lado del cliente!!!
		if defaultDemoSubscriptionKey == demoSubscription.subscriptionKey then
			isCustomDemo = false
			demoSubscription.subscriptionKey = nil
		else
			isCustomDemo = true
		end
	end

	--
	-- AZURE ACCOUNT & SUBSCRIPTIONS
	--  subscriptions
	local _, luisTotalSubscriptions, luisSubscriptionsKeys = Luis().LoadVCCSubscriptionsWithoutDemo(vccId)
	local isCustomAccount = luisTotalSubscriptions > 0
	--  azure account
	local vccAzureAccounts, totalAzureAccounts = Luis().LoadVCCAzureAccounts(vccId)
	local isAssignedAzureAccount = totalAzureAccounts > 0
	local _, vccAzureAccount
	if isAssignedAzureAccount then
		_, vccAzureAccount = next(vccAzureAccounts)
	end
	--
	-- TENANT
	local isCustomTenant = false
	local vccTenantAzureAccounts, totalVCCTenantAzureAccounts = Luis().LoadVCCTenantAzureAccount(vccId)
	-- Aca valido si los tentans han sido modificados, es decir, si NO son los por default de i6.
	if totalVCCTenantAzureAccounts > 0 then
		local _, vccTenant = next(vccTenantAzureAccounts)
		if vccTenant then
			local defaultTenantInfo = Luis().LoadDefaultTenantAzureAccount()
			for name, value in pairs(defaultTenantInfo) do
				if not vccTenant[name] then
					isCustomTenant = true
					break
				end
				if vccTenant[name] ~= value then
					isCustomTenant = true
					break
				end
			end
		end
	end

	assert(type(vccId) == "string")

	local vccAzureLanguageStudioAccounts, totalVCCAzureLanguageStudioAccounts = Luis().LoadVCCAzureLanguageStudioAccount(vccId) 
	local azureLanguageStudio = {} 
	if totalVCCAzureLanguageStudioAccounts > 0 then 
		local firstKey = next(vccAzureLanguageStudioAccounts) 
  		azureLanguageStudio = vccAzureLanguageStudioAccounts[firstKey] 
	else
		AzureLanguageStudio = {} 
	end 

	----
	--LogDebug("----> currentVcc: '%s'", currentVcc)
	currentVcc.Luis = {
		Demo = {
			Enabled = isDemo,
			IsCustom = isCustomDemo,
			DemoSubscription = demoSubscription
		},
		Tenant = {
			IsCustom = isCustomTenant,
			Total = totalVCCTenantAzureAccounts,
			AzureAccounts = vccTenantAzureAccounts
		},
		Account = {
			IsCustom = isCustomAccount,
			isAssignedAzureAccount = isAssignedAzureAccount,
			Azure = vccAzureAccount,
			Subscriptions = luisSubscriptionsKeys
		},
		AzureLanguageStudio = azureLanguageStudio 
	}

	-- Load QnA information
	local _, qnaMakerTotalSubscriptions, qnaMakerSubscriptionsKeys = QnA().LoadVCCSubscriptions(vccId)
	currentVcc.QnAMaker = {
		TotalSubscriptions = qnaMakerTotalSubscriptions,
		SubscriptionsKeys  = qnaMakerSubscriptionsKeys
	}

	-- Load QnA information V2 fede
	local _, qnaMakerTotalSubscriptionsV2, qnaMakerSubscriptionsKeysV2 = QnA().LoadVCCSubscriptionsV2(vccId) -- fede
	currentVcc.QnAMakerV2 = { -- fede
		TotalSubscriptions = qnaMakerTotalSubscriptionsV2,
		SubscriptionsKeys  = qnaMakerSubscriptionsKeysV2
	}

	if HasRepositoryService() then
		local prefixesConfig       = {}
		local lastS3StorageConfig  = {}
		local currentStorageConfig = {}
		
		-- Check interaction prefix config
		if currentVcc.InteractionPrefixConfig then
			-- Decode to a table
			prefixesConfig = json.decode(currentVcc.InteractionPrefixConfig)
			prefixesConfig.TreeString = ""

			for index, data in ipairs(prefixesConfig.PrefixTree) do
				prefixesConfig.TreeString = prefixesConfig.TreeString .. data .. "/"
			end
		else
			prefixesConfig.Action     = 2
			prefixesConfig.TreeString = "date/"
		end

		currentVcc.InteractionPrefixConfig = prefixesConfig

		-- Return just the last StorageConfig or the default value
		if not currentVcc.StorageConfig then
			--Default configuration to check if it has changed
			currentStorageConfig.StorageLocation = "cassandra"
			currentStorageConfig.StorageLocationParams = {}
		else
			--current Storage config is the newest one, based on timestamp
			local fullStorageConfig   = json.decode(currentVcc.StorageConfig)
			local maxTimestampIndex   = 0
			local maxTimestampIndexS3 = 0
			for timestampIndex, config in pairs(fullStorageConfig) do
				if (type(timestampIndex) == "number" or tonumber(timestampIndex) ~= nil) and maxTimestampIndex < tonumber(timestampIndex) and type(config) == "table" then
					currentStorageConfig = config
					maxTimestampIndex = tonumber(timestampIndex)
				end
				if (type(timestampIndex) == "number" or tonumber(timestampIndex) ~= nil) and maxTimestampIndexS3 < tonumber(timestampIndex) and type(config) == "table" and config.StorageLocation and config.StorageLocation == "S3"  then
					lastS3StorageConfig = config
					maxTimestampIndexS3 = tonumber(timestampIndex)
				end
			end

			currentStorageConfig = fullStorageConfig[tostring(maxTimestampIndex)]

			if lastS3StorageConfig and lastS3StorageConfig.StorageLocationParams and currentStorageConfig.StorageLocation ~= "S3" then
				currentStorageConfig.StorageLocationParams  = lastS3StorageConfig.StorageLocationParams
			end
		end

		currentVcc.StorageConfig = currentStorageConfig
	end

	-- Add StorageLifecycle
	if not currentVcc.StorageLifecycle then
		local currentStorageLifecycle = {}
		currentStorageLifecycle.S3Files         = "90"
		currentStorageLifecycle.S3MetaData      = "90"
		currentStorageLifecycle.RepositoryCall  = "90"
		currentStorageLifecycle.RepositoryMail  = "90"
		currentStorageLifecycle.RepositoryOther = "90"
		currentVcc.StorageLifecycle             = currentStorageLifecycle
	else
		currentVcc.StorageLifecycle             = json.decode(currentVcc.StorageLifecycle)
	end

	return success, currentVcc
end

function GetSystemConfiguration (elementId)
	assert(elementId)
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
	local success, systemConfiguration = pcall(PM.GetSystemConfiguration, ksProv, elementId)
	ksProv:Close()
	if not success then
		local errMsg = systemConfiguration
		LogWarning("[provisioning:models.vccs.GetSystemConfiguration] Error loading System Configuration for element '%s'.", elementId, errMsg)
		return false
	end
	return true, systemConfiguration
end

local function SendRpositoryRequest (action, vccInfo, vccId, newStorageConfig)
	local requestBody  = {}
	local reqError     = "S3_BUCKET_NOT_EXISTS"

	-- Check the action
	if action == "existsbk" or action == "createbk" then
		requestBody.S3Action    = action
		requestBody.S3PubKey    = vccInfo.StorageConfig.StorageLocationParams.S3PubKey
		requestBody.S3PrivKey   = vccInfo.StorageConfig.StorageLocationParams.S3PrivKey
		requestBody.S3Region    = vccInfo.StorageConfig.StorageLocationParams.S3Region
		requestBody.S3Bucket    = vccInfo.StorageConfig.StorageLocationParams.S3Bucket
		if action == "createbk" then
			reqError = "S3_BUCKET_CREATION_FAILED"
			requestBody.S3Lifecycle = vccInfo.StorageLifecycle.S3Files
		end
	elseif action == "createlc" then
		requestBody.S3Action    = action
		requestBody.STNewCfg    = tostring(newStorageConfig)
		requestBody.S3CurVcc    = vccId
		requestBody.S3Lifecycle = vccInfo.StorageLifecycle.S3Files
	else
		LogWarning("Unknow repository action request: %s", action)
		return false, "S3_UNKNOW_REQUEST_ACTION"
	end

	-- Do the request
	if inconcert and inconcert.environment and inconcert.environment.apps and inconcert.environment.apps.provisioning and
		inconcert.environment.apps.provisioning.repository and #inconcert.environment.apps.provisioning.repository
	then
		local repositoryServiceIP   = inconcert.environment.apps.provisioning.repository[1].ip
		local repositoryServicePORT = tostring(inconcert.environment.apps.provisioning.repository[1].port)
		local url = "http://" .. repositoryServiceIP .. ":" .. repositoryServicePORT .. "/inconcert/apps/repository/" .. action .. "/"
		local string_body  = json.encode(requestBody)
		local responseBody = {}
		local result, code, headers, httpStatus = socket.http.request {
			url = url,
			method = "POST",
			headers = {
				["Content-Type"] = "application/json",
				["Content-Length"] = tostring(#string_body),
			},
			sink = ltn12.sink.table(responseBody),
			source = ltn12.source.string(string_body)
		}
		if code ~= 200 then
			if code ~= 206 then
				LogWarning("[provisioning:models.vccs.Save] Error in request: %s for vcc '%s' -> '%s'", action, vccId, json.encode(responseBody))
				return false, reqError
			end
			LogWarning("[provisioning:models.vccs.Save] Error creating S3 bucket lifecycle rule for vcc '%s' -> '%s'", vccId, json.encode(responseBody))
			return false, "S3_BUCKET_LIFECYCLE_FAILED"
		end
	else
		LogWarning("[provisioning:models.vccs.Save] Error retrieving repository data for vcc '%s'", vccId)
		return false, "S3_NO_REPOSITORY_DATA"
	end

	return true, ""
end

function GetVCC (ksAgent, vccId, withCustomFields)
	assert(ksAgent)
	assert(vccId)
	withCustomFields = withCustomFields or false
	InitHelpersAgentServiceOptions()

	local success, vccInfo = pcall(PM.GetVCC, ksAgent, vccId, withCustomFields)
	if not success then
		local err = vccInfo
		return false, err
	end
	if utils.TableIsEmpty(vccInfo) then
		return false, "VCC DOES NOT EXIST"
	end
	if withCustomFields then
		-- las campanas 
		local vccCampaigns = utils.GetVCCCampaigns(vccId, ksAgent)
		--LogDebug("[provisioning.models.GetVCC] --> vcc: %s Campaigns: '%s'.", vccId, utils.JSON_Encode(vccCampaigns))
		local campaigns = {}
		for campaignId in pairs(vccCampaigns) do
			local isAllegroCampaign, callAllegroAccount = utils.IsAllegroCampaign(campaignId, ksAgent, vccId)
			--LogDebug("[provisioning.models.GetVCC]    --> campaignId: %s@%s, isAllegroCampaign: %s,  callAllegroAccount: '%s'.", campaignId, vccId, isAllegroCampaign, utils.JSON_Encode(callAllegroAccount))
			local campaign = {
				Id = campaignId,
				isAllegroCampaign = isAllegroCampaign
			}
			if isAllegroCampaign then
				campaign.callAllegroAccount = callAllegroAccount
			end
			campaigns[campaignId] = campaign
		end
		vccInfo.Campaigns = campaigns
	end
	return true, vccInfo
end

function UpdateVCC (ksAgent, vccId, vccInfo, ksProv)
	assert(ksAgent)
	assert(vccId)
	assert(vccInfo)
	assert(ksProv)
	-- desde el provisioningmanager el error llega en un JSON.
	local success, encodedErr = pcall(PM.UpdateVCC, ksAgent, vccId, vccInfo, ksProv)
	if not success then
		LogWarning("[provisioning:models.vccs.UpdateVCC] Error Updating VCC '%s' -> '%s'", vccId, encodedErr)
		return false, encodedErr
	end
	return true
end

function Save (vccId, vccInfo, ksAgent, ksProv)
	local s3ErrMsg           = ""
	local s3Success          = true
	local release_connection = false
	local currentProcMaxRate = 0

	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection("Agent")
		release_connection = true
	end
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		release_connection = true
	end
	local function CloseKeyspaces ()
		if not release_connection then
			return
		end
		if ksAgent then
			ksAgent:Close()
		end
		if ksProv then
			ksProv:Close()
		end
	end
	--------------------
	-- Clean Up de valores que llegan y no se tienen que guardar!!
	-- Estos son valores que llegan client side (campos calculados que NO se deben almacenar en cassandra)
	vccInfo.isNew = nil
	vccInfo.HasAllegroIntegration = nil
	vccInfo.HasWorkForceManagementIntegration = nil
	vccInfo.HasMarketingAndSalesIntegration = nil
	vccInfo.HasHelpDeskIntegration = nil
	vccInfo.Luis = nil
	vccInfo.QnAMaker = nil
	vccInfo.MaxOnlineUsers = nil
	--------------------
	-- Hay que convertir tablas a JSON, ya que la rocka provisioning manager falla NO los convierte.
	--  TODO: a esto quizás lo debería realizar la misma rocka.
	--     pero si lo dejamos por afuera a la conversión, garantizamos de que NO se guarde basura en la CF !! Hay que ver!!
	if type(vccInfo.AllegroIntegration) == "table" then
		vccInfo.AllegroIntegration = json.encode(vccInfo.AllegroIntegration)
	end
	-------------------
	
	-- Get Current values from VCC
	local vccSucess, currentVcc = GetVCC(ksAgent, vccId, false)
	-- If have Vcc data and TelephonyConfig values check process manual rate
	if vccSucess and currentVcc.TelephonyConfig then
		currentVcc.TelephonyConfig = json.decode(currentVcc.TelephonyConfig)
		if currentVcc.TelephonyConfig.ProcessMaxManualRate then
			local maxRate = tonumber(currentVcc.TelephonyConfig.ProcessMaxManualRate)
			if maxRate then
				currentProcMaxRate = maxRate
			end
		end
	end

	-- Check Repository service configs
	if HasRepositoryService() and vccInfo.StorageLifecycle and vccInfo.StorageConfig then
		local newStorageConfig     = false
		local currS3FilesLifecycle = ""
		local s3StorageBucket      = ""
		local newIsS3Config        = (vccInfo.StorageConfig and vccInfo.StorageConfig.StorageLocation and vccInfo.StorageConfig.StorageLocation == "S3") or false
		local s3StorageConfig      = {}
		local currentStorageConfig = {}

		-- Prepare Prefixes configuration before save
		if (vccInfo.InteractionPrefixConfig and vccInfo.InteractionPrefixConfig.TreeString) then
			local sep ="/"
			local prefixConfig = {}

			prefixConfig.Action     = 2
			prefixConfig.PrefixTree = {}
			for str in string.gmatch(vccInfo.InteractionPrefixConfig.TreeString, "([^"..sep.."]+)") do
				table.insert(prefixConfig.PrefixTree, str)
			end

			vccInfo.InteractionPrefixConfig = prefixConfig
		end

		-- Get current StorageLifecycle configuration if exists
		if currentVcc.StorageLifecycle then
			currentVcc.StorageLifecycle = json.decode(currentVcc.StorageLifecycle)
			currS3FilesLifecycle        = currentVcc.StorageLifecycle.S3Files
		end

		--Check current Storage Configuration
		if not currentVcc.StorageConfig then
			--Default configuration to check if it has changed
			currentStorageConfig.StorageLocation = "cassandra"
			currentStorageConfig.StorageLocationParams = {}
		else
			--current Storage config is the newest one, based on timestamp
			local maxTimestampIndex = 0
			local fullStorageConfig = json.decode(currentVcc.StorageConfig)
			for timestampIndex, config in pairs(fullStorageConfig) do
				if (type(timestampIndex) == "number" or tonumber(timestampIndex) ~= nil) and maxTimestampIndex < tonumber(timestampIndex) and type(config) == "table" then
					currentStorageConfig = config
					maxTimestampIndex = tonumber(timestampIndex)
				end
				-- Is the new configuration an S3 Storage? Then retrieve the last bucket used if not doing yet.
				if newIsS3Config and s3StorageBucket == "" then
					if config.StorageLocationParams.S3PubKey  == vccInfo.StorageConfig.StorageLocationParams.S3PubKey  and 
					   config.StorageLocationParams.S3PrivKey == vccInfo.StorageConfig.StorageLocationParams.S3PrivKey and 
					   config.StorageLocationParams.S3Region  == vccInfo.StorageConfig.StorageLocationParams.S3Region  and not config.StorageLocationParams.IsPredBucket
					then
						s3StorageBucket = config.StorageLocationParams.S3Bucket
					end
				end
			end
		end

		local currentS3Bucket         = currentStorageConfig.StorageLocationParams.S3Bucket      and currentStorageConfig.StorageLocationParams.S3Bucket      or ""
		local newS3Bucket             = vccInfo.StorageConfig.StorageLocationParams.S3Bucket     and vccInfo.StorageConfig.StorageLocationParams.S3Bucket     or ""
		local currentPredefinedBucket = currentStorageConfig.StorageLocationParams.IsPredBucket  and currentStorageConfig.StorageLocationParams.IsPredBucket  or false
		local newPredefinedBucket     = vccInfo.StorageConfig.StorageLocationParams.IsPredBucket and vccInfo.StorageConfig.StorageLocationParams.IsPredBucket or false
		
		-- Somethig change in the configuration
		if currentStorageConfig.StorageLocation ~= vccInfo.StorageConfig.StorageLocation or (currentStorageConfig.StorageLocation == "S3" and 
		   (currentStorageConfig.StorageLocationParams.S3PubKey ~= vccInfo.StorageConfig.StorageLocationParams.S3PubKey  or 
		   currentStorageConfig.StorageLocationParams.S3PrivKey ~= vccInfo.StorageConfig.StorageLocationParams.S3PrivKey or 
		   currentStorageConfig.StorageLocationParams.S3Region  ~= vccInfo.StorageConfig.StorageLocationParams.S3Region)) or
		   (currentPredefinedBucket ~= newPredefinedBucket or (newPredefinedBucket and (currentS3Bucket ~= newS3Bucket)))
		then
			local currentTimeStamp   = os.time()
			currentVcc.StorageConfig = currentVcc.StorageConfig and json.decode(currentVcc.StorageConfig) or {}

			if newIsS3Config then	-- The new configuration it's an S3.
				if newPredefinedBucket then       -- It's user predefined bucket created in AWS console.
					-- Check that new predefined bucket exists.
					s3Success, s3ErrMsg = SendRpositoryRequest("existsbk",vccInfo,vccId,newStorageConfig)

					if s3Success == false then
						CloseKeyspaces()
						return false, s3ErrMsg
					end

					-- S3 files lifecycle should be preserved if it exists or set to default.
					vccInfo.StorageLifecycle.S3Files = currS3FilesLifecycle ~= '' and currS3FilesLifecycle or "90"
				else
					if s3StorageBucket == "" then -- It's a new configuration, then create a new bucket and apply StorageLifecycle
						vccInfo.StorageConfig.StorageLocationParams.S3Bucket = CreateS3BucketName(vccId, currentTimeStamp)
						s3Success, s3ErrMsg = SendRpositoryRequest("createbk",vccInfo,vccId,newStorageConfig)

						if s3Success == false then
							if s3ErrMsg == "S3_BUCKET_LIFECYCLE_FAILED" then
								vccInfo.StorageLifecycle.S3Files = currS3FilesLifecycle
							else
								CloseKeyspaces()
								return false, s3ErrMsg
							end
						else
							LogInfo("Bucket %s created and lifecycle applied to %s days", vccInfo.StorageConfig.StorageLocationParams.S3Bucket, vccInfo.StorageLifecycle.S3Files)
						end
					else                          -- Copy the configuration with a new timestamp, but check if the bucket still exists, otherwise recreate it.
						vccInfo.StorageConfig.StorageLocationParams.S3Bucket = s3StorageBucket
						s3Success, s3ErrMsg = SendRpositoryRequest("createbk",vccInfo,vccId,newStorageConfig)

						if s3Success == false then
							if s3ErrMsg == "S3_BUCKET_LIFECYCLE_FAILED" then
								vccInfo.StorageLifecycle.S3Files = currS3FilesLifecycle
							else
								CloseKeyspaces()
								return false, s3ErrMsg
							end
						else
							LogInfo("Reusing bucket %s", vccInfo.StorageConfig.StorageLocationParams.S3Bucket)
						end
					end
				end
			end
			newStorageConfig = true
			currentVcc.StorageConfig[tostring(currentTimeStamp)] = vccInfo.StorageConfig -- Something change so we have to create a new record.
		end

		-- S3 Lifecycle configuration was changed and if new bucket was created and not fail to apply lifecycle
		if not newPredefinedBucket and currS3FilesLifecycle ~= vccInfo.StorageLifecycle.S3Files and s3Success == true then
			s3Success, s3ErrMsg = SendRpositoryRequest("createlc",vccInfo,vccId,newStorageConfig)
			-- Not save the Lifecycle configuration
			if s3Success == false then
				vccInfo.StorageLifecycle.S3Files = currS3FilesLifecycle
			else
				LogInfo("S3 storage lifecycle request completed successfully")
			end
		end

		vccInfo.StorageConfig = currentVcc.StorageConfig -- Set the new config or maintenance the old one.
	end
	
	-- Set Defaults to StorageLifecycle if not have
	if not vccInfo.StorageLifecycle then
		vccInfo.StorageLifecycle = {}
		vccInfo.StorageLifecycle.S3Files         = "90"
		vccInfo.StorageLifecycle.S3MetaData      = "90"
		vccInfo.StorageLifecycle.RepositoryCall  = "90"
		vccInfo.StorageLifecycle.RepositoryMail  = "90"
		vccInfo.StorageLifecycle.RepositoryOther = "90"
	end

	-- Save all configurations
	local success, encodedErr = UpdateVCC(ksAgent, vccId, vccInfo, ksProv)
	if not success then
		CloseKeyspaces()
		return success, encodedErr
	end

	-- Is not an allegro Vcc and have ProcessMaxManualRate and change it, then check if it's necessary for each process update it to the max value
	if not vccInfo.AllegroIntegration and vccInfo.TelephonyConfig and type(vccInfo.TelephonyConfig) == "table" and vccInfo.TelephonyConfig.ProcessMaxManualRate then
		local manualRate = tonumber(vccInfo.TelephonyConfig.ProcessMaxManualRate)
		if manualRate then
			if currentProcMaxRate ~= manualRate then
				success = UpdateProcessMaxRate(vccId, manualRate, ksAgent)
			end
		else
			success = false
		end
	end

	CloseKeyspaces()

	-- Check final result.
	if not success and not s3Success then -- If fail both: change ProcessMaxManualRate and Lifecycle
		return false, "S3_LIFECYLE_MAX_RATE_FAILED"
	elseif not success then               -- Fail to change ProcessMaxManualRate
		return false, "MANUAL_RATE_CHANGE_FAILED"
	elseif not s3Success then             -- Fail Lifecycle
		return false, s3ErrMsg
	end

	return success, errMsg
end

function Delete (vcc_id, ksProv, ksAgent)
	local release_agent_connection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection( "Agent" )
		release_agent_connection = true
	end
	local release_prov_connection = false
	if not ksProv then
		ksProv = utils.GetDatasource():CassandraConnection( "Provisioning" )
		release_prov_connection = true
	end
	InitHelpersAgentServiceOptions()

	local res, result = pcall(PM.DeleteVCC, ksProv, ksAgent, vcc_id)
	if not res or not result.success then
		LogWarning("[provisioning:models.vccs.Delete] Error deleting vccs '%s' -> '%s'.", vcc_id, result.errMsg)
	end

	if release_agent_connection then
		ksAgent:Close()
	end
	if release_prov_connection then
		ksProv:Close()
	end

	return result
end


--[[
	Esta funcion se encarga de limpiar las claves públicas y privadas de Captcha de la config de inconcert.
	Esto se realiza cuando la entidad del VCC va a ser enviada hacia el navegador, para que las claves de inconcert NO queden expuestas!!
]]--
function CleanUpInconcertCaptchaInfo (vccInfo)
	if not vccInfo or not vccInfo.Security then
		return
	end
	-- Security captcha
	if type(vccInfo.Security) == "string" then
		vccInfo.Security = json.decode(vccInfo.Security)
	end
	if not vccInfo.Security.captcha then
		return
	end

	-- Las claves publicas y privadas de la cuenta de inconcert
	local inconcertCaptcha = LoadInconcertCaptchaInfo()
	if not inconcertCaptcha or not inconcertCaptcha.keys then
		return
	end
	-- Si las claves configuradas en el VCC son iguales a las claves de inconcert, las limpio.
	if vccInfo.Security.captcha.publicKey == inconcertCaptcha.keys.publicKey then
		vccInfo.Security.captcha.publicKey = nil
	end
	if vccInfo.Security.captcha.secretKey == inconcertCaptcha.keys.secretKey then
		vccInfo.Security.captcha.secretKey = nil
	end
end


--[[
	Esta funcion se encarga de limpiar el password de acceso a servidor externo de autenticacion de la config de inconcert.
	Esto se realiza cuando la entidad del VCC va a ser enviada hacia el navegador, para que la clave NO quede expuesta!!
]]--
function CleanUpExternalAuthenticationPassword (vccInfo)
	if not vccInfo or not vccInfo.Security or not vccInfo.Security.authentication then
		return
	end
	local authType = vccInfo.Security.authentication.type
	if authType == "BASIC" or authType == "MFA" then	
		return
	end
	authType = authType:lower()
	if not vccInfo.Security[authType] then
		return
	end 
	vccInfo.Security[authType].adminPassword = m_externalAuthFakePassword
end

function GetMaxNumberConfigurationsAvailableByVcc (keyspace)

	local closeKeyspace
	if not keyspace then
		closeKeyspace = true
		keyspace = utils.GetDatasource():CassandraConnection("Agent")
	end
	--
	local vccsConfigLimits = utils.GetAllColumns {
		keyspace     = keyspace or "Agent",
		startKey = "",
		endKey = "",
		columnFamily = "VCCs",
		columns = {"ConfigLimits"}
	} or {}

	LogInfo("[provisioning:models.vccs.GetMaxNumerConfigurationsAvailableByVcc] vccs --> %s", json.encode(vccsConfigLimits))
	if closeKeyspace then
		keyspace:Close()
	end
	local _, envConfigLimit = GetSystemConfiguration("EnvironmentConfigLimits")
	
	if not envConfigLimit then
		return nil, "ERROR_SYSTEM_CONFIG"
	end

	for key, value in pairs(envConfigLimit) do
		envConfigLimit[key] = tonumber(value)
	end
	
	LogInfo("[provisioning:models.vccs.GetMaxNumerConfigurationsAvailableByVcc] envConfigLimit --> %s", json.encode(envConfigLimit))
	for vcc, value in pairs(vccsConfigLimits) do
		if value.ConfigLimits and vcc ~= "system" then
			envConfigLimit.MaxCampaigns = envConfigLimit.MaxCampaigns - value.ConfigLimits.MaxCampaigns
			envConfigLimit.MaxCampaignsUsers = envConfigLimit.MaxCampaignsUsers - value.ConfigLimits.MaxCampaignsUsers
			envConfigLimit.MaxAttsLevel = envConfigLimit.MaxAttsLevel - value.ConfigLimits.MaxAttsLevel
			envConfigLimit.MaxAttsLevelUsers = envConfigLimit.MaxAttsLevelUsers - value.ConfigLimits.MaxAttsLevelUsers
			envConfigLimit.MaxWorkgroups = envConfigLimit.MaxWorkgroups - value.ConfigLimits.MaxWorkgroups
			envConfigLimit.MaxWorkgroupsUsers = envConfigLimit.MaxWorkgroupsUsers - value.ConfigLimits.MaxWorkgroupsUsers
		end
	end
	return true, envConfigLimit
end

--[[
	Sample:
]]--
function LoadVCCSessionTimeouts (vccInfo)
	assert(type(vccInfo) == "table")
	return vccInfo.SessionTimeouts
end

--[[
	Sample:
	{
		keys = {
			publicKey = "6Lf-J0UUAAAAAGmR_IKYvcsFhjsu9v_AXeTnsg87",
			secretKey = "6Lf-J0UUAAAAAFFGy0pjfQgF0tZRtaGYOc2NuWdw"
		}
	}
]]--
function LoadInconcertCaptchaInfo ()
	local inconcertCaptcha = PM.LoadInconcertCaptchaInfo() -- Las claves publicas y privadas de la cuenta de inconcert
	return inconcertCaptcha
end

--[[
	Sample:
		{"captcha":{"enabled":"0","secretKey":"6Lf-J0UUAAAAAFFGy0pjfQgF0tZRtaGYOc2NuWdw","publicKey":"6Lf-J0UUAAAAAGmR_IKYvcsFhjsu9v_AXeTnsg87"}}
]]--
function LoadVCCSecurity (vccInfo)
	assert(type(vccInfo) == "table")

	local security = vccInfo.Security or {}
	if type(security) == "string" then
		security = json.decode(vccInfo.Security)
	end

	-- captcha
	local captchaConfig = security.captcha or {enabled = "0"}
	local captchaType =  nil
	if tostring(captchaConfig.enabled) == "1" then
		local inconcertCaptcha = LoadInconcertCaptchaInfo() -- Las claves publicas y privadas de la cuenta de inconcert
		-- En el caso en el cual NO se hayan definido claves, se deben asignar las claves de inconcert.
		--

		-- Public key
		local publicKey = captchaConfig.publicKey
		if not captchaConfig.publicKey or utils.alltrim(captchaConfig.publicKey) == "" then
			publicKey = inconcertCaptcha.keys.publicKey
			captchaType = "DEFAULT" -- Para no mantener cuando se crea el registro de auditoria
		end
		captchaConfig.publicKey = publicKey

		-- Secret key
		local secretKey = captchaConfig.secretKey
		if not captchaConfig.publicKey or utils.alltrim(captchaConfig.secretKey) == "" then
			secretKey = inconcertCaptcha.keys.secretKey
		end
		captchaConfig.secretKey = secretKey
	end
	pcall(PM.MakeCaptchaConfig, security, captchaConfig)

	if captchaType then
		security.captcha.type = captchaType
	end
	return security
end

--[[
	Sample:
	{Enabled = "1"}
]]--
function LoadVCCInteractionNotifications (vccInfo)
	assert(type(vccInfo) == "table")
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local _, isEnabledNotifications = pcall(PM.LoadVCCInteractionNotifications, vccInfo, mustAsk)
	return isEnabledNotifications
end

--[[
	Sample:
		{"HelpDesk":{"Enabled":"1","UserAuthentication":{"Password":"Inc0nc3r710","User":"admin","Scope":"FullControl","GrantType":"password","ClientId":"51507d0c6a714f54bb23375c5eb4cabc"},"Notifications":{"Url":"https://168.62.161.232:2443/ichelpdeskdesarrollo/notifications"},"Licenses":{"Users":"1"},"Server":{"HOST":"https://168.62.161.232:2443/ichelpdeskdesarrollo"}},"MarketingAndSales":{"Notifications":{"Url":"https://webapp01.inconcertcc.com/public/occ/process"},"AppRender":{"Url":"https://webapp01.inconcertcc.com/form/render"},"AccessToken":{"Token":"eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJ1c2VySW5zdGFuY2UiOiJkZWZhdWx0IiwidXNlcklkIjoiX19BdXhJbnRlZ3JhdGlvblVzZXIiLCJpYXQiOjE1ODgxNzg4NDF9.NaPbner0OXLcgc4Mt4iV_x5ZVrPmC2q3t6OpNRyRZb4","Url":"https://webapp01.inconcertcc.com/public/occ/get_session_token"},"Enabled":"1"},"WorkForceManagement":{"Enabled":"1","Notifications":{"Url":"https://wfm_site.com/notifications"}}}
]]--
function LoadVCCAppsIntegrations (vccInfo, isEnabledNotifications, widgetsDefinitions)
	assert(type(vccInfo) == "table")
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	pcall(PM.LoadVCCAppsIntegrations, vccInfo, isEnabledNotifications, mustAsk, widgetsDefinitions)
end

--------------
-- rasa
---------------
local rasa = {}

function Rasa ()
	return rasa
end

--[[
   Importa una App de LUIS a RASA
]]--
rasa.ImportLuisApp = function(info)
	assert(type(info) == "table")
	assert(info.vcc)
	assert(info.luisAppId)
	local success, response = PM.rasa.ImportLuisApp(info)
	return success, response
end

--[[
	Obtiene la lista Apps de LUIS que no han sido importadas a  RASA
]]--
rasa.LoadUnimportedLuisApps = function(info)
	assert(type(info) == "table")
	assert(info.vcc)
	return PM.rasa.LoadUnimportedLuisApps(info)
end

--------------
-- luis
---------------
local luis = {}

function Luis ()
	return luis
end

--[[ LUIS DEMO SUBSCRIPTIONS KEY ---------------]]--

luis.StartAssignDemoSubscriptionKeyToVCC = function (vccInfo)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local info = {
		vcc = vccInfo.vcc,
		subscriptionKey = vccInfo.demo and vccInfo.demo.subscriptionKey or nil
	}
	local success, err = PM.luis.StartAssignDemoSubscriptionKeyToVCC(info, mustAsk)
	return success, err
end

luis.StartUnassignDemoSubscriptionKeyOfVCC = function (vccInfo)
	local isDemo, demoSubscription = PM.luis.HasConfiguredDemoSubscriptionsInVcc(vccInfo.vcc)
	if not isDemo then
		return false, "NOT CONFIGURED DEMO SUBSCRIPTION"
	end

	local noQuestions = true -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	return PM.luis.ExecuteUnAssignSubscriptionKeyOfVCC({
			subscriptionKey = demoSubscription.subscriptionKey,
			vcc = vccInfo.vcc
		},
		isDemo,
		noQuestions
	)
end

--[[ LUIS TENANT ACCOUNTS ---------------]]--

luis.StartAssignTenantAccountToVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local success, err = PM.luis.StartAssignTenantAzureAccountToVCC(vccInfo, mustAsk)
	return success, err
end

luis.StartUnassignTenantAccountOfVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local success, err = PM.luis.StartUnAssignTenantAzureAccountOfVCC(vccInfo, mustAsk)
	return success, err
end

luis.LoadDefaultTenantAzureAccount = function ()
	return PM.luis.LoadDefaultTenantAzureAccount()
end

luis.RestoreDefaultTenantAzureAccountInVCC = function (vcc)
	assert(vcc)
	return PM.luis.RestoreDefaultTenantAzureAccountInVCC(vcc)
end

--[[ LANGUAGE STUDIO ACCOUNTS ---------------]]--  

luis.StartAssignLanguageStudioAccountToVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local success, err = PM.luis.StartAssignLanguageStudioAccountToVCC(vccInfo, mustAsk)
	return success, err
end

luis.RestoreDefaultLanguageStudioAccountInVCC = function (vcc)
	assert(vcc)
	return PM.luis.RestoreDefaultLanguageStudioAccountInVCC(vcc)
end

--[[ LUIS SUBSCRIPTIONS KEY ---------------]]--

luis.StartProcessAssignSubscriptionToVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	assert(vccInfo.azure)
	assert(vccInfo.azure.AzureSubscriptionId)
	assert(vccInfo.azure.ResourceGroup)
	assert(vccInfo.azure.AccountName)
	assert(type(vccInfo.subscriptions) == "table")

	local vccData = {
		vcc = vccInfo.vcc,
		AzureSubscriptionId = vccInfo.azure.AzureSubscriptionId,
		ResourceGroup       = vccInfo.azure.ResourceGroup,
		AccountName         = vccInfo.azure.AccountName,
		subscriptions       = vccInfo.subscriptions
	}

	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local success, err = PM.luis.StartProcessAssignSubscriptionToVCC(vccData, mustAsk)
	return success, err
end

luis.StartProcessUnAssignSubscriptionOfVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	return PM.luis.StartProcessUnAssignSubscriptionOfVCC(vccInfo, mustAsk)
end

--[[
]]--
luis.HasConfiguredDemoSubscriptionsInVcc = function (vcc)
	assert(vcc)
	return PM.luis.HasConfiguredDemoSubscriptionsInVcc(vcc)
end

--[[
]]--
luis.GetDefaultDemoSubscriptionKey = function ()
	return PM.luis.GetDefaultDemoSubscriptionKey()
end

--[[
]]--
luis.LoadVCCSubscriptions = function (vcc)
	assert(vcc)
	local encodedList, total, list = PM.luis.LoadVCCSubscriptions(vcc)
	return encodedList, total, list
end

--[[
]]--
luis.LoadVCCSubscriptionsWithoutDemo = function (vcc)
	assert(vcc)
	local encodedList, total, list = PM.luis.LoadVCCSubscriptionsWithoutDemo(vcc)
	return encodedList, total, list
end

--[[
]]--
luis.LoadVCCAzureAccounts = function (vcc)
	assert(vcc)
	return PM.luis.LoadVCCAzureAccounts(vcc)
end

--[[
]]--
luis.LoadVCCTenantAzureAccount = function (vcc)
	assert(vcc)
	return PM.luis.LoadVCCTenantAzureAccount(vcc)
end


--[[
]]--
luis.LoadVCCAzureLanguageStudioAccount = function (vcc) 
	assert(vcc)
	assert(type(vcc) == "string")
	return PM.luis.LoadVCCAzureLanguageStudioAccount(vcc)
end

--------------
-- QnA
---------------
local qna = {}

function QnA ()
	return qna
end

--[[ QnA SUBSCRIPTIONS KEY ---------------]]--

qna.StartProcessAssignSubscriptionToVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	assert(type(vccInfo.subscriptions) == "table")

	local vccData = {
		vcc = vccInfo.vcc,
		subscriptions = vccInfo.subscriptions
	}

	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local success, err = PM.qnaMaker.StartProcessAssignSubscriptionToVCC(vccData, mustAsk)
	return success, err
end

qna.StartProcessAssignSubscriptionToVCCV2 = function (vccInfo) -- fede
	assert(vccInfo)
	assert(vccInfo.vcc)
	assert(type(vccInfo.subscriptionsV2) == "table")

	local vccData = {
		vcc = vccInfo.vcc,
		subscriptionsV2 = vccInfo.subscriptionsV2
	}

	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	local success, err = PM.qnaMaker.StartProcessAssignSubscriptionToVCCV2(vccData, mustAsk)
	return success, err
end

qna.StartProcessUnAssignSubscriptionOfVCC = function (vccInfo)
	assert(vccInfo)
	assert(vccInfo.vcc)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	return PM.qnaMaker.StartProcessUnAssignSubscriptionOfVCC(vccInfo, mustAsk)
end

qna.StartProcessUnAssignSubscriptionOfVCCV2 = function (vccInfo) -- fede
	assert(vccInfo)
	assert(vccInfo.vcc)
	local mustAsk = false -- para que NO realice preguntas (prompt), ya que el mismo código se utiliza desde el pm-cli.lua
	return PM.qnaMaker.StartProcessUnAssignSubscriptionOfVCCV2(vccInfo, mustAsk)
end

--[[
]]--
qna.LoadVCCSubscriptions = function (vcc)
	assert(vcc)
	local encodedList, total, list = PM.qnaMaker.LoadVCCSubscriptions(vcc)
	return encodedList, total, list
end

qna.LoadVCCSubscriptionsV2 = function (vcc) -- fede
	assert(vcc)
	local encodedList, total, list = PM.qnaMaker.LoadVCCSubscriptionsV2(vcc)
	return encodedList, total, list
end

-------------
-- PACKAGES
---------------
local packages = {}

function Packages ()
	return packages
end

--[[ PACKAGES UPDATING IN VCC ---------------]]--

packages.UpdatePackagesInVCC = function (vcc, packagesToUpdate, ksAgent, ksProv)
	local releaseConnection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection("Agent")
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConnection = true
	end

	-- 1 - obtengo los currents packages asignados al vcc
	local modelPackages = require("apps.provisioning.models.packages")
	local currentVccPackages = {}
	local success, list = modelPackages.GetVCCPackages(vcc, ksProv)
	if success then
		for packageId in pairs(list or {}) do
			currentVccPackages[packageId:lower()] = packageId
		end
		--LogDebug("---> currentVccPackages: %s", json.encode(currentVccPackages))
	end

	InitHelpersAgentServiceOptions()

	local warningList = {assign = {}, remove = {}}
	-- 2 - asigno los Nuevos packages que me han llegado desde el request
	for _, package in pairs(packagesToUpdate or {}) do
		local lowerId = package.Id:lower()
		if not currentVccPackages[lowerId] then
		-- Si NO es un current packages en el VCC, lo asigno ya que es nuevo.
			local success, err = PM.AssignPackageToVCC(ksProv, ksAgent, package.Id, vcc)
			if not success then
				LogWarning("[provisioning:models.vccs.UpdatePackagesInVCC] + Error assigning package '%s' to vcc: '%s'. --> '%s'.", package.Id, vcc, json.encode(err))
				table.insert(warningList.assign, err)
			end
		else
		-- Si el package a agregar ya existe en la BD, lo elimino de la lista temporal.
			currentVccPackages[lowerId] = nil
		end
	end

	-- 3 - recorro los elementos de la lista temporal que quedaron y los libero del VCC, ya que los mismos no han llegado en el request.
	for _, packageId in pairs(currentVccPackages) do
		local success, err = PM.FreePackageFromVCC(ksProv, packageId)
		if not success then
			LogWarning("[provisioning:models.vccs.UpdatePackagesInVCC] - Error removing package '%s' to vcc: '%s'. --> '%s'.", packageId, vcc, json.encode(err))
			table.insert(warningList.remove, err)
		end
	end

	if releaseConnection then
		ksAgent:Close()
		ksProv:Close()
	end

	return true, warningList
end

--[[

]]--
packages.FreePackageFromVCC = function (vcc, ksAgent, ksProv)
	return packages.UpdatePackagesInVCC(vcc, nil, ksAgent, ksProv)
end

-------------
-- PEERS
---------------
local peers = {}

function Peers ()
	return peers
end

--[[ PEERS UPDATING IN VCC ---------------]]--

peers.UpdatePeersInVCC = function (vcc, peersToUpdate, assignationType, ksAgent, ksProv)
	local releaseConnection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection("Agent")
		ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
		releaseConnection = true
	end

	-- 1 - obtengo los currents peers asignados al vcc
	local modelPeers = require("apps.provisioning.models.peers")
	local currentVccPeers = {}
	local success, list = modelPeers.GetVCCPeers(vcc, ksProv)
	if success then
		for peerId, peer in pairs(list or {}) do
			currentVccPeers[peerId:lower()] = peer
		end
		--LogDebug("---> currentVccPeers: %s", json.encode(currentVccPeers))
	end

	InitHelpersAgentServiceOptions()

	local warningList = {assign = {}, remove = {}}
	-- 2 - asigno los Nuevos peers que me han llegado desde el request
	for _, peer in pairs(peersToUpdate or {}) do
		local lowerId = peer.Id:lower()
		if not currentVccPeers[lowerId] then
		-- Si NO es un current peers en el VCC, lo asigno ya que es nuevo.
			local tech, name = peers.SeparatePeer(peer.Id)
			local success, err = PM.AssignPeerToVCC(ksProv, ksAgent, tech, name, vcc, assignationType)
			if not success then
				LogWarning("[provisioning:models.vccs.UpdatePeersInVCC] + Error assigning peer '%s' to vcc: '%s'. --> '%s'.", peer.Id, vcc, json.encode(err))
				table.insert(warningList.assign, err)
			end
		else
		-- Si el peer a agregar ya existe en la BD, lo elimino de la lista temporal.
			if currentVccPeers[lowerId].assignationType ~= assignationType then
				local err = {
					code = "ASSIGNED_WITH_ANOTHER_TYPE",
					errMsg = "peer already assigned with another type",
					rawData = {
						peerID = peer.Id,
						vcc = peer.VCC
					}
				}
				LogWarning("[provisioning:models.vccs.UpdatePeersInVCC] + Error assigning peer '%s' to vcc: '%s'. --> '%s'.", peer.Id, vcc, json.encode(err))
				table.insert(warningList.assign, err)
			end
			currentVccPeers[lowerId] = nil
		end
	end

	-- 3 - recorro los elementos de la lista temporal que quedaron y los libero del VCC, ya que los mismos no han llegado en el request.
	for _, peer in pairs(currentVccPeers) do
		if peer.assignationType == assignationType then
			local success, err = PM.FreePeerFromVCC(ksProv, ksAgent, peer.tech, peer.name, vcc)
			if not success then
				LogWarning("[provisioning:models.vccs.UpdatePeersInVCC] - Error removing peer '%s/%s' to vcc: '%s'. --> '%s'.", peer.tech, peer.name, vcc, json.encode(err))
				table.insert(warningList.remove, err)
			end
		end
	end

	if releaseConnection then
		ksAgent:Close()
		ksProv:Close()
	end

	return true, warningList
end

--[[

]]--
peers.FreePeerFromVCC = function (vcc, assignationType, ksAgent, ksProv)
	return peers.UpdatePeersInVCC(vcc, nil, assignationType, ksAgent, ksProv)
end

--[[ PEERS RESETING IN VCC ---------------]]--

peers.ResetPeersInRedis = function (vcc)
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")
	local ksProv = utils.GetDatasource():CassandraConnection("Provisioning")
	local warningList = {assign = {}, remove = {}}
	local success, err = PM.ResetPeerToVCC(ksProv, ksAgent, vcc)
	ksAgent:Close()
	ksProv:Close()
	return true
end

peers.SeparatePeer = function (peerId)
	assert(type(peerId) == "string")
	local tech, name
	local separatorPosition = peerId:find("_")
	if separatorPosition then
		tech = peerId:sub(1, separatorPosition - 1)
		name = peerId:sub(separatorPosition + 1)
	end
	return tech, name
end

-------------
-- ALLEGRO
-------------
local allegro = {}

function Allegro ()
	return allegro
end

allegro.FreeAllPeersFromVCC = function(vcc_id, ksProv)
	local success, assigned_peers_to_all_vcc = pcall(PM.ListPeersByVCC, ksProv)
	if not success then
		local err_msg = assigned_peers_to_all_vcc
		LogWarning("[provisioning:models.vccs.allegro.FreeAllPeersFromVCC] Allegro - Error Free all peers from vcc: '%s' --> '%s'.", vcc_id, err_msg)
		return false
	end

	local assigned_peers_to_vcc = assigned_peers_to_all_vcc["VCC:" .. vcc_id]

	for _, peer_name in ipairs(assigned_peers_to_vcc or {}) do
		local separator_position = peer_name:find("_")
		if separator_position then
			local tech = peer_name:sub(1, separator_position - 1)
			local peer_id = peer_name:sub(separator_position + 1)
			local success, err_msg = pcall( PM.FreePeerFromVCC, ksProv, tech, peer_id )
			if not success then
				LogWarning("[provisioning:models.vccs.allegro.FreeAllPeersFromVCC] Allegro - Error free peer '%s' from vcc: '%s'. --> '%s'.", peer_name, vcc_id, err_msg)
			else
				LogInfo("[provisioning:models.vccs.allegro.FreeAllPeersFromVCC] Allegro - Free peer '%s' from vcc: '%s' successfully.", peer_name, vcc_id)
			end
		end
	end

	return true
end

-------------
-- LICENSES
-------------
local licenses = {}

function Licenses ()
	return licenses
end

licenses.LoadDefaultLicensesValues = function()
	return PM.LoadDefaultLicensesValues()
end

-------------
-- Accounts
-------------
local accounts = {}

function Accounts ()
	return accounts
end

local function LoadSpecificsAccounts ( accounts, accountType, accountId )
	if not accounts then
		return
	end

	if type( accounts ) == "string" then
		accounts = json.decode( accounts )
	end

	if accountType then
		accountType = accountType:upper()
		local accountsByType = accounts[ accountType ]
		if accountId then
			accountId = accountId:upper()
			return accountsByType and accountsByType[ accountId ]
		else
			return accountsByType
		end
	else
		return accounts
	end
end

accounts.GetVCCAccounts = function(vcc, accountType, accountId, keyspace)
	local accounts = utils.GetColumnValue {
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCs",
		key          = vcc,
		column       = "Accounts"
	}
	
	return LoadSpecificsAccounts( accounts, accountType, accountId )
end

-------------
-- Campaigns
-------------
local campaigns = {}

function Campaigns ()
	return campaigns
end

campaigns.GetVCCCampaigns = function(vcc, keyspace)
	local campaigns = utils.GetAllColumns {
		keyspace     = keyspace or "Agent",
		columnFamily = "VCCCampaigns",
		key          = vcc
	} or {}
	
	return campaigns
end

-------------
-- Recording Configuration
-------------
local recordingConfiguration = {}

function RecordingConfiguration ()
	return recordingConfiguration
end

recordingConfiguration.LoadDefaults = function()
	return utils.JSON_Decode(utils.GetSystemConfig("RECORDING").RecordingConfigurationDefaultValues)
end

function GetVCCTtsAsrEngines (vccId, ksAgent)
	--LogDebug("[provisioning:models.vccs.GetVCCTtsAsrEngines] Retrieving VCCTtsAsrEngines: '%s'.", vccId)

	local releaseConnection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection( "Agent" )
		releaseConnection = true
	end

	local success, currentVccTtsAsrEngines = pcall(PM.GetVCCTtsAsrEngines, ksAgent, vccId)
	--LogDebug("[provisioning:models.vccs.GetVCCTtsAsrEngines] Retrieved vcc '%s' -> success: '%s', success: '%s'", vccId, success, currentVccTtsAsrEngines)

	if not success then
		local errMsg = currentVccTtsAsrEngines
		LogWarning("[provisioning:models.vccs.GetVCCTtsAsrEngines] Error retrieving vcc '%s' -> '%s'", vccId, errMsg)
		return false
	end

	if releaseConnection then
		ksAgent:Close()
	end

	return success, currentVccTtsAsrEngines
end

function SaveVCCTtsAsrEngines (vccId, vccInfo, ksAgent)
	local release_connection = false
	if not ksAgent then
		ksAgent = utils.GetDatasource():CassandraConnection( "Agent" )
		release_connection = true
	end

	local success, errMsg = pcall(PM.UpdateVCCTtsAsrEngines, ksAgent, vccId, vccInfo)
	if not success then
		LogWarning("[provisioning:models.vccs.SaveVCCTtsAsrEngines] Error saving vccs '%s' -> '%s'", vccId, errMsg)
	end

	if release_connection then
		ksAgent:Close()
	end

	return success, errMsg
end

-------------
-- Groups
-------------
local groups = {}

function Groups ()
	return groups
end

groups.GetVCCGroups = function(vcc, keyspace)
	local allGroups = utils.GetAllColumns {
		keyspace     = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		key          = vcc
	} or {}
	local systemGroups = utils.GetAllColumns {
		keyspace     = keyspace or "Agent",
		columnFamily = "Group",
		superColumns = true,
		key          = "system"
	} or {}
	for groupId, group in pairs(systemGroups) do
		if groupId:lower() ~= "system"
			and not allGroups[groupId]
		then
			allGroups[groupId] = group
		end
	end
	return allGroups
end

-------------
-- LDAP
-------------
local ldap = {}

function Ldap ()
	return ldap
end

ldap.GetLdapGroups = function(vcc, server, baseDN, adminUserDN, password, useTLS)
	if password == m_externalAuthFakePassword then
		password = GetCurrentExternalAuthPassword(vcc)
	end
	return pcall(PM.GetLdapGroups, server, baseDN, adminUserDN, password, useTLS)
end

-------------
-- Azure AD
-------------
local azuread = {}

function AzureAD ()
	return azuread
end

azuread.GetAzureADGroups = function(openid_configuration_url, client_id, client_secret)
	if client_secret == m_externalAuthFakePassword then
		client_secret = GetCurrentExternalAuthPassword(vcc)
	end
	return pcall(PM.GetAzureADGroups, openid_configuration_url, client_id, client_secret)
end

azuread.GetOpenIdConfiguration = function(openid_configuration_url)
	return pcall(PM.GetOpenIdConfiguration, openid_configuration_url)
end

function GetExternalAuthFakePassword()
	return m_externalAuthFakePassword
end

function GetCurrentExternalAuthPassword (vccId)
	InitHelpersAgentServiceOptions ()
	local ksAgent = utils.GetDatasource():CassandraConnection( "Agent" )
	local success, currentVcc = GetVCC(ksAgent, vccId)
	ksAgent:Close()
	if not success then
		LogError("[provisioning:models.GetCurrentExternalAuthPassword] Recovering current vcc config: '%s'", currentVcc)
		return ""
	end
	local currentVccSecurity = json.decode(currentVcc.Security)
	local authType = currentVccSecurity.authentication.type
	if authType == "BASIC" or authType == "MFA" then
		LogError("[provisioning:models.GetCurrentExternalAuthPassword] vcc '%s' authentication type is not external", vccId)
		return ""
	end
	authType = authType:lower()
	return currentVccSecurity[authType].adminPassword
end

function GetVccWithLicenseInfo(vcc)		
	local options = utils.Get_m_helpers_agentservice_options()
	return m_helpers.agentservice.GetVccWithLicenseInfo(options, vcc)
end	

function GetAvailableLicenseResources()
	local options = utils.Get_m_helpers_agentservice_options()
	return m_helpers.agentservice.GetAvailableLicenseResources(options, "")
end

function CompanyIsEditable()
	local options = utils.Get_m_helpers_agentservice_options()
	return m_helpers.agentservice.CompanyIsEditable(options, "")
end

function CompanyIsRequired()
	local options = utils.Get_m_helpers_agentservice_options()
	return m_helpers.agentservice.CompanyIsRequired(options, "")
end

function ValidateCompany(company)
	local options = utils.Get_m_helpers_agentservice_options()
	return m_helpers.agentservice.ValidateCompany(options, company)
end

function GetS3Regions ()
	local regions = {}
	table.insert(regions, { id = "us-east-1", description = "US East (N. Virginia)"})
	table.insert(regions, { id = "us-east-2", description = "US East (Ohio)"})
	table.insert(regions, { id = "us-west-1", description = "US West (N. California)"})
	table.insert(regions, { id = "us-west-2", description = "US West (Oregon)"})
	table.insert(regions, { id = "eu-west-1", description = "EU (Ireland)"})
	table.insert(regions, { id = "eu-west-2", description = "EU (London)"})
	table.insert(regions, { id = "eu-west-3", description = "EU (Paris)"})
	table.insert(regions, { id = "eu-central-1", description = "EU (Frankfurt)"})
	table.insert(regions, { id = "eu-north-1", description = "EU (Stockholm)"})
	table.insert(regions, { id = "eu-south-1", description = "EU (Milan)"})
	table.insert(regions, { id = "eu-south-2", description = "EU (Spain)"})
	table.insert(regions, { id = "ap-east-1", description = "Asia Pacific (Hong Kong)"})
	table.insert(regions, { id = "ap-south-1", description = "Asia Pacific (Mumbai)"})
	table.insert(regions, { id = "ap-southeast-1", description = "Asia Pacific (Singapore)"})
	table.insert(regions, { id = "ap-southeast-2", description = "Asia Pacific (Sydney)"})
	table.insert(regions, { id = "ap-northeast-1", description = "Asia Pacific (Tokyo)"})
	table.insert(regions, { id = "ap-northeast-2", description = "Asia Pacific (Seoul)"})
	table.insert(regions, { id = "ap-northeast-3", description = "Asia Pacific (Osaka)"})
	table.insert(regions, { id = "sa-east-1", description = "South America (Sao Paulo"})
	table.insert(regions, { id = "cn-north-1", description = "China (Beijing)"})
	table.insert(regions, { id = "cn-northwest-1", description = "China (Ningxia)"})
	table.insert(regions, { id = "ca-central-1", description = "Canada (Central)"})
	table.insert(regions, { id = "me-south-1", description = "Middle East (Bahrain)"})
	table.insert(regions, { id = "af-south-1", description = "Africa (Cape Town)"})
	table.insert(regions, { id = "us-gov-west-1", description = "AWS GovCloud (US-West)"})
	table.insert(regions, { id = "us-gov-east-1", description = "AWS GovCloud (US-East)"})
	table.insert(regions, { id = "us-iso-east-1", description = "US ISO East"})
	table.insert(regions, { id = "us-isob-east-1", description = "US ISOB East (Ohio)"})
	table.insert(regions, { id = "us-iso-west-1", description = "US ISO West"})
	return true, regions
end	

function GetallVccs()
	return utils.GetAllColumns({
		keyspace     = "Agent",
		columnFamily = "VCCs",
		startKey     = "",
		endKey       = ""
	}) or {}
end
