--[[!
-- @file inconcert/apps/reports/exporter.lua
-- @brief Report exporter wrapper
--]]

local lfs = require "lfs"
module((...), package.seeall)

-- cached array containing all supported types
local exporter_types = { }
-- map of exporter interfaces by type
local exporters = {}

--[[
-- @brief Called when the applications starts, loads all exporters from disk
--]]
function load_exporters(path)
	LogInfo("Loading report exporters from path %s", path)

	for file in lfs.dir(path) do
		if file:match(".*.lua$") then
			local file_path = string.format("%s/%s", path, file)

			local interface = dofile(file_path)

			LogInfo ("Loaded %s report exporter, supports streaming: %s, can be used: %s",
					string.upper(interface.type),
					interface.streamable and "yes" or "no",
					interface.practicable and "yes" or "no"
				)

			if interface.practicable then
				exporters[interface.type] = interface
				table.insert(exporter_types, interface.type)
			end
		end
	end
end

--[[
-- @return Returns The list of supported types
--]]
function get_types()
	return exporter_types
end

--[[
-- @brief Opens a new exporter instance
--]]
function open_exporter (type, report, _T, ...)
	local exporter_interface = exporters[type]
	if not exporter_interface then
		return nil, "Exporter type not found"
	end
	return exporter_interface.factory(report, _T, ...)
end
