--[[
-- stuff
--]]

module((...), package.seeall)


local identity;
local predNot;
local predProps;

---
-- Helpers
---

local function makeEqual(a)
	return function(b)
		return a == b;
	end
end

local function resolvePredicate(pred)
	if type(pred) == "function" then
		return pred;
	end

	if type(pred) == "table" then
		return predProps(pred);
	end

	return makeEqual(pred);
end

local function makeIf(pred)
	return function(v, k)
		if pred(v, k) then
			return v;
		end
	end
end

--
-- Real functions
--

local function each(tbl, iterator)
	for k,v in pairs(tbl) do
		if iterator(v, k) == false then
			return false;
		end
	end
	return true;
end

local function some(tbl, pred)
	local idx;
	pred = resolvePredicate(pred);
	return not each(tbl, function(v, k)
		if pred(v, k) then
			idx = k;
			return false; -- break the loop
		end
	end), idx;
end

local function map(tbl, iterator, includeNil)
	if includeNil == nil then includeNil = true end
	local result = {};
	each(tbl, function(v, k)
		local elm = iterator(v, k);
		if includeNil or elm ~= nil then
			table.insert(result, elm);
		end
	end);
	return result;
end

local function select(tbl, pred)
	return map(tbl, makeIf(resolvePredicate(pred)), false);
end

local function reject(tbl, pred)
	return map(tbl, makeIf(predNot(resolvePredicate(pred))), false);
end

local function reduce(tbl, iterator, memo)
	if not memo then
		memo = #tbl > 0 and type(tbl[1]) == "string" and "" or 0;
	end
	each(tbl, function(v, k)
		memo = iterator(memo, v, k);
	end);
	return memo;
end

local function contains(tbl, item)
	return some(tbl, makeEqual(item));
end

local function concat(tbl1, tbl2, mutate)
	local concatImpl = function(target, src)
		each(src, function(v)
			table.insert(target, v);
		end);
		return target;
	end
	return concatImpl(mutate and tbl1 or map(tbl1, identity), tbl2);
end

local function pick(tbl, ...)
	local function myReduce(args, memo)
		return reduce(args, function(memo, v)
			if type(v) == "table" then
				return myReduce(v, memo);
			end
			memo[v] = tbl[v];
			return memo;
		end, memo)
	end

	return myReduce({...}, {});
end

local function keys(tbl)
	return map(tbl, function(_, k)
		return k;
	end);
end

local function values(tbl)
	return map(tbl, function(e)
		return e;
	end);
end

local function extend(tbl, ...)
	local args = {...}
	each(args, function(tblN)
		each(tblN, function(v, k)
			tbl[k] = v
		end)
	end)
	return tbl
end

local function remove(tbl, key)
	local val = tbl[key]
	tbl[key] = nil
	return val
end

-- for generators
local function firsty(gen, t, i) -- i should be nil
	return gen(t, i);
end

-- for tables
local function first(tbl)
	return next(tbl);
end

--
--

identity = function(...)
	return ...;
end

predNot = function(fn)
	return function(...)
		return not fn(...);
	end
end

predProps = function(props)
	return function(v1)
		return each(props, function(v2, k)
			return v1[k] == v2;
		end);
	end
end


return {
	--[[
	-- _.each({ "one", "two" }, print)  => prints "one 1" and "two 2"
	--]]
	each     = each,

	--[[
	-- _.every({ 4, 5, 6 }, function(v) return v > 2 end)  => true
	-- _.every({ 4, 5, 6 }, function(v) return v > 4 end)  => false
	--]]
	every    = each,

	--[[
	-- _.some({ 4, 5, 6 }, function(v) return v > 4 end)  => true, 2
	--]]
	some     = some,

	--[[
	-- _.map({ 4, 5, 6 }, function(v) return v * 2 end)  => { 8, 10, 12 }
	--]]
	collect  = map,
	map      = map,

	--[[
	-- _.select({ 4, 5, 6 }, function(v) return v > 4 end)  => { 5, 6 }
	--]]
	filter   = select,
	select   = select,

	--[[
	-- _.reject({ 4, 5, 6 }, function(v) return v > 4 end)  => { 4 }
	-- _.reject({ 4, 5, 6 }, function(v) return v > 4 end)  => { 4 }
	--]]
	reject   = reject,
	without  = reject,

	--[[
	-- _.reduce({ 4, 5, 6 }, function(m, v) return m + v end)      => 15
	-- _.reduce({ 4, 5, 6 }, function(m, v) return m + v end, 10)  => 25
	--]]
	inject   = reduce,
	foldl    = reduce,
	reduce   = reduce,

	contains = contains,
	concat   = concat,
	pick     = pick,
	keys     = keys,
	values   = values,
	extend   = extend,
	remove   = remove,

	firsty   = firsty, -- for iterators
	first    = first,  -- for tables

	predNot   = predNot,
	predProps = predProps,
	identity  = identity,
};
