/*jshint bitwise:true, browser:true, curly:true, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, esversion: 9 */
/*jshint newcap:false, quotmark:false */
/*global $, _, EventEmitter, inConcert*/

this.inConcert = this.inConcert || {};
this.inConcert.reports = this.inConcert.reports || {};

(function () {
	"use strict";

	const app_name = "designer";
	const format = inConcert.Utils.format;
	const helpers = inConcert.reports.helpers;

	const get_report_type_desc = function(type) {
		let desc;
		if (type) {
			const reports_desc = {
				"detail"    : "<%=_T('Detail')%>",
				"livedetail": "<%=_T('Live Detail')%>",
				"summary"   : "<%=_T('Summary')%>"
			};
			desc = reports_desc[type.toLowerCase()];
		}
		return desc || "<%=_T('Unrecognized')%>";
	};

	const get_translated_scope = function(scope = "") {
		const translatedScopes = helpers.listTranslatedScopes();
		return translatedScopes[scope] || scope;
	};

	/*!
	 * Events:
	 * \li render(page, callback)
	 * \li ready()
	 * \li exit(status, message)
	 * \li status(message)
	 * \li error(message)
	 */
	function FormBase(page_url, $container, context)
	{
		const self = $.extend(this, {
			pageSrc       : page_url,
			container     : $container,
			context       : context,
			renderComplete: false,
			hNavigation   : null,
			innerForm     : null
		});

		if (self.context && self.context.saveButton) {
			self.once('exit', function() {
				self.context.saveButton.prop('disabled', false);
			});
		}
	}

	/* See event list above */
	inConcert.Utils.inherit(FormBase, EventEmitter);

	/*! virtual - called to initialize the GUI */
	FormBase.prototype.onRender = function(callback) {
		callback(true, "");
	};

	/* virtual - Called after onRender, when the form is actually shown */
	FormBase.prototype.onShow = function() {
		// stub
	};

	/*! virtual - called when Confirm button is pressed */
	FormBase.prototype.onConfirm = function(callback) {
		callback(true, "");
	};

	/*! virtual - called when Cancel button is pressed */
	FormBase.prototype.onCancel = function(callback) {
		callback(true, "");
	};

	/*! virtual - called before closing the form */
	FormBase.prototype.onExit = function(status, callback) {
		callback(true, "");
	};

	/*! render, display the form */
	FormBase.prototype.show = function() {
		var self = this;
		self.emit("render", self.pageSrc, function(content, error) {
			if (content) {
				self.container.html(content);
				self.start();
			} else {
				self.emit("exit", "error", format("<%=_T('Form rendering failed with error {0}')%>", error));
			}
		});
	};

	/*! set form's title, subtitle, and tagline (small text below the subtitle) */
	FormBase.prototype.setTitle = function(title, subtitle, tagline) {
		$('#reports_form_title').html(title);
		$('#reports_form_section_title').text(subtitle || "");
		$('#reports_form_tagline').text(tagline || "");
	};

	FormBase.prototype.hasPermissionToSave = function() {
		if ( !this.permissions || !this.permissions.creation ) {
			return true;
		}
		return inConcert.permissions.to_action( 'reports' , this.permissions.creation, 'SAVE' );
	};

	/*! Called when form's Cancel button is pressed */
	FormBase.prototype.cancel = function() {
		var self = this;
		if (self.innerForm !== null) {
			return self.innerForm.cancel();
		}
		self.onCancel(function(canceled, message) {
			if ( canceled ) {
				self.emit("exit", "cancel");
			} else {
				self.emit("error", message);
			}
		});
	};

	/*! Called when form's Confirm button is pressed */
	FormBase.prototype.confirm = function() {
		var self = this;
		if (self.innerForm !== null) {
			return self.innerForm.confirm();
		}
		self.onConfirm(function(confirmed, message) {
			if ( confirmed ) {
				self.emit("exit", "done", message);
			} else {
				self.emit("error", message);
			}
		});
	};

	/*! Called to close the form with some \p status and an optional \p message */
	FormBase.prototype.close = function(status, message) {
		var self = this;
		self.onExit(status, function(exited, errorThrown) {
			if ( exited ) {
				self.emit("exit", status, message);
			} else {
				self.emit("error", errorThrown);
			}
		});
	};

	/*! Called internally to render the form */
	FormBase.prototype.start = function() {
		var self = this;
		self.onRender(function(success, errorThrown) {
			if ( success ) {
				self.emit("ready");
				_.defer( function() {
					self.renderComplete = true;
					self.onShow();
				});
			} else {
				self.close("error", errorThrown);
			}
		});
	};

	/*! Called internaly to initialize the Wizard in case it's supported */
	FormBase.prototype.initNavigation = function(params) {
		var self = this;

		params.enterPage = params.onEnterPage || $.noop; delete params.onEnterPage;
		params.leavePage = params.onLeavePage || $.noop; delete params.onLeavePage;

		self.hNavigation = helpers.setupNavigation(self.container, _.extend({
			onLeavePage: function(pageNumber, $pageElem, direction) {
				var success = true;
				if (direction === "forward") {
					var result = self.validateSection($pageElem, _.bind(self.isFieldEnabled, self));
					if (!result.success) {
						self.emit("error", result.errorMessage);
					}
					success = result.success;
				}
				if (success) {
					params.leavePage(pageNumber, $pageElem, direction);
				}
				return success;
			},
			onEnterPage: function(pageNumber, $pageElem, allVisited) {
				self.context.saveButton.prop('disabled', !allVisited);
				params.enterPage(pageNumber, $pageElem, allVisited);
			}
		}, params));
	};

	/*
	 * Called to update the style of the label associated with the field
	 * Assume the label is right before the element
	 */
	FormBase.prototype.onFieldDisabledChange = function(disabled, $field) {
		if (disabled) {
			$field.prev('.formLabel').addClass('formLabel-ds');
		} else {
			$field.prev('.formLabel').removeClass('formLabel-ds');
		}
	};

	FormBase.prototype.isFieldEnabled = function($field) {
		const self = this;

		do {
			if ($field.length === 0 || $field.attr('ic-data-bind-disabled')) {
				return false;
			}

			$field = $field.parent();
		} while($field[0] !== self.container[0]);

		return true;
	};

	/*!
	 * This function comes handy to change the interface based in an attribute.
	 *
	 * Mostly used when the user changes an option that makes the rest of the form change
	 */
	FormBase.prototype.switchUseCaseByAttr = function(attrName, attrValue, $container) {
		const self = this;
		$container = $container || self.container;

		const $elements = $container.find('[' + attrName + ']');

		const $elementsToShow = $elements.filter('[' + attrName + '="'  + attrValue + '"]')
			.removeAttr('ic-data-bind-disabled');

		const $elementsToHide = $elements.filter('[' + attrName + '!="' + attrValue + '"]')
			.attr('ic-data-bind-disabled', "true");

		return helpers.niceFading({
			$show: $elementsToShow,
			$hide: $elementsToHide
		});
	};

	/*!
	 * Validates a single section, i.e. not the entire container
	 */
	FormBase.prototype.validateSection = function($section, iterator) {
		$section.find('.formFieldValidation').hide();

		const result = helpers.readDataModel($section, iterator);

		if (!result.success) {
			result.missing_fields.forEach(function($field) {
				$field.next('.formFieldValidation').show();
				$field.focus();
			});

			if (result.missing_fields.length === 1) {
				const $field = result.missing_fields[0];
				result.errorMessage = $field.attr('ic-data-bind-error-message') || format(
					"<%=_T('Field {0} is required')%>",
					$field.attr('ic-data-bind-alias') || $field.attr('ic-data-bind')
				);
			} else {
				result.errorMessage = "<%=_T('Please fill up all mandatory fields')%>";
			}

			if ((result.alpha || []).length === 1) {
				const $field = result.alpha[0];
				result.errorMessage = $field.attr('ic-data-bind-error-message') || format(
					"<%=_T('Only letters, numbers, underscore are allowed in the {0}.')%>",
					$field.attr('ic-data-bind-alias') || $field.attr('ic-data-bind')
				);
			} else {
				result.errorMessage = "<%=_T('Please check the indicated fields.')%>";
			}
		}

		return result;
	};

	/*!
	 * Called internally to validate the form, mostly to check for required fields.
	 * It also build the list of fields the user has entered
	 * \return \code { success: boolean, errorMessage: string, fields: object } \endcode
	 */
	FormBase.prototype.validate = function(iterator) {
		return this.validateSection(this.container, iterator);
	};

	FormBase.prototype.readonlyAllCtrls = function() {
		var $form = $( "#reports_form_data_container" ); 
		// readonly all textbox
		$form.find(".formData input")
			.prop("disabled", true)
			.addClass("readonly");

		// disabled all combos
		$form.find(".formData select")
			.prop("disabled", true)
			.addClass("readonly");

		// readonly all textarea
		$form.find(".formData textarea")
			.prop("disabled", true)
			.addClass("readonly");

		// remove all required class
		$form.find(".formData label")
			.removeClass("formLabel-rq")
			.addClass("formLabel-ds readonly");
	};

	FormBase.prototype.applyPermissions = function() {
		var $confirm = $('#reports_form_button_confirm');
		var $cancel = $('#reports_form_button_cancel');
		var $required_message = $('#reports_form_required_message');

		if ( this.hasPermissionToSave() ) {
		// has permissions
			$confirm.show();
			$cancel.val( "<%=_T('Cancel')%>" );
			$required_message.show();
			return;
		}

		// not has permissions
		$confirm.hide();
		$cancel.val( "<%=_T('Close')%>" );
		$required_message.hide();
		this.readonlyAllCtrls();
	};

	/**************************************/
	/**************************************/

	/*!
	 * Reports Designer app
	 */
	function ReportsDesigner( dashboard )
	{
		var self = $.extend( this,  {
			dashboard       : dashboard                  ,
			controller      : dashboard.get_controller() ,
			viewGrid        : null                       ,
			currentForm     : null                       ,
			currentView     : null
		});

		self.initialize_once = _.once(function() {
			self.viewGrid = $('#reports_view_grid');
			self.viewGridBody = self.viewGrid.children('tbody');
			self.viewGridHead = self.viewGrid.children('thead');
			self.setupEventHandlers();
		});

		// TODO: Use templates
		// note: row_data is passed AS-IS to appendRow()
		var insertRowOrdered = function(name, row_data) {
			var row_index;
			if (self.rowCount === 0) {
				$('#reports_view_not_results_found').hide();
				self.viewGrid.show();
			} else {
				self.viewGridBody.children().each(function() {
					var $row = $(this);
					if (row_index) {
						$row.children('td:first').text(++row_index);
						return;
					}
					var compareResult = $row.children('[ic-data-bind="name"]').text().localeCompare(name, undefined, {sensitivity: "base"});
					if ( compareResult > 0 ) {
						row_index = parseInt($row.children('td:first').text(), 10);
						self.appendRow(row_data)
							.hide()
							.detach()
							.insertBefore($row)
							.fadeIn("slow")
							.children('td:first').text(row_index);
						$row.children('td:first').text(++row_index);
					}
				});
			}

			if (!row_index) {
				self.appendRow(row_data).hide().fadeIn("slow");
			}

			if (!self.updateViewFilter()) {
				self.initializePagination(
					_.map(self.viewGridBody.children(), $)
				);
			}
		};

		var updateRow = function(id_attr, id, data) {
			const $row = self.viewGridBody.children('[' + id_attr + '="' +  id + '"]');
			let full_text_search = '';
			const $columns = $row.children('[ic-data-bind]');
			$columns.each(function() {
				let columnName = $(this).attr('ic-data-bind');
				let newval = data[columnName];
				if (newval) { // 'scope' is not valid when updating composite metrics
					full_text_search += newval.replace(/\W/g, '').toLowerCase();
					$(this).text(newval);
				}
			});
			$row.attr('rd-text-search', full_text_search);
			self.updateViewFilter();
		};

		var deleteRow = function(id_attr, id) {
			const $row_to_delete = self.viewGridBody.children('[' + id_attr + '="' +  id + '"]');
			$row_to_delete
				.fadeOut("slow").promise().done(function() {
					var $this = $(this);
					var rowIndex = parseInt($this.children('td:first').text(), 10);
					$this.nextAll().each(function() {
						$(this).children('td:first').text(rowIndex++);
					});
					self.m_paginator.remove($this);
					if ( --self.rowCount === 0 ) {
						self.viewGrid.hide();
						$('#reports_view_not_results_found')
							.text(self.currentView.noRecordsMsg)
							.show();
					}
				}
			);
		};

		self._views = {
			reports: {
				title       : "<%=_T('Custom Reports')%>",
				button_text : "<%=_T('Create report')%>",
				loading_msg : "<%=_T('Loading Custom Reports')%>",
				new_text    : "<%=_T('New Report')%>",
				i_class_name : "far fa-file-chart-line",
				permissions : {
					view    : "view_custom_reports",
					create  : "save_custom_reports",
					erase   : "delete_custom_reports"
				},
				left_section: "reports",
				grid_headers: [ "<%=_T('Report name')%>", "<%=_T('Report type')%>" ],
				noRecordsMsg: "<%=_T('No custom reports found.')%>",
				init_once   : _.once(function() {
					self.controller.on("reportAdded", function(report) {
						insertRowOrdered(report.name, {
							columns: [
								{ 'ic-data-bind': "name", text: report.name },
								{ text: get_report_type_desc(report.type)   }
							],
							attributes: {
								'rd-report-id'  : report.id,
								'rd-report-type': report.type
							}
						});
					}).on("reportUpdated", function(report) {
						updateRow('rd-report-id', report.id, report);
					}).on("reportDeleted", function(report) {
						deleteRow('rd-report-id', report.id);
					});
				}),
				enter: function(callback) {
					this.init_once();
					_.defer(_.bind(self.listReports, self), callback);
				},
				leave: function() {
				},
				on_new:  function() {
					self.renderReport(null);
				},
				on_edit: function($row) {
					self.editEntity({
						entity: "Report",
						loader: _.bind(self.controller.loadReport, self.controller, $row.attr('rd-report-type'), $row.attr('rd-report-id')),
						render: _.bind(self.renderReport, self)
					});
				},
				on_delete: function($row) {
					self.deleteEntity({
						entity : "Report",
						deleter: _.bind(self.controller.deleteReport, self.controller, $row.attr('rd-report-id'))
					});
				}
			},
			metrics: {
				title       : "<%=_T('Metrics')%>",
				button_text : "<%=_T('Create metric')%>",
				loading_msg : "<%=_T('Loading Metrics')%>",
				new_text    : "<%=_T('New Metric')%>",
				i_class_name: "far fa-chart-bar",
				permissions : {
					view    : "view_custom_metrics",
					create  : "save_custom_metrics",
					erase   : "delete_custom_metrics"
				},
				left_section: "metrics",
				grid_headers: [ "<%=_T('Name')%>", "<%=_T('Scope')%>", "<%=_T('Description')%>" ],
				//noRecordsMsg: "<%=_T('No custom metrics found.')%>",
				noRecordsMsg: "<span class='messageInfoText1'><%=_T('No results found.')%></span> <span class='messageInfoText2'><%=_T('There are no items created or matching your search yet.')%></span>",
				init_once   : _.once(function() {
					self.controller.on("metricAdded", function(metric) {
						insertRowOrdered(metric.name, {
							columns: [
								{ 'ic-data-bind': "name",        text: metric.name || metric.id },
								{ 'ic-data-bind': "scope",       text: metric.scope || (metric.composite && "<%=_T('Composite')%>") },
								{ 'ic-data-bind': "description", text: metric.description || "" }
							],
							attributes: {
								'rd-metric-id': metric.id
							}
						});
					}).on("metricUpdated", function(metric) {
						updateRow('rd-metric-id', metric.id, metric);
					}).on("metricDeleted", function(metric) {
						deleteRow('rd-metric-id', metric.id);
					});
				}),
				enter: function(callback) {
					this.init_once();
					_.defer(_.bind(self.listMetrics, self), callback);
				},
				leave: function() {
				},
				on_new:  function() {
					self.renderMetric(null);
				},
				on_edit: function($row) {
					self.editEntity({
						entity: "Metric",
						loader: _.bind(self.controller.loadMetric, self.controller, $row.attr('rd-metric-id')),
						render: _.bind(self.renderMetric, self)
					});
				},
				on_delete: function($row) {
					self.deleteEntity({
						entity : "Metric",
						deleter: _.bind(self.controller.deleteMetric, self.controller, $row.attr('rd-metric-id'))
					});
				}
			},
			filters: {
				title       : "<%=_T('Filters')%>",
				button_text : "<%=_T('Create filter')%>",
				loading_msg : "<%=_T('Loading Filters')%>",
				new_text    : "<%=_T('New Filter')%>",
				i_class_name: "far fa-filter",
				permissions : {
					view    : "view_custom_filters",
					create  : "save_custom_filters",
					erase   : "delete_custom_filters"
				},
				left_section: "filters",
				grid_headers: [ "<%=_T('Name')%>", "<%=_T('Type')%>", "<%=_T('Scope')%>", "<%=_T('Description')%>" ],
				noRecordsMsg: "<%=_T('No custom filters found.')%>",
				init_once   : _.once(function() {
					self.controller.on("filterAdded", function(filter) {
						insertRowOrdered(filter.name, {
							columns: [
								{ 'ic-data-bind': "name",        text: filter.name        },
								{ text: get_report_type_desc(filter.type)                 },
								{ text: get_translated_scope(filter.scope)                },
								{ 'ic-data-bind': "description", text: filter.description }
							],
							attributes: {
								'rd-filter-id'  : filter.id,
								'rd-filter-type': filter.type
							}
						});
					}).on("filterUpdated", function(filter) {
						updateRow('rd-filter-id', filter.id, filter);
					}).on("filterDeleted", function(filter) {
						deleteRow('rd-filter-id', filter.id);
					});
				}),
				enter: function(callback) {
					this.init_once();
					_.defer(_.bind(self.listFilters, self), callback);
				},
				leave: function() {
				},
				on_new:  function() {
					self.renderFilter(null);
				},
				on_edit: function($row) {
					self.editEntity({
						entity: "Filter",
						loader: _.bind(self.controller.loadFilter, self.controller, $row.attr('rd-filter-id'), $row.attr('rd-filter-type')),
						render: _.bind(self.renderFilter, self)
					});
				},
				on_delete: function($row) {
					self.deleteEntity({
						entity : "Filter",
						deleter: _.bind(self.controller.deleteFilter, self.controller, $row.attr('rd-filter-id'), $row.attr('rd-filter-type'))
					});
				}
			},
			dimensions: {
				title       : "<%=_T('Custom Fields')%>",
				button_text : "<%=_T('Create custom field')%>",
				loading_msg : "<%=_T('Loading Custom Fields')%>",
				new_text    : "<%=_T('New Custom Field')%>",
				i_class_name: "far fa-poll-h",
				permissions : {
					view    : "view_custom_fields",
					create  : "save_custom_fields",
					erase   : "delete_custom_fields"
				},
				left_section: "dimensions",
				grid_headers: [ "<%=_T('Name')%>", "<%=_T('Scope')%>", "<%=_T('Type')%>", "<%=_T('Description')%>" ],
				noRecordsMsg: "<%=_T('No Custom Fields found.')%>",
				init_once   : _.once(function() {
					self.controller.on("dimensionAdded", function(dimension) {
						insertRowOrdered(dimension.name, {
							columns: [
								{ 'ic-data-bind': "name",        text: helpers.getCustomDimensionName(dimension.name) },
								{ 'ic-data-bind': "scope",       text: dimension.scopeLabel || dimension.scope || "" },
								{ 'ic-data-bind': "type",        text: dimension.type              },
								{ 'ic-data-bind': "description", text: dimension.description || "" }
							],
							attributes: {
								'rd-dimension-id'  : dimension.id,
								'rd-dimension-type': dimension.type
							}
						});
					}).on("dimensionUpdated", function(dimension) {
						updateRow('rd-dimension-id', dimension.id, _.extend({}, dimension, {
							name: helpers.getCustomDimensionName(dimension.name)
						}));
					}).on("dimensionDeleted", function(dimension) {
						deleteRow('rd-dimension-id', dimension.id);
					});
				}),
				enter: function(callback) {
					this.init_once();
					_.defer(_.bind(self.listDimensions, self), callback);
				},
				leave: $.noop,
				on_new:  function() {
					self.renderDimension(null);
				},
				on_edit: function($row) {
					self.editEntity({
						entity: "Dimension",
						loader: _.bind(self.controller.loadDimension, self.controller, $row.attr('rd-dimension-id')),
						render: _.bind(self.renderDimension, self)
					});
				},
				on_delete: function($row) {
					self.deleteEntity({
						entity : "Custom Field",
						deleter: _.bind(self.controller.deleteDimension, self.controller, $row.attr('rd-dimension-id'))
					});
				}
			}
		};
	}

	/*!
	 * App method
	 */
	ReportsDesigner.prototype.enter = function() {
		this.initialize_once();
	};

	/*!
	 * App method
	 */
	ReportsDesigner.prototype.leave = function() {
		//$('#btn_show_manage').show();
		//$("#report_designer_main_container").hide();
	};

	/*!
	 * App method
	 */
	ReportsDesigner.prototype.onMenuItemClick = function(/*$item*/) {
		// interface requirement, nothing here..
	};

	ReportsDesigner.prototype.showOverlay = function(message) {
		this.dashboard.block_right_panel({ with_animation: true, with_message: message });
	};

	ReportsDesigner.prototype.hideOverlay = function() {
		this.dashboard.unblock_right_panel();
	};

	ReportsDesigner.prototype.renderForm = function(form_constructor, params) {
		var self = this;

		var form_ready = false;
		self.showOverlay("<%=_T('Setting up...')%>");

		
		self.currentForm = new form_constructor(params, self.controller, $('#reports_form_data_container'), {
				saveButton  : $('#reports_form_button_confirm'),
				cancelButton: $('#reports_form_button_cancel')
			}).
			on("render", function(page_uri, callback) {
				self.dashboard.load_inner_page(page_uri, callback);
			}).
			on("ready", function() {
				form_ready = true;
				$('#reports_view_container').hide();
				$('#reports_form_container').show();
				self.currentForm.applyPermissions();
				self.hideOverlay();
			}).
			on("exit", function(status, message) {
				self.dashboard.hide_message();
				$('#reports_form_container').hide();
				$('#reports_view_container').show();
				if (!form_ready) {
					self.hideOverlay();
				}
				switch (status) {
					case "done":
						if (message) {
							self.dashboard.show_message("messageInfo", message);
						}
						break;
					case "cancel":
						break;
					case "error":
						self.dashboard.show_message("messageError", message);
						break;
				}
				self.currentForm = null;
			}).
			on("status", function(message) {
				self.dashboard.show_message("messageInfo", message);
			}).
			on("error", function(message, persistent) {
				self.dashboard.show_message("messageError", message, persistent && -1);
			});

		self.currentForm.show();
	};

	ReportsDesigner.prototype.renderReport = function(report) {
		window.__report = report;  // TODO: Remove this code
		this.renderForm(inConcert.reports.FormEditReport, report);
	};

	ReportsDesigner.prototype.renderMetric = function(metric) {
		window.__metric = metric;  // TODO: Remove this code
		this.renderForm(inConcert.reports.FormEditMetric, metric);
	};

	ReportsDesigner.prototype.renderFilter = function(filter) {
		window.__filter = filter;  // TODO: Remove this code
		this.renderForm(inConcert.reports.FormEditFilter, filter);
	};

	ReportsDesigner.prototype.renderDimension = function(dimension) {
		window.__dimension = dimension;  // TODO: Remove this code
		this.renderForm(inConcert.reports.FormEditDimension, dimension);
	};

	ReportsDesigner.prototype.makeSearchRegExp = function(word)
	{
		if (typeof word !== "string") {
			return "";
		}
		let search_pattern = inConcert.Utils.removeSpecialChars(word);
		return new RegExp(search_pattern.replace(/[^\w\s]/gi, ""), "i");
	};

	ReportsDesigner.prototype.applyViewFilter = function() {
		const self = this;
		const $reset = $('#reports_filter_reset');
		const $paginator = $('#reports_designer_paginator');
		const $not_found = $('#reports_view_not_results_found');
		const filterText = $('#reports_filter_text').val().trim();

		let $rows = this.viewGridBody.children();
		let total_rows = $rows.length;

		const showNotResultFound = function() {
			self.viewGrid.hide();
			let message;
			if (total_rows === 0) {
				message = self.currentView.noRecordsMsg;
			}
			else {
				//message = format("<%=_T('No results found for')%> <span class=\"noResultsQuery\">{0}</span>.", filterText);
				message = format("<span class='messageInfoText1'><%=_T('No results found.')%></span> <span class='messageInfoText2'><%=_T('There are no items created or matching your search yet.')%></span>", filterText);
			}
			$not_found.html(message).show();
		};

		if (total_rows === 0) {
			showNotResultFound();
			return;
		}

		if (filterText === "") {
			$reset.hide();
			$rows.show();
			$paginator.show();
			if (this.m_paginator) {
				this.m_paginator.refresh_page();
			}
			this.viewGrid.show();
			$not_found.hide();
			return;
		}

		let hidden_rows = 0;
		$rows.each(function(_, row) {
			let $row = $(row);
			let row_value = $row.attr("rd-text-search");
			let search_regexp = self.makeSearchRegExp(filterText);
			if (row_value.search(search_regexp) === -1) {
				$row.hide();
				hidden_rows++;
			}
			else {
				$row.show();
			}
		});

		$reset.show();
		$paginator.hide(); // oculto la botonera del paginado

		if (total_rows === hidden_rows) {
			showNotResultFound();
			return;
		}

		this.viewGrid.show();
		$not_found.hide();
	};

	ReportsDesigner.prototype.updateViewFilter = function() {
		const filterText = $('#reports_filter_text').val().trim();
		if (filterText !== "") {
			this.applyViewFilter();
			return true;
		}
		return false;
	};

	ReportsDesigner.prototype.setupViewFilter = function() {
		const self = this;
		$('#reports_filter_text').on("keyup", function(event) {
			// if user press Enter trigger the search inmediatly
			if (event.keyCode === 13) {
				self.applyViewFilter();
			}
		});
	};

	ReportsDesigner.prototype.initializePagination = function(initialItems, options) {
		var self = this;

		if (self.m_paginator) {
			self.m_paginator = self.m_paginator.destroy();
		}

		if (initialItems.length === 0) {
			return;
		}

		self.m_paginator = $.Paginator.init(_.extend({
			pages_container : $('#reports_designer_paginator'),
			data_container  : self.viewGridBody,
			previous        : "",
			next            : "",
			show_at_least   : 12,
			rows_per_page   : 10,
			row_count       : initialItems.length,
			items           : initialItems,
			text_format     : "<%=_T('{page_row_start}-{page_row_end} of {many}')%>",
			translated_many : "<%=_T('many')%>",
			callback: function(page, need_data, on_done) {
				on_done();
			}
		}, options));

		self.m_paginator.no_more_pages();
	};

	ReportsDesigner.prototype.menuChangeExternalOption = function( rd_section ) {
		var self = this;

		var doSwitch = function() {
			self.rowCount = 0;

			self.viewGrid
				.hide(); // for faster cleanup
			self.viewGridBody
				.empty();
			self.viewGridHead
				.empty()
				.append('<th></th>'); // counter

			var old_view = self.currentView;
			self.currentView = self._views[ rd_section ];// self._views[$section.attr('rd-section')];

			if (old_view) {
				old_view.leave();
				if (self.m_paginator) {
					self.m_paginator = self.m_paginator.destroy();
				}
			}


			self.showOverlay(self.currentView.loading_msg);

			// setup and show grid headers
			self.currentView.grid_headers.forEach(function(header) {
				self.viewGridHead.append('<th>' + header + '</th>');
			});
			self.viewGridHead.append('<th></th>'); // action buttons

			$('#reports_view_title').text(self.currentView.title);
			$('#reports_view_button_new').text(self.currentView.button_text);
			$('#reports_view_img').removeClass().addClass(self.currentView.i_class_name);
			$('#reports_form_img').removeClass().addClass(self.currentView.i_class_name);

			$('#reports_view_button_new').val( self.currentView.new_text)
				[
					inConcert.permissions.to_action('reports' , self.currentView.permissions.create, 'SAVE') ?
					'show' : 'hide'
				]();

			$('#reports_filter_text').val('');

			self.currentView.enter(function() {
				if (self.rowCount > 0) {
					$('#reports_view_not_results_found').hide();
					self.initializePagination(
						_.map(self.viewGridBody.children(), $)
					);
					self.viewGrid.show();
				} else {
					$('#reports_view_not_results_found')
						//.text(self.currentView.noRecordsMsg)
						.html(self.currentView.noRecordsMsg)
						.show();
				}
				self.hideOverlay();
			});
		};

		doSwitch();
	};

	ReportsDesigner.prototype.setupEventHandlers = function() {
		const self = this;

		$('#reports_view_container').on("click", "#reports_filter_reset", function() {
			$('#reports_filter_text').val('');
			self.applyViewFilter();
		});

		$('#reports_view_container').on("click", "#reports_filter_search", function() {
			self.applyViewFilter();
		});

		$('#reports_view_container').on("click", "#reports_view_button_new", function() {
			if (!inConcert.permissions.to_action('reports' , self.currentView.permissions.create, 'SAVE')) {
				let name = ( $( this ).attr("value" ) || "" ).toLowerCase();
				let error_message = format( "<%=_T('Unable to create {0}: You dont have permission for this action.')%>", name );
				self.dashboard.show_message( "messageError", error_message );
				return;
			}
			self.currentView.on_new();
		});

		$('#reports_form_buttons_actions').on("click", "#reports_form_button_confirm", function() {
			self.currentForm.confirm();
		});

		$('#reports_form_buttons_actions').on("click", "#reports_form_button_cancel", function() {
			self.currentForm.cancel();
		});

		const on_edit = function( $row ) {
			if (!inConcert.permissions.to_action( 'reports' , self.currentView.permissions.view)) {
				let error_message = "<%=_T('Unable to edit: You dont have permission for this action.')%>" ;
				self.dashboard.show_message( "messageError", error_message );
				return;
			}
			self.currentView.on_edit( $row );
		};

		this.viewGridBody.on("click", '.gridActionButton', function() {
			const $row = $(this).parent().parent();
			switch ($(this).attr('rd-view-action')) {
				case "edit":
					on_edit( $row );
					break;
				case "delete":
					self.currentView.on_delete($row);
					break;
			}
		}).on("dblclick", 'tr', function() {
			on_edit( $(this) );
		});

		this.setupViewFilter();
	};

	ReportsDesigner.prototype.editEntity = function(params) {
		params = _.extend({
			entity    : "entity",
			loader    : $.noop,
			render    : $.noop,
			loadingMsg: "<%=_T('Loading {0}')%>",
			failMsg   : "<%=_T('Failure loading {0}: {1}')%>"
		}, params);

		var self = this;
		self.showOverlay(format(params.loadingMsg, params.entity));
		params.loader()
			.done(function(entity) {
				params.render(entity);
			})
			.fail(function(errorThrown) {
				self.hideOverlay();
				self.dashboard.show_message("messageError", format(params.failMsg, params.entity, errorThrown));
			});
	};

	ReportsDesigner.prototype.deleteEntity = function(params) {
		params = _.extend({
			entity     : "item",
			deleter    : $.noop,
			title      : "<%=_T('About to delete {0}')%>",
			question   : "<%=_T('Are you sure you want to delete this {0}?')%>",
			deletingMsg: "<%=_T('Deleting {0} ...')%>",
			successMsg : "<%=_T('{0} successfully deleted.')%>",
			failMsg    : "<%=_T('Unable to delete {0}: {1}')%>"
		}, params);

		var self = this;
		helpers.askForConfirmation(format(params.question, params.entity), format(params.title, params.entity)).done(function() {
			self.showOverlay(format(params.deletingMsg, params.entity));
			params.deleter()
				.always(function() {
					self.hideOverlay();
				})
				.done(function() {
					self.dashboard.show_message("messageInfo", format(params.successMsg, params.entity));
				})
				.fail(function( errorThrown ) {
					var message_error;
					switch ( errorThrown ) {
						case "forbidden":
							message_error = "<%=_T('You dont have permission to remove')%>";
						break;
						default:
							message_error = errorThrown;
					}
					self.dashboard.show_message("messageError", format( params.failMsg, params.entity, message_error ));
				});
		});
	};

	ReportsDesigner.prototype.appendRow = function(options) {
		const self = this;
		options = _.extend({
			editable : true,
			removable: true,
			tags     : []
		}, options);

		const $row = $('<tr></tr>');
		$row.attr("row_type", "main");
		$row.append('<td>' + (++self.rowCount) + '</td>'); // counter

		// columns
		options.columns.forEach(function(column) {
			if (typeof(column) !== "object") {
				column = { text: column };
			}
			options.tags.push(column.text.replace(/\W/g, '').toLowerCase());

			const $td = $('<td></td>')
				.attr('title', column.text)
				.text(column.text)
				.appendTo($row);
			delete column.text; // so we can iterate without bottering

			$.each(column, function(k, v) {
				$td.attr(k, v);
			});
		});

		var p = inConcert.permissions;
		// get edit grid button 
		var get_edit_button = function() {
			if ( !p.to_action( 'reports' , self.currentView.permissions.view ) ) {
				return '';
			}
			var button_label = "<%=_T('Edit')%>";
			if ( !p.to_action( 'reports' , self.currentView.permissions.create ) ) {
				button_label = "<%=_T('View')%>";
			}
			return '<button rd-view-action="edit" type="button" class="buttonColorGray gridActionButton edit" ' + (options.editable  ? '' : 'disabled ') + 'value="' + button_label + '" title="'+ button_label +'"><i class="buttons-icons-edit"></i>' + button_label + '</button>';
		// get d grid button 
		};
		var get_del_button = function() {
			if ( !p.to_action( 'reports' , self.currentView.permissions.view ) ) {
				return '';
			}
			if ( !p.to_action( 'reports' , self.currentView.permissions.erase ) ) {
				return '';
			}
			return '<button rd-view-action="delete" type="button" class="buttonColorGray gridActionButton delete" ' + (options.removable ? '' : 'disabled ') + 'value="' + "<%=_T('Delete')%>" + '" title="'+ "<%=_T('Erase')%>" +'"><i class="buttons-icons-delete"></i>' + "<%=_T('Delete')%>" + '</button>';
		};

		// action buttons
		$row.append(
			'<td class="gridColumnActions">' +
				get_edit_button() +
				get_del_button() +
			'</td>'
		);
		$row.attr('rd-text-search', options.tags.join(','));

		// attributes for the row
		$.each(options.attributes || {}, function(k, v) {
			$row.attr(k, v);
		});

		if (self.m_paginator) {
			self.m_paginator.append($row);
		} else {
			// finally, append the row
			this.viewGridBody.append($row);
		}

		return $row; // return the new row so new data can be attached to it
	};

	ReportsDesigner.prototype.listReports = function(callback) {
		var self = this;
		self.controller.listReports("custom").done(function(reports) {
			reports.forEach(function(report) {
				self.appendRow({
					columns: [
						{ 'ic-data-bind': "name", text: report.name },
						{ text: get_report_type_desc(report.type)   }
					],
					attributes: {
						'rd-report-id'  : report.id,
						'rd-report-type': report.type
					}
				});
			});
		}).fail(function(errorThrown) {
			self.dashboard.show_message("messageError", format("<%=_T('Unable to list reports: {0}')%>", errorThrown));
		}).always(callback);
	};

	ReportsDesigner.prototype.listMetrics = function(callback) {
		var self = this;
		self.controller.listMetrics("custom").done(function(metrics) {
			metrics.forEach(function(metric) {
				self.appendRow({
					editable: metric.mutable !== false,
					tags    : metric.mutable === false ? [ "disabled" ] : [],
					columns : [
						{ 'ic-data-bind': "name",        text: metric.name || metric.id },
						{ 'ic-data-bind': "scope",       text: metric.scope || (metric.composite && "Composite") },
						{ 'ic-data-bind': "description", text: metric.description || "" }
					],
					attributes: { 'rd-metric-id': metric.id }
				});
			});
		}).fail(function(errorThrown) {
			self.dashboard.show_message("messageError", format("<%=_T('Unable to list metrics: {0}')%>", errorThrown));
		}).always(callback);
	};

	ReportsDesigner.prototype.listFilters = function(callback) {
		var self = this;

		var onError = function(type, errorThrown) {
			self.dashboard.show_message("messageError", format("<%=_T('Unable to list {0} filters: {1}')%>", type, errorThrown));
			callback();
		};

		var filterTypeTranslated;
		var appendFilter = function(filter) {
			self.appendRow({
				columns: [
					{ 'ic-data-bind': "name", text: filter.name || filter.id  },
					{ text: filterTypeTranslated                              },
					{ text: get_translated_scope(filter.scope)                },
					{ 'ic-data-bind': "description", text: filter.description }
				],
				attributes: {
					'rd-filter-id'  : filter.id,
					'rd-filter-type': filter.type
				}
			});
		};

		self.controller.listSummaryFilters("custom").done(function(summaryFilters) {
			filterTypeTranslated = "<%=_T('Summary')%>";
			summaryFilters.forEach(appendFilter);

			self.controller.listDetailFilters("custom").done(function(detailFilters) {
				filterTypeTranslated = "<%=_T('Detail')%>";
				detailFilters.forEach(appendFilter);
				callback();
			}).fail(_.partial(onError, "<%=_T('Detail')%>"));

		}).fail(_.partial(onError, "<%=_T('Summary')%>"));
	};

	ReportsDesigner.prototype.listDimensions = function(callback) {
		var self = this;
		self.controller.listDimensions("custom").done(function(dimensions) {
			dimensions = _.select(dimensions, helpers.makeFilter({ __designer: true }));
			_.each(dimensions, function(d) {
				self.appendRow({
					columns : [
						{ 'ic-data-bind': "name",        text: helpers.getCustomDimensionName(d.name) },
						{ 'ic-data-bind': "scope",       text: d.scopeLabel || d.scope || "" },
						{ 'ic-data-bind': "type",        text: d.type        },
						{ 'ic-data-bind': "description", text: d.description || "" }
					],
					attributes: { 'rd-dimension-id': d.id }
				});
			});
		}).fail(function(errorThrown) {
			self.dashboard.show_message("messageError", format("<%=_T('Unable to list dimensions: {0}')%>", errorThrown));
		}).always(callback);
	};

	inConcert.reports.FormBase = FormBase;

	inConcert.reports.external = inConcert.reports.external || {};
	if ( typeof inConcert.reports.external.designer !== "function" ) {
		inConcert.reports.external.designer = ReportsDesigner;
	}

	inConcert.reports.dashboard.registerApp ({
		name       : app_name,
		permissions: "manage",
		factory    : function(dashboard) {
			return new ReportsDesigner(dashboard);
		}
	});
}());
