/*jshint bitwise:true, browser:true, curly:true, eqeqeq:true, noarg:true, noempty:true, strict:true, undef:true, unused:true, esversion: 9 */
/*jshint newcap:false, quotmark:false */
/*global $, _, inConcert*/

this.inConcert = this.inConcert || {};
this.inConcert.reports = this.inConcert.reports || {};

(function () {
	"use strict";

	// scopes that are considered custom
	var customScopes = {
		webchat			: true,
		webcall			: true,
		cci				: true,
		contactform		: true,
		survey			: true,
		api				: true,
		batch_data		: true,
		process_flow	: true
	};

	var customFieldsScopes = {
		api				: true,
		batch_data		: true,
		process_flow	: true
	};

	var _T = inConcert.i18n.getString;
	var format = inConcert.Utils.format;

	var SCOPES = {
		Interaction					: "<%=_T('Conversation')%>",
		InteractionAgent			: "<%=_T('Conversation Agent')%>",
		InteractionAttentionLevel	: "<%=_T('Conversation Att Level')%>",
		InteractionCount			: "<%=_T('Conversation Count')%>",
		Agent						: "<%=_T('Agent')%>",
		AgentStateLapse				: "<%=_T('Agent State Lapse')%>",
		CustomComposite				: "<%=_T('Custom Composite')%>",
		InteractionRule				: "<%=_T('Rules')%>",
		InteractionVoiceBlaster		: "<%=_T('Voice Blaster')%>"
	};

	// relations between scopes, or at least specifies whether or not
	// dimensions from one scope can be used in another scope
	var scopeCompatibility = {
		Interaction      : _.extend({}, customScopes),
		InteractionAgent : _.extend({
			Interaction: true
		}, customScopes),
		InteractionAttentionLevel : _.extend({
			Interaction: true
		}, customScopes)
	};

	/*!
	 * Call this function in any method that is intended to be abstract
	 */
	function abstractFunctionCalled(fn) {
		throw "" + fn + " is abstract";
	}

	/*!
	 * Set a property deep-inside an object
	 * Example:
	 *    \code deepSet({}, "foo.bar.baz", true) \endcode
	 * will return:
	 * \code
	 * {
	 *   foor: {
	 *     bar: {
	 *       baz: true
	 *     }
	 *   }
	 * }
	 * \endcode
	 */
	function deepSet(obj, prop, value) {
		var base = obj;
		prop = prop.split('.');

		var index = 0;
		var len = prop.length;
		for ( ; index < len - 1; ++index ) {
			var tmp = obj[ prop[index] ];
			if ( !tmp ) {
				tmp = { };
				obj[ prop[index] ] = tmp;
			}
			obj = tmp;
		}
		obj[ prop[index] ] = value;
		return base;
	}

	/*!
	 * Gets a property from deep-inside obj, see \c deepSet for details
	 */
	function deepGet(obj, prop) {
		prop = prop.split('.');

		var index = 0;
		var len = prop.length;
		for ( ; index < len - 1; ++index) {
			obj = obj[ prop[index] ];
			if (obj === undefined || obj === null) {
				return undefined;
			}
		}
		return obj[ prop[index] ];
	}

	/*
	 * Returns a function that, when called, will invoke
	 * the function received as parameter.
	 *
	 * This function is intended to be used as an agument
	 * for _.compose; usually after _.partial or _.bind
	 */
	function future() {
		var args = arguments;
		return function(fn) {
			return fn.apply(this, args);
		};
	}


	/*!
	 * Creates and returns a function than can be used as
	 * an iterator in a find/filter/select call
	 *
	 * Example: Select all the elements in \c customers
	 * having the name 'Jhon' without using \c _.findWhere
	 *
	 * \code _.select(names, makeFilter({name: "Jhon"})); \endcode
	 *
	 * Example: Populate the list \c $list with the
	 * elements from \c datasource having the attribute \c scope set to \c 'Public'
	 *
	 * \code populateList($list, datasource, makeFilter({scope: 'Public'})); \endcode
	 */
	function makeFilter(properties) {
		return (function(item) {
			return _.every(properties, function(value, key) {
				return item[key] === value;
			});
		});
	}

	/*!
	 * Hides \p options.$hide, and then shows \p options.$show
	 *
	 * \return A promise object which provides the \c done() event
	 *
	 * Example:
	 *
	 * \code
	 *   niceFading({
	 *       $hide: $('#loginForm'),
	 *       $show: $('#dashboard')
	 *   }).done(function() {
	 *       alert("Animation complete!");
	 *   });
	 * \endcode
	 */
	function niceFading(options) {
		var deferred = new $.Deferred();
		var speed = options.speed || 200;

		options.$hide.fadeOut(speed).promise().done(function() {
			options.$show.fadeIn(speed).promise().done(function() {
				deferred.resolve();
			});
		});

		return deferred;
	}

	/*!
	 * Returns \c true is \p $parent has a children matching \p selector
	 */
	function hasChildren($parent, selector) {
		var $children = $parent.children(selector);
		if ($children.length > 0) {
			return $children;
		}
		return false;
	}

	/*!
	 * A function that can ask a simple Yes/No question
	 *
	 * \return A promise object which raises the \c done() event when the
	 * user's answer is \c Yes, otherwise, the event veing raised is \c fail().
	 *
	 * Example:
	 *
	 * \code
	 *   helpers.askForConfirmation(
	 *       "So you wanna do 'The Thing', don't ya?", // message
	 *       "Confirm"                                  // title
	 *   ).done(function() {
	 *       alert("So your answer was 'Yes'");
	 *   }).fail(function() {
	 *       alert("cluck cluck cluuuuoooock");
	 *   });
	 * \endcode
	 */
	function askForConfirmation(question, title) {
		var deferred = new $.Deferred();

		$.inconcertPopup.Show(
			title || "<%=_T('Confirm')%>",
			question,
			$.inconcertPopup.ButtonSets.BS_YESNO,
			function(response) {
				if (response === "Yes") {
					deferred.resolve();
				} else {
					deferred.reject();
				}
			},
			undefined,
			{
				popup_id        : "rpt_inconcertPopup",
				container_class : "confirmationPopUp",
				overlay_class   : "confirmationOverlay"
			}
		);

		return deferred;
	}

	/*!
	 * Builds a function that will be used to append an single item to a container
	 *
	 * Note there is a performance gain in using this approach since the 'append'
	 * function can be used from multiple places and we cannot afford to re-create variables
	 * when appending lots of items in one shot
	 *
	 * Look for \c populateList to see an example
	 */
	function makeAppender($target, canAdd)
	{
		canAdd = canAdd || function() {
			return true;
		};
		function getTransform(str) {
			var cmd = str.split(':'); // e.g: {value:lower}
			if (cmd.length === 1) {
				return  { str: str, transform: _.identity };
			}
			var modifier = cmd[1];
			return {
				str: cmd[0],
				transform: function(v) {
					switch(modifier) {
						case "lower":
							return v.toLowerCase();
						case "upper":
							return v.toUpperCase();
						case "translate":
							return _T(v);
						default:
							return v;
					}
				}
			};
		}

		var $template_html = $('<div>').append(
				$target.children('[ic-role="template"]')
					.clone()
					.show()
					.removeAttr('ic-role')
			).html();

		var appender = function(item, key, $beforeThan)
		{
			if (!canAdd(item, key)) {
				return false;
			}
			var item_html = $template_html.replace(/\{([^{}]*)\}/g, function(a, b) {
				var modifier = getTransform(b);
				b = modifier.str.split('|'); // support for fallback properties
				for (var i = 0; i < b.length; ++i) {
					switch( b[i] ) {
						case "empty":
							return "";
						case "key":
							return modifier.transform(key || "");
						case "value":
							if ( typeof(item) !== "object" ) {
								return modifier.transform(item || "");
							}
							/* falls through */
						default:
							if (typeof(item) === "object") {
								var value = deepGet(item, b[i]);
								if (value !== undefined) {
									return modifier.transform(value);
								}
							}
					}
				}
				return modifier.transform(a);
			});

			if ($beforeThan) {
				return $(item_html).insertBefore($beforeThan);
			}

			return $(item_html).appendTo($target);
		};

		return appender;
	}

	/*!
	 * Populate the list \p $target using the given datasource
	 * $target can be a HTMl ul, select, or anything else
	 *
	 * The only requirement is that the first child in $target should be a
	 * hidden object playing as template.
	 *
	 * Example template for an ul element
	 * \code
	 *   <li ic-role="template" style="display:none">
	 *      <label>{key}</label>
	 *      <input type="text" value={value}/>
	 *   </li>
	 * \endcode
	 *
	 * \note When $target is a \c select element the first element after the
	 * template is selected, this action triggers the \c change event in $target
	 *
	 * \param iterator A function(item, key) to call before appending an alement. The element
	 * is not appended to the list when this function returns \c false. This function can also
	 * overwrite the element being inserted (i.e. \c item) by returning an \c object.
	 *
	 * \param options An object with the following properties:
	 * \li sort boolean (default: \c false) If \c true, items are sorted alphabetically
	 * \li disableOnEmpty boolean (default: \c false) The element $list will be disabled if no element is appended
	 * \li defaultItem object (default: \c undefined) That little guy that is there when no one else is
	 * \li firstItem object (defaylt: \c undefined) This item is added at the beggining of the list, but only if there are elements
	 *
	 * \return The number of elements appended to the list (not including \p defaultItem)
	 */
	function populateList($target, datasource, iterator, options)
	{
		// var uid = inConcert.Utils.generateGuid();
		// var startDate = new Date();
		// console.log("A) populateList - START --> UID: '" + uid + "' - Start date: '" + startDate + "' - List: '" + $target.attr("id") + "'");

		if (typeof(iterator) === "object") {
			options = iterator;
			iterator = $.noop;
		} else {
			iterator = iterator || $.noop;
		}

		options = _.extend({
			sort          : false,     // sort the list after being populated
			disableOnEmpty: false,     // disable the element if no items were added (even having defaultItem)
			defaultItem   : undefined, // insert this item if no items were added
			candidateValue: undefined,
			firstItem     : undefined,
			triggerChange : true
		}, options);

		var isVisible = false;
		if (options.hideList) {
			isVisible = $target.is(':visible');
			if (isVisible) {
				$target.hide(); // avoid repaints
			}
		}

		// // (*) Realizo un detach de la lista para que el procesamiento de la misma sea en memoria y no directa sobre el DOM.
		var $parent;
		if (options.detachContainer) {
			$parent = $target.parent();
			$target.detach();
		}
		$target.children(':not([ic-role])').remove();

		// create the appender
		var itemCount = 0;
		var appender = makeAppender($target);

		for (var key in datasource)
		{
			if (datasource.hasOwnProperty(key))
			{
				var item = datasource[key];
				var iterator_result = iterator(item, key);
				if (iterator_result === false) {
					continue;
				}

				if (iterator_result && typeof(iterator_result) === "object") {
					item = iterator_result;
				}

				var $element = appender(item, key);
				itemCount += 1;

				if (iterator_result && typeof(iterator_result) === "function") {
					iterator_result($element);
				}
			}
		}

		if (options.disableOnEmpty) {
			$target.prop('disabled', itemCount === 0);
		}

		switch(itemCount) {
			case 0:
				if (options.defaultItem) {
					appender(options.defaultItem, options.defaultItem.key); // pass the key in case there is one
				}
				break;
			default:
				if (options.sort) {
					$target.children(':not([ic-role])').sort(function(a, b) {
						return a.innerHTML.localeCompare(b.innerHTML, undefined, {sensitivity: "base"});
					}).appendTo($target);
				}
				if (options.firstItem) {
					appender(options.firstItem, options.firstItem.key, $target.children(':first')); // pass the key in case there is one
				}
		}

		if ($target.is('select')) { // make sure template is not the selected one
			var $options = $target.children('[ic-role!="template"]');
			if (options.candidateValue && $options.filter('[value="' + options.candidateValue + '"]').length > 0) {
				$target.val(options.candidateValue);
			} else {
				$target.val($options.first().val());
			}
			if (options.triggerChange) {
				$target.trigger('change');
			}
		}

		if (isVisible) {
			$target.show();
		}

		// // Aca realiza un attach de la lista (*) a la cual le hemos realizado detach.
		if (options.detachContainer) {
			$parent.append($target);
		}

		// var endDate = new Date();
		// var elapsedTimeSec = (endDate.getTime() - startDate.getTime()) / 1000;
		// console.log("B) populateList - END --> UID: '" + uid + "' - End date" + endDate + "' - Elapsed time (sec): '" + elapsedTimeSec + "' - List: '" + $target.attr("id") + "' --> TOTAL ITEMS: " + itemCount);

		return itemCount;
	}

	/*!
	 * Esta funcion realiza lo mismo que la funcion "populateList"
	 * con la diferencia de que realiza el recorrido de la lista de forma asincronica.
	 * De esta forma evitamos que el navegador quede colgado.
	 */
	function populateListAsync($target, datasource, iterator, options, callback)
	{
		callback = callback || $.noop;

		// var uid = inConcert.Utils.generateGuid();
		// var startDate = new Date();
		// console.log("A) populateListAsync - START --> UID: '" + uid + "' - Start date: '" + startDate + "' - List: '" + $target.attr("id") + "'");

		if (typeof(iterator) === 'object') {
			options = iterator;
			iterator = $.noop;
		} else {
			iterator = iterator || $.noop;
		}

		options = _.extend({
			sort              : false,     // sort the list after being populated
			disableOnEmpty    : false,     // disable the element if no items were added (even having defaultItem)
			defaultItem       : undefined, // insert this item if no items were added
			candidateValue    : undefined,
			firstItem         : undefined,
			triggerChange     : true,
			maxAppendPerLoop  : 2,         // default value
			pauseBetweenLoopMs: 100,       // default value
			waitingFn         : $.noop     // default function
		}, options);

		var isVisible = false;
		if (options.hideList) {
			isVisible = $target.is(':visible');
			if (isVisible) {
				$target.hide(); // avoid repaints
			}
		}

		// (*) Realizo un detach de la lista para que el procesamiento de la misma sea en memoria y no directa sobre el DOM.
		var $parent;
		if (options.detachContainer) {
			$parent = $target.parent();
			$target.detach();
		}

		$target.children(':not([ic-role])').remove();

		// create the appender
		var itemCount = 0;
		var appender = makeAppender($target, options.canAddFn);

		inConcert.Utils.foreach_timeout({
			rows : datasource,
			max_append_per_loop: options.maxAppendPerLoop,
			pause_between_loop_ms: options.pauseBetweenLoopMs,
			waiting_message: options.waitingMsg,
			// por cada registro que procesamos se invoca esta funcion
			append_fn: function( item, added, key )
			{
				var iterator_result = iterator(item, key);

				if (iterator_result === false) {
					return;
				}

				if (typeof(iterator_result) === "object") {
					item = iterator_result;
				}

				var $element = appender(item, key);
				if ($element) {
					itemCount += 1;

					if (typeof(iterator_result) === "function") {
						iterator_result($element);
					}
				}
			},
			// la funcion, para que muestre en pantalla un waiting durante la carga o un log.
			waiting_fn : function( message ) {
				options.waitingFn( message );
			},
			// Esta funcion es invocada cuando finaliza el recorrido de la lista y la carga del DOM.
			complete_fn: function()
			{
				// Aca realiza un attach de la lista (*) a la cual le hemos realizado detach.
				if (options.detachContainer) {
					$parent.append($target);
				}

				if (options.disableOnEmpty) {
					$target.prop('disabled', itemCount === 0);
				}

				switch(itemCount) {
					case 0:
						if (options.defaultItem) {
							appender(options.defaultItem, options.defaultItem.key); // pass the key in case there is one
						}
						break;
					default:
						if (options.sort) {
							$target.children(':not([ic-role])').sort(function(a, b) {
								return a.innerHTML.localeCompare(b.innerHTML, undefined, {sensitivity: 'base'});
							}).appendTo($target);
						}
						if (options.firstItem) {
							appender(options.firstItem, options.firstItem.key, $target.children(':first')); // pass the key in case there is one
						}
				}

				if ($target.is('select')) { // make sure template is not the selected one
					var $options = $target.children('[ic-role!="template"]');
					if (options.candidateValue && $options.filter('[value="' + options.candidateValue + '"]').length > 0) {
						$target.val(options.candidateValue);
					} else {
						$target.val($options.first().val());
					}
					if (options.triggerChange) {
						$target.trigger('change');
					}
				}

				if (isVisible) {
					$target.show();
				}

				// var endDate = new Date();
				// var elapsedTimeSec = (endDate.getTime() - startDate.getTime()) / 1000;
				// console.log("B) populateListAsync - END --> UID: '" + uid + "' - End date" + endDate + "' - Elapsed time (sec): '" + elapsedTimeSec + "' - List: '" + $target.attr("id") + "'");

				callback(itemCount);
			}
		});
	}

	/*!
	 * Esta funcion realiza lo mismo que la funcion "populateList"
	 * con la diferencia de que realiza el recorrido de la lista de forma asincronica.
	 * De esta forma evitamos que el navegador quede colgado.
	 */
	function filterListAsync($target, datasource, iterator, options, callback)
	{
		callback = callback || $.noop;

		// var uid = inConcert.Utils.generateGuid();
		// var startDate = new Date();
		// console.log("A) filterListAsync - START --> UID: '" + uid + "' - Start date: '" + startDate + "' - List: '" + $target.attr("id") + "'");

		if (typeof(iterator) === 'object') {
			options = iterator;
			iterator = $.noop;
		} else {
			iterator = iterator || $.noop;
		}

		options = _.extend({
			sort              : false,     // sort the list after being populated
			disableOnEmpty    : false,     // disable the element if no items were added (even having defaultItem)
			defaultItem       : undefined, // insert this item if no items were added
			candidateValue    : undefined,
			firstItem         : undefined,
			triggerChange     : true,
			maxAppendPerLoop  : 20,        // default value
			pauseBetweenLoopMs: 100,       // default value
			waitingFn         : $.noop     // default function
		}, options);

		var isVisible = false;
		if (options.hideList) {
			isVisible = $target.is(':visible');
			if (isVisible) {
				$target.hide(); // avoid repaints
			}
		}

		// (*) Realizo un detach de la lista para que el procesamiento de la misma sea en memoria y no directa sobre el DOM.
		var $parent;
		if (options.detachContainer) {
			$parent = $target.parent();
			$target.detach();
		}

		var $children = $target.children(':not([ic-role])');
		$children.hide();

		var itemCount = 0;

		inConcert.Utils.foreach_timeout({
			rows : datasource,
			max_append_per_loop: options.maxAppendPerLoop,
			pause_between_loop_ms: options.pauseBetweenLoopMs,
			waiting_message: options.waitingMsg,
			// por cada registro que procesamos se invoca esta funcion
			append_fn: function( item, added, key )
			{
				var iterator_result = iterator(item, key);
				if (iterator_result === false) {
					return;
				}
				var $item = options.$getItem(item);
				if ($item && $item.length > 0) {
					$item.show();
					itemCount++;
				}
			},
			// la funcion, para que muestre en pantalla un waiting durante la carga o un log.
			waiting_fn : function( message ) {
				options.waitingFn( message );
			},
			// Esta funcion es invocada cuando finaliza el recorrido de la lista y la carga del DOM.
			complete_fn: function()
			{
				// Aca realiza un attach de la lista (*) a la cual le hemos realizado detach.
				if (options.detachContainer) {
					$parent.append($target);
				}

				if (options.disableOnEmpty) {
					$target.prop('disabled', itemCount === 0);
				}

				switch(itemCount) {
					case 0:
						// if (options.defaultItem) {
							// appender(options.defaultItem, options.defaultItem.key); // pass the key in case there is one
						// }
						break;
					default:
						if (options.sort) {
							$target.children(':not([ic-role])').sort(function(a, b) {
								return a.innerHTML.localeCompare(b.innerHTML, undefined, {sensitivity: 'base'});
							}).appendTo($target);
						}
						// if (options.firstItem) {
							// appender(options.firstItem, options.firstItem.key, $target.children(':first')); // pass the key in case there is one
						// }
				}

				if ($target.is('select')) { // make sure template is not the selected one
					var $options = $target.children('[ic-role!="template"]');
					if (options.candidateValue && $options.filter('[value="' + options.candidateValue + '"]').length > 0) {
						$target.val(options.candidateValue);
					} else {
						$target.val($options.first().val());
					}
					if (options.triggerChange) {
						$target.trigger('change');
					}
				}

				if (isVisible) {
					$target.show();
				}

				// var endDate = new Date();
				// var elapsedTimeSec = (endDate.getTime() - startDate.getTime()) / 1000;
				// console.log("B) filterListAsync - END --> UID: '" + uid + "' - End date" + endDate + "' - Elapsed time (sec): '" + elapsedTimeSec + "' - List: '" + $target.attr("id") + "'");

				callback(itemCount);
			}
		});
	}

	/*!
	 * Similar to \c populateList but with only one element
	 */
	function appendToList($target, source, $beforeThan) {
		return makeAppender($target)(source, undefined, $beforeThan);
	}

	/*!
	 * This function setups/fills the view with the values
	 * presented in \p model.
	 *
	 * This function looks for any element inside \p $container having the 'ic-data-bind' attribute
	 * and then tries to set its value.
	 *
	 * \param $container jQuery The container
	 *
	 * \param model object Te object holding the data that's going to be displayed
	 *
	 * \param options.disabledFields Array Useful for values that cannot change, any elements having
	 * a ic-data-bind attribute that is present in this array will be disabled.
	 *
	 * \param options.onFieldDisabled function($field) Called when an element is about to be disabled
	 *
	 * \param options.listIterator function(attr, $list, ite, key) A function to call each time an item is
	 * appended to a list. This function is wrapped an used as a iterator of \c populateList.
	 * Very useful when you must to transform some value
	 */
	function displayDataModel($container, model, options) {
		options = _.extend({
			disabledFields : [],
			customLoaders  : {},
			listIterator   : $.noop,
			onFieldDisabled: $.noop
		}, options);

		$container.find('[ic-data-bind]').each(function() {
			var $field = $(this);
			var propertyName = $field.attr('ic-data-bind');

			if (_.contains(options.disabledFields, propertyName)) { // value cannot change, element will be disabled
				$field.prop('disabled', true);
				options.onFieldDisabled($field);
			}

			var value = deepGet(model, propertyName);
			if (value === undefined || value === null) {
				if ($field.is('input') && ($field.is(':checkbox') || $field.is(':radio'))) {
					value = false; // be tolerant with booleans, makes sense to assume false
				} else {
					return; // leave as-is
				}
			}

			if ($field.is('input') || $field.is('select') || $field.is('textarea')) {
				if ($field.is(':checkbox') || $field.is(':radio')) { // support for boolean values
					$field.prop('checked', value);
				} else {
					$field.val(value);
				}
				$field.trigger('change'); // http://api.jquery.com/change/
			} else if ($field.is('ul')) {
				// this one is a list, must have a template element
				populateList($field, value, function(item, key) {
					return options.listIterator(propertyName, $field, item, key);
				});
			} else {
				if (options.customLoaders[propertyName]) {
					options.customLoaders[propertyName]($field, value);
				} else {
					$field.text(value); // any other element
				}
			}
		});
	}

	/*!
	 * This is the inverse of \p displayDataModel, looks in \p $container for
	 * all the elements having a \c ic-data-bind attribute and then tries to get
	 * their value.
	 * \return object having the followin properties
	 * \lu success boolean A value indicating if all values could be read or not
	 * \li fields The object resulting of the lecture
	 * \li missing_fields jQuery[] In case a \c required field is not set
	 *
	 * \remarks ul elements are treaten like arrays of objects. Children elements
	 * should have the attribute ic-inner-data-bind set. This is a limitation of
	 * the current API and is expected to be improved in the future
	 */
	function readDataModel($container, iterator) {
		var result = {
			success       : true,
			missing_fields: [],
			alpha         : [],
			fields        : {}
		};

		var options = _.extend({
				iterator     : function($field) {
					return $field.is(':visible');
				},
				customLoaders: {}
			}, typeof(iterator) === "object" ? iterator : {
				iterator: iterator
			});

		function getValueFromField($field, propertyName) {
			if ($field.is('input') || $field.is('select') || $field.is('textarea')) {
				if ($field.is(':checkbox') || $field.is(':radio')) { // support for boolean values
					return $field.is(':checked');
				}
				return $field.val().trim();
			} else if ($field.is('ul')) {
				var items = [];
				$field.children('[ic-role!="template"]').each(function() {
					var $li = $(this);
					var $properties = $li.find('[ic-inner-data-bind]');
					if ($properties.length === 0) {
						items.push($li.text());
					} else {
						var obj = {};
						$properties.each(function() {
							var $inner = $(this);
							var attr = $inner.attr('ic-inner-data-bind');
							obj[attr] = getValueFromField($inner);
						});
						items.push(obj);
					}
				});
				return items;
			} else {
				if (options.customLoaders[propertyName]) {
					return options.customLoaders[propertyName]($field);
				}
				return $field.text(); // any other element
			}
		}

		$container.find('[ic-data-bind]').each(function() {
			var $field = $(this);
			var propertyName = $field.attr('ic-data-bind');
			if (options.iterator($field) === false) {
				return;
			}
			var value = getValueFromField($field, propertyName);
			if (!value && $field.is('[required]')) {
				result.success = false;
				result.missing_fields.push($field);
			}
			else {
				if (value && $field.is('[letters_number_underscore]')) {
					let validation_result = inConcert.Utils.validate(value, "letters_number_underscore");
					if (!validation_result.success) {
						result.success = false;
						result.alpha.push($field);
					}
				}
			}
			// value is valid
			deepSet(result.fields, propertyName, value);
		});

		return result;
	}

	/*!
	 * Setups navigation in a container
	 *
	 */
	function setupNavigation($container, options) {
		options = _.extend({
			onLeavePage       : $.noop,
			onEnterPage       : $.noop,
			allVisited        : false,
			indexCurrentClass : 'navigationIndexCurrent',
			indexDisabledClass: 'navigationIndexDisabled'
		}, options);

		var currentPage;
		var $index = $container.find('[ic-role="navigation-index"]');
		var $pages = $container.find('[ic-role="navigation-page"]');
		var $next  = $container.find('[ic-role="navigation-next"]');
		var $prev  = $container.find('[ic-role="navigation-prev"]');

		function pageCount() {
			return $index.children(':visible').length;
		}

		function allVisited() {
			return $index.children('.' + options.indexDisabledClass + ':visible').length === 0;
		}

		function jumpToPage(page, direction) {
			if (currentPage !== undefined) {
				var $indexElem = $index.children('[ic-navigation-page="' + currentPage + '"]:visible');
				var $pageElem  = $pages.filter('[ic-navigation-page="' + currentPage + '"]');

				var canLeave = options.onLeavePage(currentPage, $pageElem, direction, $indexElem);
				if (canLeave === false) {
					return false;
				}

				$indexElem.removeClass(options.indexCurrentClass);
				$pageElem.hide();
			}

			currentPage = page;
			var $nextIndex = $index.children('[ic-navigation-page="' + currentPage + '"]:visible');
			var $nextPage  = $pages.filter('[ic-navigation-page="'   + currentPage + '"]');

			$nextPage.show();
			$nextIndex
				.addClass(options.indexCurrentClass)
				.removeClass(options.indexDisabledClass);

			options.onEnterPage(currentPage, $nextPage, allVisited(), $nextIndex);

			$prev.prop('disabled', currentPage === 1);
			$next.prop('disabled', currentPage === pageCount());
		}

		$next.on('click', function() {
			var nextPage = currentPage + 1;
			if (nextPage <= pageCount()) {
				jumpToPage(nextPage, "forward");
			}
		});

		$prev.on('click', function() {
			var nextPage = currentPage - 1;
			if (nextPage > 0) {
				jumpToPage(nextPage, "backward");
			}
		});

		$index.on('click', 'li', function()
		{
			var $tab = $(this);

			var isDisabled = $tab.prop('diabled');
			if (isDisabled) {
				return;
			}
			if ($tab.hasClass(options.indexDisabledClass) || $tab.hasClass(options.indexCurrentClass)) {
				return;
			}

			var pageNumber = parseInt($tab.attr('ic-navigation-page'), 10);
			if ( isNaN(pageNumber) ) {
				return;
			}

			jumpToPage(pageNumber, pageNumber > currentPage ? 'forward' : 'backward');
		});

		if (options.allVisited === true) {
			$index.children()
				.removeClass(options.indexDisabledClass);
		}

		_.defer(function() {
			jumpToPage(1);
		});

		return {
			pageCount  : pageCount,
			allVisited : allVisited,
			enablePage : function(pageNumber) {
				if (isNaN(pageNumber)) {
					return;
				}
				var $tab = $index.find('[ic-navigation-page="' + pageNumber.toString() + '"]');
				$tab.prop("diabled", false);
			},
			disablePage : function(pageNumber){
				if (isNaN(pageNumber)) {
					return;
				}
				var $tab = $index.find('[ic-navigation-page="' + pageNumber.toString() + '"]');
				$tab.prop("diabled", true);
			},
			currentPage: function() {
				return currentPage;
			}
		};
	}

	/*!
	 * On-the-fly text filtering
	 *
	 * \param container The container where to bind the \c keyup event
	 * \param inputFieldSelector A more-specific selector, allows to bind to
	 * multiple filters at the same time
	 * \param targetSelector Selector for the target container (default: \c ul)
	 *
	 * The target container must be a sibling of the input field, unless it
	 * starts with a sharp, aka pound sign, aka number sign, aka hash key :P
	 *
	 * Possible ussages:
	 * \li setupTextFiltering($container)
	 * \li setupTextFiltering($container, 'inputFieldSelector')
	 * \li setupTextFiltering($container, 'inputFieldSelector', 'ul')
	 * \li setupTextFiltering('inputFieldSelector')
	 * \li setupTextFiltering('inputFieldSelector', 'ul')
	 */
	function setupTextFiltering (container, inputFieldSelector, targetSelector) {
		if ( typeof(container) === "string" ) {
			container = $(container);
			targetSelector = inputFieldSelector;
		}

		if (!targetSelector) {
			targetSelector = 'ul';
		}

		var getTarget;
		if (targetSelector[0] === '#') {
			getTarget = function() {
				const $target = $(targetSelector);
				return (function() {
					return $target;
				});
			}();
		}
		else {
			getTarget = function($this) {
				return $this.nextAll(targetSelector);
			};
		}

		const applySearch = function(_, force) {
			const $this = $(this);

			let thisSearch = $this.val().toLowerCase();
			if (thisSearch === $this.data('last-search') && !force) {
				return;
			}
			$this.data('last-search', thisSearch);

			let $target;
			let targetId = $this.attr("ic-list-id");
			if (targetId) {
				$target = $("#" + targetId);
			}
			else {
				$target = getTarget($this);
			}

			let $items = $target.children('[ic-role!="template"]');
			if (thisSearch === "") {
				$items.show();
			} else {
				$items.not   ('[ic-text-search*="' + thisSearch + '"]').hide();
				$items.filter('[ic-text-search*="' + thisSearch + '"]').show();
			}
		};

		if (inputFieldSelector === undefined) {
			container.on('keyup', _.debounce(applySearch, 300));
		} else {
			container.on('keyup', inputFieldSelector, _.debounce(applySearch, 300));
		}

		return {
			apply: function($inputField) {
				$inputField = $inputField || container;
				applySearch.call($inputField[0], null, true);
			}
		};
	}
	
	var scopeDescriptions = {
		Interaction	: {
			all		: "<%=_T('Int.')%>"
		},
		InteractionAgent	: {
			text	: "<%=_T('{0} by Agent')%>",
			all		: "<%=_T('Int.')%>"
		},
		InteractionAttentionLevel	: {
			text	: "<%=_T('{0} by Att Lev.')%>",
			all		: "<%=_T('Int.')%>"
		},
		InteractionCount	: {
			text	: "<%=_T('{0} Counter')%>",
			all		: "<%=_T('Int.')%>"
		},
		Agent	: {
			all		: "<%=_T('Agent')%>"
		},
		AgentStateLapse	: {
			all		: "<%=_T('Ag. States')%>"
		},
		CustomComposite	: {
			all		: "<%=_T('Composite')%>"
		},
		InteractionRule	: {
			all		: "<%=_T('Rules')%>"
		},
		InteractionVoiceBlaster	: {
			all		: "<%=_T('Voice Blaster')%>"
		}
	};
	
	function addScopeToMetricName(metric) {
			var name = metric.name || metric.id;
			if (metric.scope) {
				var conf = scopeDescriptions[metric.scope] || {};
				var text = conf.text || "{0}";
				var all = conf.all || metric.scope;
				name += " - " + format(text, metric.sub_scope? _T(metric.sub_scope) : all);
			}
			return name;
		}

	/*!
	 * Initializes the metric selector, which consist of a list and a set of filters
	 *
	 * \param $container Container where to start looking for elements
	 * \param controller The reports controller
	 * \param iterator Function to call before adding a metric to the list. If \p iterator
	 * returns \c false the metric is ignored
	 * \param callback A function(success, errorThrown) to call when done
	*/
	function setupMetricsSelector(params)
	{
		var container            = params.container;
		var controller           = params.controller;
		var iterator             = params.iterator || $.noop;
		var constructor_callback = params.callback || $.noop;
		var update               = params.update   || $.noop;

		var $filters_container   = container.children('[ic-role="filters"]');
		var $filter_by_scope     = $filters_container.find('[ic-role="filter-by-scope"]');
		var $filter_by_sub_scope = $filters_container.find('[ic-role="filter-by-sub-scope"]');
		var $list_container      = container.children('[ic-role="list-container"]');
		var $metrics_available   = $list_container.find('[ic-role="metrics-available"]');

		var hSearch;
		var availableMetrics;
		var availableScopes;

		function transformComposite(metric) {
			if (!metric.scope && metric.composite) {
				return $.extend({ // make sure we do not modify the metric object (this may be done server-side, btw)
					scope: "CustomComposite"
				}, metric);
			}
		}

		function wrapIterator(func) {
			return function(metric) {
				var result = transformComposite(metric);
				metric = result || metric;
				metric.name_with_scope = addScopeToMetricName(metric);
				if (func(metric) === false) {
					return false;
				}
				var r = iterator(metric);
				if (typeof(r) === "function") {
					return r;
				}
				return r !== false ? result : false;
			};
		}

		var bindTypeFiltering = $.noop;
		var bindScopeFiltering = $.noop;

		var $allFilterByType = $filters_container.find('[ic-role="filter-by-type"] [name="metric-type"]');
		var disabledFilterControls = function(disabled)
		{
			// los radio buttons (ALL, BUILT-IN, CUSTOM)
			$allFilterByType.prop('disabled', disabled);
			// combo scope
			$filter_by_scope.prop('disabled', disabled);
			// combo sub-scope
			$filter_by_sub_scope.prop('disabled', disabled);
		};

		var unbindScopeFiltering = function() {
			$filter_by_scope.off('change');
			$filter_by_sub_scope.off('change');
		};

		var fillScopeList = function(list, callback)
		{
			callback = callback || $.noop;
			//console.log("--> FILL SCOPE LIST");
			unbindScopeFiltering();
			populateList($filter_by_scope, list, { sort: true });
			bindScopeFiltering();
			callback();
		};

		var fillSubScopeList = function(callback)
		{
			callback = callback || $.noop;
			var currentScope = $filter_by_scope.val();

			//console.log("--> FILL SUB-SCOPE LIST '" + currentScope + "'");

			var list = availableScopes[currentScope];
			unbindScopeFiltering();
			populateList(
				$filter_by_sub_scope,
				list, 
				function(subscope) {
					return { value : subscope, name : _T(subscope) };
				},
				{disableOnEmpty: true, sort : true}
			);
			bindScopeFiltering();
			callback();
		};

		var filtering = false;
		var filterAsyncMetrics = function(callback)
		{
			callback = callback || $.noop;

			if (filtering) {
				callback(false);
				return;
			}
			filtering = true;

			//var uid = inConcert.Utils.generateGuid();
			var scope = $filter_by_scope.val();
			var subScope = $filter_by_sub_scope.val();

			// console.log("(+)--> [filterAsyncMetrics] START FILTERING scope: '" + scope + "', subScope: '" + subScope + "' --> UID: '" + uid + "'");

			filterListAsync(
				$metrics_available,
				availableMetrics,

				// la funcion Iteradora
				wrapIterator(function(metric) {
					var isValidScope = (scope === "all" || metric.scope === scope);
					var isValidSubScope = (subScope === "all" || metric.sub_scope === subScope);
					return isValidScope && isValidSubScope;
				}),

				// Los options
				{
					detachContainer: false,
					hideList: false,
					maxAppendPerLoop: 100,
					pauseBetweenLoopMs: 0,
					waitingMsg: "<%=_T('Please wait, filter Metrics rows by Sub-Scope, {0} of {1}.')%>",
					waitingFn: function(/*message*/) {
						// Aca se podria mostrar un waiting si fuese necesario
						// console.log("[filterAsyncMetrics] " + message + "' --> UID: '" + uid + "'");
					},
					// Valida si la fila se puede agregar
					$getItem: function(metric) {
						return $metrics_available.find("[rd-metric='" + metric.id + "']");
					}
				},

				// Esta es la funcion que se invoca cuando la carga de la lista asincronica ha finalizado.
				function(/*itemCount*/)
				{
					filtering = false;
					//console.log("(-)--> [filterAsyncMetrics] END FILTERING scope: '" + scope + "', subScope: '" + subScope + "' --> UID: '" + uid + "'");
					//hSearch.apply(); // reset search
					update();
					callback(true);
				}
			);
		};

		var loading = false;
		var loadAsyncMetrics = function(type, callback = $.noop)
		{
			if (loading) {
				callback(false);
				return;
			}
			loading = true;

			//var uid = inConcert.Utils.generateGuid();
			//console.log("--> [loadAsyncMetrics] START Execute Request type: '" + type + "' --> UID: '" + uid + "'");

			controller.listMetrics(type).done(function(metrics)
			{
				availableMetrics = metrics;
				availableScopes = {};
				var availableScopesList = {};

				//console.log("(+)-- [loadAsyncMetrics] START ADDING Async Metrics type: '" + type + "' --> UID: '" + uid + "'");
				populateListAsync(
					$metrics_available,
					availableMetrics,

					// la funcion Iteradora
					wrapIterator(function(metric) {
						if (metric.scope) {
							availableScopes[metric.scope] = availableScopes[metric.scope] || [];
							if (metric.sub_scope && metric.sub_scope !== "all" && !_.contains(availableScopes[metric.scope], metric.sub_scope)) {
								availableScopes[metric.scope].push(metric.sub_scope);
							}
							availableScopesList[metric.scope] = { name : SCOPES[metric.scope] || metric.scope };
						}
					}),

					// Los options
					{
						maxAppendPerLoop: 100,
						pauseBetweenLoopMs: 5,
						waitingMsg: "<%=_T('Please wait, Adding Metrics rows, {0} of {1}.')%>",
						waitingFn: function(/*message*/) {
							// Aca se podria mostrar un waiting si fuese necesario
							//console.log("[loadAsyncMetrics] " + message + "' --> UID: '" + uid + "'");
						}
					},

					// Esta es la funcion que se invoca cuando la carga de la lista asincronica ha finalizado.
					function(/*itemCount*/)
					{
						//console.log("(-)-- [loadAsyncMetrics] END ADDING Async Metrics type: '" + type + "' --> UID: '" + uid + "'");
						fillScopeList(availableScopesList, function() {
							fillSubScopeList(function() {
								loading = false;
								callback(true);
								update();
								//console.log("-- [loadAsyncMetrics] FINISHED REQUEST type: '" + type + "' --> UID: '" + uid + "'");
							});
						});
					}
				);
			}).fail(function(errorThrown) {
				loading = false;
				callback(false, errorThrown);
			});
		};

		bindTypeFiltering = function()
		{
			// RADIO BUTTON (ALL, BUILT-IN, CUSTOM)
			$filters_container.find('[ic-role="filter-by-type"]')
				.on('change', '[name="metric-type"]', function()
				{
					var $this = $(this);
					disabledFilterControls(true);
					if ($this.is(':checked')) { // seems to be always true in this scope
						loadAsyncMetrics($this.val(), function(success) {
							if (success) {
								disabledFilterControls(false);
							}
						});
					}
				});
		};

		bindScopeFiltering = function()
		{
			$filter_by_scope.on('change', function()
			{
				//var scope = $(this).val();
				//console.log("--> ONCHANGE - FILTER SCOPE scope: '" + scope + "'");
				disabledFilterControls(true);
				fillSubScopeList(function() {
					filterAsyncMetrics(function(success) {
						if (success) {
							disabledFilterControls(false);
						}
					});
				});
			});

			$filter_by_sub_scope.on('change', function()
			{
				//var scope = $filter_by_scope.val();
				//var subScope = $(this).val();
				//console.log("--> ONCHANGE - FILTER SUB SCOPE '" + subScope + "', scope: '" + scope + "'");
				disabledFilterControls(true);
				filterAsyncMetrics(function(success) {
						if (success) {
							disabledFilterControls(false);
						}
				});
			});
		};

		unbindScopeFiltering();
		disabledFilterControls(true);

		// callback is used only the first time
		loadAsyncMetrics('all', function(success, errorThrown) {
			constructor_callback(success, errorThrown, $metrics_available);
			bindTypeFiltering();
			bindScopeFiltering();
			disabledFilterControls(false);
			hSearch = setupTextFiltering($list_container.find('[ic-role="filter-by-content"]'));
		});

		// this handle will allow the user to re-list metrics anytime he wants to
		return {
			reset: function() {
				var $selected = $filters_container.find('[ic-role="filter-by-type"] [name="metric-type"]:checked');
				$selected.trigger('change');
			}
		};
	}

	/*!
	 * \return a list with custom scopes
	 */
	function listCustomScopes() {
		return _.keys(customScopes);
	}

	/*!
	 * \return a list with custom scopes
	 */
	function listTranslatedScopes() {
		return SCOPES;
	}

	/*!
	 * \return \c true if the scope \p scope is custom, otherwise \c false.
	 */
	function isScopeCustom(scope) {
		return customScopes[scope] || false;
	}

	/*!
	 * \return \c true if the scope \p scope is custom, otherwise \c false.
	 */
	function isScopeCustomField(scope) {
		return customFieldsScopes[scope] || false;
	}
	
	/*!
	 * \return dimension name without scope
	 */
	function getCustomDimensionName(name) {
		return name.replace(/^(api|batch_data|process_flow)\./, '');
	}

	/*!
	 * \return true if the \childScope is compatible with \p parentScope
	 *
	 * \remarks This operation is not Commutative.
	 */
	function isScopeCompatible(parentScope, childScope) {
		return (
			parentScope === childScope ||
			(scopeCompatibility[parentScope] && scopeCompatibility[parentScope][childScope]) ||
			false
		);
	}

	/*!
	 * Checks if the dimension \p is compatible with the
	 * scope \p scope, i.e. checks if the \p can be used
	 * in the scpecified scope.
	 */
	function isDimensionCompatible(d, scope) {
		return (
			!d.scope ||
			isScopeCompatible(scope, d.scope) ||
			(d.compatibleWith && d.compatibleWith[scope]) ||
			false
		);
	}

	/*!
	 * Filters dimensions list and gets a unique list of
	 * groups, which is used to fill the list \p $target.
	 *
	 * \return a \c promise object (\c $.Deferred)
	 */
	function loadCustomFieldGroups(controller, $target, scope) {
		return controller.listDimensions("custom").then(_.compose(function(projects) {
			populateList($target, projects, {
				sort: true,
				disableOnEmpty: true,
				defaultItem: {
					value: "",
					label: "<%=_T('No Groups')%>"
				}
			});
			return undefined; // re-using deferred object
		}, _.compact, _.uniq, function(dimensions) {
			return _.map(dimensions, function(dimension) {
				return (scope? dimension.scope === scope : isScopeCustomField(dimension.scope)) && dimension.projectId;
			});
		}));
	}

	/*
	 * Loads tweety projects and puts them in the specified containers
	 *
	 * \param params.controller A reference to reports controller
	 * \param params.tweetyApi A instance of Tweety Client API
	 * \param params.$targetWebchat Element where to append Web Chat projects
	 * \param params.$targetCci Element where to append CCI projects
	 *
	 * \return A promise object providing \c .done(errorThrown)
	 */
	function loadIntegrationProjects(params) {
		var deferred = new $.Deferred();

		params.tweetyApi.load_all_projects(function(success, projects) {
			var errorThrown;

			if (!success) {
				errorThrown = projects;
				projects = [];
			}

			var options = {
				sort: true,
				disableOnEmpty: true,
				defaultItem: {
					id: "",
					name: "<%=_T('No Projects')%>"
				}
			};

			if (params.$targetWebchat) {
				populateList(params.$targetWebchat, projects, makeFilter({type: "chatcampaign" }), options);
			}
			if (params.$targetWebcall) {
				populateList(params.$targetWebcall, projects, makeFilter({type: "webcallcampaign" }), options);
			}
			if (params.$targetCci) {
				populateList(params.$targetCci, projects, makeFilter({ type: "cci" }), options);
			}
			if (params.$targetContactForm) {
				populateList(params.$targetContactForm, projects, makeFilter({type: "contactform"}), options);
			}
			if (params.$targetEmailSurvey) {
				populateList(params.$targetEmailSurvey, projects, makeFilter({type: "survey_form"}), options);
			}
			
			var resolveFn = _.bind(deferred.resolve, deferred);
			var loadCustom = function(callback) {
				if (params.$targetCustomFields) {
					loadCustomFieldGroups(params.controller, params.$targetCustomFields, "api")
						.done(callback)
						.fail(callback);
				} else {
					deferred.resolve(errorThrown);
				}
			};
			var loadBatchData = function(callback) {
				if (params.$targetBatchDataFields) {
					loadCustomFieldGroups(params.controller, params.$targetBatchDataFields, "batch_data")
						.done(callback)
						.fail(callback);
				} else {
					deferred.resolve(errorThrown);
				}
			};
			var loadProcessFlow = function(callback) {
				if (params.$targetProcessFlowFields) {
					loadCustomFieldGroups(params.controller, params.$targetProcessFlowFields, "process_flow")
						.done(callback)
						.fail(callback);
				} else {
					deferred.resolve(errorThrown);
				}
			};
			loadCustom(function() {
				loadBatchData(function() {
					loadProcessFlow(function() {
						resolveFn();
					});
				});
			});
		});

		return deferred;
	}

	/*!
	 * This function setups a set of DOM elements in order to
	 * allow the user to select the available integrations options.
	 *
	 * \param params.container Root element
	 * \param params.controller An instance of \c inConcert.reports.ReportsController, needed
	 * by \c loadIntegrationProjects()
	 * \param params.tweetyApi An instance of \c inConcert.TweetyAPI, needed
	 * by \c loadIntegrationProjects()
	 * \param params.beforeChange A function(category, newScope, newValue), should return \c false to cancel
	 * \param params.afterChanfe A function(category, newScope, newValue), just notifies the change
	 *
	 * \return A \c promise object which provides the \c done() event
	 */
	function configureIntegration(params) {
		params = _.extend({
			container   : undefined,
			controller  : undefined, // needed by loadIntegrationProjects
			tweetyApi   : undefined, // needed by loadIntegrationProjects
			beforeChange: $.noop, //function(category, newScope, newValue)
			afterChange : $.noop //function(category, newScope, newValue)
		}, params);

		var status = {
			frontend: {
				scope   : undefined,
				projectId: undefined
			},
			backend: {
				scope   : "cci",
				projectId: undefined
			},
			batch_data: {
				scope    : 'batch_data',
				projectId: undefined
			},
			process_flow: {
				scope    : 'process_flow',
				projectId: undefined
			}
	};

		// Handle switching between (frontend) integration types (webchat, contactform, api)
		params.container.find('[ic-data-bind="integrationType"]')
			.on('change', function() {
				var $this = $(this);
				var newScope = $this.val();
				var scopeProject = params.container.find('[rd-for-integration-type="' + newScope + '"]')
					.find('[ic-role="select-integration-project"]').val();

				params.beforeChange('frontend', newScope, scopeProject, function(success) {
					if (success === false) {
						$this.val(status.frontend.scope);  // have to revert only this one
					} else {
						status.frontend.scope = newScope;
						status.frontend.projectId = scopeProject; // update project ID with the correct value (the one that matches the scope)
						params.afterChange('frontend', newScope, scopeProject);
					}
				});
			});

		return loadIntegrationProjects({
			controller         : params.controller,
			tweetyApi          : params.tweetyApi,
			$targetWebchat     : params.container.find('[ic-data-bind="webchatProject"]'),
			$targetWebcall     : params.container.find('[ic-data-bind="webcallProject"]'),
			$targetCci         : params.container.find('[ic-data-bind="cciProject"]'),
			$targetContactForm : params.container.find('[ic-data-bind="contactformProject"]'),
			$targetEmailSurvey : params.container.find('[ic-data-bind="surveyProject"]'),
			$targetCustomFields: params.container.find('[ic-data-bind="apiProject"]'),
			$targetBatchDataFields: params.container.find('[ic-data-bind="batch_dataProject"]'),
			$targetProcessFlowFields: params.container.find('[ic-data-bind="process_flowProject"]')
		}).then(function(errorThrown) {
			// operation may fail but we keep going
			var handleChange = function() {
				var $this = $(this);
				var newValue = $this.val();
				var category = $this.parents('[rd-integration-category]').attr('rd-integration-category');

				params.beforeChange(category, status[category].scope, newValue, function(success) {
					if (success === false) {
						$this.val($this.data('rd-current-value'));
					} else {
						$this.data('rd-current-value', newValue);
						status[category].projectId = newValue;
						params.afterChange(category, status[category].scope, newValue);
					}
				});
			};

			params.container.find('[ic-role="select-integration-project"]')
				.on('change', handleChange)
				.each(function() {
					var $this = $(this);
					$this.data('rd-current-value', $this.val());
				});

			return errorThrown;
		});
	}

	/*!
	 * This function setups a boolean expression generator inside the specified container
	 *
	 * There are two tipes of boolean expressions: simple and complex.
	 * Simple expressions are those consisting of a field, a comparator, and a value.
	 * Complex expressions consist of a group of simple or complex expression grouped with
	 * the follwing attributes:
	 * \li A logical connector (or logical operator), i.e. and, or
	 * \li An expected truth value: true or false
	 * A complex expression starts with the following sentence
	 * \li <All|Some> of the following conditions must evaluate to <true|false>
	 * Where 'All' maps to 'and', 'Some' maps to 'or', and then the expression must
	 * evaluate to \c false all sub-expressions are negated.
	 *
	 * \param container jQuery The container where to build the generator
	 * \param options object Options, parameters to pass to the generator
	 *
	 * \li \c options.fieldTemplate  string A themplate \c <option> item to put inside the field selector
	 *
	 * \li \c options.loadAvailableFields function($target, callback, expectedField) This function is
	 * called to fillup a specific field selector. \c $target is the selector, \c callbacks is a
	 * function to call when done, and \c expectedField is the value that is going to be set on
	 * that field (this happens only when loading an expression). \ callback can take an
	 * argument \c disabled, which will indicate the generator that the expression cannot be modified.
	 *
	 * \li \c options.selectedFieldChanged function(params, callback) Called when the user changes the
	 * selection in the field selector. \c params is an object with the following properties: value (string),
	 * source (jQuery), selectedItem (jQuery), expression (jQuery), and \c callback is a function that can
	 * take three parameters: fieldType, placeholder, domain. \c fieldType specifies the type of the
	 * field, \c placeholder matches the \c placeholder of the input value, and \c domain specifies the
	 * allowed values in case \c fieldType is \c enum
	 *
	 * \li \c options.formulaChanged function(expr) Called when the formula changes, takes an string parameter
	 * which is the string representation of the formula being built.
	 *
	 * \return object A handle which provides the follwing functions:
	 * \li \code init : function(expr) \endcode Allows to inject an already generated formula
	 * \li \code reset: function() \endcode Resets the container, therefore the expression, has the same effect than calling init({})
	 * \li \code get  : function() \endcode Returns the formula being built as an object suitable to be read by \c init()
	 * \li \code str  : function() \endcode Returns the formula being built as string
	 */
	function booleanExpressionGenerator(container, options) {
		options = _.extend({
			readonly              : false,
			fieldTemplate         : '<option ic-role="template" value="">Select a field</option>',
			loadAvailableFields   : function($target, callback/*, expectedField*/) {
				callback();
			},
			selectedFieldChanged: function(params, callback) {
				// params: { value, source, selectedItem, expression }
				callback("string", '<%=_T("value...")%>'); // function(fieldType, placeholder)
			},
			formulaChanged: undefined //function(exprstr)
		}, options);

		// anytime the selected field changes, one of these elements takes place
		var fieldsByType = {
			string   : "<input type='text' ic-type='string' class='form-control' placeholder='" + "<%=_T('value...')%>" + "' />",
			number   : "<input type='number' ic-type='number' class='form-control' placeholder='" + "<%=_T('value...')%>" +"' />",
			'boolean': "<select ic-type='boolean' class='form-control'>" +
							"<option value='true' selected>" + "<%=_T('True')%>" + "</option>" +
							"<option value='false'>" + "<%=_T('False')%>" + "</option>" +
						"</select>",
			'enum'   : "<select ic-type='enum' class='form-control'>" +
							"<option ic-role='template' value='{value}' style='display:none'>{label|value}</option>" +
						"</select>"
		};

		// template for action buttons (add, remove, switch)
		var expressionButtonsTemplate = "" +
			"<div ic-role='expression-buttons' class='expression-buttons'>" +
				"<button ic-role='add' class='expression-action-button buttonColorWhite iconOnlyBtn addCondition' type='button' value='' title='" + "<%=_T('Add another condition')%>" + "'><i class='far fa-check'></i></button>" +
				"<button ic-role='switch' class='expression-action-button buttonColorWhite iconOnlyBtn switchCondition' type='button' value='' title='" + "<%=_T('Switch between simple and composite')%>" + "'><i class='far fa-folder-tree'></i></button>" +
				"<button ic-role='remove' class='expression-action-button buttonColorWhite iconOnlyBtn deleteCondition' type='button' value='' title='" + "<%=_T('Remove this condition')%>" + "'><i class='far fa-trash-alt'></i></button>" +
			"</div>";

		// template form simple expressions
		var simpleExpressionTemplate = "" +
			"<div ic-role='expression' ic-expression-type='simple' class='booleanExpression simpleExpression'>" +
				"<div ic-role='expression-andOr' class='expressionAndOr'><span ic-role='expression-and'><%=_T('And')%></span><span ic-role='expression-or'><%=_T('Or')%></span><span ic-role='expression-not'><%=_T('Not')%></span></div>"+
				"<select ic-role='expression-field' class='form-control expression-field'>" + options.fieldTemplate + "</select>" +
				"<select ic-role='expression-comparator' class='form-control expression-comparator'>" +
					"<option value='==' ic-supported-types='number,string,boolean,enum' selected>" + "<%=_T('equal to')%>" + "</option>" +
					"<option value='<>' ic-supported-types='number,string,boolean,enum'>" + "<%=_T('not equal to')%>" + "</option>" +
					"<option value='<'  ic-supported-types='number'>" + "<%=_T('less than')%>" + "</option>" +
					"<option value='<=' ic-supported-types='number'>" + "<%=_T('less than or equal to')%>" + "</option>" +
					"<option value='>'  ic-supported-types='number'>" + "<%=_T('greater than')%>" + "</option>" +
					"<option value='>=' ic-supported-types='number'>" + "<%=_T('greater than or equal to')%>" + "</option>" +
				"</select>" +
				"<div ic-role='expression-value' class='expression-value'>" + fieldsByType.string + "</div>" + expressionButtonsTemplate +
			"</div>";

		// template for complex (aka composite) expressions
		var complexExpressionTemplate = '' +
			"<div ic-role='expression' ic-expression-type='complex' class='booleanExpression complexExpression complexExpressionAnd'>" +
				"<div ic-role='expression-andOr' class='complexAndOr expressionAndOr'><span ic-role='expression-and'><%=_T('And')%></span><span ic-role='expression-or'><%=_T('Or')%></span><span ic-role='expression-not'><%=_T('Not')%></span></div>"+
				"<div ic-role='expression-header' class='complexExpressionHeader'>" + 
					"<select ic-role='logical-operator-connector' class='form-control logical-operator-connector'>" +
						"<option value='and' selected>" + "<%=_T('All')%>" + "</option>" +
						"<option value='or'>" + "<%=_T('Some')%>" + "</option>" +
					"</select>" +
					"<div class='complexExpressionSeparator'>" + "<%=_T('of the following conditions must evaluate to')%>" + "</div>" +
					"<select ic-role='logical-operator-value' class='form-control logical-operator-value'>" +
						"<option value='true' selected>" + "<%=_T('True')%>" + "</option>" +
						"<option value='false'>" + "<%=_T('False')%>" + "</option>" +
					"</select>" +
					expressionButtonsTemplate +
				"</div>" +
				"<ul ic-role='expression-body' class='complexExpressionBody formList'></ul>" +
			"</div>";

		function fadeOut($element, callback) {
			$element.fadeOut().promise().done(callback);
		}

		function createExpressionElement(template) {
			return $('<li></li>')
				.append(template);
		}

		function bindEvent(event, selector, callback) {
			container
				.off(event, selector)
				.on(event, selector, callback);
		}

		// converts formula to string
		function serializeFormula(formula) {
			function not(expr) {
				return 'not(' + expr + ')';
			}

			function quote(value) {
				return typeof(value) === "string" ? ('"' + value + '"') : value;
			}

			switch(formula.type) {
				case 'simple':
					return [formula.field.name, formula.comparator, quote(formula.value)].join(' ');
				case 'complex':
					var expr = _.compact(_.map(formula.childs,
							_.compose(formula.positive ? _.identity : not, serializeFormula) // recursive call
						)).join(' ' + formula.connector + ' ');
					return expr.length > 0 && ( "( " + expr + " )" ) || expr;
			}
		}

		// build the formula from the DOM
		function buildFormula() {

			function getFieldName($field) {
				return {
					id  : $field.val(),
					name: $field.children(':selected').attr('ic-field-name') || $field.val()
				};
			}

			function getFieldValue($field) {
				var value = $field.val();
				return value.match(/^[-+]?\d+$/) ? parseInt(value, 10) : // only integers
					value.match(/^(true|false)$/) ? (value === "true") :
						value;
			}

			function processNode($node) {
				switch ($node.attr('ic-expression-type')) {
					case 'simple':
						var $field = $node.children('[ic-role="expression-value"]')
							.children(':first');
						var result = {
							type      : 'simple',
							field     : getFieldName($node.children('[ic-role="expression-field"]')),
							comparator: $node.children('[ic-role="expression-comparator"]').val(),
							value     : getFieldValue($field)
						};
						if (result.field.name && (result.value || result.value === 0 || !$field.is('input[type="number"]')) ) {
							return result;
						}
						break;
					case 'complex':
						var $header = $node.children('[ic-role="expression-header"]');
						var $subExpressions = $node.children('[ic-role="expression-body"]');
						return {
							type     : 'complex',
							connector: $header.find('[ic-role="logical-operator-connector"]').val(),
							positive : $header.find('[ic-role="logical-operator-value"]').val() === "true",
							childs   : _.compact( _.map($subExpressions.children(),
								_.compose(processNode, future(':first'), _.partial(_.bind, $.fn.children), $) // recursive call
							))
						};
				}
			}

			return processNode(
				container.children('[ic-role="expression"]') // there should be only one child
			);
		}

		function makeExpressionProcessor($targetList, template) {
			return function(expr) {
				var $element = createExpressionElement(template)
					.appendTo($targetList);
				processExpression(expr, $element.children(':first')); // there's only one children
			};
		}

		function appendEmptyExpression($node) {
			makeExpressionProcessor($node.find('[ic-role="expression-body"]'), simpleExpressionTemplate)(
					{ type: 'simple', field: {id: ''}, comparator: '==', value: '' }
				);
		}

		// the node for the expression is already created
		function processExpression(expr, $node) {

			// leave as is
			if ( _.isEmpty(expr) ) {
				return appendEmptyExpression($node);
			}

			switch (expr.type) {
				case 'simple':
					var $expressionField      = $node.children('[ic-role="expression-field"]');
					var $expressionComparator = $node.children('[ic-role="expression-comparator"]');
					var $expressionValue      = $node.children('[ic-role="expression-value"]');
					options.loadAvailableFields($expressionField, function(disabled) {
						if ( options.readonly ) {
							disabled = true;
						}
						$expressionField.val(expr.field.id)
							.prop('disabled', disabled === true)
							.trigger('change');
						$expressionComparator.val(expr.comparator)
							.prop('disabled', disabled === true);
						// use children here 'cause the control changes when setting $expressionField
						$expressionValue.children().val("" + expr.value)
							.prop('disabled', disabled === true);
						if ( disabled ) {
							$node.attr('ic-expression-disabled', true);
							var $buttons;
							if ( options.readonly ) { // all buttons
								$buttons = $node.find('[ic-role="expression-buttons"] > input[ic-role]');
							} else {// without add button
								$buttons = $node.find('[ic-role="expression-buttons"] > input[ic-role!="add"]');
							}
							$buttons.prop('disabled', true);
						}
					}, expr.field.id);
					break;
				case 'complex':
					if ( expr.childs.length === 0) {
						appendEmptyExpression($node);
					} else {
						var $header = $node.children('[ic-role="expression-header"]');
						$header.find('[ic-role="logical-operator-connector"]').val(expr.connector)
							.trigger('change');
						$header.find('[ic-role="logical-operator-value"]').val("" + expr.positive)
							.trigger('change');

						var $subExpressions = $node.children('[ic-role="expression-body"]');

						_.each( _.where(expr.childs, {type: 'simple' }),
							makeExpressionProcessor($subExpressions, simpleExpressionTemplate)
						);

						_.each( _.where(expr.childs, {type: 'complex' }),
							makeExpressionProcessor($subExpressions, complexExpressionTemplate)
						);
					}
					break;
			}
		}

		// avoid updating the formula repeatedly
		var updateFormula = options.formulaChanged && _.debounce(function() {
			var str = serializeFormula(
				buildFormula()
			);
			options.formulaChanged(str);
		}, 250) || _.identity;

		// when the user changes from All to Some or vice versa
		bindEvent('change', '[ic-role="logical-operator-connector"]', function() {
			var $expression = $(this).parents('[ic-role="expression"]:first');
			switch($(this).val()) {
				case "and":
					$expression.removeClass('complexExpressionOr')
						.addClass('complexExpressionAnd');
					break;
				case "or":
					$expression.removeClass('complexExpressionAnd')
						.addClass('complexExpressionOr');
					break;
			}
			updateFormula();
		});

		// when the users changes the expexted result of an expression (true, false)
		bindEvent('change', '[ic-role="logical-operator-value"]', function() {
			var $expression = $(this).parents('[ic-role="expression"]:first');
			switch($(this).val()) {
				case "true":
					$expression.removeClass('complexExpressionFalse'); break;
				case "false":
					$expression.addClass('complexExpressionFalse'); break;
			}
			updateFormula();
		});

		// when changing the selected field, must advice the user in case the type changes
		bindEvent('change', '[ic-role="expression-field"]', function() {
			var $this       = $(this);
			var $selected   = $this.children(':selected');
			var $expression = $this.parents('[ic-role="expression"]:first');

			options.selectedFieldChanged({
				value        : $this.val(),
				source       : $this,
				selectedItem : $selected,
				expression   : $expression
			}, function(fieldType, placeholder, domain) {
				var $valueContainer = $expression.children('[ic-role="expression-value"]')
					.html(fieldsByType[fieldType] || fieldsByType.string);
				if (placeholder) {
					$valueContainer.children().attr('placeholder', placeholder);
				}
				if (fieldType === "enum" && domain) {
					populateList($valueContainer.children(), domain);
				}

				var $comparator = $expression.children('[ic-role="expression-comparator"]');
				$comparator.children(':not([ic-supported-types*="' + fieldType + '"])').hide();
				$comparator.val(
					$comparator.children('[ic-supported-types*="' + fieldType + '"]').show().val()
				);
			});

			updateFormula();
		});

		// when the value changes
		bindEvent('change', '[ic-role="expression-value"]', updateFormula);
		bindEvent('change', '[ic-role="expression-comparator"]', updateFormula);

		// When a new condition is added
		bindEvent('click', '[ic-role="add"]', function() {
			var $this = $(this)
				.prop('disabled', true);

			$this.siblings('[ic-role="switch"]')
				.prop('disabled', true);

			var $container = $this.parents('[ic-role="expression-body"]:first');

			var $elem = createExpressionElement(simpleExpressionTemplate)
				.appendTo($container);

			options.loadAvailableFields(
				$elem.find('[ic-role="expression-field"]'), _.identity
			);

			//updateFormula();
		});

		// When a condition is deleted
		bindEvent('click', '[ic-role="remove"]', function() {
			var $thisExpression = $(this).parents('li:first');
			if ($thisExpression.siblings().length === 0) {
				return;
			}

			if ($thisExpression.next().length === 0) { // enable buttons in previus expression
				var $prevExpression = $thisExpression.prev().children('[ic-role="expression"]'); // there's only one
				$prevExpression.find('[ic-role="add"]')
					.prop('disabled', false);
				if (!$prevExpression.attr('ic-expression-disabled')) {
					$prevExpression.find('[ic-role="switch"]')
						.prop('disabled', false);
				}
			}

			fadeOut($thisExpression, function() {
				$thisExpression.remove();
				updateFormula();
			});
		});

		// When switching from simple to complex conditions, or from complex to simple
		bindEvent('click', '[ic-role="switch"]', function() {
			var previousType = $(this).parents('[ic-role="expression"]:first')
				.attr('ic-expression-type');

			fadeOut($(this).parents('li:first'), function() {
				var $exprli = $(this).html(
					previousType === "simple" ? complexExpressionTemplate : simpleExpressionTemplate
				);
				if (previousType === "simple") {
					// expression is now complex, must append a default item
					appendEmptyExpression($exprli);
				} else {
					options.loadAvailableFields(
						$exprli.find('[ic-role="expression-field"]'), _.identity
					);
				}

				$exprli.fadeIn();
				updateFormula();
			});
		});

		function initFromExpression(expr) {

			var $mainExpression = $(complexExpressionTemplate)
				.appendTo(container.empty());

			$mainExpression.find('[ic-role="expression-header"] > select')
				.prop('disabled', true);
			$mainExpression.find('[ic-role="expression-header"] > [ic-role="expression-buttons"]')
				.remove();

			processExpression(expr, $mainExpression);
		}

		return {

			// setups the container according to the given expression
			init: function(expr) {
				initFromExpression(_.isArray(expr) && {
						type     : 'complex',
						connector: 'and',
						positive : true,
						childs   : expr
				} || expr);
				return this;
			},

			// resets the container, therefore the expression, it's a shortcut for init({})
			reset: function() {
				return this.init([]);
			},

			// returns the boolean expression as an object suitable to be read by \c init()
			get: function() {
				return buildFormula();
			},

			// returns the boolean expression as string
			str: function() {
				return serializeFormula(buildFormula());
			}
		}.init({ });
	}

	if (typeof(inConcert.reports.helpers) !== "object") {
		inConcert.reports.helpers = {
			abstractFunctionCalled     : abstractFunctionCalled,
			deepSet                    : deepSet,
			deepGet                    : deepGet,
			future                     : future,
			makeFilter                 : makeFilter,
			niceFading                 : niceFading,
			hasChildren                : hasChildren,
			askForConfirmation         : askForConfirmation,
			makeAppender               : makeAppender,
			populateList               : populateList,
			populateListAsync          : populateListAsync,
			filterListAsync            : filterListAsync,
			appendToList               : appendToList,
			displayDataModel           : displayDataModel,
			readDataModel              : readDataModel,
			setupNavigation            : setupNavigation,
			setupTextFiltering         : setupTextFiltering,
			setupMetricsSelector       : setupMetricsSelector,
			listCustomScopes           : listCustomScopes,
			isScopeCustom              : isScopeCustom,
			isScopeCustomField         : isScopeCustomField,
			getCustomDimensionName	   : getCustomDimensionName,
			isScopeCompatible          : isScopeCompatible,
			isDimensionCompatible      : isDimensionCompatible,
			loadCustomFieldGroups      : loadCustomFieldGroups,
			loadIntegrationProjects    : loadIntegrationProjects,
			configureIntegration       : configureIntegration,
			booleanExpressionGenerator : booleanExpressionGenerator,
			addScopeToMetricName       : addScopeToMetricName,
			listTranslatedScopes       : listTranslatedScopes
		};
	}
}());
