local json = require "json"
local orbit = require "orbit"
local utils = require "shared.apps.utils"
local api_utils = require "api.utils"
local ir = require "lua_iragentlink"
local base64 = require "base64"

require "inconcert.orbit.application"

local wsapi = require "wsapi"
wsapi.request = require "wsapi.request"
wsapi.request_stream = require "wsapi.request_stream"

local Repository = require "icc_repository"
local cassandra = require "lua_cassandra"
local LuaConfigurationControl = require "LuaConfigurationControl"

local modulename = select(1, ...)
module(modulename, package.seeall, orbit.new, inconcert.orbit.application.Decorate)

_M.model = nil


local model = relativeRequire("model", ...)
local m_models = model.GetModels()

local configControl
local modelBuilt = false
local configControl
local started = false
local repSrvNodes

local m_helpers = require "helpers"
local m_hasRepositoryService = nil

local function HasRepositoryService()
	if m_hasRepositoryService == nil then
		m_hasRepositoryService = m_helpers.utils.HasRepositoryService()
	end
	return m_hasRepositoryService
end

---
-- Some components expect a lualiblogger compatible logger instance.
-- The webhandler does not provide one, but we have the needed methods
-- to fake one.
--
local function make_compatible_logger ()
	return {
		debug = LogDebug,
		info = LogInfo,
		warn = LogWarning,
		error = LogError,
		fatal = LogFatal,
	}
end

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs(checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current

		--Configure RepositoryService Nodes
		if HasRepositoryService() then
			app_status.CheckDependencies(newConfig, {"redis", "cassandra", "repository"})

			local nodes = {}
			for key, value, state in newConfig["repository"]:childs() do
				if state ~= "deleted" then
					LogInfo("Connecting to repositoryservice on %s:%s", value.ip, tostring(value.port))
					table.insert(nodes, { ip = value.ip, port = value.port })
				end
			end
			repSrvNodes = next(nodes) and nodes or repSrvNodes
		else
			app_status.CheckDependencies(newConfig, {"redis", "cassandra"})
		end

		app_status.current = newConfig

		--Configure Redis
		inconcert.environment.apps.repository.redis = inconcert.environment.apps.repository.redis or {}
		inconcert.environment.apps.repository.redis.host = newConfig["redis"]["host"]
		inconcert.environment.apps.repository.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.apps.repository.redis.pwd = newConfig["redis"]["pwd"] or ""

		--Configure Cassandra
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				LogInfo("Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
				if not started and not Repository.init({ cassandra = value, logger = make_compatible_logger(), repositoryNodes = repSrvNodes }) then
					app_status.actualStatus = "invalid"
					return app_status.actualStatus
				else
					LogInfo("Done connecting to cassandra")
					started = true
				end
			end
		end

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end

		app_status.actualStatus = "ready"
		return "ready"
	end
}

---
-- Module initialization. This function is called by the webserver when initializing this application, on each VM.
--
function Initialize (env)
	LogInfo("Initializing Repository")

	-- Notify the ConfigurationServer when we're ready.

	Repository.CASSANDRA_COLUMN_SIZE = env.cassandra_column_size * 1024
	if env.cassandra_max_columns_per_row then
		Repository.CASSANDRA_MAX_COLUMNS_PER_ROW = env.cassandra_max_columns_per_row
	end

	configControl = LuaConfigurationControl.Client {
		type = "AGENT_REPOSITORY",
		address = inconcert.environment.apps.repository.config_server.ip,
		port = inconcert.environment.apps.repository.config_server.port,
		node_address = inconcert.environment.apps.repository.config_server.myIp,
		node_port = inconcert.environment.apps.repository.config_server.myPort,
	}

	if HasRepositoryService() then
		configControl:AddServiceDependency("Repository")
	end

	configControl:TriggerNodeReady()
	return true
end

function ManualStart(cassandra_host, cassandra_port)
	if started then
		return
	end
	local cassandra = {
		host = cassandra_host,
		port = cassandra_port
	}
	Repository.init({ cassandra = cassandra, logger = make_compatible_logger() }) 
	started = true
	LogInfo("Repository manual started")
end


---
-- To be used with multipart/form-data requests.
-- This returns a function that adapts the signature of multipart sinks to our regular ones.
-- The signature of the multipart sink is:
--	o Table containing decoded (name, file) and raw (headers) mime header data
--  o String value containing a chunk of the file data
--  o Boolean which indicates wheather the current chunk is the last one (eof)
--
-- And our regular sink is just:
--  o String value containing a chunk of the file data
--
local function multipart_sink_adapter (sink)
	--LogDebug("[REPOSITORY:storeFile] multipart_sink_adapter: sink %s", sink)
	local mimeHeaders -- unused
	return function (headers, chunk, last)
		--LogDebug("[REPOSITORY:storeFile]      multipart_sink_adapter: chunk %s, last: %s", sink, last)
		if not mimeHeaders then
			mimeHeaders = headers
		end
		sink(chunk)
		if last then
			sink(nil)
		end
	end
end

--
-- Repository API
--

--
-- Stores a file into the repository.
--
-- Data received from the network is divided into 'buckets' of fixed size and stored as columns in Cassandra.
--
-- Take into account that for this to work, the POST request must have a Content-Type header which cannot be
-- x-www-form-urlencoded.
--
-- example url:
-- /inconcert/apps/repository/add/cosas?id=el_id&section=la_Seccion&format=el_formato&contentType=el_content_type
--

function storeFile (wsapi_env, contentType, validateRequest, makeResponse, onEnd)
	contentType = contentType or "text/plain"
	validateRequest = validateRequest or function(request) return 200 end
	makeResponse = makeResponse or function(status, reason) return reason or "" end
	onEnd = onEnd or function() return 200, "" end

	-- Since this is a raw wsapi handler, we need to make sure the proper http verb was used
	local response = wsapi.response.new(200, { ["Content-Type"] = contentType })
	response:write("")
	response.headers["Response"] = "defaultError"
	if wsapi_env.REQUEST_METHOD:lower() ~= "post" then
		response.status = 405
		return response:finish()
	end

	local request = wsapi.request_stream.new(wsapi_env)
	local validationStatus, validationError = validateRequest(request)
	if validationStatus ~= 200 then
		if validationStatus == true then
			validationStatus = 200
		end
		response.status = validationStatus
		response:write(validationError)
		return response:finish()
	end

	--LogDebug("[REPOSITORY:storeFile] 1 - Storing file. Request: %s", json.encode(request))

	if not m_models.Repository then
		model.CreateModelTable()
		m_models = model.GetModels()
	end
	
	local vccId = m_models.Repository.getVcc(request.params)

	local newFile = {
		PartialId         = request.params.id,
		Section           = tonumber(request.params.section)        ,
		Format            = request.params.format                   ,
		ContentType       = request.params.contentType              ,
		FileName          = request.params.fileName                 ,
		ServerPath        = request.params.path                     ,
		CreateIfNotexists = request.params.createIfNotexists        ,
		FileSize          = tonumber(wsapi_env.CONTENT_LENGTH) or 0 ,
		Relation          = request.params.relation                 ,
		proxyType         = request.params.proxyType or "FilesOther",
		vcc               = vccId                                   ,
		TimeStamp         = os.time(os.date("!*t")) -- current date in UTC
	}

	--LogDebug("[REPOSITORY:storeFile] 2 - Storing file. newFile: %s", json.encode(newFile))

	-- Checks that input parameters are valid.

	if newFile.FileSize <= 0 then
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write(makeResponse(false, "Invalid file. Zero length"))
		return response:finish()
	end

	if not newFile.PartialId or not newFile.Section or not newFile.Format or not newFile.ContentType then
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write(makeResponse(false, "Invalid file data"))
		return response:finish()
	end

	-- Create our bucket sink. Provide a context to hold total number of bytes, number of buckets, etc.
	local sink_context = {}

	local function when_completed (ok, err)
		if ok then
			local status, err = onEnd()
			--LogDebug("[REPOSITORY:storeFile]   when_completed - status: %s, err: %s", is_multipart, json.encode(err or {}))
			if status == 200 then
				response.headers["Response"] = "ok"
				response.status = 200
				response:write(makeResponse(true, err))
			else
				response.status = status
				response:write(makeResponse(false, err))
			end
		else
			response.status = 500
			response:write(makeResponse(false, err.msg))
		end
	end

	local is_multipart = type(request.env.CONTENT_TYPE) == "string" and request.env.CONTENT_TYPE:match("^multipart/form%-data; boundary=(.+)$")

	--LogDebug("[REPOSITORY:storeFile] 3 - Storing file. is_multipart: %s", is_multipart)

	-- supply `size_is_known` so we can check that the size of received data matches Content-Size (avoid broken uploads in case of network failure).
	-- Can't do that check with multipart data. I don't always have the proper Content-Length
	local size_is_known = not is_multipart
	local sink = Repository.make_store_sink(newFile, size_is_known, sink_context, when_completed)

	if is_multipart then
		sink = multipart_sink_adapter(sink)
	end

	-- Start sucking data out of the pipe. It will call `when_completed` function and the http reply is
	-- produced there.
	request:parse_message_body(sink)

	-- After 'when_completed' was called and the status was properly set, return the http response
	return response:finish()
end

--[[
	Funcion invocada desde otras Apps, para almacenar archivos en el Repositorio.
	Ejemplo en los luas de migracion de VCCs Allegro a Nativos.
		Despues de obtener desde el Allegro todo el binario, debe realizar la misma logica de subida en chunk que realiza "wsapi"
			cuando los archivos son Uploaded desde la Web.
		Le llegan archivos y se los va proporcionando en partes a la rocak icc_repository.
]]--
function ExternalStoreFile (info)
	assert(type(info == "table"))
	assert(info.Id)
	assert(info.BlobFile)
	assert(info.VCC)
	assert(info.FileName)

	local blobFile = info.BlobFile
	local fileSize = #blobFile
	local fileInfo = {
		ContentType = info.ContentType or "audio/wav",
		FileName    = info.FileName                  ,
		FileSize    = fileSize                       ,
		Format      = info.Format or "wav"           ,
		PartialId   = info.Id                        ,
		proxyType   = info.ProxyType or "FilesOther" ,
		Section     = info.Section or "1"            ,
		TimeStamp   = os.time(os.date("!*t"))
	}

	LogDebug("[REPOSITORY:ExternalStoreFile] 1 - Storing file. fileInfo: %s", json.encode(fileInfo))

	-- el archivo
	local fileName = fileInfo.FileName:lower()
	local temporalFileName = string.format("/tmp/migration_%s", fileName):lower()
	utils.CreateFile(temporalFileName, blobFile)

	local sizeIsKnown = true
	local sinkContext = {}
	-- callback invocado desde la rocka icc_repository, cuando ha finalizado el upload del archivo.
	local function whenCompleted (ok, err)
		LogDebug("[REPOSITORY:ExternalStoreFile] 1--------> whenCompleted - ok: '%s', err: '%s'.", ok, err)
		-- Despues de terminar de subir el archivo al repositorio, agrega el mismo a los Sharedfiles
		local files = {isNew = true}
		files[fileName] = {
			RepositoryId           = info.Id,
			ContentType            = "audio/mpeg",
			Size                   = fileSize,
			is_new                 = true,
			Id                     = fileName,
			Name                   = fileName,
			Description            = fileName,
			Type                   = "audio",
			Language               = "none",
			AgentVisible           = "0",
			AccountMessagesVisible = "1",
			FlowDesignerVisible    = "0"
		}
		local sharedFilesController = utils.GetControllerOthersApps("administration", {controllerName = "shared_files", modelName = "SharedFiles"})
		sharedFilesController.ExternalUpdateSharedFiles(info.VCC, files)
	end
	local sinkCallback = Repository.make_store_sink(fileInfo, sizeIsKnown, sinkContext, whenCompleted)

	LogDebug("[REPOSITORY:ExternalStoreFile] 2 sinkCallback: %s, temporalFileName: %s", sinkCallback, temporalFileName)

	local blobTemporalFile = assert(io.open(temporalFileName, "rb"))
	utils.RemoveFile(temporalFileName)
	--LogDebug("=====> 1- Opened blobTemporalFile: %s", blobTemporalFile)

	-- Esta es la misma logica que utiliza WSAPI para los upload de files.
	local ltn12 = require "ltn12"
	local m = {}
	m.make_source = function ()
		local left = fileInfo.FileSize
		local chunkSize = 4096
		local blocksize = math.min(chunkSize, left)
		--LogDebug("[REPOSITORY:ExternalStoreFile.make_source] ENTRA -> chunkSize: %s, left: %s, blocksize: %s.", chunkSize, left, blocksize)
		return ltn12.source.simplify( function()
			--LogDebug("[REPOSITORY:ExternalStoreFile.make_source]       1- ltn12.source.simplify -> ENTRA left: %s..", left)
			while left > 0 do
				local chunk = blobTemporalFile:read(blocksize)
				--LogDebug("[REPOSITORY:ExternalStoreFile.make_source]       2- ltn12.source.simplify ->  chunk: %s..", chunk)
				if not chunk then
					break
				end
				left = left - #chunk
				blocksize = math.min(blocksize, left)
				--LogDebug("[REPOSITORY:ExternalStoreFile.make_source]       3- ltn12.source.simplify -> left: %s, blocksize: %s.", left, blocksize)
				return chunk, nil
			end
			--LogDebug("[REPOSITORY:ExternalStoreFile.make_source]       4- ltn12.source.simplify -> SALE ..")
			return nil
		end )
	end

	m.parse_message_body = function()
		--LogDebug("[REPOSITORY:ExternalStoreFile.parse_message_body 1]")
		local sourceFn = m.make_source()
		--LogDebug("[REPOSITORY:ExternalStoreFile.parse_message_body 2] sourceFn: %s", type(sourceFn))
		local ret, err = ltn12.pump.all(sourceFn, sinkCallback)
		--LogDebug("[REPOSITORY:ExternalStoreFile.parse_message_body 3 . ret: '%s', err: '%s']", ret, err)
	end

	-- Comienza el parse del archivo
	m.parse_message_body()
end

local function storeFileWeb (wsapi_env)
	return storeFile(wsapi_env)
end

--[[
]]--
local function storePrivateFileWeb (wsapi_env)
	return storeFileWeb(wsapi_env)
end

local function storeInteractionFile (wsapi_env)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local irRequest = nil
	return storeFile(wsapi_env, "application/json",
		function(request)
			irRequest = json.decode(request.params.p)
			if not irRequest.interactionID or not irRequest.proxyType or not irRequest.file then
				return 400, "invalid request"
			end
			local result = irLink:CanAddInteractionFile(irRequest)
			if result.result ~= "OK" then
				LogWarning("storeInteractionFile %s - %s", request.params.p, json.encode(result))
				return 403, "can not send file"
			end
			irRequest.file.id = uuid.new()
			if irRequest.file.extension ~= "" then
				irRequest.file.id = irRequest.file.id .. "." .. irRequest.file.extension
			end
			request.params.id = irRequest.file.id
			request.params.fileName = irRequest.file.name
			request.params.section = -1
			request.params.format = ""
			request.params.contentType = irRequest.file.type
			request.params.proxyType = irRequest.proxyType
			request.params.vcc = irRequest.vcc
			return 200
		end,
		nil,
		function()
			local result = irLink:AddInteractionFile(irRequest)
			if result.result ~= "OK" then
				--error adding file to ir: remove file from repository
				deleteFile(irRequest.file.id, -1)
				LogWarning("storeInteractionFile %s - %s", request.params.p, json.encode(result))
				return 500, "error sending file"
			end
			return 200, json.encode({ result = "OK", id = irRequest.file.id })
		end
	)
end

--[[

]]--
local function storePrivateInteractionFile (wsapi_env)
	return storeInteractionFile(wsapi_env)
end

---
-- Retrieves a file from the repository.
--
-- This needs to accomodate two ways that files could have been stored in the repository.
-- The old way, were the entire file was stored in Cassandra in a column named 'blob' and the new way, were
-- the file is split in N columns.
--
-- The function is a raw wsapi handler since we need to control the way we deliver the response body (piece by piece
--	instead of all at once).
--

function getFile (id, section, part, base64Content, repoKeySpace)
	if section == nil then
		section = -1
	end
	if part == nil then
		part = -1
	end
	LogDebug("[REPOSITORY:getFile] Getting file: id: %s - section: %s - part: %s. Content In Base64: %s", id, section, part, base64Content)
	local exists, metadata = Repository.file_exists(id, section, repoKeySpace)
	if not exists then
		LogError("[REPOSITORY:getFile] File not exists: id: %s - section: %s - part: %s", id, section, part)
		return false, "file not exits"
	end
	if not metadata.blob and not (metadata.NumBuckets and metadata.NumRows and metadata.MaxColsPerRow) then
		LogError("[REPOSITORY:getFile] File not available: id: %s - section: %s - part: %s", id, section, part)
		return false, "file not available"
	end
	local status, content = Repository.retrieve_file(id, section, part, nil, repoKeySpace)
	if not status then
		LogError("[REPOSITORY:getFile] Error getting file: id: %s - section: %s - part: %s", id, section, part)
		return false, "error getting file"
	end
	if base64Content then
		content = base64.encode(content)
	end
	return {
		id = id,
		section = section,
		part = part,
		metadata = metadata,
		content = content
	}
end

function retrieveFile (wsapi_env, validateRequest, withContentType)
	validateRequest = validateRequest or function(request)
		return 200
	end

	local request = wsapi.request.new(wsapi_env)
	local headers = {}
	local status = 404	-- not found by default

	--LogDebug("[REPOSITORY:retrieveFile] Retrieving files: request: '%s'", json.encode(request or {}))

	local make_empty_body = function()
		return coroutine.wrap(function() coroutine.yield("") end)
	end

	-- nota: estos headers response estan mal! para señalar el status se usa, justamente, el status
	headers["Response"] = "defaultError"
	headers["Content-Type"] = "application/octet-stream"

	local validationStatus, validationError = validateRequest(request)
	if validationStatus ~= 200 then
		if validationStatus == true then
			headers["Response"] = "ok"
			return 200, headers, validationError
		else
			LogWarning("[REPOSITORY:retrieveFile] Unable to retrieve file. Data: %s", json.encode(request or {}))
			headers["Response"] = validationError
			return validationStatus, headers, make_empty_body()
		end
	end

	local currentHash = wsapi_env.HTTP_IF_NONE_MATCH
	local id = request.params["id"]
	local section = tonumber(request.params["section"] or "-1")
	local part = request.params["part"] or -1

	if not id then
		LogError("[REPOSITORY:retrieveFile] Invalid file data")
		headers["Response"] = "paramsDataError"
		return status, headers, make_empty_body()
	end

	LogInfo("[REPOSITORY:retrieveFile] Retrieving file: id: %s - section: %s - part: %s", id, section, part)

	local exists, metadata = Repository.file_exists(id, section)
	if not exists then
		LogWarning("[REPOSITORY:retrieveFile] File '%s' not found on cassandra (or not complete yet) (missing metadata)", id)
		return status, headers, make_empty_body()
	end

	--LogDebug("[REPOSITORY:retrieveFile] ---> Retrieving file: currentHash: %s -  metadata.MD5: %s.", currentHash, metadata.MD5)

	-- if the requester has it on cache, tell her to keep it
	if currentHash and #currentHash > 0 and metadata.MD5 == currentHash then
		headers["Set-Cookie"] = "download_started=true; Path=/" -- used by the jquery download plugin
		headers["Response"] = "ok"
		headers["ETag"] = metadata.MD5
		LogInfo("[REPOSITORY:retrieveFile] File '%s' Not Modified.", id)
		return 304, headers, make_empty_body()
	end

	-- if it has no blob (old scheme) and neither has any of the new fields, then it is not there
	if not metadata.blob and not (metadata.NumBuckets and metadata.NumRows and metadata.MaxColsPerRow) then
		-- no blob field nor NumBuckets or NumRows field. Maybe those fields will eventually exist by grace of Cassandra, so
		-- reply with 502 Timeout. Someone should retry... hopefully.
		LogWarning("[REPOSITORY:retrieveFile] File '%s' not found on cassandra (or not complete yet) (missing bucket metadata)", id)
		return 502, headers, make_empty_body()
	end

	local ok, generator = Repository.retrieve_file(id, section, part, true)
	if not ok then
		LogWarning("[REPOSITORY:retrieveFile] File does not exist. id: '%s', section: '%s', part: '%s'.", id, section, part)
		-- esta la metadata, pero no esta el archivo
		return 400, headers, make_empty_body()
	end

	-- Ok, here we know for sure that we have all needed data. Fill the missing pieces...
	headers["ETag"] = metadata.MD5

	if metadata.Format then
		local format = metadata.Format:lower()
		headers["Content-Type"] = (format == "mp3" and "audio/mpeg") or (format == "wav" and "audio/x-wav") or headers["Content-Type"]
	end

	if withContentType and metadata.ContentType then
		headers["Content-Type"] = metadata.ContentType
	end

	local fileName = request.params.name or metadata.FileName
	if string.sub(fileName, -1) == "." then
		fileName = string.sub(fileName, 1, -2)
	end
	headers["Content-Length"]      = metadata.FileSize
	headers["Response"]            = "ok"
	headers["Content-Disposition"] = [[attachment; filename="]] .. fileName .. '"'
	headers["Set-Cookie"]          = "download_started=true; Path=/" -- used by the jquery download plugin

	if request.params["range"] then
		headers["Content-Range"]   = "bytes 0-" .. (metadata.FileSize-1) .. "/" .. metadata.FileSize
		return 206, headers, generator
	end

	-- and now stream the data back to the requester
	return 200, headers, generator
end

local function downloadFileWeb (wsapi_env)
	--LogDebug("[REPOSITORY:downloadFileWeb] Retrieved Web files--> wsapi_env: '%s'.", json.encode(wsapi_env))
	local code, headers, generator = retrieveFile(wsapi_env)
	LogInfo("[REPOSITORY:downloadFileWeb] Retrieved Web files. code: '%s', headers: '%s', generator: '%s'.", code, json.encode(headers or {}), generator)
	return code, headers, generator
end

local function getPublicFileWeb (wsapi_env)
	LogInfo("[REPOSITORY:getPublicFileWeb] Retrieved PUBLIC Web file...")
	return downloadFileWeb(wsapi_env)
end

local function loadInteractionFile (wsapi_env)
	return retrieveFile(wsapi_env,
		function(request)
			if not request.params.interaction or not request.params.type or not request.params.id then
				return 400, "invalid request"
			end
			local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
			local result = irLink:CanRetrieveInteractionFile({
				id            = request.params.id,
				interactionID = request.params.interaction,
				proxyType     = request.params.type
			})
			if result.result ~= "OK" then
				LogWarning("loadInteractionFile type %s interaction %s id %s - %s", request.params.type, request.params.interaction, request.params.id, json.encode(result))
				return 403, "file not available"
			end
			return 200
		end
	)
end

local function getPublicInteractionFile (wsapi_env)
	LogInfo("[REPOSITORY:getPublicInteractionFile] Retrieved PUBLIC Interaction file...")
	return loadInteractionFile(wsapi_env)
end

local function getPublicFile (wsapi_env)
	return retrieveFile(wsapi_env,
		function(request)
			if not request.params.id then
				return 400, "invalid request"
			end
			if not m_models.Repository.HasPublicFile(request.params.id) then
				LogWarning("getPublicFile file %s not available", request.params.id)
				return 403, "file not available"
			end
			return 200
		end, true
	)
end

---
--

function existsFile(file_id, section, part)
	return Repository.file_exists(file_id, section)
end

local function existsFileWeb (web)
	web.headers["Content-Type"] = "application/json"
	local parameters = web.GET
	local file_id, section
	if parameters then
		file_id = parameters.id
		section = parameters.section or -1
		part = parameters.part or -1
	end
	local exists, metadata = existsFile(file_id, section, part)
	return json.encode({ exists = exists and true or false, metadata = metadata or nil })
end

local function existsPrivateFileWeb (web)
	return existsFileWeb(web)
end

---
--

function deleteMultipleFiles (files)
	local result = { deleted = {}, failed  = {} }
	for _, file in ipairs(files or {}) do
		local repo_file_id = Repository.make_file_id(file.id, file.section)
		if Repository.erase_file(file.id, file.section) then
			table.insert(result.deleted, repo_file_id)
		else
			table.insert(result.failed, repo_file_id)
		end
	end
	return result
end

local function deleteMultipleFilesWeb (web)
	web.headers["Response"] = "defaultError"
	LogInfo("Repository: Deleting multiples files: '%s'", web.input["post_data"])
	local files = json.decode(web.input["post_data"])
	local result = deleteMultipleFiles(files)
	return json.encode(result)
end

local function deletePrivateMultipleFilesWeb (web)
	return deleteMultipleFilesWeb(web)
end

---
-- Usage: /inconcert/apps/repository/erase/?id=TU_ID&section=TU_SECTION
--

function deleteFile (file_id, section)
	return Repository.erase_file(file_id, section)
end

function deleteBatchFiles (filesToRemove)
	return Repository.erase_batch_files(filesToRemove)
end

local function deleteFileWeb (web)
	web.headers["Response"] = "defaultError"
	LogInfo("Repository: Deleting one file.")

	local parameters = web.GET
	local file_id, section
	if parameters then
		file_id = parameters.id
		section = parameters.section
	end
	local result = deleteFile(file_id, section)
	return json.encode(result)
end

--[[
]]--
local function deletePrivateFileWeb (web)
	return deleteFileWeb(web)
end

local function downloadToTemporalFile (wsapi_env, response)
	local response = wsapi.response.new(200, { ["Content-Type"]= "text/plain" })
	response:write("")
	response.headers["Response"] = "defaultError"
	if wsapi_env.REQUEST_METHOD:lower() ~= "post" then
		response.status = 405
		return
	end

	local request = wsapi.request_stream.new(wsapi_env)
	setmetatable(request.POST, nil)
	local chunkSize = 2048
	local left = tonumber(wsapi_env.CONTENT_LENGTH) or 0
	local blocksize = math.min(chunkSize, left)
	local tmpFilePath = os.tmpname()
	local tmpFile = io.open(tmpFilePath, "wb")

	LogInfo("[repository:downloadToTemporalFile] Downloading temporal file %s", tmpFilePath, left)

	while left > 0 do
		local chunk = wsapi_env.input:read(blocksize)
		if not chunk then break end

		tmpFile:write(chunk)
		left = left - #chunk
		blocksize = math.min(blocksize, left)
	end

	LogInfo("[repository:downloadToTemporalFile] Downloaded temporal file %s", tmpFilePath)
	tmpFile:close()
	return tmpFilePath
end

local function downloadFromTemporalFile (wsapi_env, response, filePath)
	local file = io.open(filePath, "rb")
	if not file then
		response.status = 405
		return response:finish()
	end
	local content = file:read("*a")
	file:close()
	response:write(content)
	response.headers["Response"] = "ok"
	response.status = 200
	return response:finish()
end

local CONVERT_FILE_OPTIONS = {
	telephony_audio = "-ac 1 -ar 8000 -acodec pcm_s16le"
}

local CONVERT_FILE_EXTENSIONS = {
	telephony_audio = "wav"
}

local function ConvertFile (filePath, fileOptions, fileExtension)
	fileOptions = CONVERT_FILE_OPTIONS[fileOptions] or fileOptions
	fileExtension = CONVERT_FILE_EXTENSIONS[fileExtension] or fileExtension
	local convertedFilePath = string.format("%s.%s", filePath, fileExtension)
	local command = string.format("ffmpeg -i %s %s %s", filePath, fileOptions, convertedFilePath)
	LogInfo("[REPOSITORY:ConvertFile] running %s", command)
	os.execute(command)
	LogInfo("[REPOSITORY:ConvertFile] done %s", command)
	return convertedFilePath
end

local function convertTelephonyAudioFile (wsapi_env)
	local response = wsapi.response.new(200, { ["Content-Type"]= "audio/wav" })
	response:write("")
	response.headers["Response"] = "defaultError"
	--LogDebug("[REPOSITORY:convertTelephonyAudioFile] convert audio to telehpony audio file")
	local tmpFilePath = downloadToTemporalFile(wsapi_env, response)
	if not tmpFilePath then
		return response:finish()
	end
	local convertedFilePath = ConvertFile(tmpFilePath, "telephony_audio", "telephony_audio")
	return downloadFromTemporalFile(wsapi_env, response, convertedFilePath)
end

--[[
	funcion utilizada desde las migraciones de VCCs Allegro a OCC.
		convierte audios al formato permitido en la asterisk de OCC.
]]--
function ConvertAndRetrieveTelephonyAudioFile (file, fileOptions, fileExtension)
	fileOptions = fileOptions or "telephony_audio"
	fileExtension = fileExtension or "telephony_audio"
	local tmpFilePath = os.tmpname()
	utils.CreateFile(tmpFilePath, file)
	local convertedFilePath = ConvertFile(tmpFilePath, fileOptions, fileExtension)
	local success, convertedFile = utils.ReadFile(convertedFilePath)
	utils.RemoveFile(convertedFilePath)
	if not success then
		local err = convertedFile
		LogError("[REPOSITORY:ConvertAndRetrieveTelephonyAudioFile] Unable to convert telephony audio file: '%s'. Err: %s.", filePath, json.encode(err))
		return false, err
	end
	-- ok
	LogInfo("[REPOSITORY:ConvertAndRetrieveTelephonyAudioFile] Successfully converted telephony audio file.")
	return true, convertedFile
end

local function serviceOk (web)
	web.headers["Response"] = "ok"
	return ""
end

---------------------------------------------------------------------------------
-- URLs routed to functions
---------------------------------------------------------------------------------
--
--[[
	URLs Publicas "NO necesitan autenticacion" (inicio de sesion en OCC)
		Para ser publicas tienen que estar definidas en la tabla "excludedUrls" donde se monta la app "repository" en "\serverconfig\xavante\config.lua"
		Uso:
			Estas URLs se utilizan:
				- cuando realizamos request desde otros "repos" (IR, etc)
				- cuando se las envia a los clientes (ejemplo: "".../repository/get/.. en un email para descargar una conversacion)
]]--
_M:dispatch_get  ( deleteFileWeb               , "/delete/(.*)"                  )
_M:dispatch_get  ( existsFileWeb               , "/exists/(.*)"                  )
_M:dispatch_post ( deleteMultipleFilesWeb      , "/multiplesdelete/([^.]*)"      )
_M:dispatch_post ( serviceOk                   , "/serviceok/"                   )
_M:dispatch_wsapi( getPublicFile               , "/public/get/(.*)"              )
_M:dispatch_wsapi( getPublicFileWeb            , "/get/(.*)"                     )
_M:dispatch_wsapi( getPublicInteractionFile    , "/interaction/get/(.*)"         )
_M:dispatch_wsapi( storeFileWeb                , "/add/([^.]*)"                  )
_M:dispatch_wsapi( storeInteractionFile        , "/interaction/add/([^.]*)"      )
---------------------------------------------------------------------------------
--[[
	URLs Privadas "SI necesitan autenticacion" (inicio de sesion en OCC)
		Uso:
			Estas URLs se utilizan:
				- se utilizan para realizar request desde el navegador desde la pagina de OCC.
]]--
_M:dispatch_get  ( deletePrivateFileWeb         , "/erase/(.*)"                   )
_M:dispatch_get  ( existsPrivateFileWeb         , "/is_found/(.*)"                )
_M:dispatch_post ( deletePrivateMultipleFilesWeb, "/multipleserase/([^.]*)"       )
_M:dispatch_wsapi( convertTelephonyAudioFile    , "/convert_telephony_audio/(.*)" )
_M:dispatch_wsapi( downloadFileWeb              , "/download/(.*)"                )
_M:dispatch_wsapi( loadInteractionFile          , "/interaction/load/(.*)"        )
_M:dispatch_wsapi( storePrivateFileWeb          , "/new/([^.]*)"                  )
_M:dispatch_wsapi( storePrivateInteractionFile  , "/interaction/new/([^.]*)"      )
---------------------------------------------------------------------------------
