module(..., package.seeall, lunit.testcase)

local cjson = require "cjson"

local uuid = require "uuid"

local mock = require "wsapi.mock"
local repository = require "apps.repository"

--repository.Initialize({})

local app = mock.make_handler(repository.run)

function test_serviceok()
	
	local response, request = app:post("/serviceok/", nil, {})
	assert_equal(200, response.code)
	assert_equal("ok", response.headers["Response"])
end

function test_missing_file_length()
	
	-- Missing file length
	local response, request = app:post("/add/stuff", nil, {})
	assert_equal(500, response.code)
	assert_equal("paramsError", response.headers["Response"])
end

function test_upload_invalid_content_length()
	
	-- Missing file length
	local response, request = app:post("/add/stuff", nil, { ["Content-Length"] = -1})
	assert_equal(500, response.code)
	assert_equal("paramsError", response.headers["Response"])
	assert_equal("Invalid file. Zero length", response.body)
end

function test_upload_invalid_file_data()
	
	-- Missing metadata fields
	local data = "123"
	local response, request = app:post("/add/stuff", data, { ["Content-Type"] = "text/plain"})
	assert_equal(500, response.code)
	assert_equal("paramsError", response.headers["Response"])
	assert_equal("Invalid file data", response.body)

	--print(cjson.encode(response))
end

function test_upload_small()
	
	local data = ("A"):rep(1024)
	
	local response, request = app:post("/add/?id=some_id&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	response, request = app:get("/get/", { id = "some_id", section="1", format="format", contentType="content_type", part=1})
	assert_equal(200, response.code)
	assert_equal(data, response.body)
	assert_equal([[attachment; filename="foo"]], response.headers["Content-Disposition"])
end

function test_upload_regular()
	
	local data = ("A"):rep(1024) .. ("B"):rep(1024) .. ("C"):rep(1024) .. ("D"):rep(1024)
	
	local response, request = app:post("/add/?id=some_id&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	response, request = app:get("/get/", { id = "some_id", section="1", format="format", contentType="content_type", part=1})
	assert_equal(200, response.code)
	assert_equal(data, response.body)
	assert_equal([[attachment; filename="foo"]], response.headers["Content-Disposition"])
end


function test_upload_big()

	-- 12 megabytes of random data
	local data = {}
	for i = 1,1024 do
		table.insert(data, string.char(math.random(0,255)))
	end
	data = table.concat(data):rep( 1024 * 12 )
	
	local response, request = app:post("/add/?id=other_id&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	response, request = app:get("/get/", { id = "other_id", section="1", format="format", contentType="content_type", part=1})
	assert_equal(200, response.code)
	assert_equal(data, response.body)
	assert_equal([[attachment; filename="foo"]], response.headers["Content-Disposition"])
end
--]=]

function test_upload_huge()

	-- 64 megabytes of random data
	local data = {}
	for i = 1,1024 do
		table.insert(data, string.char(math.random(0,255)))
	end
	data = table.concat(data):rep( 1024 * 64 )
	
	local response, request = app:post("/add/?id=other_id3&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	response, request = app:get("/get/", { id = "other_id3", section="1", format="format", contentType="content_type", part=1})
	assert_equal(200, response.code)
	assert_equal(#data, #response.body)
	assert_equal([[attachment; filename="foo"]], response.headers["Content-Disposition"])
end

function test_upload_incomplete_file()
	-- 6 megabytes of random data
	local data = {}
	for i = 1,1024 do
		table.insert(data, string.char(math.random(0,255)))
	end
	data = table.concat(data):rep( 1024 * 6 )
	
	local response, request = app:post("/add/?id=other_id2&section=1&format=format&contentType=content_type&fileName=foo", 
		data:sub(1, #data / 2), { ["Content-Type"] = "text/plain", ["Content-Length"] = #data * 2 }
	)
	assert_equal(500, response.code)

	response, request = app:get("/get/", { id = "other_id2", section="1", format="format", contentType="content_type", part=1})
	assert_equal(404, response.code)
end

function test_upload_and_delete_file()
	
	local data = ("A"):rep(1024)

	local id = uuid.new()
	
	local response, request = app:post("/add/?id="..id.."&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	response, request = app:get("/delete/"..id)
	assert_equal(200, response.code)

	response, request = app:get("/get/", { id = id, section="1", format="format", contentType="content_type", part=1})
	assert_equal(404, response.code)
end

function test_upload_and_delete_multiple_files()
	
	local data = ("A"):rep(1024)

	local id1, id2 = uuid.new(), uuid.new()
	
	local response, request = app:post("/add/?id="..id1.."&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	response, request = app:post("/add/?id="..id2.."&section=1&format=format&contentType=content_type&fileName=foo", data, { ["Content-Type"] = "text/plain"})
	assert_equal(200, response.code)

	local data = cjson.encode{ id1, id2 }
	response, request = app:post("/multiplesdelete/", data, { ["Content-Type"] = "application/json" })
	assert_equal(200, response.code)

	response, request = app:get("/get/", { id = id1, section="1", format="format", contentType="content_type", part=1})
	assert_equal(404, response.code)
	
	response, request = app:get("/get/", { id = id2, section="1", format="format", contentType="content_type", part=1})
	assert_equal(404, response.code)
end
