local oo = require "loop.base"
local date = require "LuaOrm.Date"
require "LuaOrm.Mapper"
require "inconcert.translators"
local lua_globaltimer = require "lua_globaltimer"
local ir = require "lua_iragentlink"
local permissions = require "shared.permissions.permissions"
local http_mapping = require "apps.agent.result_codes_to_http_codes"
local helpers = require "apps.agent.helpers"
local model_perm = require "shared.permissions.models.permissions"
local snModel = require "apps.social_networks.model"
local models = snModel:GetModels()
local utils = require "shared.apps.utils"
local moses = require "moses"
local model_campaigns = require "apps.agent.models.campaigns"
local utf8 = require "icu.utf8"

local wsapi = require "wsapi"
wsapi.request = require "wsapi.request"
wsapi.request_stream = require "wsapi.request_stream"

require "uuid"

module(..., package.seeall, inconcert.orbit.application.Decorate)

m_htmlRender = nil

local m_globaltimer
local function getNow()
	if not m_globaltimer then
		m_globaltimer = lua_globaltimer.LuaGlobalTimer(inconcert.environment.apps.agent.cluster)
		local timeServers = inconcert.environment.apps.agent.timeServers
		for _, server in pairs(timeServers or {}) do
			m_globaltimer:AddKnownServer(server.IP, server.port)
		end
	end
	return LuaOrm.Date(m_globaltimer:GetTime())
end

local function hasLicensesForChannel (channel, ksAgent)
	local agent = require "apps.agent.agent"
	return agent.hasLicensesForChannel(channel, ksAgent)
end

local function to_string (value)
	return moses.isTable(value) and table.concat(value, " ") or tostring(value)
end

-- wraps a single value in an array (or returns the array)
local function to_table (value)
	return moses.isTable(value) and value or { tostring(value) }
end

local function is_nonempty_string (value)
	return (moses.isString(value) and not moses.isEmpty(value))
end

local function IsPinned (post)
	return post and post.properties and post.properties.pinned or false
end

local function SortPinnedPosts (posts, last, rows, sort)
	local sort_type, sort_order = next(sort or {})
	table.sort( posts, function( a, b )

		if IsPinned(a) and IsPinned(b) then
			return a.publish_timestamp > b.publish_timestamp
		elseif IsPinned(a) then
			return true
		elseif IsPinned(b) then
			return false
		elseif sort_type and a and b then
			if sort_order.order == "desc" then
				return a[sort_type] > b[sort_type]
			else
				return a[sort_type] < b[sort_type]
			end
		end
		return false
	end )

	local filtered = {}
	for i = 1, rows do
		local index = i + last
		if index > #posts then
			break
		end
		table.insert(filtered, posts[index])
	end
	return filtered
end

local function SearchPostsHelper (request)
	LogInfo("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] (START) Executing social networks request: %s", json.encode(request))

	local _T = m_htmlRender.extension.makeTextTranslator()
	local user_name, vcc = inconcert.authUser.userName, inconcert.authUser.virtualContactCenter
	local ksAgent = utils.GetDatasource():CassandraConnection("Agent")

	-- validate the user is filtering for valid campaigns, in case empty add all the campaigns available
	local userCampaigns = model_campaigns.GetUserCampaigns(user_name, vcc, ksAgent)
	if not next(userCampaigns) then
		ksAgent:Close()
		LogInfo("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] User %s does not belong to a campaign, no posts are returned", user_name)
		return 403, { reason = _T("User does not belong to any campaign") }
	end

	local request_query = request.query
	if not request_query then
		ksAgent:Close()
		LogWarning("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] Missing query")
		return 400, { reason = _T("Missing query") }
	end

	local query = {
		query = {
			bool = {
				must = {},
				filter = { }
			}
		}
	}

	local sortPinned = false

	--view filter
	local viewPermission
	local now = getNow():toString("%Y-%m-%d %H:%M:%S")
	if request.view == "ACTIVE" then
		viewPermission = "sn_active_posts"
		request_query.state = { "ACTIVE", "PAUSED" }
		request_query.publish_timestamp = { format = "yyyy-MM-dd HH:mm:ss", lte = now }
		sortPinned = true
	elseif request.view == "SCHEDULED" then
		viewPermission = "sn_scheduled_posts"
		request_query.state = { "ACTIVE", "PAUSED" }
		request_query.publish_timestamp = { format = "yyyy-MM-dd HH:mm:ss", gte = now }
		sortPinned = true
	elseif request.view == "ACTIVE_SCHEDULED" then
		viewPermission = nil
		request_query.state = { "ACTIVE", "PAUSED" }
	elseif request.view == "ARCHIVE" then
		viewPermission = "sn_posts_history"
		request_query.state = { "CLOSED", "DELETED" }
		request_query.publish_timestamp = nil
	else
		ksAgent:Close()
		LogWarning("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] Unrecognized view: %s", request.view)
		return 400, { reason = _T("Invalid view") }
	end

	local applicationsActions = model_perm.GetApplicationsActions(ksAgent, vcc)

	--conditions
	local mustFilters = {
		type = {
			default = models.Posts.GetTypes(),
			filter = function(channel)
				return hasLicensesForChannel(channel, ksAgent)
			end,
			empty = "There are no licenses to any social network channel"
		},
		campaign = {
			default = moses.keys(userCampaigns),
			filter = function(campaign)
				return not viewPermission or model_perm.IsPermissionEnabledForUser("social_networks", viewPermission, campaign, ksAgent, applicationsActions)
			end,
			empty = "User has not permissions to any campaign in this view"
		},
		account = {},
		page = {},
		state = {},
		created_by_user = {},
		publish_timestamp = {
			condition = function(name, filter)
				return { range = { [name] = filter } }
			end
		}
	}
	--add all accounts and pages to avoid getting deleted pages
	if not request_query.account then
		mustFilters.account.default = {}
		mustFilters.page.default = {}
		for _, campaignId in pairs(mustFilters.campaign.default) do
			local accounts = model_campaigns.GetCampaignAccounts(campaignId, ksAgent)
			for _, account in pairs(accounts) do
				if account.proxyType == "TWITTER" or account.proxyType == "YOUTUBE" then
					table.insert(mustFilters.account.default, account.id)
					mustFilters.page.emptyCondition = true
				elseif account.proxyType == "FACEBOOK" or account.proxyType == "INSTAGRAM" then
					table.insert(mustFilters.account.default, account.id)
					if type(account.pages) == "string" then
						account.pages = json.decode(account.pages)
					end
					for _, pageId in pairs(account.pages) do
						table.insert(mustFilters.page.default, pageId)
					end
				end
			end
		end
	end

	local must = query.query.bool.must
	for name, config in pairs(mustFilters) do
		local filter = request_query[name] and to_table(request_query[name]) or config.default or {}
		if config.filter then
			filter = moses.filter(filter, function(_, value)
				return config.filter(value)
			end)
		end
		if not utils.TableIsEmpty(filter) then
			local condition
			if config.condition then
				condition = config.condition(name, filter)
			else
				local filterType
				if #filter > 1 then
					filterType = "terms"
				else
					filterType = "term"
					filter = filter[1]
				end
				condition = { [filterType] = { [name] = filter } }
			end
			if config.emptyCondition then
				condition = {
					["bool"] = {
						["should"] = {
							{
								["bool"] = {
									must_not = {
										exists = { field = name }
									}
								}
							},
							condition
						}
					}
				}	
			end
			table.insert(must, condition)
		elseif config.empty then
			ksAgent:Close()
			LogWarning("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] Empty config: %s", config.empty)
			return 403, { reason = _T(config.empty) }
		end
	end

	--content
	if request_query.content then
		local value = to_string(request_query.content)
		if is_nonempty_string(value) then
			table.insert(query.query.bool.must, {
				wildcard = {
					content = string.format("%s*", value)
				}
			})
			if request.highlight then
				query.highlight = { fields = { content = { foo = 1 } } } -- foo is there to avoid {} to be encoded as []
			end
		end
	end

	-- This field can't be empty, so remove it no query was added
	if #query.query.bool.must == 0 then
		query.query.bool = nil
	end

	--sort
	query.sort = request.sort or { { publish_timestamp = { order = "desc" } } }

	-- Determine if pagination is needed
	query.size = request.rows or 20
	query.from = (tonumber(request.from) or 0)
	if sortPinned then
		query.size = 500
		query.from = 0
	end

	local ok, result = models.Posts.search_elastic(inconcert.authUser.virtualContactCenter, query)
	if not ok then
		ksAgent:Close()
		LogWarning("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] Elasticsearch result: %s", json.encode(result))
		return 500, { result = "REQUEST_FAILED", reason = result.error }
	end
	if result.timed_out then
		ksAgent:Close()
		LogWarning("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] Query against elasticsearch timed out. %s", json.encode(query))
		return 502, { result = "REQUEST_TIMEDOUT", reason = _T("Query timed out") }
	end
	LogInfo("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] Query took %s ms and returned %s hits out of %s", result.took, #result.hits.hits, result.hits.total.value)

	local posts, highlights = {}, {}
	local emptyPost = 0
	local ksInteractions = utils.GetDatasource():CassandraConnection( "Interactions" )
	-- retrieve the full post
	moses.each(result.hits.hits, function(_, hit)
		local post = models.Posts.GetPost(hit._id, true, ksInteractions)
		if not utils.TableIsEmpty(post) then
			table.insert(posts, post)
		else
			-- obtengo el total de post vacios para restarlos al total.
			emptyPost = emptyPost + 1
		end
		if request.highlight then
			highlights[hit._id] = hit.highlight.content
		end
	end)
	if sortPinned then
		posts = SortPinnedPosts(posts, tonumber(request.from) or 0, request.rows or 20, query.sort[1])
	end
	local last = (tonumber(request.from) or 0) + #posts

	local account_statistics = nil
	if request.account_statistics then
		local accountId = request_query.account .. (request_query.page and (":" .. request_query.page) or "")
		account_statistics = models.Posts.GetAccountStatistics(accountId, ksInteractions) or {}
	end
	ksInteractions:Close()
	ksAgent:Close()

	local res = {
		posts = posts,
		total = result.hits.total.value - emptyPost,
		last = last,
		highlights = request.highlight and highlights,
		account_statistics = account_statistics
	}
	--LogDebug("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] --> (END) Executed social networks. res: %s", json.encode(res))

	LogInfo("[SOCIAL_NETWORKS.POSTS:SearchPostsHelper] (END) Executed social networks. total: %d", res.total)

	return 200, res
end

local function SearchPosts (web)
	local status, response = SearchPostsHelper(web.request)
	web.status = status
	return response
end

local function GetPosts (web)
	local post_data =  web.input["post_data"]
	local res, ids = pcall(json.decode, post_data)
	if not res or not ids then
		LogWarning("[SNPost:GetPosts] Invalid request: %s", post_data)
		web.status = 400
		return {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	local posts = {}
	local ksInteractions = utils.GetDatasource():CassandraConnection( "Interactions" )
	for _, id in pairs(ids) do
		local post = models.Posts.GetPost(id, true, ksInteractions)
		if not utils.TableIsEmpty(post) then 
			posts[id] = post
		end
	end
	ksInteractions:Close()
	web.status = 200
	return posts
end

--Actions

local actions = {
	PAUSE = {
		permission		= "sn_pause",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "PauseSNPost"
	},
	RESUME = {
		permission		= "sn_resume",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "ResumeSNPost"
	},
	CLOSE = {
		permission		= "sn_close",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "CloseSNPost"
	},
	REOPEN = {
		permission		= "sn_reopen",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "ReopenSNPost"
	},
	DELETE = {
		permission		= "sn_delete",
		types			= { FACEBOOK = true, TWITTER = true, YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "DeleteSNPost"
	},
	HIDE = {
		permission		= "sn_hide",
		types			= { FACEBOOK = true, YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "HideSNPost"
	},
	PRIVATE = {
		permission		= "sn_hide",
		types			= { YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "PrivateSNPost"
	},
	SHOW = {
		permission		= "sn_hide",
		types			= { FACEBOOK = true, YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "ShowSNPost"
	},
	DISABLE_COMMENTS = {
		permission		= "sn_hide",
		types			= { YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "DisableCommentsSNPost"
	},
	ENABLE_COMMENTS = {
		permission		= "sn_hide",
		types			= { YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "EnableCommentsSNPost"
	},
	EDIT = {
		permission		= "sn_edit",
		types			= { FACEBOOK = true, YOUTUBE = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId", "post" },
		method			= "EditSNPost"
	},
	EDIT_PROPERTIES = {
		permission		= "sn_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId", "properties" },
		method			= "EditSNPostProperties"
	},
	CREATE = {
		permission		= "sn_create",
		required_params = { "proxyType", "campaignAtVCC", "account", "page", "post" },
		method			= "NewSNPost"
	},
	CREATE_FROM_SEARCH = {
		permission		= "sn_twitter_search",
		types			= { TWITTER = true },
		required_params = { "proxyType", "campaignAtVCC", "account", "postId", "message" },
		method			= "CreateFromSearch"
	},
	PIN = {
		permission		= "sn_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "PinSNPost"
	},
	UNPIN = {
		permission		= "sn_edit",
		required_params = { "proxyType", "campaignAtVCC", "account", "postId" },
		method			= "UnpinSNPost"
	},
}

local function RunAction (data, action, applicationsActions)
	data.user = utils.GetCurrentLoggedUser()
	local definition = actions[action]
	if not definition then
		LogWarning("[SNPost] Invalid action: %s for data %s", action, json.encode(data))
		return 400, {result = "INVALID_REQUEST", reason = "Invalid action: " .. action}
	end
	--validate params
	for _, param in pairs(definition.required_params or {}) do
		if not data[param] then
			LogWarning("[SNPost] Invalid request: %s in action %s", json.encode(data), action)
			return 400, {result = "INVALID_REQUEST", reason = "Required param not found: " .. param}
		end
	end
	--validate type
	if definition.types and not definition.types[data.proxyType] then
		LogWarning("[SNPost] Invalid request: %s in action %s", json.encode(data), action)
		return 400, {result = "INVALID_REQUEST", reason = "Action " .. action .. " not supported for type: " .. data.proxyType}
	end
	--validate permissions
	if not model_perm.IsPermissionEnabledForUser("social_networks", definition.permission, data.campaignAtVCC, nil, applicationsActions) then
		LogWarning("[SNPost] Invalid action: %s for data %s", action, json.encode(data))
		return 403, {result = "FORBIDDEN", reason = "User has not permissions to action: " .. action}
	end
	--run action
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local response = irLink[definition.method](irLink, data)
	local status = http_mapping.getHTTPCode(response["result"])
	return status, response
end

local function GetInstagramPost(web)
	local post_data =  web.input["post_data"]
	local res, data = pcall(json.decode, post_data)
	if not res or not data or not data.campaignAtVCC or not data.account or not data.permalink then
		LogWarning("[SNPost] Invalid request: %s in action GetInstagramPost", post_data)
		web.status = 400
		return {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	LogInfo("[SNPost] Searching instagram post %s", post_data)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:GetInstagramPost({
		proxyType = "INSTAGRAM",
		campaignAtVCC = data.campaignAtVCC,
		account = data.account,
		permalink = data.permalink
	})
	local status = http_mapping.getHTTPCode(result["result"])
	if status ~= 200 then
		LogError("[SNPost] Error searching instagram post %s - %s", post_data, json.encode(result))
	end
	web.status = status
	return result
end

local function RunActionFromWeb (web, action)
	local post_data =  web.input["post_data"]
	local res, data = pcall(json.decode, post_data)
	if not res or not data then
		LogWarning("[SNPost] Invalid request: %s in action %s", post_data, action)
		return 400, {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	return RunAction(data, action)
end

local function RunBatchAction (batch, action)
	LogInfo("[SNPost] Running batch action %s for %d posts", action, #batch)
	local vcc = inconcert.authUser.virtualContactCenter
	local applicationsActions = model_perm.GetApplicationsActions(nil, vcc)
	local successful_nodes = {}
	local unsuccessful_nodes = {}

	for _, data in pairs(batch) do
		local id = (data.account or "") .. ":" .. (data.postId or "")
		local status, response = RunAction(data, action, applicationsActions)
		if status == 200 then
			table.insert(successful_nodes, id)
		else
			response.id = id
			table.insert(unsuccessful_nodes, { id = id, status = status, result = response })
		end
	end
	LogInfo("[SNPost] Batch action %s finished with %d successful posts and %d unsuccessful", 
		action, #successful_nodes, #unsuccessful_nodes)
	return 200, { result = "OK", successful_nodes = successful_nodes, unsuccessful_nodes = unsuccessful_nodes }
end

local function RunBatchActionFromWeb(web, action)
	local post_data =  web.input["post_data"]
	local res, batch = pcall(json.decode, post_data)
	if not res or #batch == 0 then
		LogWarning("[SNPost] Invalid request: %s in action %s", post_data, action)
		return 400, {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	return RunBatchAction(batch, action)
end

local function PausePosts (web)
	local status, response = RunBatchActionFromWeb(web, "PAUSE")
	web.status = status
	return response
end

local function ResumePosts (web)
	local status, response = RunBatchActionFromWeb(web, "RESUME")
	web.status = status
	return response
end

local function ClosePosts (web)
	local status, response = RunBatchActionFromWeb(web, "CLOSE")
	web.status = status
	return response
end

local function ReopenPosts (web)
	local status, response = RunBatchActionFromWeb(web, "REOPEN")
	web.status = status
	return response
end

local function DeletePosts (web)
	local status, response = RunBatchActionFromWeb(web, "DELETE")
	web.status = status
	return response
end

local function HidePosts (web)
	local status, response = RunBatchActionFromWeb(web, "HIDE")
	web.status = status
	return response
end

local function PrivatePosts (web)
	local status, response = RunBatchActionFromWeb(web, "PRIVATE")
	web.status = status
	return response
end

local function ShowPosts (web)
	local status, response = RunBatchActionFromWeb(web, "SHOW")
	web.status = status
	return response
end

local function DisablePostsComments (web)
	local status, response = RunBatchActionFromWeb(web, "DISABLE_COMMENTS")
	web.status = status
	return response
end

local function EnablePostsComments (web)
	local status, response = RunBatchActionFromWeb(web, "ENABLE_COMMENTS")
	web.status = status
	return response
end

local function EditPosts(web)
	local status, response = RunBatchActionFromWeb(web, "EDIT")
	web.status = status
	return response
end

local function EditPostsProperties(web)
	local status, response = RunBatchActionFromWeb(web, "EDIT_PROPERTIES")
	web.status = status
	return response
end

local function PinPosts (web)
	local status, response = RunBatchActionFromWeb(web, "PIN")
	web.status = status
	return response
end

local function UnpinPosts (web)
	local status, response = RunBatchActionFromWeb(web, "UNPIN")
	web.status = status
	return response
end

local function CreateFromSearch(web)
	local status, response = RunBatchActionFromWeb(web, "CREATE_FROM_SEARCH")
	web.status = status
	return response
end


local function NewPost(web)
	local status, response = RunActionFromWeb(web, "CREATE")
	web.status = status
	return response
end

local function GetPostRequest(web, requestId)
	local request = models.Posts.GetPostRequest(requestId)
	web.status = 200
	return { result = "OK", request = request }
end

local function ExistsInElastic(web, postId)
	local exists = models.Posts.ExistsInElastic(inconcert.authUser.virtualContactCenter, postId)
	web.status = 200
	return { result = "OK", exists = exists }
end

local function PostImage(web)

	local response = wsapi.response.new(200, { ["Content-Type"]= "text/plain" })
	response:write("")
	response.headers["Response"] = "defaultError"

	local request = wsapi.request_stream.new(web)
	local fileSize = tonumber(web.CONTENT_LENGTH)
	
	if fileSize <= 0 then
		LogError("[SNPost] Unable to upload image. Current file size is %d", fileSize)
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("Invalid file data")
		return response:finish()
	end

	local res, data = pcall(json.decode, request.params.p)
	if not res or not data or not data.proxyType or not data.campaignAtVCC or not data.account or not data.post then
		LogError("[SNPost] Unable to upload image. Incomplete file information.")
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("Invalid file data")
		return response:finish()
	end
	data.page = data.page or ""

	if not model_perm.IsPermissionEnabledForUser( "social_networks", "sn_create", data.campaignAtVCC ) then
		LogError("[SNPost] Unable to upload image. User has not permissions.")
		response.headers["Response"] = "paramsError"
		response.status = 500
		response:write("User has not permissions")
		return response:finish()
	end
	
	LogInfo("[SNPost] Creating post from image in campaign %s account %s page %s", 
		data.campaignAtVCC, data.account, data.page)
	
	-- aca piso la metatabla para poder encargarme yo de la leida del POST data
	setmetatable(request.POST, nil)

	local chunkSize = 2048
	local table_file = {}

	local left = fileSize
	local blocksize = math.min(chunkSize, left)
	while left > 0 do
		local chunk = web.input:read(blocksize)
		if not chunk then break end
		table.insert(table_file, chunk)
		left = left - #chunk
		blocksize = math.min(blocksize, left)
	end
	
	local fileId = uuid.new() .. "." .. data.post.file.extension
	local fileContent = table.concat(table_file)
	models.Posts.SetTemporalFile(fileId, fileContent)
	data.post.file.id = fileId

	local status, data = RunAction(data, "CREATE")
	response.status = status
	response:write(json.encode(data))
	
	return response:finish()
end

local function SearchTweets(web)
	local post_data =  web.input["post_data"]
	local res, data = pcall(json.decode, post_data)
	if not res or not data or not data.campaignAtVCC or not data.account or not data.conditions then
		LogWarning("[SNPost] Invalid request: %s in action SearchTweets", post_data)
		web.status = 400
		return {result = "INVALID_REQUEST", reason = "Invalid request: " .. tostring(post_data)}
	end
	LogInfo("[SNPost] Searching tweets %s", post_data)
	local irLink = ir.IRAgentLink(inconcert.environment.apps.agent.cluster)
	local result = irLink:SearchTweets({
		proxyType = "TWITTER",
		campaignAtVCC = data.campaignAtVCC,
		account = data.account,
		conditions = data.conditions
	})
	local status = http_mapping.getHTTPCode(result["result"])
	if status ~= 200 then
		LogError("[SNPost] Error searching tweets %s - %s", post_data, json.encode(result))
	end
	web.status = status
	return result
end

function MapUris( m, htmlRender, prefix )
	local perm = permissions
	local p = prefix
	local Log = require("apps.agent.requestFilters").LogRequest
	local jsonValidator = utils.JsonValidator
	local vr = utils.JsonValidatorReply
	local r = utils.JsonReply

	m:dispatch_post( Log .. vr	.. SearchPosts, p ..				"/search_posts/(.*)")
	m:dispatch_post( Log .. vr	.. GetPosts, p ..					"/get_posts/(.*)")
	m:dispatch_post( Log .. vr	.. PausePosts, p ..					"/pause_posts/(.*)")
	m:dispatch_post( Log .. vr	.. ResumePosts, p ..				"/resume_posts/(.*)")
	m:dispatch_post( Log .. vr	.. ClosePosts, p ..					"/close_posts/(.*)")
	m:dispatch_post( Log .. vr	.. ReopenPosts, p ..				"/reopen_posts/(.*)")
	m:dispatch_post( Log .. vr	.. DeletePosts, p ..				"/delete_posts/(.*)")
	m:dispatch_post( Log .. vr	.. HidePosts, p ..					"/hide_posts/(.*)")
	m:dispatch_post( Log .. vr	.. PrivatePosts, p ..				"/private_posts/(.*)")
	m:dispatch_post( Log .. vr	.. ShowPosts, p ..					"/show_posts/(.*)")
	m:dispatch_post( Log .. vr	.. DisablePostsComments, p ..		"/disable_posts_comments/(.*)")
	m:dispatch_post( Log .. vr	.. EnablePostsComments, p ..		"/enable_posts_comments/(.*)")
	m:dispatch_post( Log .. vr	.. EditPosts, p ..					"/edit_posts/(.*)")
	m:dispatch_post( Log .. vr	.. EditPostsProperties, p ..		"/edit_posts_properties/(.*)")
	m:dispatch_post( Log .. vr	.. PinPosts, p ..					"/pin_posts/(.*)")
	m:dispatch_post( Log .. vr	.. UnpinPosts, p ..					"/unpin_posts/(.*)")
	m:dispatch_post( Log .. vr	.. NewPost, p ..					"/new_post/(.*)")
	m:dispatch_post( Log .. vr	.. SearchTweets, p ..				"/search_tweets/(.*)")
	m:dispatch_post( Log .. vr	.. CreateFromSearch, p ..			"/create_from_search/(.*)")
	m:dispatch_get ( Log .. r	.. GetPostRequest, p ..				"/get_post_request/(.+)")
	m:dispatch_get ( Log .. r	.. ExistsInElastic, p ..			"/exists_in_elastic/(.+)")
	m:dispatch_wsapi(			   PostImage, p ..					"/post_image/([^.]*)")
	m:dispatch_post( Log .. vr	.. GetInstagramPost, p ..			"/get_instagram_post/(.*)")

	m_htmlRender = htmlRender
end
