local m_models = require("apps.supervisor.model"):GetModels()
local m_utils = require "shared.apps.utils"
local m_permissions = require "shared.permissions.permissions"
local m_modelPermissions = require "shared.permissions.models.permissions"
local m_adminModel, m_adminCampaigns

module(..., package.seeall, inconcert.orbit.application.Decorate)

--[[
	Esta funcion arma los detalles del Usuario.
	Se formatea según lo espera la función "m_adminCampaigns.UpdateMembersList" del controller de administration. En sí es ídem a como se postea desde javascript desde la app "adminisration->campaigns->members"
]]--
local function GetAvailableUserDetail (userId, campaignId, cachedPermissions, keyspace, applicationsActions)
	local vcc = inconcert.authUser.virtualContactCenter
	-- En este array se encuentran unicamente la lista de "columns" que se retornaran de la CF "Users". Esto para que sea mas optimizado.
	local userColumns = {"Name", "FirstSurname", "UserGroup", "IsSystemUser", "Status", "SupervisorWorkgroups", "DefaultCampaign"}
	local member = m_models.Users.Get(vcc, userId, keyspace, userColumns)

	if not member then
		LogWarning("[Supervisor.Campaigns:GetAvailableUserDetail] User '%s' does not exist in vcc '%s'.", userId, vcc)
		return
	end
	if tonumber(member.IsSystemUser) == 1 then
		LogWarning("[Supervisor.Campaigns:GetAvailableUserDetail] Unable to Retrieve SYSTEM User '%s' of vcc '%s'.", userId, vcc)
		return
	end
	if tonumber(member.Status) ~= 1 then
		LogWarning("[Supervisor.Campaigns:GetAvailableUserDetail] Unable to Retrieve DISABLED User '%s' of vcc '%s'.", userId, vcc)
		return
	end

	local simpleUserId = m_utils.GetId(userId)
	local fullUserId = m_utils.AppendVcc(simpleUserId, vcc)
	member.Id = fullUserId

	if not cachedPermissions.RolesPermissions[member.UserGroup] then
	-- cacheo los permisos del rol, para que no tenga que ir a buscarlo por cada usuario a cassandra
		cachedPermissions.RolesPermissions[member.UserGroup] = m_modelPermissions.GetUserGroupPermissions(member.UserGroup, keyspace)
	end

	-- carga si tiene acceso a una app determinada
	local hasPermissionToApp = function(appId, fullCampId)
		-- paso las campana a la cual pertenece el usuario, para que no tenga que ir contra cassandra
		local userCampaigns
		if userCampaigns then
			userCampaigns = {}
			userCampaigns[fullCampId] = true
		end
		-- los override de permisos
		local overrideVccPermissions = m_modelPermissions.GetUserPermissions(fullUserId, keyspace) or m_utils.EMPTY_OBJECT

		return m_permissions.HasPermissionToApplication(
			appId,
			member.Id,
			member.UserGroup,
			fullCampId,
			keyspace,
			{
				group_permissions     = cachedPermissions.RolesPermissions[member.UserGroup],
				campaigns_permissions = cachedPermissions.CampaignsPermissions,
				user_vcc_permissions  = overrideVccPermissions,
				user_campaigns        = userCampaigns,
				applicationsActions   = applicationsActions
			}
		)
	end
	-- de momento se valida contra las apps "interaction" y "supervisor"
	local simpleCampaignId = m_utils.GetId(campaignId)
	local fullCampaignId = m_utils.AppendVcc(simpleCampaignId, vcc)
	member.apps_permisson = {
		interaction = hasPermissionToApp("interaction", fullCampaignId),
		supervisor  = hasPermissionToApp("supervisor", nil) -- si la campana es nil, es porque tiene que buscar el permiso contra todas las campanas
	}

	-- los supervisor workgroups
	if member.apps_permisson.supervisor then
		member.supervisor_workgroups = (member.SupervisorWorkgroups or {})[simpleCampaignId] or {}
	end
	--LogDebug("  %s", json.encode(member))
	return member
end

local function LoadAvailablesMembersDetail (campaignId, keyspace, currentMembers)
	local vcc = inconcert.authUser.virtualContactCenter
	currentMembers = currentMembers or m_models.Campaigns.GetCampaignMembers(vcc, campaignId, keyspace)

	local cachedPermissions = {
		-- Aqui cacheamos los permisos de todas las campanas, para que en la validacion de permisos sobre apps, no tenga que ir a buscar estos permisos a cassandra todas las veces, por cada usuario.
		CampaignsPermissions = m_modelPermissions.GetAllVCCCampaignsPermissions(vcc, keyspace),
		-- En esta variable quedaran cacheados los permisos por roles a los cuales pertenezcan los usuarios.
		RolesPermissions = {}
	}

	local members = {}
	local applicationsActions = m_modelPermissions.GetApplicationsActions(keyspace, vcc)

	for memberId, roleId in pairs(currentMembers) do
		local member = GetAvailableUserDetail (memberId, campaignId, cachedPermissions, keyspace, applicationsActions)
		if member then
			members[m_utils.GetId(member.Id):lower()] = member
		end
	end

	return members, cachedPermissions, applicationsActions
end


local function AdmUpdateMembers (campaignId, membersToUpdated, keyspace, skipUpdate, currentCampaignMembers)
	LogDebug("[SUPERVISOR.CAMPAIGNS:AdmUpdateMembers] Invoke Administration Update members. Campaign: '%s'.", campaignId)
	m_adminModel = m_adminModel or require "apps.administration.model"
	if not m_adminModel.IsCreatedModel() then
		m_adminModel:BuildModels()
	end
	m_adminCampaigns = m_adminCampaigns or require "apps.administration.controllers.campaigns"
	return m_adminCampaigns.UpdateMembersList(campaignId, membersToUpdated, keyspace, nil, nil, nil, nil, skipUpdate, currentCampaignMembers)
end

--[[
	Esta función es la encargada de procesar la lista de members a agregar y/o quitar en una campaña.
	La misma es invocada por el controller member en la función "UpdateSupervisorMembers"; esta función se invoca mediante request desde el lado del cliente.
	Acá recibimos la campaña y la lista de members a agregar o quitar.
	Ej del JSON recibido:
		agents = [
			{userId: 'usuario66', action: 'add'},
			{userId: 'admin', action: 'remove'},
		]
	Como se puede ver lo que llega en el Id del usuario y la acción a realizar.
]]--
function UpdateMembers (campaignId, sourceMembers, skipUpdate)
	assert(campaignId)
	assert(sourceMembers)

	local vcc = inconcert.authUser.virtualContactCenter
	local completeUsername = m_utils.AppendVcc(inconcert.authUser.userName, vcc)
	local startSubProcessTime = os.time()
	LogInfo("[Supervisor.Campaigns:UpdateMembers] User: '%s' Updating supervisor agents on campaign: '%s'.", completeUsername, campaignId)

	--LogDebug("--> sourceMembers: %s", json.encode(sourceMembers))

	-- Esta funcion es la encargada de Crear un member con la estructura exacta que necesita el controller de Administration para poder realizar el ABM.
	local function MakeMemberInfo (simpleMemberId, member)
		return {
			id = simpleMemberId,
			role = member.UserGroup,
			name = ("%s %s"):format(member.Name or "", member.FirstSurname or ""),
			apps_permisson = member.apps_permisson,
			supervisor_workgroups = member.supervisor_workgroups
		}
	end

	-- Se obtiene el array de members que hoy existen en la BD y lo convertimos en un hash con la clave del Id del usuario.
	local keyspace = m_utils.GetDatasource():CassandraConnection("Agent")
	local currentMembers = m_models.Campaigns.GetCampaignMembers(vcc, campaignId, keyspace)
	local databaseAvailablesMembers, cachedPermissions, applicationsActions
	databaseAvailablesMembers, cachedPermissions, applicationsActions = LoadAvailablesMembersDetail(campaignId, keyspace)
	--LogDebug("  databaseAvailablesMembers-- > %s@%s--> %s", campaignId, vcc, json.encode(currentAvailablesMembers))

	local countRemovedMembers = 0
	local canUpdateDB = false
	local membersToUpdated = {}

	--[[
		1- (Removiendo members)
		Acá se remueven los members de la BD según los members que se hayan especificado en el POST(source) con la action="remove".
		Se reocrre la lista de members de la BD, y si el mismo no se encuentra en el POST(source), se debe de incluir en la lista actualizar, ya que sino el mismo sera removido.
		Si el member de la BD existe en el POST (source), se debe de validar que la acción de esté sea "remove" y si es así, el mismo NO se agrega a la lista a actualizar, y con esto,
		estaríamos removiendo el member de la BD.
	]]--
	LogDebug("")
	LogDebug("----------------------------")
	LogDebug("- Start REMOVE Validation")
	LogDebug("----------------------------")
	local i = 0
	for simpleMemberId, databaseMember in pairs(databaseAvailablesMembers) do
		i = i + 1
		local mustAdd
		local membersToInclude
		local sourceMemberInfo = sourceMembers[simpleMemberId]

		--LogDebug("")
		--LogDebug("%d-->  Verificando para REMOVER- > simpleMemberId '%s'  sourceMemberInfo '%s'", i, simpleMemberId, sourceMemberInfo)

		if not sourceMemberInfo then
			--LogDebug("   - No realiza cambios al databaseMember: '%s' '%s'", simpleMemberId, json.encode(databaseMember))
			membersToInclude = MakeMemberInfo(simpleMemberId, databaseMember)
			mustAdd = true
		else
			if sourceMemberInfo.action:lower() == "remove" then
				countRemovedMembers = countRemovedMembers + 1
				LogInfo("[Supervisor.Campaigns:UpdateMembers] %d - REMOVE from BD the User '%s' of campaign: '%s'.", i, simpleMemberId, campaignId)
				canUpdateDB = true
				mustAdd = false
				sourceMembers[simpleMemberId] = nil
				-- no lo incluyo en el array "membersToUpdated" y con esto cuando se invoque a la funcion para que actualice la BD, realiza el remove del mismo.
			--else
				--LogDebug("   - No se debe eliminar este member: '%s', ya que la accion a realizar es; '%s'", simpleMemberId, sourceMemberInfo.action)
			end
		end

		if mustAdd then
			table.insert(membersToUpdated, membersToInclude)
		end
	end
	LogDebug("")
	LogDebug("----------------------------")
	LogDebug("+ Start ADD Validation")
	LogDebug("----------------------------")
	--[[
		2- (Agregando members)
		Acá se agregan los members a la BD según los members que se hayan especificado en el POST(source) con la action="add".
		Se reocrre la lista de members que se han recibido en POST(source), y si el mismo NO se encuentra en la lista de la BD, se debe de incluir en la lista a agregar.
	]]--
	i = 0
	local countAddedMembers = 0

	for simpleMemberId, sourceMemberInfo in pairs(sourceMembers) do
		i = i + 1
		--LogDebug("sourceMemberInfo: '%s'", json.encode(sourceMemberInfo))
		if sourceMemberInfo.action:lower() == "add" then
			local mustAdd
			local membersToInclude
			local databaseMember = databaseAvailablesMembers[simpleMemberId]
			
			--LogDebug("%d-->  Verificando para AGREGAR - > simpleMemberId '%s' databaseMember '%s'", i, simpleMemberId, databaseMember)
			
			if not databaseMember then
				-- Ahora obtengo de la BD los detalles del Usuario
				applicationsActions = applicationsActions or m_modelPermissions.GetApplicationsActions(keyspace, vcc)
				local user = GetAvailableUserDetail(simpleMemberId, campaignId, cachedPermissions, keyspace, applicationsActions)
				LogInfo("[Supervisor.Campaigns:UpdateMembers] %d - ADD to BD the User '%s' in campaign: '%s'.", i, simpleMemberId, campaignId)
				if user then
					-- Creo el usuario segun lo espera el controller de adminsiration
					membersToInclude  = MakeMemberInfo(simpleMemberId, user)
					mustAdd           = true
					canUpdateDB       = true
					countAddedMembers = countAddedMembers + 1
				else
					LogWarning("[Supervisor.Campaigns:UpdateMembers] Unable to add member '%s' in campaigns '%s'.", simpleMemberId, campaignId)
				end
			else
				--LogDebug(" --> member: '%s' already exist", simpleMemberId)
				mustAdd          = true
				membersToInclude = MakeMemberInfo(simpleMemberId, databaseMember)
			end

			if mustAdd then
				--LogDebug("5--> membersToInclude '%s' '%s'", simpleMemberId, json.encode(membersToInclude))
				table.insert(membersToUpdated, membersToInclude)
			end
		end
	end

	if canUpdateDB then
	-- Si exiten cambios a realizar, invoco al controller de administration para que este se encargue de realizar la actualizacion en la BD.
		LogInfo("[Supervisor.Campaigns:UpdateMembers] Adding '%d' and Removing '%d' agents to campaign: '%s'.", countAddedMembers, countRemovedMembers, campaignId)
		LogDebug("[Supervisor.Campaigns:UpdateMemberss] Elapsed time %s seconds, before AdmUpdateMembers", m_utils.GetElapsedTimeSec(startSubProcessTime))
		--LogDebug("===> membersToUpdated: '%s'", json.encode(membersToUpdated))
		-- Le paso el keyspace, ya que el controller de administration en esta funcion, realiza varias conexiones a cassandra.
		--El ultimo param es para skipear el update, cuando se agregan o eliminan usuarios. Esto para mejorar el performance.
		local ok, err = AdmUpdateMembers(campaignId, membersToUpdated, keyspace, skipUpdate, currentMembers)
		if ok then
			LogInfo("[Supervisor.Campaigns:UpdateMembers] Success updated '%d' agents on campaign: '%s'.", #membersToUpdated, campaignId)
		else
			keyspace:Close()
			LogWarning("[Supervisor.Campaigns:UpdateMembers] Error Updating agents on campaign: '%s'. Result: '%s'", campaignId, json.encode(err))
			m_utils.CollectAllMemory()
			return false, err
		end
	else
		LogWarning("[Supervisor.Campaigns:UpdateMembers] There are no updates to be made in campaign: '%s'.", campaignId)
		--LogDebug("===> membersToUpdated: '%s'", json.encode(membersToUpdated))
	end

	keyspace:Close()
	-- X las dudas nos cuidamos ya que acá se realizan varias idas a cassandra y quizas con muchos datos
	m_utils.CollectAllMemory()
	LogDebug("[Supervisor.Campaigns:UpdateMembers] ...END sub process . Elapsed time %s seconds.", m_utils.GetElapsedTimeSec(startSubProcessTime))
	return true
end

function UpdateAttentionLevelsMembers(campaignId, attentionLevelsUpdate)
    assert(campaignId)
    assert(attentionLevelsUpdate)

    local vcc = inconcert.authUser.virtualContactCenter
    local completeUsername = m_utils.AppendVcc(inconcert.authUser.userName, vcc)

    LogInfo("[Supervisor.Campaigns:UpdateAttentionLevelsMembers] User: '%s' is updating agents the attention levels defined in the campaign: '%s'", completeUsername, campaignId)
    m_adminModel = m_adminModel or require "apps.administration.model"

    if not m_adminModel.IsCreatedModel() then
        m_adminModel:BuildModels()
    end
    
	m_adminCampaigns = m_adminCampaigns or require "apps.administration.controllers.campaigns"
	
	return m_adminCampaigns.UpdateAttentionLevelsMembers(campaignId, attentionLevelsUpdate)
end

m_htmlRender = nil
function MapUris( m, prefix, htmlRender)
	-- POR AHORA NO MAPEO NINGUNA URL->FUNCION, pero lo agrega la funcion por compatibilidad.
	m_htmlRender = htmlRender
end
