require("orbit")
require("inconcert.orbit.application")
require("inconcert.translators")
require("lfs")
require("json")
local supervisorshared = require("supervisorshared")
require "reporter"
local LuaConfigurationControl = require "LuaConfigurationControl"
local cassandra = require "lua_cassandra"
local utils = require "shared.apps.utils"
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")

module(..., package.seeall, orbit.new, inconcert.orbit.application.Decorate)

_M.model = nil

local model = relativeRequire("model", ...)
local m_models = model.GetModels()
local configControl
local modelBuilt = false

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	actualStatus = "waiting",

	CheckDependencies = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			if newConfig[dep] == nil or newConfig:DiffState(dep) == "deleted" then
				LogError("Dependency '%s' is not met", dep)
				app_status.actualStatus = "invalid"
				error("Unmet dependency "..dep)
			end
		end
	end,

	CheckDependencyList = function (newConfig, dep)
		for key, value, state in newConfig[dep]:childs() do
			if state ~= "deleted" then
				return
			end
		end
		LogError("Dependency '%s' is empty", dep)
		app_status.actualStatus = "invalid"
		error("Empty dependency "..dep)
	end,

	CheckDependenciesList = function (newConfig, checkList)
		for _, dep in ipairs (checkList) do
			app_status.CheckDependencyList(newConfig, dep)
		end
	end,

	--Actually this is the only function called from the appReadyFilter
	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged and app_status.actualStatus ~= "invalid" then
			return app_status.actualStatus
		end

		newConfig = newConfig or app_status.current

		app_status.CheckDependencies(newConfig, {"agent", "redis", "cassandra", "elasticsearch"})

		if not inconcert.environment.apps.supervisor then
			inconcert.environment.apps.supervisor = {}
		end

		app_status.current = newConfig

		--Configure Redis
		inconcert.environment.apps.supervisor.redis = inconcert.environment.apps.supervisor.redis or {}
		inconcert.environment.apps.supervisor.redis.host = newConfig["redis"]["host"]
		inconcert.environment.apps.supervisor.redis.port = tonumber(newConfig["redis"]["port"])
		inconcert.environment.apps.supervisor.redis.pwd = newConfig["redis"]["pwd"] or ""

		inconcert.environment.apps.supervisor.agentRedis = inconcert.environment.apps.supervisor.agentRedis or {}
		inconcert.environment.apps.supervisor.agentRedis.host = newConfig["agentredis"]["host"]
		inconcert.environment.apps.supervisor.agentRedis.port = tonumber(newConfig["agentredis"]["port"])
		inconcert.environment.apps.supervisor.agentRedis.pwd = newConfig["redis"]["pwd"] or ""

		inconcert.environment.apps.supervisor.elasticSearch = inconcert.environment.apps.supervisor.elasticSearch or {}
		for key, value, state in newConfig["elasticsearch"]:childs() do
			if state ~= "deleted" then
				table.insert(inconcert.environment.apps.supervisor.elasticSearch, value)
			end
		end

		--Configure Cassandra
		inconcert.environment.apps.supervisor.cassandra = inconcert.environment.apps.supervisor.cassandra or {}
		local cassandraIP, cassandraPort
		for key, value, state in newConfig["cassandra"]:childs() do
			if state ~= "deleted" then
				table.insert(inconcert.environment.apps.supervisor.cassandra, value)
				LogInfo("Connecting to cassandra on %s:%s", value.host, tostring(value.port))
				cassandra.seed(value.host, value.port)
				cassandra.ip = value.host
				cassandra.port = value.port
				if not cassandraIP then
					cassandraIP = value.host
					cassandraPort = value.port
				end
			end
		end

		-- Global Timer
		inconcert.environment.apps.supervisor.timeServers = inconcert.environment.apps.supervisor.timeServers or {}
		for key, value, state in newConfig["timeServers"]:childs() do
			if state ~= "deleted" then
				table.insert(inconcert.environment.apps.supervisor.timeServers, value)
			end
		end

		--Configure outboundengineservice
		inconcert.environment.apps.supervisor.outboundengineservice = {}
		local outboundengineservice = inconcert.environment.apps.supervisor.outboundengineservice
		for key, value, state in newConfig["outboundengineservice"]:childs() do
			if state ~= "deleted" then
				outboundengineservice[#outboundengineservice + 1] = {host = value.ip, port = value.port}
			end
		end

		-- Reporter
		if cassandraIP then
			reporter.ConnectCassandra(cassandraIP, cassandraPort)
			reporter.SetGlobalTimer(inconcert.environment.apps.agent.cluster)
		end

		-- Initialize supervisor shared
		supervisorshared.DataFilter():Initialize(inconcert.environment.apps.supervisor)

		if modelBuilt == false then
			model:BuildModels()
			modelBuilt = true
		end

		app_status.actualStatus = "ready"
		return "ready"
	end
}

function Initialize ()
	LogInfo("Starting application 'supervisor'")

	configControl = LuaConfigurationControl.Client {
		type = "SUPERVISOR",
		address = inconcert.environment.apps.agent.config_server.ip,
		port = inconcert.environment.apps.agent.config_server.port,
		node_address = inconcert.environment.apps.agent.config_server.myIp,
		node_port = inconcert.environment.apps.agent.config_server.myPort
	}

	configControl:AddServiceDependency("Agent")
	configControl:AddServiceDependency("Administration")
	configControl:TriggerNodeReady()

	return true
end

local function makeTextTranslator ()
	local translator = inconcert.translators.supervisor
	return function(...)
		return translator:Text(...)
	end, i18nlib.LCID[translator.language]
end

function GetApplicationTitle ()
	local _T = makeTextTranslator()
	return _T("supervisor")
end

local htmlRender = require"inconcert.render".new("supervisor", {
	MainTitle = _M.GetApplicationTitle,
	Title = function() return _T("supervisor") end,
	model = nil,
	makeTextTranslator = makeTextTranslator
})

-----------------------
-- (cotrolles) urls/apis
function GetControllers ()
	local controllerDir = m_appLocation .. "/controllers"
	local controllers = {}
	for file in lfs.dir(controllerDir) do
		if file ~= "." and file ~= ".." then
			local f = controllerDir..'/'..file
			local attr = lfs.attributes (f)
			if attr.mode ~= "directory" then
				local controllerName = string.match(file, [[(.+).lua]])
				if controllerName then
					controllers[controllerName] = { url = controllerName }
				end
			end
		end
	end
	return controllers
end

for key, controllerDef in pairs(GetControllers()) do
	--requires y mapeo funciones
	local controller = relativeRequire("controllers." .. key, ...)
	controller.MapUris(_M, "/"..controllerDef.url, htmlRender)
end

function SummaryGetRealtimeData (web)
	return true
end

local function CheckDependencies(web)
	app_status.CheckStatus()

	return true
end

--scripts and pages

--paginas
local function render_pages (web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = { _T = _T, _L = _L }
	return htmlRender:RenderPageLP("/apps/supervisor/pages/" .. name .. ".html", web, env)
end

--scripts
function render_script (web, name, versionCode, languageCode)
	if (languageCode ~= "") then
		name = name .. "." .. versionCode
	else
		languageCode = versionCode
	end
	inconcert.authUser = { languageNumber = tonumber(languageCode) }
	local _T, _L = makeTextTranslator()
	local env = 
		{_T = _T,
		_L = _L,
		m_utils = utils
	}
	return htmlRender:ScriptLp("/apps/supervisor/scripts/" .. name .. ".js", web, env)
end

_M:dispatch_get(render_script, "/scripts/([^.]+)%.([^.]+)%.?([^.]*)%.js")
_M:dispatch_get(render_pages, "/pages/([^.]+)%.([^.]+)%.?([^.]*)%.html")

function serve_file (app_module, cache)
	return function (web)
		local filename = string.sub(web.path_info, 2, #web.path_info)
		if not cache then
			app_module.response.DontCache(web)
		end
		return _M:serve_static(web, m_appLocation.."/"..filename)
	end
end

-------------------------------------------------------------------------
_M:dispatch_post(SummaryGetRealtimeData, "/summary/get_realtime_data/(.*)")
_M:dispatch_post(CheckDependencies, "/check_dependencies/")
_M:dispatch_get(_M:serve_file(), "/images/.+", "/styles/.+", "/css/.+")
