module(..., package.seeall)

local utils = require "shared.apps.utils"
require "LuaOrm.Date"

local m_datasource

function GetName()
	return "Indicators"
end

function GetDefinition (datasource)
	m_datasource = datasource
	return _M
end

function CampaignExists (campaign_at_vcc, keyspace)
	return utils.GetColumnValue({
		keyspace = keyspace or "Agent",
		columnFamily = "Campaigns",
		key = campaign_at_vcc,
		column = "Id"
	}) or {}
end

function GetCampaignAccounts (campaign_without_vcc, vcc, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		superColumns = true,
		columnFamily = "CampaignAccounts",
		key = utils.AppendVcc(campaign_without_vcc, vcc),
	} or {}
end

function GetCampaignAccount (campaign_without_vcc, vcc, account_id, keyspace)
	local accounts = GetCampaignAccounts(campaign_without_vcc, vcc, keyspace)
	for _, account in pairs(accounts) do
		if account.id == account_id then
			return account
		end
	end
end

function GetCampaignAccountTypes (campaignId, vcc, keyspace)
	local types = {}
	for _, account in pairs(GetCampaignAccounts(campaignId, vcc, keyspace)) do
		if account.proxyType == "CALL" then
			if account.incomingConfig and account.incomingConfig.enabled then
				types["INBOUND_CALL"] = true
				types["VOICEMAIL"] = true
			end
			if account.outgoingConfig and account.outgoingConfig.enabled then
				types["OUTBOUND_CALL"] = true
			end
		else
			types[account.proxyType] = true
		end
	end
	return types
end

function GetCampaignAttentionLevels (campaign_with_vcc, keyspace)
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "CampaignAttentionLevels",
		key = campaign_with_vcc
	} or {}
end

-- function GetHistoricalSerie (indicator, campaignAtVcc, slice, startDate, endDate, keyspace)
	-- return utils.GetAllColumns {
		-- keyspace = keyspace or "Historical",
		-- columnFamily = "InteractionCounters",
		-- key = indicator..":"..campaignAtVcc..":"..slice,
		-- startColumn = startDate,
		-- endColumn = endDate,
		-- valueAsNumber = true,
		-- columnCount = 500, --TODO: calculate correct count
	-- }
-- end

--
-- params = { campaign = "camp1", vcc = "vcc1", cut = "cut1", slice = "slice1", hours_ago = "hours_ago", indicators = {}, account_type = "account_type1"}
--
function GetHistoricalReport (params)
	--LogDebug("[Supervisor:GetHistoricalReport] - Model params: '%s'", json.encode(params))
	local vcc             = params.vcc
	local campaign        = params.campaign
	local account         = params.account
	local attention_level = params.attention_level
	local cut             = params.cut
	local slice           = params.slice
	local hours_ago       = params.hours_ago
	local indicators      = params.indicators
	local account_type    = params.account_type
	local initiative      = params.initiative

	-- Generate the report
	local now = LuaOrm.Date():toDifferentTimezone(inconcert.authUser.timezone)
	local sliceStartTable = {
		year  = now:year(),
		month = now:month(),
		day   = now:day(),
		hour  = now:hour(),
		min   = now:minute()
	}

	local increment = { 1 }
	if slice == "15Min" then
		sliceStartTable.min = sliceStartTable.min - (sliceStartTable.min % 15)
		increment = { 0, 0, 0, 0, 15 }
	elseif slice == "30Min" then
		sliceStartTable.min = sliceStartTable.min - (sliceStartTable.min % 30)
		increment = { 0, 0, 0, 0, 30 }
	elseif slice == "Hour" then
		sliceStartTable.min = 0
		increment = { 0, 0, 0, 1 }
	elseif slice == "Day" then
		sliceStartTable.min = 0
		sliceStartTable.hour = 0
		increment = { 0, 0, 1 }
	end

	local nowSliceStart = LuaOrm.Date(sliceStartTable)
	local fromSliceStart = nowSliceStart:add(0, 0, 0, -hours_ago)
	local sliceHours = {
		["15Min"] = 0.25,
		["30Min"] = 0.5,
		["Hour"]  = 1,
		["Day"]   = 24,
		["Month"] = 24 * 30
	}

	local report = {
		cut = cut,
		sliceSize = slice,
		indicators = {},
		filter = {
			VCC = vcc,
			Campaign       = campaign,
			Account        = account,
			AttentionLevel = attention_level,
			DateFrom       = fromSliceStart:toString(),
			DateTo         = nowSliceStart:toString(),
			Type           = account_type
		},
		vcc = vcc,
		reverse = true,
		resultSize = hours_ago / (sliceHours[slice] or 0.25) + 1,
		timezone = inconcert.authUser.timezone
	}

	for _, indicator in ipairs(indicators) do
		local scope = indicator.scope[initiative] or indicator.scope.all
		if scope then
			local summaryId
			if (not attention_level or attention_level == "") then
				summaryId = scope.summaryId
			else
				summaryId = scope.summaryIdAttentionLevel
			end
			table.insert(report.indicators, summaryId)
		end
	end

	-- Ejecuto contra los Reportes el Query de Summary
	local _T_Dummy = function(s) return s end, {}
	local idsToNames = nil
	local redis = utils.GetRedisClient()
	local authFilter = nil
	local result, handle, err = reporter.QuerySummaryFromCache(report, utils, _T_Dummy, idsToNames, redis, authFilter)
	--local result, handle, err = reporter.QuerySummary(report, utils, _T_Dummy, idsToNames, redis, authFilter)
	if not result then
		err = err or {}
		err.result = err.errType
		LogWarning("[Supervisor.inidicaors:GetHistoricalReport] Unable to retrieve reporter.QuerySummaryFromCache. err: %s", err.result)
		return false, err
	end

	-- Fill processedData
	local processedData = {}
	for _, indicator in ipairs(indicators) do
		processedData[indicator.id] = {
			id = indicator.id,
			description = indicator.description,
			data = {}
		}
	end

	local sliceStart = fromSliceStart
	while sliceStart:time() < nowSliceStart:time() do
		local candidateRow
		while #result.rows > 0 do
			local lastRow = result.rows[#result.rows] -- This is the earliest one
			local slicePeriod = lastRow[1]
			if slice == "15Min" or slice == "30Min" or slice == "Hour" then
				-- It comes as "YYYY-mm-dd HH:MM - HH':MM'". Convert it to a YYYY-mm-dd HH:MM:00.
				slicePeriod = slicePeriod:match("^(.*) %- ")..":00"
			end
			local timeDate = LuaOrm.Date(slicePeriod)
			local timestamp = timeDate:time()
			if timestamp > sliceStart:time() then
				break
			end
			--LogDebug("[Supervisor] - Processed data for %s (at %s)", tostring(timeDate), tostring(sliceStart))
			result.rows[#result.rows] = nil
			if timestamp == sliceStart:time() then
				--LogDebug("[Supervisor] - Candidate row is %s (at %s)", tostring(timeDate), tostring(sliceStart))
				candidateRow = lastRow
				break
			end
		end
		if candidateRow then
			--LogDebug("[Supervisor] - Candidate row for %s is %s", tostring(sliceStart), json.encode(candidateRow))
		end
		for idx, indicator in ipairs(indicators) do
			-- The indicator value is in the column idx+1 of the candidate row, or the default if it does not exist.
			table.insert(processedData[indicator.id].data, { sliceStart:time() * 1000, candidateRow and candidateRow[idx+1] or indicator.default })
		end
		sliceStart = sliceStart:add(unpack(increment))
	end

	return true, processedData
end


function GetProcessFlows (vcc, flowsIds, keyspace)
	local flows = utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "ProcessFlows",
		superColumns = true,
		key = vcc,
		columns = flowId and {flowId} or {}
	}
	if flowsIds then
		local filteredFlows = {}
		for _,flowId in pairs(flowsIds) do
			filteredFlows[flowId] = flows[flowId] or {}
		end
		return filteredFlows
	else
		return flows or {}
	end
end

function GetAccountsByFlow (vcc, flowId, keyspace)
	flowId = utils.AppendVcc( flowId, vcc )
	return utils.GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "AccountsByProcessFlow",
		key = flowId
	}
end