/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:true, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global $, inConcert, _ */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.

if (!this.inConcert) {
	this.inConcert = {};
}


/**
 * @namespace
 * @name SupervisorConfig
 */
(function () {

	'use strict';

	const THROTTLE_WAIT = 1000; // millis

	/**
	 * Creates a new supervisor configuration instance.
	 *
	 * @author Sebastián Moreno (smoreno@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorConfig
	 * @param dashboard_instance
	 */
	function SupervisorConfiguration(dashboard_instance) {
		this.m_model = dashboard_instance.get_model_manager();
		this.m_current_configuration = null;
		this.m_dashboard = dashboard_instance;
	}

	if (typeof inConcert.sup_configuration !== 'function') {
		inConcert.sup_configuration = SupervisorConfiguration;
	}

	SupervisorConfiguration.prototype.get_configuration = function(callback)
	{
		var self = this;
		if (this.m_current_configuration){
			callback(this.m_current_configuration);
		}
		else {
			this.m_model.get_configuration(function(config, error) {
				if (!error){
					self.m_current_configuration = config;
					callback(config);
				}
				else{
					callback(null, error);
				}
			});
		}
	};

	/**
	 * Will update the cached config
	 *
	 * @public
	 * @memberOf SupervisorConfig
	 * @param  {object} config - The new configuration.
	 */
	SupervisorConfiguration.prototype.update_configuration_cache = function(config){
		this.m_current_configuration = config;
	};

	/**
	 * Get the current panel config for a given campaign
	 *
	 * @public
	 * @param  {string} campaign_id - The campaign id.
	 * @param  {string} panel_id    - The panel id
	 * @return {object|null}
	 */
	SupervisorConfiguration.prototype.get_panel_config = function(campaign_id, panel_id){
		if (!this.m_current_configuration || !this.m_current_configuration[campaign_id]){
			return null;
		}
		return this.m_current_configuration[campaign_id].dashboard.panels[panel_id];
	};

	/** @public */
	SupervisorConfiguration.prototype.update_campaign_visibility_config = function(campaign_id, visible){
		if (this.m_current_configuration && this.m_current_configuration[campaign_id]){
			if(this.m_current_configuration[campaign_id].visible !== visible) {
				this.m_current_configuration[campaign_id].visible_timestamp = Date.now() / 1000;
			}
			this.m_current_configuration[campaign_id].visible = visible;
		}
	};

	/**
	 * Build members panel configuration based on the visibility of elements.
	 * This function is idempotent, will not touch previous configuration
	 *
	 * @private 
	 * @param  {string} campaign_id     - The campaign identificator
	 * @param  {object} campaign_node   - A jQuery object for this campaign
	 * @param  {object} panel_node      - A jQuery object for the panel
	 * @param  {object} previous_config - Previous panel config (if any)
	 * 
	 * @return {object}                 The new panel configuration
	 */
	SupervisorConfiguration.prototype.__build_members_config = function(campaign_id, campaign_node, panel_node, previous_config) {
		// Clonamos la configuracion previa, para que no cambie tras la copia por referencia
		var prev_config = Object.assign({}, previous_config);
		var members_config = previous_config || {};
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);

		// set panel visibility
		members_config.visible = panel_node.is(':visible');

		members_config.options = {
			show_offline   : $('#sup_dash_show_offline_members_' + campaign_dom_id).prop('checked'),
			show_kickouts  : $('#sup_dash_show_kickouts_' + campaign_dom_id).prop('checked'),
			show_avatars   :  $('#sup_dash_show_avatars_members_option_' + campaign_dom_id).prop('checked'),
			show_names     : $('#sup_dash_show_names_members_' + campaign_dom_id).prop('checked'),
			show_state_time: $('#sup_dash_show_state_time_members_option_' + campaign_dom_id).prop('checked'),
			show_state_name: $('#sup_dash_show_state_name_members_option_' + campaign_dom_id).prop('checked'),
			show_state_visual_indicator: $('#sup_dash_show_state_visual_indicator_members_option_' + campaign_dom_id).prop('checked'),
			show_all:       $('#sup_dash_show_all_' + campaign_dom_id).prop('checked'),
			show_agents_id:       $('#sup_dash_show_id_members_' + campaign_dom_id).prop('checked'),
			show_workgroups:       $('#sup_dash_show_workgroups_option_' + campaign_dom_id).prop('checked'),
			show_total_conversations:       $('#sup_dash_show_total_conversations_option_' + campaign_dom_id).prop('checked'),
			show_total_conversations_by_channel:       $('#sup_dash_show_total_conversations_by_channel_option_' + campaign_dom_id).prop('checked')
		};

		// Si existen opciones de configuración previa
		if(prev_config.options)
		{
			// Actualizamos los timestamps asociados a las opciones que afectan a las alarmas por kickout
			if (members_config.options.show_kickouts !== prev_config.options.show_kickouts) {
				members_config.options.show_kickouts_timestamp = Date.now() / 1000;
			}
			else {
				members_config.options.show_kickouts_timestamp = prev_config.options.show_kickouts_timestamp || 0;
			}
			if(members_config.options.show_offline !== prev_config.options.show_offline) {
				members_config.options.show_offline_timestamp = Date.now() / 1000;
			}
			else {
				members_config.options.show_offline_timestamp = prev_config.options.show_offline_timestamp || 0;
			}
		}
		else {
			//Inicializamos los timestamps asociados a las opciones que afectan a las alarmas por kickout
			members_config.options.show_kickouts_timestamp = 0;
			members_config.options.show_offline_timestamp = 0;
		}
		return members_config;
	};

	/**
	 * Build indicators (charts) panel configuration based on the visibility of elements.
	 * This function is idempotent, will not touch previous configuration
	 *
	 * @private 
	 * @param  {string} campaign_id     - The campaign identificator
	 * @param  {object} $campaign        - A jQuery object for this campaign
	 * @param  {object} $panel          - A jQuery object for the panel
	 * @param  {object} previous_config - Previous panel config (if any)
	 * 
	 * @return {object}                   The new panel configuration
	 */
	SupervisorConfiguration.prototype.__build_indicators_config = function(campaign_id, $campaign, $panel, previous_config)
	{
		let indicators_config = previous_config || {};

		// set panel visibility
		indicators_config.visible = $panel.is(':visible');

		// visible charts
		indicators_config.charts = indicators_config.charts || {};

		// go and save the config for each chart
		let chart_id;
		let $chks = $campaign.find('.chartOptionCheck');
		$chks.each(function(_, chart_opt) {
			const $chart_opt = $(chart_opt);
			chart_id = $chart_opt.parent().attr('id');
			indicators_config.charts[chart_id] = $chart_opt.is(':checked');
		});

		// slice / timeago configuration
		const campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		indicators_config.slice = $('#sup_dash_indicators_slice_select_' + campaign_dom_id).val();
		indicators_config.time_ago = $('#sup_dash_indicators_time_ago_select_' + campaign_dom_id).val();

		// override panel visible attr, panel should be hidden if all charts are unchecked.
		const at_least_one_checked = $campaign.find('.chartOptionCheck:checked').length !== 0;
		if (!at_least_one_checked){
			indicators_config.visible = false;
		}

		return indicators_config;
	};

	/**
	 * Build summary panel configuration based on the visibility of elements.
	 * This function is idempotent, will not touch previous configuration
	 *
	 * @private 
	 * @param  {string} campaign_id     - The campaign identificator
	 * @param  {object} campaign_node   - A jQuery object for this campaign
	 * @param  {object} panel_node      - A jQuery object for the panel
	 * @param  {object} previous_config - Previous panel config (if any)
	 * 
	 * @return {object}                 The new panel configuration
	 */
	SupervisorConfiguration.prototype.__build_summary_config = function(campaign_id, campaign_node, panel_node, previous_config) {
		let summary_config = previous_config || {};
		// set panel visibility
		summary_config.visible = panel_node.is(':visible');
		// Create groups object
		summary_config.groups = summary_config.groups || {};
		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		const $campaign_summary_time_ago = $('#sup_dash_summary_time_ago_select_' + campaign_dom_id);
		summary_config.realtime_window = $campaign_summary_time_ago.children(':selected').val();
		// visible groups
		campaign_node.find('.summaryGroupCheck').each(function() {
			// Save group configuration
			let group_checkbox = $(this);
			let group_container = group_checkbox.parent().parent(); //container is two hops up,
			let group_id = group_container.attr('id');
			summary_config.groups[group_id] = {
				visible: false,
				// TO-DO: Re enable this when you fix those stupids bugs, you moron.
				// collapsed : !group_container.find('.summaryFieldsetRow').is(':visible'),
				indicators : {}
			};
			// visible indicators
			let atLeastOneChecked = false;
			group_container.find('.summaryRowCheck').each(function(){
				let indicator_checkbox = $(this);
				let indicator_node = indicator_checkbox.parent().parent(); // indicator container is the parent of parent of the option.
				let indicator_id = indicator_node.attr('id');
				let isChecked = indicator_checkbox.is(':checked');
				if (isChecked) {
					atLeastOneChecked = true;
				}
				summary_config.groups[group_id].indicators[indicator_id] = isChecked;
			});
			summary_config.groups[group_id].visible = atLeastOneChecked;
		});
		return summary_config;
	};

	/**
	 * Will build the configuration for every campaign panel.
	 *
	 * @private
	 * @return {Object} The supervisor config
	 */
	SupervisorConfiguration.prototype.__build_configuration = function()
	{
		const self = this;

		let full_screen_selected_tab_id;
		let full_screen = this.m_dashboard.is_full_screen_mode();
		if (full_screen) {
			full_screen_selected_tab_id = this.m_dashboard.full_screen_selected_tab_id();
		}
		let campaigns_config = !_.isEmpty(this.m_current_configuration) ? this.m_current_configuration : {};
		let $campapings_panels = $('#supervisor_main .supervisorPanelContainer');

		// 1 - save each campaign node of the dashboard.
		$campapings_panels.each(function()
		{
			var $container = $(this);
			var campaign_id = $container.attr('campaign_id');
			campaigns_config[campaign_id] = campaigns_config[campaign_id] || {};

			// Start saving configuration for this campaign
			campaigns_config[campaign_id].visible = campaigns_config[campaign_id].visible !== undefined ? campaigns_config[campaign_id].visible : $container.is(':visible');
			// si la app se encuentra en modo full screen
			campaigns_config[campaign_id].full_screen = full_screen;
			campaigns_config[campaign_id].full_screen_selected_tab_id = full_screen_selected_tab_id;

			var current_view = self.m_dashboard.get_current_view_id(campaign_id);

			campaigns_config[campaign_id].dashboard = campaigns_config[campaign_id].dashboard || {};
			campaigns_config[campaign_id].dashboard.current_view = current_view;
			campaigns_config[campaign_id].dashboard.panels = campaigns_config[campaign_id].dashboard.panels || {}; 

			// Panels config
			$container.find('[viewer_type="operations"] .panelBodyColumnAux').each(function(i, panel) {
				var panel_node = $(panel);
				var panel_id = panel_node.attr('panel_id');
				var panel_config = campaigns_config[campaign_id].dashboard.panels[panel_id] || {};

				var panel_conf;

				// Save config acording to current the panel id (Summary/Indicators/Members)
				switch(panel_id){
					case 'sup_dash_summary_panel':
						panel_conf = self.__build_summary_config(campaign_id, $container, panel_node, panel_config);
						break;

					case 'sup_dash_indicators_panel':
						panel_conf = self.__build_indicators_config(campaign_id, $container, panel_node, panel_config);
						break;

					case 'sup_dash_members_panel':
						panel_conf = self.__build_members_config(campaign_id, $container, panel_node, panel_config);
						break;
					case 'default': break;
				}

				campaigns_config[campaign_id].dashboard.panels[panel_id] = panel_conf;
			});
		});

		// 2 - configuracion seleccionada en el selector campaign / account / AL y en la configuracion de las alarmas por campana
		$.each(campaigns_config, function( selector_campaign_id, campaign_config )
		{
			// 2.1 configuracion seleccionada en el selector campaign / account / AL
			campaign_config.dashboard.config = self.m_dashboard.get_selected_campaign_account_al(selector_campaign_id);
			// 2.2 configuracion seleccionada en el selector campaign / account / AL
			campaign_config.alarms = self.m_dashboard.get_alarms_config(selector_campaign_id);
		});

		// save the member
		this.m_current_configuration = campaigns_config;

		// all built, return config
		return campaigns_config;
	};

	/**
	 * Will save the current configuration for this supervisor,
	 * this means all panels, grafics, visible indicators, team member options,
	 * and current dashboard view or interactions view.
	 *
	 * Note: It's doing the save in a throttling process, to avoid 
	 * multiple saves at the same time.
	 *
	 * For a config example see:
	 * https://gist.github.com/kernelp4nic/4c04e67e6ce5dbbe1a97
	 * 
	 * @public
	 * @param  {Function} callback
	 */
	SupervisorConfiguration.prototype.save_current_configuration = _.throttle(function(callback)
	{
		var self = this;
		callback = callback || $.noop;

		var save;

		save = function() {
			// Para poder realizar el Save, la configuracion de los paneles del dashboard tienen que estar ready
			//  Por ej: en la configuracion de alarmas se deben haber completado sus request, etc.
			if (!self.m_dashboard.is_ready_config()) {
				window.setTimeout(function() {
					save();
				}, 500);
				return;
			}
			var config = self.__build_configuration();
			self.m_model.save_configuration(inConcert.app.getLoggedUserId(), config, function()
			{
				self.m_dashboard.set_current_config(config);
				callback();
			});
		};

		save();

	}, THROTTLE_WAIT);
}());
