/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 9 */

/* global inConcert, _ */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}


/**
 * @namespace
 * @name SupervisorDashboardAppointment
 */
(function () {

	'use strict';

	const _T = inConcert.i18n.getString;
	var format = inConcert.Utils.format;

	/**
	 * Creates a new appointment dashboard instance.
	 *
	 * @author Sebastián Moreno (smoreno@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorDashboardAppointment
	 * @param dashboard_instance
	 */
	function DashboardAppointment(dashboard_instance) {

		DashboardAppointment.uber.constructor.call(this, 'appointment', "<%=_T('Appointment')%>", dashboard_instance);

		this.permission_id = 'sup_view_appointment';
		this.navigation_folder_id = 'appointment';
		this.m_view_classes_name = 'appointmentViewer';
		this.m_model = dashboard_instance.get_model_manager();
		this.m_dashboard = dashboard_instance ? dashboard_instance : null;

		// Search
		this.m_page_by_campaign = {};
		this.m_request_search_appointment = null;

		// current config
		this.m_config = dashboard_instance.get_config_manager();

		// finite state machine
		// possible states by campaign: idle, refreshing, processing
		this.__state = {};
	}

	inConcert.Utils.inherit(DashboardAppointment, inConcert.sup_dashboard_interactions_view);

	if (typeof inConcert.sup_dashboard_appointment !== 'function') {
		inConcert.sup_dashboard_appointment = DashboardAppointment;
	}

	DashboardAppointment.prototype.__get_interaction_gui_config = function()
	{
		var self = this;

		var config = DashboardAppointment.uber.__get_interaction_gui_config.call(this);
		config.columns = $.extend(true, config.columns || {}, {
			user_owner : { show: true }
		});

		config.get_specifics_interaction_classes = function( interaction, $interaction ) {
			return self.get_specifics_interaction_classes( interaction, $interaction );
		};

		config.callbacks = {
			can_open_client_window: function( interaction )
			{
				if ( interaction.belongsToMe() ) {
					return false;
				}
				return true;
			}
		};

		return config;
	};

	DashboardAppointment.prototype.__must_show_alarm = function (campaign_id, interaction)
	{
		var category = "conversation_time_on_state";
		return this.m_dashboard.eval_notifications_alarms(campaign_id, category, interaction);
	};

	DashboardAppointment.prototype.get_specifics_interaction_classes = function( interaction, $interaction )
	{
		if ( typeof interaction !== 'object' ) {
			return '';
		}

		var classes = this.__get_alarm_classes(interaction, $interaction);
		return classes;
	};

	DashboardAppointment.prototype.on_change_config_selector = function( campaign_id )
	{
		DashboardAppointment.uber.on_change_config_selector.call(this, campaign_id);
	};

	DashboardAppointment.prototype.__get_prefix = function( /*interaction*/ )
	{
		return 'supervisor_int_appointment_';
	};

	DashboardAppointment.prototype.__get_dom_interaction_params = function( interaction )
	{
		return {
			prefix: this.__get_prefix( interaction ),
			allowed_bulk_actions: false // TODO: agregar permiso ?
		};
	};

	DashboardAppointment.prototype.process_interactions = function(campaign_id, interactions)
	{
		var self = this;

		if (this.view_has_changed(campaign_id)){
			return;
		}
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var interactions_container = this.get_interactions_container(campaign_dom_id);
		var scroll_container = interactions_container.parent();
		var scroll_top = scroll_container.scrollTop();
		var detached_container = interactions_container.detach();

		// build markup for every conversation.
		interactions.forEach(function(interaction)
		{
			// create and append node.
			var parmams_interaction = self.__get_dom_interaction_params(interaction);
			var $interaction_node = $( self.m_interaction_gui.$build_interaction_node( interaction, self.m_templates.interaction_node_tpl, parmams_interaction ) );
			self.insert_interaction_in_panel_list( interaction, $interaction_node, detached_container );
			// update interaction on panel
			self.m_interaction_gui.update_interaction_data( interaction, $interaction_node );
		});

		scroll_container.append(detached_container);
		scroll_container.scrollTop(scroll_top);
	};

	DashboardAppointment.prototype.search_appointment = function ($current_node, init_in_page, callback = $.noop)
	{
		const self = this;

		console.info("[DashboardAppointment.search_appointment] START ");

		if (this.m_request_search_appointment) {
			console.info("[DashboardAppointment.search_appointment] --> Unable to start search appointments. Already exist another request: " + this.m_request_search_appointment);
			callback(false);
			return;
		}

		let campaign_id = $current_node.attr('campaign_id');
		let search_query = this.m_dashboard.getQueryBuilderFilter(campaign_id);
		if (!search_query) {
			return;
		}
		// Dejo disabled la navegacion, hasta que culmine de procesar el request
		let navigation = this.get_navigation(campaign_id);
		if (navigation) {
			navigation.disable_navigation(true);
		}
		const onSearchFinished = function(filtered_interactions) {
			callback(filtered_interactions);
			if (navigation) {
				// Dejo enabled la navegacion, ya que ha terminado de procesar el Request.
				navigation.disable_navigation(false);
			}
			self.m_request_search_appointment = null;
		};

		// configuracion seleccionada de la campana en el tree selector
		//var campaign_config = this.m_dashboard.get_campaign_selected_config(campaign_id);
		//var account_id = campaign_config.account_id;
		//var attention_level_id = campaign_config.al_id;

		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var $dynamic = $('#sup_dash_interaction_viewer_dynamic_' + campaign_dom_id);

		//Pills
		var $pillContainer = $dynamic.find(".form-group-pill-card-outlined");
		var $noPill = $pillContainer.find(".inlineMessagesContainer");
		var $pillGroup = $pillContainer.find(".form-group-pill-container");
		$pillGroup.empty();
		if(search_query && search_query.rules && $.isArray(search_query.rules) && search_query.rules.length > 0)
		{
			var addRemoveButtonClick;
			addRemoveButtonClick = function(ruleId) {
				$("#" + ruleId + "_remove").click(function() {
					var ruleIdx = parseInt(this.id.replace("sup_appointment" + campaign_dom_id + "_pill_", ""));
					//Eliminar del queryBuilder
					if(self.m_dashboard.queryBuilderObj[campaign_dom_id].deleteRule(ruleIdx))
					{
						//Eliminar del contenedor de pills
						$(this).parent().remove();
					}
				});
			};

			for (var i = 0; i < search_query.rules.length; i++) {
				var ruleId = "sup_appointment" + campaign_dom_id + "_pill_" + i;
				var rulePillText = _T(search_query.rules[i].label || search_query.rules[i].id) + " | " + _T(search_query.rules[i].operator) + " | " + "<strong>" + _T(search_query.rules[i].value) + "</strong>";
				const $lbl = "" +
					"<label id='" + ruleId + "_label' class='form-group-pill'>" +
						"<span>" + rulePillText + "</span>" +
						"<button id='" + ruleId + "_remove' class='pillButtonDelete'></button>" +
					"</label>";
				$pillGroup.append($lbl);
				$pillGroup.show();
				$noPill.hide();

				addRemoveButtonClick(ruleId);
			}
		}
		else {
			$pillGroup.hide();
			$noPill.show();
		}

		//labels
		let selected_labels = navigation.get_selected_labels();
		if (selected_labels && $.isArray(selected_labels) && selected_labels.length > 0) {
			var labelsRule = {
				id : "labels",
				operator : "in",
				value : selected_labels
			};
			search_query.rules.push(labelsRule);
		}
		
		if(!search_query.rules.find(rule => rule.id === "campaign")) {
			//campaign
			var campaignRule = {
				id : "campaign",
				operator : "equal",
				value : campaign_id
			};
			search_query.rules.push(campaignRule);
		}
		search_query.rules = this.m_dashboard.getCommonConfigRules(search_query.rules, campaign_id);

		var hasCustomFilter = function() {
			return (search_query.rules.find(rule => rule.id === "agent") || search_query.rules.find(rule => rule.id === "labels") || search_query.rules.find(rule => rule.id === "type") !== undefined);
		};

		// if custom filter is set, get all pages.
		var default_page = hasCustomFilter ? 0 : 1;
		var current_page = init_in_page || ( this.m_page_by_campaign[campaign_id] ? this.m_page_by_campaign[campaign_id].filter.page : default_page );

		var filter = {
			//campaign_id        : campaign_id,
			//account_id         : campaign_config.account_id,
			//attention_level_id : campaign_config.al_id,
			page               : current_page,
			rows_per_page      : this.INTERACTIONS_PER_PAGE,
			query              : search_query,
			sort               : [{state_timestamp: { order: "asc" }, priority: {order: "desc"}}]
		};

		this.m_page_by_campaign[campaign_id] = { paginating : false, filter : filter };

		// Los appointment se ordenan por fecha de creacion.
		filter.sort = [{start_timestamp: { order: "asc" }, priority: {order: "desc"}}];
		
		this.m_request_search_appointment = this.m_model.search_open_interactions(filter, function(status, result)
		{
			if (status) {
				$dynamic.find("[query_builder_clean]").prop("disabled", !self.m_dashboard.queryBuilderObj[campaign_dom_id].hasRules());
				// Update pagination info.
				self.m_page_by_campaign[campaign_id].totalRows = result.total || 0;
				self.m_page_by_campaign[campaign_id].last_page = Math.ceil(result.total / self.INTERACTIONS_PER_PAGE);
				self.update_current_pagination(campaign_id, result.interactions.length || 0);
				onSearchFinished(result.interactions);
			}				
			else{
				self.m_request_search_appointment = null;
				inConcert.app.getMainWindow().showBannerMessage({ message : format("<%=_T('Unable to get campaign conversations for appointment view on campaign: {0}.')%>", campaign_id), timeout : 5000});
			}
		});
	};

	DashboardAppointment.prototype.get_view_paged_info = function()
	{
		return this.m_page_by_campaign;
	};

	DashboardAppointment.prototype.execute_new_search = function(campaign_id, callback, cleanup_search_text)
	{
		callback = callback || $.noop;
		var self = this;

		if (cleanup_search_text) {
			this.cleanup_search_text(campaign_id);
		}

		this.__state[campaign_id] = 'processing';
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		this.show_waiting(campaign_dom_id, "<%=_T('Searching conversations')%>");
		var $current_node = this.m_dashboard.get_campaign_node(campaign_id);

		this.search_appointment(
			$current_node,
			undefined,
			function(interactions){
				if (interactions) {
					self.clear_list(campaign_dom_id);
					if (interactions.length === 0){
						self.toggle_no_conversations_msg(campaign_dom_id, "<%=_T('No appointment conversations to show. Try another search.')%>");
					}
					else{
						self.process_interactions(campaign_id, interactions);
					}	
				}
				self.hide_waiting(campaign_dom_id);
				self.__state[campaign_id] = 'idle';
				callback();
			});
	};

	DashboardAppointment.prototype.init = function($current_node, member_id)
	{
		DashboardAppointment.uber.init.call(this, $current_node, { has_batch_actions : false });

		var self = this;
		var campaign_id = $current_node.attr('campaign_id');
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);

		//var $search = $current_node.find('[id="sup_dash_interaction_viewer_dynamic_' + campaign_dom_id + '"]');
		//var $search_button = $search.find('[filter_search_execute]');

		// Set agent if it was specified.
		var init_on_page;
		if (member_id){
			init_on_page = 1;
		}

		// setup custom appointment pagination.
		this.setup_view_pagination($current_node, campaign_id);
		// initial search
		this.__state[campaign_id] = 'processing';
		this.show_waiting(campaign_dom_id, "<%=_T('Searching conversations')%>"); // Show waiting

		this.search_appointment(
			$current_node,
			init_on_page,
			function(interactions) {
				if (interactions) {
					self.clear_list(campaign_dom_id); // Clear conversations list and setup menu actions.
					if ( interactions.length === 0 ) {
						self.toggle_no_conversations_msg(campaign_dom_id, "<%=_T('No appointment conversations to show.')%>");
					}
					else {
						self.process_interactions(campaign_id, interactions);
					}
				}
				self.hide_waiting(campaign_dom_id);
				self.on_view_load_finished(campaign_id);
				self.__state[campaign_id] = 'idle';
			});
	};

	DashboardAppointment.prototype.update_current_pagination = function(campaignId, pageInteractions)
	{
		let pageInfo = this.m_page_by_campaign[campaignId];
		if (!pageInfo) {
			return;
		}
		let currentPage = pageInfo.filter.page;
		let lastPage = pageInfo.last_page;
		let totalRows = pageInfo.totalRows;

		DashboardAppointment.uber.update_current_pagination.call(this, campaignId, currentPage, lastPage, pageInteractions, totalRows);
	};

	DashboardAppointment.prototype.refresh = function ($current_node)
	{
		const self = this;

		let campaign_id = $current_node.attr('campaign_id');
		if (this.view_has_changed(campaign_id)){
			return;
		}
		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		// go and search for conversations
		this.search_appointment(
			$current_node,
			undefined, 
			function(interactions)
			{
				if (!self.allowed_to_refresh_view(campaign_id)){
					self.hide_waiting(campaign_dom_id);
					return;
				}
				// if no conversations, show message to the user and return.
				if (interactions) {
					if ( interactions.length === 0 ){
						self.clear_list(campaign_dom_id);
						self.toggle_no_conversations_msg(campaign_dom_id, "<%=_T('No appointment conversations to show.')%>");
					}
					else {
						// we got conversations, remove no conversations message (if exists.)
						self.remove_no_conversations_msg(campaign_dom_id);
					}
					self.process_off_dom_interaction_update(
						campaign_dom_id,
						interactions,
						self.m_templates.interaction_node_tpl,
						function( interaction, $interaction_node /*, previous_interaction_data*/ ) {
							// update interaction on panel
							self.m_interaction_gui.update_interaction_data( interaction, $interaction_node );
						}
					);
				}
				self.hide_waiting(campaign_dom_id);
			}
		);
	};
}());
