/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:true, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/*global $, inConcert, Mustache, _, inconcert */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name DashboardIndicators
 * @author Sebastian Moreno (smoreno@inconcertcc.com)
 */
(function () {

	'use strict';
	const FORMAT = inConcert.Utils.format;
	const nid = inConcert.Utils.getNormalizedId;
	const capitalize = inConcert.Utils.capitaliseFirstLetter;
	const _T = inConcert.i18n.getString;

	/**
	 * Creates a new indicators dashboard instance.
	 *
	 * @constructor
	 * @memberOf DashboardIndicators
	* @param dashboard_instance
	 */
	function DashboardIndicators(dashboard_instance)
	{
		const self = this;

		this.m_id = 'indicators';
		this.m_templates = inConcert.sup_templates.dashboard;
		this.m_model = dashboard_instance.get_model_manager();
		this.m_dashboard = dashboard_instance ? dashboard_instance : null;

		// current config.
		this.m_config = dashboard_instance.get_config_manager(); // current config

		// current charts by id
		this.m_charts_by_id = {};

		// map to know if i'm editing
		this.m_in_edit_mode_by_campaign = {};
		this.m_hnd_building_chart = null;

		// message bus
		inConcert.SupervisorMessageBus.subscribe('/supervisor/dashboard/' + this.m_id, function(notification) {
			self.__on_panel_visibility_changed(notification.campaign_id, notification.visible);
		});

		this.m_chart_serie_colors = {
			'MAIL' :       '#FF18F6',
			'CHAT' :       '#0FBBBB',
			'WEBCONTACT' : '#C56DFF',
			'CALL' :       '#9900FF',
			'INBOUND_CALL': '#25933D', // defino ambos colores para los nuevos canales
			'OUTBOUND_CALL': '#16009B',
			'IM':          '#F7CF97',
			'SMS':         '#00B1FD',
			'WHATSAPP':    '#00CC66',
			'FACEBOOK':    '#4267B2',
			'INSTAGRAM':   '#FD03B2',
			'YOUTUBE':     '#FF0000',
			'APPSTORE':    '#2072F3',
			'GOOGLEPLAY':  '#3C4A5F',
			'RCS':	       '#FCC400',
			'TWITTER':     '#1DA1F2',
			'VOICEMAIL' :  '#FFA449',
			'TOTAL':       '#333333'
		};
	}

	if (typeof inConcert.sup_dashboard_indicators !== 'function') {
		inConcert.sup_dashboard_indicators = DashboardIndicators;
	}

	DashboardIndicators.prototype.on_change_config_selector = function( campaign_id )
	{
		this.refresh(campaign_id);
	};

	/**
	 * Get the indicators panel jQuery node
	 * 
	 * @private
	 * @param  {string} campaign_id - The campaign id (i.e. campaign@vcc)
	 * @return {object}             - The jQuery panel node
	 */
	DashboardIndicators.prototype.__get_panel_node = function(campaign_id){
		return $('#sup_dash_indicators_panel_' + nid(campaign_id));
	};

	DashboardIndicators.prototype.__$get_legends_container = function(campaign_id){
		return $('#sup_dash_indicators_legend_' + nid(campaign_id));
	};

	DashboardIndicators.prototype.__get_serie_selector_container = function(campaign_id){
		return $('#sup_dash_indicators_series_' + nid(campaign_id));
	};

	DashboardIndicators.prototype.__$get_legend_item = function(campaign_id, legend_name){
		return $('#sup_dash_indicators_legend_item_'+nid(campaign_id)+'_'+legend_name);
	};

	DashboardIndicators.prototype.__legend_already_exists = function(campaign_id, legend_name){
		return this.__$get_legend_item(campaign_id, legend_name).length > 0;
	};

	DashboardIndicators.prototype.__get_serie_selector_item = function(campaign_id, serie_id){
		return $('#sup_dash_indicators_serie_option_' + nid(campaign_id) + '_' + serie_id);
	};

	DashboardIndicators.prototype.__serie_already_exists = function(campaign_id, serie_id){
		return this.__get_serie_selector_item(nid(campaign_id), serie_id).length > 0;
	};

	DashboardIndicators.prototype.__get_panel_container = function(campaign_id, panel_node){
		return panel_node ? panel_node.find('.columnContent') : this.__get_panel_node(campaign_id).find('.columnContent');
	};

	DashboardIndicators.prototype.__get_trend_container = function(campaign_id){
		return this.__get_panel_container(campaign_id).find('.trendContainer');
	};

	DashboardIndicators.prototype.__on_panel_visibility_changed = function(campaign_id, visible) {
		if ( visible ){
			var charts = this.__get_trend_container(campaign_id);
			var chart_checkboxes = charts.find(':checkbox');

			var total_charts = charts.length;
			var total_charts_hidden = chart_checkboxes.not(':checked').length;

			if (total_charts === total_charts_hidden){
				this.__enter_edit_mode(campaign_id);
			}
		}
	};

	/**
	 * Called when the user leaves edit mode
	 *
	 * NOTE: When the user enters edit mode, 
	 * all charts are automatically shown by css.
	 * 
	 * @param  {String} campaign_id The campaign identificator (campaign@vcc)
	 */
	DashboardIndicators.prototype.__leave_edit_mode = function(campaign_id)
	{
		const self = this;
		var panel = this.__get_panel_node(campaign_id);
		var container = this.__get_panel_container(campaign_id, panel);
		var campaign_dom_id = nid(campaign_id);

		var chart_checkboxes = this.__get_trend_container(campaign_id).find(':checkbox');

		// if user uncheck some indicator, hide it.
		chart_checkboxes.not(':checked').parent().hide();

		// hide all checkboxes
		chart_checkboxes.hide();
		this.__redraw_chart_messages(campaign_id);

		// hide serie selector
		this.__get_serie_selector_container(campaign_id).hide();

		// if all groups are unchecked, hide column.
		if (!chart_checkboxes.is(':checked')){
			// TO-DO: This could be done with pubsub
			this.m_dashboard.trigger_column_update(campaign_id, 'indicators', false);
		}

		// update legends visualization
		var legends_container = this.__$get_legends_container(campaign_id);
		if (legends_container.children().length > 0){
			panel.addClass('showingLegends');
			this.__$get_legends_container(campaign_id).show();
		}

		// update edit/done button
		$('#sup_dash_column_edit_indicators_' + campaign_dom_id)
			.addClass('butColumnEdit')
			.removeClass('butColumnDone');

		// save configuration.
		this.m_dashboard.save_config(function() {
			self.__hide_slice_configuration(campaign_dom_id);

			container.removeClass('editMode');
			self.m_in_edit_mode_by_campaign[campaign_id] = false;

			//fetch data and redraw charts
			self.setup(campaign_id);
		});
	};

	/**
	 * Called when the user enters in edit mode
	 *
	 * NOTE: When the user enters edit mode, 
	 * all charts are automatically hidden by css.
	 * 
	 * @param  {String} campaign_id The campaign identificator (campaign@vcc)
	 */
	DashboardIndicators.prototype.__enter_edit_mode = function(campaign_id){
		var campaign_dom_id = nid(campaign_id);
		var panel = this.__get_panel_node(campaign_id);
		var container = this.__get_panel_container(campaign_id, panel);

		// charts options and slice configuration
		container.children().show();
		container.find('.chartOptionCheck').show();
		this.__show_slice_configuration(campaign_dom_id);

		// update legends visualization
		panel.removeClass('showingLegends');
		this.__$get_legends_container(campaign_id).hide();

		// show serie selector
		this.__get_serie_selector_container(campaign_id).show();

		// update edit/done button
		$('#sup_dash_column_edit_indicators_' + campaign_dom_id ).removeClass('butColumnEdit').addClass('butColumnDone');

		// add custom class to the container
		container.addClass('editMode');
		this.m_in_edit_mode_by_campaign[campaign_id] = true;
	};

	DashboardIndicators.prototype.__get_indicator_chart_dom_id = function(campaign_id, indicator_id){
		return '#' + nid(campaign_id) + '_' + indicator_id;
	};

	// if i am in edit mode, do not refresh the view.
	DashboardIndicators.prototype.allowed_to_refresh_view = function (campaign_id) {
		var current_view = this.m_dashboard.get_current_view_id(campaign_id);
		if ( current_view !== 'dashboard' ) {
			return false;
		}
		return this.m_in_edit_mode_by_campaign.hasOwnProperty(campaign_id) ? !this.m_in_edit_mode_by_campaign[campaign_id] : true;
	};

	DashboardIndicators.prototype.__get_name_from_indicator_id = function(indicator_id){
		return indicator_id.replace(/^sup_dash_indicators_(.*?)_(.*?)_(.*?)$/, '$3');
	};

	/**
	 * Return true if chart needs to be visible
	 * based on user configuration
	 *
	 * @private
	 * @param  {Object} config    - Panel configuration
	 * @param  {String} chart_id  - Chart identificator to check.
	 * @return {Boolean}
	 */
	DashboardIndicators.prototype.__chart_should_be_visible = function (config, chart_id){
		if (config && config.charts && config.charts.hasOwnProperty(chart_id)){
			// if the chart is collapsed, all indicators should be hidden.
			if ( config.charts[chart_id].hasOwnProperty('collapsed') ){
				return !config.charts[chart_id].collapsed;
			}
			// otherwise, return the chart indicator visibility
			return config.charts[chart_id];
		}
		// visible by default
		return true;
	};

	DashboardIndicators.prototype.__redraw_chart_messages = function(campaign_id){
		var self = this;
		if (self.m_charts_by_id[campaign_id]){
			for (var chart_id in self.m_charts_by_id[campaign_id]){
				if (self.m_charts_by_id[campaign_id].hasOwnProperty(chart_id)){
					var no_data_msg = Mustache.render(self.m_templates.indicator_no_data_tpl, {message: "<%=_T('Not enough data to display chart')%>"});
					$(self.__get_indicator_chart_dom_id(campaign_id, chart_id)).find('.formIndicatorPanelContent').prepend(no_data_msg);
				}
			}
		}
	};

	DashboardIndicators.prototype.__redraw_charts = function(campaign_id){
		var self = this;
		var charts = self.__get_trend_container(campaign_id);
		// redraw charts
		charts.each(function(i, e){
			if (self.m_charts_by_id[campaign_id] && self.m_charts_by_id[campaign_id][self.__get_name_from_indicator_id(e.id)]){
				self.m_charts_by_id[campaign_id][self.__get_name_from_indicator_id(e.id)].chart.Redraw();
			}
		});
	};

	/**
	 * Will setup the handlers to toggle each chart
	 *
	 * @private
	 * @param {object} current_node -  Current campaign node.
	 */
	DashboardIndicators.prototype.setup_title_collapse_handlers = function(current_node){
		current_node.find('.formIndicatorPanelTitle')
			.unbind('click')
			.bind('click', function(){
				$(this).toggleClass('collapsed').next().toggle();
			});
	};

	DashboardIndicators.prototype.setup_edit_mode = function (campaign_id)
	{
		var self = this;

		var campaign_dom_id = nid(campaign_id);
		var container = this.__get_panel_container(campaign_id);

		// build and hide slice configuration.
		this.__build_slice_configuration(campaign_id);
		this.__hide_slice_configuration(campaign_dom_id);

		// edit button
		$('#sup_dash_column_edit_indicators_' + campaign_dom_id)
			.unbind('click')
			.bind('click', function()
			{
				// *done* behaviour -> if the user was already in edit mode.
				if (container.hasClass('editMode')){
					self.__leave_edit_mode(campaign_id);
				}
				else {
					self.__enter_edit_mode(campaign_id);
				}
			});
	};

	DashboardIndicators.prototype.__get_slice_combo = function(campaign_dom_id){
		return $('#sup_dash_indicators_slice_select_' + campaign_dom_id);
	};

	DashboardIndicators.prototype.__get_time_ago_combo = function(campaign_dom_id){
		return $('#sup_dash_indicators_time_ago_select_' + campaign_dom_id);
	};

	DashboardIndicators.prototype.__get_slice_config_container = function(campaign_dom_id){
		return $('#sup_dash_indicators_slice_' + campaign_dom_id);
	};

	DashboardIndicators.prototype.__build_slice_configuration = function(campaign_id) {
		var self = this;
		var campaign_dom_id = nid(campaign_id);
		var slice_configuration = self.__get_slice_config_container(campaign_dom_id);

		// if it was already built. return
		if ( slice_configuration.length !== 0 ) {
			return;
		}

		var container = self.__get_panel_container(campaign_id);

		slice_configuration = $(Mustache.render(self.m_templates.indicator_slice_configuration_tpl,
			{
				id         : campaign_dom_id ,
				two_points : "<%=_T(':')%>"         ,
				slice: {
					label: "<%=_T('Slice')%>",
					data: {
						'15_min': "<%=_T('15 minutes')%>",
						'30_min': "<%=_T('30 minutes')%>",
						'1_hour': "<%=_T('1 hour')%>",
						'1_day' : "<%=_T('1 day')%>"
					}
				},
				time_ago: {
					label: "<%=_T('Time Ago')%>",
					data: {
						'2_hours': "<%=_T('2 hours')%>",
						'4_hours': "<%=_T('4 hours')%>",
						'8_hours': "<%=_T('8 hours')%>",
						'1_week' : "<%=_T('1 week')%>"
					}
				}
			}
		));
		container.prepend(slice_configuration);

		//Load configuration and select slice and time_ago combos
		var current_config = self.m_config.get_panel_config(campaign_id, 'sup_dash_indicators_panel');
		var slice_combo = self.__get_slice_combo(campaign_dom_id);
		var time_ago_combo = self.__get_time_ago_combo(campaign_dom_id);

		if (current_config && current_config.slice && current_config.time_ago) {
			slice_combo.find('[value="' + current_config.slice + '"]').prop('selected', true);
			time_ago_combo.find('[value="' + current_config.time_ago + '"]').prop('selected', true);
		}
		else {
			slice_combo.find('[value="15Min"]').prop('selected', true);
			time_ago_combo.find('[value="2Hour"]').prop('selected', true);
		}
		
		//helper objects to correct inconsistencies from combo selector
		var sliceToTimeAgo = {
			'15Min': '2Hour',
			'30Min': '4Hour',
			'Hour': '8Hour',
			'Day': 'Week'
		};

		var timeAgoToSlice = {
			'Week': 'Day',
			'8Hour': 'Hour',
			'4Hour': '30Min',
			'2Hour': '15Min'
		};

		slice_combo.change(function(){
			var time_ago_combo = self.__get_time_ago_combo(campaign_dom_id);
			var selected_slice = $(this).val();
			time_ago_combo.find('[value="' + sliceToTimeAgo[selected_slice] + '"]').prop('selected', true);
		});
		
		time_ago_combo.change(function(){
			var slice_combo = self.__get_slice_combo(campaign_dom_id);
			var selected_interval = $(this).val();
			
			slice_combo.find('[value="' + timeAgoToSlice[selected_interval] + '"]').prop('selected', true);
		});
	};

	DashboardIndicators.prototype.__show_slice_configuration = function(campaign_dom_id) {
		this.__get_slice_config_container(campaign_dom_id).show();
	};

	DashboardIndicators.prototype.__hide_slice_configuration = function(campaign_dom_id) {
		this.__get_slice_config_container(campaign_dom_id).hide();
	};

	DashboardIndicators.prototype.__get_slice_minutes = function(slice) {
		var sliceMinutes = {
			'15Min'	:	15,
			'30Min'	:	30,
			'Hour'	:	60,
			'2Hour'	:	60 * 2,
			'4Hour'	:	60 * 4,
			'8Hour'	:	60 * 8,
			'Day'	:	60 * 24,
			'Week'	:	60 * 24 * 7
		};
		return sliceMinutes[slice];
	};

	DashboardIndicators.prototype.__chart_Y_formatter = function(n){
		if (Math.floor(n) !== n){
			return '';
		}
		//Scales the value and adds the corresponding unit prefix
		//so instead of displaying '1000000' this will display '1M'
		var suffix = ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y', '*10^27', '*10^30', '*10^33', '*10^36', '*10^39'];
		var sign = n < 0 ? '-' : '';

		n = Math.abs(n);
		var i = 0;

		while (n >= 1000) {
			n /= 1000;
			i++;
		}
		//round to 2 decimals
		n = Math.round(n * 100) / 100;
		return sign + n + suffix[i];
	};

	DashboardIndicators.prototype.__chart_X_need_ticks = function(min, max, campaign_id){
		if (!min || !inConcert.Utils.isNumber(min)) {
			return;
		}

		var self = this;
		var zeropad = inConcert.Utils.zeropad;

		//create ticks so they are slice aligned
		var current_config = self.m_config.get_panel_config(campaign_id, 'sup_dash_indicators_panel');
		if (!current_config) {
			current_config = {
				slice: 'Hour',
				time_ago: '8Hour'
			};
		}

		var slice = self.__get_slice_minutes(current_config.slice) * 60 * 1000;
		var ticks = [];
		var init = Math.floor(min);
		init = init - (init % slice);
		var tick = init;

		while (tick <= max) {
			if (current_config.slice === 'Day') {
				var date = new Date(tick);
				ticks.push([tick, zeropad(date.getUTCDate()) + '/' + zeropad(date.getUTCMonth() + 1)]);
			}
			else {
				//use default label (specified in xFormatter)
				ticks.push([tick, null]);
			}
			tick = tick + slice;
			if (ticks.length > 20) {
				//we don't want more than 20 labels
				//because it can slow things down
				return ticks;
			}
		}
		return ticks;
	};

	/**
	 * Adds a new legend item to the legends container only if not exist
	 *
	 * @private
	 * @param  {String} campaign_id The campaign identificator (campaign@vcc)
	 * @param  {String} channel     The channel id (MAIL)
	 * @param  {String} name        The channel name (Mail)
	 */
	DashboardIndicators.prototype.__add_legend = function(campaign_id, channel, name)
	{
		var $legends_container = this.__$get_legends_container(campaign_id);

		if ( !this.__legend_already_exists(campaign_id, channel) ) {
			$legends_container.append(Mustache.render(this.m_templates.indicator_chart_legend, {
				channel : channel,
				name : name,
				campaign_id: nid(campaign_id)
			}));
			this.__get_panel_node(campaign_id).addClass('showingLegends');
		}
	};

	DashboardIndicators.prototype.__remove_legend = function(campaign_id, leyend_id) {
		this.__$get_legend_item(campaign_id, leyend_id).remove();
	};

	DashboardIndicators.prototype.__update_channels_legend = function(campaign_id)
	{
		var self = this;

		var campaign = inConcert.app.getCampaign(campaign_id);
		if (!campaign) {
			return;
		}
		var campaign_config = this.m_dashboard.get_campaign_selected_config(campaign_id);

		$.each( this.m_chart_serie_colors, function( channel )
		{
			var $channel_legend = self.__$get_legend_item(campaign_id, channel);
			var channelPart = channel.replace("INBOUND_", "").replace("OUTBOUND_", "");
			if (campaign.has_accounts_of_proxy( channelPart ) ) {
				var can_show_legend = true;
				if (channel !== 'TOTAL') {
					var $chk_serie = self.__get_serie_selector_item(campaign_id, channel);
					if ( campaign_config.account_id && campaign_config.channel !== "CALL" ) {
						can_show_legend = false;
						$chk_serie.prop("disabled", true);
					}
					else {
						can_show_legend = $chk_serie.is(":checked");
						$chk_serie.prop("disabled", false);
					}
				}
				if ( can_show_legend ) {
					if ( $channel_legend.length === 0 ) {
						self.__add_legend(campaign_id, channel, channel);
					}
					else {
						$channel_legend.show();
					}
				}
				else {
					$channel_legend.hide();
				}
			}
			else {
				if ( channel !== 'TOTAL' ) {
					$channel_legend.hide();
				}
			}
		});
	};

	/**
	 * Adds a new serie item to the series container only if not exist
	 *
	 * @private
	 * @param  {String} campaign_id   The campaign identificator (campaign@vcc)
	 * @param  {String} serie_id      The serie id (MAIL)
	 * @param  {String} serie_name    The serie name (this will be showed to the user) (Mail)
	 */
	DashboardIndicators.prototype.__add_serie_selector = function(campaign_id, serie_id, serie_name){
		var self = this;
		// just allow one serie at the same time
		if ( this.__serie_already_exists(campaign_id, serie_id) ){
			return;
		}
		var $serie = $( Mustache.render(this.m_templates.chart_serie_selector_tpl, {
			label: serie_name, 
			serie: serie_id,
			campaign_id: nid(campaign_id)
		}));
		var $series_container = this.__get_serie_selector_container(campaign_id);
		$series_container.append( $serie );
		// si no tiene licencia para el canal, lo oculta
		var proxies = { CALL: true, CHAT: true, FACEBOOK: true, INSTAGRAM: true, YOUTUBE: true, APPSTORE: true, GOOGLEPLAY: true, RCS: true, IM: true, MAIL: true, SMS: true, WHATSAPP: true, TWITTER: true, WEBCONTACT: true, VOICEMAIL: true };
		var proxy_id = ( serie_id || '' ).toUpperCase().replace("INBOUND_", "").replace("OUTBOUND_", "");
		if ( !proxies[ proxy_id ] || inConcert.app.hasLicensesForChannel( proxy_id ) ) {
			$serie.show();
		}
		else {
			$serie.hide();
		}

		var $chk_serie = this.__get_serie_selector_item(campaign_id, serie_id);
		$chk_serie.change(function(){
			var serie_checkbox = $(this);
			self[ serie_checkbox.is(':checked') ? '__show_serie' : '__hide_serie' ](campaign_id, serie_checkbox.attr('serie_id'));
		});
	};

	DashboardIndicators.prototype.__show_serie = function(campaign_id, serie_id)
	{
		var self = this;

		_.each(self.m_charts_by_id[campaign_id], function(chart_object)
		{
			var series_names = chart_object.chart.Series.Secondary.GetSecondarySeriesNames();

			if (series_names.length === 0 || !_.contains(series_names, serie_id)) {
				// make sure this serie was previously created (and is not a main serie)
				if (chart_object.secondary_series[serie_id]){
					self.__add_legend(campaign_id, serie_id, capitalize(_T(serie_id)));
					chart_object.chart.Series.Secondary.AddSerie(serie_id, {
						source: chart_object.chart_data[serie_id].data
					},
					{
						// TO-DO: Get this from a function
						type : chart_object.chart_type,
						drawPoints : true,
						lineWidth : 2,
						color : self.m_chart_serie_colors[serie_id],
						hintString: '<div>['+chart_object.tooltip+']</div>'
					});
				}
			}
		});
	};

	DashboardIndicators.prototype.__hide_serie = function(campaign_id, serie_id)
	{
		var self = this;
		if ( !this.m_charts_by_id.hasOwnProperty(campaign_id) ){
			return;
		}
		_.each(this.m_charts_by_id[campaign_id], function(chart_object)
		{
			var series_names = chart_object.chart.Series.Secondary.GetSecondarySeriesNames();
			if (series_names.length > 0 && _.contains(series_names, serie_id)){
				chart_object.chart.Series.Secondary.RemoveSerie(serie_id);
				self.__remove_legend(campaign_id, serie_id);
			}
		});
	};

	DashboardIndicators.prototype.__get_default_chart_properties = function(data, type, tooltip, serie_id){
		return {
			DataSource: {
				source: data
			},
			Properties : {
				type : type,
				drawPoints : true,
				lineWidth : 2,
				color : serie_id ? this.m_chart_serie_colors[serie_id] : '#4B606A',
				hintString: '<div>['+tooltip+']</div>'
			}
		};
	};

	DashboardIndicators.prototype.__build_chart = function (chart_data, campaign_id, onAfterBuild)
	{
		var self = this;

		var main_serie = this.__get_default_chart_properties(chart_data.data, chart_data.chart_type, chart_data.tooltip);

		var secondary_series = {};

		// dealing with secondary series?
		if (chart_data.containsSecondarySeries){
			// show leyends container
			this.__$get_legends_container(campaign_id).show();

			_.each(chart_data.data, function(serie, serie_id){
				// create legend
				self.__add_legend(campaign_id, serie_id, capitalize(_T(serie_id)));

				// build basic chart data
				var serie_data = self.__get_default_chart_properties(serie.data, chart_data.chart_type, chart_data.tooltip, serie_id);

				// we always need to specify a main serie for this chart
				// so we make 'TOTAL' indicator our main serie
				if (serie_id === 'TOTAL'){
					main_serie = serie_data; // override previous declared main serie
				} else {
					secondary_series[serie_id] = serie_data;
					// TOTAL indicator is not shown on serie selection
					self.__add_serie_selector(campaign_id, serie_id, capitalize(_T(serie_id)));
				}
			});
		}

		// initialize the chart member
		this.m_charts_by_id[campaign_id][chart_data.indicator_id] = {
			chart_data : chart_data.data,
			secondary_series : secondary_series,
			chart_type : chart_data.chart_type,
			tooltip : chart_data.tooltip
		};

		var chart = new inconcert.Chart({
			div: chart_data.indicator_container_id,
			ChartProperties : {
				title : chart_data.description,
				allowZoom: false,
				useUTC : true,
				minExtreme: 0,
				yFormatter: function(n) {
					var format = self.__chart_Y_formatter(n);
					return format;
				},
				xFormatter: function(timestamp) {
					var date = new Date(timestamp);
					return inConcert.Utils.zeropad(date.getUTCHours()) + ':' + inConcert.Utils.zeropad(date.getUTCMinutes());
				},
				xNeedsTicks: function (min, max) {
					var ticks = self.__chart_X_need_ticks(min, max, campaign_id);
					return ticks;
				}
			},
			Series: {
				Properties: {
					showSeriesSelector: false
				},
				Main: main_serie,
				Secondary : secondary_series
			},
			onAfterBuild : onAfterBuild
		});
		return chart;
	};

	/**
	 * this function will build the chart in a timeout with a closure,
	 * this is done to give the runtime time to process when he is free of operations.
	 *
	 * @private
	 * @param  {String}             campaign_id                        The campaign identificator (campaign@vcc)
	 * @param  {Object}             chart_data                         The chart internal data
	 * @param  {Object}             chart_data.indicator_container_id  The chart container id
	 * @param  {Object|Array}       chart_data.data                    The chart data points
	 * @param  {Function} callback                                     Called when the creation it's done
	 * @return {Function}                                              The closure
	 */
	DashboardIndicators.prototype.__build_chart_closure = function(campaign_id, chart_data, callback = $.noop)
	{
		const self = this;

		var cleanup_timeout = function() {
			if (self.m_hnd_building_chart) {
				window.clearTimeout(self.m_hnd_building_chart);
				self.m_hnd_building_chart = null;
			}
		};
		cleanup_timeout();

		return function()
		{
			cleanup_timeout();
			// update waiting message
			self.m_dashboard.update_waiting_message(FORMAT("<%=_T('Building charts for campaign: {0}')%>", campaign_id));

			self.m_hnd_building_chart = window.setTimeout(function() {
				// Timeout en espera para poder dejar de visualizar el waiting de charts
				self.m_dashboard.hide_waiting();
				cleanup_timeout();
			}, 3000);

			const $indicator_node = $('#' + chart_data.indicator_container_id);

			// chart already exists?
			if ( $indicator_node.length === 0 ){
				return;
			}

			// build the real chart, this will call the external charts library
			var current_chart = self.__build_chart(chart_data, campaign_id, function() { 
				callback();
			});

			// redraw chart
			current_chart.Redraw();

			// Toggle *not enough data* message.
			if ( chart_data.data.length === 0 ){
				let no_data_msg = Mustache.render(self.m_templates.indicator_no_data_tpl, { message: "<%=_T('Not enough data to display chart')%>" });
				$indicator_node.find('.formIndicatorPanelContent').prepend(no_data_msg);
				$indicator_node.find('.indicatorContentPanelChart').hide();
			} else {
				$indicator_node.find('.trendChartMessageContainer').remove();
				$indicator_node.find('.indicatorContentPanelChart').show();
			}

			// update member with the chart from the library
			self.m_charts_by_id[campaign_id][chart_data.indicator_id].chart = current_chart;
		};
	};

	DashboardIndicators.prototype.__create_new_chart_node = function(indicator_id, indicator_complete_id, description) {
		var chart_template = this.m_templates.indicator_chart_tpl;
		return $(Mustache.render(chart_template, {
			title: description || indicator_id,
			id: indicator_complete_id 
		}));
	};

	DashboardIndicators.prototype.__update_chart_data = function(campaign_id, indicator_id, $indicator_node, new_data, contains_secondary_series){
		var self = this;
		var chart_object  = self.m_charts_by_id[campaign_id][indicator_id],
			previous_data = chart_object.chart_data,
			current_chart = chart_object.chart;

		// update data inside the chart only if has changed.
		if (!_.isEqual(previous_data, new_data)){

			// update chart hash data.
			self.m_charts_by_id[campaign_id][indicator_id].chart_data = new_data;

			// if the indicators is serie data, we need to handle differently
			if (contains_secondary_series){
				var secondary_series = self.m_charts_by_id[campaign_id][indicator_id].secondary_series;
				// update each serie data
				_.each(new_data, function(serie, channel){
					var series_names = current_chart.Series.Secondary.GetSecondarySeriesNames();
					if (series_names.length > 0 && _.contains(series_names, channel)){
						current_chart.Series.Secondary.SetDataSource (channel, {
							source : serie.data
						});
					} else if (channel === "TOTAL") {
						current_chart.Series.Main.SetDataSource ({
							source : serie.data
						});
					}
					else {
						self.__add_legend(campaign_id, channel, capitalize(_T(channel)));
						secondary_series[channel] = self.__get_default_chart_properties(serie.data, 
								chart_object.chart_type, chart_object.tooltip, channel);
						self.__add_serie_selector(campaign_id, channel, capitalize(_T(channel)));
						var $chk_serie = self.__get_serie_selector_item(campaign_id, channel);
						if ($chk_serie.is(":checked")) {
							current_chart.Series.Secondary.AddSerie(channel, {
								source: serie.data
								},
								{
									// TO-DO: Get this from a function
									type : chart_object.chart_type,
									drawPoints : true,
									lineWidth : 2,
									color : self.m_chart_serie_colors[channel],
									hintString: '<div>['+chart_object.tooltip+']</div>'
								});
						}
					}
				});
				_.each(previous_data, function(serie, channel) {
					if (!new_data[channel]) {
						if (secondary_series[channel]){
							current_chart.Series.Secondary.RemoveSerie(channel);
						}
					}
				});
			}
			else {
				current_chart.Series.Main.SetDataSource ({
					source : new_data
				});
			}

			// handle chart without previous data.
			var previous_message = $indicator_node.find('.trendChartMessageContainer');
			if ( previous_message.length > 0 ){
				previous_message.remove();
				$indicator_node.find('.indicatorContentPanelChart').show();
			}
		}
		//always do a redraw
		current_chart.Redraw();
	};

	DashboardIndicators.prototype.__update_chart_visibilty = function(current_config, $current_indicator_node, indicator_dom_id){
		// update indicator config.
		if ( this.__chart_should_be_visible(current_config, indicator_dom_id)){
			// this is done this way because of a high cpu consume
			// detected making show on every refresh tick
			if ( $current_indicator_node.is(':hidden')){
				$current_indicator_node.show();
			}
			$current_indicator_node.find('.chartOptionCheck').prop('checked', true);
		}
		// hide indicator node and uncheck checkbox
		else{
			// this is done this way because of a high cpu consume
			// detected making hide on every refresh tick
			if ( $current_indicator_node.is(':visible')){
				$current_indicator_node.hide();
			}
			$current_indicator_node.find(':checkbox').prop('checked', false);
		}
	};

	DashboardIndicators.prototype.process_indicators = function(campaign_id, indicators, on_indicators_processed = $.noop)
	{
		var self = this;

		// handle desconnections from inConcert server
		if (!campaign_id || _.isEmpty(indicators)) {
			on_indicators_processed();
			return;
		}

		var current_config = this.m_config.get_panel_config(campaign_id, 'sup_dash_indicators_panel');

		//assume default values
		current_config = current_config || {slice: 'Hour', time_ago: '8Hour'};

		if ( !this.m_charts_by_id[campaign_id] ){
			this.m_charts_by_id[campaign_id] = {};
		}

		var indicators_left_to_process = indicators.length;
		var check_if_ready_to_finish = function() {
			indicators_left_to_process--;
			if (indicators_left_to_process === 0){
				on_indicators_processed();
			}
		};

		// process every chart
		var campaign_dom_id = nid(campaign_id);
		var $indicators_container = this.__get_panel_container(campaign_id);

		indicators.forEach(function(indicator)
		{
			var data = indicator.data;
			var slice = self.__get_slice_minutes(current_config && current_config.slice ? current_config.slice : '15Min') * 60 * 1000;

			//slice >= Day
			if (slice >= 60 * 24 * 1000) {
				slice = 0;
			}

			//add slice duration to timestamps so each point represents the end of a slice
			if (data && data[0] && data[0][0] && typeof(data[0][0]) !== 'string') {
				for (var j = 0; j < data.length; j++) {
					data[j][0] += slice;
				}
			}

			var indicator_id = nid(indicator.id);
			var description = indicator.description;
			var indicator_complete_id = campaign_dom_id + '_' + indicator_id;
			var indicator_dom_id = 'sup_dash_indicators_' + indicator_complete_id;
			var $current_indicator_node = $('#' + indicator_dom_id);

			// create chart if does not exists.
			if ( $current_indicator_node.length === 0 ) {
				$current_indicator_node = self.__create_new_chart_node(indicator_id, indicator_complete_id, description);

				// append node and build
				$indicators_container.append($current_indicator_node);

				// if the first element of the data array is an string
				// I will assume that will be a bar chart (string is the legend)
				var chart_type = data[0] && data[0][0] && typeof(data[0][0]) === 'string' ? 'bar' : 'line';
				var tooltip = chart_type === 'bar' ? 'label' : 'y';

				var chart_data = {
					indicator_id : indicator_id,
					indicator_container_id : 'sup_dash_indicators_container_' + indicator_complete_id,
					description : description,
					data : data,
					chart_type : chart_type,
					tooltip : tooltip,
					containsSecondarySeries : indicator.containsSecondarySeries || false
				};

				// build chart in a closure to avoid freezes
				setTimeout(
					self.__build_chart_closure(campaign_id, chart_data, check_if_ready_to_finish),
				0);
			}
			else if (self.m_charts_by_id[campaign_id][indicator_id]) {
				self.__update_chart_data(campaign_id, indicator_id, $current_indicator_node, data, indicator.containsSecondarySeries || false);
			}

			// update visibility based on configuration
			self.__update_chart_visibilty(current_config, $current_indicator_node, indicator_dom_id);
		});
	};

	DashboardIndicators.prototype.on_dashboard_opened = function(campaign_id){
		this.__redraw_charts(campaign_id);
	};

	DashboardIndicators.prototype.refresh_realtime_charts = function (campaign_id, realtime_indicators, callback) {
		callback = callback || $.noop;
		if (this.allowed_to_refresh_view(campaign_id)){
			this.process_indicators(campaign_id, realtime_indicators, function(){
				callback();
			});
		}
	};

	DashboardIndicators.prototype.setup_realtime_charts = function (campaign_id, realtime_indicators, callback) {
		callback = callback || $.noop;
		this.process_indicators(campaign_id, realtime_indicators, function(){
			callback();
		});
	};

	DashboardIndicators.prototype.refresh = function (campaign_id, callback, indicators)
	{
		callback = callback || $.noop;
		if (this.allowed_to_refresh_view(campaign_id)){
			this.setup(campaign_id, function(success, info) {
				callback(success, info);
			}, indicators);
			this.__update_channels_legend(campaign_id);
		}
	};

	DashboardIndicators.prototype.get_historical_data_filter = function (campaign_id)
	{
		let data = this.m_config.get_panel_config(campaign_id, 'sup_dash_indicators_panel');
		if ( !data ) {
			data = {
				slice: '15Min',
				time_ago: '2Hour'
			};
		}

		data.campaign_id = campaign_id.replace(/^(.*?)@(.*?)$/, '$1');

		let campaign_config = this.m_dashboard.get_campaign_selected_config(campaign_id);
		data.account = !this.m_dashboard.isAllegro(campaign_id) && campaign_config.account_name || null;
		if (data.account) {
			data.channel = campaign_config.channel || null;
		}
		data.attention_level = campaign_config.al_name;
		return data;
	};

	DashboardIndicators.prototype.setup = function (campaign_id, on_setup_finished = $.noop, indicators)
	{
		const self = this;

		const $current_node = this.m_dashboard.get_campaign_node(campaign_id);
		if ($current_node.length === 0) {
			return;
		}
		this.setup_edit_mode(campaign_id);

		const cbGet_historical_data = function(indicators, err) {
			if (indicators && !err) {
				self.process_indicators(campaign_id, indicators, function(){
					self.setup_title_collapse_handlers($current_node);
					on_setup_finished(true, indicators);
				});
			}
			else {
				on_setup_finished(false);
			}
		};

		if (typeof indicators !== "undefined") {
			return cbGet_historical_data(indicators);
		}
		let data = this.get_historical_data_filter(campaign_id);
		this.m_model.get_historical_data( data, cbGet_historical_data);
	};
}());
