/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global $, inConcert, Mustache, _ */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name SupervisorDashboardInteractions
 */
(function () {

	'use strict';
	const format = inConcert.Utils.format;
	const nid = inConcert.Utils.getNormalizedId;

	/**
	 * Creates a new interactions dashboard instance.
	 * Contains common operations for archive, queue, monitoring views.
	 *
	 * @author Sebastian Moreno (smoreno@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorDashboardInteractions
	 * @param view_id
	 * @param view_name
	 * @param dashboard
	 */
	function DashboardInteractionsView(view_id, view_name, dashboard)
	{
		this.m_type = 'interactions';
		this.m_id = view_id;
		this.m_name = view_name;
		this.m_interaction_gui = new inConcert.interaction_gui( inConcert.supervisor_actions, this.__get_interaction_gui_config() );
		this.m_templates = inConcert.sup_templates.dashboard;
		this.m_templates.interaction_node_tpl = this.m_interaction_gui.get_node_tpl();
		this.m_dashboard = dashboard;
		this.m_params = null;
		this.$campaign = null;
		this.m_view_classes_name = '';
		// lista de agentes indexados por el nombre completo del agente
		this.agents_by_name  = {};
		this.initialized = false;
		// app jquery selector, used for inconcert tooltip
		this.m_interaction_gui.setTooltips( '#supervisor_main', '.columnContentAux [tooltip]' );
		// workgroups
		this.m_workgroups = dashboard.get_workgroup_manager();
		// current config
		this.m_config = inConcert.sup_configuration;
		// Este valor se obtiene del LUA "/apps/agent/shared/utils.lua", funcion "GetMaxAllowedRowsInES"
		this.MAX_ALLOWED_ROWS_IN_ES = inConcert.app.getServerInfo().maxAllowedRowsInES;
		this.INTERACTIONS_PER_PAGE = 20;
		this.m_can_refresh_interactions = null;
	}

	if (typeof inConcert.sup_dashboard_interactions_view !== 'function') {
		inConcert.sup_dashboard_interactions_view = DashboardInteractionsView;
	}

	DashboardInteractionsView.prototype.__get_interaction_gui_config = function()
	{
		return {
			from_app : 'supervisor',
			owner    : this,
			columns  : {
				user_owner : { show: true }
			}
		};
	};

	DashboardInteractionsView.prototype.init = function( $campaign, params )
	{
		if ( !$campaign || $campaign.length === 0 ) {
			throw 'DashboardInteractionsView.init - Error DOM campaign is undefined!!';
		}
		this.update_viewer_classes($campaign);
		var campaign_id = $campaign.attr('campaign_id');
		this.m_dashboard.set_current_view_id( campaign_id, this.m_id );
		this.m_params = params || {};
		var $batch_actions_container = this.$get_batch_actions_container(campaign_id);
	
		if ( !this.has_configured_batch_actions(campaign_id) ) {
			$batch_actions_container.hide();
		}
		else {
			var allowed_batch_actions = this.has_permission_to_batch_actions();
			var $marked_interactions = this.$get_marked_interactions(campaign_id);
			var selected_interactions = $marked_interactions.length >= 1;

			if ( allowed_batch_actions && selected_interactions ) {
				this.disable_batch_actions(campaign_id);
				$batch_actions_container.show();
				$batch_actions_container.find('[tooltip]').inconcert_tooltip();
			}
			else {
				$batch_actions_container.hide();
			}
		}

		this.showOrHideColumns(campaign_id);
		this.set_events(campaign_id);
		this.initialized = true;
	};

	DashboardInteractionsView.prototype.is_initialized = function() { return this.initialized; };

	DashboardInteractionsView.prototype.on_change_config_selector = function( /*campaign_id*/ ) {};

	DashboardInteractionsView.prototype.get_campaign_normalized_id = function( campaign_id )
	{
		return this.m_dashboard.get_normalized_id(campaign_id);
	};

	DashboardInteractionsView.prototype.$get_campaign_node = function( campaign_id )
	{
		var campaign_dom_id = this.get_campaign_normalized_id(campaign_id);
		var $campaign = this.m_dashboard.get_campaign_node_from_dom_id(campaign_dom_id);
		return $campaign;
	};

	DashboardInteractionsView.prototype.cleanup_search_text = $.noop;

	DashboardInteractionsView.prototype.set_values_search_text = $.noop;

	DashboardInteractionsView.prototype.disable_batch_actions = function( campaign_id )
	{
		const $check_all = this.$get_batch_actions_mark_all(campaign_id);
		if ( $check_all.prop('checked') ) {
			$check_all.prop('checked', false);
		}
		$check_all.trigger('click_mark_all');
	};

	DashboardInteractionsView.prototype.hide_show_select_all_batch_actions = function( campaign_id )
	{
		const $interactions = this.get_interactions(null, campaign_id);
		const $mark_all = this.$get_batch_actions_mark_all(campaign_id);
		if ($interactions && $interactions.find('[int_batch_selector]').length > 0) {
			$mark_all.show();
			return;
		}
		$mark_all.hide();
	};

	DashboardInteractionsView.prototype.$get_batch_actions_container = function( campaign_id )
	{
		var $campaign = this.$get_campaign_node(campaign_id);
		var $batch_actions_container = $campaign.find('[batch_actions_container]');
		return $batch_actions_container;
	};
	
	DashboardInteractionsView.prototype.$get_batch_actions_mark_all = function( campaign_id )
	{
		const $campaign = this.$get_campaign_node(campaign_id);
		return $campaign.find('[interactions_table_headers] [batch_actions_mark_all]');
	};

	DashboardInteractionsView.prototype.has_configured_batch_actions = function(campaign_id)
	{
		if (this.m_dashboard.isAllegro(campaign_id)) {
			return false;
		}
		var $batch_actions_container = this.$get_batch_actions_container(campaign_id);
		if ( $batch_actions_container.length === 0 ) {
			return false;
		}
		return this.m_params.has_batch_actions;
	};

	DashboardInteractionsView.prototype.set_events = function( campaign_id )
	{
		this.set_batch_events(campaign_id);
		var campaign_dom_id = this.get_campaign_normalized_id(campaign_id);
		var $container = this.get_interactions_container(campaign_dom_id);
		this.m_interaction_gui.set_events($container);
	};

	DashboardInteractionsView.prototype.set_batch_events = function(campaign_id)
	{
		var self = this;
		if ( !this.has_configured_batch_actions(campaign_id) ) {
			return;
		}

		var allowed_batch_actions = this.has_permission_to_batch_actions();
		if ( !allowed_batch_actions ) {
			return;
		}
		var $batch_actions_container = this.$get_batch_actions_container(campaign_id);

		// select/unselect all interactions button
		var $batch_actions_mark_all = this.$get_batch_actions_mark_all(campaign_id);
		$batch_actions_mark_all
			.unbind('click_mark_all')
			.bind('click_mark_all', function()
			{
				var $chk = $(this);
				var mark = $chk.prop('checked');
				$chk.attr('tooltip', mark ? "<%=_T('Unselect all conversations')%>" : "<%=_T('Select all conversations')%>" );
				self.mark_unmark_batch_interactions(mark, campaign_id);
				self.update_batch_buttons(campaign_id);
			})
			.unbind('click')
			.bind('click', function()
			{
				$(this).trigger('click_mark_all');
			});

		$batch_actions_container.find('[batch_action_id]')
			.unbind('click')
			.bind('click', function()
			{
				if ($(this).hasClass("disabledInfo")) {
					return;
				}
				var $button = $(this);
				self.execute_batch_action($button, campaign_id);
			});
	};

	DashboardInteractionsView.prototype.execute_batch_action = function( $button, campaign_id )
	{
		var self = this;

		if ($button.hasClass("disabledInfo")) {
			return;
		}

		if ( !this.has_permission_to_batch_actions() ) {
			var message = "<%=_T('You have not permissions to batch actions.')%>";
			inConcert.app.getMainWindow().showBannerMessage({ message : message, messageType : 'Error', timeout : 5000});
			return;
		}

		var marked_interactions = [];
		var update_interactions = function( action_fn )
		{
			marked_interactions.forEach(function(interaction)
			{
				var $interaction = self.$find_interaction( interaction );
				if ( $interaction.length > 0 ) {
					action_fn( interaction, $interaction );
				}
			});
		};

		var batch_actions = {
			TRANSFER: function()
			{
				var popup_callbacks = {
					transfer: {
						on_transfer_success : function() {
							self.disable_batch_actions(campaign_id);
						},
						on_before_transfer_execute : function(transfer_to, interactions) {
							$.each(interactions, function(_, interaction){
								const view = interaction && inConcert.app.getInteractionManager().getInteractionClient(interaction);
								if (view && view.on_before_transfer_execute) {
									view.on_before_transfer_execute(transfer_to);
								}
							});
						},
						on_transfer_error : $.noop,
						onOpen: function() {
							update_interactions(function(_, $interaction) {
								$interaction.prop('selected_to', 'TRANSFER').addClass('selectedToTransfer');
							});
						},
						onClose: function() {
							update_interactions(function(_, $interaction) {
								$interaction.prop('selected_to', '').removeClass('selectedToTransfer');
							});
						}
					}
				};
				// instancia la clase "interaction_gui" para cargar el submenu de transfer ("member", "campaign", "att level" y "number")
				var interaction_gui = new inConcert.interaction_gui(inConcert.supervisor_actions,
					{
						from_app : 'supervisor',
						owner    : this,
						callbacks: popup_callbacks,
						custom_allowed_actions: {
							transfer: true
						}
					}
				);
				interaction_gui.show_submenu_actions(marked_interactions, {
					$parent: $button,
					action_type: 'Transfer',
					align: null // para que no quede alineado a la izquierda, ya que los submenus se alineas siempre a la izquierda.
				});
			}
		};

		var action = $button.attr('batch_action_id');
		var action_fn = batch_actions[action];
		if ( typeof action_fn !== 'function' ) {
			throw 'DashboardInteractionsView.execute_batch_action. Unrecognized batch action: ' + action;
		}
		var $marked_interactions = this.$get_marked_interactions(campaign_id);
		$.each($marked_interactions, function( _, marked_interaction )
		{
			var interaction_data = $(marked_interaction).data('interaction');
			marked_interactions.push(interaction_data);
		});
		action_fn();
	};

	DashboardInteractionsView.prototype.mark_unmark_batch_interactions = function( mark, campaign_id )
	{
		var $interactions = this.get_interactions(null, campaign_id);
		$interactions.find('[int_batch_selector]').prop('checked', mark);
		$interactions.removeClass('interactionSelectToBatch');
		if (mark) {
			$interactions.addClass('interactionSelectToBatch');
		}
		return $interactions;
	};

	DashboardInteractionsView.prototype.$get_marked_interactions = function( campaign_id )
	{
		var $interactions = this.get_interactions(null, campaign_id);
		var $marked_interactions = [];

		$.each($interactions, function(_, interaction)
		{
			var $interaction = $(interaction);
			var $batch_selector = $interaction.find('[int_batch_selector]');
			// las interacciones mias, no tiene el selector (chk) visible
			if ($batch_selector.is(':visible')){
				if ( $batch_selector.prop('checked') ) {
					$marked_interactions.push($interaction);
				}
			}
		});

		return $marked_interactions;
	};

	DashboardInteractionsView.prototype.update_batch_buttons = function( campaign_id )
	{
		var $batch_actions_container = this.$get_batch_actions_container(campaign_id);
		var $actions = $batch_actions_container.find('[batch_action_id]');
		var $marked_interactions = this.$get_marked_interactions(campaign_id);
		var count_marked = $marked_interactions.length;
		var disabled = count_marked === 0;
		var $batch_actions_container_count_text = this.$get_batch_actions_container(campaign_id).find('[batch_actions_container_number]');

		$batch_actions_container.hide();

		$.each($actions, function(_, action)
		{
			var $action = $(action);
			if ( disabled ) {
				$action.prop('disabled', true);
				$action.removeClass('disabledInfo');
				$batch_actions_container.hide();
			}
			else {
				var button_state = 'enabled';
				var action_id = $action.attr('batch_action_id');
				var has_permission_to_action = true;
				var disabled_tootip = '';
				var channel;
				$.each($marked_interactions, function(_, interaction)
				{
					var $interaction = $(interaction);
					var interaction_data = $interaction.data( 'interaction' );
					if ( interaction_data ) {
						has_permission_to_action = interaction_data.hasPermission( action_id, 'supervisor' );
						if ( !has_permission_to_action ) {
							button_state = 'disabled';
							disabled_tootip = "<%=_T('You have not permissions to transfer conversations.')%>";
							return false; // break
						}
						if ( channel === undefined ) {
							channel = interaction_data.type;
						}
						else {
							if ( channel !== interaction_data.type ) {
								button_state = 'disabled';
								disabled_tootip = "<%=_T('You can not transfer conversations from different channels!')%>";
								return false; // break
							}
						}
					}
				});

				$action.prop('disabled', false);
				if ( button_state === 'enabled' ) {
					$batch_actions_container.show();
					$action
						.removeClass('disabledInfo')
						.attr('tooltip', format("<%=_T('{0} Selected')%>", count_marked)); 
					$batch_actions_container_count_text.text(format("<%=_T('{0}')%>", count_marked));
				}
				else {
					$action
						.addClass('disabledInfo')
						.attr('tooltip', disabled_tootip); 
					$batch_actions_container.hide();
				}
			}
		});
	};

	DashboardInteractionsView.prototype.has_permission_to_batch_actions = function()
	{
		// TODO: Validar este permiso de accion en batch si deberia estar en la app supervisor.
		//return inConcert.permissions.to_action_on_any_campaign( 'interaction', 'int_bulk_actions' );
		return true;
	};

	DashboardInteractionsView.prototype.get_navigation = function (campaign_id)
	{
		return this.m_dashboard.get_navigation(campaign_id);
	};

	DashboardInteractionsView.prototype.__must_show_alarm = function (/*campaign_id, interaction*/)
	{
		var response = {must_show_alarm: false};
		return response;
	};

	DashboardInteractionsView.prototype.__get_alarm_classes = function ( interaction, $interaction )
	{
		var classes = '';
		var campaign_id = interaction.campaign;
		var response = this.__must_show_alarm(campaign_id, interaction);
		if (!response.must_show_alarm) {
			return classes;
		}

		var alarm = this.m_dashboard.get_notification_alarm_by_indicator({
			indicator_id : response.indicator_id,
			campaign_id  : response.campaign_id,
			account_id   : response.account_id,
			al_id        : response.al_id
		});

		$interaction.attr('title', '')
			.data('alarm', {});
		if (alarm) {
			var alarm_message = this.m_dashboard.get_notification_alarm_message(alarm, response.current_value);
			var tooltip = alarm_message ? alarm_message.tooltip : '';
			$interaction.attr('title', tooltip)
				.addClass('alarmed')
				.data('alarm', alarm);
			classes = 'alarmed ';
		}

		return classes;
	};

	DashboardInteractionsView.prototype.get_specifics_interaction_classes = function( interaction, $interaction )
	{
		if ( typeof interaction !== 'object' ) {
			return '';
		}
		var classes = this.__get_alarm_classes(interaction, $interaction);

		if ( !interaction.belongsToMe() ) {
			return classes;
		}
		return classes + 'userInteraction';
	};

	DashboardInteractionsView.prototype.get_static_interactions_container = function(campaign_dom_id) {
		return $(`#sup_dash_static_interactions_container_${campaign_dom_id}`);
	};

	DashboardInteractionsView.prototype.get_interactions_viewer_node = function(campaign_dom_id){
		return $('#sup_dash_interaction_viewer_' + campaign_dom_id);
	};

	DashboardInteractionsView.prototype.get_operations_viewer_node = function(campaign_dom_id){
		return $('#sup_dash_operations_viewer_' + campaign_dom_id);
	};

	DashboardInteractionsView.prototype.get_interactions_container = function(campaign_dom_id) {
		const $viewer = $('#sup_dash_interaction_viewer_container_' + campaign_dom_id);
		return $viewer;
	};

	DashboardInteractionsView.prototype.get_interactions = function( $container, campaign_id )
	{
		if ( !$container ) {
			var campaign_dom_id = this.get_campaign_normalized_id(campaign_id);
			$container = this.get_interactions_container(campaign_dom_id);
		}
		return $container.find( 'div[node_type="interaction"]' );
	};

	DashboardInteractionsView.prototype.$get_all_groups_interactions = function( $container ) {
		if ( !$container ) {
			return;
		}
		return $container.find( 'div[node_type="group_interactions"]' );
	};

	DashboardInteractionsView.prototype.__get_face_by_name = function(name)
	{
		name = name.toLowerCase();
		if (!this.facets_by_name) {
			return name;
		}
		if (this.facets_by_name[name]) {
			return this.facets_by_name[name];
		}
		return name;
	};

	DashboardInteractionsView.prototype.__make_faces_by_name = function(facets)
	{
		var self = this;
		this.facets_by_name = {};
		var facet_labels = [];
		$.each(facets, function(_, facet) {
			self.facets_by_name[facet.label.toLowerCase()] = facet.value;
			facet_labels.push({ value : facet.label, label : facet.label });
		});
		return facet_labels;
	};

	DashboardInteractionsView.prototype.get_filter_facets = $.noop;

	DashboardInteractionsView.prototype.show_waiting = function (campaign_dom_id, message) {
		$('#sup_dash_interactions_waiting_' + campaign_dom_id).show();
		$('#sup_dash_interactions_waiting_msg_' + campaign_dom_id).text(message || '');
	};

	DashboardInteractionsView.prototype.__remove_viewer_classes = function($campaign) {
		$campaign.find('[viewer_type="interactions"]').removeClass('archiveViewer queuedViewer waitingViewer abandonedViewer bot_interactionsViewer monitoringViewer appointmentViewer globalViewer');
		$campaign.find('[campaign_box_title]').removeClass('archiveViewerTitle queuedViewerTitle waitingViewerTitle abandonedViewerTitle interactionsViewerTitle monitoringViewerTitle appointmentViewerTitle globalViewerTitle global_searchViewerTitle BotsViewer');
	};

	DashboardInteractionsView.prototype.__add_viewer_classes = function($campaign, classes_name) {
		classes_name = classes_name || this.m_view_classes_name;
		$campaign.find('[viewer_type="interactions"]').addClass(classes_name);
		$campaign.find('[campaign_box_title]').addClass(classes_name+"Title");
	};

	DashboardInteractionsView.prototype.update_viewer_classes = function($campaign) {
		this.__remove_viewer_classes($campaign);
		this.__add_viewer_classes($campaign);
	};

	DashboardInteractionsView.prototype.hide_waiting = function (campaign_dom_id) {
		$('#sup_dash_interactions_waiting_' + campaign_dom_id).hide();
	};

	DashboardInteractionsView.prototype.remove_no_conversations_msg = function (campaign_dom_id) {
		return $(`#sup_dash_campaign_${campaign_dom_id}_views_no_interactions_msg`).remove();
	};

	DashboardInteractionsView.prototype.toggle_no_conversations_msg = function (campaign_dom_id, message) {
		this.remove_no_conversations_msg(campaign_dom_id);
		var container = this.get_static_interactions_container(campaign_dom_id);
		container.append($(Mustache.render(this.m_templates.no_interactions_msg, { normalized_id: campaign_dom_id, message: message })));
	};

	DashboardInteractionsView.prototype.clear_list = function(campaign_dom_id)
	{
		// remove previous messages
		this.remove_no_conversations_msg(campaign_dom_id);
		var interactions_container = this.get_interactions_container(campaign_dom_id);
		// remove scroll bindings.
		interactions_container.parent().unbind();
		// remove previous interactions and group interactions
		interactions_container.children().remove();
	};

		/*
		Esta funcion valida si se puede refrescar las interacciones.
		Se usa en el on_ready_to_refresh del inconcert_querybuilder
	*/
	DashboardInteractionsView.prototype.can_refresh_interactions = function() {
		return this.m_can_refresh_interactions !== false;
	};

	DashboardInteractionsView.prototype.set_can_refresh_interactions = function(can_refresh) {
		this.m_can_refresh_interactions = can_refresh;
	};

	/**
	 * Returns true if the current interactions view has changed
	 * @param  {object} campaign_dom_id - The campaign interaction node
	 * @return {boolean}
	 */
	DashboardInteractionsView.prototype.view_has_changed = function(campaign_id)
	{
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var interactions_container = this.get_interactions_container(campaign_dom_id);
		var has_changed = interactions_container.is(':visible');
		if ( has_changed ) {
			var navigation = this.get_navigation(campaign_id);
			if ( navigation ) {
				var selected_folder_id = navigation.get_selected_folder_id();
				has_changed = selected_folder_id !== this.navigation_folder_id;
			}
		}
		return has_changed;
	};

	DashboardInteractionsView.prototype.has_permission_to = function ( action_id, campaign_id ) {
		return inConcert.permissions.to_action( 'supervisor', action_id, undefined /* category */, campaign_id, true /* no log */ );
	};

	DashboardInteractionsView.prototype.open_view = function( campaign_id )
	{
		const self = this;
		if ( this.permission_id ) {
			let has_permission = this.has_permission_to( this.permission_id, campaign_id );
			if ( !has_permission ) {
				// TODO: Message ?
				return false;
			}
		}
		let $campaign = this.m_dashboard.get_campaign_node(campaign_id);
		$campaign.addClass('interactionsViewerCondition');
		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);

		// hide dashboard and menu
		$campaign.find('[viewer_type="operations"]').hide();
		$('#sup_dash_panel_menu_config_view_' + campaign_dom_id).hide();
		// show current view
		// FIX-ME: This should be done via css
		$campaign.removeClass('twoPanels').addClass('threePanels');
		$campaign.find('[viewer_type="interactions"]').addClass(this.m_id + 'Viewer').show();
		$campaign.find('[campaign_box_title]').addClass(this.m_id + 'ViewerTitle');

		//
		//TODO: DESACTIVAR navegacion ??

		$campaign.find('input[view_id="' + this.m_id + '"]').prop('checked', true);

		let allowed_panels = []; // panel con permisos
		// permisos ids de cada panel (indicadores, team, summary)
		let panels_permissions = this.m_dashboard.panels_permissions;
		$.each( panels_permissions, function( _, permission_id )
		{
			let has_permission = self.has_permission_to( permission_id, campaign_id );
			if ( has_permission ) {
				allowed_panels.push( true );
			}
		});
		// show or hide back to dashboard button
		let $back = this.m_dashboard.$back_button(campaign_id);
		let $full_screen_back_button = this.m_dashboard.$full_screen_back_button(campaign_id);
		if (allowed_panels.length > 0) {
			$back.show();
			$full_screen_back_button.show();
		}
		else {
			$back.hide();
			$full_screen_back_button.hide();
		}

		$campaign.trigger('view_opened');
		this.m_dashboard.on_dashboard_closed(campaign_id, this.m_id);

		return true;
	};

	DashboardInteractionsView.prototype.on_view_load_finished = function(_){
		//
		//TODO: ACTIVAR navegacion ??
	};

	DashboardInteractionsView.prototype.setup_pagination = function(campaign_id, search_fn)
	{
		var self = this;

		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var page_buttons = [
			{ selector: '#interactions_page_first_' + campaign_dom_id , tooltip: "<%=_T('First Page')%>"    },
			{ selector: '#interactions_page_next_'  + campaign_dom_id , tooltip: "<%=_T('Next Page')%>"     },
			{ selector: '#interactions_page_prev_'  + campaign_dom_id , tooltip: "<%=_T('Previous Page')%>" },
			{ selector: '#interactions_page_last_'  + campaign_dom_id , tooltip: "<%=_T('Last Page')%>"     }
		];

		page_buttons.forEach(function(button)
		{
			var $button_node = $( button.selector );

			$button_node
				.unbind('click_navigate_to')
				.bind('click_navigate_to', function() {
					// Disable until finish pagination.
					$(this).prop('disabled', true);

					// show waiting and clear interactions list.
					self.show_waiting(campaign_dom_id);
					self.clear_list(campaign_dom_id);
					self.disable_batch_actions(campaign_id);
					// Call the search function with the button clicked and a callback function.
					search_fn( this, function() {
						self.hide_waiting(campaign_dom_id);
					});
				});

			$button_node
				.attr('tooltip', button.tooltip)
				.unbind('click')
				.bind('click', function(){
					$(this).trigger('click_navigate_to');
				});
		});
	};

	DashboardInteractionsView.prototype.get_interaction_dom_id = function( interaction ) {
		return this.__get_prefix() + nid( interaction.id );
	};

	DashboardInteractionsView.prototype.$find_interaction = function( interaction )
	{
		var interaction_dom_id = this.get_interaction_dom_id(interaction);
		var $campaign = this.$get_campaign_node(interaction.campaign);
		var $interaction = $campaign.find( '#' + interaction_dom_id );
		return $interaction;
	};

	DashboardInteractionsView.prototype.__get_prefix = function( /*interaction*/ )
	{
		throw('DashboardInteractionsView.prototype.__get_prefix. It must implemented in derived classes.');
	};

	DashboardInteractionsView.prototype.__get_dom_interaction_params = function( /*interaction*/ )
	{
		throw('DashboardInteractionsView.prototype.__get_dom_interaction_params. It must implemented in derived classes.');
	};

	DashboardInteractionsView.prototype.process_off_dom_interaction_update = function(campaign_dom_id, new_interactions, interaction_tpl, custom_update_fn)
	{
		var self = this;

		custom_update_fn = custom_update_fn || $.noop;
		var $container = self.get_interactions_container(campaign_dom_id);
		var $scroll_container = $container.parent();
		var container_top = $scroll_container.scrollTop();

		// detach the node from the DOM to make faster udpates.
		$container = $container.detach();

		// get previous ids
		var previous_interactions_ids = [];
		var interactions_nodes = this.get_interactions( $container );

		interactions_nodes.each(function() {
			if ( this.id ){
				previous_interactions_ids.push(this.id);
			}
		});

		// 1 - lista con los ids de las nuevas interacciones.
		//     el mismo se valida contra la lista de interacciones existentes.
		var new_interactions_map = {};
		// recorro y
		$.each( new_interactions, function( _, interaction )
		{
			var interaction_dom_id = self.get_interaction_dom_id( interaction );
			// build a hash to check for removal later
			new_interactions_map[interaction_dom_id] = true;
		});

		// 2 - valido si las interacciones agregadas en (1), existen en las interacciones previas del DOM
		//     si no existen, las elimina del mismo.
		previous_interactions_ids.forEach(function(interaction_dom_id){
			if (!new_interactions_map[interaction_dom_id]){
				self.remove_interaction( $container, interaction_dom_id );
			}
		});

		// 3 - procesa las nuevas interacciones
		$.each( new_interactions, function( order, interaction )
		{
			var interaction_dom_id = self.get_interaction_dom_id( interaction );

			// get the interaction node
			var $interaction_node = $container.find( '#' + interaction_dom_id );

			if ( $interaction_node.length === 0 ){
			// new interaction, add it to the panel
				var parmams_interaction = self.__get_dom_interaction_params(interaction);
				$interaction_node = self.m_interaction_gui.$build_interaction_node( interaction, interaction_tpl, parmams_interaction );
				// el "order" es opcional, y es el orden en el cual se deben de agregar las nuevas interacciones.
				// este orden es calculado server side.
				self.insert_interaction_in_panel_list( interaction, $interaction_node, $container, order );
			}
			else {
				var previous_interaction_data = $interaction_node.data('interaction');
				self.attach_to_group_interaction( $container, previous_interaction_data, interaction );
			}

			// Update interaction node data.
			$interaction_node.data('interaction', interaction);
			// call a custom update function for each interaction node.
			custom_update_fn( interaction, $interaction_node );
		});

		// put back to the DOM and adjust scroll top position.
		$scroll_container.append($container);
		
		$container.find("[int_labels_disposition]").each(function(_, element) {
			self.m_interaction_gui.toggleInteractionLabels($(element));
		});

		$scroll_container.scrollTop(container_top);
	};

	DashboardInteractionsView.prototype.attach_to_group_interaction = function( $container, previous_interaction_data, interaction )
	{
		var it_id = this.get_interaction_dom_id( interaction );
		var $interaction = $container.find( '#' + it_id );
		if ( $interaction.length === 0 ) {
			return;
		}
		var prev_it_group = previous_interaction_data.getGroupId();
		if ( prev_it_group ) {
			return;
		}
		var it_group = interaction.getGroupId();
		if ( !it_group ) {
			return;
		}
		$interaction.detach();
		$interaction.attr( 'belongs_to_group', it_group );
		this.insert_interaction_in_panel_list( interaction, $interaction, $container );
	};

	DashboardInteractionsView.prototype.remove_interaction = function( $container, interaction_dom_id, has_remove_all_group_interactions )
	{
		var self = this;

		var remove = function( dom_id ) {
			var $it = $container.find( '#' + dom_id );
			if ( $it.length === 0 ) {
				return;
			}
			var it = $it.data( 'interaction' );
			$it.remove();
			return it;
		};

		var interaction = remove( interaction_dom_id );
		if ( !interaction ) {
			return;
		}

		if ( has_remove_all_group_interactions ) {
			var $group = this.$get_group_interactions( $container, interaction );
			if ( $group && $group.length > 0 ) {
				var children = $group.children();
				if ( children && children.length > 0 ) {
					$.each( children, function( _, it ) {
						var $it = $( it );
						var current_it = $it.data( 'interaction' );
						var dom_id = self.get_interaction_dom_id( current_it );
						remove( dom_id );
					});
				}
			}
		}

		this.remove_group_interactions( $container, interaction );
	};

	DashboardInteractionsView.prototype.detach_interaction_of_group = function(  $container, $interaction, $group )
	{
		if ( !$interaction || $interaction.length === 0 ) {
			return;
		}
		$interaction.detach();

		var int_id = $interaction.data( 'interaction' ).id;
		var interaction = inConcert.app.getInteractionManager().findInteraction( int_id );

		if ( $group ) {
			$group.before( $interaction );
			return;
		}
		else {
			$group = this.$get_group_interactions( $container, interaction );
			if ( $group && $group.length > 0 ) {
				$group.before( $interaction );
				return;
			}
		}
		$container.append( $interaction );
	};

	DashboardInteractionsView.prototype.remove_group_interactions = function( $container, interaction )
	{
		var $group = this.$get_group_interactions( $container, interaction );
		if ( !$group || $group.length === 0 ) {
		//  If the group does not exist on the DOM, return
			return;
		}
		var children = $group.children();
		if ( !children || children.length === 0 ) {
			// if there are no interactions in the group; remove group interactions of the DOM
			$group.remove();
			return;
		}

		if ( children.length === 1 ) {
		// if there is only one group interaction:
		//   1 detach group interaction,
		//   2 attach interaction in panel
		//   3 remove empty group interaction of the DOM
			var $last_interaction = $( children[ 0 ] );
			this.detach_interaction_of_group( $container, $last_interaction, $group );
			$group.remove();
			return;
		}
	};

	DashboardInteractionsView.prototype.sort_interaction_on_group = function( fi /* first interaction */, si /* second interaction */ ) {
		var date_compare = function() {
			return fi.currentTimeRaw() < si.currentTimeRaw();
		};

		if ( fi.type === si.type ) {
		// to the same type of interactions, the comparison is done by date
			return date_compare();
		}

		// types of interactions with greater weight, are located below
		// Ej: webcontact, go under every interaction type.
		var interactions_weight = {
			'WEBCONTACT': 100
			// "MAIL": 50   // sample
		};
		var fi_weight = interactions_weight[ fi.type ] || 0;
		var si_weight = interactions_weight[ si.type ] || 0;

		if ( fi_weight === si_weight ) {
		// compared by date interactions with the same weight
			return date_compare();
		}
		// aca ordeno segun el peso del tipo de interaccion
		return fi_weight > si_weight;
	};

	DashboardInteractionsView.prototype.$get_group_interactions = function( $container, interaction ) {
		if ( !$container || !interaction.getGroupId ) {
			return false;
		}
		var group_interaction_id = interaction.getGroupId();
		return $container.find( '[group_interaction_id="' + group_interaction_id + '"]' );
	};

	DashboardInteractionsView.prototype.__insert_interaction_in_group = function( interaction, $new_interaction, $container )
	{
		var self = this;
		if (!interaction) {
			return false;
		}
		var belongsToSomeGroup = interaction.belongsToSomeGroup();
		if ( !belongsToSomeGroup ) {
			return false;
		}

		var $group = this.$get_group_interactions( $container, interaction );

		var response = { new_group: false, $group: false, sort_group: false };
		var group_interaction_id = interaction.getGroupId();
		var $interactions_group = $container.find( '[belongs_to_group="' + group_interaction_id + '"]' );

		if ( $interactions_group.length === 0 ) {
		// si aun no existe ninguna interaccion en el grupo.
		// inserto la primer interaccion del grupo directo en el panel
			return false;
		}
		else if ( $interactions_group.length === 1 ) {
			$group = $( Mustache.render( this.m_templates.group_interactions_node_tpl, { group_interaction_id: group_interaction_id } ) );
			$group.data( 'interaction', interaction ); // data from the first interaction are used to sort the group node list
			response.new_group = true;

			$.each( $interactions_group, function( _, dom_it ) {
				var $it = $( dom_it );
				$it.detach();
				$it.attr( 'belongs_to_group', group_interaction_id );
				$group.append( $it );
			});
		}

		var inserted = false;
		var group_interactions = $group.children();
		$.each( group_interactions, function( i, group_interaction )
		{
			var $group_interaction = $( group_interaction );
			var current_interaction = $group_interaction.data('interaction');
			if ( self.sort_interaction_on_group( current_interaction, interaction ) ) {
				$group_interaction.before( $new_interaction );
				if ( i === 0 ) {
					$group.data('interaction', interaction); // data from the first interaction are used to sort the group node list
					response.sort_group = true;
				}
				inserted = true;
				return false; // break
			}
		});

		if ( !inserted ) {
			$group.append( $new_interaction );
		}

		response.$group = $group;
		return response;
	};

	DashboardInteractionsView.prototype.already_exist_interaction_in_panel_list = function(interaction, $container)
	{
		if (!interaction) {
			return false;
		}
		let dom_contanier_id = $container.attr('id');
		let dom_interacton_id = inConcert.Utils.getNormalizedId(this.__get_prefix() + interaction.id);
		const $already_exist = $container.find(`[id='${dom_interacton_id}']`);
		if ($already_exist && $already_exist.length > 0) {
			console.info("[DashboardInteractionsView.already_exist_interaction_in_panel_list] 1 ---> (ALREADY EXIST) interaction.id: " + interaction.id + ", dom_interacton_id: " + dom_interacton_id + ", dom_contanier_id: " + dom_contanier_id);
			return true;
		}
		//console.info("[DashboardInteractionsView.already_exist_interaction_in_panel_list] 2 ---> (NOT EXIST) interaction.id: " + interaction.id + ", dom_interacton_id: " + dom_interacton_id + ", dom_contanier_id: " + dom_contanier_id);
		return false;
	};

	// This function inserts the interaction or group interactions in the list panel (queued, archived, assigned)
	DashboardInteractionsView.prototype.insert_interaction_in_panel_list = function( interaction, $new_interaction, $container, order /*optional*/ )
	{
		const self = this;
		if (!interaction) {
			return;
		}
		let group_response = this.__insert_interaction_in_group( interaction, $new_interaction, $container );

		// funcion que agrega una interaccion en un orden determinado
		const append = function( $interaction )
		{
			const already_exist = self.already_exist_interaction_in_panel_list(interaction, $container);
			if (already_exist) {
				return false;
			}
			const $interactions = $container.children();
			if ( isNaN(order) || $interactions.length === 0 ) {
			// si no se ha definido un orden o es la primera interaccion de la lista
				$container.append( $interaction );
				return true;
			}
			else {
			// si se ha definido un orden o no es la primera interaccion de la lista,
			//  busco si en la posicion actual(indice) ya existe una interaccion,
			//  y si la misma existe, tengo que agregar la nueva interaccion en esa posicion,
			//  y si no existe una interaccion en esa posicion, debe de agregarla en la ultima posicion.
				const $current_interaction = $interactions.eq(order);
				if ($current_interaction.length > 0 ) {
					$current_interaction.before($interaction); 
				}
				else {
					$container.append( $interaction );
				}
				return true;
			}
		};

		if ( !group_response ) {
		// insert a simple new interaction in list panel
			append( $new_interaction );
		}
		else if ( group_response.new_group ) {
		// insert a new group interaction in list panel
			append( group_response.$group );
		}
		// TODO: reordenar grupo en la lista, con los datos de la primera interaccion.
		// group_response.sort_group
		this.on_inserted_interaction_in_panel_list( $new_interaction );
	};

	DashboardInteractionsView.prototype.on_inserted_interaction_in_panel_list = function( $interaction )
	{
		var self = this;
		// bind checkbox for batch actions
		var $batch_selector = $interaction.find('[int_batch_selector]');
		$batch_selector
			.unbind('click')
			.bind('click', function(e) {
				self.on_interaction_marked($interaction);
				// TODO: revisar tema de bulling. Se esta ejecutando y abriendo la ventana.
				e.stopPropagation();
				$("body").trigger('click');
			});
	};

	DashboardInteractionsView.prototype.is_marked_all_interactions = function( campaign_id )
	{
		var $all_interactions = this.get_interactions(null, campaign_id);
		if ( $all_interactions.length === 0 ) {
			return false;
		}
		var total_interactions = 0;
		$.each($all_interactions, function(_, interaction)
		{
			var $interaction = $(interaction);
			var $batch_selector = $interaction.find('[int_batch_selector]');
			// las interacciones mias, no tiene el selector (chk) visible y no las debo contar como marcadas
			if ($batch_selector.is(':visible')){
				total_interactions++;
			}
		});
		var $marked_interactions = this.$get_marked_interactions(campaign_id);
		return total_interactions === $marked_interactions.length;
	};

	DashboardInteractionsView.prototype.on_interaction_marked = function( $interaction )
	{
		var interaction = $interaction.data('interaction');
		var campaign_id = interaction.campaign;
		var $mark_all = this.$get_batch_actions_mark_all(campaign_id);
		var has_mark_all = false;
		// Selected to batch is true if the checkbox is selected.
		var marked_to_batch = $interaction.find('[int_batch_selector]').prop('checked') ? true : false;
		$interaction.removeClass('interactionSelectToBatch');
		if ( marked_to_batch ) {
			$interaction.addClass('interactionSelectToBatch');
			has_mark_all = this.is_marked_all_interactions(campaign_id);
		}
		$mark_all.prop('checked', has_mark_all);
		this.update_batch_buttons(campaign_id);
	};

	DashboardInteractionsView.prototype.__get_search_facet_values = function( campaign_id, type, callback = $.noop )
	{
		const self = this;

		let agents_workgroups_of_campaign = this.m_workgroups.get_agents_workgroups_of_campaign(campaign_id);

		/*
			Funcion que se encarga de agregar al array (lista de id de agentes) cada agente,
				tambien agrega en la lista ordenada por nombre completo.
		*/
		var add_agent = function(arr, agent)
		{
			let agent_id = agent.Id.toLowerCase();
			let must_include = false;
			// Filtra el campo agente NAME segun, los agentes de los workgroups a los cuales pertenece el supervisor.
			if (!agents_workgroups_of_campaign || agents_workgroups_of_campaign[agent_id]) {
				must_include = true;
			}

			if (!must_include) {
				return;
			}

			var user_name_params = {
				include_user_name: true,
				include_user_last_name: true,
				include_user_first_surname: true,
				include_user_last_surname: true,
				include_user_id: true
			};

			var complete_user_name = inConcert.Utils.formatUserName(agent, user_name_params);
			arr.push(complete_user_name);
			self.agents_by_name[complete_user_name.toLowerCase()] = agent_id;
		};

		// Filtra el campo agente ID segun, los agentes de los workgroups a los cuales pertenece el supervisor.
		var callback_agents = function (agents_ids, options = {}) {
			let filtered_agents = [];
			for (let i = 0; i < (agents_ids || []).length; i++) {
				let agent_id = agents_ids[i].toLowerCase();
				if (!agents_workgroups_of_campaign || agents_workgroups_of_campaign[agent_id]) {
					filtered_agents.push(agent_id);
				}
			}
			callback(filtered_agents, options);
		};

		var facets_values = {
			agent : function()
			{
				var campaign = inConcert.app.getCampaign(campaign_id);
				if (campaign) {
					var campaign_config = self.m_dashboard.get_campaign_selected_config(campaign_id);
					if ( campaign_config.account_id && campaign_config.al_id ) {
					// 1 - si en el selector de campanas ha seleccionado algun AL debe de visualizar los users de ese AL
						if ( campaign.has_attention_levels_configured_in_account(campaign_config.account_id) ) {
							var al_users = campaign.getUsersInAttentionLevel(campaign_config.al_id);
							var users_ids = [];
							$.each(al_users, function( user_id ) {
								users_ids.push(user_id);
							});
							// Push current member Id.
							users_ids.push(inConcert.app.getLoggedUserId());
							callback_agents(users_ids);
							return;
						}
					}

					// 2 - si en el selector de campanas no ha seleccionado "attention levels", trae todos los members de la campana
					campaign.getTeamMembers(function( members )
						{
							if (members.length === 0){
								callback_agents([]);
							}
							// Extract member's ids.
							var members_ids = _.pluck(members, 'Id');
							// Push current member Id.
							members_ids.push(inConcert.app.getLoggedUserId());
							callback_agents(members_ids);
						},
						true // from cache
					);
				}
			},

			agent_name : function()
			{
				var campaign = inConcert.app.getCampaign(campaign_id);
				if (campaign) {
					var campaign_config = self.m_dashboard.get_campaign_selected_config(campaign_id);
					if ( campaign_config.account_id && campaign_config.al_id ) {
					// 1 - si en el selector de campanas ha seleccionado algun AL debe de visualizar los users de ese AL
						if ( campaign.has_attention_levels_configured_in_account(campaign_config.account_id) )
						{
							campaign.getTeamMembers(function( members )
								{
									if (!members || members.length === 0){
										callback([]);
										return;
									}

									var agents_by_id = {};
									for (var i = 0; i < members.length; i++) {
										var agent = members[i];
										add_agent([], agent);
										agents_by_id[agent.Id] = agent;
									}

									var al_users = campaign.getUsersInAttentionLevel(campaign_config.al_id);
									var agents_names = [];
									$.each(al_users, function( user_id ) {
										var agent = agents_by_id[user_id];
										add_agent(agents_names, agent);
									});

									// Push current member Id.
									var logged_user_info = inConcert.app.getLoggedUserInfo();
									add_agent(agents_names, logged_user_info);

									var sorted_agents = inConcert.Utils.sortArray(agents_names);
									callback(sorted_agents, {preserveOrder: true});
								},
								true // from cache
							);

							return;
						}
					}

					// 2 - si en el selector de campanas no ha seleccionado "attention levels", trae todos los members de la campana
					campaign.getTeamMembers(function( members )
						{
							var agents_names = [];
							if (!members || members.length === 0){
								callback(agents_names);
								return;
							}

							for (var i = 0; i < members.length; i++) {
								var agent = members[i];
								add_agent(agents_names, agent);
							}

							// Push current member Id.
							var logged_user_info = inConcert.app.getLoggedUserInfo();
							add_agent(agents_names, logged_user_info);

							var sorted_agents = inConcert.Utils.sortArray(agents_names);
							callback(sorted_agents, {preserveOrder: true});
						},
						true // from cache
					);
				}
			},

			attention_level : function()
			{
				var displayed_attention_levels = [];
				var campaign_config = self.m_dashboard.get_campaign_selected_config(campaign_id);

				if ( campaign_config.account_id && !campaign_config.al_id ) {
					var campaign = inConcert.app.getCampaign(campaign_id);

					if ( campaign && campaign.has_attention_levels_configured_in_account(campaign_config.account_id)  ) {
						var account = campaign.getAccount(campaign_config.account_id);

						if ( account ) {
							$.each( account.AttentionLevels || {}, function(_, assigned_al) {
								var current_al = campaign.getAttentionLevel(assigned_al.attention_level_id);
								if ( current_al ) {
									displayed_attention_levels.push(current_al.name);
								}
							});
						}
					}
				}

				callback( displayed_attention_levels );
			},

			type: function() {
				var displayed_channels = [];
				$.each( ['CALL', 'CHAT', 'FACEBOOK', 'INSTAGRAM', 'YOUTUBE', 'APPSTORE', 'GOOGLEPLAY', 'IM', 'MAIL', 'SMS', 'RCS', 'WHATSAPP', 'TWITTER', 'WEBCONTACT', 'VOICEMAIL'],
					function( _, channel ) {
						if ( inConcert.app.hasLicensesForChannel( channel ) ) {
							displayed_channels.push( channel );
						}
					});
				callback( displayed_channels );
			},

			workgroup: function() {
				callback( self.m_workgroups.get_campaign_workgroups(campaign_id) );
			}
		};

		if ( facets_values[type] ) {
			facets_values[type]();
			return;
		}

		callback();
	};

	DashboardInteractionsView.prototype.hideMenuActions = function() {
		if (this.m_interaction_gui) {
			this.m_interaction_gui.hide_menu_actions();
		}
	};

	DashboardInteractionsView.prototype.showOrHideColumns = function(campaign_id)
	{
		const campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		const $dynamic = $('#sup_dash_interaction_viewer_dynamic_' + campaign_dom_id);
		$dynamic.find("[header_column_name='attention_time']").hide();
	};

	DashboardInteractionsView.prototype.allowed_to_refresh_view = function(/*campaign_id*/)
	{
		return true;
	};

	DashboardInteractionsView.prototype.setup_view_pagination = function($current_node, campaign_id)
	{
		const self = this;

		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		// call the parent class (base conversations viewer)
		this.setup_pagination(campaign_id, function(button, on_pagination_finished)
		{
			let currentViewPagination = self.get_view_paged_info();
			currentViewPagination[campaign_id].paginating = true;
			// acording to the button clicked, set correct page.
			switch (button.id)
			{
				case 'interactions_page_first_' + campaign_dom_id:
					currentViewPagination[campaign_id].filter.page = 1;
					break;
				case 'interactions_page_next_' + campaign_dom_id:
					currentViewPagination[campaign_id].filter.page++;
					break;
				case 'interactions_page_prev_' + campaign_dom_id:
					currentViewPagination[campaign_id].filter.page--;
					break;
				case 'interactions_page_last_' + campaign_dom_id:
					let page;
					if (currentViewPagination[campaign_id].totalRows > self.MAX_ALLOWED_ROWS_IN_ES) {
						// Si ha llegado al maximo de rows permitidos por ES, me quedo con la pagina previa
						page = self.MAX_ALLOWED_ROWS_IN_ES / self.INTERACTIONS_PER_PAGE;
						page = Math.floor(page) < page ? Math.floor(page) + 1 : page;
					}
					else {
						page = currentViewPagination[campaign_id].last_page;
					}
					currentViewPagination[campaign_id].filter.page = page;
					break;
				default:
					break;
			}
			// go and search for conversations
			self.refresh($current_node, on_pagination_finished);
		});
	};

	DashboardInteractionsView.prototype.update_current_pagination = function(campaign_id, current_page, last_page, page_interactions, total_interactions)
	{
		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);

		if ( page_interactions === 0 && total_interactions > 0 ) {
			$('#interactions_page_prev_' + campaign_dom_id).trigger('click_navigate_to');
			return;
		}

		let previous_page = ( current_page === 0 || isNaN( current_page ) ) ? 0 : current_page - 1;

		// calculate start member counter.
		let start = null;
		if (total_interactions === 0){
			start = 0;
		}
		else if (current_page === 1){
			start = 1;
		}
		else{
			start = ( this.INTERACTIONS_PER_PAGE * previous_page ) + 1;
		}

		let finish = null;
		if (current_page === last_page || page_interactions === total_interactions){
			finish = total_interactions;
		}
		else {
			// just calculate given the current page number the last member
			finish = page_interactions * current_page;
		}

		let previous_button_disabled = current_page === 1;
		let next_button_disabled = false;
		let pager_view_hidden = false;
		if (total_interactions === 0 || total_interactions <= this.INTERACTIONS_PER_PAGE) {
			// Si no tiene registros o el total de registros es menor o igual a la cantidad de registros por pagina, desactivo la navegacion para obtener mas Rows.
			next_button_disabled = true;
			previous_button_disabled = true;
		}
		else if ((current_page * this.INTERACTIONS_PER_PAGE) >= this.MAX_ALLOWED_ROWS_IN_ES) {
			// Si llego al maximo de Rows permitidas desactivo la navegacion para obtener mas Rows.
			next_button_disabled = true;
		}
		else if (current_page === last_page) {
			// Si estoy en la ultima pagina desactivo la navegacion para obtener mas Rows.
			next_button_disabled = true;
		}
		
		if (next_button_disabled === true && previous_button_disabled === true) {
			pager_view_hidden = true;
		}

		// update counters and button status
		let pagination_text = format("<%=_T('{0} - {1} of {2} conversations')%>", start, finish, total_interactions);
		$('#interactions_page_text_' + campaign_dom_id).text(pagination_text);
		$('#interactions_page_prev_' + campaign_dom_id).prop('disabled', previous_button_disabled);
		$('#interactions_page_first_' + campaign_dom_id).prop('disabled', previous_button_disabled);
		$('#interactions_page_next_' + campaign_dom_id).prop('disabled', next_button_disabled);
		$('#interactions_page_last_' + campaign_dom_id).prop('disabled', next_button_disabled);
		$('#interactions_list_pager_' + campaign_dom_id).prop('hidden', pager_view_hidden);
	};
}());