/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
  strict:true, undef:true, unused:true, curly:true, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global $, inConcert, Mustache */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name SupervisorDashboardSummary
 */
(function () {

	'use strict';

	/**
	 * Creates a new summary dashboard instance.
	 *
	 * @author Sebastián Moreno (smoreno@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorDashboardSummary
	 * @param dashboard_instance
	 */
	function DashboardSummary(dashboard_instance)
	{
		var self = this;
		this.m_id = 'summary';

		this.m_templates = inConcert.sup_templates.dashboard;
		this.m_dashboard = dashboard_instance ? dashboard_instance : null;
		this.m_config = dashboard_instance.m_config;

		// Hash to know if i'm editing
		this.m_in_edit_mode_by_campaign = {};

		// message bus
		inConcert.SupervisorMessageBus.subscribe('/supervisor/dashboard/' + this.m_id, function(notification) {
			self.__on_panel_visibility_changed(notification.campaign_id, notification.visible);
		});
	}

	if (typeof inConcert.sup_dashboard_summary !== 'function') {
		inConcert.sup_dashboard_summary = DashboardSummary;
	}

	DashboardSummary.prototype.on_change_config_selector = function( /*campaign_id*/ ) {};

	DashboardSummary.prototype.__get_panel_node = function(campaign_id){
		return $('#sup_dash_summary_panel_' + this.m_dashboard.get_normalized_id(campaign_id));
	};

	DashboardSummary.prototype.__get_panel_container = function(campaign_id){
		return this.__get_panel_node(campaign_id).find('.columnContent');
	};

	/**
	 * Return true if the user is editing the column, otherwise, false.
	 * 
	 * @param  {String} campaign_id - The campaign id.
	 * @return {Boolean}            - True if the user is editing, otherwise, false.
	 */
	DashboardSummary.prototype.__user_inside_edit_mode = function(campaign_id){
		return this.m_in_edit_mode_by_campaign[campaign_id] || false;
	};

	/**
	 * If I'm in edit mode, do not refresh the view.
	 * 
	 * @param  {String} campaign_id - The campaign id.
	 * @return {Boolean}            - True if the user can refresh the view, otherwise, false.
	 */
	DashboardSummary.prototype.__allowed_to_refresh_view = function (campaign_id){
		return !this.__user_inside_edit_mode(campaign_id);
	};

	DashboardSummary.prototype.__on_column_changed = function (campaign_id) {
		var container = this.__get_panel_container(campaign_id);
		var groups = container.children();
		var groups_checkboxes = groups.find('.summaryGroupCheck');

		var total_groups = groups.length;
		var total_groups_hidden = groups_checkboxes.not(':checked').length;

		if (total_groups === total_groups_hidden){
			// show groups
			groups.show();
			groups_checkboxes.prop('checked', true);

			// show indicators
			groups.children().show();
			groups.children().find('.summaryRowCheck').prop('checked', true);
		}
	};

	/**
	 * Called when the users leaves the summary edit mode. Will setup the panel view
	 * and save the configuration
	 *
	 * @private
	 * @param  {String} campaign_id - The campaign identificator. (eg. campaign@vcc)
	 */
	DashboardSummary.prototype.__leave_edit_mode = function(campaign_id)
	{
		// toggle edit mode
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		$('#sup_dash_column_edit_summary_' + campaign_dom_id)
			.addClass('butColumnEdit')
			.removeClass('butColumnDone');

		var $container = this.__get_panel_container(campaign_id);

		// if user uncheck some indicator, hide it.
		let $summaryGroupCheck = $container.find('[summaryGroupCheck="true"]');
		let countUncheckGroups = 0;
		$.each($summaryGroupCheck, function(_, groupCheck) {
			let $sumaryFieldSet = $(groupCheck).parent().parent();
			let indictors = $sumaryFieldSet.find('[indicatorRowCheck="true"]');
			let indictors_checkeds = indictors.filter((_, item) => {
				return $(item).prop('checked');
			});
			if (indictors_checkeds.length === 0) {
				$sumaryFieldSet.hide();
				countUncheckGroups++;
			}
		});
		$container.find('.summaryRowCheck').not(':checked').parent().parent().hide();

		// remove edit mode
		$container.removeClass('editMode');

		// If all groups are unchecked, hide column.
		if ( countUncheckGroups ===  $summaryGroupCheck.length ){
			// TO-DO: Change this to PubSub
			var success = this.m_dashboard.trigger_column_update(campaign_id, 'summary', false);
			if ( success ) {
				this.__on_column_changed(campaign_id);
			}
		}

		// hide slice combo
		$container.find('.sliceConfiguration').hide();

		// Save configuration.
		// TO-DO: Double check this, this is an async call and should have a callback.
		this.m_dashboard.save_config();

		// set member for edit mode
		this.m_in_edit_mode_by_campaign[campaign_id] = false;
	};

	/**
	 * Called when the user enters in panel edit mode, to add/remove groups and indicators.
	 *
	 * @private
	 * @param  {String} campaign_id - The campaign identificator. (eg. campaign@vcc)
	 */
	DashboardSummary.prototype.__enter_edit_mode = function(campaign_id)
	{
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);

		$('#sup_dash_column_edit_summary_' + campaign_dom_id)
			.removeClass('butColumnEdit')
			.addClass('butColumnDone');

		var $container = this.__get_panel_container(campaign_id);
		$container.addClass('editMode');

		// show summary groups with license
		var proxies = { CALL: true, CHAT: true, FACEBOOK: true, INSTAGRAM: true, YOUTUBE: true, APPSTORE: true, GOOGLEPLAY: true, RCS: true, IM: true, MAIL: true, SMS: true, WHATSAPP: true, TWITTER: true, WEBCONTACT: true, VOICEMAIL: true };
		var $group_nodes = $container.children();

		$.each($group_nodes, function( _, group_node )
		{
			var $group_node = $( group_node );
			var group_id = ( $group_node.attr( 'simple_group_id') || '' ).toUpperCase();
			if ( !proxies[ group_id ] || inConcert.app.hasLicensesForChannel( group_id ) ) {
				$group_node.show();
			}
			else {
				$group_node.hide();
			}
		});

		$container.find('.sliceConfiguration').show(); // show time ago
		$container.find('.summaryFieldsetRow').show(); // show indicators

		// set member for edit mode
		this.m_in_edit_mode_by_campaign[campaign_id] = true;
	};

	DashboardSummary.prototype.__setup_edit_mode = function (campaign_id) {
		var self = this;
		var campaign_dom_id = self.m_dashboard.get_normalized_id(campaign_id);
		var container = self.__get_panel_container(campaign_id);

		// all campaigns does no start in edit mode.
		self.m_in_edit_mode_by_campaign[campaign_id] = false;

		var update_last_visible_child = function(){
			// add last visible child.
			container.find('.summaryFieldsetRow').removeClass('last');
			container.find('.summaryFieldset')
				.each(function(i,e){
					$(e).find('.summaryFieldsetRow:visible:last').addClass('last');
				});
		};

		// add 'last' class to last visible child row,
		// this is done to remove the border via css.
		update_last_visible_child();

		// edit button
		$('#sup_dash_column_edit_summary_' + campaign_dom_id)
			.unbind('click')
			.bind('click', function(){
				// *done* behaviour -> if the user was already in edit mode.
				if (container.hasClass('editMode')){
					self.__leave_edit_mode(campaign_id);
				}
				else {
					self.__enter_edit_mode(campaign_id);
				}
				update_last_visible_child();
			});

		// group checkbox
		container.find('.summaryGroupCheck')
			.unbind('change').bind('change', function( /*event*/ ){
				// this is kind of a hack, this change could be triggered
				// clicking the label (because the label is associated with the checkbox id (for attr))
				// and the markup it's reusing the same things from edit and normal mode.
				// so we make sure if the user is inside normal mode, we change back the checkbox state
				if (!self.__user_inside_edit_mode(campaign_id)){
					$(this).prop('checked', !this.checked);
				}
				$(this.parentNode).siblings().find(':checkbox').prop('checked', this.checked);
			});

		// indicator checkbox
		container.find('.summaryRowCheck')
			.unbind('change')
			.bind('change', function(){
				if (self.__user_inside_edit_mode(campaign_id)){
					var summary_group = $(this).parent().parent().parent();
					const totalRowCheck = summary_group.find('.summaryRowCheck').length;
					var total_checked = summary_group.find('.summaryRowCheck:checked').length;
					summary_group.find('[summaryGroupCheck="true"]').prop('checked', total_checked === totalRowCheck);
				}
		});
	};

	/**
	 * Will setup the handlers to toggle each summary group
	 *
	 * @private 
	 * @param {object} current_node -  Current campaign node.
	 * @param {String} campaign_id
	 */
	DashboardSummary.prototype.__setup_group_collapse_handlers = function(current_node, campaign_id){
		var self = this;

		current_node.find('.summaryFieldsetTitleBefore')
			.unbind('click')
			.bind('click', function(){
				if (!self.__user_inside_edit_mode(campaign_id)){
					$(this).toggleClass('collapsed').parent().find('.summaryFieldsetRow').toggle();
				}
			});

		current_node.find('.fieldsetTilteText')
			.unbind('click')
			.bind('click', function(){
				if (!self.__user_inside_edit_mode(campaign_id)){
					var title = $(this);
					title.parent().toggleClass('collapsed');
					title.parent().parent().find('.summaryFieldsetRow').toggle();
				}
			});
	};

	/**
	 * Triggered when the visibility of the panel is changed, it's a callback
	 * for a PubSub operation.
	 *
	 * @private
	 * @param  {String} campaign_id - The campaign identificator
	 * @param  {Boolean} visible    - Panel is visible or hidden
	 */
	DashboardSummary.prototype.__on_panel_visibility_changed = function(campaign_id, visible) {
		if ( !visible ){
			return;
		}
		if ( !this.__get_panel_node(campaign_id).find('.summaryGroupCheck').is(':checked') ){
			this.__enter_edit_mode(campaign_id);
		}
	};

	/**
	 * Returns true if summary group is collapsed based on configuration
	 *
	 * @private
	 * @param  {Object} config   - Current panel configuration
	 * @param  {String} group_id - Group id to be checked
	 * @return {Boolean}         - True if should be visible, otherwise false.
	 */
	DashboardSummary.prototype.__group_is_collapsed = function(config, group_id){
		if ( config && config.groups && config.groups[group_id] && config.groups[group_id].hasOwnProperty('collapsed')) {
			return config.groups[group_id].collapsed;
		}
		// not collapsed is default if not set.
		return false;
	};

	/**
	 * Returns true if summary group should be created based on user configuration
	 *
	 * @private
	 * @param  {Object} config   - Current panel configuration
	 * @param  {String} group_id - Group id to be checked
	 * @return {Boolean}         - True if should be visible, otherwise false.
	 */
	DashboardSummary.prototype.__group_should_be_visible = function (config, group_id){
		if (config && config.groups && config.groups[group_id]){
			return config.groups[group_id].visible;
		}
		// create by default
		return true;
	};

	/**
	 * Returns true if summary indicator should be created based on user configuration
	 *
	 * @private
	 * @param  {Object} config        - Current panel configuration
	 * @param  {String} group_id      - Group id to be checked
	 * @param  {String} indicator_id  - The indicator id to be checked.
	 * @return {Boolean}              - True if should be visible, otherwise false.
	 */
	DashboardSummary.prototype.__indicator_should_be_visible = function (config, group_id, indicator_id){
		// if group config exists
		if ( config && config.groups && config.groups[group_id] ) {

			// otherwise, return the current indicator visibility
			if (config.groups[group_id].indicators.hasOwnProperty(indicator_id)){
				return config.groups[group_id].indicators[indicator_id];
			}
		}
		// create by default
		return true;
	};

	/**
	 * Will create or update if already exists a summary group.
	 *
	 * @private
	 * 
	 * @param  {String} campaign_id                - The campaign identificator.
	 * 
	 * @param  {Object} group                      - A summary group
	 * @param  {String} group.ID                   - The summary group identificator.
	 * @param  {String} group.description          - The summary group description
	 * 
	 * @param  {Object} container                  - A jQuery container to add this group.
	 * @param  {Object} panel_config               - Current panel configuration
	 * 
	 * @return {Object} new_summary_group          - A new/updated summary group
	 * @return {String} new_summary_group.dom_id   - The DOM Id for the new/updated summary group.
	 * @return {Object} new_summary_group.node     - The jQuery Node id for the new/updated summary group.
	 */
	DashboardSummary.prototype.__create_or_update_summary_group_node = function(campaign_id, group, container, panel_config)
	{
		var show_account_summary_group = function(account, groupId)
		{
			let account_type = account && account.proxyType.toLowerCase();
			return !account || account_type === groupId || account_type === "call" && (groupId.indexOf("call") >= 0 || groupId === "voicemail");
		};

		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var summary_group_id = group.ID;

		// Create Summary Group node
		var group_complete_id = campaign_dom_id + '_' + summary_group_id;
		var group_dom_id = 'sup_dash_group_summary_' + group_complete_id;
		var $group_node = container.find('#' + group_dom_id);

		var campaign_config = this.m_dashboard.get_campaign_selected_config(campaign_id);
		var group_totals_name = group.description || summary_group_id;

		// el titulo del grupo total (campaign/account), cambia a ACCOUNTS TOTALS, si es que en "campaign_selector" se ha seleccionado una cuenta/AL
		if ( group.type === 'entity' && campaign_config.account_id ) {
			group_totals_name = "<%=_T('ACCOUNT TOTALS')%>";
		}

		const is_checked_group = function(config, group_id){
			if (config && config.groups && config.groups[group_id] && config.groups[group_id].indicators){
				return Object.values(config.groups[group_id].indicators).every((item) => {
					return item;
				});
			}
			// create by default
			return true;
		};

		// if group node exists, update info.
		if ( $group_node.length > 0 ) {
			$group_node.find('.fieldsetTilteText').text(group_totals_name);
		} 
		else {
			// create and append complete markup
			var group_view = { 
				full_group_id   : group_complete_id,
				simple_group_id : summary_group_id ,
				group_name      : group_totals_name,
				checked         : is_checked_group(panel_config, group_dom_id)
			};
			$group_node = $(Mustache.render(this.m_templates.group_summary_tpl, group_view));
			container.append($group_node);
		}

		// Si en campaign_selector se ha seleccionado una cuenta/AL, se debe de visualizar unicamente el panel de ese channel.
		var hidden_by_channel = false;
		if ( group.type === 'channel' && campaign_config.account_id ) {
			var campaign = inConcert.app.getCampaign(campaign_id);
			if ( campaign ) {
				var account = campaign.getAccount(campaign_config.account_id);
				hidden_by_channel = !show_account_summary_group(account, summary_group_id);
			}
			else {
				hidden_by_channel = true;
			}
		}

		if ( this.__user_inside_edit_mode(campaign_id) ) {
			// si no tiene licencia para el canal, lo oculta
			var proxies = { CALL: true, CHAT: true, FACEBOOK: true, INSTAGRAM: true, YOUTUBE: true, APPSTORE: true, GOOGLEPLAY: true, RCS: true, IM: true, MAIL: true, SMS: true, WHATSAPP: true, TWITTER: true, WEBCONTACT: true, VOICEMAIL: true };
			if ( !hidden_by_channel && (!proxies[ summary_group_id ] || inConcert.app.hasLicensesForChannel( summary_group_id )) ) {
				$group_node.show();
			}
			else {
				$group_node.hide();
			}
		}
		else {
			if ( !hidden_by_channel ) {
				$group_node.show();
			}
			else {
				$group_node.hide();
			}
		}

		return {
			dom_id : group_dom_id,
			node : $group_node
		};
	};

	DashboardSummary.prototype.__show_alarm = function ( campaign_id, $node, indicator_id )
	{
		// (ALARMS) Marco/desmarco el indicador como alarmado.
		var campaign_config = this.m_dashboard.get_campaign_selected_config(campaign_id);
		var alarm_config = this.m_dashboard.get_notification_alarm_by_indicator({
			indicator_id : indicator_id,
			campaign_id  : campaign_id,
			account_id   : campaign_config.account_id,
			al_id        : campaign_config.al_id
		});
		return this.m_dashboard.show_alarm($node, alarm_config);
	};


	/**
	 * Will create or update if already exists an indicator node.
	 *
	 * @private
	 * 
	 * @param  {String} campaign_id                - The campaign identificator.
	 * 
	 * @param  {Object} indicator                  - An indicator node
	 * @param  {String} indicator.ID               - The indicator group identificator.
	 * @param  {String} indicator.description      - The indicator group description
	 * @param  {String} indicator.total            - The total of the indicator to be shown.
	 * @param  {String} indicator.ofWorkgroups     - The total by workgroup of the indicator to be shown.
	 * 
	 * @param  {Object} $group_node                 - A jQuery group container to add this indicator.
	 * 
	 * @return {Object} new_indicator              - A new/updated indicator
	 * @return {String} new_indicator.dom_id       - The DOM Id for the new/updated indicator.
	 * @return {Object} new_indicator.node         - The jQuery Node id for the new/updated indicator.
	 */
	DashboardSummary.prototype.__create_or_update_summary_indicator_node = function(campaign_id, indicator, $group_node)
	{
		var campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		var indicator_complete_id = campaign_dom_id + '_' + indicator.ID;
		var indicator_dom_id = 'sup_dash_summary_indicator_' + indicator_complete_id;
		var $indicator = $group_node.find('#' + indicator_dom_id);

		// workgroup values have precedence over normal values
		var row_value =  indicator.ofWorkgroups ? indicator.ofWorkgroups : indicator.total;

		// check if indicator node already exists.
		if ($indicator.length > 0){
			$indicator.find('.summaryNumberTotal').text(row_value);
		} 
		else {
			var view = {
				id       : indicator_complete_id,
				row_name : indicator.description,
				row_value: row_value
			};
			$indicator = $(Mustache.render(this.m_templates.row_summary_tpl, view));
			$group_node.append($indicator);
		}

		// muestra la alarma en el indicador (si corresponde visualizarla)
		this.__show_alarm(campaign_id, $indicator, indicator.ID);

		return {
			dom_id : indicator_dom_id,
			node : $indicator
		};
	};

	//
	//	Indicator Example:
	//
	//	[
	//		{
	//			ID : "agents",
	//			description : "Agents",
	//			indicators : [
	//				{
	//					ID : "abandoned_percentage",
	//					description : "Abandoned Percentage",
	//					total : "00:00:05"
	//				},
	//				{
	//					ID : "agents_active",
	//					description : "Active Agents",
	//					total : 2,
	//					ofWorkgroups : 1
	//				}
	//			]
	//		}
	//	]
	DashboardSummary.prototype.__setup_summary = function (campaign_id, summary)
	{
		// there is a odd case when summary is undefined (get supervisor view request),
		// immediately we protect us from this wrong behaviour.
		if ( !summary ){
			return;
		}

		// console.log("DashboardSummary.prototype.__setup_summary --> campaign: '" + campaign_id + "', summary info: " + JSON.stringify(summary));

		var self = this;
		var config = self.m_config.get_panel_config(campaign_id, 'sup_dash_summary_panel');

		// set realtime window combo value
		var realtime_window = config && config.realtime_window ? config.realtime_window : 'Last15Minutes';
		let campaign_dom_id = this.m_dashboard.get_normalized_id(campaign_id);
		const $campaign_summary_time_ago = $('#sup_dash_summary_time_ago_select_' + campaign_dom_id);
		const $value = $campaign_summary_time_ago.find('[value="' + realtime_window + '"]');
		$value.attr('selected', true);

		var oldContainer = self.__get_panel_container(campaign_id);
		var parent = oldContainer.parent();
		var top = parent.scrollTop();
		var $container = oldContainer.detach();
		$container.find('[simple_group_id]').hide();

		let total_visible_summary_panels = summary.length;
		summary.forEach(function(raw_group)
		{
			// create/update summary group
			var processed_group = self.__create_or_update_summary_group_node(campaign_id, raw_group, $container, config);

			// hide node if visible is set to false in configuration.
			if ( !self.__group_should_be_visible(config, processed_group.dom_id )){
				processed_group.node.hide();
				total_visible_summary_panels--;
			}

			// create markup for each indicator, if indicator exists, just update info.
			raw_group.indicators.forEach(function (indicator) {
				var processed_indicator = self.__create_or_update_summary_indicator_node(campaign_id, indicator, processed_group.node, config);
				// hide node if 'visible' property is set to false in configuration.
				if ( !self.__indicator_should_be_visible(config, processed_group.dom_id, processed_indicator.dom_id )){
					processed_indicator.node.hide().find(':checkbox').prop('checked', false);
				}
			});

			// TO-DO: if group is collapsed, hide indicators
		});

		/*
			Aca agrego una clase al contendor de paneles, para que sepa la cantidad de paneles que contiene.
			Ej:
				Si NO tiene paneles se agrega la clase "withoutPanels".
				Si tiene 1 panel se agrega la clase "singlePanels".
				Si tiene 2 paneles se agrega la clase "twoPanels".
				Si tiene mas de 2 paneles se agrega la clase "manyPanels".
		*/
		// 1 - Remuevo las clases de este tipo que se han agregado previamente.
		// Ej: withoutPanels singlePanels twoPanels manyPanels
		//
		const top_panels = 2;
		const panels_prefix = {
			"0": "withoutPanels",
			"1": "singlePanel",
			"2": "twoPanels"
		};
		let classes_to_remove = "";
		for (let i = 0; i <= top_panels; i++) {
			if (classes_to_remove.length > 0) {
				classes_to_remove = classes_to_remove + " ";
			}
			let current_class = panels_prefix[i.toString()];
			if (current_class) {
				classes_to_remove = classes_to_remove + current_class;
			}
		}
		const many_class = "manyPanels";
		classes_to_remove = classes_to_remove + many_class;
		//console.debug("--> supervisor remove panel clases: '" + classes_to_remove + "', in campaign: " + campaign_id);
		$container.removeClass(classes_to_remove);

		// 2 - Agrego la clase segun la cantidad de paneles.
		// Ej: withoutPanels, o sino singlePanels, o sino twoPanels, o en el caso que supere los paneles agrega manyPanels
		//
		let class_total_summary_panels;
		if (total_visible_summary_panels > top_panels) {
			class_total_summary_panels = many_class;
		}
		else {
			let current_class = panels_prefix[total_visible_summary_panels.toString()];
			if (current_class) {
				class_total_summary_panels = current_class;
			}
		}
		//console.debug("--> supervisor add panel clases: '" + class_total_summary_panels + "', in campaign: " + campaign_id);
		$container.addClass(class_total_summary_panels);

		parent.append($container);
		parent.scrollTop(top);
	};

	/**
	 * Will refresh the panel view.
	 *
	 * @public
	 * @param  {Object} $current_node - Current jQuery campaign node.
	 * @param  {Object} summary       - The supervisor view summary, this comes from the server.
	 */
	DashboardSummary.prototype.refresh = function ($current_node, summary)
	{
		let campaign_id = $current_node.attr('campaign_id');
		if (this.__allowed_to_refresh_view(campaign_id)){
			this.__setup_summary(campaign_id, summary);
		}
		// actualizo la informacion en modo full screen
		this.__full_screen_update({
			campaign_id: campaign_id,
			data: summary
		});
	};

	/**
	 * Will setup or refresh summary column for a given campaign node.
	 * 
	 * @param  {Object} $current_node - Current jQuery campaign node.
	 * @param  {Object} summary       - The supervisor view summary, this comes from the server.
	 */
	DashboardSummary.prototype.setup = function ($current_node, summary)
	{
		var campaign_id = $current_node.attr('campaign_id');
		this.__setup_summary(campaign_id, summary);
		this.__setup_group_collapse_handlers($current_node, campaign_id);
		this.__setup_edit_mode(campaign_id);
		// actualizo la informacion en modo full screen
		this.__full_screen_update({
			campaign_id: campaign_id,
			data: summary
		});
	};

	DashboardSummary.prototype.__full_screen_update = function (info)
	{
		return this.m_dashboard.full_screen_update_by_summary(info);
	};
}());
