/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:true, browser:true, indent:false, maxerr:50, quotmark:single, esversion: 11 */

/* global inConcert*/

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name SupervisorModel
 */
(function() {

	'use strict';

	/**
	 * Creates a new supervisor model instance.
	 *
	 * @author Sebastián Moreno (smoreno@inconcertcc.com)
	 * @constructor
	 * @memberOf SupervisorModel
	 */
	function SupervisorModel(dashboard_instance)
	{
		var base_url = inConcert.app.getBaseDomain();
		this.m_dashboard = dashboard_instance;
		this.m_url_base = base_url + '/inconcert/apps/supervisor';
		this.m_url_agent = base_url + '/inconcert/apps/agent';
		this.m_current_members_by_campaign = {};
		this.m_cached_system_indicators = null;
		this.m_all_users_with_access_to_interaction_app = null;
		this.m_loading_all_users = false;
		this.m_pubsub_new_user_handle = null;
	}

	if (typeof inConcert.sup_model !== 'function') {
		inConcert.sup_model = SupervisorModel;
	}

	/*
		Members Methods
	*/
	SupervisorModel.prototype.change_member_state = function( memberId, newState, campaign_id, callback )
	{
		inConcert.logger.debug({action_name : 'change_member_state', member : memberId, new_state: newState, action_type : 'members', app: 'supervisor'});
		callback = callback || $.noop;

		var request = inConcert.invoke({
			url: this.m_url_base + '/members/change_state/',
			type: 'POST',
			data: JSON.stringify( { memberId : memberId, state : newState, campaignId : campaign_id }),
			dataType: 'json',
			contentType: 'application/json',
			success: function(response) {
				callback(true, response);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};


	SupervisorModel.prototype.logout_member = function(memberId, campaign_id, supervisorId, callback)
	{
		inConcert.logger.debug({action_name : 'logout_member', member : memberId, action_type : 'members', app: 'supervisor'});
		callback = callback || $.noop;

		var url = this.m_url_base + '/members/logout/';
		var timestamp = ((new Date().getTime()) / 1000).toString();
		var kickoutData = {
			Message : `<%=_T('You have been logged out by supervisor')%>`,
			UserID : supervisorId,
			Forced : true,
			timestamp : timestamp
		};

		var request = inConcert.invoke({
			url: url,
			type: 'POST',
			data: JSON.stringify({ memberId : memberId, campaignId : campaign_id, kickout : kickoutData }),
			dataType: 'json',
			contentType: 'application/json',
			success: function() {
				callback(true);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};

	SupervisorModel.prototype.broadcast_message = function( message, users_ids, campaign_id, callback )
	{
		inConcert.logger.debug({action_name : 'broadcast_message', action_type:'members', app: 'supervisor' });
		callback = callback || $.noop;

		var request = inConcert.invoke({
			url: this.m_url_base + '/poller/broadcast_message/',
			type: 'POST',
			contentType: 'application/json',
			dataType: 'json',
			data: JSON.stringify({ campaignId: campaign_id, message : message, users : users_ids }),
			success: function(response) {
				callback(response);
			},
			error: function(_, error) {
				callback(null, error.result);
			}
		});

		return request;
	};

	SupervisorModel.prototype.is_member_logged = function(memberId, callback)
	{
		inConcert.logger.debug({action_name : 'is_member_logged', member : memberId, action_type : 'members', app: 'supervisor'});
		callback = callback || $.noop;

		var request = inConcert.invoke({
			url: this.m_url_base + '/members/is_supervisor_member_logged/',
			type: 'POST',
			data: JSON.stringify( { memberId : memberId }),
			dataType: 'json',
			contentType: 'application/json',
			success: function(data) {
				callback(data.logged);
			},
			error: function(_/*, error*/) {
				callback(false);
			}
		});

		return request;
	};

	SupervisorModel.prototype.start_screen_streaming = function(memberId, campaign_id, callback)
	{
		inConcert.logger.debug({action_name : 'start_screen_streaming', member : memberId, action_type : 'members', app: 'supervisor'});
		callback = callback || $.noop;

		var url = this.m_url_base + '/members/start_screen_streaming/';
		var request = inConcert.invoke({
			url: url,
			type: 'POST',
			data: JSON.stringify({ memberId : memberId, campaignId : campaign_id }),
			dataType: 'json',
			contentType: 'application/json',
			success: function(response) {
				callback(true, response.data);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};


	SupervisorModel.prototype.stop_screen_streaming = function(memberId, campaign_id, callback)
	{
		inConcert.logger.debug({action_name : 'stop_screen_streaming', member : memberId, action_type : 'members', app: 'supervisor'});
		callback = callback || $.noop;

		var url = this.m_url_base + '/members/stop_screen_streaming/';
		var request = inConcert.invoke({
			url: url,
			type: 'POST',
			data: JSON.stringify({ memberId : memberId, campaignId : campaign_id }),
			dataType: 'json',
			contentType: 'application/json',
			success: function() {
				callback(true);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};

	/**
	* Searches for queued interactions using a given filter.
	*/
	SupervisorModel.prototype.search_queued_interactions = function (filter, callback)
	{
		inConcert.logger.debug({action_name : 'search_queued', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchQueuedInteractions(filter, callback, 'supervisor');
	};

	/**
	* Searches for waiting interactions using a given filter.
	*/
	SupervisorModel.prototype.search_waiting_interactions = function (filter, callback)
	{
		inConcert.logger.debug({action_name : 'search_waiting', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchWaitingInteractions(filter, callback, 'supervisor');
	};

	/**
	* Searches for abandoned interactions using a given filter.
	*/
	SupervisorModel.prototype.search_abandoned_interactions = function (filter, callback)
	{
		inConcert.logger.debug({action_name : 'search_abandoned', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchAbandonedInteractions(filter, callback, 'supervisor');
	};

	//executes a search using the specified filter
	SupervisorModel.prototype.search_archive_interactions = function( filter, callback )
	{
		inConcert.logger.debug({action_name : 'search_archive', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchArchiveInteractions(filter, callback, 'supervisor');
	};

	//executes a search using the specified filter
	SupervisorModel.prototype.search_open_interactions = function( filter, callback )
	{
		inConcert.logger.debug({action_name : 'search_open', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchOpenInteractions(filter, callback, 'supervisor');
	};

	//executes a search using the specified filter
	SupervisorModel.prototype.search_monitor_interactions = function( filter, callback )
	{
		inConcert.logger.debug({action_name : 'search_monitor', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchOpenInteractions(filter, callback, 'supervisor');
	};

	SupervisorModel.prototype.search_bot_interactions = function (filter, callback)
	{
		inConcert.logger.debug({action_name : 'search_bot_interactions', filter : filter, action_type : 'interactions', app: 'supervisor'});
		inConcert.app.getInteractionManager().searchBotInteractions(filter, callback, 'supervisor');
	};

	// searches interactions that have the given ticket
	SupervisorModel.prototype.search_ticket = function (filter, callback = $.noop)
	{
		inConcert.logger.debug({action_name: 'search_ticket', filter: filter, action_type: 'interactions', app: 'supervisor'});
		let api = inConcert.app.getTicketsAPI();
		api.search({ tickets : filter.tickets }, function(status, result) {
			let list = [];
			if (status && result.status) {
				result.interactions.forEach(function(interaction){
					list.push(inConcert.app.getInteractionManager().buildInteractionObject(interaction));
				});
			}
			callback(list);
		});
	};

	/*
		Configuration Methods
	*/

	SupervisorModel.prototype.get_configuration = function( callback )
	{
		var self = this;
		var request = inConcert.invoke({
			url: self.m_url_base + '/configuration/get/',
			type: 'GET',
			dataType: 'json',
			contentType: 'application/json',
			success: function(response) {
				if (typeof(callback) === 'function'){
					callback(response);
				}
			},
			error: function(_, error) {
				if (typeof(callback) === 'function'){
					callback(null, error.result);
				}
			}
		});

		return request;
	};

	SupervisorModel.prototype.save_configuration = function(user_id, configuration, callback)
	{
		callback = callback || $.noop;

		inConcert.logger.debug({action_name : 'save_configuration', user_id: user_id, action_type:'campaign', app: 'supervisor' });

		var request = inConcert.invoke({
			url: this.m_url_base + '/configuration/save/',
			type: 'POST',
			data: JSON.stringify({
				memberId: user_id,
				configuration: configuration
			}),
			dataType: 'json',
			contentType: 'application/json',
			success: function() {
				callback(true);
			},
			error: function(_, error) {
				callback(false, error.result);
			}
		});

		return request;
	};

	/*
		Workgroups Methods
	*/

	SupervisorModel.prototype.get_workgroups = function(callback)
	{
		callback = callback || $.noop;

		var request = inConcert.invoke({
			url: this.m_url_base + '/workgroups/get/',
			type: 'GET',
			dataType: 'json',
			contentType: 'application/json',
			success: function(response) {
				callback(response);
			},
			error: function(_, error) {
				callback(null, error.result);
			}
		});

		return request;
	};

	SupervisorModel.prototype.__get_channels_properties = function(campaign_id, account_id)
	{
		var campaign = inConcert.app.getCampaign(campaign_id);
		var properties = {};
		if (!campaign) {
			return properties;
		}

		var addAccountProperties = function(account) {
			if (!account) {
				return;
			}
			var channel_properties = properties[account.proxyType] || {
				inbound_part : false,
				outbound_part : false
			};
			if (account.proxyType === 'CALL') {
				channel_properties.inbound_part = channel_properties.inbound_part || campaign.isInboundAccount(account);
				channel_properties.outbound_part = channel_properties.outbound_part || campaign.isOutboundAccount(account);
			}
			properties[account.proxyType] = channel_properties;
		};

		if (account_id) {
			addAccountProperties(campaign.getAccount(account_id));
		}
		else {
			$.each(campaign.get_accounts(), function(_, account) {
				addAccountProperties(account);
			});
		}
		return properties;
	};

	/**
	 * Given a campaign id, and an optional filter options,
	 * will return the current view for that supervisor applied by the filter.
	 *
	 * @param  {String}   campaign_id
	 * @param  {Object}   filter_options
	 * @param  {String}   filter_options.account
	 * @param  {String}   filter_options.attention_level
	 * @param  {String}   filter_options.agentNameFilter
	 * @param  {String}   filter_options.agentOrder
	 * @param  {String}   filter_options.agentStateFilter
	 * @param  {String}   filter_options.page
	 * @param  {Number}   filter_options.rows_per_page
         * @param  {Boolean}  filter_options.showOfflineAgents
         * @param  {Boolean}  filter_options.showKickouts
	 * @param  {Array}    filter_options.workgroups
	 * @param  {Function} callback
	 */
	SupervisorModel.prototype.get_supervisor_view = function(campaign_id, filter_options, callback)
	{
		callback = callback || $.noop;

		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		var campaign_without_vcc = campaign_id.replace(/^(.*?)@(.*?)$/, '$1');
		// obtengo las alarms del supervisor para luego pasarlas en el request.
		var alarms = this.m_dashboard.get_alarms_config(campaign_id);

		var request = inConcert.invoke({
			url: this.m_url_base + '/indicators/get_supervisor_panel/',
			type: 'POST',
			contentType: 'application/json',
			dataType: 'json',
			data: JSON.stringify({
				//parameters
					campaignId         : campaign_without_vcc,
					rows_per_page      : filter_options.rows_per_page,
					showOfflineAgents  : filter_options.showOfflineAgents,
					showKickouts       : filter_options.showKickouts,
				//filters
					account            : filter_options.account,
					attention_level    : filter_options.attention_level,
					agentNameFilter    : filter_options.agentNameFilter,
					agentOrder         : filter_options.agentOrder,
					agentStateFilter   : filter_options.agentStateFilter,
					page               : filter_options.page,
					realtimeWindow     : filter_options.realtimeWindow,
					workgroups         : filter_options.workgroups,
				//configuration
					channel_properties : this.__get_channels_properties(campaign_id, filter_options.account),
				//supervisor alarms
					alarms             : alarms
				}
			),
			success: function(response) {
				callback(response);
			},
			error: function(_, error) {
				callback(null, error.result);
			}
		});

		return request;
	};

	SupervisorModel.prototype.getSupViewAndMembersByCampaigns = function(data, callback)
	{
		callback = callback || $.noop;

		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}
		let campaignFilters;
		let campaign_batch = data.campaigns;
		for (const full_campaign_id in campaign_batch) {
			if (Object.hasOwnProperty.call(campaign_batch, full_campaign_id)) {
				const element = campaign_batch[full_campaign_id];
				const campaign_without_vcc = full_campaign_id.replace(/^(.*?)@(.*?)$/, '$1');
				// obtengo las alarms del supervisor para luego pasarlas en el request.
				const alarms = this.m_dashboard.get_alarms_config(full_campaign_id);
				let data = {
					//parameters
					campaignId         : campaign_without_vcc,
					rows_per_page      : element.filter.rows_per_page,
					showOfflineAgents  : element.filter.showOfflineAgents,
					showKickouts       : element.filter.showKickouts,
				//filters
					account            : element.filter.account,
					attention_level    : element.filter.attention_level,
					agentNameFilter    : element.filter.agentNameFilter,
					agentOrder         : element.filter.agentOrder,
					agentStateFilter   : element.filter.agentStateFilter,
					page               : element.filter.page,
					realtimeWindow     : element.filter.realtimeWindow,
					workgroups         : element.filter.workgroups,
				//configuration
					channel_properties : this.__get_channels_properties(full_campaign_id, element.filter.account),
				//supervisor alarms
					alarms             : alarms
				};

				//Para obtener members (alarmas)
				if (element.filterMembers) {
					data.filterMembers = element.filterMembers;
					data.filterMembers.campaignId = campaign_without_vcc;
					data.filterMembers.alarms = alarms;
				}

				//Para obtener los indicadores
				if (element.filterIndicators) {
					data.filterIndicators = element.filterIndicators;
				}

				if (!campaignFilters) {
					campaignFilters = {};
				}
				campaignFilters[full_campaign_id] = data;
			}
		}

		if (!campaignFilters) {
			return callback(null);
		}
		data.campaigns = campaignFilters;

		return inConcert.invoke({
			url: this.m_url_base + '/indicators/get_supview_and_members_by_campaigns/',
			type: 'POST',
			contentType: 'application/json',
			dataType: 'json',
			data: JSON.stringify(data),
			success: function(response) {
				callback(response);
			},
			error: function(_, error) {
				callback(null, error.result);
			}
		});
	};

	SupervisorModel.prototype.get_historical_data = function(data, callback = $.noop)
	{
		const self = this;
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		let campaign_without_vcc = data.campaign_id.replace(/^(.*?)@(.*?)$/, '$1');
		let request, executeRequestFn;
		let retry = 0;
		let timeout_hnd = null;
		const max_retry = 10; // maxima cantidad de reintentos
		const timemout_ms = 3000; // espero 3 segundos y reintento  (max_retry * max_retry = 30 segundos) 

		// funcion que se invoca nuevamente si el request falla con el result = "UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING", hasta que este ready.
		executeRequestFn = function() {
			retry++;
			console.info('SupervisorModel.get_historical_data - Retry ' + retry + ' executing request /indicators/get_historical_data/. Campaign: ' + campaign_without_vcc + ', retry: ' + retry);
			request = inConcert.invoke({
				url: self.m_url_base + '/indicators/get_historical_data/',
				type: 'POST',
				contentType: 'application/json',
				dataType: 'json',
				data: JSON.stringify({
						campaign_id    : campaign_without_vcc,
						account        : data.account,
						channel        : data.account && data.channel || null,
						attention_level: data.attention_level,
						slice          : data.slice,
						time_ago       : data.time_ago
					}
				),
				success: function(response) {
					if (timeout_hnd) {
						window.clearTimeout(timeout_hnd);
						timeout_hnd = null;
					}
					callback(response);
				},
				error: function(_, error) {
					console.warn('SupervisorModel.get_historical_data - Request error. Campaign: ' + campaign_without_vcc + ', error: ' + JSON.stringify(error));
					if (error.result === 'UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING') { // codigo de error que lo retorna la rocka "reporter"
						if (timeout_hnd) {
							window.clearTimeout(timeout_hnd);
							timeout_hnd = null;
						}
						if (retry < max_retry) {
							timeout_hnd = window.setTimeout(function() {
								executeRequestFn();
							}, timemout_ms);
						}
						else {
							console.warn('SupervisorModel.get_historical_data - Unable to rertrive historical info campaign: ' + campaign_without_vcc + ', has exceeded the maximum number of retries allowed. Max retry ' + max_retry);
							callback(null, error.result);
						}
					}
					else {
						callback(null, error.result);
					}
				}
			});
		};

		executeRequestFn();

		return request;
	};


	SupervisorModel.prototype.get_historical_flows_data = function(data, callback = $.noop)
	{
		const self = this;
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		//let campaign_without_vcc = data.campaign_id.replace(/^(.*?)@(.*?)$/, '$1');
		let request, executeRequestFn;
		let retry = 0;
		let timeout_hnd = null;
		const max_retry = 10; // maxima cantidad de reintentos
		const timemout_ms = 3000; // espero 3 segundos y reintento  (max_retry * max_retry = 30 segundos) 

		// funcion que se invoca nuevamente si el request falla con el result = "UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING", hasta que este ready.
		executeRequestFn = function() {
			retry++;
			console.info('SupervisorModel.get_historical_flows_data - Retry ' + retry + ' executing request /indicators/get_historical_flows_data/. Campaign: ' + data.campaign + ', retry: ' + retry);
			request = inConcert.invoke({
				url: self.m_url_base + '/indicators/get_historical_flows_data/',
				type: 'POST',
				contentType: 'application/json',
				dataType: 'json',
				data: JSON.stringify(data),
				success: function(response) {
					if (timeout_hnd) {
						window.clearTimeout(timeout_hnd);
						timeout_hnd = null;
					}
					callback(response);
				},
				error: function(_, error) {
					console.warn('SupervisorModel.get_historical_flows_data - Request error. Campaign: ' + data.campaign + ', error: ' + JSON.stringify(error));
					if (error.result === 'UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING') { // codigo de error que lo retorna la rocka "reporter"
						if (timeout_hnd) {
							window.clearTimeout(timeout_hnd);
							timeout_hnd = null;
						}
						if (retry < max_retry) {
							timeout_hnd = window.setTimeout(function() {
								executeRequestFn();
							}, timemout_ms);
						}
						else {
							console.warn('SupervisorModel.get_historical_flows_data - Unable to rertrive historical info campaign: ' + data.campaign + ', has exceeded the maximum number of retries allowed. Max retry ' + max_retry);
							callback(null, error.result);
						}
					}
					else {
						callback(null, error.result);
					}
				}
			});
		};

		executeRequestFn();

		return request;
	};

	SupervisorModel.prototype.get_available_process_flows = function(data, callback = $.noop)
	{
		const self = this;
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		//let campaign_without_vcc = data.campaign_id.replace(/^(.*?)@(.*?)$/, '$1');
		let request, executeRequestFn;
		let retry = 0;
		let timeout_hnd = null;
		const max_retry = 10; // maxima cantidad de reintentos
		const timemout_ms = 3000; // espero 3 segundos y reintento  (max_retry * max_retry = 30 segundos) 

		// funcion que se invoca nuevamente si el request falla con el result = "UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING", hasta que este ready.
		executeRequestFn = function() {
			retry++;
			console.info('SupervisorModel.get_available_process_flows - Retry ' + retry + ' executing request /indicators/get_historical_flows_data/. Campaign: ' + data.campaign + ', retry: ' + retry);
			request = inConcert.invoke({
				url: self.m_url_base + '/indicators/get_available_process_flows/',
				type: 'POST',
				contentType: 'application/json',
				dataType: 'json',
				data: JSON.stringify(data),
				success: function(response) {
					if (timeout_hnd) {
						window.clearTimeout(timeout_hnd);
						timeout_hnd = null;
					}
					callback(response);
				},
				error: function(_, error) {
					console.warn('SupervisorModel.get_available_process_flows - Request error. Campaign: ' + data.campaign + ', error: ' + JSON.stringify(error));
					if (error.result === 'UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING') { // codigo de error que lo retorna la rocka "reporter"
						if (timeout_hnd) {
							window.clearTimeout(timeout_hnd);
							timeout_hnd = null;
						}
						if (retry < max_retry) {
							timeout_hnd = window.setTimeout(function() {
								executeRequestFn();
							}, timemout_ms);
						}
						else {
							console.warn('SupervisorModel.get_available_process_flows - Unable to rertrive available process flows campaign: ' + data.campaign + ', has exceeded the maximum number of retries allowed. Max retry ' + max_retry);
							callback(null, error.result);
						}
					}
					else {
						callback(null, error.result);
					}
				}
			});
		};

		executeRequestFn();

		return request;
	};

	SupervisorModel.prototype.get_system_indicators = function( callback )
	{
		var self = this;
		callback = callback || $.noop;

		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		if (this.m_cached_system_indicators) {
			return callback(this.m_cached_system_indicators);
		}

		var request = inConcert.invoke({
			url: this.m_url_base + '/indicators/get_system_indicators/',
			type: 'POST',
			contentType: 'application/json',
			dataType: 'json',
			data: JSON.stringify({}),
			success: function(response) {
				self.m_cached_system_indicators = response;
				callback(response);
			},
			error: function() {
				callback(null);
			}
		});

		return request;
	};

	/**
	 * Retorna los members de la campana, o del attention level,
	 * will return the current view for that supervisor applied by the filter.
	 *
	 * @param  {String}   campaign_id
	 * @param  {Object}   filter_options
	 * @param  {String}   filter_options.account
	 * @param  {String}   filter_options.attention_level
	 * @param  {String}   filter_options.agentNameFilter
	 * @param  {String}   filter_options.agentOrder
	 * @param  {Boolean}  filter_options.orderDesc
	 * @param  {String}   filter_options.agentStateFilter
	 * @param  {String}   filter_options.page
	 * @param  {Number}   filter_options.rows_per_page
	 * @param  {Boolean}  filter_options.showOfflineAgents
	 * @param  {Boolean}  filter_options.showKickoutss
	 * @param  {Boolean}  filter_options.showInteractionDetails
	 * @param  {Array}    filter_options.workgroups
	 * @param  {Function} callback
	 * @param  {Boolean}  from_cache
	 */
	SupervisorModel.prototype.get_supervisor_agents = function(campaign_id, filter_options, callback, from_cache)
	{
		var self = this;
		callback = callback || $.noop;

		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		if ( from_cache ) {
			return callback(self.m_current_members_by_campaign[campaign_id]);
		}
		var campaign_without_vcc = campaign_id.replace(/^(.*?)@(.*?)$/, '$1');
		// obtengo las alarms del supervisor para luego pasarlas en el request.
		var alarms = this.m_dashboard.get_alarms_config(campaign_id);

		var request = inConcert.invoke({
			url: this.m_url_base + '/members/get_supervisor_members/',
			contentType: 'application/json',
			type: 'POST',
			dataType: 'json',
			data: JSON.stringify({
				//parameters
					campaignId                         : campaign_without_vcc,
					rows_per_page                      : filter_options.rows_per_page,
					showOfflineAgents                  : filter_options.showOfflineAgents,
					showKickouts                       : filter_options.showKickouts,
					showInteractionDetails             : filter_options.showInteractionDetails || false,

				// filters
					account_info                       : filter_options.account_info,
					agentNameFilter                    : filter_options.agentNameFilter,
					agentOrder                         : filter_options.agentOrder,
					orderDesc                          : filter_options.orderDesc,
					agentStateFilter                   : filter_options.agentStateFilter,
					agentStateTimeFilter               : filter_options.agentStateTimeFilter,
					agentStateTimeOperatorFilter       : filter_options.agentStateTimeOperatorFilter,
					page                               : filter_options.page,
					workgroups                         : filter_options.workgroups,

					interactionsStateFilter            : filter_options.interactionsStateFilter,
					interactionsStateTimeFilter        : filter_options.interactionsStateTimeFilter,
					interactionsStateTimeOperatorFilter: filter_options.interactionsStateTimeOperatorFilter,
				//supervisor alarms
					alarms                             : alarms
				}
			),
			success: function(response) {
				self.m_current_members_by_campaign[campaign_id] = response;
				callback(response);
			},
			error: function(_, error) {
				callback(null, error.result);
			}
		});

		return request;
	};

	SupervisorModel.prototype.update_campaign_agents = function(campaign_id, agents, total_agent_to_update, callback = $.noop)
	{
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		let campaign_without_vcc = campaign_id.replace(/^(.*?)@(.*?)$/, '$1');

		let request = inConcert.invoke({
			url: this.m_url_base + '/members/update_supervisor_members/',
			contentType: 'application/json',
			type: 'POST',
			dataType: 'json',
			data: JSON.stringify(
			{
				//parameters
					campaignId: campaign_without_vcc,
					members   : agents,
					total_members: total_agent_to_update
				}
			),
			success: function(response) {
				callback(true, response);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};

	
	SupervisorModel.prototype.update_attention_levels_members = function(campaign_id, attLevelsToUpdate, callback = $.noop)
	{
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(null);
		}

		let request = inConcert.invoke({
			url: this.m_url_base + '/members/update_attention_levels_members/',
			contentType: 'application/json',
			type: 'POST',
			dataType: 'json',
			data: JSON.stringify(
				{
					campaignId: campaign_id,
					attLevelsToUpdate : attLevelsToUpdate
				}
			),
			success: function(response) {
				callback(true, response);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};

	/*
	POST info:
		{
		  "query": {
			"Id": "",
			"Enabled": "1",
		  },
		  "columns": [
			"Id",
			"Name",
			"FirstSurname",
			"UserGroup"
		  ]
		}
	*/
	SupervisorModel.prototype.load_users_with_access_to_interaction_app = function(campaign_id, callback = $.noop, require_als)
	{
		const self = this;

		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(false);
		}

		let post_data = {
			campaignId: campaign_id,
			query: {
				Id: '',
				Enabled: '1'
			},
			columns: [
				'Id',
				'Name',
				'FirstSurname',
				'UserGroup'
			]
		};

		if (require_als) {
			post_data.require_als = true;
		}

		this.m_loading_all_users = true;
		var request = inConcert.invoke({
			url: this.m_url_base + '/members/load_users_with_access_to_interaction_app/',
			contentType: 'application/json',
			type: 'POST',
			dataType: 'json',
			data: JSON.stringify(post_data),
			success: function(response) {
				self.m_loading_all_users = false;
				self.m_all_users_with_access_to_interaction_app = response?.users;
				callback(true, self.m_all_users_with_access_to_interaction_app, response?.als);
			},
			error: function(_, error) {
				self.m_loading_all_users = false;
				callback(false, error.result);
			}
		});

		return request;
	};

	SupervisorModel.prototype.loadAttLevelsByCampaign = function(campaign_id, callback = $.noop)
	{
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors) {
			return callback(false);
		}
		const url = inConcert.app.getBaseDomain() + '/inconcert/apps/administration/campaigns/get_attention_levels/' + campaign_id;
		let request = inConcert.invoke({
			url: url,
			contentType: 'application/json',
			type: 'GET',
			dataType: 'json',
			success: function(attention_levels) {			
				callback(true, attention_levels);
			},
			error: function(_, error) {
				callback(false);
			}
		});
		return request;
	};

	SupervisorModel.prototype.getMemberCampaigns = function(member_id, callback = $.noop)
	{
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			return callback(false);
		}
		const url = this.m_url_base + '/members/get_member_campaigns/';
		var request = inConcert.invoke({
			url: url,
			type: 'POST',
			contentType: 'application/json',
			data: JSON.stringify({memberId: member_id}),
			dataType: 'json',
			success: function(campaigns) {			
				callback(true, campaigns);
			},
			error: function(_, error) {
				callback(false, error);
			}
		});

		return request;
	};

	SupervisorModel.prototype.setCurrentMembersByCampaign = function(campaign_id, members)
	{
		this.m_current_members_by_campaign[campaign_id] = members;
	};

}());
