--[[
-- @file inconcert/apps/visitortracking/init.lua
-- @brief Main file for application 'visitor tracking'
--]]
require("orbit")
require "inconcert.orbit.application"
local json = require "cjson"
local wrap_redis = require "wrapredis"

local modulename = select(1, ...)
local m_appLocation = select(2, ...)
assert(m_appLocation, "should be running under inconcertWebHandler")
module(modulename, package.seeall, orbit.new, inconcert.orbit.application.Decorate)

_M.model = nil
local models = relativeRequire("models", ...)

local REDIS_KEY_LAST_VISITS   = "TrackMe::LastVisits"
local REDIS_KEY_VISITED_PAGES = "TrackMe::VisitedPages"
local REDIS_KEY_CLIENT_INFO   = "TrackMe::VisitorInfo"

local redis_cli;
local configControl;
local expire_seconds = 3600;
-- limits the amount of cookies being sent to the database in one-shot
local max_cookies_per_shot = 20;

--App status control table for the appReadyFilter
--It validates the configuration received from the config server
app_status = {
	currentStatus = "waiting",

	CheckStatus = function ()
		local hasChanged, newConfig = configControl:ConfigurationHasChanged()
		if not hasChanged then
			return app_status.currentStatus
		end

		local cassandra_config = newConfig["cassandra"];
		if cassandra_config:DiffState() ~= "none" then
			local ok, err = models.reconfigure(cassandra_config);
			if not ok then
				LogError("%s", err)
				app_status.currentStatus = "invalid"
				return app_status.currentStatus
			end
		end

		local redis_config = newConfig["redis"];
		if redis_config:DiffState() ~= "none" then
			if redis_config:DiffState() == "deleted" then
				LogWarning("Configuration lost")
				redis_cli = nil;
				app_status.currentStatus = "invalid"
				return app_status.currentStatus
			else
				
				LogDebug("Connecting to redis [%s:%s]", redis_config.host, redis_config.port)
				redis_cli = wrap_redis(redis_config.host, tonumber(redis_config.port), redis_config.pwd)
			end
		end

		app_status.currentStatus = "ready"
		return app_status.currentStatus
	end
}

function Initialize()
	LogInfo("Initializing application 'visitor tracking'");

	-- load config from apps_config.lua
	local base_config = inconcert.environment.apps.visitortracking;

	-- load variables (there's only one at the moment)
	expire_seconds = base_config.expire_seconds or expire_seconds;

	-- config control has its own table
	configControl = LuaConfigurationControl.Client({
			type         = "VisitorTracking",
			address      = base_config.config_server.ip,
			port         = base_config.config_server.port,
			node_address = base_config.config_server.myIp,
			node_port    = base_config.config_server.myPort,
		});

	configControl:TriggerNodeReady();
	return true;
end

local function handle_move_old_history(web)
	local redis = redis_cli;
	if not redis then
		web.status = 503;
		return "Service currently Unavailable";
	end

	local current_date = os.time(os.date("!*t")); -- current date in UTC

	local check_next_records = function(count)
		local pipeline = models.create_insert_pipeline()
		local delete_cookies = {};

		local results = redis:zrange(REDIS_KEY_LAST_VISITS, 0, count-1, "WITHSCORES");
		for _,field in ipairs(results) do
			local cookie = field[1];
			local score = field[2];

			if current_date - score < expire_seconds then
				break; -- this one is still new, so are all of them
			end

			table.insert(delete_cookies, cookie);
			LogDebug("Visitor Tracking: Moving cookie [%s] to cassandra", cookie);

			local visitor_info = redis:hgetall(REDIS_KEY_CLIENT_INFO..":"..cookie);
			if visitor_info then
			-- push visitor data into the pipeline
				pipeline:append_visitor_data(cookie, visitor_info);
			end

			local raw_list = redis:lrange(REDIS_KEY_VISITED_PAGES..":"..cookie, 0, -1);

			local last_timestamp;
			for _,raw_entry in pairs(raw_list) do
				local page_entry = json.decode(raw_entry);

				-- make sure the column name is unique based in the following premises
				-- * timestamp  is in seconds
				-- * the list is ordered
				-- so we are covered in case a crazy user visits more than one page in the same seconds
				-- we will cover our ass by testing for "greater than or equal to" instead of a plain equal
				if last_timestamp and last_timestamp >= page_entry.timestamp then
					page_entry.timestamp = last_timestamp + 0.001;
				end

				-- push the page into the pipeline
				pipeline:append_page(cookie, page_entry.page, page_entry.timestamp);

				-- save timestamp so we can check the next page
				last_timestamp = page_entry.timestamp;
			end
		end

		return pipeline, delete_cookies;
	end

	local pipeline;
	LogDebug("Visitor Tracking: Moving old history");
	repeat
		pipeline, delete_cookies = check_next_records(max_cookies_per_shot);
		if not pipeline:empty() then

			LogDebug("Visitor Tracking: Moving %d records (for %d cookies) from redis to cassandra", pipeline:count(), #delete_cookies);

			-- commit inserts
			local ok, err = pipeline:commit();
			if not ok then
				LogError("Failed to save visitor data: %s", err);
				web.status = 500;
				return "Internal Server Error";
			end

			-- and remove the keys from redis
			redis:multi();
			for _,cookie in pairs(delete_cookies) do
				redis:zrem(REDIS_KEY_LAST_VISITS, cookie);
				redis:del(REDIS_KEY_VISITED_PAGES ..":".. cookie);
				redis:del(REDIS_KEY_CLIENT_INFO ..":".. cookie);
			end
			redis:exec();
		end
	until pipeline:count() < max_cookies_per_shot
	LogDebug("Visitor Tracking: No more cookies to move from redis to cassandra");

	web.status = 200;
	return "";
end

local function handle_get_visitor_history(web, cookie)
	web.status = 200;
	web.headers["Content-Type"] = "application/json";

	local history = models.get_visitor_history(cookie);
	return json.encode(history);
end


_M:dispatch_post(handle_move_old_history, "/move_old_history")
_M:dispatch_get(handle_get_visitor_history, "/get_visitor_history/(.*)")
