--[[
-- @file inconcert/apps/visitortracking/models.lua
-- @brief Data acces layer for application 'visitor tracking'
--]]
local cassandra = require "lua_cassandra"
local json = require "cjson"

module((...), package.seeall)

--[[
-- @brief class pipeline, helper for BatchInsert
-- Methods:
-- @li @c append_visitor_data(cookie, visitor_data)
-- @li @c append_page(cookie, page, timestamp)
-- @li @c commit()
-- @li @c count() : @c integer
-- @li @c empty() : @c boolean
--]]
insert_pipeline = {}
__pipeline_mt = {}
__pipeline_mt.__index = insert_pipeline

-- called to update cassandra pool
function reconfigure(cassandra_config)
	for key, value, state in cassandra_config:childs() do
		if state ~= "deleted" then
			LogInfo("Connecting to cassandra on %s:%d", value.host, value.port);
			cassandra.seed(value.host, value.port)
		end
	end

	local ok, connection = pcall(cassandra.CassandraConnection, cassandra, "Historical")
	if not ok then
		return false, "Error while retrieving connection to Cassandra: ".. tostring(err)
	end

	connection:Close()
	return true
end

-- creates an insert pipeline in order to perform a BatchInsert
function create_insert_pipeline()
	return insert_pipeline:new();
end

-- returns the visitor history
function get_visitor_history(cookie)
	local result = { }

	local keyspace = cassandra:CassandraConnection("Historical");

	local columns = keyspace:GetSliceRange(cookie, "VisitorTrackingHistory", "", {
			column_names = {},
			slice_range = {
					start = "",
					finish = "",
					reversed = true,
					count = 200
				}
			}
		);

	if next(columns) then
		result.page_list = {};
		local user_data_found = false;

		for _,v in pairs(columns) do
			if not user_data_found and v.name == "_" then
				local user_data = json.decode(v.value);
				for k, v in pairs(user_data) do
					result[k] = v;
				end
				user_data_found = true;
			else
				table.insert(result.page_list, {
						page = v.value,
						timestamp = string.match(v.name, "timestamp_(.*)")
					});
			end
		end
	end

	return result;
end


-- constructor
function insert_pipeline:new()
	return setmetatable({
			_inserts = {},
		}, __pipeline_mt);
end

-- returns how many records exist in the array
function insert_pipeline:count()
	return #self._inserts;
end

-- returns true if there are no records to insert
function insert_pipeline:empty()
	return not next(self._inserts);
end

-- appends user data
function insert_pipeline:append_visitor_data(cookie, visitor_data)
	table.insert(self._inserts, {
			column_family = "VisitorTrackingHistory",
			column_key    = cookie,
			column_name   = "_",
			column_value  = json.encode(visitor_data)
		});
end

-- appends a single page
function insert_pipeline:append_page(cookie, page, timestamp)
	table.insert(self._inserts, {
			column_family = "VisitorTrackingHistory",
			column_key    = cookie,
			column_name   = "timestamp_"..string.format("%.3f", timestamp),
			column_value  = page
		});
end

-- sends queued records to cassandra
function insert_pipeline:commit()
	local keyspace = cassandra:CassandraConnection("Historical");

	local ok, err = pcall(function()
		keyspace:BatchInsert(self._inserts)
	end);

	keyspace:Close();
	return ok, err;
end
