local json = require "cjson"
local icuregex = require "icu.regex"
local ustring  = require "icu.ustring"
require("socket")
local http = require "socket.http"
local m_helpers = require "helpers"
local m_appLocation = select(2, ...)
module(..., package.seeall)
local m_dataSourceConnections = nil
local m_cassandra = nil

--Nozzle filters
_M.JsonValidator = require "nozzle.stock".json_validator()
--should I customize this filter to raise an error if something distinct a table is passed??
_M.JsonReply = require "nozzle.stock".json_reply()
--this is the correct order of composition for these filters
_M.JsonValidatorReply = JsonValidator..JsonReply

--an empty object used for json encoding purposes
_M.EMPTY_OBJECT = setmetatable ({}, {
	__toJson = function()
		return "{}"
	end,
	__newindex = function(table, key, value)
		error("Attempt to modify read-only table")
	end,
})

function noop () return end -- empty function
function ltrim (s) return m_helpers.string.ltrim(s) end
function rtrim (s) return m_helpers.string.rtrim(s) end
function alltrim (s) return m_helpers.string.alltrim(s) end

function SanitizeId (id)
	assert(type(id) == "string")
	id = id or ""
	id = id:gsub('"', '')
	id = id:gsub("'", "")
	return id
end

function SanitizeFileName (name)
	assert(type(name) == "string")
	name = name or ""
	name = name:gsub('"', '')
	name = name:gsub("'", "")
	name = name:gsub(" ", "_")
	return name
end

function AreEqualStrings (s1, s2)
	if type(s1) ~= "string" or type(s2) ~= "string" then
		return false
	end
	return alltrim(s1:lower()) == alltrim(s2:lower())
end

function RoundNumber (num, numDecimalPlaces)
	local mult = 10^(numDecimalPlaces or 0)
	return math.floor(num * mult + 0.5) / mult
end

local function __MakeGetFn ( fn )
	local LogRequest = require("apps.agent.requestFilters").LogRequest
	return _M.JsonReply .. LogRequest .. fn
end
function DispatchGet (m, fn, url) m:dispatch_get( __MakeGetFn(fn), url ) end

function GetCurrentLoggedUser ()
	if not inconcert.authUser then
		return
	end
	return AppendVcc(inconcert.authUser.userName, inconcert.authUser.virtualContactCenter)
end

--[[
]]--
function IsUserFullSystemAccess (user)
	return m_helpers.utils.IsUserFullSystemAccess(user)
end

--[[
]]--
function CurrentLoggedUserIsFullSystemAccess ()
	if not inconcert.authUser then 
		return false
	end
	return IsUserFullSystemAccess({
		Id        = inconcert.authUser.userName,
		UserGroup = inconcert.authUser.userGroup,
		VCC       = inconcert.authUser.virtualContactCenter,
		VccGroup  = inconcert.authUser.groupVCC
	})
end

local function SetAllegroIntegration (allegro_integration)
	if inconcert and inconcert.authUser and inconcert.authUser.userName then
		inconcert.authUser.allegro_integration = allegro_integration or { enabled = false }
	end
end

local function GetAllegroIntegration ()
	return inconcert and inconcert.authUser and inconcert.authUser.allegro_integration or nil
end

function TypeBelongsToProxy (interactionType, proxy)
	return interactionType == proxy or interactionType == "VOICEMAIL" and proxy == "CALL"
end

--[[
 Crea las funciones de _T (traductor) para la app/translator que se ha recibido por parámetro.

La función que se retorna (la _T), cuando se invoca recibe como segundo parámetro el Id de otra app/translator del cual se necesite traducir.
Sirve para realizar traducciones cruzadas entre diferentes Apps.

Ejemplos:
	1 - Traducción en la misma App:
		_T("hello word")
			Acá realiza la traducción obteniendo el traductor del .MO de la App actual.
	2 - Traducción cruzada entre Apps:
		_T("hello word", "integrations")
			Supongamos que estamos en la App 'Provisioning' en este la traducción se realiza obteniendo el traductor del .MO de la App 'integrations'.
]]--
function MakeTextTranslator (translatorId)
	local translator = inconcert.translators[translatorId]
	return function(...)
			local args = {...}
			local translatedWord
			local otherTranslatorId = args[2]
			if otherTranslatorId then
				local word = args[1]
				local otherTranslator = inconcert.translators[otherTranslatorId]
				translatedWord = otherTranslator:Text(word)
			else
				translatedWord = translator:Text(...)
			end
			--LogDebug("[UTILS.MakeTextTranslator._T] ---> args: %s. translatedWord: '%s'", JSON_Encode(args), translatedWord)
			return translatedWord
		end,
		i18nlib.LCID[translator.language]
end

function CassandraColumnExists (columnFamily, column, key, keyspace)
	assert(columnFamily)
	assert(column)
	assert(key)
	assert(keyspace)

	return utils.GetColumnValue({
		keyspace     = keyspace,
		columnFamily = columnFamily,
		key          = key,
		column       = column
	})
end


--[[

]]--
function GetAllegroVCC (vcc)
	local allegroVccModule = GetAllegroController({controllerName = "provisioning_vccs", modelName = nil})
	local success, result = allegroVccModule.GetVCC(vcc)
	if not success then
		local err = result
		return false, err
	end
	return true, result
end

--[[

]]--
function IsReadyAllegroVcc (vcc, ksAgent)
	if not HasAllegroIntegration(vcc, ksAgent) then
		local result =  {result = "VCC NOT HAS ALLEGRO INTEGRATION", vcc = vcc}
		--LogDebug("[UTILS.IsReadyAllegroVcc] - vcc: %s -> result: '%s'", vcc, json.encode(result))
		return false, result
	end
	---------------------------------------------------------------------------------------
	--  Se ejecuta un metodo de Algoritmos para realizar la validacion de que el Servidor de Allegro, el Webhandler y MsSQL
	--
	local allegroResourcesModule = GetAllegroController({controllerName = "resources", modelName = nil})
	local success, result = allegroResourcesModule.LoadAlgorithms()
	---------------------------------------------------------------------------------------
	if not success then
		local err = {
			result = "UNABLE TO CONNECT WITH ALLEGRO",
			vcc = vcc,
			err = result
		}
		LogError("[UTILS.IsReadyAllegroVcc] '%s'", json.encode(err))
		return false, err
	end
	return true, result
end

--[[

]]--
function HasAllegroIntegration (vcc_id, keyspace)
	--from cache
	--LogDebug("[UTILS.HasAllegroIntegration] vcc: %s", vcc_id)
	local allegro_integration = GetAllegroIntegration()
	if allegro_integration then
		--LogDebug("[UTILS.HasAllegroIntegration] 1 - vcc: %s -> allegro_integration: '%s'", vcc_id, json.encode(allegro_integration))
		return (allegro_integration.enabled or false), allegro_integration
	end
	--from database
	vcc_id = vcc_id or (inconcert and inconcert.authUser and inconcert.authUser.virtualContactCenter)
	if vcc_id then
		allegro_integration = GetColumnValue {
			keyspace     = keyspace or "Agent",
			columnFamily = "VCCs",
			key          = vcc_id,
			column       = "AllegroIntegration"
		}
	end
	if type(allegro_integration) == "string" then
		allegro_integration = json.decode(allegro_integration)
	else
		allegro_integration = nil
	end
	--LogDebug("[UTILS.HasAllegroIntegration] 2 - vcc: %s -> allegro_integration: '%s'", vcc_id, json.encode(allegro_integration or {}))
	SetAllegroIntegration(allegro_integration)
	return (allegro_integration and allegro_integration.enabled or false), allegro_integration
end

function MakeLogAllegroVccMsg (vcc, ksAgent)
	if not vcc then
		return ""
	end
	return HasAllegroIntegration(vcc, ksAgent) and " with Allegro Integration" or ""
end

function IsAllegroAuthenticationRequired (vcc_id, keyspace)
	local has, allegro_integration = HasAllegroIntegration (vcc_id, keyspace)
	if not has then
		return false
	end
	local authModule = allegro_integration and allegro_integration.IntegratedAuthentication and
		allegro_integration.IntegratedAuthentication.AuthenticationModule or nil
	return authModule == "ACTIVE DIRECTORY" or authModule == "CUSTOM"
end

function IsAllegroChannel (channel, vcc_id, keyspace)
	return HasAllegroIntegration( vcc_id, keyspace ) and (channel == "CALL" or channel == "VOICEMAIL")
end

--[[
	Valida si la campana es de Allegro. Es decir si en un vcc de allegro tiene una cuenta de telefonia.
	retorna true/false y la cuenta de telefonia.
]]--
function IsAllegroCampaign (campaignId, keyspace, currentVcc)
	assert(campaignId)
	local isAllegroCampaign = false
	local vcc = currentVcc
	if not vcc then
		if inconcert and inconcert.authUser and inconcert.authUser.virtualContactCenter then
			vcc = inconcert.authUser.virtualContactCenter
			campaignId = AppendVcc(GetId(campaignId), vcc)
		else
			local _
			_, vcc = SeparateVCC(campaignId)
		end
	end
	if not HasAllegroIntegration(vcc, keyspace) then
		return false
	end
	local accounts = GetCampaignAccounts(campaignId, keyspace, vcc)
	--LogDebug("[UTILS.IsAllegroCampaign]  1 - campaignId: %s@%s -> accounts: '%s'", campaignId, vcc, json.encode(accounts or {}))
	local callAllegroAccount
	for _, account in pairs(accounts) do
		if IsAllegroChannel(account.proxyType, vcc, keyspace) then
			--LogDebug("[UTILS.IsAllegroCampaign]  2 - accountId: %s -> isAllegroCampaign: true.", account.id)
			isAllegroCampaign = true
			callAllegroAccount = account
			break
		else
			--LogDebug("[UTILS.IsAllegroCampaign]  3 - accountId: %s -> isAllegroCampaign: false.", account.id)
		end
	end
	return isAllegroCampaign, callAllegroAccount, accounts
end

--[[
	Estos valores se utilizan en VCC de Allegro y si la campana es de tipo CALL.
	Los mismos se utilizan para determinar si la sincronizacion de Usuarios entre OCC vs Allegro ha fallado (ejemplo por caidas de los ambientes, webhandlers, etc.)
	Estos valores se utilizan en:
		- /administration/controllers/campaigns funciones "Search" y "UpdateMembersList"
		- \administration\scripts\adm_form_campaigns_section_members.js funcion "validate_entity"
		- /agent/agent.lua funcion "AgentStartup"
	Sample:
		{
			warnings = {
				stepNumber     = 4, -- o menor (Si el valor del step number)
				expiredTimeSec = 10
			}
		}
	
	El valor del stepNumber 4 es el step donde ha culminado la actualizacion contra Allegro.
	Si es menor o igual al stepNumber 4, y ha fallado es necesario realizar un rollback sobre allegro.
]]--
function GetUpdateMembersInfo ()
	return {
		-- allegro campaigns
		warnings = {
			stepNumber     = 4,  -- o menor que cuatro
			expiredTimeSec = 120 -- 2 minutos (este es un valor maximo en segundos en el cual la operacion de actualizacion de members puede estar corriendo)
		}
	}
end

--[[
	Esta funcion calcula los warnings de las campanas de Allegro, cuando se realiza la actualizacion de Members.
	La misma es invocada en:
		1- /administration/controllers/campaigns funciones "Search" y "UpdateMembersList"
			Aca sirve para visualizar en la view de campanas un error de las campanas con este tipo de error (siempre se va a ver este error, hasta que no se arregle)
		2- /agent/agent.lua funcion "AgentStartup"
			En el login del usuario, sirve para darle feedback en el navegador.
			Tambien sirve en el login para realizar sincronizaciones de forma automatica.

]]--
function AddCampaignsWarning (vcc, campaign, campaignsWarnings, now, callback, ksAgent, mustExcludeMembersInErr)
	assert(vcc)
	assert(campaign)
	assert(campaignsWarnings)
	now = now or os.time()
	callback = callback or function() end
	local updateMembersResult = campaign.UpdateMembersResult
	if not updateMembersResult then
		--LogDebug("[UTILS.AddCampaignsWarning] 1---AddWarning----> updateMembersResult: %s", json.encode(updateMembersResult or {}))
		return
	end
	if not HasAllegroIntegration(vcc, ksAgent) then
		--LogDebug("[UTILS.AddCampaignsWarning] 2---AddWarning----> vcc: %s. NO ES UN VCC DE ALLEGRO.", vcc)
		return
	end
	-- comienza la validacion
	--LogDebug("[UTILS.AddCampaignsWarning] 3---AddWarning----> updateMembersResult: %s", json.encode(updateMembersResult))
	local updateMemberElapsedTimeSec = now - updateMembersResult.time
	--LogDebug("[UTILS.AddCampaignsWarning] 4---AddWarning----> updateMemberElapsedTimeSec: %s", updateMemberElapsedTimeSec)
	local updateMembersInfo = GetUpdateMembersInfo()
	if (updateMembersResult.finish == "ERROR" and updateMembersResult.errType == "ALLEGRO_INTEGRATION_ERROR")
		or (
			updateMembersResult.stepNumber <= updateMembersInfo.warnings.stepNumber
			--and updateMemberElapsedTimeSec > updateMembersInfo.warnings.expiredTimeSec
		)
	then
		--[[
			El valor del stepNumber 4 es el step donde ha culminado la actualizacion contra Allegro.
			Si es menor o igual al stepNumber 4, y ha fallado es necesario realizar un rollback sobre allegro.
		]]--
		if mustExcludeMembersInErr then
			updateMembersResult.members = nil -- debo excluir los members en la respuesta (el detalle, para que no viaje al navegador)
		end
		local warning = {
			campaignId = GetId(campaign.Id),
			error      = updateMembersResult
		}
		--LogDebug("[UTILS.AddCampaignsWarning] 4.1---AddWarning-SI---> campaign.Id: %s.", campaign.Id)
		table.insert(campaignsWarnings, warning)
		callback(warning)
	--else
		--LogDebug("[UTILS.AddCampaignsWarning] 4.2---AddWarning-NO---> campaign.Id: %s.", campaign.Id)
	end
end

local function SetAppsIntegrations (appsIntegrations)
	if inconcert and inconcert.authUser and inconcert.authUser.userName then
		inconcert.authUser.appsIntegrations = appsIntegrations
	end
end

local function GetAppsIntegrations (vccId, keyspace)
	--from cache
	local appsIntegrations = inconcert and inconcert.authUser and inconcert.authUser.appsIntegrations or nil
	if appsIntegrations then
		return appsIntegrations
	end
	--from database
	vccId = vccId or (inconcert and inconcert.authUser and inconcert.authUser.virtualContactCenter)

	if vccId then
		appsIntegrations = GetColumnValue {
			keyspace     = keyspace or "Agent",
			columnFamily = "VCCs",
			key          = vccId,
			column       = "AppsIntegrations"
		}
		if type(appsIntegrations) == "string" then
			appsIntegrations = json.decode(appsIntegrations)
			if m_helpers.utils.HasLicenseSystem() then
				local options = {}
				if inconcert and inconcert.environment and inconcert.environment.apps and inconcert.environment.apps.agent and inconcert.environment.apps.agent.agentservice and inconcert.environment.apps.agent.cluster then
					options.host = inconcert.environment.apps.agent.agentservice.host
					options.port = inconcert.environment.apps.agent.agentservice.port
					options.cluster = inconcert.environment.apps.agent.cluster
				end	
				local vccLicense = m_helpers.agentservice.GetVccLicense(options, vccId)
				for appId, app in pairs(appsIntegrations) do
					if app.Licenses and app.Licenses.Users then
						app.Licenses.Users = vccLicense and vccLicense.AppsIntegrations and vccLicense.AppsIntegrations[appId] or 0
					end
				end
			end
		else
			appsIntegrations = nil
		end
	end
	SetAppsIntegrations(appsIntegrations)
	return appsIntegrations
end

function HasHelpDeskIntegration (vccId, keyspace)
	local appsIntegrations = GetAppsIntegrations(vccId, keyspace)
	if appsIntegrations and appsIntegrations.HelpDesk then
		local enabled = tonumber(appsIntegrations.HelpDesk.Enabled) == 1
		return enabled, appsIntegrations.HelpDesk
	end
	return false
end

function GetCampaignAccounts (campaign_id, keyspace, vcc)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	local campaign_id_with_vcc = AppendVcc(GetId(campaign_id), vcc)

	local params = {
		keyspace = keyspace or "Agent",
		superColumns = true,
		columnFamily = "CampaignAccounts",
		key = campaign_id_with_vcc
	}

	return GetAllColumns( params ) or {}
end

function GetVCCCampaigns (vcc, keyspace)
	return GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCCampaigns",
		key = vcc
	} or {}
end

function GetCampaignMembers (vcc, campaign_id, ksAgent)
	return GetAllColumns {
		keyspace = ksAgent or "Agent",
		columnFamily = "CampaignMembers",
		key = AppendVcc(campaign_id, vcc)
	} or {}
end

---
-- OJO! este modulo es compartido entre distintas aplicaciones, las que tienen que llamar este SetupConnections
-- antes de poder usar funciones como GetAllColumns, SaveRows o RemoveColumns, pasando la tabla 'lib_cassandra'
-- a la que se le tiene que haber hecho 'seed' previamente, etc.
-- Si alguien por error setea algo incorrecto acá, rompe el funcionamiento de cualquier otra aplicación.
--
-- Los módulos así, como este, tipo "utils" NO DEBEN TENER ESTADO.
-- En SetupConnections se estaba pasando Redis, que por suerte no se estaba usando, lo cual hubiese sido mucho peor
-- porque ya se está pasando una instancia conectada.
-- Si otra aplicación pasa una conexion de redis a un server distinto, todas las demas aplicaciones dejan de andar, y
-- andá a encontrar el porqué del bug.
--

function SetupConnections (connections)
	if not m_dataSourceConnections then
		m_dataSourceConnections = connections
	else
		for name, connection in pairs(connections) do
			if not m_dataSourceConnections[name] then
				m_dataSourceConnections[name] = connection
			end
		end
	end
	m_cassandra = connections.cassandra
end

function GetDatasource ()
	return m_cassandra
end

local m_cassandraCachedKeySpaces = {}
function GetCachedCassandraKeySpace (ksName)
	assert(ksName)
	assert(m_cassandra)
	if m_cassandraCachedKeySpaces[ksName] then
		--LogDebug("[UTILS.GetCachedCassandraKeySpace] 1- Retrieving Cassandra Ks: %s from CACHED Connection -> keyspace: '%s.", ksName, m_cassandraCachedKeySpaces[ksName])
		return m_cassandraCachedKeySpaces[ksName]
	end
	m_cassandraCachedKeySpaces[ksName] = m_cassandra:CassandraConnection("Agent")
	--LogDebug("[UTILS.GetCachedCassandraKeySpace] 2- Retrieving Cassandra Ks: %s from NEW Connection -> keyspace: '%s.", ksName, m_cassandraCachedKeySpaces[ksName])
	return m_cassandraCachedKeySpaces[ksName]
end

function GetConnections ()
	return m_dataSourceConnections
end

function GetRedisClient ()
	local agentModel = require "apps.agent.model"
	local redis = agentModel:GetRedisClient()
	return redis
end

function GetAgentRedisClient ()
	local agentModel = require "apps.agent.model"
	local agentRedis = agentModel:GetAgentRedisClient()
	return agentRedis
end

function GetColumnValue ( options )
	assert( options, "You must supply options" )
	--assert( type( options.keyspace ) == "string" or type( options.keyspace ) == "table", "keyspace id must be a string or table")
	assert( type( options.columnFamily ) == "string", "columnFamily id must be a string")
	assert( type( options.key ) == "string", "columnFamily id must be a string")
	assert( type( options.column ) == "string", "columnFamily id must be a string")

	if type( options.keyspace ) == "string" then
		local keyspace = m_cassandra:CassandraConnection( options.keyspace )
		local columnValue = keyspace:GetColumnValue( options.key, options.columnFamily, options.column )
		keyspace:Close()
		return columnValue
	else
		return options.keyspace:GetColumnValue( options.key, options.columnFamily, options.column )
	end
end

function GetSystemConfig (elementId, keyspace, columns)
	local ksProv = keyspace
	local release = false
	if not ksProv then
		ksProv = GetDatasource():CassandraConnection("Provisioning")
		release = true
	end
	local result = m_helpers.system:GetSystemConfig(elementId, ksProv, columns)
	if release then
		ksProv:Close()
	end
	return result
end

function JSON_Decode (value)
	return m_helpers.table:JSON_Decode(value)
end

function JSON_Encode (value)
	return m_helpers.table:JSON_Encode(value)
end

function Uuid (withoutSeparator, length)
	return m_helpers.utils:Uuid(withoutSeparator, length)
end

--[[
	Generic function to retrieve all columns in a keyspace and columnFamily

	ATENCION:
	---------
	ESTA FUNCION ES UNA COPIA EXACTA DE LA QUE EXISTE EN EL REPO "FULLCONTACT" EN "helpers.lua"
	Hay que mantenerla igual en ambos lados, es decir si la modificamos acá hay que actualizarla en "helpers.lua" y viceversa
	
	Se agrega filtrado de supercolumns, para mantener compatibilidad con codigos legados se debe agregar en options 
	la opcion de filterSuperColumns = true y pasar el columns como un array con las columnas deseadas: { Id, Name, Type }
--]]
function GetAllColumns (options)
	assert(options, "You must supply options")

	local function XOR (A, B)
		return (A and not B) or (B and not A)
	end

	assert(XOR(options.key, options.startKey or options.endKey), "You must supply key or <startKey, endKey> but not both")

	--LogDebug(" 1 ENTRA GetAllColumns " .. tostring(options.columnFamily))

	local has_close_keyspace = true
	local keyspace
	if type( options.keyspace ) == "string" then
		keyspace = m_cassandra:CassandraConnection( options.keyspace )
	else
		keyspace = options.keyspace
		has_close_keyspace = false
	end

	local func, specific_columns
	if options.superColumns then
		func = keyspace.GetSuperRangeSlices
		specific_columns = {}
	else
		func = keyspace.GetRangeSlices
		specific_columns = options.columns or {}
	end

	local CanAddRow = type(options.CanAddRow) == "function" and options.CanAddRow or function() return true end

	local get_rows_chunk = function( start_key, end_key, load_chunk_rows )
		local ok, rows = pcall(func,
			keyspace,
			options.columnFamily, --column family
			"", --supercolumn
			start_key, --startkey
			end_key, --endkey
			{ --predicate
				column_names = specific_columns,
				slice_range = {
					start = options.startColumn or "",
					finish = options.endColumn or "",
					reversed = options.reversed or false,
					count = options.columnCount or 10000
				}
			},
			load_chunk_rows --total keys
		)
		if ok then
			return rows
		end
	end

	local get_value = function( column, super_column, key )
		if type(options.transform_value) == "function" then
			return options.transform_value( column, super_column, key )
		else
			return options.valueAsNumber and column.value_as_number or JSON_Decode(column.value)
		end
	end

	local all_result = {}
	local chunk_result = {}
	local previsous_chunk_result = {}
	local start_chunk_key = options.key or options.startKey or ""
	local end_chunk_key = options.key or options.endKey or ""
	local previous_chunk_key
	local top_rows = options.rowCount
	local total_rows = 0
	local continue = true
	local chunks_loop = 0

	-- Funcion encargada de determinar si la funcion tiene que retornar todos los registros de los chunks, o si tiene que invocar al callback
	-- con los rows que se han obtenido en el chunk
	local function HasRetrieveChunks ()
		return options.callbacks and options.callbacks.chunk_iter
	end

	-- To use with supercolumns to check if add columns to result or not
	local function IsAllowedSuperColumn( superColsArray, value )
		if options.filterSuperColumns and superColsArray and #superColsArray > 0
		then
			for i = 1, #superColsArray do
				if superColsArray[i] == value then return true end
			end

			return false
		else
			return true
		end
	end

	local rows_per_chunk = options.rows_per_chunk or 30000
	local sleep_between_chunk_sec = options.sleep_between_chunk_sec or nil
	--LogDebug("rows_per_chunk: " .. tostring(rows_per_chunk) .. ", sleep_between_chunk_sec: " .. tostring(sleep_between_chunk_sec))

	while continue do
		chunks_loop = chunks_loop + 1
		if type(sleep_between_chunk_sec) == "number" and chunks_loop > 1 then
			LogInfo("[GetAllColumns] Wait moment please, there is a " .. tostring(sleep_between_chunk_sec) .. " seconds pause between chunk. Chunk " .. tostring(chunks_loop) .. " of " .. tostring(rows_per_chunk) .. " rows.")
			socket.sleep(sleep_between_chunk_sec)
		end

		local load_chunk_rows = rows_per_chunk
		-- El primer chunk siempre retorna la totalidad especificada en el parametro recibido "rows_per_chunk",
		--  pero a partir del segundo chunk tengo que incrementar en 1 el valor recibido como parametro (solo una vez),
		--  para que los siguientes, chunks retornen la misma cantidad.
		-- Esto es asi ya que el punto de partida para obtener informacion es el ultimo row del chunk anterior, y si no hacemos esto,
		--   en el siguiente chunk va a venir repetido el ultimo row del chunk anterior.
		if chunks_loop > 1 then
			load_chunk_rows = rows_per_chunk + 1
		end
		previous_chunk_key = start_chunk_key

		local rows = get_rows_chunk( start_chunk_key, end_chunk_key, load_chunk_rows )

		--LogDebug("Chunk Number: " .. tostring(chunks_loop) .. ", Column Family" .. tostring(options.columnFamily))
		if not rows then
			break
		end

		local total_rows_in_chunk = 0

		for key, columns in rows:Iterator() do
			-- Si el row fue agregado en el chunk anterior, no lo tiene que agregar.
			if not previsous_chunk_result[key] then
				local row = {}
				for _, column in pairs(columns) do
					if not options.superColumns then
						row[column.name] = get_value(column, nil, key)
					else
						local super_column = column
						row[super_column.name] = {}
						for _, col_of_super_column in pairs(super_column.columns) do
							if IsAllowedSuperColumn(options.columns, col_of_super_column.name) then
								row[super_column.name][col_of_super_column.name] = get_value(col_of_super_column, super_column, key)
							end
						end
					end
				end

				if CanAddRow(row) then
					all_result[key] = row
					chunk_result[key] = DeepCopy(all_result[key])
				end
			end

			total_rows = total_rows + 1
			total_rows_in_chunk = total_rows_in_chunk + 1
			--LogDebug(tostring(total_rows) .. "' - CHUNK_KEY: '" .. tostring(start_chunk_key) .. "' - chunk row: " .. tostring(total_rows_in_chunk))

			start_chunk_key = key -- actualizo el chunk con el que se ha procesado, para cuando se vuela a pedir los siguientes chunks.

			local is_top_rows = total_rows == top_rows
			if is_top_rows then
				--LogDebug("BREAK - is_top_rows: " .. tostring(is_top_rows))
				if HasRetrieveChunks() then
				-- Invoco al callback con los rows del chunk y luego limpio la tabla, para continuar con los chunks siguientes.
					--LogDebug("1-->RETORNO al callback del chunk: " .. tostring(total_rows_in_chunk))
					options.callbacks.chunk_iter(chunk_result, chunks_loop)
				end
				previsous_chunk_result = nil
				previsous_chunk_result = DeepCopy(chunk_result)
				chunk_result = {}
				continue = false
				break
			end
		end

		if continue then
			if HasRetrieveChunks() then
			-- Invoco al callback con los rows del chunk y luego limpio la tabla, para continuar con los chunks siguientes.
				--LogDebug("2-->RETORNO al callback del chunk: " .. tostring(total_rows_in_chunk))
				options.callbacks.chunk_iter(chunk_result, chunks_loop)
			end
			previsous_chunk_result = nil
			previsous_chunk_result = DeepCopy(chunk_result)
			chunk_result = {}

			if total_rows_in_chunk < rows_per_chunk then
				--LogDebug("BREAK - total_rows_in_chunk: " .. tostring(total_rows_in_chunk) .. " rows_per_chunk: " .. tostring(rows_per_chunk))
				continue = false
			else
				local is_equal_chunks = previous_chunk_key == start_chunk_key

				if is_equal_chunks then
					--LogDebug("BREAK - is_equal_chunks: " .. tostring(is_equal_chunks) .. " start_chunk_key: " .. start_chunk_key)
					continue = false
				end
			end
		end
	end

	if has_close_keyspace then
		keyspace:Close()
	end

	if HasRetrieveChunks() then
	-- Como la funcion que retorna los rows de cada chunk se encuentra en el callback, me voy y no retorno nada mas.
		return
	end

	-- Como no se fueron retornando los row por cada callback, retorno todo de una.

	if options.key then
		all_result = all_result[options.key]
	end

	--LogDebug("all_result: " .. JSON_Encode(all_result))
	--LogDebug(" 2 SALE GetAllColumns " .. tostring(options.columnFamily))

	return all_result
end

--[[
	Funcion encargada de procesar todas las columnas por chunk y retornar al callback de chunks.
	La funcion en si termina invocando a la funcion "GetAllColumns" con otros parametros que son requeridos para esta tarea.
]]--
function ProcessBatchColumns (options)
	assert(type(options) == "table", "Options not a table")
	assert(type(options.rows_per_chunk) == "number", "Options.rows_per_chunk not a number")
	assert(type(options.callbacks) == "table", "Options.callbacks not a table")
	assert(type(options.callbacks.chunk_iter) == "function", "Options.callbacks.chunk_iter not a function")

	GetAllColumns(options)
end

--[[
	Función genérica que almacena en cassandra.

	ATENCION:
	---------
	ESTA FUNCION ES UNA COPIA EXACTA DE LA QUE EXISTE EN EL REPO "FULLCONTACT" EN "helpers.lua"
	Hay que mantenerla igual en ambos lados, es decir si la modificamos acá hay que actualizarla en "helpers.lua" y viceversa
--]]
function SaveRows (options)
	assert(options, "You must supply options")
	assert(options.keyspace, "You must supply a keyspace")
	assert(options.columnFamily, "You must supply a columnFamily")
	assert(options.rows, "You must supply one or more rows")

	local keyspace, close_connection
	if type(options.keyspace) == "string" then
		close_connection = true
		keyspace = m_cassandra:CassandraConnection(options.keyspace)
	else
		close_connection = false
		keyspace = options.keyspace
	end

	local columnFamily = options.columnFamily
	local superColumns = options.superColumns
	local removeBefore = options.removeBefore

	for rowId, row in pairs(options.rows) do
		if removeBefore and not superColumns then
			keyspace:RemoveColumn(rowId, columnFamily)
		end
		for columndId, column in pairs(row) do
			if superColumns then
				if removeBefore then
					keyspace:RemoveSuperColumn(rowId, columnFamily, columndId)
				end
				for k, v in pairs(column) do
					local value = type(v) == "table" and json.encode(v) or tostring(v)
					keyspace:InsertSuperColumn(rowId, columnFamily, columndId, k, value)
				end
			else
				--standard column
				local value = type(column) == "table" and json.encode(column) or tostring(column)
				keyspace:InsertColumn(rowId, columnFamily, columndId, value)
			end
		end
	end

	if close_connection then
		keyspace:Close()
	end
end

--[[
	If parameter row has the form: ["row1", "row2", ...]
	This function will remove one or more rows,
	
	If parameter row has the form: {
		row1 : ["column1", "column2", ...],
		row2 : ["column1", "column3", ...],
	}
	This function will remove a supercolumn
	
	
	If parameter row has the form: {
		row1: {
			supercolumn1: ["column1", "column2", ...],
			supercolumn2: ["column1", "column2", ...],
		},
		row2: {...}
	}
	
	This function will remove only each supercolumn
]]
function RemoveColumns (options)
	assert(options, "You must supply options")
	assert(options.keyspace, "You must supply a keyspace")
	assert(options.columnFamily, "You must supply a columnFamily")
	assert(options.rows, "You must supply one or more rows")

	local keyspace, close_connection
	if type(options.keyspace) == "string" then
		close_connection = true
		keyspace = m_cassandra:CassandraConnection(options.keyspace)
	else
		close_connection = false
		keyspace = options.keyspace
	end
	--LogDebug("[UTILS.RemoveColumns] -> options.columnFamily: %s, options.rows: %s", options.columnFamily, json.encode(options.rows or {}))

	for rowId, row in pairs(options.rows) do
		if type(row) == "string" then
			--LogDebug("[UTILS.RemoveColumns]  a-> rowId: %s, options.columnFamily: %s, row: %s", rowId, options.columnFamily, row)
			keyspace:RemoveColumn(row, options.columnFamily)
		elseif type(row) == "table" then
			for columnName, column in pairs(row) do
				if type(column) == "string" then
					if options.superColumns then
						--LogDebug("[UTILS.RemoveColumns]  b-> (superColumn) rowId: %s, options.columnFamily: %s, column: %s", rowId, options.columnFamily, column)
						keyspace:RemoveColumn(rowId, options.columnFamily, nil, column)
					else
						--LogDebug("[UTILS.RemoveColumns]  c-> rowId: %s, options.columnFamily: %s, column: %s", rowId, options.columnFamily, column)
						keyspace:RemoveColumn(rowId, options.columnFamily, column)
					end
				elseif type(column) == "table" then
					if not options.superColumns then
						error("You must specify options.superColumns to remove in this level")
					end
					for k, v in pairs(column) do
						--LogDebug("[UTILS.RemoveColumns]  d-> rowId: '%s', options.columnFamily: '%s', (k: '%s',  v: '%s'), column: '%s'", rowId, options.columnFamily, k, v, columnName)
						keyspace:RemoveColumn(rowId, options.columnFamily, v, columnName)
					end
				else
					error("Unexpected column type in RemoveColumn")
				end
			end
		else
			error("Unexpected row type in RemoveColumn")
		end
	end

	if close_connection then
		keyspace:Close()
	end
end

--[[
	Attempts to detect the language using the language Cookie (set by the login process) or the Accept-Language header sent by the browser
--]]
function DetectLanguage (web)
	LogInfo("Trying to detect language => Accept-Language: %s - Language Cookie: %s", web.vars.HTTP_ACCEPT_LANGUAGE, web.cookies.LULID)
	local acceptHeader, languageCookie = web.vars.HTTP_ACCEPT_LANGUAGE or "", web.cookies.LULID
	if languageCookie then
		return i18nlib.LCID[languageCookie], languageCookie
	else
		local langCode = string.match(acceptHeader, "^([^,]+)")
		if langCode then
			langCode = langCode:match("(%a%a)[%-.+]?")
			web:set_cookie("LULID", {
				value = langCode, 
				path = "/", 
				domain = m_cookies_domain,
				httponly = true,
				sameSite = "None",
				secure = true
			})
			return i18nlib.LCID[langCode], langCode
		else
			web:set_cookie("LULID", {
				value = "en_US", 
				path = "/", 
				domain = m_cookies_domain,
				httponly = true,
				sameSite = "None",
				secure = true
			})
			return i18nlib.LCID.en_US, "en"	-- defaulting to english
		end
	end
end

--[[
	Validates that the origin table has the required fields
	and that fields are of the type that are supposed to be
	(only in the first level)
]]
function HasFields (origin, fields)
	for field, typeName in pairs(fields) do
		if type(origin[field]) ~= typeName then
			LogWarning("Table %s does not match with field %s with type %s", json.encode(origin), field, typeName)
			return false, "You need to provide field "..field.." as "..typeName
		end
	end
	return true
end

--[[
	Appends @vcc to id only if it hasn't
]]
function AppendVcc (id_, vcc_)
	id_ = id_ or ""
	vcc_ = vcc_ or ""
	local id, vcc = id_:match("(.+)@(.+)")
	if id and vcc then
		return id.."@"..vcc_
	else
		return id_.."@"..vcc_
	end
end

--[[
	Removes vcc from idAtVcc if it has a vcc
]]
function GetId (idAtVcc)
	local id = (idAtVcc or ""):match("(.+)@.+")
	return id or idAtVcc
end

function GetAppLocation ()
	return m_appLocation
end

function Separate (s, separator)
	return m_helpers.utils:Separate(s, separator)
end

function SeparateVCC (s)
	separator = "@"
	return Separate(s, separator)
end

function TableIsEmpty (t)
	return m_helpers.table:TableIsEmpty(t)
end

function StringIsEmpty (s)
	return type(s) ~= "string" or alltrim(s) == ""
end
 
function SplitString (s, separator)
	return m_helpers.utils.SplitString(s, separator)
end

--[[
Filter entities by fields or predicate
each field must match exactly and by conjunction
You can pass a function with parameters (key, entity) as a predicate
]]
function FilterEntities (entities, fields)
	local result = {}
	local matchFn = type(fields) == "function" and fields or function(key, entity)
		for field, value in pairs(fields) do
			if not (entity[field] and entity[field]:lower() == value:lower()) then
				return false
			end
		end
		return true
	end
	for key, entity in pairs(entities) do
		if matchFn(key, entity) then
			result[key] = entity
		end
	end
	return result
end

--[[
Functional programming style map
]]
function Map (original, fn)
	local result = {}
	for k, v in pairs(original or {}) do
		result[k] = fn(k, v) or v
	end
	return result
end

--[[
Remaps keys applying a function to each one and returning a new map
]]
function Remap (original, fn)
	local result = {}
	for k, v in pairs(original) do
		result[fn(k, v) or k] = v
	end
	return result
end
--[[
Reduce 
	Retorna el resultado de aplicar fn al valor inicial y a cada elemento de la tabla.
	Primero llama a la funcion con el valor inicial y el primer elemento de la tabla
	se queda con el resultado de la funcion y la vuelve a llamar con este y el siguiente elemento de la tabla
	hasta recorrerla toda
]]
function Reduce(tbl, fn, init_val,iterator)
	iterator = iterator or ipairs
	local result = init_val;
	for k, v in iterator(tbl) do
		result = fn(result, v,k);
	end
	return result;
end
function Filter (tbl,fn)
	local result = {}
	for k,v in pairs(tbl) do
		if fn(v,k) then
			result[k]=v
		end
	end
	return result
end

-- functional if 
function iif(expr,trueres,falseres)
	if expr then
		return trueres
	else
		return falseres
	end
end

--[[
Receive two tables and calculates adds and removes
invoking the corresponding function
]]
function ProcessDiffs (oldSet, newSet, onAddFn, onRemoveFn, onCmpFunction)
	for id, entity in pairs(newSet) do
		if not oldSet[id] then
			onAddFn(id, entity)
		else
			onCmpFunction(id, oldSet[id], entity)
		end
	end
	for id, entity in pairs(oldSet) do
		if not newSet[id] then
			onRemoveFn(id, entity)
		end
	end
end

function IsArray (t)
	return m_helpers.table:IsArray(t)
end

function TruncateArray (sourceArray, from, to)
	return m_helpers.table:TruncateArray(sourceArray, from, to)
end

function ArrayToHash (array, fieldId, transformKeyFn, transformValueFn)
	return m_helpers.table:ArrayToHash(array, fieldId, transformKeyFn, transformValueFn)
end

--[[
	params
		hash        : (requerido)    lista origen, con la cual se creara un array.
		pFieldId    : (NO requerido) nombre del atributo en el cual se almacenara el indice de la lista origen en el array.
		pResponeType: (NO requerido) (DEFAULT/RAW) el tipo de respuesta en el array.
			DEFAULT     : Por defecto cada item del array es un objeto.
			RAW: Cada item de un array se devuelve sin ninguna modificacion.
]]--
function HashToArray (hash, pFieldId, pResponeType)
	return m_helpers.table:HashToArray(hash, pFieldId, pResponeType)
end

function SortArray (array, MakeSortingValue, desc)
	return m_helpers.table:SortArray(array, MakeSortingValue, desc)
end

--[[
	Return a response for the Search API for various entities

	Response must have the form:
	{
		rows = {...},
		last = "...",
	}

	And postData must have the form:
	{
		limit = n,
	}
]]
function MakeSearchResponse (response, postData, params)
	local parameters = params or {}
	--LogDebug("[UTILS.MakeSearchResponse] 1 - response: %s", json.encode(response))
	-- LogDebug("[UTILS.MakeSearchResponse] 1.1 - response.total_unfiltered_rows: %s", response.total_unfiltered_rows)
	-- LogDebug("[UTILS.MakeSearchResponse] 1.2 - #response.rows: %s", #response.rows)
	-- LogDebug("[UTILS.MakeSearchResponse] 1.3 - response.total_rows: %s", response.total_rows)
	-- LogDebug("[UTILS.MakeSearchResponse] 1.4 - response.last: %s", response.last)
	-- LogDebug("[UTILS.MakeSearchResponse] 2 - postData: %s", json.encode(postData))
	-- LogDebug("[UTILS.MakeSearchResponse] 3 - params.MakeSortingValue: %s", parameters.MakeSortingValue)
	-- LogDebug("[UTILS.MakeSearchResponse] 4 - params.noOrder: %s", parameters.noOrder)
	-- LogDebug("[UTILS.MakeSearchResponse] 5 - params.TransformRowFn: %s", parameters.TransformRowFn)

	local attrId = postData.__attr_id or "Id"
	local startSubProcessTime = os.time()

	if not parameters.noOrder then
		local orderAttrName = postData.__order_attr_name or "Name"
		local MakeSortingValue = parameters.MakeSortingValue
		if not MakeSortingValue then
			MakeSortingValue = function(row)
				--LogDebug("[UTILS.MakeSearchResponse] ---> MakeSortingValue")
				assert(row)
				return (row[orderAttrName] or ""):lower()
			end
		end
		response.rows = SortArray(response.rows, MakeSortingValue, parameters.desc)
	end
	--LogDebug("[UTILS.MakeSearchResponse]      END Subprocess (SortArray) ELAPSED Time '%s' seconds. Total rows: %s", GetElapsedTimeSec(startSubProcessTime), #response.rows)

	if postData.limit then
		local from, to
		if postData.numberPage then
			--LogDebug("[UTILS.MakeSearchResponse]    a. ")
			local numberPage = tonumber(postData.numberPage)
			if numberPage == 1 then
				from = 1
			else
				from = ((numberPage - 1) * postData.limit) + 1
			end
			to = from + (postData.limit - 1)
		elseif postData.last and postData.last ~= "" then
			--LogDebug("[UTILS.MakeSearchResponse]    b. ")
			for index, row in pairs(response.rows) do
				if row[attrId] == postData.last then
					from = index + 1
					break
				end
			end
			if from then
				to = from + postData.limit -1
			else
				to = 0
			end
		else
			--LogDebug("[UTILS.MakeSearchResponse]    c. ")
			to = postData.limit
		end

		--LogDebug("[UTILS.MakeSearchResponse] 2 - postData: %s", json.encode(postData))
		--LogDebug("[UTILS.MakeSearchResponse]    - from: %s", from)
		--LogDebug("[UTILS.MakeSearchResponse]    - to: %s", to)
	
		response.rows = TruncateArray(response.rows, from, to)
		--LogDebug("[UTILS.MakeSearchResponse] 1.1 - response: %s", json.encode(response))
	end

	--LogDebug("[UTILS.MakeSearchResponse] 3 - #response.rows: %s", #response.rows)
	response.last = #response.rows > 0 and response.rows[#response.rows][attrId] or ""
	--LogDebug("[UTILS.MakeSearchResponse] 4 - response.last: %s", response.last)

	if #(postData.columns or {}) > 0 or parameters.TransformRowFn then
		--project columns
		parameters.TransformRowFn = parameters.TransformRowFn or function (_row)
			return _row
		end
		for i, row in ipairs(response.rows) do
			local newRow = {}
			
			row = parameters.TransformRowFn(row)
			for _, columnName in ipairs(postData.columns or {}) do
				newRow[columnName] = row[columnName]
			end
			response.rows[i] = newRow
		end
	end

	startSubProcessTime = os.time()
	--LogDebug("UTILS.MakeSearchResponse]      END Subprocess (ProcessedColumns) ELAPSED Time '%s' seconds.", GetElapsedTimeSec(startSubProcessTime))
	--LogDebug("[UTILS.MakeSearchResponse] 5 - #response.rows: %s", #response.rows)
	--LogDebug("[UTILS.MakeSearchResponse] 4 - response: %s", json.encode(response))

	return response
end

--[[
	Función que ejecuta contra ES una consulta para las Vistas de la App Configuration (Administration, OEConfig, etc)
		Recibe:
			- info: (post request)
			- info.listName: el nombre de lista en el cual retorna ES. (*)
			- instance: instancia del objeto que va a ejecutar ES (ej: Fullcontact.Dialer)
			- searchFnName: nombre de la función de la instance que será ejecutada (ej Fullcontact.Dialer[searchFnName](....))
		Retorna:
			- el formato esperado del lado del cliente para todas las vistas (conf_form_view.js).

		(*) Ej de una respuesta de ES:
				{
				"addresses": [
					{
					"Id": "Phone:+59898100199",
					"Type": "Phone",
					"Key": "vcc1:01blacklistimportid:Phone:+59898100199",
					"VCC": "vcc1",
					"Scope": "vcc1:01blacklistimportid",
					"Channels": [
						"CALL"
					],
					"AddressInfo": {
						"LocalNumber": "098100199",
						"OriginalId": "+59898100199",
						"Number": "59898100199"
					},
					"ImportId": "01blacklistimportid"
					},
					{
					"Id": "Phone:+59898100200",
					"Type": "Phone",
					"Key": "vcc1:01blacklistimportid:Phone:+59898100200",
					"VCC": "vcc1",
					"Scope": "vcc1:01blacklistimportid",
					"Channels": [
						"CALL"
					],
					"AddressInfo": {
						"LocalNumber": "098100200",
						"OriginalId": "+59898100200",
						"Number": "59898100200"
					},
					"ImportId": "01blacklistimportid"
					}
				],
				"last": 2,
				"total": {
					"value": 2,
					"relation": "eq"
				}
			}

		-- Como se puede ver info.listName tiene que coincidir con la lista de datos del array.
]]--
function ExecuteESQueryForConfigurationViews (info, instance, searchFnName)
	assert(info)
	assert(info.listName)
	assert(instance)
	assert(searchFnName)
	--LogDebug("[utils.ExecuteESQueryForConfigurationViews] 1-> instance: %s, searchFnName: %s, info: %s", instance, searchFnName, JSON_Encode(info))

	local vcc = inconcert.authUser.virtualContactCenter
	local rowsPerPage = info.limit
	local start = ""
	if info.numberPage and info.numberPage > 1 then
		start = info.last --rowsPerPage * info.numberPage
	end
	local filter = {
		--["VCC"] = {exact = true, values = {vcc}}
	}
	for name, value in pairs(info.query or {}) do
		if type(value) == "string" or type(value) == "number" then
			value = alltrim(value)
			if value:len() > 0 then
				if not filter[name] then
					filter[name] = {exact = false, values = {value}}
				end
			end
		elseif type(value) == "table" then
			if not filter[name] then
				filter[name] = value
			end
		else
			assert(false, "ExecuteESQueryForConfigurationViews - unsupported type " .. tostring(type(value)) .. " in value " .. tostring(value))
		end
	end
	local sort = info.sort --{{Name = {order = "asc"}}}
	local params = {
		filter      = filter,
		startRow    = start,
		rowsPerPage = rowsPerPage,
		vcc         = vcc,
		sort        = sort
	}
--	SAMPLE: Fullcontact.Dialer:Search(params)
	local success, data = instance[searchFnName](instance, params)
	--LogDebug("[utils.ExecuteESQueryForConfigurationViews] 2 -> success: %s, data: %s", success, JSON_Encode(data))
	if not success then
		return false, data
	end
	data.rows = {}
	if not info.columns then
		data.rows = data[info.listName]
	else
		--project columns
		for i, row in pairs(data[info.listName]) do
			local newRow = {}
			for _, columnName in ipairs(info.columns) do
				newRow[columnName] = row[columnName]
			end
			table.insert(data.rows, newRow)
		end
	end
	data[info.listName] = nil
	if type(data.total) then
		data.total_rows = data.total.value
		data.total = nil
	end
	--LogDebug("[utils.ExecuteESQueryForConfigurationViews] 3 -> RESPONSE data: %s", JSON_Encode(data))
	return true, data
end

--[[
	Match entity's fields against
	a query table, taking in account pagination 
	
	postData must have the form:
	{
		last = "xxx",
		query = {
			field1 = "value1",
			field2 = "value2",
			...
		}
	}
]]

local function GetNameOrId (entity)
	if entity.Name and entity.Name ~= '' then
		return entity.Name
	elseif entity.name and entity.name ~= '' then
		return entity.name
	elseif entity.Id and entity.Id ~= '' then
		return entity.Id
	elseif entity.id and entity.id ~= '' then
		return entity.id
	end
	error("entity hasn't Name or Id")
end

function MatchFilter (entity, postData, logicalOperator, excludeColumns, customMatchFilter)
	logicalOperator = logicalOperator and logicalOperator:upper() or "OR"
	local filter = postData.query
	excludeColumns = excludeColumns or {}

	--LogDebug("[utils.MatchFilter]-> logicalOperator: %s", logicalOperator)
	--LogDebug("[utils.MatchFilter]-> filter: %s", filter)

	if type(filter) ~= "table" then
		--LogDebug("[utils.MatchFilter]  -> MATCH - NO FILTER")
		return true
	end
	--LogDebug("[utils.MatchFilter]-> entity: %s", json.encode(entity))
	--LogDebug("[utils.MatchFilter]-> filter: %s", json.encode(filter))

	local function CanFilterColumn (columnName)
		if excludeColumns[columnName] then
			--LogDebug("[UTILS:MatchFilter] (CanFilterColumn) NO filtro por la columna - excludeColumns.columnName: '%s'.", columnName)
			return false
		end
		return true
	end

	local function Match (entityValue, filterValue, fieldId, row)
		if type(customMatchFilter) == "function" then
			local isOk, customMatch = customMatchFilter(filterValue, fieldId, row)
			if isOk ~= nil then
				return isOk, customMatch
			end
		end
		entityValue = entityValue or ""
		local startPos, endPos = tostring(entityValue):lower():find(filterValue, nil, true)
		local isOk = startPos and true or false
		return isOk, startPos, endPos
	end

	local mustInclude = false
	--LogDebug("[utils.MatchFilter]  ---------------------- filter: %s", json.encode(filter))
	for filterName, filterValue in pairs(filter) do
		--LogDebug("[utils.MatchFilter]  ---------------------- filterName: %s", filterName)
		local canFilterColumn = CanFilterColumn(filterName)
		if canFilterColumn then
			filterValue = tostring(filterValue):lower()

			local entityValue = entity[filterName]
			if type(entityValue) == "table" then
				entityValue = JSON_Encode(entityValue)
			end
			--LogDebug("--")
			--LogDebug("[utils.MatchFilter]  -> filterName: %s", filterName)
			--LogDebug("[utils.MatchFilter]  -> filterValue: %s", filterValue)
			--LogDebug("[utils.MatchFilter]  -> entityValue: %s", entityValue)
	
			--if entityValue then
				--local match = tostring(entityValue):lower():match(filterValue)
				local canInclude, startPos, endPos = Match(entityValue, filterValue, filterName, entity)
				--LogDebug("[utils.MatchFilter]  -> canInclude '%s' startPos: '%s' endPos: '%s' logicalOperator: '%s'", canInclude, startPos, endPos, logicalOperator)
				--LogDebug("1    ..")
				if canInclude then
					--LogDebug("1.1    ..")
					mustInclude = true
					-- SI Machea
					if logicalOperator == "OR" then
						-- OR OPERATOR
						--LogDebug("[utils.MatchFilter]  -> SIIII MATCH - logicalOperator OR")
						break
					--else
						-- AND OPERATOR -
						-- continua
					end
				else
					--LogDebug("1.2    ..")
					-- NO Machea
					if logicalOperator == "AND" then
						-- OR OPERATOR
						--LogDebug("[utils.MatchFilter]  -> NO MATCH - logicalOperator AND")
						mustInclude = false
						break
					--else
						-- AND OPERATOR -
						-- continua
					end
				end
				--LogDebug("2    ..")
			--end
		end
	end
	--LogDebug("")
	-- NO Machea
	--LogDebug("[utils.MatchFilter]  -> mustInclude: %s - logicalOperator: %s", mustInclude, logicalOperator)
	return mustInclude
end

function TreeFilter (row, treeQuery)
	assert(type(row) == "table")
	assert(type(treeQuery) == "table")

	--LogDebug("")
	--LogDebug("[utils.TreeFilter] Start Tree Filter ...")
	--LogDebug("[utils.TreeFilter]   row: %s", json.encode(row))

	local excludeFilterColumns = {items = true, childs = true}

	local function RecursiveFilter (row, nodeTreeFilter)
		--LogDebug("[utils.TreeFilter] 0 (RecursiveFilter) -------------------------")
		if not nodeTreeFilter then
			--LogDebug("[utils.TreeFilter] 1 (RecursiveFilter) - NO nodeTreeFilter")
			return true
		end
		local logicalOperator = nodeTreeFilter.operator or "OR"
		local mustAdd = false
		for _, filterItem in pairs(nodeTreeFilter.items or {}) do
			local filterInfo = {
				query = filterItem or {}
			}
			--LogDebug("[utils.TreeFilter] 2 (RecursiveFilter) - filterInfo: %s", json.encode(filterInfo))
			local isMatch = MatchFilter(row, filterInfo, logicalOperator, excludeFilterColumns)
			if isMatch then
				--LogDebug("[utils.TreeFilter] 3 (RecursiveFilter) - SI Match invoco a la fn recursiva logicalOperator: %s", logicalOperator)
				mustAdd = RecursiveFilter(row, filterItem.childs)
				if mustAdd then
				--	mustAdd = true
					--LogDebug("[utils.TreeFilter] 3.1 (RecursiveFilter) - SI Match la fn recursiva logicalOperator: %s", logicalOperator)
					if logicalOperator == "OR" then
						--LogDebug("   3.1.1 (RecursiveFilter) - break logicalOperator: %s", logicalOperator)
						break
					end
				else
				--	mustAdd = false
					--LogDebug("[utils.TreeFilter] 3.2 (RecursiveFilter) - NO Match la fn recursiva logicalOperator: %s", logicalOperator)
					if logicalOperator == "AND" then
						--LogDebug("[utils.TreeFilter]   3.2.1 (RecursiveFilter) - break logicalOperator: %s", logicalOperator)
						break
					end
				end
			else
				--LogDebug("[utils.TreeFilter] 4 (RecursiveFilter) - NO Match logicalOperator: %s", logicalOperator)
				mustAdd = false
				if logicalOperator == "AND" then
					--LogDebug("[utils.TreeFilter]   4.1 (RecursiveFilter) - break logicalOperator: %s", logicalOperator)
					break
				end
			end
		end
		--LogDebug("[utils.TreeFilter] 5 (RecursiveFilter) ----------> Me voy - mustAdd %s logicalOperator: %s", mustAdd, logicalOperator)
		return mustAdd
	end

	return RecursiveFilter(row, treeQuery)
end

function GetRowCount (rows)
	return m_helpers.table:GetRowCount(rows)
end

function Cardinality (t)
	local count = 0
	for k in pairs(t or {}) do
		count = count + 1
	end
	return count
end

function Id_n (id, n)
	if n == 1 then
		return id
	else
		return id.."_"..n
	end
end

-- Helper function to know if an interaction is a live one
function InteractionIsLive (int_type)
	return int_type == "CALL" or int_type == "IM" or int_type == "CHAT"
end

--[[
	Process agent status and determine his campaign state
	Possible states are:
	BUSY: when agent has an interaction (it doesn't mather if agent is not logged)
	BUSY_OTHER_CAMPAIGN: when agent has an interaction in other campaign
	BREAK: agent has no interactions and isn't available
	OFFLINE: agent is offline
	IDLE: None of the previous states
]]
function ProcessCampaignState (status, campaignId)
	if not status.state or status.state == "" or status.state == "OFFLINE" then
		return "OFFLINE"
	end
	if status.interactions then
		local candidate_state = nil
		for id, interaction in pairs(status.interactions) do
			if InteractionIsLive(interaction.type) then
				if interaction.campaign == campaignId then
					candidate_state = "BUSY"
					break
				end
				candidate_state = "BUSY_OTHER_CAMPAIGN"
			end
		end
		if candidate_state then
			return candidate_state
		end
	end
	if status.campaigns and status.campaigns[campaignId] == false then
		return "BREAK"
	end
	--by default return idle
	return "IDLE"
end

--[[
	recursively compare two tables.
	returns true if they are equal
	if different returns false.
]]
function StructuralCompare (v1, v2, CanValidateFn, current_attr)
	return m_helpers.table:StructuralCompare(v1, v2, CanValidateFn, current_attr)
end

--=============================================================================
--[[
	Funcciones de fechas - conversiones de UTC a Timezone y viceversa.
--]]

function ConvertHourToUTCHour (stringHour, formatDate, timezone)
	return m_helpers.date:ConvertHourToUTCHour(stringHour, formatDate, timezone)
end

function ConvertUTCHourToHour (stringHour, formatDate, timezone)
	return m_helpers.date:ConvertUTCHourToHour(stringHour, formatDate, timezone)
end

--=============================================================================
--[[
	Esta función convierte a UTC la fecha que se encuentra en el timezone del Usuario.
	La misma invoca a la función "IsEnabledDateBetweenDaylightSavingTime" para determinar si debe o no ajustar la fecha resultante al Horario de Verano.
--]]
function ConvertDateToUTCDate (stringDate, stringHour, formatDate, isFullDate, timezone)
	return m_helpers.date:ConvertDateToUTCDate(stringDate, stringHour, formatDate, isFullDate, timezone)
end

--=============================================================================
--[[
	Esta función convierte al Timezone del Usuario la fecha que se encuentra en UTC.
	La misma invoca a la función "IsEnabledDateBetweenDaylightSavingTime" para determinar si debe o no ajustar la fecha resultante al Horario de Verano.
--]]
function ConvertUTCDateToDate (stringDate, stringHour, formatDate, isFullDate, timezone)
	return m_helpers.date:ConvertUTCDateToDate(stringDate, stringHour, formatDate, isFullDate, timezone)
end

function DeepCopy (t)
	return m_helpers.table:DeepCopy(t)
end

--[[
	Finds an unused TCP port
--]]
function get_free_port()
	local port
	local free = false
	while not free do
		port = math.random(2000,30000)
		local file = io.popen('netstat -an|grep ":'.. port .. ' "', "r")
		free = string.len(file:read("*a") or "") == 0
	end
	return port
end

-- ( START ) EMAIL VALIDATION ---------------------------

--- Email regexes table
local emails_regexp = {
	parse    = icuregex.compile([[^\s*("<[^<>]*>"\s+|[^<>]*\s+)?(<[^<>\s]*>|[^<>\s]*)$]]),
	validate = icuregex.compile([[^((([a-z]|\d|[\\!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+(\.([a-z]|\d|[\\!#\$%&'\*\+\-\/=\?\^_`{\|}~]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])+)*)|((\x22)((((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(([\x01-\x08\x0b\x0c\x0e-\x1f\x7f]|\x21|[\x23-\x5b]|[\x5d-\x7e]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(\\([\x01-\x09\x0b\x0c\x0d-\x7f]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))))*(((\x20|\x09)*(\x0d\x0a))?(\x20|\x09)+)?(\x22)))@((([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|\d|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.)+(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])|(([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])([a-z]|\d|-|\.|_|~|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])*([a-z]|[\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])))\.?$]], "i")
}

--- Validate email address against regex.
function ValidateEmail (address)
	if type(address) == "string" then address = ustring(address) end
	for icumatch in icuregex.gmatch(emails_regexp.validate, address) do
		return true
	end
	return false
end

--- this function extracts the mail address (ellio@hotmail.com) from complete address ("Ellio de Mello" <ellio@hotmail.com>)
-- It also returns the Name, if any
--
local function Extract (address)
	local extractedAddress = address
	local extractedName
	for icumatch in icuregex.gmatch(emails_regexp.parse, ustring(address)) do
		if icumatch[1] == false then
			extractedAddress = tostring(icumatch.value)
		else
			extractedName = tostring(icumatch[1].value)
			extractedAddress = tostring(icumatch[2].value)
		end
		break
	end
	-- remove optional < > from mail address
	extractedAddress = extractedAddress and extractedAddress:match("^<?([^>]+)>?$")
	if extractedName then
		extractedName = alltrim(extractedName)
		extractedName = extractedName:match([[^["']?([^"']+)["']?$]])
	end
	extractedAddress = extractedAddress and extractedAddress:match("^<?([^>]+)>?$")
	return extractedAddress, extractedName
end

--- this function extracts the email address and verifies
local function Validate (completeAddress)
	completeAddress = alltrim(completeAddress)
	local address = Extract(completeAddress)
	if not ValidateEmail(address) then
		LogWarning("Invalid email address: '%s'", completeAddress)
		return false
	end
	return true
end
--- Parses string mail lists and return a table with the parsed values.
--
-- @tparam string addressList
-- @treturn table an array of addresses
function ParseMailList (addressList)
	if not addressList or #addressList == 0 then
		return {}
	end

	local values = {}
	if string.find(addressList, ";", nil, true) == nil then
		if Validate(addressList) then
			table.insert(values, addressList)
		end
		return values
	end

	local address
	local start = 1
	local position = 1
	local quotes = 0
	local length = #addressList

	-- parse the mail list
	while position <= length do
		local char = string.sub(addressList, position, position)

		if char == ';' then
			if quotes == 0 or quotes == 2 then
				address = string.sub(addressList, start, position -1)
				if Validate(address) then
					table.insert(values, address)
				end
				start = position + 1
				quotes = 0
				if position == length then
					break
				end
			elseif quotes > 2 then
				address = string.sub(addressList, start, position -1)
				LogWarning("Invalid email address: '%s'", address)
				start = position + 1
				quotes = 0
				if position == length then
					break
				end
			end
		elseif char == '"' then
			quotes = quotes + 1
		end

		if position == length then
			if quotes ~= 0 and quotes ~= 2 then
				address = string.sub(addressList, start, position -1)
				LogWarning("Invalid email address: '%s'", address)
			else
				address = string.sub(addressList, start, position)
				if Validate(address) then
					table.insert(values, address)
				end
			end
		end

		position = position + 1
	end

	return values
end
-- ( END ) EMAIL VALIDATION ---------------------------

function CanRunJob (redis, jobId, intervalSeconds)
	local now = os.time()
	local lastRun = redis:get("LastJobExecution:" .. jobId)
	lastRun = lastRun and lastRun ~= "" and tonumber(lastRun) or 0
	if lastRun > 0 and now - lastRun < intervalSeconds - 5 then
		--LogInfo("Job %s was already executed at %d", jobId, lastRun)
		return false
	end
	redis:set("LastJobExecution:" .. jobId, now)
	return true
end

function CanRunRealtimeJob (redis, jobId, timeout, autoStart)
	local mustAutoStart = true
	if type(autoStart) == "boolean" then
		mustAutoStart = autoStart
	end
	local now = os.time()
	local lastRun = redis:get("LastJobExecution:" .. jobId)
	lastRun = lastRun and lastRun ~= "" and tonumber(lastRun) or 0
	
	--LogDebug("[utils.CanRunRealtimeJob]---> lastRun %s, now - lastRun %s, lastRun %s", lastRun, now - lastRun, timeout)
	if lastRun > 0 and now - lastRun < timeout then
		--LogInfo("Job %s was already executed at %d", jobId, lastRun)
		return false
	end

	if mustAutoStart then
		StartRealtimeJob(redis, jobId)
	end

	return true
end

function StartRealtimeJob (redis, jobId)
	local now = os.time()
	--LogDebug("[utils.StartRealtimeJob]---> now %s jobId %s", now, jobId)
	redis:set("LastJobExecution:" .. jobId, now)
end

function EndRealtimeJob (redis, jobId)
	--LogDebug("[utils.EndRealtimeJob]---> jobId %s", jobId)
	redis:del("LastJobExecution:" .. jobId)
end

--[[
	Estas funciones validan la conexion a internet.
	Realiza la validacion contra una URL conocida "www.google.com".
	Al resultado lo almacena en Redis, para que un siguiente request no tenga que realizar la validacion nuevamente.
	La validacion expira en unos 2 minutos (EXPIRED_ELAPSED_TIME_SEC)
]]--

local EXPIRED_ELAPSED_TIME_SEC = 120 -- 2 MINUTOS
local m_cachedLastTestInternetConnection = nil

--[[
	Obtiene desde Redis el resultado de la ultima validacion que se ha almacenado.
]]--
local function GetLastTestInternetConnection (redis)
	assert(redis)
	local lastTestInternetConnection = redis:get("LastTestInternetConnection")
	if not lastTestInternetConnection then
		lastTestInternetConnection = {success = false, err = false, lastTestTime = nil}
	else
		lastTestInternetConnection = json.decode(lastTestInternetConnection)
	end
	return lastTestInternetConnection
end

--[[
	Almacena en Redis el resultado de la ultima validacion.
]]--
local function SetLastTestInternetConnection (redis, lastTestInternetConnection)
	assert(redis)
	m_cachedLastTestInternetConnection = lastTestInternetConnection
	local encodedLastTestInternetConnection = json.encode(lastTestInternetConnection)
	redis:set("LastTestInternetConnection", encodedLastTestInternetConnection)
end

--[[
	Realiza una vaildacion para saber si es posible realizar la validacion de conexion a internet.
	Para que la validacion se pueda realizar, se pueda dar por 2 motivos:
	1 - Si quien invoca a la funcion fuerza la acción.
	2 - Si el tiempo de la ultima validacion ha expirado (EXPIRED_ELAPSED_TIME_SEC).
]]--
local function MustTestInternetConnection (redis, force)
	local mustTest = false
	local lastTestInternetConnection

	if not m_cachedLastTestInternetConnection then
		lastTestInternetConnection = GetLastTestInternetConnection(redis)
	else
		lastTestInternetConnection = m_cachedLastTestInternetConnection
	end

	local now = os.time()

	if lastTestInternetConnection.lastTestTime then
		local difference = now - lastTestInternetConnection.lastTestTime
		--LogDebug("----> 3 MustTestInternetConnection force: %s, difference: %s", force, difference)
		if force or difference > EXPIRED_ELAPSED_TIME_SEC then
			lastTestInternetConnection.lastTestTime = now
			mustTest = true
		end
	else
		lastTestInternetConnection.lastTestTime = now
		mustTest = true
	end

	--LogDebug("--------> 1 MustTestInternetConnection now: %s", json.encode(lastTestInternetConnection))

	return mustTest, lastTestInternetConnection
end

--[[
	Funcion publica que puede ser invocada para forzar la actualizacion.
]]--
function UpdateTestInternetConnection (redis, success, err)
	local _, lastTest = MustTestInternetConnection(redis)
	lastTest.success = success
	lastTest.err = err
	SetLastTestInternetConnection(redis, lastTest)
end

function TestInternetConnection (redis, force)
	local mustTest, lastTest = MustTestInternetConnection(redis, force)
	if not mustTest then
		--LogDebug("[UTILS:TestInternetConnection] Not a Must Test Internet Connection")
		SetLastTestInternetConnection(redis, lastTest)
		return lastTest.success
	end

	--LogDebug("[UTILS:TestInternetConnection] Testing Internet Connection")

	local connection = socket.tcp()
	connection:settimeout(1000)
	local success, err = connection:connect("www.google.com", 80) -- nos conectamos a una URL conocida para validar la conexion a internet
	connection:close()

	if not success then
		lastTest.success = false
		lastTest.err = err
		SetLastTestInternetConnection(redis, lastTest)
		LogWarning("[UTILS:TestInternetConnection] Error Internet Connection. '%s'", err)
		return lastTest.success
	end

	lastTest.success = true
	lastTest.err = nil
	SetLastTestInternetConnection(redis, lastTest)

	return lastTest.success
end

--------------------------------------------------------------------------
-- CSV
--------------------------------------------------------------------------
function ParseCSVHeader (csv)
	return m_helpers.csv.ParseCSVHeader(csv)
end

function Csv2Table (csv, ProcessColumn)
	return m_helpers.csv.Csv2Table(csv, ProcessColumn)
end

function Table2Csv (tableRows, allowedColumns, excludeColumns, separator, orderHeaders)
	return m_helpers.csv.Table2Csv(tableRows, allowedColumns, excludeColumns, separator, orderHeaders)
end

function ParseLine (text, fieldSeparator)
	return m_helpers.csv.ParseCSVLine(text, fieldSeparator)
end

function ParseWebQueryString (text)
	--LogWarning("[UTILS:ParseWebQueryString] 1 - text '%s'", text)
	local parsedLine = ParseLine(text, "&")
	--LogWarning("[UTILS:ParseWebQueryString] 2 - parsedLine '%s'", JSON_Encode(parsedLine))
	local fieldValues = {}
	for _, fieldValue in ipairs(parsedLine) do
		local parsedFieldValue = ParseLine(fieldValue, "=")
		--LogWarning("[UTILS:ParseWebQueryString]   3 - parsedFieldValue '%s'", JSON_Encode(parsedFieldValue))
		local name = parsedFieldValue[1]
		local value = parsedFieldValue[2]
		fieldValues[name] = value
	end
	--LogWarning("[UTILS:ParseWebQueryString] 4 - fieldValues '%s'", JSON_Encode(fieldValues))
	return fieldValues
end

-------------------------------------------------------------------------------------------------------------

function AvailableUserCapabilityForChannel (user, channel, fullCampaignId, models)
	assert(type(user) == "table")
	assert(user.Id)
	assert(user.VCC)
	assert(channel)
	assert(fullCampaignId)
	assert(models)
	local fullUserId = AppendVcc(user.Id, user.VCC)
	local agentRedis = GetAgentRedisClient()

	--LogInfo("[UTILS:AvailableUserCapabilityForChannel] Validating User '%s' Available Capability for Channel '%s', campaign: '%s'.", fullUserId, channel, fullCampaignId)

	local keyspace = GetDatasource():CassandraConnection("Agent")
	local simpleUserId = user.Id
	local simpleCampaignId = GetId(fullCampaignId)

	local available_slots = -1
	-- 1 Busco si hay capacidad definida para el VCC. Si la hay, ningun member la debe sobrepasar
	local capabilityCurrentVCC = models.Channels.GetInteractionTypeByChannelCurrentVCC(channel, keyspace)
	user.Capability = DeepCopy(capabilityCurrentVCC)
	if user.Capability and tonumber(user.Capability.attentionCapability) >= 0 then
		user.BusySlots = tonumber(agentRedis:get("AgentService::AgentBusySlots:" .. channel .. ":" .. fullUserId) or "0")
		available_slots = user.Capability.attentionCapability - user.BusySlots
		if available_slots <= 0 then
			LogWarning("[UTILS:AvailableUserCapabilityForChannel] User '%s' Has NOT Available Capability for Channel '%s' on VCC '%s'", fullUserId, channel, user.VCC)
			return false	--VCC capability exceeded for channel 
		end
	end
	
	-- 2 Busco si hay capacidad definida para el member de la campaign, en cuyo caso no se debe sobrepasar
	local capabilityMember = models.Channels.GetInteractionTypeByChannelCampaignMember(channel, simpleCampaignId, simpleUserId, keyspace)
	user.Capability = DeepCopy(capabilityMember)
	if user.Capability and tonumber(user.Capability.attentionCapability) >= 0 then
		user.BusySlots = tonumber(agentRedis:hget("AgentService::AgentBusySlotsOnCampaign:"..fullUserId..":"..fullCampaignId, channel) or "0")
		available_slots = user.Capability.attentionCapability - user.BusySlots
		if available_slots <= 0 then
			LogWarning("[UTILS:AvailableUserCapabilityForChannel] User '%s' Has NOT Available Capability for Channel '%s' on campaign '%s'", fullUserId, channel, fullCampaignId)
			return false	--Member capability exceeded for channel 
		else
			return true
		end
	end
	
	-- 3 Busco si hay capacidad definida para la campaign, en cuyo caso no se debe sobrepasar
	local capabilityCampaign = models.Channels.GetInteractionTypeByChannelCampaign(channel, simpleCampaignId, keyspace)
	user.Capability = DeepCopy(capabilityCampaign)
	if user.Capability and tonumber(user.Capability.attentionCapability) >= 0 then
		user.BusySlots = tonumber(agentRedis:hget("AgentService::AgentBusySlotsOnCampaign:"..fullUserId..":"..fullCampaignId, channel) or "0")
		available_slots = user.Capability.attentionCapability - user.BusySlots
		if available_slots <= 0 then
			LogWarning("[UTILS:AvailableUserCapabilityForChannel] Campaign '%s' Has NOT Available Capability for Channel '%s'", fullCampaignId, channel)
			return false	--Campaign capability exceeded for channel 
		end
	end
	
	-- 4 me quedo con la capacidad de system por defecto
	local capabilitySystemVCC = models.Channels.GetInteractionTypeByChannelSystemVCC(channel, keyspace)
	user.Capability = DeepCopy(capabilitySystemVCC)
	return true
	
end

function ParseWebDesignerURL (url)
	--print("      URL: " .. tostring(url))
	if not url then 
		return
	end
	local parsedURL = {}

	-- PROTOCOL
	local startPos, endPos = string.find(url, ":")
	if not endPos then
		return
	end
	parsedURL.protocol = string.sub(url, 1, endPos - 1)
	--print("PROTOCOL: " .. parsedURL.protocol)

	-- SERVER ADDRESS
	local startPos = endPos + 3
	local _, endPos = string.find(url, "/", startPos)
	if not endPos then
		return
	end
	parsedURL.serverAddress = string.sub(url, startPos, endPos - 1)
	--print("SERVER ADDRESS: " .. parsedURL.serverAddress)

	-- DOMAIN
	local startPos = endPos
	local _, endPos = string.find(url, "webdesigner", startPos)
	if not endPos then
		return
	end
	parsedURL.domain = string.sub(url, startPos, endPos)
	--print("DOMAIN: " .. parsedURL.domain)

	-- APP NAME
	local startPos = endPos + 2
	parsedURL.appName = string.sub(url, startPos)
	--print("APP NAME: " .. parsedURL.appName)

	--print("      Parsed URL: " .. json.encode(parsedURL))

	return parsedURL
end

--[[
	parameters:
		listString = "uno|dos|tres"
		separator = "|"
		get_value_fn = optional function
	return table:
		["uno","dos","tres"] or [{value:"uno"},{value:"dos"},{value:"tres"}]
]]--
function ParseList (listString, separator, get_value_fn)
	assert(type(listString) == "string")
	assert(type(separator) == "string")
	GetValue = type(get_value_fn) == "function" and get_value_fn or function(s) return s end

	local list = {}
	local function Insert (value)
		local newValue = GetValue(value)
		if not newValue then
			return
		end
		table.insert(list, newValue)
	end

	local startPosition = 1
	local endPosition = nil

	while true do
		endPosition = string.find(listString, separator, startPosition, endPosition, true)
		if not endPosition then
			local value = string.sub(listString, startPosition)
			Insert(value)
			break
		end
		local value = string.sub(listString, startPosition, endPosition -1)
		Insert(value)
		startPosition = endPosition + 1
		endPosition = nil
	end

	return list
end

function ShowMemory (msg)
	m_helpers.utils.Memory.Show(msg)
end

function CollectAllMemory ()
	m_helpers.utils.Memory.CollectAll()
end

--[[ -------------------------------------------------------------
	jQuery y jQueryUI definiciones
]]--
local function GetJQueryDefinition (minimized)
	local jQueryDefinitions = {
		-- last version
		["3.7.1"] = {
			jsFile = minimized and
				"/inconcert/lib/jquery/jquery-3.7.1.min.js" or
				"/inconcert/lib/jquery/jquery-3.7.1.js"
		},
		-- other versions
		["3.6.0"] = {
			jsFile = minimized and
				"/inconcert/lib/jquery/jquery-3.6.0.min.js" or
				"/inconcert/lib/jquery/jquery-3.6.0.js"
		},
		["3.4.1"] = {
			jsFile = minimized and
				"/inconcert/lib/jquery/jquery-3.4.1.min.js" or
				"/inconcert/lib/jquery/jquery-3.4.1.js"
		},
		["1.9.1"] = {
			jsFile = minimized and
				"/inconcert/lib/jquery/jquery-1.9.1.min.js" or
				"/inconcert/lib/jquery/jquery-1.9.1.js"
		},
		["1.3.2"] = { -- version utilizada por el frm "changeAvatar.html" (el plugin jquery.Jcrop.js, necesita esta versión)
			jsFile = minimized and
				"/inconcert/lib/jquery/jquery-1.3.2.min.js" or
				"/inconcert/lib/jquery/jquery-1.3.2.js"
		}
	}
	local uiDefinitions = {
		-- last version
		["1.13.2"] = {
			jsFile = (minimized and
				"/inconcert/lib/jquery-ui-1.13.2/jquery-ui.min.js" or
				"/inconcert/lib/jquery-ui-1.13.2/jquery-ui.js"),
			cssFile = (minimized and
				"/inconcert/lib/jquery-ui-1.13.2/jquery-ui.min.css" or
				"/inconcert/lib/jquery-ui-1.13.2/jquery-ui.css")
		},
		-- other versions
		["1.13.0"] = {
			jsFile = (minimized and
				"/inconcert/lib/jquery-ui-1.13.0/jquery-ui.min.js" or
				"/inconcert/lib/jquery-ui-1.13.0/jquery-ui.js"),
			cssFile = (minimized and
				"/inconcert/lib/jquery-ui-1.13.0/jquery-ui.min.css" or
				"/inconcert/lib/jquery-ui-1.13.0/jquery-ui.css")
		},
		["1.12.1"] = {
			jsFile = (minimized and
				"/inconcert/lib/jquery-ui-1.12.1/jquery-ui.min.js" or
				"/inconcert/lib/jquery-ui-1.12.1/jquery-ui.js"),
			cssFile = (minimized and
				"/inconcert/lib/jquery-ui-1.12.1/jquery-ui.min.css" or
				"/inconcert/lib/jquery-ui-1.12.1/jquery-ui.css")
		},
		["1.10.2"] = {
			jsFile = (minimized and
				"/inconcert/lib/jquery-ui-1.10.2/js/minified/jquery-ui.min.js" or
				"/inconcert/lib/jquery-ui-1.10.2/js/jquery-ui.js"),
			cssFile = (minimized and
				"/inconcert/lib/jquery-ui-1.10.2/themes/base/minified/jquery-ui.min.css" or
				"/inconcert/lib/jquery-ui-1.10.2/themes/base/jquery-ui.css")
		}
	}
	return jQueryDefinitions, uiDefinitions
end

function jQueryVersion (versionNumber, uiVersionNumber)
	local i6_type = os.getenv("INCONCERT_ENV_MODE")
	local is_production = i6_type == "PRODUCTION"
	local minimized = is_production and "MINIMIZED" or "NORMAL"

	--LogInfo("[APPS.UTILS:jQuery] ----- Loading Versions - jQuery '%s' & jQuery UI: '%s'. File types: '%s'. -----", versionNumber, uiVersionNumber, minimized)

	local jQueryDefinitions, jQueryUIDefinitions = GetJQueryDefinition(is_production)

	local jQueryVersionInfo = jQueryDefinitions[versionNumber]
	if not jQueryVersionInfo then
		LogWarning("[APPS.UTILS:jQuery] Unrecognized jQuery version '%s'", versionNumber)
	end

	local jQueryUIVersionInfo = jQueryUIDefinitions[uiVersionNumber]
	if not jQueryUIVersionInfo then
		LogWarning("[APPS.UTILS:jQueryVersion] Unrecognized jQuery UI version '%s'", uiVersionNumber)
	end

	return jQueryVersionInfo, jQueryUIVersionInfo
end

--[[-----------------------------------------------------------------------------------------------------------
	Estas son las versiones de hemos utilizado desde que hemos comenzado con i6
		jQuery   : 1.9.1  (Feb 04, 2013)
		jQuery UI: 1.10.2 (Mar 14, 2013)
	Estas son las versiones ESTABLES al 15/10/2021.
		jQuery   : 3.4.1  (May 01, 2019)
		jQuery UI: 1.12.1 (Sep 14, 2016)
		--
		jQuery   : 3.6.0  (Mar 02, 2021)
		jQuery UI: 1.13.0 (Oct 07, 2021)
-------------------------------------------------------------------------------------------------------------]]
function jQueryStableVersion ()
	local stableVersion = {
		jQuery = "3.6.0",
		jQueryUI = "1.13.2"
	}
	-- local stableVersion = {
	-- 	jQuery = "1.9.1",
	-- 	jQueryUI = "1.10.2"
	-- }
	--LogInfo("[APPS.UTILS:jQueryStableVersion] ###### LOADING STABLE VERSIONS ######")
	return jQueryVersion(stableVersion.jQuery, stableVersion.jQueryUI)
end

--[[-----------------------------------------------------------------------------------------------------------
	Estas son las últimas versiones publicadas de jQuery
		jQuery   : 3.7.1  (2023-08-28T13:37Z)
		jQuery UI: 1.13.2 (2022-07-14)
-------------------------------------------------------------------------------------------------------------]]
function jQueryLastVersion ()
	local lastVersions = {
		jQuery = "3.7.1",
		jQueryUI = "1.13.2"
	}
	--LogInfo("[APPS.UTILS:jQueryLastVersion] ***** LOADING LAST VERSIONS *****")
	return jQueryVersion(lastVersions.jQuery, lastVersions.jQueryUI)
end

-------------------------------------------------------------------------------------------------------------
--
-------------------------------------------------------------------------------------------------------------
function jQuery (version)
	if type(version) == "string" then
		version = version:upper()
		if version == "LAST" then
			return jQueryLastVersion()
		elseif version == "STABLE" then
			return jQueryStableVersion()
		else
			assert(false, "Unrecognized jQuery type: '" .. version .. "'")
		end
	elseif type(version) == "table" then
		return jQueryVersion(version.jQuery, version.jQueryUI)
	else
		assert(false, "Unrecognized jQuery type: '" .. type(version) .. "'")
	end
end

function GetElapsedTimeSec (start)
	return m_helpers.utils.ElapsedTime.GetElapsedTimeSec(start)
end

--[[
	Función utilizada desde la rocka "reporter"
]]--
function GetAgentService ()
	local agentservice = require "apps.agent.agentservice"
	return agentservice
end

--[[
	Función que genera el prefijo saliente de discado para cuentas de Teli6 (vccs sin integracion con Allegro)
	Obtiene de redis el último existente y le incrementa 1.
	Retorna el número con un sufijo.
	Ej: 26*
]]--
function MakeOutgoingPrefixForCalli6 (vcc)
	if HasAllegroIntegration(vcc) then
		return
	end
	local agentRedis = GetAgentRedisClient()
	local newpPrefix = agentRedis:incr('Agent::CampaignPrefixes')
	-- prefix should end with a number sing in order to be easily parsed
	local outboundPrefix = ("%s*"):format(newpPrefix)
	return outboundPrefix
end

function RandomNumber(length)
	local result = ""
	for i = 1, length do
		result = result .. math.random(0, 9)
	end
	return result
end

function ConvertValueToString (sourceValue, tableFieldSeparator)
	--assert(sourceValue)
	tableFieldSeparator = tableFieldSeparator or ","
	local sourceType = type(sourceValue)

	--LogDebug("[APPS.UTILS:ConvertValueToString] 1 - sourceValue: '%s', sourceType: '%s', tableFieldSeparator: '%s'.", sourceValue, sourceType, tableFieldSeparator)

	local stringValue = ""
	if sourceType == "table" then
		if TableIsEmpty(sourceValue) then
			return ""
		end
		if #sourceValue > 0 then
			stringValue = table.concat(sourceValue, tableFieldSeparator)
		else
			stringValue = JSON_Encode(sourceValue)
		end
	else
		stringValue = tostring(sourceValue)
	end

	--LogDebug("[APPS.UTILS:ConvertValueToString] 2 - sourceValue: '%s', sourceType: '%s', stringValue: '%s'.", sourceValue, sourceType, stringValue)

	return stringValue
end

--[[
	Valida que el VCC tenga configurado el NLU LUIS o de RASA y el mismo se encuentre activo.
]]--
function IsAvailableNLUInVCC (nluType, vcc, keyspace)
	vcc = vcc or inconcert.authUser.virtualContactCenter
	assert(nluType)
	assert(vcc)
	nluType = nluType:upper()
	local allowedNLUTypes = {LUIS = true, RASA = true}
	assert(allowedNLUTypes[nluType])

	local vccInfo = GetAllColumns {
		keyspace = keyspace or "Agent",
		columnFamily = "VCCs",
		key = vcc,
		columns = {"BotInfo"}
	}

	--LogDebug("[UTILS:IsAvailableNLUInVCC] ------------>    Is Available %s for vcc '%s'. vccInfo '%s'.", vcc, nluType, json.encode(vccInfo or {}))

	local vccBotInfo = (vccInfo or {}).BotInfo
	if not vccBotInfo then
		local errType = ("%s INFO DOES NOT EXIST IN VCC"):format(nluType)
		LogError("[UTILS:IsAvailableNLUInVCC] %s Info not exist in DataBase for vcc '%s'. Err: %s\nBotInfo: %s", nluType, vcc, errType, json.encode(vccBotInfo or {}))
		return false, errType
	end

	local currentBOT = (vccBotInfo.Type or ""):upper()
	if currentBOT ~= nluType then
		local errType = ("VCC HAS CONFIGURED %s NLU"):format(currentBOT)
		LogError("[UTILS:IsAvailableNLUInVCC] Error the NLU configured in vcc '%s' is '%s'. Err: %s\nBotInfo: %s", vcc, currentBOT, errType, json.encode(vccBotInfo or {}))
		return false, errType
	end

	if not vccBotInfo.Enabled then
		local errType = "NLU IT IS DISABLED IN THE VCC"
		LogError("[UTILS:IsAvailableNLUInVCC] The NLU it is disabled in the vcc '%s'. Err: %s\nBotInfo: %s", vcc, errType, json.encode(vccBotInfo or {}))
		return false, errType
	end

	return true
end

--[[ -------------------------
	FILE SYSTEM
]]--
function CreateFile (fullFileName, content)
	return m_helpers.filesystem.CreateFile(fullFileName, content)
end

function RemoveFile (fullFileName)
	return m_helpers.filesystem.RemoveFile(fullFileName)
end

function ReadFile (fullFileName, returnType)
	assert(type(fullFileName) == "string")
	--LogDebug("[UTILS:ReadFile] Read file '%s' return type: '%s'.", fullFileName, returnType)
	fullFileName = webHandler and webHandler.path.normalize(fullFileName) or fullFileName
	return m_helpers.filesystem.FileRead(fullFileName, returnType)
end

function RemoveDirectory (directoryName)
	assert(directoryName)
	if not directoryName then
		return false, {}
	end
	LogInfo("[UTILS:RemoveDirectory] Removing directory. %s.", directoryName)
	local exist = isDirectory(directoryName)
	if not exist then
		LogWarning("[UTILS:RemoveDirectory] Unable to remove. Directory does not exist. '%s'", directoryName)
		local err = {result = "UNABLE TO REMOVE DIRECTORY. DIRECTORY DOES NOT EXIST", directory = directoryName}
		return false, err
	end
	local commandLine = ("rm -rf %s"):format(directoryName)
	--LogDebug("[UTILS:RemoveDirectory] Removing directory. %s.", commandLine)
	os.execute(commandLine)
	return true
end

function GetFileExtension (fileName)
	local ext = fileName:match("^.+%.(.+)$") or ""
	local name
	local extLength = ext:len()
	if extLength == 0 then
		name = fileName
	else
		name = fileName:sub(1, fileName:len() - (extLength + 1))
	end
	return ext, name
end

function isDirectory (directoryName)
	assert(type(directoryName) =="string")
	--LogDebug("[UTILS:isDirectory] Validating Directory. directoryName: %s.", directoryName)
	return m_helpers.filesystem.IsDirectory(directoryName)
end

function CreateDirectory (directoryName)
	assert(type(directoryName) =="string")
	--LogDebug("[UTILS:CreateDirectory] Creating Directory. directoryName: %s.", directoryName)
	return m_helpers.filesystem.CreateDirectory(directoryName)
end

function FileSize (fullFileName)
	assert(type(fullFileName) =="string")
	local lfs = require "lfs"
	local size = lfs.attributes(fullFileName, "size")
	--LogDebug("[UTILS:FileSize] Size: %s, File: %s.", size, fullFileName)
	return size
end

function GetDirectories (directoryBase)
	--LogDebug("[UTILS:GetDirectories] Loading directories. directoryBase: %s", directoryBase)
	local directories = {}
	if not isDirectory(directoryBase) then
		LogWarning("[UTILS:GetDirectories] Unable to load files. Is not a directory: %s", directoryBase)
		return directories
	end
	for file in lfs.dir(directoryBase) do
		if file ~= "." and file ~= ".." then
			--LogDebug("[UTILS:GetDirectories]    file: %s", file)
			local f = directoryBase..'/'..file
			local attr = lfs.attributes(f)
			if attr.mode == "directory" then
				local directory = file
				table.insert(directories, directory)
			end
		end
	end
	--LogDebug("[UTILS.GetDirectories] Loaded directories: %s", json.encode(directories))
	return directories
end

function GetFiles (directoryBase)
	--LogDebug("[UTILS:GetFiles] Loading files of directory: %s", directoryBase)
	local files = {}
	if not isDirectory(directoryBase) then
		LogWarning("[UTILS:GetFiles] Unable to load files. Is not a directory: %s", directoryBase)
		return files
	end
	for file in lfs.dir(directoryBase) do
		if file ~= "." and file ~= ".." then
			--LogDebug("[UTILS:GetFiles]    file: %s", file)
			local f = directoryBase..'/'..file
			local attr = lfs.attributes(f)
			if attr.mode ~= "directory" then
				table.insert(files, file)
			end
		end
	end
	--LogDebug("[UTILS.GetFiles] Loaded files of directory. %s", json.encode(files))
	return files
end

function CompressFiles (fileName, sourceFullDirectory, loadContent)
	assert(fileName)
	assert(sourceFullDirectory)

	local zipFileName = fileName .. ".zip"
	local zipFullFileName = sourceFullDirectory .. zipFileName
	LogInfo("[UTILS:CompressFiles] Compressing Files. Source directory:  '%s'. File Destination: '%s', loadContent: '%s'.", sourceFullDirectory, zipFullFileName, loadContent)

	local command = ("cd %s; zip -m -r -9 %s *"):format(sourceFullDirectory, zipFullFileName)
	--LogDebug("[UTILS:CompressFiles] Compress Files (ZIP) command %s.", command)
	os.execute(command)

	local zipInfo = {
		fileName     = zipFileName,
		fullFileName = zipFullFileName,
		directory    = sourceFullDirectory,
		size         = FileSize(zipFullFileName)
	}
	LogInfo("[UTILS:CompressFiles] Compress Result:  '%s'", json.encode(zipInfo))

	if loadContent then
		local success, zipContent = ReadFile(zipInfo.fullFileName)
		if not success then
			LogWarning("[UTILS:CompressFiles] Unable to load content Compressed File. '%s'.", zipInfo.fullFileName)
			local err = zipContent
			return false, err
		end
		zipInfo.content = zipContent
		--LogDebug("[UTILS:CompressFiles] Compress Content:  '%s'", zipInfo.content)
	end

	return true, zipInfo
end

function GetCampaign (campaignIdWithVcc, columns, keyspace)
	assert(type(campaignIdWithVcc) == "string")
	return GetAllColumns {
		keyspace     = keyspace or "Agent",
		columnFamily = "Campaigns",
		key	         = campaignIdWithVcc,
		columns      = columns
	}
end

local m_cachedCampaignsInfo = {}
function GetCachedCampaignInfo (campaignIdWithVcc, keyspace)
	assert(type(campaignIdWithVcc) == "string")

	campaignIdWithVcc = campaignIdWithVcc:lower()
	--LogDebug("[UTILS:GetCachedCampaignInfo] Loading cached campign %s.", campaignIdWithVcc)

	if m_cachedCampaignsInfo and m_cachedCampaignsInfo[campaignIdWithVcc] then
		local cachedCampaignInfo = m_cachedCampaignsInfo[campaignIdWithVcc]
		local nowTime = os.time()
		local elapsedTime = nowTime - cachedCampaignInfo.cachedTime
		local EXPIRED_TIME_SEC = 30 -- 30 segundos

		--LogDebug("---> 1 nowTime '%s', cachedTime '%s', elapsedTime: '%s' seconds -->  EXPIRED_TIME_SEC '%s'.", nowTime, cachedCampaignInfo.cachedTime, elapsedTime, EXPIRED_TIME_SEC)

		if elapsedTime <= EXPIRED_TIME_SEC then
			--LogDebug("[UTILS:GetCachedCampaignInfo] 1.1 ===>  Retrieving cached campign %s.", campaignIdWithVcc)
			return cachedCampaignInfo.campaign
		--else
			--LogDebug("[UTILS:GetCachedCampaignInfo] 1.2 ===> EXPIRED cached %s.", campaignIdWithVcc)
		end
	end

	local campaign = GetCampaign(campaignIdWithVcc, keyspace)
	if not campaign then
		LogWarning("[UTILS:GetCachedCampaignInfo] Error campaign %s does not exist.", campaignIdWithVcc)
		return
	end

	--LogDebug("[UTILS:GetCachedCampaignInfo] 2 -----> Now Caching & Retrieving campign %s.", campaignIdWithVcc)

	m_cachedCampaignsInfo[campaignIdWithVcc] = {cachedTime = os.time(), campaign = campaign}
	return campaign
end

-------
local m_timezone = {}
m_timezone.cachedTimezones = false
m_timezone.cachedExpiredTimeSec = 60 * 60 -- 1 hora
m_timezone.cachedTime = nil

m_timezone.FindAll = function(self, keyspace)
	if self.cachedTimezones then
		local nowTime = os.time()
		local elapsedTimeSec = nowTime - self.cachedTime
		if elapsedTimeSec <= self.cachedExpiredTimeSec then
			--LogDebug("[UTILS:m_timezone.FindAll] 1.1 -----> Loading Timezones from CACHE. Expired in %s minutes.", (self.cachedExpiredTimeSec - elapsedTimeSec) / 60)
			return self.cachedTimezones
		--else
			--LogDebug("[UTILS:m_timezone.FindAll] 1.2 ===> EXPIRED cache.")
		end
	end
	self.cachedTime = os.time()
	--LogDebug("[UTILS:m_timezone.FindAll] 2 -----> Loading Timezones from DATABASE. Time: %s, expired in %s minutes.", self.cachedTime, self.cachedExpiredTimeSec /)
	local closeKeyspace
	if not keyspace then
		keyspace = GetDatasource():CassandraConnection("Agent")
		closeKeyspace = true
	end
	self.cachedTimezones = {}
	local timezoneList = keyspace:GetRangeSlices("Timezones", --column family
		"",  --supercolumn
		"", --startkey
		"", --endkey
		{ --predicate
			column_names = {},
			slice_range = {
				start = "",
				finish = "",
				reversed = false,
				count = 100
			}
		},
		1000 --total keys
	)
	if timezoneList then
		for timezoneId, timezoneRow in timezoneList:Iterator() do
			if #timezoneRow > 0 then
				--Ignoro posible registros borrados.
				local timezone = {}
				for k, v in ipairs(timezoneRow) do
					timezone[v.name] = v.value
					if v.name == "HistoricalDaylight" then
						local success, historicalDaylight = pcall(json.decode, v.value)
						if success then
							timezone[v.name] = historicalDaylight
						end
					end
				end
				timezone.Id = timezoneId
				self.cachedTimezones[timezoneId] = timezone
			end
		end
	end
	if closeKeyspace then
		keyspace:Close()
	end
	return self.cachedTimezones
end

m_timezone.Get = function(self, timezoneId, keyspace)
	assert(timezoneId)
	local timezones = self:FindAll(keyspace)
	assert(timezones)
	local timezone = timezones[timezoneId]
	assert(timezone, "[UTILS.m_timezone.Get] Timezone " .. timezoneId .. " does not exst !!")
	return timezone
end

--[[
	Retorna con en el mismo formato que se encuentra en "inconcert.authUser.timezone"
]]--
m_timezone.GetEnvironmentFormat = function(self, timezoneId, keyspace)
	local timezone = self:Get(timezoneId, keyspace)
	local enviormentTimezone = {
		id                 = timezoneId,                                   -- ej: "UYT",
		offset             = tonumber(timezone.TimeZone),                  -- ej: -3,
		minutes            = tonumber(timezone.TimeZoneMinutes),           -- ej: 0,
		dst                = (tonumber(timezone.DaylightSavingTime) ~= 0), -- ej: true
		HistoricalDaylight = timezone.HistoricalDaylight                   -- el historico
	}
	--LogDebug("[UTILS.m_timezone.GetEnvironmentFormat] timezoneId '%s', enviormentTimezone: '%s'", timezoneId, json.encode(enviormentTimezone))
	return enviormentTimezone
end

--[[
	Funcion que calcula el timezone offset, retorna en minutos.
]]--
m_timezone.CalculateTimezoneOffsetMin = function(self, timezoneId, keyspace)
	local timezone = self:GetEnvironmentFormat(timezoneId, keyspace)
	local offset, minutes, dst = assert(timezone.offset, "missing 'offset' field"), timezone.minutes or 0, timezone.dst or false

	--LogDebug("[UTILS.m_timezone.CalculateTimezoneOffsetMin] timezoneId '%s', offset: '%s', minutes: '%s', dst: '%s'", timezoneId, offset, minutes, dst)

	-- ajusta el signo de los minutos si el desfase es negativo 
	if offset < 0 and minutes > 0 then
		minutes = minutes * -1
	end
	offset = offset * 60
	if dst then
		offset = offset + 60
	end

	local timezoneOffset = offset + minutes
	--LogDebug("[UTILS.m_timezone.CalculateTimezoneOffsetMin] timezoneId '%s', calculated timezoneOffset: '%s'", timezoneId, timezoneOffset)
	return timezoneOffset
end

function GetTimeZoneInstance()
	return m_timezone
end
-------------------

--[[
	ELASTIC SEARCH
]]--

local m_elasticSearch = {}

--[[
	 Tope máximo de registros permitidos por ES.
--]]
m_elasticSearch.GetMaxAllowedRowsInES = function(self, fromAPI)
	return m_helpers.elasticsearch:GetMaxAllowedRowsInES(fromAPI)
end

--[[
	 Funcion encargada de validar el tope maximo de registros a retornar por ES.
	 Si el request esta solicitando mas registros de los soportados, los mismos se topean al maximo permitido.
	 En si se ubica en la ultima pagina soportada.
--]]
m_elasticSearch.MakeElasticSearchPaged = function (self, query, fromAPI)
	return m_helpers.elasticsearch:MakeElasticSearchPaged(query, fromAPI)
end

--[[
]]--
m_elasticSearch.MakeConnectionConfig = function (self, info)
	assert(type(info) == "table")
	if not info.credentials then
		local elasticSearchConnectionConfig = GetConnectionConfig("elasticsearch", info.appId or nil)
		if elasticSearchConnectionConfig then
			info.credentials = elasticSearchConnectionConfig.credentials
			info.host = info.host or elasticSearchConnectionConfig.host
			info.port = info.port or elasticSearchConnectionConfig.port
		end
	end
	return info
end

--[[
	Sample:
		credentials: 
			info.credentials = {
				userId  = "elastic",
				userPwd = "2U96vhR9AsJ5Hyj9YnjT"
			}
--]]
m_elasticSearch.MakeRequestInfo = function (self, info)
	local requestInfo = m_elasticSearch:MakeConnectionConfig(info)
	return m_helpers.elasticsearch:MakeRequestInfo(requestInfo)
end

--[[
--]]
m_elasticSearch.MakeRequestInfoForDelete = function (self, info)
	local requestInfo = m_elasticSearch:MakeConnectionConfig(info)
	return m_helpers.elasticsearch:MakeRequestInfoForDelete(requestInfo)
end

--[[
--]]
m_elasticSearch.PitElasticSearch = function (self, vccId, connectionConfig)
	--LogDebug("[UTILS.PitElasticSearch] Get PIT id from elasticsearch. vccId: %s, connectionConfig: %s", vccId, connectionConfig)
	assert(type(connectionConfig) == "table")
	assert(connectionConfig.host)
	assert(connectionConfig.port)

	local requestInfo = m_elasticSearch:MakeRequestInfo({
		query = query
	})

	local response_body = {}
	local index_name = ("vcc-%s"):format(vccId)

	local ok, response_code, response_headers, response_status_line = http.request(
	{
		host = connectionConfig.host,
		port = connectionConfig.port,
		url = ("/%s/_pit?keep_alive=180s"):format(index_name),
		headers = requestInfo.headers,
		method = "POST",
		sink = ltn12.sink.table(response_body),
		redirect = false
	})
	if not ok then
		LogError("[UTILS.PitElasticSearch]: %s\n%s\n%s\n%s", response_code, json.encode(response_headers), response_status_line,
			json.encode(response_body))
		return false, { error = response_status_line, status = response_code }
	end

	--LogWarning("=-=-=-=-= %s", json.encode(response_headers, true))

	local raw_response = table.concat(response_body)
	if not response_headers["content-type"]:match("^application/json") then
		LogError("[UTILS.PitElasticSearch] Unknown response from elasticsearch. headers: %s\nresponse: %s", json.encode(response_headers), json.encode(raw_response))
		return false, { error = "inconcert unknown", status = 500 }
	end
	local response = json.decode(raw_response)
	if response.error then
		LogError("[UTILS.PitElasticSearch] Request failed with the following error: '%s'", json.encode(response.error))
		return false, response
	end
	LogInfo("[UTILS.PitElasticSearch] PIT id from elasticsearch (vcc: %s): %s", vccId, response.id)
	return true, response.id
end

--[[
--]]
m_elasticSearch.CountElasticSearch = function (self, query, vccId, connectionConfig)
	--LogDebug("[UTILS.PitElasticSearch] CountElasticSearch - elasticsearch. vccId: %s, connectionConfig: %s", vccId, connectionConfig)
	assert(type(connectionConfig) == "table")
	assert(connectionConfig.host)
	assert(connectionConfig.port)

	local requestInfo = m_elasticSearch:MakeRequestInfo({
		query = query
	})

	local response_body = {}
	local index_name = "vcc-" .. vccId
	local urlString = "/"..index_name.."/_count"

	local ok, response_code, response_headers, response_status_line = http.request({
		host = connectionConfig.host,
		port = connectionConfig.port,
		url = urlString,
		headers = requestInfo.headers,
		method = "POST",
		source = requestInfo.source,
		sink = ltn12.sink.table(response_body),
		redirect = false
	})
	if not ok then
		if response_body then
			return false, response_code
		end
		if response_code then
			return false, response_code
		end
		if response_headers then
			return false, json.encode(response_headers)
		end
		if response_status_line then
			return false, json.encode(response_status_line)
		end
		return false, ""
	end
	response_body = table.concat(response_body)
	if response_headers["content-type"]:match("^application/json") then
		local response = json.decode(response_body)
		return true, response_body
	end
	return false, response_body
end

--[[
--]]
function GetElasticSearchInstance ()
	return m_elasticSearch
end
-----------------------------------------------

--[[
Función encargada de obtener  la configuración de alguna app del objeto "inconcert.environment.apps".
Si NO existe la configuración de la app que se ha recibido como parámetro, se obtendrá la configuración por defecto de la app "agent".
--]]
function GetConnectionConfig (connectionType, appId)
	--LogDebug("[UTILS:GetConnectionConfig] -> Get connectionType '%s' of app: '%s'.", connectionType, appId or "ANY")
	assert(connectionType)
	if not inconcert or not inconcert.environment then
		--LogWarning("[UTILS:GetConnectionConfig] 1--> Invalid inconcert.environment instance.")
		return
	end

	--[[
	]]--
	local function GetConnection (appId, env)
		if not env then
			--LogDebug("[UTILS:GetConnectionConfig]   2--> App inconcert.environment does not exist.")
			return false
		end
		if not env[appId] then
			---- El default si es que aun no se ha cargado la VM
			--LogDebug("[UTILS:GetConnectionConfig]   3--> App inconcert.environment.apps['%s'] does not exist in instance. Will be assigned: inconcert.environment.apps.agent: %s", appId, JSON_Encode(inconcert.environment.apps.agent))
			env[appId] = inconcert.environment.apps.agent
			if not env[appId] then
				LogWarning("[UTILS:GetConnectionConfig]   4--> App inconcert.environment.apps['%s'] is invalid.", appId)
				return false
			end
		end
		local conn = env[appId][connectionType]
		if not conn then
			--LogDebug("[UTILS:GetConnectionConfig] 6 --> inconcert.environment.app['%s']['%s'] has not initialized connection type.", appId, connectionType)
			return false
		end
		-- ok
		return conn
	end

	local environmentDef = {
		api = true,
		apps = {
			"agent",
			"outbound_engine",
			"supervisor"
		}
	}

	if appId then
		local conn = GetConnection(appId, inconcert.environment.apps)
		if conn then
			-- ok
			--LogDebug("[UTILS:GetConnectionConfig] 1 OK - return inconcert.environment[%s][%s] .", appId, connectionType)
			--m_helpers.table:Log(conn)
			return conn
		end
		environmentDef[appId] = nil
	end

	local i = 0
	for appId, info in pairs(environmentDef) do
		i = i + 1
		if type(info) == "boolean" then
			local conn = GetConnection(appId, inconcert.environment.apps)
			if conn then
				-- ok
				--LogDebug("[UTILS:GetConnectionConfig] 2.%s OK - return inconcert.environment[%s][%s] .", i, appId, connectionType)
				--m_helpers.table:Log(conn)
				return conn
			end
		elseif type(info) == "table" then
			for j, appId in ipairs(info) do
				local conn = GetConnection(appId, inconcert.environment.apps)
				if conn then
					-- ok
					--LogDebug("[UTILS:GetConnectionConfig] 3.%s.%s OK - return inconcert.environment.apps[%s][%s] .", i, j, appId, connectionType)
					--m_helpers.table:Log(conn)
					return conn
				end
			end
		end
	end

	-- Error
	LogWarning("[UTILS:GetConnectionConfig] ---> Connection type '%s' does not configured in apps of inconcert.environment.apps..", connectionType)
	return
end

--[[
--]]
local m_modelsOthersApps = {}

--[[
--]]
local function GetModelOthersApps (appId, entityName)
	assert(type(appId) == "string")
	assert(type(entityName) == "string")
	appId = appId:lower()
	--LogDebug("[UTILS:GetModelOthersApps] 1 ---> Get Model. appId: '%s', entityName: '%s'.", appId, entityName)
	local appModel = m_modelsOthersApps[appId]
	if not appModel or not appModel[entityName] then
	-- Si el modelo no se ha creado aun, debo crearlo.
	--  Esto sucede en algunas ocasiones si es que me han invocado desde otra app Ej: "Allegro" y aun no se ha creado el modelo.
	--  Entonces tengo que crear el modelo.
		local moduleName = string.format("apps.%s.model", appId)
		--LogDebug("[UTILS:GetModelOthersApps] 2 ---> require '%s', appId: '%s', entityName: '%s'.", moduleName, appId, entityName)
		local model = require(moduleName)  -- apps.administration.model, apps.administration.agent, apps.batch_management.model
		model:BuildModels()
		m_modelsOthersApps[appId] = model:GetModels()
		appModel = m_modelsOthersApps[appId]
	end
	local entityModel = appModel[entityName]
	assert(type(entityModel) == "table")
	return entityModel
end

function GetAdmModel (entityName)
	return GetModelOthersApps("administration", entityName)
end

function GetAgentModel (entityName)
	return GetModelOthersApps("agent", entityName)
end

function GetBatchModel (entityName)
	return GetModelOthersApps("batch_management", entityName)
end

function GetOEModel (entityName)
	return GetModelOthersApps("outbound_engine", entityName)
end

function GetControllerOthersApps (appId, controllerInfo)
	assert(type(appId) == "string")
	assert(type(controllerInfo) == "table")

	appId = appId:lower()
	--LogDebug("[UTILS:GetControllerOthersApps] 1 ---> Start App: '%s', controllerInfo: '%s'.", appId, json.encode(controllerInfo))

	-- 1 el modelo
	local modelInstance
	if controllerInfo.modelName then
		modelInstance = GetModelOthersApps(appId, controllerInfo.modelName)
	end

	-- 2 el controller
	local controllerInstance
	if controllerInfo.controllerName then
		local moduleName
		if controllerInfo.controllerPath then
			moduleName = string.format("apps.%s%s%s", appId, controllerInfo.controllerPath, controllerInfo.controllerName) -- Ej: "apps.outbound_engine.outboundengineclient"
		else
			moduleName = string.format("apps.%s.controllers.%s", appId, controllerInfo.controllerName) -- Ej: "apps.agent.controllers.interactions"
		end
		assert(moduleName)
	--LogDebug("[UTILS:GetControllerOthersApps] 2 ---> require '%s', appId: '%s', name: '%s'.", moduleName, appId, controllerInfo.controllerName)
		controllerInstance = require(moduleName)
	end
	
	--LogDebug("[UTILS:GetControllerOthersApps] 3 ---> controllerInstance '%s', modelInstance: '%s'.", controllerInstance, modelInstance)

	return controllerInstance, modelInstance
end

function GeAdmController (controllerInfo)
	return GetControllerOthersApps("administration", controllerInfo)
end

function GetAgentController (controllerInfo)
	return GetControllerOthersApps("agent", controllerInfo)
end

function GetOEController (controllerInfo)
	return GetControllerOthersApps("outbound_engine", controllerInfo)
end

function GetAllegroController (controllerInfo)
	return GetControllerOthersApps("allegro", controllerInfo)
end

function IsRunningOEProcess (processId)
	assert(type(processId) == "string")
	local outboundEngineClient = GetOEController({controllerName = "outboundengineclient", modelName = "OutboundProcesses", controllerPath = "."})
	local isRunning, result = outboundEngineClient.processes:IsRunning(processId)
	LogInfo("[UTILS:IsRunningOEProcess] --> OE Process '%s' %sIs RUNNING.", processId, (isRunning and "" or "NOT "))
	return isRunning, result
end

function GetOEApi (moduleApiName)
	-- Obtengo el controller de la app "outbound_engine" y las funciones que expone. Por defecto "Procesos"
	moduleApiName = (moduleApiName or "processes"):lower()
	local allowedApiModules = {
		contacts = {
			controllerName = "contacts",
			modelName      = nil -- no tiene archivo de modelo
		},
		-- por defecto
		processes = {
			controllerName = "processes",
			modelName      = "OutboundProcesses"
		},
		system = {
			controllerName = "system",
			modelName      = nil -- no tiene archivo de modelo
		}
	}
	local apiModuleDefinition = allowedApiModules[moduleApiName]
	assert(type(apiModuleDefinition) == "table")
	local controller, model = GetOEController(apiModuleDefinition)
	assert(controller)
	return controller.GetApi(), model
end

function GetOEProcessesSystemChannels (_T)
	--LogDebug("[UTILS:GetOEProcessesSystemChannels] .---------------> _T: %s", _T)
	local oeProcessChannelsSystemConfig = GetSystemConfig("OUTBOUND_ENGINE_PROCESSES_CHANNELS")
	local systemChannels = oeProcessChannelsSystemConfig.Channels
	assert(systemChannels)
	local allowedChannels = {}
	for channelId, info in pairs(systemChannels) do
		if info.enabled then
			allowedChannels[channelId] = info
		end
	end
	return allowedChannels
end

function GetOEProcessesSystemRulesActions (_T)
	--LogDebug("[UTILS:GetOEProcessesSystemRulesActions] .---------------> _T: %s", _T)
	_T = _T or function(text) return text end
	local systemChannels = GetOEProcessesSystemChannels(_T)
	local oeProcessSystemRulesActions = GetSystemConfig("OUTBOUND_ENGINE_PROCESSES_RULES_SYSTEM_ACTIONS")
	local systemActions = oeProcessSystemRulesActions.Actions
	assert(systemActions)
	local allowedActions = {}
	local hashAllowedActionsByResultNumber = {}
	for _, systemAction in ipairs(systemActions) do
		if systemChannels[systemAction.channel] then
			systemAction.translatedResult = _T(systemAction.result)
			--LogDebug("[UTILS:GetOEProcessesSystemRulesActions] ---Translated Result------------> '%s' = '%s'.", systemAction.result, systemAction.translatedResult)
			table.insert(allowedActions, systemAction)
			-- se agrega en un hash de acciones por canal y por numero de resultado
			hashAllowedActionsByResultNumber[systemAction.channel] = hashAllowedActionsByResultNumber[systemAction.channel] or {}
			local resultNumberId = tostring(systemAction.result_number)
			hashAllowedActionsByResultNumber[systemAction.channel][resultNumberId] = systemAction
		end
	end
	return allowedActions, hashAllowedActionsByResultNumber
end

function GetOSDistro()
	return m_helpers.utils:GetOSDistro()
end

function Get_m_helpers_agentservice_options()
	local options = {}
	if inconcert and inconcert.environment and inconcert.environment.apps and inconcert.environment.apps.agent and inconcert.environment.apps.agent.agentservice and inconcert.environment.apps.agent.cluster then
		options.host = inconcert.environment.apps.agent.agentservice.host
		options.port = inconcert.environment.apps.agent.agentservice.port
		options.cluster = inconcert.environment.apps.agent.cluster
	end
	return options
end

function Sleep (timeSeconds)
	m_helpers.utils:Sleep(timeSeconds)
end

function AsyncTasksWithAuthToken (url, taskId, method, headers, jsonData, cookies)
	assert(type(url) == "string")
	--LogDebug("[UTILS:AsyncTasksWithAuthToken] ---> url '%s', taskId '%s'.",url, taskId)
	local cookies = cookies or {}
	if inconcert and inconcert.authUser then
		cookies["AUTH_SID_AGENT"] = inconcert.authUser.authToken
	end
	m_helpers.utils:ExecuteAsyncTasks(url, taskId, method, headers, jsonData, cookies)
end

local function ValidateJsonData (data, validations)
	--LogDebug("[Utils.ValidateJsonData] json_data '%s' - validations '%s'", json.encode(data), json.encode(validations))
	return m_helpers.validation:StartValidateJsonValues(data, validations)
end

-- local function ValidateValue (value, validations )
-- 	return m_helpers.validation:ValidateValue(value, validations)
-- end

--[[
	Funciones para realizar distintos tipos de validaciones sobre valores de campos.
	Parametros:
		validations: validaciones a realizar sobre los datos de una entidad.
		functionToCall: Funcion callback
		isSimple: Cuando se espera un nodo ejemplo.
			{"Name":"<img src=x onerror=alert(1)>","Id":"Habilidad2","isNew":false,"DefaultValue":1,"Description":"Habilidad1"}
			
		-Por lo general es una arreglo o tabla de varios nodos de una misma entidad:
		{"Habilidad2":{"Name":"<img src=x onerror=alert(1)>","Id":"Habilidad2","isNew":false,"DefaultValue":1,"Description":"Habilidad1"}}
]]--
function WebJsonValidatorData (validations, functionToCall)
	--LogDebug("[UTILS:WebJsonValidatorData] 1--> validations '%s', functionToCall: '%s': '%s'.", json.encode(validations), functionToCall)
	--[[
	]]--
	return function(web, ...)
		local ok, entities = pcall(json.decode, web.input["post_data"])
		--LogDebug("[UTILS:WebJsonValidatorData.callback] 2--> validations '%s', data: '%s'.", json.encode(validations), web.input["post_data"])

		local args = { ... }
		if not ok then
			return functionToCall(web, args)
		end

		local isValid, err = ValidateJsonData(entities, validations)
		if not isValid then
			LogError("[UTILS:WebJsonValidatorData.callback] Error in field '%s', value: '%s'. %s", err.field_id, err.value, err.reason)
			web.status = "400"
			return json.encode(err)
		end

		return functionToCall(web, args)
	end
end

function ParseQueryString (queryStr, TransformValuesFn)
	return m_helpers.utils:ParseUrlQueryString(queryStr, TransformValuesFn)
end

--[[-----------------------------------------------------
	GOOGLE ANALYTICS
---------------------------------------------------------]]
local m_analytics = {}

--[[
--]]
m_analytics.IsEnabled = function(self)
	return m_helpers.utils.Analytics:Get():IsEnabled()
end

--[[
--]]
m_analytics.IncludeHeadLibResources = function(self, include)
	return m_helpers.utils.Analytics:Get():IncludeHeadLibResources(include)
end

--[[
--]]
m_analytics.IncludeBodyLibResources = function(self, include)
	return m_helpers.utils.Analytics:Get():IncludeBodyLibResources(include)
end

--[[
--]]
function Analytics ()
	return m_analytics
end

--[[-----------------------------------------------------
	asdfaaaaaaaaaaaaaaaaaaaaaaaaa
---------------------------------------------------------]]
--[[
	Funcion encargada de Incluir los JS y CSS de APPS, LIB y SHARED, segun los permisos del Usuario.
	params:
		-----------------------------
		- CalculateIncludes: (funcion) La cual se encarga de realizar un calculo personalizado de los includes a insertar.
										Este parametro se utiliza solo, es decir si viene este parametro, NO se tienen en cuenta los demas parametros que llegan (**).
		--(**)---------------------------
		- MustInclude      : (funcion) Funcion que valida si los includes Customs se deben de incluir (ej: validando permisos).
		- customInclusions : (string/function)  Aca llegan los includes (markup de js y css) que se van a insertar si es que la respuesta de la funcion "MustInclude" ha dado true.
		- alwaysIncludes   : (string)  Aca llegan los includes (markup de js y css) que se van a insertar SIEMPRE, sin importar la respuesta de la funcion "MustInclude".
		- notIncludeMessage: (string)  Texto que se va a incluir en la MasterPage indicando que estos resources (js & css) no fueron incluidos.
		-L                 : (number)  Numero que se agrega como sufijo en los JS y CSS en los builds.
		-----------------------------
]]--
function IncludeLibResources (params, customInfo)
	local includes
	if type(params.CalculateIncludes) == "function" then
		includes = params.CalculateIncludes()
	else
		local customInclusions = ""

		if params.customInclusions and params.MustInclude() then
			customInclusions = type(params.customInclusions) == "function" and params.customInclusions(customInfo) or params.customInclusions
		else
			local message = params.notIncludeMessage
			--LogDebug(message)
			customInclusions = string.format([[<!-- %s -->]], message):upper()
		end

		if (params.alwaysIncludes or ""):len() > 0 then
			includes = params.alwaysIncludes .. "\n" .. customInclusions
		else
			includes = customInclusions
		end
	end

	return (includes or ""):gsub("L_NUMBER", params._L)
end

--[[
	Incluye los SNIPPETS para: GOOGLE ANALYTICS, GOOGLE TAG MANAGER, HOTJAR.
	En el HEAD y en el BODY.
]]--
function InclueAnalyticsLibResources (params)
	assert(type(params) == "table")
	assert(type(params.includeType) == "string")
	assert(params._L)

	local m_analytics = Analytics()
	local isEnabled = true
	--[[
		Funcion que valida si debe agregar los snippets
	]]--
	local function MustInclude ()
		if not isEnabled then
			return false
		end
		return m_analytics:IsEnabled()
	end
	--LogDebug("[UTILS:InclueAnalyticsLibResources]----> params.include: %s", json.encode(params.include or {}))
	local includeByDefault = {
		analytics = true,
		hotjar    = true
	}
	local include = params.include or includeByDefault

	if params.includeType:upper() == "BODY" then
		return IncludeLibResources({
			MustInclude       = MustInclude,
			notIncludeMessage = "No Include IN BODY HTML the SNIPPETS for GOOGLE ANALYTICS, GOOGLE TAG MANAGER, HOTJAR",
			customInclusions  = m_analytics:IncludeBodyLibResources(include),
			_L                = params._L
		})
	elseif params.includeType:upper() == "HEAD" then
		return IncludeLibResources({
			MustInclude       = MustInclude,
			notIncludeMessage = "No Include IN HEAD HTML the SNIPPETS for GOOGLE TAG MANAGER",
			customInclusions  = m_analytics:IncludeHeadLibResources(include),
			_L                = params._L
		})
	end
	assert("[UTILS:WebJsonValidatorData]--> Unrecognized includeType: " .. params.includeType)
end
