/*jshint quotmark:false*/
/*global $, inConcert, d3, Mustache, _*/

if ( !this.inConcert ) {
  this.inConcert = {};
}

(function () {
/*jshint validthis:true*/
"use strict";

	var format = inConcert.Utils.format;

	var colors = ["#1E72AA", "#FF8B4D", "#8CBB4E", "#ae81ff", "#C44440", "#46b3e6", "#E00000", "#FE9B0C", "#F27405",
		"#F15A29", "#D93D04", "#855C68", "#603642", "#DDE9CA", "#85C6A9", "#7EAF9F", "#25517F", "#7A7C95",
		"#999EBB", "#ACBFDF", "#DDDDDD", "#EDC685", "#472348", "#3F3F3F"];
	
	function getColor(index) {
		index = index % colors.length;
		return colors[index];
	}
	
	function makeProperties(defaultProperties, customProperties) {
		customProperties = customProperties || {};
		var properties = {};
		$.each(defaultProperties, function(id, value) {
			var custom = customProperties[id];
			properties[id] = (custom !== undefined && custom !== null)? custom : value;
		});
		return properties;
	}
	
	function makePercentages(elements) {
		var total = 0;
		$.each(elements, function(_, element) {
			if (element.total) {
				return;
			}
			total += element.value || 0;
		});
		$.each(elements, function(_, element) {
			element.sourceValueLabel = element.valueLabel || element.value;
			element.sourceValue = element.value;
			element.value = element.total? 100 : (total >0? (element.value * 100 / total) : 0);
			element.valueLabel = Math.round(element.value) + "%";
		});
	}
	
	function hoverElement($element, $row, hover) {
		$row.toggleClass("hover", !!hover);
		var className = $element.attr("class");
		className = className.replace(" hover", "") + (hover? " hover" : "");
		$element.attr("class", className);
	}
	
	function cleanView(viewName) {
		$(format("div[chart-view={0}]", viewName)).remove();
	}

	var pie_default = {
		viewName : "",
		size : null,
		size_percentage : 100,
		percentages : false,
		id_key : "indicator",
		$references : null,
		tooltip : {
			enabled : true,
			className : "piechartTooltip",
			make : function(data, properties) {
				if (properties.percentages) {
					return format("<strong>{0}: {1} ({2})</strong>",
					data.name || data.id || "",
					data.sourceValueLabel || data.sourceValue || 0,
					data.valueLabel || data.value || 0);
				}
				return format("<strong>{0} ({1})</strong>",
					data.name || data.id || "",
					data.valueLabel || data.value || 0);
			}
		}
	};

	/**
	 * Makes pie chart (container must exists and has assigned size)
	 *
	 * @public
	 *
	 * @param	{JQuery}			$container
	 * @param	{array<element>}	elements_t				sample: [ { name : "prueba1", value : 50, color : "red" }, { name : "prueba2", value : 100, color : "blue" } ]
	 *			{string}			element.id			
	 *			{string}			element.name		
	 *			{integer}			element.value
	 *			{string}			element.valueLabel		optional: value to be shown
	 *			{string}			element.color			optional: fixed color to be assigned
	 * @param	{object}			props					optional
	 *			{string}			props.viewName			view name
	 *			{integer}			props.size				chart size
	 *			{integer}			props.size_percentage	percentage in "donut" chart
	 *			{bool}				props.percentages		indicates if percentages will be shown
	 *			{string}			props.id_key			key of id property
	 *			{JQuery}			props.$references		list of reference elements
	 *			{object}			props.tooltip			
	 *			{bool}				props.tooltip.enabled	indicates if tooltips will be shown
	 *			{string}			props.tooltip.className	tooltips class name
	 *			{function}			props.tooltip.make		function to make tooltips content
	 */	
	function pie($container, elements_t, props) {
		var properties = makeProperties(pie_default, props);
		var totalElement = null;
		var elements = _.reduce(elements_t, function(memo, element) {
			if (element.total) {
				totalElement = element;
			}
			else {
				memo.push(element);
			}
			return memo;
		}, []);
		if (!elements.length && totalElement) {
			elements.push(totalElement);
		}
		
		if (properties.percentages) {
			makePercentages(elements);
		}
				
		$container.html("");
		var container = d3.select($container[0]);
		
		var radius;
		if (properties.size) {
			$container
				.width(properties.size)
				.height(properties.size);
			radius = properties.size / 2;
		}
		else {
            radius = Math.min($container.width(), $container.height()) / 2;
		}
	
		var outerRadious = radius;
		var innerRadius = outerRadious - outerRadious * properties.size_percentage / 100;
		
		var arc = d3.arc()
            .outerRadius(outerRadious)
            .innerRadius(innerRadius);

		var pie = d3.pie()
            .sort(null)
            .value(function (data) {
                return data.value || 0;
            });
		
		var svg = container
			.append("svg")
            .attr("width", "100%")
            .attr("height", "100%")
            .append("g")
            .attr("transform", "translate(" + radius + "," + radius + ")");
	
		var tip = null;
		if (properties.tooltip.enabled) {
			tip = d3.tip()
			   .attr("class", "d3-tip " + properties.tooltip.className)
			   .attr("chart-view", properties.viewName)
			   .offset([-10, 0])
			   .html(function (d) {
				   return properties.tooltip.make(d.data, properties);
			   });
			svg.call(tip);
		}
	
		var g = svg.selectAll(".arc")
            .data(pie(elements))
            .enter()
			.append("g")
            .attr("class", "arc")
            .attr(properties.id_key, function (d, i) {
                return d.data.id || d.data.name || (properties.id_key + "_" + i);
            });
			
		if (properties.tooltip.enabled) {
			g
				.on("mouseover", tip.show)
				.on("mouseout", tip.hide);
		}
			
		g
			.append("path")
            .attr("d", arc)
            .style("fill", function (d, i) {
                return d.data.color || getColor(i);
            });
			
		if (properties.$references) {
			$container.find(".arc")
				.off("mouseover")
				.off("mouseout")
				.on("mouseover", function () {
					hoverElement($(this).children(), 
						properties.$references.filter(format("[{0}={1}]", properties.id_key, $(this).attr(properties.id_key))), true);
				})
				.on("mouseout", function () {
					hoverElement($(this).children(), 
						properties.$references.filter(format("[{0}={1}]", properties.id_key, $(this).attr(properties.id_key))), false);
				});
			properties.$references
				.off("mouseover")
				.off("mouseout")
				.on("mouseover", function () {
					hoverElement($container.find(format("[{0}={1}]", properties.id_key, $(this).attr(properties.id_key))).children(), 
						$(this), true);
				})
				.on("mouseout", function () {
					hoverElement($container.find(format("[{0}={1}]", properties.id_key, $(this).attr(properties.id_key))).children(), 
						$(this), false);
				});
		}
	}
	
	function makeAxis(axis, props, size, length) {
		//ticks number
		var ticks = null;
		switch(props.ticks) {
			case "all":
				ticks = length - 1;
				break;
			case "size":
				ticks = Math.floor(size / props.pixelsPerTick);
				break;
			case "split":
				ticks = length - 1;
				var split = props.split || 2;
				var max = Math.floor(size / props.pixelsPerTick);
				while (ticks > max) {
					ticks = Math.floor(ticks / split);
				}
		}
		if (ticks) {
			axis = axis.ticks(ticks);
		}
		//ticks format
		if (props.format) {
			axis = axis.tickFormat(props.format);
		}
		return axis;
	}

	var bars_default = {
		viewName : "",
		size : null,
		size_percentage : 100,
		percentages : false,
		id_key : "indicator",
		$references : null,
		tooltip : {
			enabled : true,
			className : "piechartTooltip",
			make : function(data, properties) {
				if (properties.percentages) {
					return format("<strong>{0}: {1} ({2})</strong>",
					data.name || data.id || "",
					data.sourceValueLabel || data.sourceValue || 0,
					data.valueLabel || data.value || 0);
				}
				return format("<strong>{0} ({1})</strong>",
					data.name || data.id || "",
					data.valueLabel || data.value || 0);
			}
		}
	};

	/**
	 * Makes bar chart (container must exists and has assigned size)
	 *
	 * @public
	 *
	 * @param	{JQuery}			$container
	 * @param	{array<element>}	elements				sample: [ { name : "prueba1", value : 50, color : "red" }, { name : "prueba2", value : 100, color : "blue" } ]
	 *			{string}			element.id			
	 *			{string}			element.name		
	 *			{integer}			element.value
	 *			{string}			element.valueLabel		optional: value to be shown
	 *			{string}			element.color			optional: fixed color to be assigned
	 * @param	{object}			props					optional
	 *			{string}			props.viewName			view name
	 *			{integer}			props.size				chart size
	 *			{bool}				props.percentages		indicates if percentages will be shown
	 *			{string}			props.id_key			key of id property
	 *			{JQuery}			props.$references		list of reference elements
	 *			{object}			props.tooltip			
	 *			{bool}				props.tooltip.enabled	indicates if tooltips will be shown
	 *			{string}			props.tooltip.className	tooltips class name
	 *			{function}			props.tooltip.make		function to make tooltips content
	 */	
	function bars($container, elements, props) {
		var properties = makeProperties(bars_default, props);
		
		if (properties.percentages) {
			makePercentages(elements);
		}
		
		$container.html("");
		var container = d3.select($container[0]);
		
		var width;
		var height;
		if (properties.size) {
			width = properties.size.width;
			height = properties.size.height;
			$container
				.width(width)
				.height(height);
		}
		else {
			width = $container.width();
			height = $container.height();
		}
		
		var margin = {
			left : properties.margin && properties.margin.left || 0,
			right : properties.margin && properties.margin.right || 0,
			top : properties.margin && properties.margin.top || 0,
			bottom : properties.margin && properties.margin.bottom || 0
		};
		
		width = width - margin.left - margin.right;
		height = height - margin.top - margin.bottom;

		var x = d3.scaleBand().rangeRound([0, width]).padding(0.1);
		var y = d3.scaleLinear().rangeRound([height, 0]);
		
		x.domain(elements.map(function(d, i) { return d.name || d.id; }));
		y.domain([0, d3.max(elements, function(d) { return d.value; })]);
		
		var svg = container.append("svg")
            .attr("width", "100%")
            .attr("height", "100%");
		
		var g = svg.append("g")
			.attr("transform", "translate(" + margin.left + "," + margin.top + ")");
		
		if (properties.xAxis) {
			g.append("g")
				.attr("class", "axis axis--x")
				.attr("transform", "translate(0," + height + ")")
				.call(makeAxis(d3.axisBottom(x), properties.xAxis, width, elements.length));
		}
		if (properties.yAxis) {
			g.append("g")
				.attr("class", "axis axis--y")
				.call(makeAxis(d3.axisLeft(y), properties.yAxis, height, elements.length));
		}

		g.selectAll(".bar")
		  .data(elements)
		  .enter().append("rect")
			.attr("class", "bar")
			.attr("x", function(d) { return x(d.name || d.id); })
			.attr("y", function(d) { return y(d.value); })
			.attr("width", x.bandwidth())
			.attr("height", function(d) { return height - y(d.value); })
            .style("fill", function (d, i) {
                return d.color || getColor(i);
            })
            .attr(properties.id_key, function (d, i) {
                return d.id || d.name || (properties.id_key + "_" + i);
            });
			
		var tip = null;
		if (properties.tooltip.enabled) {
			tip = d3.tip()
			   .attr("class", "d3-tip " + properties.tooltip.className)
			   .attr("chart-view", properties.viewName)
			   .offset([-10, 0])
			   .html(function (d) {
				   return properties.tooltip.make(d, properties);
			   });
			svg.call(tip);
		}

		svg.append("rect")
			.attr("transform", "translate(" + margin.left + "," + margin.top + ")")
			.attr("class", "overlay")
			.attr("width", width)
			.attr("height", height)
			.on("mouseover", function() {
				if (properties.$references) {
					hoverElement($container.find(format(".bar[{0}]", properties.id_key)), 
						properties.$references.filter(format("[{0}]", properties.id_key)), false);
				}
				if (tip) {
					tip.hide();
				}
			})
			.on("mouseout", function() {
				if (properties.$references) {
					hoverElement($container.find(format(".bar[{0}]", properties.id_key)), 
						properties.$references.filter(format("[{0}]", properties.id_key)), false);
				}
				if (tip) {
					tip.hide();
				}
			})
			.on("mousemove", function() {
				var x0 = d3.mouse(this)[0],
					eachBand = x.step(),
					i = Math.floor(x0 / eachBand),
					d = elements[i];
				if (!d) {
					return;
				}
				if (properties.$references) {
					var key = d.id || d.name || (properties.id_key + "_" + i);
					hoverElement($container.find(format(".bar[{0}]", properties.id_key)), 
						properties.$references.filter(format("[{0}]", properties.id_key)), false);
					hoverElement($container.find(format(".bar[{0}={1}]", properties.id_key, key)), 
						properties.$references.filter(format("[{0}={1}]", properties.id_key, key)), true);
				}
				if (tip) {
					tip.show.call(this, d);
				}
			});
			
		if (properties.$references) {
			properties.$references
				.off("mouseover")
				.off("mouseout")
				.on("mouseover", function () {
					hoverElement($container.find(format(".bar[{0}={1}]", properties.id_key, $(this).attr(properties.id_key))), 
						$(this), true);
				})
				.on("mouseout", function () {
					hoverElement($container.find(format(".bar[{0}={1}]", properties.id_key, $(this).attr(properties.id_key))), 
						$(this), false);
				});
		}
	}
	
	var line_default = {
		viewName : "",
		size : null,
		xAxis : {
			ticks : "split"
		},
		yAxis : {
			ticks : "size",
			pixelsPerTick : 30
		},
		margin : {
			left : 40,
			right : 5,
			top : 5,
			bottom : 30
		},
		area : true,
		tooltip : {
			enabled : true,
			className : "linechartTooltip",
			make : function(data, properties) {
				return format("{0} ({1})",
					data.x.label || data.x.value || "",
					data.y.label || data.y.value || 0);
			}
		},
		color : "#202121"
	};
	
	function line($container, data, props) {
		var properties = makeProperties(line_default, props);
		
		$container.html("");
		
		var width;
		var height;
		if (properties.size) {
			width = properties.size.width;
			height = properties.size.height;
			$container
				.width(width)
				.height(height);
		}
		else {
			width = $container.width();
			height = $container.height();
		}
		
		var margin = {
			left : properties.margin && properties.margin.left || 0,
			right : properties.margin && properties.margin.right || 0,
			top : properties.margin && properties.margin.top || 0,
			bottom : properties.margin && properties.margin.bottom || 0
		};
		
		width = width - margin.left - margin.right;
		height = height - margin.top - margin.bottom;

		var bisector = d3.bisector(function(d) { return d.x.value; }).left;

		var x = d3.scaleLinear().range([0, width]);
		var y = d3.scaleLinear().range([height, 0]);
	
		x.domain(d3.extent(data, function(d) { return d.x.value; }));
		y.domain([d3.min(data, function(d) { return d.y.value; }) / 1.005, 
			d3.max(data, function(d) { return d.y.value; }) * 1.005]);
		
		var container = d3.select($container[0]);
		var svg = container.append("svg")
            .attr("width", "100%")
            .attr("height", "100%");
	
		var g = svg.append("g")
			.attr("transform", "translate(" + margin.left + "," + margin.top + ")");
		
		if (properties.xAxis) {
			g.append("g")
				.attr("class", "axis axis--x")
				.attr("transform", "translate(0," + height + ")")
				.call(makeAxis(d3.axisBottom(x), properties.xAxis, width, data.length));
		}
		if (properties.yAxis) {
			g.append("g")
				.attr("class", "axis axis--y")
				.call(makeAxis(d3.axisLeft(y), properties.yAxis, height, data.length));
		}
	
		if (properties.area) {
			var area = d3.area()
				.x(function(d) { return x(d.x.value); })
				.y0(height)
				.y1(function(d) { return y(d.y.value); });

			g.append("path")
				.datum(data)
				.attr("class", "area")
				.attr("fill", properties.color)
				.attr("d", area);
		}

		var line = d3.line()
			.x(function(d) { return x(d.x.value); })
			.y(function(d) { return y(d.y.value); });
		
		g.append("path")
			.datum(data)
			.attr("class", "line")
			.attr("stroke", properties.color)
			.attr("d", line);

		var focus = g.append("g")
			.attr("class", "focus")
			.style("display", "none");

		focus.append("line")
			.attr("class", "x-hover-line hover-line")
			.attr("stroke", properties.color)
			.attr("y1", 0)
			.attr("y2", height);

		focus.append("circle")
			.attr("class", "bigCircle")
			.attr("stroke", properties.color);

		focus.append("circle")
			.attr("class", "smallCircle")
			.attr("stroke", properties.color);
	
		var tip = null;
		if (properties.tooltip.enabled) {
			tip = d3.tip()
			   .attr("class", "d3-tip " + properties.tooltip.className)
			   .attr("chart-view", properties.viewName)
			   .offset([-10, 0])
			   .html(function (d) {
				   return properties.tooltip.make(d, properties);
			   });
			svg.call(tip);
		}

		svg.append("rect")
			.attr("transform", "translate(" + margin.left + "," + margin.top + ")")
			.attr("class", "overlay")
			.attr("width", width)
			.attr("height", height)
			.on("mouseover", function() { 
				focus.style("display", null); 
			})
			.on("mouseout", function() {
				focus.style("display", "none");
				if (tip) {
					tip.hide();
				}
			})
			.on("mousemove", function() {
				var x0 = x.invert(d3.mouse(this)[0]),
					i = bisector(data, x0, 1),
					d0 = data[i - 1],
					d1 = data[i],
					d = x0 - d0.x.value > d1.x.value - x0 ? d1 : d0;
				focus.attr("transform", "translate(" + x(d.x.value) + "," + y(d.y.value) + ")");
				focus.select(".x-hover-line").attr("y2", height - y(d.y.value));
				if (tip) {
					tip.show.call(this, d);
				}
			});
	}
	
	var multiline_default = {
		viewName : "",
		size : null,
		xAxis : {
			ticks : "split"
		},
		yAxis : {
			ticks : "size",
			pixelsPerTick : 30
		},
		margin : {
			left : 40,
			right : 5,
			top : 5,
			bottom : 30
		},
		area : true,
		tooltip : {
			enabled : true,
			className : "multilinechartTooltip",
			containerTemplate : "".concat(
				"<div class='multilineTooltipContainer'>",
					"<div class='multilineTooltipX'>",
						"{{ xLabel }}",
					"</div>",
					"<div tooltipIndicators class='multilineTooltipYs'>",
						"{{{ indicators }}}",
					"</div>",
				"</div>"
			),
			indicatorTemplate : "".concat(
				"<div class='multilineTooltipIndicator'>",
					"<div class='multilineTooltipIndicatorRef' style='background-color: {{ color }}'>",
					"</div>",
					"<div class='multilineTooltipIndicatorName'>",
						"{{ name }}",
					"</div>",
					"<div class='multilineTooltipIndicatorValue'>",
						"{{ value }}",
					"</div>",
				"</div>"
			),
			make : function(data, indicators, properties) {
				return Mustache.render(properties.tooltip.containerTemplate, {
					xLabel : data.x.label || data.x.value || "",
					indicators : _.reduce(indicators, function(memo, indicator, index) {
						return memo + Mustache.render(properties.tooltip.indicatorTemplate, {
							name : indicator.name || indicator.id,
							value : data[indicator.id].label || data[indicator.id].value || 0,
							color : indicator.color || getColor(index)
						});
					}, "")
				});
			}
		},
		color : "#000"
	};
	
	function multiline($container, indicators, data, props) {
		var properties = makeProperties(multiline_default, props);
		
		$container.html("");
		
		var width;
		var height;
		if (properties.size) {
			width = properties.size.width;
			height = properties.size.height;
			$container
				.width(width)
				.height(height);
		}
		else {
			width = $container.width();
			height = $container.height();
		}
		
		var margin = {
			left : properties.margin && properties.margin.left || 0,
			right : properties.margin && properties.margin.right || 0,
			top : properties.margin && properties.margin.top || 0,
			bottom : properties.margin && properties.margin.bottom || 0
		};
		
		width = width - margin.left - margin.right;
		height = height - margin.top - margin.bottom;

		var bisector = d3.bisector(function(d) { return d.x.value; }).left;

		var x = d3.scaleLinear().range([0, width]);
		var y = d3.scaleLinear().range([height, 0]);
	
		x.domain(d3.extent(data, function(d) { return d.x.value; }));
		y.domain([
			props.min !== undefined? props.min : d3.min(data, function(d) {
				var min = 0;
				$.each(indicators, function(_, indicator) {
					min = Math.min(min, d[indicator.id].value);
				});
				return min;
			}) / 1.005, 
			props.max !== undefined? props.max : d3.max(data, function(d) { 
				var max = 0;
				$.each(indicators, function(_, indicator) {
					max = Math.max(max, d[indicator.id].value);
				});
				return max;
			}) * 1.005
		]);
		
		var container = d3.select($container[0]);
		var svg = container.append("svg")
            .attr("width", "100%")
            .attr("height", "100%");
	
		var g = svg.append("g")
			.attr("transform", "translate(" + margin.left + "," + margin.top + ")");
		
		if (properties.xAxis) {
			g.append("g")
				.attr("class", "axis axis--x")
				.attr("transform", "translate(0," + height + ")")
				.call(makeAxis(d3.axisBottom(x), properties.xAxis, width, data.length));
		}
		if (properties.yAxis) {
			g.append("g")
				.attr("class", "axis axis--y")
				.call(makeAxis(d3.axisLeft(y), properties.yAxis, height, data.length));
		}
		
		var focusList = [];
	
		$.each(indicators, function(index, indicator) {
			var color = indicator.color || getColor(index);
			
			if (properties.area) {
				var area = d3.area()
					.x(function(d) { return x(d.x.value); })
					.y0(height)
					.y1(function(d) { return y(d[indicator.id].value); });

				g.append("path")
					.datum(data)
					.attr("class", "area")
					.attr("fill", color)
					.attr("d", area);
			}

			var line = d3.line()
				.x(function(d) { return x(d.x.value); })
				.y(function(d) { return y(d[indicator.id].value); });

			g.append("path")
				.datum(data)
				.attr("class", "line")
				.attr("stroke", color)
				.attr("d", line);
		
			var focus = g.append("g")
			.attr("class", "focus")
			.style("display", "none");

			focus.append("line")
				.attr("class", "x-hover-line hover-line")
				.attr("stroke", color)
				.attr("y1", 0)
				.attr("y2", height);

			focus.append("circle")
				.attr("class", "bigCircle")
				.attr("stroke", color);

			focus.append("circle")
				.attr("class", "smallCircle")
				.attr("stroke", color);
		
			focusList.push(focus);
		});
	
		var tip = null;
		if (properties.tooltip.enabled) {
			tip = d3.tip()
			   .attr("class", "d3-tip " + properties.tooltip.className)
			   .attr("chart-view", properties.viewName)
			   .offset([-10, 0])
			   .html(function (d) {
				   return properties.tooltip.make(d, indicators, properties);
			   });
			svg.call(tip);
		}

		svg.append("rect")
			.attr("transform", "translate(" + margin.left + "," + margin.top + ")")
			.attr("class", "overlay")
			.attr("width", width)
			.attr("height", height)
			.on("mouseover", function() {
				$.each(focusList, function(_, focus) {
					focus.style("display", null);
				}); 
			})
			.on("mouseout", function() {
				$.each(focusList, function(_, focus) {
					focus.style("display", "none");
				});
				if (tip) {
					tip.hide();
				}
			})
			.on("mousemove", function() {
				var x0 = x.invert(d3.mouse(this)[0]),
					i = bisector(data, x0, 1),
					d0 = data[i - 1],
					d1 = data[i],
					d = x0 - d0.x.value > d1.x.value - x0 ? d1 : d0;
				$.each(focusList, function(index, focus) {
					var indicator = indicators[index];
					focus.attr("transform", "translate(" + x(d.x.value) + "," + y(d[indicator.id].value) + ")");
					focus.select(".x-hover-line").attr("y2", height - y(d[indicator.id].value));
				});
				if (tip) {
					tip.show.call(this, d);
				}
			});
	}

	if (!inConcert.charts) {
		inConcert.charts = {
			cleanView : cleanView,
			pie : pie,
			bars : bars,
			line : line,
			multiline : multiline
		};
	}

}());