/*jshint quotmark:false, esversion: 11*/
/* global jQuery */

/*!
 * Plug-In				:	jQuery ToolTip
 * Version				:	1.0
 * Autor					:	Ellio de Mello
 * Descripcion			:	Crea un tooltip en un elemento del html!!
 * Liberado				:	2010-08-23 GMT-500
 * Archivo				:	tooltip.js
 * Dependencias			:	jQuery v1.9.1      (JS)
 
 * ----------------------------------------------------------------------------------------------------------
 * Opcion		Tipo		Default					Descripcion
 * ----------------------------------------------------------------------------------------------------------
 * className		(alpha)		"genericTooltip"	Clase a asignar al tooltip.
 * alignment		(alpha)		"right-bottom"		Ubicacion donde se mostrará el tooltip, con respecto a su duenio.
 * show-delay		(num)		500					Tiempo en milisegundos que debe estar posicionado con el mouse antes de que se muestre el tooltip.
 * remove-delay		(num)			2000			Tiempo en milisegundos que se visualiza el tooltip.
 * detach			(boolean)		undefined		Si tiene el valor en true indica que deben de eliminarse los eventos que muestran/ocultan el tooltip.
 * bind_selector	(string)		undefined		Si el valor no se recibe por parametro, el bindeo se realiza estatico (viejo metodo bind/unbind). Si viene un selector, el bindeo se realiza de forma dinamica (viejos metodos live/die)
 
 * ----------------------------------------------------------------------------------------------------------
 *
 * Sintaxis
 * ---------
 * HTML		:	<div id="demo" />
 * JS		:	$(selector).tooltip([{opc1: "val1", .., .., ..}]);
 *
 * Ejemplos
 * ---------
 *	1)
 *	HTML	:	<input id="id1" type="button" value="boton1" />
 *				<input id="id2" type="button" value="boton2" />
 *	JS		:	$("#id1").inconcert_tooltip({message: "tooltip boton1"});
 *				$("#id2").inconcert_tooltip({message: "tooltip boton2"});
 *	2)
 *	HTML	:	<input tooltip="tooltip boton1" id="id1" type="button" value="boton1" />
 *				<input tooltip="tooltip boton2" type="button" value="boton2" />
 *	JS		:	$("input [tooltip]").inconcert_tooltip();
 *
*/

(function($) {
	"use strict";

	$.fn.inconcert_tooltip = function(user_options) {
		//opciones x default del tooltip
		var default_options = {
			"className": "genericTooltip",
			"alignment": "right-center",
			"show-delay": 50,
			"remove-delay": 2000,
			message: "",
			content: function(ownerControl) {
				return typeof this.message === "function" ? this.message(ownerControl) : this.message;
			}
		};
		var options = $.extend(default_options, user_options);
		var timeoutId;

		//------------------------------------------
		//retorna el nombre de la clase a aplicar al tooltip
		var getTooltipClass = function(ownerControl) {
			return ownerControl.attr("tooltipclass") ? ownerControl.attr("tooltipclass") : options.className;
		};

		//------------------------------------------
		//calcula y retorna la posicion donde se ubicará el tooltip
		var getPosition = function(ownerControl) {
			var tooltipClass = getTooltipClass(ownerControl);
			//alineacion del control
			var alignmentMatch = options.alignment.match(new RegExp("(.*)-(.*)"));
			if (!alignmentMatch) {
				return;
			}
			var alignmentHorizontal = alignmentMatch[1].toLowerCase();
			var alignmentVertical = alignmentMatch[2].toLowerCase();
			var tooltip = $("." + tooltipClass); // el tooltip posta!!
			var ownerPos = ownerControl.position(); // las posiciones del control duenio del tooltip
			var tooltipPos = {top: 0, left: 0};
			if (alignmentVertical === "bottom") {
				// parte inferior del control owner (es donde va a comenzar a visualizarse el tooltip)
				// the magic number is to not get sticked to the owner control
				tooltipPos.top = ownerControl.outerHeight(true) + 5;
			}
			else {
				// parte superior del control owner (es donde va a comenzar a visualizarse el tooltip)
				tooltipPos.top = - tooltip.outerHeight(true);
			}
			if (alignmentHorizontal === "right") {
				// right aligned owner control.
				tooltipPos.left = ownerControl.outerWidth(true) / 2; // divided by 2, to begin in the middle of owner control
				// Aca valido que el tooltip no quede fuera del borde de la derecha de la ventana.
				// - ownerPos.left            = es la posicion a la izquierda donde se encuentra ubicado el control owner!!!
				// - tooltipPos.left          = contiene el ancho del control owner!!!
				// - tooltip.outerWidth(true) = el ancho del tooltip!!!
				// *** la suma de estos 3 nos da la posicion donde termina el tooltip.
				if (window.innerWidth < (ownerPos.left + tooltipPos.left + tooltip.outerWidth(true))) {
					// en caso de que se cumpla lo anterior, al tooltip lo ubico desde donde comienza el control owner - el ancho del tooltip
					// y un pequenio numero magico!
					// ergo, tengo que restar en vez de sumar.
					tooltipPos.left = - (tooltip.outerWidth(true) / 2);
				}
			}
			else {
				// left aligned owner control.
				tooltipPos.left = - (tooltip.outerWidth(true) / 2); // divided by 2, to begin in the middle of owner control
			}
			//-------
			tooltipPos.top = ownerPos.top + tooltipPos.top;
			tooltipPos.left = ownerPos.left + tooltipPos.left;
			return tooltipPos;
		};

		//------------------------------------------
		//funcion que remueve el markup del tooltip
		var detachTooltipFrom = function(ownerControl) {
			var tooltipClass = getTooltipClass(ownerControl);
			$('.' + tooltipClass).remove();
			clearTimeout(timeoutId);
		};

		//------------------------------------------
		// retorna el mensaje del tooltip a visualizar. A este mensaje lo obtiene en este orden:
		//		1 - Se muestra el mensaje recibido por parametro por el usuario del tooltip.
		//		2 - Se muestra el mensaje que tenga asociado el markup (duenio del tooltip).
		//		3 - Da un error y muestra el contenido del html donde falta el atributo "tooltip".
		var getTooltipMessage = function(ownerControl) {
			var userMessage = options.content(ownerControl);
			if (userMessage && userMessage.length > 0) {
				return userMessage;
			}
			var ownerMessage = ownerControl.attr("tooltip");
			if (ownerMessage && ownerMessage.length > 0) {
				return ownerMessage;
			}
			return false;
		};

		//------------------------------------------
		//funcion que genera el markup del tooltip y lo muestra en la ubicacion indicada
		var showTooltipFor = function(ownerControl) {
			detachTooltipFrom(ownerControl);
			// obtengo el mensaje del tooltip.
			var tooltipMessage = getTooltipMessage(ownerControl);
			if (!tooltipMessage) {
				return;
			}
			// obtengo la clase a aplicar en el tooltip
			var tooltipClass = getTooltipClass(ownerControl);
			//al padre del control le agrego el markup del tooltip.
			ownerControl.parent().append('<div class="' + tooltipClass + '" style="display:none;">' + tooltipMessage + '</div>');
			//obtengo la posicion del tooltip
			var tooltipPosition = getPosition(ownerControl);
			if (!tooltipPosition) {
				detachTooltipFrom(ownerControl);
				return;
			}
			//posiciono y muestro el tooltip
			var tooltip = $("." + tooltipClass); // el tooltip posta!!
			tooltip.css(tooltipPosition).show();
			//elimino el timeout del show
			clearTimeout(timeoutId);
			//agrego un timeout para removerlo luego de superado el dealy de remove!!
			timeoutId = setTimeout(function(){
				detachTooltipFrom(ownerControl);
			}, options["remove-delay"]);
		};

		// if the user use the plugin from an already 
		// built node, we use that, otherwise, go and select it.
		// also this works with detached from dom nodes too.
		var $selector = this.length > 0 ? this : $(this.selector);

		if (options.detach) {
			if ( !options.bind_selector ) {
			// unbind estatico (viejos metodos bind/unbind)
				$selector
					.off( "mouseover.ic_tooltip" )
					.off( "mouseout.ic_tooltip"  );
			}
			else {
			// unbind dinamico (viejos metodos live/die)
				$selector
					.off( "mouseover.ic_tooltip", options.bind_selector )
					.off( "mouseout.ic_tooltip" , options.bind_selector );
			}
		}
		else {
			//--------------------------------------------------------
			//handlers que manipulan el mouse in/ mouse out!!!!
			var mouseover = function() 
			{
				var ownerControl = $(this);
				timeoutId = setTimeout(function() {
					showTooltipFor(ownerControl);
				}, options["show-delay"]);
			};
			var mouseout = function() 
			{
				detachTooltipFrom($(this));
			};

			if ( !options.bind_selector ) {
			// unbind estatico (viejos metodos bind/unbind)
				$selector
					.off( "mouseover.ic_tooltip"            )
					.on ( "mouseover.ic_tooltip", mouseover )
					.off( "mouseout.ic_tooltip"             )
					.on ( "mouseout.ic_tooltip" , mouseout  );
			}
			else {
			// unbind dinamico (viejos metodos live/die)
				$selector
					.off( "mouseover.ic_tooltip", options.bind_selector            )
					.on ( "mouseover.ic_tooltip", options.bind_selector, mouseover )
					.off( "mouseout.ic_tooltip" , options.bind_selector            )
					.on ( "mouseout.ic_tooltip" , options.bind_selector, mouseout  );
			}
		}

		return this;
	};
})(jQuery);
