/*!
 * \file jquery-paginator.js
 * \author ichramm
 *
 * Created on July 26, 2013, 05:16 PM
 */
/*jshint quotmark:false, esversion: 11*/
/*global jQuery*/

(function($, undefined) {
	"use strict";

	/*!
	 * The plugin name is also the function the user will invoke in the jQuery object
	 */
	var plugin_name = "paginator";

	/*!
	 * Paginator's default options
	 */
	var default_options = {

		/*!
		 * jQuery object where to append rows - required
		 */
		data_container: undefined,

		/*!
		 * Where to place the paginator
		 */
		pages_container: undefined,

		/*!
		 * Format string for display
		 *
		 * The \c text_format variable accepts the following variables:
		 *
		 * \li \c {page_number}    Replaced with the current page number
		 * \li \c {page_row_start} Replaced with the index of the first row in the current page
		 * \li \c {page_row_end}   Replaced with the index of the last row in the current page
		 * \li \c {row_count}      Replaced with the total number of rows
		 * \li \c {page_count}     Replaced with the total number of pages
		 * \li \c {many}           Replaced with the total number of rows when done, or with the value of \c translated_many
		 *
		 * Example:
		 *
		 * \code 'Showing rows from {page_row_start} to {page_row_end} of {many}' \endcode
		 *
		 * If row_count is 83 and rows_per_page is set to 10, when the user enters the 5th page
		 * the text will be shown as:
		 *
		 * \code 'Showing rows from 51 to 60 of 83' \endcode
		 */
		text_format: undefined,

		/*! Set the corresponding localized text for this option */
		translated_many: "many",

		/*! Accepted values: back, front */
		text_position: 'back',

		/*! how many records? if known */
		row_count: undefined,

		/*! how many rows per page? */
		rows_per_page: 10,

		/*! html rows number per db record */
		rows_number_per_record: 1,

		/*! how many page buttons must show at the left (accepts number or the string 'all') */
		pages_at_left: 2,

		/*! how many page buttons must show at the right (accepts number or the string 'all') */
		pages_at_right : 2,

		/*!
		 * Sets how many pages must be kept visible, default: pages_at_left + pages_at_right - 1
		 * The paginator will be force to pad with pages from the left when there are no more pages
		 * at the right, and, to pad with pages from the right when there are no more pages at the left
		 *
		 * Imagine the following scenario:
		 * H. There are 12 pages.
		 * H. Configuration: pages_at_left = pages_at_right = 4
		 * H. Current visible page is 10
		 * Q. What pages will be shown?
		 * A. -> The paginator will show pages 6,7,8,9,10,11,12 (total: 7)
		 * Q. What if show_at_least to 8?
		 * A. -> The paginator will also show page 5, because there are no more pages to the right
		 */
		show_at_least: undefined,

		/*! show button to first page? ignored when \c pages_at_left = 'all' */
		show_first_page: true,

		/*! show button to last page? ignored when \c pages_at_right = 'all' */
		show_last_page : true,

		/*! When this is set, pages_at_left, pages_at_right, show_first_page, and show_last_page have no effect */
		show_all_pages : false,

		/*! show 'next' button? */
		show_next      : true,

		/*! show 'prev' button? **/
		show_prev      : true,

		/*!  // 'prev' button text */
		previous       : "← previous",

		/*! 'next' button text */
		next           : "next →",

		/*! If you already have a dataset, put it here! */
		items          : [],

		/*!
		 * Selector used to exclued elements when hiding elements.
		 * Must be a valid jQuery selector
		 *
		 * Example:
		 * \li ignore elements having the 'header' class
		 * \code ignored_elements: '.header' \endcode
		 *
		 * \li ignore elements of type input
		 * \code ignored_elements: '[type="input"]' \endcode
		 *
		 * \li Combine both selectors
		 * \code ignored_elements: '.header, [type="input"]' \endcode
		 */
		ignored_elements : "",

		/*!
		 * A function to call when the user enters a new page:
		 *
		 * callback: function(page_number, need_data, on_done) {
		 *    on_done();
		 * }
		 *
		 * \remarks You must call \c on_done(), or the paginator will not work properly
		 */
		callback       : null
	};

	/*!
	 * jQuery Pagination plugin
	 *
	 * Initialization:
	 *
	 * \code
	 * $('#paginator_container').paginator({
	 *     container: $('#mytable').find('tbody')
	 *     // more options ...
	 * });
	 * \endcode
	 *
	 * Appending rows:
	 *
	 * \code
	 * $('#paginator_container').paginator('append', $row);
	 * \endcode
	 *
	 * And when running out of data:
	 *
	 * \code
	 * $('#paginator_container').paginator('done');
	 * \endcode
	 *
	 *
	 * \remarks DO NOT append rows directly to your container, use
	 * the paginator instead.
	 *
	 * <p>
	 * <b>CSS Classes:</b>
	 *
	 * \li \c pg-main-container    Class applied to the main container
	 * \li \c pg-display-text      Class applied to the display text (\see text_format)
	 * \li \c pg-buttons-container Class applied to the div containing all the buttons
	 * \li \c pg-button            Class applied to each button, including previous and next buttons
	 * \li \c pg-button-disabled   Class applied to a button when disabled
	 * \li \c pg-button-page       Class applied to all button that represent a page
	 * \li \c pg-button-current    Class applied to the button representing the current page
	 * \li \c pg-button-prev       Class applied to the 'previous' button, i.e: The button that turns back the page
	 * \li \c pg-button-next       Class applied to the 'next' (aka forward) button
	 * \li \c pg-button-hidden     Class applied to the buttons being hidden
	 * \li \c pg-button-dotted     Class applied to the dotted area filling blank spaces
	 * </p>
	 */
	function Paginator( options )
	{
		var self = this;

		/*! pages indexed by page number (row 0 is empty) */
		this.pages = [];

		this.options = $.extend({}, default_options, options);

		// esta propiedad tiene el total real de registros existentes en el server (aunque la grilla nunca proyecte la totalidad)
		this.m_total_rows = null; // opcional

		/*! set by the user when calling \c no_more_pages() */
		this.no_more_data = this.options.row_count && this.total_number_of_items() >= this.options.row_count;

		// validate containers are properly set
		if (!this.options.data_container) {
			throw ("Data Container must not be empty");
		}
		if (!this.options.pages_container) {
			throw ("Pages Container must not be empty");
		}

		this.options.on_changed_page = this.options.on_changed_page || $.noop;

		if (this.options.show_all_pages) {
			this.options.show_first_page = true;
			this.options.show_last_page = true;
			this.options.pages_at_left  = 'all';
			this.options.pages_at_right = 'all';
		} else if (!this.options.show_at_least) {
			this.options.show_at_least = -1;
			if ( typeof(this.options.pages_at_left) === "number") {
				this.options.show_at_least += this.options.pages_at_left;
			}
			if ( typeof(this.options.pages_at_right) === "number") {
				this.options.show_at_least += this.options.pages_at_right;
			}
		}

		// show_at_least includes first and last page
		if (this.options.show_first_page) {
			--this.options.show_at_least;
		}
		if (this.options.show_last_page) {
			--this.options.show_at_least;
		}

		this.$pages_container   = this.options.pages_container;
		this.$data_container    = this.options.data_container;
		// para que el append de los rows se realice en memoria y luego se vuelquen todos de una unica vez en el DOM.
		this.append_with_detach = this.options.append_with_detach || false; // Despues del primer congelado del 2024 esta opcion debe de ser por defecto en true !!!!!!!!!!!!

		// bind this instance to the containe in case it's needed
		var old_instance = this.$pages_container.data(plugin_name);
		if (old_instance) {
			old_instance = old_instance.destroy();
		}
		this.$pages_container.data(plugin_name, this);

		// empty paginator container in order to insert required elements
		this.$pages_container.html('');
		if (!this.$pages_container.hasClass('pg-main-container')) {
			this.$pages_container.addClass('pg-main-container');
		}

		var append_display_text = function() {
			// display text, in case is set
			self.$pages_container.append(
				$('<div class="pg-display-text"></div>')
					.append('<span></span>')
			);
		};

		// append text container if it goes first
		if (this.options.text_format && this.options.text_position === "front") {
			append_display_text();
		}

		// container for prev, next and all page buttons
		this.$buttons_container = $('<div class="pg-buttons-container"></div>');
		this.$pages_container.append(this.$buttons_container);

		this.$buttons_container.on("click", '.pg-button', function()
		{
			var $button = $(this);
			if ($button.hasClass('pg-button-disabled')) {
				return;
			}
			if ($button.hasClass('pg-button-page') || $button.hasClass('pg-button-dotted')) {
				var page = parseInt( $button .attr('page-number'), 10);
				if (page !== self.current_page) {
					self.__jump_to_page(page);
				}
			} else if ($button.hasClass('pg-button-prev')) {
				if (self.current_page > 1) {
					self.__disable_button($button);
					self.__jump_to_page(self.current_page - 1, function() {
						if (self.current_page > 1)
							self.__enable_button($button);
					});
				}
			} else if ($button.hasClass('pg-button-next')) {
				if (!self.no_more_data || self.current_page < self.page_count()) {
					self.__disable_button($button);
					self.__jump_to_page(self.current_page + 1, function() {
						if (self.current_page < self.page_count())
							self.__enable_button($button);
					});
				}
			}
		});

		// append text container if it goes last
		if (this.options.text_format && this.options.text_position === "back") {
			append_display_text();
		}

		// previous button, in case is set
		if (this.options.show_prev) {
			var $prev_button = $('<a href="javascript:void(0)" class="pg-button pg-button-prev"></a>')
				.text(this.options.previous);
			this.$buttons_container.append($prev_button);
		}

		// append first page
		this.__append_page();

		// next button, in case is set
		if (this.options.show_next) {
			var $next_button = $('<a href="javascript:void(0)" class="pg-button pg-button-next"></a>')
				.text(this.options.next);
			this.$buttons_container.append($next_button);
		}

		// append all initial items
		this.options.row_count = 0; // este valor se incrementa en la funcion "__append"
		// si el append se debe de realizar en memoria quito el contenedor de rows del DOM.
		let $data_container_parent = null;
		if (this.append_with_detach) {
			$data_container_parent = this.$data_container.parent();
			this.$data_container.detach();
		}
		// realizo el append
		$.each( this.options.items || [], function( i, $element ) {
			self.__append( $element, i );
		});
		// si el append se realizo en memoria, debo volver a instar el contenedor de rows en el DOM.
		if (this.append_with_detach) {
			$data_container_parent.append(this.$data_container);
		}

		// enter first row
		this.__jump_to_page(1);
	}

	/*!
	 * Destroys the paginator
	 *
	 * \remarks Does not remove elements form the container
	 */
	Paginator.prototype.destroy = function()
	{
		this.$pages_container.data(plugin_name, null);

		// delete everything (dom objects, event handlers, data, etc)
		var $children = this.$pages_container.children();
		$children.remove();

		// and cleanup the div (just to be sure)
		this.$pages_container
			.removeClass('pg-main-container')
			.html(''); // we prefer not to hide the container

		this.pages = [ [] ];

		// S eliminan los rows del DOM
		var $rows = this.$data_container.children();
		$rows.remove();

		return null;
	};

	Paginator.prototype.$get_rows = function()
	{
		var $rows = this.$data_container.children();
		return $rows;
	};

	/*!
	 * \return main rows items count ( no subgrid )
	 */
	Paginator.prototype.total_number_of_items = function() {
		// se divide entre 2 para que no cuentes los rows de la subgrid
		var count = this.options.items.length / this.options.rows_number_per_record;
		return parseInt( count, 10 );
	};

	/*!
	 * \return main rows items count ( no subgrid )
	 */
	Paginator.prototype.number_items_per_page = function( page_number ) {
		// se divide entre 2 para que no cuentes los rows de la subgrid
		if (!this.pages) {
			return 0;
		}
		var page_rows = this.get_page_rows(page_number);
		if (!page_rows) {
			return 0;
		}
		var count = page_rows.length / this.options.rows_number_per_record;
		return parseInt( count, 10 );
	};

	/*!
	 * \return The current page count
	 */
	Paginator.prototype.page_count = function() {
		return this.pages.length;
	};

	/*!
	 * Inserts the element \p $element at the end of the container
	 */
	Paginator.prototype.append = function( $element, index ) {
		var page_added = this.__append( $element, index );
		if ( page_added ) {
			this.__update_page_navigator();
		} else {
			this.update_footer_total_text();
		}
	};

	/*!
	 * Removed the row \p $row from the container
	 */
	Paginator.prototype.remove = function( $row, is_main_row )
	{
		const coordinates = $row.data('paginator.coordinates');
		is_main_row = is_main_row === undefined ? true : is_main_row;
		$row.remove();
		if ( this.options.row_count && is_main_row ) {
			this.options.row_count--;
		}
		if (coordinates) {
			this.__on_item_removed(coordinates.page, coordinates.row_position);
		}
		if ( this.options.row_count === 0 ) {
			this.destroy();
		}
	};

	/*!
	 * Called to tell the paginator there will be no more rows
	 *
	 * The means the paginator will disable 'next' button and
	 * will not ask for more rows.
	 */
	Paginator.prototype.no_more_pages = function() {
		this.no_more_data = true;
		this.update_footer_total_text();
	};

	/*!
	 * refresh de current page.
	 */
	Paginator.prototype.refresh_page = function( on_refresh ) {
		var force = true;
		var current_page = this.current_page;
		this.__jump_to_page( current_page, on_refresh, force );
	};

	/*!
	 * return the current page number.
	 */
	Paginator.prototype.get_current_page_number = function() {
		return this.current_page;
	};

	/*!
	 * return the total rows
	 * 	Al total que retorna el paginador hay que restarle 1.
	 */
	Paginator.prototype.get_total_rows = function() {
		if (this.m_total_rows) {
			return this.m_total_rows;
		}
		return this.options.row_count - 1;
	};

	/*!
	 * set the total rows
	 * 	Puede cambiar el total de rows, cuando en la grilla no vamos a mostrar mas de N rows paginadas, pero total verdadero es mayor (el real).
	 */
	Paginator.prototype.set_total_rows = function(total_rows) {
		this.m_total_rows = total_rows || null;
	};

	/* Private functions */

	/*! low level version, does not update GUI */
	Paginator.prototype.__append = function( $row, index )
	{
		const row_type = $row.attr("row_type");
		if ("main" === row_type) {
			this.options.row_count++;
		}

		if ( $row.parent()[0] !== this.$data_container[0] ) {
			this.$data_container.append( $row );
		}

		var page_rows;
		var page_count = this.page_count();
		var page_added = false;

		if ( page_count === 0 || (this.number_items_per_page( page_count ) === this.options.rows_per_page) ) {
			page_rows = this.__append_page();
			page_count++;
			page_added = true;
		}

		if ( !page_rows ) {
			page_rows = this.get_page_rows( page_count );
		}

		page_rows.push( $row );

		// note: coordinates are zero-based (this.pages[0] is always empty)
		var coordinates = {
			page: page_count,
			row_position : page_rows.length - 1
		};
		$row.data( 'paginator.coordinates', coordinates );

		// if the element belongs to a page different than the current one, then it must be hidden
		if ( this.current_page !== page_count ) {
			$row.hide();
		}
		if ( typeof this.options.on_append_one_row === 'function' ) {
			this.options.on_append_one_row( $row, index );
		}

		return page_added;
	};

	Paginator.prototype.__append_page = function()
	{
		this.pages.push([]);
		var page_number = this.page_count();

		var $new_page_button = $('<a href="javascript:void(0)" class="pg-button pg-button-page"></a>')
			.attr('page-number', page_number)
			.text(page_number);

		if (page_number === 1) {
			this.$buttons_container.append($new_page_button);
		} else {
			var $last_page = this.$buttons_container.children('.pg-button-page').last();
			$last_page.after($new_page_button);
		}
		return this.pages[this.pages.length - 1];
	};

	Paginator.prototype.__remove_last_page = function()
	{
		var last_page = this.pages.length - 1;
		if ( typeof(last_page) !== "number" || !$.isArray( this.pages ) || !$.isArray( this.pages[last_page] ) ) {
			return false;
		}
		this.pages.splice(last_page);
		return true;
	};

	/*!
	 * Updates all elements to the right of the element being removed
	 * May become a very expensive function
	 */
	Paginator.prototype.__on_item_removed = function( page_number, row_position )
	{
		var self = this;

		var $page_rows = this.get_page_rows( page_number );

		if ( !$page_rows ) {
			this.__jump_to_page( 1 );
			return;
		}

		var update_coordinates = function( current_page, row_pos )
		{
			var $current_page_rows = self.get_page_rows( current_page );
			if (!$current_page_rows) {
				return;
			}
			row_pos = row_pos || 0;
			for ( ; row_pos < $current_page_rows.length; row_pos++ ) {
				var $row = $current_page_rows[ row_pos ];
				var coordinates = $row.data( 'paginator.coordinates' );
				if (coordinates) {
					coordinates.page = current_page;
					coordinates.row_position = row_pos;
					$row.data( 'paginator.coordinates', coordinates );
				}
			}
		};

		// remove element from page
		$page_rows.splice( row_position, 1 );

		// this is the element that will be moved form page x+1 to page x
		var $last_element_in_x;

		// and now start padding the whole table (may be painful with a big one)
		var x_coord = page_number;
		var page_count = this.page_count();
		for ( ; x_coord < page_count; ++x_coord) {
			var $next_page_rows = this.get_page_rows( x_coord + 1 );
			if ($next_page_rows && $next_page_rows.length > 0) {
				// grab the first element from the next page
				var $shifted = $next_page_rows.shift();
				$page_rows.push($shifted);
				if (!$last_element_in_x) {
					$last_element_in_x = $shifted; // see? this is the first one from the next page
				}
			}

			// update the whole page, we could update less element when i == x but...
			update_coordinates( x_coord, row_position );
			row_position = undefined;
		}

		$page_rows = this.get_page_rows( x_coord );

		if ($page_rows && $page_rows.length > 0) {
			update_coordinates(x_coord);
		} else { // last page was empty or having only one element
			--page_count;
			page_count = page_count < 1 ? 1 : page_count;
			this.__remove_last_page();
			this.$buttons_container.children('.pg-button-page').last().remove();
		}

		if ( $last_element_in_x && this.current_page === page_number ) {
			// this should always be the case assuming the element gets removed due to a user click
			$last_element_in_x.show();
		}

		if (this.current_page <= page_count) {
		// it's safe to update the navigator since the are more pages than the current one
			this.__update_page_navigator();
		} else {
		// the last (and current) page was removed, jump to the previous one
			this.__jump_to_page(page_count);
		}
	};

	Paginator.prototype.__disable_button = function($button) {
		$button.addClass('pg-button-disabled');
		$button.prop('disabled', true);
	};

	Paginator.prototype.__enable_button = function($button) {
		$button.removeClass('pg-button-disabled');
		$button.prop('disabled', false);
	};

	Paginator.prototype.__jump_to_page = function( page_number, on_complete, force )
	{
		var self = this;

		// Tiene que comenzar en la primer página si el page_number es < 1
		page_number = page_number <= 1 ? 1 : page_number;

		var on_jump_complete = function() {
			if ( on_complete ) {
				on_complete();
			}
			var $page_rows = self.get_page_rows( page_number );
			self.options.on_changed_page( page_number, $page_rows );
		};

		if ( !force && this.current_page === page_number ) {
			return;
		}

		if (!this.options.callback && this.page_count() < page_number) {
			return;
		}

		var do_change_page = function() {

			var $prev_button = self.$buttons_container.children('.pg-button-prev');
			var $next_button = self.$buttons_container.children('.pg-button-next');
			const totalPages = self.page_count();

			if (totalPages >= page_number) {
				// hide all current elements (first-level only)
				var $visible_elements = self.$data_container.children(':visible');
				if (self.options.ignored_elements !== "") {
					$visible_elements = $visible_elements.not(self.options.ignored_elements);
				}
				$visible_elements.hide();

				var $page_rows = self.get_page_rows( page_number );
				// and display the correct ones
				if ($page_rows) {
					 $page_rows.forEach(function( $row ) {
						let is_colapsed = $row.prop("colapsed");
						if (is_colapsed) {
						 	$row.hide();
						}
						else {
							$row.show();
						}
					});
				}

				// previous button
				if (self.options.show_prev) {
					if (totalPages === 1 || page_number === 1)
						self.__disable_button( $prev_button );
					else
						self.__enable_button( $prev_button );
				}

				// next button
				if (self.options.show_next) {
					if (totalPages === 1 || page_number === totalPages)
						self.__disable_button( $next_button );
					else
						self.__enable_button( $next_button );
				}

				// now update page index
				self.current_page = page_number;
				self.__update_page_navigator();
			} else {
				// no more data, Why are we here then? next button?
				if (self.options.show_next) {
					self.__disable_button( $next_button );
				}
			}

			on_jump_complete();
		};

		if ( this.options.callback ) {
			// the user should request and append more data or no more pages will be shown
			var page_count = this.page_count();
			var need_data = function() {
			// became a function 'cause condition is a little bit longer than expected
				if ( self.no_more_data || page_count > page_number ) {
					return false;
				}

				if ( page_count === page_number ) {
					let items_per_page = self.number_items_per_page( page_count );
					return items_per_page < self.options.rows_per_page;
				}

				return ( page_count < page_number );
			};

			this.last_requested_page = page_number;
			this.options.callback(page_number, need_data(), function() {
				if (self.last_requested_page === page_number) {
					do_change_page();
				} else {
					on_jump_complete();
				}
			});
		} else {
			do_change_page();
		}
	};

	Paginator.prototype.update_footer_total_text = function() {
		var self = this;

		// in case we're being called before being fully initialized
		if ( !self.options.text_format || !self.current_page ) {
			return;
		}
		let page_count = self.page_count();
		let row_count;
		if (this.m_total_rows) {
			// el total real de registros (server side) no proyectados en el paginado.
			row_count = this.get_total_rows();
		}
		else {
			row_count = self.options.rows_per_page * ( page_count - 1) + self.number_items_per_page( page_count );
			if ( self.options.row_count && self.options.row_count > row_count ) {
				row_count = self.options.row_count;
			}	
		}

		var previous_page = self.current_page - 1;
		var vars = {
			page_number   : self.current_page,
			page_row_start: Math.min( self.options.rows_per_page * ( previous_page ) + 1, row_count ),
			page_row_end  : self.options.rows_per_page * ( previous_page ) + self.number_items_per_page( self.current_page ),
			row_count     : row_count,
			page_count    : page_count
		};
		vars.many = self.no_more_data ? vars.row_count : self.options.translated_many;

		self.$pages_container.find('.pg-display-text span').text(
			self.options.text_format.replace(/\{([^{}]*)\}/g, function( a, b ) {
				var res = vars[ b ];
				if ( typeof( res ) !== "undefined" ) {
					return res;
				}
				return a;
			})
		);
	};

	Paginator.prototype.__update_page_navigator = function() {
		var self = this;

		var show_page_buttons = function($elements) {
			return $elements.removeClass('pg-button-hidden').show();
		};

		var hide_page_buttons = function($elements) {
			return $elements.addClass('pg-button-hidden').hide();
		};

		var insert_dotted_area = function($sibling, rel_position) {
			var pivot;
			var sibling_position = parseInt($sibling.attr('page-number'), 10);
			var $dots = $('<a href="javascript:void(0)" class="pg-button pg-button-dotted"> ... </a>');

			if (rel_position === "before") {
				pivot = 1;
				$sibling.before($dots);
			} else {
				pivot = self.page_count();
				$sibling.after($dots);
			}

			$dots.attr('page-number', Math.floor((sibling_position + pivot) / 2));
		};

		// it's assumed pages_at_left !== 'all' and pages_at_right !== 'all'
		var calculate_visible_pages = function() {
			var result = { at_left: 0, at_right: 0 };

			while (result.at_left < self.options.pages_at_left && (self.current_page - result.at_left) > 1) {
				++result.at_left;
			}

			while (result.at_right < self.options.pages_at_right && self.current_page + result.at_right < self.page_count()) {
				++result.at_right;
			}

			var page_limit = Math.min(self.options.show_at_least, self.page_count());
			if (result.at_left + result.at_right < page_limit) {
				var done_left = false, done_right = false;
				while (result.at_left + result.at_right < page_limit && !(done_left && done_right)) {
					if (done_left === false) {
						if (self.current_page - result.at_left > 1) {
							++result.at_left;
						} else {
							done_left = true;
						}
					}
					if (done_right === false) {
						if (self.current_page + result.at_right < self.page_count()) {
							++result.at_right;
						} else {
							done_right = true;
						}
					}
				}
			}

			return result;
		};

		self.$buttons_container.children('.pg-button-dotted').remove();
		self.$buttons_container.children('.pg-button-current').removeClass('pg-button-current');

		var $selected_page = self.$buttons_container.children('.pg-button-page')
			.filter('[page-number="' + self.current_page + '"]')
			.addClass('pg-button-current');
		show_page_buttons($selected_page); // just to be sure

		// calculate how many pages to show at left and right
		var visible_pages = calculate_visible_pages();

		//show_page_buttons($selected_page.prevAll('.pg-button-hidden'));
		if (self.options.pages_at_left !== 'all') {

			var $left_most = $selected_page;
			for ( var i = 0; i < visible_pages.at_left; ++i) {
				$left_most = show_page_buttons($left_most.prev());
			}

			// don't hide page 1 if it is the only one left
			if ( self.current_page - visible_pages.at_left > 2 ) {
				var $prev_buttons = $left_most.prevAll('.pg-button-page');
				if (self.options.show_first_page) {
					$prev_buttons = $prev_buttons.filter('[page-number!=1]');
				}

				hide_page_buttons($prev_buttons);
				insert_dotted_area($left_most, "before");

			} else if (!self.options.show_first_page && (self.current_page - visible_pages.at_left) === 2) {
				// page 1 may be already hidden, lets fix it!
				show_page_buttons(self.$buttons_container.children('.pg-button-page').first());
			}
		}

		//show_page_buttons($selected_page.nextAll('.pg-button-hidden'));
		if (self.options.pages_at_right !== 'all') {

			var $right_most = $selected_page;
			for (var j = 0; j < visible_pages.at_right; ++j) {
				$right_most = show_page_buttons($right_most.next());
			}

			if ( self.current_page + visible_pages.at_right < self.page_count() - 1 ) {
				var $next_buttons = $right_most.nextAll('.pg-button-page');
				if (self.options.show_last_page) {
					$next_buttons = $next_buttons.filter('[page-number!="' + self.page_count() + '"]');
				}

				hide_page_buttons($next_buttons);
				insert_dotted_area($right_most, "after");

			} else if (!self.options.show_last_page && (self.current_page + visible_pages.at_right) === (self.page_count() - 1)) {
				// last page may be already hidden, lets fix it!
				show_page_buttons(self.$buttons_container.children('.pg-button-page').last());
			}
		}

		self.update_footer_total_text();
	};

	/*!
	 * \return total rows in page
	 */
	Paginator.prototype.total_rows_in_page = function(page_number)
	{
		if ( typeof(page_number) !== "number" || !$.isArray( this.pages ) ) {
			return 0;
		}
		var $page_rows = this.get_page_rows( page_number );
		if ( !$.isArray( $page_rows ) ) {
			return 0;
		}
		return $page_rows.length;
	};

	/*!
	 * \retorna los rows de una pagina
	 */
	Paginator.prototype.get_page_rows = function( page_number ) {
		if ( typeof( page_number ) !== "number" || !$.isArray( this.pages ) ) {
			return;
		}
		page_number--;
		page_number = page_number < 0 ? 1 : page_number;
		return this.pages[ page_number ];
	};

	/*!
	 * \retorna el total de filas de páginas

	 */
	Paginator.prototype.get_total_page_rows = function( ) {
		
		let total = 0;
		this.pages.forEach((page) => total += page.length);
		return total;
	};

	/*!
	 * Call this function to initialize the paginator
	 * using the specified options.
	 *
	 * \remarks Don't forget to set \c data_container and \c pages_container
	 */
	$.Paginator = {
		init: function(options) {
			return new Paginator(options);
		}
	};

})(jQuery);
