local utf8 = require "icu.utf8"
local model_perm = require "shared.permissions.models.permissions"
local utils = require "shared.apps.utils"
local m_moduleLocation = select(2, ...)

module(..., package.seeall, inconcert.orbit.application.Decorate)

--[[
	If the campaign has the form campaign@vcc extracts the vcc part and uses the vcc of the logged user
--]]
local function SanitizeCampaignName (id)
	local vcc = inconcert.authUser.virtualContactCenter
	local campaignId, campaignVCC = string.match(id, "([^@]+)@([^@]+)")
	campaignId =  utf8.lower((campaignId or id).. "@" .. vcc)
	return campaignId
end

function Initialize (vcc, keyspace, cleanUpCache)
	LogInfo("Start initializing permissions")
	vcc = vcc or inconcert.authUser.virtualContactCenter
	model_perm.LoadAllGroupPermissions(vcc, keyspace, cleanUpCache)
	model_perm.LoadAllVccUsersPermissions(vcc, keyspace)
	LogInfo("End initializing permissions")
end

--[[
	Returns a function used in the url mappings wich need to check for a specific campaign permission
	Assumes the mapping is a JSON post request, with a campaignId member of the json, any other case returns 403/forbidden
--]]
function CheckCampaignPermission (self, appId, action, functionToCall)
	local validate = function(web, fullCampaignId, ...)
		if not fullCampaignId then
			LogWarning("[SHARED.PERMISSIONS] CheckCampaignPermission - Invalid Request campaign id: '%s'", fullCampaignId)
			web.status = "400"
			return json.encode{result = "Invalid request"}
		end
		local hasPermission = model_perm.IsPermissionEnabledForUser(appId, action, fullCampaignId)
		--LogDebug("[SHARED.PERMISSIONS] CheckCampaignPermission --- appId: '%s', action: '%s', fullCampaignId: '%s', hasPermission: '%s'.", appId, action, fullCampaignId, hasPermission)
		if hasPermission then
			return functionToCall(web, fullCampaignId, ...)
		end
		LogWarning("[SHARED.PERMISSIONS] CheckCampaignPermission - Status Forbidden.")
		web.status = "403"
		return json.encode{result = "forbidden"}
	end

	return function(web, ...)
		local simpleCampaignId
		local postedData = web.input["post_data"]
		local res, request = pcall(json.decode, postedData)
		local args = {...}

		--LogDebug("[SHARED.PERMISSIONS] CheckCampaignPermission - Start Validation: postedData: '%s', args: %s", postedData, args)

		if request and request.campaignId then
		-- por aca vienen los POST
			simpleCampaignId = request.campaignId
		else
		-- por aca vienen los GET
			simpleCampaignId = SanitizeCampaignName(args[1]) -- en la posicion 1 del array tiene que venir el id de la campana
			table.remove(args, 1)
		end

		local fullCampaignId
		local vcc = inconcert.authUser.virtualContactCenter
		if vcc then
			fullCampaignId = utils.AppendVcc(utils.GetId(simpleCampaignId), vcc)
		else
			fullCampaignId = simpleCampaignId
		end
		fullCampaignId = fullCampaignId:lower()

		return validate(web, fullCampaignId, args)
	end
end

--[[
	Check if the action is allowed for this user
--]]
function CheckPermission (self, appId, action, functionToCall)
	return function(web, ...)
		if type(action) == "string" then
			--LogDebug("[SHARED.PERMISSIONS:CheckPermission] 1---> appId: '%s', action: '%s'.", appId, action)
			-- valida contra una accion determinada que llega como string
			if model_perm.IsPermissionEnabledForUser(appId, action) then
				-- si tiene permisos
				return functionToCall(web, ...)
			end
		elseif type(action) == "table" then
			-- valida contra una lista de acciones que llegan en un array (todas tienen que ser true)
			--LogDebug("[SHARED.PERMISSIONS:CheckPermission] 2---> appId: '%s', action: '%s'.", appId, json.encode(action))
			local actions = action
			local hasPermission
			for i, actionId in ipairs(actions) do
				hasPermission = model_perm.IsPermissionEnabledForUser(appId, actionId)
				--LogDebug("[SHARED.PERMISSIONS:CheckPermission]  2.%d---> appId: '%s', action: '%s' hasPermission: %s.", i, appId, actionId, hasPermission)
				if not hasPermission then
					break
				end
			end
			if hasPermission then
				-- si tiene permisos
				return functionToCall(web, ...)
			end
		end
		-- NO tiene permisos !!
		web.status = "403"
		return json.encode{result = "forbidden"}
	end
end

--[[
	Valida si el Usuario tiene permisos sobre una App.
	Funcion utilizada para el mapeo de funciones en los controllers
--]]
function CheckAppPermission (self, appId, fullCampaignId, functionToCall)
	assert(appId)
	assert(functionToCall)
	-- fullCampaignId es opcional

	return function(web, ...)
		local hasPermissionToApp = true
		if not inconcert.authUser then
			return functionToCall(web, ...)
		end

		local vcc = inconcert.authUser.virtualContactCenter
		local simpleUserId = inconcert.authUser.userName
		local fullUserId = utils.AppendVcc(simpleUserId, vcc)
		local roleId = inconcert.authUser.userGroup

		-- aca realizo la validación si el usuario tiene permisos para la app
		hasPermissionToApp = HasPermissionToApplication(
			appId,
			fullUserId,
			roleId,
			fullCampaignId--,
			--keyspace
		)

		if hasPermissionToApp then
			return functionToCall(web, ...)
		end
		web.status = "403"
		return json.encode{result = "forbidden"}
	end
end

--[[
	application,
	action,
	campaignId  -- opcional
--]]
function IsPermissionEnabledForLoggedUser (application, action, campaignId, keyspace, applicationsActions)
	return model_perm.IsPermissionEnabledForUser(application, action, campaignId, keyspace, applicationsActions)
end

function IsPermissionEnabledForSpecificUser (application, action, fullUserId, userGroup, userVccPermissions, groupPermissions, userCampaignsPermissions, keyspace, applicationsActions, skipCanValidationPermission)
	return model_perm.IsPermissionEnabledForSpecificUser(application, action, fullUserId, userGroup, userVccPermissions, groupPermissions, userCampaignsPermissions, keyspace, applicationsActions, skipCanValidationPermission)
end

function GetApplicationsActions (keyspace, vcc)
	return model_perm.GetApplicationsActions(keyspace, vcc)
end

--[[
campaign_id, keyspace (opcionales)
--]]
function HasPermissionToApplication (app_id, user_id, user_group, campaign_id, keyspace, permissions, skipCanValidationPermission)
	assert(type(app_id) == "string" or type(app_id) == "table", "The 'app_id' parameter can not be string or table")
--LogDebug("0.5 permissions.user_campaigns %s", permissions.user_campaigns)
	permissions = permissions or {}
	local HasPermission = function( app )
		return model_perm.HasPermissionToApplication(
			app,
			user_id,
			user_group,
			permissions.user_campaign_permissions,
			permissions.group_permissions,
			campaign_id,
			keyspace,
			permissions.user_vcc_permissions,
			permissions.campaigns_permissions,
			permissions.user_campaigns,
			permissions.applicationsActions,
			permissions.customPermissions,
			skipCanValidationPermission
		)
	end

	if type(app_id) == "string" then
		return HasPermission( app_id )
	end

	local apps = app_id
	-- valida que al menos tenga permisos en una de las aplicaciones recibidas 
	local has_permission = false
	for _, app in ipairs(apps) do
		has_permission = HasPermission( app )
		if has_permission then
			break
		end
	end
	return has_permission
end

--[[

--]]
function RoleHasPermissionToApplication (app_id, role_id, role_permissions, keyspace)
	assert(type(app_id) == "string" or type(app_id) == "table", "The 'app_id' parameter can not be string or table")

	permissions = permissions or {}
	local HasPermission = function( app_id )
		return model_perm.RoleHasPermissionToApplication(
			app_id,
			role_id,
			role_permissions,
			keyspace
		)
	end

	if type(app_id) == "string" then
		return HasPermission( app_id )
	end

	local apps = app_id
	-- valida que al menos tenga permisos en una de las aplicaciones recibidas
	local has_permission = false
	for _, app in ipairs(apps) do
		has_permission = HasPermission( app )
		if has_permission then
			break
		end
	end
	return has_permission
end

--[[

--]]
function RoleHasPermissionToApplicationAction (appId, actionId, roleId, rolePermissions, keyspace)
	assert(type(appId)    == "string", "The 'appId' parameter can not be string"   )
	assert(type(actionId) == "string", "The 'actionId' parameter can not be string")
	assert(type(roleId)   == "string", "The 'roleId' parameter can not be string"  )

	local HasPermission = function()
		return model_perm.RoleHasPermissionToApplicationAction(
			appId,
			actionId,
			roleId,
			rolePermissions,
			keyspace
		)
	end

	return HasPermission()
end

