/* jshint browser:true, quotmark:false, smarttabs:true, -W014, esversion: 11 */
/* global $, _, window, inConcert, InconcertRecorder, inConcertLanguages, GoldenLayout */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

(function()
{
	"use strict";

	// This function is attached as a handler for the Copy event in the body
	// overrides default copy to clipboard with a text/plain version
	// by now it s only used inside webchatClientWindow.html
	window.customCopy = function(event) {
		var selection = window.getSelection();
		event.clipboardData.setData("text/plain", selection);
		event.preventDefault();
	};

	const _T = inConcert.i18n.getString;
	const format = inConcert.Utils.format;
	// Estas dimensiones deben de coincidir con lo que se ha definido en "base_panel.js" funcion "openInteractionClient"
	const DEFAULT_WINDOW_HEIGHT_PX = 500;
	const DEFAULT_WINDOW_WIDTH_PX = 950;
	/**
	Class ClientWindow.
	@param options
	*/
	function ClientWindow (options)
	{
		const self = this;
		options = options || {};

		this.m_popupLabels = null;
		this.m_commentPopup = null;
		this.m_selectedContact = false;
		this.sidebar = null;
		this.m_popupTickets = null;
		this.m_finalization = null;
		this.m_languages = null;
		this.m_cci_application = null;          /* instancia de la app de CC Integration*/
		this.m_helpdesk_integration_app = null; /* instancia de la app de HelpDesk Integration*/
		this.m_user_default_campaign_id = null;
		/*
			el estado actual de la ventana del cliente 
				(connected, disconnected, pending, invalid, monitoring, wrapping_up, ..)
				el mismo se actualiza en la funcion "change_client_state"
				esta funcion es invocda locamente en el mismo archivo o escuchando el evento onApplicationStatusChanged que ha subscrito en la escucha de eventos.
		*/
		this.m_client_state = null;
		/*
			Variables utiles para los clientes:
				twitter, instagram, sms, facebook, youtube, whatsapp.
			Aunque no aplica para todos los clientes, (ejemplo: webchat, mail, etc) igualmente se deja en la clase base,
				ya que sino tiene que estar repetida con el mismo codigo en los clientes que si lo necesitan.
		*/
		this.MARK_ALL_READ_BTN_TEXT   = "<%=_T('Mark All As Read')%>";
		this.MARK_ALL_UNREAD_BTN_TEXT = "<%=_T('Mark All As Unread')%>";

		ClientWindow.uber.constructor.call( this, {
				move: ["topDragBar", "bottomDragBar"]
			},
			options
		);

		this.events_count = 0;
		this.is_loading = true;
		this.from_app = options.from_app || window.from_app || "interaction";
		// TODO: #9778 - CANY - Docked.
		this.panel_name = options.panel_name || window.panel_name || null;

		// Botones de la TOOLBAR
		this.m_interactionToolbarButtons = {
			Finish: { 
				id    : "finishInteraction", 
				method: function() { self.finishInteraction(); }
			},
			Reopen: { 
				id    : "reopenInteraction", 
				method: function() { self.reopenInteraction(); }
			},
			Take: { 
				id    : "takeInteraction", 
				method: function() { self.takeInteraction(); }
			},
			Hangup: { 
				id    : "hangupInteraction", 
				method: function() { self.hangupInteraction(); }
			},
			Queue: { 
				id    : "queueInteraction", 
				method: function() { self.queueInteraction(); }
			},
			Transfer: { 
				id    : "transferInteraction", 
				method: function($button) { self.transferInteraction($button); }
			},
			Cancel: { 
				id    : "cancelInteraction", 
				method: function() { self.cancelInteraction(); }
			},
			Conference: { 
				id    : "conferenceInteraction", 
				method: function($button) { self.conferenceInteraction($button); }
			},
			Schedule: { 
				id    : "scheduleInteraction", 
				method: function($button) { self.scheduleInteraction($button); }
			}
		};

		$.extend(true, this.m_interactionToolbarButtons, options.customInteractionToolbarButtons || {});

		// El menu contextual y sus acciones lo genera y ejecuta la clase "interaction_gui".
		//  asi se reutiliza el codigo ya existente para las bandejas.
		//
		const actions = this.from_app === "interaction" ? inConcert.client_window_attention_actions : inConcert.client_window_supervisor_actions;
		this.m_interaction_gui = new inConcert.interaction_gui(actions,
			{
				from_app: this.from_app,
				owner   : this,
				callbacks:{
					transfer:{
						on_before_transfer_execute : function(transfer_to) {
							self.on_before_transfer_execute(transfer_to);
						},
					}
				},
				// Si se pasa este atributo, solo se visualizaran estas acciones en el menu contextual
				custom_allowed_actions: this.getMenuContextualActions()
			}
		);

		this.is_readonly = false;
		if ( options.readonly === true ) {
			this.is_readonly = true;
		}

		this.is_archive = false;
		if ( options.archive ) {
			this.is_archive = true;
		}
		this.is_queued = false;
		if ( options.queued ) {
			this.is_queued = true;
		}
		this.is_waiting = false;
		if ( options.waiting ) {
			this.is_waiting = true;
		}
		this.change_client_state("connected");

		// TO-DO: This is even possible? If so, should be a comment explaining the case.
		if (!this.m_interaction) {
			return;
		}

		this.m_interaction.opened = true;
		this.m_interaction.setView(this);

		if (this.m_interaction.hasPendingLabel()) {
			this.change_client_state("pending");
		}

		this.makeMarkupOwnerInformation();
		this.is_monitor = false;
		if (options.monitor) {
			this.is_monitor = true;
			this.loadOwnerInformation();
		}

		this.initialize(options);
		this.setVerticalResizingLimits(DEFAULT_WINDOW_HEIGHT_PX);

		// TO-DO: Whats the case for an "invalid" interaction? If so, should be a comment explaining the case.
		if (!this.m_interaction.isValid()) {
			this.invalidInteraction();
			return false;
		}
		this.eventsSuscribe();
		this.events_count = this.m_interaction.getEventCount();
		this.updateContactDisplay();
		return true;
	}

	inConcert.Utils.inherit( ClientWindow, inConcert.Dialog );

	//Define functions and objects in the namespace in case they don t already exist
	if (typeof inConcert.ClientWindow !== 'function') {
		inConcert.ClientWindow = ClientWindow;
	}

	ClientWindow.prototype.clientUpdate = $.noop;
	ClientWindow.prototype.updateSideBar = $.noop;

	ClientWindow.prototype.canMinimizeWhenOutside = function(target)
	{
		//#18807 No se minimiza cuando es click outside en la barra principal del top
		if (this.__isEventClickInTopMainHeader(target)) {
			return false;
		}

		// TODO: #9778 - CANY - Docked.
		//
		return true;
		//return false;
	};

	ClientWindow.prototype.__isEventClickInTopMainHeader = function(target)
	{
		//Click en la barra principal del top.
		if (!target) {
			return false;
		}


		return inConcert.app.getMainWindow().belongsEventTargetToContainer("inboxMainHeader", target) ||
		inConcert.app.getMainWindow().belongsEventTargetToContainer("inboxMainHeaderMenus", target);
	};
	
	ClientWindow.prototype.getMenuContextualActions = function()
	{
		const actions = {
		// acciones solo para el cliente
			// attachfile   : true, // (EJEMPLO)
		// acciones sobre la interaccion
			comment      : true,
			label        : true,
			emit_ticket  : true,
			search_ticket: true,
			queue        : true,
			transfer     : true,
			schedule	 : true
		// estas NO tienen que visualizarse en el menu, ya que las mismas se encuentran en la toolbar
		};
		return actions;
	};

	ClientWindow.prototype.getWindowDefaultTitle = function()
	{
		// esta funcion se debe de implementar en la clase hija
		return "<%=_T('Conversation Window')%>";
	};

	ClientWindow.prototype.on_before_transfer_execute = $.noop;

	ClientWindow.prototype.setTitle = function( window_title )
	{
		if ( !window_title && this.m_interaction && !this.m_interaction.isBeingComposed() ) {
			window_title = this.m_interaction.getClientName();
			if ( !window_title || window_title.length === 0 ) {
				var first_outgoing_event = this.m_interaction.getFirstOutgoingEvent();
				if ( first_outgoing_event ) {
					window_title = this.m_interaction.get_agent_name( first_outgoing_event );
				}
			}
		}
		if ( !window_title || window_title.length === 0 ) {
			window_title = this.getWindowDefaultTitle();
		}
		ClientWindow.uber.setTitle.call( this, window_title );  //invokes the base class method
	};

	ClientWindow.prototype.updateWindowClasses = function( classes )
	{
		classes = classes || "";
		if (this.m_interaction) {
			classes += " interaction" + (this.m_interaction.isOutgoing() ? "Outbund" : "Inbound");
			classes += " interactionType" + (this.m_interaction.type === "CALL" && this.m_interaction.isWeb()? "WEBCALL" : this.m_interaction.type);
		}
		classes += " scope" + this.from_app.toUpperCase();
		this.addClass( classes );
	};

	ClientWindow.prototype.onMaximize = function(/*event*/)
	{
		ClientWindow.uber.onMaximize.call(this);
		if (this.sidebar) {
			this.sidebar.onMaximizeWindow();
		}
		this.adjust_panel_positions();
		this.updateWindowClasses();
	};

	ClientWindow.prototype.onMinimize = function(/*event*/)
	{
		ClientWindow.uber.onMinimize.call(this);
		if (this.sidebar) {
			this.sidebar.onMinimizeWindow();
		}
		this.updateWindowClasses();
	};

	ClientWindow.prototype.onRestore = function(/*event*/)
	{
		ClientWindow.uber.onRestore.call(this);
		if (this.sidebar) {
			var size;
			if (this.sidebar.isVisible()) {
				size = this.getWindowSizeOnSidebarShown();
			}
			else {
				size = this.getWindowSizeOnSidebarHidden();
			}
			this.setHorizontalResizingLimits(size.width);
			this.setWidth(size.width);
			this.sidebar.onRestoreWindow();
		}
		this.adjust_panel_positions();
		this.updateWindowClasses();
	};

	ClientWindow.prototype.dispose = function() {
		// before call onClose, unsubscribe from pubsub notifications
		this.eventsUnsuscribe();

		if (this.sidebar) {
			this.sidebar.dispose();
		}

		if(this.m_interaction) {
			this.m_interaction.deflate();
		}

		if (this.m_btnToggleContainerTopTime) {
			clearInterval(this.m_btnToggleContainerTopTime);
			delete this.m_btnToggleContainerTopTime;
		}
	};

	ClientWindow.prototype.eventsSuscribe = function() {
		var self = this;
		// application connected
		var pub_conn_id = "/application/connected";
		this.pubsub_connected_handle = inConcert.app.messagebus.subscribe(
			pub_conn_id,
			function(ok) {
				self.onApplicationStatusChanged(ok);
			});

		if ( this.sidebar ) {
			var isArchived = self.m_interaction.isFinished();
			if (!isArchived) {
				var campaign = inConcert.app.getCampaign(self.m_interaction.campaign);
				if ( campaign && campaign.hasCannedMessages() ) {
					// campaign - canned changed
					var pub_cann_id =  "/canned_messages/" + this.m_interaction.campaign + "/updated";
					this.pubsub_canned_messages_handle = inConcert.app.messagebus.subscribe(
						pub_cann_id,
						function(differences) {
							campaign.getCannedMessages(function(canned_messages) {
								self.sidebar.update_canned_messages(canned_messages, differences);
							});
					});
				}
			}
		}
	};

	ClientWindow.prototype.eventsUnsuscribe = function() {
		// application connected
		if (this.pubsub_connected_handle) {
			inConcert.app.messagebus.unsubscribe(this.pubsub_connected_handle);
		}
		if (this.sidebar) {
			// campaign - canned changed
			if (this.pubsub_canned_messages_handle) {
				inConcert.app.messagebus.unsubscribe(this.pubsub_canned_messages_handle);
			}
		}
	};

	ClientWindow.prototype.onWindowClosed = function()
	{
		ClientWindow.uber.onWindowClosed.call( this );
		if (this.m_interaction) {
			this.m_interaction.opened = false;
			this.m_interaction.setView(null);
		}
		var res = this.m_interaction;
		this.dispose();
		delete this.m_interaction;
		return res;
	};

	ClientWindow.prototype.onApplicationStatusChanged = function(ok)
	{
		const text_agent_disconnected = "<%=_T('inConcert Agent is disconnected, changing to read-only mode')%>";
		if (!ok) {
			const previous_banner_text = $('.clientParticularMessages .clientMessagesTextAux').text();
			if (previous_banner_text !== text_agent_disconnected){
				console.warn("ClientWindow.onApplicationStatusChanged - " + text_agent_disconnected);
			}
			this.change_client_state("disconnected");
			$(".clientDisconnectedOverlay").show();
			this.displayStatusMessage(text_agent_disconnected, "Alert", true);
			return;
		}
		this.change_client_state("connected");
		$(".clientDisconnectedOverlay").hide();
		// if we show a disconnected banner previously, hide it.
		const previous_banner_text = $('.clientParticularMessages .clientMessagesTextAux').text();
		if (previous_banner_text === text_agent_disconnected){
			this.hideStatusMessage();
		}
	};

	ClientWindow.prototype.invalidInteraction = function() {
		this.change_client_state("invalid");
		this.displayStatusMessage("<%=_T('Current conversation is invalid')%>", "Info", false, 3000);
	};

	ClientWindow.prototype.hasPermissionToAction = function( action, app_id )
	{
		action = ( action || "" ).toUpperCase();
		app_id = app_id || "interaction";

		var canCreateOutgoing = function() {
			var response = inConcert.app.getInteractionManager().hasPermissionToCreateOutgoingInteractions();
			return response.success;
		};

		var actions = {
			CREATE_OUTGOING        : { interaction: canCreateOutgoing                     , supervisor: false                     },

			MARK_AS_SPAM           : { interaction: "int_mark_as_spam"                    , supervisor: false                     },
			MARK_AS_NOT_SPAM       : { interaction: "int_not_spam"                        , supervisor: false                     },

			SPELL_CHECKER_ADD_WORD : { interaction: "int_spell_checker_add_to_dictionary" , supervisor: false                     },
			CHAT_FILE_TRANSFER     : { interaction: "int_chat_file_transfer"              , supervisor: false                     },
			MAIL_PRINT             : { interaction: "int_mail_print"                      , supervisor: "sup_mail_print"          },
			MOVE_MAIL_TO_FOLDER    : { interaction: "int_mail_move_to_folder"             , supervisor: "sup_mail_move_to_folder" },

			TICKET_EMIT            : { interaction: "int_emit_ticket"                     , supervisor: "sup_emit_ticket"         },
			TICKET_SEARCH          : { interaction: "int_search_ticket"                   , supervisor: "sup_search_ticket"       }
		};

		var action_def = actions[action];
		if (!action_def) {
			return false;
		}
		var action_id = action_def[app_id];
		if ( !action_id ) {
			return false;
		}

		if (typeof action_id === "function") {
			return canCreateOutgoing();
		}
		// string
		var campaign_id = this.m_interaction ? this.m_interaction.campaign : undefined;
		return inConcert.permissions.to_action( app_id, action_id, undefined /* category */, campaign_id, true /* no log */ );
	};

	ClientWindow.prototype.updateBodyClasses = function()
	{
		if (!this.m_interaction) {
			return;
		}
		const $body = $("body");
		$body.removeClass("newConversation");

		if (this.m_interaction.isBeingComposed()) {
			$body.addClass("newConversation");
		}

		$body.removeClass("withTemplatesMessages");
		if (this.canSendTemplateMessages()) {
			$body.addClass("withTemplatesMessages");
		}

		$body.removeClass("withRecordingAudio");
		if (this.canRecordAudio()) {
			$body.addClass("withRecordingAudio");
		}
	};

	/*
		Si la conversacion se encuentra en conflictos por los contactos, abrimos el dialogo para solucionar el mismo.
	*/
	ClientWindow.prototype.showDialogContactConflict = function() {
		if (!this.m_interaction) {
			return;
		}
		if (!this.m_interaction.hasContactConflict()) {
			return;
		}
		if (this.m_interaction.isAssigned() || this.m_interaction.isLocked()) {
			let options ={
				from_app   : this.from_app,
				interaction: this.m_interaction
			};
			if (this.m_interaction) {
				if (this.m_interaction.contact && this.m_interaction.contact.length > 0) {
					options.contact_id = this.m_interaction.contact;
				}
				else {
					options.addresses = this.m_interaction.interactionData.contactAddresses;
				}
			}
			const contactWindow = new inConcert.ContactWindow(options);
			contactWindow.show();
		}
	};

	ClientWindow.prototype.initialize = function()
	{
		if (!this.m_interaction) {
			return;
		}

		this.setCachedHTMLInteractionToolbarButtons();
		this.setTitle(); // window title
		this.load_languages();
		this.setupCampaignSelector();
		this.toggleActionsIfPending();
		this.bindGenericToolbarActionsButtons();
		this.intializeComments();
		this.intializeLabels();
		this.initializeTicketInfo();
		this.initializeFinalization();
		this.updateToolbarButtons();
		this.setupTransferInfo();
		this.setTooltips();
		this.updateBodyClasses();
		this.update_header_contact_info();
		this.hideStatusMessage();

		this.m_currentEvent = -1;

		// interactions that are not event based do not get the special treatment
		// also, if they're being compsed, disable
		if (!this.m_interaction.supportsEvents() || this.m_interaction.isBeingComposed()) {
			this.disableRelatedSections();
		}
		else {
			/*
				En el cliente de Mail tengo que esperar, a que el layout se haya inicializado, para poder visualizar el body del mail.
				A esto lo realiza en la funcion/callback "onSidebarIntialisedLayout" de la clase de mail.
				Ref #8533
			*/
			if (this.m_interaction.type !== "MAIL") {
				this.navigateToBestEvent(undefined, true /* is_initialized*/);
			}
		}
		this.validateCampaignAndAccountOperational();
		this.showDialogContactConflict();
		this.firstLoadDraftMessage();
		this.m_initialized = true;
	};

	ClientWindow.prototype.__resize_container_top_info = function(forceFixed, forceAuto) {
		const $container = $("#headerGeneralInfo");
			let newClass = "";
			if ((forceAuto || $container.attr("HeightFixed") === "true") && !forceFixed) { //Esta en fixed, cambio a auto
				newClass = "heightAuto";
				$container.attr("HeightFixed", "false");
			}
			else{ //Esta en auto, cambio a fixed
				newClass = "heightFixed";
				$container.attr("HeightFixed", "true");
			}

			$container.removeClass("heightFixed heightAuto").addClass(newClass);
	};

	ClientWindow.prototype.__toggle_btn_container_top_info = function()
	{
		/*
			Calculo si la info entra en headerGeneralInfo:
			-Para que la info quepa, el ancho de hijos (interactionContextInfo) debe ser meno a:
				* Padre (headerGeneralInfo) - ancho de botones ( butShowRelated + butCloseRelated ) - WIDTH_MAGIC_NUMBER (otros) 
		*/
		const WIDTH_MAGIC_NUMBER = 0; //Cambiar a necesidad.
		const $btnToggle = $("#headerGeneralInfoBtnToggle");
		const $parent = $("#headerGeneralInfo");

		const parentHWidht = $parent.width();
		let need_to_expand = false;

		//Suma de ancho del interactionContextInfo (Se hace en ciclo por si a futuro agregan mas)
		let width_info_children = 0;
		const $contextInfo = $parent.find("[interactionContextInfo]");
		$contextInfo.children().each(function(_, node) {
			// Suma el ancho de cada hijo al ancho total
			width_info_children += $(node).width(); 
		  });

		//Suma de ancho de los botones de accion
		let width_action_btns = 0;
		const $acction_btns = $parent.find("[headerGeneralInfoBtnAction]");
		$.each($acction_btns, function (_, btn) {
			const $btn = $(btn);
			if ($btn.is(":visible")) {
				width_action_btns += $btn.innerWidth(); // Utilizamos innerWidth para incluir el relleno
			}
		});
		
		//Si necesita expandir muestro boton de expandir.
		need_to_expand = width_info_children > (parentHWidht - width_action_btns - WIDTH_MAGIC_NUMBER);
		if (need_to_expand) {
			$btnToggle.show();
		}
		else
		{
			$btnToggle.hide();
			//Si esta expandido cambio a fixed
			if ($parent.attr("HeightFixed") === "false") {
				this.__resize_container_top_info(true);
			}
		}
	};

	ClientWindow.prototype.__events_client_container_top = function()
	{
		const self = this;
		$("#headerGeneralInfoBtnToggle")
		.off("click")
		.on("click", function() {
			self.__resize_container_top_info();
		});

		this.m_btnToggleContainerTopTime = setInterval(() => {
			//Cada un seg, valido si mostrar o no el boton para dejar en auto el top
			self.__toggle_btn_container_top_info();
		}, 1000);
	};

	ClientWindow.prototype.update_header_contact_info = function(info)
	{
		info = info || {};
		if (!this.m_interaction) {
			return;
		}

		if (this.m_interaction) {
			$("#contactTopID").text(this.m_interaction.getDisplayName());
		}

		if (!info.contactID) {
			if (this.m_interaction.interactionData) {
				info.contactID = this.m_interaction.interactionData.contactName;
			}
			if (!info.contactID) {
				if (this.m_interaction.getEventCount() === 0) {
					info.contactID = inConcert.app.getLoggedUserName();
				}
			}
		}

		const first_interaction_event = (this.m_interaction.events|| [])[0];
		if (!info.contactFromAccount) {
			if (this.m_interaction.getEventCount() === 0) {
				info.contactFromAccount = this.m_interaction.getAccountDisplayName();
			}
			if (!info.contactFromAccount) {
				if (first_interaction_event) {
					info.contactFromAccount = this.m_interaction.getEventUserlIdByChannel(0);
				}
			}
		}

		if (!info.avatarURI) {
			if (this.m_interaction.getEventCount() > 0) {
				if (first_interaction_event) {
					info.avatarURI = this.m_interaction.getEventUserAvatarURI(first_interaction_event);
				}
			}
			if (!info.avatarURI) {
				info.avatarURI = this.getGenericAvatar();
			}
		}

		$("#contactID").text(info.contactID || "");
		$("#contactFromAccount").text(info.contactFromAccount || "");
		$("#contactFromAvatar").attr("data", info.avatarURI);
	};

	ClientWindow.prototype.loading = function(is_loading, timeout_ms, callback = $.noop)
	{
		const self = this;
		let $screenWaiting;

		const hide_loading = function(timeout) {
			// Pequeno timeout para cubrirnos al cargar la pantalla, ya que puede ser muy rapida la carga de la misma.
			window.setTimeout(function() {
				var interID = "";
				if (self.m_interaction) {
					interID = self.m_interaction.id;
				}
				console.debug("[ClientWindow.loading] =======================> - HIDE WAITING - interaction " + interID);
				$screenWaiting.hide();
				callback(true);
				self.is_loading = false;
			}, timeout);
		};

		if(!this.m_interaction) {
			console.error("[ClientWindow.loading] -----> NO EXISTE interaccion");
			callback(false);
			return;
		}

		console.debug("[ClientWindow.loading] -----> ENTRA - interaction " + this.m_interaction.id + ", is_loading " + is_loading);
		$screenWaiting = $("#screenWaiting");
		if ($screenWaiting.length === 0) {
			callback(false);
			return;
		}
		if (is_loading) {
			console.debug("[ClientWindow.loading] -----> + SHOW WAITING - interaction " + this.m_interaction.id);
			$screenWaiting.show();
			if (timeout_ms) {
				hide_loading(timeout_ms);
			}
			else {
				callback(true);
			}
		}
		else {
			hide_loading(undefined);
		}

		this.is_loading = is_loading;
	};

	ClientWindow.prototype.show_waiting = function(timeout_ms, callback) {
		this.loading(true, timeout_ms, callback);
	};

	ClientWindow.prototype.hide_waiting = function(callback) {
		this.loading(false, undefined, callback);
	};

	/*
		Mark as read/unread, uno o varios eventos.
	*/
	ClientWindow.prototype.mark_as_read = function( mark_as_read, events, callback ) {
		var self = this;
		callback = callback || $.noop;

		var $mark_all = $( "#mark_all_read_unread" );
		$mark_all.prop( "disabled", true ); // desactivo el boton que marca a todos como read/urnead

		var $all_links_mark_as = $( "#thread_container" ).find( "[mark_as_read_unread]" );
		$all_links_mark_as.prop( "disabled", true ); // desactivo todos los links que marca el comment como read/urnead

		var message;
		if ( mark_as_read ) {
			message = "<%=_T('Please wait, marking as read.')%>";
		}
		else {
			message = "<%=_T('Please wait, marking as unread.')%>";
		}
		this.displayStatusMessage( message, "Process", false );

		this.m_interaction.setRead(
			mark_as_read,
			events,
			function( success ) {
				self.set_mark_all_unread_read_button_caption();
				callback(success);
				$mark_all.prop( "disabled", false ); // activo el boton que marca a todos como read/urnead
				$all_links_mark_as.prop( "disabled", false ); // activo todos los links que marca el comment como read/urnead
				self.hideStatusMessage();
			}
		);
	};

	ClientWindow.prototype.markAllReadUnread = function(callback = $.noop) {
		const self = this;
		const mark_as_read = $("#mark_all_read_unread").val() === self. MARK_ALL_READ_BTN_TEXT;
		let events_to_mark = [];

		if ( !mark_as_read ) {
			$.each( self.m_interaction.getEventList() || [], function( _, event ) {
				if ( event.creator === "contact" ) {
					let original_index = event.__index;
					if ( typeof original_index === undefined ) {
						throw "ClientWindow.prototype.markAllReadUnread - event.__index is required";
					}
					// en el array de eventos a marcar como "unread" tiene que ir el indice posta del evento, segun se encuentra en el server.
					events_to_mark.push( original_index );
				}
			});
		}

		self.mark_as_read(mark_as_read, events_to_mark, function(success) {
			if ( !success ) {
				callback(false);
				return;
			}
			callback(true);

			const all_events = self.m_interaction.getEventList();
			$.each( all_events || [], function( _, event ) {
				if ( event.creator !== "contact" || !event.message || !event.message.id ) {
					return true; // continue
				}
				let id = event.message.id;
				let $message = $("#thread_container").find("[data-id='" + id + "']");
				if ( !$message || $message.length === 0 ) {
					return true; // continue
				}
				self.on_readed_event( true /* success */, $message, mark_as_read );
			});
		});
	};

	ClientWindow.prototype.is_ready = function()
	{
		return this.is_loading;
	};
	
	ClientWindow.prototype.setEventsHandlers = function()
	{
		const self = this;

		if (!this.m_interaction) {
			return;
		}

		window.id = this.m_interaction.id;

		// TODO: #9778 - CANY - Docked.
		if (inConcert.app.is_enable_docked_windows_interactions()) {
			// DOCKED WINDOWS
			const $buttonGoBack = $("#buttonGoBack");
			$buttonGoBack
				.off("click")
				.on("click", function() {
					self.close();
				}
			);
		}
		// NORMAL WINDOWS

		// related events ------------
		$("#butShowRelated")
			.off("click")
			.on("click", function() {
				inConcert.logger.guiAction({action_name : "open_related", window_id: window.id});
				self.toggleRelatedEvents();
				$("body").removeClass("withRelatedEvents").addClass("withRelatedEvents");
			}
		);

		$("#butCloseRelated")
			.off("click")
			.on("click", function() {
				inConcert.logger.guiAction({action_name : "close_related", window_id: window.id});
				self.toggleRelatedEvents();
				$("body").removeClass("withRelatedEvents");
			});

		var closeRelated = function ($element) {
			let selected_event = $element.data("index");
			selected_event = parseInt( selected_event, 10 );
			self.closeRelated(selected_event);
		};
		const $related = $("#relatedInteractionsList");

		// doble click en una linea de los eventos no leidos.
		$related
			.off("dblclick", "[unread_event]")
			.on("dblclick", "[unread_event]", function() {
				closeRelated($(this));
			});

		// doble click en una linea de los eventos no leidos.
		$related
			.off("click", "[unread_event_action]")
			.on("click", "[unread_event_action]", function() {
				closeRelated($(this));
			});

		/* webchat y whatsapp*/
		self.$attachmentPicker()
			.off("click")
			.on("click", function() {
				const $parent = $(this);
				self.showFilesMenu($parent);
			}
		);

		//client container top
		this.__events_client_container_top();
	};

	ClientWindow.prototype.setTooltips = function()
	{
		// general tooltip -- TODO: ver de realizar un selector mas especifico para este caso
		$("[tooltip]").inconcert_tooltip();
	};

	ClientWindow.prototype.clientUnusable = function()
	{
		this.is_client_unusuable = true;
		// toolbar de acciones
		$(".clientContainerActions input").hide();                   // oculto las acciones existentes en cada cliente. (siempre son buttons)
		$(".clientContainerActions .actionButtonsSeparator").hide(); // oculto los separadores
		this.enableRelatedSections();                                // si tiene mas de un evento tienen que quedar visible la navegacion
		this.displayConflictCampaignMessage( true /*infinite*/ );
	};

	/*
		Esta funcion recupera del HTML correspondiente al cliente, el DOM correspondiente a los botones de acciones propias de la interaccion (finish, take, etc)
		y lo cachea en una variable de la clase, para que luego sea renderizado nuevamente en la funcion que actualiza los botones de la toolbar.
	*/
	ClientWindow.prototype.setCachedHTMLInteractionToolbarButtons = function()
	{
		const self = this;

		if (typeof(this.m_cachedHTMLInteractionToolbarButtons) === "string") {
		// Si ya fue cacheado, no tiene que volver a cachear.
			return;
		}
		
		const $interactionToolbarButtons = this.$getActionsButtonsContainer();
		if ($interactionToolbarButtons.length > 0) {
			// Al DOM correspondiente a cada boton le agrega un atributo "btn_action=??" con el id del hash donde se define el comportamiento de cada boton de la toolbar.
			// Este atributo es utilizado para asignar el evento Click dinamico.
			$.each($interactionToolbarButtons.find("[id]"), function(_, btn)
			{
				var $btn = $(btn);
				var btn_id = $btn.attr("id").toLowerCase();
				$.each(self.m_interactionToolbarButtons, function(definition_id, definition)
				{
					if (definition.id.toLowerCase() === btn_id) {
						$btn.attr("btn_action", definition_id);
						return false; //break
					}
				});
			});
		}
		this.m_cachedHTMLInteractionToolbarButtons = $interactionToolbarButtons.html() || "";
	};

	ClientWindow.prototype.getCachedHTMLInteractionToolbarButtons = function()
	{
		return this.m_cachedHTMLInteractionToolbarButtons;
	};

	ClientWindow.prototype.isTransferredInteraction = function()
	{
		if (!this.m_interaction) {
			return false;
		}
		let transfer_info = this.m_interaction.check_is_transfer();
		if (!transfer_info || !transfer_info.origin_data) {
			return false;
		}
		return transfer_info;
	};

	ClientWindow.prototype.setupTransferInfo = function()
	{
		const self = this;

		//first clean transfer info
		const $header_transfer_info = $('#headerTransferInfo');
		$header_transfer_info.hide();

		const $transfer_container = $('#transferContainer');
		$transfer_container.html("");

		const $body = $("body");
		$body.removeClass("interactionTransferred");

		let transfer_info = this.isTransferredInteraction();
		if (!transfer_info) {
			return;
		}
		$body.addClass("interactionTransferred");

		let origin_data = transfer_info.origin_data;
		let target_data = transfer_info.target_data || {};
		let target_type = (target_data.type || "").toLowerCase();

		let transferrerType;
		let transferrerId;

		if (transfer_info.transferrer) {
			if (transfer_info.transferrer === "bot") {
				transferrerType = "bot";
				transferrerId = "bot";
			}
			else {
				transferrerType = "agent";
				let transferrerId_a = transfer_info.transferrer.match(/(.+)@([^@]+)$/);
				transferrerId = transferrerId_a.length > 1? transferrerId_a[1] : null;
			}
		}
		else if (origin_data.from_agent) {
			transferrerType = "agent";
			let transferrerId_a = origin_data.from_agent.match(/(.+)@([^@]+)$/);
			transferrerId = transferrerId_a.length > 1? transferrerId_a[1] : null;
		}
		else if (origin_data.from_rule) {
			transferrerType = "rule";
			transferrerId = origin_data.from_rule;
		}
		else if (origin_data.from_bot) {
			transferrerType = "bot";
			transferrerId = origin_data.from_bot;
		}
		if (!transferrerType || !transferrerId) {
			return;
		}

		/*
			Funcion que retorna un string con la descripcion de la campana+cuenta+al
			Tanto para campana source o target de la transferencia a la campana.
		*/
		var get_campaign_definition = function (campaign_info, account_info, al_info) {
			// obtengo el origen de la transferencia. TODO: ver si se factoriza a otra funcion!!
			let transfer_camp_info = "";
			let campaign, campaign_name, account_name, attention_level_name;

			// campana origen
			let full_campaign_id = typeof(campaign_info) === "object" && campaign_info.id ? campaign_info.id : campaign_info;
			if (full_campaign_id) {
				campaign = inConcert.app.getCampaign(full_campaign_id);
				if (self.m_interaction.campaign !== full_campaign_id) {
					if (campaign) {
						campaign_name = campaign.getDescription();
					}
					campaign_name = campaign_name || (typeof(campaign_info) === "object" && campaign_info.name ? campaign_info.name : inConcert.app.get_id(full_campaign_id));
				}
			}

			// cuenta origen
			if ( account_info && account_info.id ) {
				if ( self.m_interaction.account !== account_info.id ) {
					if ( campaign ) {
						let account = campaign.getAccount( account_info.id );
						if ( account ) {
							account_name = campaign.getAccountDescription( account );
						}
					}
					account_name = account_name || account_info.name;
				}
			}

			// attention level origen
			if ( al_info && al_info.id ) {
				if ( !self.m_interaction.attention_level || self.m_interaction.attention_level.id !== al_info.id ) {
					if ( campaign ) {
						let attention_level = campaign.getAttentionLevel( al_info.id );
						if ( attention_level ) {
							attention_level_name = attention_level.name;
						}
					}
					attention_level_name = attention_level_name || al_info.name;
				}
			}

			if ( campaign_name ) {
				if ( !account_name && !attention_level_name ) {
					transfer_camp_info = format("<%=_T('Campaign: {0}.')%>", campaign_name);
				}
				else {
					if ( account_name ) {
						if ( attention_level_name ) {
							transfer_camp_info = format("<%=_T('Campaign: {0} - Account: {1} - Attention level: {2}.')%>", campaign_name, account_name, attention_level_name);
						}
						else {
							transfer_camp_info = format("<%=_T('Campaign: {0} - Account: {1}.')%>", campaign_name, account_name);
						}
					}
					else {
						transfer_camp_info = format("<%=_T('Campaign: {0} - Attention level: {1}.')%>", campaign_name, attention_level_name);
					}
				}
			}
			else if ( account_name ) {
				if ( attention_level_name ) {
					transfer_camp_info = format("<%=_T('Account: {0} - Attention level: {1}.')%>", account_name, attention_level_name);
				}
				else {
					transfer_camp_info = format("<%=_T('Account: {0}.')%>", account_name);
				}
			}
			else if ( attention_level_name ) {
				transfer_camp_info = format("<%=_T('Attention level: {0}.')%>", attention_level_name);
			}

			// retorna la info de la campana, cuenta y al
			return transfer_camp_info;
		};

		/*
			Crea el markup destino de la transferencia, tanto del source como del target.
		*/
		var $create_transfer_markup = function()
		{
			var transfer_origin = get_campaign_definition(origin_data.from_campaign, origin_data.from_account, origin_data.from_attention_level);

			var transfer_target;
			if ( window.monitor && transfer_info.to ) { // show different target when monitoring
				transfer_target = format("<%=_T('Transferred to {0}')%>", transfer_info.to.match(/(.+)@([^@]+)$/)[1]);
			} else {
				transfer_target = "<%=_T('Transferred to you')%>";
				// el step CREATED se utiliza cuando se ha transferido a campana/cuenta
				if (transfer_info.step === "CREATED" || target_type !== "member") {
					transfer_target = "<%=_T('Transferred')%>";
				}
			}

			let transfer_date = inConcert.Utils.getDateFromString(transfer_info.timestamp);
			let tooltip = format("{0} {1}", transfer_target, inConcert.Utils.formatDate(transfer_date || ""));

			const time_ago = inConcert.Utils.timeAgo(transfer_date);
			transfer_target = format("{0} {1}", transfer_target, time_ago);

			const can_show_transfer_origin = (transfer_origin || "").length > 0;
			if (can_show_transfer_origin) {
				tooltip = tooltip + "\n" + transfer_origin;
			}
			if ((transfer_info.message || "").length > 0) {
				tooltip = tooltip + "\n" + transfer_info.message;
			}

			/*
				Crea el markup destino de la transferencia destino.
			*/
			var make_html_transfer_destination = function ()
			{
				if (!target_type) {
					return "";
				}
				let display_transfer_type = target_type;

				if (target_type === "member") {
					display_transfer_type = "<%=_T('Agent')%>";
				}
				else if (target_type === "campaign") {
					display_transfer_type = get_campaign_definition(target_data.to_campaign);
					display_transfer_type = display_transfer_type || "<%=_T('Campaign')%>";
				}
				else if (target_type === "attention_level") {
					display_transfer_type = "<%=_T('Attention Level')%>";
				}
				else if (target_type === "number") {
					display_transfer_type = "<%=_T('Number')%>";
					if (target_data.to_number) {
						let phone_number = typeof(target_data.to_number) === "object" && target_data.to_number.number ? target_data.to_number.number : target_data.to_number;
						display_transfer_type = format("<%=_T('{0}: {1}')%>", display_transfer_type, phone_number || "");
						if (target_data.to_number.contactInfo) {
							display_transfer_type = format("<%=_T('{0} - {1}')%>", display_transfer_type, target_data.to_number.contactInfo.Name);
						}
					}
				}
				let transfer_type = format("{0}", display_transfer_type);
				let trsanfer_type_css_class = "interactionTransferredTo" + target_type.toUpperCase();

				let html = "" +
					"<div class='destinyInfo " + trsanfer_type_css_class + "'>" +
						"<i class=''></i>" +
						"<div class='interactionDestinyName'>" + transfer_type + "</div>" +
					"</div>";

				return html;
			};

			const transfer_owner_name = format("{0} ({1})", transfer_info.fromName, time_ago);
			const $transfer = $("".concat(
				"<div title='", tooltip ,"' class='interactionOwnerContainerAux interactionOwnerContainerAuxTransfer'>",
					make_html_transfer_destination(),
					"<div class='ownerInfo'>",
						"<div class='interactionOwnerName'>", transfer_owner_name, "</div>",
					"</div>",
				"</div>"
			));

			return $transfer;
		};

		var get_member_name = function( member )
		{
			var name = member.Id;
			var can_name = false;
			if ( ( member.Name || "" ).trim().length > 0 ) {
				name = member.Name;
				can_name = true;
			}
			if ( ( member.FirstSurname || "" ).trim().length > 0 ) {
				name = ( can_name ? ( name + " " ) : "" ) + member.FirstSurname;
			}
			return name;
		};

		var show_info = function(memberName, avatar)
		{
			transfer_info.fromName = memberName;
			transfer_info.avatar = avatar;
			const $html = $create_transfer_markup();
			$transfer_container.html($html);
			$header_transfer_info.show();
		};

		switch(transferrerType) {
			case "agent":
				var avatar = inConcert.app.getBaseUrl() + "/users/get_avatar/" + transferrerId;
				if (transferrerId === inConcert.app.getLoggedUserId()) {
					show_info(format("{0}", inConcert.app.getLoggedUserName()), avatar);
				}
				else {
					inConcert.app.getTeamMember(
						transferrerId,
						function(member)
						{
							if ( !member ) {
								return;
							}
							show_info(format("{0}", get_member_name(member)), avatar);
						},
						true // from cache
					);
				}
				break;
			case "rule":
				show_info(format("<%=_T('Rule: {0}')%>", transferrerId), inConcert.app.getBaseUrl() + "/images/ruleAvatar.png");
				break;
			case "bot":
				show_info("<%=_T('Bot')%>", inConcert.app.getBaseUrl() + "/images/botAvatar.png");
				break;
		}
	};

	/*
		Shows current contact detail
	*/
	ClientWindow.prototype.viewContactDetails = function() {
		if (this.m_interaction.contact && this.m_interaction.contact !== "") {
			inConcert.app.getMainWindow().showContact(this.m_interaction.contact);
		}
	};

	/*
		toggle client actions if pending
	*/
	ClientWindow.prototype.toggleActionsIfPending = function()
	{
		if (!this.m_interaction.isLive()) {
			var pendingLabeled = this.m_interaction.hasPendingLabel();
			if (pendingLabeled) {
				this.change_client_state("pending");
			}
			else {
				this.change_client_state("connected");
			}
		}
	};

	/*
		Shows current labels
	*/
	ClientWindow.prototype.showLabels = function( labelMng, $container )
	{
		var self = this;
		var labelHtml = labelMng.getOwnerLabelMarkup( this.m_interaction );
		if (labelHtml.length === 0) {
			return false;
		}
		$container.append( labelHtml ).show();
		var $delete = $container.find(".butDeleteTag");

		$delete.unbind("click").bind("click", function() {
			labelMng.deleteLabelOnInteractions( $(this), [ self.m_interaction ] );
		});

		if ( this.m_interaction.hasPermission( "LABEL", this.from_app ) ) {
			$delete.show();
		}
		else {
			$delete.hide();
		}
		return true;
	};

	/*
		Shows current disposition
	*/
	ClientWindow.prototype.showDisposition = function(labelMng, $container)
	{
		var disposition = this.m_interaction.dispositionCode;
		if (!disposition) {
			return;
		}
		var dispositionHtml = labelMng.getSingleLabelMarkup({
			name       : disposition,
			className  : "interactionResultTag",
			hasActions : false,
			tooltip : ""
		});
		if (dispositionHtml.length > 0) {
			$container.append(dispositionHtml);
			return true;
		}
		return false;
	};

	/*
		Shows current labels and disposition
	*/
	ClientWindow.prototype.showLabelsAndDisposition = function()
	{
		if (!this.m_popupLabels) {
			return;
		}
		var labelMng = this.m_popupLabels.getLabelManagementObject();
		if (!labelMng) {
			return;
		}
		var $container = $("#labelContainer");
		const $dispositionCodeContainer = $("#dispositionCodeContainer");
		$container.html("");
		$dispositionCodeContainer.html("");
		var showingDisposition = this.showDisposition(labelMng, $dispositionCodeContainer);
		var showingLabels = this.showLabels(labelMng, $container);
		$dispositionCodeContainer[showingDisposition?"show":"hide"]();
		if (showingLabels) {
			$container.show();
			$("body").addClass("interactionWithLabels");
		}
		else {
			$("body").removeClass("interactionWithLabels");
		}
		if (this.adjustBodyTop) {
			this.adjustBodyTop();
		}
		this.toggleActionsIfPending();
	};

	/*
		Initialize controls labels
	*/
	ClientWindow.prototype.intializeLabels = function() {
		if (!inConcert.LabelsPopup) {
			return;
		}
		this.m_popupLabels = new inConcert.LabelsPopup({ from_app: this.from_app });
		//shows current labels
		this.showLabelsAndDisposition();
	};

	/*
		Initialize controls comments
	*/
	ClientWindow.prototype.intializeComments = function()
	{
		var self = this;
		//load comments for the interaction and setup comment handler
		var comments = this.m_interaction.getComments();
		var totalComments = comments ? comments.length : 0;

		var commentsClass;
		if (totalComments === 0) {
			commentsClass = "interactionCommentsNew";
		}
		else {
			//check if the comments are unread for the current user
			if (this.m_interaction.youHasUnreadComments()) {
				commentsClass = "interactionCommentsUnread";
			}
			else {
				commentsClass = "interactionCommentsRead";
			}
		}

		$(".interactionComments")
			.addClass(commentsClass)
			.find('span').text(totalComments > 0 ? totalComments : "")
			.click(function() {
				self.__openComment($(this));
			});
	};

	ClientWindow.prototype.__openComment = function($parentNode)
	{
		var self = this;

		this.m_commentPopup = new inConcert.CommentsPopup({ from_app: this.from_app });

		this.m_commentPopup.show({
			interaction: this.m_interaction,
			parentNode: $parentNode,
			callbacks: {
				onOpen: function() {
					//mark interaction comments as read
					self.m_interaction.readComments(function(success) {
						if (!success) {
							return;
						}
						$(".interactionComments")
							.removeClass("interactionCommentsUnread")
							.addClass("interactionCommentsRead");
					});
				},
				onClose: function(params) {
					if (typeof params !== "object") {
						return;
					}
					if (params.errorMessage) {
						self.displayStatusMessage(params.errorMessage, "Error", false, 5000);
					}
				}
			}
		});
	};

	ClientWindow.prototype.initializeFinalization = function()
	{
		let self = this;
		if ( !inConcert.FinalizationUI ) {
			return;
		}
		this.m_finalization = new inConcert.FinalizationUI({
			client_window : this,
			container_id  : 'finalization-container',
			on_ready      : function() {
				self.onReadyFinalization();
			}
		});
	};

	ClientWindow.prototype.onReadyFinalization = $.noop;
	
	ClientWindow.prototype.updateFinalization = function()
	{
		if (this.m_finalization) {
			this.m_finalization.set_values();
		}
	};

	/**
	* If the current interaction has a ticket assigned, show it and enable assorted buttons and stuff.
	*/
	ClientWindow.prototype.initializeTicketInfo = function(ticket)
	{
		if (!this.m_interaction) {
			return;
		}

		const $tickets_info = $("[tickets_info='true']");
		$tickets_info.html("");

		if (this.m_interaction.isInternal() || this.m_interaction.isBeingComposed()) {
			return;
		}

		ticket = ticket || this.m_interaction.getAssignedTicket();
		if (ticket) {
			$tickets_info.append( $("<div>" + ticket + "</div>") );
		}
	};

	ClientWindow.prototype.updateTicketInfo = function(ticket)
	{
		this.initializeTicketInfo(ticket);
		this.updateToolbarButtons();
	};

	ClientWindow.prototype.checkTicketInfo = function()
	{
		let ticket = this.m_interaction.getAssignedTicket();
		if (ticket && !$("[tickets_info='true']").html()) {
			this.initializeTicketInfo();
		}
	};

	ClientWindow.prototype.set_cached_availables_campaigns = function( campaigns )
	{
		this.cachedAvailablesCampaigns = inConcert.Utils.cloneObject( campaigns || {} );
	};

	ClientWindow.prototype.getCampaign = function(campaignId) {
		var self = this;
		var campaign = inConcert.app.getCampaign(campaignId);
		if (campaign) {
			return campaign;
		}
		// add of the cached campaigns list
		if (this.cachedAvailablesCampaigns) {
			for (var ci = 0; ci < this.cachedAvailablesCampaigns.length; ci++) {
				campaign = this.cachedAvailablesCampaigns[ci];
				if (campaign.Id === campaignId) {
					this.cachedCampaign = campaign;
					break;
				}
			}
		}
		/*
		A campaign object is created with the basic functions:
			if the campaign was deleted and the user closes and reopens the client window
		*/
		return this.cachedCampaign || {
			Id: campaignId,
			getDescription: function() {
				return campaignId;
			},
			getAccount: function() {
				return {id: self.m_interaction.account, state: "0"};
			},
			getAccountDescription: function() {
				return self.m_interaction.account;
			},
			accounts: [{id: self.m_interaction.account}]
		};
	};

	ClientWindow.prototype.displayCampaignInformation = function(campaignId)
	{
		if (window.taskbar && window.interaction) {
			window.taskbar.onLocalInteractionUpdate(window.interaction, true);
		}
		if (this.m_interaction && !this.m_interaction.contact) {
			var campaign = this.getCampaign(campaignId);
			if (!campaign) {
				return;
			}
			let info = {
				contactID         : campaign.getDescription(),
				contactFromAccount: null,
				avatarURI         : inConcert.app.getBaseUrl() + this.m_interaction.getAvatarURI()
			};
			this.update_header_contact_info(info);
		}
	};

	ClientWindow.prototype.campaignIsOperational = function()
	{
		var self = this;
		var permissionFn;

		if ( ( this.from_app || "" ).toLowerCase() === "interaction" ) {
			permissionFn = function() {
				var permission = {
					enabled: self.hasPermissionToAction( "CREATE_OUTGOING" )
				};
				if ( !permission.enabled ) {
					permission.message = "<%=_T('Unable to create outgoing conversation. You not have permission.')%>";
				}
				return permission;
			};
		}

		return this.m_interaction.campaignIsOperational( permissionFn );
	};

	ClientWindow.prototype.change_account = function(campaignId, new_account_id)
	{
		var old_account_id = this.m_interaction.getAccount();
		this.m_interaction.setAccount(new_account_id || "");
		var campaignIsOperational = this.campaignIsOperational();
		if (!campaignIsOperational.success) {
			//if campaign is not operational I don t care about the account statuses
			this.hideConflictAccountIcon();
			return;
		}
		if (!new_account_id) {
			return;
		}
		var campaign = this.getCampaign(campaignId);
		var account = campaign.getAccount(new_account_id, this.m_interaction.type);
		if (!account) {
			this.showConflictAccountIcon("<%=_T('This account has been removed.')%>");
			return;
		}
		if (account.state === "0" || !account.status) {
			this.showConflictAccountIcon();
		}
		else if (account.status.status !== "OK") {
			this.showConflictAccountIcon(account.status.status);
		}
		else {
			this.hideConflictAccountIcon();
		}
		this.on_account_changed( old_account_id, new_account_id );
	};

	ClientWindow.prototype.on_account_changed = function(/*old_account_id, new_account_id*/)
	{
		this.show_attention_level_section();
		inConcert.app.m_mainWindow.updateInteractions([this.m_interaction]);
	};

	ClientWindow.prototype.change_campaign = function(campaign, currentAccountId)
	{
		var campaignId = campaign.Id;
		var currentCampaignId = this.m_interaction.campaign;
		if (currentCampaignId !== campaignId) {
			this.m_interaction.setCampaign(campaignId);
		}

		var campaignIsOperational = this.campaignIsOperational();
		if (!campaignIsOperational.success) {
			this.hideConflictAccountIcon();
			this.showConflictCampaignIcon(campaignIsOperational.errorMessage);
		}
		else {
			this.hideConflictCampaignIcon();
		}

		var availables_accounts = 0;
		var $cboAccounts = $('#cboAccounts');
		var addAccountToCombo = function(accountId, accountName) {
			if (!accountId || !accountName) {
				return;
			}
			$cboAccounts
				.append($("<option></option>")
				.attr("value", accountId)
				.text(accountName));
			availables_accounts++;
		};

		currentAccountId = currentAccountId || $cboAccounts.val();
		$cboAccounts.html('');
		var hashAddedAccounts = {};
		var accounts = campaign.accounts;
		for (var ai = 0; ai < accounts.length; ai++) {
			var account = accounts[ai];
			if (account.proxyType !== this.m_interaction.type) {
				continue;
			}
			hashAddedAccounts[account.id] = true;
			addAccountToCombo(account.id, campaign.getAccountDescription(account));
			if (!currentAccountId) {
				currentAccountId = account.id;
			}
		}

		if (currentCampaignId === campaignId) {
			if (hashAddedAccounts[currentAccountId]) {
				// select current account
				$cboAccounts.val(currentAccountId);
			}
			else {
				// if the current account has been removed, it is added again to display in the list.
				addAccountToCombo(currentAccountId, currentAccountId);
			}
			this.change_account(campaignId, currentAccountId);
		}
		else {
			// campaign has changed, select the first account
			 $cboAccounts.find("first-child").attr("selected", "selected");
			this.change_account(campaignId, $cboAccounts.val());
		}

		// desactiva el cbo de accounts si:
			// 1 - la campana o accounts no se encuentran operacionales (disabled, etc)
			// 2 - si la saliente de pending se va a generar a partir de una appointment
			// 3 - si existe una sola cuenta en el cbo
		var disabled_accounts_selector = false;
		if ( !campaignIsOperational.success || this.m_interaction.appointment || availables_accounts <= 1 ) {
			disabled_accounts_selector = true;
		}
		$cboAccounts.prop( "disabled", disabled_accounts_selector );

		this.on_account_changed();
		this.displayCampaignInformation(campaignId);
	};

	ClientWindow.prototype.showConflictCampaignIcon = function(errorMessage) {
		$(".icoGroupConflict")
			.addClass("showConflict")
			.inconcert_tooltip({
				message: errorMessage ? errorMessage : "<%=_T('Campaign is not available.')%>"
			}
		);
	};

	ClientWindow.prototype.hideConflictCampaignIcon = function() {
		$(".icoGroupConflict").removeClass("showConflict").inconcert_tooltip({detach: true});
	};

	ClientWindow.prototype.showConflictAccountIcon = function( errorMessage, $conflict ) {
		$conflict = $conflict || $("#headerGeneralInfo .icoAccountConflict");

		$conflict
			.addClass("showConflict")
			.inconcert_tooltip({
				message: errorMessage ? errorMessage : "<%=_T('Account is disabled.')%>"
			}
		);
	};

	ClientWindow.prototype.hideConflictAccountIcon = function( $conflict ) {
		$conflict = $conflict || $("#headerGeneralInfo .icoAccountConflict");
		$conflict.removeClass("showConflict").inconcert_tooltip({detach: true});
	};

	ClientWindow.prototype.displayConflictCampaignMessage = function(inifinite, message)
	{
		var timeout_ms = inifinite ? undefined : 5000;
		if (!message) {
			var campaignId = $("#cboCampaigns").val();
			var campaign = this.getCampaign(campaignId);
			var campaignName = campaign ? campaign.getDescription() : campaignId;
			message = format("<%=_T('{0} campaign is not available.')%>", campaignName || ""); 
		}
		this.displayStatusMessage(
			message,
			"Alert",
			false,
			timeout_ms);
	};

	ClientWindow.prototype.displayConflictAccountMessage = function() {
		var campaign = this.getCampaign($("#cboCampaigns").val());
		var accountId = $("#cboAccounts").val() || this.m_interaction.getAccount();
		var account = campaign.getAccount(accountId, this.m_interaction.type);
		var accountName = account ? campaign.getAccountDescription(account) : accountId;
		this.displayStatusMessage(
			format("<%=_T('{0} account is disabled.')%>", accountName),
			"Alert",
			false,
			5000);
	};

	/*
	 * Add campaigns list to the control (pending interactions)
	 */
	ClientWindow.prototype.addCampaignsToList = function(selectedCampaign, selectedAccount)
	{
		var self = this;
		if (!this.m_interaction) {
			return;
		}
		var availableCampaigns = this.m_interaction.getAllowedCampaigns(inConcert.app.getCampaigns());
		if (!availableCampaigns) {
			this.set_cached_availables_campaigns( availableCampaigns );
			return;
		}

		const $campaigns = $('#cboCampaigns');

		var addCampaignToCombo = function(campaign)
		{
			/*
				Cuando estamos creando una interaccion saliente, se valida si el permiso de creacion se encuentra activo para la campana.
				Asi sabemos si se debe de agregar o no al combo de campanas.
			*/
			if (self.m_interaction.isBeingComposed()) {
				const creation_permissions = {
					"CALL"      : "int_quick_create_call",
					"CHAT"      : false,
					"FACEBOOK"  : true,
					"IM"        :"int_quick_create_im",
					"INSTAGRAM" : true,
					"MAIL"      : "int_quick_create_mail",
					"SMS"       : "int_quick_create_sms",
					"TWITTER"   : true,
					"WHATSAPP"  : "int_quick_create_whatsapp",
					"RCS" 	    : "int_quick_create_rcs",
					"YOUTUBE"   : true,
					"APPSTORE"  : true,
					"GOOGLEPLAY": true
				};
				let permission_id = creation_permissions[self.m_interaction.type.toUpperCase()];
				if (typeof(permission_id) === "string") {
					const has_validate_all_overrides = true;
					const has_creation_permission = inConcert.permissions.to_action("interaction", permission_id, undefined, campaign.Id, true /* no log */, has_validate_all_overrides);
					if (!has_creation_permission) {
						return false;
					}
				}
			}

			let is_default_campaign = self.m_user_default_campaign_id === campaign.Id;
			let campaign_desc = campaign.getDescription() + (is_default_campaign ? "<%=' (' .. _T('Default') .. ')'%>" : "");
			const $campaign = $("<option></option>");
			$campaign.attr("value", campaign.Id);
			$campaign.text(campaign_desc);

			$campaigns.append($campaign);

			return true;
		};

		var enabledCampaignControls = function(campaigns) {
			//if only one campaign available, do not show a combo for the campaigns(this gets done by CSS against the disabled style)
			$campaigns.prop("disabled", campaigns <= 1 || self.m_selectedContact);
			$('#cboAccounts')[campaigns === 0 ? "hide" : "show"]();
		};

		$campaigns.html('');
		var isAvailableSelectedCampaign = false;
		var availableCampaignsLength = availableCampaigns.length;
		let countAddedCampaigns = 0;
		// add availables campaigns
		var campaign;
		for (var ci = 0; ci < availableCampaignsLength; ci++) {
			campaign = availableCampaigns[ci];
			let is_added_campaign = addCampaignToCombo(campaign);
			if (is_added_campaign) {
				countAddedCampaigns++;
				if (!isAvailableSelectedCampaign && (!selectedCampaign || campaign.Id === selectedCampaign)) {
					isAvailableSelectedCampaign = true;
				}
			}
		}
		if (!isAvailableSelectedCampaign && self.m_interaction.isBeingComposed()) {
			selectedCampaign = null;
		}
		var selectedCampaignId = selectedCampaign || $campaigns.val();
		campaign = this.getCampaign(selectedCampaignId);
		if (selectedCampaign && !isAvailableSelectedCampaign) {
			// add unavailable campaign
			addCampaignToCombo(campaign);
		}
		enabledCampaignControls(countAddedCampaigns + (!isAvailableSelectedCampaign ? 1 : 0));
		$("#cboCampaigns option[value='" + selectedCampaignId + "']").attr("selected", "selected");

		if (selectedCampaignId && selectedCampaignId.length > 0) {
			this.change_campaign(campaign, selectedAccount);
		}
		else {
			//default on the first campaign of the list
			if (countAddedCampaigns > 0) {
				var availableCampaignId = availableCampaigns[0].Id;
				$campaigns.val(availableCampaignId);
				this.change_campaign(campaign, selectedAccount);
			}
		}
		this.set_cached_availables_campaigns( availableCampaigns );
	};

	/*
	 * Assign campaign and account to the controls (not pending interactions)
	 */
	ClientWindow.prototype.setCampaignAndAccount = function()
	{
		if (!this.m_interaction) {
			return;
		}
		var campaignId = this.m_interaction.campaign;
		if ( campaignId ) {
			var campaign = this.getCampaign(campaignId);
			$('#cboCampaigns').append(
				$("<option></option>").attr("value", campaignId).text(campaign.getDescription())
			).val(campaignId);
			var accountId = this.m_interaction.getAccount();
			if (accountId) {
				var account = campaign.getAccount(accountId, this.m_interaction.type);
				var accountName = account ? campaign.getAccountDescription(account) : accountId;
				$('#cboAccounts').append(
					$("<option></option>").attr("value", accountId).text(accountName)
				).val(accountId);
			}
		}
		this.disableCampaignAccountsControls();
		this.show_attention_level_section();
	};

	ClientWindow.prototype.show_attention_level_section = function()
	{
		const self = this;
		const $section = $("#attention_level_section");
		const $combo = $("#cboAttentionLevels");
		if (!$section.length || !$combo.length) {
			return;
		}
		
		function setAttentionLevels(als = []) {
			if (!als.length) {
				$combo.html("");
				$section.hide();
				return;
			}
			let current_al = $combo.val();
			if (self.m_interaction.draftAttentionLevel) {
				current_al = self.m_interaction.draftAttentionLevel;
				self.m_interaction.draftAttentionLevel = null;
			}
			$combo
				.html(als.map(al => 
					`<option value="${al.id}" ${al.id === current_al && "selected" || ""}>${al.name}</option>`))
				.prop("disabled", als.length === 1);
			$section.show();
		}
		
		if (!self.m_interaction) {
			return setAttentionLevels();
		}
		const account_id = $("#cboAccounts").val();
		const als = self.m_interaction.getAvailableAttentionLevels( account_id );
		setAttentionLevels(als);
	};
	
	ClientWindow.prototype.getSelectedAttentionLevel = function() {
		return $("#cboAttentionLevels").val() || null;
	};

	/*
		Setups combos and handlers used to change campaign and account used in an outgoing interaction
	*/
	ClientWindow.prototype.setupCampaignSelector = function() {
		var self = this;
		var creationDate;
		if (this.m_interaction) {
			creationDate = this.m_interaction.creationDate();
		}
		else {
			creationDate = inConcert.Utils.userFormatDate(new Date());
		}
		$("#creationDate").text(creationDate);
		//build the list of available campaign/accounts for this interaction type and add the campaigns to the combo
		$("#cboCampaigns")
			.html("")
			.change(function() {
				var campaign = self.getCampaign($(this).val());
				self.change_campaign(campaign);
				inConcert.app.m_mainWindow.updateInteractions([self.m_interaction]);
			});
		$("#cboAccounts")
			.html("")
			.change(function() {
				self.change_account($("#cboCampaigns").val(), $(this).val());
			});
		if (this.m_interaction.isBeingComposed()) {
			this.m_user_default_campaign_id = this.m_user_default_campaign_id || inConcert.app.getLoggedUserDefaultCampaignByChannel(this.m_interaction.type);
			let selected_campaign_id = this.m_interaction.campaign;
			if (!selected_campaign_id) {
			// Si aun NO tiene campana, asigno la campana por default del usuario para el canal.
				if (this.m_user_default_campaign_id) {
					selected_campaign_id = this.m_user_default_campaign_id;
				}
			}
			this.addCampaignsToList(selected_campaign_id, this.m_interaction.account);

			// si la interaction pending va a ser creada a partir de un contact frm
			this.m_interaction.hasContactFormInteractionParent( function( is_contactform_parent_it ){
				if ( !is_contactform_parent_it ) {
					return;
				}
				self.disableCampaignAccountsControls(); // desactivo el cbo campana y account
			});
		}
		else {
			var availableCampaigns = this.m_interaction.getAllowedCampaigns( inConcert.app.getCampaigns() );
			this.set_cached_availables_campaigns( availableCampaigns );
			this.setCampaignAndAccount();
		}
	};

	ClientWindow.prototype.enableCampaignControls = function() {
		var self = this;
		if (!self.m_interaction) {
			return;
		}
		var availableCampaigns = self.m_interaction.getAllowedCampaigns(inConcert.app.getCampaigns());
		if (!availableCampaigns) {
			return;
		}
		if (availableCampaigns.length === 1) {
			return;
		}
		$('#cboCampaigns').prop("disabled", false);
	};

	ClientWindow.prototype.disableCampaignAccountsControls = function() {
		$('#cboCampaigns').prop("disabled", true);
		$('#cboAccounts').prop("disabled", true);
		$('#cboAttentionLevels').prop("disabled", true);
	};

	ClientWindow.prototype.updateFirstOutgoingEvent = function()
	{
		// si es el primer evento que se genera de una saliente debo mostrar el mismo!!
		if (this.m_interaction && this.m_interaction.initiative === "OUTBOUND" &&
				this.m_interaction.events && this.m_interaction.events.length === 1 && !this.m_interaction.isFinished())
		{
			this.m_currentEvent = 0;
			this.readAndOpenEvent();
		}
	};

	ClientWindow.prototype.updateContactDisplay = function(interaction)
	{
		var self = this;
		var currentInteraction = interaction ? interaction : this.m_interaction;
		if (!currentInteraction) {
			return;
		}
		var avatarURI = inConcert.app.getBaseUrl() + currentInteraction.getAvatarURI();
		$("#contactFromAvatar").attr("data", avatarURI);

		var displayName;
		var clientName = currentInteraction.getClientName();
		var displayAddress = currentInteraction.getDisplayAddress();

		// if we have contact name and an address, compose a display name with both.
		if (displayAddress && clientName) {
			if (clientName !== displayAddress && currentInteraction.type !== 'TWITTER') {
				displayName = displayAddress;
			}
			else{
				displayName = clientName;
			}
		}
		// if we don t have an address, just show the client name.
		else if (clientName){
			displayName = clientName;
		}
		// if we don t have an client name, just show the address.
		else if (displayAddress){
			displayName = displayAddress;
		}
		else{
			displayName = "";
		}

		// Make contact linkable to open popup window if:
		// + is not an internal chat
		// + interaction account has contact management enabled
		// + visualize_address permission is enabled

		const visualize_address = inConcert.app.canVisualizeAddress(null, currentInteraction.campaign, null, true);

		if ((displayName || (currentInteraction.type === "CALL" && currentInteraction.hasContactConflict())) && !currentInteraction.isInternal() && currentInteraction.accountHasContactManagementEnabled( this.from_app ) && visualize_address){
			if(clientName && clientName.length) {
				this.setTitle(clientName);
			}
			$("#contactID").replaceWith("<span id='contactID' href='javascript:void(0);'>" + displayName + "</span>");
			$("#iconContactId").show()
				.unbind('click')
				.bind('click', function() {
					let options = {
						from_app: self.from_app,
						interaction : currentInteraction,
						name : clientName,
						avatar : avatarURI,
						external_id : currentInteraction.interactionData.gatheredContactData && currentInteraction.interactionData.gatheredContactData.external_id || null
					};
					// interaction<->contact relation is 1-1,
					// but we are safe to assume this here because this
					// function is only used on 1-1 communication with the client
					// (i.e. interactions without events).
					if (currentInteraction.contact && currentInteraction.contact.length > 0) {
						options.contact_id = currentInteraction.contact;
					}
					else {
						options.addresses = currentInteraction.interactionData.contactAddresses;
					}
					const contactWindow = new inConcert.ContactWindow(options);
					contactWindow.show();
				});
		}
		else {
			const $iconContactId = $("#iconContactId");
			$iconContactId.hide();
			$("#contactID").replaceWith("<span id='contactID'>" + displayName + "</span>");
		}
	};

	ClientWindow.prototype.getContactInfoCallback = function() {
		return null;
	};

	ClientWindow.prototype.onInteractionGone = function(undefined /* interaction */, showMessage)
	{
		const invokeId = inConcert.Utils.generateGuid();
		console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". START -> Interaction.id: " + this.m_interaction?.id + ", showMessage: " + showMessage);
		if (this.m_popupLabels) {
			console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". close labels popup");
			this.m_popupLabels.close();
		}
		if (this.m_commentPopup) {
			console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". close comments popup");
			this.m_commentPopup.close();
		}
		$("body").removeClass("interactionEnded");
		const deferredClose = _.bind(function(interaction) {
			console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". END -> deferredClose fn");
			if (interaction === this.m_interaction) {
				console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". 1 deferredClose fn - YES close window");
				this.close();
			}
			else {
				console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". 2 deferredClose fn - NO close window");
			}
		}, this, this.m_interaction);
		if (showMessage) {
			this.displayStatusMessage("<%=_T('Current conversation has ended')%>", "Info");
		}
		console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". zeroTimeout execute");
		inConcert.Utils.zeroTimeout(function() {
			console.debug("[ClientWindow.onInteractionGone] invokeId: " + invokeId + ". zeroTimeout callback");
			deferredClose();
		});
	};

	ClientWindow.prototype.makeMarkupOwnerInformation = function() {
		const $owner_node = $("#ownerSection");
		$owner_node.html(
			'<div class="interactionOwnerContainerAux">' +
				'<div class="ownerInfo">' +
					'<div class="interactionOwnerName"></div>' +
					'<div class="interactionOwnerRol"></div>' +
				'</div>' +
			'</div>'
		);
	};
	
	/*
	 This is to know who the fuck owns the conversation when the user is looking at a conversation owned by another user
	*/
	ClientWindow.prototype.loadOwnerInformation = function()
	{
		const self = this;

		var get_member_name = function( member )
		{
			let name = member.Id;
			let can_name = false;
			if ( ( member.Name || "" ).trim().length > 0 ) {
				name = member.Name;
				can_name = true;
			}
			if ( ( member.FirstSurname || "" ).trim().length > 0 ) {
				name = ( can_name ? ( name + " " ) : "" ) + member.FirstSurname;
			}
			return name;
		};
		
		const $owner_node = $("#ownerSection");

		this.change_client_state("monitoring");
		//a conversation monitored may have no assigned agent, maybe it s on the IVR yet
		if ( !this.m_interaction.assignedAgents || this.m_interaction.assignedAgents.length === 0 ) {
			$owner_node.hide();
			return;
		}

		//TODO, handle special case here if trying to monitor a CALL conversation in the middle of a transfer
		//the agent to me monitored should be selected from the interaction s event list
		const agentAtVCC = this.m_interaction.assignedAgents[0];
		let agentID = agentAtVCC;
		if (agentAtVCC.split("@") > 0) {
			agentID = agentAtVCC.match(/(.+)@([^@]+)$/)[1];
		}

		inConcert.app.getTeamMember(
			agentID,
			function(teamMember) {
				if ( teamMember ) {
					self.m_monitoredMember = teamMember;
					$owner_node.find(".interactionOwnerName").text(get_member_name(teamMember));
					$owner_node.find(".interactionOwnerRol").text(teamMember.UserGroup);
					$owner_node.show();
				}
				else {
					$owner_node.hide();
				}
			},
			true // from cache
		);
	};

	ClientWindow.prototype.showScreenReadOnlyAfterLoading = function () {
		if (!this.m_interaction) {
			return;
		}
		const lastStep = this.m_interaction.getLastStep();
		const beforeLastStep = this.m_interaction.getBeforeLastStep();
		if ((this.m_interaction.isTransferring() && lastStep && lastStep.origin_data && lastStep.origin_data.type === "bot") ||
			this.m_interaction.isBot() && beforeLastStep && beforeLastStep.action === "TRANSFERRING" && beforeLastStep.origin_data && beforeLastStep.origin_data.type === "bot") {
			this.showScreenReadOnly(true);
		}
	};

	ClientWindow.prototype.onUpdateSidebarInformation = function () {

		if (this.sidebar)
		{
			this.sidebar.onInformationUpdated();
		}
	};

	ClientWindow.prototype.showScreenReadOnly = function (show) {
		const $screenReadOnly = $("#transferToBotWaiting");
		$screenReadOnly.hide();
		if (show) {
			$screenReadOnly.show();
		}
	};

	ClientWindow.prototype.change_client_state = function( new_state )
	{
		if (!this.m_client_state || this.m_client_state !== new_state) {
			this.m_client_state = new_state;
		}
		else {
			// si me han invocado y el estado no ha cambiado me voy.
			return new_state;
		}
		var $body = $("body");
		var $state_flap = $(".clientStateFlap");
		$body.removeClass("interactionArchived interactionQueued interactionWaiting interactionDisconnected interactionInvalid interactionClientPending");
		$body.removeClass("gossipingClient monitoringClient monitoringQueuedClient monitoringWaitingClient coachingClient");

		if ( this.is_archive ) {
			new_state = "archived";
		}
		else if ( this.is_queued ) {
			new_state = "queued";
		}
		else if ( this.is_waiting ) {
			new_state = "waiting";
		}
		else if ( this.is_monitor && new_state === "connected" ) {
			new_state = "monitoring";
		}

		switch( new_state )
		{
			case "archived":
				$state_flap.text("<%=_T('Archived')%>");
				$body.addClass("interactionArchived");
			break;

			case "connected":
				$state_flap.text("");
			break;

			case "disconnected":
				$state_flap.text("<%=_T('Disconnected')%>");
				$body.addClass("interactionDisconnected");
			break;

			case "ended":
				$state_flap.text("<%=_T('Ended')%>");
				$body.removeClass("interactionEnded").addClass("interactionEnded");
			break;

			case "invalid":
				$state_flap.text("<%=_T('Invalid')%>");
				$body.addClass("interactionInvalid");
			break;

			case "monitoring":
				var label = "<%=_T('Monitoring')%>";
				var monitoring_class = "monitoringClient";

				if ( this.from_app === "interaction" ) {
					if ( (this.m_interaction.isTaken() && !this.m_interaction.belongsToMe()) || (this.m_interaction.isFinished() && !this.m_interaction.archivedBelongsToMe()) ) {
						label = "<%=_T('Belongs to another user')%>";
						monitoring_class = "gossipingClient";
					}
				}
				//a conversation monitored may have no assigned agent, maybe it s on the IVR yet
				//if ( !this.m_interaction.assignedAgents || this.m_interaction.assignedAgents.length === 0 ) {
				if ( this.m_interaction.isQueued() ) {
					monitoring_class = "monitoringQueuedClient";
				}

				$state_flap.text(label);
				$body.addClass(monitoring_class);
			break;

			case "pending":
				$state_flap.text("<%=_T('Pending')%>");
				$body.addClass("interactionClientPending");
			break;

			case "queued":
				$state_flap.text("<%=_T('Queued')%>");
				$body.addClass("interactionQueued");
			break;

			case "waiting":
				$state_flap.text("<%=_T('Waiting')%>");
				$body.addClass("interactionWaiting");
			break;

			case "wrapping_up":
				$state_flap.text("<%=_T('Wrapping Up')%>");
				$body.addClass("interactionWrapping");
			break;
		}

		this.onChangedClientState(new_state);
		return new_state;
	};

	ClientWindow.prototype.onNewEvents = $.noop;

	ClientWindow.prototype.onChangedClientState = $.noop;

	ClientWindow.prototype.isLiveWindow = function() {
		return false;
	};

	/*
		Ajusta la ventana en modo MAXIMIZED o NORMAL, solo para interacciones SALIENTES, ya que cuando la interaccion esta en construccion y es PENDING la misma
		no tiene una campana ni una cuenta, y por ende no se puede llegar a la configuracion de la campana/cuenta.

		En el momento en que la interaccion SALIENTE ya tiene una campana y cuenta asociada ahi si ya se puede realizar la validacion.
	*/
	ClientWindow.prototype.openWindowMode = function() {
		if (!this.m_interaction) {
			return;
		}
		if (this.m_alreadyOpenWindowMode) {
			return true;
		}
		const account_gui_config = this.m_interaction.getAccountConfigGUI();
		if (this.isMaximized()) {
			if (account_gui_config.window_open_mode !== "MAXIMIZED") {
				this.restore();
				this.m_alreadyOpenWindowMode = true;
			}
		}
		else {
			if (account_gui_config.window_open_mode === "MAXIMIZED") {
				this.maximize();
				this.m_alreadyOpenWindowMode = true;
			}
		}
	};

	/**
	* The interaction we are viewing has been updated.
	* @param interaction
	* @param callback
	*/
	ClientWindow.prototype.onInteractionUpdated = function(interaction, callback = $.noop)
	{
		const self = this;

		var inflate = function()
		{
			if (self.inflating) {
				callback(false);
				return;
			}
			self.inflating = true;
			self.m_interaction = interaction;

			self.m_interaction.inflate(function()
			{
				if (!self.m_interaction) {
					self.inflating = false;
					callback(false);
					return;
				}
				var $body = $("body");
				$body.removeClass("interactionAbandoned interactionArchived interactionQueued interactionLocked interactionWaiting");
				if (self.m_interaction.isFinished()) {
					$body.addClass("interactionArchived");
				}
				else if (self.m_interaction.isQueued()) {
					$body.addClass("interactionQueued");
				}
				else if (self.m_interaction.isWaiting()) {
					$body.addClass("interactionWaiting");
				}
				else if (self.m_interaction.isLocked()) {
					$body.addClass("interactionLocked");
				}
				else if (self.m_interaction.isAbandoned()) {
					$body.addClass("interactionAbandoned");
				}
				else if (self.m_interaction.isWrappedUp()) {
					self.updateLayout("ClientWindow.onInteractionUpdated - isWrappedUp");
				}
				else if (self.m_interaction.interactionData && self.m_interaction.interactionData.internal_consultation_flag) {
					self.m_interaction.setView(self);
				}
				
				if (self.m_interaction.isUpdateTransferBot()) {
					self.showScreenReadOnly(false);
				}

				const isOutgoing = self.m_interaction.isOutgoing();
				if (!self.m_interaction.isBeingComposed()) {
					self.disableCampaignAccountsControls();
					/*
						Para el caso de las salientes, cuando se han creado y han dejado de ser pending (ej: telefonia),
						Se debe actualizar en la sidebar el historico de interacciones para el contacto (si es que tiene un contacto definido).
						A esto lo realiza una unica vez, y para eso cachea el contacto en la interaccion.
					*/
					if (isOutgoing) {
						if (self.m_interaction.contact && (self.m_interaction.contact !== self.m_interaction.__cached_contact_id)) {
							self.m_interaction.__cached_contact_id = self.m_interaction.contact;
							inConcert.app.messagebus.publish('/application/contacts/contact_updated', {Id: self.m_interaction.contact});
						}
					}
				}


				self.checkTicketInfo();
				self.showLabelsAndDisposition();
				self.updateNavigationButtons();
				self.updateToolbarButtons();
				self.updateFirstOutgoingEvent();
				self.updateBodyClasses();
				self.onNewEvents();
				self.onUpdateSidebarInformation();
				self.events_count = self.m_interaction.getEventCount();

				const isInteractionCallAllegro = self.m_interaction.isAllegro();

				self.enableIntegrations(self.m_interaction,
					false, /* Forzar*/
					function( shown_integration, integration )
					{
						if ( !shown_integration || !integration ) {
							return;
						}
						if (isOutgoing && !isInteractionCallAllegro) {
							// Visualizamos los Tabs de integraciones, cuando es una saliente (en modo PENDING NO se visualizan, hasta que se crea).
							// Si es una interaccion saliente de CALL de Allegro, el updateLayout, se realiza en la clase "PhoneClient" "phone_client.js" al momento de crear la outgoing.
							self.updateLayout("ClientWindow.onInteractionUpdated - enableIntegrations");
						}
					}
				);

				if (self.m_interaction.isWrappedUp()) {
					self.change_client_state("wrapping_up");
				}
				else {
					$body.removeClass("interactionEnded");
				}
				if (self.clientUpdate) {
					self.clientUpdate();
				}
				self.inflating = false;
				callback(true);
			});
		};

		inflate();
	};

	/**
	* If the client window is bound to an interaction whose account or campaign are not
	* operational, render the client unusable.
	*/
	ClientWindow.prototype.validateCampaignAndAccountOperational = function() {
		// internal interactions are skipped because they aren t bound to any account
		if (this.m_interaction.isInternal() || this.m_interaction.isFinished()) {
			return;
		}

		// si la campana se ha eliminado, deja el cliente inutilizable
		var campaign_id = this.m_interaction.getCampaign() || "";
		var campaign_obj = inConcert.app.getCampaign( campaign_id );
		if ( campaign_obj === null || campaign_obj === undefined ) {
			this.clientUnusable();
			return;
		}

		var campaignIsOperational = this.campaignIsOperational();
		if (!campaignIsOperational.success) {
			$('#cboAccounts').prop('disabled', true);
			this.hideConflictAccountIcon();
			this.showConflictCampaignIcon(campaignIsOperational.errorMessage);
			return;
		}
		this.hideConflictCampaignIcon();
		var accountIsOperational = this.m_interaction.accountIsOperational();
		if (!accountIsOperational.success) {
			this.showConflictAccountIcon(accountIsOperational.errorMessage);
		}
		else {
			this.hideConflictAccountIcon();
		}
	};

	ClientWindow.prototype.canExecuteInteractionAction = function()
	{
		var result = this.campaignIsOperational();
		if (!result.success) {
			this.displayConflictCampaignMessage(false, result.errorMessage);
			return false;
		}
		if (!this.m_interaction.accountIsOperational().success) {
			this.displayConflictAccountMessage();
			return false;
		}
		return true;
	};

	ClientWindow.prototype.hideConflictAccountIcon = function() {
		$(".icoAccountConflict").removeClass("showConflict").inconcert_tooltip({detach: true});
	};

	ClientWindow.prototype.onCampaignAccountUpdated = function() {
		if (!this.m_interaction) {
			return;
		}
		if (this.m_interaction.isBeingComposed()) {
			var actualAvailableCampaigns = this.m_interaction.getAllowedCampaigns(inConcert.app.getCampaigns());
			var campaignDiffereces = inConcert.Utils.diffObjects(this.cachedAvailablesCampaigns, actualAvailableCampaigns);
			if (campaignDiffereces) {
				this.addCampaignsToList($("#cboCampaigns").val(), $("#cboAccounts").val());
			}
		}
		else {
			this.validateCampaignAndAccountOperational();
			// actualiza los botones de la toolbar correspondiente a acciones sobre la interaccion (finish, take, etc)
			this.updateInteractionToolbarButtons();
		}
	};

	ClientWindow.prototype.afterRenderRelatedMarkup = function() {
		return;
	};

	ClientWindow.prototype.getInteractionRelatedMarkup = function( properties )
	{
		// los unread se tienen que validar contra el id posta ( como se encuentran ordenados en el server ) ya que en el cliente los eventos son reordenados x fecha de creacion.
		const readClass = this.m_interaction.isRead( properties.original_index ) ? "relatedInteractionRead" : "relatedInteractionUnread";
		return $("".concat(
			'<div unread_event="true" class="relatedInteraction ', readClass, '">',
				'<div class="interactionAvatar">',
					'<object data="', properties.avatarURI, '" type="image/png" width="36px" height="36px">',
						'<img src="', this.getGenericAvatar(), '" width="48px" height="36px"/>',
					'</object>',
				'</div>',
				'<div class="interactionInfoContainer">',
					"<label class='interactionContactName'>", properties.fromName, "</label>",
					"<label class='interactionDate'>", this.m_interaction.getLocaleDateString( properties.sorted_index ), "</label>",
				'</div>',
				'<a unread_event_action="true" href="javascript:void(0);" class="interactionButton">',
					'<i class="buttons-icons-chevron-circle-right-solid"></i>',
				'</a>',
			'</div>')
		);
	};

	ClientWindow.prototype.getCaptionRelatedButton = function()
	{
		return "<%=_T('All Related')%>";
	};

	ClientWindow.prototype.closeRelated = function(event_index, callback = $.noop)
	{
		if (event_index !== undefined) {
			this.m_currentEvent = event_index;
			this.readAndOpenEvent(undefined, callback);
		}
		else {
			if (this.m_currentEvent !== undefined && this.m_currentEvent > -1) {
				this.readAndOpenEvent(undefined, callback);
			}
			else {
				this.navigateToBestEventByChannel(callback);
			}
		}
		$("#butCloseRelated").hide();
		$("#butShowRelated")
			.attr("value", this.getCaptionRelatedButton())
			.show();
		$("#relatedInboxContainer").hide();
		$("body").removeClass("withRelatedEvents");

		if (this.sidebar) {
			this.sidebar.toggle_popup_tabs_button();
		}
	};

	ClientWindow.prototype.is_visible_related_events = function() {
		var is_visible = $("#relatedInboxContainer").is(":visible");
		return is_visible;
	};
	
	ClientWindow.prototype.getGenericAvatar = function() {
		return "/inconcert/apps/agent/images/genericAvatar.png";
	};

	ClientWindow.prototype.getUserDataFromEvent = function(currentEvent, sorted_index, members){
		var fromName = "", fromAccount = "", avatarUri;
		var it = this.m_interaction;
		avatarUri = it.getEventUserAvatarURI(currentEvent) || this.getGenericAvatar();

		//if is a response from someone in the callcenter use it s avatar and name if available
		if (it.isInternalMessage(sorted_index)) {
			var sourceId = this.m_interaction.getSource(sorted_index);
			if ( typeof currentEvent.creator === "string" && currentEvent.creator.toLowerCase() === "automatic") {
				fromName = currentEvent.name;
				fromAccount = currentEvent.from;
				// if the interaction have a contact associated, set the proper info to the client window
				if (it.contact && it.contact.length > 0 && it.contactAddress){
					// only if we are working with the same event address
					if (it.contactAddress && it.contactAddress === fromAccount){
						fromName = it.interactionData.contactName;
						fromAccount = it.contactAddress;
					}
				}
			}
			else {
				var parsed_id = inConcert.app.separate_vcc( sourceId );
				if ( parsed_id ) {
					var member_without_vcc = parsed_id.id;
					if (member_without_vcc === inConcert.app.getLoggedUserId()) {
						fromName = inConcert.app.getLoggedUserName();
						fromAccount = inConcert.app.getLoggedUserId();
					}
					else {
						var member = members.find(member_without_vcc);
						if (member) {
							fromName = member.Name;
							fromAccount = member_without_vcc;
						}
					}
				}
			}
		}
		else {
			fromName = currentEvent.name;
			fromAccount = currentEvent.from;
		}
		if (!fromName) {
			fromName = it.getUserName(currentEvent);
		}
		return {
			avatarUri   : avatarUri,
			fromName    : fromName,
			fromAccount : fromAccount
		};
	};

	ClientWindow.prototype.getRelatedEvents = function() {
		return this.m_interaction.getEventList();
	};

	ClientWindow.prototype.fillRelatedEvents = function( events, members, $list )
	{
		const self = this;
		// Append markup for each event.
		events.forEach( function( current_event )
		{
			const original_index = current_event.__index;
			if ( typeof original_index === undefined ) {
				throw "ClientWindow.prototype.fillRelatedEvents - current_event.__index is required";
			}
			// a partir del indice original del evento (como esta en el server), obtengo el indice del evento en la lista ordenada
			const sorted_index = self.m_interaction.get_sorted_index_by_original_index( original_index );
			if ( sorted_index === false ) {
				throw "ClientWindow.prototype.fillRelatedEvents - sorted_index is required";
			}

			const userData = self.getUserDataFromEvent( current_event, sorted_index, members );

			const $relatedEvent = self.getInteractionRelatedMarkup({
				sorted_index   : sorted_index,
				original_index : original_index, 
				avatarURI      : userData.avatarUri,
				fromName       : userData.fromName,
				fromAccount    : userData.fromAccount,
				event          : current_event
			});

			if ( $relatedEvent && $relatedEvent.length > 0 ) {
				// acta tengo asignar el valor del evento original, ya que cuando da doble click sobre la interaccion (no leida) tiene que marcar la misma contra los readers_v2, con el indice posta del array.
				$relatedEvent.data("index", original_index);
				$relatedEvent.find("[unread_event_action]").data("index", original_index);
				// first element is the oldest, but we want the newest firts on the list,
				// so we use prependTo instead of appendTo.
				$relatedEvent.prependTo( $list );
			}
		});
	};

	ClientWindow.prototype.toggleRelatedEvents = function(is_initialized)
	{
		const self = this;
		const $container = $("#relatedInboxContainer");

		if ( $container.length === 0 ) {
			return;
		}

		let force_hide_include_tab_btn = false;
		if ( this.is_visible_related_events() ) {
			this.closeRelated();
		}
		else if (is_initialized) {
			this.disableRelatedSections();
			if ( this.canShowRelatedSections() ) {
				$("#butShowRelated")
					.attr("value", this.getCaptionRelatedButton())
					.show();
			}
		}
		else {
			force_hide_include_tab_btn = true;
			$("#butCloseRelated").show();
			$("#butShowRelated").hide();
			this.loading(true);
			var $list = $("#relatedInteractionsList");
			$list.detach();
			$list.children().remove();

			inConcert.app.getTeamMembers(
				function(members) {
					if ( !members ) {
						self.loading(false);
						return;
					}
					// Get event list sorted by date.
					var events = self.getRelatedEvents();
					self.fillRelatedEvents( events, members, $list );

					$list.appendTo( $container );
					$container.show();
					self.afterRenderRelatedMarkup();
					self.loading(false);
				},
				true // from cache
			);
		}

		if (this.sidebar) {
			this.sidebar.toggle_popup_tabs_button(force_hide_include_tab_btn);
		}
	};

	ClientWindow.prototype.isEditable = function(custom_verify /*optional*/)
	{
		var verify = {
			is_being_composed: true,
			is_finished      : true,
			is_monitor       : true,
			is_queued        : true,
			is_waiting       : true
		};
		$.extend(true, verify, custom_verify);

		// Valida que la interaccion exista.
		if (!this.m_interaction) {
			return false;
		}
		// Valida que la interaccion no se este componiendo. Que no se encuentre en los estados "state=PENDING" o "state=CREATED".
		if (verify.is_being_composed && this.m_interaction.isBeingComposed()) {
			return false;
		}
		// Valida que la interaccion no este finalizada. Que no se encuentre en los estados "state=FINISHED" o "state=CANCELED", o que no sea SPAM o se haya descartado.
		if (verify.is_finished && this.m_interaction.isFinished()) {
			return false;
		}
		// Valida que la ventana no se haya abierto desde la app supervisor y se este monitoreando la misma.
		if (verify.is_monitor && this.is_monitor) {
			return false;
		}
		// Valida que la interaccion no se encuentre encolada. Que no se encuentre en los estados "state=QUEUED" o "state=TRANSFER QUEUED".
		if (verify.is_queued && this.m_interaction.isQueued()) {
			return false;
		}
		// Valida que la interaccion no se encuentre en espera. Que no se encuentre en el estado "state=WAITING".
		if (verify.is_waiting && this.m_interaction.isWaiting()) {
			return false;
		}

		// es editable la interaccion
		return true;
	};

	ClientWindow.prototype.readAndOpenEvent = function(event_index, callback = $.noop)
	{
		event_index = event_index || this.m_currentEvent;
		if (!this.m_interaction.isRead(event_index) && this.isEditable()) {
			this.m_interaction.setRead(true, event_index);
			window.taskbar.onLocalInteractionUpdate(window.interaction);
		}
		this.openCurrentEvent(callback);
	};

	ClientWindow.prototype.openCurrentEvent = function(callback = $.noop)
	{
		callback(true);
	};

	ClientWindow.prototype.showNavigation = $.noop;
	ClientWindow.prototype.hideNavigation = $.noop;

	/**
	* Related sections. Explicar que son.
	*/
	ClientWindow.prototype.disableRelatedSections = function() {
		if ( this.is_visible_related_events() ) {
			return;
		}
		$("#butShowRelated").hide();
		$("#butCloseRelated").hide();
		$("#relatedInboxContainer").hide();
		this.hideNavigation();
	};

	ClientWindow.prototype.canShowRelatedSections = function()
	{
		return this.m_interaction.getEventCount() > 1;
	};

	// TO-DO: This function does a lot more than the name says,
	// why is counting events and deciding this? It should ONLY enable stuff as the name says.
	ClientWindow.prototype.enableRelatedSections = function()
	{
		if ( this.is_visible_related_events() ) {
			return;
		}
		if ( this.canShowRelatedSections() ) {
			$("#butShowRelated")
				.attr("value", this.getCaptionRelatedButton())
				.show();
			this.showNavigation();
		}
		else {
			this.disableRelatedSections();
		}
	};

	ClientWindow.prototype.openParentClientWindow = function( close_current_window, callback )
	{
		callback = callback || $.noop;

		var it = this.m_interaction;
		if ( !it.isBeingComposed() ) {
			return callback();
		}
		it.hasContactFormInteractionParent( function( is_contactform_parent_it )
		{
			if ( !is_contactform_parent_it ) {
				return callback();
			}
			it.getParentInteraction( function( parent_it ) {
				if ( !parent_it ) {
					return callback();
				}
				if ( close_current_window ) {
					inConcert.app.getInteractionManager().cancelInteraction( it ); // cuando se abre el parent interaction se cierra y cancela la interaccion actual
				}
				if ( parent_it ) {
					inConcert.app.getMainWindow().getPanel("AssignedInteractions").openInteractionClient( parent_it );
				}
				callback();
			});
		});
	};

	ClientWindow.prototype.__setExternalMessage = function(message = "")
	{
		if (!this.m_interaction) {
			return;
		}
		this.appendToSendMessageText(message);
		this.focusSendMessageText();
	};

	ClientWindow.prototype.setQnAMakerAnswer = function(answer)
	{
		this.__setExternalMessage(answer.answer);
	};

	/*
		Por defecto todos los canales soportoan los Canneds messages en modo "texto sin formato"
	*/
	ClientWindow.prototype.setCannedMessage = function(canned) {
		this.__setExternalMessage(canned.plain_message);
	};

	ClientWindow.prototype.setMessageInSideBar = function(message) {
		if (!this.sidebar) {
			return;
		}
		this.sidebar.set_message(message);
	};

	ClientWindow.prototype.hideSideBar = function()
	{
		if (this.hasConfiguredLayout()) {
			return;
		}
	};

	ClientWindow.prototype.removeSideBar = function()
	{
	};

	/*
		Valida el cliente tiene configurado GoldenLayout
	*/
	ClientWindow.prototype.hasConfiguredLayout = function()
	{
		const layout_config = this.getLayoutConfig();
		if (layout_config && layout_config.enabled) {
			return true;
		}
		return false;
	};

	/*
		Recupera informacion del GoldenLayout para el cliente.
		Si algun cliente especifico necesita alguna config especial, deberia implementar esta funcion.
	*/
	ClientWindow.prototype.getLayoutConfig = function()
	{
		let enabled = true;
		if (typeof(GoldenLayout) === "undefined") {
			enabled = false;
		}
		const layout_config = {
			enabled       : enabled,
			client_main_id: "mainContainer"
		};
		return layout_config;
	};

	ClientWindow.prototype.onSidebarIntialisedLayout = function()
	{
		const myInt = this.m_interaction || {};
		console.debug('ClientWindow.onSidebarIntialisedLayout --->  CALLBACK INVOKED -> interaction.id: ' + myInt.id + ", interaction.state: " +  myInt.state);
		const self = this;
		// Dejo de visualizar el waiting del frm
		this.loading(false, undefined, function () {
			self.showScreenReadOnlyAfterLoading();
			self.setEventsHandlers();
			self.openWindowMode();
			self.validateAutomaticCloseWindow();
			self.onWindowIsReady(myInt);
		});
	};

	ClientWindow.prototype.onWindowIsReady = $.noop;

	ClientWindow.prototype.initializeSideBar = function()
	{
		const self = this;

		var fatalError = function(messageErr) {
			// Dejo de visualizar el waiting del frm
			self.loading(false);
			self.displayStatusMessage(messageErr, "Error", true);
			console.error(messageErr);
		};

		if (!this.m_interaction) {
			const messageErr = "<%=_T('Unable to initialize sidebar. Conversation not found.')%>";
			fatalError(messageErr);
			return false;
		}

		// Internal interactions have no sidebar
		if (this.m_interaction.isInternal()) {
			let $screenWaiting = $("#screenWaiting");
			$screenWaiting.hide();
			if(this.showMinimizeButton){
				this.showMinimizeButton();
			}
			this.setDraggable(true);
			//Campaign hidden on internal calls
			$(".clientContainerTop").hide();
			return false;
		}

		// This check is to know if the Client Window
		// included the client sidebar script.
		if (!inConcert.ClientSidebar) {
			const messageErr = "<%=_T('Unable to initialize sidebar. Sidebar no se ha instanciado')%>";
			fatalError(messageErr);
			return false;
		}

		this.m_interaction.interactionData = this.m_interaction.interactionData || {};
		var data             = this.m_interaction.interactionData.data || {};
		var contact_addresses = this.m_interaction.interactionData.contactAddresses;
		var client_id        = data.client_id  || "";
		var login            = this.m_interaction.getReportFields( "login"       );
		var pre_survey       = this.m_interaction.getReportFields( "pre_survey"  );
		var params           = this.m_interaction.getReportFields( "params"      );
		var post_survey      = this.m_interaction.getReportFields( "post_survey" );
		var batch            = this.m_interaction.getReportFields( "batch_agent_data" );
		var process_flow     = this.m_interaction.getReportFields( "process_flow_data" );
		var token            = data.token;
		var account_id       = this.m_interaction.account;
		var campaign_id      = this.m_interaction.campaign;

		// if we have a sidebar already initialized,
		// just update information.
		if ( this.sidebar ) {
			var changed_campaign = this.sidebar.m_campaign_id !== campaign_id;
			var changed_account  = this.sidebar.m_account_id !== account_id;
			this.sidebar.update_information(
				campaign_id       ,
				account_id        ,
				client_id         ,
				contact_addresses  ,
				login             ,
				pre_survey        ,
				params            ,
				post_survey       ,
				batch             ,
				process_flow      ,
				token             ,
				this.m_interaction,
				this.getLayoutConfig()

			);
			if (changed_campaign || changed_account) {
				// Esto sucede unicamente cuando estamos creando una interaccion saliente.
				this.sidebar.on_changed_campaign_account(campaign_id, account_id);
			}
			// ok
			return true;
		}

		this.sidebar = new inConcert.ClientSidebar(
		{
			owner   : this,
			from_app: this.from_app,
			actions : this.from_app === "interaction" ? inConcert.attention_actions : inConcert.supervisor_actions,
			onSidebarShown : function()
			{
 				// Callback on Sidebar Shown
				var is_maximized = self.isMaximized();
				var is_minimized = self.isMinimized();
				if ( !is_maximized && !is_minimized ) {
					var size = self.getWindowSizeOnSidebarShown();
					self.setHorizontalResizingLimits( size.width );
					self.setWidth( size.width );
				}
			},
			onSidebarHidden : function()
			{
			// Callback on Sidebar Hidden
				var is_maximized = self.isMaximized();
				var is_minimized = self.isMinimized();
				if ( !is_maximized && !is_minimized ) {
					var size = self.getWindowSizeOnSidebarHidden();
					self.setHorizontalResizingLimits( size.width );
					self.setWidth( size.width );
				}
			},
			onSidebarChangeTab: function(/*tab_id*/)
			{
				self.updateSideBar();
			},
			onSidebarSetCannedMessage: function(canned)
			{
				canned = canned || {};
				self.setCannedMessage(canned);
			},
			onSidebarSetQnaAMakerAnswer: function(answer)
			{
				answer = answer || {};
				self.setQnAMakerAnswer(answer);
			},
			onSidebarIntialisedLayout: function()
			{
				self.onSidebarIntialisedLayout();
			},
			onShowMinimizeButton: function(){
				if(self.showMinimizeButton){
					self.showMinimizeButton();
				}
			},
			campaign_id     : campaign_id       ,
			account_id      : account_id        ,
			client_id       : client_id         ,
			contact_addresses: contact_addresses  ,
			login           : login             ,
			pre_survey      : pre_survey        ,
			params          : params            ,
			post_survey     : post_survey       ,
			batch           : batch             ,
			process_flow    : process_flow      ,
			token           : token             ,
			interaction     : this.m_interaction,
			layout_config   : this.getLayoutConfig()
		});

		// button initial state is visible but disabled
		// $show_sidebar_btn.prop("disabled", true).show();

		// initialize the sidebar, this is done to pre-fetch the info
		// before showing anything to the user
		this.sidebar.initialize(campaign_id, client_id, function(ok, error_message)
		{
			if ( error_message ){
				inConcert.app.m_mainWindow.showBannerMessage({ message: error_message, messageType: "Error", timeout: 5000 }, null, true);
			}

			if (!self.m_interaction) {
				return;
			}

			self.m_interaction.hasContactFormInteractionParent( function( is_contactform_parent_it ){
				if ( !is_contactform_parent_it ) {
					return;
				}
			});

			self.onInitializedSidebar();
		});

		// ok
		return true;
	};

	ClientWindow.prototype.onInitializedSidebar = function() {
		// Enable CCI popup integration for live conversations
		if (!this.m_is_intializedSidebar) {
			this.m_is_intializedSidebar = true;
			this.enableIntegrations(this.m_interaction);
			this.updateSideBar();
		}
	};

	ClientWindow.prototype.getAppsIntegrationsDOMIds = function()
	{
		let domIds = [];
		if (this.m_cci_application) {
			let cciDomIds = this.m_cci_application.getDOMIds();
			domIds.push(cciDomIds);
		}
		if (this.m_helpdesk_integration_app) {
			let cciDomIds = this.m_helpdesk_integration_app.getDOMIds();
			domIds.push(cciDomIds);
		}
		return domIds;
	};

	ClientWindow.prototype.enableIntegrations = function(interaction, force, callback = $.noop)
	{
		const self = this;
		let has_invoked_callback = false;
		this.enableCCI(interaction, force, function(success, integrationData, err) {
			if (success) {
				callback(success, integrationData, err);
				has_invoked_callback = true;
			}
			self.enableHelpDeskIntegration(interaction, force, function(success, integrationData, err) {
				if (!has_invoked_callback) {
					callback(success, integrationData, err);
				}
				//self.updateLayout();
			});
		});
	};

	ClientWindow.prototype.enableCCI = function(interaction, force, callback = $.noop)
	{
		if ( this.m_cci_application && !force ) {
			callback( false );
			return;
		}
		interaction = interaction || this.m_interaction;
		if (!interaction || !interaction.hasCCIntegration()) {
			callback(false);
			return;
		}
 		if ( this.is_monitor ) {
			callback( false );
			return;
		}
		if ( !interaction.belongsToMe() ) {
			callback( false );
			return;
		}

		// Will setup CCI integration if configured.
		var enable_integration = function (interaction)
		{
			const self = this;

			const must_show_cc_integration = self.sidebar && self.sidebar.can_show_cc_integration();
			if (!must_show_cc_integration) {
				callback( false );
				return this;
			}

			// Bail out if the integration trigger is not 'embed'
			const integrationData = interaction.getIntegrationData();
			if ( !integrationData ) {
				callback( false );
				return this;
			}

			// If no application, there s nothing we need to do
			self.m_cci_application = inConcert.IntegrationManager.newAppIntegration(integrationData.type, integrationData);
			if (!self.m_cci_application) {
				callback(false);
				return this;
			}

			inConcert.logger.action({
				action_name: 'open_application',
				interaction: interaction.id,
				name: integrationData.application
			});

			// the client sidebar MUST be opened before because it holds the container div
			//if ( this.sidebar ) {
				//this.updateLayout();
			//}

			// Prevent this window to be closed interactively (see onWindowClosing)...
			// ...and supply a function to undo that when the popup needs to be closed
			var onClose = function() {
				self.m_cci_application = null;
				self.close();
			};

			if (interaction.contact && interaction.contact.length > 0) {
				inConcert.app.getContactManager().getContact(interaction.contact, function(contact) {
					self.m_cci_application.load(interaction, contact, {
						host_window: window,
						close_handler: onClose,
						on_load: function() {
							if(self.sidebar) {
								self.sidebar.enable_cc_integration();
							}
							callback( true, integrationData );
						},
						on_ready: function(success/*, err*/) {
							if (!success) {
								if(self.sidebar) {
									self.sidebar.on_error_cc_integration();
								}
								callback( true, integrationData );
							}
						}
					});
				});
			}
			else {
				self.m_cci_application.load(interaction, undefined, {
					host_window: window,
					close_handler: onClose,
					on_load: function() {
						if(self.sidebar) {
							self.sidebar.enable_cc_integration();
						}
						callback( true, integrationData );
					},
					on_ready: function(success/*, err*/) {
						if (!success) {
							if(self.sidebar) {
								self.sidebar.on_error_cc_integration();
							}
							callback( true, integrationData );
						}
					}
				});
			}
		};

		enable_integration.apply(this, [interaction]);
	};

	ClientWindow.prototype.enableHelpDeskIntegration = function(interaction, force, callback = $.noop)
	{
		if ( this.m_helpdesk_integration_app && !force ) {
			callback( false );
			return;
		}
		interaction = interaction || this.m_interaction;
		if (!interaction || !interaction.hasIntegrationWithHelpDesk()) {
			callback(false);
			return;
		}
		if ( this.is_monitor ) {
			callback( false );
			return;
		}
		if ( !interaction.belongsToMe() ) {
			callback( false );
			return;
		}
	
		// Will setup CCI HELPDESK integration if configured.
		var enable_integration = function (interaction)
		{
			const self = this;
	
			const must_show_cc_integration = self.sidebar && self.sidebar.can_show_helpdesk_integration();
			if (!must_show_cc_integration) {
				callback( false );
				return this;
			}
	
			// Bail out if the integration trigger is not 'embed'
			const helpdeskIntegrationData = interaction.getHelpDeskIntegrationData();
			if (!helpdeskIntegrationData || !helpdeskIntegrationData.Config) {
				callback( false );
				return this;
			}
	
			// If no application, there s nothing we need to do
			self.m_helpdesk_integration_app = inConcert.IntegrationManager.newAppIntegration("helpdesk_integration", helpdeskIntegrationData.Config);
			if (!self.m_helpdesk_integration_app) {
				callback(false);
				return this;
			}
	
			inConcert.logger.action({
				action_name: 'open_application',
				interaction: interaction.id,
				name: helpdeskIntegrationData.application
			});
	
			// the client sidebar MUST be opened before because it holds the container div
			//if (this.sidebar) {
				//this.updateLayout();
			//}

			// Prevent this window to be closed interactively (see onWindowClosing)...
			// ...and supply a function to undo that when the popup needs to be closed
			var onClose = function() {
				self.m_helpdesk_integration_app = null;
				self.close();
			};
	
			if (interaction.contact && interaction.contact.length > 0) {
				inConcert.app.getContactManager().getContact(interaction.contact, function(contact) {
					self.m_helpdesk_integration_app.load(interaction, contact, {
						host_window: window,
						close_handler: onClose,
						on_load: function() {
							if(self.sidebar) {
								self.sidebar.enable_helpdesk_integration();
							}
							callback( true, helpdeskIntegrationData );
						},
						on_ready: function(success/*, err*/) {
							if (!success) {
								if(self.sidebar) {
									self.sidebar.on_error_helpdesk_integration();
								}
								callback( true, helpdeskIntegrationData );
							}
						}
					});
				});
			}
			else {
				self.m_helpdesk_integration_app.load(interaction, undefined, {
					host_window: window,
					close_handler: onClose,
					on_load: function() {
						if(self.sidebar) {
							self.sidebar.enable_helpdesk_integration();
						}
						callback( true, helpdeskIntegrationData );
					},
					on_ready: function(success/*, err*/) {
						if (!success) {
							if(self.sidebar) {
								self.sidebar.on_error_helpdesk_integration();
							}
							callback( true, helpdeskIntegrationData );
						}
					}
				});
			}
		};
	
		enable_integration.apply(this, [interaction]);
	};

	ClientWindow.prototype.hasCCIntegration = function()
	{
		return this.m_cci_application !== null && this.m_cci_application !== undefined;
	};

	ClientWindow.prototype.hasHelpDeskIntegration = function()
	{
		return this.m_helpdesk_integration_app !== null && this.m_helpdesk_integration_app !== undefined;
	};

	ClientWindow.prototype.hasMasIntegration = function()
	{
		if (this.sidebar) {
			const has_integration = this.sidebar.can_show_mas_integration();
			return has_integration;
		}
		return false;
	};

	ClientWindow.prototype.hasAnyIntegration = function()
	{
		if (this.hasCCIntegration()) {
			return true;
		}
		if (this.hasHelpDeskIntegration()) {
			return true;
		}
		if (this.hasMasIntegration()) {
			return true;
		}
		return false;
	};

	ClientWindow.prototype.getWindowSizeOnSidebarShown  = function()
	{
		/* hay que implementarla en la clase hija segun el cliente */
		var size = { width: null, height: null };
		size.width = DEFAULT_WINDOW_WIDTH_PX;
		return size;
	};

	ClientWindow.prototype.getWindowSizeOnSidebarHidden = function()
	{
		/* hay que implementarla en la clase hija segun el cliente */
		var size = { width: null, height: null };
		size.width = DEFAULT_WINDOW_WIDTH_PX;
		return size;
	};

	ClientWindow.prototype.isVisibleToolbarActionButton = function(action_id) {
		var not_must_validate_custom_action = true;
		return this.m_interaction_gui.is_visible_action(action_id, this.m_interaction, not_must_validate_custom_action);
	};

	ClientWindow.prototype.bindGenericToolbarActionsButtons = function()
	{
		const self = this;

		// Asigna los eventos dinamicos a los botones de la TOOLBAR. Acciones sobre la interaccion (finish, reopen, etc)
		const event_selector = "[btn_action]";
		const $actionsButtonsContainer = this.$getActionsButtonsContainer();
		$actionsButtonsContainer
			.off("click", event_selector)
			.on("click", event_selector, function() {
				const $this = $(this);
				const actionId = $(this).attr("btn_action");
				const action = self.m_interactionToolbarButtons[actionId];
				action.method($this);
			});

		$("#showMoreActions")
			.off("click")
			.on("click", function() {
				self.showMoreActions();
			}
		);
	};

	ClientWindow.prototype.showMoreActions = function()
	{
		if (!this.m_interaction || (this.m_interaction && this.m_interaction.state === "CREATED")) {
			return;
		}
		var cloned_interaction = inConcert.Utils.cloneObject(this.m_interaction);
		var $button = $("#showMoreActions");
		this.m_interaction_gui.show_menu_actions(cloned_interaction, {
			$parent: $button
		});
	};

	ClientWindow.prototype.updateLayout = function(invoked_from = "") {
		if (!this.sidebar) {
			return;
		}
		invoked_from = "ClientWindow.updateLayout " + invoked_from;
		this.sidebar.update_sidebar_tabs(invoked_from);
		this.sidebar.refreshLayout();
	};

	ClientWindow.prototype.toggleToolbBarContextualMenu = function()
	{
		var counter_visible_actions = this.m_interaction_gui.count_visible_actions(this.m_interaction);
		$("#showMoreActions").toggle(counter_visible_actions > 0 && true || false);
	};

	ClientWindow.prototype.$getActionsButtonsContainer = function()
	{
		return $("#interactionToolbarButtons");
	};

	ClientWindow.prototype.updateInteractionToolbarButtons = function()
	{
		const self = this;
		if (!this.m_interaction) {
			return;
		}

		// obtengo el markup de la toolbar tal cual se encuentra en el archivo .html
		const cachedHTML = this.getCachedHTMLInteractionToolbarButtons();
		if (cachedHTML && cachedHTML.length > 0) {
			// Si el mismo existe, lo reenderizo para luego procesarlo.
			const $actionsButtonsContainer = this.$getActionsButtonsContainer();
			$actionsButtonsContainer.html(cachedHTML);
		}

		//console.debug("ClientWindow.prototype.updateInteractionToolbarButtons - toolbar html: " + cachedHTML);
		let i = 0;
		$.each(this.m_interactionToolbarButtons, function(actionId, action)
		{
			i++;
			const $btn = $("#" + action.id);
			//console.debug(i + " - ClientWindow.prototype.updateInteractionToolbarButtons - action.id '" + action.id + "', exist in dom: " + ($btn.length > 0 ? "SI" : "NO"));

			if ($btn.length === 0) {
				return true; // continue
			}

			let is_visible = false;

			if (!action.disabled) {
				const has_permission = action.permission && action.permission();
				//console.debug(i + " - ClientWindow.prototype.updateInteractionToolbarButtons - action.id '" + action.id + "', has_permission: " + has_permission);
				if (has_permission) {
					is_visible = true;
				}
				else {
					const can_shown = !!self.isVisibleToolbarActionButton(actionId);
					//console.debug(i + " - ClientWindow.prototype.updateInteractionToolbarButtons - actionId '" + actionId + "', !!self.isVisibleToolbarActionButton: " + can_shown);
					if (can_shown) {
						is_visible = true;
					}
				}
			}
			// else {
			// 	console.debug(i + " - ClientWindow.prototype.updateInteractionToolbarButtons - action.id '" + action.id + "', action.disabled: " + action.disabled);
			// }

			//console.debug(i + " - ClientWindow.prototype.updateInteractionToolbarButtons - action.id '" + action.id + "', is_visible: " + is_visible);

			if (is_visible) {
				// Si se tiene que visualizar lo muestro.
				if (action.show) {
					action.show($btn);
				}
				$btn.show();
			}
			else {
				// Si no se tiene que visualizar oculto el boton.
				if (action.hide) {
					action.hide($btn);
				}
				$btn.hide();
			}
		});

		this.toggleDragAndDropFiles();
		this.toggleAudioRecord();
		this.toggleTemplateMessages();
		// Muestra u oculta el boton que abre las acciones del menu contextual, dependiendo de si tiene o no items el menu.
		this.toggleToolbBarContextualMenu();
	};

	// adjust panel positions dynamically
	ClientWindow.prototype.adjust_panel_positions = function ()
	{
		const $offset_top = $(".clientOffsetTop");
		if ( $offset_top.length === 0 ) {
			return;
		}
		const $offset_bottom = $(".clientOffsetBottom");
		if ( $offset_bottom.length === 0 ) {
			return;
		}
		const adjust_fn = _.debounce(function (/*mutations*/)
		{
			const height = $offset_top.css("height");
			//console.log("[ClientWindow.adjust_panel_positions] Adjust Clients Window, bottom positions '.clientOffsetTop height' to ->'clientOffsetBottom.top'. : '" + height + "");
			if (height) {
				$offset_bottom.css("top", height);
			}
		}, 10);
		adjust_fn();

		if (!this.m_position_observer) {
			this.m_position_observer = new MutationObserver( adjust_fn );
			this.m_position_observer.observe( $offset_top[0], { attributes: true, childList: true, characterData: true, subtree: true } );
		}
	};

	ClientWindow.prototype.reopenInteraction = function()
	{
		var self = this;

		var app_id = this.from_app;
		var interaction = this.m_interaction;
		this.displayStatusMessage("<%=_T('Reopening...')%>", "Process", true);

		var reopen_callback = function( reopened, messageError )
		{
			var main_window = inConcert.app.getMainWindow();
			if ( reopened ) {
				self.is_archive = false;
				self.is_queued = false;
				self.is_waiting = false;
				self.is_readonly = false;
				self.hideStatusMessage();
				self.displayStatusMessage("<%=_T('Reopened successfully')%>", "Info", false, 3000);
				self.close();
			}
			else {
				//handle error
				self.hideStatusMessage();
				main_window.showBannerMessage({message: messageError, messageType: "Error", timeout: 5000 }, null, true);
				self.displayStatusMessage(messageError, "Error", false, 5000);
			}

			if ( ( app_id || "" ).toLowerCase() === "supervisor" ) {
				var sup_app = main_window.getAppSupervisor();
				if ( sup_app ) {
					var sup_archive = sup_app.get_dashboard().get_archive_manager();
					if ( sup_archive ) {
						sup_archive.on_reopen_conversation( reopened, interaction );
					}
				}
			}
		};

		this.m_interaction.reopen( reopen_callback, app_id );
	};

	ClientWindow.prototype.finishInteraction = function()
	{
		const self = this;

		let message;
		if (this.m_interaction.hasContactConflict()) {
			message = format("<%=_T('This conversation contains a conflict, you must first resolve before finish it')%>", this.m_interaction.campaign);
			this.displayStatusMessage( message, "Error", false, 5000 );
			return;
		}
		if ( !this.m_interaction.hasPermission( "FINISH", this.from_app ) ) {
			message = format("<%=_T('You have not permissions to finish conversations on {0} campaign.')%>", this.m_interaction.campaign);
			this.displayStatusMessage( message, "Error", false, 5000 );
			return;
		}
		const campaign = inConcert.app.getCampaign(this.m_interaction.campaign);
		if (!campaign) {
			throw "Campaign has been deleted";
		}
		const dispositionCodes = campaign.getDispositionCodes();
		if (dispositionCodes ) {
			const dispositionCodesPopup = new inConcert.DispositionCodesPopup( { from_app: this.from_app } );
			dispositionCodesPopup.show({
				interactions:     [this.m_interaction],
				parentNode:       $("#finishInteraction"),
				dispositionCodes: dispositionCodes,
				callbacks: {
					on_finish : function( success, error ) {
						if (success) {
							self.close();
						}
						else {
							self.displayStatusMessage( error, "Error", false, 5000 );
						}
					},
					onOpen: $.noop,
					onClose: $.noop
				}
			});
		}
		else {
			const archive_callback = function(success, successfulNodes, unsuccessfulNodes)
			{
				let error_msg;
				if (!success) {
					error_msg = successfulNodes;
					self.displayStatusMessage( error_msg, "Error", false, 5000 );
				}
				else if (success && $.isArray(successfulNodes) && $.isArray(unsuccessfulNodes)) {

					let unsuccessful_length = unsuccessfulNodes.length;
					if (unsuccessful_length > 0) {
						if (unsuccessful_length === 1) {
							let unsuccessfulNode = unsuccessfulNodes[0];
							switch (unsuccessfulNode.result)
							{
								case "INVALID_REQUEST":
									error_msg =  "<%=_T('Invalid request please try again later')%>";
								break;

								case "FORBIDDEN":
									error_msg = "<%=_T('You have not permissions to finish conversations on campaign.')%>";
								break;

								case "UNABLE_TO_FINISH_INTERACTION_WITH_PENDING_EVENTS":
									let total_pending_events;
									if (unsuccessfulNode.raw_info) {
										total_pending_events = unsuccessfulNode.raw_info.total_pending_events;
									}
									error_msg = format("<%=_T('Unable to finish conversation. Conversation has {0} sending events.')%>", total_pending_events);
								break;

								default:
									error_msg = "<%=_T('Failed finished conversations, please try again later')%>";
							}
						}
						else {
							error_msg = format("<%=_T('{0} conversations failed to finish')%>", unsuccessfulNodes.length);
						}
						self.displayStatusMessage( error_msg, "Error", false, 5000 );
						inConcert.app.getMainWindow().getPanel("AssignedInteractions").updateInteractionNode( self.m_interaction );
					}
					// handle the single finish from supervisor.
					else if (successfulNodes.length === 1) {
						self.close();
					}
				}
			};

			this.m_interaction.archive( undefined, undefined, archive_callback, this.from_app );
		}
	};

	ClientWindow.prototype.hangupInteraction = function()
	{
		this.m_interaction_gui.execute_action("Hangup", this.m_interaction);
	};

	ClientWindow.prototype.cancelInteraction = function()
	{
		this.m_interaction_gui.execute_action("Cancel", this.m_interaction);
		this.close();
	};

	ClientWindow.prototype.takeInteraction = function()
	{
		const self = this;
		const app_id = this.from_app;

		this.displayStatusMessage("<%=_T('Taking...')%>", "Process", true);

		const take_callback = function( taked, messageError )
		{
			const main_window = inConcert.app.getMainWindow();
			if ( taked ) {
				self.is_archive = false;
				self.is_queued = false;
				self.is_waiting = false;
				self.is_readonly = false;
				self.hideStatusMessage();
				self.displayStatusMessage("<%=_T('Taked successfully')%>", "Info", false, 3000);
			}
			else {
				//handle error
				self.hideStatusMessage();
				main_window.showBannerMessage({message: messageError, messageType: "Error", timeout: 5000 }, null, true);
				self.displayStatusMessage(messageError, "Error", false, 5000);
			}
		};

		this.m_interaction.take( take_callback, app_id );
	};

	ClientWindow.prototype.takeBotInteraction = function()
	{
		const self = this;
		const app_id = this.from_app;

		this.displayStatusMessage("<%=_T('Taking...')%>", "Process", true);

		const take_callback = function( taked, messageError )
		{
			const main_window = inConcert.app.getMainWindow();
			if ( taked ) {
				self.is_archive = false;
				self.is_queued = false;
				self.is_waiting = false;
				self.is_readonly = false;
				self.hideStatusMessage();
				self.displayStatusMessage("<%=_T('Taked successfully')%>", "Info", false, 3000);
			}
			else {
				//handle error
				self.hideStatusMessage();
				main_window.showBannerMessage({message: messageError, messageType: "Error", timeout: 5000 }, null, true);
				self.displayStatusMessage(messageError, "Error", false, 5000);
			}
		};

		this.m_interaction.take_bot_interaction( take_callback, app_id );
	};

	ClientWindow.prototype.queueInteraction = function()
	{
		const self = this;
		const app_id = this.from_app;

		this.displayStatusMessage("<%=_T('Queueing...')%>", "Process", true);

		var queue_callback = function( queued, messageError )
		{
			const main_window = inConcert.app.getMainWindow();
			if ( queued ) {
				self.is_archive = false;
				self.is_queued = false;
				self.is_waiting = false;
				self.is_readonly = false;
				self.hideStatusMessage();
				self.displayStatusMessage("<%=_T('Queued successfully')%>", "Info", false, 3000);
			}
			else {
				//handle error
				self.hideStatusMessage();
				main_window.showBannerMessage({message: messageError, messageType: "Error", timeout: 5000 }, null, true);
				self.displayStatusMessage(messageError, "Error", false, 5000);
			}
		};

		this.m_interaction.requeue(queue_callback, app_id);
	};

	ClientWindow.prototype.transferInteraction = function($button)
	{
		const cloned_interaction = inConcert.Utils.cloneObject(this.m_interaction);
		this.m_interaction_gui.show_submenu_actions(cloned_interaction, {
			$parent: $button,
			action_type: 'Transfer',
			align: null // para que no quede alineado a la izquierda, ya que los submenus se alinean siempre a la izquierda.
		});
	};

	ClientWindow.prototype.conferenceInteraction = function($button)
	{
		const cloned_interaction = inConcert.Utils.cloneObject(this.m_interaction);
		const params = {
			positionNode: $button
		};
		this.m_interaction_gui.execute_action("Conference", cloned_interaction, params);
	};

	ClientWindow.prototype.scheduleInteraction = function($button)
	{
		const cloned_interaction = inConcert.Utils.cloneObject(this.m_interaction);
		const params = {
			positionNode: $button
		};
		this.m_interaction_gui.execute_action("Schedule", cloned_interaction, params);
	};

	ClientWindow.prototype.updateToolbarButtons = function()
	{
		if (!this.m_interaction) {
			return false;
		}
		// initialize client sidebar
		this.initializeSideBar();
		var interactionPendingOrFinished = this.m_interaction.isBeingComposed() || this.m_interaction.isFinished(); //|| this.is_queued

		if (!interactionPendingOrFinished) {
			var comments = this.m_interaction.getComments();
			var totalComments = comments ? comments.length : 0;
			var $commentsNode = $('.interactionComments');
			if (totalComments > 0) {
				$commentsNode.css("display", "");
				$commentsNode.find('span').text(totalComments);
				$commentsNode.removeClass("interactionCommentsNew");
				//check if the comments are unread for the current user
				if (this.m_interaction.youHasUnreadComments()) {
					$commentsNode.removeClass("interactionCommentsRead").addClass("interactionCommentsUnread");
				}
				else {
					$commentsNode.removeClass("interactionCommentsUnread").addClass("interactionCommentsRead");
				}
			}
			else {
				$commentsNode.hide();
			}
		}
		// actualiza los botones de la toolbar correspondiente a acciones sobre la interaccion (finish, take, etc)
		this.updateInteractionToolbarButtons();
		// solo para los canales que tengan este tipo de mensajes.
		this.updateReadUnreadButtons();
	};

	ClientWindow.prototype.updateReadUnreadButtons = function()
	{
		let can_show = false;
		if (this.isEditable()) {
			can_show = this.m_interaction.getContactsEvents().length > 0;
		}

		if ( can_show ) {
		//	$("#mark_all_read_unread").show();
			this.set_mark_all_unread_read_button_caption();
		}
	//	else {
	//	//	$("#mark_all_read_unread").hide();
	//	}
	};

	ClientWindow.prototype.updateNavigationButtons = function()
	{
		if ( this.m_interaction.getEventCount() ) {
			this.enableRelatedSections();
		} else {
			this.disableRelatedSections();
		}
	};

	// navigation in events
	ClientWindow.prototype.get_current_event = function() { return this.m_interaction.getEvent( this.m_currentEvent ); };

	ClientWindow.prototype.navigateToEvent = function(index, callback = $.noop) {
		this.m_currentEvent = index;
		this.readAndOpenEvent(undefined, callback);
	};

	ClientWindow.prototype.navigateFirst = function(callback = $.noop) {
		var first_index_event = this.m_interaction.getFirstIndexEvent();
		this.navigateToEvent(first_index_event, callback);
	};

	ClientWindow.prototype.navigateToBestEventByChannel = function(callback = $.noop) {
		this.navigateFirst(callback);
	};

	ClientWindow.prototype.navigateToBestEvent = function(callback, is_initialized)
	{
		callback = callback || $.noop;
		var related_events_count = this.m_interaction.getEventCount();
		if (related_events_count === 1) {
			this.disableRelatedSections();
		}
		var unread_events = this.m_interaction.getUnreadEventCount();
		if ( unread_events >= 1 && related_events_count > 1 ) {
			this.navigateToBestEventByChannel(callback);
			this.toggleRelatedEvents(is_initialized);
		} else if (unread_events === 1) {
			var unread_event = this.m_interaction.getFirstUnreadEventIndex();
			if (!unread_event) {
				unread_event = this.m_interaction.getFirstIndexEvent();
			}
			this.navigateToEvent(unread_event, callback);
		} else {
			this.navigateToBestEventByChannel(callback);
		}
	};

	ClientWindow.prototype.is_supported_language = function( current_language )
	{
		current_language = (current_language || "").toLowerCase();
		var is_supported = false;

		$.each(this.m_languages || {}, function( _, language)
		{
			if ( language.id.toLowerCase() === current_language ) {
				is_supported = true;
				return false; // break
			}
		});

		return is_supported;
	};

	ClientWindow.prototype.get_language = function()
	{
		var self = this;
		var default_language = "es";

		if ( !this.m_interaction ) {
			// aca no deberia pasar nunca..... pero!!!
			return default_language;
		}

		var $language = $("#cboLanguage");
		if ( $language.length > 0 ) {
			if ( $language.find("option").length > 0 ) {
		// 1 si la ventana tiene un combo language y al menos tiene un item en este combo, me quedo con el idioma que se encuentre seleccionado.
				return $language.val();
			}
		}

		var is_supported_language = function( language )
		{
			// valida si el lenguaje que se ha detectado, es soportado por nuestro spell checker, y si no es soportado configura el x defecto.
			if ( self.is_supported_language(language) ) {
				return language;
			}
			return;
		};

		var interaction_language = this.m_interaction.get_language(); // obtengo el lenguaje de la interaccion detectado x el IR
		if ( interaction_language ) {
		// 2 valido si el lenguaje detectado x el IR es soportado x el sistema, y en el caso que si lo retorno.
			if ( is_supported_language(interaction_language) ) {
				return interaction_language;
			}
		}

		// 3 si no entra en ninguno de los casos anteriores, me quedo con el default (es).
		return default_language;
	};

	ClientWindow.prototype.fill_languages = function()
	{
		var $language = $("#cboLanguage");
		if ( $language.length === 0 ) {
			return;
		}
		var interaction_language = (this.get_language() || "").toLowerCase();

		$language.html();
		$.each(this.m_languages, function( _, language)
		{
			var $option = $("<option></option>").attr("value", language.id).html(_T(language.name));
			if ( language.id.toLowerCase() === interaction_language ) {
				$option.attr("selected", "selected");
			}
			$language.append($option);
		});
	};

	ClientWindow.prototype.load_languages = function()
	{
		var self = this;
		this.m_languages = [];
		inConcert.app.loadLanguages(function( success, languages )
		{
			if ( success ) {
				self.m_languages = languages;
				self.fill_languages();
			}
			else {
				self.m_languages = [];
			}
		});
	};

	//triggered to update the view, timers, data and whatsoever
	ClientWindow.prototype.onTimer = function() {
		if (this.m_interaction) {
			var time = this.m_interaction.currentTime();
			$('#elapsedTime').text(time);
		}
	};

	ClientWindow.prototype.clearTimeoutStatusMessage = function() {
		if (this.m_closeStatusMessageTimer) {
			clearTimeout(this.m_closeStatusMessageTimer);
			this.m_closeStatusMessageTimer = null;
		}
	};

	ClientWindow.prototype.hideStatusMessage = function() {
		this.clearTimeoutStatusMessage();
		const $message_container = $("#message_container");
		$message_container.find("[message='true']").text("");
		$message_container.hide();
	};

	/*
		Funcion que crea le markup del mensaje una uncia vez en el <DIV> contenedor id="message_container"; en un siguiente mensaje no lo crea nuevamente sino que lo visualiza.
		Esta funcion tambien contactena mensajes existentes, si es que ya se encuentra en pantalla otro mensaje.
		Para que concatene el mensaje, debe recibir en el parametro "others" un objeto con el atributo "concatenateValue" en true.
		Opcionalmente para concatenar los mensajes existe un atributo messageSeparator (string), el cual nos sirve para utilizar un separador de mensajes. (Por defecto es el valor ";")
		
		Markup Ej:
			<div id="message_container" class="clientContainerMessages" style="">
				<div message_type_container='true' class="clientParticularMessages messageInfo">
					<a class="clientMessagesClose" href="javascript:void(0);" role="close">cerrar</a>
					<div class="clientMessagesText">
						<label message='true' class="clientMessagesTextAux">Por favor, ingrese el correo del receptor antes de enviar</label>
					</div>
				</div>
			</div>
	*/
	ClientWindow.prototype.displayStatusMessage = function(text, type, closeDisabled, timeout_ms, onHide, others = {})
	{
		const self = this;

		this.clearTimeoutStatusMessage();

		let concatenateValue = false;
		if (others.concatenateValue) {
			concatenateValue = true;
		}

		let hideStatusMessage = _.bind(this.hideStatusMessage, this);
		if (typeof(onHide) === "function") {
			hideStatusMessage = _.compose(onHide, hideStatusMessage);
		}

		const $message_container = $("#message_container");
		if ($message_container.length === 0) {
			throw("ClientWindow.prototype.displayStatusMessage - Unable display message. DOM container Id 'message_container' not found.");
		}
		if ($message_container.html().length === 0) {
			let htmlContent = "" +
				"<div message_type_container='true'>" +
					"<a class='clientMessagesClose' href='javascript:void(0);' role='close'></a>" + //<%=_T('close')%>
					"<div class='clientMessagesText'>" +
						"<label message='true' class='clientMessagesTextAux'></label>" +
					"</div>" +
				"</div>";
			$message_container.html(htmlContent);
			const $close = $message_container.find("[role='close']");
			$close.on("click", function(){
				self.clearTimeoutStatusMessage();
				hideStatusMessage();
			});
		}
		else {
			if ($message_container.is(":visible")) {
				if (concatenateValue) {
					let currentMessage = $message_container.find("[message='true']").text();
					if (currentMessage) {
						let messageSeparator = others.messageSeparator || ";";
						text = currentMessage + messageSeparator + " " + text;
					}
				}
			}
		}

		$message_container.find("[message='true']").text(text);
		$message_container.find("[message_type_container='true']")
			.removeClass()
			.addClass(`clientParticularMessages message${type}`);
		$message_container.show();

		if (timeout_ms) {
			this.m_closeStatusMessageTimer = setTimeout(function() {
				hideStatusMessage();
				self.clearTimeoutStatusMessage();
			}, timeout_ms);
		}

		const $close = $message_container.find("[role='close']");
		if (closeDisabled) {
			$close.hide();
		} else {
			$close.show();
		}

		return text;
	};

	ClientWindow.prototype.onContactsClose = function() {
		inConcert.logger.guiAction({action_name : "close_contact_window", window_id: window.id});
	};
	
	ClientWindow.prototype.$attachmentPicker = function() {
		return $("#sendFile");
	};
	
	ClientWindow.prototype.$recordStart = function() {
		return $("#recordAudioStartButton");
	};
	
	ClientWindow.prototype.$recordSend = function() {
		return $("#recordAudioSendButton");
	};
	
	ClientWindow.prototype.$recordCancel = function() {
		return $("#recordAudioCancelButton");
	};
	
	ClientWindow.prototype.$recordContainer = function() {
		return $("#recordAudioContainer");
	};
	
	ClientWindow.prototype.$recordDuration = function() {
		return $("#recordAudioDuration");
	};

	ClientWindow.prototype.getAttachmentDescriptionLimit = function() {
		return 500;
	};

	ClientWindow.prototype.getAttachmentSizeLimitMb = function() {
		return 16;
	};

	ClientWindow.prototype.getAttachmentSizeLimit = function() {
		return this.getAttachmentSizeLimitMb() * 1024 * 1024;
	};

	ClientWindow.prototype.hasAttachment = function() {
		return false;
	};

	ClientWindow.prototype.hasAttachmentDescriptionEmojies = function() {
		return true;
	};

	ClientWindow.prototype.canSendAttachment = function() {
		if (!this.m_interaction) {
			return false;
		}
		if (this.is_monitor) {
			return false;
		}
		if (!this.hasAttachment()) {
			return false;
		}
		if (!this.m_interaction.canSendAttachment()) {
			return false;
		}
		return true;
	};

	ClientWindow.prototype.hasDragAndDropFiles = function() {
		return false;
	};
	
	ClientWindow.prototype.canDragAndDropFiles = function() {
		return this.hasDragAndDropFiles() && this.canSendAttachment();
	};

	ClientWindow.prototype.hasSharedFiles = function() {
		return true;
	};
	
	ClientWindow.prototype.canSendSharedFile = function() {
		return this.hasSharedFiles() && this.canSendAttachment();
	};

	ClientWindow.prototype.hasAudioRecord = function() {
		return false;
	};

	ClientWindow.prototype.canRecordAudio = function() {
		if (!this.m_interaction) {
			return false;
		}
		if (this.is_monitor) {
			return false;
		}
		if (!this.hasAudioRecord()) {
			return false;
		}
		if (!this.m_interaction.canSendAttachment()) {
			return false;
		}
		return true;
	};

	ClientWindow.prototype.getRecordingTimeLimit = function() {
		return 600;
	};

	ClientWindow.prototype.getAttachmentOptions = function() {
		let isWhatsappCloudApiAccount = inConcert.app.getCampaign(this.m_interaction.campaign).getAccount(this.m_interaction.account).integrationType === "ICC_CLOUD_API";
		let options = {
			image : {
				name : "<%=_T('Image')%>",
				order : 1,
				description : true,
				type : JSON.parse(JSON.stringify(inConcert.FileUploaderTypes[isWhatsappCloudApiAccount ? "whatsapp_cloud_image" : "image"])),
				search_files_type_id : isWhatsappCloudApiAccount ? "whatsapp_cloud_image" : "image"
			},
			video : {
				name : "<%=_T('Video')%>",
				order : 2,
				description : true,
				type : JSON.parse(JSON.stringify(inConcert.FileUploaderTypes[isWhatsappCloudApiAccount ? "whatsapp_cloud_video" : "video"])),
				search_files_type_id : isWhatsappCloudApiAccount ? "whatsapp_cloud_video" : "video"
			},
			audio : {
				name : "<%=_T('Audio')%>",
				order : 3,
				description : true,
				type : JSON.parse(JSON.stringify(inConcert.FileUploaderTypes.audio))
			},
			file : {
				name : "<%=_T('File')%>",
				order : 4,
				description : true,
				type : JSON.parse(JSON.stringify(inConcert.FileUploaderTypes[isWhatsappCloudApiAccount ? "whatsapp_cloud_file" : "file"])),
				search_files_type_id : isWhatsappCloudApiAccount ? "whatsapp_cloud_file" : "file"
			}
		};
		$.each(options, function(id, option) {
			option.Id = id;
		});
		return options;
	};

	ClientWindow.prototype.getSortedAttachmentOptions = function() {
		return inConcert.Utils.hash_to_sorted_array(this.getAttachmentOptions(), function(option) {
			return option.order;
		});
	};
	
	ClientWindow.prototype.getAttachmentOption = function(id) {
		let options = this.getAttachmentOptions();
		return options && id && options[id] || null;
	};
	
	ClientWindow.prototype.openAttachmentSelector = function(message, callback = $.noop) {
		let self = this;
		let option = self.getAttachmentOption(message.fileType);
		if (!option) {
			return callback(false, "file type not supported");
		}
		if (message.fileId) {
			let repositoryAPI = inConcert.app.getRepositoryAPI();
			repositoryAPI.getMetadata({ id : message.fileId }, function(status, repositoryFile) {
				if (status && repositoryFile) {
					message.option = option;
					message.repositoryFile = repositoryFile;
					self.__showAttachmentPreview(option, { repositoryFile : repositoryFile, text : message.text }, callback);
				}
				else {
					callback(false, "repository file not found");
				}
			});
		}
		else {
			//open shared files list
			const $parent = null; // lo dejamos como referencia
			self.showAttachmentOptions(option, message, $parent, callback);
		}
	};
	
	ClientWindow.prototype.__makeOptionUploader = function(option, $option, message = null) {
		let self = this;
		let uploader = new inConcert.FileUploaderControl({
			uploader : $option? {
				$control : $option
			} : null,
			type : option.type,
			sizeLimitMB : self.getAttachmentSizeLimitMb()
		});

		uploader.onFileAdded = function(fileData, callback) {
			uploader.__loadFilePreviewImage(fileData, function() {
				self.__showAttachmentPreview(option, { uploader : uploader, fileData : fileData, text : message && message.text });
				callback();
			});
		};

		uploader.onFileAddError = function(file, result, callback) {
			self.displayStatusMessage(result.reason, "Error", false, 3000);
			callback();
		};
		
		return uploader;
	};

	ClientWindow.prototype.__makeAttachmentOption = function(option, message = null) {
		let self = this;

		let $option = $("".concat(
			"<div option='" + option.Id + "'  class='interactionMenuItem " + option.Id + "AttachmentOption'>",
				option.name,
			"</div>"
		));
		self.m_uploaders = self.m_uploaders || {};
		if (self.m_uploaders[option.Id]) {
			self.m_uploaders[option.Id].destroy();
		}
		self.m_uploaders[option.Id] = self.__makeOptionUploader(option, $option, message);
		return $option;
	};
	
	ClientWindow.prototype.__selectAttachmentOption = function(option) {
		this.m_uploaders[option.Id].openUploadPopup();
	};

	ClientWindow.prototype.__showAttachmentOptions = function(selectedOption, message, $parent, callback, popup_main_menu) {
		const self = this;
		callback = callback || $.noop;

		if (!this.canSendAttachment()) {
			this.hideStatusMessage();
			this.displayStatusMessage("<%=_T('You cannot send files.')%>", "Error", false, 3000);
			return;
		}

		// (*) = "menu" principal de "interaction_gui->interaction_actions"
		if (popup_main_menu) {
			//
			// Invoco a la funcion de la clase base, para que no cierre el menu (*) de forma automatica, al dar click fuera de este.
			// Esto sirve para cuando el popup se ha abierto desde el (*).
			// 
			popup_main_menu.setClosedAutomatic(false);
		}

		$parent = $parent || self.$attachmentPicker();

		const menu_id = 'attachment_menu_actions';
		$('#' + menu_id).remove();
		let $menu = $(''.concat(
			'<div id="' + menu_id + '" class="interactionMenuBlind" hidden></div>'
		));

		const attachmentOptions = this.getSortedAttachmentOptions();
		$.each(attachmentOptions, function(_, option) {
			if (option.disabled) {
				return;
			}
			const $option = self.__makeAttachmentOption(option, message);
			$option.click(function() {
				window.setTimeout(function() {
					self.disposeAttachmentOptions();
					// if (popup_main_menu) {
					// 	// Este callback se invoca cuando el menu no se ha cerrado de forma automatica, cuando se ha dado click fuera de este.
					// 	// Al invocar a la funcion "setClosedAutomatic" con el valor en true, estamos dejando nuevamente el menu (*) para que se cierre de forma automatica.
					// 	popup_main_menu.setClosedAutomatic(true);
					// 	popup_main_menu.finalize();
					// }
				}, 50);
			});
			$menu.append($option);
		});

		$('body').append($menu);
		this.attachmentMenu = new inConcert.PopupWindow({ownerId: menu_id});
		this.attachmentMenu.show(
		{
			ownerId       : menu_id,
			eventName     : menu_id,
			canReposition : true,
			container     : $menu,
			parentNode    : $parent,
			align         : "LEFT",
			callbacks: {
				onOpen: function() {
					$('#' + menu_id).show();
					callback(true);
					if (selectedOption) {
						window.setTimeout(function() {
							self.__selectAttachmentOption(selectedOption);
						}, 50);
					}
				},
				onClose: function( onClose ) {
					onClose = onClose || $.noop;
					if (self.attachmentMenu) {
						self.attachmentMenu.hide(function() {
							onClose();
						});
					}
					else {
						onClose();
					}
				}
			}
		});
	};

	ClientWindow.prototype.showAttachmentOptions = function(selectedOption, message, $parent, callback, popup_main_menu) {
		if ( this.attachmentMenu && this.attachmentMenu.isVisible() ) {
			this.disposeAttachmentOptions();
		}
		else {
			this.__showAttachmentOptions(selectedOption, message, $parent, callback, popup_main_menu);
		}
	};

	ClientWindow.prototype.disposeAttachmentOptions = function() {
		let self = this;
		if (this.attachmentMenu) {
			this.attachmentMenu.finalize(function() {
				self.attachmentMenu = null;
			});
		}
	};

	ClientWindow.prototype.hideAttachmentOptions = function() {
		if (this.attachmentMenu) {
			this.attachmentMenu.hide();
		}
	};

	ClientWindow.prototype.__addAttachmentPreviewControl = function(option, data, $popup) {
		//get data
		let imageUrl = null;
		let fileName = null;
		if (data.sharedFile) {
			fileName = data.sharedFile.Name;
			if (option.type.preview) {
				imageUrl = inConcert.app.getRepositoryAPI().getUrl({ id : data.sharedFile.RepositoryId });
			}
		}
		else if (data.repositoryFile) {
			fileName = data.repositoryFile.FileName;
			if (option.type.preview) {
				imageUrl = inConcert.app.getRepositoryAPI().getUrl({ id : data.repositoryFile.PartialId });
			}
		}
		else {
			fileName = data.uploader.getFileName(data.fileData.file);
			if (data.uploader.__hasFilePreviewImage(data.fileData) && data.fileData.blobContent) {
				imageUrl = data.fileData.blobContent;
			}
		}
		//set preview
		let $preview = $("<div class='attachmentPreview'></div>");
		if (imageUrl) {
			$preview.css("background-image", "url(\"" + imageUrl + "\")");
		}
		else {
			$preview.text(fileName);
		}
		$popup.find("[middle_container]").append($preview);
	};

	ClientWindow.prototype.__updateAttachmentCharsCounter = _.throttle(function() {
		let currentText = $("#attachmentText").val();
		let currentChars = currentText && currentText.length || 0;
		let remainingChars = this.getAttachmentDescriptionLimit() - currentChars;
		$("#attachmentDescriptionCounter")
			.toggleClass("commentClientCharacterCounterNegative", remainingChars < 0)
			.text(remainingChars);
		$("#attachmentSend").prop("disabled", remainingChars < 0);
	}, 100);

	ClientWindow.prototype.__addAttachmentPreviewDescriptionControl = function(option, data, $popup) {
		let self = this;
		if (!option.description) {
			return;
		}
		let $text = $("".concat(
			"<div class='attachmentTextContainer'>",
				"<textarea id='attachmentText' class='attachmentText' data-emojiable='true'></textarea>",
			"</div>"
		));
		let $control = $text.find("textarea");
		$control
			.val(data.text || "")
			.change(function() {
				self.__updateAttachmentCharsCounter();
			});
		let $counter = $("<label class='attachmentDescriptionCounter' id='attachmentDescriptionCounter'></label>");
		$popup.find("[middle_container]").append($text);
		$popup.find("[bottom_container]").append($counter);
		if (this.hasAttachmentDescriptionEmojies()) {
			window.emojiPicker.discover();
		}
		self.__updateAttachmentCharsCounter();
	};
	
	ClientWindow.prototype.__hasAttachmentPreview = function(option, data) {
		return true;
	};

	ClientWindow.prototype.__showAttachmentPreview = function(option, data, callback = $.noop) {
		let self = this;
		
		function onConfirm(callback) {
			if (data.sharedFile) {
				self.sendSharedFile(option, data.sharedFile, callback);
			}
			else if (data.repositoryFile) {
				self.sendRepositoryFile(option, data.repositoryFile, callback);
			}
			else {
				self.sendAttachment(option, data.fileData, callback);
			}
		}
		
		if (!self.__hasAttachmentPreview(option, data)) {
			return onConfirm(callback);
		}
		
		let popupId = "windowAttachmentPreviewContainer";
		$('#' + popupId).remove();
		let $popup = $("".concat(
			"<div class='systemPopWindow windowContainer attachmentPreviewPopup " + 
				option.Id + "Preview' id='" + popupId + "'>",
				"<div class='windowTop'>",
					"<div class='windowTitle'>",
						"<%=_T('Preview')%>",
					"</div>",
				"</div>",
				"<div middle_container class='windowMiddle'>",
				"</div>",
				"<div bottom_container class='windowBottom internalPopUpButtons d-flex'>",
					"<input class='pageButtonGeneric buttonColorWhite' id='attachmentCancel' type='button' value='" + "<%=_T('CANCEL')%>" + "'>",
					"<input class='pageButtonGeneric buttonColorBlue ' id='attachmentSend' type='button' value='" + "<%=_T('Send')%>" + "'>",
				"</div>",
			"</div>"
		));
		$popup.find("#attachmentSend").click(function() {
			onConfirm();
		});
		$popup.find("#attachmentCancel").click(function() {
			self.__closeAttachmentPreview();
		});

		$('body').append($popup);
		self.__addAttachmentPreviewControl(option, data, $popup);
		self.__addAttachmentPreviewDescriptionControl(option, data, $popup);
		
		self.attachmentPreview = new inConcert.PopupWindow({ownerId: popupId});
		self.attachmentPreview.show(
		{
			ownerId       : popupId,
			eventName     : popupId,
			canReposition : true,
			closeOnFocusLost : false,
			parentNode    : popupId,
			callbacks: {
				onOpen: function() {
					$('#' + popupId).show();
					callback(true);
				},
				onClose: function( callback ) {
					callback = callback || $.noop;
					if (self.attachmentPreview) {
						self.attachmentPreview.hide(function() {
							callback();
						});
					}
					else {
						callback();
					}
				}
			}
		});
	};

	ClientWindow.prototype.__closeAttachmentPreview = function() {
		let self = this;
		if (this.attachmentPreview) {
			this.attachmentPreview.finalize(function() {
				self.attachmentPreview = null;
			});
		}
	};

	ClientWindow.prototype.sendAttachment = function(option, fileData, callback = $.noop) {
		let self = this;
		let data = self.processAttachmentData({
			type : option.Id,
			description : (option.description && $("#attachmentText").val() || "").trim()
		});
		$("#attachmentSend, #attachmentCancel").prop("disabled", true);
		self.m_interaction.sendAttachment(data, fileData.file, function(status, result) {
			if (!status) {
				self.displayStatusMessage(result, "Error", false, 3000);
			}
			self.__closeAttachmentPreview();
			callback(status, result);
		});
	};

	ClientWindow.prototype.sendRepositoryFile = function(option, repositoryFile, callback = $.noop) {
		let self = this;
		let data = self.processAttachmentData({
			type : option.Id,
			description : (option.description && $("#attachmentText").val() || "").trim()
		});
		let fileData = {
			id : repositoryFile.PartialId,
			name : repositoryFile.FileName,
			type : repositoryFile.ContentType,
			extension : inConcert.Utils.getFileExtension(repositoryFile.FileName)
		};
		$("#attachmentSend, #attachmentCancel").prop("disabled", true);
		self.m_interaction.sendRepositoryFile(data, fileData, function(status, result) {
			if (!status) {
				self.displayStatusMessage(result, "Error", false, 3000);
			}
			self.__closeAttachmentPreview();
			callback(status, result);
		});
	};

	ClientWindow.prototype.sendSharedFile = function(option, sharedFile, callback = $.noop) {
		let self = this;
		let data = self.processAttachmentData({
			fileId : sharedFile.Id,
			description : (option.description && $("#attachmentText").val() || "").trim()
		});
		$("#attachmentSend, #attachmentCancel").prop("disabled", true);
		self.m_interaction.sendSharedFile(data, function(status, result) {
			if (!status) {
				self.displayStatusMessage(result, "Error", false, 3000);
			}
			self.__closeAttachmentPreview();
			callback(status, result);
		});
	};
	
	ClientWindow.prototype.processAttachmentData = function(data) {
		return data;
	};

	ClientWindow.prototype.toggleDragAndDropFiles = function() {
		if (this.canDragAndDropFiles()) {
			this.enableDragAndDropFiles();
		}
		else {
			this.disableDragAndDropFiles();
		}
	};

	ClientWindow.prototype.getDragAndDropControl = function() {
		return null;
	};

	ClientWindow.prototype.enableDragAndDropFiles = function() {
		const self = this;
		this.disableDragAndDropFiles();
		let $control = this.getDragAndDropControl();
		if (!$control) {
			return;
		}

		let options = this.getSortedAttachmentOptions();
		let type = {
			id : "multiple",
			list : options.map(function(current_opt) {
				return current_opt.type;
			}),
			invalid : "<%=_T('Invalid file type')%>"
		};

		this.m_draggerUploader = new inConcert.FileUploaderControl({
			dragger : {
				$control : $control
			},
			type : type,
			sizeLimitMB : self.getAttachmentSizeLimitMb()
		});

		this.m_draggerUploader.onFileAdded = function(fileData, callback) {
			self.m_draggerUploader.__loadFilePreviewImage(fileData, function() {
				let fileType = self.m_draggerUploader.getValidFileTypeFromList(fileData.file, type.list);
				let opt = options.find(function(current_opt) {
					current_opt.type.id = fileType.id;
					return current_opt;
				});
				self.__showAttachmentPreview(opt, { uploader : self.m_draggerUploader, fileData : fileData });
				callback();
			});
		};

		this.m_draggerUploader.onFileAddError = function(file, result, callback) {
			self.displayStatusMessage(result.reason, "Error", false, 3000);
			callback();
		};
	};

	ClientWindow.prototype.disableDragAndDropFiles = function() {
		if (this.m_draggerUploader) {
			this.m_draggerUploader.destroy();
			this.m_draggerUploader = null;
		}
	};

	ClientWindow.prototype.__makeSharedFileOption = function(option) {
		let self = this;
		
		let $option = $("".concat(
			"<div option='" + option.Id + "' class='interactionMenuItem " + option.Id + "AttachmentOption'>",
				option.name,
			"</div>"
		));
		
		$option.click(function() {
			self.__showSharedFileSelector(option);
		});
		
		return $option;
	};

	ClientWindow.prototype.__showSharedFilesOptions = function($parent, popup_main_menu) {
		const self = this;

		if (!this.hasSharedFiles()) {
			this.hideStatusMessage();
			this.displayStatusMessage("<%=_T('You cannot send files.')%>", "Error", false, 3000);
			return;
		}

		if (popup_main_menu) {
			// (*) = "menu" principal de "interaction_gui->interaction_actions"
			//
			// Invoco a la funcion de la clase base, para que no cierre el menu (*) de forma automatica, al dar click fuera de este.
			// Esto sirve para cuando el popup se ha abierto desde el (*).
			// 
			popup_main_menu.setClosedAutomatic(false);
		}

		$parent = $parent || self.$attachmentPicker();
		const menu_id = 'shared_files_menu_actions';
		$('#' + menu_id).remove();
		const $menu = $(''.concat(
			'<div id="' + menu_id + '" class="interactionMenuBlind" hidden></div>'
		));

		const sharedFilesOptions = this.getSortedAttachmentOptions();
		$.each(sharedFilesOptions, function(_, option) {
			if (option.disabled) {
				return;
			}
			const $option = self.__makeSharedFileOption(option);
			$menu.append($option);
		});
		$('body').append($menu);
		this.sharedFilesMenu = new inConcert.PopupWindow({ownerId: menu_id});
		this.sharedFilesMenu.show(
		{
			ownerId       : menu_id,
			eventName     : menu_id,
			canReposition : true,
			container     : $menu,
			parentNode    : $parent,
			align         : "LEFT",
			callbacks: {
				onOpen: function() {
					$('#' + menu_id).show();
				},
				onClose: function( callback ) {
					callback = callback || $.noop;
					if (self.sharedFilesMenu) {
						self.sharedFilesMenu.hide(function() {
							callback();
						});
					}
					else {
						callback();
					}
				}
			}
		});
	};

	ClientWindow.prototype.showSharedFilesOptions = function($parent, popup_main_menu) {
		if ( this.sharedFilesMenu && this.sharedFilesMenu.isVisible() ) {
			this.disposeSharedFilesOptions();
		}
		else {
			this.__showSharedFilesOptions($parent, popup_main_menu);
		}
	};

	ClientWindow.prototype.disposeSharedFilesOptions = function() {
		let self = this;
		if (this.sharedFilesMenu) {
			this.sharedFilesMenu.finalize(function() {
				self.sharedFilesMenu = null;
			});
		}
	};

	ClientWindow.prototype.hideSharedFilesOptions = function() {
		if (this.sharedFilesMenu) {
			this.sharedFilesMenu.hide();
		}
	};
	
	ClientWindow.prototype.__makeSharedFilesList = function(option) {
		let self = this;
		self.m_sharedFilesList = new inConcert.SharedFilesList({
			fileTypes : [{
				id : option.Id,
				name : option.name,
				extensions : option.type.extensions
			}],
			fileTypesFilter : false,
			maxSize : self.getAttachmentSizeLimit(),
			agentVisible : true,
			search_files_type_id : option.search_files_type_id
		});
		let $list = self.m_sharedFilesList.make();
		self.m_sharedFilesList.start();
		return $list;
	};
	
	ClientWindow.prototype.openSharedFileSelector = function(message, callback = $.noop) {
		let self = this;
		let option = self.getAttachmentOption(message.fileType);
		if (!option) {
			return callback(false, "file type not supported");
		}
		if (!message.fileId) {
			//open shared files list
			return self.__showSharedFileSelector(option, message, callback);
		}
		let filesAPI = inConcert.app.getSharedFilesAPI();
		filesAPI.getFile({ fileId : message.fileId }, function(status, sharedFile) {
			if (status) {
				message.option = option;
				message.sharedFile = sharedFile;
				self.__showAttachmentPreview(option, { sharedFile : sharedFile, text : message.text }, callback);
			}
			else {
				callback(false, "shared file not found");
			}
		});
	};

	ClientWindow.prototype.__showSharedFileSelector = function(option, message = null, callback = $.noop) {
		let self = this;
		let popupId = "windowSharedFileSelectorContainer";
		$('#' + popupId).remove();
		let $popup = $("".concat(
			"<div class='systemPopWindow windowContainer sharedFileSelectorPopup " + 
				option.Id + "SharedFileSelector' id='" + popupId + "'>",
				"<div class='windowTop'>",
					"<div class='windowTitle'>",
						"<%=_T('Select a file')%>",
					"</div>",
				"</div>",
				"<div middle_container class='windowMiddle'>",
				"</div>",
				"<div bottom_container class='windowBottom'>",
					"<input class='pageButtonGeneric buttonColorWhite mr-1' id='sharedFileCancel' type='button' value='" + "<%=_T('Cancel')%>" + "'>",
					"<input class='pageButtonGeneric buttonColorBlue' id='sharedFileSelect' type='button' value='" + "<%=_T('Attach')%>" + "'>",
				"</div>",
			"</div>"
		));
		
		$popup.find("#sharedFileSelect").click(function() {
			self.selectSharedFile(option, message);
		});
		$popup.find("#sharedFileCancel").click(function() {
			self.__closeSharedFileSelector();
		});
		$popup.find("[middle_container]").append(self.__makeSharedFilesList(option));

		$('body').append($popup);
		
		self.sharedFileSelectPreview = new inConcert.PopupWindow({ownerId: popupId});
		self.sharedFileSelectPreview.show(
		{
			ownerId       : popupId,
			eventName     : popupId,
			canReposition : true,
			closeOnFocusLost : false,
			parentNode    : popupId,
			callbacks: {
				onOpen: function() {
					$('#' + popupId).show();
					callback(true);
				},
				onClose: function( onClose ) {
					onClose = onClose || $.noop;
					if (self.sharedFileSelectPreview) {
						self.sharedFileSelectPreview.hide(function() {
							onClose();
						});
					}
					else {
						onClose();
					}
				}
			}
		});
	};

	ClientWindow.prototype.__closeSharedFileSelector = function() {
		let self = this;
		if (self.m_sharedFilesList) {
			self.m_sharedFilesList = null;
		}
		if (this.sharedFileSelectPreview) {
			this.sharedFileSelectPreview.finalize(function() {
				self.sharedFileSelectPreview = null;
			});
		}
	};

	ClientWindow.prototype.selectSharedFile = function(option, message = null) {
		let sharedFile = this.m_sharedFilesList.getSelectedFile();
		if (!sharedFile) {
			this.displayStatusMessage("<%=_T('Must select a file')%>", "Error", false, 3000);
			return;
		}
		this.__closeSharedFileSelector();
		this.__showAttachmentPreview(option, { sharedFile : sharedFile, text : message && message.text });
	};

	ClientWindow.prototype.showFilesMenu = function($parent, options) {
		if ( this.m_filesMenu && this.m_filesMenu.isVisible() ) {
			this.disposeFilesMenu();
		}
		else {
			this.__showFilesMenu($parent, options);
		}
	};

	ClientWindow.prototype.__showFilesMenu = function($parent, options = {}) {
		const self = this;

		$parent = $parent || self.$attachmentPicker();
		const menu_id = "files_menu_actions";
		$('#' + menu_id).remove();
		const $menu = $(''.concat(
			'<div id="' + menu_id + '" class="interactionMenuBlind" hidden></div>'
		));

		const sharedFilesOptions = this.getSortedFileOptions();

		$.each(sharedFilesOptions, function(_, option) {
			if (option.disabled) {
				return;
			}

			let $option;
			if (option.$get_menu_item_html) {
				$option = option.$get_menu_item_html(option);
			}
			else {
				$option = $("".concat(
					"<div option='" + option.Id + "' class='interactionMenuItem " + (option.className || "") + "'>",
						option.name,
					"</div>"
				));
			}

			if (option.bind_click) {
				option.bind_click(option, $option, options.popup_main_menu);
			}
			else {
				$option.click(function() {
					option.execute($option, options.popup_main_menu);
				});
			}

			$menu.append($option);
		});

		let align = options.align ? options.align : undefined; // LEFT, or undefined

		$('body').append($menu);

		this.m_filesMenu = new inConcert.PopupWindow({ownerId: menu_id});
		this.m_filesMenu.show(
		{
			ownerId       : menu_id,
			eventName     : menu_id,
			canReposition : true,
			container     : $menu,
			parentNode    : $parent,
			align         : align,
			callbacks: {
				onOpen: function() {
					$('#' + menu_id).show();
				},
				onClose: function(callback, is_auto_close) {
					callback = callback || $.noop;
					var onClose = function() {
						callback();
						if (options.popup_main_menu) {
							// Este callback se invoca cuando el menu no se ha cerrado de forma automatica, cuando se ha dado click fuera de este.
							// Al invocar a la funcion "setClosedAutomatic" con el valor en true, estamos dejando nuevamente el menu para que se cierre de forma automatica.
							options.popup_main_menu.setClosedAutomatic(true);
							if (is_auto_close) {
								options.popup_main_menu.finalize();
							}
						}
					};
					if (self.m_filesMenu) {
						self.m_filesMenu.hide(function() {
							onClose();
						});
					}
					else {
						onClose();
					}
				}
			}
		});
	};

	ClientWindow.prototype.disposeFilesMenu = function() {
		const self = this;
		if (!this.m_filesMenu) {
			return;
		}
		this.m_filesMenu.finalize(function() {
			self.m_filesMenu = null;
		});
	};

	ClientWindow.prototype.getFileOptions = function() {
		const self = this;

		let options = {
			file_system : {
				Id: "file_system",
				name : "<%=_T('From file system')%>",
				order : 1,
				description : true,
				className: "actionAttachFileFromFileSystem",
				execute: function($parent, popup_main_menu) {
					self.showSharedFilesOptions($parent, popup_main_menu);
				}
			},
			device : {
				Id: "device",
				name : "<%=_T('From PC')%>",
				order : 2,
				description : true,
				className: "actionAttachFileFromPC",
				execute: function($parent, popup_main_menu) {
					self.showAttachmentOptions(null, null, $parent, null, popup_main_menu);
				}
			}
		};

		return options;
	};

	ClientWindow.prototype.getSortedFileOptions = function() {
		return inConcert.Utils.hash_to_sorted_array(this.getFileOptions(), function(option) {
			return option.order;
		});
	};

	ClientWindow.prototype.onAudioRecordStarted = $.noop;
	ClientWindow.prototype.onAudioRecordStopped = $.noop;
	
	ClientWindow.prototype.toggleAudioRecord = function() {
		if (this.canRecordAudio()) {
			this.enableAudioRecord();
		}
		else {
			this.disableAudioRecord();
		}
	};

	ClientWindow.prototype.enableAudioRecord = function() {
		const self = this;
		if (this.m_recorder) {
			return;
		}
		this.m_recorder = new InconcertRecorder();
		self.$recordStart()
			.show()
			.off("click")
			.on("click", function() {
				self.__startAudioRecord();
			});
		self.$recordSend()
			.off("click")
			.on("click", function() {
				self.__sendAudioRecord();
			});
		self.$recordCancel()
			.off("click")
			.on("click", function() {
				self.__cancelAudioRecord();
			});
	};

	ClientWindow.prototype.disableAudioRecord = function() {
		if (!this.m_recorder) {
			return;
		}
		this.m_recorder.stop();
		this.m_recorder = null;
		this.$recordStart().off("click").hide();
		this.$recordSend().off("click");
		this.$recordCancel().off("click");
		this.$recordContainer().hide();
	};

	var audioRecordErrors = {
		"HTTPS" : "<%=_T('HTTPS is required to record audio.')%>",
		"IE" : "<%=_T('Internet Explorer is not supported to record audio.')%>"
	};

	ClientWindow.prototype.__startAudioRecord = function() {
		const self = this;
		const error = self.m_recorder.validate();
		if (error) {
			self.displayStatusMessage(audioRecordErrors[error] || error, "Error", false, 3000);
			return;
		}
		self.$recordStart().prop("disabled", true);
		self.m_recorder.start(function(status, error) {
			self.$recordStart().prop("disabled", false);
			if (!status) {
				self.displayStatusMessage(format("<%=_T('Error starting audio record: {0}')%>", error), 
					"Error", false, 3000);
				return;
			}
			self.m_recordingTime = 0;
			self.$recordDuration().text("00:00");
			self.$recordStart().hide();
			self.$recordContainer().show();
			self.toggleSendMessage(false);
			self.m_recordingInterval = window.setInterval(function() {
				self.__recordingInterval();
			}, 1000);
			self.onAudioRecordStarted();
		});
	};

	ClientWindow.prototype.__recordingInterval = function() {
		var self = this;
		if (!self.m_recorder) {
			window.clearInterval(self.m_recordingInterval);
			return;
		}
		self.m_recordingTime++;
		var limit = self.getRecordingTimeLimit();
		if (self.m_recordingTime > limit) {
			self.displayStatusMessage(format("<%=_T('Audio record must last less than {0} seconds.')%>", limit), 
				"Error", false, 3000);
			self.__cancelAudioRecord();
			return;
		}
		var text = inConcert.Utils.zeropad(Math.floor(self.m_recordingTime / 60)) + ":" + 
				inConcert.Utils.zeropad(self.m_recordingTime % 60);
		self.$recordDuration().text(text);
	};

	ClientWindow.prototype.__stopAudioRecord = function(callback = $.noop) {
		const self = this;
		window.clearInterval(this.m_recordingInterval);
		self.$recordSend().prop("disabled", true);
		self.$recordCancel().prop("disabled", true);
		this.m_recorder.stop(function(status, error) {
			self.$recordSend().prop("disabled", false);
			self.$recordCancel().prop("disabled", false);
			if (!status) {
				self.displayStatusMessage(format("<%=_T('Error stopping audio record: {0}')%>", error), "Error", false, 3000);
				return callback(false);
			}
			self.$recordStart().show();
			self.$recordContainer().hide();
			self.toggleSendMessage(true);
			callback(true);
		});
	};

	ClientWindow.prototype.__sendAudioRecord = function() {
		const self = this;
		self.$recordStart().prop("disabled", true);
		this.__stopAudioRecord(function(status) {
			if (!status) {
				self.$recordStart().prop("disabled", false);
				return;
			}
			self.__uploadAudioRecord(function() {
				self.$recordStart().prop("disabled", false);
				self.onAudioRecordStopped();
			});
		});
	};

	ClientWindow.prototype.__cancelAudioRecord = function() {
		const self = this;
		self.$recordStart().prop("disabled", true);
		this.__stopAudioRecord(function() {
			self.$recordStart().prop("disabled", false);
			self.onAudioRecordStopped();
		});
	};

	ClientWindow.prototype.__getRecordFileName = function() {
		return "audio";
	};

	ClientWindow.prototype.__uploadAudioRecord = function(callback = $.noop) {
		const self = this;
		this.m_recorder.getRecordFile(self.__getRecordFileName(), function(file, error) {
			if (!file) {
				self.displayStatusMessage(format("<%=_T('Error processing audio record file: {0}')%>", error), 
					"Error", false, 3000);
				return callback(false);
			}
			var data = self.processAttachmentData({
				type : "audio",
				description : "",
				record : true
			});
			self.m_interaction.sendAttachment(data, file, function(status, error) {
				if (!status) {
					self.displayStatusMessage(error, "Error", false, 3000);
				}
				callback(true);
			});
		});
	};

	ClientWindow.prototype.hasTemplateMessages = function() {
		return false;
	};

	ClientWindow.prototype.canSendTemplateMessages = function() {
		return this.hasTemplateMessages() && this.m_interaction && this.m_interaction.canSendTemplateMessages();
	};

	ClientWindow.prototype.toggleTemplateMessages = function() {
		if (this.canSendTemplateMessages()) {
			this.enableTemplateMessages();
		}
		else {
			this.disableTemplateMessages();
		}
	};

	ClientWindow.prototype.enableTemplateMessages = function() {
		const self = this;
		$("#messageAddTemplateButton")
			.show()
			.off("click")
			.on("click", function() {
				self.openTemplateMessagesPopup();
			});
	};

	ClientWindow.prototype.disableTemplateMessages = function() {
		$("#messageAddTemplateButton")
			.hide()
			.off("click");
	};

	ClientWindow.prototype.appendTemplateMessagesPopup = function($popup) {
		$("body").append($popup);
	};

	ClientWindow.prototype.openTemplateMessagesPopup = function(message, callback = $.noop) {
		const self = this;
		const popupId = "windowTemplateMessagesPopupContainer";
		$('#' + popupId).remove();
		const $popup = $("".concat(
			"<div id='" + popupId + "' class='systemPopWindow windowContainer templateMessagesPopup' style='display:none;'>",
				"<div class='windowTop'>",
					"<div class='windowTitle'>",
						"<%=_T('Template Message')%>",
					"</div>",
				"</div>",
				"<div middle_container class='windowMiddle'>",
					"<div class='card-block'>",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label for='templateMessageGroup' class='required'>",
										"<%=_T('Namespace')%>:",
									"</label>",
									"<select id='templateMessageGroup' class='form-control'>",
									"</select>",
								"</div>",
							"</div>",
						"</div>",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label for='templateMessageId' class='required'>",
										"<%=_T('Name')%>:",
									"</label>",
									"<select id='templateMessageId' class='form-control'>",
									"</select>",
								"</div>",
							"</div>",
						"</div>",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label for='templateMessageLanguage' class='required'>",
										"<%=_T('Language')%>:",
									"</label>",
									"<select id='templateMessageLanguage' class='form-control'>",
									"</select>",
								"</div>",
							"</div>",
						"</div>",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label class=''>",
										"<%=_T('Header')%>:",
									"</label>",
									"<div id='templateHeaderText' class='form-control'>",
									"</div>",
								"</div>",
							"</div>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label class=''>",
										"<%=_T('Content')%>:",
									"</label>",
									"<div id='templateMessageText' class='form-control'>",
									"</div>",
								"</div>",
							"</div>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label class=''>",
										"<%=_T('Resource')%>:",
									"</label>",
									"<div id='templateMessageButton' class='form-control'>",
									"</div>",
								"</div>",
							"</div>",
						"</div>",
					"</div>",
				"</div>",
				"<div bottom_container class='windowBottom internalPopUpButtons d-flex '>",
					"<input class='pageButtonGeneric buttonColorWhite' id='templateMessageCancel' type='button' value='" + "<%=_T('Cancel')%>" + "'>",
					"<input class='pageButtonGeneric buttonColorBlue ' id='templateMessageSend' type='button' value='" + "<%=_T('Send')%>" + "'>",
				"</div>",
			"</div>"
		));
		$popup.find("#templateMessageSend").click(function() {
			self.sendTemplateMessage();
		});
		$popup.find("#templateMessageCancel").click(function() {
			self.closeTemplateMessagePopup();
		});

		this.appendTemplateMessagesPopup($popup);

		$("#messageAddTemplateButton").prop("disabled", true);

		this.loadTemplateMessageOptions(message, function(status, error) {
			$("#messageAddTemplateButton").prop("disabled", false);
			if (!status) {
				return callback(false, error);
			}
			self.templateMessagePopup = new inConcert.PopupWindow({ownerId: popupId});
			self.templateMessagePopup.show(
			{
				ownerId       : popupId,
				eventName     : popupId,
				canReposition : true,
				closeOnFocusLost : true,
				parentNode    : popupId,
				callbacks: {
					onOpen: function() {
						$('#' + popupId).show();
					},
					onClose: function( callback ) {
						callback = callback || $.noop;
						if (self.templateMessagePopup) {
							self.templateMessagePopup.hide(function() {
								callback();
							});
						}
						else {
							callback();
						}
					}
				}
			});
			callback(true);
		});
	};

	ClientWindow.prototype.loadTemplateMessageOptions = function(message, callback) {
		const self = this;
		this.m_interaction.loadTemplateMessageOptions(function(options) {
			if (!options) {
				self.displayStatusMessage("<%=_T('Error loading template messages')%>", "Error", false, 3000);
				return callback(false, "Error loading template messages");
			}
			self.templateMessageOptions = {};
			var templateMessageOptions_a = [];
			$.each(options, function(_, option) {
				if (!option.available) {
					return;
				}
				var group;
				if (self.templateMessageOptions[option.groupId]) {
					group = self.templateMessageOptions[option.groupId];
				}
				else {
					group = {
						groupId : option.groupId,
						groupName : option.groupName || option.groupId,
						messages : {}
					};
					self.templateMessageOptions[option.groupId] = group;
					templateMessageOptions_a.push(group);
				}
				group.messages[option.id] = option;
			});
			if (!templateMessageOptions_a.length) {
				self.displayStatusMessage("<%=_T('Has not configured template messages')%>", "Error", false, 3000);
				return callback(false, "Has not configured template messages");
			}
			templateMessageOptions_a.sort(function(a, b) {
				return a.groupName.localeCompare(b.groupName);
			});
			var $groups = [];
			$.each(templateMessageOptions_a, function(_, group) {
				$groups.push($(format("<option value='{0}'>{1}</option>",
					group.groupId, group.groupName)));
			});

			$("#templateMessageLanguage").change(function() {
				self.loadTemplateMessageText();
			});

			$("#templateMessageId").change(function() {
				var groupId = $("#templateMessageGroup").val();
				var messageId = $(this).val();
				var message = (groupId && messageId)? self.templateMessageOptions[groupId].messages[messageId] : {};
				var languages_a = [];
				$.each(message.languages || {}, function(id, language) {
					languages_a.push({
						id : id,
						label : inConcertLanguages.getLanguageLabel(id),
						text : language.text
					});
				});
				languages_a.sort(function(a, b) {
					return a.label.localeCompare(b.label);
				});
				var $languages = [];
				$.each(languages_a, function(_, language) {
					var $language = $(format("<option value='{0}'>{1}</option>",
						language.id, language.label));
					$language.attr("title", language.text);
					$languages.push($language);
				});
				if ($languages.length === 0 || $languages.length > 1) {
					var $language = $("<option value='USER'><%=_T('User Language')%></option>");
					$language.attr("title", "<%=_T('User language is used to select which message send')%>");
					$languages.unshift($language);
				}
				$("#templateMessageLanguage")
					.html("")
					.append($languages)
					.change()
					.parent()
					.toggle($languages.length > 1);
			});

			$("#templateMessageGroup")
				.append($groups)
				.change(function() {
					var groupId = $(this).val();
					var messages = groupId && self.templateMessageOptions[groupId].messages || {};
					var messages_a = [];
					$.each(messages, function(_, message) {
						let defaultTemplateMessageLanguage = self.getDefaultTemplateMessageLanguage(message);
						if(defaultTemplateMessageLanguage) {
							messages_a.push({
								id : message.id,
								name : message.name || message.id,
								text : defaultTemplateMessageLanguage.text
							});
						}
					});
					messages_a.sort(function(a, b) {
						return a.name.localeCompare(b.name);
					});
					var $messages = [];
					$.each(messages_a, function(_, message) {
						var $message = $(format("<option value='{0}'>{1}</option>",
							message.id, message.name));
						$message.attr("title", message.text);
						$messages.push($message);
					});
					$("#templateMessageId")
						.html("")
						.append($messages)
						.change();
				})
				.change();
			if ($groups.length < 2) {
				$("#templateMessageGroup").parent().hide();
			}
			//preset message
			let template = message && message.template || {};
			if (template.group) {
				$("#templateMessageGroup").val(template.group).change();
			}
			if (template.id) {
				$("#templateMessageId").val(template.id).change();
			}
			if (template.language) {
				$("#templateMessageLanguage").val(template.language).change();
			}
			if (self.m_textTemplate && template.example_header_text_value) {
				self.m_textTemplate.setHeader(template.example_header_text_value);
			}
			if (self.m_textTemplate && template.parameters) {
				self.m_textTemplate.set(template.parameters);
			}
			if (self.m_textTemplate && template.hsm_buttons) {
				let ctaButton = template.hsm_buttons.find(button => button.type === "call_to_action" && button.urlButtonParam);
				if (ctaButton) {
					self.m_textTemplate.setCtaButton(ctaButton.payload);
				}
			}
			callback(true);
		});
	};

	ClientWindow.prototype.getDefaultTemplateMessageLanguage = function(message) {
		const defaultTexts = [
			"en",
			"en_US"
		];
		if (message.defaultText) {
			return { text : message.defaultText };
		}
		var defaultLanguage = null;
		let userLanguage = inConcert.app.getLoggedUserLanguage();
		if(message.languages[userLanguage]) {
			return message.languages[userLanguage];
		}
		$.each(defaultTexts, function(_, id) {
			if (message.languages[id]) {
				defaultLanguage = message.languages[id];
				return false;
			}
		});
		if (!defaultLanguage) {
			var keys = Object.keys(message.languages);
			keys.sort();
			defaultLanguage = message.languages[keys[0]];
		}
		return defaultLanguage;
	};

	ClientWindow.prototype.loadTemplateMessageText = function() {
		var self = this;
		var groupId = $("#templateMessageGroup").val();
		var messageId = $("#templateMessageId").val();
		var message = (groupId && messageId)? self.templateMessageOptions[groupId].messages[messageId] : null;
		if (!message) {
			return;
		}
		var languageId = $("#templateMessageLanguage").val();
		let language = languageId === "USER"? self.getDefaultTemplateMessageLanguage(message) :
			message.languages[languageId];
		if (!language) {
			return;
		}

		//caso ICC_CLOUD_API: 
		//	el texto y posible param de la cabecera esta en el header
		//	los buttons estan en los components
		languageId = languageId === "USER" ? inConcert.app.getLoggedUserLanguage() : languageId;
		let header_text;
		let example_header_text;
		message.hsm_buttons = [];
		for (const language in message.languages) {
			if (Object.hasOwn(message.languages, language)) {
				if(languageId === language && message.languages[language].components) {
					for (const component in message.languages[language].components) {
						if (Object.hasOwn(message.languages[language].components, component)) {
							if(message.languages[language].components[component]) {
								if(message.languages[language].components[component].type === "HEADER" && message.languages[language].components[component].format === "TEXT" && message.languages[language].components[component].header_param) {
									header_text = message.languages[language].components[component].text;
									example_header_text = message.languages[language].components[component].header_param;
								}
								else if(message.languages[language].components[component].type === "BUTTONS") {
									for (let i = 0; i < message.languages[language].components[component].buttons.length; i++) {										
										let button = {};									
										button.type = message.languages[language].components[component].buttons[i].type === "QUICK_REPLY" ? "quick_reply" : "call_to_action";
										if(button.type === "call_to_action" && message.languages[language].components[component].buttons[i].type === "URL" && message.languages[language].urlButtonParam) {
											button.title = message.languages[language].components[component].buttons[i].text;
											button.url = message.languages[language].components[component].buttons[i].url;
											button.urlButtonParam = true;
											button.otp_type = message.languages[language].components[component].buttons[i].otp_type;
										}
										button.payload = message.languages[language].components[component].buttons[i].text;
										message.hsm_buttons.push(button);
									}
								}
							}
						}
					}
				}
			}
		}

		this.m_textTemplate = new inConcert.TextTemplate({
			header_control : $("#templateHeaderText"),
			header_text : header_text,
			example_header_text : example_header_text,
			control : $("#templateMessageText"), 
			content : language.text,
			placeholders : language.params || message.params,
			messages : {
				"required" : "<%=_T('Item value is required')%>",
				"resource" : "<%=_T('Resource must be a valid URL resource')%>"
			},
			button_control: $("#templateMessageButton"),
			hsm_buttons: message.hsm_buttons
		});
	};

	ClientWindow.prototype.closeTemplateMessagePopup = function() {
		var self = this;
		if (this.templateMessagePopup) {
			this.templateMessagePopup.finalize(function() {
				self.templateMessagePopup = null;
			});
		}
	};

	ClientWindow.prototype.getAddressToSendTemplateMessage = function() {
		return {};
	};

	ClientWindow.prototype.sendTemplateMessage = function(callback = $.noop) {
		var self = this;
		var groupId = $("#templateMessageGroup").val();
		if (!groupId) {
			self.displayStatusMessage("<%=_T('Must select a group')%>", "Error", false, 3000);
			return callback(false, "Must select a group");
		}
		var messageId = $("#templateMessageId").val();
		if (!messageId) {
			self.displayStatusMessage("<%=_T('Must select a message')%>", "Error", false, 3000);
			return callback(false, "Must select a message");
		}
		var languageId = $("#templateMessageLanguage").val();
		if (!languageId) {
			self.displayStatusMessage("<%=_T('Must select a language')%>", "Error", false, 3000);
			return callback(false, "Must select a language");
		}
		languageId = languageId === "USER" ? inConcert.app.getLoggedUserLanguage() : languageId;
		var textData = self.m_textTemplate.get(true);
		if (textData.error) {
			self.displayStatusMessage(textData.error.message, "Error", false, 3000);
			return callback(false, textData.error.message);
		}
		var address = self.getAddressToSendTemplateMessage();
		if (address && address.error) {
			self.closeTemplateMessagePopup();
			return callback(false, "Must enter an address");
		}

		$("#templateMessageSend").prop("disabled", true);

		let message = (groupId && messageId) ? self.templateMessageOptions[groupId].messages[messageId] : null;
		let buttons = [];
		if (message && message.hsm_buttons) {
			buttons = message.hsm_buttons;
		}		
		let mediaUrl = "";
		if (message && message.mediaUrl) {
			mediaUrl = message.mediaUrl;
		}
		let mediaFileName = "";
		if (message && message.mediaFileName) {
			mediaFileName = message.mediaFileName;
		}
		let mediaContentType = "";
		if (message && message.mediaContentType) {
			mediaContentType = message.mediaContentType;
		}

		//caso ICC_CLOUD_API: los datos de media estan en los components
		let headerMediaTypes = {
			"IMAGE"    : true,
			"VIDEO"    : true,
			"DOCUMENT" : true
		};
		for (const language in message.languages) {
			if (Object.hasOwn(message.languages, language)) {
				if(languageId === language && message.languages[language].components) {
					for (const component in message.languages[language].components) {
						if (Object.hasOwn(message.languages[language].components, component)) {
							if(message.languages[language].components[component]) {
								if(message.languages[language].components[component].type === "HEADER" && headerMediaTypes[message.languages[language].components[component].format]) {
									mediaUrl = message.languages[language].components[component].mediaUrl;
									if(message.languages[language].components[component].sharedFile) {
										mediaFileName = message.languages[language].components[component].sharedFile.fileName;
										mediaContentType = message.languages[language].components[component].sharedFile.contentType;
									}
								}
							}
						}
					}
				}
			}
		}
		self.m_interaction.sendTemplateMessage({
			attention_level_id : self.getSelectedAttentionLevel(),
			address : address,
			groupId : groupId,
			messageId : messageId,
			languageId : languageId,
			textData: textData,
			buttons: buttons,
			mediaUrl: mediaUrl,
			mediaFileName: mediaFileName,
			mediaContentType: mediaContentType,
			callback : function(status, result) {
				if (!status) {
					self.displayStatusMessage(result, "Error", false, 3000);
				}
				self.closeTemplateMessagePopup();
				callback(status, result);
			}
		});
	};

	ClientWindow.prototype.getContactAddressActions = function(addressType) {
		const actionsByAddressType = {
			Phone : [
				{ type : "CALL", label : "<%=_T('Call')%>" },
				{ type : "SMS", label : "<%=_T('Sms')%>" },
				{ type : "WHATSAPP", label : "<%=_T('Whatsapp')%>" }
			],
			IM : [
				{ type : "IM", label : "<%=_T('IM')%>" }
			],
			Mail : [
				{ type : "MAIL", label : "<%=_T('E-mail')%>" }
			]
		};
		if (!inConcert.permissions.to_action_on_any_campaign("interaction", "int_create_outgoing")) {
			return;
		}
		let actions = [];
		$.each(actionsByAddressType[addressType] || [], function(_, action) {
			if (!inConcert.app.hasLicensesForChannel(action.type)) {
				return;
			}
			actions.push(action);
		});
		return actions;
	};

	ClientWindow.prototype.startContactAddressInteraction = function(addressType, proxyType, addressInfo, contact, callback) {
		let defaultError = "<%=_T('Unable to create outgoing conversation, please try again later')%>";
		inConcert.app.getContactManager().buildAddress(addressType, addressInfo, function(address) {
			if (!address) {
				return callback(false, defaultError);
			}
			let panel = inConcert.app.getMainWindow().getPanel("AssignedInteractions");
			let response = panel.createOutgoingInteraction(proxyType, contact, address, inConcert.app.getCampaigns());
			if (response.success) {
				callback(true);
			}
			else {
				let error = response.conflicts ? response.conflicts.message : defaultError;
				callback(false, error);
			}
		});
	};
	
	ClientWindow.prototype.startForward = function(index) {
		const self = this;
		this.iterateEventsToForward(index);
		$("body").addClass("forwardEnabled");
		$("[message_control]").addClass("forwardHiddenControl");
		$("[forward_message_counter], [forward_message_button], [forward_message_cancel_button]").show();
		$("[forward_message_button]")
			.off("click")
			.on("click", function() {
				self.openForwardMessagePopup();
			});
		$("[forward_message_cancel_button]")
			.off("click")
			.on("click", function() {
				self.stopForward();
			});
		this.onForwardSelectorChange();
	};
	
	ClientWindow.prototype.stopForward = function() {
		$("[forward_message_button], [forward_message_cancel_button]").off("click");
		$("[forward_message_counter], [forward_message_button], [forward_message_cancel_button]").hide();
		$("[forward_selector_container]").remove();
		$("[message_control]").removeClass("forwardHiddenControl");
		$("body").removeClass("forwardEnabled");
	};

	ClientWindow.prototype.enableEventForward = function($event, event_index, selected) {
		let self = this;
		let $selector = $(format(
			"".concat(
				"<div forward_selector_container class='forwardSelectorContainer'>",
					"<input forward_selector='{0}' type='checkbox' class='forwardSelector', {1}>",
				"</div>"
			),
			event_index,
			selected? "checked" : ""
		));
		$selector.change(function() {
			self.markForwardSelector($(this), $(this).is(":checked"));
			self.onForwardSelectorChange();
		});
		$event.addClass("forwardEnabled");
		self.addForwardSelector($event, $selector);
		self.markForwardSelector($selector, selected);
	};
	
	ClientWindow.prototype.onForwardSelectorChange = function() {
		let selected = $("[forward_selector]:checked").length;
		$("[forward_message_counter]").text(format("<%=_T('{0} selected')%>", selected));
		$("[forward_message_button]").prop("disabled", selected === 0);
	};

	ClientWindow.prototype.appendForwardMessagePopup = function($popup) {
		$("body").append($popup);
	};
	
	ClientWindow.prototype.getForwardMessagePopupAddressLabel = function() {
		return "<%=_T('Address')%>";
	};

	ClientWindow.prototype.openForwardMessagePopup = function() {
		const self = this;
		const popupId = "windowForwardMessagesPopupContainer";
		const addressLabel = this.getForwardMessagePopupAddressLabel();
		$('#' + popupId).remove();
		const $popup = $("".concat(
			"<div id='" + popupId + "' class='systemPopWindow windowContainer forwardMessagePopup' style='display:none;'>",
				"<div class='windowTop'>",
					"<div class='windowTitle'>",
						"<%=_T('Forward Messages')%>",
					"</div>",
				"</div>",
				"<div middle_container class='windowMiddle pageFormContainer'>",
					"<div class='card-block'>",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div class='form-group mb-1'>",
									"<label class='form-check-label-title pl-0'>",
										"<%=_T('Forward to')%>:",
									"</label>",
								"</div>",
							"</div>",
							"<div class='col-12'>",
								"<div class='form-group mb-0'>",
									"<div class='d-flex'>",
										"<div class='form-radiobutton'>",
											"<label for='forward_message_to_address' class='d-flex align-items-center pl-0'>",
												"<input type='radio' name='forward_message_to' id='forward_message_to_address' class='form-radiobutton-input' checked>",
												"<span>",
													addressLabel,
												"</span>",
											"</label>",
										"</div>",
										"<div class='form-radiobutton'>",
											"<label for='forward_message_to_contact' class='d-flex align-items-center'>",
												"<input type='radio' name='forward_message_to' id='forward_message_to_contact' class='form-radiobutton-input'>",
												"<span>",
													"<%=_T('Contact')%>",
												"</span>",
											"</label>",
										"</div>",
									"</div>",
								"</div>",
							"</div>",
						"</div>",
						"<hr />",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div forward_message_to_container='address' class='form-group mb-0' hiddden>",
									"<label for='templateMessageId' class='required'>",
										addressLabel + ":",
									"</label>", 
									"<input type='text' id='forward_message_address' class='form-control'>",
								"</div>",
							"</div>",
						"</div>",
						"<div class='row'>",
							"<div class='col-12'>",
								"<div forward_message_to_container='contact' class='form-group mb-0' hiddden>",
									"<label for='templateMessageId' class='required'>",
										"<%=_T('Enter contact')%>:",
									"</label>", 
								"</div>",
							"</div>",
						"</div>",
					"</div>",
				"</div>",
				"<div bottom_container class='windowBottom customPageButtonContainer'>",
					"<input class='pageButtonGeneric buttonColorWhite' id='forwardMessageCancel' type='button' value='" + "<%=_T('Cancel')%>" + "'>",
					"<input class='pageButtonGeneric buttonColorBlue' id='forwardMessageSend' type='button' value='" + "<%=_T('Send')%>" + "'>",
				"</div>",
			"</div>"
		));
		
		$popup.find("[name=forward_message_to]").change(function() {
			self.selectFowardMessagePopupTo();
		});
		$popup.find("#forwardMessageSend").click(function() {
			self.forwardMessages();
		});
		$popup.find("#forwardMessageCancel").click(function() {
			self.closeForwardMessagePopup();
		});

		this.appendForwardMessagePopup($popup);

		$("[forward_message_button], [forward_message_cancel_button]").prop("disabled", true);

		this.forwardMessagesPopup = new inConcert.PopupWindow({ownerId: popupId});
		this.forwardMessagesPopup.show(
		{
			ownerId       : popupId,
			eventName     : popupId,
			canReposition : true,
			closeOnFocusLost : false,
			parentNode    : popupId,
			callbacks: {
				onOpen: function() {
					$('#' + popupId).show();
				},
				onClose: function(callback = $.noop) {
					if (self.forwardMessagesPopup) {
						self.forwardMessagesPopup.hide(function() {
							callback();
						});
					}
					else {
						callback();
					}
				}
			}
		});

		this.selectFowardMessagePopupTo();
	};
	
	ClientWindow.prototype.selectFowardMessagePopupTo = function() {
		$("[forward_message_to_container]").hide();
		if ($("#forward_message_to_address").prop("checked")) {
			$("[forward_message_to_container=address]").show();
			this.loadForwardMessagePopupToAddress();
		}
		else {
			$("[forward_message_to_container=contact]").show();
			this.loadForwardMessagePopupToContact();
		}
	};
	
	ClientWindow.prototype.loadForwardMessagePopupToAddress = function() {
		
	};
	
	ClientWindow.prototype.loadForwardMessagePopupToContact = function() {
		if (this.m_forwardContacts) {
			return;
		}
		this.m_forwardContacts = this.makeForwardMessageContactsList();
		$("[forward_message_to_container='contact']").append(this.m_forwardContacts.make());
		this.m_forwardContacts.start();
	};
	
	ClientWindow.prototype.makeForwardMessageContactsList = function() {
		return new inConcert.ForwardMessageContactsList(this.getForwardMessageAddressType());
	};

	ClientWindow.prototype.closeForwardMessagePopup = function() {
		let self = this;
		this.m_forwardContacts = null;
		if (this.forwardMessagesPopup) {
			this.forwardMessagesPopup.finalize(function() {
				self.templateMessagePopup = null;
			});
		}
		$("[forward_message_button], [forward_message_cancel_button]").prop("disabled", false);
	};
	
	ClientWindow.prototype.getForwardMessagePopupTo = function() {
		if ($("#forward_message_to_address").prop("checked")) {
			return this.getForwardMessagePopupToAddress();
		}
		else {
			return this.getForwardMessagePopupToContact();
		}
	};
	
	ClientWindow.prototype.getForwardMessagePopupToAddress = function() {
		let address = $("#forward_message_address").val();
		if (!address) {
			this.displayStatusMessage( "<%=_T('Must enter an address')%>", "Error", true, 5000 );
			return false;
		}
		return this.validateForwardMessageAddress(address);
	};
	
	ClientWindow.prototype.validateForwardMessageAddress = function(address) {
		return address;
	};
	
	ClientWindow.prototype.getForwardMessagePopupToContact = function() {
		let address = this.m_forwardContacts.getSelectedAddress();
		if (!address) {
			this.displayStatusMessage( "<%=_T('Must select an address')%>", "Error", true, 5000 );
			return false;
		}
		return this.validateForwardMessageAddress(address);
	};
	
	ClientWindow.prototype.getForwardMessages = function() {
		let self = this;
		let messages = [];
		$("[forward_selector]:checked").each(function() {
			let message = self.m_interaction.getForwardMessage($(this).attr("forward_selector"));
			if (message) {
				messages.push(message);
			}
		});
		return messages;
	};
	
	ClientWindow.prototype.forwardMessages = function() {
		let self = this;
		self.hideStatusMessage();
		let address = self.getForwardMessagePopupTo();
		let messages = self.getForwardMessages();
		if (!address || !messages.length) {
			return;
		}
		$("#forwardMessageSend, #forwardMessageCancel").prop("disabled", true);
		self.m_interaction.forward(address, messages, function(status, message, error) {
			$("#forwardMessageSend, #forwardMessageCancel").prop("disabled", false);
			if (status) {
				self.closeForwardMessagePopup();
				self.stopForward();
			}
			else {
				self.displayStatusMessage(message, "Error", false, 5000);
			}
		});
	};
	
	ClientWindow.prototype.iterateEventsToForward = function(index) {
		throw "not implemented";
	};
	
	ClientWindow.prototype.addForwardSelector = function($event, $selector) {
		throw "not implemented";
	};
	
	ClientWindow.prototype.markForwardSelector = function($selector, selected) {
		throw "not implemented";
	};
	
	ClientWindow.prototype.getForwardMessageAddressType = function() {
		throw "not implemented";
	};
	
	ClientWindow.prototype.firstLoadDraftMessage = function() {
		this.loadDraftMessage();
	};
	
	ClientWindow.prototype.loadDraftMessage = function() {
		let self = this;
		if (self.m_interaction && self.m_interaction.hasDraftMessage()) {
			let message = self.m_interaction.getDraftMessage();
			let autosend = self.m_interaction.autosendDraftMessage();
			let callback = self.m_interaction.getDraftMessageCallback() || $.noop;
			self.m_interaction.removeDraftMessage(); //TODO: draft messages store policy
			self.__addDraftMessage(message, function(status, error) {
				if (status && autosend) {
					window.setTimeout(function() {
						self.__sendDraftMessage(message, callback);
					}, 500);
				}
				else {
					callback(status, error);
				}
			});
		}
	};
	
	ClientWindow.prototype.__addDraftMessage = function(message, callback) {
		switch(message.msgtype) {
			case "template":
				this.__addDraftTemplateMessage(message, callback);
				break;
			case "attachment":
				this.__addDraftAttachmentMessage(message, callback);
				break;
			case "shared_file":
				this.__addDraftSharedFileMessage(message, callback);
				break;
			default:
				this.__addDraftTextMessage(message, callback);
		}
	};
	
	ClientWindow.prototype.__addDraftTemplateMessage = function(message, callback) {
		let self = this;
		if (self.canSendTemplateMessages()) {
			self.openTemplateMessagesPopup(message, callback);
		}
		else {
			callback(false, "not supported");
		}
	};
	
	ClientWindow.prototype.__addDraftAttachmentMessage = function(message, callback) {
		let self = this;
		if (self.canSendAttachment()) {
			self.openAttachmentSelector(message, callback);
		}
		else {
			callback(false, "not supported");
		}
	};
	
	ClientWindow.prototype.__addDraftSharedFileMessage = function(message, callback) {
		let self = this;
		if (self.canSendSharedFile()) {
			self.openSharedFileSelector(message, callback);
		}
		else {
			callback(false, "not supported");
		}
	};
	
	ClientWindow.prototype.__addDraftTextMessage = function(message, callback) {
		callback(false, "not supported");
	};
	
	ClientWindow.prototype.__sendDraftMessage = function(message, callback) {
		switch(message.msgtype) {
			case "template":
				this.__sendDraftTemplateMessage(message, callback);
				break;
			case "attachment":
				this.__sendDraftAttachmentMessage(message, callback);
				break;
			case "shared_file":
				this.__sendDraftSharedFileMessage(message, callback);
				break;
			default:
				this.__sendDraftTextMessage(message, callback);
		}
	};
	
	ClientWindow.prototype.__sendDraftTemplateMessage = function(message, callback) {
		this.sendTemplateMessage(callback);
	};
	
	ClientWindow.prototype.__sendDraftSharedFileMessage = function(message, callback) {
		if (message.sharedFile) {
			this.sendSharedFile(message.option, message.sharedFile, callback);
		}
		else {
			callback(false, "must enter a file id");
		}
	};
	
	ClientWindow.prototype.__sendDraftAttachmentMessage = function(message, callback) {
		if (message.repositoryFile) {
			this.sendRepositoryFile(message.option, message.repositoryFile, callback);
		}
		else {
			callback(false, "must enter a file id");
		}
	};
	
	ClientWindow.prototype.__sendDraftTextMessage = function(message, callback) {
		callback(false, "not supported");
	};

	/*
		Funcion util para los clientes:
			twitter, instagram, sms, facebook, youtube, whatsapp.
		Aunque no aplica para todos los clientes, (ejemplo: webchat, mail, etc) igualmente se deja en la clase base,
			ya que sino tiene que estar repetida con el mismo codigo en los clientes que si lo necesitan.
	*/
	ClientWindow.prototype.set_mark_all_unread_read_button_caption = function() {
		const is_there_any_unread = this.m_interaction.getUnreadEventCount() > 0;
		let caption;
		let class_name;
		let action_id;

		if (is_there_any_unread) {
			caption = this.MARK_ALL_READ_BTN_TEXT;
			class_name = "buttons-icons-mail-closed";
			action_id = "READ";
		}
		else {
			caption = this.MARK_ALL_UNREAD_BTN_TEXT;
			class_name = "buttons-icons-mail-open";
			action_id = "UNREAD";
		}

		const $btn = $("#mark_all_read_unread");
		$btn.attr("action_id", action_id);
		$btn.val(caption);
		$btn.attr("title", caption);
		$btn.find("i").removeClass().addClass(class_name);
	};

	ClientWindow.prototype.get_mark_all_unread_read_button_caption = function() {
		const $btn = $("#mark_all_read_unread");

		return {
			caption: $btn.attr("title"),
			action_id: $btn.attr("action_id")
		};
	};

	/* -------------------------
		Funciones utilizadas para la busqueda de contactos.
	----------------------------*/

	ClientWindow.prototype.addNewDestinationContact = function($btn, field, search_value, callback = $.noop)
	{
		if (!this.m_addresses_book_model) {
			this.m_addresses_book_model = new inConcert.AddressesBookSelectorModel(this.m_interaction.type);
		}

		//this contact selector allows to create new contacts
		if (!this.m_contact_auto_search_popup) {
			let address_type = this.m_addresses_book_model.getAddressType();

			this.m_contact_auto_search_popup = new inConcert.AutoSearchContactSelector(
				$("#contactSelector"),
				address_type, //"Mail", "Phone"
				function(onContactCreatedFn) {
					inConcert.app.getMainWindow().createContact(onContactCreatedFn);
				},
				function() {
					return $("#cboCampaigns").val();
				},
				search_value,
				this.m_addresses_book_model
			);
		}
		else {
			this.m_contact_auto_search_popup.set_search_value(search_value);
		}

		const interaction_type = this.m_interaction.type.toLowerCase();
		const $field = $(field);

		this.m_contact_auto_search_popup.show({
			parentNode: $btn,
			callbacks: {
				onOpen : $.noop,
				onClose: $.noop
			},
			onContactSelectedFn: function(contact, address) {
				let value;
				if (interaction_type === "mail") {
					let currentContacts = $field.val();
					currentContacts = currentContacts.replace(search_value, "");
					value = currentContacts + contact.Name + " <" + address + ">";
				}
				else {
					value = address;
				}
				$field.val(value).focus();
				callback(true);
			}
		});
	};

	ClientWindow.prototype.search_contact = function(search_value, $ctrl, callback = $.noop) {
		const ctrl_id = $ctrl.attr("id");
		this.addNewDestinationContact($ctrl, "#" + ctrl_id, search_value, callback);
	};

	ClientWindow.prototype.reset_search_contact = function($ctrl) {
		$ctrl.val("").focus();
		if (this.m_contact_auto_search_popup) {
			this.m_contact_auto_search_popup.hide();
		}
	};

	ClientWindow.prototype.search_contact_keypress = function(evt, $ctrl, callback = $.noop) {
		const self = this;
		if (this.m_contact_search_timeout) {
			window.clearTimeout(this.m_contact_search_timeout);
			this.m_contact_search_timeout = undefined;
		}
		if (evt.keyCode === 27) { //Escape reset search
			this.reset_search_contact($ctrl);
			return;
		}
		let search_value = $ctrl.val();
		if (search_value) {
			if (this.m_interaction.type === "MAIL") {
				// Las interacciones de MAIL permiten mas de un destinatario (separado por ;) por lo que para este caso se realiza una nueva busqueda en blanco (sin filtrar el valor insertado previamente)
				let pos = search_value.lastIndexOf(";");
				if (pos > -1) {
					search_value = search_value.substring(pos + 1);
				}
			}
			// se encuentra en un timeout de 800 ms, para que no realice la busqueda mientras escribis.
			const elapsed_time_ms = 800;
			this.m_contact_search_timeout = window.setTimeout(function() {
				window.clearTimeout(self.m_contact_search_timeout);
				self.m_contact_search_timeout = undefined;
				self.search_contact(search_value, $ctrl, callback = $.noop);
			}, elapsed_time_ms);
		}
		else {
			if (evt.keyCode === 8 || evt.keyCode === 13 || evt.keyCode === 46) { // backspace/enter and no text, reset search.
				this.reset_search_contact($ctrl);
				return;
			}
		}
	};
	/* ------------------------- */

	/*
		Para utilizar las confirmaciones, debe incluir en los .html de cada cliente los siguientes JavaScripts
			<script type="text/javascript" src="/inconcert/lib/jqmodal/jqmodal.js"></script>
			<script type="text/javascript" src="/inconcert/apps/agent/scripts/modalpopups.<%= _L %>.js"></script>
	*/
	ClientWindow.prototype.confirm = function(message, callback = $.noop)
	{
		$.inconcertPopup.Show(
			"<%=_T('Confirm')%>",
			message,
			$.inconcertPopup.ButtonSets.BS_YESNO,
			function(response) {
				if (response.toLowerCase() !== "yes") {
					callback(false);
					return;
				}
				window.setTimeout(function() {
					callback(true);
				});
			},
			undefined,
			{
				popup_id       : "conf_inconcertPopup",
				container_class: "confirmationPopUp",
				overlay_class  : "confirmationOverlay"
			}
		);
	};

/*
-----------
Send Message
-----------
TODO: (REFACTORING) Revisar todas estas funciones referentes a "send message" y otras son utilizadas en los clientes "webchat", "whatsapp", "sms", "rcs".
Hay que revisar si pueden ser pasadas para aqui, o se les puede encontrar otra solucion.
*/
	ClientWindow.prototype.$sendMessageContainer = function() {
		return $("#sendMessageContainer");
	};

	ClientWindow.prototype.toggleSendMessage = function(show) {
		this.$sendMessageContainer().toggle(!!show);
	};
	//
	//--Fin ClientWindow----------------------------------------------------------------------------

	function ForwardMessageContactsList(addressType) {
		ForwardMessageContactsList.uber.constructor.call(this, inConcert.app.getContactManager(), {
			className : "forwardMessageContacts",
			addressTypes : [ addressType ],
			addressRows : true,
			addressRowsAutoSelectFirst : true
		});
	}

	if (inConcert.ContactList) {
		inConcert.Utils.inherit(ForwardMessageContactsList, inConcert.ContactList);
	}
	
	if (typeof inConcert.ForwardMessageContactsList !== 'function') {
		inConcert.ForwardMessageContactsList = ForwardMessageContactsList;
	}

	//disable drag and drop redirect
	$(document)
		.on("dragover", function(e) {
			e.preventDefault();
			e.stopPropagation();
		})
		.on("dragenter", function(e) {
			e.preventDefault();
			e.stopPropagation();
		})
		.on("drop", function(e) {
			e.preventDefault();
			e.stopPropagation();
		});
}());
