/* jshint forin:true, noarg:true, jquery:true, noempty:true, eqeqeq:true, bitwise:true,
	strict:true, undef:true, unused:vars, curly:false, browser:true, indent:false, maxerr:50, esversion: 11 */

/* global $, inConcert, Mustache, _, EventEmitter */

// Create an inConcert object only if one does not already exist. We create the
// methods in a closure to avoid creating global variables.
if (!this.inConcert) {
	this.inConcert = {};
}

/**
 * @namespace
 * @name SupervisorDashboard
 * @author Sebastian Moreno (smoreno@inconcertcc.com)
 */
(function () {
	'use strict';

	const _T = inConcert.i18n.getString;
	const format = inConcert.Utils.format;
	const nid = inConcert.Utils.getNormalizedId;
	const REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS = 10000; // 10 seconds
	const REFRESH_INTERVAL_MS                      = 3000;  // 3 seconds
	const HISTORICAL_CHART_REFRESH_INTERVAL_MS     = 60000; // 1 minute (min slice 15 min)
	const QUEUE_PROCESS_INTERVAL_MS                = 200;   // 200 millis
	const REFRESH_ALARMS_ANOTHER_APP_MS            = 30000; // 30 seconds
	const ALERTED_ALARMS_TIMEOUT_MS                = 10000; // 10 seconds
	const MAXIMUM_NUMBER_OF_MONITORABLE_CAMPAIGNS  = 20;     // numero maximo de campanas monitoreables

	/**
	 * Creates a new dashboard instance.
	 *
	 * @constructor
	 * @memberOf SupervisorDashboard
	 */
	function SupervisorDashboard()
	{
		const self = this;

		this.latest_refresh = {};

		// Model/Config/Templates members
		this.m_templates = inConcert.sup_templates.dashboard;
		this.m_model = new inConcert.sup_model(this);
		this.m_config = new inConcert.sup_configuration(this);
		this.m_workgroups = new inConcert.sup_workgroups(this);
		// dashboard configuration
		this.m_dash_config = new inConcert.sup_dashboard_config(this);
		// notificaciones de las alarmas
		this.m_alarms_notifications = new inConcert.sup_alarms_notifications({dashboard: this, container_id: 'supervisor_main'});
		// lista de indicadores del sistema (alarmas) categorizados.
		this.m_categorized_indicators = {};

		// Call this members with dashboard instance as parameter.
		this.m_dash_archive = new inConcert.sup_dashboard_archive(this);
		this.m_dash_monitor = new inConcert.sup_dashboard_monitor(this);
		this.m_dash_appointment = new inConcert.sup_dashboard_appointment(this);
		this.m_dash_queued = new inConcert.sup_dashboard_queued(this);
		this.m_dash_waiting = new inConcert.sup_dashboard_waiting(this);
		this.m_dash_abandoned = new inConcert.sup_dashboard_abandoned(this);
		this.m_dash_bot_interactions = new inConcert.sup_dashboard_bot_interactions(this);
		this.m_dash_global_search = new inConcert.sup_dashboard_global_search(this);
		this.m_dash_members = new inConcert.sup_dashboard_members(this);
		this.m_dash_indicators = new inConcert.sup_dashboard_indicators(this);
		this.m_dash_summary = new inConcert.sup_dashboard_summary(this);
		// navigation by campaign
		this.navigation = {};

		// campaigns helpers members.
		this.m_rendered_campaigns = {};
		this.m_campaigns_being_changed = {};

		this.m_supervisor_leave = false;
		this.m_dashboard_render_state = "NONE";
		this.m_user_without_config = true;
		// TO-DO: Remove this member, use config js
		this.m_current_configuration = {};
		this.m_current_view_by_campaign = {};

		// refresh and timers
		this.m_on_refresh_timer = null;
		// refresh and timers alarms counter
		this.m_on_refresh_alarms_counter_timer = null;
		// start chart refresh timer.
		this.__start_charts_refresh_timer();

		// event queue starts empty
		this.m_event_queue = [];
		this.m_event_process_timer = null;
		this.m_current_processed_event_queue = null;
		this.m_hnd_current_processed_event_queue = null;

		// array to keep pending operations that
		// need the UI loaded
		this.m_pending_operations = [];

		// start queue processing right away
		this.__process_events_queue_alt();

		// permisos ids de cada panel (indicadores, team, summary)
		this.panels_permissions = {
			indicators: 'sup_view_indicators',
			members   : 'sup_view_team',
			summary   : 'sup_view_summary'
		};

		this.on('QueueBeingCompleted', function(alters_campaign_amount) {
			self.__on_queue_beign_completed(alters_campaign_amount);
		});

		this.m_members_selector = null;
		this.m_full_screen = null;
		this.pubsub = {};
		this.events_suscribe();
		// alarms notify counter
		this.m_hnd_alerted_timeout = null;
		this.initialize_alarms_counter_refresh();

		// Fecha y hora UTC de login del supervisor. Se usa en las comparaciones de fecha de notificaciones de kickouts
		this.m_userLoggedUTCdate = new Date((new Date()).toUTCString().slice(0, -4));

		// querybuilder object list
		this.queryBuilderObj = {};
		// global_searc querybuilder object list
		this.global_search_queryBuilderObj = {};

		this.m_dash_container_scroll_is_moving = false;
	}

	// inherits from event emitter.
	inConcert.Utils.inherit(SupervisorDashboard, EventEmitter);

	if (typeof inConcert.sup_dashboard !== 'function') {
		inConcert.sup_dashboard = SupervisorDashboard;
	}

	SupervisorDashboard.prototype.on_error = $.noop;

	SupervisorDashboard.prototype.get_maximum_number_of_monitorable_campaigns = function()
	{
		return MAXIMUM_NUMBER_OF_MONITORABLE_CAMPAIGNS;
	};

	SupervisorDashboard.prototype.initialize_full_screen = function()
	{
		if (this.m_full_screen) {
			return;
		}
		this.m_full_screen = new inConcert.sup_full_screen({dashboard: this});
	};

	SupervisorDashboard.prototype.events_suscribe = function()
	{
		const self = this;
		inConcert.app.messagebus.subscribe('/application/campaigns/bulk_add', function(campaigns) {
			// Se reciben todas las campanas nuevas que se le han asignado al Supervisor, y se van procesando de a una.
			campaigns.forEach((campaign_data, i) => {
				if (typeof campaign_data !== 'object' || !campaign_data.Id) {
					return;
				}
				var campaign = inConcert.app.getCampaign( campaign_data.Id );
				if (!campaign) {
					return;
				}
				let new_campaign_number = i + 1;
				self.__on_campaign_has_been_assigned_to_you(campaign, new_campaign_number);
			});
		});
		inConcert.app.messagebus.subscribe('/application/campaigns/removed', function(campaign_data) {
			self.__on_your_campaign_has_been_removed(campaign_data.Id);
		});
	};

	SupervisorDashboard.prototype.is_rendered_dashboard = function()
	{
		return this.m_dashboard_render_state === "RENDERED";
	};

	SupervisorDashboard.prototype.get_current_view_id = function( campaign_id )
	{
		return this.m_current_view_by_campaign[campaign_id] || 'dashboard';
	};

	SupervisorDashboard.prototype.set_chip_name = function (campaign_id,current_view_id) {
		let name="";
		let nameInside="";
		let current_dashboard;
		if (current_view_id) {
			switch ( current_view_id ) {
				case 'dashboard':
					name="<%=_T('Main dashboard')%>";
				break;
				case 'monitoring':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Active_Int')%>";
				break;
				case 'appointment':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Appointment_Int')%>";
				break;
				case 'archive':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Archived_Int')%>";
				break;
				case 'queued':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Queued')%>";
				break;
				case 'waiting':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Waiting_Int')%>";
				break;
				case 'global_search':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Global search')%>";
				break;
				case 'abandoned':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Abandoned_Int')%>";
				break;
				case 'bot_interactions':
					name="<%=_T('Interactions')%>";
					nameInside="<%=_T('Bot_interactions_Int')%>";
				break;
				case 'dashboard_config':
					name="<%=_T('Alarm management')%>";
				break;
			}
			current_dashboard = this.__get_current_dashboard(campaign_id, current_view_id);
		 }
		 const normalized_id = this.get_normalized_id(campaign_id);
		 const $chip = $(`#sup_dash_panel_child_${normalized_id}`);
		
		 let is_bots_view = this.__is_campaign_view_bots(campaign_id);
		 if (is_bots_view) {
			name = "<%=_T('Dashboard de Bots')%>";
			current_dashboard = {
				m_view_classes_name : "BotsViewer"
			};
		}

		let is_interactions = current_view_id !== "dashboard_config";
		 $chip.text(name);
		 $chip.hide();
		 if (this.m_full_screen) {
			  this.m_full_screen.set_chip_name(name, campaign_id, current_dashboard, current_view_id);
		}

		if (is_interactions)  {
			$(`#sup_dash_panel_interaction_title_${normalized_id}`).text(nameInside);
			$chip.show();
		}
		else if(name && name !==""){
			$chip.show();
		}
	};

	SupervisorDashboard.prototype.set_current_view_id = function( campaign_id, current_view_id )
	{
		this.set_chip_name(campaign_id,current_view_id);
		this.m_current_view_by_campaign[campaign_id] = current_view_id || 'dashboard';
	};

	SupervisorDashboard.prototype.set_current_config = function( config )
	{
		this.m_current_configuration = config;
	};

	SupervisorDashboard.prototype.get_stored_campaign_config = function(campaign_id)
	{
		if (!this.m_current_configuration || !campaign_id) {
			return;
		}
		return this.m_current_configuration[campaign_id];
	};

	/**
	 * Hide waiting animation
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.hide_waiting = function(){
		$('#sup_dash_main_waiting').hide();
	};

	/**
	 * Shows waiting animation
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.show_waiting = function(msg) {
		this.__hide_message();
		$('#sup_dash_main_waiting').show();
		if (msg) {
			this.update_waiting_message(msg);
		}
	};

	SupervisorDashboard.prototype.is_ready_config = function()
	{
		var self = this;

		// 1 Valida que la lista de campanas que el supervisor esta visualizando, se encuentren reenderizadas correctamente.
		var ready_count = 0;
		var camp_count = 0;
		$.each(this.m_rendered_campaigns, function(campaign_id) {
			camp_count++;
			if (self.__is_ready_rendered_campaign(campaign_id)) {
				ready_count++;
			}
		});
		if (camp_count === 0 || camp_count > ready_count) {
			// Si no existen campanas en pantalla, o la cantidad de campanas en pantalla no estan todas listas,
			// la configuracion aun no esta lista. 
			return false;
		}

		// 2 Valida contra la clase del dashboard de configuracion, que dicha clase este lista.
		//   Es decir que hayan culminado todos sus request, etc.
		if (!this.m_dash_config) {
			return;
		}
		var is_ready = this.m_dash_config.is_ready();

		return is_ready;
	};

	SupervisorDashboard.prototype.has_configured_alarms = function( campaign_id )
	{
		if (!this.m_dash_config) {
			return;
		}
		return this.m_dash_config.has_configured_alarms(campaign_id);
	};

	/**
	 * Retorna la configuracion de las alarmas definidas para la campana
	 *
	 * @public
	 * @param  {String} campaign_id - id de campana
	 * @return {object} - campaign alarms config: Ej:
		{
			indicators: {
				"campaign": {
					"indActiveAgents": {
					  "enabled": true,
					  "value": 14041996,
					  "operator": "DISTINCT"
					},
					..
				},
				accounts : {..},
				als : {..}
			}
	 */
	SupervisorDashboard.prototype.get_alarms_config = function( campaign_id )
	{
		if (!this.m_dash_config) {
			return;
		}
		return this.m_dash_config.get_alarms_config(campaign_id);
	};

	SupervisorDashboard.prototype.get_system_indicators = function()
	{
		if (!this.m_dash_config) {
			return;
		}
		var system_indicators = this.m_dash_config.get_system_indicators();
		this.__categorize_indicators(system_indicators);
		return system_indicators;
	};

	SupervisorDashboard.prototype.__categorize_indicators = function(system_indicators)
	{
		var self = this;
		if (!system_indicators) {
			return;
		}
		$.each(system_indicators.categories, function (category_id) {
			self.m_categorized_indicators[category_id] = {};
		});

		$.each(system_indicators.list, function (indicator_id, indicator) {
			self.m_categorized_indicators[indicator.category_id][indicator_id] = indicator;
		});
	};

	SupervisorDashboard.prototype.get_categorized_indicators = function()
	{
		if (!this.m_dash_config) {
			return;
		}
		return this.m_categorized_indicators;
	};

	/**
	 * Retorna la configuracion seleccionada en el tree selector de la campana: /cuentas/al.
	 *
	 * @public
	 * @param  {String} campaign_id - id de campana
	 * @return {object} - campaign selector config: Ej: { account_id: 'tu_cuenta', al_id: 'tu_al' }
	 * @memberOf SupervisorDashboard
	 */
	SupervisorDashboard.prototype.get_campaign_selected_config = function(campaign_id)
	{
		if ( typeof campaign_id !== 'string' ) {
			return {};
		}
		var campaign_config = this.get_stored_campaign_config(campaign_id);
		if ( campaign_config && campaign_config.dashboard ) {
			campaign_config = campaign_config.dashboard.config || {};
			return campaign_config;
		}
		return {};
	};

	SupervisorDashboard.prototype.change_campaign_selected_config = function(new_campaign_config, callback)
	{
		callback = callback || $.noop;
		if ( typeof new_campaign_config !== 'object' ) {
			callback(false);
			return;
		}
		var current_config = this.get_campaign_selected_config(new_campaign_config.campaign_id);
		if (!current_config) {
			callback(false);
			return;
		}
		current_config.account_id = new_campaign_config.account_id;
		current_config.al_id = new_campaign_config.al_id;

		var normalized_campaign_id = this.get_normalized_id(new_campaign_config.campaign_id);
		var $campaign = $('#sup_dash_campaign_selector_item_' + normalized_campaign_id);

		this.__set_campaign_selector_config($campaign);
		$('#sup_dash_campaign_selector_ok').trigger('click');
		callback(true);
	};

	/**
	 * Check if the dashboard is currently active
	 *
	 * @private
	 * @return {Boolean} - Returns true if the dashboard is active, otherwise false
	 */
	SupervisorDashboard.prototype.__dashboard_is_active = function(){
		return $('#supervisor_main').is(':visible');
	};

	/**
	 * Will process the supervisor pending operations, sometimes
	 * we need to delay some actions (like adding/removing campaigs)
	 * because we need to make sure the UI is loaded.
	 *
	 */
	SupervisorDashboard.prototype.__process_pendings_operations = function()
	{
		const self = this;
		this.m_pending_operations.forEach(function(operation){
			operation.fn.apply(self, operation.args);
		});
		this.m_pending_operations = [];
	};

	/**
	 * It's a callback for a PubSub subscribe called
	 * when the user has been removed from a campaign.
	 *
	 * @private
	 * @param  {String} campaign_id - The campaign identificator (eg. campaign@vcc)
	 * @memberOf SupervisorDashboard
	 */
	SupervisorDashboard.prototype.__on_your_campaign_has_been_removed = function(campaign_id)
	{
		if (!this.is_rendered_dashboard()) {
			return;
		}
		// execution should be delayed if the user is in another of application.
		// this is done this way because the calculation of each panel position
		// cannot be done with hidden panels
		if ( !this.__dashboard_is_active() ){
			this.m_pending_operations.push({ fn : this.__on_your_campaign_has_been_removed, args : [campaign_id] });
			return;
		}
		// add remove campaign event
		this.__add_event_to_queue({
				campaign_id : campaign_id,
				event_type : 'REMOVE'
			},
			"FIRST"
		);
		// remove item from campaign list selector.
		let $campaign_popup = $('#sup_dash_campaign_selector_popup');
		let $campaign_container = $campaign_popup.find('.campaignArea');
		$campaign_container.find('.campaignRowCheck[campaign_id*="'+ campaign_id +'"]').parent().remove();
	};

	/**
	 * It's a callback for a PubSub subscribe called
	 * when the user has been added to a campaign.
	 *
	 * @private
	 * @param  {String} campaign - The campaign identificator (eg. campaign@vcc)
	 * @memberOf SupervisorDashboard
	 */
	SupervisorDashboard.prototype.__on_campaign_has_been_assigned_to_you = function(campaign, campaign_number)
	{
		if (!this.is_rendered_dashboard()) {
			return;
		}
		// execution should be delayed if the user is in another application.
		// this is done this way because the calculation of each panel position
		// cannot be done with hidden panels
		if (!this.__dashboard_is_active()){
			this.m_pending_operations.push({fn: this.__on_campaign_has_been_assigned_to_you, args: [campaign]});
			return;
		}

		// calculo la cantidad de campanas que van a quedar reenderizas
		let total_rendered_campaigns = this.__get_total_campaigns_rendered() + campaign_number;
		// valido si la campana puede ser monitoreada, es decir si no ha pasado el numero maximo de campanas que el supervisor puede monitorear al mismo tiempo.
		let result = this.validate_number_of_campaigns_to_monitor(total_rendered_campaigns);

		if (result.success) {
			// Si la campana puede ser monitoreada, la encolo para que sea procesada (se agregue el panel, el tab, en la grilla del dashboard, etc).
			this.__add_event_to_queue({
				campaign_id: campaign.Id,
				event_type : 'ADD'
			});
		}
		// La agrego al selector de campanas (unckecked), sin importar que NO haya pasado el tope maximo monitoreables.
		this.__add_campaign_selector_node(campaign, false);
	};

	/**
	 * Start the process event queue timer
	 *
	 * @private
	 * @memberOf SupervisorDashboard
	 */
	SupervisorDashboard.prototype.__start_process_queue_timer = function() {
		const self = this;
		this.m_event_process_timer = window.setTimeout(function(){
			self.__process_events_queue_alt();
		}, QUEUE_PROCESS_INTERVAL_MS);
	};

	/**
	 * Trigger when a queue event is beign completed.
	 *
	 * @private
	 * @param {boolean} alters_campaign_amount - If current process alters the amount of campaigns (add or remove)
	 * @memberOf SupervisorDashboard
	 */
	SupervisorDashboard.prototype.__on_queue_beign_completed = function(alters_campaign_amount){
		var self = this;
		// if add or remove events was present on the queue
		// hide waiting and save config
		if (alters_campaign_amount){
			this.save_config(function(){
				self.hide_waiting();
			});
		}
		//else {
			//this.hide_waiting();
		//}
	};

	/**
	 * Get a campaign filter to send a request for a supervisor view
	 *
	 * @private
	 * @param {string} campaign_id - The campaign id.
	 * @return {object} - The campaign filter
	 */
	SupervisorDashboard.prototype.__get_campaign_filter = function( campaign_id )
	{
		var members_config = this.m_config.get_panel_config(campaign_id, 'sup_dash_members_panel');
		var show_offline = members_config ? members_config.options.show_offline : true;
		var show_kickouts = members_config ? members_config.options.show_kickouts : true;

		// use default realtime window if the user does not have config
		var summary_config = this.m_config.get_panel_config(campaign_id, 'sup_dash_summary_panel');
		var realtime_window = summary_config && summary_config.realtime_window ? summary_config.realtime_window : 'Last15Minutes';

		// campaign selector configuration ( account_id, attention_level_id )
		var campaign_config = this.get_campaign_selected_config(campaign_id);

		return {
			account           : !this.isAllegro(campaign_id) && campaign_config.account_id || null,
			attention_level   : campaign_config.al_id,
			page              : 0,
			realtimeWindow    : realtime_window,
			showOfflineAgents : show_offline,
			showKickouts      : show_kickouts,
			workgroups        : this.m_workgroups.get_campaign_workgroups(campaign_id)
		};
	};

	SupervisorDashboard.prototype.__cleanup_processed_event = function()
	{
		this.m_current_processed_event_queue = null; // ha finalizado de procesar el evento, entonces lo libero
	};

	SupervisorDashboard.prototype.__validate_expired_processed_event = function()
	{
		if (!this.m_current_processed_event_queue) {
			return;
		}
		//console.info("SupervisorDashboard.prototype.__validate_expired_processed_event - Validando si ha expirado el evento encolado. Event: " + this.m_current_processed_event_queue.uid + " - " + this.m_current_processed_event_queue.timestamp);
		let now = new Date();
		let difference_sec = (now - this.m_current_processed_event_queue.timestamp) / 1000;
		if (difference_sec >= 10) { // si pasaron N segundos realiza cleanup, para que continue con la cola de eventos (esto no deberia pasar, pero nos cubrimos)
			if (this.m_current_processed_event_queue.isBatch) {
				console.info("SupervisorDashboard.prototype.__validate_expired_processed_event - Los eventos en lote han expirado. Tiempo en segundos: " + difference_sec);
			}
			else{
				console.info("SupervisorDashboard.prototype.__validate_expired_processed_event - El evento '"  + this.m_current_processed_event_queue.uid + "' encolado ha expirado. Tiempo en segundos: " + difference_sec);
			}
			this.__cleanup_processed_event();
		}
	};

	SupervisorDashboard.prototype.__process_events_queue_alt = function()
	{
		const self = this;

		var on_process_completed = function(){
			self.__start_process_queue_timer();
		};

		// empty list, nothing to do.
		if ( this.m_event_queue.length === 0 ){
			return on_process_completed();
		}

		// Hasta que no termine el evento que se esta procesando, no puedo continuar con el siguiente evento encolado
		if (this.m_current_processed_event_queue) {
			this.__validate_expired_processed_event();
			return on_process_completed();
		}

		const start_cleanup_event = function(event) {
			self.__cleanup_processed_event();
			event.state = "RUNNING";
			self.m_current_processed_event_queue = event; // me quedo con el evento que se comienza a procesar (Puede ser en batch)
			return event;
		};

		// ADD and REMOVE are events alters the campaign amount,
		// this events take more time to process and the configuration needs to be saved.
		const process_alters_campaign_amount = this.m_event_queue.some(function(e){ return e.event_type === 'ADD' || e.event_type === 'REMOVE'; });

		////console.info("SupervisorDashboard.prototype.__process_events_queue - Total Events: '" + this.m_event_queue.length);

		// Primero tengo que procesar los eventos "REMOVE" por lo que busco el primero que aparezca en el array.
		let current_event;
		for (let i = 0; i < this.m_event_queue.length; i++) {
			let ev = this.m_event_queue[i];
			if (ev.event_type === 'REMOVE') {
				current_event = ev;
				this.m_event_queue.splice(i, 1);
				break;
			}
		}

		////console.info("SupervisorDashboard.prototype.__process_events_queue - After remove events: '" + this.m_event_queue.length);
		if (current_event) {
			current_event = start_cleanup_event(current_event);
			// call remove
			this.__remove_campaign(current_event.campaign_id, current_event.campaign, current_event, function (_) {
				self.__cleanup_processed_event();
				// this was the last event on the queue?
				if (self.m_event_queue.length === 0) {
					self.emit('QueueBeingCompleted', process_alters_campaign_amount);
					on_process_completed();
				}
				else {
					// more events? go and call myself again
					self.__process_events_queue_alt();
				}
			});
			return;
		}
		//console.info("SupervisorDashboard.prototype.__process_events_queue - Start processing the 'update' and 'add' events. Total Events: " + this.m_event_queue.length + ", datetime " + (new Date()).toUTCString());

		//Se procesaron todos los remove. Ahora proceso los refresh y add en batch
		//Convierto a map para procesar y Agrego campos necesarios (campaign & members filters) para obtener la informacion de las vistas.
	
		let campaign_events = {};
		let last_event;
		for (let index = 0; index < this.m_event_queue.length; index++) {
			let ev = this.m_event_queue[index];
			if (ev.event_type === 'ADD' || ev.event_type === 'REFRESH') {
				const campaign_id = ev.campaign_id;

				ev.filter = self.__get_campaign_filter(campaign_id);
				self.m_dash_members.initializePageByCampaign(campaign_id);
				ev.filterMembers = self.m_dash_members.get_members_filters(campaign_id);
				/*
					Procesar indicadores:
					- Si es refresh (Ya anda renderizada)
					- Si esta prendida la flag (m_fire_charts_refresh)
					- Cuando se selecciona por primera vez una campaign se pide refrescar, solo para la campanna especifica.
						o
					* Si es ADD   
				*/
				const refresh_charts = self.m_fire_charts_refresh && self.m_fire_charts_refresh.enable && (!self.m_fire_charts_refresh.campaign_id || self.m_fire_charts_refresh.campaign_id === campaign_id);
				
				if (((ev.event_type === 'REFRESH' && refresh_charts) ||  ev.event_type === 'ADD')) {
					ev.filterIndicators = self.m_dash_indicators.get_historical_data_filter(campaign_id);
				}

				campaign_events[campaign_id] = ev;
				last_event = ev;
			}
		}

		delete self.m_fire_charts_refresh;
		
		if (!last_event) {
			self.__cleanup_processed_event();
			on_process_completed();
			return;
		}

		this.m_event_queue = [];

		//Obtengo el ultimo elemento como referencia para el handle de tiempo de proceso y el force de 
		last_event.isBatch = true;
		start_cleanup_event(last_event);

		/*
		En una misma llamada, obtengo toda la informacion relacionada a las vistas de las campannas (alarmas, members, etc). 
		Tambien se obtienen los workgroups para ser actualizados.
		Todo esto porque se enviaban 3 request cada 3 seg al server y no es performante.
		*/
		let data = {
			campaigns : campaign_events,
			addWorkgroups : true,         // Flag para obtener los worgroups
		};

		self.m_model.getSupViewAndMembersByCampaigns(data, function (response, error) {
		
			// Check if we are good to continue.
			if (inConcert.app.m_hadErrors || (!response && !error)){
				on_process_completed();
				return;
			}

			let supervisor_campaign_views = response && response.supervisor_views;
			//console.info("SupervisorDashboard.prototype.__process_events_queue - Request completed, get campaigns info (view, members, alarms) Total Events: " + Object.keys(supervisor_campaign_views || {}).length + ", datetime " + (new Date()).toUTCString());
			
			//Actualizo los workgroups
			let campaign_workgroups = response && response.campaign_workgroups;
			if (campaign_workgroups) {
				self.m_workgroups.update_workgroups(campaign_workgroups);
			}

			if (error || !supervisor_campaign_views || supervisor_campaign_views.length === 0) {
				console.info("SupervisorDashboard.prototype.__process_events_queue " + format("<%=_T('Could not get supervisor info for campaigns, error: {0}')%>", error || ""));

				self.__show_default_error_message();
				self.__cleanup_processed_event();
				on_process_completed();
				return;
			}

			//Despues que tenga toda la info, proceso los add y refresh
			var process_add_refresh_events = function (sup_campaign_views) {
				if (!sup_campaign_views || Object.keys(sup_campaign_views).length === 0) {
					//Termino de procesar todas las vistas de las campannas
					//console.info("SupervisorDashboard.prototype.__process_events_queue - COMPLETED processing the 'update' and 'add' events. datetime " + (new Date()).toUTCString());
					self.__cleanup_processed_event();
					self.emit('QueueBeingCompleted', process_alters_campaign_amount);
					on_process_completed();
				}

				const callback_render_or_refresh = function(campaign_id) {
					//Se elimina la info de la campanna, para procesar la siguinte
					delete sup_campaign_views[campaign_id];
					process_add_refresh_events(sup_campaign_views);
				};

				for (const campaign_id in sup_campaign_views) {
					if (Object.hasOwnProperty.call(sup_campaign_views, campaign_id)) {

						let current_event = campaign_events[campaign_id];
						let current_supervisor_campaign_view = sup_campaign_views[campaign_id];
						if (current_supervisor_campaign_view.result && current_supervisor_campaign_view.result !== "OK") {
							switch(current_supervisor_campaign_view.result){
								case 'NOT_LOADED':
									const hide_close_button = self.__get_total_campaigns_rendered() === 0;
									self.show_info_message(format("<%=_T('Campaign: {0} is not loaded yet, be patient')%>", campaign_id), hide_close_button);
									break;
								case 'NOT_FOUND':
									// let this error pass through
									break;
								default:
									self.show_error_message(format("<%=_T('Could not get supervisor info for campaign: {0}, error: {1}')%>", campaign_id, error));
									break;
							}
				
							return callback_render_or_refresh(campaign_id, current_event.event_type);
						}

						//Renderizo
						if (current_event.event_type === "ADD") {
							//Renderizo pasandole la info de la vista para la campanna y el evento.
							let view = {
								supervisor_view: current_supervisor_campaign_view
							};

							//Para los add se actualizan tambien los indicadores.
							self.__render_campaign(campaign_id, undefined, current_event, function (_) {
								callback_render_or_refresh(campaign_id, "ADD");
							}, view);
						}
						else {
							//Refresco vista
							/* jshint loopfunc: true */
							self.__refresh_campaign_from_view(campaign_id, current_event, function (_) {
								//Refresco Indicadores si se precisa (cada 60 seg)
								if (current_supervisor_campaign_view.indicatorsData) {
									self.__start_refresh_indicator(campaign_id, current_supervisor_campaign_view.indicatorsData, function() {
										callback_render_or_refresh(campaign_id, "REFRESH");
									});
									return;
								}
								callback_render_or_refresh(campaign_id, "REFRESH");
							}, current_supervisor_campaign_view);
						}
					}
				}
			};

			process_add_refresh_events(supervisor_campaign_views);

		});
		
	};

	SupervisorDashboard.prototype.__start_refresh_indicator = function(campaign_id, indicators, callback = $.noop)
	{
		// Check if we are good to continue.
		if (inConcert.app.m_hadErrors){
			callback(false);
			return;
		}
		if (!indicators) {
			return callback(true);
		}

		if (indicators.result && indicators.result !== "OK") {
			//Error
			let result = indicators.result;
			let reason = indicators.reason || result;
			/*
			 - codigo de error que lo retorna la rocka "reporter" cuando se esta ejecutando un reporte.
			 - Si es el caso se busca la info nuevamente y se refresca.
			 - Otro tipo de error se deja log
			*/
			if (result === "UNABLE TO RETRIEVE CACHE INFO - ANOTHER REPORTS IS RUNNING") {
				this.m_dash_indicators.refresh(campaign_id);
				
			}
			else{
				console.warn("Error refreshing indicators, reason: " + reason);
			}
		}
		else{
			this.m_dash_indicators.refresh(campaign_id, undefined, indicators);
		}
		callback();
	};

	/**
	 * Processes the event queue one by one, until the queue is empty.
	 * Schedules a new call to the same function within a timeout (defined as QUEUE_PROCESS_INTERVAL_MS)
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.__process_events_queue = function()
	{
		const self = this;

		var on_process_completed = function(){
			self.__start_process_queue_timer();
		};

		// empty list, nothing to do.
		if ( this.m_event_queue.length === 0 ){
			return on_process_completed();
		}

		// Hasta que no termine el evento que se esta procesando, no puedo continuar con el siguiente evento encolado
		if (this.m_current_processed_event_queue) {
			this.__validate_expired_processed_event();
			return on_process_completed();
		}

		// ADD and REMOVE are events alters the campaign amount,
		// this events take more time to process and the configuration needs to be saved.
		const process_alters_campaign_amount = this.m_event_queue.some(function(e){ return e.event_type === 'ADD' || e.event_type === 'REMOVE'; });

		////console.info("SupervisorDashboard.prototype.__process_events_queue - Total Events: '" + this.m_event_queue.length);

		// Primero tengo que procesar los eventos "REMOVE" por lo que busco el primero que aparezca en el array.
		let current_event;
		for (let i = 0; i < this.m_event_queue.length; i++) {
			let ev = this.m_event_queue[i];
			if (ev.event_type === 'REMOVE') {
				current_event = ev;
				this.m_event_queue.splice(i, 1);
				break;
			}
		}
		////console.info("SupervisorDashboard.prototype.__process_events_queue - After remove events: '" + this.m_event_queue.length);
		if (!current_event) {
			// Si no existe ningun evento de tipo "REMOVE", obtengo el primer elemento de la lista de eventos encolados
			current_event = this.m_event_queue.shift();
			//// console.info("SupervisorDashboard.prototype.__process_events_queue - After add o refresh events: '" + this.m_event_queue.length);
		}
		////console.info("SupervisorDashboard.prototype.__process_events_queue - Event to processed: '" + current_event.event_type + "', on campaign: " + current_event.campaign_id);

		const mapping = {
			'ADD'     : this.__render_campaign,
			'REFRESH' : this.__refresh_campaign,
			'REMOVE'  : this.__remove_campaign
		};

		this.__cleanup_processed_event();
		current_event.state = "RUNNING";
		this.m_current_processed_event_queue = current_event; // me quedo con el evento que se comienza a procesar
		self.m_hnd_current_processed_event_queue = window.setTimeout(function() {
			self.__cleanup_processed_event();
		}, 5000);

		// call with new args
		mapping[current_event.event_type].call(this, current_event.campaign_id, current_event.campaign, current_event, function(_)
		{
			self.__cleanup_processed_event();
			// this was the last event on the queue?
			if (self.m_event_queue.length === 0) {
				self.emit('QueueBeingCompleted', process_alters_campaign_amount);
				on_process_completed();
			}
			else {
				// more events? go and call myself again
				self.__process_events_queue();
			}
		});
	};

	/**
	 * Add an event to the end of the queue.
	 *
	 * @private
	 * @param {Object} new_event             - The event to be added.
	 * @param {String} new_event.event_type  - The event type (ADD,REMOVE,REFRESH)
	 * @param {String} new_event.campaign_id - The campaign id.
	 */
	SupervisorDashboard.prototype.__add_event_to_queue = function(new_event, position)
	{
		new_event.uid = (new_event.campaign_id + " - " + new_event.event_type).toUpperCase();
		new_event.state = "QUEUED";
		new_event.timestamp = new Date();

		let already_exist = false;
		for (let i = 0; i < this.m_event_queue.length; i++) {
			let current_event = this.m_event_queue[i];
			if (current_event.uid === new_event.uid) {
				this.m_event_queue[i] = new_event;
				already_exist = true;
				break;
			}
		}

		if (already_exist) {
			return false;
		}

		if (position === "FIRST") {
			this.m_event_queue.unshift(new_event);
		}
		else {
			this.m_event_queue.push(new_event);
		}
		return true;
	};

	/**
	 * Will make and return attention level selector node.
	 *
	 * @private
	 * @param  {Object} campaign        - The campaign object.
	 * @param  {String} account_id      - The account string.
	 * @param  {Object} attention level - The al object.
	 * @param  {Boolean} checked        - If the al checkbox will be checked or not
	 */
	SupervisorDashboard.prototype.__$make_attention_level_selector_node = function( campaign, account_id, al )
	{
		var self = this;

		var campaign_id = campaign.Id;
		var al_id = al.id;
		var al_def = campaign.getAttentionLevel(al_id);
		var normalized_account_id = this.get_normalized_id(account_id);
		var normalized_al_id = this.get_normalized_id(al_id);

		var al_item_view = {
			campaign_id             : campaign_id,
			account_id              : account_id,
			attention_level_id      : al_id,
			normalized_id           : normalized_account_id + '_' + normalized_al_id,
			normalized_account_id   : normalized_account_id,
			attention_level_name    : al_def.name
		};

		var $attention_level = $(Mustache.render(this.m_templates.popup_attention_level_item, al_item_view));

		var event_selector = '[attention_level_selector]';
		$attention_level
			.off('click', event_selector)
			.on('click', event_selector, function()
			{
				var $radio = $(this);
				var account_id = $radio.attr('account_id');
				var normalized_account_id = self.get_normalized_id( account_id );
				var $account_radio = $('#sup_dash_account_selector_item_' + normalized_account_id);
				self.__mark_unmark_campaign_account( $account_radio, true /* mark */, 'attention_level' );
				$radio.prop('checked', true);
				$radio.parent().removeClass('selected').addClass('selected');
				var campaign_id = $radio.attr('campaign_id');
				self.__set_selected_campaign_account_al( true, campaign_id );
			});

		var campaign_config = this.get_campaign_selected_config(campaign_id);
		if ( al_id === campaign_config.al_id ) {
			$attention_level.removeClass('selected').addClass('selected');
			var $attention_level_radio = $attention_level.find('[attention_level_selector]');
			$attention_level_radio.prop('checked', true);
		}

		return $attention_level;
	};

	SupervisorDashboard.prototype.__mark_unmark_campaign_account = function( $radio, mark, from )
	{
		var unmark_account_als = function( $assigned_als )
		{
			$.each($assigned_als, function( _, assigned_al )
			{
				var $assigned_al = $(assigned_al);
				$assigned_al.removeClass('selected');
				var $al_selector = $assigned_al.find('[attention_level_selector]');
				$al_selector.prop('checked', false);
			});
		};

		var hide_account_als = function( $current_account_selector, $current_als_container )
		{
			$current_account_selector.prop('checked', false);
			$current_als_container.hide();
		};

		var $current_account = $radio.parent();
		var $all_campaign_accounts = $current_account.parent().find('[account_item_type]');

		$.each($all_campaign_accounts, function(_, acc)
		{
			var $other_account = $(acc);
			$other_account.removeClass('selected childSelected');
			var $other_account_selector = $other_account.find('[account_selector]');
			var $als_container = $other_account.find('[attention_levels_list]');
			var $assigned_als = $als_container.find('li.attentionLevelListItem');
			hide_account_als( $other_account_selector, $als_container );
			unmark_account_als( $assigned_als );
		});

		if ( mark ) {
			$current_account.removeClass('selected').addClass('selected');
			if ( from === 'attention_level' ) {
				$current_account.removeClass('childSelected').addClass('childSelected');
			}
			$current_account.find('[attention_levels_list]').show();
			$radio.prop('checked', true);
		}
	};

	/**
	 * Will make and return account selector node.
	 *
	 * @private
	 * @param  {Object} campaign        - The campaign object.
	 * @param  {Object} account         - The account object.
	 */
	SupervisorDashboard.prototype.__$make_account_selector_node = function( campaign, account, $accounts_list )
	{
		var self = this;

		var account_id = account.id;
		var normalized_account_id = this.get_normalized_id(account_id);
		var account_selector_item_id = 'sup_dash_account_selector_item_' + normalized_account_id;
		if ( $('#' + account_selector_item_id).length > 0 ) {
			return;
		}

		var account_name = campaign.getAccountDescription(account);
		var account_desc = campaign.isAllegro() ? "<%=_T('Phone Account')%>" : account_name;

		var campaign_id = campaign.Id;
		var account_item_view = {
			campaign_id             : campaign_id,
			normailized_campaign_id : this.get_normalized_id(campaign_id),
			account_id              : account_id,
			id                      : normalized_account_id,
			account_name            : account_name,
			account_desc            : account_desc,
			channel                 : account.proxyType
		};

		var $account = $(Mustache.render(this.m_templates.popup_account_item, account_item_view));
		$account.addClass("accountItem" + account.proxyType.toUpperCase());
		$accounts_list.append($account);

		var campaign_config = this.get_campaign_selected_config(campaign_id);
		var event_selector = '[id="' + account_selector_item_id + '"]';
		$account
			.off('click', event_selector)
			.on('click', event_selector, function()
			{
				var $radio = $(this);
				self.__mark_unmark_campaign_account( $radio, true /* mark */ );
				var campaign_id = $radio.attr('campaign_id');
				self.__set_selected_campaign_account_al( true, campaign_id );
			});
		if ( campaign_config.account_id === account_id ) {
			$account.addClass('selected');
			if ( campaign_config.al_id ) {
				$account.addClass('childSelected');
			}
			var $account_radio = $account.find('[account_selector]');
			$account_radio.prop('checked', true).trigger('click');
		}

		var $als_list = $account.find('[attention_levels_list]');
		$.each(account.AttentionLevels || {}, function(al_id)
		{
			var al_def = campaign.getAttentionLevel(al_id);
			var $al = self.__$make_attention_level_selector_node(campaign, account_id, al_def);
			$als_list.append($al);
		});

		return $account;
	};

	/**
	 * Will make and return campaign all accounts selector node.
	 *
	 * @private
	 * @param  {Object} campaign        - The campaign object.
	 */
	SupervisorDashboard.prototype.__$make_header_campaign_accounts_selector_node = function( campaign )
	{
		var self = this;

		var campaign_id = campaign.Id;
		var normalized_campaign_id = this.get_normalized_id(campaign_id);

		var header_all_accounts_selector_item_id = 'sup_dash_header_campaign_accounts_selector_item_' + normalized_campaign_id;
		if ( $('#' + header_all_accounts_selector_item_id).length > 0 ) {
			return;
		}

		var header_campaign_accounts_item_view = {
			campaign_id  : campaign_id,
			id           : normalized_campaign_id,
			header_title : "<%=_T('Campaign (All Accounts)')%>"
		};

		var $header = $(Mustache.render(this.m_templates.popup_header_campaign_accounts_item, header_campaign_accounts_item_view));

		var event_selector = '[id="' + header_all_accounts_selector_item_id + '"]';
		$header
			.off('click', event_selector)
			.on('click', event_selector, function()
			{
				var $radio = $(this);
				var $header = $radio.parent();
				var marked = $header.hasClass('selected');
				var $marked_account = $header.parent().find('[account_item_type="account"] > [account_selector]:checked');
				$.each($marked_account, function(_, acc_radio)
				{
					var $acc_radio = $(acc_radio);
					self.__mark_unmark_campaign_account( $acc_radio, false /* mark */ );
				});
				if ( !marked ) {
					$header.addClass('selected');
					$radio.prop('checked', true);
				}
				var campaign_id = $radio.attr('campaign_id');
				self.__set_selected_campaign_account_al( true, campaign_id );
			});

		var campaign_config = this.get_campaign_selected_config(campaign_id);
		var marked_account_header = campaign_config.account_id === undefined;
		if ( marked_account_header ) {
			$header.addClass('selected');
			var $header_radio = $header.find('[account_selector]');
			$header_radio.prop('checked', true);
		}

		return $header;
	};

	SupervisorDashboard.prototype.get_selected_campaign_account_al = function( campaign_id, force)
	{
		var config = {};
		var normalized_campaign_id = this.get_normalized_id(campaign_id);
		var $campaign = $('#sup_dash_campaign_selector_item_' + normalized_campaign_id);
		if ( !force && !$campaign.prop('checked') ) {
			return config;
		}

		var is_allegro_campaign = false;
		var campaign = inConcert.app.getCampaign(campaign_id);
		if (campaign) {
			is_allegro_campaign = campaign.isAllegro();
		}

		var $account_selector;
		// asigna la configuracion de la cuenta, ya sea de cuentas de integracion con allegro o sin integracion
		var set_account_config = function(account_id) {
			config.campaign_id = campaign_id;
			config.account_id = account_id || $account_selector.attr('account_id');
			config.account_name = $account_selector.attr('account_name');
			config.channel = $account_selector.attr('channel');
		};

		const $selector = $(`#sup_dash_campaign_accounts_config_${normalized_campaign_id}`);

		var $campaign_accounts_container = $selector.find('[accounts_list]');
		config.campaign_id = campaign_id;
		if ( is_allegro_campaign ) {
			// Si la campana tiene integracion con allegro, si o si agregar la configuracion de la unica cuenta de la campana
			$account_selector = $campaign_accounts_container.find('[account_item_type="account"] > [account_selector]');
			
			if ( $account_selector.length > 0 ) {
				var is_checked_call_account = $account_selector.prop('checked');
				if (!is_checked_call_account) {
					// la dejo siempre marcada ya que es la unica
					$account_selector.prop('checked', true);
				}
				set_account_config();
			}
			else if((campaign.get_accounts() || []).length > 0) //Caso en que no se ha agregado el markup de accounts aun, dado que se agregan una vez se renderice
			{
				set_account_config(campaign.get_accounts()[0].id);
			}
		}
		else {
			// Si es una campana sin integracion con allegro, agrega los valores seleccionados en el campaign selector (campaign, account, al)
			$account_selector = $campaign_accounts_container.find('[account_item_type] > [account_selector]:checked');
			var account_item_type = $account_selector.length > 0 ? $account_selector.parent().attr('account_item_type') :"";
			if ( account_item_type === 'account') {
				set_account_config();
					var $als_container = $account_selector.parent().find('[attention_levels_list]');
					var $al_selector = $als_container.find('[attention_level_selector]:checked');
					if ( $al_selector.length > 0 ) {
						config.al_id = $al_selector.attr('attention_level_id');
						config.al_name = $al_selector.attr('attention_level_name');
					}
			}
		}

		return config;
	};

	SupervisorDashboard.prototype.__set_selected_campaign_account_al = function( add, campaign_id, selection_config)
	{
		var action;
		if ( !add ) {
			action = 'REMOVE';
		}
		else {
			var $campaign_node = this.get_campaign_node(campaign_id);
			action = $campaign_node.length === 0 ? 'ADD' : 'REFRESH';
		}
		var config = { action : action };
		if ( action !== 'REMOVE' ) {
			selection_config = selection_config || this.get_selected_campaign_account_al(campaign_id);
			config = $.extend(config, selection_config);
		}
		this.m_campaigns_being_changed[ campaign_id ] = config;
	};

	SupervisorDashboard.prototype.__make_campaign_accounts_selector = function (campaign, marked_campaign, $popup_campaign_accounts, $selector_campaigns) {
		// Limpiar contenedor del selector de cuentas por campannas
		const campaign_id = campaign.Id;
		const normalized_id = this.get_normalized_id(campaign_id);
		const self = this;
		$popup_campaign_accounts = $popup_campaign_accounts || $(`#sup_dash_campaign_accounts_config_${normalized_id}`);
		$popup_campaign_accounts.html("");
		$selector_campaigns = $selector_campaigns || $("#sup_dash_campaign_selector_popup").find(`li[selector_campaign_id='${campaign_id}']`);
		// Creo el markup de las cuentas/als para la campanna y lo agrego al contenedor
		const $accounts_selector = $(this.m_templates.popup_campaigns_accounts_container);
		$popup_campaign_accounts.append($accounts_selector);

		const $accounts_list = $accounts_selector.find('[accounts_list]');

		const is_allegro_campaign = campaign.isAllegro();
		let $allegro_account;

		// agrega una cuenta al tree
		var add = function( account )
		{
			if (!is_allegro_campaign) {
				const $all_campaign = self.__$make_header_campaign_accounts_selector_node(campaign);
				if ( $all_campaign ) {
					$accounts_list.append($all_campaign);
				}
			}
			const $account = self.__$make_account_selector_node(campaign, account, $accounts_list);
			if ( $account ) {
				if (is_allegro_campaign) {
					$allegro_account = $account;
				}
			}
		};

		// recorro la lista de cuentas y las agrego
		let accounts = campaign.get_accounts();

		this.enable_disable_btn_campaign_accounts_selector(campaign_id);
		
		$.each(accounts || [], function(_, account) { add(account); });
		// append campaign new node
		if ( marked_campaign ) {
			$selector_campaigns.find('.campaignRowCheck').trigger('change');
			if ($allegro_account) {
				// si es una allegro account, tengo que dejarla marcada
				const $account_chk = $allegro_account.find('input[account_selector]');
				if ($account_chk.length > 0) {
					$account_chk.trigger('click');
				}
			}
		}

		const $ok = $('#sup_dash_campaign_selector_ok');
		this.pubsub[campaign_id] = this.pubsub[campaign_id] || {};
		//subscribe to added new account
		this.pubsub[campaign_id].added_account = inConcert.app.messagebus.subscribe(
			'/application/campaign/' + campaign_id + '/account/added/',
			function(new_account)
			{
				// Agrego la cuenta en el tree selector
				add(new_account);
				self.__set_selected_campaign_account_al( true, campaign_id );
				self.enable_disable_btn_campaign_accounts_selector(campaign_id);
				$ok.trigger('click');
			});

		//subscribe to updated account
		this.pubsub[campaign_id].updated_account = inConcert.app.messagebus.subscribe(
			'/application/campaign/' + campaign_id + '/account/updated/',
			function(account)
			{
				// Cuando se actualiza una cuenta valido si han cambiados los ALs de la misma, y realizo las actualizaciones pertinentes.
				let account_id = account.id;
				let normalized_account_id = self.get_normalized_id( account_id );
				const $account_radio = $('#sup_dash_account_selector_item_' + normalized_account_id);
				const $account = $account_radio.parent();
				const $als_list = $account.find('[attention_levels_list]');

				// 1 - Recorro la nueva lista de attention levels, buscando nuevos ALs asignados en la cuenta.
				//     Si encuentro alguno nuevo AL, lo agrego al tree bajo la cuenta.
				let account_als = account.AttentionLevels || {};
				$.each(account_als, function(al_id)
				{
					let normalized_al_id = self.get_normalized_id(al_id);
					let dom_al_id = normalized_account_id + '_' + normalized_al_id;
					let $al = $('#sup_dash_attention_level_selector_item_' + dom_al_id);

					if ( $al.length === 0 ) {
						let al_def = campaign.getAttentionLevel(al_id);
						$al = self.__$make_attention_level_selector_node(campaign, account_id, al_def);
						$als_list.append($al);
					}
				});

				// 2 - Recorro en el DOM bajo la cuenta la lista de ALs existentes, y valido si es que existen en la nueva lista de ALs de la cuenta/campana.
				//     Si encuentra alguno que existe en el DOM y no existe en asignacion de ALs de la cuenta, lo agrega a una lisat temporal para luego removerlo del DOM (*)
				let als_to_remove = {};
				$.each($als_list.find('li[attention_level_item_type="al"]'), function(_, al)
				{
					const $al = $(al);
					const $al_radio = $al.find('[attention_level_id]');
					let attention_level_id = $al_radio.attr('attention_level_id');
					if ( !account_als[attention_level_id] ) {
						als_to_remove[attention_level_id] = true;
					}
				});

				// 3 - Recorre la lista temporal de ALs a eliminar (*) si es que existe alguno a eliminar.
				//     Cuando encuentra alguno a eliminar, lo remueve valida si el mismo se encuentra marcado/seleccionado.
				//     Si se encuentra seleccionado, marca la cuenta padre en el tree como seleccionada, y luego ejecuta el codigo de configuracion (**).
				let has_change_config = false;
				$.each(als_to_remove, function(al_to_remove)
				{
					let normalized_al_id = self.get_normalized_id(al_to_remove);
					let dom_al_id = normalized_account_id + '_' + normalized_al_id;
					const $al_radio = $('#sup_dash_attention_level_selector_item_' + dom_al_id);
					var is_selected_al = $al_radio.prop('checked');
					$al_radio.parent().remove();
					if ( is_selected_al ) {
						$account_radio.trigger('click');
						has_change_config = true;
					}
				});
				if ( has_change_config ) {
					// (**) codigo de configuracion
					self.__set_selected_campaign_account_al( true, campaign_id );
					$ok.trigger('click');
				}
			});

		//subscribe to removed account
		this.pubsub[campaign_id].removed_account = inConcert.app.messagebus.subscribe(
			'/application/campaign/' + campaign_id + '/account/removed/',
			function(account_id)
			{
				// Si se ha removido la cuenta, tengo que validar si la misma se encuentra seleccionada, y si se encuentra seleccionada debe dejar seleccionada la campana.
				let normalized_account_id = self.get_normalized_id(account_id);
				let account_selector_item_id = 'sup_dash_account_selector_item_' + normalized_account_id;
				const $account_radio = $accounts_list.find('[id="' + account_selector_item_id + '"]');

				if ($account_radio.length > 0) {
					if ($account_radio.prop('checked')) {
						const $all_accounts_radio = $accounts_list.find('[account_item_type="header"] > [account_selector]');
						if ($all_accounts_radio.length > 0 ) {
							$all_accounts_radio.trigger('click');
						}
					}
					self.__set_selected_campaign_account_al( true, campaign_id );
					self.enable_disable_btn_campaign_accounts_selector(campaign_id);
					$ok.trigger('click');
					$account_radio.parent().remove();
				}
			});
	};

	/**
	 * Will add a campaign to the campaigns selector popup.
	 *
	 * @private
	 * @param  {Object} campaign - The campaign object.
	 * @param  {Boolean} marked_campaign - If the campaign checkbox will be checked or not
	 */
	SupervisorDashboard.prototype.__add_campaign_selector_node = function( campaign, marked_campaign )
	{
		const $campaign_selector_list = $('#sup_dash_campaign_selector_list');
		let campaign_id = campaign.Id;
		const normalized_campaign_id = this.get_normalized_id(campaign_id);
		let $current_selector_campaign = $campaign_selector_list.find("[selector_campaign_id='" + campaign_id + "']");
		const $popup_campaign_accounts = $(`#sup_dash_campaign_accounts_config_${normalized_campaign_id}`);
		if ($current_selector_campaign.length > 0) { // if already exists campaign node
			
			const $chk = $('#sup_dash_campaign_selector_item_' + normalized_campaign_id);
			if (marked_campaign) {
				if (!$chk.prop('checked')) {
					$chk.prop('checked', true);
				}
			}
			else {
				if ($chk.prop('checked')) {
					$chk.prop('checked', false);
				}
			}
			if ($popup_campaign_accounts.html() === "") {
				this.__make_campaign_accounts_selector(campaign, marked_campaign, $popup_campaign_accounts);
			}
			return false;
		}

		// append campaign new node
		const normalized_id = this.get_normalized_id(campaign_id);
		const campaign_item_view = {
			campaign_id  : campaign_id,
			id           : normalized_id,
			campaign_name: campaign.getDescription(),
			checked      : marked_campaign
		};

		const $new_selector_campaign = $(Mustache.render(this.m_templates.popup_campaign_item, campaign_item_view));
		$campaign_selector_list.append($new_selector_campaign);
		this.__make_campaign_accounts_selector(campaign, marked_campaign, $popup_campaign_accounts, $new_selector_campaign);
		return true;
	};

	SupervisorDashboard.prototype.full_screen_goto_tab_campaign = function (tab_id, on_changed = $.noop)
	{
		if (!this.m_full_screen) {
			return;
		}
		this.m_full_screen.goto_tab_campaign(tab_id, on_changed = $.noop);
	};

	SupervisorDashboard.prototype.full_screen_update_by_summary = function (info)
	{
		if (!this.m_full_screen) {
			return;
		}
		this.m_full_screen.update_by_summary(info);
	};

	SupervisorDashboard.prototype.full_screen_update_by_members = function (info)
	{
		if (!this.m_full_screen) {
			return;
		}
		this.m_full_screen.update_by_members(info);
	};

	SupervisorDashboard.prototype.show_alarm = function ( $node, alarm_info, calculated_current_value )
	{
		alarm_info = alarm_info || {};
		// (ALARMS) Marco/desmarco el indicador como alarmado.
		var alarm = this.get_notification_alarm_by_indicator({
			indicator_id : alarm_info.indicator_id,
			campaign_id  : alarm_info.campaign_id,
			account_id   : alarm_info.account_id,
			al_id        : alarm_info.al_id,
			agentId      : alarm_info.agentId, // opcional (solo para los request que van contra el panel de members)
			timestamp      : alarm_info.timestamp	//Opcional. Solo para alarmas de kickout
		});

		$node.attr('title', '')
			.removeClass('alarmed')
			.data('alarm', {});

		if (alarm) {
			var tooltip;
			if (calculated_current_value) {
				var alarm_message = this.get_notification_alarm_message(alarm, calculated_current_value);
				tooltip = alarm_message ? alarm_message.tooltip : '';
			}
			else {
				tooltip = alarm.tooltip || '';
			}
			$node.attr('title', tooltip)
				.addClass('alarmed')
				.data('alarm', alarm);
			return true;
		}

		return false;
	};

	SupervisorDashboard.prototype.members_change_status_filter = function (campaign_id, status, order, must_clean_member_text)
	{
		this.m_dash_members.change_status_filter(campaign_id, status, order, must_clean_member_text);
	};

	SupervisorDashboard.prototype.members_change_member_filter = function (campaign_id, member)
	{
		this.m_dash_members.change_member_filter(campaign_id, member);
	};

	SupervisorDashboard.prototype.validate_number_of_campaigns_to_monitor = function(total_rendered_campaigns)
	{
		let selected_campaigns = total_rendered_campaigns || $('#sup_dash_campaign_selector_popup .campaignRowCheck:checkbox').filter(':checked').length;
		let result = {
			success: true,
			selected_campaigns: selected_campaigns,
			reason: null
		};
		if (result.selected_campaigns > MAXIMUM_NUMBER_OF_MONITORABLE_CAMPAIGNS) {
			//console.warn("SupervisorDashboard.prototype.validate_number_of_campaigns_to_monitor - Overflow monitoring campaigns. Total campaigns: " + result.selected_campaigns);
			result.success = false;
			result.reason = "OVERFLOW";
		}
		else if (selected_campaigns === 0) {
			//console.warn("SupervisorDashboard.prototype.validate_number_of_campaigns_to_monitor - You have to select at least one campaign.");
			result.success = false;
			result.reason = "EMPTY";
		}
		return result;
	};

	/**
	 * Build the campaign list for add/remove campaigns from the dashboard.
	 *
	 * @private
	 * @param {Array} campaigns - A list of campaigns
	 */
	SupervisorDashboard.prototype.__build_campaign_selector = function(campaigns)
	{
		const self = this;

		let $campaign_popup = $('#sup_dash_campaign_selector_popup');
		// first time? Create it
		if ($campaign_popup.length === 0){
			$campaign_popup = $(Mustache.render(this.m_templates.popup_campaigns,
				{
					// translated messages
					title     : "<%=_T('Campaigns Configuration')%>",
					ok_message: "<%=_T('Refresh')%>"
				}
			));
			// append popup to main supervisor view
			$('#supervisor_main').append($campaign_popup);
		}

		var validate_number_of_campaigns_to_monitor = function()
		{
			let result = self.validate_number_of_campaigns_to_monitor();
			if (result.success) {
				return result;
			}
			if (result.reason === "OVERFLOW") {
				inConcert.app.getMainWindow().showBannerMessage({
					message: format("<%=_T('You can not monitor more than {0} campaigns.')%>", MAXIMUM_NUMBER_OF_MONITORABLE_CAMPAIGNS),
					timeout: 5000
				});
			}
			// EMPTY
			return result;
		};

		let $ok = $('#sup_dash_campaign_selector_ok');
		// event handling when the campaigns checkboxes changes
		const event_selector = '.campaignRowCheck';
		$campaign_popup
			.off('change', event_selector)
			.on('change', event_selector, function()
			{
				$ok.prop('disabled', false);
				const $checkbox = $(this);
				let campaign_id = $checkbox.attr('campaign_id');
				let result = validate_number_of_campaigns_to_monitor();
				// update ok button status
				if (!result.success) {
					if (result.reason === "OVERFLOW") {
						// NO esta permitido seleccionar la campana
						$checkbox.prop('checked', false);
					}
					else if (result.reason === "EMPTY") {
						// Al menos tiene que quedar una campana seleccionada
						$checkbox.prop('checked', true);
					}
					else {
						throw("SupervisorDashboard.prototype.__build_campaign_selector - Unrecognized result.reason: " + result.reason);
					}
					return;
				}
				// SI esta permitido seleccionar la campana
				let is_marked_campaign = $checkbox.prop('checked');
				self.__set_selected_campaign_account_al( is_marked_campaign, campaign_id );
			});

		let $campaign_container = $campaign_popup.find('.campaignArea');
		// add every campaign to the list
		campaigns.forEach(function(campaign){
			let campaign_node = $campaign_container.find('.campaignRowCheck[campaign_id="'+ campaign.Id +'"]');
			if ( campaign_node.length === 0 ) {
				// check/uncheck campaign checkbox based on configuration
				let campaign_config = self.get_stored_campaign_config(campaign.Id);
				let checked = campaign_config && campaign_config.visible;
				let marked = checked !== undefined ? checked : true;
				self.__add_campaign_selector_node(campaign, marked);
			}
		});

		// OK Button behaviour
		$ok.unbind('click').bind('click', function()
		{
			let result = validate_number_of_campaigns_to_monitor();
			if (!result.success) {
				return;
			}
			// hide popup
			$campaign_popup.unbind('clickoutside').hide();
			$('[action_id="campaign_selector"]').removeClass('menu-show');
			let campaigns_being_changed_by_actions = {
				ADD    : {},
				REFRESH: {},
				REMOVE : {}
			};
			for (let campaign_id in self.m_campaigns_being_changed) {
				if (self.m_campaigns_being_changed.hasOwnProperty(campaign_id)) {
					let config = self.m_campaigns_being_changed[campaign_id];
					if (!campaigns_being_changed_by_actions[config.action]) {
						throw("Unrecognized action: " + config.action);
					}
					campaigns_being_changed_by_actions[config.action][campaign_id] = config;
				}
			}
			self.m_campaigns_being_changed = {};

			var execute_actions = function(list, position)
			{
				for (let campaign_id in list) {
					if (list.hasOwnProperty(campaign_id)) {
						let config = list[campaign_id];
						self.__add_event_to_queue({
								campaign_id: campaign_id,
								config     : config,
								event_type : config.action,
								refresh_indicators : config.action === "REFRESH",
								refresh_by_change_config: true
							},
							position
						);
						self.save_config();
						// panels detail
						self.m_dash_summary.on_change_config_selector(campaign_id);
						self.m_dash_members.on_change_config_selector(campaign_id);

						// interactions
						let current_view = self.get_current_view_id(campaign_id);

						switch(current_view)
						{
							case 'dashboard_config':
								self.m_dash_config.on_change_config_selector(campaign_id);
							break;
							case 'monitoring':
								self.m_dash_monitor.on_change_config_selector(campaign_id);
							break;
							case 'appointment':
								self.m_dash_appointment.on_change_config_selector(campaign_id);
							break;
							case 'archive':
								self.m_dash_archive.on_change_config_selector(campaign_id);
							break;
							case 'queued':
								self.m_dash_queued.on_change_config_selector(campaign_id);
							break;
							case 'waiting':
								self.m_dash_waiting.on_change_config_selector(campaign_id);
							break;
							case 'abandoned':
								self.m_dash_abandoned.on_change_config_selector(campaign_id);
							break;
							case 'bot_interactions':
								self.m_dash_bot_interactions.on_change_config_selector(campaign_id);
							break;
							case 'global_search':
							// como el global search es algo independiente, y funciona como los demas panels del supervisor
							// lo actualizo desde la navegacion.
								var navigation = self.get_navigation(campaign_id);
								if (navigation) {
									var folder_id = navigation.get_selected_folder_id();
									if (folder_id === 'global_search') {
										navigation.apply_filter_labels();
									}
								}
							break;
						}
					}
				}
			};

			self.m_event_queue = [];
			console.info("SupervisorDashboard.prototype.__build_campaign_selector - Queued REMOVE & REFRESH Campaign events.");
			execute_actions(campaigns_being_changed_by_actions.REMOVE, "FIRST"); // Para que primero encole y procese los REMOVE DE CAMPANAS.
			execute_actions(campaigns_being_changed_by_actions.REFRESH);

			var async_queued_events;
			let hnd_queued_events = null;
			// funcion asincronica, para que ejecute los eventos de tipo ADD despues que haya termiando de procesar los remove de las campanas.
			async_queued_events = function() {
				if (hnd_queued_events) {
					window.clearTimeout(hnd_queued_events);
					hnd_queued_events = null;
				}
				if (self.m_event_queue.length === 0) {
					console.info("SupervisorDashboard.prototype.__build_campaign_selector - Queued ADD Campaign events.");
					execute_actions(campaigns_being_changed_by_actions.ADD);
					return;
				}
				else {
					// continuo esperando a que se termine el current event.
					self.__validate_expired_processed_event();
				}
				hnd_queued_events = window.setTimeout(function() {
					async_queued_events();
				});
			};
			async_queued_events();

			// full screen event (NO necesita pasar la campana ya que por ahora solo refresca todas las rows de la grilla del dashboard)
			self.full_screen_on_change_config_selector();
		});
	};

	SupervisorDashboard.prototype.__set_campaign_selector_config = function($campaign)
	{
		var campaign_id = $campaign.attr('campaign_id');
		const normalized_id = this.get_normalized_id(campaign_id);
		const $selector = $(`#sup_dash_campaign_accounts_config_${normalized_id}`);
		// 1 - desmarcado todas las campanas, cuentas, als
		if ( $campaign.prop('checked' ) ) {
			var $marked_account =  $selector.find('ul[accounts_list] input[account_selector]:checked');
			if ( $marked_account.length > 0 ) {
				// al desmarcar la cuenta, ya desmarca los als
				this.__mark_unmark_campaign_account( $marked_account, false /* mark */ );
				$marked_account.prop('checked', false);
			}
			$campaign.trigger('click'); // lo desmarca
		}

		// 2 - dejo marcado las campanas, cuentas, als que se encuentran configuradas.
		
		var campaign_config = this.get_campaign_selected_config(campaign_id);

		if ( campaign_config.campaign_id ) {
		// marco la campana
			$campaign.trigger('click');

			var $accounts_list = $selector.find('[accounts_list]');
			if ( campaign_config.account_id ) {
			// marco una cuenta especifica
				var $account_radio = $accounts_list.find('[account_item_type="account"] > [account_selector][account_id="' + campaign_config.account_id + '"]');
				this.__mark_unmark_campaign_account( $account_radio, true /* mark */ );
				$account_radio.prop('checked', true);
				if ( campaign_config.al_id ) {
				// marco un attention level especifico
					var $als_list = $account_radio.parent().find('[attention_levels_list]');
					var $al_radio = $als_list.find('[attention_level_selector][attention_level_id="' + campaign_config.al_id + '"]');
					$al_radio.trigger('click');
				}
			}
			else {
			// marco el chk (campaign all accounts)
				var $all_accounts_radio = $accounts_list.find('[account_item_type="header"] > [account_selector]');
				$all_accounts_radio.trigger('click');
			}
		}
	};

	/**
	 * Toggle the campaign selector to add/remove campaigns from dashboard.
	 *
	 * @public
	 */
	SupervisorDashboard.prototype.toggle_campaign_selector = function()
	{
		const self = this;

		const $campaign_popup = $('#sup_dash_campaign_selector_popup');
		const $campaign_button_popup = $('[action_id="campaign_selector"]');
		const $campaign_selector_list = $('#sup_dash_campaign_selector_list');
		var bind_event_outside;

		var show = function()
		{
			bind_event_outside();
			let campaign_id_in_full_screen = self.get_campaign_id_in_full_screen();
			const $campaigns = $campaign_selector_list.find('[selector_campaign_id]');
			// Recorro las campanas del selector y:
			$.each($campaigns, function(_, campaign) {
				let $campaign = $(campaign);
				if (!campaign_id_in_full_screen || "dashboard" === campaign_id_in_full_screen) {
					// 1 - Visualizo el item de la lista si NO esta en modo full screen, o si esta en modo full screen y posicionado en el dashboard (lista de campanas)
					$campaign.show();
				}
				else {
					let selector_campaign_id = $campaign.attr('selector_campaign_id');
					if (campaign_id_in_full_screen.toUpperCase() === selector_campaign_id.toUpperCase()) {
					// 2 - Visualizo el item de la lista, si se encuentra en modo Full Screen y estamos posicionados en el TAB de esta campana
						$campaign.show();
					}
					else {
					// 3 - NO lo visualizo el item de la lista, si se encuentra en modo Full Screen y NO estamos posicionados en el TAB de esta campana
						$campaign.hide();
					}
				}
			});
			$campaign_popup.show();
			$campaign_button_popup.addClass('menu-show');
		};
		var hide = function() {
			$campaign_popup.unbind('clickoutside');
			$campaign_popup.hide();
			$campaign_button_popup.removeClass('menu-show');
		};

		bind_event_outside = function() {
			// Se agrega en un pequeno timeout antes de realizar el bindeo para que el mismo se biendee despues de visualizarlo correctamente, sino el mismo se cierra
			window.setTimeout(function() {
				$campaign_popup.bind('clickoutside', function(_) {
						hide();
					}
				);
			});
		};

		let is_visible = $campaign_popup.is(':visible');

		if (!is_visible) {
			//Oculto todos los selectores de cuentas.
			this.__hide_all_account_selectors();
		}

		// restore checkbox statuses based on campaign configuration
		$campaign_selector_list.find('.campaignRowCheck').each(function(_, campaign) {
			let $campaign = $(campaign);
			self.__set_campaign_selector_config($campaign);
		});
		if (is_visible) {
			hide();
		}
		else {
			show();
		}
	};

	/** @public */
	SupervisorDashboard.prototype.get_monitor_manager = function() {
		return this.m_dash_monitor;
	};

	/** @public */
	SupervisorDashboard.prototype.get_appointment_manager = function() {
		return this.m_dash_appointment;
	};

	/** @public */
	SupervisorDashboard.prototype.get_model_manager = function() {
		return this.m_model;
	};

	/** @public */
	SupervisorDashboard.prototype.get_archive_manager = function() {
		return this.m_dash_archive;
	};

	/** @public */
	SupervisorDashboard.prototype.get_config_manager = function() {
		return this.m_config;
	};

	/** @public */
	SupervisorDashboard.prototype.get_workgroup_manager = function() {
		return this.m_workgroups;
	};

	/** @public */
	SupervisorDashboard.prototype.find_element_inside_campaign = function(campaign_id, selector) {
		return $('#sup_dash_campaign_'+ this.get_normalized_id(campaign_id) + ' ' + selector);
	};

	/**
	 * Get a normalized campaign dom id. Replaces @ and spaces with underscores.
	 *
	 * @public
	 * @param {string} campaign_id - The campaign id (i.e: some campaign@inconcert).
	 * @return (string) (i.e: some_campaign_inconcert)
	 *
	 */
	SupervisorDashboard.prototype.get_normalized_id = function (campaign_id) {
		return nid(campaign_id);
	};

	/**
	 * Given a campaign id, will return a campaign node.
	 *
	 * @public
	 * @param {string} campaign_id - The campaign id (i.e: some campaign@inconcert).
	 * @returns {object} The campaign node (jquery)
	 *
	 */
	SupervisorDashboard.prototype.get_campaign_node = function (campaign_id) {
		return $('#sup_dash_campaign_' + this.get_normalized_id(campaign_id));
	};

	SupervisorDashboard.prototype.get_campaign_bots_node = function( campaign_id )
	{
		return $(`#sup_dash_bots_info_container_${this.get_normalized_id(campaign_id)}`);
	};

	SupervisorDashboard.prototype.get_campaign_title_bar_node = function( campaign_id )
	{
		return $(`#sup_dash_campaign_title_${this.get_normalized_id(campaign_id)}`);
	};


	/**
	 * Given a normalized campaign id, will return a campaign node.
	 *
	 * @private
	 * @param {string} campaign_dom_id - The normalized campaign id (i.e: some campaign_inconcert).
	 * @returns {object} The campaign node (jquery)
	 *
	 */
	SupervisorDashboard.prototype.get_campaign_node_from_dom_id = function (campaign_dom_id) {
		return $('#sup_dash_campaign_' + campaign_dom_id);
	};

	SupervisorDashboard.prototype.__get_campaign_body_node = function (campaign_dom_id){
		return this.get_campaign_node_from_dom_id(campaign_dom_id).find('.panelBody');
	};

	/**
	 * Given a jQuery node, will return the calculated css top position
	 *
	 * @private
	 * @param  {Object} node - The jQuery node
	 * @return {Number}      - Top position of the node
	 */
	SupervisorDashboard.prototype.__get_node_top_position = function (node){
		return parseInt(node.css('top'), 10) || 0;
	};

	SupervisorDashboard.prototype.__set_body_panel_top_position = function (campaign_dom_id, how_much){
		this.__get_campaign_body_node(campaign_dom_id).css('top', how_much);
	};

	SupervisorDashboard.prototype.__reduce_body_panel_top_position = function (campaign_dom_id, how_much){
		var $body_panel_node = this.__get_campaign_body_node(campaign_dom_id);
		var body_panel_top = this.__get_node_top_position($body_panel_node);
		$body_panel_node.css('top', body_panel_top - how_much);
	};

	SupervisorDashboard.prototype.__increase_body_panel_top_position = function (campaign_dom_id, how_much){
		var $body_panel_node = this.__get_campaign_body_node(campaign_dom_id);
		var body_panel_top = this.__get_node_top_position($body_panel_node);
		$body_panel_node.css('top', body_panel_top + how_much);
	};

	/** @private */
	SupervisorDashboard.prototype.__get_total_campaigns_rendered = function () {
		if (this.m_full_screen) {
			return this.m_full_screen.get_total_grid_rows();
		}
		return _.size(this.m_rendered_campaigns);
	};

	SupervisorDashboard.prototype.get_rendered_campaigns = function () {
		return this.m_rendered_campaigns || {};
	};

	/*
		Valida si la reenderizacion de la campana en pantalla ha termiando exitosamente
	*/
	SupervisorDashboard.prototype.__is_ready_rendered_campaign = function (campaign_id) {
		if (!this.m_rendered_campaigns) {
			return false;
		}
		var rendered_campaign = this.m_rendered_campaigns[campaign_id];
		if (!rendered_campaign) {
			return false;
		}
		return rendered_campaign.ready;
	};

	/** @public */
	SupervisorDashboard.prototype.on_leave = function () {
		this.m_supervisor_leave = true;
		// stop refreshing view.
		this.stop_refresh();
		this.__set_counter_alarms();
		this.start_alarms_counter_refresh(false);
	};

	/*
		Esta funcion actualiza los atributos de los contadores del boton de acceso de la app "supervisor".
		Dicha actualizacion se realiza cuando nos moveos hacia otra app, es decir cuando salimos de la app "supervisor".
		Esto sirve para que el blink no se visualice si no es necesario, ya que el numero de notificaciones pueden haber cambiado y el blink
		  no se debe de visualizar al instante de salir de la app "supervisor"
	*/
	SupervisorDashboard.prototype.__set_counter_alarms = function ()
	{
		var totals_alarms = this.get_totals_alarms();
		var total_live = totals_alarms.total_live;
		var unread = totals_alarms.unread;

		var $supervisor_btn = this.__$get_app_button();
		var $alarm_number_container = this.__$get_app_alarm_number($supervisor_btn);

		var $alarm_unread_number = $alarm_number_container.find("[alarm_unread_number]");
		$alarm_unread_number.attr("alarm_unread_number", unread);

		var $alarm_live_number = $alarm_number_container.find("[alarm_live_number]");
		$alarm_live_number.attr("alarm_live_number", total_live);
	};

	/*
		Instancia el "selector" de members para las campanas
	*/
	SupervisorDashboard.prototype.__initialize_members_selector_popup = function()
	{
		const self = this;
		if (this.m_members_selector) {
			return;
		}
		this.m_members_selector = new inConcert.users_selector({
			from_app: "supervisor",
			container_id: "sup_members_selector_container",
			container_classes: "",
			callbacks: {
				update: function(info, callback = $.noop) {
					self.__update_campaign_members(info.campaign_id, info.selected_users, info.current_users, info.campaign_att_levels, callback);
				},
				show_waiting: function(msg) {
					self.show_waiting(msg);
				},
				hide_waiting: function() {
					self.hide_waiting();
				}
			}
		});
	};
	SupervisorDashboard.prototype.__add_att_levels_members = function(campaign_id, selected_users, campaign_al, callback = $.noop)
	{
		/*
		Solo se usa para agregar agentes que se hayan agregado nuevos a campanna, en los AL
		*/
		var self = this;
		let levels_to_update = campaign_al;
		let needUpdate = false;
		levels_to_update.forEach( level => {
			for (let user_key in selected_users.users){
				if (selected_users.users.hasOwnProperty(user_key)) {
					let user_selected = selected_users.users[user_key];
					if(user_selected.att_levels && user_selected.att_levels[level.id] && !level.users[user_key])
					{
						if(Array.isArray(level.users)){
							level.users = {};
						}
		
						level.users[user_key] = {
							id : user_key
						};
						needUpdate = true;
					}
				}		
			}
		});

		if(!needUpdate){
			callback(true);
			return;
		}

		let msg = "<%=_T('Please wait... Updating attention levels users.')%>";
		this.show_waiting(msg);
		// - Envio request al server asignar / desasignar a los agentes los attention levels definidos en la campaña
		this.m_model.update_attention_levels_members(campaign_id, levels_to_update, function(success, result) {
			self.hide_waiting();
			if (success) {
				let msg = "<%=_T('Attention levels users have been updated')%>";
				self.show_info_message(msg, false, true);
			}
			callback(success, result);
		});

	};
	SupervisorDashboard.prototype.__update_att_levels_members = function(campaign_id, current_users, selected_users, campaign_al, callback = $.noop)
	{
		var self = this;
		let levels_to_update = campaign_al;
		let needUpdate = false;
		levels_to_update.forEach( level => {
			for (let current_user_id in current_users) {
				if (current_users.hasOwnProperty(current_user_id)) {
				
					let user_selected = selected_users.users[current_user_id];
					let exists_user = level.users[current_user_id];
					/*
					Eliminar, si el current user no existe en la lista de seleccionados (se deselecciono)
					 y si pertece al AL
					*/
					if (!user_selected )
					{
						if(exists_user) {
							delete level.users[current_user_id];
							needUpdate = true;
						}
					}
					/*
					Eliminar, si el usuario seleccionado esta en la lista de AL
					y el AL.id no esta en la lista del AL del usuario seleccionado 
					*/
					else if(exists_user && user_selected.att_levels && !user_selected.att_levels[level.id]){
						delete level.users[current_user_id];
						needUpdate = true;
					}
					/*
					Agrego, si el AL.id esta en la lista de AL seleccionados por el usuario
					y el usuario no esta agregado en el AL.
					*/
					else if(user_selected && !exists_user && user_selected.att_levels && user_selected.att_levels[level.id]){
						needUpdate = true;
						if(Array.isArray(level.users)){
							level.users = {};
						}

						level.users[current_user_id] = {
							id : current_user_id
						};
					}
				}
			}
		});

		if(!needUpdate){
			callback(true, undefined, levels_to_update);
			return;
		}

		let msg = "<%=_T('Please wait... Updating attention levels users.')%>";
		this.show_waiting(msg);
		// - Envio request al server asignar / desasignar a los agentes los attention levels definidos en la campaña
		this.m_model.update_attention_levels_members(campaign_id, levels_to_update, function(success, result) {
			
			let msg;

			self.hide_waiting();

			if (success)
			{
				msg = "<%=_T('Attention levels users have been updated')%>";
				self.show_info_message(msg, false, true);
				callback(success, result, levels_to_update);
			}
			else
			{
				if (result.result === "PENDING_INT")
				{
					msg = "<%=_T('Attention levels users have been updated but some users cannot be removed because they have active interactions.')%>";
					self.show_info_message(msg, false, true);
					callback(true, result, levels_to_update);
				}
				else if(result.result === "ERROR_ALS_USERS_EXCEEDS_LIMIT_ALLOWED"){
					const limitAlsUsers = inConcert.app.getMaxConfigurationsAllowed("MaxAttsLevelUsers");
					msg = format("<%=_T('The total number of users assigned to the att. levels exceeds the limit allowed {0}.')%>", limitAlsUsers);
					inConcert.app.getMainWindow().showBannerMessage({message: msg, timeout: 8000});
					callback(success, result, levels_to_update);
				}
				else
				{
					msg = "<%=_T('Invalid request. Unable to update attention levels.')%>";
					self.show_error_message(msg);
					callback(success, result, levels_to_update);
				}
			}
		});
	};

	SupervisorDashboard.prototype.__update_campaign_members = function(campaign_id, selected_users, current_users, att_levels, callback = $.noop)
	{
		const self = this;

		let agents_to_update = {}; // (**) Este object va a tener la lista de members a agregar o quitar de la campana
		let total_agent_to_update = 0;
		let save_selected_users = inConcert.Utils.cloneObject(selected_users);
		// 1 - REMOVIENDO Members existentes.
		//
		// Recorro la lista current members de la campana, y valido si existe en la lista de members.
		// Si el current member NO existe, lo agrego en el array (**) marcandolo para que el mismo sea REMOVIDO de la campana.
		for (let current_user_id in current_users) {
			if (current_users.hasOwnProperty(current_user_id)) {
				if (!selected_users.users[current_user_id]) {
					total_agent_to_update++;
					agents_to_update[current_user_id] = {
						userId: current_user_id,
						Id: current_user_id,
						action: "remove"
					};
				}
				else {
					delete selected_users.users[current_user_id];
				}
			}
		}

		// 2 - AGREGANDO Members.
		//
		// Recorro la lista de los members a incluir en la campana, y valido si existe en la lista de current members.
		// Si el member a incluir NO existe, lo agrego en el array (**) marcandolo para que el mismo sea AGREGADO de la campana.
		for (let selected_user_id in selected_users.users) {
			if (selected_users.users.hasOwnProperty(selected_user_id)) {
				if (!current_users[selected_user_id]) {
					total_agent_to_update++;
					agents_to_update[selected_user_id] = {
						Id: selected_user_id,
						userId: selected_user_id,
						action: "add"
					};
				}
			}
		}

		//Remover Agentes de los AL, si es necesario
		this.__update_att_levels_members(campaign_id, current_users, save_selected_users, att_levels, function(success, result, att_levels_updated) {
			
			if(!success){
				callback(success, result);
				return;
			}

			// Si no hay cambios a efectuar, me voy
			if (total_agent_to_update === 0) {
				callback(success, result);
				return;
			}
			let camp_name;
			let campaign = inConcert.app.getCampaign(campaign_id);
			if (campaign) {
				camp_name = campaign.getDescription();
			}
			else {
				camp_name = campaign_id;
			}

			let msg = format("<%=_T('Please wait... Updating {0} users in campaign {1}.')%>", total_agent_to_update, camp_name);
			self.show_waiting(msg);
			// 3 - Envio request al server para que actualice los members en la campana
			self.m_model.update_campaign_agents(campaign_id, agents_to_update, total_agent_to_update, function(success, result) {
				if (!success)
				{
					let msg = "<%=_T('Failed updating users list, please try again in a few minutes...')%>"; // default message
					if (result.result === "ALLEGRO_INTEGRATION_ERROR") {
						msg = "<%=_T('Update Problems.')%>";
						msg = msg + "<%=_T('There is a users synchronization problem with Allegro. To fix the problem you need to validate the services in Allegro and update the users again.')%>";
					}
					else if (result.result === "ERROR_CONNECTION_REDIS") {
						msg = "<%=_T('Error validating campaigns/users relationship limits.')%>";
					}
					else if (result.result === "ERROR_USERS_EXCEEDS_LIMIT_ALLOWED") {
						const limitCampUsers = inConcert.app.getMaxConfigurationsAllowed("MaxCampaignsUsers");
						msg = format("<%=_T('The total number of users assigned to the campaigns exceeds the limit allowed {0}.')%>", limitCampUsers);
					}
					else if (result.result === "THERE_IS_ANOTHER_USERS_UPDATE_IN_PROGRESS_FOR_THIS_CAMPAIGN") {
						msg = "<%=_T('Unable to update users list. There is another users update in progress for this campaign.')%>";
					}
					else {
						if (result.problematicUsers && result.problematicUsers[0]) {
							msg = format("<%=_T([[Error updating users from campaign {0}. The user {1} has assigned interactions.]])%>", camp_name, result.problematicUsers[0].id);
						}
					}

					self.hide_waiting();
					inConcert.app.getMainWindow().showBannerMessage({message: msg, timeout: 10000});
					callback(success, result);

					return;
				}
				
				//Agregar agentes a los AL. que se hayan actualizado recien en la campanna
				self.__add_att_levels_members(campaign_id, selected_users, att_levels_updated, function(success, result) {
					callback(success, result);
					self.hide_waiting();
					if (success) {
						let msg = format("<%=_T('{0} users have been updated in {1} campaign.')%>", total_agent_to_update, camp_name);
						self.show_info_message(msg, false, true);
					}

				}, true, false);

			});
		
		}, false, true);
		
	};

	/*
		Esta funcion es la encargada de marcar las alarmas sin visualizar en el boton de acceso a la app "supervisor",
		cuando el usuario recien ha iniciado sesion en "i6", y por defecto queda posicionado en una app la cual no es la app "supervisor"
	*/
	SupervisorDashboard.prototype.initialize_alarms_counter_refresh = function ()
	{
		const self = this;

		//console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - START");
		this.m_hnd_alerted_timeout = null;
		// si no tiene permisos sobre la app "supervisor", se va
		let allowed_sup_app = inConcert.permissions.to_application("supervisor");
		if (!allowed_sup_app) {
			//console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - CANCEL - YOU NOT HAS PERMISSION TO SUPERVISOR APP.");
			return;
		}

		// funcion de reintentar que se define mas abajo y se utiliza dentro de "get_config" si es que ha fallado el request que obtiene la configuracion del supervisor.
		var get_config_retry; 

		// ejecuta la funcion del get config del supervisor
		var get_config = function()
		{
			//console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - get_config.");
			self.get_config(function(config, err)
			{
				if (err) {
					let error_message = format('[Supervisor] (initialize_alarms_counter_refresh) - Error getting configuration from server: {0}', err && err.message ? err.message : err);
					inConcert.logger.error(error_message);
					//console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - get_config - ERROR.");
					self.__show_default_error_message();
					// voy por el retry si es que ha fallado
					get_config_retry();
					return;
				}

				//console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - get_config - OK.");

				// SUCCESS get config
				if (config && !_.isEmpty(config)) {
					self.set_current_config(config);
					self.m_user_without_config = false;
				}
	
				// busca todas la campanas a las cuales el usuario pertenece para iniciar el dashboard de configuracion
				// asi despues le puede solicitar las alarmas configuradas 
				let user_campaigns = inConcert.app.getCampaigns();
				let selected_campaigns = [];
				$.each(user_campaigns, function(_, campaign)
				{
					// TODO: realizar validacion contra las campanas que esta visualizando, NO contra todas las campanas
					// a las cuales el supervisor pertenece.
					let campaign_config = config[campaign.Id];
					if (!campaign_config || (campaign_config && !campaign_config.visible)) {
						return true; // continue
					}
					self.m_dash_config.setup(campaign.Id, "CONFIGURATION-ONLY-DATA"); // inicio el dashboard de config
					selected_campaigns.push(campaign);
				});
	
				// La invocacion a la funcion "start_alarms_counter_refresh", la cual comienza a actualizar el
				// el contador, se ejecuta recien despues que TODOS los dashboards de configuracion se encuentras listos (request, etc),
				// ya que se sino la configuracion no es completa y los datos serian erroneos.
				let interval_ms = 500;
				let total_retries = 20; // 10 segundos
				let retrieve_number = 0;
				let interval_hnd;
	
				interval_hnd = window.setInterval(function()
				{
					let count_ready_campaigns = 0;
					// Recorro las campanas valiado que la configuracion de las alarmas este lista en el dashboard.
					$.each(selected_campaigns, function(_, campaign) {
						// TODO: realizar validacion contra las campanas que esta visualizando, NO contra todas las campanas
						// a las cuales el supervisor pertenece.
						let is_ready = self.m_dash_config.is_ready_alarms_campaign_config(campaign.Id);
						if (is_ready) {
							count_ready_campaigns++;
						}
					});
					// Si todas las alarmas estan lista, o si se ha llegado al numero maximo de reintenos, en el timer
					// realiza la ejecucion de la funcion "start_alarms_counter_refresh" que acualizar el contador.
					if (count_ready_campaigns === selected_campaigns.length || retrieve_number === total_retries) {
						window.clearInterval(interval_hnd);
						self.start_alarms_counter_refresh(true);
					}
					else {
						retrieve_number++;
					}
				}, interval_ms);
			});
		};

		// funciones de retry por si falla la ejecucion inicial de la configuracion de las campanas de supervisor.
		const elapsed_time_retry_ms = 2000;
		const max_retries = 5;
		let current_retry = 0;

		get_config_retry = function() {
			current_retry++;
			console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - get_config_retry - RETRY " + current_retry.toString());
			if (current_retry > max_retries) {
				let error_message = '[Supervisor] (initialize_alarms_counter_refresh) - Error getting configuration !!!!!';
				inConcert.logger.error(error_message);
				console.debug("SupervisorDashboard.prototype.initialize_alarms_counter_refresh - get_config_retry - RETRY ERROR !!.");
				self.__show_default_error_message();
				return;
			}
			window.setTimeout(function() {
				get_config();
			}, elapsed_time_retry_ms);
		};

		// comienzo el get config
		get_config();
	};

	/*
		Esta funcion va obteniendo la cantidad de alarmas live por campanas y actualiza el contador de campanas en el boton de acceso
		a la applicacion "supervisor"
	*/
	SupervisorDashboard.prototype.start_alarms_counter_refresh = function (force)
	{
		const self = this;

		//console.debug("SupervisorDashboard.prototype.start_alarms_counter_refresh - START ");

		if (!this.__can_start_alarms_counter_refresh(force)) {
			//console.debug("SupervisorDashboard.prototype.start_alarms_counter_refresh - NO PUEDE INICIAR START");
			this.stop_alarms_counter_refresh();
			return;
		}

		var process = function()
		{
			self.get_config(function(config)
			{
				let visible_campaigns;
				var campaigns = inConcert.app.getCampaigns();
				for (let index = 0; index < campaigns.length; index++) {
					const campaign = campaigns[index];
					const campaign_config = config[campaign.Id];
					if (campaign_config && campaign_config.visible && ( force || self.has_configured_alarms(campaign.Id))) {
						if (!visible_campaigns) {
							visible_campaigns = {};
						}

						//Se busca en un mismo request los members y acciones de panel
						self.m_dash_members.initializePageByCampaign(campaign.Id);
						visible_campaigns[campaign.Id] = {
							filter : self.__get_campaign_filter(campaign.Id),
							filterMembers : self.m_dash_members.get_members_filters(campaign.Id)
						};
					}
				}

				if (!visible_campaigns) {
					return;
				}
				let data = {
					campaigns : visible_campaigns
				};
				self.m_model.getSupViewAndMembersByCampaigns(data, function (response, error) {
					
					let supervisor_campaign_views = response && response.supervisor_views;
					if (error || !supervisor_campaign_views) {
						self.__hide_alarms_counter();
						return;
					}

					for (const campaign_id in supervisor_campaign_views) {
						if (Object.hasOwnProperty.call(supervisor_campaign_views, campaign_id)) {
							const element = supervisor_campaign_views[campaign_id];
							self.__refresh_alarms_notifications(campaign_id, element);
						}
					}

					if (inConcert.app.getCurrentApp() === "supervisor") {
						self.stop_alarms_counter_refresh();
						return;
					}
					if (!self.__can_start_alarms_counter_refresh(force)) {
						self.stop_alarms_counter_refresh();
						return;
					}
					// muestro la cantidad de alarmas en el boton de la app "supervisor"
					self.__show_alarms_counter();

				});

			});
		};

		// 1 - la primera vez que ingresa, lo ejecuta y despues queda en el timer
		process();

		// 2 - ahora queda en el timer
		this.m_on_refresh_alarms_counter_timer = window.setInterval(function () {
			process();
		}, REFRESH_ALARMS_ANOTHER_APP_MS);
	};

	SupervisorDashboard.prototype.stop_alarms_counter_refresh = function ()
	{
		window.clearInterval(this.m_on_refresh_alarms_counter_timer);
		this.m_on_refresh_alarms_counter_timer = null;
		this.__hide_alarms_counter();
	};

	SupervisorDashboard.prototype.__can_start_alarms_counter_refresh = function (force)
	{
		const allowed_sup_app = inConcert.permissions.to_application("supervisor");
		if (!allowed_sup_app) {
			//console.debug("SupervisorDashboard.prototype.__can_start_alarms_counter_refresh - NO PUEDE VISUALIZAR EL COUNTER - NO HAS PERMISSION TO SUPERVISOR APP");
			return false;
		}
		if (!force && !this.m_supervisor_leave) {
			//console.debug("SupervisorDashboard.prototype.__can_start_alarms_counter_refresh - NO PUEDE VISUALIZAR EL COUNTER - !this.m_supervisor_leave  && !force");
			return false;
		}
		//console.debug("SupervisorDashboard.prototype.__can_start_alarms_counter_refresh - SI PUEDE VISUALIZAR EL COUNTER");
		return true;
	};

	SupervisorDashboard.prototype.__$get_app_button = function ()
	{
		var $dashboard_apps_tab = $("#dashboard_apps_tab");
		var $supervisor = $dashboard_apps_tab.find('[dashboard_item="supervisor"]');
		return $supervisor;
	};

	SupervisorDashboard.prototype.__$get_app_alarm_number = function ($supervisor_btn)
	{
		var $alarm_number_container = $supervisor_btn.find("[alarm_number_container='true']");
		return $alarm_number_container;
	};

	SupervisorDashboard.prototype.__clear_alerted_alarm_timeout = function ()
	{
		window.clearTimeout(this.m_hnd_alerted_timeout);
		this.m_hnd_alerted_timeout = null;
	};

	SupervisorDashboard.prototype.__show_alarms_counter = function ()
	{
		const self = this;

		if (inConcert.app.getCurrentApp() === "supervisor") {
		// Si nos encontramos en la app supervisor, ocultamos los contadores.
			this.__hide_alarms_counter();
			return;
		}

		let totals_alarms = this.get_totals_alarms();
		let total_live = totals_alarms.total_live;
		let unread = totals_alarms.unread;

		const $supervisor_btn = this.__$get_app_button();
		const $alarm_number_container = this.__$get_app_alarm_number($supervisor_btn);

		// El numero de unread
		const $alarm_unread_number = $alarm_number_container.find("[alarm_unread_number]");
		$alarm_unread_number
			.html(unread || 0)
			.attr("alarm_unread_number", unread);

		// El numero total de live
		const $alarm_live_number = $alarm_number_container.find("[alarm_live_number]");
		const old_total_live = parseInt($alarm_live_number.attr("alarm_live_number"), 10) || 0;
		$alarm_live_number
			.html(total_live || 0)
			.attr("alarm_live_number", total_live);

		// El tooltip del contenedor de valores live y unread
		$alarm_number_container
			.attr("title", format(_T("You have {0} alarms, {1} unview."), total_live, unread))
			.show();

		// Esta validacion debe estar posterior a la actualizacion de los atributos y html del contador.
		// Esto para que el valor old_total_live actualice los valores en 0 si es que no hay mas alertas.
		// Entonces si estamos posicionados en otra app que no sea "supervisor", y nos llega una nueva alerta,
		//   se va a visualizar correctamente el blink.
		if (unread < 1 && total_live < 1) {
		// Si no existen unread ni lives, ocultamos los contadores.
			this.__hide_alarms_counter();
			return;
		}

		// Si el numero de live se ha incrementado debo de mostrar el blink
		$supervisor_btn.removeClass("alerted");
		if (total_live > old_total_live) {
			$supervisor_btn.addClass("alerted");
			this.__clear_alerted_alarm_timeout();
			// El blink dura unos N segundos, despues la clase que realiza el mismo es removida.
			this.m_hnd_alerted_timeout = window.setTimeout(function() {
				self.__clear_alerted_alarm_timeout();
				$supervisor_btn.removeClass("alerted");
			}, ALERTED_ALARMS_TIMEOUT_MS);
		}
	};

	SupervisorDashboard.prototype.__hide_alarms_counter = function ()
	{
		this.__clear_alerted_alarm_timeout();
		var $supervisor_btn = this.__$get_app_button();
		$supervisor_btn.removeClass("alerted");
		var $alarm_number_container = this.__$get_app_alarm_number($supervisor_btn);
		$alarm_number_container.hide();
	};

	/**
	 * Given a number of current panels (1, 2 or 3), will add/remove clases to
	 * to the current campaign node.
	 *
	 * @private
	 * @param {object} current_node - Current campaign node (jquery node).
	 * @param {number} total_panels - Number of panels left.
	*/
	SupervisorDashboard.prototype.__update_panel_class = function (current_node, total_panels) {
		if (typeof (total_panels) !== 'number') {
			return;
		}
		switch (total_panels) {
			case 1:
				current_node.removeClass('threePanels twoPanels');
				break;
			case 2:
				current_node.removeClass('threePanels').addClass('twoPanels');
				break;
			case 3:
				current_node.removeClass('twoPanels').addClass('threePanels');
				break;
			case 'default':
				break;
		}
	};

	/**
	 * This function will calculate and set the 'left' position of each campaign panel
	 * on the left side if the current node (dom three representation)
	 *
	 * @private
	 * @param {Object}  current_node    - Current campaign node
	 * @param {Boolean} need_space      - A bool switch to know if I need to add or substract space to others nodes.
	 * @param {Number}  how_much_space  - Default is 261px (a column width) but you can pass another value (i.e. entire campaign node in px)
	 */
	SupervisorDashboard.prototype.__set_next_campaigns_panels_positions = function (current_node, need_space, how_much_space) {
		var space = how_much_space ? how_much_space : 261;
		// Calculate position of the next panels
		current_node.nextAll().each(function (_, e) {
			var current_sibling = $(e);
			var current_left = parseInt(current_sibling.css('left'), 10);
			var final_left = need_space ? current_left + space : current_left - space;
			current_sibling.css('left', final_left);
		});
	};

	SupervisorDashboard.prototype.open_view = function ( view_id, params )
	{
		var self = this;
		var campaign_id = params.campaign_id;
		var navigation = this.get_navigation(campaign_id);
		var opened, dash_view;
		var campaign_dom_id = this.get_normalized_id( campaign_id );

		switch ( view_id )
		{

			case 'monitoring':
				
				this.m_dash_monitor.set_can_refresh_interactions(false);
				opened = this.m_dash_monitor.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('active', undefined, {can_refresh_interactions: false});
				}
				dash_view = this.m_dash_monitor;
				if(params && params.agent_id) {
					this.queryBuilderObj[campaign_dom_id].AddRuleAgent(params.agent_id);
				}
				this.m_dash_monitor.set_can_refresh_interactions(true);
				break;
			case 'appointment':
				opened = this.m_dash_appointment.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('appointment');
				}
				dash_view = this.m_dash_appointment;
				break;
			case 'queued':
				opened = this.m_dash_queued.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('queued');
				}
				dash_view = this.m_dash_queued;
				break;
			case 'waiting':
				opened = this.m_dash_waiting.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('waiting');
				}
				dash_view = this.m_dash_waiting;
				break;
			case 'abandoned':
				opened = this.m_dash_abandoned.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('abandoned');
				}
				dash_view = this.m_dash_abandoned;
				break;
			case 'bot_interactions':
				opened = this.m_dash_bot_interactions.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('bot_interactions');
				}
				dash_view = this.m_dash_bot_interactions;
				break;
			case 'archive':
				this.m_dash_archive.set_can_refresh_interactions(false);
				opened = this.m_dash_archive.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('archive', undefined, {can_refresh_interactions: false});
				}
				dash_view = this.m_dash_archive;
				if(params && params.agent_id) {
					this.queryBuilderObj[campaign_dom_id].AddRuleAgent(params.agent_id);
				}
				this.m_dash_archive.set_can_refresh_interactions(true);
				break;
			case 'global_search':
				navigation.select_folder('global_search');
				break;
			default:
				// by default, open monitoring.
				opened = this.m_dash_monitor.open_view(campaign_id);
				if ( opened ) {
					navigation.select_folder('active');
				}
				dash_view = this.m_dash_monitor;
		}

		if (params.must_execute_new_search && dash_view.execute_new_search) {
			var cleanup_search_text = true;
			dash_view.execute_new_search(campaign_id, undefined, cleanup_search_text, params);
		}

		window.setTimeout(function()
		{
			self.onReadyToRefresh(campaign_id);
		}, 1000);
		
		return opened;
	};

	/** @private */

	/**
	 * Based on the current campaign configuration, will setup the corresponding panels/options
	 *
	 * @private
	 * @param  {Object} current_node - The current jQuery campaign node.
	 * @param  {String} campaign_id  - The campaign identificator.
	 */
	SupervisorDashboard.prototype.__configure_node_visibility = function (current_node, campaign_id)
	{
		var self = this;
		var campaign_config = this.get_stored_campaign_config(campaign_id);

		current_node.unbind('view_opened').bind('view_opened', function () {
			self.update_panel_positions();
		});

		// If last view was monitoring, open view and return.
		if ( campaign_config && campaign_config.dashboard.current_view !== 'dashboard' )
		{
			this.open_view( campaign_config.dashboard.current_view, { campaign_id: campaign_id } );
			return;
		}
		else {
			if ( campaign_config ) {
				var panels = campaign_config.dashboard.panels;
				for (var panel_id in panels) {
					if (panels.hasOwnProperty(panel_id)){
						var panel_is_visible = panels[panel_id].visible;
						// Toggle current panel
						current_node.find('#' + panel_id)
							.toggle(panel_is_visible)
							.toggleClass('columnHidden', !panel_is_visible);

						current_node.find('#' + panel_id.replace('panel', 'option')).prop('checked', panel_is_visible);
						this.__set_next_campaigns_panels_positions(current_node, panels[panel_id].visible);
					}
				}
			}
			// update panel class
			var visible_panels = current_node.find('.panelBodyColumnAux:visible').length;
			this.__update_panel_class(current_node, visible_panels);
			// trigger on dashboard opened.
			this.on_dashboard_opened(campaign_id);
		}
	};

	/** @public */
	SupervisorDashboard.prototype.on_dashboard_opened = function(campaign_id)
	{
		this.set_current_view_id(campaign_id, 'dashboard');
		var current_node = this.get_campaign_node(campaign_id);

		// configure columns
		var total_visible_panels = current_node.find('.panelBodyColumnAux:visible').length;
		this.__update_panel_class(current_node, total_visible_panels);

		// notify columns
		// TO-DO: Use PubSub
		this.m_dash_indicators.on_dashboard_opened(campaign_id);
		this.update_panel_positions();
		this.m_full_screen.update_toolbar_buttons();
	};

	/** @public */
	SupervisorDashboard.prototype.on_dashboard_closed = function(campaign_id, opened_view_id)
	{
		this.set_current_view_id(campaign_id, opened_view_id);
		this.update_panel_positions();
		this.m_full_screen.update_toolbar_buttons();
		this.__clean_main_dashboard_layout_classes(campaign_id);
	};

	/** @private */
	SupervisorDashboard.prototype.__toggle_panel = function(campaign_dom_id, column_id, showOrClose) {
		$('#sup_dash_' + column_id + '_panel_' + campaign_dom_id).toggle(showOrClose).toggleClass('columnHidden', !showOrClose);
	};

	/** @public */
	SupervisorDashboard.prototype.trigger_column_update = function ( campaign_id, column_id, showOrClose )
	{
		if ( showOrClose ) {
		// el permiso solo se valida, si se quiere mostrar el panel
			var panel_permission = {
				indicators : 'sup_view_indicators',
				members    : 'sup_view_team',
				summary    : 'sup_view_summary'
			};
			var permission_id = panel_permission[ column_id ];
			if ( permission_id ) {
				var has_permission = this.has_permission_to( permission_id, campaign_id );
				if ( !has_permission ) {
					return false;
				}
			}
		}

		var campaign_dom_id = this.get_normalized_id( campaign_id );

		// publish status for this column
		inConcert.SupervisorMessageBus.publish('/supervisor/dashboard/'+ column_id, {
			campaign_id : campaign_id,
			campaign_dom_id : campaign_dom_id,
			visible : showOrClose
		});

		// show or hide panel
		var current_node = this.get_campaign_node(campaign_id);
		this.__toggle_panel(campaign_dom_id, column_id, showOrClose);

		// adjust next panel positions and classes
		var total_visible_panels = current_node.find('.panelBodyColumnAux:visible').length;
		current_node.find('#panelViewsConfig:not([style="display:none"]').attr('class','panelOptionsContainer panelViewsConfig panelViewsConfigPanels'+total_visible_panels);
		this.__update_panel_class(current_node, total_visible_panels);
		this.__set_next_campaigns_panels_positions(current_node, showOrClose);

		// update panel menu checkbox
		current_node.find('input[panel_option_id="'+ column_id +'"]').prop('checked', showOrClose);

		// at least one panel should be visible, disable last checkbox and label inside the menu.
		var options_checked = current_node.find('.panelOptionCheck:checked');
		options_checked.prop('disabled', options_checked.length === 1 ? true : false);
		options_checked.siblings().prop('disabled', options_checked.length === 1 ? true : false);

		//Update main dashboard layout
		this.update_main_dashbord_layout(campaign_id);
		
		return true;
	};

	SupervisorDashboard.prototype.full_screen_on_change_config_selector = function()
	{
		if (!this.m_full_screen) {
			return;
		}
		return this.m_full_screen.on_change_config_selector();
	};

	SupervisorDashboard.prototype.full_screen_toggle_campaign_detail = function ($btn)
	{
		let campaign_id = this.get_campaign_id_in_full_screen();
		if (!campaign_id) {
			return;
		}
		this.toggle_campaign_detail(campaign_id, $btn);
	};

	SupervisorDashboard.prototype.full_screen_toggle_campaign_menu = function ()
	{
		this.m_full_screen.toggle_campaign_menu();
	};

	SupervisorDashboard.prototype.full_screen_toggle_campaign_menu_config_view = function ()
	{
		this.m_full_screen.toggle_campaign_menu_config_view();
	};

	SupervisorDashboard.prototype.full_screen_back_dashboard = function ()
	{
		let campaign_id = this.get_campaign_id_in_full_screen();
		if (!campaign_id) {
			return;
		}
		var $back = this.$back_button(campaign_id);
		$back.trigger('click');
	};

	SupervisorDashboard.prototype.toggle_campaign_detail = function (campaign_id, $btn)
	{
		const self = this;
		const campaign_dom_id = this.get_normalized_id(campaign_id);
		const $current_node_body = this.__get_campaign_body_node(campaign_dom_id);
		// toggle campaign container
		this.__get_campaign_detail_container(campaign_dom_id).toggle();
		if ($btn) {
			$btn.toggleClass('btnActive');
		}
		// update panel body position
		const $detail_container = this.__get_campaign_detail_container(campaign_dom_id);
		if ($detail_container.is(':visible')){
			const detail_container_height = $detail_container.outerHeight();
			const default_title_container_height = $('#sup_dash_panel_name_' + campaign_dom_id).parent().outerHeight();
			this.__set_body_panel_top_position(campaign_dom_id, detail_container_height + default_title_container_height);
			$btn.bind('clickoutside', function() {
				self.toggle_campaign_detail(campaign_id, $btn);
			});
		}
		else {
			const default_body_top = this.__get_node_top_position($current_node_body);
			this.__set_body_panel_top_position(campaign_dom_id, default_body_top);
			$btn.unbind('clickoutside');
		}
	};

	/*
	Ejemplo de la lista de members que se envia al server.
		agents = [
			{userId: 'roxana066', action: 'add'},
			{userId: 'admin', action: 'add'},
			{userId: 'usuario203', action: 'add'},
			{userId: 'usuario70', action: 'add'},
			{userId: 'usuario1001', action: 'remove'},
		];
	*/
	SupervisorDashboard.prototype.open_popup_members = function (campaign_id)
	{
		const self = this;

		if (!this.m_members_selector) {
			throw("Error!!");
		}

		const has_permission = this.has_permission_to('sup_update_members', campaign_id);
		if (!has_permission) {
			inConcert.app.getMainWindow().showBannerMessage({
				message: "<%=_T([[You don't have permissions for this action.]])%>",
				timeout: 5000
			});
			return;
		}

		let campaign = inConcert.app.getCampaign(campaign_id);
		let campaign_name = campaign ? campaign.getDescription(): campaign_id;

		this.show_waiting("<%=_T('Please wait...')%>");
		// Lo agrego en un pequeno timeout para que se vea el waiting, ya que puede ser muy rapida la carga y hay un juego de pantalla
		window.setTimeout(function()
		{
			let msg = format("<%=_T('Please wait... loading users of {0} campaign')%>", campaign_name);
			self.show_waiting(msg);
			// Como la lista de members aun no se ha reenderizado, obtengo todos los usuarios del vcc y luego creo el markup por unica vez.
			// Por ultimo obtengo los members de la campana y los marco.
			self.m_model.load_users_with_access_to_interaction_app(campaign_id,
				function(success, vcc_users, att_levels) {
					if (!success) {
						// TODO: visualizar error
						self.hide_waiting();
						return;
					}

					// Creo el markup con cada usuario.
					att_levels = inConcert.Utils.hash_to_sorted_array(att_levels || {}, undefined, undefined, { without_id : true, attr : "name", order : "asc" });
					// obtengo los members de la campana y luego los marco los checkboxes correspondientes.
					
					self.m_members_selector.make_users_markup(
						vcc_users,
						att_levels,
						function(success) {
							if (!success) {
							// TODO: visualizar error
								self.hide_waiting();
								return;
							}
							let members_belong_to_campaign = 0;
							vcc_users.forEach(user => {
								if (user.BelongToCampaigns[campaign_id]) {
									members_belong_to_campaign++;
								}
							});
							self.m_members_selector.set_total_campaign_members(members_belong_to_campaign);
							self.m_members_selector.mark_users(()=> {
								self.m_members_selector.show();
								self.hide_waiting();
							});
						},
						campaign_id
					);
				},
				true
			);
		}, 500);
	};

	/**
	 * Will setup the campaign detail toggle button behaviour.
	 *
	 * @private
	 * @param {Object} current_node - Current campaign jQuery node.
	 */
	SupervisorDashboard.prototype.__setup_campaign_detail_toggle_handlers = function ($current_node)
	{
		const self = this;
		const campaign_id = $current_node.attr('campaign_id');
		const campaign_dom_id = this.get_normalized_id(campaign_id);
		const $campaign_detail_button = $('#sup_dash_panel_detail_' + campaign_dom_id);
		$campaign_detail_button
			.unbind('click')
			.bind('click', function() {
				self.toggle_campaign_detail(campaign_id, $campaign_detail_button);
			})
			.unbind('clickoutside');
	};

	SupervisorDashboard.prototype.has_permission_to = function ( action_id, campaign_id ) {
		return inConcert.permissions.to_action('supervisor', action_id, undefined /* category */, campaign_id, true /* no log */);
	};

	SupervisorDashboard.prototype.has_permission_to_any_campaign = function ( action_id ) {
		return inConcert.permissions.to_action_on_any_campaign('supervisor', action_id);
	};

	SupervisorDashboard.prototype.__get_current_dashboard = function( campaign_id, current_view_id)
	{
		current_view_id = current_view_id || this.get_current_view_id(campaign_id);

		switch(current_view_id)
		{
			case 'dashboard_config':
				return this.m_dash_config;

			case 'monitoring':
				return this.m_dash_monitor;

			case 'appointment':
				return this.m_dash_appointment;

			case 'archive':
				return this.m_dash_archive;

			case 'queued':
				return this.m_dash_queued;

			case 'waiting':
				return this.m_dash_waiting;

			case 'abandoned':
				return this.m_dash_abandoned;

			case 'bot_interactions':
				return this.m_dash_bot_interactions;

			case 'global_search':
				// TODO: se hace algo con este ?
				return this.m_dash_global_search;
		}
	};

	SupervisorDashboard.prototype.__set_events_back_button = function( campaign_id )
	{
		const self = this;

		// // nothing to do if view has changed.
		// if (this.view_has_changed(campaign_id)){
			// return;
		// }

		let campaign_dom_id = this.get_normalized_id( campaign_id );
		// back to dashboard button
		let $back = this.$back_button(campaign_id);
		$back
			.unbind('click')
			.bind('click', function()
			{
				let $campaign = self.get_campaign_node(campaign_id);
				$campaign.removeClass('alarmsCondition interactionsViewerCondition');

				const panels_permissions = {
				// checkboxes
					indicators : 'sup_view_indicators',
					members    : 'sup_view_team',
					summary    : 'sup_view_summary'
				};
				let count_checked_shown_panels = 0;
				let shown_panels = [];
				let hidden_panels = [];
				$.each( panels_permissions, function( panel_id, permission_id ) {
					let has_permission = self.has_permission_to( permission_id, campaign_id );
					if ( has_permission ) {
						let $chk = $campaign.find('input[panel_option_id="'+ panel_id +'"]');
						if ( $chk.prop('checked') ) {
							count_checked_shown_panels++;
						}
						shown_panels.push( panel_id );
					}
					else {
						hidden_panels.push( panel_id );
					}
				});

				// si no existen paneles visibles, se va
				if ( shown_panels.length === 0 ) {
					inConcert.app.getMainWindow().showBannerMessage({
						message : "<%=_T('NOT HAS PERMISSIONS TO VIEW PANELS')%>",
						timeout : 5000
					});
					return;
				}
				// si existen paneles sin permisos, los oculto
				$.each( hidden_panels, function( _, panel_id ) {
					self.trigger_column_update( campaign_id, panel_id, false /* hide */ );
				});

				let $first_chk;
				if ( shown_panels.length === 1 ) {
					// si solo existe un panel visible, y el chk se encuentra desmarcado, me quedo en el chk para luego, invocar el trigger
					let visible_panel_id = shown_panels[ 0 ];
					let $chk = $campaign.find('input[panel_option_id="'+ visible_panel_id +'"]');
					if ( !$chk.prop('checked') ) {
						$first_chk = $chk;
					}
				}
				else if ( shown_panels.length > 1 ) {
					// si existe mas de un panel visible los recorro. 
					$.each( shown_panels, function( _, visible_panel_id ) {
						let $chk = $campaign.find('input[panel_option_id="'+ visible_panel_id +'"]');
						let is_checked = $chk.prop('checked');
						if ( count_checked_shown_panels === 1 && is_checked ) {
							$chk.prop('disabled', true);
						}
						else {
							$chk.prop('disabled', false);
						}
						// me quedo con el primero que exista no checked, para luego realizar la validacion de (+)
						if ( !is_checked ) {
							$first_chk = $chk;
						}
					});
				}
				// (+) si no existe ningun checkbox que visualice los panels marcados, tengo que dejar el primero de todos marcado y visualizando el panel
				if ( $first_chk && count_checked_shown_panels === 0 ) {
					$first_chk.prop('checked', true).trigger('change');
				}

				// clear interactions and messages
				let current_view_id = self.get_current_view_id(campaign_id);
				if (current_view_id !== 'dashboard') {
					let current_dashboard = self.__get_current_dashboard(campaign_id);
					if (current_dashboard) {
						if (current_dashboard.m_type === 'interactions') {
							current_dashboard.clear_list(campaign_dom_id);
						}
						// hide current interaction view and show dashboard
						current_dashboard.get_interactions_viewer_node(campaign_dom_id).hide();
						current_dashboard.get_operations_viewer_node(campaign_dom_id).show();
					}
				}

				// hide back to dashboard and show edit panels
				let $full_screen_back_button = self.$full_screen_back_button(campaign_id);
				$full_screen_back_button.hide();
				$(this).hide();
				$('#sup_dash_panel_menu_' + campaign_dom_id).show();
				$('#sup_dash_panel_menu_config_view_' + campaign_dom_id).show();

				//Si ando bot
				self.leave_bot_view(campaign_id);

				// hide alarm ABM config
				$campaign.find('[viewer_type="alarms_config"]').hide();
				// trigger on dashboard opened.
				self.on_dashboard_opened(campaign_id);
				// update main dashboard layout
				self.update_main_dashbord_layout(campaign_id);
				// save configuration.
				self.save_config();
			});

		// // full screen back to dashboard button
		// let $full_screen_back_button = this.$full_screen_back_button(campaign_id);
		// $full_screen_back_button
			// .unbind('click')
			// .bind('click', function() {
				// $back.trigger('click');
			// });
	};

	// boton de back desde el panel de interacciones al dashboard (summary+indicators+members)
	SupervisorDashboard.prototype.$back_button = function (campaign_id)
	{
		var campaign_dom_id = this.get_normalized_id(campaign_id);
		var $back = $('#sup_dash_back_dashboard_' + campaign_dom_id);
		return $back;
	};

	// boton de back desde el panel de interacciones al dashboard (summary+indicators+members)
	SupervisorDashboard.prototype.$full_screen_back_button = function (campaign_id)
	{
		return this.m_full_screen.$full_screen_back_button(campaign_id);
	};

	SupervisorDashboard.prototype.save_config = function (callback = $.noop)
	{
		this.m_config.save_current_configuration(callback);
	};

	SupervisorDashboard.prototype.get_config = function (callback = $.noop)
	{
		this.m_config.get_configuration(callback);
	};

	SupervisorDashboard.prototype.full_screen_notify_alarms = function(campaign_id, all_campaign_alarms)
	{
		if (!campaign_id || !this.m_full_screen) {
			return;
		}
		this.m_full_screen.notify_alarms(campaign_id, all_campaign_alarms);
	};

	SupervisorDashboard.prototype.notify_alarms = function (campaign_id, alarms, notification_request_type)
	{
		let all_campaign_alarms = this.m_alarms_notifications.queue_alarms_notifications(campaign_id, alarms, notification_request_type);
		if (all_campaign_alarms) {
			this.full_screen_notify_alarms(campaign_id, all_campaign_alarms);
		}
	};

	SupervisorDashboard.prototype.update_alarms = function ()
	{
		return this.m_alarms_notifications.update_notifications();
	};

	SupervisorDashboard.prototype.get_notification_alarm_by_indicator = function (info)
	{
		if (!this.m_alarms_notifications) {
			return;
		}
		return this.m_alarms_notifications.get_notification_alarm_by_indicator(info);
	};

	SupervisorDashboard.prototype.get_notification_alarm_message = function (info, calculated_current_value)
	{
		if (!this.m_alarms_notifications) {
			return;
		}
		return this.m_alarms_notifications.get_alarm_msg(info, calculated_current_value);
	};

	SupervisorDashboard.prototype.eval_notifications_alarms = function (campaign_id, category_id, entity)
	{
		if (!this.m_alarms_notifications) {
			return;
		}
		return this.m_alarms_notifications.eval_notifications_alarms(campaign_id, category_id, entity);
	};

	SupervisorDashboard.prototype.get_totals_alarms = function ()
	{
		if (!this.m_alarms_notifications) {
			return 0;
		}
		return this.m_alarms_notifications.get_totals_alarms();
	};

	SupervisorDashboard.prototype.toggle_alarms_notifications = function ()
	{
		this.m_alarms_notifications.toggle_popup();
	};

	SupervisorDashboard.prototype.__set_events_alarms_notifications_button = function ()
	{
		const self = this;
		const $toolbar_campaign_alarms = $('#sup_toolbar_options [action_id="campaign_alarms"]');
		$toolbar_campaign_alarms
			.unbind('click')
			.bind('click', function() {
				self.toggle_alarms_notifications();
			});
	};

	SupervisorDashboard.prototype.__show_or_hide_alarms_notifications_button = function ()
	{
		const $toolbar_campaign_alarms = $('#sup_toolbar_options [action_id="campaign_alarms"]');
		const $full_screen_campaign_alarms = $('#sup_full_screen_header_container [full_screen_header_buttons] [action_id="campaign_alarms"]');
		const has_permission = this.has_permission_to_any_campaign('sup_view_alarms_config');
		if (!has_permission) {
			$toolbar_campaign_alarms.hide();
			$full_screen_campaign_alarms.hide();
		}
		else {
			$toolbar_campaign_alarms.show();
			$full_screen_campaign_alarms.show();
		}
	};

	SupervisorDashboard.prototype.__set_handles_alarms_notifications = function ()
	{
		this.__set_events_alarms_notifications_button();
		this.__show_or_hide_alarms_notifications_button();
	};

	SupervisorDashboard.prototype.can_update_members_button = function (campaign_id = "")
	{
		return this.has_permission_to('sup_update_members', campaign_id);
	};

	/**
	 * Will setup the campaign menu options, show panels, close current campaign.
	 *
	 * @private
	 * @param {Object} current_node - Current campaign jQuery node.
	 */
	SupervisorDashboard.prototype.__setup_campaign_menu_handlers = function (current_node)
	{
		const self = this;

		var campaign_id = current_node.attr('campaign_id');
		var campaign_dom_id = this.get_normalized_id(campaign_id);
		var navigation = this.get_navigation(campaign_id);
		this.__set_events_back_button(campaign_id);
		this.__set_handles_alarms_notifications();

		const $members = self.find_element_inside_campaign(campaign_id, '.panelOptionItem [option_type="update_members"]');

		// Show/Hide options list handler
		var $campaign_option_list_nav = current_node.find('[panelOptionsNavegation]');
		var $campaign_option_list_config_view = current_node.find('[panelOptionsConfigView]');

		var hide_menu_options = function($container, callback = $.noop) {
			$container = $container || $campaign_option_list_nav;
			$container.unbind('clickoutside');
			$container.hide();
			callback();
		};

		var shown_panel_options = [];
		
		var show_menu_options = function( has_hide_panel )
		{
			let $back = self.$back_button(campaign_id);
			let $full_screen_back_button = self.$full_screen_back_button(campaign_id);
			shown_panel_options = [];

			var show_view = function( dashboard )
			{
				$back.trigger("click"); // Cuando selecciono una vista de interacciones (monitoring, archive, etc), previamente realizo "back" por si tenemos abierto la config de alarmas
				// open and initialize view
				dashboard.open_view(campaign_id);
				if ( dashboard.navigation_folder_id ) {
					let forceShow = dashboard.navigation_folder_id === "archive";
					navigation.select_folder( dashboard.navigation_folder_id, false, {forceShow : forceShow});
				}
				if ( shown_panel_options.length === 0 ) {
					$back.hide();
					$full_screen_back_button.hide();
				}
				else {
					$back.show();
					$full_screen_back_button.show();
				}
				// close menu, hide and adjust panel positions
				hide_menu_options();
				self.update_panel_positions();
				self.save_config();
			};

			var shown_view_options = [];
			var views_interactions_permissions = [
				{view_id: "monitor", permission_id: 'sup_view_monitoring', is_view_interactions: true},
				{view_id: "archive", permission_id:'sup_view_archive', is_view_interactions: true},
				{view_id: "appointment", permission_id:'sup_view_appointment', is_view_interactions: true},
				{view_id: "queued", permission_id:'sup_view_queued', is_view_interactions: true},
				{view_id: "waiting", permission_id:'sup_view_waiting', is_view_interactions: true},
				{view_id: "abandoned", permission_id:'sup_view_abandoned', is_view_interactions: true},
				{view_id: "global_search", permission_id:'sup_view_global_search'},
				{view_id: "bot_interactions", permission_id:'sup_monitoring_bot_interaction', is_view_interactions: true}
			];

			var get_dashboard = function( view_id )
			{
				var dashboard = self['m_dash_' + view_id];
				if ( dashboard ) {
					return dashboard;
				}
				var mapped_dashboards = {
					global_search: 'global_search'
				};
				var mapped_id = mapped_dashboards[view_id];
				return self['m_dash_' + mapped_id];
			};

			var has_view_permission = function (view, permission_id) {
				var dashboard = get_dashboard(view);
				if ( !dashboard || !dashboard.navigation_folder_id ) {
					return;
				}

				return self.has_permission_to( permission_id, campaign_id )? dashboard : undefined;
			};

			// Show campaign monitoring/archive/queued view

			let can_show_interactions = false;
			$.each( views_interactions_permissions, function( _ , view_data )
				{
					let view_id = view_data.view_id;
					let permission_id = view_data.permission_id;
					let is_view_interactions = view_data.is_view_interactions;
					let dashboard = has_view_permission(view_id, permission_id);

					if (dashboard) {
						shown_view_options.push( dashboard );
						navigation.show_folder_item( dashboard.navigation_folder_id );

						if (is_view_interactions) {
							can_show_interactions = true;
						}
						else{
							let $options_view = self.find_element_inside_campaign( campaign_id, `.panelOptionItem [view_id="${view_id}"]` );
							$options_view.parent().show();
							$options_view
								.off('click')
								.on('click', function () {
									show_view( dashboard );
								});
						}
					}
					else{
						//navigation.hide_folder_item( dashboard.navigation_folder_id );
						if (!is_view_interactions) {
							let $options_view = self.find_element_inside_campaign( campaign_id, `.panelOptionItem [view_id="${view_id}"]` );
							$options_view.parent().hide();
							$options_view.off('click');	
						}
					}
				});

				let $options_view_interactions = self.find_element_inside_campaign( campaign_id, `.panelOptionItem [view_id="interactions"]` );
				if (can_show_interactions && shown_view_options.length > 0) {
					$options_view_interactions.parent().show();
							$options_view_interactions
								.off('click')
								.on('click', function () {
									show_view( shown_view_options[0] ); // Siempre entra por monitoring
								});
				}
				else{
					$options_view_interactions.parent().hide();
					$options_view_interactions.off('click');
				}

			// Show campaign indicators/members/summary
			var $chks = self.find_element_inside_campaign( campaign_id, '.panelOptionItem [option_type="panel"]' );
			$.each( $chks, function( _, chk )
			{
				var $chk = $( chk );
				var panel_id = $chk.attr('panel_option_id');
				var $container = $chk.parent();

				var permission_id = self.panels_permissions[ panel_id ] || '';
				var has_permission = self.has_permission_to( permission_id, campaign_id );

				if ( has_permission ) {
					shown_panel_options.push( panel_id );
					$container.show();
					$chk.show();
				}
				else {
					$container.hide();
					$chk.hide();
					if ( has_hide_panel ) {
						self.trigger_column_update( campaign_id, panel_id, false );
					}
				}
			});

			// Si se visualizan mas de dos paneles y es fullscreen se muestran las opciones de layout
			$(`#sup_dash_show_main_dashboard_layout_${campaign_dom_id}`)[shown_panel_options.length > 1 && self.get_current_view_id(campaign_id) === "dashboard" && self.m_full_screen && self.m_full_screen.get_campaign_id() === campaign_id? "show":"hide"]();

			if ( shown_view_options.length === 0 ) {
			// if you not have permission any view, back to the main view
				$back.trigger( 'click' ); // esto sirve tanto en modo simple screen como full screen
			}
			else if ( shown_view_options.length === 1 ) {
				navigation.hide_folder_item( shown_view_options[ 0 ].navigation_folder_id );
			}
			else if ( shown_panel_options.length === 0 ) {
			// si no tiene acceso a ningun panel (summary, indicators, team), regresa a los view (monitoring, archive, etc)
				show_view( shown_view_options[ 0 ] );
			}
		};

		var bind_event_outside = function($container, callback = $.noop)
		{
			let is_visible = $container.is(':visible');
			if (is_visible) {
				return;
			}
			// Se agrega en un pequeno timeout el bindeo para que se biendee despues de visualizarlo correctamente, sino el mismo se cierra
			window.setTimeout(function() {
				$container.bind('clickoutside', function(e) {
						hide_menu_options($container, callback);
					}
				);
			});
		};

		$('#sup_dash_panel_menu_' + campaign_dom_id)
			.unbind('click')
			.bind('click', function() {
				const $this = $(this);
				$this.removeClass("btnActive").addClass("btnActive");
				show_menu_options();
				bind_event_outside($campaign_option_list_nav, function() {
					$this.removeClass("btnActive");
					$("#sup_full_screen_header_container").find('[action_id="campaign_menu"]').removeClass("btnActive");
				});
				$members.parent()[self.can_update_members_button(campaign_id)? "show":"hide"]();
				$campaign_option_list_nav.toggle();
				self.__select_dashboard_nav_option(campaign_id);
			});

		$('#sup_dash_panel_menu_config_view_' + campaign_dom_id)
			.unbind('click')
			.bind('click', function() {
				const $this = $(this);
				$this.removeClass("btnActive").addClass("btnActive");
				show_menu_options();
				bind_event_outside($campaign_option_list_config_view, function() {
					$this.removeClass("btnActive");
					$("#sup_full_screen_header_container").find('[action_id="campaign_menu_config_view"]').removeClass("btnActive");
				});
				$campaign_option_list_config_view.toggle();
			});

		$('#sup_dash_panel_show_full_screen_' + campaign_dom_id)
			.unbind('click')
			.bind('click', function() {
				self.full_screen_toggle_visibility({full_screen: true, campaign_id: campaign_id});
			});

		// Checkboxes -> Show/Hide columns (summary, indicators, team)
		current_node.find('.panelOptionCheck').unbind('change').bind('change', function () 
		{
			var $current_checkbox = $(this);
			var current_checked = $current_checkbox.prop('checked');
			var panel_id = $current_checkbox.attr('panel_option_id');
			var success = self.trigger_column_update(campaign_id, panel_id, current_checked);

			if ( success ) {
				// Save configuration
				self.save_config();
			}
			else {
				inConcert.app.getMainWindow().showBannerMessage({
					message : "<%=_T('NOT HAS PERMISSIONS FOR THIS ACTION IN VIEW PANELS')%>",
					timeout : 5000
				});
				$current_checkbox.prop( 'checked', false );
			}
		});


		// Show campaign alarm
		const $buttons_alarm = self.find_element_inside_campaign( campaign_id, '.panelOptionItem [option_type="alarms_config"]' );
		$.each( $buttons_alarm, function( _, button )
		{
			var $button = $( button );
			var $container = $button.parent();

			var has_permission = self.has_permission_to( 'sup_view_alarms_config', campaign_id );

			if ( has_permission ) {
				$container.show();
				$button
					.off('click')
					.on('click', function () {

						self.leave_bot_view(campaign_id);
						self.m_dash_config.show(campaign_id);
						// close menu, hide and adjust panel positions
						hide_menu_options();
					})
					.show();
			}
			else {
				$container.hide();
				$button.hide().off('click');
			}
		});


		// Show campaign bots
		const $buttons_bots = self.find_element_inside_campaign( campaign_id, '.panelOptionItem [option_type="bots"]' );
		$buttons_bots
				.off('click')
				.on('click', function () {
					if (!self.__is_campaign_view_bots(campaign_id)) {
						self.load_bots_dashboard(campaign_id);
					}
				});

		// Show main dashboard
		const $main_dashboard = self.find_element_inside_campaign(campaign_id, '.panelOptionItem [option_type="main_dashboard"]');
		$main_dashboard
			.off('click')
			.on('click', function () {
				const $back = self.$back_button(campaign_id);
				$back.trigger('click');
			});	
			
		// Show popup update members
		$members
			.off('click')
			.on('click', function () {
				self.open_popup_members(campaign_id);
			});	
		
		// hide panels without permissions
		show_menu_options(true);

		//Event: Configurations view menu layout opcion  
		$.each($('#sup_dash_show_main_dashboard_layout_' + campaign_dom_id).find('[option_type="layout"]'), function (_, layout) {
			const $layout = $(layout);
			$layout
				.off('click')
				.on('click', function () {
					const $this = $(this);
					let value = $this.val(); // layout1, layout2, layout3
					var $siblings = $this.parent().siblings();
					const isChecked = $this.prop('checked');
					if (isChecked) {
						$siblings.find('input').prop('checked', false);
						self.update_main_dashbord_layout(campaign_id, value);
					}
				});

			//Cuando se inicia el renderizado de la campanna (chequeamos el layout por defecto)
			if ((shown_panel_options.length === 2 && $layout.val() === "layout1") || (shown_panel_options.length === 3 && $layout.val() === "layout2")) {
				$layout.prop('checked', true).click();
			}
		});

		// Event show popup campain account/als
		const $accounts_selector = $(`#sup_dash_panel_campaign_accounts_${campaign_dom_id}`);
		$accounts_selector
			.off('click')
			.on('click', function () {
				const $parent = $(this);
				self.show_campaign_accounts_selector(campaign_id, $parent);
			});
			
		// Event show popup bots filters
		const $bots_filters_selector = $(`#sup_dash_panel_campaign_bots_filter_${campaign_dom_id}`);
		$bots_filters_selector
			.off('click')
			.on('click', function () {
				const $parent = $(this);
				self.show_campaign_bots_filters(campaign_id, $parent);
			});
	};

	SupervisorDashboard.prototype.__select_dashboard_nav_option = function(campaign_id) {

		//Se remueve la clase 
		$.each(this.find_element_inside_campaign( campaign_id, `[panelOptionsNavegation] .panelOptionItem` ), function(_, listPanelItem) {
			$(listPanelItem).removeClass("selected");
		});

		//Obtengo la vista actual para marcar e indicar la opcion del menu actual visible.
		const campaign_dom_id = this.get_normalized_id(campaign_id);
		switch(this.m_current_view_by_campaign[campaign_id]) {
			case "monitoring":
			case "appointment":
			case "archive":
			case "queued":
			case "waiting":
			case "abandoned":
				// show_conversations
				$(`#sup_dash_show_view_conversations_${campaign_dom_id}`).parent().removeClass("selected").addClass("selected");
				break;
			case "global_search":
				// show_global_search
				$(`#sup_dash_show_view_global_search_${campaign_dom_id}`).parent().removeClass("selected").addClass("selected");
					break;
			case "dashboard_config":
				//show_alarms_config
				$(`#sup_dash_show_alarm_config_${campaign_dom_id}`).parent().removeClass("selected").addClass("selected");
				break;
			case "dashboard":
				if (this.__is_campaign_view_bots(campaign_id)) {
					$(`#sup_dash_show_bots_${campaign_dom_id}`).parent().removeClass("selected").addClass("selected");
				}
				else{
					//main dashboard
				$(`#sup_dash_show_main_dashboard_${campaign_dom_id}`).parent().removeClass("selected").addClass("selected");
				}
				break;
			
		}
	};

	SupervisorDashboard.prototype.__get_dash_current_layout_selected = function(campaign_id) {

		return $('#sup_dash_show_main_dashboard_layout_' + this.get_normalized_id(campaign_id)).find('[option_type="layout"]:checked').val();
	};

	SupervisorDashboard.prototype.update_main_dashbord_layout = function (campaign_id, layout) {
		//Total paneles visibles
		const self = this;
		let panelChks = self.find_element_inside_campaign(campaign_id, '[panelOptionsConfigView] [option_type="panel"]').filter(function (_, panel) {
			const $panel = $(panel);
			return $panel.prop('checked');
		});

		const total_panels = panelChks.length;
		//Si no existe mas de 2 paneles o no es main dashboard o no es fullscreen, dejo clase col y salgo
		if (total_panels < 2 || this.get_current_view_id(campaign_id) !== "dashboard" || !(this.m_full_screen && this.m_full_screen.get_campaign_id() === campaign_id)) {
			return self.__clean_main_dashboard_layout_classes(campaign_id);
		}

		//Si no viene layout por param, obtengo el seleccionado
		layout = layout || this.__get_dash_current_layout_selected(campaign_id) || (total_panels === 3? "layout2": "layout1");
		/*
		total_panels === 3:
					layout1                     layout2                           layout3
			[col-6 / col-3 / col-3] o [col-3 / col-6 / col-3] (default) o [col-3 / col-3 / col-6]
		
		total_panels === 2:
					layout1                     layout2                           layout3
			[col-6 / col-6] (default)      o [col-9 / col-3]                  o [col-3 / col-9]
		
		Segun el layout y total de paneles cambiar las clases de los paneles
		*/
		let layouts = {
			layout1 : {
				three_cols: ["col-6", "col-3", "col-3"],
				two_cols: ["col-6", "col-6"]
			},
			layout2 : {
				three_cols: ["col-3", "col-6", "col-3"],
				two_cols: ["col-9", "col-3"]
			},
			layout3 : {
				three_cols: ["col-3", "col-3", "col-6"],
				two_cols: ["col-3", "col-9"]
			}
		};

		let cols = layouts[layout][total_panels > 2? "three_cols": "two_cols"];
		$.each(panelChks, function(index, panel) {
			const panel_id = $(panel).attr('panel_option_id');
			const colClass = cols[index];
			const $panel = self.find_element_inside_campaign(campaign_id, `[viewer_type="operations"] [panel_id="sup_dash_${panel_id}_panel"]`);
			$panel.removeClass("col col-3 col-6 col-9").addClass(colClass);
		});

	};

	SupervisorDashboard.prototype.__clean_main_dashboard_layout_classes = function(campaign_id)
	{
		$.each(this.find_element_inside_campaign(campaign_id, '[viewer_type="operations"] [panel_id]'), function(_, panel) {
			const $panel = $(panel);
			$panel.removeClass("col col-3 col-6 col-9").addClass("col");
		});
	};

	/*
		Alternar entre los modos Full Screen y Simple Screen
	*/
	SupervisorDashboard.prototype.full_screen_toggle_visibility = function(params)
	{
		const campaign_id = params.campaign_id || this.full_screen_selected_tab_id();
		const is_view_bot = this.__is_campaign_view_bots(campaign_id);
		if (is_view_bot) {
			params.campaign_id = campaign_id;
			const $sup_dashboard_container = this.get_campaign_bots_node(params.campaign_id);
			const appFormBot = $sup_dashboard_container.find('#icc-processflow-dashboard')[0];
			params.fromBotDashboard = true;
			appFormBot.setFullScreenMode(params.fullscreen);
			this.m_full_screen.toggle_visibility(params);
			appFormBot.resizeCharts();
			return;
		}

		this.m_full_screen.toggle_visibility(params);
	};

	SupervisorDashboard.prototype.is_full_screen_mode = function()
	{
		return this.m_full_screen.is_full_screen_mode();
	};

	SupervisorDashboard.prototype.full_screen_selected_tab_id = function()
	{
		return this.m_full_screen.get_selected_tab_id();
	};

	/**
	 * Will set the correct left position of the campaign node.
	 * The calculation is: (Summation of each panel width) + (total panels * 10px of offset)
	 *
	 * @private
	 * @param {object} $current_node - Current campaign node.
	 */
	SupervisorDashboard.prototype.__set_campaign_node_position = function ($current_node)
	{
		var left_value = this.update_panel_positions();
		$current_node.css('left', left_value);
	};

	/**
	 * When a campaign node was deleted, will set the correct left position of remainings campaigns nodes.
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.update_panel_positions = function ()
	{
		var left_value = 0;
		$('#supervisor_main .supervisorPanelContainer').each(function (_, e) {
			var $camp = $(e);
			if ($camp.is(':visible')) {
				$camp.css('left', left_value);
				left_value += $camp.width() + 10;
			}
		});
		return left_value;
	};

	/**
	 * Get the campaign timestamp for a given campaign,
	 * this value come from the get_supervisor_view request
	 *
	 * @public
	 * @param  {string} campaign_id - The campaign id
	 * @return {number} - The campaign timestamp
	 */
	SupervisorDashboard.prototype.get_campaign_timestamp = function(campaign_id) {
		return this.m_rendered_campaigns[campaign_id] ? this.m_rendered_campaigns[campaign_id].campaign_timestamp : 0;
	};

	/**
	 * Get the campaign title for a given campaign.
	 * Will add workgroups to the name is the supervisor has at least one configured.
	 *
	 * @private
	 * @param  {string} campaign_id - The campaign id.
	 * @param {object} campaign_object
	 * @param {array} campaign_workgroups
	 * @param {object} event_queue
	 * @return {string} - The campaign title.
	 */
	SupervisorDashboard.prototype.get_campaign_title = function(campaign_id, campaign_object, campaign_workgroups, event_queue)
	{
		const is_allegro_campaign = campaign_object.isAllegro();
		let campaign_name = campaign_object.getDescription();
		let campaign_title = '';
		let config = {};
		
		let campaign_dom_id = this.get_normalized_id(campaign_id);		
		$("#sup_dash_campaign_" + campaign_dom_id).removeClass("showChips");
		if ( event_queue && event_queue.config ) {
			config = event_queue.config;
			if (!is_allegro_campaign) {
			// unicamente se visualiza en el titulo de la ventana de la campana, la informacion de la cuenta y de los ALs, 
			//  cuando la campana NO es de integracion con allegro
				if ( config.account_id ) {
					var account = campaign_object.getAccount(config.account_id);
					var account_name = campaign_object.getAccountDescription(account);
					var channel_class = "accountSelect" + account.proxyType.toUpperCase();
					$("#sup_dash_campaign_" + campaign_dom_id).addClass("showChips");
					if ( config.al_id ) {
						var al = campaign_object.getAttentionLevel(config.al_id);
						var al_name = al ? al.name : config.al_id;
						campaign_title = format('<span class="campaignNameSpan" title="{0}">{0}</span> <span class="accountName {1}" title="{2}">{2}</span> <span class="attentionLevelName" title="{3}">{3}</span>', campaign_name, channel_class, account_name, al_name);
					}
					else {
						campaign_title = format('<span class="campaignNameSpan" title="{0}">{0}</span> <span class="accountNameSelect {1}" title="{2}">{2}</span>', campaign_name, channel_class, account_name);
					}
				}
			}
		}

		if ( campaign_title === '' ) {
			campaign_title = format('<span class="campaignNameSpan" title="{0}">{0}</span>',campaign_name);
		}

		campaign_workgroups = campaign_workgroups ? campaign_workgroups : (this.m_workgroups.get_campaign_workgroups(campaign_id) || []);
		if (campaign_workgroups.length > 0) {
			// Si la campana tiene Workgropus, y es de tipo "call allegro" o NO tiene cuentas en el selector de campanas (es decir ha seleccionado solamente campana),
			//	ahi SI debe de mostrar el wkg en el titulo de la campana.
			if (is_allegro_campaign || !config.account_id) {
				var workgroups_names = '';
				campaign_workgroups.forEach(function(wg_id, index){
					var delimiter = index + 1 === campaign_workgroups.length ? '' : ' - ';
					workgroups_names += wg_id + delimiter;
				});
				campaign_title = format('{0} <span class="campaignWorkgroup">( {1} )</span>', campaign_title, workgroups_names);
			}
		}

		return campaign_title;
	};

	/**
	 * Given a campaign account, will return the status id to show
	 * and the status detail.
	 *
	 * @param  {Object} account       - campaign account returned from application.
	 * @return {Object} status
	 * @return {String} status.id     - Status Id (used inside the markup class)
	 * @return {String} status.detail - Status Detail (ok, disabled, or error if any)
	 */
	SupervisorDashboard.prototype.__get_campaign_account_status = function(account){
		var status = {
			id : 'DISABLED',
			detail : "<%=_T('Account disabled')%>"
		};
		if (account.state === '1'){
			if (account.status) {
				status.id = account.status.status !== 'OK' ? 'ERROR' : 'OK';
				status.detail = _T(account.status.status);
			} else {
				status.detail = "<%=_T('Proxy is not checking this account yet, please wait, it is coming up!')%>";
			}
		}
		return status;
	};

	/**
	 * Will return a Mustache rendered template for an account
	 * to use inside the campaign detail information.
	 *
	 * @private
	 * @param  {String} campaign_id     - The campaign id
	 * @param  {Object} account         - The account object from campaign
	 * @return {String}                 - Account template rendered
	 */
	SupervisorDashboard.prototype.__get_campaign_account_detail_node = function( campaign_id, account)
	{
		var campaign_config = this.get_campaign_selected_config(campaign_id);

		var hidden = '';
		if ( campaign_config.account_id && campaign_config.account_id !== account.id ) {
			hidden = 'hidden';
		}

		var attention_level_list = this.__get_campaign_account_attention_levels(campaign_id, account);
		var campaign_dom_id = this.get_normalized_id(campaign_id);

		return Mustache.render(this.m_templates.campaign_detail_account_tpl, {
			two_points             : "<%=_T(':')%>",
			distribution_name      : "<%=_T('Distribution')%>",
			status_name            : "<%=_T('Status')%>",
			attention_level_name   : "<%=_T('Attention levels')%>",
			attention_level_list   : attention_level_list,
			normalized_id          : campaign_dom_id,
			account_id             : nid(account.id),
			account_type           : account.proxyType,
			account_name           : account.name,
			distribution           : _T(account.CCAlgorithm),
			status                 : status.id,
			status_detail          : status.detail,
			hidden                 : hidden
		});
	};

	SupervisorDashboard.prototype.__get_campaign_account_attention_levels = function( campaign_id, account )
	{
		var none = "<%=_T('None')%>";
		if ( !campaign_id || !account ) {
			return none;
		}
		var campaign = inConcert.app.getCampaign(campaign_id);
		if ( !campaign ) {
			return none;
		}
		var campaign_config = this.get_campaign_selected_config(campaign_id);
		var attention_level_list;
		if ( campaign.has_attention_levels_configured_in_account( account.id ) ) {
			$.each(account.AttentionLevels || {}, function(_, assigned_al )
			{
				var al_id = assigned_al.attention_level_id;
				// Muestra el AL si:
				//  No tiene seleccionado AL en el menu "campaign selector", o si el que se encuentra seleccionado es igual al current (assigned_al)
				if ( !campaign_config.al_id || campaign_config.al_id === al_id ) {
					var att_level = campaign.getAttentionLevel( al_id );
					if ( att_level ) {
						if ( !attention_level_list ) {
							attention_level_list = att_level.name;
						}
						else {
							attention_level_list = attention_level_list + ' / ' + att_level.name;
						}
					}
				}
			});
		}
		return attention_level_list || none;
	};

	SupervisorDashboard.prototype.get_slices = function()
	{
		return {
			'Last15Minutes': "<%=_T('15 minutes')%>",
			'LastHour'     : "<%=_T('1 hour')%>",
			'Daily'        : "<%=_T('Today')%>"
		};
	};

	/**
	 * Given a campaign will build a complete markup using Mustache templating.
	 *
	 * @param  {String} campaign_id - The campaign identificator.
	 * @param  event_queue
	 * @return {jQuery object}      - The complete markup for this campaign.
	 */
	SupervisorDashboard.prototype.__$get_campaign_markup = function(campaign_id, event_queue)
	{
		var self = this;

		var campaign = inConcert.app.getCampaign(campaign_id);
		if (!campaign ) {
			return;
		}
		var valid_until = campaign.AlwaysValid ? "<%=_T('Always Valid.')%>" : campaign.ValidationEndTime;

		var account_list = [];
		if (campaign.accounts && campaign.accounts.length > 0){

			var campaign_config = this.get_campaign_selected_config(campaign_id);
			campaign.accounts.forEach(function(account)
			{
				var hidden = '';
				if ( campaign_config.account_id && campaign_config.account_id !== account.id ) {
					hidden = 'hidden';
				}
				var attention_level_list = self.__get_campaign_account_attention_levels(campaign_id, account);
				var status = self.__get_campaign_account_status(account);

				account_list.push( {
					account_id             : nid(account.id),
					account_type           : account.proxyType,
					account_name           : account.name,
					distribution_name      : "<%=_T('Distribution')%>",
					distribution           : _T(account.CCAlgorithm),
					status_name            : "<%=_T('Status')%>",
					status                 : status.id,
					status_detail          : status.detail,
					attention_level_name   : "<%=_T('Attention levels')%>",
					attention_level_list   : attention_level_list,
					hidden                 : hidden
				});
			});
		}

		var workgroups = this.m_workgroups.get_campaign_workgroups(campaign_id);
		var campaign_view = {
			campaign_id         : campaign_id,
			normalized_id       : this.get_normalized_id(campaign_id),
			hidden_batch_actions: 'hidden',

			// campaign info
			valid_until         : valid_until,
			on_schedule         : campaign.isOnSchedule(),
			timezone            : campaign.ValidationTimeZone,
			workgroups          : workgroups.length > 0,
			workgroups_names    : workgroups,
			accounts            : account_list,

			// messages to be translated.
			two_points          : "<%=_T(':')%>"                   ,
			valid_until_label   : "<%=_T('Valid Until')%>"         ,
			schedule_label      : "<%=_T('On Schedule')%>"         ,
			timezone_label      : "<%=_T('Timezone')%>"            ,
			workgroups_label    : "<%=_T('Workgroups')%>"          ,
			back                : "<%=_T('Back')%>"                ,
			summary             : "<%=_T('Summary')%>"             ,
			indicators          : "<%=_T('Indicators')%>"          ,
			members             : "<%=_T('Members')%>"             ,
			team                : "<%=_T('Team')%>"                ,
			show_global_search  : "<%=_T('Global search')%>"       ,
			show_alarms_config  : "<%=_T('Alarms Config')%>"       ,
			hide_campaign       : "<%=_T('Hide Campaign')%>"       ,
			monitoring          : "<%=_T('Monitoring')%>"          ,
			appointment         : "<%=_T('Appointment')%>"         ,
			archive             : "<%=_T('Archive')%>"             ,
			queued              : "<%=_T('Queued')%>"              ,
			broadcast_message   : "<%=_T('Broadcast Message')%>"   ,
			broadcast_send      : "<%=_T('SEND')%>"                ,
			broadcast_cancel    : "<%=_T('CANCEL')%>"              ,
			show_avatars        : "<%=_T('Avatar')%>"        ,
			show_names          : "<%=_T('Name')%>"          ,
			show_offline_agents : "<%=_T('Show offline agents')%>" ,
			show_kickouts       : "<%=_T('Show Kickout Alarms')%>" ,
			show_state_time     : "<%=_T('Time in state')%>"  ,
			show_state_name     : "<%=_T('Show State')%>"          ,
			show_id             : "<%=_T('Id')%>"          ,
			show_state_description     : "<%=_T('Description')%>"          ,
			show_state_visual_indicator     : "<%=_T('Visual indicator')%>"          ,
			show_workgroups     : "<%=_T('Workgroups')%>"          ,
			show_total_conversations     : "<%=_T('Interactions')%>"          ,
			show_total_conversations_by_channels     : "<%=_T('Interactions by channel')%>"          ,
			show_bots           : "<%=_T('Bots')%>"			       ,
			show_main_dashboard : "<%=_T('Main dashboard')%>"			       ,
			show_conversations   :  "<%=_T('Show conversations')%>"			       ,
			show_main_bots		: "<%=_T('Bots dashboard')%>"			       ,
			show_alarm_management : "<%=_T('Alarm management')%>"			       ,
			show_all_interactions	: "<%=_T('Interactions')%>"			       ,
			show_agent_configuration	: "<%=_T('Update agents')%>"			       ,

			// header buttons tooltips
			refresh_tooltip     : "<%=_T('Refresh')%>"             ,
			campaign_tooltip    : "<%=_T('Campaign Selector')%>"   ,
			alarms_tooltip      : "<%=_T('Alarms')%>"              ,
			config_tooltip      : "<%=_T('Update agents')%>"       ,
			menu_tooltip          : "<%=_T('View navegation menu')%>",
			menu_config_view_tooltip        : "<%=_T('View configurations menu')%>"                ,
			dashboard_bots_tooltip: "<%=_T('Bots')%>"			   ,
			detail_tooltip      : "<%=_T('Detail')%>"              ,
			full_screen_tooltip : "<%=_T('Full Screen')%>"         ,
			simple_screen_tooltip:"<%=_T('Normal Screen')%>"       ,
			dashboard_tooltip	: "<%=_T('Dashboard')%>"           ,
			filter_tooltip	    : "<%=_T('Filter')%>"              ,
			set_view_tooltip	    : "<%=_T('Set view')%>"              ,
			configview_tooltip	: "<%=_T('View settings')%>"              ,
			back_tooltip: "<%=_T('Back')%>"              ,
			time_ago: {
				label: "<%=_T('Time Ago')%>",
				data: this.get_slices()
			}
		};

		// Create node and calculate left position
		var $campaign_node = $( Mustache.render(this.m_templates.campaign_panel_tpl, campaign_view));
		// set title
		var title = this.get_campaign_title(campaign_id, campaign, workgroups, event_queue);
		var campaign_dom_id = this.get_normalized_id(campaign_id);
		$campaign_node.find('[id="sup_dash_panel_name_' + campaign_dom_id + '"]').html(title);
		//add allegro classes
		if (this.isAllegro(campaign_id)) {
			$campaign_node.addClass("allegroCampaign");
		}

		return $campaign_node;
	};

	SupervisorDashboard.prototype.get_navigation = function (campaign_id)
	{
		return this.navigation[campaign_id];
	};

	SupervisorDashboard.prototype.getCommonConfigRules = function (rules, campaign_id) 
	{
		const campaign_config = this.get_campaign_selected_config(campaign_id);
		rules = rules || [];

		if (campaign_config.account_id && !rules.find(rule => rule.id === "account")) {
			rules.push({
				id : "account",
				operator : "equal",
				value : campaign_config.account_id
			});
		}

		if (campaign_config.al_id && !rules.find(rule => rule.id === "attention_level")) {
			rules.push({
				id : "attention_level",
				operator : "equal",
				value : campaign_config.al_id
			});
		}

		return rules;
	};

	SupervisorDashboard.prototype.getQueryBuilderFilter = function (campaign_id) 
	{
		var campaign_dom_id = this.get_normalized_id(campaign_id);
		return this.queryBuilderObj[campaign_dom_id].getFilter();
	};

	/**
	 */
	SupervisorDashboard.prototype.__build_navigation = function (campaign_id)
	{
		var self = this;

		var campaign_dom_id = this.get_normalized_id(campaign_id);
		var container_id = 'sup_dash_interaction_viewer_navigation_' + campaign_dom_id;
		var $dynamic = $('#sup_dash_interaction_viewer_dynamic_' + campaign_dom_id);		
		var $queryBuilderRefreshButon = $dynamic.find("[query_builder_refresh]");
		var $queryBuilderCleanButon = $dynamic.find("[query_builder_clean]");
		var $queryBuilderCloseButon = $dynamic.find("[query_builder_close_times]");
		var $queryBuilderMessagesContainer = $dynamic.find(".inlineMessagesContainer");

		var campaign = inConcert.app.getCampaign(campaign_id);
		if (!campaign) {
			return;
		}

		var on_ready_to_refresh = function() {
			/*
				Se agrega validacion para determinar si se puede refrescar.
				Por defecto true, se mantiene como siempre estuvo.
				Esto se hace para cuando se cambie de panel, no refrescar cuando se este armando el filtro.
				De esta forma no se refresca 3 o 4 veces, solo una ya cuando este creado el filtro.
			*/
			let current_dashboard = self.__get_current_dashboard(campaign_id);
			let can_refresh = true;
			if (current_dashboard && typeof current_dashboard.can_refresh_interactions === "function") {
				can_refresh = current_dashboard.can_refresh_interactions();
			}
			if (can_refresh) {
				self.onReadyToRefresh(campaign_id);	
			}
		};

		var on_rule_added = function() {
			 $queryBuilderCleanButon.prop("disabled", false);
		};

		var $queryBuilder = $dynamic.find('[query_builder]');
		var inconcert_querybuilder_config = {
			container: $queryBuilder,
			fromApp: "supervisor",
			supervisorCampaignId: campaign_id,
			onReadyToRefresh: on_ready_to_refresh,
			onRuleAdded: on_rule_added,
			panels: [
				{
					panelId: "active",
					filters:["agent", "agentname", "start_date", "contact_address", "content", "ticket", "type", "events_unanswered", "events_unread", "has_attachments", "initiative", "has_draft"]
				},
				{
					panelId: "appointment",
					filters:["agent", "agentname", "start_date", "contact_address", "content", "ticket", "type", "events_unanswered", "events_unread", "has_attachments", "initiative"]
				},
				{
					panelId: "archive",
					filters:["agent", "agentname", "businessdata", "start_date", "contact_address", "contact", "contact_name", "disposition_code", "content", "ticket", "type", "events_unread", "has_attachments", "initiative"]
				},
				{
					panelId: "queued",
					filters:["start_date", "contact_address", "contact", "contact_name", "content", "ticket", "type", "events_unanswered", "events_unread", "has_attachments", "initiative"]
				},
				{
					panelId: "waiting",
					filters:["start_date", "contact_address", "contact", "contact_name", "content", "ticket", "type", "events_unanswered", "events_unread", "has_attachments", "initiative"]
				},
				{
					panelId: "abandoned",
					filters:["start_date", "contact_address", "contact", "contact_name", "content", "ticket", "type", "has_attachments", "initiative"]
				},
				{
					panelId: "bot_interactions",
					filters:["type", "need_help"]
				}
			]
		};
		this.queryBuilderObj[campaign_dom_id] = new inConcert.inconcert_querybuilder(inconcert_querybuilder_config);

		$queryBuilderRefreshButon.on('click', function() {
			self.onReadyToRefresh(campaign_id);
		});

		$queryBuilderCleanButon.on('click', function() {
			self.queryBuilderObj[campaign_dom_id].cleanAll();
		});

		$queryBuilderCloseButon.on('click', function() {
			self.queryBuilderObj[campaign_dom_id].toggle();
		});		

		$queryBuilderMessagesContainer.on('click', function() {
			self.queryBuilderObj[campaign_dom_id].toggle();
		});		

		var $static = $('#sup_dash_interaction_viewer_static_container_' + campaign_dom_id);
		var $global_search_queryBuilderRefreshButon = $static.find("[query_builder_refresh]");
		var $global_search_queryBuilderCleanButon = $static.find("[query_builder_clean]");
		var $global_search_queryBuilderCloseButon = $static.find("[query_builder_close_times]");
		var $global_search_queryBuilder = $static.find('[query_builder]');

		var on_global_search_rule_added = function() {
			 $global_search_queryBuilderCleanButon.prop("disabled", false);
		};

		var global_search_inconcert_querybuilder_config = {
			container: $global_search_queryBuilder,
			fromApp: "supervisor",
			supervisorCampaignId: campaign_id,
			onReadyToRefresh: on_ready_to_refresh,
			onRuleAdded: on_global_search_rule_added,
			panels: [
				{
					panelId: "global_search",
					filters:["agent", "agentname", "businessdata", "start_date", "contact_address", "contact", "contact_name", "disposition_code", "content", "ticket", "type", "account", "state", "attention_level", "events_unanswered", "events_unread", "has_attachments", "initiative", "has_draft", "need_help", "id" /*interaction_id*/]
				}
			]
		};
		this.global_search_queryBuilderObj[campaign_dom_id] = new inConcert.inconcert_querybuilder(global_search_inconcert_querybuilder_config);

		$global_search_queryBuilderRefreshButon.on('click', function() {
			self.onReadyToRefresh(campaign_id);
		});		
		
		$global_search_queryBuilderCleanButon.on('click', function() {
			self.global_search_queryBuilderObj[campaign_dom_id].cleanAll();
		});

		$global_search_queryBuilderCloseButon.on('click', function() {
			self.global_search_queryBuilderObj[campaign_dom_id].toggle();
			$("#sup_dash_campaign_" + campaign_dom_id).toggleClass("showingQueryBuilder");
		});		

		this.navigation[campaign_id] = new inConcert.ConversationsNavigation(
		{
			from_app: 'supervisor',
			queryBuilderObj: this.global_search_queryBuilderObj[campaign_dom_id],
			container_id: container_id,
			global_search_container_id: 'sup_dash_interaction_viewer_static_' + campaign_dom_id, // debe existir en inboxmaster.html
			// El valor del campo "id" debe coincidir con lo que se almacena en el IR para ES, en la funcion store_in_elasticsearch en la variable "data" del archivo (ir.lua)
			global_search_columns_sort : {
				creation_date                 : { order: 'desc', 'default': false, id: 'start_timestamp' },
				last_event_date_or_state_date : { order: 'desc', 'default': true , id: 'state_timestamp' },
				contact_name                  : { order: 'asc' , 'default': false, id: 'contact_name'    },
				owner                         : { order: 'asc' , 'default': false, id: 'agent'           }
			// este se ordena segun la traduccion asi que puede quedar mal ordenado
				//state    : { order: 'asc' , 'default': false, id: 'state'           }
			// A estos no los ordena correctamente, ya sea porque es una imagen o es un campo calculado cliente side.
				//type     : { order: 'asc' , 'default': false, id: 'type'            }
			},
			campaigns: [campaign],
			parent : this,
			global_search_filter_by_campaign: { type: 'UNIQUE_CAMPAIGN', campaign_id: campaign.Id },
			permissions: {
				active        : 'sup_view_monitoring',
				appointment   : 'sup_view_appointment',
				archive       : 'sup_view_archive',
				queued        : 'sup_view_queued',
				tickets       : 'sup_tickets_mng',
				waiting       : 'sup_view_waiting',
				abandoned     : 'sup_view_abandoned',
//				spam          : 'int_view_spam'//,
				global_search : 'sup_view_global_search',
				bot_interactions : 'sup_monitoring_bot_interaction'
			},
			callbacks: {
				active: {
					show: function(options) {
						$static.hide();

						//valido se no refrescar las interacciones hasta q termine de levantarse el filtro.
						let skip_set_can_refresh_interactions = false;
						if (options && options.can_refresh_interactions === false) {
							//Si ya esta seteado para no refrescar, es porque viene de open_view
							skip_set_can_refresh_interactions = true;
						}
						else{
							self.m_dash_monitor.set_can_refresh_interactions(false);
						}
						
						self.queryBuilderObj[campaign_dom_id].switchToPanel("active");
						$dynamic.show();
						var $campaign = self.get_campaign_node(campaign_id);

						if (!skip_set_can_refresh_interactions) {
							self.m_dash_monitor.set_can_refresh_interactions(true);
						}
						
						self.m_dash_monitor.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_monitor.execute_new_search(campaign_id);
					}
				},
				appointment: {
					show: function() {
						$static.hide();
						self.queryBuilderObj[campaign_dom_id].switchToPanel("appointment");
						$dynamic.show();
						var $campaign = self.get_campaign_node(campaign_id);
						self.m_dash_appointment.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_appointment.execute_new_search(campaign_id);
					}
				},
				archive: {
					show: function(options) {
						var $campaign = self.get_campaign_node(campaign_id);
						if ( $campaign && $campaign.length > 0 ) {
							self.set_current_view_id( campaign_id, self.m_dash_archive.m_id );
						}
						$static.hide();
						
						//valido se no refrescar las interacciones hasta q termine de levantarse el filtro.
						let skip_set_can_refresh_interactions = false;
						if (options && options.can_refresh_interactions === false) {
							//Si ya esta seteado para no refrescar, es porque viene de open_view
							skip_set_can_refresh_interactions = true;
						}
						else{
							self.m_dash_archive.set_can_refresh_interactions(false);
						}

						self.queryBuilderObj[campaign_dom_id].switchToPanel("archive"); //Se arma todo el filtro para archivados
						$dynamic.show();

						if (!skip_set_can_refresh_interactions) {
							self.m_dash_archive.set_can_refresh_interactions(true);
						}
						self.m_dash_archive.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_archive.refresh(campaign_id);
					}
				},
				queued: {
					show: function() {
						$static.hide();
						self.queryBuilderObj[campaign_dom_id].switchToPanel("queued");
						$dynamic.show();
						var $campaign = self.get_campaign_node(campaign_id);
						self.m_dash_queued.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_queued.execute_new_search(campaign_id);
					}
				},
				waiting: {
					show: function() {
						$static.hide();
						self.queryBuilderObj[campaign_dom_id].switchToPanel("waiting");
						$dynamic.show();
						var $campaign = self.get_campaign_node(campaign_id);
						self.m_dash_waiting.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_waiting.execute_new_search(campaign_id);
					}
				},
				abandoned: {
					show: function() {
						$static.hide();
						self.queryBuilderObj[campaign_dom_id].switchToPanel("abandoned");
						$dynamic.show();
						var $campaign = self.get_campaign_node(campaign_id);
						self.m_dash_abandoned.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_abandoned.execute_new_search(campaign_id);
					}
				},
				bot_interactions: {
					show: function() {
						$static.hide();
						self.queryBuilderObj[campaign_dom_id].switchToPanel("bot_interactions");
						$dynamic.show();
						var $campaign = self.get_campaign_node(campaign_id);
						self.m_dash_bot_interactions.init($campaign);
						self.save_config();
						return true;
					},
					filter_label: function(_) {
						self.m_dash_bot_interactions.execute_new_search(campaign_id);
					}
				},
				spam: {
					show: function() {
						return true;
					},
					filter_label: $.noop
				},
				global_search: {
					show: function()
					{
						$dynamic.hide();
						//Aqui se pierde el campaign_dom_id definido arriba, pero tengo el campaign_id
						self.global_search_queryBuilderObj[self.get_normalized_id(campaign_id)].switchToPanel("global_search");
						$static.show();
						var $campaign = self.get_campaign_node(campaign_id);
						if ( !self.m_dash_global_search.is_initialized() ) {
							self.m_dash_global_search.init($campaign);
						}
						else{
							self.m_dash_global_search.update_viewer_classes($campaign);
						}
						var campaign_dom_id = self.get_normalized_id(campaign_id);
						// la clase al contenedor del viewer la tengo que agregar de forma manual ya que la busqueda global no es un dashboard
						var $panel = $('#sup_dash_interaction_viewer_' + campaign_dom_id);
						$panel.removeClass('archiveViewer queuedViewer waitingViewer abandonedViewer bot_interactionsViewer monitoringViewer appointmentViewer globalViewer');
						$panel.addClass('globalViewer');

						// como el global search es algo independiente, y NO funciona como los demas paneles del supervisor
						// tengo que actualizar la current view.
						self.set_current_view_id( campaign_id, 'global_search' );
						return true;
					},
					filter_label: function(_) {
						self.queryBuilderObj[campaign_dom_id].onReadyToRefresh();
					}
				}
			}
		});
	};

	/**
	 * It's called by the initialize method, given a campaign,
	 * will render and setup all necesary markup and events.
	 *
	 * @private
	 * @async
	 * @param  {String} campaign_id - The campaign id.
	 * @param _ -dummy
	 * @param event_queue
	 * @param  {Function} on_campaign_rendered - a callback will be called after the campaign creation.
	 */
	SupervisorDashboard.prototype.__render_campaign = function (campaign_id, _, event_queue, on_campaign_rendered, supervisor_view_object)
	{
		const self = this;

		console.info("SupervisorDashboard.prototype.__render_campaign - Render campaign: '" + campaign_id + "'");

		var can_render_campaign = function() {
			// calculo la cantidad de campanas que van a quedar reenderizas
			let total_rendered_campaigns = self.__get_total_campaigns_rendered();
			let rendered_campaigns = self.get_rendered_campaigns();
			if (!rendered_campaigns[campaign_id]) {
				total_rendered_campaigns++;
			}
			// valido si la campana puede ser monitoreada, es decir si no ha pasado el numero maximo de campanas que el supervisor puede monitorear al mismo tiempo.
			let result = self.validate_number_of_campaigns_to_monitor(total_rendered_campaigns);
			if (!result.success) {
				console.warn("SupervisorDashboard.prototype.__render_campaign - Unable to render " + total_rendered_campaigns + " campaigns. Current campaign: '" + campaign_id + "'");
				return false;
			}
			return true;
		};
		if (!can_render_campaign()) {
			return on_campaign_rendered(true);
		}

		const get_supervisor_view_callback = function(supervisor_view, error) {
			if ( !supervisor_view || error ) {
				switch(error){
					case 'NOT_LOADED':
						const hide_close_button = self.__get_total_campaigns_rendered() === 0;
						self.show_info_message(format("<%=_T('Campaign: {0} is not loaded yet, be patient')%>", campaign_id), hide_close_button);
						break;
					case 'NOT_FOUND':
						// let this error pass through
						break;
					default:
						self.show_error_message(format("<%=_T('Could not get supervisor info for campaign: {0}, error: {1}')%>", campaign_id, error));
						break;
				}
				return on_campaign_rendered(false);
			}

			if (self.m_rendered_campaigns[campaign_id]){
				return on_campaign_rendered(true);
			}

			// calculo la cantidad de campanas que van a quedar reenderizas
			if (!can_render_campaign()) {
				return on_campaign_rendered(true);
			}

			// hide previous messages, show and update waiting animation
			let waiting_msg = format("<%=_T('Building campaign: {0}')%>", campaign_id);
			self.show_waiting(waiting_msg);

			// La agrego al selector de campanas
			let campaign = inConcert.app.getCampaign(campaign_id);
			if (!campaign) {
				on_campaign_rendered(false);
				return;
			}
			
			let $campaign = self.get_campaign_node(campaign_id);
			if ($campaign.length === 0)
			{
				// Create campaign node
				$campaign = self.__$get_campaign_markup(campaign_id, event_queue);

				self.__set_campaign_node_position($campaign);
				$('#sup_container').append($campaign); // append to main container.

				// hash to know wich campaigns should be rendered again or not.
				self.m_rendered_campaigns[campaign_id] = {
					campaign_timestamp: supervisor_view.nowTimestamp,
					ready: false
				};
			}

			self.__add_campaign_selector_node(campaign, true);
			self.show_or_hide_campaign_node(campaign_id);

			// genero la navegacion
			self.__build_navigation(campaign_id);

			// update config
			self.m_config.update_campaign_visibility_config(campaign_id, true);

			// Menu handlers and campaign info
			self.__setup_campaign_menu_handlers($campaign);
			self.__setup_campaign_detail_toggle_handlers($campaign);
			// Summary Column
			self.m_dash_summary.setup($campaign, supervisor_view.realtimeSummary);

			// TeamMembers Column
			self.m_dash_members.setup($campaign, supervisor_view.memberData);

			// Start campaign configuration
			self.m_dash_config.setup(campaign_id);

			// Based on current config (if exist), will setup campaign node and panels visibility.
			self.__configure_node_visibility($campaign, campaign_id);

			// agrega el tab y la row correspondiente a la campana de modo full screen
			self.full_screen_on_added_campaign(campaign_id);
			// SI el dashboard se esta reenderizando (es decir se estan cargando por primera vez las campanas que el supervisor tiene asignadas, ahi realizo las visualizaciones full screen)
			if (!self.is_rendered_dashboard()) {
				// Muestra u oculta el modo full screen de la campana
				let full_screen = false;
				let full_screen_selected_tab_id = campaign_id;
				let campaign_config = self.get_stored_campaign_config(campaign_id);
				if (campaign_config && campaign_config.full_screen) {
					// En modo full screen siempre dejamos posicionado en dashboard.
					full_screen = true;
					full_screen_selected_tab_id = 'dashboard';
				}
				self.full_screen_toggle_visibility({full_screen: full_screen, campaign_id: full_screen_selected_tab_id});
			}

			let $header_buttons = $campaign.find('[header_buttons]');
			$header_buttons.inconcert_tooltip({alignment: 'left-bottom', bind_selector: '[tooltip]'});

			// Indicators
			// Note: Indicators is the heaviest operation of the all business.
			// so, we call on_campaign_rendered on the setup indicators callback
			// Busco si viene los indicadores en los datos de la vista
			let indicators;
			if (supervisor_view.indicatorsData && (!supervisor_view.indicatorsData.result || supervisor_view.indicatorsData.result === "OK")) {
				// Se valida que no existe error en la busqueda de los indicadores. Si es caso, se vuelve a buscar.
				indicators = supervisor_view.indicatorsData;
			}
			
			self.m_dash_indicators.setup_realtime_charts(campaign_id, supervisor_view.realtimeCharts, function()
			{
				let is_realtime = true;
				self.full_screen_setup_charts(campaign_id, supervisor_view.realtimeCharts, is_realtime);
				self.m_dash_indicators.setup(campaign_id, function(success, historical_charts)
				{
					//inConcert.logger.debug(format('Finished rendering campaign: {0}', campaign_id));
					if (success) {
						is_realtime = false;
						self.full_screen_setup_charts(campaign_id, historical_charts, is_realtime);
					}
					if (self.m_rendered_campaigns[campaign_id]) {
						self.m_rendered_campaigns[campaign_id].ready = true;
					}
					on_campaign_rendered(true);
				}, indicators);
			});
		};

		if (supervisor_view_object) { //Uso la data pasada por parametro obtenida con anterioridad y no vuelvo a buscar al server
			return get_supervisor_view_callback(supervisor_view_object.supervisor_view, supervisor_view_object.error);
		}

		let filter = this.__get_campaign_filter(campaign_id);

		// go and get the view for this campaign from the server
		this.m_model.get_supervisor_view(campaign_id, filter, get_supervisor_view_callback);
	};

	SupervisorDashboard.prototype.full_screen_setup_charts = function (campaign_id, charts_info, is_realtime)
	{
		const self = this;
		var setup_charts;
		setup_charts = function() {
			if (!self.m_full_screen) {
				window.setTimeout(function() {
					setup_charts();
				}, 100);
				return;
			}
			self.m_full_screen.get_chart_instance().setup_charts(campaign_id, charts_info, is_realtime);
		};
		setup_charts();
	};

	SupervisorDashboard.prototype.full_screen_refresh_realtime_charts = function (campaign_id, charts_info)
	{
		if (!this.m_full_screen) {
			return;
		}
		this.m_full_screen.get_chart_instance().refresh_realtime_charts(campaign_id, charts_info);
	};

	/**
	 * Will refresh the campaign dashboard info (title and detail)
	 *
	 * @private
	 * @param  {String} campaign_id - The campaign identificator
	 *
	 * @param  {Object}   campaign                     - A campaign object (from application)
	 * @param  {Boolean}  campaign.AlwaysValid         - Campaign is always valid? (speaking of dates)
	 * @param  {String}   campaign.ValidationEndTime   - Campaign validation end date. (related to AlwaysValid)
	 * @param  {Function} campaign.isOnSchedule        - Tells us if the campaign is on schedule right now.
	 * @param  {Array}    campaign.accounts            - A list of accounts for this campaign.
	 */
	SupervisorDashboard.prototype.__refresh_campaign_info = function (campaign_id, campaign, event_queue)
	{
		var self = this;

		campaign = campaign || inConcert.app.getCampaign(campaign_id);
		var campaign_dom_id = this.get_normalized_id(campaign_id);
		var workgroups = this.m_workgroups.get_campaign_workgroups(campaign_id);
		var title = this.get_campaign_title(campaign_id, campaign, workgroups, event_queue);
		var valid_until = campaign.AlwaysValid ? "<%=_T('Always Valid.')%>" : campaign.ValidationEndTime;
		var schedule_class = campaign.isOnSchedule() ? 'gridImageTrue' : 'gridImageFalse';

		// update all the things
		$('#sup_dash_panel_name_' + campaign_dom_id).html(title);
		$('#sup_dash_detail_valid_until_' + campaign_dom_id).text(valid_until);
		$('#sup_dash_detail_schedule_' + campaign_dom_id).removeClass().addClass(schedule_class);
		$('#sup_dash_detail_timezone_' + campaign_dom_id).text(campaign.ValidationTimeZone);
		$('#sup_dash_detail_workgroups_' + campaign_dom_id).text(workgroups);

		// nothing else to do without accounts
		if (!campaign.accounts && campaign.accounts.length === 0){
			return;
		}

		var previous_accounts = [];
		var current_accounts_map = {};

		// store previous accounts ids
		$('#sup_dash_panel_detail_container_' + campaign_dom_id + ' div[id^="sup_dash_detail_account_status_"]')
			.each(function(_){
				if (this.id){
					previous_accounts.push(this.id);
				}
			});

		campaign.accounts.forEach(function(account)
		{
			var account_id = nid(account.id);
			var account_dom_id = 'sup_dash_detail_account_status_'+ campaign_dom_id +'_'+ account_id;
			var status = self.__get_campaign_account_status(account);
			var $account_node = $('#' + account_dom_id);

			// populate hour hash for this account to check for removal later
			current_accounts_map[account_dom_id] = true;

			// check if account is new
			if ( $account_node.length === 0 ){
				// append new node
				$account_node = $(self.__get_campaign_account_detail_node(campaign_id, account));
				$('#sup_dash_panel_detail_container_' + campaign_dom_id).append($account_node);
				// update campaign dashboard top position
				self.__increase_body_panel_top_position(campaign_dom_id, $account_node.outerHeight());
			}
			// already exists? update data
			else {
				var campaign_config = self.get_campaign_selected_config(campaign_id);
				// oculta la info de la cuenta si:
				//  tiene configurada una cuenta en "campaign_selector", y la misma es distinta a la current
				if ( campaign_config.account_id && campaign_config.account_id !== account.id ) {
					$account_node.hide();
				}
				else {
					$account_node
						.removeClass(function (_, class_name) {
							// this is the class we are looking for?
							return (class_name.match (/\bcampaignAccount\S+/g) || []).join(' ');
						})
						.addClass('campaignAccount' + status.id);

					var suffix_id = campaign_dom_id +'_' + account_id;
					var attention_level_list = self.__get_campaign_account_attention_levels(campaign_id, account);

					$('#sup_dash_detail_account_status_attention_level_' + suffix_id).text(attention_level_list);
					$('#sup_dash_detail_account_status_detail_' + suffix_id).text(status.detail);
					$('#sup_dash_detail_account_name_' + suffix_id).text(account.name);
					$('#sup_dash_detail_account_distribution_' + suffix_id).text(_T(account.CCAlgorithm));

					$account_node.show();
				}
			}
		});

		// check for account removal
		previous_accounts.forEach(function(previous_id){
			if ( !current_accounts_map[previous_id] ){
				// save account height and remove the node
				var $account_node = $('#'+previous_id);
				var account_node_height = $account_node.outerHeight();
				$account_node.remove();
				// update campaign dashboard top position
				self.__reduce_body_panel_top_position(campaign_dom_id, account_node_height);
			}
		});
	};

	SupervisorDashboard.prototype.__refresh_alarms_notifications = function (campaign_id, supervisor_view, callback = $.noop)
	{
		// 1- Se valida si se obtuvieron las alarmas de summary
		if (!supervisor_view ) {
			return callback(false);
		}

		//
		if (!supervisor_view.alarms || supervisor_view.alarms.length === 0) {
			return callback(true);
		}

		// 2) - Las notifico, pero se van a encolar a la espera de las alarms (3)
		this.notify_alarms(campaign_id, supervisor_view.alarms, "REQUEST_SUPERVISOR_PANEL");

		//console.debug("SupervisorDashboard.prototype.__retrieve_alarms_notifications - VOY POR LAS ALARMAS DE MEMBERS");

		// 3) - Ahora se obtienen las alarmas de los members
		this.m_dash_members.get_members(campaign_id, callback, supervisor_view.memberData);
	};

	SupervisorDashboard.prototype.__retrieve_alarms_notifications = function (campaign_id, force, callback = $.noop)
	{
		const self = this;

		//console.debug("SupervisorDashboard.prototype.__retrieve_alarms_notifications - START");

		if (!force && !this.has_configured_alarms(campaign_id)) {
			callback(false);
			console.debug("SupervisorDashboard.prototype.__retrieve_alarms_notifications - NO TIENE ALARMAS CONFIGURADA EN LA CAMPANA " + campaign_id);
			return;
		}
		let filter = this.__get_campaign_filter(campaign_id);
		// 1) - Se realiza request para obtener las alarmas de summary.
		//console.debug("SupervisorDashboard.prototype.__retrieve_alarms_notifications - VOY POR LAS ALARMAS DE SUMMARY");
		this.m_model.get_supervisor_view(campaign_id, filter, function(supervisor_view, error)
		{
			if (!supervisor_view || error) {
				callback(false);
				//console.debug("SupervisorDashboard.prototype.__retrieve_alarms_notifications - ERROR OBTENIENDO ALARMAS DE SUMMARY");
				return;
			}
			// 2) - Las notifico, pero se van a encolar a la espera de las alarms 3) (las del request de members)
			self.notify_alarms(campaign_id, supervisor_view.alarms, "REQUEST_SUPERVISOR_PANEL");

			//console.debug("SupervisorDashboard.prototype.__retrieve_alarms_notifications - VOY POR LAS ALARMAS DE MEMBERS");

			// 3) - Ahora se realiza request para obtener las alarmas de los members
			self.m_dash_members.get_members(campaign_id, function(success) {
				callback(success);
			});
		});
	};

	SupervisorDashboard.prototype.get_campaign_id_in_full_screen = function()
	{
		if (!this.m_full_screen) {
			return;
		}
		return this.m_full_screen.get_campaign_id();
	};

	SupervisorDashboard.prototype.show_or_hide_campaign_node = function (campaign_id, $current_node)
	{
		$current_node = $current_node || this.get_campaign_node(campaign_id);
		if ($current_node.length === 0){
			return;
		}
		if (inConcert.permissions.to_application('supervisor', campaign_id)) {
			if (!this.m_full_screen || !this.m_full_screen.is_full_screen() || this.get_campaign_id_in_full_screen() === campaign_id) {
				$current_node.show();
				return;
			}
		}
		$current_node.hide();
	};

	SupervisorDashboard.prototype.on_toggle_full_screen = function(campaign_id)
	{
		this.m_dash_members.on_toggle_full_screen(campaign_id);
		this.show_or_hide_campaign_node(campaign_id);

		this.set_chip_name(campaign_id, this.get_current_view_id(campaign_id));

		if (!(this.m_full_screen && this.m_full_screen.is_full_screen())) {

			// Cuando se cambie a simple screen refresco los indicadores
			this.m_fire_charts_refresh = { enable: true};

			// Muestro/Oculto botones del toolbar
			this.update_simple_screen_toolbar_buttons(campaign_id);

			//Se actualiza el layout del dashboard principal.
			this.__clean_main_dashboard_layout_classes(campaign_id);
			
		}
	};

	SupervisorDashboard.prototype.__refresh_campaign_from_view = function (campaign_id, campaign_event, callback, supervisor_view)
	{
		const self = this;
		callback = callback || $.noop;

		// Si nos encontramos en modo "full screen" panel dashboard
		const full_screen_selected_tab_dashboard = self.is_full_screen_mode() && self.full_screen_selected_tab_id() === "dashboard";

		const $current_node = self.get_campaign_node(campaign_id);
		if ($current_node.length === 0) {
			return this.__refresh_alarms_notifications(campaign_id, supervisor_view, function(success) {
				callback(success);
			});
		}

		// Esta funcion valida el tiempo transcurrido desde la ultima actualizacion que se realizo en panel actual de conversaciones
		//  si paso los N segundos (REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS), puede realizar el request y la actualizacion correspondiente.
		// Esta validacion se realiza aca, ya que todos los paneles de todas las campanas estan enganchados en el mismo loop de actualizacion,
		//   y el cambio puede ser muy muy grande y de gran impacto en toda la app
		//
		const can_refresh = function( campaign_id, panel_id )
		{
			let current_time = new Date().valueOf();
			// 1 - Si es la primera vez que se realiza el refresh para la campana, crea toda la informacion y retorna true para que se realice la actualizacion.
			if ( !self.latest_refresh[campaign_id] ) {
				self.latest_refresh[campaign_id] = {};
				self.latest_refresh[campaign_id][panel_id] = {};
				self.latest_refresh[campaign_id][panel_id].time = current_time;
				return true;
			}
			// 2 - Si es la primera vez que se realiza el refresh para el panel (monitoring, queud, waiting) de la campana, crea toda la informacion y retorna true para que se realice la actualizacion.
			if ( !self.latest_refresh[campaign_id][panel_id] ) {
				self.latest_refresh[campaign_id][panel_id] = {};
				self.latest_refresh[campaign_id][panel_id].time = current_time;
				return true;
			}
			// 3 - Si el tiempo transcurrido desde la ultima actualizacion para la campana/panel supera los N segundos (REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS)
			//      actualiza el timestamp y retorna true para que se realice la actualizacion.
			let lastes_refresh_time = self.latest_refresh[campaign_id][panel_id].time;
			if ( (lastes_refresh_time + REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS) < current_time ) {
				self.latest_refresh[campaign_id][panel_id].time = current_time;
				return true;
			}
			// 4 - No puede realizar el refresh
			return false;
		};

		/*
			Funcion que actualiza los Paneles "Summary" y "dashboard/full screen"
			Si estamos en el dashboard se ejecuta directamente la funcion callback y no la funcion refresh_alarms, ya que este request que va contra las alarmas del propio request.
		*/
		const update_summary = function () {

			// primero tengo que actualizar las alarmas en el popup de notificaciones y alarmas
			self.notify_alarms(campaign_id, supervisor_view.alarms, "REQUEST_SUPERVISOR_PANEL");

			// update columns data.
			self.m_dash_indicators.refresh_realtime_charts(campaign_id, supervisor_view.realtimeCharts);
			self.full_screen_refresh_realtime_charts(campaign_id, supervisor_view.realtimeCharts);
			self.m_dash_summary.refresh($current_node, supervisor_view.realtimeSummary);
			
			//Refresh members pasando la vista de miembros obtenido
			self.m_dash_members.refresh($current_node, supervisor_view.memberData);

			// update campaign now timestamp
			self.m_rendered_campaigns[ campaign_id ].campaign_timestamp = supervisor_view.nowTimestamp;
			// update detail info
			self.__refresh_campaign_info( campaign_id, campaign_event.campaign, campaign_event);
			self.update_panel_positions();
			callback(true);
		};

		
		self.show_or_hide_campaign_node(campaign_id, $current_node);


		/*
		  - Si nos encontramos en modo "full screen->dashboard" (en la grilla), debemos actualizar el summary.
		  - Se actualiza el summary y se continua al siguiente hasta que concluya.
		*/
		if (full_screen_selected_tab_dashboard) {
			return update_summary();
		}

		//Caso que no es "full screen->dashboard", se refresca la vista
		let current_view = self.get_current_view_id(campaign_id);

		//Se actualiza los titles de los tabs
		if (this.m_full_screen) {

			if (this.m_full_screen.is_full_screen_mode()) {
				this.m_full_screen.make_tab_caption(campaign_id);
			}
			else {
				const workgroups = this.m_workgroups.get_campaign_workgroups(campaign_id);
				const title = this.get_campaign_title(campaign_id, campaign_event.campaign || inConcert.app.getCampaign(campaign_id), workgroups, campaign_event);
				$('#sup_dash_panel_name_' + this.get_normalized_id(campaign_id)).html(title);
			}
		}

		switch (current_view) {
			case 'monitoring':
				if (can_refresh(campaign_id, current_view)) {
					self.m_dash_monitor.refresh($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'appointment':
				if (can_refresh(campaign_id, current_view)) {
					self.m_dash_appointment.refresh($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'queued':
				if (can_refresh(campaign_id, current_view)) {
					self.m_dash_queued.refresh($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'waiting':
				if (can_refresh(campaign_id, current_view)) {
					self.m_dash_waiting.refresh($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'abandoned':
				if (can_refresh(campaign_id, current_view)) {
					self.m_dash_abandoned.refresh($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'bot_interactions':
				if (can_refresh(campaign_id, current_view)) {
					self.m_dash_bot_interactions.refresh($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'archive':
				// Cuando se actualice la configuracion, si es archive se refresca la vista
				if (campaign_event.refresh_by_change_config) {
					self.m_dash_archive.clear_list( self.get_normalized_id(campaign_id));
					self.m_dash_archive.search_archive($current_node);
				}
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'global_search':
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'dashboard_config':
				self.__refresh_alarms_notifications(campaign_id, supervisor_view, callback);
				break;

			case 'dashboard':
				// Si estamos en el panel "dashboard", NO "full screen"
				if (self.m_dash_container_scroll_is_moving) {
					callback(true);
				}
				else{
					update_summary();
				}
				break;

			default:
				throw ('SupervisorDashboard.prototype.__refresh_campaign - Unrecognized panel view: ' + current_view);
		}	
	};

	/**
	 * Refresh a campaign node.
	 *
	 * @private
	 * @async
	 * @param  {String} campaign_id - The campaign id.
	 * @param  {Object} campaign    - The campaign object from application.
	 * @param event_queue
	 * @param  {Function} callback  - a callback will be called after the campaign is refreshed.
	 */
	SupervisorDashboard.prototype.__refresh_campaign = function (campaign_id, campaign, event_queue, callback = $.noop)
	{
		const self = this;

		var refresh_callback = function() {
			self.__retrieve_alarms_notifications(campaign_id, true, callback);
			return;
		};

		let filter = this.__get_campaign_filter(campaign_id);

		// Esta funcion valida el tiempo transcurrido desde la ultima actualizacion que se realizo en panel actual de conversaciones
		//  si paso los N segundos (REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS), puede realizar el request y la actualizacion correspondiente.
		// Esta validacion se realiza aca, ya que todos los paneles de todas las campanas estan enganchados en el mismo loop de actualizacion,
		//   y el cambio puede ser muy muy grande y de gran impacto en toda la app
		//
		var can_refresh = function( campaign_id, panel_id )
		{
			let current_time = new Date().valueOf();
			// 1 - Si es la primera vez que se realiza el refresh para la campana, crea toda la informacion y retorna true para que se realice la actualizacion.
			if ( !self.latest_refresh[campaign_id] ) {
				self.latest_refresh[campaign_id] = {};
				self.latest_refresh[campaign_id][panel_id] = {};
				self.latest_refresh[campaign_id][panel_id].time = current_time;
				return true;
			}
			// 2 - Si es la primera vez que se realiza el refresh para el panel (monitoring, queud, waiting) de la campana, crea toda la informacion y retorna true para que se realice la actualizacion.
			if ( !self.latest_refresh[campaign_id][panel_id] ) {
				self.latest_refresh[campaign_id][panel_id] = {};
				self.latest_refresh[campaign_id][panel_id].time = current_time;
				return true;
			}
			// 3 - Si el tiempo transcurrido desde la ultima actualizacion para la campana/panel supera los N segundos (REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS)
			//      actualiza el timestamp y retorna true para que se realice la actualizacion.
			let lastes_refresh_time = self.latest_refresh[campaign_id][panel_id].time;
			if ( (lastes_refresh_time + REFRESH_CONVERSATIONS_PANELS_INTERVAL_MS) < current_time ) {
				self.latest_refresh[campaign_id][panel_id].time = current_time;
				return true;
			}
			// 4 - No puede realizar el refresh
			return false;
		};

		const $current_node = this.get_campaign_node(campaign_id);
		if ( $current_node.length === 0 ){
			refresh_callback();
			return;
		}

		this.show_or_hide_campaign_node(campaign_id);

		/*
			Funcion que actualiza los Paneles "Summary" y "dashboard/full screen"

			Si estamos en el dashboard se ejecuta directamente la funcion callback y no la funcion refresh_callback, ya que este request que va contra las alarmas del propio request.
		*/
		var update_summary = function () {
			// update columns data.
			self.m_model.get_supervisor_view(campaign_id, filter, function(supervisor_view, error)
			{
				if ( !supervisor_view || error ) {
					callback(false);
					return;
				}
				// primero tengo que actualizar las alarmas en el popup de notificaciones y alarmas
				self.notify_alarms(campaign_id, supervisor_view.alarms, "REQUEST_SUPERVISOR_PANEL");

				// update columns data.
				self.m_dash_indicators.refresh_realtime_charts(campaign_id, supervisor_view.realtimeCharts);
				self.full_screen_refresh_realtime_charts(campaign_id, supervisor_view.realtimeCharts);
				self.m_dash_summary.refresh($current_node, supervisor_view.realtimeSummary);
				self.m_dash_members.refresh($current_node);

				// update campaign now timestamp
				self.m_rendered_campaigns[ campaign_id ].campaign_timestamp = supervisor_view.nowTimestamp;
				// update detail info
				self.__refresh_campaign_info( campaign_id, campaign, event_queue );
				self.update_panel_positions();

				callback(true);
			});
		};

		const full_screen = this.is_full_screen_mode();
		if (full_screen) {
			// Si nos encontramos en modo "full screen"
			const full_screen_selected_tab_id = this.full_screen_selected_tab_id();
			if (full_screen_selected_tab_id === "dashboard") {
				// Si nos encontramos en modo "full screen->dashboard" (en la grilla), debemos actualizar el summary.
				update_summary();
				return;
			}
		}

		let current_view = this.get_current_view_id(campaign_id);
		switch ( current_view )
		{
			case 'monitoring':
				if ( !can_refresh( campaign_id, current_view ) ) {
					refresh_callback();
					return;
				}
				this.m_dash_monitor.refresh( $current_node );
				refresh_callback();
				break;

			case 'appointment':
				if ( !can_refresh( campaign_id, current_view ) ) {
					refresh_callback();
					return;
				}
				this.m_dash_appointment.refresh( $current_node );
				refresh_callback();
				break;

			case 'queued':
				if ( !can_refresh( campaign_id, current_view ) ) {
					refresh_callback();
					return;
				}
				this.m_dash_queued.refresh( $current_node );
				refresh_callback();
				break;

			case 'waiting':
				if ( !can_refresh( campaign_id, current_view ) ) {
					refresh_callback();
					return;
				}
				this.m_dash_waiting.refresh( $current_node );
				refresh_callback();
				break;

			case 'abandoned':
				if ( !can_refresh( campaign_id, current_view ) ) {
					refresh_callback();
					return;
				}
				this.m_dash_abandoned.refresh( $current_node );
				refresh_callback();
				break;

			case 'bot_interactions':
				if ( !can_refresh( campaign_id, current_view ) ) {
					refresh_callback();
					return;
				}
				this.m_dash_bot_interactions.refresh( $current_node );
				refresh_callback();
				break;

			case 'archive':
				refresh_callback();
				break;

			case 'global_search':
				refresh_callback();
				break;

			case 'dashboard_config':
				refresh_callback();
				break;

			case 'dashboard':
				// Si estamos en el panel "dashboard", NO "full screen"
				update_summary();
				break;

			default:
				throw('SupervisorDashboard.prototype.__refresh_campaign - Unrecognized panel view: ' + current_view);
		}
	};

	/**
	 * Removes a campaign node. Update other campaigns position
	 *
	 * @private
	 * @param  {String} campaign_id - The campaign id.
	 * @param  {String} campaign_obj - The campaign object (from application).
	 * @param _ - dummy
	 * @param  {Function} callback - a callback will be called after the campaign is refreshed.
	 */
	SupervisorDashboard.prototype.__remove_campaign = function (campaign_id, campaign_obj, _, callback)
	{
		const self = this;

		console.info("SupervisorDashboard.prototype.__remove_campaign - Unassign campaign: '" + campaign_id + "'");
		// remove node
		this.get_campaign_node(campaign_id).remove();
		// position next nodes correctly
		this.update_panel_positions();

		// remueve el tab correspondiente a la campana de modo full screen
		this.full_screen_on_remove_campaign(campaign_id, function(success) {
			// remove from rendered campaigns
			delete self.m_rendered_campaigns[campaign_id];
			self.m_config.update_campaign_visibility_config(campaign_id, false);
			// calling back
			callback(success);
		});
	};

	/**
	 * Removes all campaign nodes.
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.__remove_all_campaigns = function () {
		// remove all campaigns nodes.
		$('#supervisor_main .supervisorPanelContainer').remove();

		// reinitialize rendered campaigns to an empty object
		this.m_rendered_campaigns = {};
	};

	/**
	 * Will process every available campaign,
	 * adding the campaign if is not rendered yet or refreshing it.
	 *
	 * @private
	 * @param  {Array} campaigns - The campaigns to be processed.
	 */
	SupervisorDashboard.prototype.__process_campaigns = function(campaigns)
	{
		const self = this;

		// nothing to do without campaigns, make some housekeeping before exit.
		if ( !campaigns || campaigns.length === 0 ) {
			self.__remove_all_campaigns();
			const hide_close = true;
			self.show_info_message("<%=_T('You should add at least one campaign to view the supervisor dashboard.')%>", hide_close);
			return;
		}

		// cantidad de campanas que se encuentran reenderizadas.
		let total_rendered_campaigns = this.__get_total_campaigns_rendered();
		let counter_new_campaigns = 0;
		// go an process campaigns
		campaigns.forEach(campaign =>
		{
			let campaign_config = self.get_stored_campaign_config(campaign.Id);

			// there is nothing to do if the campaign is configured to be hidden.
			if (campaign_config && campaign_config.visible !== undefined && !campaign_config.visible){
				return;
			}

			let event_type;
			if (self.__is_ready_rendered_campaign(campaign.Id)) {
				const full_screen_campaign = this.m_full_screen && this.m_full_screen.get_campaign_id();
				if (!full_screen_campaign || full_screen_campaign === "dashboard" || full_screen_campaign === campaign.Id || self.has_configured_alarms(campaign.Id)) {
					event_type = 'REFRESH';
				}
				else{
					// Si existe una campanna maximizada, no se refrescan las demas, si no tiene alarmas configuradas
					return;
				}
			}
			else {
				// valido si la campana puede ser monitoreada, es decir si no ha pasado el numero maximo de campanas que el supervisor puede monitorear al mismo tiempo.
				counter_new_campaigns++;
				let total_campaigns = total_rendered_campaigns + counter_new_campaigns;
				let result = self.validate_number_of_campaigns_to_monitor(total_campaigns);
				if (!result.success) {
					return;
				}
				event_type = 'ADD';
			}

			let selected_config;
			if (campaign_config && campaign_config.dashboard) {
				selected_config = campaign_config.dashboard.config;
			}
			// first campaign time? render it, otherwise refresh it!
			self.__add_event_to_queue({
				campaign_id: campaign.Id,
				campaign   : campaign,
				config     : selected_config,
				event_type : event_type
			});
		});
	};

	/*
	Funcion encargada de reenderizar las campanas que el supervisor tiene asignadas, solo se invoca la primera vez que ingresa a la app Supervisor
	*/
	SupervisorDashboard.prototype.__start_render_campaigns = function (campaigns, callback = $.noop)
	{
		const self = this;
		let total_campaigns = campaigns ? campaigns.length : 0;

		if (total_campaigns === 0) {
			this.__remove_all_campaigns();
			const hide_close = true;
			this.show_info_message("<%=_T('You should add at least one campaign to view the supervisor dashboard.')%>", hide_close);
			callback(false);
			return;
		}

		let total_rendered_campaigns = 0;
		let current_campaign_number = 0;
		let campaigns_filters = {};
		let selected_config_by_campaign = {};
		let is_process_campaign = false;
		for (let i = 0; i < total_campaigns; i++) {
			let campaign = campaigns[i];
			let campaign_config = self.get_stored_campaign_config(campaign.Id);
			// there is nothing to do if the campaign is configured to be hidden.
			if (campaign_config && campaign_config.visible !== undefined && !campaign_config.visible){
				continue;
			}
			if (campaign_config && campaign_config.dashboard) {
				selected_config_by_campaign[campaign.Id] = campaign_config.dashboard.config;
			}

			is_process_campaign = true;
			self.m_dash_members.initializePageByCampaign(campaign.Id);
			campaigns_filters[campaign.Id] = {
				filter : self.__get_campaign_filter(campaign.Id),
				filterMembers : self.m_dash_members.get_members_filters(campaign.Id),
				filterIndicators : self.m_dash_indicators.get_historical_data_filter(campaign.Id)
			};
		}

		if (is_process_campaign) {
			let data = {
				campaigns : campaigns_filters
			};

			self.m_model.getSupViewAndMembersByCampaigns(data, function (response, error) {
				let supervisor_campaign_views = response && response.supervisor_views;
				
				if (error || !supervisor_campaign_views || supervisor_campaign_views.length === 0) {
					console.info("SupervisorDashboard.prototype.__start_render_campaigns " + format("<%=_T('Could not get supervisor info for campaigns, error: {0}')%>", error || ""));
					self.show_error_message("<%=_T('Could not get supervisor info for campaigns.')%>");
					return callback(false);
				}

				for (const campaign_id in supervisor_campaign_views) {

					if (Object.hasOwnProperty.call(supervisor_campaign_views, campaign_id)) {
						const current_supervisor_campaign_view = supervisor_campaign_views[campaign_id];
						
						let view = {
							supervisor_view: current_supervisor_campaign_view
						};
						if (current_supervisor_campaign_view.result && current_supervisor_campaign_view.result !== "OK") {
							view.error = current_supervisor_campaign_view.result;
						}

						current_campaign_number++;
						if (current_campaign_number <= MAXIMUM_NUMBER_OF_MONITORABLE_CAMPAIGNS) {
							/* jshint loopfunc: true */
							self.__render_campaign(campaign_id, undefined, selected_config_by_campaign[campaign_id] , function (success) {
								if (success) {
									total_rendered_campaigns++;
									if (total_rendered_campaigns === current_campaign_number || total_rendered_campaigns === MAXIMUM_NUMBER_OF_MONITORABLE_CAMPAIGNS) {
										callback(true);
									}
								}
							}, view);
							continue;
						}
						break;
					}
				}
			});
		}
		else{
			callback(false);
		}
	};

	/**
	 * Get the dashboard message node
	 *
	 * @private
	 * @return {Object} - The jQuery message node
	 */
	SupervisorDashboard.prototype.__get_message_node = function(){
		return $('#sup_content_message');
	};

	/**
	 * Get the dashboard message container node
	 *
	 * @private
	 * @return {Object} - The jQuery container message node
	 */
	SupervisorDashboard.prototype.__get_message_container_node = function(){
		return $('#sup_content_message_container');
	};

	/**
	 * Get the campaign detail information container.
	 *
	 * @private
	 * @param  {String} campaign_dom_id - The campaign DOM id (i.e. support_inconcert)
	 * @return {Object}                 - The jQuery campaign info container
	 */
	SupervisorDashboard.prototype.__get_campaign_detail_container = function(campaign_dom_id){
		return $('#sup_dash_panel_detail_container_' + campaign_dom_id);
	};

	/**
	 * Hides current dashboard message
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.__hide_message = function(){
		this.__get_message_node().text('');
		this.__get_message_container_node()
			.hide()
			.removeClass('messageError messageInfo messageOk')
			.parent().removeClass('showingMessage');
	};

	/**
	 * Setup dashboard message close button handlers.
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.__setup_close_message_handler = function(){
		var self = this;
		$('#sup_message_close').text("<%=_T('close')%>").click(function() {
			self.__hide_message();
		});
	};

	/**
	 * Shows an error message inside the dashboard
	 *
	 * @private
	 * @param  {String} message - Message to be shown.
	 */
	SupervisorDashboard.prototype.show_error_message = function(message){
		this.__get_message_node().text(message);
		this.__get_message_container_node()
			.removeClass('messageInfo messageOk')
			.addClass('messageError')
			.show()
			.parent().addClass('showingMessage');
	};

	/**
	 * Shows an info message inside the dashboard
	 *
	 * @private
	 * @param  {String} message - Message to be shown.
	* @param {Boolean} hide_close_button
	 */
	SupervisorDashboard.prototype.show_info_message = function(message, hide_close_button, auto_close)
	{
		const self = this;

		this.__get_message_node().text(message);
		this.__get_message_container_node()
			.removeClass('messageError messageOk')
			.addClass('messageInfo')
			.show()
			.parent().addClass('showingMessage');

		// close button visibility, shown by default.
		$('#sup_message_close').show();
		if (hide_close_button){
			$('#sup_message_close').hide();
		}
		if (auto_close) {
			window.setTimeout(function() {
				self.__hide_message();
			}, 5000);
		}
	};

	/**
	 * Will show a default error message inside the dashboard.
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.__show_default_error_message = function(){
		this.show_error_message("<%=_T('We are having trouble contacting our servers, supervisor information may be outdated...')%>");
	};

	/**
	 * Will update the main waiting message
	 *
	 * @public
	 * @param  {String} message - Message to be updated
	 */
	SupervisorDashboard.prototype.update_waiting_message = function(message) {
		var $waiting = $('#sup_dash_main_waiting');
		$waiting.find('[waiting_message]').text(message || '');
	};

	/**
	 * Will clear all timeouts that are turned on.
	 *
	 * @public
	 */
	SupervisorDashboard.prototype.stop_refresh = function () {
		window.clearTimeout(this.m_on_refresh_timer);
		window.clearTimeout(this.m_on_refresh_chart_timer);
		window.clearTimeout(this.m_event_process_timer);
		this.m_on_refresh_timer = null;
		this.m_on_refresh_chart_timer = null;
		this.m_event_process_timer = null;
		delete this.m_fire_charts_refresh;
	};

	/**
	 * Return true if refresh timers are alive
	 *
	 * @public
	 * @return {Boolean}
	 */
	SupervisorDashboard.prototype.is_refresh_alive = function () {
		return this.m_on_refresh_timer && this.m_on_refresh_chart_timer && this.m_event_process_timer;
	};

	/**
	 * Will start the refresh dashboard timer
	 *
	 * @public
	 */
	SupervisorDashboard.prototype.start_refresh_timer = function ()
	{
		var self = this;
		if (this.m_supervisor_leave){
			return;
		}
		// clear the previous timeout before create a new one.
		window.clearTimeout(this.m_on_refresh_timer);
		this.m_on_refresh_timer = window.setTimeout(function () {
			self.refresh();
		}, REFRESH_INTERVAL_MS);
	};

	/**
	 * Will start the timer for the charts refreshing, it's done
	 * in a different function because the timeout is different.
	 *
	 * @private
	 */
	SupervisorDashboard.prototype.__start_charts_refresh_timer = function ()
	{
		var self = this;
		if (this.m_on_refresh_chart_timer) {
			return;
		}
		window.clearTimeout(this.m_on_refresh_chart_timer);
		this.m_on_refresh_chart_timer = window.setTimeout(function () {
			self.m_on_refresh_chart_timer = null;
			self.__start_charts_refresh_timer();

			// Refresh indicators for each campaign.
			if (Object.keys(self.m_rendered_campaigns).length > 0) {
				self.m_fire_charts_refresh = {
					enable : true
				};
			}
		}, HISTORICAL_CHART_REFRESH_INTERVAL_MS);
	};

	/**
	 * Refresh current campaigns
	 *
	 * @public
	 * @async
	 * @param {Function} onRefreshFinished - Callback
	 */
	SupervisorDashboard.prototype.refresh = function(on_refresh_finished)
	{
		var self = this;

		// callback (if any)
		on_refresh_finished = on_refresh_finished || $.noop;

		// get campaigns and order by Id.
		var campaigns = _.sortBy(inConcert.app.getCampaigns(), function(campaign){ return campaign.Id; });
		// process every campaign
		self.__process_campaigns(campaigns);

		// setup refresh timer.
		self.start_refresh_timer();
		// callback
		on_refresh_finished();

	};

	SupervisorDashboard.prototype.refresh_bots_dashboard = function(){
		let self = this;
		let campaigns = _.sortBy(inConcert.app.getCampaigns(), function(campaign){ return campaign.Id; });
		campaigns.forEach((campaign)=>{
			//console.log(campaign, "ITERATION");
			let nid = self.get_normalized_id(campaign.Id);
			let $dashboardContainer = $(`#sup_dash_bots_info_container_${nid}`);
			const appForm = $dashboardContainer.find('icc-processflow-dashboard')[0];
			if(appForm && appForm.reloadInfo){
				appForm.reloadInfo();
			}
			
		});
	};

	SupervisorDashboard.prototype.__is_campaign_view_bots = function (campaign_id, $sup_dashboard_container) {
		$sup_dashboard_container = $sup_dashboard_container || this.get_campaign_bots_node(campaign_id);
		return $sup_dashboard_container && $sup_dashboard_container.css('display') && $sup_dashboard_container.css('display') !== "none";
	};
	
	SupervisorDashboard.prototype.leave_bot_view = function (campaign_id) {
		
		if (this.__is_campaign_view_bots(campaign_id)) {
			this.__toggle_bot_view(campaign_id, false);
		}
	};

	SupervisorDashboard.prototype.__toggle_bot_view = function(campaignId, showBotView, $dashboardContainer)
	{
		const self = this;
		let campaignSelectedId = campaignId || self.get_campaign_id_in_full_screen();
		let nid = self.get_normalized_id(campaignSelectedId);
		let $viewer_operations = $(`#sup_dash_operations_viewer_${nid}`);
		$dashboardContainer = $dashboardContainer || $(`#sup_dash_bots_info_container_${nid}`);
		let $supContainer = $("#sup_container");
		let $back = this.$back_button(campaignSelectedId);

	   if (showBotView) {
		   $back.click();
		   $viewer_operations.hide();
		   $('#sup_dash_campaign_' + nid).removeClass("showBots").addClass("showBots");
		   $dashboardContainer.show();
		   self.__clean_main_dashboard_layout_classes(campaignSelectedId);

		   if (self.is_full_screen_mode()) {
				self.m_full_screen.update_toolbar_buttons();
		   }
		   else{
			self.toggle_btns_toolbar_bots(campaignSelectedId, true);
		   }
		   self.set_chip_name(campaignSelectedId, self.get_current_view_id(campaignSelectedId));

	   } else {
		   $viewer_operations.show();
		   $('#sup_dash_campaign_' + nid).removeClass("showBots");
		   $dashboardContainer.hide();
		   self.update_main_dashbord_layout(campaignSelectedId);
		   if (self.is_full_screen_mode()) {
				$supContainer.removeClass("supBotsDashContainer");
				self.m_full_screen.update_toolbar_buttons();
			}
			else{
			 self.toggle_btns_toolbar_bots(campaignSelectedId, false);
			}
	   }

		if (self.is_full_screen_mode()) {
			$dashboardContainer.removeClass("supBotsDashFullScreen").addClass("supBotsDashFullScreen");
			$dashboardContainer.removeClass("supBotsDashSimpleScreen");
		} else {
			$dashboardContainer.removeClass("supBotsDashFullScreen");
			$dashboardContainer.removeClass("supBotsDashSimpleScreen").addClass("supBotsDashSimpleScreen");
		}

	};
	/**
	 * Refresh current bot info for campaigns
	 *
	 * @public
	 * @async
	 * @param {Function} onRefreshFinished - Callback
	 */
	SupervisorDashboard.prototype.load_bots_dashboard = function(campaignId)
	{
	   var self = this;
	   let campaignSelectedId = campaignId || self.get_campaign_id_in_full_screen();
	   let nid = self.get_normalized_id(campaignSelectedId);
	   let $dashboardContainer = $(`#sup_dash_bots_info_container_${nid}`);
	   let viewDashboard = function(viewBotDashboard){
			self.__toggle_bot_view(campaignSelectedId, viewBotDashboard, $dashboardContainer);
	   };
	   viewDashboard(true);

	   let campaignSelected = inConcert.app.getCampaign(campaignSelectedId);
   
	   const appForm = $dashboardContainer.find('#icc-processflow-dashboard')[0];
	   appForm.init({
		   campaignSelected: campaignSelected,
		   m_model:self.m_model,
		   viewDashboard,
		   isFullScreen: self.is_full_screen_mode(),
		   m_full_screen:self.m_full_screen,
		   initDateAndTimeInputs:function(inputs){				
			   let timeOptions = {
				   datepicker:false,
				   format:'H:i',
				   allowTimes:(()=>{
					   let options = [];
					   for(let h = 0; h < 24; h++){
						   for(let m = 0; m < 59; m=m+15){
							   options.push(`${h<10?'0':''}${h}:${m<10?'0':''}${m}`);
						   }
					   }
					   return options;
				   })()
			   };
			   let localLanguage = inConcert.app.getLoggedUserLanguage(); // en|es|fr|pt
			   
			   let dateOptions = {
					dayOfWeekStart: localLanguage === "en" ? 0:1,
				    timepicker:false,
				    format:"d/m/Y"
			   };
			   $.datetimepicker.setLocale(localLanguage);
			   $(inputs.dateFrom).datetimepicker(dateOptions);
			   $(inputs.dateTo).datetimepicker(dateOptions);
			   $(inputs.timeFrom).datetimepicker(timeOptions);
			   $(inputs.timeTo).datetimepicker(timeOptions);
		   },
		   onSelectFilterOption: function() {
			self.campaignBotsSelectFilterOption(campaignSelectedId);
		   }
	   });
	};

	SupervisorDashboard.prototype.full_screen_on_added_campaign = function(campaign_id)
	{
		if (!campaign_id || !this.m_full_screen) {
			return;
		}
		this.m_full_screen.on_added_campaign(campaign_id);
	};

	SupervisorDashboard.prototype.full_screen_on_remove_campaign = function(campaign_id, callback = $.noop)
	{
		if (!campaign_id || !this.m_full_screen) {
			callback(false);
			return;
		}
		this.m_full_screen.on_removed_campaign(campaign_id, callback);
	};

	SupervisorDashboard.prototype.__initialize = function()
	{
		this.initialize_full_screen();
		this.__initialize_members_selector_popup();
		this.update_alarms();
	};

	SupervisorDashboard.prototype.on_render = function(success) {
		if (!success) {
			this.m_dashboard_render_state = "FAIL";
			return;
		}
		// set rendered flag
		this.m_dashboard_render_state = "RENDERED";
		this.__initialize();
		// setup refresh timer.
		this.start_refresh_timer();
	};

	/**
	 * Initialize method, will setup markup for all user campaigns.
	 *
	 * @public
	 * @async
	 * @param {Function} on_render_finished - Callback
	 */
	SupervisorDashboard.prototype.render = function (on_render = $.noop)
	{
		const self = this;

		this.m_supervisor_leave = false;
		this.stop_alarms_counter_refresh();

		// check if timers are still valid
		if ( !this.is_refresh_alive() ){
			this.__start_charts_refresh_timer();
			this.__start_process_queue_timer();
		}

		var on_render_finished = function(success) {
			self.on_render(success);
			on_render(success);
		};

		// if the dashboard is already rendered,
		// just start the refresh timer and return.
		if (this.is_rendered_dashboard()) {
			this.start_refresh_timer();
			this.__process_pendings_operations();
			on_render_finished(true);
			return;
		}

		this.__initialize(); // inicializa controles

		this.m_dashboard_render_state = "RENDERING";
		// remove previous messages.
		this.__hide_message();
		this.__setup_close_message_handler();
		this.__process_pendings_operations();

		var error_fn = function(error_message) {
			inConcert.logger.error(error_message);
			self.__show_default_error_message();
			// callback
			on_render_finished(false);
		};

		// Get config for this user if exists.
		this.get_config(function(config, error)
		{
			if (error) {
				let error_message = format('[Supervisor] (render) - Error getting configuration from server: {0}', error && error.message ? error.message : error);
				error_fn(error_message);
				return;
			}
			// SUCCESS get config
			if ( config && !_.isEmpty(config) ) {
				self.set_current_config(config);
				self.m_user_without_config = false;
			}
			
			// get campaigns and order by Id.
			let campaigns = _.sortBy(inConcert.app.getCampaigns(), function(campaign){ return campaign.Id; });
			// start render campaigns
			self.__start_render_campaigns(campaigns, function() {

				// Process campaigns - Render/Refresh
				self.__build_campaign_selector(campaigns);

				// configuration is saved on first render
				if (self.m_user_without_config){
					self.save_config();
				}
				// callback
				on_render_finished(true);
			});
		});
	};

	SupervisorDashboard.prototype.hideMenuActions = function() {
		this.m_dash_monitor.hideMenuActions();
		this.m_dash_appointment.hideMenuActions();
		this.m_dash_archive.hideMenuActions();
		this.m_dash_queued.hideMenuActions();
		this.m_dash_waiting.hideMenuActions();
		this.m_dash_abandoned.hideMenuActions();
		this.m_dash_global_search.hideMenuActions();
		this.m_dash_bot_interactions.hideMenuActions();
	};

	/**
	 * Indicates if a campaign is from allegro
	 *
	 * @public
	 * @param campaign_id
	 */
	SupervisorDashboard.prototype.isAllegro = function(campaign_id) {
		let campaign = inConcert.app.getCampaign(campaign_id);
		if (!campaign) {
			return false;
		}
		return campaign.isAllegro();
	};

	SupervisorDashboard.prototype.get_html_templates = function() {
		return this.m_templates;
	};

	SupervisorDashboard.prototype.sanitize_value = function(value = "")
	{
		if (typeof value !== "string") {
			return value;
		}
		// 1 - Elimino todo el markup HTML y me quedo con texto
		value = inConcert.Utils.htmlBodyToPlainText(value);
		// 2 - Convierto los Enters en </BR>
		value = inConcert.Utils.nl2br(value);
		return value;
	};

	SupervisorDashboard.prototype.onReadyToRefresh = function(campaign_id)
	{
		switch(this.m_current_view_by_campaign[campaign_id]) {
			case "monitoring":
				this.m_dash_monitor.execute_new_search(campaign_id);
				break;
			case "appointment":
				this.m_dash_appointment.execute_new_search(campaign_id);
				break;
			case "archive":
				this.m_dash_archive.refresh(campaign_id);
				break;
			case "queued":
				this.m_dash_queued.execute_new_search(campaign_id);
				break;
			case "waiting":
				this.m_dash_waiting.execute_new_search(campaign_id);
				break;
			case "abandoned":
				this.m_dash_abandoned.execute_new_search(campaign_id);
				break;
			case "bot_interactions":
				this.m_dash_bot_interactions.execute_new_search(campaign_id);
				break;
			case "global_search":
				this.navigation[campaign_id].global_search.search(false);
				break;
		}
	};

	SupervisorDashboard.prototype.update_simple_screen_toolbar_buttons = function(campaign_id)
	{
		let campaigns = campaign_id && {[campaign_id] : true} || this.get_rendered_campaigns();
		const self = this;
		$.each(campaigns, function(campaign_id) {

			const normalized_camp = self.get_normalized_id(campaign_id);
			const btn_menu_config_view = $(`#sup_dash_panel_menu_config_view_${normalized_camp}`);
			let current_view = self.get_current_view_id(campaign_id);
			//Se muestra el menu de config de vistas (summary, members, indicators)
			btn_menu_config_view[current_view === "dashboard"? "show": "hide"]();

			const is_view_bot = self.__is_campaign_view_bots(campaign_id);
			if (is_view_bot) {
				$(`#sup_dash_panel_campaign_bots_filter_${normalized_camp}`).show();
				$(`#sup_dash_panel_detail_${normalized_camp}`).hide();
				$(`#sup_dash_panel_menu_config_view_${normalized_camp}`).hide();
				$(`#sup_dash_panel_campaign_accounts_${normalized_camp}`).hide();
			}
			
		});
	};

	SupervisorDashboard.prototype.__hide_all_account_selectors = function()
	{
		const self = this;
		$.each(this.m_rendered_campaigns, function(campaign_id) {
			const normalized_id = self.get_normalized_id(campaign_id);
			const $popup = $(`#sup_dash_campaign_accounts_config_${normalized_id}`);
			self.__hide_account_selector(campaign_id, $popup);
		});
	};

	SupervisorDashboard.prototype.__hide_account_selector = function(campaign_id, $selector, $btn_parent)
	{
		if (!$selector.is(':visible')) {
			return;
		}

		if ($btn_parent) {
			$btn_parent.removeClass("btnActive");
		}
		else{
			const normalized_id = this.get_normalized_id(campaign_id);
			$(`#sup_dash_panel_campaign_accounts_${normalized_id}`).removeClass("btnActive");
			$("#sup_full_screen_header_navigation_campaign_accounts_btn").removeClass("btnActive");
		}
		
		$selector.hide();
		//Guardo los cambios que pudieran existir
		let old_config = this.get_campaign_selected_config(campaign_id);
		let current_config = this.get_selected_campaign_account_al(campaign_id, true);
		if (inConcert.Utils.diffObjects(old_config, current_config)) {
			this.__set_selected_campaign_account_al(true, campaign_id, current_config);
			$('#sup_dash_campaign_selector_ok').trigger('click');	
		}
	};

	SupervisorDashboard.prototype.show_campaign_accounts_selector = function(campaign_id, $parent)
	{
		const normalized_id = this.get_normalized_id(campaign_id);
		const $popup = $(`#sup_dash_campaign_accounts_config_${normalized_id}`);
		const self = this;

		const fnRePosition = function() {
			const botonRect = $parent[0].getBoundingClientRect();
			const isbtnFullScreen = $parent[0].id === "sup_full_screen_header_navigation_campaign_accounts_btn";
			let left = 0;
			let top = 0;
			if (isbtnFullScreen) {
				left = botonRect.left + 16;
			}
			else {
				top = botonRect.height + 10;
				let campaign_node_offset = self.get_campaign_node(campaign_id).offset();
				left = botonRect.left - campaign_node_offset.left;
			}
			$popup.css({ left: left, top: top});
		};

		//Si ya esta visible, salgo
		if ($popup.is(':visible')) {
			return;
		}

		const show = function() {
			
			var campaign_config = self.get_campaign_selected_config(campaign_id);
			var $accounts_list = $popup.find('[accounts_list]');
			if ( campaign_config.account_id ) {
			// marco una cuenta especifica
				var $account_radio = $accounts_list.find('[account_item_type="account"] > [account_selector][account_id="' + campaign_config.account_id + '"]');
				self.__mark_unmark_campaign_account( $account_radio, true /* mark */ );
				$account_radio.prop('checked', true);
				if ( campaign_config.al_id ) {
				// marco un attention level especifico
					var $als_list = $account_radio.parent().find('[attention_levels_list]');
					var $al_radio = $als_list.find('[attention_level_selector][attention_level_id="' + campaign_config.al_id + '"]');
					$al_radio.trigger('click');
				}
			}
			else {
			// marco el chk (campaign all accounts)
				var $all_accounts_radio = $accounts_list.find('[account_item_type="header"] > [account_selector]');
				$all_accounts_radio.trigger('click');
			}
			//Muestro y recalculo la posicion.
			$popup.show();
			fnRePosition();

			//Agrego clase btnActive
			$parent.removeClass("btnActive").addClass("btnActive");

			//Evento de clickOutside.
			$popup.off("clickoutside")
				.on("clickoutside", function (e) {
					const $target = $(e.target);
					if ($target.is($parent) || $target.closest($parent).length > 0) {
						return;
					}
					self.__hide_account_selector(campaign_id, $popup);
				});
		};

		show();
	};

	SupervisorDashboard.prototype.enable_disable_btn_campaign_accounts_selector = function (campaign_id) {
		// Hibilitar/Deshabilitar el boton del selector de cuentas.
		const campaign = inConcert.app.getCampaign(campaign_id);
		if (!campaign) {
			return;
		}
		const normalized_id = this.get_normalized_id(campaign_id);
		const accounts = campaign.get_accounts();
		$(`#sup_dash_panel_campaign_accounts_${normalized_id}`).prop('disabled', accounts.length === 0);
		if (this.m_full_screen && this.m_full_screen.get_campaign_id() === campaign_id) {
			$("#sup_full_screen_header_navigation_campaign_accounts_btn").prop('disabled', accounts.length === 0);
		}
	};

	SupervisorDashboard.prototype.toggle_btns_toolbar_bots = function(campaign_id, is_view_bot)
	{
		const normalized_camp = this.get_normalized_id(campaign_id);
		if (is_view_bot) {
			$(`#sup_dash_panel_campaign_bots_filter_${normalized_camp}`).css("display", "inline-flex");
		}
		else{
			$(`#sup_dash_panel_campaign_bots_filter_${normalized_camp}`).hide();
		}
		$(`#sup_dash_panel_detail_${normalized_camp}`)[is_view_bot?"hide":"show"]();
		$(`#sup_dash_panel_menu_config_view_${normalized_camp}`)[is_view_bot?"hide":"show"]();
		$(`#sup_dash_panel_campaign_accounts_${normalized_camp}`)[is_view_bot?"hide":"show"]();
	};

	SupervisorDashboard.prototype.show_campaign_bots_filters = function(campaign_id, $parent)
	{
		const is_view_bot = this.__is_campaign_view_bots(campaign_id);
		if (!is_view_bot) {
			return;
		}

		$parent.removeClass("btnActive").addClass("btnActive");
		const $bots_container = this.get_campaign_bots_node(campaign_id);
		const $popup = $bots_container.find("#sup_dash_panel_bots_filter_popup");
		//Si ya esta visible, salgo
		if ($popup.is(':visible')) {
			return;
		}
		const self = this;
		const fnRePosition = function() {
			const botonRect = $parent[0].getBoundingClientRect();
			let campaign_node_offset = self.get_campaign_node(campaign_id).offset();
			let left = botonRect.left - campaign_node_offset.left;
			const isbtnFullScreen = $parent[0].id === "sup_dash_full_screen_panel_bots_filter";
			let top = isbtnFullScreen? botonRect.top : botonRect.height + 10;
			$popup.css({ left: left, top: top});
			
		};
		//Evento de clickOutside.
		$popup.off("clickoutside")
		.on("clickoutside", function (e) {
			const $target = $(e.target);
			if ($target.is($parent) || $target.closest($parent).length > 0) {
				return;
			}
			$parent.removeClass("btnActive");
			$(this).hide();
		});

		$popup.show();
		fnRePosition();

	};

	SupervisorDashboard.prototype.campaignBotsSelectFilterOption = function(campaign_id)
	{
		const $bots_node = this.get_campaign_bots_node(campaign_id);
		const is_view_bot = this.__is_campaign_view_bots(campaign_id, $bots_node);
		if (!is_view_bot) {
			return;
		}
		const appForm = $bots_node.find('#icc-processflow-dashboard')[0];
		let filterTitle = appForm.getfilterTitle();
		let nid = this.get_normalized_id(campaign_id);
		let html = `<span><b><%=_T("Viewing")%>:</b> ${filterTitle} </span><div class="separator" style="margin-right: 10px;"></div><i class="far fa-calendar-alt"></i>`;
		$("#sup_dash_full_screen_panel_bots_filter").html(html);
		$(`#sup_dash_panel_campaign_bots_filter_${nid}`).html(html);
	};

}());
